/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a simple C++ wrapper to collect all black channel calls provided by QNX, and will be used to integrate with
 * the message infrastructure
 */

#include "BlackChannel/BlackChannel.h"

#include <cstring>
#include <random>
#include <sstream>
#include <uuid/uuid.h>

namespace bc {

// shared secret between publisher and subscriber
static const BCSharedSecret g_ikm = {{0xad, 0xc5, 0x8f, 0xcd, 0xc7, 0x1a, 0xf4, 0x1e, 0x03, 0x72, 0x95, 0x2a, 0x6b,
                                      0xf6, 0x02, 0x7b, 0x66, 0xda, 0xec, 0x97, 0xc9, 0x24, 0x39, 0xf0, 0xaa, 0x5c,
                                      0x49, 0x95, 0x57, 0x8d, 0x3e, 0x0f, 0x44, 0x7b, 0x74, 0xba, 0xc5, 0x69, 0x22,
                                      0x43, 0xf7, 0xe0, 0x68, 0xd5, 0x97, 0x37, 0xbb, 0x33, 0x0a, 0x5e, 0xb5, 0x8e,
                                      0x75, 0xbd, 0x69, 0xf9, 0x05, 0x81, 0x57, 0x21, 0xd3, 0xb0, 0x53, 0x1d}};

BCKeyTable calcSessionKeyTable(const BlackChannelKey& salt) noexcept
{
    BCKeyTable table{};

    rekeyCalculateKs(salt.getBCKey(), table, BC_KEY_TABLE_SIZE);

    return table;
}

static std::mt19937& getEngineInstance()
{
    static std::random_device rd;
    static std::seed_seq seed{rd(), rd(), rd(), rd(), rd(), rd(), rd(), rd()}; // (up to) 256 bits of entropy
    static std::mt19937 engine{seed};

    return engine;
}

BCByteArray getNewSalt() noexcept
{
    std::uniform_int_distribution<> keyValDis(BC_KEY_ELEM_MIN, BC_KEY_ELEM_MAX - 1);

    BCByteArray newSalt;
    for (size_t i = 0; i < BC_KEY_SIZE; ++i) {
        newSalt[i] = static_cast<uint8_t>(keyValDis(getEngineInstance()));
    }

    return newSalt;
}

std::string getNewUUId() noexcept
{
    uuid_t uuid;
    uuid_generate(uuid); // NOLINT(cppcoreguidelines-pro-bounds-array-to-pointer-decay)

    // +1 for the NUL termination
    char uuid_cstr[BC_SENDER_UUID_SIZE + 1];
    uuid_unparse(uuid, static_cast<char*>(uuid_cstr)); // NOLINT(cppcoreguidelines-pro-bounds-array-to-pointer-decay)

    return std::string(static_cast<char*>(uuid_cstr));
}

std::string bc_result_to_str(const bc_result_t result)
{
    std::string bcResultStr;

    switch (result) {
        case BC_OK:
            bcResultStr = "BC_OK";
            break;
        case BC_E_INPUTERR_WRONG:
            bcResultStr = "BC_E_INPUTERR_WRONG";
            break;
        case BC_E_INPUTERR_NULL:
            bcResultStr = "BC_E_INPUTERR_NULL";
            break;
        case BC_E_WRONG_STATE:
            bcResultStr = "BC_E_WRONG_STATE";
            break;
        case BC_E_INTERR:
            bcResultStr = "BC_E_INTERR";
            break;
        default:
            break;
    }

    return bcResultStr;
}

std::string bc_p17_status_to_str(const bc_p17_check_status_t status)
{
    std::string bcStatusStr;

    switch (status) {
        case BC_P17STATUS_OK:
            bcStatusStr = "BC_P17STATUS_OK";
            break;
        case BC_P17STATUS_NONEWDATA:
            bcStatusStr = "BC_P17STATUS_NONEWDATA";
            break;
        case BC_P17STATUS_ERROR:
            bcStatusStr = "BC_P17STATUS_ERROR";
            break;
        case BC_P17STATUS_REPEATED:
            bcStatusStr = "BC_P17STATUS_REPEATED";
            break;
        case BC_P17STATUS_OKSOMELOST:
            bcStatusStr = "BC_P17STATUS_OKSOMELOST";
            break;
        case BC_P17STATUS_WRONGSEQUENCE:
            bcStatusStr = "BC_P17STATUS_WRONGSEQUENCE";
            break;
        default:
            break;
    }

    return bcStatusStr;
}

bool rekeyCalculateKs(const BCByteArray& salt, BCKeyTable& okm, uint8_t iterations) noexcept
{
    bool res = true;

    uint8_t prk[BC_KEY_SIZE];
    hmac256_t ctx;

    /* hkdf_extract */
    res &=
        (hmac_sha256(g_ikm.data(), BC_SHARED_SECRET_SIZE, salt.data(), BC_KEY_SIZE, static_cast<uint8_t*>(prk)) == 1);

    /* hkdf_expand */
    for (uint8_t i = 1; i <= iterations; i++) {
        res &= (hmac_sha256_init(static_cast<uint8_t*>(prk), BC_KEY_SIZE, &ctx) == 1);

        res &= (hmac_sha256_add(
                    (i > 1) ? okm[static_cast<size_t>(i - 2)].data() : nullptr, (i > 1) ? BC_KEY_SIZE : 0, &ctx) == 1);

        res &= (hmac_sha256_add(&i, 1, &ctx) == 1);

        res &= (hmac_sha256_done(&ctx, okm[static_cast<size_t>(i - 1)].data()) == 1);
    }
    return res;
}

} // namespace bc
