/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a simple BlackChannelKey class implementation, both salt and secret are examples of BlackChannelKey
 */

#include "BlackChannel/BlackChannelKey.h"

#include <algorithm>
#include <cctype>
#include <cstdint>
#include <iomanip>
#include <sstream>

using namespace bc;

BlackChannelKey::BlackChannelKey(const BCByteArray& key) noexcept
    : m_key{key}
    , m_validness{true}
{}

std::string BlackChannelKey::toString() const
{
    std::stringstream ss;
    ss << std::hex << std::setfill('0');
    for (auto byte : m_key) {
        ss << std::setw(HEX_DIGITS_PER_BYTE) << static_cast<unsigned>(byte);
    }
    return ss.str();
}

BlackChannelKey BlackChannelKey::fromString(const std::string& s)
{
    // This implementation assumes that [min, max) bounds encompass the entire range of an 8-bit byte. Here, we check
    // that the assumption holds.
    static_assert(BC_KEY_ELEM_MIN == 0, "fromString() should check byte lower bound");
    static_assert(BC_KEY_ELEM_MAX == 256, "fromString() should check byte upper bound");

    BCByteArray bytes;

    // Return an error key if (1) we have an incorrect number of digits, or (2) there is a non-hex digit.
    if ((s.size() != bytes.size() * HEX_DIGITS_PER_BYTE) ||
        std::any_of(s.begin(), s.end(), [](unsigned char digit) { return (std::isxdigit(digit) == 0); })) {
        return BlackChannelKey{};
    }

    for (std::size_t i = 0; i < bytes.size(); ++i) {
        const auto digits = s.substr(i * HEX_DIGITS_PER_BYTE, HEX_DIGITS_PER_BYTE);
        const auto byte = std::stoi(digits, nullptr, BASE_16);
        bytes.at(i) = static_cast<std::uint8_t>(byte);
    }

    return BlackChannelKey{bytes};
}

bool BlackChannelKey::operator==(const BlackChannelKey& other) const noexcept
{
    return m_key == other.getBCKey();
}

bool BlackChannelKey::operator!=(const BlackChannelKey& other) const noexcept
{
    return m_key != other.getBCKey();
}
