/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a simple BlackChannelMsg class implementation
 */

#include "BlackChannel/BlackChannelMsg.h"

#include <bc/bc_qnx_util.h>

using namespace bc;

BlackChannelMsg::BlackChannelMsg(const std::string& msg,
                                 const std::string& senderId,
                                 MESSAGE_TYPE msgType,
                                 SHA2Type sha2Type)
{
    if (sha2Type == SHA2Type::SHA256) {
        size_t cap = BC_MSG_HEADER_SHA256_SIZE + msg.size();
        m_msgBCMsg.reserve(cap);
        m_msgBCMsg.resize(cap);
        m_shaHeaderSize = HEADER_SIZE_SHA256_FULL;
    } else if (sha2Type == SHA2Type::SHA512) {
        throw std::runtime_error("SHA512 is not supported yet!");
    } else {
        throw std::runtime_error("Invalid SHA2 type specified!");
    }

    if (senderId.length() != BC_SENDER_UUID_SIZE) {
        throw std::runtime_error("Invalid sender uuid specified!");
    }
    for (size_t i = 0; i < BC_SENDER_UUID_SIZE; i++) {
        m_msgBCMsg[i + m_shaHeaderSize] = senderId[i];
    }

    if (msgType == MESSAGE_TYPE::METADATA) {
        m_msgBCMsg[m_shaHeaderSize + BC_SENDER_UUID_SIZE] = BC_MSG_TYPE_META;
    } else if (msgType == MESSAGE_TYPE::MESSAGE) {
        m_msgBCMsg[m_shaHeaderSize + BC_SENDER_UUID_SIZE] = BC_MSG_TYPE_MSG;
    } else {
        throw std::runtime_error("Invalid msg type specified!");
    }

    for (size_t i = 0; i < msg.size(); i++) {
        m_msgBCMsg[i + m_shaHeaderSize + BC_MSG_SENDER_ID_TYPE_SIZE] = msg[i];
    }
}

BCMsgParseResult BlackChannelMsg::parseBlackChannelMsg(const std::string& bcMsg)
{
    // currently, only SHA256 is supported
    if (bcMsg.length() < BC_MSG_HEADER_SHA256_SIZE) {
        return {BCMsgParseResultCode::INVALID_LENGTH};
    }

    MESSAGE_TYPE msgType{MESSAGE_TYPE::INVALID};
    char metaOrMsg = bcMsg[HEADER_SIZE_SHA256_FULL + BC_SENDER_UUID_SIZE];
    switch (metaOrMsg) {
        case BC_MSG_TYPE_META:
            msgType = MESSAGE_TYPE::METADATA;
            break;
        case BC_MSG_TYPE_MSG:
            msgType = MESSAGE_TYPE::MESSAGE;
            break;
        default:
            return {BCMsgParseResultCode::INVALID_TYPE};
    }
    const auto uuid = bcMsg.substr(HEADER_SIZE_SHA256_FULL, BC_SENDER_UUID_SIZE);

    const auto msg = bcMsg.substr(BC_MSG_HEADER_SHA256_SIZE);

    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    const auto bcMsgU8 = reinterpret_cast<const std::uint8_t*>(bcMsg.data());
    const auto counterOffset = static_cast<std::uint16_t>(COUNTER_OFFSET_SHA256);
    const auto counter = bc_hdr_get_u32(bcMsgU8, counterOffset);

    return {BCMsgParseResultCode::OK, uuid, msgType, msg, counter};
}
