/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a simple BlackChannelPublisher class implementation
 */

#include "BlackChannel/BlackChannelPublisher.h"

#include "BlackChannel/BlackChannelConstants.h"
#include "BlackChannel/BlackChannelSaltMetadata.h"
#include "BlackChannel/BlackChannelSaltUpdateMetadata.h"

#include "lib/avlog/avlog.hpp"

#include <array>

using namespace bc;
using namespace std;

BlackChannelPublisher::BlackChannelPublisher()
    : m_salt{getNewSalt()}
{
    protectInit();
}

void BlackChannelPublisher::protectInit()
{
    m_keyTable = calcSessionKeyTable(m_salt);

    memcpy(&m_tx_config, &g_config_default_256, sizeof(g_config_default_256));

    std::array<uint8_t*, BC_KEY_TABLE_SIZE> key_ptr_tbl{};
    for (size_t i = 0; i < BC_KEY_TABLE_SIZE; i++) {
        key_ptr_tbl.at(i) = m_keyTable.at(i).data();
    }
    m_tx_config.key_table = static_cast<std::uint8_t**>(key_ptr_tbl.data());
    m_tx_config.key_length = BC_KEY_SIZE;
    m_tx_config.key_table_size = BC_KEY_TABLE_SIZE;

    bc_result_t result = bc_p17_protect_init(&m_tx_config, &m_tx_state);
    if (result != BC_OK) {
        throw runtime_error("Black Channel profile 17 protect init failed!");
    }

    AVLOG_INFO("(Re-)initialized Black Channel protect data structures.");
}

bool BlackChannelPublisher::protect(const std::string& bcMsg) noexcept
{
    bool ret = true;

    // Handle transitions to a new salt value
    {
        lock_guard<mutex> lock(m_saltPublishingMutex);
        if (m_saltPublishingState == SaltPublishingState::PublishSaltUpdate) {
            if (hasSaltUpdateCompleted()) {
                AVLOG_INFO("Salt update is ready, counter: {}, initial: {}, through: {}",
                           AVLOG_ARGS("counter", "initial", "through"),
                           m_tx_state.counter,
                           m_validInitial,
                           m_validThrough);

                // We have completed the salt update
                protectInit();
                m_saltPublishingState = SaltPublishingState::PublishSalt;
            }
        }
    }

    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast, cppcoreguidelines-pro-type-const-cast)
    auto msg = const_cast<uint8_t*>(reinterpret_cast<const uint8_t*>(bcMsg.data()));
    bc_result_t result = bc_p17_protect(&m_tx_config, &m_tx_state, msg, static_cast<uint32_t>(bcMsg.size()));
    if (result != BC_OK) {
        AVLOG_ERROR("Black Channel profile 17 protect failed!");
        ret = false;
    }

    return ret;
}

bool BlackChannelPublisher::hasSaltUpdateCompleted() const noexcept
{
    bool updateCompleted{false};
    // Check whether the counter is outside of the range [m_validInitial..m_validThrough].
    if (m_validInitial <= m_validThrough) {
        updateCompleted = ((m_tx_state.counter > m_validThrough) || (m_tx_state.counter < m_validInitial));
    } else {
        // the counter wrapped around while calculating validThrough
        updateCompleted = ((m_tx_state.counter > m_validThrough) && (m_tx_state.counter < m_validInitial));
    }
    return updateCompleted;
}

void BlackChannelPublisher::scheduleSaltUpdate(const std::uint32_t validMessageCount)
{
    lock_guard<mutex> lock(m_saltPublishingMutex);

    // A salt update can only be scheduled when we are in the PublishSalt state. Otherwise, we are either not
    // publishing, stopping, or already in the process of publishing a salt update.
    if (m_saltPublishingState != SaltPublishingState::PublishSalt) {
        AVLOG_WARN("Attempted to schedule a salt update while in the wrong state.");
        return;
    }

    scheduleSaltUpdateInternal(validMessageCount);

    // Wake up the salt publishing thread
    m_saltPublishingCondition.notify_one();
}

void BlackChannelPublisher::scheduleSaltUpdateInternal(const std::uint32_t validMessageCount)
{
    m_saltPublishingState = SaltPublishingState::PublishSaltUpdate;
    m_salt.setBCKey(getNewSalt());
    m_validInitial = m_tx_state.counter;

    // Note that validInitial (the tx state counter) is the counter for the _next_ message, so if validMessageCount is
    // zero, this still ensures that we will send at least one message to the subscriber.
    m_validThrough = m_validInitial + validMessageCount;

    AVLOG_INFO("Scheduled a salt update; initial: {}, through: {}",
               AVLOG_ARGS("validInitial", "validThrough"),
               m_validInitial,
               m_validThrough);
}

void BlackChannelPublisher::startPublishingSalt(
    const std::chrono::milliseconds resaltInterval,
    const std::function<void(const std::string&)>& metadataWriter,
    const std::function<std::uint32_t()>& estimateSaltUpdateValidMessageCount)
{
    // If we are in a valid state to do so, start publishing
    {
        lock_guard<mutex> lock(m_saltPublishingMutex);
        if (m_saltPublishingState != SaltPublishingState::NotPublishing) {
            AVLOG_WARN("Attempted to start publishing salt while already running.");
            return;
        }

        m_saltPublishingState = SaltPublishingState::PublishSalt;
    }

    // Create a thread from which to publish salt and salt updates
    const auto saltPublishingThread = [this, resaltInterval, metadataWriter, estimateSaltUpdateValidMessageCount]() {
        publishSalt(resaltInterval, metadataWriter, estimateSaltUpdateValidMessageCount);
    };
    m_saltPublishingThread = make_unique<thread>(saltPublishingThread);
}

void BlackChannelPublisher::publishSalt(const std::chrono::milliseconds resaltInterval,
                                        const std::function<void(const std::string&)>& metadataWriter,
                                        const std::function<std::uint32_t()>& estimateSaltUpdateValidMessageCount)
{
    AVLOG_INFO("Started salt publishing.");

    chrono::steady_clock::time_point resaltTime = chrono::steady_clock::now() + resaltInterval;
    for (;;) {
        unique_lock<mutex> lock(m_saltPublishingMutex);

        // Handle a request to stop
        if (m_saltPublishingState == SaltPublishingState::StoppingPublishing) {
            m_saltPublishingState = SaltPublishingState::NotPublishing;
            AVLOG_INFO("Stopped salt publishing.");
            return;
        }

        // Determine whether to schedule a salt update. If resaltInterval is 0ms, then updates are disabled.
        auto resaltIntervalRemaining =
            chrono::duration_cast<chrono::milliseconds>(resaltTime - chrono::steady_clock::now());
        if (resaltIntervalRemaining <= 0ms) {
            if (m_saltPublishingState == SaltPublishingState::PublishSalt && resaltInterval > 0ms) {
                scheduleSaltUpdateInternal(estimateSaltUpdateValidMessageCount());
            }
            resaltTime = chrono::steady_clock::now() + resaltInterval;
            resaltIntervalRemaining = resaltInterval;
        }

        string saltMsg;
        std::chrono::milliseconds interval;

        std::tie(saltMsg, interval) = getPublisherMetadataAndInterval();
        metadataWriter(saltMsg);

        // If resalting is enabled, ensure that we don't sleep longer than the next scheduled resalt.
        if (resaltIntervalRemaining > 0ms) {
            interval = std::min(interval, resaltIntervalRemaining);
        }

        // Sleep until the next scheduled event, or we are woken up.
        m_saltPublishingCondition.wait_for(lock, interval);
    }
}

void BlackChannelPublisher::stopPublishingSalt()
{
    // If we are in a valid state to do so, stop publishing.
    {
        lock_guard<mutex> lock(m_saltPublishingMutex);
        if (m_saltPublishingState == SaltPublishingState::NotPublishing ||
            m_saltPublishingState == SaltPublishingState::StoppingPublishing) {
            AVLOG_WARN("Attempted to stop salt publishing thread from an invalid state.");
            return;
        }

        AVLOG_INFO("Stopping salt publishing.");
        m_saltPublishingState = SaltPublishingState::StoppingPublishing;
    }

    // Wake up the salt publishing thread
    m_saltPublishingCondition.notify_one();
    m_saltPublishingThread->join();
}

std::tuple<std::string, std::chrono::milliseconds> BlackChannelPublisher::getPublisherMetadataAndInterval()
{
    std::string metadata;
    chrono::milliseconds interval;

    if (m_saltPublishingState == SaltPublishingState::PublishSalt) {
        metadata = BlackChannelSaltMetadata{m_salt}.toString();
        interval = SALT_PUBLISHING_INTERVAL;
    } else if (m_saltPublishingState == SaltPublishingState::PublishSaltUpdate) {
        metadata = BlackChannelSaltUpdateMetadata{m_validThrough, m_salt}.toString();
        interval = SALT_UPDATE_PUBLISHING_INTERVAL;
    }

    return make_tuple(metadata, interval);
}
