/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a simple BlackChannelSubscriber class implementation
 */

#include "BlackChannel/BlackChannelSubscriber.h"

#include "BlackChannelMetadataType.h"
#include "lib/avlog/avlog.hpp"

#include <array>
#include <vector>

using namespace bc;
using namespace std;

void BlackChannelSubscriber::checkInit()
{
    m_keyTable = calcSessionKeyTable(m_salt);

    memcpy(&m_rx_config, &g_config_default_256, sizeof(g_config_default_256));

    std::array<uint8_t*, BC_KEY_TABLE_SIZE> key_ptr_tbl{};
    for (size_t i = 0; i < BC_KEY_TABLE_SIZE; i++) {
        key_ptr_tbl.at(i) = m_keyTable.at(i).data();
    }

    m_rx_config.key_table = static_cast<std::uint8_t**>(key_ptr_tbl.data());
    m_rx_config.key_length = BC_KEY_SIZE;
    m_rx_config.key_table_size = BC_KEY_TABLE_SIZE;

    bc_result_t result = bc_p17_check_init(&m_rx_config, &m_rx_state);
    if (result != BC_OK) {
        throw runtime_error("Black Channel profile 17 check init failed!");
    }

    AVLOG_INFO("(Re-)initialized Black Channel check data structures.");
}

bool BlackChannelSubscriber::check(const std::string& bcMsg) noexcept
{
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    const auto data = reinterpret_cast<const uint8_t*>(bcMsg.data());
    const auto size = bcMsg.size();

    bool checkSucceeded{false};
    bc_result_t bcResult{BC_E_INTERR};

    std::tie(checkSucceeded, bcResult) = check(data, size);
    if (!checkSucceeded) {
        if (m_state == SubscriberState::Resalting) {
            // It is likely that the salt update has completed, and that we're now using the new salt.
            AVLOG_INFO(
                "Salt update may be complete; last counter: {}, valid through: {}", m_rx_state.counter, m_validThrough);

            const auto currentSalt = m_salt;
            m_salt = m_pendingSalt;
            checkInit();

            std::tie(checkSucceeded, bcResult) = check(data, size);
            if (checkSucceeded) {
                // Commit the new salt.
                m_state = SubscriberState::Salted;
                AVLOG_INFO("Salt update complete; new salt is in use");
            } else {
                // Rollback to the old salt -- the received message was invalid.
                m_salt = currentSalt;
                checkInit();
            }
        }
    }

    if (!checkSucceeded) {
        AVLOG_ERROR("check result: {}, m_rx_state status: {}",
                    bc_result_to_str(bcResult),
                    bc_p17_status_to_str(m_rx_state.status));
    }

    return checkSucceeded;
}

std::tuple<bool, bc_result_t> BlackChannelSubscriber::check(const std::uint8_t* const data,
                                                            const std::size_t size) noexcept
{
    bc_result_t bcResult = bc_p17_check(&m_rx_config, &m_rx_state, data, static_cast<uint32_t>(size));

    /**
     * The optimum value for Max Delta Count is 0xFFFFFFFF - receiver count.
     * This permits all messages with a sequence number greater than the receiver count to be classified as
     * OK/OKSOMELOST. This not only solves the ‘lost first message’ problem introduced by a small Max Delta Count but
     * also mitigates the risk of a replay attack.
     */
    m_rx_config.max_delta_counter = BC_OPTIMUM_MAX_DELTA_COUNTER - m_rx_state.counter;

    const auto checkSucceeded = (bcResult == BC_OK) && ((m_rx_state.status == BC_P17STATUS_OK) ||
                                                        (m_rx_state.status == BC_P17STATUS_OKSOMELOST));
    return std::make_tuple(checkSucceeded, bcResult);
}

bool BlackChannelSubscriber::processMetadata(const std::string& metadata)
{
    const auto metadataType = getBlackChannelMetadataType(metadata);
    switch (metadataType) {
        case BlackChannelMetadataType::Salt:
            return processSaltMetadata(BlackChannelSaltMetadata::fromString(metadata));

        case BlackChannelMetadataType::SaltUpdate:
            return processSaltUpdateMetadata(BlackChannelSaltUpdateMetadata::fromString(metadata));

        case BlackChannelMetadataType::Unknown:
            AVLOG_INFO("bc msg unknown metadata received");
            return false;
    }

    // This should be unreachable code. If we get here, the classification from getBlackChannelMetadataType() is wrong.
    AVLOG_ERROR("bc msg uncategorized metadata received");
    return false;
}

bool BlackChannelSubscriber::processSaltMetadata(const BlackChannelSaltMetadata& saltMetadata)
{
    if (!saltMetadata.key().isBCKeyValid()) {
        AVLOG_WARN("Invalid salt received");
        return false;
    }

    if (saltMetadata.key() != m_salt) {
        m_salt = saltMetadata.key();
        checkInit();
    }

    return true;
}

bool BlackChannelSubscriber::processSaltUpdateMetadata(const BlackChannelSaltUpdateMetadata& saltUpdateMetadata)
{
    if (!saltUpdateMetadata.key().isBCKeyValid()) {
        AVLOG_WARN("Invalid salt update received");
        return false;
    }

    if (m_state == SubscriberState::Salted ||
        ((m_state == SubscriberState::Resalting) && (saltUpdateMetadata.key() != m_pendingSalt))) {
        if (m_state == SubscriberState::Resalting) {
            AVLOG_INFO("Salt update value changed before previous salt update was handled");
        }

        // This is the first of a new round of salt updates
        m_pendingSalt = saltUpdateMetadata.key();
        m_validThrough = saltUpdateMetadata.validThrough();
        m_state = SubscriberState::Resalting;

        AVLOG_INFO("Salt Update processed; valid through: {}", m_validThrough);
    }

    return true;
}
