/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a C++ helper class which provides useful APIs to integrate with the message infrastructure
 */

#ifndef APTIVWARECORE_BLACKCHANNEL_H
#define APTIVWARECORE_BLACKCHANNEL_H

#include "BlackChannelConstants.h"
#include "BlackChannelKey.h"

extern "C"
{
#include <bc/bc_qnx_p17.h>
#include <hmac/bc_hmac.h>
}

#include <string>

namespace bc {
using BCSharedSecret = std::array<uint8_t, BC_SHARED_SECRET_SIZE>;

using BCKeyTable = std::array<BCByteArray, BC_KEY_TABLE_SIZE>;

static const bc_p17_config_t g_config_default_256 = {0,
                                                     HEADER_SIZE_SHA256_FULL,
                                                     DEFAULT_MAX_PAYLOAD_SIZE,
                                                     BC_OPTIMUM_MAX_DELTA_COUNTER,
                                                     nullptr,
                                                     0,
                                                     0,
                                                     BC_P17_HMAC_SIZE_SHA256,
                                                     BC_P17_HMAC_SHA256};

/**
 * @brief calculate the session key table based on the given salt
 *
 * @param salt - a salt provided as a BlackChannelKey
 * @return session key table
 */
BCKeyTable calcSessionKeyTable(const BlackChannelKey& salt) noexcept;

/**
 * @brief generate a random salt for publisher
 *
 * @return BCByteArray
 */
BCByteArray getNewSalt() noexcept;

/**
 * @brief generate a new uuid (36 bytes), used as a unique sender id
 *
 * @return a string style uuid
 */
std::string getNewUUId() noexcept;

/**
 * @brief helper function to convert bc_result to string
 *
 * @param result - bc_result_t
 * @return string
 */
std::string bc_result_to_str(const bc_result_t result);

/**
 * @brief helper function to convert bc_p17_check_status_t to string
 *
 * @param status - bc_p17_check_status_t
 * @return string
 */
std::string bc_p17_status_to_str(const bc_p17_check_status_t status);

/**
 * @brief a simplified implementation of HKDF, provided by karl.robinson@aptiv.com.
 *        - optional context/application info is ignored
 *        - we only need OKM large enough to key HMAC-SHA256 (N=1)
 *
 * @param salt - salt message
 * @param okm - output keying material
 * @param iterations - iteration times, same as BC_KEY_TABLE_SIZE
 * @return true - if successful, otherwise false
 */
bool rekeyCalculateKs(const BCByteArray& salt,
                      std::array<BCByteArray, BC_KEY_TABLE_SIZE>& okm,
                      uint8_t iterations) noexcept;
} // namespace bc

#endif // APTIVWARECORE_BLACKCHANNEL_H
