/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a helper constants header file to provide all the black channel related values.
 */

#ifndef APTIVWARECORE_BLACKCHANNELCONSTANTS_H
#define APTIVWARECORE_BLACKCHANNELCONSTANTS_H

#include <chrono>
#include <cstdint>
#include <limits>
#include <string>

namespace bc {
using namespace std::literals;

enum class MESSAGE_TYPE : uint8_t
{
    INVALID = 0,
    METADATA = 1, // metadata added to implement black channel protocol
    MESSAGE = 2   // normal message sent/received at SCChannel layer
};

constexpr uint32_t BC_KEY_SIZE = 32; // salt size

// used to generate random salt values in this range [0, 256)
constexpr uint32_t BC_KEY_ELEM_MIN = 0;
constexpr uint32_t BC_KEY_ELEM_MAX = 256;

constexpr uint32_t BC_SHARED_SECRET_SIZE = 64; // shared secret size

// black channel key table size, will be used to determine bc_p17_config_t.key_table_size
constexpr uint32_t BC_KEY_TABLE_SIZE = 1;

constexpr std::chrono::milliseconds SALT_PUBLISHING_INTERVAL = 1000ms;       // interval between SALT messages
constexpr std::chrono::milliseconds SALT_UPDATE_PUBLISHING_INTERVAL = 250ms; // interval between SALT UPDATE messages

constexpr uint32_t HEADER_PADDING_SIZE = sizeof(std::uint32_t) + sizeof(std::uint32_t); // 8 bytes; length + counter
constexpr uint32_t HEADER_SIZE_SHA256_FULL = (32 + HEADER_PADDING_SIZE);                // 40 bytes
constexpr uint32_t HEADER_SIZE_SHA512_FULL = (64 + HEADER_PADDING_SIZE);                // 72 bytes
constexpr uint32_t COUNTER_OFFSET_SHA256 = (32 + sizeof(std::uint32_t));                // 36 bytes; 32 + length
constexpr uint32_t COUNTER_OFFSET_SHA512 = (64 + sizeof(std::uint32_t));                // 68 bytes; 64 + length

// the maximum allowed gap between the counter values of two consecutively received valid safety frames
constexpr uint32_t BC_OPTIMUM_MAX_DELTA_COUNTER = 0xFFFFFFFF;

// the max payload to support for our system, use uint32_t max for now.
constexpr uint32_t DEFAULT_MAX_PAYLOAD_SIZE = std::numeric_limits<uint32_t>::max();

// 36 bytes for the UUID string representation, see the man page for uuid_unparse for more info
constexpr uint32_t BC_SENDER_UUID_SIZE = 36;

// 1 byte is used to differentiate metadata and normal protected messages in black channel message header
constexpr uint32_t BC_MSG_TYPE_SIZE = 1;
constexpr char BC_MSG_TYPE_MSG = '&';
constexpr char BC_MSG_TYPE_META = '#';

// more offset for black channel message header
constexpr uint32_t BC_MSG_SENDER_ID_TYPE_SIZE = BC_SENDER_UUID_SIZE + BC_MSG_TYPE_SIZE;              // 37 bytes
constexpr uint32_t BC_MSG_HEADER_SHA256_SIZE = HEADER_SIZE_SHA256_FULL + BC_MSG_SENDER_ID_TYPE_SIZE; // 77 bytes
constexpr uint32_t BC_MSG_HEADER_SHA512_SIZE = HEADER_SIZE_SHA512_FULL + BC_MSG_SENDER_ID_TYPE_SIZE; // 109 bytes

// constants for conversions to/from hex encodings
constexpr std::size_t BASE_16 = 16;
constexpr std::size_t HEX_DIGITS_PER_BYTE = 2;
constexpr std::size_t HEX_DIGITS_PER_UINT32 = sizeof(std::uint32_t) * HEX_DIGITS_PER_BYTE;
constexpr std::size_t HEX_DIGITS_PER_KEY = BC_KEY_SIZE * HEX_DIGITS_PER_BYTE;

} // namespace bc

#endif // APTIVWARECORE_BLACKCHANNELCONSTANTS_H
