/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a simple BlackChannelMsg class implementation
 */

#ifndef APTIVWARECORE_BLACKCHANNELMSG_H
#define APTIVWARECORE_BLACKCHANNELMSG_H

#include "BlackChannelConstants.h"
#include "lib/avlog/avlog.hpp"
#include "lib/base/enum/Trait.hpp"

#include <array>

namespace bc {
/**
 * @brief enum class for SHA2Type
 */
enum class SHA2Type : uint8_t
{
    UNKNOWN = 0,
    SHA256 = 1,
    SHA512 = 2
};

/**
 * @brief enum class for Black Channel message parse result
 */
enum class BCMsgParseResultCode : uint8_t
{
    OK = 0,
    INVALID_LENGTH,
    INVALID_TYPE,
};

/**
 * @brief Result structure for parseBlackChannelMsg(). The members below \a code are only valid if <em>code == OK</em>.
 */
struct BCMsgParseResult
{
    BCMsgParseResultCode code{BCMsgParseResultCode::INVALID_TYPE};
    std::string uuid{};
    MESSAGE_TYPE msgType{MESSAGE_TYPE::INVALID};
    std::string msg{};
    std::uint32_t counter{0};
};

/**
 * @brief BlackChannelMsg - a helper class to construct the black channel msg used by protect/check pair in QBCCT,
 * since the format of the black channel message header is fixed in size, so this class is able to allocate the memory
 * in advance and put all the related field in the correct position
 *
 * Black channel message format introduction:
 * Exchanging the protocol metadata requires the channels are able to differentiate between metadata and normal
 * messages, so type is used to indicate the message type. Detailed format for both are listed as below:
 *
 * normal message: [bc : 40 bytes] [sender Id: 36 bytes] [type : 1 byte] [interface header] [type info] [blob]
 * metadata message: [bc : 40 bytes] [sender Id: 36 bytes] [type : 1 byte] [metadata msg]
 *
 * Please note the bc size 40 bytes is only applied to SHA256 which is the current one supported
 */
class BlackChannelMsg final
{
public:
    /**
     * delete default ctor
     */
    BlackChannelMsg() = delete;

    /**
     * @brief delete copy ctor
     */
    BlackChannelMsg(const BlackChannelMsg&) = delete;

    /**
     * @brief delete copy assignment
     */
    BlackChannelMsg& operator=(const BlackChannelMsg&) = delete;

    /**
     * @brief construct the black channel msg
     *
     * @param msg - original msg to send from the upper-level application
     * @param senderId - sender id in the uuid format
     * @param msgType - type of msg, normal message or metadata
     * @param sha2Type - SHA2Type, only SHA256 is supported as for now
     */
    explicit BlackChannelMsg(const std::string& msg,
                             const std::string& senderId,
                             MESSAGE_TYPE msgType,
                             SHA2Type sha2Type);

    /**
     * @brief get the string representation of the internal black channel msg buffer
     *
     * @return string
     */
    std::string getMsgBuf() const noexcept { return m_msgBCMsg; };

    /**
     * @brief parse a black channel message
     *
     * @param bcMsg - the black channel message to parse
     * @return BCMsgParseResult containing the status of the parse, and the parsed data.
     */
    static BCMsgParseResult parseBlackChannelMsg(const std::string& bcMsg);

private:
    std::string m_msgBCMsg{};    ///< internal buffer to hold all the information
    uint32_t m_shaHeaderSize{0}; ///< SHA2 header size
};
} // namespace bc

template<>
struct base::EnumTrait<bc::SHA2Type>
{
    static constexpr std::array namedEnumerators{
        base::Enumerator{bc::SHA2Type::UNKNOWN, "UNKNOWN"},
        base::Enumerator{bc::SHA2Type::SHA256, "SHA256"},
        base::Enumerator{bc::SHA2Type::SHA512, "SHA512"},
    };
};

template<>
struct base::EnumTrait<bc::BCMsgParseResultCode>
{
    static constexpr std::array namedEnumerators{
        base::Enumerator{bc::BCMsgParseResultCode::OK, "OK"},
        base::Enumerator{bc::BCMsgParseResultCode::INVALID_LENGTH, "INVALID_LENGTH"},
        base::Enumerator{bc::BCMsgParseResultCode::INVALID_TYPE, "INVALID_TYPE"},
    };
};
AVLOG_ENUM_FORMAT(bc::SHA2Type)
AVLOG_ENUM_FORMAT(bc::BCMsgParseResultCode)

#endif // APTIVWARECORE_BLACKCHANNELMSG_H
