/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a simple BlackChannelPublisher class implementation
 */

#ifndef APTIVWARECORE_BLACKCHANNELPUBLISHER_H
#define APTIVWARECORE_BLACKCHANNELPUBLISHER_H

#include "BlackChannel.h"

#include <chrono>
#include <condition_variable>
#include <functional>
#include <memory>
#include <mutex>
#include <thread>
#include <tuple>

namespace bc {
/**
 * @brief BlackChannelPublisher class
 *
 * A wrapper class provides APIs for our message infrastructure to integrate QNX Black Channel Communication Technology
 * (QBCCT) APIs.
 */
class BlackChannelPublisher final
{
public:
    /**
     * @brief Constructor.
     */
    BlackChannelPublisher();

    /**
     * @brief default dtor
     */
    ~BlackChannelPublisher() = default;

    /**
     * @brief protect the black channel message to transmit using profile 17
     *
     * @param bcMsg - black channel message which provides a data buffer to insert the bc header.
     * Please use BlackChannelMsg to help construct the bcMsg with the correct format.
     * Currently, only SHA256 is applied, so the bcMsg now leaves the first HEADER_SIZE_SHA256_FULL (40) bytes empty.
     * @return - true if the protect operation is successful, false otherwise
     */
    bool protect(const std::string& bcMsg) noexcept;

    /**
     * @brief Schedule a salt update. If a salt update is already in progress, this is a no-op.
     * @param validMessageCount The number of messages to send before a salt update completes.
     */
    void scheduleSaltUpdate(const std::uint32_t validMessageCount);

    /**
     * @brief Start publishing salt messages using the given \a metadataWriter.
     * @param resaltInterval The interval between salt updates.
     * @param metadataWriter A writer, usually encapsulating a channel, on which to publish salt metadata.
     * @param estimateSaltUpdateValidMessageCount (Optional) A function estimating the number of messages to send before
     *                                            a salt update completes.
     */
    void startPublishingSalt(const std::chrono::milliseconds resaltInterval,
                             const std::function<void(const std::string&)>& metadataWriter,
                             const std::function<std::uint32_t()>& estimateSaltUpdateValidMessageCount);

    /**
     * @brief Stop publishing salt messages.
     */
    void stopPublishingSalt();

    // Our m_saltPublishingMutex implicitly makes us unmoveable and uncopyable; make this explicit
    BlackChannelPublisher(const BlackChannelPublisher&) = delete;
    BlackChannelPublisher& operator=(const BlackChannelPublisher&) = delete;
    BlackChannelPublisher(BlackChannelPublisher&&) = delete;
    BlackChannelPublisher& operator=(BlackChannelPublisher&&) = delete;

private:
    ///< The salt publishing state: whether it is salt, or a salt update being published.
    enum class SaltPublishingState
    {
        NotPublishing,      ///< Salt metadata is not being published (initial/final state)
        StoppingPublishing, ///< Publication of salt metadata is ceasing
        PublishSalt,        ///< We are publishing salt
        PublishSaltUpdate   ///< We are publishing salt updates
    };

    /**
     * @brief Initialize/reinitialize our protection data structures.
     */
    void protectInit();

    /**
     * @brief Schedule a salt update. The caller is responsible for state checking and locking.
     * @param validMessageCount The number of messages to send before a salt update completes.
     */
    void scheduleSaltUpdateInternal(const std::uint32_t validMessageCount);

    /**
     * @brief Publish salt messages using the given \a metadataWriter.
     * @param resaltInterval The interval between salt updates.
     * @param metadataWriter A writer, usually encapsulating a channel, on which to publish salt metadata.
     * @param estimateSaltUpdateValidMessageCount (Optional) A function estimating the number of messages to send before
     *                                            a salt update completes.
     */
    void publishSalt(const std::chrono::milliseconds resaltInterval,
                     const std::function<void(const std::string&)>& metadataWriter,
                     const std::function<std::uint32_t()>& estimateSaltUpdateValidMessageCount);

    /**
     * @brief Determine whether a salt update has completed.
     * @returns True if the salt update has completed (the required number of messages have been sent), otherwise false.
     */
    bool hasSaltUpdateCompleted() const noexcept;

    /**
     * @brief Given the \a key, and our internal publishing state, return a string-encoded metadata message.
     * @param key The salt.
     * @returns The encoded metadata.
     */
    std::tuple<std::string, std::chrono::milliseconds> getPublisherMetadataAndInterval();

    bc_p17_config_t m_tx_config{};                       ///< Configuration for transmitting data using QBCCT profile 17
    bc_p17_protect_state_t m_tx_state{};                 ///< Control data for a QBCCT profile 17 protection header
    std::mutex m_saltPublishingMutex{};                  ///< Mutex for salt publishing state changes
    std::condition_variable m_saltPublishingCondition{}; ///< Condition variable for state change notifications
    SaltPublishingState m_saltPublishingState{SaltPublishingState::NotPublishing}; ///< Salt publishing state
    std::unique_ptr<std::thread> m_saltPublishingThread{nullptr};                  ///< Salt publishing thread
    bc::BlackChannelKey m_salt{};                                                  ///< Generated salt to publish
    bc::BCKeyTable m_keyTable{};                                                   ///< Generated session key table
    std::uint32_t m_validInitial{0}; ///< The message counter at the time of scheduleSaltUpdate().
    std::uint32_t m_validThrough{0}; ///< The black channel sequence number provided to scheduleSaltUpdate().
};
} // namespace bc

#endif // APTIVWARECORE_BLACKCHANNELPUBLISHER_H
