/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief Class representing a Black Channel Salt Update message.
 */

#ifndef APTIVWARECORE_BLACKCHANNELSALTUPDATEMETADATA_H
#define APTIVWARECORE_BLACKCHANNELSALTUPDATEMETADATA_H

#include "BlackChannel/BlackChannelKey.h"

#include <memory>

namespace bc {

/**
 * \brief Class representing a Black Channel Salt Update message.
 *
 * Salt Message {
 *     uint32 valid_through
 *     uint8 salt[SALT_SZ]
 * }
 */
class BlackChannelSaltUpdateMetadata final
{
public:
    /**
     * \brief Construct a BlackChannelSaltUpdateMetadata instance.
     * \param validThrough The sequence number of the last Black Channel protected message that is protected by the
     *        current session keys.
     * \param key The salt.
     */
    BlackChannelSaltUpdateMetadata(const std::uint32_t validThrough, const BlackChannelKey& key)
        : validThrough_{validThrough}
        , key_{key}
    {}

    /**
     * \brief Destructor.
     */
    ~BlackChannelSaltUpdateMetadata() = default;

    /**
     * \brief Accessor for our validThrough value.
     * \returns The validThrough value.
     */
    std::uint32_t validThrough() const noexcept { return validThrough_; }

    /**
     * \brief Accessor for our key (salt).
     * \returns The salt value.
     */
    const BlackChannelKey& key() const noexcept { return key_; }

    /**
     * \brief Encode a string represention of this instance.
     * \returns The encoded string.
     */
    std::string toString() const;

    /**
     * \brief Determine whether the given string represents encoded Salt Update metadata.
     * \param s The string to classify.
     * \returns True if the string, \a s, represents Salt Update metadata,false otherwise.
     */
    static bool isSaltUpdateMetadata(const std::string& s);

    /**
     * \brief Decode a string representation of Salt Update metadata.
     * \param s The string to decode.
     * \returns A new BlackChannelSaltUpdateMetadata instance.
     */
    static BlackChannelSaltUpdateMetadata fromString(const std::string& s);

private:
    std::uint32_t validThrough_{0};
    BlackChannelKey key_{};
};

} // namespace bc

#endif // APTIVWARECORE_BLACKCHANNELSALTUPDATEMETADATA_H
