/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief a simple BlackChannelSubscriber class implementation
 */

#ifndef APTIVWARECORE_BLACKCHANNELSUBSCRIBER_H
#define APTIVWARECORE_BLACKCHANNELSUBSCRIBER_H

#include "BlackChannel/BlackChannel.h"
#include <BlackChannel/BlackChannelSaltMetadata.h>
#include <BlackChannel/BlackChannelSaltUpdateMetadata.h>

#include <tuple>

namespace bc {
/**
 * @brief BlackChannelSubscriber class
 *
 * A wrapper class provides APIs for our message infrastructure to integrate QNX Black Channel Communication Technology
 * (QBCCT) APIs.
 */
class BlackChannelSubscriber final
{
public:
    /**
     * @brief Constructor.
     */
    BlackChannelSubscriber() = default;

    /**
     * @brief Destructor
     */
    ~BlackChannelSubscriber() = default;

    /**
     * @brief Determine whether the subscriber is awaiting its first message.
     * @returns A boolean indicating whether we are waiting for the first (non-error) message.
     */
    bool isAwaitingFirstMessage() const noexcept { return (m_state == SubscriberState::AwaitingFirstMessage); }

    /**
     * @brief Update the subscriber state after having processed the first message.
     */
    void processedFirstMessage() noexcept { m_state = SubscriberState::Salted; }

    /**
     * @brief verify a black channel message using QBCCT profile 17
     *
     * @param bcMsg - black channel message to check, usually a message produced by protect() operation
     * @return - true if the check operation is successful, false otherwise
     */
    bool check(const std::string& bcMsg) noexcept;

    /**
     * @brief Process incoming metadata, updating our state.
     * @param metadata The metadata received for this subscriber.
     * @returns A boolean indicating whether the metadata was successfully processed.
     */
    bool processMetadata(const std::string& metadata);

private:
    ///< The subscriber state
    enum class SubscriberState
    {
        AwaitingFirstMessage, ///< Awaiting receipt of the first (non-error) message on the channel
        Salted,               ///< We have a valid salt, and have seen our first message
        Resalting             ///< We are awaiting completion of the salt update process
    };

    /**
     * @brief Initialize/reinitialize our check data structures.
     */
    void checkInit();

    /**
     * @brief verify a black channel message using QBCCT profile 17
     *
     * @param data - black channel message to check, usually a message produced by protect() operation
     * @param size - the size of the message
     * @return - true if the check operation is successful, false otherwise
     */
    std::tuple<bool, bc_result_t> check(const std::uint8_t* const data, const std::size_t size) noexcept;

    /**
     * @brief Process incoming salt metadata, updating our state.
     * @param metadata The salt metadata received for this subscriber.
     * @returns A boolean indicating whether the salt metadata was successfully processed.
     */
    bool processSaltMetadata(const BlackChannelSaltMetadata& saltMetadata);

    /**
     * @brief Process incoming salt update metadata, updating our state.
     * @param metadata The salt update metadata received for this subscriber.
     * @returns A boolean indicating whether the salt update metadata was successfully processed.
     */
    bool processSaltUpdateMetadata(const BlackChannelSaltUpdateMetadata& saltUpdateMetadata);

    bc_p17_config_t m_rx_config{};     ///< Configuration for transmitting data using QBCCT profile 17.
    bc_p17_check_state_t m_rx_state{}; ///< Control data for a QBCCT profile 17 protection header.
    BlackChannelKey m_salt{};          ///< the current salt
    BlackChannelKey m_pendingSalt{};   ///< the next salt to use upon reaching the validThrough counter
    std::uint32_t m_validThrough;      ///< the upper bound of the valid range of counter values for a salt update
    BCKeyTable m_keyTable{};           ///< the session key table
    SubscriberState m_state{SubscriberState::AwaitingFirstMessage}; ///< Our state
};

} // namespace bc

#endif // APTIVWARECORE_BLACKCHANNELSUBSCRIBER_H
