/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Short description of the contents of the file.

#include "platform/middleware/libs/BlackChannel/include/BlackChannel/BlackChannelKey.h"
#include <gtest/gtest.h>

using namespace bc;

const BCByteArray fullKey{{0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF, 0x01, 0x23, 0x45,
                           0x67, 0x89, 0xAB, 0xCD, 0xEF, 0x01, 0x23, 0x45, 0x67, 0x89, 0xAB,
                           0xCD, 0xEF, 0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF}};

static BCByteArray createKey(std::uint8_t fillValue)
{
    BCByteArray key{};
    key.fill(fillValue);
    return key;
}

TEST(BlackChannelKeyTest, constructor)
{
    EXPECT_NO_THROW(BlackChannelKey());
    EXPECT_NO_THROW(BlackChannelKey(createKey(0x00)));
}

TEST(BlackChannelKeyTest, getBCKey)
{
    // Verify object stores the key passed during construction.
    for (const auto& key : {createKey(0xAA), createKey(0x00)}) {
        EXPECT_EQ(BlackChannelKey(key).getBCKey(), key);
    }
}

TEST(BlackChannelKeyTest, setBCKey)
{
    // Verify object updates its key with the key passed to setBCKey
    BlackChannelKey objectUnderTest;
    for (const auto& key : {createKey(0xAA), createKey(0x00)}) {
        objectUnderTest.setBCKey(key);
        EXPECT_EQ(objectUnderTest.getBCKey(), key);
    }
}

TEST(BlackChannelKeyTest, isBCKeyValid)
{
    // Verify key is valid only when set during construction.
    {
        BlackChannelKey objectUnderTest;
        EXPECT_FALSE(objectUnderTest.isBCKeyValid());
    }
    {
        BlackChannelKey objectUnderTest(createKey(0x00));
        EXPECT_TRUE(objectUnderTest.isBCKeyValid());
    }
    {
        // TODO: This tests captures current behavior, but doesn't seem to be what you'd
        // expect.  You'd expect setting a key would update the validity.
        BlackChannelKey objectUnderTest;
        objectUnderTest.setBCKey({});
        EXPECT_FALSE(objectUnderTest.isBCKeyValid());
    }
}

TEST(BlackChannelKeyTest, toString)
{
    // Verify toString creates a hex digit formatted string for the key.
    EXPECT_EQ("0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef", BlackChannelKey(fullKey).toString());

    EXPECT_EQ("0000000000000000000000000000000000000000000000000000000000000000", BlackChannelKey().toString());

    EXPECT_EQ("0000000000000000000000000000000000000000000000000000000000000000",
              BlackChannelKey(createKey(0x00)).toString());
}

TEST(BlackChannelKeyTest, fromString)
{
    // Verify default constructed key is returned for invalid number of digits.
    EXPECT_EQ(BlackChannelKey::fromString(""), BlackChannelKey());
    EXPECT_EQ(BlackChannelKey::fromString("0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcde"),
              BlackChannelKey());
    EXPECT_EQ(BlackChannelKey::fromString("0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0"),
              BlackChannelKey());

    // Verify default constructed key is return if any digit is invalid.
    EXPECT_EQ(BlackChannelKey::fromString("g123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef"),
              BlackChannelKey());
    EXPECT_EQ(BlackChannelKey::fromString("0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdeg"),
              BlackChannelKey());

    EXPECT_EQ(BlackChannelKey::fromString("0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef"),
              BlackChannelKey(fullKey));
}

TEST(BlackChannelKeyTest, comparisonOperators)
{
    EXPECT_TRUE(BlackChannelKey() == BlackChannelKey());
    EXPECT_TRUE(BlackChannelKey(createKey(0x00)) == BlackChannelKey(createKey(0x00)));

    EXPECT_TRUE(BlackChannelKey(createKey(0x01)) != BlackChannelKey(createKey(0x00)));
    EXPECT_TRUE(BlackChannelKey(createKey(0x00)) != BlackChannelKey(createKey(0x01)));

    BlackChannelKey operand;
    operand.setBCKey(createKey(0xAA));
    EXPECT_TRUE(operand == BlackChannelKey(createKey(0xAA)));
    EXPECT_TRUE(BlackChannelKey(createKey(0xAA)) == operand);
}
