/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief unit tests for BlackChannel related classes
 */

#include <Clock/SystemClock.h>
#include <Comms/ChannelFactory.h>
#include <Comms/SCChannel.h>
#include <config/ConfigParser.h>
#include <gtest/gtest.h>
#include <interfaces/baseTypes/Header.h>

#include <chrono>
#include <cstring>

const std::string TEST_DATA_DIR = std::getenv("TEST_DATA_DIR");

using namespace std::literals;

namespace bcTest {

const auto saltPublishingDelay = 1100ms; // Salt is published every 1000ms
const auto messageDelay = 100ms;         // Allow for wire delays

inline void addInterfaceHeaderInfo(const std::string& msg, std::string& rawMsg)
{
    interfaces::baseTypes::InterfaceHeader hdr;
    struct otimeSpec sendTime = SystemClock::getSystemTime().to_otimeSpec();
    interfaces::baseTypes::setHeaderVersion(hdr, OTTO_INTERFACE_HEADER_VERSION);
    interfaces::baseTypes::setSendTimeStamp(hdr, sendTime);
    std::size_t hdrSize = interfaces::baseTypes::getInterfaceHeaderSize(OTTO_INTERFACE_HEADER_VERSION);
    const std::string header((char*)&hdr, hdrSize);

    rawMsg.append(header);
    rawMsg.append(msg);
}

inline void removeInterfaceHeaderInfo(std::string& rawMsg, std::string& msg)
{
    interfaces::baseTypes::InterfaceHeader hdr;
    std::size_t hdrSize = interfaces::baseTypes::getInterfaceHeaderSize(OTTO_INTERFACE_HEADER_VERSION);
    std::memcpy((char*)&hdr, rawMsg.c_str(), hdrSize);
    rawMsg = rawMsg.erase(0, hdrSize);

    msg = rawMsg;
}

class BlackChannelBasicTest : public ::testing::Test
{
public:
    void SetUp() override
    {
        std::string fileName = std::string(TEST_DATA_DIR) + "/BlackChannelTestConfig.cfg";
        std::string sectionName = "sampleBlackChannelTest";
        ASSERT_TRUE(parser_.parseFile(fileName)) << "Failed to parse file: " << fileName;
        ASSERT_TRUE(parser_.hasSection(sectionName)) << sectionName << " Not found!";

        auto configSection = parser_.getSection(sectionName, section_);
        ASSERT_TRUE(configSection) << "Cannot parse section: " << sectionName;

        testBCPub_ = dynamic_cast<task::SCChannel*>(task::ChannelFactory::createChannel(section_, task::Access::Write));
        EXPECT_NE(nullptr, testBCPub_);
        testBCSub_ = dynamic_cast<task::SCChannel*>(task::ChannelFactory::createChannel(section_, task::Access::Read));
        EXPECT_NE(nullptr, testBCSub_);

        // The salt update interval comes from the configuration file. (This is used internally by our channels.)
        const std::string resaltEntryName{"resaltIntervalSeconds"};
        double resaltIntervalSeconds{0.0};
        auto resaltEntry = section_.get(resaltEntryName, resaltIntervalSeconds);
        ASSERT_EQ(resaltEntry, ConfigSection::GS_OK) << "Cannot retrieve value: " << resaltEntryName;

        saltUpdateInterval_ =
            std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::duration<double>(resaltIntervalSeconds));
    }

    void TearDown() override
    {
        delete testBCPub_;
        delete testBCSub_;
    }

protected:
    ConfigParser parser_{};
    ConfigSection section_{};
    task::SCChannel* testBCPub_{nullptr};
    task::SCChannel* testBCSub_{nullptr};
    bool stopReceiving{false};
    std::chrono::milliseconds saltUpdateInterval_{0u};
};

TEST_F(BlackChannelBasicTest, BasicConnectionTest)
{
    testBCSub_->connect();
    testBCPub_->connect();

    // wait to make sure salt is delivered to subscriber
    std::this_thread::sleep_for(saltPublishingDelay);

    testBCPub_->disconnect();
    testBCSub_->disconnect();
}

TEST_F(BlackChannelBasicTest, BasicOnePublisherAndOneSubscriberTest)
{
    const std::string messageToSend{"this is a black channel test!"};
    std::string messageToReceive;

    testBCSub_->connect();
    testBCPub_->connect();

    // wait to make sure salt is delivered to subscriber
    std::this_thread::sleep_for(saltPublishingDelay);

    std::string tmpSendMsg;
    addInterfaceHeaderInfo(messageToSend, tmpSendMsg);
    testBCPub_->write(tmpSendMsg);

    std::this_thread::sleep_for(messageDelay);

    std::string tmpRecvMsg;
    task::ChannelStatus readStatus = testBCSub_->read(tmpRecvMsg);
    EXPECT_EQ(task::ChannelStatus::CS_NewData, readStatus);

    removeInterfaceHeaderInfo(tmpRecvMsg, messageToReceive);
    EXPECT_EQ(messageToSend, messageToReceive);

    testBCPub_->disconnect();
    testBCSub_->disconnect();
}

TEST_F(BlackChannelBasicTest, BasicSaltUpdateTest)
{
    const std::string messageToSend{"this is a black channel salt update test!"};
    std::string messageToReceive;

    const auto exchangeMessage = [&, this]() {
        std::string tmpSendMsg;
        addInterfaceHeaderInfo(messageToSend, tmpSendMsg);
        testBCPub_->write(tmpSendMsg);

        std::this_thread::sleep_for(messageDelay);

        std::string tmpRecvMsg;
        task::ChannelStatus readStatus = testBCSub_->read(tmpRecvMsg);
        EXPECT_EQ(task::ChannelStatus::CS_NewData, readStatus);

        removeInterfaceHeaderInfo(tmpRecvMsg, messageToReceive);
        EXPECT_EQ(messageToSend, messageToReceive);
    };

    testBCSub_->connect();
    testBCPub_->connect();

    // wait to make sure salt is delivered to subscriber
    std::this_thread::sleep_for(saltPublishingDelay);

    // here, we are using the initial salt value
    exchangeMessage();

    // wait for the salt update to be published (wait _at least_ the messageDelay)
    std::this_thread::sleep_for(std::max(saltUpdateInterval_ - saltPublishingDelay, messageDelay));

    // here we are still using the initial salt value, as we need to read at least one message before the update can be
    // processed
    exchangeMessage();

    // here, we are using the updated salt value
    exchangeMessage();

    testBCPub_->disconnect();
    testBCSub_->disconnect();
}

class BlackChannelAdvancedTest : public ::testing::Test
{
public:
    void SetUp() override
    {
        std::string fileName = std::string(TEST_DATA_DIR) + "/BlackChannelTestConfig.cfg";
        std::string sectionName = "sampleBlackChannelTest";
        ASSERT_TRUE(parser_.parseFile(fileName)) << "Failed to parse file: " << fileName;
        ASSERT_TRUE(parser_.hasSection(sectionName)) << sectionName << " Not found!";

        auto configSection = parser_.getSection(sectionName, section_);
        ASSERT_TRUE(configSection) << "Cannot parse section: " << sectionName;

        const std::string bcModule1 = "sampleBlackChannelModule1";
        ASSERT_TRUE(task::ChannelFactory::connectAs(bcModule1))
            << "Failed to connect module: " << bcModule1 << "to channel factory!";
        testBCPub1_ =
            dynamic_cast<task::SCChannel*>(task::ChannelFactory::createChannel(section_, task::Access::Write));
        EXPECT_NE(nullptr, testBCPub1_);

        const std::string bcModule2 = "sampleBlackChannelModule2";
        ASSERT_TRUE(task::ChannelFactory::connectAs(bcModule2))
            << "Failed to connect module: " << bcModule2 << "to channel factory!";
        testBCPub2_ =
            dynamic_cast<task::SCChannel*>(task::ChannelFactory::createChannel(section_, task::Access::Write));
        EXPECT_NE(nullptr, testBCPub2_);

        const std::string bcModule3 = "sampleBlackChannelModule3";
        ASSERT_TRUE(task::ChannelFactory::connectAs(bcModule3))
            << "Failed to connect module: " << bcModule3 << "to channel factory!";
        testBCSub_ = dynamic_cast<task::SCChannel*>(task::ChannelFactory::createChannel(section_, task::Access::Read));
        EXPECT_NE(nullptr, testBCSub_);
    }

    void TearDown() override
    {
        delete testBCPub1_;
        delete testBCPub2_;
        delete testBCSub_;
    }

protected:
    ConfigParser parser_{};
    ConfigSection section_{};
    task::SCChannel* testBCPub1_{nullptr};
    task::SCChannel* testBCPub2_{nullptr};
    task::SCChannel* testBCSub_{nullptr};
};

TEST_F(BlackChannelAdvancedTest, TwoPublisherUsingSeparateChannelsAndOneSubscriberSingleMessageTest)
{
    const std::string msgToSend1{"this is a black channel test from publisher 1!"};
    const std::string msgToSend2{"this is a black channel test from publisher 2!"};
    std::string msgToRecv;

    testBCSub_->connect();
    testBCPub1_->connect();
    testBCPub2_->connect();

    // wait to make sure salt is delivered to subscriber
    std::this_thread::sleep_for(saltPublishingDelay);

    std::string tmpSendMsg1{""};
    addInterfaceHeaderInfo(msgToSend1, tmpSendMsg1);
    EXPECT_EQ(task::CS_Success, testBCPub1_->write(tmpSendMsg1));

    std::string tmpSendMsg2{""};
    addInterfaceHeaderInfo(msgToSend2, tmpSendMsg2);
    EXPECT_EQ(task::CS_Success, testBCPub2_->write(tmpSendMsg2));

    task::ChannelStatus readStatus;
    do {
        std::string tmpRecvMsg1{""};
        readStatus = testBCSub_->read(tmpRecvMsg1);

        if (readStatus == task::CS_NewData) {
            removeInterfaceHeaderInfo(tmpRecvMsg1, msgToRecv);
            EXPECT_TRUE(msgToSend1 == msgToRecv || msgToSend2 == msgToRecv);
        }
        std::this_thread::sleep_for(messageDelay);
    } while (readStatus != task::CS_NewData);

    do {
        std::string tmpRecvMsg2{""};
        readStatus = testBCSub_->read(tmpRecvMsg2);

        if (readStatus == task::CS_NewData) {
            removeInterfaceHeaderInfo(tmpRecvMsg2, msgToRecv);
            EXPECT_TRUE(msgToSend1 == msgToRecv || msgToSend2 == msgToRecv);
        }
        std::this_thread::sleep_for(messageDelay);
    } while (readStatus != task::CS_NewData);

    testBCPub1_->disconnect();
    testBCPub2_->disconnect();
    testBCSub_->disconnect();
}

TEST_F(BlackChannelAdvancedTest, TwoPublisherUsingSeparateChannelsAndOneSubscriberMultipleMessageTest)
{
    const std::string msgToSend1Base{"this is a black channel test from publisher 1: "};
    const std::string msgToSend2Base{"this is a black channel test from publisher 2: "};
    const uint32_t msgNum = 5;
    std::vector<std::string> msgList1ToSend, msgList2ToSend;

    for (size_t i = 0; i < msgNum; i++) {
        msgList1ToSend.push_back(msgToSend1Base + std::to_string(i));
    }

    for (size_t i = 0; i < msgNum; i++) {
        msgList2ToSend.push_back(msgToSend2Base + std::to_string(i));
    }

    std::string msgToRecv;

    testBCSub_->connect();
    testBCPub1_->connect();
    testBCPub2_->connect();

    // wait to make sure salt is delivered to subscriber
    std::this_thread::sleep_for(saltPublishingDelay);

    for (size_t i = 0; i < msgNum; i++) {
        std::string tmpSendMsg1{""};
        addInterfaceHeaderInfo(msgList1ToSend[i], tmpSendMsg1);
        EXPECT_EQ(task::CS_Success, testBCPub1_->write(tmpSendMsg1));

        std::this_thread::sleep_for(messageDelay);

        std::string tmpSendMsg2{""};
        addInterfaceHeaderInfo(msgList2ToSend[i], tmpSendMsg2);
        EXPECT_EQ(task::CS_Success, testBCPub2_->write(tmpSendMsg2));

        std::this_thread::sleep_for(messageDelay);

        std::string tmpRecvMsg1{""};
        task::ChannelStatus readStatus = testBCSub_->read(tmpRecvMsg1);
        EXPECT_EQ(task::ChannelStatus::CS_NewData, readStatus);
        removeInterfaceHeaderInfo(tmpRecvMsg1, msgToRecv);
        EXPECT_EQ(msgList1ToSend[i], msgToRecv);

        std::string tmpRecvMsg2{""};
        readStatus = testBCSub_->read(tmpRecvMsg2);
        EXPECT_EQ(task::ChannelStatus::CS_NewData, readStatus);
        removeInterfaceHeaderInfo(tmpRecvMsg2, msgToRecv);
        EXPECT_EQ(msgList2ToSend[i], msgToRecv);
    }

    testBCPub1_->disconnect();
    testBCPub2_->disconnect();
    testBCSub_->disconnect();
}

} // namespace bcTest
