/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief Black Channel benchmark test for APIs provided by QBCCT
 */

extern "C"
{
#include <bc/bc_qnx_p17.h>
}

#include <benchmark/benchmark.h>
#include <cstdint>
#include <cstring>
#include <random>

using namespace std;

static constexpr uint32_t UT_KEY_TABLE_SIZE = 2;
static constexpr uint32_t UT_KEY_LENGTH = 32;

static constexpr uint32_t HEADER_SIZE_SHA256_FULL = 40;
static constexpr uint32_t DEFAULT_MAX_PAYLOAD_SIZE = 4194304; // 4M bytes

static std::array<std::array<uint8_t, UT_KEY_LENGTH>, UT_KEY_TABLE_SIZE> ut_key_tbl{
    {{{0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15,
       16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31}},
     {{32, 33, 35, 34, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
       48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63}}}};

static uint8_t* ut_key_ptr_tbl[UT_KEY_TABLE_SIZE] = {ut_key_tbl[0].data(), ut_key_tbl[1].data()};

const bc_p17_config_t g_config_default_256 = {0,
                                              HEADER_SIZE_SHA256_FULL,
                                              HEADER_SIZE_SHA256_FULL + DEFAULT_MAX_PAYLOAD_SIZE,
                                              2,
                                              ut_key_ptr_tbl,
                                              UT_KEY_LENGTH,
                                              UT_KEY_TABLE_SIZE,
                                              BC_P17_HMAC_SIZE_SHA256,
                                              BC_P17_HMAC_SHA256};

class Data : public ::benchmark::Fixture
{
    static constexpr uint32_t RANDOM_DATA_SIZE = 10000;

public:
    void SetUp(const ::benchmark::State& state) override
    {
        ran_data_min_len_ = state.range(0);
        ran_data_max_len_ = state.range(1);

        data_.clear();
        vector<uint8_t> s = generateUInt8Vec();

        for (size_t i = 0; i < RANDOM_DATA_SIZE; ++i) {
            std::shuffle(s.begin(), s.end(), getEngineInstance());
            data_.emplace_back(s);
        }

        std::memcpy(&tx_config_, &g_config_default_256, sizeof(g_config_default_256));

        auto result = bc_p17_protect_init(&tx_config_, &tx_state_);
        if (result != BC_OK) {
            throw std::runtime_error("Black Channel profile 17 protect init failed!");
        }

        std::memcpy(&rx_config_, &g_config_default_256, sizeof(g_config_default_256));
        result = bc_p17_check_init(&rx_config_, &rx_state_);
        if (result != BC_OK) {
            throw std::runtime_error("Black Channel profile 17 check init failed!");
        }
    }

    void TearDown(const ::benchmark::State& state) override {}

    std::vector<uint8_t> nextData()
    {
        static size_t i = 0;

        if (data_.empty()) {
            return generateUInt8Vec();
        }
        return data_[i++ % data_.size()];
    }

private:
    std::vector<uint8_t> generateUInt8Vec()
    {
        uniform_int_distribution<> lenGenerator(ran_data_min_len_, ran_data_max_len_);

        string val = generate(lenGenerator(getEngineInstance()));

        return {val.begin(), val.end()};
    }

    inline static std::string generate(size_t max_length)
    {
        string possible_characters = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
        uniform_int_distribution<> dist(0, possible_characters.size() - 1);
        string ret;

        for (std::size_t i = 0; i < max_length; i++) {
            std::size_t random_index =
                dist(getEngineInstance()); // get index between 0 and possible_characters.size()-1
            ret += possible_characters[random_index];
        }
        return ret;
    }

    static mt19937& getEngineInstance()
    {
        static random_device rd;
        static mt19937 engine(rd());

        return engine;
    }

private:
    uint32_t ran_data_min_len_{};
    uint32_t ran_data_max_len_{};

public:
    static std::vector<std::vector<uint8_t>> data_;
    static bc_p17_config_t tx_config_;
    static bc_p17_protect_state_t tx_state_;
    static bc_p17_config_t rx_config_;
    static bc_p17_check_state_t rx_state_;
};

std::vector<std::vector<uint8_t>> Data::data_{};
bc_p17_config_t Data::tx_config_{};
bc_p17_protect_state_t Data::tx_state_{};
bc_p17_config_t Data::rx_config_{};
bc_p17_check_state_t Data::rx_state_{};

BENCHMARK_DEFINE_F(Data, bc_p17_protect)(benchmark::State& state)
{
    for (auto _ : state) {
        std::vector<uint8_t> data = Data::nextData();
        benchmark::DoNotOptimize(bc_p17_protect(&Data::tx_config_, &Data::tx_state_, &data[0], data.size()));
    }
}

// the length of data used by bc_p17_protect() call wil be 1-10B, 10-100B, 100-1KB, 1k-10KB, 10-100KB, 100KB-1MB, 1-4MB
// bytes
BENCHMARK_REGISTER_F(Data, bc_p17_protect)
    ->Args({1, 10})
    ->Args({10, 100})
    ->Args({100, 1024})
    ->Args({1024, 10 * 1024})
    ->Args({10 * 1024, 100 * 1024})
    ->Args({100 * 1024, 1024 * 1024});
//  ->Args({1024 * 1024, 4 * 1024 * 1024});

BENCHMARK_DEFINE_F(Data, bc_p17_check)(benchmark::State& state)
{
    for (auto _ : state) {
        std::vector<uint8_t> data = Data::nextData();
        benchmark::DoNotOptimize(bc_p17_check(&Data::rx_config_, &Data::rx_state_, &data[0], data.size()));
    }
}

// the length of data used by bc_p17_check() call wil be 1-10B, 10-100B, 100-1KB, 1k-10KB, 10-100KB, 100KB-1MB, 1-4MB
// bytes
BENCHMARK_REGISTER_F(Data, bc_p17_check)
    ->Args({1, 10})
    ->Args({10, 100})
    ->Args({100, 1024})
    ->Args({1024, 10 * 1024})
    ->Args({10 * 1024, 100 * 1024})
    ->Args({100 * 1024, 1024 * 1024});
//  ->Args({1024 * 1024, 4 * 1024 * 1024});

BENCHMARK_MAIN();