/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file SystemClock.cc
 * @product common
 * @version
 * @released
 * @updated 02/05/2019
 * @brief System clock wrapper utility
 *
 * @section License
 *
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#include <Clock/ClockSource.h>
#include <Clock/SystemClock.h>
#include <Clock/SystemClockUtils.h>
#include <utility/utility.h>

#include "lib/avlog/avlog.hpp"

namespace ottotime {

bool SystemClock::initialized_ = false;
SystemClock::SystemTimeMode SystemClock::mode_ = SystemClock::SystemTimeMode::TimeModeSteady;
ClockParamsClock* SystemClock::externalClock_ =
    nullptr;                                // if an interface is driving the clock externally, this is a pointer to it.
Clock* SystemClock::customClock_ = nullptr; // if an interface is driving the clock externally, this is a pointer to it.
std::optional<PlaybackClockParams> SystemClock::externalClockParams_;
ottotime::otimePoint SystemClock::localBaseTime_;
ottotime::otimePoint SystemClock::synchronizedBaseTime_;
ottotime::otimePoint SystemClock::steadyBaseTime_;
std::shared_ptr<Clock> SystemClock::s_frozenClock{};

SystemClock::SystemClock()
{
    if (!initializeLocalClocks_()) {
        AVLOG_CRITICAL("SystemClock failed to initialize local clocks!");
    }
}

SystemClock::~SystemClock() {}

/* static */ void SystemClock::initialize(Clock* userClock)
{
    AVLOG_ERROR("Something called the deprecated legacy SystemClock::initialize()!  The main() for this "
                "module/app needs to be updated!");

    if (userClock != nullptr) {
        initialize(SystemTimeMode::TimeModeCustomClock, userClock);
    } else {
        if (SystemClockUtils::getEnvironmentTimeMode() == SystemTimeMode::TimeModeExternallyDriven) {
            AVLOG_ERROR("Deprecated legacy SystemClock::initialize() was called when the environment wants the "
                        "clock to be externally driven for simulation/resim!  This won't work, the playback "
                        "clock will be broken for this process!");
        }

        initialize(SystemTimeMode::TimeModeSteady);
    }
}

/* static */ bool SystemClock::initialize(const SystemTimeMode& mode, Clock* userClock /* default: nullptr */)
{
    if (initialized_) {
        AVLOG_ERROR("SystemClock::initialize() called more than once!");
        return false;
    }

    return reinitialize(mode, userClock);
}

namespace {
class StoppedClock : public Clock
{
public:
    StoppedClock(ottotime::otimePoint fixedTime)
        : fixedTime_(std::move(fixedTime))
    {}

    ottotime::otimePoint getCurrentTime() override { return fixedTime_; }

private:
    const ottotime::otimePoint fixedTime_;
};

} // namespace

/* static */ void SystemClock::freeze()
{
    s_frozenClock = std::make_shared<StoppedClock>(getSystemTime());
    reinitialize(SystemTimeMode::TimeModeCustomClock, s_frozenClock.get());
}

/* static */ bool SystemClock::reinitialize(const SystemTimeMode& mode, Clock* userClock)
{
    bool initStatus = true;

    initialized_ = true;
    mode_ = mode;

    if (!initializeSteadyClocks_()) {
        AVLOG_CRITICAL("SystemClock failed to initialize steady clocks!");
        initStatus = false;
    }

    switch (mode_) {
        case SystemTimeMode::TimeModeSteady: {
            if (userClock != nullptr) {
                AVLOG_CRITICAL("Something tried to pass a user-specified clock while asking for TimeModeSteady!");
                initStatus = false;
            }
            AVLOG_INFO("SystemClock in mode TimeModeSteady.");
            break;
        }
        case SystemTimeMode::TimeModeExternallyDriven: {
            if (userClock == nullptr) {
                AVLOG_CRITICAL("Something asked for TimeModeExternallyDriven but did NOT specify a clock!");

                // fall back to... something.
                AVLOG_CRITICAL("SystemClock falling back to TimeModeSteady.  The clock for this process may not "
                               "work as intended!");
                mode_ = SystemTimeMode::TimeModeSteady;
                initStatus = false;
            } else {
                const bool extStatus = initializeExternalClock_(userClock);
                if (!extStatus) {
                    AVLOG_CRITICAL("SystemClock failed to initialize externally driven clock!");
                    // fall back to... something.
                    AVLOG_CRITICAL("SystemClock falling back to TimeModeSteady.  Simulation/playback may not "
                                   "work as intended!");
                    mode_ = SystemTimeMode::TimeModeSteady;
                    initStatus = false;
                } else {
                    AVLOG_INFO("SystemClock in mode TimeModeExternallyDriven.");
                }
            }
            break;
        }
        case SystemTimeMode::TimeModeCustomClock: {
            if (userClock == nullptr) {
                AVLOG_CRITICAL("Something asked for TimeModeCustomClock but did NOT specify a clock!");

                // fall back to... something.
                AVLOG_CRITICAL("SystemClock falling back to TimeModeSteady.  The clock for this process may not "
                               "work as intended!");
                mode_ = SystemTimeMode::TimeModeSteady;
                initStatus = false;
            } else {
                AVLOG_INFO("SystemClock in mode TimeModeCustomClock.");
                customClock_ = userClock;
            }
            break;
        }
        default: {
            AVLOG_CRITICAL("FATAL ERROR: SystemClock didn't get a valid initialization mode! (Did you link against "
                           "the wrong version of the library?) Value: {:#010x} ({})",
                           AVLOG_ARGS("hex", "decimal"),
                           static_cast<uint64_t>(mode_),
                           static_cast<int64_t>(mode_));
            // If you managed to do this, you deserve immediate death.
            // (Probably you're linking against the wrong version of the library or something really bad like that.)
            initStatus = false;
            abort();
        }
    }
    if (!initStatus) {
        AVLOG_CRITICAL("Something went wrong while initializing SystemClock!");
    }

    AVLOG_INFO("SystemClock current system time: {} ({})",
               AVLOG_ARGS("time", "timeUTC"),
               SystemClockUtils::getTimePointString(getSystemTime()),
               SystemClockUtils::getTimePointString(getSystemTime(),
                                                    SystemClockUtils::TimeZone::TimeZoneUTC,
                                                    SystemClockUtils::TimeFormat::TimeFormatHuman));

    AVLOG_DEBUG("SystemClock current wall clock time: {}", SystemClockUtils::getWallClockTimeString());
    AVLOG_DEBUG("SystemClock current local steady time duration: {}",
                SystemClockUtils::getTimeDurationString(getLocalSteadyTime()));
    AVLOG_DEBUG("SystemClock current synchronized steady time: {}",
                SystemClockUtils::getTimePointString(getSynchronizedSteadyTime()));
    AVLOG_DEBUG("SystemClock current synchronized system time: {}",
                SystemClockUtils::getTimePointString(getSynchronizedSystemTime()));

    return initStatus;
}

/* static */ bool SystemClock::isInitialized()
{
    return initialized_;
}

/* static */ SystemClock::SystemTimeMode SystemClock::getTimeMode()
{
    return mode_;
}

/* static */ bool SystemClock::getPlaybackParams(PlaybackClockParams& outParams)
{
    outParams = PlaybackClockParams();
    bool retVal{false};
    bool isOK{true};

    if (mode_ != SystemTimeMode::TimeModeExternallyDriven) {
        AVLOG_ERROR("Something tried to ask SystemClock for clock playback parameters when not in the externally "
                    "driven clock mode!");
        isOK = false;
    }

    if (isOK && (externalClock_ == nullptr)) {
        AVLOG_ERROR("Something tried to ask SystemClock for clock playback parameters, but we can't find an "
                    "externally driven clock!");
        isOK = false;
    }

    if (isOK) {

        externalClockParams_ = externalClock_->getParams();
        isOK = externalClockParams_.has_value();

        if (isOK) {
            externalClockParams_->normalize();

            if (externalClockParams_->isValid()) {
                outParams = *externalClockParams_;
                retVal = true;
            } else {
                AVLOG_ERROR("SystemClock got a PlaybackClockParams from the playback interface, but it doesn't "
                            "seem to be valid!\n"
                            "   Base time: {}\n"
                            "   Offset: {}\n"
                            "   Playback speed: {}",
                            AVLOG_ARGS("baseTime", "offset", "scaleFactor"),
                            SystemClockUtils::getTimePointString(externalClockParams_->baseTime),
                            SystemClockUtils::getTimeDurationString(externalClockParams_->offset),
                            externalClockParams_->scaleFactor);
                isOK = false;
            }
        }
    }

    return retVal;
}

/* static */ bool SystemClock::initializeExternalClock_(Clock* externalClock)
{
    bool result{true};
    // Make sure the interface we instantiated is actually the class we expect,
    // and that it implements all the interfaces we expect it to fill.
    // On QNX we compile without RTTI, so we just have to assume this is a valid clock interface.
    ClockParamsClock* asPlaybackInterface = dynamic_cast<ClockParamsClock*>(externalClock);
    if (asPlaybackInterface == nullptr) {
        AVLOG_CRITICAL("SystemClock external clock interface wasn't actually a PlaybackClockParamsInput!");
        result = false;
    }

    if (result == true) {
        externalClock_ = asPlaybackInterface;
        AVLOG_INFO("SystemClock intialized externally driven clock interface.");
    }
    return result;
}

/* static */ bool SystemClock::initializeLocalClocks_()
{
    localBaseTime_.getCurrentSteadyTime();
    return true;
}

/* static */ bool SystemClock::initializeSteadyClocks_()
{
    synchronizedBaseTime_.getCurrentWallClockTime();
    steadyBaseTime_.getCurrentSteadyTime();
    return true;
}

/* static */ ottotime::otimePoint SystemClock::getSystemTime()
{
    return getSynchronizedSystemTime();
}

/* static */ ottotime::otimeDuration SystemClock::getSystemTimeSince(const ottotime::otimePoint& before)
{
    const auto& now = getSystemTime();
    if (mode_ == SystemTimeMode::TimeModeSteady) {
        if (now < before) {
            AVLOG_ERROR("SystemClock system time moved backwards in TimeModeSteady! now: '{}', before: '{}'",
                        AVLOG_ARGS("now", "before"),
                        SystemClockUtils::getTimePointString(now),
                        SystemClockUtils::getTimePointString(before));
        }
    }
    return (now - before);
}

/* static */ ottotime::otimeDuration SystemClock::getLocalSteadyTime()
{
    // call out to std::chrono to get the local steady time
    const ottotime::otimePoint now = ottotime::otimePoint::buildCurrentSteadyTime();
    if (now < localBaseTime_) {
        AVLOG_ERROR("SystemClock local steady time moved backwards! now: '{}', base: '{}'",
                    AVLOG_ARGS("now", "base"),
                    SystemClockUtils::getTimePointString(now),
                    SystemClockUtils::getTimePointString(localBaseTime_));
    }

    // get the difference from localBaseTime_ and return that
    return (now - localBaseTime_);
}

/* static */ ottotime::otimeDuration SystemClock::getLocalSteadyTimeSince(const ottotime::otimeDuration& before)
{
    const auto& now = getLocalSteadyTime();
    if (now < before) {
        AVLOG_ERROR("SystemClock local steady time moved backwards (or caller is confused)! now: '{}', before: '{}'",
                    AVLOG_ARGS("now", "before"),
                    SystemClockUtils::getTimeDurationString(now),
                    SystemClockUtils::getTimeDurationString(before));
    }
    return (now - before);
}

/* static */ ottotime::otimePoint SystemClock::getSynchronizedSteadyTime()
{
    ottotime::otimePoint returnValue{};
    // call out to std::chrono to get the steady time
    if (!initialized_) {
        AVLOG_ERROR("SystemClock::getSynchronizedSteadyTime() called while not initialized!");
        returnValue = ottotime::otimePoint();
    } else {
        const ottotime::otimePoint now = ottotime::otimePoint::buildCurrentSteadyTime();
        if (now < steadyBaseTime_) {
            AVLOG_ERROR("SystemClock synchronized steady time moved backwards! now: '{}', base: '{}'",
                        AVLOG_ARGS("now", "base"),
                        SystemClockUtils::getTimePointString(now),
                        SystemClockUtils::getTimePointString(steadyBaseTime_));
        }
        returnValue = (synchronizedBaseTime_ + (now - steadyBaseTime_));
    }
    return returnValue;
}

/* static */ ottotime::otimeDuration SystemClock::getSynchronizedSteadyTimeSince(const ottotime::otimePoint& before)
{
    ottotime::otimeDuration returnValue{};
    if (!initialized_) {
        AVLOG_ERROR("SystemClock::getSynchronizedSteadyTime() called while not initialized!");
        returnValue = ottotime::otimeDuration();
    } else {
        const auto& now = getSynchronizedSteadyTime();
        if (now < before) {
            AVLOG_ERROR(
                "SystemClock synchronized steady time moved backwards (or caller is confused)! now: '{}', before: '{}'",
                AVLOG_ARGS("now", "before"),
                SystemClockUtils::getTimePointString(now),
                SystemClockUtils::getTimePointString(before));
        }
        returnValue = now - before;
    }
    return returnValue;
}

/* static */ ottotime::otimePoint SystemClock::getSynchronizedSystemTime()
{
    ottotime::otimePoint returnValue{};
    if (!initialized_) {
        AVLOG_ERROR("SystemClock::getSynchronizedSystemTime() called while not initialized!");
        returnValue = ottotime::otimePoint();
    } else {
        switch (mode_) {
            case SystemTimeMode::TimeModeSteady:
                returnValue = getSynchronizedSteadyTime();
                break;

            case SystemTimeMode::TimeModeExternallyDriven:
                if (externalClock_ == nullptr) {
                    AVLOG_CRITICAL("SystemClock in externally driven mode, but couldn't find external clock!");
                    returnValue = ottotime::otimePoint();
                } else {
                    returnValue = externalClock_->getCurrentTime();
                }
                break;

            case SystemTimeMode::TimeModeCustomClock:
                if (customClock_ == nullptr) {
                    AVLOG_CRITICAL("SystemClock in custom clock mode, but couldn't find custom clock!");
                    returnValue = ottotime::otimePoint();
                } else {
                    returnValue = customClock_->getCurrentTime();
                }
                break;

            default:
                AVLOG_CRITICAL("SystemClock doesn't seem to have a valid mode after initialization! (Memory "
                               "corruption?) mode: {:#010x} ({})",
                               AVLOG_ARGS("hex", "decimal"),
                               static_cast<uint64_t>(mode_),
                               static_cast<int64_t>(mode_));
                // ...again, you've somehow done something extremely bad.  We should have caught this in initialize()
                // and killed you.  Time to die.
                returnValue = ottotime::otimePoint();
                abort();
        }
    }
    return returnValue;
}

/* static */ ottotime::otimeDuration SystemClock::getSynchronizedSystemTimeSince(const ottotime::otimePoint& before)
{
    const auto& now = getSynchronizedSystemTime();
    if (mode_ == SystemTimeMode::TimeModeSteady) {
        if (now < before) {
            AVLOG_ERROR("SystemClock system time moved backwards in steady mode (or caller is confused)! now: "
                        "'{}', before: '{}'",
                        AVLOG_ARGS("now", "before"),
                        SystemClockUtils::getTimePointString(now),
                        SystemClockUtils::getTimePointString(before));
        }
    }
    return (now - before);
}

} // end namespace ottotime
