/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file SystemClockUtils.cc
 * @product common
 * @version
 * @released 01/17/2020
 * @updated 01/17/2020
 * @brief System clock utility functions/implementations
 *
 * @section License
 *
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#include <Clock/SystemClock.h>
#include <Clock/SystemClockUtils.h>
#include <cinttypes>
#include <cmath>
#include <cstdlib>
#include <iomanip>
#include <sstream>
#include <utility/utility.h>

namespace SystemClockUtils {

SystemClock::SystemTimeMode getEnvironmentTimeMode()
{
    // TODO: make this a fatal error/assertion once everything has been updated.
    std::string systemClockType{};
    if (const char* clockEnvStr = getenv("SYSTEM_CLOCK_TYPE")) {
        systemClockType = clockEnvStr;
    }

    if (systemClockType.empty()) {
        std::cerr << "SystemClock: ERROR: SYSTEM_CLOCK_TYPE is not set.  Assuming a steady clock.  Simulation/Playback "
                     "will be broken, all launchers need to specify this!"
                  << std::endl;
        return SystemClock::SystemTimeMode::TimeModeSteady;
    }

    if (systemClockType == "local" || systemClockType == "steady") {
        return SystemClock::SystemTimeMode::TimeModeSteady;
    } else if (systemClockType == "external") {
        return SystemClock::SystemTimeMode::TimeModeExternallyDriven;
    } else {
        std::cerr << "SystemClock: ERROR: SYSTEM_CLOCK_TYPE is not set to a valid value.  Assuming a steady clock.  "
                     "Simulation/Playback will be broken, all launchers need to specify this!"
                  << std::endl;
        return SystemClock::SystemTimeMode::TimeModeSteady;
    }
}

class SystemTimeClock final : public Clock
{
    virtual ottotime::otimePoint getCurrentTime() override { return SystemClock::getSystemTime(); }
};

static SystemTimeClock sysTimeClock_;

Clock& getSystemTimeClock()
{
    return sysTimeClock_;
}

class SteadyTimeClock final : public Clock
{
    virtual ottotime::otimePoint getCurrentTime() override { return SystemClock::getSynchronizedSteadyTime(); }
};

static SteadyTimeClock steadyTimeClock_;

Clock& getSynchronizedSteadyTimeClock()
{
    return steadyTimeClock_;
}

class SynchronizedSystemTimeClock final : public Clock
{
    virtual ottotime::otimePoint getCurrentTime() override { return SystemClock::getSynchronizedSystemTime(); }
};

static SynchronizedSystemTimeClock synchronizedSystemTimeClock_;

Clock& getSynchronizedSystemTimeClock()
{
    return synchronizedSystemTimeClock_;
}

std::string getWallClockTimeString(SystemClockUtils::TimeZone zone /* default: TimeZone::TimeZoneUTC */,
                                   SystemClockUtils::TimeFormat fmt /* default: TimeFormat::TimeFormatISO8601 */)
{
    const ottotime::otimePoint now = ottotime::otimePoint::buildCurrentWallClockTime();
    return getTimePointString(now, zone, fmt);
}

std::string getTimePointString(const ottotime::otimePoint& point,
                               TimeZone /*zone*/ /* default: TimeZone::TimeZoneUTC */,
                               TimeFormat fmt /* default: TimeFormat::TimeFormatISO8601 */)
{
    if (!point.isTimeValid()) {
        return "<invalid timepoint>";
    }

    std::stringstream s;

    // TODO: handle timezone stuff!
    const tm pointAsTM = point.to_tm();

    // struct tm only goes to second resolution, so store off ms separately if we want to print those.
    constexpr int64_t NANOSECONDS_PER_MS = 1000000ULL;
    const auto pointMillis = static_cast<uint32_t>((point.to_otimeSpec().tv_nsec / NANOSECONDS_PER_MS) % 1000);

    switch (fmt) {
        case TimeFormat::TimeFormatISO8601: {
            s << std::put_time(&pointAsTM, "%FT%T%z");
            break;
        }
        case TimeFormat::TimeFormatISO8601Precise: {
            s << std::put_time(&pointAsTM, "%FT%T.") << std::setw(3) << std::setfill('0') << pointMillis
              << std::put_time(&pointAsTM, "%z");
            break;
        }
        case TimeFormat::TimeFormatHuman: {
            s << std::put_time(&pointAsTM, "%b %d %Y %T %Z");
            break;
        }
        case TimeFormat::TimeFormatHumanPrecise: {
            s << std::put_time(&pointAsTM, "%b %d %Y %T.") << std::setw(3) << std::setfill('0') << pointMillis
              << std::put_time(&pointAsTM, " %Z");
            break;
        }

        default:
            s << "<time format error>";
            break;
    }

    return s.str();
}

std::string getTimeDurationString(const ottotime::otimeDuration& duration,
                                  TimeFormat /*fmt*/ /* default: TimeFormat::TimeFormatISO8601 */)
{
    // TODO: support human-readable format!

    const int64_t totalMillis = duration.totalMilliSeconds();
    bool isNegative = totalMillis < 0;

    const auto posMillis = static_cast<uint64_t>(std::abs(totalMillis));
    const uint64_t printMillis = posMillis % 1000;

    const uint64_t totalSeconds = posMillis / 1000;
    const uint64_t printSeconds = totalSeconds % 60;

    const uint64_t totalMinutes = totalSeconds / 60;
    const uint64_t printMinutes = totalMinutes % 60;

    const uint64_t totalHours = totalMinutes / 60;
    const uint64_t printHours = totalHours % 24;

    const uint64_t totalDays = totalHours / 24;

    constexpr size_t DURATION_STRING_BUF_SIZE = 128;
    char durationStringBuffer[DURATION_STRING_BUF_SIZE];
    int printStatus = snprintf(durationStringBuffer,
                               DURATION_STRING_BUF_SIZE,
                               "P%s%" PRIu64 "DT%02" PRIu64 "H%02" PRIu64 "M%02" PRIu64 ".%03" PRIu64 "S",
                               isNegative ? "-" : "",
                               totalDays,
                               printHours,
                               printMinutes,
                               printSeconds,
                               printMillis);

    if ((printStatus < 0) || (printStatus >= static_cast<int>(DURATION_STRING_BUF_SIZE))) {
        return "<duration format error>";
    } else {
        return durationStringBuffer;
    }
}

} // end namespace SystemClockUtils
