/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file SystemClock.h
 * @product common
 * @version
 * @released
 * @updated 01/14/2020
 * @brief System clock wrapper utility
 *
 * @section License
 *
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#ifndef CLOCK_SYSTEMCLOCK_H
#define CLOCK_SYSTEMCLOCK_H

#include <Clock/ClockSource.h>
#include <interfaces/Clock/Playback/PlaybackClockParams.h>

/*!
 * @class SystemClock
 * @brief This class shall work as system wide clock and provides system's current time.
 * \n This will be same for all users and initialized only once with read only configured clock
 * \n or with default system clock if clock is not supplied in the configuration.
 * \n Task process, initializes all the interfaces during start up so System Clock shall be
 * \n initialized only after all the interfaces initialized as underlying clock implementation
 * \n is an interface. This is further coupled to playback since that uses a different clock Interface,
 * \n so the dependencies are significant to break.
 *
 * \n A separate class is necessary in order to decouple Task and the communications infrastructure.
 */

namespace ottotime {
// An internal thing for unit testing.  Please ignore.
class SystemClockScopedInitializer;

class SystemClock final
{
public:
    /*!
     * @brief Default constructor.
     */
    SystemClock();

    /*!
     * @brief Destructor.
     */
    ~SystemClock();

    /*!
     * @brief The system time modes
     */
    enum class SystemTimeMode
    {
        // synchronized, steady clock that increases monotonically.  The absolute time given is not
        // guaranteed to match the "wall clock" time in the real world, but in this mode you can
        // measure real-world durations accurately.
        TimeModeSteady,

        // synchronized clock that is being driven by something else.  Typically this is a simulation
        // or resim/playback task that may be running slower or faster than real time, synchronizing
        // the clock with the presentation of simulated or previously logged sensor data.
        //
        // The original plan was for SystemClock to internally create an interface-driven clock...
        // but doing that created a circular dependency between SystemClock and ottocomms that
        // was difficult to deal with.  So now you can call comms::getExternalClock() to create
        // this clock, and pass it to SystemClock::initialize() when using this mode.  (So essentially
        // this is identical to TimeModeCustomClock except that we require that the clock is really a
        // ClockPlaybackClient.)
        //
        // In the future we may move away from this type of clock using the interface system at all.
        //
        // This mode is not currently available on QNX platforms.
        TimeModeExternallyDriven,

        // Uses an arbitrary user-provided Clock.  This can be used to mock out the clock for testing,
        // or if a different externally-driven clock is required than the one provided by comms::getExternalClock().
        TimeModeCustomClock,
    };
    /*!
     * @brief Sets the behavior of SystemClock::getSystemTime() and SystemClock::getSynchronizedSystemTime().
     *        Currently this must be called *after* the communication interfaces are set up,
     *        since the external clock driver mechanism is an Interface.
     *
     * @param mode : the desired time mode.
     * @param userClock: for TimeModeCustom, the clock to use.  Otherwise this must be null.
     *
     * @return true if the SystemClock initialized properly, false otherwise.
     */
    static bool initialize(const SystemTimeMode& mode, Clock* userClock = nullptr);
    /*!
     * @brief DEPRECATED -- old initialize() API.  If you pass a Clock to this,
     *        sets TimeModeCustomClock using that clock.  Otherwise,
     *        uses TimeModeSteady.  Things requiring TimeModeExternallyDriven will not
     *        work with this unless the caller passes in a ClockPlaybackClient.
     *
     * @param userClock: the clock to use.  If null, uses TimeModeSteady.
     *
     * @return void
     */
    // TODO [Matt Ivester 2/11/2020]: should be deprecated but this breaks ottomatika build-missing builds.
    // [[deprecated("Use initialize(SystemTimeMode, Clock*) instead.  Contact the SI team if you have questions about
    // how this should be set up.")]]
    static void initialize(Clock* userClock);

    /*!
     * @brief returns whether SystemClock::initialize() has been called.
     *
     * @return true if SystemClock::initialize() has already been called.
     */
    static bool isInitialized(void);
    /*!
     * @brief returns the SystemTimeMode that was set in the call to initialize().
     *
     * @return the current SystemTimeMode.
     */
    static SystemTimeMode getTimeMode(void);

    /*!
     * @brief When in TimeModeExternallyDriven, reads out the current playback clock parameters.
     *        Otherwise does nothing.
     *
     * @param outParams : where the playback params will be copied to.
     *
     * @return true if we read out params, false otherwise.
     */
    static bool getPlaybackParams(PlaybackClockParams& outParams);

    /*!
     * @brief equivalent to calling getSynchronizedSystemTime().
     *
     *        in TimeModeSteady, returns getSynchronizedSteadyTime().
     *        in TimeModeExternallyDriven, returns the time from the externally driven clock.
     *        in TimeModeCustomClock, calls getSystemTime() on the user-specified Clock.
     *
     *        Consider replacing calls to this with the desired underlying API if relevant.
     *
     * @return the "system time" based on the time mode that was set.
     */
    static ottotime::otimePoint getSystemTime(void);
    /*!
     * @brief Given an earlier timestamp returned from getSystemTime(), returns
     *        how much time has passed (in terms of the system clock) since then.
     *
     *        Equivalent to (SystemClock::getSystemTime() - before).
     *
     * @param before : the earlier time to compare to.
     *
     * @return the amount of elapsed time.  This may be negative if the clock has moved backwards!
     */
    static ottotime::otimeDuration getSystemTimeSince(const ottotime::otimePoint& before);

    /*!
     * @brief A monotonic, steady clock that is not guaranteed to be synchronized to anything else.
     *        Returns the amount of time passed since the SystemClock constructor was called.
     *        This clock is always available, even before SystemClock::initialize() has been called.
     *        It may have higher resolution than the other clocks.
     *
     * @return the amount of time since the SystemClock system was started.
     */
    static ottotime::otimeDuration getLocalSteadyTime(void);
    /*!
     * @brief Given an earlier timestamp returned from getLocalSteadyTime(), returns
     *        how much time has passed for that clock since then.
     *
     *        Equivalent to (SystemClock::getLocalSteadyTime() - before).
     *
     * @param before : the earlier time to compare to.
     *
     * @return the amount of elapsed time.
     */
    static ottotime::otimeDuration getLocalSteadyTimeSince(const ottotime::otimeDuration& before);

    /*!
     * @brief A monotonic, steady clock that is synchronized between the various processes and nodes
     *        in the AV stack.  This clock is not available until SystemClock::initialize() has been called.
     *
     *        This clock is set equal to some master node's system time at initialization, but
     *        there is no guarantee that this will be a correct real-world wall clock time.  Because
     *        it is guaranteed to be steady it cannot be corrected after initialization.
     *
     * @return the current time of the synchronized steady clock.
     */
    static ottotime::otimePoint getSynchronizedSteadyTime(void);
    /*!
     * @brief Given an earlier timestamp returned from getSynchronizedSteadyTime(), returns
     *        how much time has passed for that clock since then.
     *
     *        Equivalent to (SystemClock::getSynchronizedSteadyTime() - before).
     *
     * @param before : the earlier time to compare to.
     *
     * @return the amount of elapsed time.
     */
    static ottotime::otimeDuration getSynchronizedSteadyTimeSince(const ottotime::otimePoint& before);

    /*!
     * @brief in TimeModeSteady, returns getSynchronizedSteadyTime().
     *        in TimeModeExternallyDriven, returns the time from the externally driven clock.
     *        in TimeModeCustomClock, calls getSystemTime() on the user-specified Clock.
     *
     *        This clock is synchronized between the various processes and nodes in the
     *        AV stack.  This clock is *not* guaranteed to be steady (or possibly not even monotonic)
     *        when executing in a simulation environment, as it may be externally controlled
     *        (and might pause, jump, move backwards, etc. arbitrarily).
     *        This clock is not available until SystemClock::initialize() has been called.
     *
     * @return the current time of the synchronized system clock.
     */
    static ottotime::otimePoint getSynchronizedSystemTime(void);
    /*!
     * @brief Given an earlier timestamp returned from getSynchronizedSystemTime(), returns
     *        how much time has passed for that clock since then.
     *
     *        Equivalent to (SystemClock::getSynchronizedSystemTime() - before).
     *
     * @param before : the earlier time to compare to.
     *
     * @return the amount of elapsed time.  This could be negative if the clock has moved backwards!
     */
    static ottotime::otimeDuration getSynchronizedSystemTimeSince(const ottotime::otimePoint& before);

    /*!
     * @brief The system clock is a singleton, prohibit copies
     */
    SystemClock(const SystemClock& other) = delete;
    SystemClock& operator=(const SystemClock& other) = delete;

    // Set the clock to stop advancing; that is, we'll get the current
    // time, remember it, and return it whenever we're asked for the
    // time. Used at shutdown, so that we stop reading from the
    // external clock (whose interface is being shut down).
    static void freeze();

private:
    static bool initialized_;
    static SystemTimeMode mode_;

    // Initialize the clock as specified, even if it's already been
    // initialized. Called from initialize() (unless we've been
    // initialized) and freeze() (which needs to switch to a stopped
    // clock).
    static bool reinitialize(const SystemTimeMode& mode, Clock* userClock);

    // Currently we're assuming external creation/ownership of this Clock,
    // and that they will outlive the SystemClock interface (or at least you won't
    // call any SystemClock time functions after the clock has been destroyed.)
    static ClockParamsClock* externalClock_;
    static Clock* customClock_;
    static std::optional<PlaybackClockParams> externalClockParams_;
    static std::shared_ptr<Clock> s_frozenClock;

    static bool initializeExternalClock_(Clock* externalClock);
    static bool initializeLocalClocks_();
    static bool initializeSteadyClocks_();

    static ottotime::otimePoint localBaseTime_;
    static ottotime::otimePoint synchronizedBaseTime_;
    static ottotime::otimePoint steadyBaseTime_;

    // Don't call this or create this class in regular code.
    // For unit testing purposes only.
    static void uninitializeSystemClock();
    friend class SystemClockScopedInitializer;
};
} // namespace ottotime

#endif // CLOCK_SYSTEMCLOCK_H
