/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file SystemClockUtils.h
 * @product common
 * @version
 * @released 01/17/2020
 * @updated 01/17/2020
 * @brief System clock utility functions
 *
 * @section License
 *
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#ifndef CLOCK_SYSTEMCLOCK_UTILS_H
#define CLOCK_SYSTEMCLOCK_UTILS_H

#include <Clock/ClockSource.h>
#include <Clock/SystemClock.h>

namespace SystemClockUtils {

/*!
 * @brief returns the SystemTimeMode is requested by the execution environment.
 *        Currently this looks at the SYSTEM_CLOCK_TYPE environment variable.
 *
 * @return the SystemTimeMode that is specified based on the execution environment.
 */
SystemClock::SystemTimeMode getEnvironmentTimeMode();

/*!
 * @brief Returns a Clock whose getCurrentTime() method calls SystemClock::getSystemTime().
 *
 * @return a Clock interface implementation.
 */
Clock& getSystemTimeClock();

/*
 * @brief Returns a Clock whose getCurrentTime() method calls Systemclock::getSynchronizedSteadyTime().
 *
 * @return a Clock interface implementation.
 */
Clock& getSynchronizedSteadyTimeClock();

/*!
 * @brief Returns a Clock whose getCurrentTime() method calls SystemClock::getSynchronizedSystemTime().
 *
 * @return a Clock interface implementation.
 */
Clock& getSynchronizedSystemTimeClock();

enum class TimeZone
{
    TimeZoneUTC,
    TimeZoneLocal
};
enum class TimeFormat
{
    // ISO 8601 time.  Guaranteed to sort lexicographically.
    // YYYY-MM-DDTHH:MM:SS[time zone code], ex. "2020-01-13T17:01:43Z"
    // For durations, the largest unit will be either hours or days:
    // PnDTnHnMnS or PTnHnMnS, ex. "P1DT4H32M1S" or "PT28H32M1S"
    TimeFormatISO8601,
    // YYYY-MM-DDTHH:MM:SS.SSS[time zone code], ex. "2020-01-13T18:02:43.987-05"
    // For durations:
    // PnDTnHnMnS or PTnHnMnS with fractional seconds, ex. "P1DT4H32M1.987S"
    TimeFormatISO8601Precise,

    // A time format that is intended for human consumption.
    // Exact formatting may vary, use the ISO format if you intend to machine parse timestamps.
    TimeFormatHuman,        // ex: "2020 Jan 13 16:37.15 UTC", or a duration: "1 Day 4h32m1s"
    TimeFormatHumanPrecise, // ex: "2020 Jan 13 16:37.15.567 EST", or a duration: "28h32m1.987s"
};
/*!
 * @brief Attempts to give you the "real world" time, typically for logging purposes.
 *        This clock may be adjusted arbitrarily, meaning it could jump (due to time zone changes
 *        or something changing the system clock) or move backwards (due to DST adjustments
 *        or something changing the system clock).  You can call this before initialize() has been called,
 *        but the returned time may change or jump when SystemClock::initialize() is called.
 *
 *        An ottotime::otimepoint is deliberately *not* supplied for this, because you should never compare
 *        these times to time points returned by the other functions.
 *
 * @param zone : which time zone to output in (UTC or local).
 * @param fmt: the format that the string should be in.
 *
 * @return a string representing the current wall clock time.
 */
std::string getWallClockTimeString(TimeZone zone = TimeZone::TimeZoneUTC,
                                   TimeFormat fmt = TimeFormat::TimeFormatISO8601);

/*!
 * @brief Prints a timepoint in a formatted way.
 *
 * @param zone : which time zone to output in (UTC or local).
 * @param fmt: the format that the string should be in.
 *
 * @return a string representing the passed-in timepoint.
 */
std::string getTimePointString(const ottotime::otimePoint& point,
                               TimeZone zone = TimeZone::TimeZoneUTC,
                               TimeFormat fmt = TimeFormat::TimeFormatISO8601);

/*!
 * @brief Prints a time duration in a formatted way.
 *
 * @param fmt: the format that the string should be in.
 *
 * @return a string representing the passed-in duration.
 */
std::string getTimeDurationString(const ottotime::otimeDuration& duration,
                                  TimeFormat fmt = TimeFormat::TimeFormatISO8601);

} // namespace SystemClockUtils

#endif // end #ifndef CLOCK_SYSTEMCLOCK_UTILS_H