/*!
 *****************************************************************************
 * @file      testSystemClock.cc
 * @author    Matthew Ivester
 * @date      01/17/2020
 * @attention Unit tests for SystemClock.
 *
 * @attention Copyright (c) 2020- Aptiv. All rights reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Matthew Ivester
 * @date 01/17/2020
 * @remark initial
 *
 *****************************************************************************
 */

#include <Clock/SystemClock.h>
#include <Clock/SystemClockUtils.h>
#include <Comms/InterfaceUtils.h>
#include <TestUtils/SettableClock.hpp>

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <stdlib.h>

#include <optional>

using namespace ::testing;

class MockClockParamsClock : public ClockParamsClock
{
public:
    MOCK_METHOD0(getCurrentTime, ottotime::otimePoint());

    MOCK_METHOD0(getParams, std::optional<PlaybackClockParams>());

    MOCK_METHOD0(hasParams, bool());
};

void SystemClock::uninitializeSystemClock(void)
{
    if (!SystemClock::initialized_) {
        return;
    }

    if (SystemClock::externalClock_ != nullptr) {
        SystemClock::externalClock_ = nullptr;
    }

    SystemClock::externalClockParams_ = PlaybackClockParams();
    SystemClock::localBaseTime_ = ottotime::otimePoint();
    SystemClock::synchronizedBaseTime_ = ottotime::otimePoint();
    SystemClock::steadyBaseTime_ = ottotime::otimePoint();

    SystemClock::mode_ = SystemClock::SystemTimeMode::TimeModeSteady;
    SystemClock::initialized_ = false;
}

namespace ottotime {

class SystemClockScopedInitializer final
{
public:
    SystemClockScopedInitializer() = delete;
    SystemClockScopedInitializer(SystemClock::SystemTimeMode mode, Clock* externalClock = nullptr)
    {
        initStatus_ = SystemClock::initialize(mode, externalClock);
    }

    static void uninitialize(void) { SystemClock::uninitializeSystemClock(); }

    ~SystemClockScopedInitializer() { uninitialize(); }

    const bool& getInitStatus(void) const { return initStatus_; }

private:
    bool initStatus_;
};

} // end namespace ottotime

#define SYSTEM_CLOCK_INIT(mode, clock)                                                                                 \
    const ottotime::SystemClockScopedInitializer _sys_clock_initer((mode), (clock));                                   \
    EXPECT_TRUE(_sys_clock_initer.getInitStatus())
#define SYSTEM_CLOCK_STEADY() SYSTEM_CLOCK_INIT(ottotime::SystemClock::SystemTimeMode::TimeModeSteady, nullptr)
#define SYSTEM_CLOCK_EXTERNAL(clock)                                                                                   \
    SYSTEM_CLOCK_INIT(ottotime::SystemClock::SystemTimeMode::TimeModeExternallyDriven, clock)
#define SYSTEM_CLOCK_SETTABLE(x)                                                                                       \
    TestUtils::SettableClock _settable_clock{x};                                                                       \
    SYSTEM_CLOCK_INIT(ottotime::SystemClock::SystemTimeMode::TimeModeCustomClock, &_settable_clock)

TEST(SystemClockBasics, testIsInitializedNegative)
{
    EXPECT_FALSE(SystemClock::isInitialized());
}

TEST(SystemClockBasics, testIsInitialized)
{
    SYSTEM_CLOCK_STEADY();
    EXPECT_TRUE(SystemClock::isInitialized());
}

TEST(SystemClockBasics, testIsInitializedCustom)
{
    SYSTEM_CLOCK_SETTABLE();
    EXPECT_TRUE(SystemClock::isInitialized());
}

TEST(SystemClockBasics, testIsInitializedCustomNull)
{
    TestUtils::SettableClock _settable_clock{};
    const ottotime::SystemClockScopedInitializer _sys_clock_initer(
        ottotime::SystemClock::SystemTimeMode::TimeModeCustomClock, nullptr);
    EXPECT_FALSE(_sys_clock_initer.getInitStatus());
}

TEST(SystemClockBasics, testIsInitializedCustomAbort)
{
    TestUtils::SettableClock _settable_clock{};
    // If invalid SystemTimeMode is set, Abort() will be called
    ASSERT_DEATH(
        {
            const ottotime::SystemClockScopedInitializer _sys_clock_initer(
                static_cast<ottotime::SystemClock::SystemTimeMode>(5), nullptr);
        },
        "");
}

TEST(SystemClockBasics, testIsInitializedReset)
{
    EXPECT_FALSE(SystemClock::isInitialized());

    SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeSteady);

    EXPECT_TRUE(SystemClock::isInitialized());

    ottotime::SystemClockScopedInitializer::uninitialize();

    EXPECT_FALSE(SystemClock::isInitialized());
}

TEST(SystemClockBasics, test_initializeSetsTimeModeToCustom)
{
    TestUtils::SettableClock _settable_clock;
    ottotime::SystemClock::initialize(&_settable_clock);
    EXPECT_EQ(SystemClock::getTimeMode(), SystemClock::SystemTimeMode::TimeModeCustomClock);
    ottotime::SystemClockScopedInitializer::uninitialize();
}

TEST(SystemClockBasics, test_initializeNull)
{
    ottotime::SystemClock::initialize(nullptr);
    EXPECT_EQ(SystemClock::getTimeMode(), SystemClock::SystemTimeMode::TimeModeSteady);
    ottotime::SystemClockScopedInitializer::uninitialize();
}

TEST(SystemClockBasics, testGetTimeModeSteady)
{
    SYSTEM_CLOCK_STEADY();
    EXPECT_EQ(SystemClock::getTimeMode(), SystemClock::SystemTimeMode::TimeModeSteady);
}

TEST(SystemClockBasics, testGetTimeModeCustom)
{
    SYSTEM_CLOCK_SETTABLE();
    EXPECT_EQ(SystemClock::getTimeMode(), SystemClock::SystemTimeMode::TimeModeCustomClock);
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock.
//                   This could behave inconsistently on, e.g. heavily loaded Jenkins instances.
TEST(SystemClockBasics, DISABLED_testInitSteady)
{
    SYSTEM_CLOCK_STEADY();

    const ottotime::otimePoint sysTime = ottotime::SystemClock::getSystemTime();
    const ottotime::otimePoint steadyTime = ottotime::SystemClock::getSynchronizedSteadyTime();
    const ottotime::otimeDuration diff = sysTime - steadyTime;
    const int64_t diffMs = std::abs(diff.totalMilliSeconds());
    EXPECT_LE(diffMs, 100);
}

/*
 * This test was previously disabled since the SCS connection would hang indefinitely if run
 * without a SimpleComms server. With DDS channel specified in the SYSTEM_CLOCK_EXTERNAL macro,
 * the channel will connect but not produce a valid time. Perhaps a @TODO would be to provide
 * a valid timesource.
 */
TEST(SystemClockBasics, testInitExternal)
{
    MockClockParamsClock mockClock{};
    otimePoint time(otimeSpec(12345, 67890));

    // The system time is fetched three times in the initialization function, plus the one for the test.
    EXPECT_CALL(mockClock, getCurrentTime()).Times(AtLeast(1)).WillRepeatedly(Return(time));

    SYSTEM_CLOCK_EXTERNAL(&mockClock);

    const ottotime::otimePoint sysTime = ottotime::SystemClock::getSystemTime();

    EXPECT_TRUE(sysTime.isTimeValid());
    EXPECT_EQ(time, sysTime);
}

TEST(SystemClockBasics, testGetTimeModeExternal)
{
    MockClockParamsClock mockClock{};
    SYSTEM_CLOCK_EXTERNAL(&mockClock);
    EXPECT_EQ(SystemClock::getTimeMode(), SystemClock::SystemTimeMode::TimeModeExternallyDriven);
}

TEST(SystemClockBasics, testGetParamsExternalValid)
{
    MockClockParamsClock mockClock{};

    PlaybackClockParams setParams{};
    PlaybackClockParams getParams{};
    otimePoint offset(otimeSpec(12345, 67890));
    otimePoint baseTime(otimeSpec(67890, 12345));
    otimePoint stopTime(otimeSpec(9876, 54321));

    setParams.offset = offset;
    setParams.baseTime = baseTime;
    setParams.stopTime = stopTime;
    setParams.scaleFactor = 5;

    std::optional<PlaybackClockParams> setParamsOpt{setParams};

    EXPECT_CALL(mockClock, getParams()).Times(AtLeast(1)).WillRepeatedly(Return(setParamsOpt));

    SYSTEM_CLOCK_EXTERNAL(&mockClock);

    // System should have params second time
    EXPECT_TRUE(SystemClock::getPlaybackParams(getParams));
    EXPECT_EQ(getParams, setParams);
}

TEST(SystemClockBasics, testGetParamsExternalInvalid)
{
    MockClockParamsClock mockClock{};

    PlaybackClockParams getParams{};
    std::optional<PlaybackClockParams> emptyParamsOpt{};

    EXPECT_CALL(mockClock, getParams()).Times(AtLeast(1)).WillRepeatedly(Return(emptyParamsOpt));

    SYSTEM_CLOCK_EXTERNAL(&mockClock);

    // System "hasn't received params yet"
    EXPECT_FALSE(SystemClock::getPlaybackParams(getParams));
    EXPECT_EQ(getParams, PlaybackClockParams{});
}

TEST(SystemClockBasics, testInitLocal)
{
    // only check SystemClock default constructor
    SystemClock sysClock{};
    EXPECT_FALSE(sysClock.isInitialized());
}

TEST(SystemClockSteady, test_getSynchronizedSystemTimeSince)
{
    SYSTEM_CLOCK_STEADY();

    const ottotime::otimePoint systemTime = SystemClock::getSynchronizedSystemTime();
    sleep(1);
    const ottotime::otimeDuration sinceFirst = SystemClock::getSynchronizedSystemTimeSince(systemTime);

    const int64_t diffMs = std::abs(sinceFirst.totalMilliSeconds());

    // check getSynchronizedSystemTimeSince function.
    EXPECT_GE(diffMs, 1000);
}

TEST(SystemClockBasics, testInitSettable)
{
    SYSTEM_CLOCK_SETTABLE();

    _settable_clock.setCurrentTime(ottotime::otimePoint::buildCurrentWallClockTime());
    const ottotime::otimePoint sysTime = ottotime::SystemClock::getSystemTime();

    EXPECT_EQ(_settable_clock.getCurrentTime(), sysTime);
}

TEST(SystemClockBasics, testGetParamsSettableNegative)
{
    PlaybackClockParams getParams{};

    SYSTEM_CLOCK_SETTABLE();

    EXPECT_FALSE(SystemClock::getPlaybackParams(getParams));
}

TEST(SystemClockBasics, testNegativeMultiInit)
{
    SYSTEM_CLOCK_STEADY();

    EXPECT_FALSE(SystemClock::initialize(ottotime::SystemClock::SystemTimeMode::TimeModeSteady));
}

TEST(SystemClockBasics, test_getLocalSteadyTimeSince)
{
    SYSTEM_CLOCK_STEADY();

    const ottotime::otimeDuration localSteadyTime = ottotime::SystemClock::getLocalSteadyTime();
    sleep(1);
    const ottotime::otimeDuration sinceFirst = ottotime::SystemClock::getLocalSteadyTimeSince(localSteadyTime);

    const int64_t diffMs = std::abs(sinceFirst.totalMilliSeconds());

    // check getLocalSteadyTimeSince function.
    EXPECT_GE(diffMs, 1000);
}

TEST(SystemClockBasics, test_getSynchronizedSteadyTimeSince)
{
    SYSTEM_CLOCK_SETTABLE(ottotime::otimePoint::buildCurrentWallClockTime());

    const ottotime::otimePoint sysStartSteadyTime = ottotime::SystemClock::getSynchronizedSteadyTime();
    sleep(1);
    const ottotime::otimeDuration sinceFirst =
        ottotime::SystemClock::getSynchronizedSteadyTimeSince(sysStartSteadyTime);

    const int64_t diffMs = std::abs(sinceFirst.totalMilliSeconds());

    // check getSynchronizedSteadyTimeSince function.
    EXPECT_GE(diffMs, 1000);
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock and sleep() behavior.
//                   Actually calling sleep() behaves inconsistently on, e.g. heavily loaded Jenkins instances.
TEST(SystemClockSteady, DISABLED_testSeemsSteady)
{
    SYSTEM_CLOCK_STEADY();

    const ottotime::otimePoint sysTime = ottotime::SystemClock::getSystemTime();
    sleep(1);
    const ottotime::otimeDuration sinceFirst = ottotime::SystemClock::getSystemTimeSince(sysTime);
    const int64_t diffMs = std::abs(sinceFirst.totalMilliSeconds());

    // If we really slept for 1s, make sure the clock reports roughly that!
    EXPECT_LE(diffMs, 1100);
    EXPECT_GE(diffMs, 900);
}

TEST(SystemClockSteady, testNegativeInitWithPassedClock)
{
    TestUtils::SettableClock _settable_clock;
    const bool status =
        ottotime::SystemClock::initialize(ottotime::SystemClock::SystemTimeMode::TimeModeSteady, &_settable_clock);
    ottotime::SystemClockScopedInitializer::uninitialize();
    EXPECT_FALSE(status); // should have failed!
}

TEST(SystemClockExternal, testNegativeInitExternalNoClock)
{
    const bool status =
        ottotime::SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeExternallyDriven, nullptr);
    ottotime::SystemClockScopedInitializer::uninitialize();
    EXPECT_FALSE(status); // should have failed!
}

TEST(SystemClockExternal, testNegativeInitExternalWrongClock)
{
    TestUtils::SettableClock _settable_clock;
    const bool status = ottotime::SystemClock::initialize(
        ottotime::SystemClock::SystemTimeMode::TimeModeExternallyDriven, &_settable_clock);
    ottotime::SystemClockScopedInitializer::uninitialize();
    EXPECT_FALSE(status); // should have failed!
}

TEST(SystemClockExternal, test_getSystemTimeSince)
{

    MockClockParamsClock mockClock{};
    otimePoint time(otimeSpec(12345, 67890));

    // The system time is fetched three times in the initialization function, plus the one for the test.
    EXPECT_CALL(mockClock, getCurrentTime()).Times(AtLeast(1)).WillRepeatedly(Return(time));

    SYSTEM_CLOCK_EXTERNAL(&mockClock);

    const ottotime::otimePoint sysTime = ottotime::SystemClock::getSystemTime();

    EXPECT_TRUE(sysTime.isTimeValid());
    EXPECT_EQ(time, sysTime);

    // After 1 sec
    time = otimeSpec(12346, 67890);
    EXPECT_CALL(mockClock, getCurrentTime()).Times(AtLeast(1)).WillRepeatedly(Return(time));

    const ottotime::otimeDuration sinceFirst = ottotime::SystemClock::getSystemTimeSince(sysTime);
    const int64_t diffMs = std::abs(sinceFirst.totalMilliSeconds());

    // check getSystemTimeSince function.
    EXPECT_EQ(diffMs, 1000);
}

TEST(SystemClockCustom, testNegativeInitCustomNoClock)
{
    const bool status =
        ottotime::SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeExternallyDriven, nullptr);
    ottotime::SystemClockScopedInitializer::uninitialize();
    EXPECT_FALSE(status); // should have failed!
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock and sleep() behavior.
//                   Actually calling sleep() behaves inconsistently on, e.g. heavily loaded Jenkins instances.
TEST(SystemClockCustom, DISABLED_testDoesntAffectSteadyClock)
{
    SYSTEM_CLOCK_SETTABLE(ottotime::otimePoint::buildCurrentWallClockTime());

    const ottotime::otimePoint sysStartTime = ottotime::SystemClock::getSystemTime();
    const ottotime::otimePoint sysStartSteadyTime = ottotime::SystemClock::getSynchronizedSteadyTime();
    sleep(1);
    const ottotime::otimePoint sysAfterTime = ottotime::SystemClock::getSystemTime();
    const ottotime::otimePoint sysAfterSteadyTime = ottotime::SystemClock::getSynchronizedSteadyTime();

    // Our custom clock hasn't moved, so sleeping shouldn't have affected the "system time"
    EXPECT_EQ(sysStartTime, sysAfterTime);

    // But the steady time still should have moved forward by 1s.
    const ottotime::otimeDuration steadySinceStart = sysAfterSteadyTime - sysStartSteadyTime;
    const int64_t diffMs = std::abs(steadySinceStart.totalMilliSeconds());

    // If we really slept for 1s, make sure the clock reports roughly that!
    EXPECT_LE(diffMs, 1100);
    EXPECT_GE(diffMs, 900);
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock.
//                   This could behave inconsistently on, e.g. heavily loaded Jenkins instances.
TEST(SystemClockUtils, DISABLED_testSystemTimeClock)
{
    SYSTEM_CLOCK_STEADY();

    Clock& sysClockInterface = SystemClockUtils::getSystemTimeClock();

    const ottotime::otimePoint systemTime = SystemClock::getSystemTime();
    const ottotime::otimePoint interfaceTime = sysClockInterface.getCurrentTime();

    const ottotime::otimeDuration timeDiff = systemTime - interfaceTime;
    const int64_t diffMs = std::abs(timeDiff.totalMilliSeconds());

    // Allow a small amount of difference in case we got swapped out, etc.
    EXPECT_LE(diffMs, 50);
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock.
//                   This could behave inconsistently on, e.g. heavily loaded Jenkins instances.
// TEST(SystemClockUtils, testSteadyTimeClock)
TEST(SystemClockUtils, DISABLED_testSteadyTimeClock)
{
    SYSTEM_CLOCK_STEADY();

    Clock& sysClockInterface = SystemClockUtils::getSynchronizedSteadyTimeClock();

    const ottotime::otimePoint systemTime = SystemClock::getSynchronizedSteadyTime();
    const ottotime::otimePoint interfaceTime = sysClockInterface.getCurrentTime();

    const ottotime::otimeDuration timeDiff = systemTime - interfaceTime;
    const int64_t diffMs = std::abs(timeDiff.totalMilliSeconds());

    // Allow a small amount of difference in case we got swapped out, etc.
    EXPECT_LE(diffMs, 50);
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock.
//                   This could behave inconsistently on, e.g. heavily loaded Jenkins instances.
// TEST(SystemClockUtils, testSystemSyncTimeClock)
TEST(SystemClockUtils, DISABLED_testSystemSyncTimeClock)
{
    SYSTEM_CLOCK_STEADY();

    Clock& sysClockInterface = SystemClockUtils::getSynchronizedSystemTimeClock();

    const ottotime::otimePoint systemTime = SystemClock::getSynchronizedSystemTime();
    const ottotime::otimePoint interfaceTime = sysClockInterface.getCurrentTime();

    const ottotime::otimeDuration timeDiff = systemTime - interfaceTime;
    const int64_t diffMs = std::abs(timeDiff.totalMilliSeconds());

    // Allow a small amount of difference in case we got swapped out, etc.
    EXPECT_LE(diffMs, 50);
}

TEST(SystemClockUtils, testCustomTimeClock)
{
    SYSTEM_CLOCK_SETTABLE();

    _settable_clock.setCurrentTime(ottotime::otimePoint::buildCurrentWallClockTime());
    Clock& sysClockInterface = SystemClockUtils::getSystemTimeClock();

    // These should return exactly the same thing!
    const ottotime::otimePoint systemTime = SystemClock::getSystemTime();
    const ottotime::otimePoint interfaceTime = sysClockInterface.getCurrentTime();

    EXPECT_EQ(systemTime, interfaceTime);
}

/**
 * @brief Group of tests for environment variable.
 */
class SystemClockUtilsTest : public ::testing::Test
{
protected:
    const char* clockEnvStr = "";
    std::string clockEnvStrCopy{};
    void SetUp() override final
    {
        // saving current environment variable
        clockEnvStr = getenv("SYSTEM_CLOCK_TYPE");
        if (clockEnvStr == nullptr) {
            // When SYSTEM_CLOCK_TYPE is not set
            EXPECT_EQ(SystemClockUtils::getEnvironmentTimeMode(), SystemClock::SystemTimeMode::TimeModeSteady);
        } else {
            clockEnvStrCopy = clockEnvStr;
        }
    }

    void TearDown() override final
    {
        // restoring original value
        if (clockEnvStr == nullptr) {
            unsetenv("SYSTEM_CLOCK_TYPE");
        } else {
            setenv("SYSTEM_CLOCK_TYPE", clockEnvStrCopy.c_str(), 1);
        }
    }
};

TEST_F(SystemClockUtilsTest, test_getEnvironmentTimeMode)
{

    // set SYSTEM_CLOCK_TYPE as local
    setenv("SYSTEM_CLOCK_TYPE", "local", 1);
    EXPECT_EQ(SystemClockUtils::getEnvironmentTimeMode(), SystemClock::SystemTimeMode::TimeModeSteady);

    // set SYSTEM_CLOCK_TYPE as steady
    setenv("SYSTEM_CLOCK_TYPE", "steady", 1);
    EXPECT_EQ(SystemClockUtils::getEnvironmentTimeMode(), SystemClock::SystemTimeMode::TimeModeSteady);

    // set SYSTEM_CLOCK_TYPE as external
    setenv("SYSTEM_CLOCK_TYPE", "external", 1);
    EXPECT_EQ(SystemClockUtils::getEnvironmentTimeMode(), SystemClock::SystemTimeMode::TimeModeExternallyDriven);

    // set SYSTEM_CLOCK_TYPE as invalid value
    setenv("SYSTEM_CLOCK_TYPE", "unknown", 1);
    EXPECT_EQ(SystemClockUtils::getEnvironmentTimeMode(), SystemClock::SystemTimeMode::TimeModeSteady);
}

TEST(SystemClockUtils, test_getTimePointString)
{
    MockClockParamsClock mockClock{};
    // Epoch 1637580153 = GMT:2021-11-22 11:22:33
    otimePoint time(otimeSpec(1637580153, 0));
    // Previously these had the milliseconds as .001, but that's
    // clearly wrong, when the nanoseconds in time, above, are set to
    // 0.
    constexpr static auto isoFormatStr = "2021-11-22T11:22:33.000+0000";
    constexpr static auto humanFormatStr = "Nov 22 2021 11:22:33.000 UTC";

    // The system time is fetched three times in the initialization function, plus the one for the test.
    EXPECT_CALL(mockClock, getCurrentTime()).Times(AtLeast(1)).WillRepeatedly(Return(time));

    SYSTEM_CLOCK_EXTERNAL(&mockClock);

    const ottotime::otimePoint sysTime = ottotime::SystemClock::getSystemTime();

    EXPECT_TRUE(sysTime.isTimeValid());
    EXPECT_EQ(time, sysTime);

    SystemClockUtils::TimeZone timeZone{};
    SystemClockUtils::TimeFormat timeFmt{};
    std::string testPointString{};

    // check TimeFormatISO8601Precise
    timeZone = SystemClockUtils::TimeZone::TimeZoneUTC;
    timeFmt = SystemClockUtils::TimeFormat::TimeFormatISO8601Precise;
    testPointString = SystemClockUtils::getTimePointString(sysTime, timeZone, timeFmt);

    EXPECT_EQ(testPointString, isoFormatStr);

    // check TimeFormatHumanPrecise
    timeZone = SystemClockUtils::TimeZone::TimeZoneUTC;
    timeFmt = SystemClockUtils::TimeFormat::TimeFormatHumanPrecise;
    testPointString = SystemClockUtils::getTimePointString(sysTime, timeZone, timeFmt);

    EXPECT_EQ(testPointString, humanFormatStr);
}

int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
