/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief A file to implement the Channel class
 */

#include <Comms/Channel.h>
#include <TimeStamp/otimeConversions.h>

#include <optional>

namespace task {

task::stats::MessageCallbackRegistry Channel::callbacks_{};

namespace {
/// @brief Convenience function for computing an optional latency,
/// given a receive time and an optional send time.
///
/// @param sendTime The time the message was sent, if known.
/// @param recvTime The time the message was received.
///
/// @return If sendTime is present, returns the difference between
/// sendTime and recvTime. Otherwise, std::nullopt.
std::optional<av::Duration> computeLatency(std::optional<av::TimePoint> sendTime, av::TimePoint recvTime)
{
    return sendTime.has_value() ? std::optional<av::Duration>(recvTime - sendTime.value()) : std::nullopt;
}
} // namespace

using task::stats::MessageReport;

void Channel::reportConnected()
{
    reportConnected(getName());
}

void Channel::reportConnected(const std::string& channelName)
{
    const MessageReport report{channelName,
                               MessageReport::EventType::Connected,
                               0,
                               av::timePointFromOtimePoint(SystemClock::getSystemTime()),
                               std::nullopt,
                               std::nullopt};
    reportEvent(report);
}

void Channel::reportDisconnected()
{
    reportDisconnected(getName());
}

void Channel::reportDisconnected(const std::string& channelName)
{
    const MessageReport report{channelName,
                               MessageReport::EventType::Disconnected,
                               0,
                               av::timePointFromOtimePoint(SystemClock::getSystemTime()),
                               std::nullopt,
                               std::nullopt};
    reportEvent(report);
}

void Channel::reportMessage(size_t messageSize, std::optional<av::TimePoint> sendTime)
{
    reportMessage(getName(), messageSize, sendTime);
}

void Channel::reportMessage(const std::string& channelName, size_t messageSize, std::optional<av::TimePoint> sendTime)
{
    const av::TimePoint recvTime{av::timePointFromOtimePoint(SystemClock::getSystemTime())};
    const MessageReport report{channelName,
                               sendTime.has_value() ? MessageReport::EventType::Received
                                                    : MessageReport::EventType::Sent,
                               messageSize,
                               recvTime,
                               sendTime,
                               computeLatency(sendTime, recvTime)};

    reportEvent(report);
}

void Channel::reportEvent(const MessageReport& report)
{
    callbacks_.invokeCallbacks(report);
}
} // namespace task
