/*!
 ***********************************************************
 * @file ChannelFactory.cc
 * @brief A file implementing the ChannelFactory class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * ChannelFactory.cc
 * Tugrul Galatali
 * 10/31/2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */

#include <Comms/ChannelFactory.h>

#if !defined(__aarch64__)
#include <Comms/DDSChannel.h>
#endif

#include <Comms/SCChannel.h>
#include <Comms/SQLLogChannel.h>

#include <SimpleComms/UnixClient.h>

#include "lib/avlog/avlog.hpp"

using namespace SimpleComms;

namespace task {

bool ChannelFactory::connectAs(const std::string& moduleName)
{
    get().moduleName_ = moduleName;

    return true;
}

void ChannelFactory::useClient(const std::shared_ptr<SimpleComms::Client>& client)
{
    get().sc_ = client;
}

std::unique_ptr<Channel> ChannelFactory::createChannelManaged(ConfigSection& channelConfig,
                                                              const Access mode,
                                                              std::optional<std::string> nameOpt)
{
    ChannelFactory& cf(get());
    std::string name;
    if (nameOpt.has_value()) {
        name = nameOpt.value();
    } else {
        const ConfigSection::GetStatus nameParamStatus = channelConfig.get(std::string("name"), name);

        if (nameParamStatus != ConfigSection::GS_OK) {
            AVLOG_DEBUG("name field not in config section");
            name = "unused";
        }
    }

    std::unique_ptr<Channel> channel{nullptr};

    std::string type;
    const ConfigSection::GetStatus typeParamStatus = channelConfig.get(std::string("type"), type);

    if (typeParamStatus == ConfigSection::GS_OK) {
        AVLOG_INFO("Creating channel {} of type {}", channelConfig.getName(), type);

        const auto it = cf.builders_.find(type);
        if (it != cf.builders_.end()) {
            channel = std::move(it->second->build(channelConfig, mode, name));
        } else {
            AVLOG_INFO("Unknown channel type {}", type);
        }
    } else {
        AVLOG_WARN("type invalid or undefined!");
    }

    return channel;
}

ChannelFactory& ChannelFactory::get()
{
    static ChannelFactory factory{};

    return factory;
}

std::shared_ptr<SimpleComms::Client> ChannelFactory::connectToSCS(std::optional<int> maxRetries)
{
    if (sc_ == nullptr) {
        if (moduleName_.empty()) {
            AVLOG_WARN("ChannelFactory module name not set.");
        }

        const auto serverPath = UnixClient::getDefaultScsPath();
        const std::string scsName = UnixClient::makeUniqueClientName(moduleName_);
        sc_ = std::make_shared<SimpleComms::UnixClient>(serverPath, scsName);

        int numRetries = maxRetries.has_value() ? maxRetries.value() : 0;
        while (((!maxRetries.has_value()) || numRetries >= 0) && (sc_->connect() != SimpleComms::Status::Ok)) {
            AVLOG_WARN("Unable to connect to scs as \"{}\"", scsName);
            static_cast<void>(sleep(1U));
            --numRetries;
        }

        if (maxRetries.has_value() && (numRetries < 0)) {
            sc_ = nullptr;
        }
    }

    return sc_;
}

ChannelFactory::ChannelFactory()
{
    registerBuilder("SimpleComms", std::make_unique<SCChannel::Builder>());

#if !defined(__aarch64__)
    registerBuilder("DDS", std::make_unique<DDSChannel::Builder>());
#endif

    registerBuilder("LOG", std::make_unique<SQLLogChannel::Builder>());
}

void ChannelFactory::cleanup()
{
    get().sc_ = nullptr;
}

} // namespace task
