#include "DDSConfiguratorHelper.hpp"
#include "lib/avlog/avlog.hpp"

#include "DDSConfiguratorHelper.hpp"
#include "platform/middleware/libs/LatencyLogger/include/LatencyLogger/LatencyLogger.h"
#include <Comms/DDSChannel.h>
#include <TimeStamp/otimeConversions.h>

namespace task {
DDSChannel::DDSChannel(std::string name,
                       Access mode,
                       std::uint32_t _recvQueueDepth,
                       DomainId _domainId,
                       std::shared_ptr<DDSComms::DDSConfigurator> config)
    : Channel(mode)
    , config_(std::move(config))
    , domainId(_domainId)
    , channelName(std::move(name))
    , blockWhenNoData(true)
    , recvQueueDepth(_recvQueueDepth)
    , receivingQueue(_recvQueueDepth)
    , reader(dds::core::null)
    , writer(dds::core::null)
    , readerListener(this)
    , writerListener(this)
    , domainParticipant(dds::core::null)
    , topic(dds::core::null)
{}

DDSChannel::~DDSChannel()
{
    disconnect();
}

void DDSChannel::setupDDSDomainAndTopic()
{
    domainParticipant = dds::domain::find(domainId);
    if (domainParticipant == dds::core::null) {
        auto qos = config_->createParticipantQos();
        domainParticipant = DomainParticipant(domainId, qos);
    }
    topic = dds::topic::find<BytesTopic>(domainParticipant, channelName);
    if (topic == dds::core::null)
        topic = BytesTopic(domainParticipant, channelName);
}

void DDSChannel::createReader()
{
    setupDDSDomainAndTopic();
    if (reader == dds::core::null) {
        auto qos = config_->createReaderQos(std::nullopt, channelName);
        subscriber = Subscriber(domainParticipant);
        using dds::core::status::StatusMask;
        StatusMask readerMask = StatusMask::none();
        readerMask |= StatusMask::requested_deadline_missed();
        readerMask |= StatusMask::requested_incompatible_qos();
        readerMask |= StatusMask::sample_lost();
        readerMask |= StatusMask::sample_rejected();
        readerMask |= StatusMask::liveliness_changed();
        readerMask |= StatusMask::subscription_matched();
        readerMask |= StatusMask::data_available();
        AVLOG_INFO("Creating DDS reader {}; ", channelName.c_str());
        reader = BytesReader(subscriber, topic, qos, &readerListener, readerMask);
    }
}

void DDSChannel::createWriter()
{
    setupDDSDomainAndTopic();
    if (writer == dds::core::null) {
        auto qos = config_->createWriterQos(channelName);
        publisher = Publisher(domainParticipant);
        using dds::core::status::StatusMask;
        StatusMask writerMask = StatusMask::none();
        writerMask |= StatusMask::offered_deadline_missed();
        writerMask |= StatusMask::offered_incompatible_qos();
        writerMask |= StatusMask::liveliness_lost();
        writerMask |= StatusMask::publication_matched();
        AVLOG_INFO("Creating DDS writer {}; ", channelName.c_str());
        writer = BytesWriter(publisher, topic, qos, &writerListener, writerMask);
    }
}

ChannelStatus DDSChannel::connect()
{
    {
        std::lock_guard<std::mutex> lock(channelMutex);
        channelConnected = true;
    }
    chmod(mode_);
    reportConnected();
    return CS_Success;
}

ChannelStatus DDSChannel::disconnect()
{
    {
        std::lock_guard<std::mutex> lock(channelMutex);
        if (!channelConnected) {
            return CS_Disconnected;
        }
        channelConnected = false;

        auto mask = dds::core::status::StatusMask::none();
        if (reader != dds::core::null) {
            reader.listener(nullptr, mask);
            reader = dds::core::null;
        }
        if (writer != dds::core::null) {
            writer.listener(nullptr, mask);
            writer = dds::core::null;
        }
    }
    unblock();
    clear();
    reportDisconnected();
    return CS_Disconnected;
}

ChannelStatus DDSChannel::readWithPred(std::string& msg, av::latency::Pred& pred, bool blocking)
{
    ChannelStatus cs{CS_Success};
    ottotime::otimePoint sendTime{};
    ottotime::otimePoint recvTime{};

    {
        std::unique_lock<std::mutex> lock(channelMutex);
        if ((!channelConnected) || (reader == dds::core::null)) {
            AVLOG_WARN("DDSChannel is not connected while reading messages!");
            cs = CS_Error;
        } else {
            if (blocking && receivingQueue.empty()) {
                blockWhenNoData = true;
                waitDataConVar.wait(
                    lock, [this] { return !blockWhenNoData || !receivingQueue.empty() || !channelConnected; });
            }

            if (channelConnected && !receivingQueue.empty()) {
                auto sample = receivingQueue.pop();
                sendTime = sample->sendTime;
                recvTime = sample->recvTime;
                msg = DDSComms::getMsgData(sample->data);

                cs = CS_NewData;
            }
        }
    }

    // Now we've released the mutex, we can log the latency event and
    // invoke the callbacks.
    if (cs == CS_NewData) {
        // the actual time of reception of the message is cached and set here because the comms
        // layer does not control how long a user waits to call `readWithPred`. Therefore, for
        // meaningful comms-layer latency statistics, we use the timestamp of when it is available to
        // be read by a user.
        pred = av::LatencyLogger::getLogger().ReceiveEvent(channelName, sendTime, recvTime);
        reportMessage(msg.size(), av::timePointFromOtimePoint(sendTime));
    }

    return cs;
}

ChannelStatus DDSChannel::writeV(const std::string& msg, std::vector<av::latency::Pred>&& predecessors)
{
    if ((!channelConnected) || (writer == dds::core::null)) {
        return CS_Error;
    }

    const ottotime::otimePoint sendTime = SystemClock::getSystemTime();
    ottotime::otimeSpec sendTimeSpec = sendTime.to_otimeSpec();

    if (!sendTimeSpec.isValidAsTimestamp()) {
        AVLOG_ERROR("send time otimeSpec not valid. Clock probably not initialized");
    }

    const dds::core::Time sendTimeDDS{static_cast<int32_t>(sendTimeSpec.tv_sec),
                                      static_cast<uint32_t>(sendTimeSpec.tv_nsec)};

    auto outputMsg = DDSComms::createMsg(msg);
    {
        std::lock_guard<std::mutex> lock(channelMutex);
        if (!channelConnected) {
            return CS_Error;
        }
        DDSComms::setSequenceNumber(outputMsg, ++sequenceNumber);
        // When using Reliable mode, dds::core::TimeoutError may be thrown.
        // allow other errors to propagate
        try {
            writer.write(outputMsg, sendTimeDDS);
        } catch (const dds::core::TimeoutError& e) {
            AVLOG_WARN("cannot send msg: channel {}; seq {}; err {};", channelName, sequenceNumber, e.what());
            return CS_Error;
        }
    }

    AVLOG_TRACE("sent msg: {}; {}", channelName, sequenceNumber);
    // Now we've released the mutex, we can run the callbacks.
    reportMessage(msg.size());
    av::LatencyLogger::getLogger().SendEvent(channelName, sendTime, std::move(predecessors));

    return CS_Success;
}

ChannelStatus DDSChannel::chmod(const Access mode)
{
    std::lock_guard<std::mutex> lock(channelMutex);
    if (mode == Read || mode == ReadWrite) {
        createReader();
    }
    if (mode == Write || mode == ReadWrite || mode == Standby) {
        createWriter();
    }
    return CS_Success;
}

bool DDSChannel::getQueueProperties(unsigned long& currentDepth, unsigned long& maxDepth)
{
    currentDepth = receivingQueue.size();
    maxDepth = recvQueueDepth;
    return true;
}

void DDSChannel::clear()
{
    std::lock_guard<std::mutex> lock(channelMutex);
    receivingQueue.clear();
}

void DDSChannel::unblock()
{
    std::lock_guard<std::mutex> lock(channelMutex);
    blockWhenNoData = false;
    waitDataConVar.notify_all();
}

std::unique_ptr<Channel> DDSChannel::Builder::build(ConfigSection& channelConfig, Access mode, std::string name) const
{
    uint32_t queueSize = 0;
    DomainId domainId = 0;
    DDSComms::DDSOptionsType ddsOptions;

    std::string ddsWriterQoSProfileName;
    std::string ddsReaderQoSProfileName;

    channelConfig.get(std::string("queueSize"), queueSize, 1U);
    channelConfig.get(std::string("domainId"), domainId, 0);
    channelConfig.get(std::string("ddsConfigFile"), ddsOptions.configFileName, ddsOptions.configFileName);

    channelConfig.get(std::string("ddsWriterQoSProfileName"), ddsWriterQoSProfileName, "");
    channelConfig.get(std::string("ddsReaderQoSProfileName"), ddsReaderQoSProfileName, "");

    if (!ddsWriterQoSProfileName.empty()) {
        ddsOptions.writeChannelProfileMap.emplace(name, ddsWriterQoSProfileName);
    }
    if (!ddsReaderQoSProfileName.empty()) {
        ddsOptions.readChannelProfileMap.emplace(name, ddsReaderQoSProfileName);
    }

    return std::make_unique<DDSChannel>(
        name, mode, queueSize, domainId, std::make_shared<DDSComms::DDSConfiguratorType>(ddsOptions));
}

DDSChannel::LoanedQueue::LoanedQueue(uint32_t size)
    : range{}
    , maxSize(std::min(size, MAX_CAPACITY))
{
    buffer.resize(maxSize + 1);
    if (size > MAX_CAPACITY) {
        AVLOG_WARN("DDSChannel doesn't accept any message queue with size > 1000. By default, it set as "
                   "1000. If you have special need for this queue size, please change it in DDSChannel.h");
    }
}

bool DDSChannel::LoanedQueue::empty() const
{
    std::scoped_lock<std::mutex> lock(queueMutex);
    return range.begin == range.end;
}

std::size_t DDSChannel::LoanedQueue::size() const
{
    std::scoped_lock<std::mutex> lock(queueMutex);
    return range.begin <= range.end ? range.end - range.begin : maxSize - range.begin + range.end;
}

void DDSChannel::LoanedQueue::clear()
{
    std::scoped_lock<std::mutex> lock(queueMutex);
    range = Range{};
}

DDSChannel::LoanedQueue::ItemType DDSChannel::LoanedQueue::pop()
{
    std::scoped_lock<std::mutex> lock(queueMutex);

    ItemType result = buffer[range.begin];
    buffer[range.begin] = ItemType();
    range.begin = (range.begin + 1) % (maxSize + 1);

    return result;
}

void DDSChannel::LoanedQueue::push(ItemType item)
{
    std::scoped_lock<std::mutex> lock(queueMutex);

    uint32_t newBegin = range.begin;
    uint32_t newEnd = (range.end + 1) % (maxSize + 1);
    // if the queue is full
    if (newEnd == range.begin) {
        newBegin = (range.begin + 1) % (maxSize + 1);
        newEnd = range.end;
    }
    buffer[range.end] = std::move(item);

    range.begin = newBegin;
    range.end = newEnd;
}

DDSChannel::Publisher DDSChannel::getPublisher() const
{
    return publisher;
}

DDSChannel::Subscriber DDSChannel::getSubscriber() const
{
    return subscriber;
}

DDSChannel::BytesWriter DDSChannel::getWriter() const
{
    return writer;
}

DDSChannel::BytesReader DDSChannel::getReader() const
{
    return reader;
}

} // namespace task
