/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file InterfaceUtils.cc
 * @product common
 * @version
 * @released
 * @updated 02/05/2019
 * @brief Interface utility functions
 *
 * These utility functions are for using ottoware communications in arbitrary C++ that does not
 * inherit from Task
 *
 * @section License
 *
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#include <Clock/SystemClockUtils.h>
#include <Comms/CommsPlaybackClock.h>
#include <Comms/InterfaceUtils.h>
#include <array>
#include <config/ConfigParser.h>
#include <interfaces/Generic/Input/Channel/Impl.h>

#include "lib/avlog/avlog.hpp"

using namespace task;

namespace comms {

std::shared_ptr<GenericInput<PlaybackClockParams>> getInterfaceForExternalClock(ChannelType channelType);

Clock* getExternallyDrivenClock(const ChannelType channelType)
{
    static std::map<std::shared_ptr<GenericInput<PlaybackClockParams>>, std::shared_ptr<Clock>> externalClocks{};

    Clock* extClock{nullptr};
    static auto intf = getInterfaceForExternalClock(channelType);

    if (externalClocks.find(intf) == externalClocks.end()) {
        if (intf != nullptr) {
            std::shared_ptr<Clock> commsClock =
                std::make_shared<CommsPlaybackClock>(intf.get(), &SystemClockUtils::getSynchronizedSteadyTimeClock());

            externalClocks[intf] = commsClock;
            extClock = commsClock.get();
        } else {
            AVLOG_CRITICAL("comms::getExternallyDrivenClock() could not create interface!");
        }
    } else {
        extClock = externalClocks.at(intf).get();
    }

    return extClock;
}

std::shared_ptr<GenericInput<PlaybackClockParams>> getInterfaceForExternalClock(const ChannelType channelType)
{
    const auto name{channelTypeName(channelType)};
    if (!name.has_value()) {
        AVLOG_CRITICAL("comms::getInterfaceForExternalClock channel type is invalid!");
        return nullptr;
    }
    auto c = getSystemPlaybackClockConfigSection(*name);
    if (!c.has_value()) {
        return nullptr;
    }

    return comms::getInterface<GenericInputChannel<PlaybackClockParams>>((*c).getName(), *c, task::Access::Read);
}

std::optional<ConfigSection> getSystemPlaybackClockConfigSection(const std::string& channelName)
{
    // Tasks used to have this off in their config somewhere, but we don't
    // really want the task to tell the system clock how to behave.
    // (Or if they really want to do that they can arrange to set the custom clock mode.)
    // For the ExternallyDriven clock mode, we always want this interface.
    const std::string clockConfigString = "SystemPlaybackClock" + channelName +
                                          " =\n"
                                          "{\n"
                                          "    interfaceName = \"PlaybackClockParamsInput" +
                                          channelName +
                                          "\";\n"
                                          "    implementation = \"Channel\";\n"
                                          "    channel =\n"
                                          "    {\n"
                                          "        ipcMsgQueueLength = 1;\n"
                                          "        name = \"PlaybackClockParams\";\n"
                                          "        queueSize = 1;\n"
                                          "        priority = 1;\n"
                                          "        type = \"" +
                                          channelName +
                                          "\";\n"
                                          "    };\n"
                                          "    interface = \"PlaybackClockParamsInput\";\n"
                                          "}\n";
    ConfigParser parser;
    ConfigSection config;
    if (!parser.parseString(clockConfigString) || !parser.getSection("SystemPlaybackClock" + channelName, config)) {
        AVLOG_CRITICAL("comms::getExternallyDrivenClock() SystemPlaybackClock configuration invalid!");
        return std::nullopt;
    }

    return config;
}

const static std::array<std::string, 3> channelTypes{"SimpleComms", "DDS", "LOG"};

std::optional<std::string> channelTypeName(ChannelType channelType)
{
    const auto index = static_cast<size_t>(channelType);
    if (index < channelTypes.size()) {
        return channelTypes.at(index);
    }

    return std::nullopt;
}

std::optional<ChannelType> channelTypeEnum(const std::string& channelTypeName)
{
    const auto iter = std::find(channelTypes.cbegin(), channelTypes.cend(), channelTypeName);
    if (iter != channelTypes.cend()) {
        return static_cast<ChannelType>(iter - channelTypes.begin());
    }

    return std::nullopt;
}

} // namespace comms
