#include "include/Comms/LatencyStats.h"
#include "lib/avlog/avlog.hpp"

using nanoseconds_long_double = std::chrono::duration<long double, std::nano>;
using milliseconds_long_double = std::chrono::duration<long double, std::milli>;
using microseconds_long_double = std::chrono::duration<long double, std::micro>;
using json = nlohmann::json;
LatencyStats::LatencyStats(const std::string& name,
                           int32_t latencyWarning_us,
                           const std::string& measurementName,
                           const std::string& type,
                           int32_t recvSize)
    : latencyInfo_(name, latencyWarning_us, measurementName, type, recvSize)
    , buckets_{{Range{0, 999}, 0},
               {Range{1000, 4999}, 0},
               {Range{5000, 9999}, 0},
               {Range{10000, 49999}, 0},
               {Range{50000, 99999}, 0},
               {Range{100000, 499999}, 0},
               {Range{500000, 999999}, 0},
               {Range{1000000, microseconds_long_double::max().count()}, 0}}
{}

LatStatsError LatencyStats::update(const microseconds_long_double val)
{
    if (val >= microseconds_long_double(0.0)) {
        // Handle startup latency values longer than a threshold (a second)
        // In RTI DDS we see that the first few messages are very slow to arrive.
        // This is believed to be due to the discovery process sending many messages
        // and slowing down the network performance at startup.
        //
        if ((!latencyInfo_.initialSlowdownInfo.foundUnderThreshold) &&
            (latencyInfo_.initialSlowdownInfo.numInitialOverThreshold < LATENCY_MAX_INITIAL_SLOWDOWNS)) {
            if (val < latencyInfo_.initialSlowdownInfo.thresholdInitialSlowdown) {
                latencyInfo_.initialSlowdownInfo.foundUnderThreshold = true;
                // continue on and count this entry
            } else {
                latencyInfo_.initialSlowdownInfo.numInitialOverThreshold += 1;
                latencyInfo_.initialSlowdownInfo.timeInitialLongDelayTotal += val;
                //  this is part of a long initial value, don't let it add into normal statistics
                return LatStatsError::LAT_STATS_OK;
            }
        }

        latencyInfo_.count++;
        microseconds_long_double oldMean = latencyInfo_.mean;
        latencyInfo_.mean += ((val - latencyInfo_.mean) / latencyInfo_.count);
        latencyInfo_.intermediate +=
            microseconds_long_double((val - latencyInfo_.mean).count() * (val - oldMean).count());

        if (val < latencyInfo_.minLatency) {
            latencyInfo_.minLatency = val;
        }

        if (val > latencyInfo_.maxLatency) {
            latencyInfo_.maxLatency = val;
        }

        for (auto& bucket : buckets_) {
            auto range = bucket.first;
            if ((val.count() >= range.start) && (val.count() <= range.end)) {
                // Currently the counter is a uint64_t, so even if a channel sends 10,000
                // messages per second we can run for nearly 60 million years before
                // we stop counting
                if (bucket.second < std::numeric_limits<decltype(bucket.second)>::max()) {
                    bucket.second += 1;
                }

                break;
            }
        }

        if ((latencyInfo_.latencyWarning.count() != -1) && (val >= latencyInfo_.latencyWarning)) {
            ++latencyInfo_.numLatencyWarnings;
            AVLOG_WARN("LatencyStats warning for channel {}: Threshold {}us Actual {}us",
                       latencyInfo_.channelName,
                       latencyInfo_.latencyWarning.count(),
                       val.count());
        }

        return LatStatsError::LAT_STATS_OK;
    }
    return LatStatsError::LAT_STATS_NEG_VAL;
}
