#include <Comms/OttoComms.h>
#include <Comms/OttoConfig.h>

#include <Clock/SystemClock.h>
#include <Clock/SystemClockUtils.h>
#include <Comms/ChannelFactory.h>
#include <Comms/InterfaceUtils.h>
#include <SimpleComms/UnixClient.h>
#include <TimeStamp/otime.h>
#include <chrono>
#include <config/ConfigSection.h>
#include <interfaces/Generic/Input/Channel/Impl.h>
#include <interfaces/HealthDisplay/HealthDisplay.h>
#include <thread>

using namespace std::chrono_literals;

OttoComms::OttoComms(const std::string& clientName, const std::string& configFile)
{
    avlog::setGlobalLogLevel(avlog::LogLevel::err);

    checkOttoEnvVars();

    std::cout << "OttoComms: initiating OttoConfig: " << clientName << ": " << configFile << std::endl;
    m_ottoConfig = std::make_unique<OttoConfig>(configFile);

    std::cout << "OttoComms: waiting for scs" << std::endl;
    // Connect to the IPC system (SimpleComms in this case)
    if (!task::ChannelFactory::connectAs(clientName)) {
        throw std::runtime_error(clientName + " failed to connect to IPC");
    }

    // Initialize the system clock. This must come _after_ interfaces are initialized because the
    // clock may be implemented by a configured interface
    if (useExternallyDrivenClock()) {
        auto cs = m_ottoConfig->getSection("Parameters");
        std::string channeltype{};
        cs.get("CommSystem", channeltype);

        std::optional<comms::ChannelType> ct{comms::channelTypeEnum(channeltype)};

        if (!ct.has_value()) {
            ct = comms::ChannelType::SimpleComms;
            std::cout << "OttoComms: Warning - CommSystem not defined or invalid! Using SimpleComms" << std::endl;
        }

        SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeExternallyDriven,
                                comms::getExternallyDrivenClock(*ct));
    } else {
        SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeSteady);
    }
}

OttoComms::~OttoComms() {}

std::string OttoComms::ottoConfigDir()
{
    char* const buf = getenv("OTTO_CONFIG_DIR");
    if (buf == nullptr) {
        throw std::runtime_error("OttoComms: OTTO_CONFIG_DIR not set");
    }
    return std::string{buf};
}

std::string OttoComms::activeConfigDir()
{
    char* const buf = getenv("ACTIVE_CONFIG_DIR");
    if (buf == nullptr) {
        throw std::runtime_error("OttoComms: ACTIVE_CONFIG_DIR not set");
    }
    return std::string{buf};
}

bool OttoComms::useExternallyDrivenClock()
{
    return SystemClockUtils::getEnvironmentTimeMode() == SystemClock::SystemTimeMode::TimeModeExternallyDriven;
}

ConfigSection OttoComms::getSection(const std::string& name) const
{
    return m_ottoConfig->getSection(name);
}

OttoConfig* OttoComms::getOttoConfig() const
{
    return ((m_ottoConfig != nullptr) ? m_ottoConfig.get() : nullptr);
}

void OttoComms::checkOttoEnvVars()
{
    // @todo - maybe? refactor `getenv` to be done in the constructor... but then checked here
    const char* const ottoDir = std::getenv("OTTO_DIR");
    const char* const ottoConfDir = std::getenv("OTTO_CONFIG_DIR");
    if ((ottoDir == nullptr) || (ottoConfDir == nullptr)) {
        throw std::runtime_error("OttoComms: OTTO_DIR and/or OTTO_CONFIG_DIR not set.");
    }
}
