/*!
 ***********************************************************
 * @file SCChannel.cc
 * @brief A file implementing the SCChannel class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/06/2016
 * @remark
 * Sending operation is a separate thread now.
 * @n
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/05/2016
 * @remark
 * get() supports the blocking call.
 * @n
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/05/2016
 * @remark
 * Receiving operation is a separate thread now.
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * SCChannel.cc
 * Tugrul Galatali
 * 10/30/2006
 *
 * Junsung Kim
 * 12/16/2012
 * A function chmod() is added to support fault-tolerance.
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */
#include <Comms/ChannelFactory.h>
#include <Comms/SCChannel.h>

#include "lib/avlog/avlog.hpp"
#include <Clock/SystemClock.h>
#include <SimpleComms/ControlMessages.h>
#include <Threading/PThreadLocker.h>
#include <TimeStamp/otimeConversions.h>
#include <interfaces/baseTypes/Header.h>

#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <chrono>

using namespace SimpleComms;

namespace task {

pthread_mutex_t SCChannel::channelsMutex_ = PTHREAD_MUTEX_INITIALIZER;
std::map<std::string, std::unordered_set<SCChannel*>> SCChannel::channels_;
bool SCChannel::recvThreadFlag_ = false;
pthread_t SCChannel::recvThread_;
pthread_mutex_t SCChannel::sendQMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t SCChannel::declareMutex_ = PTHREAD_MUTEX_INITIALIZER;
std::map<std::string, std::unordered_set<SCChannel*>> SCChannel::sendChannels_;
bool SCChannel::sendThreadFlag_ = false;
pthread_t SCChannel::sendThread_;
bool SCChannel::somethingToSend_ = false;

SCChannel::SCChannel(const std::shared_ptr<SimpleComms::Client> comms,
                     const std::string channelName,
                     const Access mode,
                     const uint32_t recvQueueDepth,
                     const int32_t latencyWarning_us,
                     std::chrono::milliseconds resaltInterval)
    : Channel(mode)
    , comms_(comms)
    , channelName_(channelName)
    , recvQueueDepth_(recvQueueDepth)
    , latencyWarning_us_(latencyWarning_us)
    , readFrom_(false)
    , writtenTo_(false)
    , connected_(false)
    , receivedQueue_()
    , sendQueue_()
    , resaltInterval_{resaltInterval}
    , channelId_{boost::uuids::to_string(boost::uuids::random_generator()())}
    , latencyStats_(channelName, latencyWarning_us, "Internal Queue", "SCS", static_cast<int32_t>(recvQueueDepth))
{
    recvSignal_ = PTHREAD_COND_INITIALIZER;
    AVLOG_INFO("NOT use black channel wrapper for channel: {}", channelName_);
}

SCChannel::~SCChannel()
{
    if (((mode_ == Read) || (mode_ == ReadWrite)) && (!readFrom_)) {
        size_t totalBytesCurrentlyQueued = 0;

        while (!receivedQueue_.empty()) {
            totalBytesCurrentlyQueued += receivedQueue_.front().first.size();
            receivedQueue_.pop();
        }

        AVLOG_WARN("Never read from SCChannel for channel '{}', which currently has {} bytes queued",
                   channelName_,
                   totalBytesCurrentlyQueued);
    }

    if (((mode_ == Write) || (mode_ == ReadWrite)) && (!writtenTo_)) {
        AVLOG_WARN("Never wrote to SCChannel for channel '{}'", channelName_);
    }

    if (connected_) {
        AVLOG_WARN("Destructing a connected SCChannel for channel '{}'", channelName_);
        disconnect();
    }

    // We don't strictly need to lock here since the receive thread is closed, but let's be thorough
    // in case something changes
    PThreadLocker channelsLock(channelsMutex_);

    std::string results = latencyStats_.getResults();
    AVLOG_WARN("SORT: {:<16.16} CHANNEL_STATS_BEGIN {} CHANNEL_STATS_END", channelName_, results);
}
/*
 * This helper function for removeAccess is a bit verbose. The objectives are:
 *   1. If the SCChannel 'channel' is connected to 'channelName' in the OpenChannelMap, erase it
 *   2. If 'channel' was the last SCChannel connected to 'channelName', we need to:
 *     a. Remove 'channelName' from the OpenChannelMap 'ocm' as there are no more open connections to 'channelName'
 *     b. Return 'true' to indicate that removeAccess should send an 'undeclare' message to the scs
 *
 * Return value false indicates either
 *   a. 'channel' is not connected to 'channelName' for the given access type, so undeclaring would be unnecessary
 *   b. Other SCChannels are connected to 'channelName' for the given access type, so undeclaring would be incorrect.
 */

bool SCChannel::eraseIfPresentReturnIfErasedLast(OpenChannelMap& ocm,
                                                 const std::string& channelName,
                                                 const SCChannel* const channel) const
{
    bool erasedLast = false;

    const OpenChannelMap::iterator iter(ocm.find(channelName));
    if (iter != ocm.end()) {
        const std::unordered_set<SCChannel*>::iterator c =
            std::find((*iter).second.begin(), (*iter).second.end(), channel);

        if (c != ocm[channelName].end()) {
            static_cast<void>(ocm[channelName].erase(c));

            // Maintain invariant of no empty/disconnected channels in the OpenChannelMap.

            if (ocm[channelName].empty()) {
                static_cast<void>(ocm.erase(iter));
                erasedLast = true;
            }
        }
    }

    return erasedLast;
}

ChannelStatus SCChannel::addAccess(const Access mode)
{
    ChannelStatus status = CS_Success;

    ChannelType msgChannelType = ChannelType::Unlisted;

    /*
     * Determine whether declaring read or write access necessary.
     * These checks prevent unnecessary declarations if another SCChannel already
     * connected to the same channel with overlapping access modes.
     */
    if ((mode == Write) || (mode == ReadWrite) || (mode == Standby)) {
        PThreadLocker ptl(sendQMutex_);
        if (sendChannels_.find(channelName_) == sendChannels_.end()) {
            msgChannelType = (msgChannelType == ChannelType::Subscriber) ? ChannelType::SubPub : ChannelType::Publisher;
        }
    }

    if ((mode == Read) || (mode == ReadWrite)) {
        PThreadLocker ptl(channelsMutex_);
        if (channels_.find(channelName_) == channels_.end()) {
            msgChannelType = (msgChannelType == ChannelType::Publisher) ? ChannelType::SubPub : ChannelType::Subscriber;
        }
    }

    // If msgChannelType unlisted, we already declared all the access we need. Otherwise, send the declare message

    if (msgChannelType == ChannelType::Unlisted) {
        status = CS_Success;
    } else {
        status = (comms_->declare(channelName_, msgChannelType, recvQueueDepth_) == Status::Ok) ? CS_Success : CS_Error;
    }

    // Add ourselves to the OpenChannelMaps for the access modes we are adding. Spawn send/recv threads if needed.

    if ((mode == Write) || (mode == ReadWrite) || (mode == Standby)) {
        if (status == CS_Success) {
            PThreadLocker ptl(sendQMutex_);
            sendChannels_[channelName_].insert(this);
            sendQOpen_ = true;
        }

        if (!sendThreadFlag_) {
            sendThreadFlag_ = true;
            static_cast<void>(pthread_create(&sendThread_, nullptr, startSendThread, nullptr));
        }
    }

    if ((mode == Read) || (mode == ReadWrite)) {
        if (status == CS_Success) {
            PThreadLocker ptl(channelsMutex_);
            channels_[channelName_].insert(this);
            recvQOpen_ = true;
            if (!recvThreadFlag_) {
                recvThreadFlag_ = true;
                static_cast<void>(pthread_create(&recvThread_, nullptr, startRecvThread, nullptr));
            }
        }
    }

    return status;
}

ChannelStatus SCChannel::removeAccess(const Access mode)
{
    ChannelStatus status = CS_Success;

    if (connected_ != false) {
        ChannelType msgChannelType = ChannelType::Unlisted;

        /*
         *  Remove ourself from the OpenChannelMap for the needed access modes and determine if message necessary
         *  These checks necessary to prevent unnecessary (and harmful) undeclaration if another SCChannel already
         *  connected to the same channel with overlapping access modes.
         */

        if ((mode == Write) || (mode == ReadWrite) || (mode == Standby)) {
            bool shouldJoin{false};
            {
                PThreadLocker sendlock(sendQMutex_);
                sendQOpen_ = false;

                ChannelStatus flushResult = flushSendQueue();

                if (flushResult != CS_Success) {
                    AVLOG_WARN("Error flushing send queue while closing {} write access", channelName_);
                }

                if (eraseIfPresentReturnIfErasedLast(sendChannels_, channelName_, this)) {
                    msgChannelType =
                        (msgChannelType == ChannelType::Subscriber) ? ChannelType::SubPub : ChannelType::Publisher;
                }

                shouldJoin = (sendThreadFlag_ && sendChannels_.empty());
            }

            if (shouldJoin) {
                AVLOG_WARN("waiting for the send thread to join");
                static_cast<void>(pthread_join(sendThread_, nullptr));
            }
        }

        if ((mode == Read) || (mode == ReadWrite)) {
            bool shouldJoin{false};
            {
                PThreadLocker recvlock(channelsMutex_);
                recvQOpen_ = false;
                unblock();
                if (eraseIfPresentReturnIfErasedLast(channels_, channelName_, this)) {
                    msgChannelType =
                        (msgChannelType == ChannelType::Publisher) ? ChannelType::SubPub : ChannelType::Subscriber;
                }
                shouldJoin = (recvThreadFlag_ && channels_.empty());
            }
            if (shouldJoin) {
                AVLOG_WARN("waiting for the recv thread to join");
                static_cast<void>(pthread_join(recvThread_, nullptr));
            }
        }

        // Undeclare, but only send the message if necessary (don't send if others connected or we were not connected)

        if (msgChannelType != ChannelType::Unlisted) {
            status = (comms_->undeclare(channelName_, msgChannelType) == Status::Ok) ? CS_Success : CS_Error;
        } else {
            status = CS_Success;
        }
    }

    return status;
}

ChannelStatus SCChannel::connect(void)
{
    PThreadLocker declareLock(declareMutex_);

    ChannelStatus status = CS_Success;

    if (!connected_) {
        status = addAccess(mode_);

        if (status == CS_Success) {
            AVLOG_DEBUG("Successfully connected to channel {}", channelName_);
            connected_ = true;
            reportConnected();
        } else {
            AVLOG_ERROR("Failed to connect to channel {}", channelName_);
        }
    }

    return status;
}

ChannelStatus SCChannel::disconnect(void)
{
    PThreadLocker declareLock(declareMutex_);

    ChannelStatus status = CS_Success;

    if (connected_) {
        status = removeAccess(mode_);

        if (status == CS_Success) {
            connected_ = false;
            AVLOG_DEBUG("Successfully disconnected from channel {}", channelName_);
            reportDisconnected();
        } else {
            connected_ = true;
            AVLOG_ERROR("Failed to disconnect from channel {}", channelName_);
        }
    }

    return status;
}

ChannelStatus SCChannel::readWithPred(std::string& msg, av::latency::Pred& pred, const bool blocking)
{
    readFrom_ = true;
    ChannelStatus cs{CS_Success};

    ottotime::otimePoint sendTime{};

    {
        PThreadLocker channelsLock(channelsMutex_);
        if (recvQOpen_) {
            if (blocking && receivedQueue_.empty()) {
                pthread_cond_wait(&recvSignal_, &channelsMutex_);
            }

            if (!receivedQueue_.empty()) {
                const auto& entry = receivedQueue_.front();
                msg = std::move(entry.first);
                sendTime = entry.second;
                pred = av::LatencyLogger::getLogger().ReceiveEvent(channelName_, sendTime);
                cs = CS_NewData;
                receivedQueue_.pop();
            } // fall through, errors from updateMessageQueues() are propagated
        } else {
            AVLOG_WARN("Tried to read from disconnected channel {}!", channelName_);
            cs = CS_Error;
        }
    }

    if (cs == CS_NewData) {
        reportMessage(msg.size(), av::timePointFromOtimePoint(sendTime));
    }

    return cs;
}

ChannelStatus SCChannel::chmod(const Access mode)
{
    PThreadLocker declareMutex(declareMutex_);
    const ChannelStatus cs = CS_Success;

    AVLOG_WARN("Mode is changed from {} to {}", mode_, mode);
    mode_ = mode;

    return cs;
}

void SCChannel::recvThread(void)
{
    while (true) {

        {
            PThreadLocker channelsLock(channelsMutex_);
            if (channels_.empty()) {
                recvThreadFlag_ = false;
                break;
            }

            (*channels_.begin()->second.begin())->updateMessageQueues();
        }
        static_cast<void>(usleep(100U));
    }
}

void SCChannel::sendThread(void)
{
    while (true) {

        {
            PThreadLocker sendQLock(sendQMutex_);
            if (sendChannels_.empty()) {
                sendThreadFlag_ = false;
                break;
            }

            (*sendChannels_.begin()->second.begin())->updateSendQueues();
        }
        static_cast<void>(usleep(100U));
    }
}

ChannelStatus SCChannel::updateMessageQueues(void)
{
    Status::States ret = Status::Error;

    if (((mode_ == Read) || (mode_ == ReadWrite) || (mode_ == Standby)) && comms_->hasMsg()) {
        {
            std::map<std::string, std::unordered_set<SCChannel*>>::iterator it;
            for (it = channels_.begin(); it != channels_.end(); ++it) {
                std::string msg;
                MessageInfo msgInfo;

                while ((ret = comms_->receiveMsg((*it).first, msg, msgInfo)) == Status::NewData) {
                    std::unordered_set<SCChannel*>& destChannels = (*it).second;

                    auto sendTime = std::chrono::seconds(msgInfo.sendMsgTime.tv_sec) +
                                    std::chrono::microseconds(msgInfo.sendMsgTime.tv_usec);
                    addLatencyStartTime(std::chrono::duration_cast<microseconds_long_double>(sendTime), destChannels);

                    const int64_t sendTimeMicroseconds =
                        std::chrono::duration_cast<std::chrono::duration<long long, std::micro>>(sendTime).count();

                    updateRecvQueueForChannels(msg, sendTimeMicroseconds, destChannels);
                }
            }

            updateLatencyEndTime(channels_);
        }
    }

    ChannelStatus result{CS_Error};
    if (ret == Status::Ok) {
        result = CS_Success;
    }

    return result;
}

ChannelStatus SCChannel::updateSendQueues(void)
{
    ChannelStatus ret = CS_Error;

    if ((mode_ == Write) || (mode_ == ReadWrite) || (mode_ == Standby)) {
        if (somethingToSend_) {
            std::map<std::string, std::unordered_set<SCChannel*>>::iterator it;
            for (it = sendChannels_.begin(); it != sendChannels_.end(); ++it) {
                std::unordered_set<SCChannel*>& channels = (*it).second;

                for (std::unordered_set<SCChannel*>::iterator j = channels.begin(); j != channels.end(); ++j) {
                    ChannelStatus result = (*j)->flushSendQueue();

                    if (result != CS_Success) {
                        AVLOG_WARN("Error writing to SimpleComms on channel {}", (*j)->channelName_);
                        ret = CS_Error;
                    }
                }
            }
            somethingToSend_ = false;
        } else {
            ret = CS_Success;
        }
    }

    return ret;
}

ChannelStatus SCChannel::writeV(const std::string& msg, std::vector<av::latency::Pred>&& predecessors)
{
    return this->writeTyped(msg, std::move(predecessors));
}

bool SCChannel::getQueueProperties(unsigned long& currentDepth, unsigned long& maxDepth)
{
    PThreadLocker channelsLock(channelsMutex_);
    currentDepth = receivedQueue_.size();
    maxDepth = recvQueueDepth_;
    return true;
}

void SCChannel::unblock()
{
    static_cast<void>(pthread_cond_signal(&recvSignal_));
}

void SCChannel::clear()
{
    if ((mode_ == Read) || (mode_ == ReadWrite)) {
        decltype(receivedQueue_) empty{};
        PThreadLocker channelsLock(channelsMutex_);
        receivedQueue_.swap(empty);
    }
}

ChannelStatus SCChannel::writeTyped(const std::string& msg, std::vector<av::latency::Pred>&& predecessors)
{
    writtenTo_ = true;
    Status::States ret = Status::Ok;

    if ((mode_ == Write) || (mode_ == ReadWrite)) {
        if (ret == Status::Ok) {
            PThreadLocker sendQLock(sendQMutex_);
            if (sendQOpen_) {
                sendQueue_.emplace(msg, std::move(predecessors));
                somethingToSend_ = true;
            } else {
                ret = Status::Error;
            }
        }
    } else if (mode_ == Standby) {
        ret = Status::Ok;
    } else {
        ret = Status::Error;
    }

    return ((ret == Status::Ok) ? CS_Success : CS_Error);
}

std::uint32_t SCChannel::estimateSaltUpdateValidMessageCount()
{
    // The send queue size is the number of messages that will be sent _before_ the SALT UPDATE itself gets sent.
    // (This is a potential, but harmless, overestimate if there are metadata messages pending, as these don't increment
    // BlackChannel's internal send counter.)
    PThreadLocker sendQLock(sendQMutex_);
    return static_cast<uint32_t>(sendQueue_.size());
}

void SCChannel::updateRecvQueueForChannels(const std::string& msg,
                                           const int64_t sendTimeMicroseconds,
                                           const std::unordered_set<SCChannel*>& channels) const
{
    for (const auto& channel : channels) {
        if (channel != nullptr) {
            channel->receivedQueue_.push({msg, otimePoint::fromMicrosecondsSinceEpoch(sendTimeMicroseconds)});

            if (channel->receivedQueue_.size() > channel->recvQueueDepth_) {
                channel->receivedQueue_.pop();
            }

            static_cast<void>(pthread_cond_signal(&(channel->recvSignal_)));
        }
    }
}

void SCChannel::addLatencyStartTime(const microseconds_long_double& sendTime,
                                    const std::unordered_set<SCChannel*>& channels)
{
    for (const auto& channel : channels) {
        if (channel != nullptr) {
            channel->latencyStats_.addStartTime(sendTime);
        }
    }
}

void SCChannel::updateLatencyEndTime(const std::map<std::string, std::unordered_set<SCChannel*>>& channels)
{
    const auto now{std::chrono::system_clock::now().time_since_epoch()};
    const microseconds_long_double endTime = std::chrono::duration_cast<microseconds_long_double>(now);

    for (const auto& channelSet : channels) {
        for (const auto& channel : channelSet.second) {
            if (channel != nullptr) {
                channel->latencyStats_.setEndTime(endTime);
            }
        }
    }
}

ChannelStatus SCChannel::flushSendQueue()
{
    ChannelStatus ret = CS_Success;

    while (!sendQueue_.empty()) {
        auto& entry = sendQueue_.front();
        const auto& msg = entry.first;
        auto& predecessors = entry.second;

        const Status::States commsResult = comms_->sendMsgV(channelName_, msg, std::move(predecessors));

        sendQueue_.pop();

        if (commsResult != Status::States::Ok) {
            ret = CS_Error;
        }
    }

    return ret;
}

std::unique_ptr<Channel> SCChannel::Builder::build(ConfigSection& channelConfig, Access mode, std::string name) const
{
    uint32_t queueSize;
    static_cast<void>(channelConfig.get(std::string("queueSize"), queueSize, 1U));

    std::optional<int> maxRetriesOpt{};

    {
        uint32_t maxRetries;
        auto status = channelConfig.get(std::string("maxRetries"), maxRetries);
        if (status == ConfigSection::GS_OK) {
            maxRetriesOpt = maxRetries;
        }
    }

    auto sc = ChannelFactory::get().connectToSCS(maxRetriesOpt);

    int32_t latencyWarning_us;
    static_cast<void>(channelConfig.get(std::string("latencyWarning_us"), latencyWarning_us, -1));

    double resaltIntervalSeconds{0};
    static_cast<void>(channelConfig.get(std::string("resaltIntervalSeconds"), resaltIntervalSeconds));
    const auto resaltInterval =
        std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::duration<double>(resaltIntervalSeconds));

    return std::make_unique<SCChannel>(sc, name, mode, queueSize, latencyWarning_us, resaltInterval);
}

} // namespace task
