/*!
***********************************************************
* @file SQLLogChannel.cc
* @brief A file implementing the SQLLogChannel class
*
* @section License
*
* Copyright (c) 2015- Delphi Automotive. All rights
* reserved.
*
* @section changeLogs Change Logs
*
* @author Junsung Kim (junsung.kim@delphi.com)
* @date 07/05/2016
* @remark
* get() supports the blocking call.
* @n
*
* @author Junsung Kim (junsung.kim@delphi.com)
* @date 06/09/2016
* @remark
* Using a timestamp marked when the data is
* sent for collecting logs
* @n
*
* @author Junsung Kim (junsung.kim@delphi.com)
* @date 02/22/2016
* @remark
* Timestamps are replaced with otimePoint.
* @n
*
***********************************************************
*/

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

#include <Comms/SQLLogChannel.h>

#include <cerrno>
#include <chrono>
#include <dataLog/utility.h>
#include <string_view>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <interfaces/baseTypes/Header.h>

#include <Statistics/MeanVariance.hpp>
#include <Statistics/MinMax.hpp>

#include "lib/avlog/avlog.hpp"

namespace task {

namespace {
std::string getNameFromConfigSection(ConfigSection& channelConfig)
{
    std::string res;
    channelConfig.get("name", res, "unused");
    return res;
}
} // namespace

SQLLogChannel::SQLLogChannel(ConfigSection& channelConfig, const Access mode, std::optional<std::string> nameOpt)
    : LogChannel(mode, nameOpt.has_value() ? nameOpt.value() : getNameFromConfigSection(channelConfig))
{
    AVLOG_INFO("CTOR ConfigSection, Access, nameOpt ");

    SQLLogChannelConfig config;
    if (channelConfig.get(std::string("logFileName"), config.logFileName) != ConfigSection::GS_OK) {
        config.logFileName.clear();
    }

    if (channelConfig.get(std::string("enableSQLiteTracing"), config.enableSQLiteTracing) != ConfigSection::GS_OK) {
        config.enableSQLiteTracing = false;
    }

    if (channelConfig.get(std::string("sqlProfileWarning_us"), config.sqlProfileWarning_us) != ConfigSection::GS_OK) {
        config.sqlProfileWarning_us = -1;
    }

    config_ = config;
    ls_.latencyInfo_.channelName = config.logFileName;
    ls_.latencyInfo_.latencyWarning = microseconds_long_double(config.sqlProfileWarning_us);
}

SQLLogChannel::SQLLogChannel(const SQLLogChannelConfig& config, const Access mode)
    : LogChannel(mode, config.logFileName)
    , config_(config)
    , ls_(config.logFileName, static_cast<int32_t>(config.sqlProfileWarning_us), "", "", -1)
{
    AVLOG_INFO("CTOR SQLLogChannelConfig, Access ");
}

SQLLogChannel::SQLLogChannel(SQLLogChannel&& channel)
    : LogChannel(channel.mode_, channel.getName())
    , logDb_(channel.logDb_)
    , readCurrent_(channel.readCurrent_)
    , rowId_(channel.rowId_)
    , config_(channel.config_)
    , logVersion_(channel.logVersion_)
    , ls_(channel.ls_.latencyInfo_.channelName,
          static_cast<int32_t>(channel.ls_.latencyInfo_.latencyWarning.count()),
          "",
          "",
          -1)
{
    AVLOG_INFO("CTOR SQLLogChannel");
    channel.logDb_ = nullptr;
}

void SQLLogChannel::operator=(SQLLogChannel&& channel)
{
    disconnect();

    mode_ = channel.mode_;
    logDb_ = channel.logDb_;
    readCurrent_ = channel.readCurrent_;
    rowId_ = channel.rowId_;
    config_ = channel.config_;
    logVersion_ = channel.logVersion_;
    channel.logDb_ = nullptr;
}

SQLLogChannel::~SQLLogChannel()
{
    disconnect();

    if (config_.enableSQLiteTracing) {
        const std::string results = ls_.getResults();
        AVLOG_INFO("SORT: {:<16.16} SQL_STATS_BEGIN {} SQL_STATS_END", ls_.latencyInfo_.channelName, results);
    }

    // Save any remaining data
    throughput_.addDataWritten(dataWritten_.exchange(0));
    std::string throughputResults{throughput_.getJSON().dump()};
    AVLOG_INFO(
        "SORT: {:<16.16} THROUGHPUT_STATS_BEGIN {} SQL_STATS_END", ls_.latencyInfo_.channelName, throughputResults);
}

int SQLLogChannel::tracecallback(uint32_t type, void* context, void* /*statement*/, void* nanosec64)
{
    if (context != nullptr && nanosec64 != nullptr && type == SQLITE_TRACE_PROFILE) {
        // The context variable MUST be SQLLogChannel*
        SQLLogChannel* channel = reinterpret_cast<SQLLogChannel*>(context);

        uint64_t time_ns = *reinterpret_cast<uint64_t*>(nanosec64);
        std::chrono::nanoseconds duration(time_ns);

        channel->ls_.update(std::chrono::duration_cast<microseconds_long_double>(duration));

        if (channel->ls_.latencyInfo_.latencyWarning > microseconds_long_double(0.0) &&
            duration > channel->ls_.latencyInfo_.latencyWarning) {
            const char* name = channel->ls_.latencyInfo_.channelName.c_str();
            AVLOG_WARN(
                "Channel {} exceeded log warn time {}us. Actual: {}us",
                name,
                std::chrono::duration_cast<std::chrono::microseconds>(channel->ls_.latencyInfo_.latencyWarning).count(),
                std::chrono::duration_cast<std::chrono::microseconds>(duration).count());
        }
    }

    // sqlite3_trace_v2 reference says value ignored, "implementations should return zero for future compatibility."
    return 0;
}

ChannelStatus SQLLogChannel::connect(void)
{
    if (config_.logFileName.empty()) {
        AVLOG_CRITICAL("logFileName parameter undefined/invalid, nothing to do");

        return CS_Error;
    }

    ChannelStatus connectStatus = CS_Success;
    if (mode_ == Write) {
        uint32_t oFlags = static_cast<uint32_t>(SQLITE_OPEN_READWRITE);
        oFlags |= static_cast<uint32_t>(SQLITE_OPEN_CREATE);

        if (connectStatus == CS_Success) {
            const std::string tmpLogFileName = config_.logFileName;

            // If we're about to create the db, we also need to write the version
            // stat returns 0 on success. Note that stat check must happen before the call to sqlite3_open_v2()
            struct stat statBuff;
            const bool shouldWriteVersion = !(stat(config_.logFileName.c_str(), &statBuff) == 0);

            const int status = sqlite3_open_v2(tmpLogFileName.c_str(),       // Database file name
                                               &logDb_,                      // sqlite db handle
                                               static_cast<int32_t>(oFlags), // Flags
                                               nullptr                       // Name of VFS module to use
            );

            switch (status) {
                case SQLITE_OK: {
                    int ret = SQLITE_OK;
                    char* zErrMsg = nullptr;

                    if (shouldWriteVersion) {
                        ret = dataLog::writeLogVersion(logDb_, &zErrMsg);
                        if (ret != SQLITE_OK) {
                            AVLOG_CRITICAL("Failed to write log version: {}", ret);
                            sqlite3_free(zErrMsg);
                            connectStatus = CS_Error;
                        } else {
                            logVersion_ = dataLog::DATA_LOG_VERSION;
                        }
                    } else {
                        ret = dataLog::readLogVersion(logDb_, logVersion_, &zErrMsg);
                        if (ret != SQLITE_OK) {
                            AVLOG_CRITICAL("Failed to read log version: {} {}", ret, sqlite3_errmsg(logDb_));
                            sqlite3_free(zErrMsg);
                            connectStatus = CS_Error;
                        }
                    }

                    ret = sqlite3_exec(logDb_, "DROP TABLE IF EXISTS OTTOLOG;", nullptr, nullptr, &zErrMsg);
                    if (ret != SQLITE_OK) {
                        AVLOG_CRITICAL("Failed to drop the table for {} :{}", tmpLogFileName, zErrMsg);
                        sqlite3_free(zErrMsg);
                    }

                    std::string sqlCmd{""};
                    if (logVersion_ >= 3) {
                        // version 3
                        sqlCmd = "CREATE TABLE OTTOLOG(time REAL NOT NULL, data BLOB NOT NULL, publisherId INTEGER "
                                 "NOT NULL, messageCounter INTEGER NOT NULL);";
                    } else {
                        sqlCmd = "CREATE TABLE OTTOLOG(time REAL NOT NULL, data BLOB NOT NULL);";
                    }

                    ret = sqlite3_exec(logDb_, sqlCmd.c_str(), nullptr, nullptr, &zErrMsg);
                    if (ret != SQLITE_OK) {
                        AVLOG_CRITICAL("Failed to create the table for {} :{}", tmpLogFileName, zErrMsg);
                        sqlite3_free(zErrMsg);
                        connectStatus = CS_Error;
                    } else

                    {
                        config_.logFileName = tmpLogFileName;
                        connectStatus = CS_Success;

                        // logger will not wait for the write to be over before writing the next data
                        ret = sqlite3_exec(logDb_, "PRAGMA synchronous = 0;", 0, 0, 0);

                        // temporary storage for db is memory
                        ret = sqlite3_exec(logDb_, "PRAGMA temp_store = 2;", 0, 0, 0);

                        if (config_.enableSQLiteTracing) {
                            sqlite3_trace_v2(logDb_,
                                             SQLITE_TRACE_PROFILE,
                                             SQLLogChannel::tracecallback,
                                             reinterpret_cast<void*>(this));
                        }
                    }
                    buffer_.connect(logVersion_, logDb_);
                    break;
                }
                default:
                    connectStatus = CS_Error;
                    AVLOG_CRITICAL("Unknown error creating log file, bailing {}", sqlite3_errmsg(logDb_));
                    break;
            }
        }
    } else if (mode_ == Read) {
        const int status = sqlite3_open_v2(config_.logFileName.c_str(), // Database file name
                                           &logDb_,                     // sqlite db handle
                                           SQLITE_OPEN_READONLY,        // Flags
                                           nullptr                      // Name of VFS module to use
        );

        switch (status) {
            case SQLITE_OK: {
                const std::string sqlCmd = "SELECT rowid FROM OTTOLOG ORDER BY ROWID ASC LIMIT 1";
                sqlite3_stmt* pStmt;

                int readStatus;

                readStatus = dataLog::readLogVersion(logDb_, logVersion_, NULL);

                switch (readStatus) {
                    case SQLITE_OK:
                        break;

                    default:
                        AVLOG_CRITICAL("Failed to read log version: {} {}", readStatus, sqlite3_errmsg(logDb_));
                        connectStatus = CS_Error;
                        break;
                }

                readStatus = sqlite3_prepare_v2(logDb_, sqlCmd.c_str(), -1, &pStmt, NULL);

                switch (readStatus) {
                    case SQLITE_OK:
                        break;

                    default:
                        AVLOG_CRITICAL("connect() prepare failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
                        connectStatus = CS_Error;
                        break;
                }

                readStatus = sqlite3_step(pStmt);

                switch (readStatus) {
                    case SQLITE_DONE:
                    case SQLITE_ROW:
                        rowId_ = sqlite3_column_int64(pStmt, 0);
                        AVLOG_INFO("connect() the first row id is {}", rowId_);

                        // If we have a valid rowid, then we should read from that row when
                        // the read API is called.
                        if (rowId_ > 0) {
                            readCurrent_ = true;
                        }
                        break;

                    default:
                        AVLOG_CRITICAL("connect() step failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
                        connectStatus = CS_Error;
                        break;
                }

                readStatus = sqlite3_finalize(pStmt);

                switch (readStatus) {
                    case SQLITE_OK:
                        break;

                    default:
                        AVLOG_CRITICAL("connect() finalize failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
                        connectStatus = CS_Error;
                        break;
                }
                break;
            }
            default:
                connectStatus = CS_Error;
                AVLOG_CRITICAL("Unknown error creating log file, bailing {}", sqlite3_errmsg(logDb_));
                break;
        }
    } else {
        connectStatus = CS_Error;
        AVLOG_CRITICAL("Unknown mode");
    }

    if (connectStatus == CS_Success) {
        AVLOG_INFO("Connected to {}, logVersion = {}", config_.logFileName, logVersion_);
    }

    return connectStatus;
}

ChannelStatus SQLLogChannel::disconnect()
{
    buffer_.reset();
    ChannelStatus status{CS_Success};
    // Finalize any pending transactions before closing the db
    if (logDb_ != nullptr) {
        AVLOG_INFO("Finalizing database file {} before closing", config_.logFileName);
        while (const auto stmt = sqlite3_next_stmt(logDb_, nullptr)) {
            AVLOG_INFO("Finalizing statement {}", sqlite3_expanded_sql(stmt));
            static_cast<void>(sqlite3_finalize(stmt));
        }
    } else {
        AVLOG_INFO("Database for file '{}' null, no statements to finalize", config_.logFileName);
    }

    // Try to close the db
    const auto res = sqlite3_close(logDb_);
    logDb_ = nullptr;

    if (res == SQLITE_OK) {
        // All good
        AVLOG_INFO("Database file {} closed successfully", config_.logFileName);
    } else {
        AVLOG_ERROR("Could not close database file {} cleanly.", config_.logFileName);
        status = CS_Error;
    }

    return status;
}

ChannelStatus SQLLogChannel::getLogExtents(ottotime::otimePoint& start, ottotime::otimePoint& end)
{
    ChannelStatus status = CS_Success;
    int64_t startRowId = 0;
    int64_t endRowId = 0;
    double startTime = 0.;
    double endTime = 0.;

    std::string sqlCmd = "SELECT rowid, time FROM OTTOLOG ORDER BY ROWID ASC LIMIT 1";
    sqlite3_stmt* pStmt;

    int readStatus;

    readStatus = sqlite3_prepare_v2(logDb_, sqlCmd.c_str(), -1, &pStmt, nullptr);

    switch (readStatus) {
        case SQLITE_OK:
            break;

        default:
            AVLOG_CRITICAL("getLogExtents() prepare failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
            status = CS_Error;
            break;
    }

    readStatus = sqlite3_step(pStmt);

    switch (readStatus) {
        case SQLITE_DONE:
        case SQLITE_ROW:
            startRowId = sqlite3_column_int64(pStmt, 0);
            startTime = sqlite3_column_double(pStmt, 1);
            status = keyTootime(startTime, start);
            AVLOG_INFO("getLogExtents() the start row id is {}", startRowId);
            break;

        default:
            AVLOG_CRITICAL("getLogExtents() step failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
            status = CS_Error;
            break;
    }

    readStatus = sqlite3_finalize(pStmt);

    switch (readStatus) {
        case SQLITE_OK:
            break;

        default:
            AVLOG_CRITICAL("getLogExtents() finalize failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
            status = CS_Error;
            break;
    }

    sqlCmd = "SELECT rowid, time FROM OTTOLOG ORDER BY ROWID DESC LIMIT 1";

    readStatus = sqlite3_prepare_v2(logDb_, sqlCmd.c_str(), -1, &pStmt, nullptr);

    switch (readStatus) {
        case SQLITE_OK:
            break;

        default:
            AVLOG_CRITICAL("getLogExtents() prepare failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
            status = CS_Error;
            break;
    }

    readStatus = sqlite3_step(pStmt);

    switch (readStatus) {
        case SQLITE_DONE:
        case SQLITE_ROW:
            endRowId = sqlite3_column_int64(pStmt, 0);
            endTime = sqlite3_column_double(pStmt, 1);
            status = keyTootime(endTime, end);
            AVLOG_INFO("getLogExtents() the end row id is {}", endRowId);
            break;

        default:
            AVLOG_CRITICAL("getLogExtents() step failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
            status = CS_Error;
            break;
    }

    readStatus = sqlite3_finalize(pStmt);

    switch (readStatus) {
        case SQLITE_OK:
            break;

        default:
            AVLOG_CRITICAL("getLogExtents() finalize failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
            status = CS_Error;
            break;
    }

    // if there is no data in db, return error
    if (startRowId == endRowId) {
        status = CS_Error;
    }

    return status;
}

ChannelStatus SQLLogChannel::readWithPred(std::string& msg, av::latency::Pred& pred, MessageMetadata& metadata)
{
    ChannelStatus cs{CS_Success};
    pred = {};

    if (mode_ != Read) {
        AVLOG_ERROR("read() called while not in Read mode (mode_ = {})", mode_);
        cs = CS_Error;
    } else {
        std::string sqlCmd{""};

        if (logVersion_ >= 3) {
            // version 3
            sqlCmd = "SELECT time, data, publisherId, messageCounter FROM OTTOLOG WHERE rowid = ? ORDER BY time ASC";
        } else {
            sqlCmd = "SELECT time, data FROM OTTOLOG WHERE rowid = ? ORDER BY time ASC";
        }

        sqlite3_stmt* pStmt;

        int readStatus = sqlite3_prepare_v2(logDb_, sqlCmd.c_str(), -1, &pStmt, NULL);
        if (readStatus != SQLITE_OK) {
            AVLOG_CRITICAL("read() prepare failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
            cs = CS_Error;
        } else {
            if (!readCurrent_) {
                rowId_++;
            } else {
                readCurrent_ = false;
            }

            readStatus = sqlite3_bind_int64(pStmt, 1, rowId_);
            if (readStatus != SQLITE_OK) {
                AVLOG_CRITICAL("read() key bind failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
                cs = CS_Error;
            } else {
                readStatus = sqlite3_step(pStmt);
                if (readStatus == SQLITE_ROW || readStatus == SQLITE_DONE) {
                    double key = sqlite3_column_double(pStmt, 0);
                    if (keyTootime(key, metadata.timeOfReceipt) == CS_Success) {
                        msg = std::string(static_cast<const char*>(sqlite3_column_blob(pStmt, 1)),
                                          static_cast<size_t>(sqlite3_column_bytes(pStmt, 1)));
                        if (logVersion_ >= 3) {
                            // version 3
                            metadata.publisherId = static_cast<uint32_t>(sqlite3_column_int64(pStmt, 2));
                            metadata.messageCounter = static_cast<uint32_t>(sqlite3_column_int64(pStmt, 3));
                        }
                        cs = CS_NewData;
                    }
                } else {
                    AVLOG_CRITICAL("read() step failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
                    cs = CS_Error;
                }
            }
        }

        // always call finalize
        readStatus = sqlite3_finalize(pStmt);
        if (readStatus != SQLITE_OK) {
            AVLOG_CRITICAL("read() finalize failure: {} {}", readStatus, sqlite3_errmsg(logDb_));
            cs = CS_Error;
        }
    }

    return cs;
}

ChannelStatus SQLLogChannel::seek(const ottotime::otimePoint& seekTime)
{
    if (mode_ != Read) {
        AVLOG_ERROR("seek() called while not in Read mode (mode_ = {})", mode_);
        return CS_Error;
    }

    ChannelStatus status = CS_Success;

    std::string sqlCmd = "SELECT rowid FROM OTTOLOG WHERE time >= ?";
    sqlite3_stmt* pStmt;

    double key;
    status = otimeToKey(seekTime, key);

    if (status == CS_Success) {
        int seekStatus;

        seekStatus = sqlite3_prepare_v2(logDb_, sqlCmd.c_str(), -1, &pStmt, NULL);

        switch (seekStatus) {
            case SQLITE_OK:
                break;

            default:
                AVLOG_CRITICAL("seek() prepare failure: {} {}", seekStatus, sqlite3_errmsg(logDb_));
                status = CS_Error;
        }

        seekStatus = sqlite3_bind_double(pStmt, 1, key);

        switch (seekStatus) {
            case SQLITE_OK:
                break;

            default:
                AVLOG_CRITICAL("seek() key bind failure: {} {}", seekStatus, sqlite3_errmsg(logDb_));
                status = CS_Error;
        }

        seekStatus = sqlite3_step(pStmt);

        switch (seekStatus) {
            case SQLITE_DONE:
            case SQLITE_ROW: {
                int64_t tmpId = sqlite3_column_int64(pStmt, 0);
                if (tmpId > 0) {
                    readCurrent_ = true;
                    rowId_ = tmpId;
                    AVLOG_INFO("seek() the corresponding row id is {} for {}", rowId_, key);
                } else {
                    status = CS_Error;
                }
                break;
            }
            default:
                AVLOG_CRITICAL("seek() step failure: {} {}", seekStatus, sqlite3_errmsg(logDb_));
                status = CS_Error;
        }

        seekStatus = sqlite3_finalize(pStmt);

        switch (seekStatus) {
            case SQLITE_OK:
                break;

            default:
                AVLOG_CRITICAL("seek() finalize failure: {} {}", seekStatus, sqlite3_errmsg(logDb_));
                status = CS_Error;
        }
    }

    return status;
}

ChannelStatus SQLLogChannel::readWithPred(std::string& msg, av::latency::Pred& pred, bool /*blocking*/)
{
    MessageMetadata metadata{};

    return readWithPred(msg, pred, metadata);
}

ChannelStatus SQLLogChannel::write(const std::string& msg, const MessageMetadata& metadata)
{
    if (mode_ != Write) {
        AVLOG_ERROR("write() called while not in Write mode (mode_ = {})", mode_);
        return CS_Error;
    }

    buffer_.add(msg, metadata);

    return CS_Success;
}

ChannelStatus SQLLogChannel::writeV(const std::string& msg, std::vector<av::latency::Pred>&& /*ignore predecessors*/)
{
    interfaces::baseTypes::InterfaceHeader hdr;
    std::size_t hdrSize = sizeof(hdr);
    // the assumption is that the sequence of headerVersion
    // and sendTimeStamp is preserved for good even after
    // new fields are introduced to the interface header
    memcpy(reinterpret_cast<char*>(&hdr), msg.c_str(), std::min(hdrSize, msg.size()));

    // extract the timing information to be used as a key
    struct otimeSpec time;
    interfaces::baseTypes::getSendTimeStamp(hdr, time);
    const ottotime::otimePoint sendTime = ottotime::otimePoint(time);

    // extract the header version info to delete the header
    // before storing logs
    uint32_t hdrVer;
    interfaces::baseTypes::getHeaderVersion(hdr, hdrVer);

    uint32_t publisherId, messageCounter;
    interfaces::baseTypes::getHeaderPublisherId(hdr, publisherId);
    interfaces::baseTypes::getHeaderMessageCounter(hdr, messageCounter);

    hdrSize = interfaces::baseTypes::getInterfaceHeaderSize(hdrVer);
    std::string newMsg = msg;
    newMsg = newMsg.erase(0, hdrSize);

    MessageMetadata meta{sendTime, publisherId, messageCounter};

    return write(newMsg, meta);
}

ChannelStatus SQLLogChannel::write(const std::string& msg, const ottotime::otimePoint& timeOfReceipt)
{
    MessageMetadata metadata{timeOfReceipt, 0, 0};

    return write(msg, metadata);
}

ChannelStatus SQLLogChannel::keyTootime(const double& key, ottotime::otimePoint& time) const
{
    ChannelStatus status{CS_Error};
    if (key != 0) {
        time = convertTootime(key);
        status = CS_Success;
    }

    return status;
}

ChannelStatus SQLLogChannel::otimeToKey(const ottotime::otimePoint& time, double& key)
{
    key = static_cast<double>(otimeToSeconds(time));

    return CS_Success;
}

void SQLLogChannel::updateLogFileName(const std::string& logFileName)
{
    config_.logFileName = logFileName;
}

std::unique_ptr<Channel> SQLLogChannel::Builder::build(ConfigSection& channelConfig,
                                                       Access mode,
                                                       std::string name) const
{
    return std::make_unique<SQLLogChannel>(channelConfig, mode, name);
}

void SQLLogChannel::sampleThroughputMeasurements()
{
    throughput_.addDataWritten(dataWritten_.exchange(0));
}

nlohmann::json SQLLogChannel::getThroughputMeasurements()
{
    return throughput_.getJSON();
}

SQLLogChannel::ThroughputStats::ThroughputStats()
{
    addCollector(std::make_unique<statistics::MinMax<double>>());
    addCollector(std::make_unique<statistics::MeanVariance<double>>());

    lastSampleTime_.getCurrentSteadyTime();
}

void SQLLogChannel::ThroughputStats::addDataWritten(const uint64_t dataWritten)
{
    ottotime::otimePoint otp{};
    otp.getCurrentSteadyTime();
    auto duration = static_cast<double>(elapsedTimeInSeconds(lastSampleTime_, otp));
    lastSampleTime_ = otp;
    addSample(static_cast<double>(dataWritten) / duration);
}

bool SQLLogChannel::WriteBuffer::connect(uint32_t logVersion, sqlite3* db)
{
    bool success = false;
    if (db_ == nullptr && statement_ == nullptr) {
        logVersion_ = logVersion;
        db_ = db;

        std::string sqlCmd;
        if (logVersion_ >= 3) {
            // version 3
            sqlCmd = "INSERT INTO OTTOLOG(time, data, publisherId, messageCounter) VALUES(?, ?, ?, ?)";
        } else {
            sqlCmd = "INSERT INTO OTTOLOG(time, data) VALUES(?, ?)";
        }

        int status = sqlite3_prepare_v2(db_, sqlCmd.c_str(), -1, &statement_, nullptr);
        if (status != SQLITE_OK) {
            AVLOG_CRITICAL("Failed to initialize buffer: {}, {}", status, sqlite3_errmsg(db_));
        } else {
            success = true;
        }
    } else {
        AVLOG_CRITICAL("Failed to initialize SQLLogChannel::WriteBuffer - already initialized!");
    }

    return success;
}

void SQLLogChannel::WriteBuffer::reset()
{
    // get rid of the remaining data in the buffer
    writeOut();

    if (statement_ != nullptr) {
        int writeStatus = sqlite3_finalize(statement_);
        statement_ = nullptr;
        if (writeStatus != SQLITE_OK) {
            AVLOG_CRITICAL("write() finalize failure: {} {}", writeStatus, sqlite3_errmsg(db_));
        }
    }

    // We are just borrowing the db. The channel is responsible to actually close the db.
    db_ = nullptr;
}

bool SQLLogChannel::WriteBuffer::bind(int index, double data)
{
    int writeStatus = sqlite3_bind_double(statement_, index, data);
    if (writeStatus != SQLITE_OK) {
        AVLOG_CRITICAL("write() key bind failure: {} {}", writeStatus, sqlite3_errmsg(db_));
    }
    return writeStatus == SQLITE_OK;
}

bool SQLLogChannel::WriteBuffer::bind(int index, const std::string& data)
{
    int writeStatus = sqlite3_bind_blob(statement_, index, data.c_str(), static_cast<int>(data.size()), SQLITE_STATIC);
    if (writeStatus != SQLITE_OK) {
        AVLOG_CRITICAL("write() data bind failure: {} {}", writeStatus, sqlite3_errmsg(db_));
    }
    return writeStatus == SQLITE_OK;
}

bool SQLLogChannel::WriteBuffer::bind(int index, int64_t data)
{
    int writeStatus = sqlite3_bind_int64(statement_, index, data);
    if (writeStatus != SQLITE_OK) {
        AVLOG_CRITICAL("write() data bind failure: {} {}", writeStatus, sqlite3_errmsg(db_));
    }
    return writeStatus == SQLITE_OK;
}

void SQLLogChannel::WriteBuffer::writeOut()
{
    if (db_ != nullptr && statement_ != nullptr) {
        char* zErrMsg = nullptr;
        int ret = sqlite3_exec(db_, "BEGIN TRANSACTION", nullptr, nullptr, &zErrMsg);
        if (ret != SQLITE_OK) {
            AVLOG_CRITICAL("Failed to begin transaction: {}", zErrMsg);
            sqlite3_free(zErrMsg);
        }

        while (!buffer_.empty()) {
            const logrow& currentRow = buffer_.front();

            bool status = bind(1, currentRow.time);
            status &= bind(2, currentRow.data);

            if (logVersion_ >= 3) {
                status &= bind(3, static_cast<int64_t>(currentRow.publisherId));
                status &= bind(4, static_cast<int64_t>(currentRow.messageCounter));
            }

            if (status) {
                int writeStatus = sqlite3_step(statement_);
                if (writeStatus != SQLITE_OK && writeStatus != SQLITE_DONE) {
                    AVLOG_CRITICAL("write() step failure: {} {}", writeStatus, sqlite3_errmsg(db_));
                }
            }
            sqlite3_reset(statement_);
            buffer_.pop();
        }
        ret = sqlite3_exec(db_, "COMMIT", nullptr, nullptr, nullptr);
        if (ret != SQLITE_OK) {
            AVLOG_CRITICAL("Failed to commit transaction: {}", zErrMsg);
            sqlite3_free(zErrMsg);
        }
        currentVolume_ = 0;
    }
}

void SQLLogChannel::WriteBuffer::add(const std::string& msg, const MessageMetadata& metadata)
{
    double key;
    otimeToKey(metadata.timeOfReceipt, key);

    buffer_.emplace(key, msg, metadata.publisherId, metadata.messageCounter);
    currentVolume_ += (msg.size() + sizeof(MessageMetadata));

    if (currentVolume_ >= maxVolume_ || (key - buffer_.front().time) > writeoutTimeout_) {
        writeOut();
    }
}

SQLLogChannel::WriteBuffer::~WriteBuffer()
{
    reset();
}

} // namespace task
