/*!
 ***********************************************************
 * @file Channel.h
 * @brief A file defining the Channel class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/05/2016
 * @remark
 * Adding the blocking call capability for get()
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 ***********************************************************
 * Channel.h
 * Bryan Salesky
 * 09/02/2006
 *
 * Junsung Kim
 * 12/16/2012
 * A function chmod() is added to support fault-tolerance.
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 ***********************************************************
 */
#ifndef CHANNEL_H
#define CHANNEL_H

#include "Comms/Access.h"
#include "platform/middleware/libs/LatencyLogger/include/LatencyLogger/LatencyLogger.h"
#include <avtime/Time.hpp>
#include <config/ConfigSection.h>
#include <utility/MessageCallback.hpp>

#include <functional>
#include <mutex>
#include <optional>
#include <string>

namespace task {
/*!
 * \enum ChannelStatus
 * \brief Represents the channel status.
 */
typedef enum
{
    CS_Success,
    CS_Error,
    CS_NewData,
    CS_BrokenChannel,
    CS_Connected,
    CS_Disconnected,
    CS_Reconnecting,
    CS_LogInvalidKey,
    CS_Last ///< must be last element in enum
} ChannelStatus;

/*!
 * @class Channel
 *
 * @brief This is an abstract class defining the basic building block for Channel.
 *
 * This is a base class defining channels that enable a task to
 * communicate with tasks/devices/etc.  Its derived classes include
 * SCChannel, DDSChannel, and LogChannel.
 *
 * SCChannel(SimpleCommunicationChannel) is a concrete derived class,
 * which connects to SCS(SimpleCommunicationServer) via IPC mechanism.
 *
 * DDSChannel(SimpleCommunicationChannel) is a concrete derived class,
 * which communicates via the DDS standard.
 *
 * LogChannel is an abstract class designed for storing log data.
 *
 * SQLLogChannel is a concrete class, derived from LogChannel, which
 * connects to a SQLite DB and stores the log data.
 */
class Channel
{
public:
    Channel(Access mode = Access::ReadWrite)
        : mode_(mode)
    {}

    /*!
     * @brief Default virtual  destructor.
     */
    virtual ~Channel() {}

    Channel(const Channel& newChannel) = delete;
    Channel& operator=(const Channel&) = delete;

    class Builder
    {
    public:
        virtual ~Builder() = default;

        /// @brief Build a channel.
        ///
        /// @param channelConfig The ConfigSection describing the channel.
        /// @param mode The access mode used for the channel.
        /// @param name The name of the channel.
        virtual std::unique_ptr<Channel> build(ConfigSection& channelConfig, Access mode, std::string name) const = 0;
    };

    /*!
     * @brief This method shall connect this channel to the IPC module based on the given mode
     *
     * @return ChannelStatus : CS_Success for the successful connection.Otherwise CS_Error.
     */
    virtual ChannelStatus connect(void) = 0;

    /*!
     * @brief This method shall disconnect this channel from the IPC module
     *
     * @return ChannelStatus : CS_Success for the successful disconnection.Otherwise CS_Error.
     */
    virtual ChannelStatus disconnect(void) = 0;

    /// @brief Read a message from the IPC module.
    ///
    /// @param[out] msg as string : The message read from the IPC module
    ///
    /// @param[in] blocking : if true, the call will block until a
    /// message is available.
    ///
    /// @return ChannelStatus : CS_Success for the successful execution.
    inline ChannelStatus read(std::string& msg, bool blocking = false)
    {
        av::latency::Pred p;
        return readWithPred(msg, p, blocking);
    }

    /// @brief Read a message from the IPC module.
    ///
    /// @param[out] msg as string : The message read from the IPC module
    ///
    /// @param[out] pred : The predecessor to the read event.
    ///
    /// @param[in] blocking : if true, the call will block until a
    /// message is available.
    ///
    /// @return ChannelStatus : CS_Success for the successful execution.
    virtual ChannelStatus readWithPred(std::string& msg, av::latency::Pred& pred, bool blocking = false) = 0;

    /*!
     * @brief This method shall write the messages to the IPC module to publish.
     *
     * @param[in] msg: The message to be published.
     * @param[in] predecessors: The predecessors for latency logging
     * purposes. Note that it's a universal reference (&&), so
     * writeV() takes ownership of the vector.
     *
     * @return ChannelStatus : CS_Success for the successful execution
     */
    virtual ChannelStatus writeV(const std::string& msg, std::vector<av::latency::Pred>&& predecessors) = 0;

    template<class... Preds>
    ChannelStatus write(const std::string& msg, Preds... predecessors);

    /*!
     * @brief This method shall change the mode of the channel.
     *
     * @param[in] mode as Access enum : The new mode to be changed
     *
     * @return ChannelStatus : CS_Success for the successful execution.
     */
    virtual ChannelStatus chmod(const Access mode) = 0;

    /*!
     * @brief This method shall retrieve message queue properties of this channel.
     *
     * @param[out] currentDepth : The number of messages waiting in the queue.
     *
     * @param[out] maxDepth : The max size of the message queue.
     *
     * @return bool : false implies this channel does not support queue (i.e. LogChannel).otherwise,true.
     */
    virtual bool getQueueProperties(unsigned long& currentDepth, unsigned long& maxDepth) = 0;

    /*!
     * @brief This method shall unblock the channel if it is doing a blocking read.
     *
     * @return void
     */
    virtual void unblock() {}

    /**
     * @brief This method causes the channel to take its current count of data written and update internal throughput
     * statistics.
     *
     * \note This is a no-op unless overridden on the derived classes.
     *
     * @return void
     */
    virtual void sampleThroughputMeasurements() {}

    /*!
     * @brief This method shall clear the state of the channel.
     *
     * @return void
     */
    virtual void clear() = 0;

    virtual const std::string& getName() const = 0;

    /// @brief An RAII class for registering a callback.
    class CallbackRegistration : public task::stats::MessageCallbackRegistration
    {
    public:
        /// @brief Constructor.
        ///
        /// @param callback The callback to register. It will be
        /// unregistered when this object is destructed.
        explicit CallbackRegistration(task::stats::MessageCallback callback)
            : task::stats::MessageCallbackRegistration(callbacks_, callback)
        {}
    };

    /// @brief Report that a message was sent or received.
    ///
    /// If sendTime is omitted, it's assumed that the message was
    /// received.
    ///
    /// @param channelName The name of the channel on which the
    /// message was sent/received.
    /// @param messageSize The number of bytes in the message.
    /// @param sendTime. If present, the time at which the message was sent.
    static void reportMessage(const std::string& channelName,
                              size_t messageSize,
                              std::optional<av::TimePoint> sendTime = std::nullopt);

    /// @brief Report that the specified channel was connected.
    static void reportConnected(const std::string& channelName);

    /// @brief Report that the specified channel was disconnected.
    static void reportDisconnected(const std::string& channelName);

protected:
    /// @brief Report that a message was sent or received.
    ///
    /// If sendTime is omitted, it's assumed that the message was
    /// received.
    ///
    /// @param messageSize The number of bytes in the message.
    /// @param sendTime. If present, the time at which the message was sent.
    void reportMessage(size_t messageSize, std::optional<av::TimePoint> sendTime = std::nullopt);

    /// @brief Report that this channel was connected.
    void reportConnected();

    /// @brief Report that this channel was disconnected.
    void reportDisconnected();

    Access getMode() const { return mode_; }

protected:
    Access mode_;

private:
    static task::stats::MessageCallbackRegistry callbacks_;
    static void reportEvent(const task::stats::MessageReport& report);
};

template<class... Preds>
inline ChannelStatus Channel::write(const std::string& msg, Preds... predecessors)
{
    std::vector<av::latency::Pred> v{predecessors...};
    return writeV(msg, std::move(v));
}

} // namespace task

#endif // CHANNEL_H
