/*!
 ***********************************************************
 * @file ChannelFactory.h
 * @brief A file defining the ChannelFactory class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * ChannelFactory.h
 * Tugrul Galatali
 * 10/31/2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */
#ifndef CHANNELFACTORY_H
#define CHANNELFACTORY_H

#include "Comms/Channel.h"
#include <config/ConfigSection.h>

#include <map>
#include <memory>
#include <optional>
#include <string>
#include <vector>

namespace SimpleComms {
class Client;
}

class ChannelFactoryInitTests_ConnectToSCS_Test;

namespace task {

/*!
 * @class ChannelFactory
 * @brief This class shall work as factory class, creates channels for a task to communicate with task/device/etc
 * \n \b Detailed \b Description:
 * \n This class shall create channels for different purposes based on given configuration.
 * \n For example:
 * \n Simple Communication Channel(SCChannel) for task uses IPC mechanism.
 * \n Log Channel (SQLogChannel) for logging uses SQLLite db connection.
 * \n It also maintains the list of channels created as a list.
 * \n It connects to  SCS(SimpleCommunicationServer) if the channel created is of type 'SimpleComms'
 * \n There exists only one ChannelFactory object in task which holds connection to SCS and
 * \n the task module name connected to SCS.
 * \n It can create SCChannel to use black channel wrapper if enabled in configuration.
 *
 */
class ChannelFactory
{
public:
    /*!
     * @brief This method shall set the module name connecting to SCS(Simple Communication Server) via IPC
     *
     * @param[in] moduleName The name given to be used for module identification
     *
     * @return True for success
     *
     */
    static bool connectAs(const std::string& moduleName);
    /// \brief Use this client object instead of allocating one internally.
    static void useClient(const std::shared_ptr<SimpleComms::Client>& client);

    /*!
     * @brief This method shall create a channel with the given config and access mode.
     * \n Connects the channel to the IPC or SQLite DB based on configuration channel type
     *
     * @param[in] channelConfig Channel configuration
     *
     * @param[in] mode The mode of access (read, write, slave, read and write)
     *
     * @return A unique pointer to the new channel
     *
     */
    static std::unique_ptr<Channel> createChannelManaged(ConfigSection& channelConfig,
                                                         const Access mode,
                                                         std::optional<std::string> nameOpt = std::nullopt);

    [[deprecated]] static Channel* createChannel(ConfigSection& channelConfig, const Access mode, std::string name)
    {
        return createChannelManaged(channelConfig, mode, name).release();
    }

    [[deprecated]] static Channel* createChannel(ConfigSection& channelConfig, const Access mode)
    {
        return createChannelManaged(channelConfig, mode).release();
    }

    /*!
     * @brief This method shall return ChannelFactory object.
     * \n There exists only one ChannelFactory object in task which holds connection to SCS and
     * \n also keeps the module name connected to SCS.
     *
     * @return ChannelFactory object
     */
    static ChannelFactory& get();

    /// @brief Cleanup any shared objects help by the singleton object.
    static void cleanup();

    /// @brief Connect to SCS.
    ///
    /// @param maxRetries If maxRetries.has_value(), the maximum
    /// number of times to retry the connection. Otherwise, the
    /// connection will be retried indefinitely.
    ///
    /// @return The SCS client. Two successful calls to connectToSCS()
    /// will return the same client; hence shared_ptr.
    std::shared_ptr<SimpleComms::Client> connectToSCS(std::optional<int> maxRetries = std::nullopt);

private:
    /*!
     * @brief Default constructor of the ChannelFactory class
     *
     */
    ChannelFactory();

    /*!
     * @brief Default destructor.
     *
     */
    ~ChannelFactory() = default;

    ChannelFactory(const ChannelFactory& newChannel) = delete;
    ChannelFactory& operator=(const ChannelFactory&) = delete;

    std::map<std::string, std::unique_ptr<Channel::Builder>> builders_;

    /// @brief Registers a Builder to construct a given type of channel.
    ///
    /// @param type The type name of the channel, as used in a config.
    /// @param builder A unique pointer to the Builder. Gets moved
    /// into the ChannelFactory.
    void registerBuilder(std::string type, std::unique_ptr<Channel::Builder> builder)
    {
        builders_[type] = std::move(builder);
    }

    std::string moduleName_; ///< A unique name given to module connected to SCS for the addressing purpose.

    std::shared_ptr<SimpleComms::Client> sc_{nullptr}; ///< A SimpleComms client pointer connecting to SCS

    /// A unit test that needs to be able to construct a CF for
    /// hermetic testing.
    friend class ::ChannelFactoryInitTests_ConnectToSCS_Test;
};

} // namespace task

#endif // CHANNELFACTORY_H
