/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief An external clock implementation that can be given an interface for PlaybackClockParams and a base Clock.
 *
 * Parameters from the PlaybackClockParams are used to transform the base Clock. This is intended to replace the
 * ClockPlaybackClient interface.
 */

#ifndef COMMS_COMMSSPLAYBACKCLOCK_H
#define COMMS_COMMSSPLAYBACKCLOCK_H

#include <Clock/ClockSource.h>
#include <interfaces/Clock/Playback/Client/Impl.h>
#include <interfaces/Generic/Input/Abstract.h>

#include <optional>

/**
 * @brief A clock that uses PlaybackClockParams to transform the time reported by a base clock object.
 */

class CommsPlaybackClock : public ClockParamsClock
{
public:
    typedef GenericInputIntf<PlaybackClockParams> ParamsInputIntf;
    // Delete the default constructor; it's a useless class without.
    CommsPlaybackClock() = delete;

    /**
     * @brief Parametrized Constructor
     *
     * @param[in] interface : ParamsInputInterfaceType* A pointer to an input interface for PlaybackClockParams.
     * @param[in] baseClock : Clock* A pointer to a "base clock" to be offset. Typically the local steady clock.
     */
    CommsPlaybackClock(ParamsInputIntf* interface, Clock* baseClock);

    /**
     * @brief Default destructor
     */
    virtual ~CommsPlaybackClock() = default;

    /**
     * @brief Function to return the current time.
     *
     * \return Current time, or default otimePoint if params or baseClock are invalid.
     */
    virtual ottotime::otimePoint getCurrentTime();

    /**
     * @brief Return the current PlaybackClockParams
     *
     * \return PlaybackClockParams used by the clock.
     */
    virtual std::optional<PlaybackClockParams> getParams();

private:
    /**
     * \brief Helper function to fetch the latest parameters from the input interface.
     */
    void fetchParams();

    std::optional<PlaybackClockParams> m_params; /// The most recent playback parameters

    ParamsInputIntf* m_paramsInput{nullptr}; /// An externally managed input interface to receive PlaybackClockParams

    Clock* m_baseClock{nullptr}; /// An externally managed pointer to the base clock.
};

#endif // COMMS_COMMSSPLAYBACKCLOCK_H
