#ifndef DDS_CHANNEL_H_
#define DDS_CHANNEL_H_

#include "Comms/Channel.h"
#include <condition_variable>
#include <cstdint>
#include <dds/dds.hpp>
#include <memory>
#include <mutex>
#include <queue>
#include <unordered_map>
#include <utility>

#include "DDSConfigurator.hpp"
#include "DDSConfiguratorHelper.hpp"
#include "DataReaderListener.hpp"
#include "DataWriterListener.hpp"
// forward declaration of testing friend
class DDSChannelInitializationTest;

namespace task {
namespace {
#ifdef DDS_VERSION_RTI_DDS_CPP
using DomainId = std::int32_t;
#else
using DomainId = std::uint32_t;
#endif

// MAX_CAPACITY set as 1000, user can change it if they need deeper queue
constexpr inline std::uint32_t MAX_CAPACITY = 1000;
constexpr inline DomainId defaultDomain = 0;
} // namespace

class DDSChannel : public Channel
{
public:
    DDSChannel(std::string name,
               Access mode,
               std::uint32_t recvQueueDepth,
               DomainId domainId = defaultDomain,
               std::shared_ptr<DDSComms::DDSConfigurator> config =
                   std::make_shared<DDSComms::DDSConfiguratorType>(DDSComms::DDSOptionsType{}));
    ~DDSChannel();

    class Builder : public Channel::Builder
    {
    public:
        std::unique_ptr<Channel> build(ConfigSection& channelConfig, Access mode, std::string name) const override;
    };

    ChannelStatus connect() override;
    ChannelStatus disconnect() override;

    /// @brief Read a message from the IPC module.
    ///
    /// @param[out] msg as string : The message read from the IPC module
    ///
    /// @param[out] pred : The predecessor to the read event.
    ///
    /// @param[in] blocking as bool : if true, the call will block
    /// until a message is available.
    ///
    /// @return ChannelStatus : CS_Success for the successful execution.
    ChannelStatus readWithPred(std::string& msg, av::latency::Pred& pred, bool blocking = false) override;
    ChannelStatus writeV(const std::string& msg, std::vector<av::latency::Pred>&& predecessors) override;
    ChannelStatus chmod(const Access mode) override;
    bool getQueueProperties(unsigned long& currentDepth, unsigned long& maxDepth) override;
    void clear() override;
    void unblock() override;
    const std::string& getName() const override { return channelName; }

    /// testing friend class
    friend class ::DDSChannelInitializationTest;

private:
    void setupDDSDomainAndTopic();
    void createReader();
    void createWriter();

private:
    using DomainParticipant = dds::domain::DomainParticipant;
    using PayloadType = DDSComms::PayloadType;
    using BytesTopic = dds::topic::Topic<PayloadType>;
    using Publisher = dds::pub::Publisher;
    using Subscriber = dds::sub::Subscriber;
    using BytesWriter = dds::pub::DataWriter<PayloadType>;
    using BytesReader = dds::sub::DataReader<PayloadType>;
    using SharedBytesSamples = dds::sub::SharedSamples<PayloadType>;
    using LoanedBytesSamples = dds::sub::LoanedSamples<PayloadType>;

    struct Msg
    {
        Msg(PayloadType data, const dds::core::Time& sendTime, ottotime::otimePoint recvTime)
            : data(std::move(data))
            , sendTime(ottotime::otimeSpec(sendTime.sec(), sendTime.nanosec()))
            , recvTime(std::move(recvTime))
        {}

        Msg(const Msg& other) = default;
        Msg(Msg&& other) = default;

        Msg& operator=(const Msg& other) = default;
        Msg& operator=(Msg&& other) = default;

        const PayloadType data;
        const ottotime::otimePoint sendTime;
        const ottotime::otimePoint recvTime;
    };

    class BytesReaderListener : public DDSComms::DataReaderListener<PayloadType>
    {
    public:
        explicit BytesReaderListener(DDSChannel* channel)
            : DDSComms::DataReaderListener<PayloadType>(channel->getName())
            , channel(channel)
        {}

        void on_data_available(BytesReader& reader) override
        {
            if (!channel->channelConnected) {
                return;
            }

            const auto& samples = reader.select().state(dds::sub::status::DataState::new_data()).take();
            const auto recvTime{SystemClock::getSystemTime()};
            for (const auto& sample : samples) {
                if (sample.info().valid()) {
#ifdef DDS_VERSION_RTI_DDS_CPP
                    const auto sendTime = sample.info().source_timestamp();
#else
                    const auto sendTime = sample.info().timestamp();
#endif
                    channel->receivingQueue.push(std::make_shared<Msg>(sample.data(), sendTime, recvTime));
                }
            }
            if (!channel->receivingQueue.empty()) {
                channel->waitDataConVar.notify_one();
            }
        }

    private:
        DDSChannel* channel;
    };

    class BytesWriterListener : public DDSComms::DataWriterListener<PayloadType>
    {
    public:
        explicit BytesWriterListener(DDSChannel* channel)
            : DDSComms::DataWriterListener<PayloadType>(channel->getName())
            , channel(channel)
        {}

    private:
        DDSChannel* channel;
    };

    // A specialized queue to store unread messages.
    // 2 specific features of the queue:
    //
    // - Maintain a fixed size defined on creation.
    //   Oldest messages will be removed if the queue is excess.
    // - Queue size +1 to solve circular buffer issue, when queue.begin == queue.end,
    //   distinguish an empty buffer with a full buffer
    // - A thread-safe queue, using a simple lock-free approach with the assumption that
    // there is only one single push thread and/or one single pop thread
    // at the same time.

    class LoanedQueue
    {
    public:
        using ItemType = std::shared_ptr<const Msg>;

        LoanedQueue(uint32_t size);

        bool empty() const;
        std::size_t size() const;

        void clear();
        ItemType pop();
        void push(ItemType item);

    private:
        std::vector<ItemType> buffer;
        struct Range
        {
            uint32_t begin, end;
        };
        Range range;
        mutable std::mutex queueMutex;
        uint32_t maxSize;
    };

    std::shared_ptr<DDSComms::DDSConfigurator> config_;
    DomainId domainId;
    std::string channelName;
    bool blockWhenNoData;
    std::uint32_t recvQueueDepth;
    std::mutex channelMutex;
    std::condition_variable waitDataConVar;
    std::atomic<bool> channelConnected{false};
    LoanedQueue receivingQueue;

    BytesReader reader;
    BytesWriter writer;
    BytesReaderListener readerListener;
    BytesWriterListener writerListener;
    DomainParticipant domainParticipant;
    BytesTopic topic;
    Publisher publisher{dds::core::null};
    Subscriber subscriber{dds::core::null};
    uint32_t sequenceNumber{0};

    Publisher getPublisher() const;
    Subscriber getSubscriber() const;

    BytesWriter getWriter() const;
    BytesReader getReader() const;
};
} // namespace task

#endif // DDS_CHANNEL_H_
