/*!
 ***********************************************************
 * @file Interface.h
 * @brief A file defining the Interface class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 ***********************************************************
 * Interface.h
 * Bryan Salesky
 * 09/02/2006
 *
 * Junsung Kim
 * 12/16/2012
 * A function chmod() is added to support fault-tolerance.
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 ***********************************************************
 */
#ifndef INTERFACE_H
#define INTERFACE_H

#include "Comms/ChannelFactory.h"
#include "Comms/InterfaceDb.h"
#include "Comms/TypedChannel.h"
#include <ODatum/ODatum.h>
#include <config/ConfigSection.h>

#include "lib/avlog/avlog.hpp"

#include <string>

namespace task {

/*!
 * @class Interface
 *
 * @brief The basic interconnect between elements in the system.
 * \n \b Detailed \b Description:
 * \n This is an abstract class inherited by all the Input and Output Interfaces.
 * \n It shall create a TypedChannel based on the supplied channel configuration and given mode(
 * Read/Write/ReadWrite/Standby) \n The Interface's TypedChannel can be used to communicate with task/devices/SQLiteDB
 * etc \n TypedChannel uses message data specific to the interface. \n For Example: \n HardwareHealth Interface uses
 * HardwareHealthMessage data \n HealthDisplay Interface uses HealthDisplayMessage data
 */
template<class T>
class Interface : virtual public IDbObject
{
public:
    /*!
     * @brief Parameterized constructor.
     * \n \b Operational \b Flow:
     * \n It shall get the interface name and channel configuration from the given configuration.
     * \n It shall create raw channel with given access mode.
     * \n The channel can be of different types.
     * \n For example:
     * \n Channel type - "SimComms"  SimpleCommunicationChannel (SCChannel) for IPC
     * \n Channel type - "Log"       SQLLogChannel can be created for logging ( SQLite DB )
     * \n It creates a templatized TypedChannel which wraps the raw channel.
     * \n TypedChannel uses message data specific to the interface.
     * \n Trace message shall be logged if unable to create channel.
     *
     * @param[in] cs as ConfigSection object. Configuration parameters for the interface.
     *
     * @param[in] acc as Access enum. Access mode for the interfaces like Rear/Write/ReadWrite/Standby.
     *
     */
    Interface(ConfigSection& cs, const Access& acc)
        : IDbObject()
        , channel_(nullptr)
        , interfaceName_()
        , interfaceParams_(cs)
    {
        cs.get("interface", interfaceName_, "");

        // create an appropriate channel object based on the
        // configuration.

        ConfigSection channelConfig;

        if (cs.get(std::string("channel"), channelConfig) == ConfigSection::GS_OK) {
            std::unique_ptr<Channel> rawChannel = ChannelFactory::createChannelManaged(channelConfig, acc);

            if (rawChannel != nullptr) {
                channel_ = new TypedChannel<T>(std::move(rawChannel));
            }
        }

        if (channel_ == nullptr) {
            AVLOG_INFO("Interface {} created without a channel", interfaceName_);
        }
    }

    /*!
     * @brief Destructor of the Interface class.
     * It shall delete the channel.
     *
     */
    virtual ~Interface()
    {
        if (channel_ != nullptr) {
            delete channel_;
        }
    }

    /*!
     * @brief This method shall check if the channel exists for the Interface.
     *
     * @return bool true if it exists.otherwise,false.
     *
     */
    inline bool hasChannel(void) const { return (channel_ != nullptr); }

    /*!
     * @brief This method shall delete the existing channel for the Interface
     * \n and updates Interface's channel with the given channel.It shall be used
     * \n when Interface is created with channel and want to replace the underlying channel
     * \n with given channel.
     *
     * @param[in] channelInterface TypedChannel pointer.channel which replaces
     * 						       the existing channel for the interface.
     *
     * @return void.
     */
    void replaceChannel(TypedChannel<T>* channelInterface)
    {
        const auto temp = channel_;
        channel_ = channelInterface;
        delete temp;
    }

    /*!
     * @brief This method shall get the Interface's channel.
     * \n The TypedChannel returned shall be used either to read/write message data using IPC (or)
     * \n write logging data to SQLite DB based on channel type.
     *
     * @return The TypedChannel for the Interface.
     *
     * \sa task::TypedChannel
     *
     */
    inline TypedChannel<T>& channel(void) const
    {
        // TODO: This is not safe
        return *channel_;
    }

    /*!
     * @brief Virtual initialization function.
     * \n This method shall be implemented by concrete derived classes of Interface
     *
     * @return bool : true, if channel successfully connects to IPC/SQLiteDb etc. otherwise, false.
     *
     */
    virtual bool initialize(void) = 0;

    /*!
     * @brief Virtual change mode function with default empty implementation.
     * \n This method shall be implemented by concrete derived classes of Interface.
     * \n It shall change the access mode of the interface with given mode.
     *
     * @return Void.
     *
     */
    virtual void chmod([[maybe_unused]] const task::Access& acc) { return; }

protected:
    /*!
     * @brief Virtual cleanup function.
     * \n This method shall be implemented by concrete derived classes of Interface.
     * \n It shall disconnect the channel from IPC/SQLiteDb etc..
     *
     * @return Void.
     *
     */
    virtual void cleanup(void) = 0;

    TypedChannel<T>* channel_; ///< The typed channel shall read/write messages specific to Interface.
                               ///< For ex: HardwareHealth Interface uses HardwareHealthMessage data

    std::string interfaceName_; ///< The name of the interface to be created shall be supplied from configuration.

    ConfigSection interfaceParams_; ///< The configuration parameters for the Interface.

private:
    /*!
     * @brief Copy constructor of the Interface class
     *
     */
    Interface(const Interface&);
    void operator=(const Interface&);
};

} // namespace task

#endif // INTERFACE_H
