/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file InterfaceUtils.h
 * @product common
 * @version
 * @released
 * @updated 02/05/2019
 * @brief Interface utility functions
 *
 * These utility functions are for using ottoware communications in arbitrary C++ that does not
 * inherit from Task
 *
 * @section License
 *
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#ifndef INTERFACE_UTILS_H
#define INTERFACE_UTILS_H

#include "Comms/Access.h"
#include "Comms/InterfaceDb.h"

#include <Clock/ClockSource.h>
#include <config/ConfigSection.h>

#include <memory>
#include <optional>
#include <string>

namespace comms {

/*!
 * @brief Status code for Interface utils
 *
 * \n Ideally the entire comms system has a unified error handling scheme, but
 * \n much of the code is written in pre-C++11 with C-style influence, so we'll stick
 * \n with this for now and unify when appropriate.
 */
enum class InterfaceStatus
{
    OK,
    CREATE_FAILED,
    INITIALIZE_FAILED,
    FETCH_FAILED,
    WRONG_TYPE,
    BAD_CONFIG
};

enum class ChannelType : uint8_t
{
    SimpleComms = 0,
    DDS = 1,
    SQL_LOG = 2
};

/*!
 * @brief Converts from a ChannelType enum to the channel's name.
 * Names are chosen to match the ChannelFactory strings.
 *
 * @param[in] channelType as ChannelType: The ChannelType enum
 *
 * @return an std::optional<std::string> containing the name of the channel, if ct is valid. No value if invalid.
 */
std::optional<std::string> channelTypeName(ChannelType channelType);

/*!
 * @brief Converts from a channel's name to the ChannelType enum.
 * Names are chosen to match the ChannelFactory strings.
 *
 * @param[in] channelTypeName as string: The name of the channel type
 *
 * @return an std::optional<ChannelType> containing the ChannelType, if ct is valid. No value if invalid.
 */
std::optional<ChannelType> channelTypeEnum(const std::string& channelTypeName);

/*!
 * @brief Gets an externally driven clock interface with a fixed configuration.
 *
 * This is needed to avoid a circular dependency between SystemClock and ottocomms,
 * which is a pain for the build system to work with.
 *
 * Currently this creates a ClockPlaybackClient, which expects to be driven
 * by a ClockPlaybackServer.
 *
 * @param[in] channelType as string: The channel type: "SimpleComms" or "DDS"
 *
 * @return A pointer to the Clock.
 */
Clock* getExternallyDrivenClock(const ChannelType channelType);

/*!
 * @brief Gets the ConfigSection for the SystemPlaybackClock
 *
 *
 * @param[in] channelName as string
 *
 * @return an std::optional<ConfigSection> containing the config, if the channel name is valid. No value if invalid.
 */
std::optional<ConfigSection> getSystemPlaybackClockConfigSection(const std::string& channelName);

/*!
 * @brief This method shall create and return an interface, given the interface name, config and access
 * \n \b Operational \b Flow:
 * \n It shall look for configuration parameter named 'interface' in given config object.
 * \n If it exists then it shall create interface with given access using type passed as template argument.
 * \n It shall reset the interface if the initialization fails.
 *
 * \n InterfaceType expects a type that inherits from the 'Interface' class, or any type that
 * \n provides 'InterfaceType(ConfigSection&, const Access&)' and 'bool initialize()'.
 * \n Generally in the form of ChannelType<MessageType>, with the most common uses being
 * \n GenericInputChannel<MyMessageType> and GenericOutputChannel<MyMessageType>.
 * \n It also accepts specialized non-templated interfaces (e.g. VehicleStateInterpolated).
 *
 * @param[in] interfaceName The name of the interface.
 * @param[in] config The interface config section.
 * @param[in] access The access mode (R/W/RW).
 *
 * @return Shared pointer to interface on successful creation, nullptr otherwise.
 *
 */
template<typename InterfaceType>
std::shared_ptr<InterfaceType> getInterface(const std::string& interfaceName,
                                            ConfigSection& config,
                                            task::Access access)
{
    static_cast<void>(interfaceName);
    std::shared_ptr<InterfaceType> interfaceInstance{nullptr};

    // Interface config requires parameter 'interface' to be a valid definition
    if (config.hasParameter("interface")) {
        interfaceInstance = std::make_shared<InterfaceType>(config, access);

        if (interfaceInstance->initialize() == false) {
            // Reset so the inferface isn't returned in an invalid state
            interfaceInstance.reset();
        }
    }

    return interfaceInstance;
}

} // namespace comms

#endif // INTERFACE_UTILS_H
