/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file LatencyStatistics.h
 * @product Latency Statistics
 * @version 1
 * @released 11/27/2018
 * @brief Calculation of application-to-application message latency statistics.
 *
 * This code implements statistics of application-to-application message latency, where latency is the time from when a
 * message is sent in publish(), to when it is received by another task in get(). The following statistics will be
 * implemented:
 * - mean latency
 * - standard deviation
 * - min latency
 * - max latency
 * For the mean and standard deviation, Welford's method is implemented
 * (see http://jonisalonen.com/2013/deriving-welfords-method-for-computing-variance/)
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#ifndef LATENCYSTATS_H
#define LATENCYSTATS_H

#include <chrono>
#include <cmath>
#include <limits>
#include <map>
#include <nlohmann/json.hpp>

#include "lib/avlog/avlog.hpp"

using nanoseconds_long_double = std::chrono::duration<long double, std::nano>;
using milliseconds_long_double = std::chrono::duration<long double, std::milli>;
using microseconds_long_double = std::chrono::duration<long double, std::micro>;
using json = nlohmann::json;

/*!
 * @enum LatStatsError
 *
 * @brief Error codes for the LatencyStats class
 *
 */
enum class LatStatsError
{
    LAT_STATS_OK,
    LAT_STATS_NEG_VAL
};

/*!
 * @struct InitSlowdownThresholdInfo
 *
 * @brief This struct shall hold the information about the initial slowdown
 *
 * \n This struct holds the information about the slowdown in latency at initial startup.
 * \n This is motivated by RTI DDS that shows slow latencies at startup thought to be
 * \n due to high levels of Discovery Traffic slowing down communications.
 */
struct InitSlowdownThresholdInfo
{
    bool foundUnderThreshold{false};   //  True, if we have found a time under the threshold
    size_t numInitialOverThreshold{0}; //  Number of samples found at startup have latency over the threshold
    microseconds_long_double timeInitialLongDelayTotal{
        0}; //  Total time accumulated for initial samples having latency over the threshold
    const microseconds_long_double thresholdInitialSlowdown{
        1000000}; // Threshold in microseconds for initial slow down test
};

constexpr uint32_t LATENCY_MAX_INITIAL_SLOWDOWNS =
    10U; ///< Number of max tries to look for samples under the threshold.
         ///< Stop looking for more samples, if max tries are over.

/*!
 * @struct ChannelSetupLatencyInfo
 *
 * @brief This struct holds the information about the channel setup that might have relevance to the Latency performance
 *
 */
struct ChannelSetupLatencyInfo
{
    /*!
     * @brief Parameterized constructor
     *
     * @param[in] type as string : The given channel type
     * @param[in] recvSize as int32_t : The given size
     *
     */
    explicit ChannelSetupLatencyInfo(std::string type, int32_t _recvSize)
        : channelType(std::move(type))
        , recvSize(_recvSize)
    {}

    /*!
     * @brief Default constructor
     */
    ChannelSetupLatencyInfo() = delete;

    std::string channelType{"???"}; ///< This variable represents the channel type.
    int32_t recvSize{-1};           ///< This variable represents the receive queue size.
};

/*!
 * @struct LatencyInfo
 *
 * @brief This struct shall hold the latency statistics information i.e minimum latency,max latency,mean and standard
 * deviation.
 *
 */
struct LatencyInfo
{
    /*!
     * @brief Parameterized constructor
     *
     * @param[in] name as string: The given channel name
     * @param[in] latencyWarning as int32_t: LatencyWarning threshold
     * @param[in] measurementName as string: Title given to the latency statistics
     * @param[in] type as string: The given channel type
     * @param[in] recvSize as int32_t: The given size
     */
    explicit LatencyInfo(const std::string& name,
                         const int32_t _latencyWarning,
                         std::string _measurementName,
                         const std::string& type,
                         const int32_t recvSize)
        : channelName{name}
        , count(0)
        , minLatency{microseconds_long_double::max()}
        , maxLatency(0)
        , mean(0.0)
        , intermediate(0.0)
        , stdDev(0.0)
        , initialSlowdownInfo()
        , channelSetup{type, recvSize}
        , latencyWarning{static_cast<microseconds_long_double>(_latencyWarning)}
        , numLatencyWarnings(0)
        , measurementName{std::move(_measurementName)}
    {}
    /*!
     * @brief Default constructor
     */
    LatencyInfo() = delete;

    std::string channelName{
        ""}; ///< This variable represents the name of the interface for which latency info shall be collected.
    uint32_t count{0};                                                    ///< Number of data points
    microseconds_long_double minLatency{microseconds_long_double::max()}; ///< This variable represents minimum latency.
    microseconds_long_double maxLatency{0};                               ///< This variable represents Maximum latency
    microseconds_long_double mean{0.0};                                   ///< Mean of data points
    microseconds_long_double intermediate{0.0};                           ///< Helper to calculate statistics
    microseconds_long_double stdDev{0.0};                                 ///< Standard Deviation of data points
    InitSlowdownThresholdInfo
        initialSlowdownInfo{}; ///< Holds the information about the slow down in latency at initial startup.
    ChannelSetupLatencyInfo
        channelSetup; ///< Holds the information about channel setup i.e channel type and receive queue size.
    microseconds_long_double latencyWarning{
        -1}; ///< Warning threshold set for the samples.
             ///< Sample comes under warning threshold, If it's threshold is equals or greater than latency warning.
    uint32_t numLatencyWarnings{0};  ///< Number of samples comes under warning threshold
    std::string measurementName{""}; ///< The name (title) given to the latency statistics.
};

/*!
 * @class LatencyStats
 *
 * @brief This class shall calculate the latency statistics of an interface.
 * \n \b Detailed \b Description:
 * \n This class shall calculate application to application statistics of message latency where latency is the time
 * duration between \n when a message is sent in publish() and when it is received by another task in get().
 *
 * \n The latency statistics contains the following information.
 * \n    minimum latency
 * \n    maximum latency
 * \n    mean latency
 * \n    standard deviation
 *
 * \n This class calculates the latency statistics of an interface over time, using Welford's method.
 * \n Algorithm:
 * \n Initially, mean = 0
 * \n intermediateValue = 0
 * \n FOR each given newValue,
 * \n This algorithm increments the counter by one. i.e  count = count + 1
 * \n Stores the old mean. i,e old mean =  mean
 * \n calculate the new mean. i.e mean = mean + ( given value - mean ) / count.
 * \n calculate the aggregated squared differences from mean.
 * \n i.e  intermediateValue = intermediateValue + ( ( given value - mean ) * (given value - old mean) )
 * \n END FOR
 * \n variance = intermediateValue /(count-1);
 * \n standard deviation = square root(variance)
 */
class LatencyStats
{
public:
    /**
     * @brief The parameterized constructor
     */
    explicit LatencyStats(const std::string& name,
                          int32_t latencyWarning_us,
                          const std::string& measurementName,
                          const std::string& type,
                          int32_t recvSize);
    /**
     * @brief The default constructor
     */
    LatencyStats() = delete;

    /**
     * @brief The destructor
     */
    virtual ~LatencyStats() {}

    LatencyInfo latencyInfo_; /// The struct to hold latency information

    /*!
     * @brief This method shall add a latency value to the statistics.
     * \n This method shall update the statistics of an interface  with the new value using Welford's method.
     * \n \b Operational \b Flow:
     * \n The algorithm proceeds if the given latency value is positive. Otherwise, returns error code as
     * LAT_STATS_NEG_VAL. \n This method shall check if given latency val is longer than the threshold at start up. \n
     * If so, count number of such values and accumulate. \n If the count reaches the max number set then it won't look
     * for any more such values. \n min latency = duration with max possible value \n max latency = 0 \n Initially mean
     * = 0 \n Aggregated squared differences from mean = 0 \n For every given input newValue, \n This algorithm
     * increments the counter by one. i.e  count = count + 1 \n Stores the old mean. i,e old mean =  mean \n calculate
     * the new mean. i.e mean = mean + ( given newValue - mean ) / count. \n calculate the aggregated squared
     * differences from mean. \n i.e  intermediateValue = intermediateValue + ( ( given newValue - mean ) * (given
     * newValue - old mean) ) \n IF the given newValue is less than min latency THEN \n &emsp; min latency gets updated
     * with given newValue. \n END IF \n IF the given newValue is greater than max latency THEN \n &emsp; max latency
     * gets updated with given newValue. \n END IF \n It distributes the given newValue to suitable latency bucket. \n
     * The counter value shall get incremented by one for the corresponding range. \n \n Latency bucket is a map
     * container. Where, Key is a latency range and Value is a counter for the corresponding range. \n The following are
     * the latency bucket ranges. \n 0, 999 \n 1000, 4999 \n 5000, 9999 \n 10000, 49999 \n 50000, 99999 \n 100000,
     * 499999 \n 500000, 999999 \n 1000000,max possible value \n IF the given latency 'val' is greater than or equals to
     * warning level threshold THEN \n &emsp; Increment the LatencyWarnings counter by one. \n END IF
     *
     * @param[in] val as std::chrono::duration : The value to be added
     *
     * @return LatStatsError : LAT_STATS_OK,if the given value is positive.Otherwise, LAT_STATS_NEG_VAL.
     */
    LatStatsError update(microseconds_long_double val);

    /*!
     * @brief This method shall fill a string with all the latency information.
     * \n This method shall return the string with the following latency statistics information.
     * \n data set count ( number of data points)
     * \n duration units in terms of microseconds
     * \n mean latency
     * \n standard deviation
     * \n min latency
     * \n max latency
     * @return string : The string shall be filled with latency information, if data set contains at least one data
     * point. Otherwise, empty string.
     */
    std::string getResults()
    {
        json results;
        results["channel_name"] = latencyInfo_.channelName;

        // Need more than 1 datapoint to calculate standard deviation
        if (latencyInfo_.count > 1) {
            json latencyBuckets;
            for (const auto& bucket : buckets_) {
                std::string maxValStr;
                auto maxVal = bucket.first.end;
                if (maxVal == microseconds_long_double::max().count()) {
                    maxValStr = "max";
                } else {
                    maxValStr = std::to_string(maxVal);
                }

                std::string range = std::to_string(bucket.first.start) + " - " + maxValStr;
                latencyBuckets.emplace(range, bucket.second);
            }

            finalize();

            results["latency_stats"] = {
                {"title", latencyInfo_.measurementName},
                {"msg_count", latencyInfo_.count},
                {"units", "microseconds"},
                {"mean_latency", microseconds_long_double(latencyInfo_.mean).count()},
                {"std_dev", microseconds_long_double(latencyInfo_.stdDev).count()},
                {"min_latency", microseconds_long_double(latencyInfo_.minLatency).count()},
                {"max_latency", microseconds_long_double(latencyInfo_.maxLatency).count()},
                {"found_under_threshold", latencyInfo_.initialSlowdownInfo.foundUnderThreshold},
                {"num_initial_over_threshold", latencyInfo_.initialSlowdownInfo.numInitialOverThreshold},
                {"time_initial_long_time",
                 microseconds_long_double(latencyInfo_.initialSlowdownInfo.timeInitialLongDelayTotal).count()},
                {"threshold_initial_slowdown",
                 microseconds_long_double(latencyInfo_.initialSlowdownInfo.thresholdInitialSlowdown).count()},
                {"latency_warning", microseconds_long_double(latencyInfo_.latencyWarning).count()},
                {"num_latency_warnings", latencyInfo_.numLatencyWarnings},
                {"type", latencyInfo_.channelSetup.channelType},
                {"recvSize", latencyInfo_.channelSetup.recvSize}};
            results.emplace("buckets", latencyBuckets);
        } else {
            results["latency_stats"] = {};
        }
        return results.dump();
    }

private:
    /*!
     * \struct Range
     * \brief This struct represents the latency range.
     */
    struct Range
    {
        /*!
         * @brief Parameterized constructor
         * \n This method shall initialize the start and end values of the Range with the given values.
         *
         */
        Range(long double s, long double e)
            : start(s)
            , end(e)
        {}

        long double start; ///< This variable represents the starting value of latency range.
        long double end;   ///< This variable represents the ending value of latency range.

        /*!
         * @brief Overloaded equals to operator method
         * \n This method shall check if this object and the given object are equal.
         *
         * @param[in] other : Range struct instance
         *
         * @return bool : true, if two objects are equal.Otherwise,false.
         *
         */
        bool operator==(const Range& other) const { return (this->start == other.start) && (this->end == other.end); }

        /*!
         * @brief Overloaded Not equals to operator method
         * \n This method shall check if this object and the given object are not equal.
         *
         * @param[in] other : Range struct instance
         *
         * @return bool : true, if two objects are not equal.Otherwise,false.
         *
         */
        bool operator!=(const Range& other) const { return !(*this == other); }

        // Compare based on the start of the range. Since this is just a utility range for the buckets
        // this is suitable
        /*!
         * @brief Overloaded Less than(<) operator method
         * \n This method shall check if the this object's Range is less than the RHS object's Range.
         * \n It shall compare only the starting value of the Range objects.
         *
         * @param[in] other : Range struct instance(RHS)
         *
         * @return bool : true, if this object is less than RH object. Otherwise,false.
         *
         */
        bool operator<(const Range& other) const { return this->start < other.start; }

        /*!
         * @brief Overloaded Greater than(>) operator method
         * \n This method shall check if the this object's Range is greater than the RHS object's Range.
         * \n It shall compare only the starting value of the Range objects.
         *
         * @param[in] other : Range struct instance(RHS)
         *
         * @return bool : true, if this object is greater than RH object. Otherwise,false.
         *
         */
        bool operator>(const Range& other) const { return !(*this < other); }
    };

    ///<  Latency buckets map is used for more granular tracking of the latency distribution.
    ///<  Key in map is a latency range and
    ///<  Value in map is the count for the corresponding latency range.
    std::map<Range, uint64_t> buckets_{{Range{0, 999}, 0},
                                       {Range{1000, 4999}, 0},
                                       {Range{5000, 9999}, 0},
                                       {Range{10000, 49999}, 0},
                                       {Range{50000, 99999}, 0},
                                       {Range{100000, 499999}, 0},
                                       {Range{500000, 999999}, 0},
                                       {Range{1000000, microseconds_long_double::max().count()}, 0}};

    /**
     * @brief This method shall calculate final latency results.
     * \n This method performs final step of Welford's method and calculates variance and  standard deviation.
     * \n variance = Aggregated squared differences from mean /(data points count - 1)
     * \n standard deviation = square root(variance)
     *
     * @return void
     */
    void finalize()
    {
        if (latencyInfo_.count > 1) { // Double-check to avoid dividing by zero
            microseconds_long_double variance_s2;
            variance_s2 = latencyInfo_.intermediate / (latencyInfo_.count - 1);
            latencyInfo_.stdDev = microseconds_long_double(pow(static_cast<double>(variance_s2.count()), 0.5));
        }
    }
};

/*!
 * @class BufferedLatencyStats
 *
 * @brief This class shall calculate the latency statistics of an interface based on start and end times.
 * \n \b Detailed \b Description:
 * \n This class extends the LatencyStats class, to use a different interface for adding new messages.
 * Start times for multiple messages are added, and stored internally. At the end of an interface's buffer flush cycle,
 * setEndTime is called, which will add (endTime - startTime) for all start times added.
 *
 */
class BufferedLatencyStats : public LatencyStats
{
public:
    /**
     * @brief The Default constructor
     */
    BufferedLatencyStats() = delete;

    /**
     * @brief The parameterized constructor
     *
     * @param[in] name as string: The given channel name
     * @param[in] latencyWarning as int32_t: LatencyWarning threshold
     * @param[in] measurementName as string: Title given to the latency statistics
     * @param[in] type as string: The given channel type
     * @param[in] recvSize as int32_t: The given size
     */
    explicit BufferedLatencyStats(const std::string& name,
                                  const int32_t latencyWarning_us,
                                  const std::string& measurementName,
                                  const std::string& type,
                                  const int32_t recvSize)
        : LatencyStats(name, latencyWarning_us, measurementName, type, recvSize)
    {}

    /*!
     * @brief This method shall record a message sent time, to be later used to calculate latency.
     *
     * @param[in] microseconds_long_double : start time in microseconds since epoch
     *
     * @return LAT_STATS_OK always
     *
     */
    LatStatsError addStartTime(const microseconds_long_double val)
    {
        startTimes.push_back(val);
        return LatStatsError::LAT_STATS_OK;
    }

    /*!
     * @brief This method shall use the end time and previously added start times to update latency stats
     * \n After updating the stats with (end - start) for buffered start times, it will clear all values in
     * the start vector.
     *
     * @param[in] microseconds_long_double : end time in microseconds since epoch
     *
     * @return LAT_STATS_NEG_VAL if any of the latency measurements are negative - OK otherwise.
     *
     */
    LatStatsError setEndTime(const microseconds_long_double end)
    {
        LatStatsError ret{LatStatsError::LAT_STATS_OK};

        for (const microseconds_long_double& start : startTimes) {
            if (end > start) {
                update(end - start);
            } else {
                // Simply return this as a notice - the valid entries will still be recorded.
                ret = LatStatsError::LAT_STATS_NEG_VAL;
            }
        }

        startTimes.clear();

        return ret;
    }

private:
    std::vector<microseconds_long_double>
        startTimes{}; ///< Stores start times, all of which will have the same end time.
};

#endif // LATENCYSTATS_H
