/*!
 ***********************************************************
 * @file LogChannel.h
 * @brief A file defining the LogChannel class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 02/22/2016
 * @remark
 * The timestamp format is replaced with otime.
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * LogChannel.h
 * Tugrul Galatali
 * 12/23/2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */
#ifndef LOGCHANNEL_H
#define LOGCHANNEL_H

#include "Comms/Channel.h"

namespace task {

struct MessageMetadata
{
    ottotime::otimePoint timeOfReceipt{};
    uint32_t publisherId{0};
    uint32_t messageCounter{0};
};

/*!
 * @class LogChannel
 *
 * @brief This is an abstract class inherited from Channel.
 * \n This class is designed to log the channel data.
 * \n For example :
 * \n SQLLogChannel is a concreted derived class for LogChannel, connects to SQLite DB and stores the log data.
 */
class LogChannel : public Channel
{
public:
    /*!
     * @brief Constructor.
     */
    LogChannel(Access mode, const std::string& name)
        : Channel(mode)
        , name_(name)
    {}

    /*!
     * @brief Default virtual destructor.
     */
    virtual ~LogChannel() {}

    /*!
     * @brief Deleted copy and assignment operators
     */
    LogChannel(const LogChannel& newChannel) = delete;
    LogChannel& operator=(const LogChannel&) = delete;

    /*!
     * @brief This method shall find the start and end time of the log message.
     *
     * @param[out] start as otimePoint :  The start time found from the log.
     * @param[out] end as otimePoint:  The end time found from the log.
     *
     * @return ChannelStatus : CS_Success, for the success.
     */
    virtual ChannelStatus getLogExtents(ottotime::otimePoint& start, ottotime::otimePoint& end) = 0;

    /// @brief Read data from the log. Blocking call.
    ///
    /// @param[out] msg as string : The data that was read.
    /// @param[out] pred : The predecessor to the read event.
    /// @param[out] timeOfReceipt :  The time the message was received.
    ///
    /// @return ChannelStatus
    virtual ChannelStatus readWithPred(std::string& msg,
                                       av::latency::Pred& pred,
                                       ottotime::otimePoint& timeOfReceipt) = 0;

    /// @brief Read data from the log. Blocking call.
    ///
    /// @param[out] msg : The data that was read from the log.
    ///
    /// @param[out] timeOfReceipt :  The time the message was received.
    ///
    /// @return ChannelStatus
    inline ChannelStatus read(std::string& msg, ottotime::otimePoint& timeOfReceipt)
    {
        av::latency::Pred p;
        return readWithPred(msg, p, timeOfReceipt);
    }

    /*!
     * @brief This method shall write the log data.
     *
     * @param[in] msg as string: The message to be written.
     * @param[in] timeOfReceipt as otimePoint : The time when the data is logged.
     *
     * @return ChannelStatus
     */
    virtual ChannelStatus write(const std::string& msg, const ottotime::otimePoint& timeOfReceipt) = 0;

    /// @brief Read a message from the log. Blocking.
    ///
    /// @param[out] msg : The message read from the log.
    ///
    /// @param[out] pred : The predecessor to the read event.
    ///
    /// @param[out] metadata : The metadata of the message.
    ///
    /// @return ChannelStatus
    virtual ChannelStatus readWithPred(std::string& msg, av::latency::Pred& pred, MessageMetadata& metadata) = 0;

    /*!
     * @brief This method shall write the log data and its metadata
     *
     * @param[in] msg as string: The message to be written.
     * @param[in] metadata :  The meta data.
     *
     * @return ChannelStatus
     */
    virtual ChannelStatus write(const std::string& msg, const MessageMetadata& metadata) = 0;

    /*!
     * @brief This method allows you to seek to an arbitrary position for a given time.
     *
     * @param[in] seekTime as The timotimePoint : The time given for the look up.
     *
     * @return ChannelStatus
     */
    virtual ChannelStatus seek(const ottotime::otimePoint& seekTime) = 0;

    /*!
     * @brief This method shall return the log's version
     *
     * @return uint32_t containing the log version.
     */
    virtual uint32_t logVersion() const = 0;

    /*!
     * @brief This method shall always return false as log channels doesn't support queues.
     *
     * @param[out] currentDepth : The number of messages waiting in the queue.
     * @param[out] maxDepth : The max size of the message queue.
     *
     * @return bool : false,always.
     *
     */
    bool getQueueProperties(unsigned long& currentDepth, unsigned long& maxDepth) override { return false; }
    const std::string& getName() const override { return name_; }

private:
    const std::string name_;
};

} // namespace task

#endif // LOGCHANNEL_H
