
#ifndef OTTOLINK_OTTOCOMMS_H
#define OTTOLINK_OTTOCOMMS_H

#include <Comms/InterfaceUtils.h>
#include <config/InterfaceParser.h>

class OttoConfig;

/*!
 * @class OttoComms
 * @brief This class shall be responsible for
 * \n setting the env,
 * \n loading the configuration,
 * \n initializing the clock interface,
 * \n creating interfaces,
 * \n connecting to SCS (Simple Communication Server).
 * \n It provides utility methods
 * \n to get the configuration for a given config parameter in comms and
 * \n to check if SCS and Safer Daemons are running.
 */
class OttoComms
{
public:
    OttoComms() = delete;
    /*!
     * @brief Parameterized constructor
     * \n Initialization of OttoComms.
     * \n It shall set the env variables for comms.
     * \n Initializes the configuration
     * \n Connects to SCS (Simple Communication Server) with given name
     * \n IF external clock is enabled THEN
     * \n &emsp; Initialize the SystemClock with the external clock supplied from configuration.
     * \n ELSE
     * \n &emsp; Initialize the SystemClock with the steady clock.
     * \n END IF
     *
     * @throw runtime_error : If  client is not able to connect SCS.
     *
     * @param[in] clientName The name given to client to connect with SCS
     * @param[in] configFile Configuration file
     */
    OttoComms(const std::string& clientName, const std::string& configFile);
    /*!
     * Default virtual destructor.
     */
    virtual ~OttoComms();
    /*!
     * @brief This method shall get configuration as ConfigSection object for a given configuration parameter.
     *
     * @param[in] name name of the configuration parameter
     *
     * @return ConfigSection object
     */
    ConfigSection getSection(const std::string& name) const;
    /*!
     * @brief This method shall return the ottoConfig pointer
     *
     * @return ottoConfig pointer, if exists.Otherwise, nullptr.
     */
    OttoConfig* getOttoConfig() const;

    std::vector<std::string> getChannelsFor(const std::string& module, const std::string& key = "Channels");
    /*!
     * @brief This method shall return config directory path set for communication module
     *
     * @throw runtime_error if failed getting the config directory path set for env variable OTTO_CONFIG_DIR
     *
     * @return string config directory path
     */
    static std::string ottoConfigDir();
    /*!
     * @brief This method shall return the active config directory path set for communication module
     *
     * @throw runtime_error if failed getting the active config directory path set for the env variable
     * ACTIVE_CONFIG_DIR
     *
     * @return string : config directory path
     */
    static std::string activeConfigDir();
    /*!
     * @brief This method shall return true if the environment (SYSTEM_CLOCK_TYPE)is set to use external clock.
     *
     * @return bool : true, if the environment is set to use external clock. otherwise,false.
     */
    static bool useExternallyDrivenClock();

private:
    /*!
     * @brief This method shall set the env variables for communication module.
     *
     * @throw runtime_error if failed setting env variables OTTO_DIR and OTTO_CONFIG_DIR
     *
     * @return void
     */
    static void checkOttoEnvVars();

private:
    std::unique_ptr<OttoConfig> m_ottoConfig; ///< OttoConfig pointer initialized with given configuration file.
};

#endif // OTTOLINK_OTTOCOMMS_H
