#ifndef OTTOLINK_OTTOCONFIG_H
#define OTTOLINK_OTTOCONFIG_H

#include <config/ConfigSection.h>
#include <config/InterfaceParser.h>

#include <optional>
#include <string>
#include <vector>

/*!
 * @class OttoConfig
 * @brief This class shall be responsible for loading the configuration for
 * \n communication module.It provides methods to retrieve configuration data for a
 * \n particular configuration item.
 */
class OttoConfig
{
public:
    /*!
     * @brief Parameterized constructor
     * \n This method shall parse and load the configuration data from the given configuration file.
     * \n And also it maintains two containers for storing read only interfaces ConfigSection objects
     * \n and write only interfaces ConfigSection objects.
     *
     * @throw runtime_error if parse failed for given configuration file.
     *
     * @param[in] configFile The configuration file to be parsed.
     *
     * \sa InterfaceParser::parseFile
     */
    explicit OttoConfig(const std::string& configFile);

    /*!
     * @brief This method shall return the channels listed under Config.parameter
     *
     * @throw runtime_error if the configuration doesn't exist with given parameter (or)
     * 						configuration doesn't exist for configuration named "Config"
     *
     * @return channels list as vector of strings
     */
    std::vector<std::string> getChannels(const std::string& parameter);
    /*!
     * @brief This method shall return the ConfigSection object for the given key
     * \n from read only interfaces configuration container
     *
     * @param[in] key name of the interface for which we want ConfigSection object
     *
     * @return ConfigSection object, if the interface is available in configuration. Otherwise, empty object.
     */
    std::optional<ConfigSection> getReadOnlyInterfaceConfig(const std::string& key) const;
    /*!
     * @brief This method shall return the ConfigSection object for the given key
     * \n from write only interfaces configuration container
     *
     * @param[in] key name of the write only interface for which we want ConfigSection object
     *
     * @return ConfigSection object, if the interface is available in configuration. Otherwise, empty object.
     */
    std::optional<ConfigSection> getWriteOnlyInterfaceConfig(const std::string& key) const;
    /*!
     * @brief This method shall return the ConfigSection object for the given config section name.
     *
     * @throw runtime_error  if configuration is not available for the given config section name
     *
     * @return ConfigSection object if configuration available.
     */
    ConfigSection getSection(const std::string& name) const;

private:
    // NOTE: unfortunately this has to be `mutable` because the `getSection` function in IP is not const

    mutable InterfaceParser m_interfaceParser; ///< Parser for the interfaces configuration
    std::map<std::string, ConfigSection>
        m_readOnlyInterfaces; ///< Map container is used to store read only interfaces configuration
                              ///< where key in map is a read only interface name and value in map is a ConfigSection
                              ///< object
    std::map<std::string, ConfigSection>
        m_writeOnlyInterfaces; ///< Map container is used to store write only interfaces configuration
                               ///< Where key in map is a write only interface name and value in map is a ConfigSection
                               ///< object
    std::string m_configFilePath; ///< Configuration file name
};

#endif // OTTOLINK_OTTOCONFIG_H
