#ifndef OTTOLINK_OTTOREADONLY_CHANNEL_H
#define OTTOLINK_OTTOREADONLY_CHANNEL_H

#include "OttoConfig.h"

#include <Comms/ChannelFactory.h>
#include <ODatum/ODatum.h>
#include <interfaces/baseTypes/Header.h>

#include <atomic>
#include <thread>
/*!
 * @class OttoReadOnlyChannel
 * @brief This class shall be responsible for creating ReadOnly channel with the given configuration.
 * \n Callback method shall be supplied while creating the channel.
 * \n On creation of the channel,a thread is spawned and it continuously check for new data arrival on this channel.
 * \n Callback method shall be invoked with the received message for every new data arrival on the channel.
 */
template<class T>
class OttoReadOnlyChannel
{
public:
    using OnMsgCallback = std::function<void(const T&)>;
    /*!
     * @brief Parameterized constructor
     * \n \b Operational \b Flow:
     * \n It shall read the channel configuration from the supplied configuration.
     * \n It shall create a read only channel based on the channel configuration and connects to IPC.
     * \n It spawns a thread where it keeps waiting for new data arrival on this channel(blocking mode) to read.
     * \n If new data is available then it reads the message which is in serialized form.
     * \n It checks if the received message size greater than or equals to interface header size.
     * \n If so it removes the header from received message and deserializes the received message.
     * \n It invokes the callback method with received message.
     *
     * @param[in] config as ConfigSection object. Given configuration for the ReadOnly Channel.
     * @param[in] callback as function object. Callback function shall be invoked with message when new data arrives on
     * this channel.
     */
    explicit OttoReadOnlyChannel(const ConfigSection& config, OnMsgCallback&& callback)
        : m_callback(std::move(callback))
    {
        ConfigSection localConfig{config}; // as ConfigSection::get is not const
        ConfigSection channelConfig;
        if (localConfig.get(std::string("channel"), channelConfig) == ConfigSection::GS_OK) {
            if ((m_channel = task::ChannelFactory::createChannelManaged(channelConfig, task::Access::Read)) !=
                    nullptr &&
                m_channel->connect() == task::ChannelStatus::CS_Success) {
                name_ = std::string{localConfig.getName()};
                m_thread = std::thread{([this] {
                    while (m_running.load()) {
                        std::string blob;
                        if (m_channel->read(blob, /*blocking*/ true) == task::ChannelStatus::CS_NewData) {
                            static const size_t hdrSize{
                                interfaces::baseTypes::getInterfaceHeaderSize(OTTO_INTERFACE_HEADER_VERSION)};
                            if (m_running.load() && blob.size() >= hdrSize) { // avoid decoding after unblocking
                                blob.erase(0, hdrSize);
                                T message;
                                try {
                                    aptv::deserializeODatum(message, blob);
                                    m_callback(message);
                                } catch (...) {
                                    std::cerr << "Failed to decode: " << name_ << std::endl;
                                }
                            }
                        }
                    }
                })};
                return;
            }
        }
        std::cerr << "Failed to create OttoReadOnlyChannel: " << localConfig.getName() << std::endl;
    }

    OttoReadOnlyChannel() = delete;
    OttoReadOnlyChannel(const OttoReadOnlyChannel&) = delete;
    OttoReadOnlyChannel(OttoReadOnlyChannel&&) = delete;
    OttoReadOnlyChannel& operator=(const OttoReadOnlyChannel&) = delete;
    OttoReadOnlyChannel& operator=(OttoReadOnlyChannel&&) = delete;
    /*!
     * @brief Default destructor
     * \n It shall perform cleanup activities.
     * \n It shall unblock the channel if it is doing blocking read and disconnect the channel from IPC.
     * \n It deallocates the memory created for channel.
     * \n It shall signal the thread function to finish.
     */
    ~OttoReadOnlyChannel()
    {
        m_running.store(false);
        if (m_channel != nullptr) {
            m_channel->unblock();
        }
        if (m_thread.joinable()) {
            m_thread.join();
        }
        if (m_channel != nullptr) {
            m_channel->disconnect();
        }
    }

    const std::string& getName() const { return name_; }

private:
    std::string name_;
    const OnMsgCallback m_callback; ///< Function object passed as callback method while creating the channel.
                                    ///< Callback method shall be invoked when new data arrived on the channel.

    std::unique_ptr<task::Channel> m_channel{}; ///< The Channel pointer pointing to read only channel.
                                                ///<	It reads the messages arrived on the channel.
    std::thread m_thread;                       ///< The thread is for to check new data arrival on the channel.
    std::atomic<bool> m_running{true};          ///< An atomic object of bool type
};

#endif // OTTOLINK_OTTOREADONLY_CHANNEL_H
