#ifndef OTTOLINK_OTTOWRITEONLY_CHANNEL_H
#define OTTOLINK_OTTOWRITEONLY_CHANNEL_H

#include "OttoConfig.h"

#include <Clock/SystemClock.h>
#include <Comms/ChannelFactory.h>
#include <ODatum/ODatum.h>
#include <interfaces/PublishListMessage/PublishListMessage.h>
#include <interfaces/baseTypes/Header.h>

#include <atomic>
#include <thread>
/*!
 * @class OttoWriteOnlyChannel
 * @brief This class shall be responsible for creating for write only channel which connects to IPC.
 * \n It shall convert the given message as generic type to serialized message.
 * \n It inserts header with timestamp in serialized message and publishes the message for consumers.
 */
template<class T>
class OttoWriteOnlyChannel
{
public:
    /*!
     * @brief Parameterized constructor
     * \n It shall read the channel configuration from given configuration.
     * \n It shall create write only channel with mode "write" and connects to IPC.
     * @throw runtime_error if couldn't get channel name from the configuration
     * \n					if unable to create the channel
     * \n					if failed to connect
     * \n 					if failed to read the channel from given configuration.
     *
     * @param[in] config as ConfigSection object. Given configuration
     *
     */
    explicit OttoWriteOnlyChannel(const ConfigSection& config)
    {
        ConfigSection localConfig{config}; // as ConfigSection::get is not const
        ConfigSection channelConfig;
        const std::string publishedListChannelName{"PublishListMessage"};

        if (localConfig.get(std::string("channel"), channelConfig) == ConfigSection::GS_OK) {
            name_ = config.getName();
            if (channelConfig.get(std::string("name"), name_) != ConfigSection::GS_OK) {
                throw std::runtime_error("OttoWriteOnlyChannel: Couldn't get channel name");
            }

            m_channel = task::ChannelFactory::createChannelManaged(channelConfig, task::Access::Write, name_);
            m_publishedListChannel = task::ChannelFactory::createChannelManaged(
                channelConfig, task::Access::Write, publishedListChannelName);

            if (m_channel == nullptr) {
                throw std::runtime_error("OttoWriteOnlyChannel: Couldn't create Channel: " + name_);
            }
            if (m_channel->connect() != task::CS_Success) {
                throw std::runtime_error("OttoWriteOnlyChannel: Channel couldn't connect: " + name_);
            }

            if (m_publishedListChannel == nullptr && m_publishedListChannel->connect() != task::CS_Success) {
                std::cerr << "Failed to get config for OttoWriteOnlyChannel: " << publishedListChannelName << std::endl;
            }
        } else {
            std::cerr << "Failed to get config for OttoWriteOnlyChannel: " << localConfig.getName() << std::endl;
        }

        if (!SystemClock::isInitialized()) {
            SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeSteady);
        }
    }

    OttoWriteOnlyChannel() = delete;
    OttoWriteOnlyChannel(const OttoWriteOnlyChannel&) = delete;
    OttoWriteOnlyChannel(OttoWriteOnlyChannel&&) = delete;
    OttoWriteOnlyChannel& operator=(const OttoWriteOnlyChannel&) = delete;
    OttoWriteOnlyChannel& operator=(OttoWriteOnlyChannel&&) = delete;
    /*!
     * @brief This method shall publish the message of generic type.
     * \n \b Operational \b Flow:
     * \n It shall serialize the given message.
     * \n It inserts header with timestamp in serialized message.
     * \n It publishes the serialized message for consumers.
     *
     * @param[in] msg as generic message type. Given message to be published.
     *
     * @return void
     */
    void publish(const T& msg)
    {
        if (m_channel != nullptr) {
            m_channel->write(serializeData<T>(msg));
        }

        if (m_publishedListChannel != nullptr) {
            av::msg::PublishListMessage plMsg;
            plMsg.channels.emplace_back(name_);

            m_publishedListChannel->write(serializeData<av::msg::PublishListMessage>(plMsg));
        }
    }
    /*!
     * @brief Default destructor.
     * \n It shall disconnect the channel and deallocates the memory created for channel.
     */
    ~OttoWriteOnlyChannel()
    {
        if (m_channel != nullptr) {
            m_channel->disconnect();
        }
    }

    const std::string& getName() const { return name_; }

private:
    template<class U>
    std::string serializeData(const U& msg)
    {
        std::string serializedMsg;
        aptv::serializeODatum(serializedMsg, msg);

        interfaces::baseTypes::InterfaceHeader hdr;
        struct otimeSpec sendTime = SystemClock::getSystemTime().to_otimeSpec();
        interfaces::baseTypes::setHeaderVersion(hdr, OTTO_INTERFACE_HEADER_VERSION);
        interfaces::baseTypes::setSendTimeStamp(hdr, sendTime);
        std::size_t hdrSize = interfaces::baseTypes::getInterfaceHeaderSize(OTTO_INTERFACE_HEADER_VERSION);

        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
        std::string header(reinterpret_cast<char*>(&hdr), hdrSize);
        serializedMsg.insert(0, header);
        return serializedMsg;
    }

    std::string name_;
    std::unique_ptr<task::Channel> m_channel{nullptr}; ///< The write only channel pointer publishes the messages.
    std::unique_ptr<task::Channel> m_publishedListChannel{
        nullptr}; ///< publish PublishListMessage for introspection tool
};

#endif // OTTOLINK_OTTOWRITEONLY_CHANNEL_H
