/*!
 ***********************************************************
 * @file SCChannel.h
 * @brief A file defining the SCChannel class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/06/2016
 * @remark
 * Sending operation is a separate thread now.
 * @n
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/05/2016
 * @remark
 * get() supports the blocking call.
 * @n
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/05/2016
 * @remark
 * Receiving operation is a separate thread now.
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 ***********************************************************
 * SCChannel.h
 * Bryan Salesky
 * 09/02/2006
 *
 * Junsung Kim
 * 12/16/2012
 * A function chmod() is added to support fault-tolerance.
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 ***********************************************************
 */
#ifndef SC_CHANNEL_H
#define SC_CHANNEL_H

#include <atomic>
#include <chrono>
#include <mutex>
#include <queue>
#include <string>
#include <thread>
#include <unistd.h>
#include <unordered_set>

#include <SimpleComms/Client.h>
#include <config/ConfigSection.h>

#include "Comms/Access.h"
#include "Comms/Channel.h"
#include "Comms/LatencyStats.h"

namespace task {
using namespace std::chrono_literals;

/*!
 * @class SCChannel
 *
 * @brief This class shall inherit from Channel to communicate with a task/device/etc. using SimpleComms.
 *
 * \b Detailed \b Description:
 * \n It shall work as Publisher/SubScriber/SubPub.
 * \n It maintains two containers. One for Publisher and another for Subscriber.
 * \n Key in container is a channel name,
 * \n Value in container is a vector of SCChannels.
 * \n There can be multiple SCChannels connected to one channel name to communicate to SimpleComms.
 * \n It spawns a 'send thread' which is common for all the publisher channels.
 * \n Publishers send thread picks the messages from publisher channels send queues and publishes the messages.
 * \n It spawns 'receive thread' which is common for all the subscriber channels.
 * \n Subscribers receive thread pushes the messages into Subscriber channels receive queues, if any new data is
 * available for the channels. \n It spawns a 'salt generator thread', if black channel communication is enabled. \n
 * Salt generator thread publishes salt messages(Black Channel Session Key) in publisher role. \n Publisher converts the
 * normal message into black channel message and applies protect operation on message before publishing the message. \n
 * Subscriber gets salt message from comms and creates receiver key table (Publisher ID and SubScrber Info ). \n
 * Subscriber converts the black channel message into normal message and performs check operation before pushing the
 * messages into subscribers receiving queues.
 *
 */
class SCChannel : public Channel
{
public:
    /*!
     * @brief Parameterized constructor.Initialization of SCChannel.
     *
     * @param[in] comms as SimpleComms::Client pointer :The client pointer to the SimpleComms.
     *
     * @param[in] channelName as string : The name of the current channel.
     *
     * @param[in] mode as enum : The access mode like R/W/RW/Standby.
     *
     * @param[in] recvQueueDepth as uint32_t : The receive queue size.
     *
     * @param[in] resaltInterval The interval to use between black channel resalts. If black channel is not in use, or
     * the interval is zero, resalting is disabled.
     *
     */
    SCChannel(const std::shared_ptr<SimpleComms::Client> comms,
              const std::string channelName,
              const Access mode,
              const uint32_t recvQueueDepth,
              const int32_t latencyWarning_us = -1,
              std::chrono::milliseconds resaltInterval = 0ms);

    /*!
     * @brief Virtual destructor.
     * \n This method shall perform clean up activities.
     * \n IF the messages are not yet read from SCChannel and channel mode is Read or ReadWrite THEN
     * \n &emsp; It shall count the number of bytes to be read from SCChannel's read queue and logs a message.
     * \n &emsp; It clears the messages from SCChannel's read queue.
     * \n END IF
     * \n IF the messages are not yet published from SCChannel and the channel mode is Write or ReadWrite THEN
     * \n &emsp; Log a trace message that SCChannel is not yet published message.
     * \n END IF
     * \n IF Publisher's send thread is running and  Publisher's send channel container is empty THEN
     * \n &emsp; Wait for the Publisher's send thread to exit.
     * \n END IF
     * \n IF Subscriber's receive thread is running and Subscriber's channel container is empty THEN
     * \n &emsp; Wait for the Subscriber's receive thread to exit.
     * \n END IF
     * \n IF blcak channel communication is enabled and sat generator thread is running THEN
     * \n &emsp; wait for the salt generator thread is to exit.
     * \n END IF
     * \n It collects the channels latency statistics.i.e mean,standard deviation,minimum latency and maximum latency.
     */
    virtual ~SCChannel();

    class Builder : public Channel::Builder
    {
    public:
        std::unique_ptr<Channel> build(ConfigSection& channelConfig, Access mode, std::string name) const override;
    };

    /*!
     * @brief This method shall connect SCChannel to the SimpleComms module based on the given mode.
     * \n \b Operational \b Flow:
     * \n This method shall give access to SCChannel as Publisher/Subscriber/SubPub based on the mode.
     * \n It shall add this SCChannel to publishers channel container for a given 'channel name'.
     * \n If the mode is Write/ReadWrite/Standby, It spawns a send thread which is common for all the publishers.
     * \n The send thread method sends messages to SimComms system from all the publishers if any messages present in
     * their send queues. \n If the mode is Read/ReadWrite, It spawns a receive thread which is common for all the
     * subscribers. \n The read thread shall read the messages from SimComms and push into Subscriber channels receive
     * queue. \n \n The Publisher/Subscriber channel's container is map container where \n Key in the map is a channel
     * name, \n Value in the map is a set container of SCChannel objects. \n There can be multiple SCChannels connected
     * for a given channel name to communicate with SimpleComms.
     *
     * @return ChannelStatus : CS_Success for the successful connection.Otherwise CS_Error.
     *
     * \sa SCChannel:addAccess
     *
     */
    virtual ChannelStatus connect(void);

    /*!
     * @brief This method shall disconnect this SCChannel from SimpleComms.
     * \n \b Operational \b Flow:
     * \n It shall remove the access (Publisher/Subscriber/SubPub))to this SCChannel.
     * \n It shall remove this SCChannel from Publisher's/Subscriber's channel container based on the mode.
     * \n It shall send undeclare message to SimComms if this is the only SCChannel for a given channel remained in
     * Publisher's/Subscriber's channel container.. \n \n The Publisher/Subscriber channel's container is a map
     * container where \n Key in the map is a channel name, \n Value in the  map is a set container of SCChannel
     * objects. \n There can be multiple SCChannels connected for a given channel name to communicate with SimpleComms.
     *
     * @return ChannelStatus CS_Success for the successful disconnection.Otherwise CS_Error.
     *
     * \sa SCChannel::removeAccess
     *
     */
    virtual ChannelStatus disconnect(void);

    /// @brief Read a message from the SimpleComms channel.
    ///
    /// @param[out] msg as string : The message read from the
    /// SimpleComms channel.
    ///
    /// @param[out] pred : The predecessor to the read event.
    ///
    /// @param[in] blocking : if true, the call will block until a
    /// message is available.
    ///
    /// @return ChannelStatus : CS_Success for the successful execution.
    ChannelStatus readWithPred(std::string& msg, av::latency::Pred& pred, bool blocking = false) override;

    /*!
     * @brief Publish the given message to the SimpleComms module.
     *
     * @param[in] msg: The message to be published
     *
     * @param[in] predecessors: The predecessors of this write event,
     * to be included if the event gets logged.
     *
     * @return ChannelStatus
     * \n CS_Success If the mode is Write/ReadWrite and pushed the message into channel's send queue.(or) mode is
     * Standby \n CS_Error for all other cases.
     *
     */
    ChannelStatus writeV(const std::string& msg, std::vector<av::latency::Pred>&& predecessors) override;
    /*!
     * @brief This method shall change the mode of the channel.
     *
     * @param[in] mode as Access enum : The new mode to be changed (R/W/RW/Standby)
     *
     * @return ChannelStatus : CS_Success, always.
     *
     */
    virtual ChannelStatus chmod(const Access mode);

    /*!
     * @brief This method shall retrieve receive queue properties of this channel.
     *
     * @param[out] currentDepth : The number of messages waiting in the receive queue of the channel.
     *
     * @param[out] maxDepth : The max size of the receive queue.
     *
     * @return bool : true, always.
     *
     */
    virtual bool getQueueProperties(unsigned long& currentDepth, unsigned long& maxDepth);

    /*!
     * @brief This method shall unblock the channel if it is doing a blocking read.
     * \n This method shall send a signal to channel's read method to unblock if
     * \n read method is in blocked state and waiting for the messages to be available in
     * \n channel's receive queue. So read method shall get unblocked and read if any
     * \n messages from the channel's receive queue.
     *
     *  @return void.
     */
    virtual void unblock() override;
    /*!
     * @brief This method shall make the channel's receive queue empty if the mode is Read/ReadWrite.
     *
     * @return void.
     */
    virtual void clear() override;

    const std::string& getName() const override { return channelName_; }

protected:
    using OpenChannelMap = std::map<std::string, std::unordered_set<SCChannel*>>;
    /*!
     * @brief This method shall give access [Publisher/Subscriber/SubPub] to this SCChannel based on the given mode of
     * the channel. \n Channel Access - Publisher for Write mode \n Channel Access - Subscriber for Read mode \n Channel
     * Access - SubPub for ReadWrite/Standby mode \n Two map containers shall be maintained.One for Publishers and
     * another for subscribers where \n In each container, Key in the container is a channel name, \n Value in the
     * container is a vector of SCChannels. \n There can be multiple SCChannels connected for a given channel name to
     * communicate to SimpleComms. \n If mode is Standby/ReadWrite then the SCChannel shall be present in both the
     * containers. \n \b Operational \b Flow: \n IF the channel mode is Write/ReadWrite/Standby THEN \n &emsp; It
     * shall check if the channel name exists is publisher's container. \n &emsp; If it doesn't exist then it updates
     * the message channel access type as Publisher. \n END IF \n IF the channel mode is Read/ReadWrite THEN \n &emsp;
     * It shall check if the channel name exists in the subscriber's container. \n &emsp; If it doesn't exist then it
     * updates the message channel access type as Subscriber. \n END IF \n It shall send declare message (channel name,
     * message channel access type,buffer size) to SimComms (SCS) only once for one channel name. \n It shall not send
     * declare message for every SCChannel associate with same channel name. \n IF the channel mode is
     * Write/ReadWrite/Standby THEN \n &emsp; If the response from  SimComms is success, push this SCChannel client in
     * publishers container for the given channel name. \n &emsp; It spawns a 'send thread' which is common for all the
     * SCChannels \n &emsp; Every SCChannel shall have their own send queue. \n &emsp; This thread shall be responsible
     * for sending the messages stored in each SCChannel's send queue to SimComms. \n &emsp; If black channel
     * communication is enabled, spawns a 'salt generator thread'. \n &emsp; salt generator thread generates salts
     * periodically and publishes to SimComms. \n END IF \n IF the channel mode is Read/ReadWrite/Standby THEN \n
     * &emsp; If the response from  SimComms server is success, push this SCChannel client in subscribers container for
     * the given channel name. \n &emsp; It spawns a 'receive thread' which is common for all the SCChannels. \n &emsp;
     * The receive thread is responsible for pushing the messages into SCChannel's receive queues. \n END IF \n
     * @return ChannelStatus : CS_Success, if channel name associated with this SCChannel client is registered with
     * SimComms. \n 					   Otherwise, CS_Error
     */
    ChannelStatus addAccess(const Access mode);
    /*!
     * @brief This method shall remove the access[Publisher/Subscriber/SubPub] for this SCChannel based on the mode.
     * \n \b Operational \b Flow:
     * \n IF this SCChannel is connected THEN
     * \n &emsp; IF the mode is Write/ReadWrite/Standby THEN
     * \n &emsp;&emsp; Remove this SCChannel from the publishers map container for the associated channel name.
     * \n &emsp;&emsp; If this SCChannel is the last one(remained) in publishers map container for the associated
     * channel name, \n &emsp;&emsp; update message channel type as "Publisher".Otherwise, message channel type as
     * "Unlisted". \n &emsp;END IF \n &emsp; IF the mode is Read/ReadWrite THEN \n &emsp;&emsp; Remove this SCChannel
     * from the subscribers map container for the associated channel name. \n &emsp;&emsp; If this SCChannel is the last
     * one(remained) in subscribers map container for the associated channel name, \n &emsp;&emsp; update message
     * channel type as "Subscriber".Otherwise, message channel type as "Unlisted". \n &emsp;END IF \n &emsp; IF message
     * channel type is not "Unlisted" \n &emsp;&emsp; Send undeclare message to SimComms with channel name and message
     * channel type. \n &emsp;&emsp; Undeclare message shall be sent to SimComms only once for a channel name. \n
     * &emsp;&emsp; i.e This SCChannel is the only one remained for a given channel in Publisher/Subscriber's container
     * to be removed. \n &emsp;END IF \n END IF
     *
     * @return ChannelStatus
     * \n CS_Success, If SCChannel is already removed access(not connected) (or)
     * \n			 If this SCChannel is not the last one to be removed access (or)
     * \n 			 If send request to SimComms system with undeclare message is success.
     * \n CS_Error,   If send request to SimComms system with undeclare message is failed.
     */
    ChannelStatus removeAccess(const Access mode);

    /*!
     * @brief This method shall be a helper method for subscriber's receive thread startRecvThread.
     * \n The receive thread is common for all the subscriber channels.
     * \n If mode is R/RW/Standby, this method shall keep polling for any new data from SimComms to all subscriber
     * channels. \n If any new data is available, then it updates the particular subscriber's receive queue with the
     * received message. \n It shall stop subscribers receive thread if subscribers doesn't exist.
     *
     * \sa SCChannel::updateMessageQueues
     * \sa SCChannel::startRecvThread
     *
     * @return void
     */
    static void recvThread(void);

    /*!
     * @brief This shall be a helper method for publishers send thread method startSendThread.
     * \n IF the mode is Read/ReadWrite/Standby and send queue is updated with message for any of the SCChannel THEN
     * \n &emsp; Sends the messages to SimComms from all the Publisher's send queue if messages are available.
     * \n END IF
     * \n If publishers doesn't exist, it shall stop the publisher's send thread .
     *
     * \sa SCChannel::startSendThread
     *
     * @return void
     */
    static void sendThread(void);

    /*!
     * @brief This method shall encapsulate the pull and multiple delivery of messages from comms system to the channel.
     * \n This method shall be part of the Subscriber's receive thread method and gets executed for every defined
     * interval. \n This method shall be responsible for updating the every channel's( in subsriber's channel container)
     * receive queue for \n any new data from comms system. \n \b Operational \b Flow: \n IF the mode is
     * Read/ReadWrite/Standby and new message data available from comms THEN \n &emsp; FOR for every channel(key) in
     * Subscriber's map container \n &emsp; &emsp; Keep polling for the new data. \n &emsp; &emsp; IF new data is
     * available THEN \n &emsp; &emsp; 	If black channel communication is enabled, process the black channel message. \n
     * &emsp; &emsp; else \n &emsp; &emsp; 	Retrieve the set of SCChannel's associated with the channel name in
     * subsriber's channel container \n &emsp; &emsp; 	and push the new message data into all the SCChannel's receive
     * queue. \n &emsp; &emsp; 	The data from receive queue shall be retrieved by a read method call. \n &emsp; &emsp;
     * END IF \n &emsp; END FOR \n END IF
     *
     * @return ChannelStatus : CS_Success, if success, CS_Error, if not in a readable mode or
     * for various errors from comms
     *
     * \sa SCChannel::updateRecvQueueForChannels
     */
    ChannelStatus updateMessageQueues(void);

    /*!
     * @brief This method shall encapsulate the multiple delivery of the messages to
     * \n the underlying comms system(SimComms module).
     * \n \b Operational \b Flow:
     * \n IF the mode is Write/ReadWrite/Standby and send queue is updated with message for any of the SCChannel THEN
     * \n Publish the messages to SimComms from all the Publisher's send queue if any messages.
     * \n Clears the the message from send queue after each publish.
     * \n END IF
     *
     * @return ChannelStatus : CS_Success, If publish success. CS_Error, if this channel is not in a writable mode.
     *
     */
    ChannelStatus updateSendQueues(void);

    /*!
     * @brief This method shall be a helper method for removeAccess.
     * \n It shall remove this SCChannel from Publisher/Subscriber channel container.
     * \n If this SCChannel is the last one remained for the connected channel name in Publisher/Subscriber channel
     * container, \n returns true to indicate sending undeclare message to SimComms necessary. \n There can be multiple
     * SCChannels connected one channel name to communicate to SimpleComms. \n If this SCChannel is not the last one, no
     * need to send  undeclare message to SimComms.
     *
     * @param[in] ocm as OpenChannelMap container : The given container from which SCChannel shall be deleted for a
     * given channelName
     * @param[in] channelName as string : The given channelName
     * @param[in] channel as SCChannel pointer : The SCChannel (SimComms Client pointer) to be deleted from
     * OpenChannelMap container.
     *
     * @return bool
     * \n 'true' to indicate that removeAccess should send an 'undeclare' message to the scs
     * \n 'false' indicates either
     * \n  channel' is not connected to 'channelName' for the given access type, so undeclaring would be unnecessary
     * (or) \n  Other SCChannels are connected to 'channelName' for the given access type, so undeclaring would be
     * incorrect.
     *
     * \sa SCChannel::removeAccess
     */
    bool eraseIfPresentReturnIfErasedLast(OpenChannelMap& ocm,
                                          const std::string& channelName,
                                          const SCChannel* const channel) const;

    /*!
     * @brief This method shall be a helper method to provide the write operation based on the msg type..
     * \n \b Operational \b Flow:
     * \n If black channel communication is enabled then the normal message shall be converted as black channel message.
     * \n IF the mode is Write or ReadWrite THEN
     * \n &emsp; IF black channel communication is enabled and message is a normal message THEN
     * \n &emsp; &emsp; Apply black channel publisher's Protect operation on the converted message.
     * \n &emsp; &emsp; Protect operation is not required in case of salt message.
     * \n &emsp; END IF
     * \n &emsp; If Protect operation is success,  push the message into SCChannel's send queue.
     * \n &emsp; If black channel communication is not enabled, push the message without any conversion into SCChannel's
     * send queue. \n &emsp; Set a indicator so that the message from send queue is picked up by Publisher's send thread
     * and publishes to SimComms. \n ELSE IF mode is Standby THEN \n &emsp; It shall not update SCchannel's send
     * queue. \n END IF
     *
     * @param[in] msg: The message to be written to SimpleComms module.
     * @param[in] type: type of the message - MESSAGE/METADATA
     * @param[in] predecessors: The predecessors for this write event.
     *
     * @return ChannelStatus
     * \n CS_Success If the mode is Write/ReadWrite and pushed the message into channel's send queue.(or) mode is
     * Standby \n CS_Error for all other cases.
     *
     * \sa SCChannel::updateSendQueues and SCChannel::sendThread
     *
     */
    ChannelStatus writeTyped(const std::string& msg, std::vector<av::latency::Pred>&& predecessors);

    /**
     * @brief Return an estimate of the number of messages that should be sent before a salt update concludes.
     *
     * @return Estimated message count.
     */
    std::uint32_t estimateSaltUpdateValidMessageCount();

private:
    /*!
     * @brief This method shall add a given msg to all the connected SCChannels receiving queues of the specified
     * channel. \n Subscriber's channel container is a map container where \n Key in the container is a channel name \n
     * Value in the container is set of SCChannels associated with the corresponding channel name. \n \b Operational \b
     * Flow: \n WHILE every SCChannel in input channels set \n &emsp; IF the SCChannel exist THEN \n &emsp;&emsp; Push
     * the message into SCChannel's receiving queue. \n &emsp;&emsp; If the  receiving queue size is greater than the
     * configured receiving queue depth then \n &emsp;&emsp; Remove the oldest message from the receiving queue. \n
     * &emsp;&emsp; Raise a signal that new data is available in the receiving queue. \n &emsp;&emsp; So read method
     * gets signaled if it is waiting for new data  and retrieve the message from receiving queue. \n &emsp;END IF \n
     * END WHILE
     * @param[in] msg as string : A normal message received
     * @param[in] channels as set of SCChannel pointers
     *
     * @return void
     *
     */
    void updateRecvQueueForChannels(const std::string& msg,
                                    const int64_t sendTimeMicroseconds,
                                    const std::unordered_set<SCChannel*>& channels) const;

    /*!
     * @brief This method shall insert a start time into the BufferedLatencyStats object of each channel in the set.
     * @param[in] sendTime as microseconds_long_double : The "sent" timestamp in microseconds since epoch
     * @param[in] channels as set of SCChannel pointers
     *
     * @return void
     *
     */
    void addLatencyStartTime(const microseconds_long_double& sendTime, const std::unordered_set<SCChannel*>& channels);

    /*!
     * @brief This method shall update the BufferedLatencyStats for each channel in the container by setting the end
     * time for the previously queued start times added to the BufferedLatencyStats container to the current time.
     *
     * @param[in] channels as a map of a set of SCChannel pointers.
     *
     * @return void
     *
     */
    void updateLatencyEndTime(const std::map<std::string, std::unordered_set<SCChannel*>>& channels);

    /**
     * @brief This method will flush all messages from the channel's send queue and send them over SimpleComms.
     *
     * @return ChannelStatus : CS_Success if all messages sent successfully
     */
    ChannelStatus flushSendQueue();

protected:
    static pthread_mutex_t channelsMutex_; ///< The lock is for subscriber channel's receivedQueue_.
                                           ///< The receivedQueue_ shall get updated by receive thread as and when there
                                           ///< is a new data from SCS.
    static pthread_mutex_t sendQMutex_;    ///< This lock is for channel's sendQueue_.
                                           ///< The sendQueue_ shall be read by send thread to publish as and when
                                           ///< SCChannel writes the messages into send queue.
private:
    std::shared_ptr<SimpleComms::Client>
        comms_; ///< SimpleComms::Client pointer is connected to SCS which is shared by all SCChannel instances.
                ///< This connection pointer shall be used for send/receive messages  to/from SCS.
    const std::string
        channelName_; ///< Name of the channel is supplied from the configuration while creating the channel.
                      ///< There can be multiple SCChannel instances associated with one channel name.

    const uint32_t
        recvQueueDepth_; //< The max size limit supplied from configuration set on the channel's receive queue which
                         ///< gets filled up with the messages from SCS.
                         ///< 0 means infinite
    const int32_t latencyWarning_us_; ///< Threshold for warning of latency, -1 = no warning
    bool readFrom_;  ///< This variable represents a flag to keep track of whether read operation is called on channel.
    bool writtenTo_; ///< This variable represents a flag to keep track of whether write operation is called on channel.

    bool connected_; ///< This variable represents a flag to check whether channel is connected to SCS.

    static pthread_mutex_t declareMutex_; ///< The lock shall be used for sending declare/undeclare process to SCS
                                          ///< during connect/disconnect of the call.

    static OpenChannelMap channels_; ///< The map container is for storing subscriber channels.
                                     ///< Key in map is a channel name.
                                     ///< Value in map is a set container of SCChannel pointers for SimpleComms.

    std::queue<std::pair<std::string, ottotime::otimePoint>>
        receivedQueue_; ///< The subscriber channel's receive queue.
                        ///< It shall get updated(write) by receive thread and SChannel reads the messages from queue.

    static pthread_t
        recvThread_; ///< The receive thread shall get started when channel is connected to SCS.
                     ///< This is a common thread shared by all subscriber channels in channels_.
                     ///< It shall push the messages received on the channel into the channel's receive queues.

    static bool recvThreadFlag_; ///< This variable represents a flag to indicate whether the receive thread is started.
                                 ///< Flag becomes true when the receive thread gets started.
    /*!
     * @brief This shall be a thread method started for Subscriber's channel connected to comms system.
     * \n This is a common thread for all Subscriber's.
     * \n It shall be responsible for reading the messages from comms and pushing the messages into Subscriber's
     * receiving queues.
     *
     * @return nullptr if there are no Subscriber's.
     *
     * \sa SCChannel::recvThread
     */
    static void* startRecvThread(void* const)
    {
        recvThread();
        return nullptr;
    }

    pthread_cond_t recvSignal_; ///< This is a condition variable for receivedQueue_.
                                ///< This signal shall be raised if any new data is pushed on to channels receive queue.
                                ///< And also, this signal shall unblock read method if it is doing a blocking read.

    static OpenChannelMap sendChannels_; ///< The map container for storing publisher's channels.
                                         ///< Key in container is channel name.
                                         ///< Value in container is set container of SCChannel pointers for SimpleComms.

    using OutboundMsg = std::pair<std::string, std::vector<av::latency::Pred>>;
    std::queue<OutboundMsg> sendQueue_; /// The Publisher channel's send queue.  The send
                                        /// message queue shall be read by send thread for
                                        /// publish and SCChannel writes messages into send
                                        /// message queue.

    static pthread_t
        sendThread_; ///< The send thread shall get started when channel is connected to SCS.
                     ///< This is a common thread shared by all the Publisher channels in sendChannels_.
                     ///< It shall pick up the messages from channel's sendQueue_ and publishes to the comms system.

    static bool sendThreadFlag_; ///< This variable represents a flag to indicate whether the send thread is started.
                                 ///< If true,the send thread is started.
    /*!
     * @brief This shall be a thread method started for publishers when channel is connected to comms system.
     * \n This is a common thread for all publishers.
     * \n It shall be responsible for sending the messages to comms from all the Publisher's send queues.
     *
     * @return nullptr if there are no publishers.
     *
     * \sa SCChannel::sendThread
     *
     */
    static void* startSendThread(void* const)
    {
        sendThread();
        return nullptr;
    }

    static bool
        somethingToSend_; ///< This variable represents a flag to indicate whether the message is pushed to sendQueue_.
                          ///<  true when message is pushed to sendQueue_.

    std::chrono::milliseconds resaltInterval_{0ms}; ///< the interval between black channel resalts
    std::string channelId_{""};                     ///< the uuid of the channel
    BufferedLatencyStats latencyStats_; ///< This structure holds latency statistics information of the channel.
    bool sendQOpen_{false}; ///< Protected by sendQMutex_ to stop threads from writing to disconnected channel
    bool recvQOpen_{false}; ///< Protected by channelsMutex_ to stop threads from reading from disconnected channel
};

} // namespace task

#endif // SC_CHANNEL_H
