/*!
 ***********************************************************
 * @file SQLLogChannel.h
 * @brief A file defining the SQLLogChannel class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/05/2016
 * @remark
 * get() supports the blocking call.
 * @n
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 02/22/2016
 * @remark
 * Timestamps are replaced with otimePoint.
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

#ifndef SQLLOGCHANNEL_H
#define SQLLOGCHANNEL_H

#include "Comms/Access.h"
#include "Comms/LogChannel.h"
#include <Comms/LatencyStats.h>
#include <Statistics/CombinationStatistic.hpp>
#include <config/ConfigSection.h>
#include <sqlite3.h>

#include <atomic>
#include <optional>
#include <queue>

namespace task {
/*!
 * \struct SQLLogChannelConfig
 * @brief structure to hold db file name.
 */
struct SQLLogChannelConfig
{
    std::string logFileName{"defaultlog.db"}; ///< db file name
    bool enableSQLiteTracing{false};          /// < Determine whether to enable SQL Tracing
    int64_t sqlProfileWarning_us{-1};         ///< Warning threshold in microseconds for SQLite3 profiling
};

/*!
 * @class SQLLogChannel
 *
 * @brief This class shall be a concrete derived class inherited from channel.
 * \n It shall write channel data to db files and shall read channel data from db files.
 * \n The channel data consists of log data and timestamp corresponding to it's data.
 * \n It shall find start and end time of log messages data in db.
 *
 */
class SQLLogChannel final : public LogChannel
{
public:
    /*!
     * @brief Parameterized constructor.
     * \n It shall initialize the SQLLogChannel.
     * \n It shall get the db file name (logFileName) from the passed channel configuration.
     *
     * @param[in] channelConfig The channel configuration.
     *
     * @param[in] mode Access mode to connect sqlite db
     *
     */
    SQLLogChannel(ConfigSection& channelConfig, const Access mode, std::optional<std::string> nameOpt = std::nullopt);

    /*!
     * @brief Parameterized constructor.
     * \n It shall initialize the SQLLogChannel.
     *
     * @param[in] channelConfig The channel configuration.
     *
     * @param[in] mode Access mode to connect sqlite db
     *
     */
    SQLLogChannel(const SQLLogChannelConfig& config, const Access mode);

    /*!
     * @brief Copy constructor of the SQLLogChannel class.
     *
     * @param channel The channel to be copied from.
     *
     */
    SQLLogChannel(const SQLLogChannel& channel) = delete;

    /*!
     * @brief Move constructor of the SQLLogChannel class.
     *
     * @param channel The channel to be copied from.
     *
     */
    SQLLogChannel(const SQLLogChannel&& channel) = delete;

    /*!
     * @brief Move constructor of the SQLLogChannel class.
     *
     * @param channel The channel to be copied from.
     *
     */
    SQLLogChannel(SQLLogChannel&& channel);

    class Builder : public Channel::Builder
    {
    public:
        std::unique_ptr<Channel> build(ConfigSection& channelConfig, Access mode, std::string name) const override;
    };

    /*!
     * @brief Copy assignment operator for the SQLLogChannel class.
     *
     * @param channel The channel to be copied from.
     *
     */
    void operator=(const SQLLogChannel& channel) = delete;

    /*!
     * @brief Move assignment operator for the SQLLogChannel class.
     *
     * @param channel The channel to be copied from.
     *
     */
    void operator=(const SQLLogChannel&& channel) = delete;

    /*!
     * @brief Move assignment operator for the SQLLogChannel class.
     *
     * @param channel The channel to be copied from.
     *
     */
    void operator=(SQLLogChannel&& channel);

    /*!
     * @brief Destructor. Simply calls disconnect() to end the SQL
     * server connection.
     *
     */
    virtual ~SQLLogChannel();

    /*!
     * @brief This method shall connect channel to sqlite db based on the given mode
     *
     * \b Operational \b Flow:
     * \n IF the mode is write THEN
     * \n &emsp; IF db file(log file) exists THEN
     * \n &emsp; &emsp; it shall open a database connection with ReadWrite mode.
     * \n &emsp; ELSE
     * \n &emsp; &emsp; It creates a db file and sets the default user version for db
     * \n &emsp; &emsp; It creates a table named OTTOLOG.
     * \n ELSE IF mode is read THEN
     * \n &emsp; &emsp; It shall open a database connection with Read mode.
     * \n &emsp; &emsp; It reads user version & query data from OTTOLOG table to see read access to db is OK.
     * \n ELSE
     * \n &emsp; log a trace message as invalid mode.
     * \n END IF
     *
     * @return CS_Success If db open success with read/write mode.
     * \n CS_Error If db file name is empty
     * \n (or) Unable to create db file
     * \n (or) Unable to set user version for db
     * \n (or) Unable to create/drop  table.
     * \n (or) Unable to create table.
     * \n (or) Unable to read log db version
     * \n (or) Unable to get data from OTTOLOG table
     * \n (or) Invalid mode
     * \n (or) Any db library call failure
     * \n Specific error message shall be logged for failures.
     */
    ChannelStatus connect() override;

    /*!
     * @brief This method shall close the connection to sqlite db.
     * \n It shall clear pending transactions before closing the db.
     *
     * @return CS_Success for the successful execution.Otherwise CS_Error
     *
     */
    ChannelStatus disconnect() override;

    /// @brief Read a message from the log db. Blocking.
    ///
    /// @param[out] msg : The message read from the db log.
    ///
    /// @param[out] pred : The predecessor to the read event.
    ///
    /// @param[in] blocking : if true, the call will block until a
    /// message is available.
    ///
    /// @return ChannelStatus
    ChannelStatus readWithPred(std::string& msg, av::latency::Pred& pred, bool blocking = false) override;

    /// @brief Reads the data and return the corresponging time.
    ///
    /// @param msg [out] Message to be read.
    ///
    /// @param[out] pred : The predecessor to the read event.
    ///
    /// @param timeOfReceipt [out] The corresponding received time.
    ///
    /// @return CS_NewData for the new data found.
    ChannelStatus readWithPred(std::string& msg, av::latency::Pred& pred, ottotime::otimePoint& timeOfReceipt) override
    {
        MessageMetadata meta;
        const ChannelStatus cs = readWithPred(msg, pred, meta);
        if (cs == CS_NewData) {
            timeOfReceipt = meta.timeOfReceipt;
        }
        return cs;
    }

    ChannelStatus read(std::string& msg, ottotime::otimePoint& timeOfReceipt)
    {
        av::latency::Pred pred{};
        return readWithPred(msg, pred, timeOfReceipt);
    }

    /*!
     * @brief This method shall write the messages to the db files.
     * \n It shall write send timestamp and data to db from the given message.
     *
     * @param[in] msg The message to be written
     *
     * @param[in] predecessors: The predecessors of this write event,
     * to be included if the event gets logged.
     *
     * @return CS_Success for the successful execution
     *
     */
    ChannelStatus writeV(const std::string& msg, std::vector<av::latency::Pred>&& predecessors) override;

    /*!
     * @brief Writes the data and its received time.
     *
     * @param msg Message to be written.
     *
     * @param timeOfReceipt Timestamp when the message arrived.
     *
     * @return CS_Success for the successful execution
     *
     */
    ChannelStatus write(const std::string& msg, const ottotime::otimePoint& timeOfReceipt) override;

    /*!
     * @brief This method is not relevant as chmod does not affect the db files.
     *
     * @param[in] mode The new mode to be changed.
     *
     * @return CS_Success.
     *
     */
    ChannelStatus chmod(const task::Access acc) override { return CS_Success; };

    /*!
     * @brief This method shall find the start and end time of the log message in db
     *
     * @param[out] start The start time found from the log.
     *
     * @param[out] end The end time found from the log.
     *
     * @return CS_Success for the success.CS_Error if no data in db or db library call failures.
     *
     */
    ChannelStatus getLogExtents(ottotime::otimePoint& start, ottotime::otimePoint& end) override;

    /*!
     * @brief This method shall read the data and it's corresponding time from db.
     * \n It shall read the time,data columns from OTTOLOG table for the current row.
     *
     * @param[out] msg Message to be read.
     *
     * @param[in] metadata The message metadata to be read.
     *
     * @return CS_NewData for the new data found. CS_Error if channel connection mode is not read or db library call
     * failures.
     *
     */
    ChannelStatus readWithPred(std::string& msg, av::latency::Pred& pred, MessageMetadata& metadata);

    /*!
     * @brief This method shall write the data and it's received time to OTTOLOG table in db.
     *
     * @param[in] msg Message to be written.
     *
     * @param[in] metadata The message metadata to be written.
     *
     * @return CS_Success for the success.CS_Error if channel connection mode is not write or db library call failures.
     *
     */
    ChannelStatus write(const std::string& msg, const MessageMetadata& metadata) override;

    /*!
     * @brief This method shall update the current row ID for the given time.
     * \n It finds a row ID in table OTTOLOG where time column's value is greater than or equals to given time.
     *
     * @param[in] seekTime The time given for the look up.
     *
     * @return CS_Success for the success.CS_Error if channel connection mode is not read or db library call failures.
     *
     */
    ChannelStatus seek(const ottotime::otimePoint& seekTime) override;

    /*!
     * @brief This method shall return the log's version
     *
     * @return uint32_t : The log version
     */
    inline uint32_t logVersion() const override { return logVersion_; }

    /**
     * @brief This method shall clear the state of the channel.
     *
     * @return void
     */
    void clear() override{};

    /**
     * @brief This method shall update the name of the log file(db instance file) with given file.
     * \n This API is meant for testing purposes until a more abstract test double can be created.
     *
     * @param[in] logFileName Name to change the log file name to.
     *
     * @return void
     */
    void updateLogFileName(const std::string& logFileName);

    /**
     * @brief This method causes the channel to take its current count of data written and update internal throughput
     * statistics.
     *
     * @return void
     */
    void sampleThroughputMeasurements() override;

    /**
     * @brief Get throughput measurements
     *
     * \note Returns empty json object unless overridden by the derived class.
     *
     * @return void
     */
    nlohmann::json getThroughputMeasurements();

private:
    /*!
     * @brief This method shall convert db key to otime.
     *
     * @param[in] key Given key as double to be converted
     *
     * @param[out] time Output as otimePoint object
     *
     * @return CS_Success for the successful execution.CS_Error if given key is 0
     *
     */
    ChannelStatus keyTootime(const double& key, ottotime::otimePoint& time) const;

    /*!
     * @brief This method shall convert otime to db key as double for sqlite.
     *
     * @param[in] time To be converted.
     *
     * @param[out] key Output as double.
     *
     * @return CS_Success for the successful execution.
     *
     */
    static ChannelStatus otimeToKey(const ottotime::otimePoint& time, double& key);

private:
    /// @brief Class to buffer messages so they can be written in batches to reduce sqlite overhead.
    /// @note Will flush buffer after either reaching a given data threshold, or timeout (whichever is earlier)
    class WriteBuffer
    {
    public:
        /// @brief Construct a buffer and specify how much to wait for before writeout
        /// @param maxVolume : How large to make the queue
        /// @param writeoutTimeout : Time between writeouts
        WriteBuffer(size_t maxVolume, double writeoutTimeout)
            : maxVolume_(maxVolume)
            , writeoutTimeout_(writeoutTimeout)
        {}

        /// @brief Delete the copy constructor.
        WriteBuffer(const WriteBuffer&) = delete;

        /// @brief Destructor.
        ~WriteBuffer();

        /// @brief Initialize the object
        /// @param logVersion : Specify the version, needed to know the db schema
        /// @param db : Pointer to the (already initialized) sqlite3 db
        bool connect(uint32_t logVersion, sqlite3* db);

        /// @brief write a new message to the buffer.
        /// @param msg : The data from the message
        /// @param metadata : The metadata for the message
        void add(const std::string& msg, const MessageMetadata& metadata);

        /// @brief Reset the buffer.
        /// @note Forces write-out, finalizes statements, and invalidates db ptr. Does *not* close db.
        /// @note The SQLLogChannel retains ownership over the db and is responsible for cleaning it up.
        void reset();

    private:
        /// @brief Write all buffered data in a single sqlite transaction.
        void writeOut();

        /// @brief A struct representing the exact data to put in the db.
        struct logrow
        {
            /// @brief Construct a log row based on its contents.
            /// @param time_ : The timestamp of the message (seconds since epoch)
            /// @param data_ : serialized blob of the message
            /// @param publisherId_ : Unique identifier for the publisher
            /// @param messageCounter_ : Sequence number for that channel from that publisher
            logrow(double time_, std::string data_, uint32_t publisherId_, uint32_t messageCounter_)
                : time(time_)
                , data(std::move(data_))
                , publisherId(publisherId_)
                , messageCounter(messageCounter_)
            {}
            double time{-1.0};           /// Timestamp (seconds since epoch)
            std::string data{};          /// Data (serialized message blob)
            uint32_t publisherId{-1};    /// Unique identifier for the publisher
            uint32_t messageCounter{-1}; /// Sequence number for the message from that publisher
        };

        /// @brief Binds (double) data to the sqlite statement
        /// @brief index : Index of the field to bind
        /// @brief data : data to bind
        bool bind(int index, double data);

        /// @brief Binds (string/blob) data to the sqlite statement
        /// @brief index : Index of the field to bind
        /// @brief data : data to bind
        bool bind(int index, const std::string& data);

        /// @brief Binds (int64) data to the sqlite statement
        /// @brief index : Index of the field to bind
        /// @brief data : data to bind
        bool bind(int index, int64_t data);

        sqlite3* db_{nullptr};             /// Pointer to the sqlite db, owned by SQLLogChannel but shared with us
        sqlite3_stmt* statement_{nullptr}; /// Pointer to prepared statement
        uint32_t logVersion_{0};           /// Version of the log db schema
        std::queue<logrow> buffer_{};      /// Queue of messages needing writeout
        size_t maxVolume_{5000000};        /// Buffer size threshold for writeout
        size_t currentVolume_{0};          /// Amount of data currently buffered
        double writeoutTimeout_{1.0};      /// Writeout data after this amount of time (seconds)
    };

    /**
     * \brief A tailored sub-class for measuring the throughput for a log channel.
     *
     * Each write, add a "sub-sample" of the number of bytes written
     *
     * Periodically (eg every second) "take" the sample. The amount of data written since the last sample, then divide
     * by number of seconds since last sample taken. This value (bytes/second) is the real sample.
     */
    class ThroughputStats : public statistics::CombinationStatistic<double>
    {
    public:
        ThroughputStats();

        /**
         * \brief default destructor
         */
        virtual ~ThroughputStats() = default;

        /**
         * \brief Commit a sample with the amount of data written.
         *
         * \note The ThroughputStats class will take care of the time since last sample internally
         *
         * \param[in] dataWritten: The amount of data written since the last sample
         */
        void addDataWritten(const uint64_t dataWritten);

    protected:
        // We specify the collectors
        using statistics::CombinationStatistic<double>::addCollector;
        ottotime::otimePoint lastSampleTime_{}; ///< Time of the last update
    };

    /*!
     * @brief This method shall record profiling information on runtime of SQLite transactions when profiling is
     * enabled.
     *
     * @param[in] type          SQLite trace event code.
     *
     * @param[in] context       Pointer to the SQLLogChannel initiating the transaction.
     *
     * @param[in] statement     Pointer to sqlite prepared statement profiled.
     *
     * @param[in] nanosec64     runtime to commit statement in nanoseconds. void* points to uint64_t
     *
     * @return SQLITE_OK (0)
     *
     */
    static int tracecallback(uint32_t type, void* context, void* statement, void* nanosec64);

    sqlite3* logDb_{nullptr};             ///< The pointer to sqlite DB.
    bool readCurrent_{false};             ///< Becomes true when seek() succeeds.
    int64_t rowId_{-1};                   ///< The current db row ID to be read.
    SQLLogChannelConfig config_{};        ///< This variable holds db file name supplied from configuration.
    uint32_t logVersion_{0};              ///< Corresponds to user_version
    LatencyStats ls_{"", -1, "", "", -1}; ///< Stores information on SQLite3 profiling when enabled

    ThroughputStats throughput_{};
    std::atomic<uint64_t> dataWritten_{0};
    WriteBuffer buffer_{10000000, 1.0};
};

} // namespace task

#endif // SQLLOGCHANNEL_H
