/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file ThreadedTypedChannel.def.h
 * @product common
 * @version
 * @released
 * @updated
 * @brief Threaded, typed channel that deserializes in a background thread
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#ifndef THREADEDTYPEDCHANNEL_H
#error Should only be included from ThreadedTypedChannel.h
#endif

#include <Serialization/SerializationUtilities.h>
#include <cstring>
#include <interfaces/baseTypes/Header.h>

#include <future>

namespace task {

template<typename T>
ThreadedTypedChannel<T>::ThreadedTypedChannel(Channel* rawChannel, const size_t queueSize)
    : TypedChannel<T>(rawChannel)
    , maxQueueSize_(queueSize)
    , readQueue_()
    , readQueueMutex_()
    , cv_()
{
    launchReadThread();
}

template<typename T>
ThreadedTypedChannel<T>::~ThreadedTypedChannel()
{
    // Signal the worker to stop
    run_ = false;

    // Signal the blocking read to stop
    TypedChannel<T>::unblock();

    // And finally join
    readThread_.join();
}

template<typename T>
void ThreadedTypedChannel<T>::launchReadThread()
{
    run_ = true;
    readThread_ = std::thread([this]() { this->readThreadWorker(); });
}

template<typename T>
void ThreadedTypedChannel<T>::destroy(std::unique_ptr<std::pair<std::string, T>> msgPair)
{
    // Trigger destruction
    delete msgPair.release();
}

template<typename T>
void ThreadedTypedChannel<T>::readThreadWorker()
{
    while (run_) {
        // Use a pointer here to minimize time spent with the mutex locked
        auto pairPtr = std::make_unique<std::pair<std::string, T>>();
        pairPtr->first.resize(interfaces::baseTypes::getInterfaceHeaderSize(OTTO_INTERFACE_HEADER_VERSION));

        auto readStatus = TypedChannel<T>::read(&pairPtr->first[0], pairPtr->first.size(), pairPtr->second, true);
        if (readStatus == CS_NewData) {
            // Use a unique lock here so we can manually unlock and notify later
            std::unique_lock<std::mutex> lock(readQueueMutex_);
            readQueue_.emplace(std::move(pairPtr));

            // If we're over the size limit discard the oldest data
            if (readQueue_.size() > maxQueueSize_) {
                // Defer the actual destruction since the destructor could take a while and we're
                // in a locked section
                std::async(std::launch::async, &ThreadedTypedChannel<T>::destroy, std::move(readQueue_.front()));
                readQueue_.pop();
            }

            lock.unlock();
            cv_.notify_one();
        }
    }
}

template<typename T>
ChannelStatus ThreadedTypedChannel<T>::read(char* hdr, std::size_t hdrSize, T& msg, bool blocking)
{
    std::unique_lock<std::mutex> lock(readQueueMutex_);

    // Wait in a loop to handle spurious wakeups
    while (blocking && readQueue_.empty()) {
        cv_.wait(lock);
    }

    if (readQueue_.empty()) {
        return CS_Error;
    }

    auto msgPairPtr = std::move(readQueue_.front());
    std::memcpy(hdr, msgPairPtr->first.c_str(), hdrSize);

    msg = std::move(msgPairPtr->second);

    readQueue_.pop();

    return CS_NewData;
}

} // namespace task
