/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file ThreadedTypedChannel.h
 * @product common
 * @version
 * @released
 * @updated
 * @brief Threaded, typed channel that deserializes in a background thread
 *
 * This is the header for common code used to interact with the operating system's network
 * functionality.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#ifndef THREADEDTYPEDCHANNEL_H
#define THREADEDTYPEDCHANNEL_H

#include "Comms/Channel.h"

#include <condition_variable>
#include <mutex>
#include <queue>
#include <thread>

namespace task {

/*!
 * @class ThreadedTypedChannel
 *
 * @brief This class shall be implemented as a type specific wrapper over channel with a threaded read.
 * \n \b Detailed \b Description:
 * \n On creating threaded Typed channel, it launches worker thread for reading the messages.
 * \n Threaded typed channel shall have it's own read queue and filled by the underlying channel.
 * \n The message shall be removed from the read queue after read operation.
 * \n This class supports read with blocking and non blocking mode.
 * \n blocking mode (true) - If read queue is empty,  it waits for the new data arrival in read queue.
 * \n non blocking mode(false), It won't wait, if read is empty.
 * \n
 * \n The underlying channel's thread shall get the messages in serialized form.
 * \n But in ThreadedTypedChannel's thread, messages shall be deserialized.
 **/
template<typename T>
class ThreadedTypedChannel : public TypedChannel<T>
{
public:
    /*!
     * @brief Parameterized constructor
     * \n This method shall initialize the Typed channel and launches the worker thread.
     * \n The worker thread shall read the data from the underlying channel.
     *
     * @param[in] rawChannel as Channel pointer : The underlyig channel pointer
     * @param[in] queueSize as size_t : The max size limit set on channel's receive queue.
     */
    ThreadedTypedChannel(Channel* rawChannel, size_t queueSize = 1);
    /*!
     * @brief Virtual destructor
     * \n This method shall stop the worker thread.
     * \n This method shall unblock the read method, if it is waiting for new data arrival in blocking mode.
     */
    virtual ~ThreadedTypedChannel();

    /*!
     * @brief This method shall read the data from the underlying channel.
     * \n The threaded Typed channel is a wrapper over channel whose read queue shall be filled by it's underlying
     * channel. \n The messages in read queue is stored as pair objects where first object is Interface header and
     * Second object is the message. \n \b Operational \b Flow: \n If the threaded Typed channel's read queue is empty,
     * then \n It shall wait for the new messages available in it's read queue [blocking mode] \n If the threaded Typed
     * channel's read queue is not empty, then \n It reads the messages from it's read queue [non blocking/blocking
     * mode] \n It shall extract the header and message from the received message read from read queue. \n This method
     * shall remove the message from read queue after read.
     *
     * @param[out] hdr hdr as char* pointer : The header of the received message.
     * @param[in] hdrSize as size_t : The size of the header.
     * @param[out] msg as generic type(template type): The received message type specific to the channel.
     * @param[in] blocking as bool : true, blocking read.
     * \n                            false, non blocking read.
     *
     * @return ChannelStatus
     * \n If new data is available on the channel, CS_NewData.
     * \n If there is no data, CS_Error.
     */
    virtual ChannelStatus read(char* hdr, std::size_t hdrSize, T& msg, bool blocking = false) override;

private:
    /*!
     * @brief This method shall work as a worker thread, continuously check for the new data arrival on the channel.
     * \n The threaded Typed channel is a wrapper over channel whose read queue shall be filled by it's underlying
     * channel. \n This worker thread gets the data from the underlying channel. \n The messages in read queue shall be
     * stored as pair object where first object is Interface header and Second object is the message. \n \b Operational
     * \b Flow: \n The thread shall keep checking for the new data arrival in blocking mode. \n If the underlying raw
     * channel's receive queue is empty, the thread shall wait until the new data is available in blocking mode. \n To
     * store the received message, a unique pointer to the pair object shall be created. This is required \n to minimize
     * the time spend for creating pair object and copying the new data while lock is active on read queue. \n Copy the
     * received message to pair object, if new data is available \n IF new data is received from the underlying channel
     * THEN \n &emsp; Push the new data as a pair object into thread typed channel's read queue. \n &emsp; Check if the
     * number of messages waiting in read queue to be read is greater than the limit set on the queue size. \n &emsp; If
     * so, remove the oldest message from thread Typed channel's read queue.. \n &emsp; The thread notifies the read
     * method if it is waiting for the new data. \n &emsp; Access to the thread Typed channel's read queue shall be
     * synchronized as worker thread and read method both access the read queue. \n END IF
     *
     * @return void
     */
    void readThreadWorker();
    /*!
     * @brief This method shall start the worker thread.
     * \n It shall be called to launch the worker thread when the Typed channel is created.
     * \n The worked thread shall check continuously for the new data arrival on the channel.
     *
     * @return void
     */
    void launchReadThread();
    /*!
     * @brief This method shall release the pointer associated with the given pair object.
     *
     * @param[in] msgPair : The unique pointer to pair object.
     *
     */
    static void destroy(std::unique_ptr<std::pair<std::string, T>> msgPair);

    std::thread readThread_; ///< It is a thread that reads the data from the underlying channel
    bool run_ =
        false; ///< This variable is to control the thread. If true, thread is running. If false, thread is stopped.
    size_t maxQueueSize_ =
        1; ///< The max size limit set on the read queue is supplied when the threaded Typed channel is created.
    std::queue<std::unique_ptr<std::pair<std::string, T>>>
        readQueue_;             ///< The read queue shall be filled by the read thread. A unique Pointer to the
                                ///< pair object<header,message> is used to minimize the time spend in a locked section.
                                ///< So, the pair object shall be ready before entering locking section.
    std::mutex readQueueMutex_; ///< The mutex lock is for read queue which shall be accessed by read worker thread and
                                ///< read method.
    std::condition_variable cv_; ///< This Condition variable is used to implement blocking reads on the read queue.
                                 ///< If new data is available, read worker thread sends a signal to read method if it
                                 ///< is doing blocking read
};

} // namespace task

#include "ThreadedTypedChannel.def.h"

#endif // THREADEDTYPEDCHANNEL_H
