/*!
 ***********************************************************
 * @file TypedChannel.def.h
 * @brief A file implementing TypedChannel
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 09/19/2017
 * @remark replacing ruby functions with libconfig++.
 * updated memcpy -> std::memcpy
 * @n
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/05/2017
 * @remark
 * get() supports a blocking call.
 * @n
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 06/09/2017
 * @remark
 * read() and write() are updated to use a timestamp marked
 * when the data is sent for the log playback.
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * TypedChannel.def.h
 * Tugrul Galatali (tugrul@galatali.com)
 * 10/30/2006
 *
 * Junsung Kim
 * 12/16/2012
 * fault-tolerance support is added.
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */

#ifndef TYPEDCHANNEL_H
#error Should only be included from Channel.h
#endif

#include <ODatum/ODatum.h>
#include <Serialization/SerializationUtilities.h>
#include <cstring>

namespace task {

template<typename T>
TypedChannel<T>::TypedChannel(std::unique_ptr<Channel> rawChannel)
    : rawChannel_(std::move(rawChannel))
{}

template<typename T>
ChannelStatus TypedChannel<T>::connect(void)
{
    ChannelStatus status{CS_Error};
    if (rawChannel_ != nullptr) {
        status = rawChannel_->connect();
    }

    return status;
}

template<typename T>
ChannelStatus TypedChannel<T>::disconnect(void)
{
    ChannelStatus status{CS_Error};
    if (rawChannel_ != nullptr) {
        status = rawChannel_->disconnect();
    }

    return status;
}

template<typename T>
ChannelStatus TypedChannel<T>::read(char* hdr, std::size_t hdrSize, T& msg, av::latency::Pred& pred, bool blocking)
{
    static_assert(std::is_base_of<aptv::ODatum, T>::value,
                  "All datums must inherit from aptv::ODatum, please confirm for T in the compiler message");

    ChannelStatus status{CS_Error};
    if (rawChannel_ != nullptr) {
        std::string rawMsg{};

        status = rawChannel_->readWithPred(rawMsg, pred, blocking);

        if (status == CS_NewData) {
            if (rawMsg.size() >= hdrSize) {
                std::memcpy(hdr, rawMsg.c_str(), hdrSize);
                rawMsg.erase(0, hdrSize);
                aptv::deserializeODatum(msg, rawMsg);
            } else {
                status = CS_Error;
            }
        }
    }

    return status;
}

template<typename T>
ChannelStatus TypedChannel<T>::writeV(char* hdr,
                                      std::size_t hdrSize,
                                      const T& msg,
                                      std::vector<av::latency::Pred>&& predecessors)
{
    static_assert(std::is_base_of<aptv::ODatum, T>::value,
                  "All datums must inherit from aptv::ODatum, please confirm for T in the compiler message");

    ChannelStatus status{CS_Error};
    if (rawChannel_ != nullptr) {
        std::string rawMsg{};
        const std::string header(hdr, hdrSize);

        aptv::serializeODatumMessage(header, rawMsg, msg);

        // The blob is now [ interface header ][ type header ][ datum ]

        status = rawChannel_->write(rawMsg, std::move(predecessors));
    }

    return status;
}

template<typename T>
ChannelStatus TypedChannel<T>::chmod(const task::Access& acc)
{
    ChannelStatus status{CS_Error};
    if (rawChannel_ != nullptr) {
        status = rawChannel_->chmod(acc);
    }

    return status;
}

template<typename T>
bool TypedChannel<T>::getQueueProperties(unsigned long& currentDepth, unsigned long& maxDepth) const
{
    return (rawChannel_ != nullptr) ? rawChannel_->getQueueProperties(currentDepth, maxDepth) : false;
}

template<typename T>
void TypedChannel<T>::unblock()
{
    rawChannel_->unblock();
}

template<typename T>
void TypedChannel<T>::clear()
{
    rawChannel_->clear();
}

} // namespace task
