/*!
 ***********************************************************
 * @file TypedChannel.h
 * @brief A file defining TypedChannel
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 06/09/2017
 * @remark
 * read() and write() are updated to use a timestamp marked
 * when the data is sent for the log playback.
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**********************************************************
 * TypedChannel.h
 * Tugrul Galatali
 * 10/31/2006
 *
 * Junsung Kim
 * 12/16/2012
 * fault-tolerance support is added.
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 **********************************************************
 */
#ifndef TYPEDCHANNEL_H
#define TYPEDCHANNEL_H

#include "Comms/Channel.h"

namespace task {

/*!
 * @class TypedChannel
 *
 * @brief This class shall be implemented as a type specific adapter to Channel.
 * \n The TypedChannel shall carry the messages(read/write) specific to interface.
 * \n Example:HardwareHealth Interface uses the message type HardwareHealthMessage.
 * \n Based on the mode supplied from configuration, this channel plays as Publisher/SubScriber/SubPub role.
 * \n It supports cereal serialization.
 * \n This channel connects to IPC module.
 */
template<typename T>
class TypedChannel
{
public:
    /*!
     * @brief Parameterized constructor
     * \n It shall initialize the Typed Channel with given channel.
     *
     * @param[in] rawChannel as Channel pointer : The channel to send/receive the messages from/to the interface based
     * on the mode.
     *
     */
    TypedChannel(std::unique_ptr<Channel> rawChannel);

    /*!
     * @brief Virtual destructor
     * \n This method shall disconnect the channel from IPC module.
     */
    virtual ~TypedChannel() = default;

    /*!
     * @brief This method shall connect the channel to the IPC module based on the given mode.
     * \n If the channel is connected, then it is ready to receive/send messages to/from the interface based on the
     * mode.
     *
     * @return ChannelStatus : If the connection is successful, CS_Success. Otherwise, CS_Error
     *
     */
    virtual ChannelStatus connect(void);

    /*!
     * @brief This method shall disconnect the channel from IPC module.
     * \n It shall not be able to receive/send messages to/from the interface if the channel is disconnected
     *
     * @return ChannelStatus : If the disconnection is successful, CS_Success. Otherwise, CS_Error.
     *
     */
    virtual ChannelStatus disconnect(void);

    /// @brief Read a message from the IPC module.
    ///
    /// @param[out] hdr : The interface header extracted from the
    /// received message.
    ///
    /// @param[in] hdrSize as size_t : The size of the expected
    /// interface header.
    ///
    /// @param[out] msg : The message received the IPC module.
    ///
    /// @param[out] pred : The predecessor to the read event.
    ///
    /// @param[in] blocking : if true, the call will block until a
    /// message is available.
    ///
    /// @return ChannelStatus
    virtual ChannelStatus read(char* hdr, std::size_t hdrSize, T& msg, av::latency::Pred& pred, bool blocking = false);
    ChannelStatus read(char* hdr, std::size_t hdrSize, T& msg, bool blocking = false)
    {
        av::latency::Pred pred;
        return read(hdr, hdrSize, msg, pred, blocking);
    }

    /*!
     * @brief This method shall write the messages to the IPC module to publish.
     * \n \b Operational \b Flow:
     * \n IF the channel exists for the interface THEN
     * \n &emsp; This method shall add the header to the given message to be sent.
     * \n &emsp; Serialize the message and publishes the message.
     * \n ELSE
     * \n &emsp; return error.
     * \n END IF
     *
     * @param[in] hdr as pointer to string: The interface header to be added to the send message.
     * @param[in] hdrSize as size_t : The size of the interface header.
     * @param[in] msg as message of generic type(template type) : The message to be written.
     *
     * @return ChannelStatus : If the write is successful, CS_Success. Otherwise,CS_Error.
     *
     */
    template<class... Preds>
    ChannelStatus write(char* hdr, std::size_t hdrSize, const T& msg, Preds... predecessors)
    {
        std::vector<av::latency::Pred> p{predecessors...};
        return writeV(hdr, hdrSize, msg, std::move(p));
    }

    ChannelStatus writeV(char* hdr, std::size_t hdrSize, const T& msg, std::vector<av::latency::Pred>&& predecessors);

    /*!
     * @brief This method shall change the mode of the channel.
     *
     * @param[in] mode as Access enum: The new mode to be changed
     *
     * @return ChannelStatus. CS_Success, always.
     *
     */
    virtual ChannelStatus chmod(const task::Access& acc);

    /*!
     * @brief This method shall retrieve the queue properties of the underlying channel.
     *
     * @param[out] currentDepth: The number of messages waiting in the queue to be read.
     *
     * @param[out] maxDepth: The maximum depth(limit) of the message queue size.
     *
     * @return bool : If channel supports queues, true(Ex:SCChannel). Otherwise, false (Ex: SQLLogChannel )
     *
     */
    virtual bool getQueueProperties(unsigned long& currentDepth, unsigned long& maxDepth) const;

    /*!
     * @brief This method shall unblock the channel if it is doing a blocking read.
     * \n If there are no messages on the interface, the channel shall be waiting for the messages to be available for
     * read in blocking mode. \n Invoking this method, ensures that the channel won't wait for the messages to be
     * arrived on the interface if it is doing blocking read.
     *
     * @return void
     */
    virtual void unblock();

    /*!
     * @brief This method shall clear the messages received on the channel.
     *
     * @return void
     */
    virtual void clear();

    const std::string& getName() const { return rawChannel_->getName(); }

private:
    std::unique_ptr<Channel>
        rawChannel_; ///< The interface channel shall read messages received on the interface (or) send messages
                     ///< from interface. A message type shall be specific to the interface.Ex:HardwareHealth
                     ///< Interface uses the message type HardwareHealthMessage.
};

} // namespace task

#include <Comms/TypedChannel.def.h>

#endif // TYPEDCHANNEL_H
