/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file ChannelFactoryTest.cc
 * @brief This file implements tests for the ChannelFactory class.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#include <memory>

#include <gmock/gmock-more-actions.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include "platform/middleware/libs/DDSComms/test/DDSTestPathsHelper.hpp"
#include <Comms/ChannelFactory.h>
#include <Comms/SQLLogChannel.h>
#include <fmt/format.h>
#include <task/Task.h>

using namespace ::testing;

class ConfigSectionTest : public ConfigSection
{
public:
    ConfigSectionTest()
        : ConfigSection()
    {}

    bool init(const std::string& name, ConfigPtr& cfg) { return ConfigSection::init(name, cfg); }
};

/**
 * @brief Group of initialization tests for the channel factory
 */
TEST(ChannelFactoryInitTests, GivenNoInitWhenConnectAsThenSuccess)
{
    EXPECT_EQ(true, task::ChannelFactory::connectAs("TestModule"));
}

class AllAccessTypeTest : public testing::TestWithParam<task::Access>
{};

TEST_P(AllAccessTypeTest, GivenEmptyConfigCreateChannelReturnsNullptr)
{
    ConfigSection emptyConfigSection{};

    auto managed = task::ChannelFactory::createChannelManaged(emptyConfigSection, GetParam());

    EXPECT_EQ(nullptr, managed);
}

TEST_P(AllAccessTypeTest, GivenConfigWithoutTypeCreateChannelReturnsNullptr)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "name = \"testName\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    auto managed = task::ChannelFactory::createChannelManaged(configTest, GetParam());

    EXPECT_EQ(nullptr, managed);
}

TEST_P(AllAccessTypeTest, GivenConfigWithInvalidTypeCreateChannelReturnsNullptr)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "name = \"testName\";"
                                        "type = \"NONESUCH\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    auto managed = task::ChannelFactory::createChannelManaged(configTest, GetParam());

    EXPECT_EQ(nullptr, managed);
}

TEST_P(AllAccessTypeTest, GivenValidConfigCreateChannelReturnsChannel)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "name = \"testName\";"
                                        "type = \"LOG\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    auto managed = task::ChannelFactory::createChannelManaged(configTest, GetParam());

    EXPECT_NE(nullptr, managed);
    EXPECT_EQ(std::string("testName"), managed->getName());
}

TEST_P(AllAccessTypeTest, GivenValidConfigCreateChannelReturnsChannelNameless)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "type = \"LOG\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    auto managed = task::ChannelFactory::createChannelManaged(configTest, GetParam());

    EXPECT_NE(nullptr, managed);
    EXPECT_EQ(std::string("unused"), managed->getName());
}

TEST_P(AllAccessTypeTest, RawLogChannelWithoutName)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "type = \"LOG\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    task::SQLLogChannel channel{configTest, GetParam()};

    EXPECT_EQ(std::string("unused"), channel.getName());
}

TEST_P(AllAccessTypeTest, RawLogChannelWithNameInConfig)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "name = \"testName\";"
                                        "type = \"LOG\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    task::SQLLogChannel channel{configTest, GetParam()};

    EXPECT_EQ(std::string("testName"), channel.getName());
}

TEST_P(AllAccessTypeTest, RawLogChannelWithNameInConstructor)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "type = \"LOG\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    task::SQLLogChannel channel{configTest, GetParam(), "foo"};

    EXPECT_EQ(std::string("foo"), channel.getName());
}

TEST_P(AllAccessTypeTest, RawLogChannelWithNameInBoth)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "name = \"testName\";"
                                        "type = \"LOG\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    task::SQLLogChannel channel{configTest, GetParam(), "foo"};

    EXPECT_EQ(std::string("foo"), channel.getName());
}

TEST_P(AllAccessTypeTest, RawLogChannelMoved)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "name = \"testName\";"
                                        "type = \"LOG\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    task::SQLLogChannel channel{configTest, GetParam(), "foo"};
    EXPECT_EQ(std::string("foo"), channel.getName());

    task::SQLLogChannel channel2{std::move(channel)};
    EXPECT_EQ(std::string("foo"), channel2.getName());
}

TEST_P(AllAccessTypeTest, GivenValidConfigCreateChannelReturnsChannelWithName)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "name = \"testName\";"
                                        "type = \"LOG\";"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    const std::string name{"overrideName"};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    auto managed = task::ChannelFactory::createChannelManaged(configTest, GetParam(), name);

    EXPECT_NE(nullptr, managed);
    EXPECT_EQ(name, managed->getName());
}

#if 0
// This test should be included, but appears to run into bugs in
// ConfigSection, which believes maxRetries is a group. Since config's
// unit tests are all disabled, I am going to defer fixing this bug.
TEST_P(AllAccessTypeTest, GivenValidSCConfigCreateChannelReturnsChannelWithName)
{
    std::string testCreateChannelConfig{"testCreateChannel ="
                                        "{"
                                        "name = \"testName\";"
                                        "type = \"SimpleComms\";"
                                        "maxRetries = 1;"
                                        "};"};

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testCreateChannelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    auto managed = task::ChannelFactory::createChannelManaged(configTest, GetParam());

    EXPECT_NE(nullptr, managed);
    EXPECT_EQ("testName", managed->getName());
}
#endif

INSTANTIATE_TEST_SUITE_P(AccessTypeTests,
                         AllAccessTypeTest,
                         ::testing::Values(task::Read, task::Write, task::ReadWrite, task::Standby));

std::string getDDSChannelConfig(const char* extraField = "")
{
    return fmt::format("testCreateChannel ="
                       "{{"
                       "name = \"testName\";"
                       "type = \"DDS\";"
                       "ddsConfigFile = \"{}\";"
                       "{}"
                       "}};",
                       DDSComms::testing::configFile,
                       extraField);
}

void expectCreatedChannels(const std::string& channelConfig)
{

    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(channelConfig);

    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("testCreateChannel", configPtr));

    EXPECT_NE(nullptr, task::ChannelFactory::createChannelManaged(configTest, task::Access::Read));
    EXPECT_NE(nullptr, task::ChannelFactory::createChannelManaged(configTest, task::Access::Write));
    EXPECT_NE(nullptr, task::ChannelFactory::createChannelManaged(configTest, task::Access::ReadWrite));
    EXPECT_NE(nullptr, task::ChannelFactory::createChannelManaged(configTest, task::Access::Standby));
}

TEST(ChannelFactoryInitTests, GivenDDSConfigCreateChannelReturnsDDSChannel1)
{
    std::string testCreateChannelConfigWithSizeDomain = getDDSChannelConfig("queueSize = \"100\"; domainId = \"0\";");

    SCOPED_TRACE("GivenDDSConfigCreateChannelReturnsDDSChannel1");
    expectCreatedChannels(testCreateChannelConfigWithSizeDomain);
}

TEST(ChannelFactoryInitTests, GivenDDSConfigCreateChannelReturnsDDSChannel2)
{
    std::string testCreateChannelConfig = getDDSChannelConfig();

    SCOPED_TRACE("GivenDDSConfigCreateChannelReturnsDDSChannel2");
    expectCreatedChannels(testCreateChannelConfig);
}

TEST(ChannelFactoryInitTests, ConnectToSCS)
{
    task::ChannelFactory cf;
    EXPECT_EQ(nullptr, cf.connectToSCS(1));
    EXPECT_EQ(nullptr, cf.connectToSCS(1));
}
