/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Tests for the CommsPlaybackClock class.
 */

#include <memory>

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <Comms/CommsPlaybackClock.h>

using namespace ::testing;

/**
 * \brief This class defines a fully mocked GenericInputIntf.
 */
template<typename TStruct>
class MockGenericInput : public GenericInputIntf<TStruct>
{

public:
    MockGenericInput() = default;
    virtual ~MockGenericInput() = default;
    MOCK_METHOD2_T(get, bool(TStruct& val, bool blocking));
    MOCK_METHOD0(unblock, void());
    MOCK_METHOD0(clear, void());
    MOCK_METHOD0(initialize, bool());
    MOCK_METHOD0(cleanup, void());
};

typedef MockGenericInput<PlaybackClockParams> ParamsInput;

class MockClock : public Clock
{
public:
    MOCK_METHOD0(getCurrentTime, ottotime::otimePoint());
};

TEST(CommsPlaybackClockTests, OnInitNoSourceHasNoParams)
{
    CommsPlaybackClock clock(nullptr, nullptr);
    auto params = clock.getParams();
    EXPECT_FALSE(params.has_value());
}

TEST(CommsPlaybackClockTests, OnInitNoSourceNoResult)
{
    CommsPlaybackClock clock(nullptr, nullptr);

    EXPECT_EQ(clock.getCurrentTime(), ottotime::otimePoint{});
}

TEST(CommsPlaybackClockTests, OnInitHasNoParams)
{
    ParamsInput paramsChannel{};
    MockClock base{};

    EXPECT_CALL(paramsChannel, get(_, _)).Times(Exactly(1)).WillOnce(Return(false));

    CommsPlaybackClock clock(&paramsChannel, &base);
    auto params = clock.getParams();
    EXPECT_FALSE(params.has_value());
}

TEST(CommsPlaybackClockTests, OnInitNoResult)
{
    ParamsInput paramsChannel{};
    MockClock base{};
    CommsPlaybackClock clock(&paramsChannel, &base);

    EXPECT_CALL(paramsChannel, get(_, _)).Times(Exactly(1)).WillOnce(Return(false));

    EXPECT_EQ(clock.getCurrentTime(), ottotime::otimePoint{});
}

TEST(CommsPlaybackClockTests, HasParamsWhenGiven)
{
    ParamsInput paramsChannel{};
    MockClock base{};
    CommsPlaybackClock clock(&paramsChannel, &base);

    PlaybackClockParams setParams{};
    setParams.offset = otimePoint(otimeSpec(0, 0));
    setParams.baseTime = otimeSpec(12345, 12345);
    setParams.stopTime = otimeSpec(12350, 54321);
    setParams.scaleFactor = 5;

    EXPECT_CALL(paramsChannel, get(_, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<0>(setParams), Return(true)))
        .WillRepeatedly(Return(false));

    auto params = clock.getParams();

    EXPECT_TRUE(params.has_value());
    EXPECT_EQ(params, setParams);
}

TEST(CommsPlaybackClockTests, AllZerosBaseTime)
{
    /*
     * This is the usage in the LogPlayer as of Feb 2021...
     * Setting offset and scaleFactor to 0, and with an invalid stop time (no max time)
     * results in (getCurrentTime == params.baseTime)
     */

    ParamsInput paramsChannel{};
    MockClock base{};
    CommsPlaybackClock clock(&paramsChannel, &base);

    PlaybackClockParams setParams{};
    setParams.offset = otimePoint(otimeSpec(0, 0));
    setParams.baseTime = otimePoint(otimeSpec(12345, 12345));
    setParams.stopTime = otimePoint(otimeSpec(0, 0));
    setParams.scaleFactor = 0;

    // Must invalidate the stop time to prevent it from putting a cap
    setParams.stopTime.invalidate();

    EXPECT_CALL(paramsChannel, get(_, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<0>(setParams), Return(true)))
        .WillRepeatedly(Return(false));

    EXPECT_CALL(base, getCurrentTime()).Times(Exactly(1)).WillOnce(Return(otimePoint(otimeSpec(1000000, 10000))));

    auto params{clock.getParams()};
    EXPECT_TRUE(params.has_value());
    EXPECT_EQ(params, setParams);
    EXPECT_EQ(clock.getCurrentTime(), setParams.baseTime);
}

TEST(CommsPlaybackClockTests, StopTimeCapsOutput)
{
    /*
     * Same setup as above but this time we set a stopTime, which will cap the output
     */
    ParamsInput paramsChannel{};
    MockClock base{};
    CommsPlaybackClock clock(&paramsChannel, &base);

    PlaybackClockParams setParams{};
    setParams.offset = otimePoint(otimeSpec(0, 0));
    setParams.baseTime = otimePoint(otimeSpec(12345, 12345));
    setParams.stopTime = otimePoint(otimeSpec(100, 0));
    setParams.scaleFactor = 0;

    EXPECT_CALL(paramsChannel, get(_, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<0>(setParams), Return(true)))
        .WillRepeatedly(Return(false));

    EXPECT_CALL(base, getCurrentTime()).Times(Exactly(1)).WillOnce(Return(otimePoint(otimeSpec(1000000, 10000))));

    auto params = clock.getParams();
    EXPECT_TRUE(params.has_value());
    EXPECT_EQ(params, setParams);
    EXPECT_EQ(clock.getCurrentTime(), setParams.stopTime);
}

TEST(CommsPlaybackClockTests, WithOffsetMath)
{
    /*
     * Same setup as above but this time we set a stopTime, which will cap the output
     */
    ParamsInput paramsChannel{};
    MockClock base{};
    CommsPlaybackClock clock(&paramsChannel, &base);

    PlaybackClockParams setParams{};
    setParams.offset = otimePoint(otimeSpec(10, 100000));
    setParams.baseTime = otimePoint(otimeSpec(1000, 100000));
    setParams.stopTime = otimePoint(otimeSpec(0, 0));
    setParams.scaleFactor = 0.5;
    setParams.stopTime.invalidate();

    EXPECT_CALL(paramsChannel, get(_, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<0>(setParams), Return(true)))
        .WillRepeatedly(Return(false));

    EXPECT_CALL(base, getCurrentTime()).Times(Exactly(1)).WillOnce(Return(otimePoint(otimeSpec(1010, 300000))));

    auto params = clock.getParams();
    EXPECT_TRUE(params.has_value());
    EXPECT_EQ(params, setParams);
    EXPECT_EQ(clock.getCurrentTime(), otimePoint(otimeSpec(1010, 250000)));
}
