#include "platform/middleware/libs/DDSComms/test/DDSTestHelper.hpp"
#include <TimeStamp/otimeConversions.h>
#include <chrono>
#include <dds/dds.hpp>
#include <future>
#include <gtest/gtest.h>
#include <thread>

#include <Comms/DDSChannel.h>
using namespace task;
using namespace std;
using namespace std::chrono_literals;

using DDSComms::testing::getTestCfg;
using BytesTopicType = DDSComms::BytesTopicType;
using DomainParticipant = dds::domain::DomainParticipant;
using BytesTopic = dds::topic::Topic<BytesTopicType>;
using Publisher = dds::pub::Publisher;
using Subscriber = dds::sub::Subscriber;
using BytesWriter = dds::pub::DataWriter<BytesTopicType>;
using BytesReader = dds::sub::DataReader<BytesTopicType>;

class DDSChannelSanityTest : public ::testing::Test
{
protected:
    static void SetUpTestSuite()
    {
        SystemClock::initialize(ottotime::SystemClock::SystemTimeMode::TimeModeSteady, nullptr);
    }
};

TEST_F(DDSChannelSanityTest, createExistingDP)
{
    DomainParticipant dp{0};
    DDSChannel reader("ChannelTest", Read, 100, 0, getTestCfg());
    DDSChannel writer("ChannelTest", Write, 100, 0, getTestCfg());
}

TEST_F(DDSChannelSanityTest, create)
{
    DDSChannel reader("ChannelTest", Read, 100, 0, getTestCfg());
    DDSChannel writer("ChannelTest", Write, 100, 0, getTestCfg());
}

TEST_F(DDSChannelSanityTest, writeRead)
{
    DDSChannel reader("ReadWriteTest", Read, 100, 0, getTestCfg());
    DDSChannel writer("ReadWriteTest", Write, 100, 0, getTestCfg());

    EXPECT_EQ(writer.connect(), CS_Success);
    EXPECT_EQ(reader.connect(), CS_Success);

    std::this_thread::sleep_for(2000ms);

    EXPECT_EQ(writer.write("testdata"), CS_Success);

    std::this_thread::sleep_for(1000ms);

    std::string input{};
    auto status = reader.read(input);
    EXPECT_EQ(status, CS_NewData);
}

class DDSChannelInitializationTest : public ::testing::Test
{
protected:
    struct Context
    {
        Context()
            : dp(dds::core::null)
            , topic(dds::core::null)
            , pub(dds::core::null)
            , sub(dds::core::null)
            , bWriter(dds::core::null)
            , bReader(dds::core::null)
        {}

        DomainParticipant dp;
        BytesTopic topic;
        Publisher pub;
        Subscriber sub;
        BytesWriter bWriter;
        BytesReader bReader;
    };

    DDSChannelInitializationTest()
        : reader("ReadTest", Read, 100, 0, getTestCfg())
        , writer("WriteTest", Write, 100, 1, getTestCfg())
        , rw_er("RWChannel", ReadWrite, 100, 2, getTestCfg())
    {}

    Context queryContext(int32_t id, string name, const DDSChannel& channel)
    {
        Context context{};

        context.dp = dds::domain::find(id);
        if (context.dp == dds::core::null)
            return context;

        context.topic = dds::topic::find<BytesTopic>(context.dp, name);
        if (context.topic == dds::core::null)
            return context;

        context.sub = channel.getSubscriber();
        context.pub = channel.getPublisher();
        context.bWriter = channel.getWriter();
        context.bReader = channel.getReader();

        return context;
    }
    DDSChannel reader;
    DDSChannel writer;
    DDSChannel rw_er;
};

DomainParticipant getDomainParticipant(int32_t domainId)
{
    dds::domain::qos::DomainParticipantQos qos{};
    return DomainParticipant(domainId, qos);
}

class DDSChannelFunctionalTest : public ::testing::Test
{
public:
    DDSChannelFunctionalTest(bool shouldConnect = true)
        : dp(getDomainParticipant(0))
        , readTopic(dp, "ReadTest")
        , writeTopic(dp, "WriteTest")
        , mockReader(Subscriber(dp), writeTopic, ddsConfig.createReaderQos())
        , mockWriter(Publisher(dp), readTopic, ddsConfig.createWriterQos())
        , reader("ReadTest", Read, 100, 0, getTestCfg())
        , writer("WriteTest", Write, 100, 0, getTestCfg())
    {
        if (shouldConnect) {
            reader.connect();
            writer.connect();

            while (mockWriter.publication_matched_status().current_count() == 0 ||
                   mockReader.subscription_matched_status().current_count() == 0) {
                this_thread::sleep_for(100ms);
            }
        }
    }

    void SetUp() { SystemClock::initialize(ottotime::SystemClock::SystemTimeMode::TimeModeSteady, nullptr); }

protected:
    DDSComms::DDSConfiguratorType ddsConfig{{DDSComms::testing::configFile}};
    DomainParticipant dp;
    BytesTopic readTopic, writeTopic;
    BytesReader mockReader;
    BytesWriter mockWriter;
    DDSChannel reader, writer;
};

class DDSChannelFunctionalTestNoConnect : public DDSChannelFunctionalTest
{
public:
    DDSChannelFunctionalTestNoConnect()
        : DDSChannelFunctionalTest(false)
    {}
};

class DDSChannelMultiThreadingTest : public ::testing::Test
{
public:
    DDSChannelMultiThreadingTest()
        : reader("Test", Read, 100, 0, getTestCfg())
        , writer("Test", Write, 100, 0, getTestCfg())
    {
        reader.connect();
        writer.connect();
    }

    void multipleWrite()
    {
        thread threadPool[5];
        const string data[5] = {"one", "two", "three", "four", "five"};
        for (auto idx = 0; idx < 5; idx++) {
            threadPool[idx] = thread([this, &data, idx]() { writer.write(data[idx]); });
        }

        for (auto idx = 0; idx < 5; idx++) {
            threadPool[idx].join();
        }
    }

    vector<string> multipleRead()
    {
        thread threadPool[5];
        vector<string> data(5, "");
        for (size_t idx = 0; idx < 5; idx++) {
            threadPool[idx] = thread(&DDSChannel::read, &reader, ref(data[idx]), true);
        }

        for (auto idx = 0; idx < 5; idx++) {
            threadPool[idx].join();
        }
        return data;
    }

    ChannelStatus disconnectWhileBlockRead()
    {
        string data;
        thread readThread(&DDSChannel::read, &reader, ref(data), true);
        future<ChannelStatus> f = async(launch::async, &DDSChannel::disconnect, &reader);
        readThread.join();
        return f.get();
    }

    ChannelStatus readAfterDisconnected()
    {
        string data;
        thread disconnectThread(&DDSChannel::disconnect, &reader);
        disconnectThread.join();
        future<ChannelStatus> f = async(launch::async, &DDSChannel::read, &reader, ref(data), false);
        return f.get();
    }

    ChannelStatus writeAfterDisconnected()
    {
        thread disconnectThread(&DDSChannel::disconnect, &writer);
        disconnectThread.join();
        future<ChannelStatus> f = async(launch::async, [this]() { return writer.write("test"); });
        return f.get();
    }

    ChannelStatus doubleDisconnected()
    {
        thread disconnectThread(&DDSChannel::disconnect, &writer);
        future<ChannelStatus> f = async(launch::async, &DDSChannel::disconnect, &writer);
        disconnectThread.join();
        return f.get();
    }

    ChannelStatus doubleConnected()
    {
        thread connectThread(&DDSChannel::connect, &writer);
        future<ChannelStatus> f = async(launch::async, &DDSChannel::connect, &writer);
        connectThread.join();
        return f.get();
    }

    void doubleChomd()
    {
        future<ChannelStatus> f1 = async(launch::async, &DDSChannel::chmod, &writer, ReadWrite);
        future<ChannelStatus> f2 = async(launch::async, &DDSChannel::chmod, &writer, ReadWrite);

        EXPECT_NE(f1.get(), CS_Success);
        EXPECT_NE(f2.get(), CS_Success);
    }

    vector<string> writeAfterBlockRead()
    {
        vector<string> data(2, "");
        future<ChannelStatus> f1 = async(launch::async, &DDSChannel::read, &reader, ref(data[0]), true);
        future<ChannelStatus> f2 = async(launch::async, &DDSChannel::read, &reader, ref(data[1]), true);

        const auto w = [this](const std::string& s) { writer.write(s); };
        thread writerThread1(w, "one");
        thread writerThread2(w, "two");
        writerThread1.join();
        writerThread2.join();
        f1.get();
        f2.get();
        return data;
    }

    bool writeLargeData()
    {
        vector<string> sendData(2, "");
        vector<string> receiveData(2, "");
        for (unsigned int i = 0; i < sendData.size(); ++i) {
            for (auto j = 0; j < 5000; ++j) {
                sendData[i].push_back(static_cast<char>(i));
            }
        }

        const auto w = [this](const std::string& s) { writer.write(s); };
        thread writerThread1(w, sendData[0]);
        thread writerThread2(w, sendData[1]);
        writerThread1.join();
        writerThread2.join();

        future<ChannelStatus> f1 = async(launch::async, &DDSChannel::read, &reader, ref(receiveData[0]), true);
        future<ChannelStatus> f2 = async(launch::async, &DDSChannel::read, &reader, ref(receiveData[1]), true);

        f1.get();
        f2.get();

        for (unsigned int i = 0; i < receiveData.size(); ++i) {
            if (find(receiveData.begin(), receiveData.end(), sendData[i]) == receiveData.end()) {
                return false;
            }
        }

        return true;
    }

    DDSChannel reader, writer;
};

TEST_F(DDSChannelInitializationTest, InitializeCorrectQueueDepth)
{
    unsigned long currentDepth, maxDepth;
    reader.getQueueProperties(currentDepth, maxDepth);
    EXPECT_EQ(currentDepth, 0);
    EXPECT_EQ(maxDepth, 100);
}

TEST_F(DDSChannelInitializationTest, ReaderConnect)
{
    auto err = reader.connect();
    ASSERT_EQ(err, CS_Success);

    auto context = queryContext(0, "ReadTest", reader);
    ASSERT_TRUE(context.dp != dds::core::null);
    ASSERT_TRUE(context.topic != dds::core::null);
    ASSERT_TRUE(context.sub != dds::core::null);
    ASSERT_TRUE(context.bReader != dds::core::null);
}

TEST_F(DDSChannelInitializationTest, WriterConnect)
{
    auto err = writer.connect();
    ASSERT_EQ(err, CS_Success);

    auto context = queryContext(1, "WriteTest", writer);
    ASSERT_TRUE(context.dp != dds::core::null);
    ASSERT_TRUE(context.topic != dds::core::null);
    ASSERT_TRUE(context.pub != dds::core::null);
    ASSERT_TRUE(context.bWriter != dds::core::null);
}

TEST_F(DDSChannelInitializationTest, ReadWriteConnect)
{
    auto err = rw_er.connect();
    ASSERT_EQ(err, CS_Success);

    auto context = queryContext(2, "RWChannel", rw_er);
    ASSERT_TRUE(context.dp != dds::core::null);
    ASSERT_TRUE(context.topic != dds::core::null);
    ASSERT_TRUE(context.pub != dds::core::null);
    ASSERT_TRUE(context.sub != dds::core::null);
    ASSERT_TRUE(context.bWriter != dds::core::null);
    ASSERT_TRUE(context.bReader != dds::core::null);
}

TEST_F(DDSChannelInitializationTest, Disconnect)
{
    auto err = rw_er.connect();
    ASSERT_EQ(err, CS_Success);

    err = rw_er.disconnect();
    ASSERT_EQ(err, CS_Disconnected);
    // reader and writer should be released
    auto context = queryContext(2, "RWChannel", rw_er);
    ASSERT_TRUE(context.dp != dds::core::null);
    ASSERT_TRUE(context.topic != dds::core::null);
    ASSERT_TRUE(context.pub != dds::core::null);
    ASSERT_TRUE(context.sub != dds::core::null);
    ASSERT_TRUE(context.bWriter == dds::core::null);
    ASSERT_TRUE(context.bReader == dds::core::null);
}

TEST_F(DDSChannelInitializationTest, ChmodFromReadToWrite)
{
    auto err = reader.connect();
    ASSERT_EQ(err, CS_Success);

    reader.chmod(Write);
    err = reader.connect();
    ASSERT_EQ(err, CS_Success);

    auto context = queryContext(0, "ReadTest", reader);
    ASSERT_TRUE(context.dp != dds::core::null);
    ASSERT_TRUE(context.topic != dds::core::null);
    ASSERT_TRUE(context.pub != dds::core::null);
    ASSERT_TRUE(context.bWriter != dds::core::null);
}

// Added to provide coverage for the case where we call createReader()
// a second time.
TEST_F(DDSChannelInitializationTest, ChmodFromReadToRead)
{
    EXPECT_EQ(CS_Success, reader.connect());
    reader.chmod(Read);
}

// Added to provide coverage for the case where we call createWriter()
// a second time.
TEST_F(DDSChannelInitializationTest, ChmodFromWriteToWrite)
{
    EXPECT_EQ(CS_Success, writer.connect());
    writer.chmod(Write);
}

TEST_F(DDSChannelFunctionalTest, InvalidRead)
{
    auto recvMsg = ""s;
    auto err = writer.read(recvMsg);
    ASSERT_EQ(err, CS_Error);
}

TEST_F(DDSChannelFunctionalTest, InvalidWrite)
{
    auto err = reader.write("Hello World");
    ASSERT_EQ(err, CS_Error);
}

TEST_F(DDSChannelFunctionalTest, ReadNoData)
{
    auto recvMsg = ""s;
    auto status = reader.read(recvMsg);
    EXPECT_EQ(recvMsg, ""s);
    EXPECT_EQ(status, CS_Success);
}

TEST_F(DDSChannelFunctionalTest, BlockingRead)
{
    auto sendMsg = "Hello World"s;
    auto data = DDSComms::createMsg(sendMsg);
    mockWriter.write(data);

    auto recvMsg = ""s;
    auto status = reader.read(recvMsg, true);
    ASSERT_EQ(status, CS_NewData);
    ASSERT_EQ(recvMsg, "Hello World");
}

TEST_F(DDSChannelFunctionalTest, BlockingReadWithCallback)
{
    std::vector<task::stats::MessageReport> reports{};
    const auto startTime{av::timePointFromOtimePoint(SystemClock::getSystemTime())};
    const task::stats::MessageCallback callback{
        [&reports](const task::stats::MessageReport& report) { reports.push_back(report); }};

    task::Channel::CallbackRegistration registration{callback};

    const auto sendMsg = "Hello World"s;
    auto data = DDSComms::createMsg(sendMsg);
    mockWriter.write(data);

    auto recvMsg = ""s;
    auto status = reader.read(recvMsg, true);
    EXPECT_EQ(status, CS_NewData);
    EXPECT_EQ(recvMsg, "Hello World");

    EXPECT_EQ(reports.size(), 1);
    ASSERT_GE(reports.size(), 1);

    EXPECT_EQ(reports[0].channelName, reader.getName());
    EXPECT_TRUE(reports[0].eventType == task::stats::MessageReport::EventType::Received);
    EXPECT_EQ(reports[0].messageSize, sendMsg.size());
    EXPECT_TRUE(reports[0].timestamp >= startTime);
}

TEST_F(DDSChannelFunctionalTest, BlockingReadWriteWithSeparateCallbacks)
{
    std::vector<task::stats::MessageReport> readReports{};
    const task::stats::MessageCallback readCallback{[&readReports](const task::stats::MessageReport& report) {
        if (report.eventType == task::stats::MessageReport::EventType::Received) {
            readReports.push_back(report);
        }
    }};

    std::vector<task::stats::MessageReport> writeReports{};
    const task::stats::MessageCallback writeCallback{[&writeReports](const task::stats::MessageReport& report) {
        if (report.eventType == task::stats::MessageReport::EventType::Sent) {
            writeReports.push_back(report);
        }
    }};

    task::Channel::CallbackRegistration readRegistration{readCallback};
    task::Channel::CallbackRegistration writeRegistration{writeCallback};

    const auto startTime{av::timePointFromOtimePoint(SystemClock::getSystemTime())};

    const auto readMsg = "Read World"s;
    const auto writeMsg = "Write World"s;

    {
        auto data = DDSComms::createMsg(readMsg);
        mockWriter.write(data);

        auto recvMsg = ""s;
        auto status = reader.read(recvMsg, true);
        EXPECT_EQ(status, CS_NewData);
        EXPECT_EQ(recvMsg, "Read World");
    }

    writer.write(writeMsg);

    EXPECT_EQ(readReports.size(), 1);
    ASSERT_GE(readReports.size(), 1);
    EXPECT_EQ(writeReports.size(), 1);
    ASSERT_GE(writeReports.size(), 1);

    EXPECT_EQ(readReports[0].channelName, reader.getName());
    EXPECT_TRUE(readReports[0].eventType == task::stats::MessageReport::EventType::Received);
    EXPECT_EQ(readReports[0].messageSize, readMsg.size());
    EXPECT_TRUE(readReports[0].timestamp >= startTime);
    EXPECT_TRUE(readReports[0].latencyns);
    EXPECT_TRUE(readReports[0].sendTime);
    EXPECT_TRUE(*(readReports[0].latencyns) == (readReports[0].timestamp - (*(readReports[0].sendTime))));

    EXPECT_EQ(writeReports[0].channelName, writer.getName());
    EXPECT_TRUE(writeReports[0].eventType == task::stats::MessageReport::EventType::Sent);
    EXPECT_EQ(writeReports[0].messageSize, writeMsg.size());
    EXPECT_TRUE(writeReports[0].timestamp >= startTime);
}

TEST_F(DDSChannelFunctionalTestNoConnect, ConnectDisconnectWithCallback)
{
    std::vector<task::stats::MessageReport> reports{};
    const auto startTime{av::timePointFromOtimePoint(SystemClock::getSystemTime())};
    const task::stats::MessageCallback callback{
        [&reports](const task::stats::MessageReport& report) { reports.push_back(report); }};

    task::Channel::CallbackRegistration registration{callback};

    reader.connect();
    writer.connect();

    writer.disconnect();
    reader.disconnect();

    EXPECT_EQ(reports.size(), 4);
    ASSERT_GE(reports.size(), 4);

    EXPECT_TRUE(reports[0].eventType == task::stats::MessageReport::EventType::Connected);
    EXPECT_EQ(reports[0].channelName, "ReadTest");
    EXPECT_TRUE(reports[1].eventType == task::stats::MessageReport::EventType::Connected);
    EXPECT_EQ(reports[1].channelName, "WriteTest");
    EXPECT_TRUE(reports[2].eventType == task::stats::MessageReport::EventType::Disconnected);
    EXPECT_EQ(reports[2].channelName, "WriteTest");
    EXPECT_TRUE(reports[3].eventType == task::stats::MessageReport::EventType::Disconnected);
    EXPECT_EQ(reports[3].channelName, "ReadTest");
}

TEST_F(DDSChannelFunctionalTest, Unblock)
{
    auto blockingRead = async(launch::async, [this]() {
        auto recvMsg = ""s;
        return reader.read(recvMsg, true);
    });
    auto status = blockingRead.wait_for(1000ms);
    reader.unblock();

    ASSERT_TRUE(status == future_status::timeout);
    ASSERT_EQ(blockingRead.get(), CS_Success);
}

TEST_F(DDSChannelFunctionalTest, Write)
{
    std::vector<task::stats::MessageReport> reports{};
    const auto startTime{av::timePointFromOtimePoint(SystemClock::getSystemTime())};
    const task::stats::MessageCallback callback{
        [&reports](const task::stats::MessageReport& report) { reports.push_back(report); }};

    task::Channel::CallbackRegistration registration{callback};

    auto readAsync = async(launch::async, [this]() {
        dds::sub::cond::ReadCondition readCond(mockReader, dds::sub::status::DataState::new_data());
        dds::core::cond::WaitSet waitSet;
        waitSet += readCond;
        waitSet.dispatch();
        auto samples = mockReader.select().take();
        if (samples.length() > 0) {
            const auto& data = samples.begin()->data().payload();
            return string(data.begin(), data.end());
        }
        return ""s;
    });

    const std::string msg{"Hello World"};
    writer.write(msg);
    auto status = readAsync.wait_for(1000ms);
    ASSERT_TRUE(status == future_status::ready);
    EXPECT_EQ(readAsync.get(), msg);

    EXPECT_EQ(reports.size(), 1);
    ASSERT_GE(reports.size(), 1);

    EXPECT_EQ(reports[0].channelName, writer.getName());
    EXPECT_TRUE(reports[0].eventType == task::stats::MessageReport::EventType::Sent);
    EXPECT_EQ(reports[0].messageSize, msg.size());
    EXPECT_TRUE(reports[0].timestamp >= startTime);
}

TEST_F(DDSChannelFunctionalTest, QueryCorrectQueueProperty)
{
    unsigned long currentDepth, maxDepth;
    reader.getQueueProperties(currentDepth, maxDepth);
    EXPECT_EQ(currentDepth, 0);
    EXPECT_EQ(maxDepth, 100);
}

TEST_F(DDSChannelFunctionalTest, Clear)
{
    unsigned long currentDepth, maxDepth;
    reader.clear();
    reader.getQueueProperties(currentDepth, maxDepth);
    EXPECT_EQ(currentDepth, 0);
    EXPECT_EQ(maxDepth, 100);
}

TEST_F(DDSChannelMultiThreadingTest, MultiThreadReadWrite)
{
    multipleWrite();
    vector<string> data = multipleRead();
    EXPECT_TRUE(find(data.begin(), data.end(), "one") != data.end());
    EXPECT_TRUE(find(data.begin(), data.end(), "two") != data.end());
    EXPECT_TRUE(find(data.begin(), data.end(), "three") != data.end());
    EXPECT_TRUE(find(data.begin(), data.end(), "four") != data.end());
    EXPECT_TRUE(find(data.begin(), data.end(), "five") != data.end());
}

TEST_F(DDSChannelMultiThreadingTest, DisconnectWhileBlockRead)
{
    ChannelStatus status = disconnectWhileBlockRead();
    EXPECT_EQ(status, CS_Disconnected);
}

TEST_F(DDSChannelMultiThreadingTest, ReadAfterDisconnected)
{
    ChannelStatus status = readAfterDisconnected();
    EXPECT_EQ(status, CS_Error);
}

TEST_F(DDSChannelMultiThreadingTest, WriteAfterDisconnected)
{
    ChannelStatus status = writeAfterDisconnected();
    EXPECT_EQ(status, CS_Error);
}

TEST_F(DDSChannelMultiThreadingTest, DoubleDisconnected)
{
    ChannelStatus status = doubleDisconnected();
    EXPECT_EQ(status, CS_Disconnected);
}

TEST_F(DDSChannelMultiThreadingTest, DoubleConnected)
{
    ChannelStatus status = doubleConnected();
    EXPECT_EQ(status, CS_Success);
}

TEST_F(DDSChannelMultiThreadingTest, WriteAfterBlockRead)
{
    vector<string> data = writeAfterBlockRead();
    EXPECT_TRUE(find(data.begin(), data.end(), "one") != data.end());
    EXPECT_TRUE(find(data.begin(), data.end(), "two") != data.end());
}

TEST_F(DDSChannelMultiThreadingTest, WriteLargeData)
{
    EXPECT_TRUE(writeLargeData());
}

TEST_F(DDSChannelMultiThreadingTest, WriteAfterBlockReadWithCallback)
{
    std::vector<task::stats::MessageReport> reports{};
    std::mutex reportsMutex{};

    const task::stats::MessageCallback callback{[&reports, &reportsMutex](const task::stats::MessageReport& report) {
        std::unique_lock<std::mutex> lock(reportsMutex);
        reports.push_back(report);
    }};

    task::Channel::CallbackRegistration registration{callback};

    vector<string> data = writeAfterBlockRead();
    EXPECT_TRUE(find(data.begin(), data.end(), "one") != data.end());
    EXPECT_TRUE(find(data.begin(), data.end(), "two") != data.end());

    EXPECT_EQ(reports.size(), 4);
    ASSERT_GE(reports.size(), 4);

    int numSent = 0;
    int numReceived = 0;
    for (size_t i = 0; i < 4; ++i) {
        EXPECT_EQ(reports[i].channelName, reader.getName());

        if ((reports[i].eventType == task::stats::MessageReport::EventType::Sent) ||
            (reports[i].eventType == task::stats::MessageReport::EventType::Received)) {
            ++((reports[i].eventType == task::stats::MessageReport::EventType::Sent) ? numSent : numReceived);

            EXPECT_EQ(reports[i].messageSize, 3);

            // We previously checked that the timestamps were in
            // order, but that's a race condition. There are two
            // sending threads. Let's say thread A sends before thread
            // B, so timestamp_B >= timestamp_A. But then thread A
            // might get unscheduled, while thread B continues. Since
            // the message-sent callbacks are in those threads,
            // messageReport_B will then be in the queue before
            // messageReport_A.
        }
    }

    EXPECT_EQ(numSent, 2);
    EXPECT_EQ(numReceived, 2);
}

TEST_F(DDSChannelMultiThreadingTest, WriteAfterBlockReadWithCallbackSometimes)
{
    std::vector<task::stats::MessageReport> reports{};
    std::mutex reportsMutex{};

    const auto startTime{av::timePointFromOtimePoint(SystemClock::getSystemTime())};
    const task::stats::MessageCallback callback{[&reports, &reportsMutex](const task::stats::MessageReport& report) {
        std::unique_lock<std::mutex> lock(reportsMutex);
        reports.push_back(report);
    }};

    {
        task::Channel::CallbackRegistration registration{callback};

        vector<string> data = writeAfterBlockRead();
        EXPECT_TRUE(find(data.begin(), data.end(), "one") != data.end());
        EXPECT_TRUE(find(data.begin(), data.end(), "two") != data.end());
    }

    // Now do it again, but with no callback; we'll then check that we
    // got only the callbacks expected from the first wave.
    {
        vector<string> data = writeAfterBlockRead();
        EXPECT_TRUE(find(data.begin(), data.end(), "one") != data.end());
        EXPECT_TRUE(find(data.begin(), data.end(), "two") != data.end());
    }

    auto prevTime = startTime;
    EXPECT_EQ(reports.size(), 4);
    ASSERT_GE(reports.size(), 4);

    int numSent = 0;
    int numReceived = 0;
    for (size_t i = 0; i < 4; ++i) {
        EXPECT_EQ(reports[i].channelName, reader.getName());

        if ((reports[i].eventType == task::stats::MessageReport::EventType::Sent) ||
            (reports[i].eventType == task::stats::MessageReport::EventType::Received)) {
            ++((reports[i].eventType == task::stats::MessageReport::EventType::Sent) ? numSent : numReceived);

            EXPECT_EQ(reports[i].messageSize, 3);

            const auto curTime = reports[i].timestamp;
            EXPECT_TRUE(curTime >= prevTime);
            prevTime = curTime;
        }
    }

    EXPECT_EQ(numSent, 2);
    EXPECT_EQ(numReceived, 2);
}

TEST_F(DDSChannelMultiThreadingTest, WriteAfterBlockReadWithMultipleCallbacks)
{
    std::vector<task::stats::MessageReport> reports{};
    std::mutex reportsMutex{};

    const task::stats::MessageCallback callback{[&reports, &reportsMutex](const task::stats::MessageReport& report) {
        std::unique_lock<std::mutex> lock(reportsMutex);
        reports.push_back(report);
    }};

    size_t totalSize{0};
    size_t messageCount{0};

    const task::stats::MessageCallback statsCallback{
        [&messageCount, &reportsMutex, &totalSize](const task::stats::MessageReport& report) {
            if (report.eventType == task::stats::MessageReport::EventType::Received) {
                std::unique_lock<std::mutex> lock(reportsMutex);
                totalSize += report.messageSize;
                ++messageCount;
            }
        }};

    const auto startTime{av::timePointFromOtimePoint(SystemClock::getSystemTime())};

    task::Channel::CallbackRegistration registration{callback};
    task::Channel::CallbackRegistration statsRegistration{statsCallback};

    vector<string> data = writeAfterBlockRead();
    EXPECT_TRUE(find(data.begin(), data.end(), "one") != data.end());
    EXPECT_TRUE(find(data.begin(), data.end(), "two") != data.end());

    auto prevTime = startTime;
    EXPECT_EQ(reports.size(), 4);
    ASSERT_GE(reports.size(), 4);

    EXPECT_EQ(messageCount, 2);
    EXPECT_EQ(totalSize, 6);

    int numSent = 0;
    int numReceived = 0;
    for (size_t i = 0; i < 4; ++i) {
        EXPECT_EQ(reports[i].channelName, reader.getName());

        if ((reports[i].eventType == task::stats::MessageReport::EventType::Sent) ||
            (reports[i].eventType == task::stats::MessageReport::EventType::Received)) {
            ++((reports[i].eventType == task::stats::MessageReport::EventType::Sent) ? numSent : numReceived);

            EXPECT_EQ(reports[i].messageSize, 3);

            const auto curTime = reports[i].timestamp;
            EXPECT_TRUE(curTime >= prevTime);
            prevTime = curTime;
        }
    }

    EXPECT_EQ(numSent, 2);
    EXPECT_EQ(numReceived, 2);
}
