/**
 * \copyright
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief Interface unit tests
 */

#include <memory>

#include <gmock/gmock-more-actions.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <Comms/InterfaceUtils.h>
#include <Comms/TypedChannel.h>
#include <ODatum/ODatum.h>
#include <interfaces/SampleMessage/SampleMessage.h>
#include <task/Task.h>

#include <TestUtils/CopyInterfaces.hpp>

using namespace ::testing;

class InterfaceTestType : public task::Interface<av::msg::SampleMessage>
{
public:
    InterfaceTestType(ConfigSection& cs, const task::Access& acc)
        : task::Interface<av::msg::SampleMessage>(cs, acc)
    {}

    virtual bool initialize(void) override { return true; }

    virtual void cleanup(void) override {}
};

class ConfigSectionTest : public ConfigSection
{
public:
    ConfigSectionTest()
        : ConfigSection()
    {}

    // Make init() public
    bool init(const std::string& name, ConfigPtr& cfg) { return ConfigSection::init(name, cfg); }
};

TEST(GetInterfaceTests, GivenEmptyConfigGetInterfaceReturnsNullptr)
{
    ConfigSection emptyConfigSection{};
    EXPECT_EQ(nullptr,
              comms::getInterface<InterfaceTestType>("TestInterfaceName", emptyConfigSection, task::Access::Read));
    EXPECT_EQ(nullptr,
              comms::getInterface<InterfaceTestType>("TestInterfaceName", emptyConfigSection, task::Access::Write));
    EXPECT_EQ(nullptr,
              comms::getInterface<InterfaceTestType>("TestInterfaceName", emptyConfigSection, task::Access::ReadWrite));
    EXPECT_EQ(nullptr,
              comms::getInterface<InterfaceTestType>("TestInterfaceName", emptyConfigSection, task::Access::Standby));
}

TEST(GetInterfaceTests, GivenConfigWithChannelGetInterfaceReturnsInterfaceWithChannel)
{
    std::string testInterfaceConfig{"TestInterfaceName ="
                                    "{"
                                    "interface = \"TestInterfaceName\";"
                                    "channel ="
                                    "{"
                                    "name = \"testName\";"
                                    "type = \"LOG\";"
                                    "};"
                                    "};"};
    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testInterfaceConfig);
    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("TestInterfaceName", configPtr));

    // Creating interface with channel in config should succeed
    auto interfaceInstance =
        comms::getInterface<InterfaceTestType>("TestInterfaceName", configTest, task::Access::Read);
    ASSERT_NE(nullptr, interfaceInstance);
    // And the interface will have a channel
    EXPECT_TRUE(interfaceInstance->hasChannel());
}

TEST(GetInterfaceTests, GivenConfigWithNoChannelGetInterfaceReturnsInterfaceWithNoChannel)
{
    std::string testInterfaceConfig{"TestInterfaceName ="
                                    "{"
                                    "interface = \"TestInterfaceName\";"
                                    "};"};
    auto configPtr{std::make_shared<Config>()};
    configPtr->readString(testInterfaceConfig);
    ConfigSectionTest configTest{};
    ASSERT_TRUE(configTest.init("TestInterfaceName", configPtr));

    // Creating interface with no channel in config should succeed
    auto interfaceInstance =
        comms::getInterface<InterfaceTestType>("TestInterfaceName", configTest, task::Access::Read);
    ASSERT_NE(nullptr, interfaceInstance);

    // But the interface will have no channel
    EXPECT_FALSE(interfaceInstance->hasChannel());
}

TEST(ChannelTypeNameTests, GivenValidChannelTypeCorrectResult)
{
    const std::vector<std::pair<comms::ChannelType, std::string>> pairs = {
        {comms::ChannelType::SimpleComms, "SimpleComms"},
        {comms::ChannelType::DDS, "DDS"},
        {comms::ChannelType::SQL_LOG, "LOG"}};
    for (const auto& params : pairs) {
        std::optional<std::string> strOpt{comms::channelTypeName(params.first)};
        EXPECT_EQ(strOpt, params.second);
    }
}

TEST(ChannelTypeNameTests, GivenInvalidChannelNullOpt)
{
    std::optional<std::string> strOpt{comms::channelTypeName(static_cast<comms::ChannelType>(42))};
    EXPECT_FALSE(strOpt.has_value());
}

TEST(ChannelTypeEnumTests, GivenValidChannelTypeCorrectResult)
{
    const std::vector<std::pair<comms::ChannelType, std::string>> pairs = {
        {comms::ChannelType::SimpleComms, "SimpleComms"},
        {comms::ChannelType::DDS, "DDS"},
        {comms::ChannelType::SQL_LOG, "LOG"}};
    for (const auto& params : pairs) {
        std::optional<comms::ChannelType> ctOpt{comms::channelTypeEnum(params.second)};
        EXPECT_EQ(ctOpt, params.first);
    }
}
TEST(ChannelTypeEnumTests, GivenInvalidChannelNullOpt)
{
    std::optional<comms::ChannelType> ctOpt{comms::channelTypeEnum("InvalidName")};
    EXPECT_FALSE(ctOpt.has_value());
}

/// This test is disabled because it relies on the clock interface,
/// which isn't being built at the moment.
TEST(getExternallyDrivenClockTests, DISABLED_TestGetExternallyDriven)
{
    TestUtils::CopyInterfaces interfaceEnv{"TestInterfaceUtils"};
    EXPECT_TRUE(
        interfaceEnv.copyInterface("/Clock/Playback/Input/Channel", "/PlaybackClockParamsInput", "libChannel.so"));

    // After copying, succeeds
    Clock* result = getExternallyDrivenClock(comms::ChannelType::DDS);
    EXPECT_NE(result, nullptr);

    // Subsequent calls return the same pointer
    Clock* repeatresult = getExternallyDrivenClock(comms::ChannelType::DDS);
    EXPECT_EQ(result, repeatresult);
}

TEST(getExternallyDrivenClockTests, TestWithBogusType)
{
    // After copying, succeeds
    Clock* result = getExternallyDrivenClock(static_cast<comms::ChannelType>(10));
    EXPECT_EQ(result, nullptr);
}

TEST(getExternallyDrivenClockTests, TestSystemPlaybackClockConfigSection)
{
    // Valid ConfigSection
    const auto configSectionOpt = comms::getSystemPlaybackClockConfigSection("Test");
    EXPECT_TRUE(configSectionOpt.has_value());
    EXPECT_TRUE((*configSectionOpt).hasParameter("interface"));

    // Invalid ConfigSection
    const auto configSectionInvalidOpt = comms::getSystemPlaybackClockConfigSection("Test = { = }}");
    EXPECT_FALSE(configSectionInvalidOpt.has_value());
}
