/**
 * \copyright
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief Interface unit tests
 */

#include <memory>

#include <gmock/gmock-more-actions.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <Comms/LatencyStats.h>

using namespace ::testing;

using us_double = std::chrono::duration<double, std::micro>;

TEST(LatencyStatsTests, GivenEmptyLSAllZeros)
{
    LatencyStats ls{"test", -1, "", "", -1};
    static_cast<void>(ls.getResults());

    EXPECT_EQ(ls.latencyInfo_.count, 0);
    EXPECT_EQ(ls.latencyInfo_.numLatencyWarnings, 0);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.mean).count(), static_cast<us_double>(0.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.stdDev).count(), static_cast<us_double>(0.0).count());
}

TEST(LatencyStatsTests, SingleEntryEasy)
{
    LatencyStats ls{"test", -1, "", "", -1};
    EXPECT_EQ(ls.update(static_cast<microseconds_long_double>(1.0)), LatStatsError::LAT_STATS_OK);
    static_cast<void>(ls.getResults());

    EXPECT_EQ(ls.latencyInfo_.count, 1);
    EXPECT_EQ(ls.latencyInfo_.numLatencyWarnings, 0);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.mean).count(), static_cast<us_double>(1.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.stdDev).count(), static_cast<us_double>(0.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.minLatency).count(), static_cast<us_double>(1.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.maxLatency).count(), static_cast<us_double>(1.0).count());
}

TEST(LatencyStatsTests, NegativeEntryRejected)
{
    LatencyStats ls{"test", -1, "", "", -1};
    EXPECT_EQ(ls.update(static_cast<microseconds_long_double>(-1.0)), LatStatsError::LAT_STATS_NEG_VAL);
    static_cast<void>(ls.getResults());

    EXPECT_EQ(ls.latencyInfo_.count, 0);
    EXPECT_EQ(ls.latencyInfo_.numLatencyWarnings, 0);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.mean).count(), static_cast<us_double>(0.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.stdDev).count(), static_cast<us_double>(0.0).count());
}

TEST(LatencyStatsTests, MultipleEntriesCalculateStats)
{
    LatencyStats ls{"test", -1, "", "", -1};
    EXPECT_EQ(ls.update(static_cast<microseconds_long_double>(1.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.update(static_cast<microseconds_long_double>(2.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.update(static_cast<microseconds_long_double>(3.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.update(static_cast<microseconds_long_double>(4.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.update(static_cast<microseconds_long_double>(5.0)), LatStatsError::LAT_STATS_OK);
    static_cast<void>(ls.getResults());

    EXPECT_EQ(ls.latencyInfo_.count, 5);
    EXPECT_EQ(ls.latencyInfo_.numLatencyWarnings, 0);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.mean).count(), static_cast<us_double>(3.0).count());
    EXPECT_NEAR(static_cast<double>(ls.latencyInfo_.stdDev.count()), 1.5811388301, 0.0000001);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.minLatency).count(), static_cast<us_double>(1.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.maxLatency).count(), static_cast<us_double>(5.0).count());
}

TEST(BufferedLatencyStatsTests, GivenEmptyLSAllZeros)
{
    BufferedLatencyStats ls{"test", -1, "", "", -1};
    static_cast<void>(ls.getResults());

    EXPECT_EQ(ls.latencyInfo_.count, 0);
    EXPECT_EQ(ls.latencyInfo_.numLatencyWarnings, 0);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.mean).count(), static_cast<us_double>(0.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.stdDev).count(), static_cast<us_double>(0.0).count());
}

TEST(BufferedLatencyStatsTests, SingleEntryEasy)
{
    BufferedLatencyStats ls{"test", -1, "", "", -1};
    EXPECT_EQ(ls.addStartTime(static_cast<microseconds_long_double>(0.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.setEndTime(static_cast<microseconds_long_double>(1.0)), LatStatsError::LAT_STATS_OK);
    static_cast<void>(ls.getResults());

    EXPECT_EQ(ls.latencyInfo_.count, 1);
    EXPECT_EQ(ls.latencyInfo_.numLatencyWarnings, 0);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.mean).count(), static_cast<us_double>(1.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.stdDev).count(), static_cast<us_double>(0.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.minLatency).count(), static_cast<us_double>(1.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.maxLatency).count(), static_cast<us_double>(1.0).count());
}

TEST(BufferedLatencyStatsTests, NegativeEntryRejected)
{
    BufferedLatencyStats ls{"test", -1, "", "", -1};
    EXPECT_EQ(ls.addStartTime(static_cast<microseconds_long_double>(1.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.setEndTime(static_cast<microseconds_long_double>(0.0)), LatStatsError::LAT_STATS_NEG_VAL);
    static_cast<void>(ls.getResults());

    EXPECT_EQ(ls.latencyInfo_.count, 0);
    EXPECT_EQ(ls.latencyInfo_.numLatencyWarnings, 0);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.mean).count(), static_cast<us_double>(0.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.stdDev).count(), static_cast<us_double>(0.0).count());
}

TEST(BufferedLatencyStatsTests, MultipleEntriesCalculateStats)
{
    BufferedLatencyStats ls{"test", -1, "", "", -1};
    EXPECT_EQ(ls.addStartTime(static_cast<microseconds_long_double>(1.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.addStartTime(static_cast<microseconds_long_double>(2.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.addStartTime(static_cast<microseconds_long_double>(3.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.addStartTime(static_cast<microseconds_long_double>(4.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.addStartTime(static_cast<microseconds_long_double>(5.0)), LatStatsError::LAT_STATS_OK);
    EXPECT_EQ(ls.setEndTime(static_cast<microseconds_long_double>(6.0)), LatStatsError::LAT_STATS_OK);
    static_cast<void>(ls.getResults());

    EXPECT_EQ(ls.latencyInfo_.count, 5);
    EXPECT_EQ(ls.latencyInfo_.numLatencyWarnings, 0);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.mean).count(), static_cast<us_double>(3.0).count());
    EXPECT_NEAR(static_cast<double>(ls.latencyInfo_.stdDev.count()), 1.5811388301, 0.0000001);
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.minLatency).count(), static_cast<us_double>(1.0).count());
    EXPECT_DOUBLE_EQ(static_cast<us_double>(ls.latencyInfo_.maxLatency).count(), static_cast<us_double>(5.0).count());
}