/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file SCChannelTest.cc
 * @brief This file implements tests for the SCChannel class.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#include <algorithm>
#include <chrono>
#include <memory>
#include <string>
#include <thread>

#include <gmock/gmock-more-actions.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <SimpleComms/test/MockClient.h>
#include <TestUtils/SystemClockScopedInitializer.hpp>
#include <task/Task.h>

#include <Comms/SCChannel.h>
#include <Threading/PThreadLocker.h>

using namespace ::testing;

namespace SC_CHANNEL_TEST {

/**
 * The reason for creating this new SCChannel is to use a protected function in task::SCChannel,
 * which is "updateMessageQueues(void)".
 * Why we do that?
 * The answer is: there is a thread in task::SCChannel to update internal message queue periodly;
 *   when Jenkins or local pc run a lot of tasks,
 *   this message-queue thread may not get any chance to run within "waitForSendRcvThreadsUpdate()" of time,
 *   which cause failure of this unit test;
 *   in order to prevent flakiness of this SCChannel unit test,
 *   we call a protected function in task::SCChannel, "updateMessageQueues(void)", which guarantee the message queue
 * will get updated.
 */
class SCChannel : public task::SCChannel
{
public:
    SCChannel(const std::shared_ptr<SimpleComms::Client> comms,
              const std::string channelName,
              const task::Access mode,
              const uint32_t recvQueueDepth)
        : task::SCChannel(comms, channelName, mode, recvQueueDepth)
    {}

    void forceUpdateMsgQueues()
    {
        {
            PThreadLocker sendQLock(this->sendQMutex_);
            this->updateSendQueues();
        }
        {
            PThreadLocker channelsLock(channelsMutex_);
            this->updateMessageQueues();
        }
    }
};
} // namespace SC_CHANNEL_TEST

/**
 * @brief Group of initialization tests in the write mode.
 */
class SCChannelTestBase
{
public:
    std::shared_ptr<StrictMock<SimpleComms::MockClient>> mockSCSPtr_ =
        std::make_shared<StrictMock<SimpleComms::MockClient>>();

    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> createChannelAndSetExpectations(const std::string& channelName,
                                                                                const task::Access mode,
                                                                                const uint32_t recvQueueDepth,
                                                                                bool expect_connect_disconnect)
    {
        ON_CALL(*(mockSCSPtr_.get()), declare(_, _, _)).WillByDefault(Return(SimpleComms::Status::Ok));
        ON_CALL(*(mockSCSPtr_.get()), undeclare(_, _)).WillByDefault(Return(SimpleComms::Status::Ok));
        ON_CALL(*(mockSCSPtr_.get()), receiveMsg(_, _, _)).WillByDefault(Return(SimpleComms::Status::Ok));
        ON_CALL(*(mockSCSPtr_.get()), sendMsgV(_, _, _)).WillByDefault(Return(SimpleComms::Status::Ok));

        // Unless otherwise specified in the test cases, hasMsg() always returns false.
        // If you override this expectation, you must set an expectation for receiveMsg first.
        // Otherwise, receiveMsg() will be called with no expectation and will throw an error.
        mockSCSPtr_->setHasMessages(false);
        EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(0));

        // For some this will be boilerplate, for others it will be a focus. Let's allow that distinction.
        if (expect_connect_disconnect) {
            EXPECT_CALL(*(mockSCSPtr_.get()), declare(channelName, _, _))
                .Times(1)
                .WillOnce(Return(SimpleComms::Status::Ok));

            EXPECT_CALL(*(mockSCSPtr_.get()), undeclare(channelName, _))
                .Times(1)
                .WillOnce(Return(SimpleComms::Status::Ok));
        }

        return std::make_unique<SC_CHANNEL_TEST::SCChannel>(
            std::dynamic_pointer_cast<SimpleComms::Client, StrictMock<SimpleComms::MockClient>>(mockSCSPtr_),
            channelName,
            mode,
            recvQueueDepth);
    }
};

/**
 *
 * @brief Group of initialization tests in the modes which can Read from the channel.
 */
class SCChannelReadInitTests
    : public ::testing::TestWithParam<task::Access>
    , public SCChannelTestBase
{
protected:
    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> testChannel_;
    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> testChannel2_;
    std::string channelName_;

    task::Access mode_;
    uint32_t recvQueueDepth_;

    void SetUp() override
    {
        SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeSteady);
        this->channelName_ = "TestChannel";
        mode_ = GetParam();
        recvQueueDepth_ = 1;

        // These are init tests, so specify declare/undeclare explicitly
        this->testChannel_ = createChannelAndSetExpectations(this->channelName_, mode_, recvQueueDepth_, false);
    }

    void TearDown() override {}
};

TEST_P(SCChannelReadInitTests, GivenChannelExistsWhenConnectAndDisconnectThenSubscribeUnsubscribe)
{
    EXPECT_CALL(*(mockSCSPtr_.get()), declare(this->channelName_, _, _))
        .Times(1)
        .WillOnce(Return(SimpleComms::Status::Ok));
    EXPECT_EQ(testChannel_->connect(), task::ChannelStatus::CS_Success);

    EXPECT_CALL(*(mockSCSPtr_.get()), undeclare(this->channelName_, _))
        .Times(1)
        .WillOnce(Return(SimpleComms::Status::Ok));
    testChannel_->disconnect();
}

TEST_P(SCChannelReadInitTests, GivenChannelExistsWhenConnectAndDisconnectThenSubscribeUnsubscribeTwoChannels)
{
    this->testChannel2_ = createChannelAndSetExpectations(this->channelName_, mode_, recvQueueDepth_, false);

    // Since the channels are connecting on the same mode, there should be ONE delare and ONE undeclare. NOT two!
    EXPECT_CALL(*(mockSCSPtr_.get()), declare(this->channelName_, _, _))
        .Times(1)
        .WillOnce(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), undeclare(this->channelName_, _))
        .Times(1)
        .WillOnce(Return(SimpleComms::Status::Ok));

    EXPECT_EQ(testChannel_->connect(), task::ChannelStatus::CS_Success);
    EXPECT_EQ(testChannel2_->connect(), task::ChannelStatus::CS_Success);

    testChannel_->disconnect();
    testChannel2_->disconnect();
}

void unblockTimeout(SC_CHANNEL_TEST::SCChannel* channel, uint64_t timeout_ms)
{
    std::this_thread::sleep_for(std::chrono::milliseconds(timeout_ms));
    channel->unblock();
}

TEST_P(SCChannelReadInitTests, GivenDisconnectedReadReturnsError)
{
    std::string msg{"The quick brown fox jumps over the lazy dog"};
    EXPECT_EQ(testChannel_->read(msg), task::ChannelStatus::CS_Error);
    EXPECT_EQ(msg, "The quick brown fox jumps over the lazy dog");

    // This timeout shouldn't need to be very long - 1s should be fine.
    std::thread unblockThread(unblockTimeout, testChannel_.get(), 1000);
    EXPECT_EQ(testChannel_->read(msg, true), task::ChannelStatus::CS_Error);
    EXPECT_EQ(msg, "The quick brown fox jumps over the lazy dog");
    unblockThread.join();
}

INSTANTIATE_TEST_SUITE_P(ReadAndReadWriteModeTests,
                         SCChannelReadInitTests,
                         ::testing::Values(task::Access::Read, task::Access::ReadWrite));

/**
 * @brief Group of tests relating to reading from a channel. These should pass for all modes which are allowed
 * to Read from the channel.
 */
class SCChannelSharedReadTests
    : public ::testing::TestWithParam<task::Access>
    , public SCChannelTestBase
{
protected:
    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> testChannel_;
    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> testChannel2_;
    std::string channelName_;
    uint32_t recvQueueDepth_;
    task::Access mode_;

    void SetUp() override
    {
        this->channelName_ = "TestChannel";
        this->mode_ = GetParam();
        this->recvQueueDepth_ = 5;

        this->testChannel_ = createChannelAndSetExpectations(this->channelName_, mode_, this->recvQueueDepth_, true);
    }

    void TearDown() override { this->testChannel_->disconnect(); }

    /**
     * @brief Helper function to test the behavior of the clear API.
     *
     * @param dataToPassIn Data to pass in to the receive thread.
     * @param rcvdQueueSizeExpectation Parameter to specify what we expecte the size of the queue to be
     * after filling in the passed in data. The reason it is not the same as the passed in data size
     * is because the queue depth can influence the queue size, so specifying it is better for the
     * purpose of testing.
     * @param expectedDataFromReadApi Data to check against being read out of the read API.
     */
    void clearApiCheck(int rcvdQueueSizeExpectation, SC_CHANNEL_TEST::SCChannel& channel)
    {
        unsigned long getCurrentDepth{0}, getMaxDepth{0};

        EXPECT_EQ(true, channel.getQueueProperties(getCurrentDepth, getMaxDepth));
        EXPECT_EQ(rcvdQueueSizeExpectation, getCurrentDepth);

        // Perform a clear
        channel.clear();

        // We expect no data to available in the queue now.
        EXPECT_EQ(true, channel.getQueueProperties(getCurrentDepth, getMaxDepth));
        EXPECT_EQ(0, getCurrentDepth);
    }

    /**
     * @brief Helper function to read out data from the channel and check it against the
     * expectation
     *
     * @param expectedDataFromReadApi Data to check against being read out of the read API.
     */
    void readChannelData(std::vector<std::string> const& expectedDataFromReadApi, SC_CHANNEL_TEST::SCChannel& channel)
    {
        // When we read from the channel, we should get back the messages in the same order
        // in which it was passed in.
        std::string msgVar;
        for (auto& dataToCheck : expectedDataFromReadApi) {
            channel.forceUpdateMsgQueues();
            task::ChannelStatus status = channel.read(msgVar, false);

            ASSERT_EQ(status, task::ChannelStatus::CS_NewData);
            EXPECT_EQ(dataToCheck, msgVar);
        }

        std::string emptyMessage;
        EXPECT_EQ(task::ChannelStatus::CS_Success, channel.read(emptyMessage, false));
        EXPECT_EQ("", emptyMessage);
    }
};

TEST_P(SCChannelSharedReadTests, GivenSingleEntryWhenChannelReadThenReturnsValue)
{
    // Populate the message we expect to receive from the channel.
    std::vector<std::string> dataToPassIn{"newData"};
    const auto expectedData{dataToPassIn};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
    mockSCSPtr_->setHasMessages(true);

    this->testChannel_->connect();

    this->readChannelData(expectedData, *(this->testChannel_.get()));
}

TEST_P(SCChannelSharedReadTests, GivenSingleEntryWhenChannelReadAfterClearThenReturnsNoValue)
{
    // Populate the message we expect to receive from the channel.
    std::vector<std::string> dataToPassIn{"newData"};
    const std::vector<std::string> expectedData{};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
    mockSCSPtr_->setHasMessages(true);

    this->testChannel_->connect();

    std::for_each(dataToPassIn.begin(), dataToPassIn.end(), [&](auto&& /*ignored*/) {
        this->testChannel_->forceUpdateMsgQueues();
    });

    this->clearApiCheck(static_cast<int>(dataToPassIn.size()), *(this->testChannel_.get()));
}

TEST_P(SCChannelSharedReadTests, GivenMultipleEntriesWhenChannelReadThenReturnsFifo)
{
    // Populate the messages we expect to receive from the channel.
    // The receive queue depth is set in the setup function.

    std::vector<std::string> dataToPassIn{"data1", "data2", "data3", "data4"};
    const auto expectedData{dataToPassIn};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(1)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(2)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(3)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
    mockSCSPtr_->setHasMessages(true);

    this->testChannel_->connect();

    this->readChannelData(expectedData, *(this->testChannel_.get()));
}

TEST_P(SCChannelSharedReadTests, GivenMultipleEntriesWhenChannelReadAfterClearThenReturnsNoValue)
{
    // Populate the message we expect to receive from the channel.
    std::vector<std::string> dataToPassIn{"data1", "data2", "data3", "data4"};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(1)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(2)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(3)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
    mockSCSPtr_->setHasMessages(true);

    this->testChannel_->connect();

    std::for_each(dataToPassIn.begin(), dataToPassIn.end(), [&](auto&& /*ignored*/) {
        this->testChannel_->forceUpdateMsgQueues();
    });

    this->clearApiCheck(static_cast<int>(dataToPassIn.size()), *(this->testChannel_.get()));
}

TEST_P(SCChannelSharedReadTests, GivenReceiveOverFilledWhenChannelReadThenNewestMessagesReturned)
{
    // Populate the messages we expect to receive from the channel. Fill in 2 messages more
    // than the message queue.

    std::vector<std::string> dataToPassIn{"data1", "data2", "data3", "data4", "data5", "data6", "data7"};
    std::vector<std::string> expectedData{};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(1)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(2)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(3)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(4)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(5)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(6)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
    mockSCSPtr_->setHasMessages(true);

    // The expected data starts from the extra entries past the beginning and fills in till the most recent message.
    expectedData.assign((dataToPassIn.begin() + 2), dataToPassIn.end());

    // Run the tests
    this->testChannel_->connect();

    this->readChannelData(expectedData, *(this->testChannel_.get()));
}

TEST_P(SCChannelSharedReadTests, GivenReceiveOverFilledWhenChannelReadAfterClearThenReturnsNoValue)
{
    // Populate the messages we expect to receive from the channel. Fill in 2 messages more
    // than the message queue.

    std::vector<std::string> dataToPassIn{"data1", "data2", "data3", "data4", "data5", "data6", "data7"};
    std::vector<std::string> expectedData{};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(1)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(2)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(3)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(4)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(5)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(6)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
    mockSCSPtr_->setHasMessages(true);
    // Run the tests
    this->testChannel_->connect();

    std::for_each(dataToPassIn.begin(), dataToPassIn.end(), [&](auto&& /*ignored*/) {
        this->testChannel_->forceUpdateMsgQueues();
    });

    this->clearApiCheck(static_cast<int>(this->recvQueueDepth_), *(this->testChannel_.get()));
}

TEST_P(SCChannelSharedReadTests, GivenEmptyRecvdQueueWhenGetQueuePropertiesThenRetTrueWithZeroDepth)
{
    unsigned long getCurrentDepth, getMaxDepth;
    unsigned long expectedCurrentDepth{0}, expectedMaxDepth{this->recvQueueDepth_};

    this->testChannel_->connect();

    EXPECT_EQ(true, this->testChannel_->getQueueProperties(getCurrentDepth, getMaxDepth));
    EXPECT_EQ(expectedCurrentDepth, getCurrentDepth);
    EXPECT_EQ(expectedMaxDepth, getMaxDepth);
}

TEST_P(SCChannelSharedReadTests, GivenOneDeepRecvdQueueWhenGetQueuePropertiesThenRetTrueWithDepthOfOne)
{
    SystemClockScopedInitializer clockInitializer{SystemClock::SystemTimeMode::TimeModeSteady};

    // Populate the message we expect to receive from the channel.
    std::vector<std::string> dataToPassIn{"newData"};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
    mockSCSPtr_->setHasMessages(true);

    unsigned long getCurrentDepth, getMaxDepth;
    unsigned long expectedCurrentDepth{1}, expectedMaxDepth{this->recvQueueDepth_};

    this->testChannel_->connect();

    std::for_each(dataToPassIn.begin(), dataToPassIn.end(), [&](auto&& /*ignored*/) {
        this->testChannel_->forceUpdateMsgQueues();
    });

    EXPECT_EQ(true, this->testChannel_->getQueueProperties(getCurrentDepth, getMaxDepth));
    EXPECT_EQ(expectedCurrentDepth, getCurrentDepth);
    EXPECT_EQ(expectedMaxDepth, getMaxDepth);
}

TEST_P(SCChannelSharedReadTests, GivenMultipleEntriesDeepRecvdQueueWhenGetQueuePropertiesThenRetTrueWithCorrectDepth)
{
    SystemClockScopedInitializer clockInitializer{SystemClock::SystemTimeMode::TimeModeSteady};

    // Populate the message we expect to receive from the channel.
    std::vector<std::string> dataToPassIn{"newData", "", "new_message_1"};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(1)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(2)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
    mockSCSPtr_->setHasMessages(true);

    unsigned long getCurrentDepth, getMaxDepth;
    unsigned long expectedCurrentDepth{3}, expectedMaxDepth{this->recvQueueDepth_};

    this->testChannel_->connect();

    std::for_each(dataToPassIn.begin(), dataToPassIn.end(), [&](auto&& /*ignored*/) {
        this->testChannel_->forceUpdateMsgQueues();
    });

    EXPECT_EQ(true, this->testChannel_->getQueueProperties(getCurrentDepth, getMaxDepth));
    EXPECT_EQ(expectedCurrentDepth, getCurrentDepth);
    EXPECT_EQ(expectedMaxDepth, getMaxDepth);
}

TEST_P(SCChannelSharedReadTests, GivenMoreEntriesThanMaxDepthInRecvdQueueWhenGetQueuePropertiesThenRetTrueWithMaxDepth)
{
    // Populate the messages we expect to receive from the channel. Fill in 2 messages more
    // than the message queue.

    std::vector<std::string> dataToPassIn{"data1", "data2", "data3", "data4", "data5", "data6", "data7"};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(1)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(2)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(3)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(4)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(5)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(6)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
    mockSCSPtr_->setHasMessages(true);

    // Check for queue property expectations - the received queue should return with the
    // max depth.
    unsigned long getCurrentDepth, getMaxDepth;
    unsigned long expectedMaxDepth{this->recvQueueDepth_};

    this->testChannel_->connect();

    std::for_each(dataToPassIn.begin(), dataToPassIn.end(), [&](auto&& /*ignored*/) {
        this->testChannel_->forceUpdateMsgQueues();
    });

    EXPECT_EQ(true, this->testChannel_->getQueueProperties(getCurrentDepth, getMaxDepth));
    EXPECT_EQ(expectedMaxDepth, getCurrentDepth);
    EXPECT_EQ(expectedMaxDepth, getMaxDepth);
}

TEST_P(SCChannelSharedReadTests, GivenMultipleEntriesWhenChannelReadThenReturnsFifoTwoChannels)
{
    // Populate the messages we expect to receive from the channel.
    // The receive queue depth is set in the setup function.
    std::vector<std::string> dataToPassIn{"data1", "data2", "data3", "data4"};
    const auto expectedData{dataToPassIn};

    EXPECT_CALL(*(mockSCSPtr_.get()), receiveMsg(this->channelName_, _, _))
        .Times(AtLeast(1))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(0)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(1)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(2)), Return(SimpleComms::Status::NewData)))
        .WillOnce(DoAll(SetArgReferee<1>(dataToPassIn.at(3)), Return(SimpleComms::Status::NewData)))
        .WillRepeatedly(Return(SimpleComms::Status::Ok));

    this->testChannel2_ = createChannelAndSetExpectations(this->channelName_, mode_, recvQueueDepth_, false);

    EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));

    this->testChannel_->connect();
    this->testChannel2_->connect();

    // Now it's safe for the channels to read data. If hasMsg() returned true before this,
    // it's possible that testChannel_ would receive data but not testChannel2_.
    mockSCSPtr_->setHasMessages(true);

    // SCChannel knows how many channel objects have been created for each channel,
    // and when it reads from the UnixClient it updates all their queues.
    this->readChannelData(expectedData, *(this->testChannel_.get()));
    this->readChannelData(expectedData, *(this->testChannel2_.get()));

    // The other channel disconnects in the TearDown() function.
    this->testChannel2_->disconnect();
}

INSTANTIATE_TEST_SUITE_P(ReadAndReadWriteModeRuntimeTests,
                         SCChannelSharedReadTests,
                         ::testing::Values(task::Access::Read, task::Access::ReadWrite));

/**
 * @brief Tests which are specific to only Read mode and not other modes which are capable of reading.
 */
class SCChannelReadModeSpecificTests
    : public ::testing::Test
    , public SCChannelTestBase
{
protected:
    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> testChannel_;
    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> testChannel2_;
    std::string channelName_;
    uint32_t recvQueueDepth_;

    void SetUp() override
    {
        this->channelName_ = "TestChannel";
        task::Access mode{task::Access::Read};
        this->recvQueueDepth_ = 5;

        this->testChannel_ = createChannelAndSetExpectations(this->channelName_, mode, this->recvQueueDepth_, true);
    }

    void TearDown() override { this->testChannel_->disconnect(); }
};

TEST_F(SCChannelReadModeSpecificTests, GivenReadModeWhenWriteThenError)
{
    // Populate the message we expect to receive from the channel.
    std::string dataToPassIn{"newData"};

    this->testChannel_->connect();

    // Expect an error when we try to write in read mode.
    EXPECT_EQ(task::ChannelStatus::CS_Error, testChannel_->write(dataToPassIn));

    this->testChannel_->forceUpdateMsgQueues();
}

/**
 * @brief Group of initialization tests in the write mode.
 */
class SCChannelWriteModeInitTests
    : public ::testing::Test
    , public SCChannelTestBase
{
protected:
    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> testChannel_;
    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> testChannel2_;
    std::string channelName_;

    void SetUp() override
    {
        this->channelName_ = "TestChannel";
        task::Access mode{task::Access::Write};
        uint32_t recvQueueDepth{1};

        this->testChannel_ = createChannelAndSetExpectations(this->channelName_, mode, recvQueueDepth, false);

        // This is a write channel, so the read thread should never run.
        EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(Exactly(0));
    }

    void TearDown() override {}
};

TEST_F(SCChannelWriteModeInitTests, GivenChannelExistsWhenConnectAndDisconnectThenSubscribeUnsubscribe)
{
    EXPECT_CALL(*(mockSCSPtr_.get()), declare(this->channelName_, _, _))
        .Times(1)
        .WillOnce(Return(SimpleComms::Status::Ok));

    // Disconnect the channel.
    EXPECT_CALL(*(mockSCSPtr_.get()), undeclare(this->channelName_, _))
        .Times(1)
        .WillOnce(Return(SimpleComms::Status::Ok));

    EXPECT_EQ(testChannel_->connect(), task::ChannelStatus::CS_Success);
    testChannel_->disconnect();
}

TEST_F(SCChannelWriteModeInitTests, GivenDisconnectedWriteReturnsError)
{
    std::string msg{"The quick brown fox jumps over the lazy dog"};
    EXPECT_EQ(testChannel_->write(msg), task::ChannelStatus::CS_Error);
}

/**
 * @brief Group of tests in the write mode.
 */
class SCChannelWriteModeTests
    : public ::testing::TestWithParam<task::Access>
    , public SCChannelTestBase
{
protected:
    std::unique_ptr<SC_CHANNEL_TEST::SCChannel> testChannel_;
    std::string channelName_;
    uint32_t recvQueueDepth_;

    void SetUp() override
    {
        this->channelName_ = "TestChannel";
        task::Access mode{GetParam()};
        this->recvQueueDepth_ = 5;

        this->testChannel_ = createChannelAndSetExpectations(this->channelName_, mode, this->recvQueueDepth_, true);

        if (mode == task::Access::ReadWrite) {
            EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(AtLeast(1));
        } else {
            EXPECT_CALL(*(mockSCSPtr_.get()), hasMsg_()).Times(Exactly(0));
        }
    }

    void TearDown() override { this->testChannel_->disconnect(); }

    /**
     * @brief Helper function to set EXPECT_CALL for each message to be written.
     *
     * @param expectedData All data expected to be written
     */
    void setSendExpectations(std::vector<std::string> const& expectedData)
    {
        ::testing::InSequence seq;
        for (auto& datum : expectedData) {
            // Expect SCS to be called once with the passed in data exactly once.
            EXPECT_CALL(*(mockSCSPtr_.get()), sendMsgV(this->channelName_, datum, _))
                .Times(1)
                .WillOnce(Return(SimpleComms::Status::Ok));
        }
    }

    /**
     * @brief A helper function to actually write data and confirm it is "successfully" sent.
     *
     * @param dataToWriteApi Data to check write to the channel
     * @param clearAfterWrite Perform a channel clear operation after each write.
     */
    void writeData(std::vector<std::string> const& dataToWriteApi, bool clearAfterWrite = false)
    {
        // Write the data to the channel
        for (auto& dataToWrite : dataToWriteApi) {
            EXPECT_EQ(task::ChannelStatus::CS_Success, testChannel_->write(dataToWrite));

            if (clearAfterWrite) {
                testChannel_->clear();
            }
        }
    }
};

TEST_P(SCChannelWriteModeTests, GivenValidChannelWhenWriteSingleValueThenPublishValueOnChannelOnce)
{
    // Populate the message we expect to send to the channel.
    std::vector<std::string> dataToSend{"newData1"};
    const auto expectedData{dataToSend};

    setSendExpectations(expectedData);
    this->testChannel_->connect();
    writeData(dataToSend);
    this->testChannel_->forceUpdateMsgQueues();
}

TEST_P(SCChannelWriteModeTests, GivenValidChannelWhenWriteMultipleValuesThenPublishAllValues)
{
    // Populate the message we expect to send to the channel.
    std::vector<std::string> dataToSend{"newData1", "", "randomStuff_3"};
    const auto expectedData{dataToSend};

    setSendExpectations(expectedData);
    this->testChannel_->connect();
    writeData(dataToSend);
    this->testChannel_->forceUpdateMsgQueues();
}

TEST_P(SCChannelWriteModeTests, GivenValidChannelWhenWriteLargeNumberOfValuesThenPublishAllValues)
{
    // Populate the message we expect to send to the channel.
    std::vector<std::string> dataToSend;
    for (int nextDataMember = 0; nextDataMember < 256; nextDataMember++) {
        dataToSend.push_back(std::to_string(nextDataMember));
    }
    const auto expectedData{dataToSend};

    setSendExpectations(expectedData);
    this->testChannel_->connect();
    writeData(dataToSend);
    this->testChannel_->forceUpdateMsgQueues();
}

INSTANTIATE_TEST_SUITE_P(WriteAndReadWriteModeRuntimeTests,
                         SCChannelWriteModeTests,
                         ::testing::Values(task::Access::Write, task::Access::ReadWrite));
