/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file SQLLogChannelTest.cc
 * @brief This file implements tests for the SQLLogChannel class.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#include <chrono>
#include <experimental/filesystem>
#include <fstream>
#include <memory>
#include <string>
#include <thread>
#include <vector>

#include <gmock/gmock-more-actions.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <SimpleComms/test/MockClient.h>
#include <task/Task.h>

#include <Comms/SQLLogChannel.h>

using namespace ::testing;
using namespace task;
namespace fs = std::experimental::filesystem;

#ifndef TEST_DATA_DIR
#define TEST_DATA_DIR "./"
#endif

namespace {

const fs::path TEST_DATA_DIRECTORY{fs::path(TEST_DATA_DIR)};

/**
 * @brief Writes data to the log file
 * @param[in] logFilePath Path to the log file to write
 * @param[in] msgToWrite Data to write to the log file
 * @param[out] writtenTime Time stamp of when data is written, which is the key in the log file
 */
bool writeToLogFile(fs::path logFilePath, const std::string& msgToWrite, ottotime::otimePoint& writtenTime)
{
    bool retVal{true};
    std::shared_ptr<SQLLogChannel> tempWriteChannel;
    ConfigSection emptyConfigSection{};

    tempWriteChannel = std::make_shared<SQLLogChannel>(emptyConfigSection, Access::Write);
    tempWriteChannel->updateLogFileName(logFilePath);
    EXPECT_EQ(ChannelStatus::CS_Success, tempWriteChannel->connect());

    writtenTime.getCurrentWallClockTime();

    if (ChannelStatus::CS_Success != tempWriteChannel->write(msgToWrite, writtenTime)) {
        retVal = false;
    }
    EXPECT_EQ(ChannelStatus::CS_Success, tempWriteChannel->disconnect());

    return retVal;
}

/**
 * @brief Writes a vector of data to the log file.
 * @param[in] logFilePath Path to the log file to write
 * @param[in] msgsToWrite Data to write to the log file
 * @param[out] writtenTime Time stamp of when each datum is written, which is the key in the log file
 */
bool writeToLogFile(fs::path logFilePath,
                    const std::vector<std::string>& msgsToWrite,
                    std::vector<ottotime::otimePoint>& writtenTimes)
{
    bool retVal{true};
    std::shared_ptr<SQLLogChannel> tempWriteChannel;
    ConfigSection emptyConfigSection{};

    tempWriteChannel = std::make_shared<SQLLogChannel>(emptyConfigSection, Access::Write);
    tempWriteChannel->updateLogFileName(logFilePath);
    EXPECT_EQ(ChannelStatus::CS_Success, tempWriteChannel->connect());

    for (auto& msgToWrite : msgsToWrite) {
        const ottotime::otimePoint timeToWrite = ottotime::otimePoint::buildCurrentWallClockTime();
        writtenTimes.push_back(timeToWrite);

        if (ChannelStatus::CS_Success != tempWriteChannel->write(msgToWrite, timeToWrite)) {
            retVal = false;
            break;
        }

        // This is to make sure that the difference between successive times when data is written is large enough to
        // actually notice a difference in the timeToWrite.
        std::this_thread::sleep_for(std::chrono::milliseconds(5));
    }

    EXPECT_EQ(ChannelStatus::CS_Success, tempWriteChannel->disconnect());

    return retVal;
}

/**
 * @brief Read data from the log file
 * @param[in] logFilePath Path to the log file to wrireadte
 * @param[out] readMsg Data read from the log file
 * @param[out] readTime Time stamp of the read out data, which is the key in the log file
 */
bool readFromLogFile(fs::path logFilePath, std::string& readMsg, ottotime::otimePoint& readTime)
{
    bool retVal{true};
    std::shared_ptr<SQLLogChannel> tempReadChannel;
    ConfigSection emptyConfigSection{};

    tempReadChannel = std::make_shared<SQLLogChannel>(emptyConfigSection, Access::Read);
    tempReadChannel->updateLogFileName(logFilePath);
    EXPECT_EQ(ChannelStatus::CS_Success, tempReadChannel->connect());

    if (ChannelStatus::CS_NewData != tempReadChannel->read(readMsg, readTime)) {
        retVal = false;
    }

    EXPECT_EQ(ChannelStatus::CS_Success, tempReadChannel->disconnect());
    return retVal;
}

/**
 * @brief Reads a vector of data from the log file.
 * @param[in] logFilePath Path to the log file to read from
 * @param[in] numEntriesToRead Number of entries to read from the log file.
 * @param[out] readMsgs Data read from the log file
 * @param[out] readTimes Times read from the log file
 */
bool readFromLogFile(fs::path logFilePath,
                     size_t numEntriesToRead,
                     std::vector<std::string>& readMsgs,
                     std::vector<ottotime::otimePoint>& readTimes)
{
    bool retVal{true};
    std::shared_ptr<SQLLogChannel> tempReadChannel;
    ConfigSection emptyConfigSection{};

    tempReadChannel = std::make_shared<SQLLogChannel>(emptyConfigSection, Access::Read);
    tempReadChannel->updateLogFileName(logFilePath);
    EXPECT_EQ(ChannelStatus::CS_Success, tempReadChannel->connect());

    for (size_t i = 0; i < numEntriesToRead; i++) {
        ottotime::otimePoint readTime;
        std::string readMsg;

        if (ChannelStatus::CS_NewData == tempReadChannel->read(readMsg, readTime)) {
            readMsgs.push_back(readMsg);
            readTimes.push_back(readTime);
        } else {
            // If no more new data, return false and stop reading.
            retVal = false;
            break;
        }
    }

    EXPECT_EQ(ChannelStatus::CS_Success, tempReadChannel->disconnect());

    return retVal;
}

/**
 * @brief Ensure that a read operation does not get any new data.
 * @param[in] testChannel Channel to check for data on.
 */
void checkReadEmptyData(std::shared_ptr<SQLLogChannel> testChannel)
{
    std::string actualMsg;
    ottotime::otimePoint defaultTimeOfReceipt, actualTimeOfReceipt;
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->read(actualMsg, actualTimeOfReceipt));
    EXPECT_EQ("", actualMsg);
    EXPECT_EQ(defaultTimeOfReceipt, actualTimeOfReceipt);
}

/**
 * @brief Assert 2 otime entities are almost equal, within 4 ULP's of their double representation
 * @param lhs First otime to entry to check
 * @param rhs Second otime to entry to check
 */
void checkTimeEqualityFuzzy(const ottotime::otimePoint& lhs, const ottotime::otimePoint& rhs)
{
    ASSERT_DOUBLE_EQ(lhs.toSecondsSinceEpochDouble(), rhs.toSecondsSinceEpochDouble());
}

/**
 * @brief Asserts corresponding otime entities within two vectors are almost equal, within 4 ULP's of
 * their double representation
 * @param lhs First vector of otime entries to check
 * @param rhs Second vector of otime entries to check
 */
void checkTimeEqualityFuzzy(const std::vector<ottotime::otimePoint>& lhs, const std::vector<ottotime::otimePoint>& rhs)
{
    EXPECT_EQ(lhs.size(), rhs.size());

    for (unsigned int i = 0; i < lhs.size(); i++) {
        checkTimeEqualityFuzzy(lhs.at(i), rhs.at(i));
    }
}

/**
 * @brief A wrapper class which allows us to create and manipulate log files using RAII semantics to ensure
 * that files which are created are destroyed as expected.
 */
class LogFileWrapper
{
public:
    LogFileWrapper() = delete;

    ~LogFileWrapper()
    {
        if (fs::exists(myFilePath_)) {
            fs::remove(myFilePath_);
        }
    }

    /**
     * @brief Function to create an object which encapsulates a log file.
     * @param filePath Path to the log file to create
     * @param validLog Indicator on whether we should create a valid log file or an invalid log file.
     *
     * The invalid log file is one which does not conform to the expected log file format.
     */
    static std::shared_ptr<LogFileWrapper> create(const fs::path& filePath, bool validLog)
    {
        bool logFileCreated{false};

        if (validLog) {
            ConfigSection tmpEmptyConfigSection{};
            auto writeChannel = std::make_shared<SQLLogChannel>(tmpEmptyConfigSection, Access::Write);

            writeChannel->updateLogFileName(filePath.string());

            if (ChannelStatus::CS_Success == writeChannel->connect()) {
                logFileCreated = fs::exists(filePath);
            }
        } else {
            std::ofstream ofs;
            ofs.open(filePath.c_str(), std::ofstream::out);
            ofs << "Invalid log file data.";
            ofs.close();

            logFileCreated = fs::exists(filePath);
        }

        if (logFileCreated) {
            return (std::shared_ptr<LogFileWrapper>(new LogFileWrapper(filePath)));
        } else {
            return nullptr;
        }
    }

private:
    fs::path myFilePath_{};
    LogFileWrapper(const fs::path& filePath)
        : myFilePath_{filePath}
    {}
};

} // namespace

TEST(FuzzyTimeCheck, ExactSameCopy)
{
    ottotime::otimePoint tp1 = ottotime::otimePoint::buildCurrentWallClockTime();
    ottotime::otimePoint tp2 = tp1;
    EXPECT_EQ(tp1, tp2);
    checkTimeEqualityFuzzy(tp1, tp2);
}

TEST(FuzzyTimeCheck, Difference_1ns)
{

    timespec t1{1606134488, 774999988};
    timespec t2{1606134488, 774999987};

    ottotime::otimePoint tp1 = ottotime::otimePoint(t1);
    ottotime::otimePoint tp2 = ottotime::otimePoint(t2);
    EXPECT_NE(tp1, tp2);
    checkTimeEqualityFuzzy(tp1, tp2);
}

TEST(FuzzyTimeCheck, ConvertCausedRollover_1nsPrecision)
{
    timespec t1{1606134488, 999999999};
    ottotime::otimePoint tp1 = ottotime::otimePoint(t1);
    ottotime::otimePoint tp2 = ottotime::otimePoint::fromSecondsSinceEpochDouble(tp1.toSecondsSinceEpochDouble());
    EXPECT_NE(tp1.to_timeval().tv_usec / 1000, tp2.to_timeval().tv_usec / 1000);
    EXPECT_NE(tp1, tp2);
    checkTimeEqualityFuzzy(tp1, tp2);
}

TEST(FuzzyTimeCheck, ConvertCausedRollover_100nsPrecision)
{
    timespec t1{1606134997, 373999971};
    ottotime::otimePoint tp1 = ottotime::otimePoint(t1);
    ottotime::otimePoint tp2 = ottotime::otimePoint::fromSecondsSinceEpochDouble(tp1.toSecondsSinceEpochDouble());
    EXPECT_NE(tp1.to_timeval().tv_usec / 1000, tp2.to_timeval().tv_usec / 1000);
    EXPECT_NE(tp1, tp2);
    checkTimeEqualityFuzzy(tp1, tp2);
}

/**
 * @brief Group of initialization tests in the read mode.
 */
class SQLLogChannelReadModeInitTests : public ::testing::Test
{
protected:
    Access mode_{Access::Read};
    ConfigSection emptyConfigSection_{};

    fs::path logFilePath_{TEST_DATA_DIRECTORY / fs::path{"testReadSqlLogChannelDb.sqlite"}};
};

TEST_F(SQLLogChannelReadModeInitTests, GivenNoLogAndEmptyConfigWhenCreateReadChannelThenNoErrors)
{
    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
}

TEST_F(SQLLogChannelReadModeInitTests, GivenNogLogCreateReadChannelWhenUpdateLogNameThenNoErrors)
{
    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName("tmp");
}

TEST_F(SQLLogChannelReadModeInitTests, GivenNoLogWhenConnectToLogThenFailConnect)
{
    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName("tmp");

    EXPECT_EQ(ChannelStatus::CS_Error, testChannel->connect());
}

TEST_F(SQLLogChannelReadModeInitTests, GivenInvalidLogWhenConnectToLogThenFails)
{
    // Create an invalid log file
    auto logFile = LogFileWrapper::create(logFilePath_, false);
    ASSERT_TRUE(logFile != nullptr);

    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName(logFilePath_);

    // The connection will fail as the log file is invalid
    EXPECT_EQ(ChannelStatus::CS_Error, testChannel->connect());
}

TEST_F(SQLLogChannelReadModeInitTests, GivenConnecToInvalidLogWhenDisconnectToLogThenSucceeds)
{
    // Create an invalid log file
    auto logFile = LogFileWrapper::create(logFilePath_, false);
    ASSERT_TRUE(logFile != nullptr);

    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName(logFilePath_);

    // The connection will fail as the log file is invalid
    EXPECT_EQ(ChannelStatus::CS_Error, testChannel->connect());
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->disconnect());
}

TEST_F(SQLLogChannelReadModeInitTests, GivenValidLogWhenConnectDisconnectToLogThenSucceeds)
{
    // Create a valid log file
    auto logFile = LogFileWrapper::create(logFilePath_, true);
    ASSERT_TRUE(logFile != nullptr);

    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName(logFilePath_);

    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->connect());
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->disconnect());
}

TEST_F(SQLLogChannelReadModeInitTests, GivenValidLogWhenWriteThenReturnsErrorAndNoDataWritten)
{
    // Create a valid log file
    auto logFile = LogFileWrapper::create(logFilePath_, true);
    ASSERT_TRUE(logFile != nullptr);

    // Create a channel which references a log file and connect to it
    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName(logFilePath_);
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->connect());

    // Attempt writing data and ensure it returns an error and that nothing is written.
    EXPECT_EQ(ChannelStatus::CS_Error, testChannel->Channel::write("TestMessage1"));
    checkReadEmptyData(testChannel);

    ottotime::otimePoint defaultTimeOfReceipt;
    EXPECT_EQ(ChannelStatus::CS_Error, testChannel->write("TestMessage2", defaultTimeOfReceipt));
    checkReadEmptyData(testChannel);

    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->disconnect());
}

/**
 * @brief Group of tests in the read mode.
 */
class SQLLogChannelReadModeTests : public ::testing::Test
{
protected:
    Access mode_{Access::Read};
    ConfigSection emptyConfigSection_{};
    fs::path logFilePath_{TEST_DATA_DIRECTORY / fs::path{"testReadSqlLogChannelDb.sqlite"}};
    std::shared_ptr<LogFileWrapper> logFile_;
    std::shared_ptr<SQLLogChannel> testChannel_;

    void SetUp()
    {
        // Create a valid log file
        logFile_ = LogFileWrapper::create(logFilePath_, true);
        ASSERT_TRUE(logFile_ != nullptr);

        // Create the SQL Log channel for tests and connect to the log file from there.
        testChannel_ = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
        testChannel_->updateLogFileName(logFilePath_);
        EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->connect());
    }

    void TearDown()
    {
        // Disconnect from the log channel.
        EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->disconnect());
    }
};

TEST_F(SQLLogChannelReadModeTests, GivenEmptyLogWhenReadThenEmptyMessage)
{
    std::string actualMsg;
    ottotime::otimePoint actualTimeOfReceipt;
    ottotime::otimePoint expectedTimeOfReceipt;

    EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->read(actualMsg, actualTimeOfReceipt));
    EXPECT_EQ("", actualMsg);
    EXPECT_EQ(expectedTimeOfReceipt, actualTimeOfReceipt);
}

TEST_F(SQLLogChannelReadModeTests, GivenSingleLogEntryWhenReadThenReadValidData)
{
    // Write data to log
    std::string expectedMessage("Log entry 1");
    ottotime::otimePoint expectedTimeOfReceipt;

    EXPECT_TRUE(writeToLogFile(logFilePath_, expectedMessage, expectedTimeOfReceipt));

    std::string actualMsg;
    ottotime::otimePoint actualTimeOfReceipt;

    // Read data back and ensure we got new data and validate the contents.
    EXPECT_EQ(ChannelStatus::CS_NewData, testChannel_->read(actualMsg, actualTimeOfReceipt));
    EXPECT_EQ(expectedMessage, actualMsg);
    checkTimeEqualityFuzzy(expectedTimeOfReceipt, actualTimeOfReceipt);

    // After all data, we should get  success with empty data.
    checkReadEmptyData(testChannel_);
}

TEST_F(SQLLogChannelReadModeTests, GivenMultipleLogEntriesWhenReadThenReadValidData)
{
    // Write data to log
    std::vector<std::string> expectedMessages;
    std::vector<ottotime::otimePoint> expectedTimesOfReceipt;

    expectedMessages.push_back("Log entry 1");
    expectedMessages.push_back("2nd Log entry");
    expectedMessages.push_back("Log 3 entry");

    EXPECT_TRUE(writeToLogFile(logFilePath_, expectedMessages, expectedTimesOfReceipt));

    std::string actualMsg;
    ottotime::otimePoint actualTimeOfReceipt;

    std::string expectedMessage;
    ottotime::otimePoint expectedTimeOfReceipt;

    // Read all the messages from the log and validate it.
    for (unsigned int i = 0; i < expectedMessages.size(); i++) {
        expectedMessage = expectedMessages.at(i);
        expectedTimeOfReceipt = expectedTimesOfReceipt.at(i);

        EXPECT_EQ(ChannelStatus::CS_NewData, testChannel_->read(actualMsg, actualTimeOfReceipt));
        EXPECT_EQ(expectedMessage, actualMsg);
        checkTimeEqualityFuzzy(expectedTimeOfReceipt, actualTimeOfReceipt);
    }

    // After all data, we should get  success with empty data.
    checkReadEmptyData(testChannel_);
}

/**
 * @brief Group of initialization tests in the write mode.
 */
class SQLLogChannelWriteModeInitTests : public ::testing::Test
{
protected:
    Access mode_{Access::Write};
    fs::path logFilePath_{TEST_DATA_DIRECTORY / fs::path{"testWriteSqlLogChannelDb.sqlite"}};
    ConfigSection emptyConfigSection_{};

    void SetUp() override {}

    void TearDown() override
    {
        if (fs::exists(logFilePath_)) {
            fs::remove(logFilePath_);
        }
    }
};

TEST_F(SQLLogChannelWriteModeInitTests, GivenLogFileAbsentWhenCreateAndConnectToFileThenFileCreated)
{
    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);

    testChannel->updateLogFileName(logFilePath_.string());

    // The file is not present prior to the connect call
    EXPECT_EQ(false, fs::exists(logFilePath_));
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->connect());
    // The file is present after the connect call
    EXPECT_EQ(true, fs::exists(logFilePath_));
}

TEST_F(SQLLogChannelWriteModeInitTests, GivenConnectedToNewLogWhenDisconnectThenSucceeds)
{
    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName(logFilePath_.string());
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->connect());
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->disconnect());
}

TEST_F(SQLLogChannelWriteModeInitTests, GivenLogFilePresentWhenCreateAndConnectToFileThenNoError)
{
    // Create a valid log file
    auto logFile = LogFileWrapper::create(logFilePath_, true);
    ASSERT_TRUE(logFile != nullptr);

    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName(logFilePath_.string());

    // The file is present before and after the second connect call
    EXPECT_EQ(true, fs::exists(logFilePath_));
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->connect());
    EXPECT_EQ(true, fs::exists(logFilePath_));
}

TEST_F(SQLLogChannelWriteModeInitTests, GivenConnectedToOldLogWhenDisconnectThenSucceeds)
{
    // Create a valid log file
    auto logFile = LogFileWrapper::create(logFilePath_, true);
    ASSERT_TRUE(logFile != nullptr);

    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName(logFilePath_.string());
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->connect());
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->disconnect());
}

/**
 * @brief Group of tests in the write mode.
 */
class SQLLogChannelWriteModeTests : public ::testing::Test
{
protected:
    Access mode_{Access::Write};
    ConfigSection emptyConfigSection_{};
    fs::path logFilePath_{TEST_DATA_DIRECTORY / fs::path{"testWriteSqlLogChannelDb.sqlite"}};

    std::shared_ptr<LogFileWrapper> logFile_;
    std::shared_ptr<SQLLogChannel> testChannel_;

    void SetUp()
    {
        // Create a valid log file
        logFile_ = LogFileWrapper::create(logFilePath_, true);
        ASSERT_TRUE(logFile_ != nullptr);

        // Create the SQL Log channel for tests and connect to the log file from there.
        testChannel_ = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
        testChannel_->updateLogFileName(logFilePath_);
        EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->connect());
    }

    void TearDown()
    {
        // Disconnect from the log channel.
        EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->disconnect());
    }
};

TEST_F(SQLLogChannelWriteModeTests, GivenSingleLogEntryWhenReadThenEmpty)
{
    // Write data to log
    std::string dataToWrite("Log entry 1");
    ottotime::otimePoint timeDataWritten;
    ottotime::otimePoint defaultTimeData;

    EXPECT_TRUE(writeToLogFile(logFilePath_, dataToWrite, timeDataWritten));

    std::string actualMsg;
    ottotime::otimePoint actualTimeOfReceipt;

    // Read data back and ensure that the data is empty.
    EXPECT_EQ(ChannelStatus::CS_Error, testChannel_->read(actualMsg, actualTimeOfReceipt));
    EXPECT_EQ("", actualMsg);
    checkTimeEqualityFuzzy(defaultTimeData, actualTimeOfReceipt);
}

TEST_F(SQLLogChannelWriteModeTests, GivenMultipleLogEntriesWhenReadThenEmpty)
{
    // Write data to log
    std::vector<std::string> dataToWrite;
    std::vector<ottotime::otimePoint> timesOfWrite;

    dataToWrite.push_back("Log entry 1");
    dataToWrite.push_back("2nd Log entry");
    dataToWrite.push_back("Log 3 entry");

    EXPECT_TRUE(writeToLogFile(logFilePath_, dataToWrite, timesOfWrite));

    std::string actualMsg;
    ottotime::otimePoint actualTimeOfReceipt;

    ottotime::otimePoint defaulTime;

    // Read as many messages as there should be in the log and ensure data empty all the time
    for (unsigned int i = 0; i < dataToWrite.size(); i++) {
        EXPECT_EQ(ChannelStatus::CS_Error, testChannel_->read(actualMsg, actualTimeOfReceipt));
        EXPECT_EQ("", actualMsg);
        checkTimeEqualityFuzzy(defaulTime, actualTimeOfReceipt);
    }
}

TEST_F(SQLLogChannelWriteModeTests, GivenEmptyLogFileWhenWriteThenCorrectDataWritten)
{
    // Write data to log
    std::string dataToWrite("Log entry 1");
    const ottotime::otimePoint timeDataWritten = ottotime::otimePoint::buildCurrentWallClockTime();

    EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->write(dataToWrite, timeDataWritten));
    testChannel_->disconnect(); // force write-out

    std::string actualMsg;
    ottotime::otimePoint actualTimeOfReceipt;

    // Read data back and ensure that the data is the same as what is written.
    EXPECT_EQ(true, readFromLogFile(logFilePath_, actualMsg, actualTimeOfReceipt));
    EXPECT_EQ(dataToWrite, actualMsg);
    checkTimeEqualityFuzzy(timeDataWritten, actualTimeOfReceipt);
}

TEST_F(SQLLogChannelWriteModeTests, GivenEmptyLogFileWhenWriteMultipleDataThenCorrectDataWritten)
{
    // Write data to log
    std::vector<std::string> dataToWrite;
    std::vector<ottotime::otimePoint> timeWritten;

    dataToWrite.push_back("Log entry 1");
    dataToWrite.push_back("2nd Log entry");
    dataToWrite.push_back("Log 3 entry");

    for (auto const& datumToWrite : dataToWrite) {
        const ottotime::otimePoint timeDataWritten = ottotime::otimePoint::buildCurrentWallClockTime();
        EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->write(datumToWrite, timeDataWritten));

        // Introduce a delay to ensure time granularity is maintained between messages
        timeWritten.push_back(timeDataWritten);
        std::this_thread::sleep_for(std::chrono::milliseconds(2));
    }
    testChannel_->disconnect(); // force write-out

    // Read as many entries out of the log file as written and ensure it succeeds.
    std::vector<std::string> actualMessages;
    std::vector<ottotime::otimePoint> actualTimesWritten;
    EXPECT_EQ(true, readFromLogFile(logFilePath_, dataToWrite.size(), actualMessages, actualTimesWritten));
    EXPECT_EQ(dataToWrite, actualMessages);
    checkTimeEqualityFuzzy(timeWritten, actualTimesWritten);
}

TEST_F(SQLLogChannelWriteModeTests, MultipleMessagesSameTimeStamp)
{
    // Some data
    std::vector<std::string> dataToWrite;

    dataToWrite.emplace_back("Log entry 1");
    dataToWrite.emplace_back("Log entry 2");

    const ottotime::otimePoint timeDataWritten = ottotime::otimePoint::buildCurrentWallClockTime();
    for (const auto& datumToWrite : dataToWrite) {
        EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->write(datumToWrite, timeDataWritten));
    }
}

/**
 * @brief Group of tests in the write mode where we already have an existing log.
 */
class SQLLogChannelWriteModeTestsExistingLog : public ::testing::Test
{
protected:
    Access mode_{Access::Write};
    ConfigSection emptyConfigSection_{};
    fs::path logFilePath_{TEST_DATA_DIRECTORY / fs::path{"testWriteExistingEntriesSqlLogChannelDb.sqlite"}};

    std::shared_ptr<LogFileWrapper> logFile_;
    std::shared_ptr<SQLLogChannel> testChannel_;

    std::vector<std::string> existingMessages_;
    std::vector<ottotime::otimePoint> existingTimesOfReceipt_;

    void SetUp()
    {
        // Create a valid log file
        existingMessages_.push_back("Log entry 1");
        existingMessages_.push_back("2nd Log entry");
        existingMessages_.push_back("Log 3 entry");

        logFile_ = LogFileWrapper::create(logFilePath_, true);
        ASSERT_TRUE(writeToLogFile(logFilePath_, existingMessages_, existingTimesOfReceipt_));

        ASSERT_TRUE(logFile_ != nullptr);

        // Create the SQL Log channel for tests and connect to the log file from there.
        testChannel_ = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
        testChannel_->updateLogFileName(logFilePath_);
        EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->connect());
    }

    void TearDown()
    {
        // Disconnect from the log channel.
        EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->disconnect());
    }
};

TEST_F(SQLLogChannelWriteModeTestsExistingLog, GivenExistingLogFileWhenWriteSingleEntryThenNewDataWritten)
{
    // Write data to log
    std::string dataToWrite("Random Data entry 134232");
    const ottotime::otimePoint timeDataWritten = ottotime::otimePoint::buildCurrentWallClockTime();
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel_->write(dataToWrite, timeDataWritten));
    testChannel_->disconnect(); // force write-out

    // Attempt reading more entries than should be present(old log entries + new data) - it will return false but still
    // populate as many entries as it has.
    std::vector<std::string> actualMessages;
    std::vector<ottotime::otimePoint> actualTimesWritten;
    EXPECT_EQ(false,
              readFromLogFile(
                  logFilePath_, existingMessages_.size() + dataToWrite.size(), actualMessages, actualTimesWritten));

    // Expect ONLY the new data written to be present - the previous data should have gotten overwritten.
    std::vector<std::string> expectedData{dataToWrite};
    std::vector<ottotime::otimePoint> expectedTimes{timeDataWritten};

    EXPECT_EQ(expectedData, actualMessages);
    checkTimeEqualityFuzzy(expectedTimes, actualTimesWritten);
}

// -------------- CURRENTLY FAILING TESTS --------------

// THIS TEST DOES NOT CURRENTLY PASS AS THE SQL LOG CHANNEL ATTEMPTS TO DROP A TABLE
// BUT THE LOG FILE DOES NOT GUARANTEE IT CAN RUN SQLITE COMMANDS.
TEST_F(SQLLogChannelWriteModeTests, DISABLED_GivenInvalidLogFilePresentWhenCreateAndConnectToFileThenNoError)
{
    // Create an invalid log file
    auto logFile = LogFileWrapper::create(logFilePath_, false);
    ASSERT_TRUE(logFile != nullptr);

    auto testChannel = std::make_shared<SQLLogChannel>(emptyConfigSection_, mode_);
    testChannel->updateLogFileName(logFilePath_.string());

    // The file is present before and after the second connect call
    EXPECT_EQ(true, fs::exists(logFilePath_));
    EXPECT_EQ(ChannelStatus::CS_Success, testChannel->connect());
    EXPECT_EQ(true, fs::exists(logFilePath_));
}
