/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief DDS Quality of Service Helper
 */
#include "CycloneConfigurator.hpp"
#include "lib/avlog/avlog.hpp"
#include <dds/dds.hpp>

namespace DDSComms {

dds::pub::qos::DataWriterQos CycloneConfigurator::createWriterQos(const std::optional<std::string>& channelName) const
{
    static const dds::core::Duration asyncWriterLatencyBudget{0, 100000}; // 100us
    dds::pub::qos::DataWriterQos qos{};

    auto qosProfileNameOpt = options_.defaultWriteChannelProfile;
    const auto& profileMap = options_.writeChannelProfileMap;

    if (channelName.has_value()) {
        const auto& profileItr = profileMap.find(channelName.value());

        if (profileItr != profileMap.end()) {
            qosProfileNameOpt = profileItr->second;
        } else {
            AVLOG_INFO("No writer profile registered for channel {}; using default from dds config.",
                       channelName.value());
        }
    }

    const auto& durability = qos.policy<dds::core::policy::Durability>();
    const auto& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    auto& history = qos.policy<dds::core::policy::History>();
    auto& reliability = qos.policy<dds::core::policy::Reliability>();
    auto& latencyBudget = qos.policy<dds::core::policy::LatencyBudget>();

    if (applyCommonReaderWriterQoS(qos, qosProfileNameOpt)) {
        // perform and writer-specific layering of setttings
    } else if (qosProfileNameOpt == "AVQosLib::Generic.StrictReliableAsyncWrite") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.StrictReliable");
        latencyBudget.duration(asyncWriterLatencyBudget);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast25ReliableAsyncWrite") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.KeepLast25Reliable");
        latencyBudget.duration(asyncWriterLatencyBudget);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast5ReliableAsyncWrite") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.KeepLast5Reliable");
        latencyBudget.duration(asyncWriterLatencyBudget);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast10ReliableAsyncWrite") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.KeepLast10Reliable");
        latencyBudget.duration(asyncWriterLatencyBudget);
    } else {
        throw std::runtime_error("unknown writer QoS profile " + qosProfileNameOpt.value_or(unknownProfileName) +
                                 "for channel " + channelName.value_or(unknownProfileName));
    }

    const auto depth = getHistoryDepthStr(qos);
    AVLOG_INFO("Creating writer QoS profile {} for {}; queueDepth {}; max_samples_per_instance: {}; "
               "max_samples: {}; history_kind {}; ReliablilityKind {}; DurabilityKind {};",
               qosProfileNameOpt.value_or(unknownProfileName),
               channelName.value_or(unknownProfileName),
               depth,
               resourceLimits.max_samples_per_instance(),
               resourceLimits.max_samples(),
               history.kind(),
               reliability.kind(),
               durability.kind());

    return qos;
}

dds::sub::qos::DataReaderQos CycloneConfigurator::createReaderQos(
    [[maybe_unused]] const std::optional<uint32_t>& receiveCapacityOpt,
    const std::optional<std::string>& channelName) const
{
    static const dds::core::Duration readerLatencyBudget{0, 1000000}; // 1 ms
    dds::sub::qos::DataReaderQos qos{};

    auto qosProfileNameOpt = options_.defaultReadChannelProfile;
    const auto& profileMap = options_.readChannelProfileMap;

    if (channelName.has_value()) {
        const auto& profileItr = profileMap.find(channelName.value());

        if (profileItr != profileMap.end()) {
            qosProfileNameOpt = profileItr->second;
        } else {
            AVLOG_INFO("No reader profile registered for channel {}; using default from dds config.",
                       channelName.value());
        }
    }

    const auto& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    const auto& durability = qos.policy<dds::core::policy::Durability>();
    auto& history = qos.policy<dds::core::policy::History>();
    auto& reliability = qos.policy<dds::core::policy::Reliability>();
    auto& latencyBudget = qos.policy<dds::core::policy::LatencyBudget>();

    // This is just a hint needed to receive Async published data.
    latencyBudget.duration(readerLatencyBudget);

    if (applyCommonReaderWriterQoS(qos, qosProfileNameOpt)) {
        // perform and reader-specific layering of setttings
    } else {
        throw std::runtime_error("unknown reader QoS profile " + qosProfileNameOpt.value_or(unknownProfileName) +
                                 "for channel " + channelName.value_or(unknownProfileName));
    }

    const auto depth = getHistoryDepthStr(qos);
    AVLOG_INFO("Creating reader QoS profile {} for {}; queueDepth {}; max_samples_per_instance: {}; "
               "max_samples: {}; history_kind {}; ReliablilityKind {}; DurabilityKind {};",
               qosProfileNameOpt.value_or(unknownProfileName),
               channelName.value_or(unknownProfileName),
               depth,
               resourceLimits.max_samples_per_instance(),
               resourceLimits.max_samples(),
               history.kind(),
               reliability.kind(),
               durability.kind());
    return qos;
}

/// create a DomainParticipant Qos profile.
dds::domain::qos::DomainParticipantQos CycloneConfigurator::createParticipantQos() const
{
    dds::domain::qos::DomainParticipantQos qos{};
    return qos;
}

} // namespace DDSComms
