/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief DDS Quality of Service Helper
 */
#pragma once

#include "DDSConfigurator.hpp"
#include "DDSConfiguratorOptions.hpp"
#include <dds/dds.hpp>

#include <optional>
#include <string>

namespace DDSComms {

class CycloneConfigurator : public DDSConfigurator
{

public:
    /// Constructor.
    /// @param options: The config options
    /// @throws std::runtime_error if the configFileName provided in options is invalid
    CycloneConfigurator(CycloneOptions options)
        : options_(std::move(options))
    {}

    /// Create a DataReader QoS profile
    /// This will lookup the channelName in the provided options and attempt to create the QoS with the
    /// set profile for this channel. If the channel is not found or set, defaultReadChannelProfile will
    /// be used.
    /// @param receiveCapacityOpt Depth hint of history receive queue. Must be <= than numeric_limits<int>::max()
    /// @note receiveCapacityOpt is only a hint
    dds::sub::qos::DataReaderQos createReaderQos(
        const std::optional<uint32_t>& receiveCapacityOpt = std::nullopt,
        const std::optional<std::string>& channelName = std::nullopt) const override;

    /// create a DataWriter Qos profile
    /// This will lookup the channelName in the provided options and attempt to create the QoS with the
    /// set profile for this channel. If the channel is not found or set, defaultWriteChannelProfile will
    /// be used.
    /// @param channelName the name of the channel this QoS is for
    dds::pub::qos::DataWriterQos createWriterQos(
        const std::optional<std::string>& channelName = std::nullopt) const override;

    /// create a DomainParticipant Qos profile.
    dds::domain::qos::DomainParticipantQos createParticipantQos() const override;

    template<typename T>
    static std::string getHistoryDepthStr(const T& qos);

private:
    /// configuration options
    CycloneOptions options_{};
};

/// Apply QoS settings based on profile that are common to reader and writer QoS
/// @param qos The QoS object of type T that settings are applied to.
/// @tparam T the type of QoS, either DataReaderQoS or DataWriterQoS
/// @return whether QoS settings were applied to qos based on profile
template<typename T>
bool applyCommonReaderWriterQoS(T& qos, const std::optional<std::string>& qosProfileNameOpt)
{
    auto& history = qos.template policy<dds::core::policy::History>();
    auto& reliability = qos.template policy<dds::core::policy::Reliability>();
    auto& durability = qos.template policy<dds::core::policy::Durability>();

    bool found = true;
    if (qosProfileNameOpt == "AVQosLib::Generic.StrictReliable") {
        reliability.kind(dds::core::policy::ReliabilityKind::RELIABLE);
        history.kind(dds::core::policy::HistoryKind::KEEP_ALL);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast1Reliable" ||
               qosProfileNameOpt == "AVQosLib::Generic.KeepLastReliable") {
        reliability.kind(dds::core::policy::ReliabilityKind::RELIABLE);
        history.kind(dds::core::policy::HistoryKind::KEEP_LAST);
        history.depth(1);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast5Reliable") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.KeepLastReliable");
        history.depth(5);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast10Reliable") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.KeepLastReliable");
        history.depth(10);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast25Reliable") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.KeepLastReliable");
        history.depth(25);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast50Reliable") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.KeepLastReliable");
        history.depth(50);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast100Reliable") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.KeepLastReliable");
        history.depth(100);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLastReliable.TransientLocal") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.KeepLast1Reliable");
        durability.kind(dds::core::policy::DurabilityKind::TRANSIENT_LOCAL);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.BestEffort") {
        reliability.kind(dds::core::policy::ReliabilityKind::BEST_EFFORT);
        history.kind(dds::core::policy::HistoryKind::KEEP_LAST);
        history.depth(100);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast10BestEffort") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.BestEffort");
        history.depth(10);
    } else if (qosProfileNameOpt == "AVQosLib::Generic.KeepLast5BestEffort") {
        applyCommonReaderWriterQoS(qos, "AVQosLib::Generic.BestEffort");
        history.depth(5);
    } else {
        // not a QoS profile that has shared values between readers and writers
        found = false;
    }
    return found;
}

template<typename T>
std::string CycloneConfigurator::getHistoryDepthStr(const T& qos)
{
    const auto& history = qos.template policy<dds::core::policy::History>();
    auto depth = (history.kind() == dds::core::policy::HistoryKind::KEEP_ALL) ? std::string("NA")
                                                                              : std::to_string(history.depth());
    return depth;
}

} // namespace DDSComms
