/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file DDSClient.cc
 * \brief DDSClient member function definitions.
 */

#include <chrono>
#include <sstream>

#include "DDSComms/DDSClient.hpp"
#include "lib/avlog/avlog.hpp"
#include "platform/middleware/libs/Comms/include/Comms/Channel.h"
#include "platform/middleware/libs/DDSComms/DataReaderListener.hpp"
#include "platform/middleware/libs/DDSComms/DataWriterListener.hpp"
#include <TimeStamp/otimeConversions.h>

using namespace std::chrono_literals;
using namespace SimpleComms;

namespace DDSComms {

DDSClient::DDSClient(Client::Callback callback, int32_t domainId, std::shared_ptr<DDSConfigurator> config)
    : Client(callback)
    , domainId_(domainId)
    , config_(std::move(config))
{
    resetEntities();
}

DDSClient::~DDSClient()
{
    disconnect();
}

Status::States DDSClient::connect()
{
    Status::States retStatus{Status::States::Ok};

    std::lock_guard<std::shared_mutex> lock{clientMtx_};

    if (isConnected()) {
        retStatus = Status::States::AlreadyConnected;
    } else {
        try {
            // reuse domain participant in process if available
            domainParticipant_ = dds::domain::find(domainId_);
            if (domainParticipant_ == dds::core::null) {
                auto qos = config_->createParticipantQos();
                domainParticipant_ = dds::domain::DomainParticipant(domainId_, qos);
            }
        } catch (const dds::core::Exception& e) {
            retStatus = Status::States::Error;
            domainParticipant_ = dds::core::null;
            AVLOG_ERROR("unable to create DomainParticipant: {} ", e.what());
        }
    }

    // only start the thread if we are callback receive driven.
    if (retStatus == Status::States::Ok && callback_) {
        thread_ = std::thread(&DDSClient::threadEntry, this);
    }

    return retStatus;
}

Status::States DDSClient::disconnect()
{
    {
        // must lock shared here so thread_ can also lock shared.
        std::shared_lock<std::shared_mutex> shLock{clientMtx_};
        if (isConnected()) {
            if (callback_) {
                // signal receive thread / waitset shutdown
                unblockCondition.trigger_value(true);

                if (thread_.joinable()) {
                    AVLOG_TRACE("Waiting on receive thread to join ... ");
                    thread_.join();
                    AVLOG_TRACE("receive thread joined ");
                }
            }
        }
    }

    std::lock_guard<std::shared_mutex> exLock{clientMtx_};
    resetEntities();
    return Status::Ok;
}

void DDSClient::resetEntities()
{
    // Let entities go out of scope. If this client is the only DDS user
    // in the process, reference counts will go to zero and be freed.
    // If for some reason there are users, we won't destroy them here.

    waitset_ = dds::core::cond::WaitSet{};
    unblockCondition = dds::core::cond::GuardCondition{};
    waitset_ += unblockCondition;
    for (auto& reader : readers_) {
        if (reader.second.listener != nullptr) {
            delete reader.second.listener;
            reader.second.listener = nullptr;
        }
    }

    std::vector<DataWriterListener<PayloadType>*> writerListeners;
    for (auto& writer : writers_) {
        if (writer.second.listener != nullptr) {
            writerListeners.emplace_back(writer.second.listener);
        }
    }
    readers_.clear();
    writers_.clear();
    subscriber_ = dds::core::null;
    publisher_ = dds::core::null;
    domainParticipant_ = dds::core::null;
    for (auto listener : writerListeners) {
        if (listener != nullptr) {
            delete listener;
            listener = nullptr;
        }
    }
}

Status::States DDSClient::declare(const std::string& channel, const ChannelType type, const uint32_t receiveCapacity)
{
    std::lock_guard<std::shared_mutex> lock{clientMtx_};

    if (!isConnected()) {
        AVLOG_WARN(
            "declare called while not connected; channel {}; type {} ;", AVLOG_ARGS("channel", "type"), channel, type);
        return Status::States::NotConnected;
    }

    auto retState{Status::States::Ok};

    if (type == ChannelType::SubPub || type == ChannelType::Subscriber) {
        const Status::States readState{createReaderHelper(channel, type, receiveCapacity)};
        retState = (readState == Status::States::Ok) ? retState : readState;
    }

    if (type == ChannelType::SubPub || type == ChannelType::Publisher) {
        const Status::States writeState{createWriterHelper(channel, type)};
        retState = (writeState == Status::States::Ok) ? retState : writeState;
    }

    return retState;
}

Status::States DDSClient::createReaderHelper(const std::string& channel,
                                             const ChannelType type,
                                             const uint32_t receiveCapacity)
{
    auto retState{Status::States::Ok};
    try {
        if (ensureSubscriber()) {
            if (readers_.find(channel) == readers_.end()) {
                createReader(channel, std::optional<uint32_t>(receiveCapacity));
            }
            // else if reader already exists, silently allow it as UnixClient does.
        } else {
            retState = Status::States::Error;
            AVLOG_ERROR("unable to ensure subscriber for channel {}; type {}; ", channel, type);
        }
    } catch (const dds::core::Exception& e) {
        AVLOG_ERROR("unable to create reader with channel {}: {} ", channel, e.what());
        retState = Status::States::Error;
    } catch (const std::invalid_argument& e) {
        AVLOG_ERROR("unable to create reader with channel {}: {} ", channel, e.what());
        retState = Status::States::Error;
    }

    return retState;
}

Status::States DDSClient::createWriterHelper(const std::string& channel, const ChannelType type)
{
    auto retState{Status::States::Ok};
    try {
        if (ensurePublisher()) {
            auto writer = writers_.find(channel);
            if (writer == writers_.end()) {
                createWriter(channel);
            }
            // else if writer already exists, silently allow it as UnixClient does.
        } else {
            retState = Status::States::Error;
            AVLOG_ERROR("unable to ensure publisher for channel {}; type {}; ", channel, type);
        }
    } catch (const dds::core::Exception& e) {
        AVLOG_ERROR("unable to create writer with channel {}: {} ", channel, e.what());
        retState = Status::States::Error;
    }

    return retState;
}

Status::States DDSClient::undeclare(const std::string& channel, const ChannelType type)
{
    std::lock_guard<std::shared_mutex> lock{clientMtx_};
    if (!isConnected()) {
        AVLOG_WARN("undeclare called while not connected; channel {}; type {}; ",
                   AVLOG_ARGS("channel", "type"),
                   channel,
                   type);
        return Status::States::NotConnected;
    }

    auto retState{Status::States::Ok};

    if (type == ChannelType::SubPub || type == ChannelType::Subscriber) {
        auto reader = readers_.find(channel);
        if (reader != readers_.end()) {
            waitset_ -= reader->second.condition;
            delete reader->second.listener;
            reader->second.listener = nullptr;
            readers_.erase(channel);
        } else {
            AVLOG_WARN(
                "undeclare called on not-declared channel {}; type {}; ", AVLOG_ARGS("channel", "type"), channel, type);
            // UnixClient allows this, so we will, too
        }
    }

    if (type == ChannelType::SubPub || type == ChannelType::Publisher) {
        auto writer = writers_.find(channel);
        if (writer != writers_.end()) {
            delete writer->second.listener;
            writer->second.listener = nullptr;
            writers_.erase(channel);
        } else {
            AVLOG_WARN(
                "undeclare called on not-declared channel {}; type {}; ", AVLOG_ARGS("channel", "type"), channel, type);
            // UnixClient allows this, so we will, too
        }
    }

    return retState;
}

Status::States DDSClient::sendMsgV(const std::string& channel,
                                   const std::string& data,
                                   std::vector<av::latency::Pred>&& predecessors)
{
    auto retState{Status::States::Error};
    TimeStamp ts{};
    DDSComms::PayloadType msg;
    uint32_t seqNumber{0};
    {
        std::shared_lock<std::shared_mutex> lock{clientMtx_};
        auto writer = writers_.find(channel);
        if (writer != writers_.end() && writer->second.writer != dds::core::null) {
            seqNumber = ++(writer->second.sequenceNumber);
            msg = DDSComms::createMsg(data, seqNumber);
            // building our own timestamp such that we can provide it to LatencyLogger
            ts.setNow(); // using this to keep behavior same as UnixClient
            dds::core::Time timestamp(ts.tv_sec, static_cast<uint32_t>(ts.tv_usec * 1000));
            // When using Reliable mode, dds::core::TimeoutError may be thrown.
            // allow other errors to propagate
            try {
                writer->second.writer.write(msg, timestamp);
                retState = Status::States::Ok;
            } catch (const dds::core::TimeoutError& e) {
                AVLOG_WARN("cannot send msg: channel {}; seq {}; err {};", channel, seqNumber, e.what());
                retState = Status::States::Error;
            }
        } else {
            retState = isConnected() ? Status::States::InvalidHandle : Status::States::NotConnected;
        }
    }

    // log outside of lock, debug level as to not spam log messages on vehicle
    if (retState == Status::States::NotConnected) {
        AVLOG_DEBUG("ERROR: sendMsgV on channel {} while Client not connected ", channel);
    } else if (retState == Status::States::InvalidHandle) {
        AVLOG_DEBUG("ERROR: sendMsgV called on channel {}: undeclared channel;", channel);
    } else if (retState == Status::States::Ok) {
        const ottotime::otimeSpec sendTimeSpec{ts.tv_sec, ts.tv_usec * 1000};
        av::LatencyLogger::getLogger().SendEvent(channel, sendTimeSpec, std::move(predecessors));
        task::Channel::reportMessage(channel, data.size());
        AVLOG_TRACE("sendMsgV sent msg {} {} on {}: '{}' ", ts, seqNumber, channel, data);
    }

    return retState;
}

Status::States DDSClient::receiveMsg(const std::string& channel, std::string& data, MessageInfo& msgInfo)
{
    auto retState{Status::States::Error};
    dds::sub::LoanedSamples<PayloadType> samples{};

    {
        std::shared_lock<std::shared_mutex> lock{clientMtx_};
        auto reader = readers_.find(channel);
        if (reader != readers_.end() && reader->second.reader != dds::core::null) {
            samples = reader->second.readOneSelector->take();

            if (samples.length() == 1 && samples.begin()->info().valid()) {
                auto sample = samples.begin();
                std::string tmp = DDSComms::getMsgData(sample->data());
                data = std::move(tmp);
                populateMessageInfo(sample->info(), msgInfo);
                retState = Status::States::NewData;
            } else {
                // no data available, OR we took a message and it is not DDS "valid data".
                // both are okay, not error conditions
                retState = Status::States::Ok;
            }
        } else {
            retState = isConnected() ? Status::States::InvalidHandle : Status::States::NotConnected;
        }
    }

    // log outside of lock, debug level as to not spam log messages on vehicle
    if (retState == Status::States::NotConnected) {
        AVLOG_DEBUG("ERROR: receiveMsg on channel {} while Client not connected ", channel);
    } else if (retState == Status::States::InvalidHandle) {
        AVLOG_DEBUG("ERROR: receiveMsg on channel {} undeclared channel ", channel);
    } else if (retState == Status::States::NewData) {
        const ottotime::otimeSpec sendTimeSpec{msgInfo.sendMsgTime.tv_sec, msgInfo.sendMsgTime.tv_usec * 1000};
        av::LatencyLogger::getLogger().ReceiveEvent(channel, sendTimeSpec);
        task::Channel::reportMessage(channel, data.size(), av::timePointFromOtimeSpec(sendTimeSpec));
        AVLOG_TRACE("receiveMsg on channel got msg: '{}' ", channel, data);
    } else if (retState == Status::States::Ok) {
        AVLOG_DEBUG("receiveMsg on channel {}: no data available ", channel);
    }

    return retState;
}

void DDSClient::populateMessageInfo(const dds::sub::SampleInfo& sampleInfo, MessageInfo& msgInfo)
{
    std::stringstream s{};
    s << sampleInfo.publication_handle();
    msgInfo.origin = s.str();

#ifdef DDS_VERSION_RTI_DDS_CPP
    // reception_timestamp is an RTI extension
    msgInfo.msgReadyTime.tv_sec = sampleInfo.extensions().reception_timestamp().sec();
    msgInfo.msgReadyTime.tv_usec = static_cast<int32_t>(sampleInfo.extensions().reception_timestamp().nanosec() / 1000);
    const auto& source_ts = sampleInfo.source_timestamp();
#else
    // technically this is not the "reception timestamp" but the timestamp of "now".
    // Cyclone does not currently expose reception timestamp.
    msgInfo.msgReadyTime.setNow();
    const auto& source_ts = sampleInfo.timestamp(); // this is source timestamp
#endif

    msgInfo.sendMsgTime.tv_sec = static_cast<int32_t>(source_ts.sec());
    msgInfo.sendMsgTime.tv_usec = static_cast<int32_t>(source_ts.nanosec() / 1000);
}

Status::States DDSClient::receiveMsg(const std::string& channel, std::string& data)
{
    MessageInfo msgInfo{}; // thrown away
    return receiveMsg(channel, data, msgInfo);
}

void DDSClient::allowThrottling(std::function<void(bool)> /*callback*/)
{
    // stub no-op
    AVLOG_WARN("DDSClient::allowThrottling called but not implemented ");
}

bool DDSClient::hasMsg() const
{
    // stub no-op
    AVLOG_WARN("DDSClient::hasMsg called but not implemented ");
    return true;
}

bool DDSClient::ensureSubscriber()
{
    if (subscriber_ == dds::core::null && domainParticipant_ != dds::core::null) {
        subscriber_ = dds::sub::Subscriber(domainParticipant_);
    }

    return subscriber_ != dds::core::null;
}

bool DDSClient::ensurePublisher()
{
    if (publisher_ == dds::core::null && domainParticipant_ != dds::core::null) {
        publisher_ = dds::pub::Publisher(domainParticipant_);
    }

    return publisher_ != dds::core::null;
}

DDSClient::PayloadTopicType DDSClient::ensureTopic(const std::string& channel)
{
    PayloadTopicType topic{dds::core::null};
    if (domainParticipant_ != dds::core::null) {
        topic = dds::topic::find<PayloadTopicType>(domainParticipant_, channel);
        if (topic == dds::core::null) {
            topic = PayloadTopicType{domainParticipant_, channel};
        }
    }

    return topic;
}

void DDSClient::createWriter(const std::string& channel, std::optional<std::string> qosProfileNameOpt)
{
    auto topic = ensureTopic(channel);
    auto qos = config_->createWriterQos(channel);
    auto listener = new DataWriterListener<PayloadType>(channel);
    using dds::core::status::StatusMask;
    StatusMask writerMask = StatusMask::none();
    writerMask |= StatusMask::offered_deadline_missed();
    writerMask |= StatusMask::offered_incompatible_qos();
    writerMask |= StatusMask::liveliness_lost();
    writerMask |= StatusMask::publication_matched();
    AVLOG_INFO("Creating DDS writer {}; ", channel.c_str());
    WriterContext writerContext{
        PayloadDataWriter(publisher_, topic, qos, listener, writerMask), std::move(qosProfileNameOpt), listener};
    writers_.emplace(channel, writerContext);
}

void DDSClient::createReader(const std::string& channel,
                             const std::optional<uint32_t>& receiveCapacityOpt,
                             std::optional<std::string> qosProfileNameOpt)
{
    auto topic = ensureTopic(channel);
    auto qos = config_->createReaderQos(receiveCapacityOpt, channel);
    auto listener = new DataReaderListener<PayloadType>(channel);
    using dds::core::status::StatusMask;
    StatusMask readerMask = StatusMask::none();
    readerMask |= StatusMask::requested_deadline_missed();
    readerMask |= StatusMask::requested_incompatible_qos();
    readerMask |= StatusMask::sample_lost();
    readerMask |= StatusMask::sample_rejected();
    readerMask |= StatusMask::liveliness_changed();
    readerMask |= StatusMask::subscription_matched();
    AVLOG_INFO("Creating DDS reader {}", channel.c_str());
    auto reader = PayloadDataReader(subscriber_, topic, qos, listener, readerMask);

    // Create a ReadCondition for a reader with a specific DataState

    static const dds::sub::status::DataState readConditionState(dds::sub::status::SampleState::not_read(),
                                                                dds::sub::status::ViewState::any(),
                                                                dds::sub::status::InstanceState::any());

    auto readCondition = dds::sub::cond::ReadCondition(reader, readConditionState);

    // create our reusable selector to read one sample, attached to this read condition
    auto readOneSelector = std::make_unique<PayloadDataReader::Selector>(reader);
    readOneSelector->max_samples(1).state(readConditionState);

    // create our reusable selector to read multiple samples, attached to this read condition
    auto readMultiSelector = std::make_unique<PayloadDataReader::Selector>(reader);
    readMultiSelector->state(readConditionState);

    ReaderContext readerContext{reader,
                                std::move(qosProfileNameOpt),
                                readCondition,
                                std::move(readOneSelector),
                                std::move(readMultiSelector),
                                listener};

    readers_.emplace(channel, std::move(readerContext));

    // add our condition to the waitset
    waitset_ += readCondition;
}

bool DDSClient::isConnected() const
{
    // using domainParticipant as a proxy for connection state
    // should be called under mutex protection
    return domainParticipant_ != dds::core::null;
}

Client::MessageKey DDSClient::generateMsgKey([[maybe_unused]] const PayloadType& data, const dds::sub::SampleInfo& info)
{
    std::stringstream s{};
    // publication_handle is unique only locally to this DomainParticipant / Publisher / DataWriter .
    s << info.publication_handle();

#ifdef DDS_VERSION_RTI_DDS_CPP
    const auto& seqNumber64 = info.extensions().original_publication_virtual_sequence_number();
    const auto& seqNumber = seqNumber64.low();
    if (seqNumber64.high() != 0) {
        // we have surpassed a uint32_t representation of message count.
        AVLOG_ERROR("msg sequence is too large for a MessageKey or negative. seqNumber.high()= {} ",
                    seqNumber64.high());
    }
#else
    // Cyclone does not expose publication sequence number. We need to either:
    // - use local sequence number (bad because it might not line up with data logger seq number)
    // - add sequence number to as a base field in every AV IDL type.
    // - expose message sequence number (which should be in each message due to RTPS spec) in the
    //   cyclone API
    const auto& seqNumber = data.sequenceNumber();
#endif

    MessageKey key{s.str(), seqNumber};
    AVLOG_TRACE("msg key: {} {} ", key.first, key.second);
    return key;
}

void DDSClient::threadEntry()
{
    AVLOG_TRACE("entered receive thread entry ");
    bool exitLoop{false};
    while (!exitLoop) {
        dds::core::cond::WaitSet::ConditionSeq active_conditions{};

        AVLOG_TRACE("Begining waiting on waitset ... ");
        active_conditions = waitset_.wait(); // block
        AVLOG_TRACE("waitset wait unblocked. Active conds: {}; ", active_conditions.size());
        uint32_t numMsgsProcessed{0};
        {
            std::shared_lock<std::shared_mutex> lock{clientMtx_};
            for (const auto& active_condition : active_conditions) {
                for (auto& [channel, context] : readers_) {
                    if (active_condition == context.condition) {
                        dds::sub::LoanedSamples<PayloadType> samples = context.reader->take();
                        TimeStamp ts{};
                        ts.setNow();
                        const ottotime::otimeSpec recvTimeSpec{ts.tv_sec, ts.tv_usec * 1000};
                        for (const auto& sample : samples) {
                            if (sample->info().valid()) {
                                std::string tmp = getMsgData(sample->data());
#ifdef DDS_VERSION_RTI_DDS_CPP
                                const auto& source_ts = sample->info().source_timestamp();
#else
                                const auto& source_ts = sample->info().timestamp();
#endif
                                const ottotime::otimeSpec sendTimeSpec{source_ts.sec(), source_ts.nanosec()};
                                task::Channel::reportMessage(
                                    channel, tmp.size(), av::timePointFromOtimeSpec(sendTimeSpec));
                                auto msgKey = generateMsgKey(sample->data(), sample->info());
                                if (callback_) {
                                    AVLOG_TRACE("calling msg callback_() for {}: {} ", channel, tmp);
                                    callback_(channel, tmp, msgKey);
                                    ++numMsgsProcessed;
                                }
                                av::LatencyLogger::getLogger().ReceiveEvent(channel, sendTimeSpec, recvTimeSpec);

                            } else {
                                // "invalid" samples are not an error in DDS, more of an annoyance
                                AVLOG_DEBUG("Took invalid DDS data on {} channel ", channel);
                            }
                        }
                    }
                }

                if (active_condition == unblockCondition) {
                    // we unblocked ourselves
                    AVLOG_DEBUG("receive thead waitset saw exit condition; exiting loop. ");
                    unblockCondition.trigger_value(false);
                    exitLoop = true;
                }
            }
        } // lock scope
        AVLOG_TRACE("waitset processing done; num msgs: {};", numMsgsProcessed);
    }

    AVLOG_DEBUG("exiting receive thread ");
}

Status::States DDSClient::waitForMatchedPeer(const std::string& channel,
                                             const ChannelType localType,
                                             const uint32_t expectedMatchCount,
                                             std::chrono::milliseconds timeout)
{
    auto start = std::chrono::steady_clock::now();
    std::shared_lock<std::shared_mutex> lock{clientMtx_};

    if (!isConnected()) {
        AVLOG_WARN("waitForMatchedPeer called while not connected; channel {}; localType {} ;", channel, localType);
        return Status::States::NotConnected;
    }

    auto retState{Status::States::Ok};

    if (localType == ChannelType::SubPub || localType == ChannelType::Subscriber) {
        auto readerCtx = readers_.find(channel);
        if (readerCtx != readers_.end()) {
            auto& reader = readerCtx->second.reader;
            while (static_cast<uint32_t>(reader.subscription_matched_status().current_count()) < expectedMatchCount) {
                if (std::chrono::steady_clock::now() - start > timeout) {
                    retState = Status::States::Error;
                    break;
                }
                std::this_thread::sleep_for(2ms);
            }
        } else {
            AVLOG_WARN("waitForMatchedPeer called on not-declared channel {}; localType {}; ", channel, localType);
            retState = Status::States::InvalidHandle;
        }
    }

    if (localType == ChannelType::SubPub || localType == ChannelType::Publisher) {
        auto writerCtx = writers_.find(channel);
        if (writerCtx != writers_.end()) {
            auto& writer = writerCtx->second.writer;
            while (static_cast<uint32_t>(writer.publication_matched_status().current_count()) < expectedMatchCount) {
                if (std::chrono::steady_clock::now() - start > timeout) {
                    retState = Status::States::Error;
                    break;
                }
                std::this_thread::sleep_for(2ms);
            }
        } else {
            AVLOG_WARN("waitForMatchedPeer called on not-declared channel {}; localType {}; ", channel, localType);
            retState = Status::States::InvalidHandle;
        }
    }

    return retState;
}

std::vector<std::string> DDSClient::getPublishListMessage()
{
    std::lock_guard<std::shared_mutex> lock{clientMtx_};
    std::vector<std::string> publishedChannels;

    for (const auto& writer : writers_) {
        publishedChannels.emplace_back(writer.first);
    }
    return publishedChannels;
}

} // namespace DDSComms
