/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief DDS Quality of Service Helper
 */

#include "DDSConfigurator.hpp"

namespace DDSComms {

void applyDataReaderQueueDepth(const uint32_t receiveCapacity, dds::sub::qos::DataReaderQos& qos)
{

    if (receiveCapacity > static_cast<uint32_t>(std::numeric_limits<int>::max())) {
        throw std::invalid_argument("receiveCapacity too large.");
    }

    auto& history = qos.policy<dds::core::policy::History>();
    auto& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();

    if (receiveCapacity == 0) {
        // Allow unlimited mesages in receive Queue
        history.kind(dds::core::policy::HistoryKind::KEEP_ALL);
        history.depth(dds::core::LENGTH_UNLIMITED);
        resourceLimits.max_samples_per_instance(dds::core::LENGTH_UNLIMITED);
        resourceLimits.max_samples(dds::core::LENGTH_UNLIMITED);

    } else {
        if (history.depth() > static_cast<int32_t>(receiveCapacity)) {
            AVLOG_WARN("reducing size of receive queue depth. queueDepth: old={}; new={}; ",

                       history.depth(),
                       receiveCapacity);
        }
        history.depth(static_cast<int32_t>(receiveCapacity));

        // The interaction of history and resource limits in the context of other QoS settings
        // like reliability is not trivial. More tuning can be done, but it should be in the context of
        // starting with a specific QoS profile. Here, we are taking a less-invasive strategy of ensuring
        // when we change the history depth that we also have setup _at least_ enough maximum resources.
        if (resourceLimits.max_samples_per_instance() != dds::core::LENGTH_UNLIMITED &&
            resourceLimits.max_samples_per_instance() < static_cast<int32_t>(receiveCapacity)) {
            resourceLimits.max_samples_per_instance(static_cast<int32_t>(receiveCapacity));
        }

        if (resourceLimits.max_samples() != dds::core::LENGTH_UNLIMITED &&
            resourceLimits.max_samples() < static_cast<int32_t>(receiveCapacity)) {
            resourceLimits.max_samples(static_cast<int32_t>(receiveCapacity));
        }
        if (history.kind() != dds::core::policy::HistoryKind::KEEP_LAST) {
            AVLOG_WARN("DDSClient: changing reader HistoryKind to KEEP_LAST from KEEP_ALL. ");
            history.kind(dds::core::policy::HistoryKind::KEEP_LAST);
        }
    }
}

} // namespace DDSComms
