/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief DDS Quality of Service Helper
 */
#pragma once

#include <string>

#include "lib/avlog/avlog.hpp"
#include <dds/dds.hpp>

namespace DDSComms {

template<typename T>
class DataReaderListener : public dds::sub::NoOpDataReaderListener<T>
{
public:
    DataReaderListener(std::string topicName)
        : topicName_(std::move(topicName))
    {}

    void on_sample_lost([[maybe_unused]] dds::sub::DataReader<T>& reader,
                        const dds::core::status::SampleLostStatus& status) override
    {
        AVLOG_WARN("on_sample_lost callback: reader {}; total_count = {}; total_count_change = {}",
                   topicName_,
                   status.total_count(),
                   status.total_count_change());
    }
    void on_subscription_matched([[maybe_unused]] dds::sub::DataReader<T>& reader,
                                 const dds::core::status::SubscriptionMatchedStatus& status) override
    {
        AVLOG_INFO("on_subscription_matched callback: reader {}; total_count = {}; total_count_change = {}; "
                   "current_count = {}; current_count_change = {};",
                   topicName_,
                   status.total_count(),
                   status.total_count_change(),
                   status.current_count(),
                   status.current_count_change());
    }

    void on_sample_rejected([[maybe_unused]] dds::sub::DataReader<T>& reader,
                            const dds::core::status::SampleRejectedStatus& status) override
    {
        AVLOG_WARN("on_sample_rejected callback: reader {}; total_count = {}; total_count_change = {};",
                   topicName_,
                   status.total_count(),
                   status.total_count_change());
    }

    void on_requested_incompatible_qos([[maybe_unused]] dds::sub::DataReader<T>& reader,
                                       const dds::core::status::RequestedIncompatibleQosStatus& status) override
    {
        AVLOG_ERROR("on_requested_incompatible_qos callback: reader {}; total_count = {}; total_count_change = {}; "
                    "last_policy_id = {};  ",
                    topicName_,
                    status.total_count(),
                    status.total_count_change(),
                    status.last_policy_id());

        dds::core::policy::QosPolicyCountSeq qos_seq = status.policies();
        if (!qos_seq.empty()) {
            AVLOG_ERROR("policy_id of one first incompatible Qos policy = {}", qos_seq[0].policy_id());
        }
    }

    void on_requested_deadline_missed([[maybe_unused]] dds::sub::DataReader<T>& reader,
                                      const dds::core::status::RequestedDeadlineMissedStatus& status) override
    {
        AVLOG_WARN("on_requested_deadline_missed callback: reader {}; total_count = {}; total_count_change = {}",
                   topicName_,
                   status.total_count(),
                   status.total_count_change());
    }
    void on_liveliness_changed([[maybe_unused]] dds::sub::DataReader<T>& reader,
                               const dds::core::status::LivelinessChangedStatus& status) override
    {
        AVLOG_WARN("on_liveliness_changed callback: reader {}; alive_count = {}; not_alive_count = {}; "
                   "alive_count_change = {}; not_alive_count_change = {} ",
                   topicName_,
                   status.alive_count(),
                   status.not_alive_count(),
                   status.alive_count_change(),
                   status.not_alive_count_change());
    }

    void on_data_available([[maybe_unused]] dds::sub::DataReader<T>& reader) override {}

private:
    std::string topicName_{};
};

} // namespace DDSComms