/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief DDS Quality of Service Helper
 */
#include <experimental/filesystem>
#include <limits>

#include "RtiProConfigurator.hpp"
#include "lib/avlog/avlog.hpp"
#include <dds/dds.hpp>

namespace DDSComms {

dds::sub::qos::DataReaderQos RtiProConfigurator::createReaderQos(
    [[maybe_unused]] const std::optional<uint32_t>& receiveCapacityOpt,
    const std::optional<std::string>& channelName) const
{

    // First load the default profile from the QoS provider.
    auto qos = qosProvider_->datareader_qos();

    auto qosProfileNameOpt = options_.defaultReadChannelProfile;
    const auto& profileMap = options_.readChannelProfileMap;

    if (channelName.has_value()) {
        const auto& profileItr = profileMap.find(channelName.value());

        if (profileItr != profileMap.end()) {
            qosProfileNameOpt = profileItr->second;
        }
    }

    // override it to our own specific profile if provided
    if (qosProfileNameOpt.has_value()) {
        qos = qosProvider_->datareader_qos(qosProfileNameOpt.value());
    } else {
        const auto defaultProfile = qosProvider_->default_profile();
        qosProfileNameOpt = (defaultProfile.has_value()) ? defaultProfile.value() : unknownProfileName;
    }

    auto& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    // because we are not using keyed data, only allocate initial space for one instance
    resourceLimits.extensions().initial_instances(1);
    if (receiveCapacityOpt) {
        applyDataReaderQueueDepth(receiveCapacityOpt.value(), qos);
    }

    const auto& history = qos.template policy<dds::core::policy::History>();
    const auto& reliability = qos.template policy<dds::core::policy::Reliability>();
    const auto& durability = qos.template policy<dds::core::policy::Durability>();
    const auto depth = getHistoryDepthStr(qos);

    AVLOG_INFO("Creating reader QoS profile {} for {}; queueDepth {}; max_samples_per_instance: {}; "
               "max_samples: {}; history_kind {}; ReliablilityKind {}; DurabilityKind {};",
               qosProfileNameOpt.value_or(unknownProfileName),
               channelName.value_or(unknownProfileName),
               depth,
               resourceLimits.max_samples_per_instance(),
               resourceLimits.max_samples(),
               history.kind().underlying(),
               reliability.kind().underlying(),
               durability.kind().underlying());
    return qos;
}

dds::pub::qos::DataWriterQos RtiProConfigurator::createWriterQos(const std::optional<std::string>& channelName) const
{
    // First load the default profile from the QoS provider.
    auto qos = qosProvider_->datawriter_qos();

    auto qosProfileNameOpt = options_.defaultWriteChannelProfile;
    const auto& profileMap = options_.writeChannelProfileMap;

    if (channelName.has_value()) {
        const auto& profileItr = profileMap.find(channelName.value());

        if (profileItr != profileMap.end()) {
            qosProfileNameOpt = profileItr->second;
        }
    }

    // override it to our own specific profile if provided
    if (qosProfileNameOpt.has_value()) {
        qos = qosProvider_->datawriter_qos(qosProfileNameOpt.value());
    } else {
        // Just getting the name of the actual profile we end up using
        const auto defaultProfile = qosProvider_->default_profile();
        qosProfileNameOpt = (defaultProfile.has_value()) ? defaultProfile.value() : unknownProfileName;
    }

    auto& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    // because we are not using keyed data, only allocate initial space for one instance
    resourceLimits.extensions().initial_instances(1);
    const auto& history = qos.template policy<dds::core::policy::History>();
    const auto& reliability = qos.template policy<dds::core::policy::Reliability>();
    const auto& durability = qos.template policy<dds::core::policy::Durability>();
    const auto depth = getHistoryDepthStr(qos);

    AVLOG_INFO("Creating writer QoS profile {} for {}; queueDepth {}; max_samples_per_instance: {}; "
               "max_samples: {}; history_kind {}; ReliablilityKind {}; DurabilityKind {};",
               qosProfileNameOpt.value_or(unknownProfileName),
               channelName.value_or(unknownProfileName),
               depth,
               resourceLimits.max_samples_per_instance(),
               resourceLimits.max_samples(),
               history.kind().underlying(),
               reliability.kind().underlying(),
               durability.kind().underlying());
    return qos;
}

dds::domain::qos::DomainParticipantQos RtiProConfigurator::createParticipantQos() const
{

    // First load the default profile from the QoS provider.
    auto qos = qosProvider_->participant_qos();

    auto qosProfileNameOpt = options_.domainParticipantProfile;

    // override it to our own specific profile if provided
    if (qosProfileNameOpt.has_value()) {
        qos = qosProvider_->participant_qos(qosProfileNameOpt.value());
    } else {
        // Just getting the name of the actual profile we end up using
        const auto defaultProfile = qosProvider_->default_profile();
        qosProfileNameOpt = (defaultProfile.has_value()) ? defaultProfile.value() : unknownProfileName;
    }

    const auto shutdownDurationS = 0;         // 0s
    const auto shutdownDurationNs = 10000000; // 10ms
    qos.policy<rti::core::policy::Database>().shutdown_cleanup_period(
        dds::core::Duration(shutdownDurationS, shutdownDurationNs));

    AVLOG_INFO("Creating DP QoS profile {}; ", qosProfileNameOpt.value().c_str());

    return qos;
}

dds::core::QosProvider RtiProConfigurator::getQosProvider(std::string fileName)
{

    if (std::experimental::filesystem::exists(fileName)) {
        // if the path directly exists as given, use it. pass.
    } else if (std::experimental::filesystem::exists(getOttoDirectory() + "/" + fileName)) {
        // if the filename exists within OTO_DIR, use that filename
        fileName = getOttoDirectory() + "/" + fileName;
    } else {
        AVLOG_WARN("Unable to find QoS file at absolute path nor in OTTO_DIR: {} ", fileName);
        // return whatever the deafault provider is
        // return dds::core::QosProvider::Default();
        return dds::core::null;
    }

    // ensure we have a proper filepath
    fileName = std::experimental::filesystem::canonical(fileName);

    AVLOG_INFO("Creating QosProvider using file {} ", fileName);

    // create a DDS QosProvider using our QoS file
    rti::core::QosProviderParams params;
    params.ignore_environment_profile(true);
    params.ignore_user_profile(true);
    params.ignore_resource_profile(true);
    std::vector<std::string> url_profiles = {std::string("file://") + fileName};
    params.url_profile(url_profiles);

    dds::core::QosProvider qosProvider{""};
    qosProvider.extensions().provider_params(params);
    qosProvider.extensions().reload_profiles();

    return qosProvider;
}

} // namespace DDSComms
