/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief DDS Quality of Service Helper
 */
#pragma once

#include <optional>

#include "DDSConfigurator.hpp"
#include "DDSConfiguratorOptions.hpp"
#include <dds/dds.hpp>

#include <optional>
#include <string>
#include <unordered_map>

namespace DDSComms {

class RtiProConfigurator : public DDSConfigurator
{

public:
    /// Constructor.
    /// @param options: The config options
    /// @throws std::runtime_error if the configFileName provided in options is invalid
    RtiProConfigurator(RtiProOptions options)
        : options_(std::move(options))
    {
        qosProvider_ = getQosProvider(options_.configFileName);
        if (qosProvider_ == dds::core::null) {
            throw std::runtime_error(std::string("Failed to find QoS Config file: ") + options_.configFileName);
        }
    }

    /// Create a DataReader QoS profile.
    /// This will lookup the channelName in the provided options and attempt to create the QoS with the
    /// set profile for this channel. If the channel is not found or set, defaultReadChannelProfile will
    /// be used. If that is not set, the DDS QoSProvider default will be used.
    /// @param receiveCapacityOpt Depth hint of history receive queue. Must be <= than numeric_limits<int>::max()
    /// @param channelName the name of the channel this QoS is for
    /// @note receiveCapacityOpt is only a hint
    /// @throws dds::core::Exception if options specifies a QoS Profile name that is invalid
    dds::sub::qos::DataReaderQos createReaderQos(
        const std::optional<uint32_t>& receiveCapacityOpt = std::nullopt,
        const std::optional<std::string>& channelName = std::nullopt) const override;

    /// create a DataWriter Qos profile.
    /// This will lookup the channelName in the provided options and attempt to create the QoS with the
    /// set profile for this channel. If the channel is not found or set, defaultWriteChannelProfile will
    /// be used. If that is not set, the DDS QoSProvider default will be used.
    /// @param channelName the name of the channel this QoS is for
    /// @throws dds::core::Exception if options specifies a QoS Profile name that is invalid
    dds::pub::qos::DataWriterQos createWriterQos(
        const std::optional<std::string>& channelName = std::nullopt) const override;

    /// create a DomainParticipant Qos profile.
    /// The QoS will be based on the profile indicated in options' domainParticipantProfile, if set
    /// Otherwise DDS QoSProvider default will be used
    dds::domain::qos::DomainParticipantQos createParticipantQos() const override;

    template<typename T>
    static std::string getHistoryDepthStr(const T& qos);

private:
    /// configuration options
    RtiProOptions options_{};
    /// the DDS provider configured using the options_ configFilePath
    dds::core::QosProvider qosProvider_{dds::core::null};
    /// Create a QoS provider with its profiles sourced from fileName
    dds::core::QosProvider getQosProvider(std::string fileName);
};

template<typename T>
std::string RtiProConfigurator::getHistoryDepthStr(const T& qos)
{
    const auto& history = qos.template policy<dds::core::policy::History>();
    auto depth = (history.kind().underlying() == dds::core::policy::HistoryKind::KEEP_ALL)
                     ? std::string("NA")
                     : std::to_string(history.depth());
    return depth;
}

} // namespace DDSComms
