/**
 * @copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * @file DDSClient.h
 * @brief DDS Client Declarations
 */
#pragma once

#include "DDSConfigurator.hpp"
#include "DDSConfiguratorHelper.hpp"
#include "SimpleComms/Client.h"
#include "platform/middleware/libs/LatencyLogger/include/LatencyLogger/LatencyLogger.h"
#include <dds/dds.hpp>

#include <memory>
#include <optional>
#include <shared_mutex>
#include <string>
#include <thread>
#include <unordered_map>

namespace DDSComms {

template<typename T>
class DataWriterListener;

template<typename T>
class DataReaderListener;

class DDSClient : public SimpleComms::Client
{

private:
    using DomainParticipant = dds::domain::DomainParticipant;
    using PayloadTopicType = dds::topic::Topic<PayloadType>;
    using Publisher = dds::pub::Publisher;
    using Subscriber = dds::sub::Subscriber;
    using PayloadDataWriter = dds::pub::DataWriter<PayloadType>;
    using PayloadDataReader = dds::sub::DataReader<PayloadType>;
    using SharedPayloadSamples = dds::sub::SharedSamples<PayloadType>;

public:
    /// Constructor.
    /// @param callback: the function to call when a new message is received
    /// @param domainId: the DDS domain to use
    /// @param config: the DDS configurator to use
    DDSClient(Client::Callback callback = Client::Callback(),
              int32_t domainId = 0,
              std::shared_ptr<DDSConfigurator> config = std::make_shared<DDSConfiguratorType>(DDSOptionsType{}));
    DDSClient(const DDSClient&) = delete; ///< Disable copy constructor since boost::noncopyable has been removed
    DDSClient& operator=(DDSClient const&) =
        delete; ///< Disable assignment operator since boost::noncopyable has been removed

    /// Destructor
    ~DDSClient() override;

    /// Connect to the messaging system.
    /// Initializes DDS Domain Participant and starts receive thread if a msg reception
    /// callback is set.
    /// @return Status::Ok if successful; Status::AlreadyConnected if already connected; Status::Error otherwise
    SimpleComms::Status::States connect() override;

    /// Disconnect from the messaging system.
    /// Uninitializes DDS Domain Participant, undeclares all channels, stops and joins
    /// receive thread if it is running.
    /// @return Status::Ok if successful.
    SimpleComms::Status::States disconnect() override;

    /// Declare a channel
    /// Creates a DDS DataReader and/or DataWriter with Topic named channel.
    /// @param channel The channel name.
    /// @param type The type (subscribing, publishing, or both).
    /// @param receiveCapacity For subsciptions, the maximum queue depth, or 0 for unlimited depth.
    /// @return Status::Ok if successful; Status::NotConnected if not connected; Status::Error otherwise
    SimpleComms::Status::States declare(const std::string& channel,
                                        const SimpleComms::ChannelType type,
                                        const uint32_t receiveCapacity) override;

    /// Stop subscribing or publishing to a channel.
    /// Closes DDS DataWriter / DataReader
    /// @param channel The channel name.
    /// @param type The type (subscribing, publishing, or both).
    /// @return Status::Ok if successful.
    /// @return Status::Ok if successful; Status::NotConnected if not connected; Status::Error otherwise
    SimpleComms::Status::States undeclare(const std::string& channel, const SimpleComms::ChannelType type) override;

    /// Send a message.
    /// @param channel Name of the channel to send the message to.
    /// @param data The data to send to named channel.
    /// @param predecessors Specifiers for this event's predecessors (the events that triggered this one).
    /// @return Status::Ok if successful; Status::InvalidHandle if the channel is undeclared;
    /// Status::States::NotConnected if not connected; Status::Error otherwise.
    SimpleComms::Status::States sendMsgV(const std::string& channel,
                                         const std::string& data,
                                         std::vector<av::latency::Pred>&& predecessors) override;

    /// Receive a message (non-blocking).
    /// @param channel Name of the channel to send the message to.
    /// @param[out] data A reference to a string where the received data will be placed.
    /// @param[out] msgInfo A reference to an object to store message info.
    /// @return Status::NewData if data was received; Status::Ok if no error occurred but no data was received;
    /// Status::InvalidHandle if channel is undeclared; Status::NotConnected if not connected;
    /// Status::Error otherwise.
    SimpleComms::Status::States receiveMsg(const std::string& channel,
                                           std::string& data,
                                           MessageInfo& msgInfo) override;

    /// Receive Message, ignoring MessageInfo.
    SimpleComms::Status::States receiveMsg(const std::string& channel, std::string& data);

    /// Check if a message is ready to receive.
    /// @todo not implemented at this time.
    /// @return True if receiveMsg() will return new data.
    bool hasMsg() const override;

    /// Tell SCS that this client responds to requests to pause its output. Must be called after connecting.
    /// @todo not implemented at this time.
    void allowThrottling(std::function<void(bool)> /*callback*/) override;

    /// Get all published channels from client
    std::vector<std::string> getPublishListMessage() override;

    /// @section DDSClient specific APIs

    /// Wait until the entity of localType on channel has at least expectedMatchCount peers,
    /// with a max wait of timeout.
    /// This method polls some status values on the DDS entities every 2 milliseconds and,
    /// as such, should be used carefully. Its primary use is during unit testing where we may
    /// create a reader and writer within microseconds of one another, and we need to ensure that
    /// we have matched before we send a test message. The timeout applies individually to each
    /// ChannelType such that if you use ChannelType::SubPub, it will wait for a duration of
    /// at most ~timeout*2. Threadsafe, but this does hold the clientMtx_ in shared mode for
    /// the entire duration of its wait so, again, use carefully and sparingly.
    SimpleComms::Status::States waitForMatchedPeer(const std::string& channel,
                                                   const SimpleComms::ChannelType localType,
                                                   const uint32_t expectedMatchCount = 1,
                                                   std::chrono::milliseconds timeout = std::chrono::milliseconds{1000});

private:
    /// Encapsulates references or storage to DDS Reader related entities for one channel.
    struct ReaderContext
    {
        /// The actual DDS reader.
        PayloadDataReader reader{dds::core::null};
        /// The name for our QoS profile used, if applicable.
        std::optional<std::string> qosProfileNameOpt;
        /// Condition for attachment to Waitset that indicates new data on this channel for read.
        dds::sub::cond::ReadCondition condition{dds::core::null};
        /// Special read selector which returns at most one sample matching our condition.
        std::unique_ptr<PayloadDataReader::Selector> readOneSelector;
        /// Special read selector which returns at 1 to many sample matching our condition.
        std::unique_ptr<PayloadDataReader::Selector> readMultiSelector;
        /// DDS Reader Listener
        DataReaderListener<PayloadType>* listener{nullptr};
        /// last sequence number we observed.
        uint32_t lastObservedSequenceNumber{0};
    };

    /// Encapsulates references or storage to DDS Reader related entities for one channel.
    struct WriterContext
    {
        /// The actual DDS reader.
        PayloadDataWriter writer{dds::core::null};
        /// The name for our QoS profile used, if applicable.
        std::optional<std::string> qosProfileNameOpt;
        /// DDS Writer Listener
        DataWriterListener<PayloadType>* listener{nullptr};
        /// rolling publisher sequence number
        uint32_t sequenceNumber{0};
    };

    /// Create subscriber entity if it doesn't exist
    /// @throws dds::core::Exception on DDS internal error
    bool ensureSubscriber();
    /// Create publisher entity if it doesn't exist
    /// @throws dds::core::Exception on DDS internal error
    bool ensurePublisher();
    /// Create topic entity if it doesn't exist
    /// @throws dds::core::Exception on DDS internal error
    PayloadTopicType ensureTopic(const std::string& channel);

    /// Create a reader and handle exceptions
    /// Should only be called with clientMtx_ protection
    /// @return Ok if successfully created reader, else Error
    SimpleComms::Status::States createReaderHelper(const std::string& channel,
                                                   const SimpleComms::ChannelType type,
                                                   const uint32_t receiveCapacity);

    /// Create a writer and handle exceptions
    /// Should only be called with clientMtx_ protection
    /// @return Ok if successfully created reader, else Error
    SimpleComms::Status::States createWriterHelper(const std::string& channel, const SimpleComms::ChannelType type);

    /// Create a data writer for the channel, storing it internally
    /// @throws dds::core::Exception on DDS internal error or bad qosProfileNameOpt name
    void createWriter(const std::string& channel, std::optional<std::string> qosProfileNameOpt = std::nullopt);
    /// Create a data reader for the channel, storing it internally
    /// @throws dds::core::Exception on DDS internal error or bad qosProfileNameOpt name
    void createReader(const std::string& channel,
                      const std::optional<uint32_t>& receiveCapacityOpt,
                      std::optional<std::string> qosProfileNameOpt = std::nullopt);
    /// Check if we are connected. Using domainParticipant_ as the connected indicator.
    bool isConnected() const;
    /// Method called inside of dedicated thread if a recv callback is set
    void threadEntry();
    /// Reset DDS entities. Should be done under exclusive lock.
    void resetEntities();
    /// generate a MessageKey based on DDS Data and SampleInfo
    static MessageKey generateMsgKey(const PayloadType& data, const dds::sub::SampleInfo& info);
    /// populate msgInfo fields given the DDS sampleInfo.
    static void populateMessageInfo(const dds::sub::SampleInfo& sampleInfo, MessageInfo& msgInfo);
    /// DDS domain ID. Should be non-negative, but DDS uses it as signed in their APIs
    int32_t domainId_;
    /// The DDS Configurator strategy to use.
    std::shared_ptr<DDSConfigurator> config_;
    /// Our receive thread handle. Only started if we have a receive callback.
    std::thread thread_;
    /// Shared mutex protecting our connection state and DDS entity object lifetimes.
    /// DDSClient does not maintain any queues at the application level, and it sends and receives
    /// messages via threadsafe calls to DDS DataReaders and DataWriters. Therefore, we only
    /// need to ensure that our DDS entities are not created or destroyed while using them.
    /// We lock exclusively when creating and deleting entities, but lock shared when using the
    /// entities for reading,  writing, or status checking.
    std::shared_mutex clientMtx_;

    /// storage and lookup for our readers and related entities
    std::unordered_map<std::string, ReaderContext> readers_;
    /// storage and lookup for our writers and related entities
    std::unordered_map<std::string, WriterContext> writers_;

    /// @section DDS reference type entities

    /// DDS Waitset waited on in receive thread.
    dds::core::cond::WaitSet waitset_{dds::core::null};
    /// DDS GuardCondition attached to waitset, used to signal waitset shutdown.
    dds::core::cond::GuardCondition unblockCondition{dds::core::null};
    /// DDS Quality of Service provider
    /// DDS Domain Participant. May be shared with other users in process.
    DomainParticipant domainParticipant_{dds::core::null};
    /// DDS Publisher
    Publisher publisher_{dds::core::null};
    /// DDS Subscriber
    Subscriber subscriber_{dds::core::null};
};

} // namespace DDSComms
