#include <gtest/gtest.h>

#include "CycloneConfigurator.hpp"
#include "DDSConfiguratorTest.hpp"

#include "lib/avlog/avlog.hpp"

#include <exception>
#include <iostream>
#include <memory>
#include <string>

#include <dds/dds.hpp>

namespace {

using namespace DDSComms;

TEST(QoSConfiguratorTestBasic, TestDefault)
{
    CycloneOptions options{};

    // Since we are not installed with access to the default QoS file inside of OTTO_DIR,
    // we must set the QoS file ourselves.
    EXPECT_NO_THROW(CycloneConfigurator{options});
}

//! @brief Test Fixture DDSClient Tests
class QoSConfiguratorTest : public ::testing::Test
{
protected:
    void SetUp() override
    {
        // reader and writer
        options.readChannelProfileMap.insert({bestEffortProfileName, bestEffortProfileName});
        options.readChannelProfileMap.insert({strictReliableProfileName, strictReliableProfileName});
        options.readChannelProfileMap.insert({keepLast1ReliableProfileName, keepLast1ReliableProfileName});
        options.readChannelProfileMap.insert({keepLastReliableProfileName, keepLastReliableProfileName});
        options.readChannelProfileMap.insert({keepLast10ReliableProfileName, keepLast10ReliableProfileName});
        options.readChannelProfileMap.insert({keepLast5ReliableProfileName, keepLast5ReliableProfileName});
        options.readChannelProfileMap.insert({keepLast10ReliableProfileName, keepLast10ReliableProfileName});
        options.readChannelProfileMap.insert({keepLast25ReliableProfileName, keepLast25ReliableProfileName});
        options.readChannelProfileMap.insert({keepLast50ReliableProfileName, keepLast50ReliableProfileName});
        options.readChannelProfileMap.insert({keepLast100ReliableProfileName, keepLast100ReliableProfileName});
        options.readChannelProfileMap.insert(
            {keepLastReliableTransientLocalProfileName, keepLastReliableTransientLocalProfileName});

        options.writeChannelProfileMap.insert({bestEffortProfileName, bestEffortProfileName});
        options.writeChannelProfileMap.insert({strictReliableProfileName, strictReliableProfileName});
        options.writeChannelProfileMap.insert({keepLast1ReliableProfileName, keepLast1ReliableProfileName});
        options.writeChannelProfileMap.insert({keepLastReliableProfileName, keepLastReliableProfileName});
        options.writeChannelProfileMap.insert({keepLast5ReliableProfileName, keepLast5ReliableProfileName});
        options.writeChannelProfileMap.insert({keepLast10ReliableProfileName, keepLast10ReliableProfileName});
        options.writeChannelProfileMap.insert({keepLast25ReliableProfileName, keepLast25ReliableProfileName});
        options.writeChannelProfileMap.insert({keepLast50ReliableProfileName, keepLast50ReliableProfileName});
        options.writeChannelProfileMap.insert({keepLast100ReliableProfileName, keepLast100ReliableProfileName});
        options.writeChannelProfileMap.insert(
            {keepLastReliableTransientLocalProfileName, keepLastReliableTransientLocalProfileName});

        // reader only
        options.readChannelProfileMap.insert({keepLast5BestEffortProfileName, keepLast5BestEffortProfileName});
        options.readChannelProfileMap.insert({keepLast10BestEffortProfileName, keepLast10BestEffortProfileName});

        // error case profiles
        options.readChannelProfileMap.insert({nonExistentProfileName, nonExistentProfileName});
        options.writeChannelProfileMap.insert({nonExistentProfileName, nonExistentProfileName});

        // writer only
        options.writeChannelProfileMap.insert(
            {StrictReliableAsyncWriteProfileName, StrictReliableAsyncWriteProfileName});

        options.writeChannelProfileMap.insert(
            {KeepLast5ReliableAsyncWriteProfileName, KeepLast5ReliableAsyncWriteProfileName});

        options.writeChannelProfileMap.insert(
            {KeepLast10ReliableAsyncWriteProfileName, KeepLast10ReliableAsyncWriteProfileName});

        config = std::make_shared<CycloneConfigurator>(options);
    }
    CycloneOptions options{};
    std::shared_ptr<CycloneConfigurator> config{nullptr};
    static const dds::core::Duration readerLatencyBudget;
    static const dds::core::Duration asyncWriterLatencyBudget;
};

const dds::core::Duration QoSConfiguratorTest::readerLatencyBudget{0, 1000000};     // 1ms
const dds::core::Duration QoSConfiguratorTest::asyncWriterLatencyBudget{0, 100000}; // 100us

using QoSConfiguratorReaderTest = QoSConfiguratorTest;
using QoSConfiguratorWriterTest = QoSConfiguratorTest;
using QoSConfiguratorParticipantTest = QoSConfiguratorTest;

/// @section dds-config-reader DDS Configurator Reader QoS Tests

// default comes from CycloneOptions struct default
TEST_F(QoSConfiguratorReaderTest, DefProfile)
{
    auto qos = config->createReaderQos(std::nullopt, std::nullopt);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectReliable(qos);
    expectKeepLastN(qos, 50);
}

// default comes from CycloneOptions struct default, via unregistered channel
TEST_F(QoSConfiguratorReaderTest, DefProfile_UnregisteredChannel)
{
    auto qos = config->createReaderQos(std::nullopt, "ChannelNeverRegisteredAProfile");
    // we get the default settings from the config
    expectLatencyBudget(qos, readerLatencyBudget);
    expectReliable(qos);
    expectKeepLastN(qos, 50);
}

TEST_F(QoSConfiguratorReaderTest, BestEffortProfile)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, bestEffortProfileName);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectBestEffort(qos);
    expectKeepLastN(qos, 100);
}

TEST_F(QoSConfiguratorReaderTest, StrictReliable)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, strictReliableProfileName);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectReliable(qos);
    expectKeepAll(qos);
}

TEST_F(QoSConfiguratorReaderTest, KeepLast1Reliable)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLast1ReliableProfileName);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectKeepLastN(qos, 1);
    expectReliable(qos);
}

TEST_F(QoSConfiguratorReaderTest, KeepLastReliable)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLastReliableProfileName);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectReliable(qos);
    expectKeepLastN(qos, 1);
}

TEST_F(QoSConfiguratorReaderTest, keepLast5Reliable)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLast5ReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 5);
}

TEST_F(QoSConfiguratorReaderTest, keepLast10Reliable)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLast10ReliableProfileName);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectReliable(qos);
    expectKeepLastN(qos, 10);
}

TEST_F(QoSConfiguratorReaderTest, keepLast25Reliable)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLast25ReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 25);
}

TEST_F(QoSConfiguratorReaderTest, keepLast50Reliable)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLast50ReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 50);
}

TEST_F(QoSConfiguratorReaderTest, keepLast100Reliable)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLast100ReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 100);
}

TEST_F(QoSConfiguratorReaderTest, KeepLast5BestEffort)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLast5BestEffortProfileName);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectBestEffort(qos);
    expectKeepLastN(qos, 5);
}

TEST_F(QoSConfiguratorReaderTest, KeepLast10BestEffort)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLast10BestEffortProfileName);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectBestEffort(qos);
    expectKeepLastN(qos, 10);
}

TEST_F(QoSConfiguratorReaderTest, BestEffortProfile_UpdateQueue2)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, bestEffortProfileName);
    applyDataReaderQueueDepth(2, qos);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectBestEffort(qos);
    expectKeepLastN(qos, 2);
}

TEST_F(QoSConfiguratorReaderTest, BestEffortProfile_UpdateQueue2from5)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLast5BestEffortProfileName);
    applyDataReaderQueueDepth(2, qos);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectBestEffort(qos);
    expectKeepLastN(qos, 2);
}

TEST_F(QoSConfiguratorReaderTest, BestEffortProfile_UpdateQueueUnlimited)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, bestEffortProfileName);
    applyDataReaderQueueDepth(0, qos);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectBestEffort(qos);
    expectKeepAll(qos);
}

TEST_F(QoSConfiguratorReaderTest, strictReliableProfileName_UpdateQueue5)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, strictReliableProfileName);
    applyDataReaderQueueDepth(5, qos);
    expectLatencyBudget(qos, readerLatencyBudget);
    expectReliable(qos);
    expectKeepLastN(qos, 5);
}

TEST_F(QoSConfiguratorReaderTest, keepLastReliableTransientLocal)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLastReliableTransientLocalProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 1);
    expectTransientLocal(qos);
}

TEST_F(QoSConfiguratorReaderTest, BadProfileThrowError)
{
    EXPECT_THROW(dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, nonExistentProfileName);
                 , std::runtime_error);
}

/// @section dds-config-writer DDS Configurator Writer QoS Tests

// default comes from CycloneOptions struct default
TEST_F(QoSConfiguratorWriterTest, DefProfile)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(std::nullopt);
    expectReliable(qos);
    expectKeepLastN(qos, 25);
}

// default comes from CycloneOptions struct default
TEST_F(QoSConfiguratorWriterTest, DefProfile_UnregisteredChannel)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos("ChannelNeverRegisteredAProfile");
    expectReliable(qos);
    expectKeepLastN(qos, 25);
}

TEST_F(QoSConfiguratorWriterTest, BestEffortProfile)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(bestEffortProfileName);
    expectBestEffort(qos);
    expectKeepLastN(qos, 100);
}

TEST_F(QoSConfiguratorWriterTest, StrictReliable)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(strictReliableProfileName);
    expectReliable(qos);
    expectKeepAll(qos);
}

TEST_F(QoSConfiguratorWriterTest, KeepLast1Reliable)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(keepLast1ReliableProfileName);
    expectKeepLastN(qos, 1);
    expectReliable(qos);
}

TEST_F(QoSConfiguratorWriterTest, KeepLastReliable)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(keepLastReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 1);
}

TEST_F(QoSConfiguratorWriterTest, keepLast5Reliable)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(keepLast5ReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 5);
}

TEST_F(QoSConfiguratorWriterTest, keepLast10Reliable)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(keepLast10ReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 10);
}

TEST_F(QoSConfiguratorWriterTest, keepLast25Reliable)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(keepLast25ReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 25);
}

TEST_F(QoSConfiguratorWriterTest, keepLast50Reliable)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(keepLast50ReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 50);
}

TEST_F(QoSConfiguratorWriterTest, keepLast100Reliable)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(keepLast100ReliableProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 100);
}

TEST_F(QoSConfiguratorWriterTest, keepLastReliableTransientLocal)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(keepLastReliableTransientLocalProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 1);
    expectTransientLocal(qos);
}

TEST_F(QoSConfiguratorWriterTest, StrictReliableAsyncWrite)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(StrictReliableAsyncWriteProfileName);
    expectReliable(qos);
    expectKeepAll(qos);
    expectLatencyBudget(qos, asyncWriterLatencyBudget);
}

TEST_F(QoSConfiguratorWriterTest, KeepLast5ReliableAsyncWrite)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(KeepLast5ReliableAsyncWriteProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 5);
    expectLatencyBudget(qos, asyncWriterLatencyBudget);
}

TEST_F(QoSConfiguratorWriterTest, KeepLast10ReliableAsyncWrite)
{
    dds::pub::qos::DataWriterQos qos = config->createWriterQos(KeepLast10ReliableAsyncWriteProfileName);
    expectReliable(qos);
    expectKeepLastN(qos, 10);
    expectLatencyBudget(qos, asyncWriterLatencyBudget);
}

TEST_F(QoSConfiguratorWriterTest, BadProfileThrowError)
{
    EXPECT_THROW(dds::pub::qos::DataWriterQos qos = config->createWriterQos(nonExistentProfileName);
                 , std::runtime_error);
}

TEST_F(QoSConfiguratorParticipantTest, ParticipantQoSNoThrow)
{
    EXPECT_NO_THROW(dds::domain::qos::DomainParticipantQos qos = config->createParticipantQos(););
}

} // namespace
