#include <chrono>
#include <cstdlib>
#include <iostream>
#include <thread>

#include <gtest/gtest.h>

/* Include the C++ DDS API. */
#include "UserBytesTopicType.hpp"
#include "dds/dds.hpp"

namespace {

//! @brief Test Fixture testing Cyclone DDS usability
class CycloneTest : public ::testing::Test
{
protected:
    void SetUp() override
    {
        participant = dds::domain::DomainParticipant(std::stoi(std::getenv("DDS_DOMAIN_ID")));
        topic = dds::topic::Topic<BytesMsgType>(participant, "MyBytesTopic");
    }
    dds::domain::DomainParticipant participant{dds::core::null};
    dds::topic::Topic<BytesMsgType> topic{dds::core::null};
};

TEST(CycloneDefinedTest, CycloneDefined)
{
#ifdef DDS_VERSION_RTI_DDS_CPP
    FAIL() << "This test should be compiled with Cyclone DDS";
#endif
}

/// Sanity test to ensure we can use Cyclone DDS CXX
TEST_F(CycloneTest, CycloneWrite)
{
    dds::pub::DataWriter<BytesMsgType> writer(dds::pub::Publisher(participant), topic);

    std::vector<uint8_t> payload{1, 2, 3, 4, 5};
    BytesMsgType sample{};
    sample.payload(std::move(payload));

    // write two sample, in different ways
    writer.write(sample);
    writer << sample;
}

/// Simple Pub-Sub test on same participant
TEST_F(CycloneTest, CyclonePubSub)
{
    using namespace org::eclipse::cyclonedds;
    using namespace std::chrono_literals;

    // Create publisher
    dds::pub::Publisher publisher{participant};
    dds::pub::DataWriter<BytesMsgType> dataWriter{publisher, topic};

    // Check that there are no subscribers
    ASSERT_EQ(dataWriter.publication_matched_status().current_count(), 0);

    // Create subscriber
    dds::sub::Subscriber subscriber{participant};
    dds::sub::DataReader<BytesMsgType> dataReader{subscriber, topic};

    // configure waitset to block until we receive a msg
    dds::core::cond::WaitSet waitset;
    dds::sub::cond::ReadCondition readCond(dataReader, dds::sub::status::DataState());
    waitset += readCond;

    // Wait for subscriber to be discovered
    while (dataWriter.publication_matched_status().current_count() == 0) {
        std::this_thread::sleep_for(100ms);
    }

    // Check that subscriber is discovered
    ASSERT_EQ(dataWriter.publication_matched_status().current_count(), 1);

    // Write a message
    std::vector<uint8_t> payload{1, 2, 3, 4, 5};
    BytesMsgType sentBytesMsg{};
    sentBytesMsg.payload(std::move(payload));
    dataWriter.write(sentBytesMsg);

    // wait until we receive a message, for at most 1 second
    auto active_conditions = waitset.wait(dds::core::Duration(1, 0));

    // waitset unblocked, so read a message.
    dds::sub::LoanedSamples<BytesMsgType> readSamples{};

    // read the sampled via steam API
    dataReader >> dds::sub::read >> readSamples;

    ASSERT_EQ(readSamples.length(), 1);
    auto& readInfo = readSamples.begin()->info();
    ASSERT_TRUE(readInfo.valid());
    auto& readReceivedByteMsg = readSamples.begin()->data();
    ASSERT_TRUE(readReceivedByteMsg == sentBytesMsg);

    // Take a message (msg still exists since we didn't 'take' it in last call)
    dds::sub::LoanedSamples<BytesMsgType> takeSamples = dataReader.take();
    ASSERT_EQ(takeSamples.length(), 1);
    auto& takeInfo = takeSamples.begin()->info();
    ASSERT_TRUE(takeInfo.valid());
    auto& takeReceivedByteMsg = takeSamples.begin()->data();
    ASSERT_TRUE(takeReceivedByteMsg == sentBytesMsg);

    // Verify that there is no more data to be read
    dds::sub::LoanedSamples<BytesMsgType> readSamplesAgain = dataReader.read();
    ASSERT_EQ(readSamplesAgain.length(), 0);
}

/// cyclone internal function tested for code coverage
TEST_F(CycloneTest, BytesMsgType_CdrRead)
{
    BytesMsgType sample{};
    basic_cdr_stream str;
    std::vector<unsigned char> vec(8, 0x0);
    str.set_buffer(vec.data());
    org::eclipse::cyclonedds::core::cdr::read<basic_cdr_stream>(str, sample, 1);
}

} // namespace
