#include <gtest/gtest.h>

#include "FakeReceiver.hpp"
#include "lib/avlog/avlog.hpp"
#include "platform/middleware/libs/DDSComms/test/DDSTestHelper.hpp"
#include <DDSComms/DDSClient.hpp>
#include <chrono>
#include <exception>
#include <iostream>
#include <memory>
#include <string>
#include <thread>

#include <dds/dds.hpp>

using namespace std::placeholders;

namespace {

using namespace SimpleComms;
using namespace DDSComms;

const std::string testChannelName{"TestChannel"};
const std::string testData{"testdata"};
std::vector<av::latency::Pred> preds{};

using DDSComms::testing::getTestCfg;

TEST(DDSClientTestInstantiation, ctor)
{
    DDSClient client{Client::Callback(), std::stoi(std::getenv("DDS_DOMAIN_ID")), getTestCfg()};
}

TEST(DDSClientTestInstantiation, ConstructorDestructorNoThrow)
{
    auto client =
        std::make_shared<DDSClient>(Client::Callback(), std::stoi(std::getenv("DDS_DOMAIN_ID")), getTestCfg());
    EXPECT_NO_THROW(client.reset());
}

//! @brief Test Fixture DDSClient Tests
class DDSClientTest : public ::testing::Test
{

protected:
    DDSClient client{Client::Callback(), std::stoi(std::getenv("DDS_DOMAIN_ID")), getTestCfg()};

    void SetUp() override {}

    void TearDown() override {}
};

TEST_F(DDSClientTest, ConnectDisconnect)
{
    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
}

TEST_F(DDSClientTest, DoubleConnect)
{
    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.connect(), Status::States::AlreadyConnected);
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
}

TEST_F(DDSClientTest, DisconnectOnDisconnectedSilentlyAllows)
{
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.connect(), Status::States::AlreadyConnected);
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
}

TEST_F(DDSClientTest, CannotDeclareReadBeforeConnect)
{

    EXPECT_EQ(client.declare(testChannelName, ChannelType::Subscriber, 1), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotDeclareWriteBeforeConnect)
{

    EXPECT_EQ(client.declare(testChannelName, ChannelType::Publisher, 1), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotDeclareReadWriteBeforeConnect)
{

    EXPECT_EQ(client.declare(testChannelName, ChannelType::SubPub, 1), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotUnDeclareReadBeforeConnect)
{

    EXPECT_EQ(client.undeclare(testChannelName, ChannelType::Subscriber), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotUnDeclareWriteBeforeConnect)
{

    EXPECT_EQ(client.undeclare(testChannelName, ChannelType::Publisher), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotUnDeclareReadWriteBeforeConnect)
{

    EXPECT_EQ(client.undeclare(testChannelName, ChannelType::SubPub), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotWriteBeforeConnect)
{

    EXPECT_EQ(client.sendMsgV("NeverDeclaredChannel", testData, std::move(preds)), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotReadBeforeConnect)
{

    std::string inputMsg;
    EXPECT_EQ(client.receiveMsg("NeverDeclaredChannel", inputMsg), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotUnDeclareReadWriteOnDisconnected)
{

    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.undeclare(testChannelName, ChannelType::SubPub), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotWriteOnDisconnected)
{

    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.sendMsgV("NeverDeclaredChannel", testData, std::move(preds)), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CannotReadOnDisconnected)
{

    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    std::string inputMsg;
    EXPECT_EQ(client.receiveMsg("NeverDeclaredChannel", inputMsg), Status::States::NotConnected);
}

TEST_F(DDSClientTest, CanCallReadOnReconnected)
{

    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.declare(testChannelName, ChannelType::SubPub, 1), Status::States::Ok);
    std::string inputMsg;
    EXPECT_EQ(client.receiveMsg(testChannelName, inputMsg), Status::States::Ok); // no data, but valid
}

TEST_F(DDSClientTest, ReconnectClearsChannels)
{

    std::string inputMsg;
    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.declare(testChannelName, ChannelType::SubPub, 1), Status::States::Ok);
    EXPECT_EQ(client.receiveMsg(testChannelName, inputMsg), Status::States::Ok); // no data, but valid
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.receiveMsg(testChannelName, inputMsg), Status::States::InvalidHandle);
}

//! @brief Test Fixture DDSClient Tests
class DDSClientTestBasic : public ::testing::Test
{

protected:
    DDSClient client{Client::Callback(), std::stoi(std::getenv("DDS_DOMAIN_ID")), getTestCfg()};

    void SetUp() override { EXPECT_EQ(client.connect(), Status::States::Ok); }

    void TearDown() override { EXPECT_EQ(client.disconnect(), Status::States::Ok); }
};

TEST_F(DDSClientTestBasic, CannotDeclareAfterDisconnect)
{
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.declare(testChannelName, ChannelType::SubPub, 1), Status::States::NotConnected);
}

TEST_F(DDSClientTestBasic, DeclareAfterReconnect)
{
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.declare(testChannelName, ChannelType::SubPub, 1), Status::States::NotConnected);
    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.declare(testChannelName, ChannelType::SubPub, 1), Status::States::Ok);
}

TEST_F(DDSClientTestBasic, UndeclareNonexistentSilentlyAllow)
{
    EXPECT_EQ(client.undeclare(testChannelName, ChannelType::SubPub), Status::States::Ok);
}

TEST_F(DDSClientTestBasic, DeclareWriterTwiceSilentlyAllow)
{
    EXPECT_EQ(client.declare(testChannelName, ChannelType::Publisher, 1), Status::States::Ok);
    EXPECT_EQ(client.declare(testChannelName, ChannelType::Publisher, 1), Status::States::Ok);
}

TEST_F(DDSClientTestBasic, DeclareReaderTwiceSilentlyAllow)
{
    EXPECT_EQ(client.declare(testChannelName, ChannelType::Subscriber, 1), Status::States::Ok);
    EXPECT_EQ(client.declare(testChannelName, ChannelType::Subscriber, 1), Status::States::Ok);
}

TEST_F(DDSClientTestBasic, WaitForMatchedPeerTimesOut)
{
    std::string channelName = testChannelName + "_WaitForMatchedPeerTimesOut";
    EXPECT_EQ(client.declare(channelName, ChannelType::Subscriber, 1), Status::States::Ok);
    EXPECT_EQ(client.waitForMatchedPeer(channelName, ChannelType::Subscriber), Status::States::Error);
}

TEST_F(DDSClientTestBasic, CannotWriteOnUndeclared)
{
    EXPECT_EQ(client.sendMsgV("NeverDeclaredChannel", testData, std::move(preds)), Status::States::InvalidHandle);
}

TEST_F(DDSClientTestBasic, CannotReadOnUndeclared)
{
    std::string inputMsg;
    EXPECT_EQ(client.receiveMsg("NeverDeclaredChannel", inputMsg), Status::States::InvalidHandle);
}

TEST_F(DDSClientTestBasic, SendReceivePollFullTrip)
{
    std::string sentMsg{testData};

    EXPECT_EQ(client.declare(testChannelName, ChannelType::SubPub, 2), Status::States::Ok);
    EXPECT_EQ(client.waitForMatchedPeer(testChannelName, ChannelType::Publisher), Status::States::Ok);

    TimeStamp sendTime{};
    sendTime.setNow();
    EXPECT_EQ(client.sendMsgV(testChannelName, sentMsg, std::move(preds)), Status::States::Ok);

    std::string recvdMsg{};
    MessageInfo recvdMsgInfo{};
    Status::States retState = Status::States::Error;
    auto startPolling = std::chrono::steady_clock::now();

    while (std::chrono::steady_clock::now() - startPolling <= 5s) {
        retState = client.receiveMsg(testChannelName, recvdMsg, recvdMsgInfo);
        if (retState == Status::States::NewData) {
            break; // we received our message
        }
        std::this_thread::sleep_for(1ms);
    }
    ASSERT_EQ(retState, Status::States::NewData) << "Did not receive expected message via polling ";
    EXPECT_STREQ(recvdMsg.c_str(), sentMsg.c_str());

    // sanity check our timestamps
    ASSERT_TRUE(recvdMsgInfo.msgReadyTime >= recvdMsgInfo.sendMsgTime);
    ASSERT_TRUE(recvdMsgInfo.msgReadyTime >= sendTime); // assumes DDS times using wallclock time, not monotonic
    EXPECT_TRUE(recvdMsgInfo.msgReadyTime - recvdMsgInfo.sendMsgTime < TimeStamp{.500});
    EXPECT_TRUE(recvdMsgInfo.msgReadyTime - sendTime < TimeStamp{.500});
}

TEST_F(DDSClientTestBasic, emptyPublishListMessage)
{
    EXPECT_TRUE(client.getPublishListMessage().empty());
}

TEST_F(DDSClientTestBasic, nonEmptyPublishListMessage)
{
    EXPECT_EQ(client.declare(testChannelName, ChannelType::Publisher, 1), Status::States::Ok);
    EXPECT_EQ(client.getPublishListMessage().size(), 1);
}

TEST(DDSClientTestCallback, Receive1WithWaitSet)
{
    FakeReceiver receiver;

    DDSClient client{receiver.getBoundCallback(), 0, getTestCfg()};

    // ensure we can reconnect and function
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.connect(), Status::States::Ok);
    EXPECT_EQ(client.disconnect(), Status::States::Ok);
    EXPECT_EQ(client.connect(), Status::States::Ok);

    const uint32_t expectedMsgs = 1;
    receiver.setExpectMsgs(expectedMsgs);

    std::string sentMsg{testData};

    EXPECT_EQ(client.declare(testChannelName, ChannelType::SubPub, 1), Status::States::Ok);
    EXPECT_EQ(client.waitForMatchedPeer(testChannelName, ChannelType::Publisher), Status::States::Ok);

    EXPECT_EQ(client.sendMsgV(testChannelName, sentMsg, std::move(preds)), Status::States::Ok);
    EXPECT_EQ(receiver.wait(), true);
    EXPECT_EQ(receiver.getNumReceivedData(), expectedMsgs);

    // ensure we cannot read the message via polling because it was already taken via callback
    std::string recvdMsg;
    EXPECT_EQ(client.receiveMsg(testChannelName, recvdMsg), Status::States::Ok); // Ok means no data
    client.disconnect();
}

TEST(DDSClientTestCallback, Receive5On1Channel)
{
    FakeReceiver receiver;
    DDSClient client{receiver.getBoundCallback(), 0, getTestCfg()};

    client.connect();

    const uint32_t expectedMsgs = 5;
    EXPECT_EQ(client.declare(testChannelName, ChannelType::SubPub, expectedMsgs), Status::States::Ok);
    EXPECT_EQ(client.waitForMatchedPeer(testChannelName, ChannelType::Publisher), Status::States::Ok);

    receiver.setExpectMsgs(expectedMsgs);
    for (uint32_t i = 0; i < expectedMsgs; ++i) {
        std::string sentMsg{testData + std::to_string(i)};
        EXPECT_EQ(client.sendMsgV(testChannelName, sentMsg, std::move(preds)), Status::States::Ok);
    }

    EXPECT_EQ(receiver.wait(), true);
    EXPECT_EQ(receiver.getNumReceivedData(), expectedMsgs);

    // ensure we cannot read the message via polling because it was already taken via callback
    std::string recvdMsg;
    Status::States retState = client.receiveMsg(testChannelName, recvdMsg);
    EXPECT_EQ(retState, Status::States::Ok); // Ok means no data
    client.disconnect();
}

TEST(DDSClientTestCallback, Receive5On5Channels)
{
    FakeReceiver receiver;
    DDSClient client{receiver.getBoundCallback(), 0, getTestCfg()};

    client.connect();

    const uint32_t numberChannels = 5;
    const uint32_t expectedMsgsPerChannel = 5;
    const uint32_t expectedMsgs = expectedMsgsPerChannel * numberChannels;
    receiver.setExpectMsgs(expectedMsgs);

    std::vector<std::string> channelNames{};
    for (uint32_t i = 0; i < numberChannels; ++i) {
        channelNames.push_back(testChannelName + std::to_string(i));
        EXPECT_EQ(client.declare(channelNames[i], ChannelType::SubPub, expectedMsgsPerChannel), Status::States::Ok);
        EXPECT_EQ(client.waitForMatchedPeer(channelNames[i], ChannelType::Publisher), Status::States::Ok);
    }

    for (uint32_t i = 0; i < expectedMsgsPerChannel; ++i) {
        const std::string sentMsg{testData + std::to_string(i)};
        for (const std::string& channelName : channelNames) {
            EXPECT_EQ(client.sendMsgV(channelName, sentMsg, std::move(preds)), Status::States::Ok);
        }
    }

    EXPECT_EQ(receiver.wait(), true);
    EXPECT_EQ(receiver.getNumReceivedData(), expectedMsgs);

    std::string recvdMsg;
    for (const std::string& channelName : channelNames) {
        EXPECT_EQ(client.receiveMsg(channelName, recvdMsg), Status::States::Ok)
            << "we should have gotten no data here if callback was working as expected ";
    }

    client.disconnect();
}

} // namespace
