#include <gtest/gtest.h>

#include "lib/avlog/avlog.hpp"

#include <exception>
#include <iostream>
#include <memory>
#include <string>

#include <dds/dds.hpp>

namespace {

// reader and writer profile names
static constexpr auto bestEffortProfileName = "AVQosLib::Generic.BestEffort";
static constexpr auto strictReliableProfileName = "AVQosLib::Generic.StrictReliable";
static constexpr auto keepLast1ReliableProfileName = "AVQosLib::Generic.KeepLast1Reliable";
static constexpr auto keepLastReliableProfileName = "AVQosLib::Generic.KeepLastReliable"; // == KeepLast1Reliable
static constexpr auto keepLast5ReliableProfileName = "AVQosLib::Generic.KeepLast5Reliable";
static constexpr auto keepLast10ReliableProfileName = "AVQosLib::Generic.KeepLast10Reliable";
static constexpr auto keepLast25ReliableProfileName = "AVQosLib::Generic.KeepLast25Reliable";
static constexpr auto keepLast50ReliableProfileName = "AVQosLib::Generic.KeepLast50Reliable";
static constexpr auto keepLast100ReliableProfileName = "AVQosLib::Generic.KeepLast100Reliable";

static constexpr auto keepLastReliableTransientLocalProfileName = "AVQosLib::Generic.KeepLastReliable.TransientLocal";
static constexpr auto nonExistentProfileName = "BadProfileDoesntExist";

// reader only profiles names
static constexpr auto keepLast5BestEffortProfileName = "AVQosLib::Generic.KeepLast5BestEffort";
static constexpr auto keepLast10BestEffortProfileName = "AVQosLib::Generic.KeepLast10BestEffort";

// writer only profile names
static constexpr auto StrictReliableAsyncWriteProfileName = "AVQosLib::Generic.StrictReliableAsyncWrite";
static constexpr auto KeepLast5ReliableAsyncWriteProfileName = "AVQosLib::Generic.KeepLast5ReliableAsyncWrite";
static constexpr auto KeepLast10ReliableAsyncWriteProfileName = "AVQosLib::Generic.KeepLast10ReliableAsyncWrite";

bool valueIsUnlimitedOrAtLeastAsExpected(int32_t value, int32_t expected)
{
    return value == dds::core::LENGTH_UNLIMITED || value >= expected;
}

template<class T>
void expectReliable(const T& qos)
{
    const dds::core::policy::Reliability& reliability = qos.template policy<dds::core::policy::Reliability>();
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::RELIABLE);
}

template<class T>
void expectBestEffort(const T& qos)
{
    const dds::core::policy::Reliability& reliability = qos.template policy<dds::core::policy::Reliability>();
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::BEST_EFFORT);
}

template<class T>
void expectKeepAll(const T& qos)
{
    const dds::core::policy::History& history = qos.template policy<dds::core::policy::History>();
    EXPECT_EQ(history.kind(), dds::core::policy::HistoryKind::KEEP_ALL);
    // note that history depth is ignored when using KEEP_ALL, so no we do not check it.
}

template<class T>
void expectTransientLocal(const T& qos)
{
    const auto& durability = qos.template policy<dds::core::policy::Durability>();
    EXPECT_EQ(durability.kind(), dds::core::policy::DurabilityKind::TRANSIENT_LOCAL);
}

template<class T>
void expectLatencyBudget(const T& qos, const dds::core::Duration duration)
{
    const auto& latencyBudget = qos.template policy<dds::core::policy::LatencyBudget>();
    EXPECT_TRUE(latencyBudget.duration() == duration);
}

template<class T>
void expectKeepLastN(const T& qos, const int depth = 1)
{
    const dds::core::policy::History& history = qos.template policy<dds::core::policy::History>();
    const dds::core::policy::ResourceLimits& resourceLimits = qos.template policy<dds::core::policy::ResourceLimits>();
    // history depth should be exactly as requested
    EXPECT_EQ(history.depth(), depth);
    EXPECT_EQ(history.kind(), dds::core::policy::HistoryKind::KEEP_LAST);
    // resource limits just need to be sufficient
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples_per_instance(), depth));
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples(), depth));
}

} // namespace
