#include <gtest/gtest.h>

#include "DDSConfiguratorTest.hpp"
#include "RtiProConfigurator.hpp"

#include "lib/avlog/avlog.hpp"

#include <exception>
#include <iostream>
#include <memory>
#include <string>

#include <dds/dds.hpp>

namespace {

using namespace DDSComms;

// This profile is custom, laoded from testing QoS File
static constexpr auto lowSamplesPerInstanceProfileName = "TestAVQosLib::Generic.Common.LowSamplesPerInstance";

// the path to our QoS file
static constexpr auto testQosXml = "config/common/USER_QOS_PROFILES.xml";

TEST(QoSConfiguratorTestBasic, TestWithInvalidQoSFile)
{
    RtiProOptions options{};
    options.configFileName = "I/AM/NOT/REAL.xml";
    ASSERT_THROW(RtiProConfigurator{options};, std::runtime_error);
}

TEST(QoSConfiguratorTestBasic, TestWithQoSFile)
{
    RtiProOptions options{};
    options.configFileName = testQosXml;
    RtiProConfigurator{options};
}

//! @brief Test Fixture DDSClient Tests
class QoSConfiguratorTest : public ::testing::Test
{
protected:
    void SetUp() override
    {
        options.configFileName = testQosXml;

        options.readChannelProfileMap.insert({bestEffortProfileName, bestEffortProfileName});
        options.readChannelProfileMap.insert({strictReliableProfileName, strictReliableProfileName});
        options.readChannelProfileMap.insert({keepLastReliableProfileName, keepLastReliableProfileName});
        options.readChannelProfileMap.insert({lowSamplesPerInstanceProfileName, lowSamplesPerInstanceProfileName});

        config = std::make_shared<RtiProConfigurator>(options);
    }
    RtiProOptions options{};
    std::shared_ptr<RtiProConfigurator> config{nullptr};
};

using QoSConfiguratorReaderTest = QoSConfiguratorTest;
using QoSConfiguratorWriterTest = QoSConfiguratorTest;

TEST_F(QoSConfiguratorReaderTest, DefProfile_queueSize1)
{
    const auto queueDepth = 1;
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(queueDepth, std::nullopt);

    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();

    EXPECT_EQ(history.depth(), queueDepth);
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples_per_instance(), queueDepth));
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples(), queueDepth));
}

TEST_F(QoSConfiguratorReaderTest, DefProfile_queueSize110)
{
    const auto queueDepth = 110;
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(queueDepth, std::nullopt);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();

    EXPECT_EQ(history.depth(), queueDepth);
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples_per_instance(), queueDepth));
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples(), queueDepth));
}

TEST_F(QoSConfiguratorReaderTest, DefProfile_queueSize0Unlimited)
{
    const auto queueDepth = 0;
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(queueDepth, std::nullopt);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();

    EXPECT_EQ(history.depth(), dds::core::LENGTH_UNLIMITED);
    EXPECT_EQ(resourceLimits.max_samples_per_instance(), dds::core::LENGTH_UNLIMITED);
    EXPECT_EQ(resourceLimits.max_samples(), dds::core::LENGTH_UNLIMITED);
}

TEST_F(QoSConfiguratorReaderTest, BestEffortProfile_NoChangeJustProfile)
{

    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, bestEffortProfileName);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    dds::core::policy::Reliability& reliability = qos.policy<dds::core::policy::Reliability>();

    const auto expectedQueueDepth = 100;
    EXPECT_EQ(history.depth(), expectedQueueDepth);
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples_per_instance(), expectedQueueDepth));
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples(), expectedQueueDepth));
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::BEST_EFFORT);
}

TEST_F(QoSConfiguratorReaderTest, BestEffortProfile_queueSize1)
{
    const auto queueDepth = 1;
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(queueDepth, bestEffortProfileName);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    dds::core::policy::Reliability& reliability = qos.policy<dds::core::policy::Reliability>();

    EXPECT_EQ(history.depth(), queueDepth);
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples_per_instance(), queueDepth));
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples(), queueDepth));
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::BEST_EFFORT);
}

TEST_F(QoSConfiguratorReaderTest, BestEffortProfile_queueSize110)
{
    const auto queueDepth = 110;
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(queueDepth, bestEffortProfileName);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    dds::core::policy::Reliability& reliability = qos.policy<dds::core::policy::Reliability>();

    EXPECT_EQ(history.depth(), queueDepth);
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples_per_instance(), queueDepth));
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples(), queueDepth));
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::BEST_EFFORT);
}

TEST_F(QoSConfiguratorReaderTest, StrictReliableProfile_NoChangeJustProfile)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, strictReliableProfileName);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    dds::core::policy::Reliability& reliability = qos.policy<dds::core::policy::Reliability>();

    EXPECT_EQ(history.depth(), 1);
    EXPECT_EQ(resourceLimits.max_samples_per_instance(), dds::core::LENGTH_UNLIMITED);
    EXPECT_EQ(resourceLimits.max_samples(), 256);
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::RELIABLE);
}

TEST_F(QoSConfiguratorReaderTest, StrictReliableProfile_queueSize110)
{
    const auto queueDepth = 110;
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(queueDepth, strictReliableProfileName);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    dds::core::policy::Reliability& reliability = qos.policy<dds::core::policy::Reliability>();

    EXPECT_EQ(history.depth(), queueDepth);
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples_per_instance(), queueDepth));
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples(), queueDepth));
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::RELIABLE);
}

TEST_F(QoSConfiguratorReaderTest, StrictReliableProfile_queueSize257)
{
    const auto queueDepth = 257;
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(queueDepth, strictReliableProfileName);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    dds::core::policy::Reliability& reliability = qos.policy<dds::core::policy::Reliability>();

    EXPECT_EQ(history.depth(), queueDepth);
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples_per_instance(), queueDepth));
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples(), queueDepth));
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::RELIABLE);
}

TEST_F(QoSConfiguratorReaderTest, KeepLastReliableProfile_NoChangeJustProfile)
{
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(std::nullopt, keepLastReliableProfileName);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    dds::core::policy::Reliability& reliability = qos.policy<dds::core::policy::Reliability>();

    const auto expectedQueueDepth = 1;
    EXPECT_EQ(history.depth(), expectedQueueDepth);
    EXPECT_EQ(resourceLimits.max_samples_per_instance(), dds::core::LENGTH_UNLIMITED);
    EXPECT_EQ(resourceLimits.max_samples(), dds::core::LENGTH_UNLIMITED);
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::RELIABLE);
}

TEST_F(QoSConfiguratorReaderTest, LowSamplePerInstanceProfile_queueSize100)
{
    const auto queueDepth = 100;
    dds::sub::qos::DataReaderQos qos = config->createReaderQos(queueDepth, lowSamplesPerInstanceProfileName);
    dds::core::policy::History& history = qos.policy<dds::core::policy::History>();
    dds::core::policy::ResourceLimits& resourceLimits = qos.policy<dds::core::policy::ResourceLimits>();
    dds::core::policy::Reliability& reliability = qos.policy<dds::core::policy::Reliability>();

    EXPECT_EQ(history.depth(), queueDepth);
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples_per_instance(), queueDepth));
    EXPECT_TRUE(valueIsUnlimitedOrAtLeastAsExpected(resourceLimits.max_samples(), queueDepth));
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::BEST_EFFORT);
}

TEST_F(QoSConfiguratorReaderTest, DefProfile_DefaultProvider)
{
    // default everything
    EXPECT_NO_THROW(auto qos = config->createReaderQos());
}

TEST_F(QoSConfiguratorReaderTest, InvalidQueueSize)
{
    EXPECT_THROW({ auto qos = config->createReaderQos(std::numeric_limits<int>::max() + 1UL, std::nullopt); },
                 std::invalid_argument);
}

TEST_F(QoSConfiguratorWriterTest, DefProfile_DefaultProvider)
{
    // default everything
    EXPECT_NO_THROW(auto qos = config->createWriterQos());
}

TEST_F(QoSConfiguratorWriterTest, KeepLastReliableProfile_NoChangeJustProfile)
{
    auto qos = config->createWriterQos(keepLastReliableProfileName);
    dds::core::policy::Reliability& reliability = qos.policy<dds::core::policy::Reliability>();
    EXPECT_EQ(reliability.kind(), dds::core::policy::ReliabilityKind::RELIABLE);
}

TEST(QoSConfiguratorDomainParticipantTest, CreateDomainParticipandQosDefault)
{
    RtiProOptions options{};
    options.configFileName = testQosXml;
    RtiProConfigurator config{options};
    EXPECT_NO_THROW(auto qos = config.createParticipantQos());
}

TEST(QoSConfiguratorDomainParticipantTest, CreateDomainParticipandQosSpecified)
{
    RtiProOptions options{};
    options.configFileName = testQosXml;
    options.domainParticipantProfile = bestEffortProfileName;
    RtiProConfigurator config{options};

    EXPECT_NO_THROW(auto qos = config.createParticipantQos());
}

} // namespace
