/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @brief LocalizationPose helper functions implementation
 *
 * @section License
 *
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#include <IdlHelper/LocalizationPoseHelper.h>
#include <interfaces/Vector3d/Vector3d.h>

namespace idlhelper {

inline Eigen::Vector3d getEigenVector3d(const av::geometry::Vector3d& geom3d)
{
    return {geom3d.x, geom3d.y, geom3d.z};
}

LocalizationPoseHelper::LocalizationPoseHelper() = default;

LocalizationPoseHelper::LocalizationPoseHelper(LocalizationPose inPose)
    : LocalizationPose{std::move(inPose)}
{}

Eigen::Vector3d LocalizationPoseHelper::getPosition() const
{
    return getEigenVector3d(this->pose.position);
}

Eigen::Quaterniond LocalizationPoseHelper::getQuaternion() const
{
    const auto& q = this->pose.rotation;
    return {q.w, q.x, q.y, q.z};
}

const Eigen::Matrix3d Rx(double t)
{
    double st = sin(t), ct = cos(t);
    Eigen::Matrix3d m;
    m << 1, 0, 0, 0, ct, -st, 0, st, ct;
    return m;
}

const Eigen::Matrix3d Ry(double t)
{
    double st = sin(t), ct = cos(t);
    Eigen::Matrix3d m;
    m << ct, 0, st, 0, 1, 0, -st, 0, ct;
    return m;
}

Eigen::Vector3d LocalizationPoseHelper::getRollPitchYaw() const
{
    const auto A = getQuaternion().matrix();

    const double x = -std::atan2(-A(2, 1), A(2, 2));
    const auto Qx = Rx(-x);
    const auto B = A * Qx;

    const double y = -std::atan2(B(2, 0), B(2, 2));
    const auto Qy = Ry(-y);
    const auto C = B * Qy;

    double z = -std::atan2(-C(1, 0), C(1, 1));

    return Eigen::Vector3d(x, y, z);
}

Eigen::Vector3d LocalizationPoseHelper::getVelocity() const
{
    return getEigenVector3d(this->velocity);
}

Eigen::Vector3d LocalizationPoseHelper::getAngularRate() const
{
    return getEigenVector3d(this->angular_rate);
}

Eigen::Vector3d LocalizationPoseHelper::getAcceleration() const
{
    return getEigenVector3d(this->acceleration);
}

std::chrono::system_clock::time_point LocalizationPoseHelper::getTimestamp() const
{
    return std::chrono::system_clock::time_point{std::chrono::microseconds(this->timestamp)};
}

ottotime::otimePoint LocalizationPoseHelper::getTimePoint() const
{
    return ottotime::otimePoint{ottotime::oMicroSeconds{this->timestamp}};
}

} // namespace idlhelper
