/// \copyright
/// Copyright (c) 2021- Motional. All rights reserved.
/// NOTICE: Removing this heading is prohibited.
/// All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
/// any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
/// suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
/// secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
/// unless prior written permission is obtained from Motional.

#pragma once

#include <interfaces/IntersectionTLStatus/IntersectionTLStatus.h>
#include <interfaces/MultiIntersectionTLStatus/MultiIntersectionTLStatus.h>

#include "lib/maps/src/lib/avmap/Enums.hpp"
#include "lib/maps/src/lib/avmaputils/DynamicLaneConnectorUtils.hpp"

using namespace av::map;

namespace av::msg::utils {

/// Get the status of the TrafficLight controlling a LaneConnector if it exists.
/// @param laneConnector: A @ref LaneConnector.
/// @param intersectionTlStatus: Message containing information about TrafficLights at the LaneConnector's Intersection.
/// @return Optional status of the TrafficLight controlling the LaneConnector.
std::optional<av::msg::EdgeTLStatus> getEdgeTLStatus(const av::map::LaneConnector& laneConnector,
                                                     av::msg::IntersectionTLStatus intersectionTlStatus)
{
    std::vector<av::msg::EdgeTLStatus>
        assumedStraightStatus; // INVALID_ID is assumed to refer to the main edge (straight)
    std::optional<av::msg::EdgeTLStatus> matchedEdgeTLStatus;
    const auto toRoadBlock = laneConnector.roadBlockTo();
    if (nullptr == toRoadBlock) {
        return matchedEdgeTLStatus;
    }
    const auto targetId = toRoadBlock->id().toString();
    for (auto edgeTLStatusMapping : intersectionTlStatus.topology_edges) {
        if (edgeTLStatusMapping.to_roadblock_id == targetId) {
            matchedEdgeTLStatus = edgeTLStatusMapping.tl_status;
            // If the matching status is Circle shaped, keep looking in case another overrides it, but break as soon
            // as it finds an explicit signal (non-Circle)
            if (matchedEdgeTLStatus.value().shape != av::msg::LIGHT_SHAPE::CIRCLE) {
                break;
            }
        } else if (edgeTLStatusMapping.to_roadblock_id == av::map::ID::invalid.toString()) {
            assumedStraightStatus.push_back(edgeTLStatusMapping.tl_status);
        }
    }
    if ((!matchedEdgeTLStatus.has_value()) && (assumedStraightStatus.size() >= 1)) {
        // If there are more than one main edge, our assumptions are wrong
        AV_REQUIRE(assumedStraightStatus.size() == 1);
        matchedEdgeTLStatus = assumedStraightStatus.front();
    }
    return matchedEdgeTLStatus;
}

/// Get the status of the TrafficLight controlling a LaneConnector if it exists.
/// @param laneConnector: A @ref LaneConnector.
/// @param multiIntersectionTlStatus: Message containing information about TrafficLights.
/// @return Optional status of the TrafficLight controlling the LaneConnector.
std::optional<av::msg::EdgeTLStatus> getEdgeTLStatus(const av::map::LaneConnector& laneConnector,
                                                     av::msg::MultiIntersectionTLStatus multiIntersectionTlStatus)
{
    std::optional<av::msg::EdgeTLStatus> matchedEdgeTLStatus;
    const auto fromRoadBlock = laneConnector.roadBlockFrom();
    if (nullptr == fromRoadBlock) {
        return matchedEdgeTLStatus;
    }
    const auto fromRoadBlockId = fromRoadBlock->id().toString();
    for (auto intersectionTLStatus : multiIntersectionTlStatus.multi_intersection_status) {
        if (intersectionTLStatus.from_roadblock_id == fromRoadBlockId) {
            matchedEdgeTLStatus = getEdgeTLStatus(laneConnector, intersectionTLStatus);
            break;
        }
    }
    return matchedEdgeTLStatus;
}

/// Create LaneConnectorTrafficLightState provided LaneConnector (edge) and perception message.
/// @param laneConnector: The @ref LaneConnector of interest.
/// @param multiIntersectionTlStatus: Message describing all traffic light states from perception system.
/// @return A LaneConnectorTrafficLightState representation of the perception message
inline av::map::utils::LaneConnectorTrafficLightState createLaneConnectorTrafficLightStateFromMessage(
    const LaneConnector& laneConnector,
    av::msg::MultiIntersectionTLStatus multiIntersectionTlStatus)
{
    auto color{TrafficLightItemColor::unknown};
    auto shape{TrafficLightItemShape::unknown};
    auto state{TrafficLightFlashingState::unknown};
    auto fov{TrafficLightFov::notInFov};

    const auto edgeTlStatus = getEdgeTLStatus(laneConnector, multiIntersectionTlStatus);
    if (edgeTlStatus.has_value()) {
        switch (edgeTlStatus.value().color) {
            case av::msg::LIGHT_COLOR::UNKNOWN:
                color = TrafficLightItemColor::unknown;
                break;
            case av::msg::LIGHT_COLOR::GREEN:
                color = TrafficLightItemColor::green;
                break;
            case av::msg::LIGHT_COLOR::YELLOW:
                color = TrafficLightItemColor::yellow;
                break;
            case av::msg::LIGHT_COLOR::RED:
                color = TrafficLightItemColor::red;
                break;
        }

        switch (edgeTlStatus.value().shape) {
            case av::msg::LIGHT_SHAPE::UNKNOWN:
                shape = TrafficLightItemShape::unknown;
                break;
            case av::msg::LIGHT_SHAPE::CIRCLE:
                shape = TrafficLightItemShape::circle;
                break;
            case av::msg::LIGHT_SHAPE::LEFT_ARROW:
                shape = TrafficLightItemShape::left;
                break;
            case av::msg::LIGHT_SHAPE::RIGHT_ARROW:
                shape = TrafficLightItemShape::right;
                break;
            case av::msg::LIGHT_SHAPE::STRAIGHT_ARROW:
                shape = TrafficLightItemShape::up;
                break;
        }

        switch (edgeTlStatus.value().state) {
            case av::msg::LIGHT_STATE::UNKNOWN:
                state = TrafficLightFlashingState::unknown;
                break;
            case av::msg::LIGHT_STATE::STEADY:
                state = TrafficLightFlashingState::steady;
                break;
            case av::msg::LIGHT_STATE::FLASHING:
                state = TrafficLightFlashingState::flashing;
                break;
        }

        switch (edgeTlStatus.value().fov_state) {
            case av::msg::FOV_STATE::NOT_IN_FOV:
                fov = TrafficLightFov::notInFov;
                break;
            case av::msg::FOV_STATE::IN_FOV:
                fov = TrafficLightFov::inFov;
                break;
        }
    }
    return av::map::utils::LaneConnectorTrafficLightState{color, shape, state, fov};
}

} // namespace av::msg::utils