/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 */

#ifndef _IDLHELPER_MAP_HPP
#define _IDLHELPER_MAP_HPP

#include <map>
#include <type_traits>
#include <unordered_map>
#include <vector>

namespace idlhelper {

/**
 * @brief Convert Container map representation (e.g. std::vector<Pair>) to std::unordered_map
 *
 * Usage:
 *  class Pair {
 *	    int key;
 *		double value;
 *	};
 *  std::vector<Pair> serializableMap;
 *  std::unordered_map<int, double> map = idlhelper::getUnorderedMap(map);
 *
 * @param [in] mapIn - map stored as Container
 * @return map stored as std::unordered_map with key:value as members of Pair
 */
template<typename Container>
auto getUnorderedMap(Container mapIn)
{
    // Retrieve the map's key and value types
    using Key = decltype(std::declval<typename std::remove_reference_t<Container>::value_type>().key);
    using Value = decltype(std::declval<typename std::remove_reference_t<Container>::value_type>().value);
    std::unordered_map<Key, Value> mapOut{};
    mapOut.reserve(mapIn.size());

    // Iterate through keys and values, storing key:value pair in map
    for (auto&& key_val : mapIn) {
        mapOut.emplace(std::move(key_val.key), std::move(key_val.value));
    }

    return mapOut;
}

/**
 * @brief Convert Container map representation (e.g. std::vector<Pair>) to std::map
 *
 * Usage:
 *  class Pair {
 *	    int key;
 *		double value;
 *	};
 *  std::vector<Pair> serializableMap;
 *  std::map<int, double> map = idlhelper::getUnorderedMap(map);
 *
 * @param [in] mapIn - map stored as Container
 * @return map stored as std::map with key:value as members of Pair
 */
template<typename Container>
auto getMap(Container mapIn)
{
    // Retrieve the map's key and value types
    using Key = decltype(std::declval<typename std::remove_reference_t<Container>::value_type>().key);
    using Value = decltype(std::declval<typename std::remove_reference_t<Container>::value_type>().value);
    std::map<Key, Value> mapOut{};

    // Iterate through keys and values, storing key:value pair in map
    for (auto&& key_val : mapIn) {
        mapOut.emplace(std::move(key_val.key), std::move(key_val.value));
    }

    return mapOut;
}

/**
 * @brief Convert STL map-like container to vector representation std::vector<Pair>
 *
 * Usage:
 *  class Pair {
 *	    int key;
 *		double value;
 *	};
 *  std::unordered_map<int, double> map;
 *  std::vector<Pair> serializableMap = idlhelper::getSerializableMap<Pair>(map);
 *
 * @param [in] mapIn - an STL map-like container (e.g std::map, std::unordered_map)
 * @return map stored as std::vector<Pair>
 */
template<typename Pair, typename MapIn>
std::vector<Pair> getSerializableMap(MapIn mapIn)
{
    // Compile time assertion if type mismatch
    using MapInKey = typename std::remove_reference_t<MapIn>::key_type;
    using MapInValue = typename std::remove_reference_t<MapIn>::mapped_type;
    using PairKey = decltype(std::declval<Pair>().key);
    using PairValue = decltype(std::declval<Pair>().value);
    static_assert(std::is_same<std::decay<MapInKey>, std::decay<PairKey>>::value,
                  "Type mismatch between MapIn and Pair keys.");
    static_assert(std::is_same<std::decay<MapInValue>, std::decay<PairValue>>::value,
                  "Type mismatch between MapIn and Pair values.");

    std::vector<Pair> mapOut{};
    mapOut.reserve(mapIn.size());

    // Iterate through keys and values, storing key:value pair in MapOut
    for (auto&& key_val : mapIn) {
        mapOut.emplace_back(std::move(key_val.first), std::move(key_val.second));
    }

    return mapOut;
}

} // namespace idlhelper

#endif // _IDLHELPER_MAP_HPP
