#include <interfaces/test_msg/test_msg.h>

#include <IdlHelper/Map.hpp>
#include <gtest/gtest.h>

namespace idl_map_int_double_test {
class Pair
{
public:
    Pair() {}
    Pair(int inKey, double inValue)
        : key(std::move(inKey))
        , value(std::move(inValue))
    {}

    int key{0};
    double value{0.0};
};

class idl_map_avt
{
public:
    std::vector<Pair> map{};
};

TEST(IdlHelperTest, GetUnorderedMapIntDouble)
{
    std::unordered_map<int, double> nativeMapIn = {
        {1, 3.14},
        {8, 36.12},
        {13, 8.24},
        {3, -13.5},
    };

    // Create serializable map
    // Input: 	std::unordered_map<int, double>
    // Output:	std::vector<Pair> (where Pair has members {int key, double value})
    const auto serializableMap = idlhelper::getSerializableMap<Pair>(nativeMapIn);
    EXPECT_EQ(typeid(std::vector<Pair>), typeid(serializableMap));

    // Create native map
    // Input:	std::vector<Pair> (where Pair has members {int key, double value})
    // Output: 	std::unordered_map<int, double>
    const auto nativeMapOut = idlhelper::getUnorderedMap(serializableMap);
    EXPECT_EQ(typeid(std::unordered_map<int, double>), typeid(nativeMapOut));

    // Verify nativeMapIn matches nativeMapOut
    // We expect that if we convert to serializable type and back to native type, input == output
    EXPECT_EQ(nativeMapIn, nativeMapOut);
}

TEST(IdlHelperTest, GetMapIntDouble)
{
    std::map<int, double> nativeMapIn = {
        {1, 3.14},
        {8, 36.12},
        {13, 8.24},
        {3, -13.5},
    };

    // Create serializable map
    // Input: 	std::map<int, double>
    // Output:	std::vector<Pair> (where Pair has members {int key, double value})
    const auto serializableMap = idlhelper::getSerializableMap<Pair>(nativeMapIn);
    EXPECT_EQ(typeid(std::vector<Pair>), typeid(serializableMap));

    // Create native map
    // Input:	std::vector<Pair> (where Pair has members {int key, double value})
    // Output: 	std::map<int, double>
    const auto nativeMapOut = idlhelper::getMap(serializableMap);
    EXPECT_EQ(typeid(std::map<int, double>), typeid(nativeMapOut));

    // Verify nativeMapIn matches nativeMapOut
    // We expect that if we convert to serializable type and back to native type, input == output
    EXPECT_EQ(nativeMapIn, nativeMapOut);
}

TEST(IdlHelperTest, GetUnorderedMapIntDoubleEmpty)
{
    idl_map_avt idlMapMsg{};
    const auto map = idlhelper::getUnorderedMap(idlMapMsg.map);
    EXPECT_EQ(typeid(std::unordered_map<int, double>), typeid(map));
    EXPECT_EQ(map.empty(), true);
}
} // namespace idl_map_int_double_test

namespace idl_map_class_test {
class IDLClass
{
public:
    IDLClass() = default;
    IDLClass(bool inResult, int inNumber)
        : result(std::move(inResult))
        , number(std::move(inNumber))
    {}
    IDLClass(const IDLClass& other) = default;
    IDLClass& operator=(const IDLClass& other) = default;
    IDLClass(IDLClass&& other)
        : result(std::move(other.result))
        , number(std::move(other.number))
        , moved(true)
    {}
    IDLClass& operator=(IDLClass&& other) = default;
    inline bool operator==(IDLClass i) const { return (result == i.result && number == i.number); }

    bool result{true};
    bool number{1};
    bool moved{false};
};

class Pair
{
public:
    Pair() {}
    Pair(int inKey, IDLClass inValue)
        : key(std::move(inKey))
        , value(std::move(inValue))
    {}

    int key{0};
    IDLClass value{};
};

class idl_map_avt
{
public:
    std::vector<Pair> map{};
};

TEST(IdlHelperTest, GetUnorderedMapIntClass)
{
    // Construct unordered map with native class as value
    std::unordered_map<int, IDLClass> nativeMapIn = {
        {1, IDLClass(false, 0)},
        {8, IDLClass(true, 0)},
        {13, IDLClass(true, 1)},
        {3, IDLClass(false, 1)},
    };

    // Create serializable map
    // Input: 	std::unordered_map<int, IDLClass>
    // Output:	std::vector<Pair> (where Pair has members {int key, IDLClass value})
    const auto serializableMap = idlhelper::getSerializableMap<Pair>(nativeMapIn);
    EXPECT_EQ(typeid(std::vector<Pair>), typeid(serializableMap));

    // Create native map
    // Input:	std::vector<Pair> (where Pair has members {int key, IDLClass value})
    // Output: 	std::unordered_map<int, IDLClass>
    const auto nativeMapOut = idlhelper::getUnorderedMap(serializableMap);
    EXPECT_EQ(typeid(std::unordered_map<int, IDLClass>), typeid(nativeMapOut));

    // Verify nativeMapIn matches nativeMapOut
    // We expect that if we convert to serializable type and back to native type, input == output
    EXPECT_EQ(nativeMapIn, nativeMapOut);
}

TEST(IdlHelperTest, GetMapIntClass)
{
    // Construct map with native class as value
    std::map<int, IDLClass> nativeMapIn = {
        {1, IDLClass(false, 0)},
        {8, IDLClass(true, 0)},
        {13, IDLClass(true, 1)},
        {3, IDLClass(false, 1)},
    };

    // Create serializable map
    // Input: 	std::map<int, IDLClass>
    // Output:	std::vector<Pair> (where Pair has members {int key, IDLClass value})
    const auto serializableMap = idlhelper::getSerializableMap<Pair>(nativeMapIn);
    EXPECT_EQ(typeid(std::vector<Pair>), typeid(serializableMap));

    // Create native map
    // Input:	std::vector<Pair> (where Pair has members {int key, IDLClass value})
    // Output: 	std::map<int, IDLClass>
    const auto nativeMapOut = idlhelper::getMap(serializableMap);
    EXPECT_EQ(typeid(std::map<int, IDLClass>), typeid(nativeMapOut));

    // Verify nativeMapIn matches nativeMapOut
    // We expect that if we convert to serializable type and back to native type, input == output
    EXPECT_EQ(nativeMapIn, nativeMapOut);
}

TEST(IdlHelperTest, GetMapIntClassMove)
{
    // This check verifies that getSerializableMap and getMap support moves

    // Construct map with native class as value
    const int key = 1;
    const IDLClass value(false, 0);
    std::map<int, IDLClass> nativeMapIn{{key, value}};
    EXPECT_EQ(nativeMapIn.at(1).moved, false);

    // Create serializable map
    // Input: 	std::map<int, IDLClass>
    // Output:	std::vector<Pair> (where Pair has members {int key, IDLClass value})
    auto serializableMap = idlhelper::getSerializableMap<Pair>(std::move(nativeMapIn));
    EXPECT_EQ(typeid(std::vector<Pair>), typeid(serializableMap));
    EXPECT_EQ(serializableMap.at(0).value.moved, true);

    // Reset moved to false
    serializableMap[0].value.moved = false;
    EXPECT_EQ(serializableMap.at(0).value.moved, false);

    // Copy serializableMap
    auto serializableMapCopy = serializableMap;

    // Create native map
    // Input:	std::vector<Pair> (where Pair has members {int key, IDLClass value})
    // Output: 	std::map<int, IDLClass>
    const auto nativeMapOut = idlhelper::getMap(std::move(serializableMap));
    EXPECT_EQ(typeid(std::map<int, IDLClass>), typeid(nativeMapOut));
    EXPECT_EQ(nativeMapOut.at(1).moved, true);

    // Create native unordered_map
    // Input:	std::vector<Pair> (where Pair has members {int key, IDLClass value})
    // Output: 	std::unordered_map<int, IDLClass>
    const auto nativeUnorderedMapOut = idlhelper::getUnorderedMap(std::move(serializableMapCopy));
    EXPECT_EQ(typeid(std::unordered_map<int, IDLClass>), typeid(nativeUnorderedMapOut));
    EXPECT_EQ(nativeUnorderedMapOut.at(1).moved, true);
}

} // namespace idl_map_class_test

TEST(IdlHelperTest, GetUnorderedMapTestMessage)
{
    test::msg::test_msg msg;
    msg.map.emplace_back(1, 3.14);
    msg.map.emplace_back(3, -13.5);
    msg.map.emplace_back(8, 36.12);
    msg.map.emplace_back(13, 8.24);

    // Create native map
    // Input:	std::vector<Pair> (where Pair has members {int key, double value})
    // Output: 	std::unordered_map<int, double>
    const auto nativeMap = idlhelper::getUnorderedMap(msg.map);
    EXPECT_EQ(typeid(std::unordered_map<int, double>), typeid(nativeMap));

    // Create serializable map
    // Input: 	std::unordered_map<int, double>
    // Output:	std::vector<Pair> (where Pair has members {int key, double value})
    const auto serializableMap = idlhelper::getSerializableMap<test::msg::Pair>(nativeMap);
    EXPECT_EQ(typeid(std::vector<test::msg::Pair>), typeid(serializableMap));

    // We expect that if we convert to std::unordered_map and back, input == output
    // However, std::unordered_map does not guarantee order, so we must check each
    // Pair exists, but not that they are in the same order.
    EXPECT_EQ(serializableMap.size(), 4);
    EXPECT_EQ(std::count(serializableMap.begin(), serializableMap.end(), msg.map[0]), 1);
    EXPECT_EQ(std::count(serializableMap.begin(), serializableMap.end(), msg.map[1]), 1);
    EXPECT_EQ(std::count(serializableMap.begin(), serializableMap.end(), msg.map[2]), 1);
    EXPECT_EQ(std::count(serializableMap.begin(), serializableMap.end(), msg.map[3]), 1);
}

TEST(IdlHelperTest, GetMapTestMessage)
{
    test::msg::test_msg msg;
    msg.map.emplace_back(1, 3.14);
    msg.map.emplace_back(3, -13.5);
    msg.map.emplace_back(8, 36.12);
    msg.map.emplace_back(13, 8.24);

    // Create native map
    // Input:	std::vector<Pair> (where Pair has members {int key, double value})
    // Output: 	std::map<int, double>
    const auto nativeMap = idlhelper::getMap(msg.map);
    EXPECT_EQ(typeid(std::map<int, double>), typeid(nativeMap));

    // Create serializable map
    // Input: 	std::map<int, double>
    // Output:	std::vector<Pair> (where Pair has members {int key, double value})
    const auto serializableMap = idlhelper::getSerializableMap<test::msg::Pair>(nativeMap);
    EXPECT_EQ(typeid(std::vector<test::msg::Pair>), typeid(serializableMap));

    // We expect that if we convert to std::map and back, input == output
    // Our input vector was emplaced in order of key, which the std::map will maintain
    EXPECT_EQ(msg.map, serializableMap);
}
