/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

#include "SaveImage.h"
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <fcntl.h>
#include <getopt.h>
#include <jpeglib.h>
#include <linux/types.h>
#include <png.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <vector>
#undef eightbit

void png_save(char* filename, int image_width, int image_height, unsigned char* pic, int compression)
{
    FILE* fp = NULL;
    png_structp png_ptr;
    png_infop info_ptr;
    //    png_uint_32 k, height, width;
    std::vector<png_bytep> row_pointers(static_cast<u_int64_t>(image_height));

    /* open the file */
    if (!strcmp(filename, ""))
        fp = stdout;
    else if (filename != NULL)
        fp = fopen(filename, "w");

    if (fp == NULL) {
        printf("Error - Could not open file\n");
        return;
    }

    /* Create and initialize the png_struct with the desired error handler
     * functions.  If you want to use the default stderr and longjump method,
     * you can supply NULL for the last three parameters.  We also check that
     * the library version is compatible with the one used at compile time,
     * in case we are using dynamically linked libraries.  REQUIRED.
     */
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

    if (png_ptr == NULL) {
        printf("Error - could not create PNG\n");
        fclose(fp);
        return;
    }

    /* Allocate/initialize the image information data.  REQUIRED */
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
        fclose(fp);
        png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
        return;
    }

    /* Set error handling.  REQUIRED if you aren't supplying your own
     * error hadnling functions in the png_create_write_struct() call.
     */
    if (setjmp(png_jmpbuf(png_ptr))) {
        /* If we get here, we had a problem reading the file */
        fclose(fp);
        png_destroy_write_struct(&png_ptr, &info_ptr);
        return;
    }

    /* One of the following I/O initialization functions is REQUIRED */
    /* set up the output control if you are using standard C streams */
    png_init_io(png_ptr, fp);

    /* Set the image information here.  Width and height are up to 2^31,
     * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
     * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
     * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
     * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
     * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
     * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
     */

#ifdef eightbit
    png_set_IHDR(png_ptr,
                 info_ptr,
                 image_width,
                 image_height,
                 8,
                 PNG_COLOR_TYPE_GRAY,
                 0,
                 PNG_COMPRESSION_TYPE_BASE,
                 PNG_FILTER_TYPE_BASE);
#else
    png_set_IHDR(png_ptr,
                 info_ptr,
                 static_cast<u_int32_t>(image_width),
                 static_cast<u_int32_t>(image_height),
                 8,
                 PNG_COLOR_TYPE_RGB,
                 0,
                 PNG_COMPRESSION_TYPE_BASE,
                 PNG_FILTER_TYPE_BASE);
#endif

    png_set_compression_level(png_ptr, compression);

    /* set the palette if there is one.  REQUIRED for indexed-color images */
    // palette = (png_colorp)png_malloc(png_ptr, 256 * sizeof (png_color));
    /* ... set palette colors ... */
    // png_set_PLTE(png_ptr, info_ptr, palette, 256);

    /* optional significant bit chunk */
    /* if we are dealing with a grayscale image then */
    // sig_bit.gray = true_bit_depth;
    /* otherwise, if we are dealing with a color image then */
    // sig_bit.red = true_red_bit_depth;
    // sig_bit.green = true_green_bit_depth;
    // sig_bit.blue = true_blue_bit_depth;
    /* if the image has an alpha channel then */
    // sig_bit.alpha = true_alpha_bit_depth;
    // png_set_sBIT(png_ptr, info_ptr, sig_bit);

    /* Optional gamma chunk is strongly suggested if you have any guess
     * as to the correct gamma of the image.
     */
    // png_set_gAMA(png_ptr, info_ptr, gamma);

    /* Optionally write comments into the image */
    /*
      text_ptr[0].key = "Title";
      text_ptr[0].text = "Mona Lisa";
      text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
      text_ptr[1].key = "Author";
      text_ptr[1].text = "Leonardo DaVinci";
      text_ptr[1].compression = PNG_TEXT_COMPRESSION_NONE;
      text_ptr[2].key = "Description";
      text_ptr[2].text = "<long text>";
      text_ptr[2].compression = PNG_TEXT_COMPRESSION_zTXt;
      png_set_text(png_ptr, info_ptr, text_ptr, 3);
    */
    /* other optional chunks like cHRM, bKGD, tRNS, tIME, oFFs, pHYs, */
    /* note that if sRGB is present the cHRM chunk must be ignored
     * on read and must be written in accordance with the sRGB profile */

    /* Write the file header information.  REQUIRED */
    png_write_info(png_ptr, info_ptr);

    /* Once we write out the header, the compression type on the text
     * chunks gets changed to PNG_TEXT_COMPRESSION_NONE_WR or
     * PNG_TEXT_COMPRESSION_zTXt_WR, so it doesn't get written out again
     * at the end.
     */

    /* set up the transformations you want.  Note that these are
     * all optional.  Only call them if you want them.
     */

    /* invert monocrome pixels */
    // png_set_invert_mono(png_ptr);

    /* Shift the pixels up to a legal bit depth and fill in
     * as appropriate to correctly scale the image.
     */
    // png_set_shift(png_ptr, &sig_bit);

    /* pack pixels into bytes */
    // png_set_packing(png_ptr);

    /* swap location of alpha bytes from ARGB to RGBA */
    // png_set_swap_alpha(png_ptr);

    /* Get rid of filler (OR ALPHA) bytes, pack XRGB/RGBX/ARGB/RGBA into
     * RGB (4 channels -> 3 channels). The second parameter is not used.
     */
    // png_set_filler(png_ptr, 0, PNG_FILLER_BEFORE);

    /* flip BGR pixels to RGB */
    // png_set_bgr(png_ptr);

    /* swap bytes of 16-bit files to most significant byte first */
    // png_set_swap(png_ptr);

    /* swap bits of 1, 2, 4 bit packed pixel formats */
    // png_set_packswap(png_ptr);

    /* turn on interlace handling if you are not using png_write_image() */
    /*
      if (interlacing)
        number_passes = png_set_interlace_handling(png_ptr);
      else
        number_passes = 1;
    */
    /* The easiest way to write the image (you may have a different memory
     * layout, however, so choose what fits your needs best).  You need to
     * use the first method if you aren't handling interlacing yourself.
     */
    png_uint_32 k;

    //  png_uint_32 k, height, width;
    //  png_byte image[height][width];
    //  png_bytep row_pointers[height];

#ifdef eightbit
    {
        for (k = 0; k < static_cast<u_int32_t>(image_height); k++)
            row_pointers[k] = pic + k * image_width;
    }
#else
    {
        for (k = 0; k < static_cast<u_int32_t>(image_height); k++) {
            row_pointers[k] = pic + k * static_cast<u_int32_t>(image_width) * 3;
        }
    }
#endif

    /* One of the following output methods is REQUIRED */
    png_write_image(png_ptr, &row_pointers.front());

    /* You can write optional chunks like tEXt, zTXt, and tIME at the end
     * as well.
     */

    /* It is REQUIRED to call this to finish writing the rest of the file */
    png_write_end(png_ptr, info_ptr);

    /* if you allocated any text comments, free them here */

    /* clean up after the write, and free any memory allocated */
    png_destroy_write_struct(&png_ptr, &info_ptr);

    /* close the file */
    fclose(fp);

    /* that's it */
    //  printf("Image saved to PNG\n");
    return;
}

void ppm_save(char* filename, int image_width, int image_height, unsigned char* pic)
{
    FILE* outfile = NULL;
    unsigned char buff[3];
    int i;

    if (!strcmp(filename, ""))
        outfile = stdout;
    else if (filename != NULL)
        outfile = fopen(filename, "w");

    if (outfile == NULL) {
        printf("Could not open file\n");
        return;
    }

    /*  Raw  */
    if (1) {
        fprintf(outfile, "P6\n%d %d\n%d\n", image_width, image_height, 255);
        if (0) {
            for (i = 0; i < image_width * image_height; i++) {
                buff[0] = pic[i];
                buff[1] = pic[i];
                buff[2] = pic[i];
                fwrite(buff, 1, 3, outfile);
            }
        } else {
            for (i = 0; i < image_width * image_height * 3; i++) {
                fputc(pic[i], outfile);
            }
        }
    }

    /*  Ascii  */
    else if (0) {
        fprintf(outfile, "P3\n%d %d\n%d\n", image_width, image_height, 255);
        if (0) {
            for (i = 0; i < image_width * image_height; i++) {
                fprintf(outfile, "%03d %03d %03d\n", pic[i], pic[i], pic[i]);
            }
        } else {
            for (i = 0; i < image_width * image_height * 3; i = i + 3) {
                fprintf(outfile, "%03d %03d %03d\n", pic[i], pic[i + 1], pic[i + 2]);
            }
        }
    }

    fclose(outfile);

    printf("Image saved to PPM\n");
}

void jpeg_save(char* filename, int image_width, int image_height, unsigned char* pic, int smoothness, int quality)
{
    FILE* outfile = NULL;
    //  int ls=0, x, y;
    //  int i;
    //  int lines, pixels_per_line;
    //  unsigned char buff[3];
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    JSAMPROW row_pointer[1]; // pointer to JSAMPLE row(s)
    int row_stride;          // physical row width in image buffer

    if (!strcmp(filename, ""))
        outfile = stdout;
    else if (filename != NULL)
        outfile = fopen(filename, "w");

    if (outfile == NULL) {
        printf("Could not open file\n");
        return;
    }

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    jpeg_stdio_dest(&cinfo, outfile);
    cinfo.image_width = static_cast<u_int32_t>(image_width);
    cinfo.image_height = static_cast<u_int32_t>(image_height);

    if (0) {
        cinfo.input_components = 1;
        cinfo.in_color_space = JCS_GRAYSCALE;
    } else {
        cinfo.input_components = 3;
        cinfo.in_color_space = JCS_RGB;
    }

    cinfo.smoothing_factor = smoothness;
    if (smoothness) {
        cinfo.optimize_coding = TRUE;
    }

    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, quality, TRUE);
    jpeg_start_compress(&cinfo, TRUE);
    row_stride = cinfo.input_components * image_width;
    while (cinfo.next_scanline < cinfo.image_height) {
        row_pointer[0] = &pic[cinfo.next_scanline * static_cast<u_int32_t>(row_stride)];
        jpeg_write_scanlines(&cinfo, row_pointer, 1);
    }
    jpeg_finish_compress(&cinfo);
    jpeg_destroy_compress(&cinfo);

    /*
      struct jpeg_compress_struct cinfo;
      struct jpeg_error_mgr jerr;
      JSAMPROW row_pointer[1];     // pointer to JSAMPLE row(s)
      int row_stride;              // physical row width in image buffer

      cinfo.err = jpeg_std_error(&jerr);
      jpeg_create_compress(&cinfo);

      jpeg_stdio_dest(&cinfo, outf);
      cinfo.image_width = image_width;
      cinfo.image_height = image_height;

      if (0)
      {
        cinfo.input_components = 1;
        cinfo.in_color_space = JCS_GRAYSCALE;
      }
      else
      {
        cinfo.input_components = 3;
        cinfo.in_color_space = JCS_RGB;
      }

      cinfo.optimize_coding = TRUE;
      jpeg_set_defaults(&cinfo);
      jpeg_set_quality(&cinfo, 75, TRUE );
      jpeg_start_compress(&cinfo, TRUE);
      row_stride = cinfo.input_components*image_width;
      while (cinfo.next_scanline < cinfo.image_height) {
        row_pointer[0] = &camera->pic[cinfo.next_scanline * row_stride];
        jpeg_write_scanlines(&cinfo, row_pointer, 1);
      }
      jpeg_finish_compress(&cinfo);
      jpeg_destroy_compress(&cinfo);
    */

    fclose(outfile);

    printf("Image saved to JPEG\n");
    return;
}
