/**
 * \copyright
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief Defines the order of Camera Frames and types
 *
 */

#ifndef PERC_UTILS_CAMERA_FRAMES_HPP_
#define PERC_UTILS_CAMERA_FRAMES_HPP_

#include <cstdint>
#include <string>
#include <unordered_map>

namespace aptv {
namespace perc {
namespace utils {

enum class CameraType : uint8_t
{
    None = 0,
    BASLERACA160060GC = 1,
};

enum class CameraFrame : uint8_t
{
    Invalid = 0,
    CAM_F0 = 1,
    CAM_R0 = 2,
    CAM_R1 = 3,
    CAM_R2 = 4,
    CAM_B0 = 5,
    CAM_L2 = 6,
    CAM_L1 = 7,
    CAM_L0 = 8,
};

/**
 * @brief Convert from camera type string to CameraType enum
 *
 * @param camera_type - Camera type string as read from the config
 * @return CameraType - Camera type enum
 */
// GC: does not make much sense, the camera model should be read from calibration
inline CameraType getCameraType(const std::string& camera_type)
{

    if (camera_type == "CAM_F0" || camera_type == "CAM_R0" || camera_type == "CAM_R1" || camera_type == "CAM_R2" ||
        camera_type == "CAM_B0" || camera_type == "CAM_L2" || camera_type == "CAM_L1" || camera_type == "CAM_L0") {
        return CameraType::BASLERACA160060GC;
    } else {
        return CameraType::None;
    }
}

/**
 * @brief Get the Camera Frame Id From Enum object
 *
 * @param frame - Camera frame enum
 * @return std::string - Camera Frame string
 */
inline std::string getCameraFrameIdFromEnum(const CameraFrame frame)
{
    switch (frame) {
        case CameraFrame::CAM_F0:
            return "CAM_F0";
        case CameraFrame::CAM_R0:
            return "CAM_R0";
        case CameraFrame::CAM_R1:
            return "CAM_R1";
        case CameraFrame::CAM_R2:
            return "CAM_R2";
        case CameraFrame::CAM_B0:
            return "CAM_B0";
        case CameraFrame::CAM_L2:
            return "CAM_L2";
        case CameraFrame::CAM_L1:
            return "CAM_L1";
        case CameraFrame::CAM_L0:
            return "CAM_L0";
        case CameraFrame::Invalid:
            return "Invalid Frame";
        default:
            return "Invalid Frame";
    }
}

/**
 * @brief Get the Camera Enum object from Camera Frame Id string
 *
 * @param frame_id - Camera Frame string as read from config
 * @return CameraFrame - Camera frame enum
 */
inline CameraFrame getCameraEnumFromFrameId(const std::string& frame_id)
{
    static const std::unordered_map<std::string, CameraFrame> frames = {{"CAM_F0", CameraFrame::CAM_F0},
                                                                        {"CAM_R0", CameraFrame::CAM_R0},
                                                                        {"CAM_R1", CameraFrame::CAM_R1},
                                                                        {"CAM_R2", CameraFrame::CAM_R2},
                                                                        {"CAM_B0", CameraFrame::CAM_B0},
                                                                        {"CAM_L0", CameraFrame::CAM_L0},
                                                                        {"CAM_L1", CameraFrame::CAM_L1},
                                                                        {"CAM_L2", CameraFrame::CAM_L2}};
    return frames.at(frame_id);
}

} // namespace utils
} // namespace perc
} // namespace aptv

#endif
