#include <gtest/gtest.h>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/opencv.hpp>

#include "../SaveImage.h"
#include <ImageUtility/CameraFrames.hpp>
#include <ImageUtility/CommonImageTypes.h>
#include <ImageUtility/ImageEncodeDecode.h>

#include <experimental/filesystem>
#include <iostream>

#include "lib/avlog/avlog.hpp"

namespace fs = std::experimental::filesystem;

TEST(ImageUtilityTest, CameraFrames)
{
    // getCameraType
    std::vector<std::string> validCameraTypes{
        "CAM_F0", "CAM_R0", "CAM_R1", "CAM_R2", "CAM_B0", "CAM_L2", "CAM_L1", "CAM_L0"};
    for (const auto& x : validCameraTypes) {
        EXPECT_EQ(aptv::perc::utils::getCameraType(x), aptv::perc::utils::CameraType::BASLERACA160060GC);
    }
    EXPECT_EQ(aptv::perc::utils::getCameraType("CAM_DOES_NOT_EXIST"), aptv::perc::utils::CameraType::None);

    // getCameraFrameIdFromEnum
    std::vector<std::string> getCameraFrameIdFromEnumOut{"Invalid Frame",
                                                         "CAM_F0",
                                                         "CAM_R0",
                                                         "CAM_R1",
                                                         "CAM_R2",
                                                         "CAM_B0",
                                                         "CAM_L2",
                                                         "CAM_L1",
                                                         "CAM_L0",
                                                         "Invalid Frame"};
    uint8_t i = 0;
    for (const auto& x : getCameraFrameIdFromEnumOut) {
        auto testCameraFrame = static_cast<aptv::perc::utils::CameraFrame>(i);
        EXPECT_EQ(aptv::perc::utils::getCameraFrameIdFromEnum(testCameraFrame), x);
        i++;
    }

    // getCameraEnumFromFrameId
    uint8_t j = 1;
    for (const auto& x : validCameraTypes) {
        EXPECT_EQ(aptv::perc::utils::getCameraEnumFromFrameId(x), static_cast<aptv::perc::utils::CameraFrame>(j));
        j++;
    }
    try {
        aptv::perc::utils::getCameraEnumFromFrameId("CAM_DOES_NOT_EXIST");
    } catch (const std::out_of_range& e) {
        AVLOG_INFO("{}", e.what());
    }
}

TEST(ImageUtilityTest, CommonImageTypes)
{
    // STATUS_TYPE
    std::vector<commonImageTypes::STATUS_TYPE> validStatusTypes{commonImageTypes::STATUS_TYPE::UNKNOWN_STATUS,
                                                                commonImageTypes::STATUS_TYPE::FIXED_SCHED,
                                                                commonImageTypes::STATUS_TYPE::FREE_RUN,
                                                                commonImageTypes::STATUS_TYPE::RESUMING_FIXED_SCHED,
                                                                commonImageTypes::STATUS_TYPE::CONSECUTIVE_DROP,
                                                                commonImageTypes::STATUS_TYPE::EMPTY_STT_TABLE};
    uint8_t i = 0;
    for (const auto& x : validStatusTypes) {
        auto testStatusType = static_cast<commonImageTypes::STATUS_TYPE>(i);
        EXPECT_EQ(testStatusType, x);
        i++;
    }
}

TEST(ImageUtilityTest, ImageEncodeDecode)
{
    // compressMatBgrToJpeg
    cv::Mat testCompressCVMat = cv::Mat::zeros(3, 3, CV_8UC3);
    std::vector<uint8_t> testData;
    EXPECT_TRUE(image_encoder::compressMatBgrToJpeg(testCompressCVMat, testData, 1));
    // Empty CV Mat
    cv::Mat testCompressCVMat2 = cv::Mat::zeros(0, 0, CV_8UC3);
    EXPECT_FALSE(image_encoder::compressMatBgrToJpeg(testCompressCVMat2, testData, 1));

    // decompressJpegToMat
    cv::Mat testDecompressCVMat;
    EXPECT_TRUE(image_decoder::decompressJpegToMat(testData, testDecompressCVMat, false));
    for (int row = 0; row < 3; row++) {
        for (int col = 0; col < 3; col++) {
            EXPECT_EQ(testCompressCVMat.at<cv::Vec3b>(row, col), testDecompressCVMat.at<cv::Vec3b>(row, col));
        }
    }
    // isRGB = true
    EXPECT_TRUE(image_decoder::decompressJpegToMat(testData, testDecompressCVMat, true));
    // Empty data
    std::vector<uint8_t> emptyData;
    EXPECT_FALSE(image_decoder::decompressJpegToMat(emptyData, testDecompressCVMat, false));
}

TEST(ImageUtilityTest, SaveImageEnums)
{
    // Save_Types
    std::vector<Save_Types> validSaveTypes{Save_Types::PNG, Save_Types::PPM, Save_Types::JPEG, Save_Types::GIF};
    uint8_t i = 0;
    for (const auto& x : validSaveTypes) {
        auto testSaveTypes = static_cast<Save_Types>(i);
        EXPECT_EQ(testSaveTypes, x);
        i++;
    }

    // Ppm_Types
    std::vector<Ppm_Types> validPpmTypes{Ppm_Types::RAW, Ppm_Types::ASCII};
    uint8_t j = 0;
    for (const auto& x : validPpmTypes) {
        auto testPpmTypes = static_cast<Ppm_Types>(j);
        EXPECT_EQ(testPpmTypes, x);
        j++;
    }
}

TEST(ImageUtilityTest, SaveImage)
{
    int CVMAT_ROW = 3;
    int CVMAT_COL = 3;

    cv::Mat testCVMat = cv::Mat::eye(CVMAT_ROW, CVMAT_COL, CV_8UC3);
    unsigned char* testCVMatData = testCVMat.data;

    // png_save
    std::string pngSaveFilenameString = std::string{"test_png_save.png"};
    char* pngSaveFilename = &pngSaveFilenameString[0];
    png_save(pngSaveFilename, CVMAT_ROW, CVMAT_COL, testCVMatData, 1);
    EXPECT_TRUE(fs::exists(pngSaveFilenameString));

    cv::Mat loadPngSaveCVMat = cv::imread(pngSaveFilenameString);
    cv::Mat loadPngSaveCVMatRGB;
    cv::cvtColor(loadPngSaveCVMat, loadPngSaveCVMatRGB, cv::COLOR_BGR2RGB);
    for (int row = 0; row < CVMAT_ROW; row++) {
        for (int col = 0; col < CVMAT_COL; col++) {
            EXPECT_EQ(loadPngSaveCVMatRGB.at<cv::Vec3b>(row, col), testCVMat.at<cv::Vec3b>(row, col));
        }
    }

    // ppm_save
    std::string ppmSaveFilenameString = std::string{"test_ppm_save.ppm"};
    char* ppmSaveFilename = &ppmSaveFilenameString[0];
    ppm_save(ppmSaveFilename, CVMAT_ROW, CVMAT_COL, testCVMatData);
    EXPECT_TRUE(fs::exists(ppmSaveFilenameString));

    // jpeg_save
    std::string jpegSaveFilenameString = std::string{"test_jpeg_save.jpeg"};
    char* jpegSaveFilename = &jpegSaveFilenameString[0];
    jpeg_save(jpegSaveFilename, CVMAT_ROW, CVMAT_COL, testCVMatData, 1, 1);
    EXPECT_TRUE(fs::exists(jpegSaveFilenameString));
    // smoothness = 0
    jpeg_save(jpegSaveFilename, CVMAT_ROW, CVMAT_COL, testCVMatData, 0, 1);

    // NOTE: stdout stops printing after calling the following functions
    // Empty filename
    // std::string emptyString = std::string{""};
    // char* emptyFilename = &emptyString[0];
    // jpeg_save(emptyFilename, CVMAT_ROW, CVMAT_COL, testCVMatData, 1, 1);
    // ppm_save(emptyFilename, CVMAT_ROW, CVMAT_COL, testCVMatData);
    // png_save(emptyFilename, CVMAT_ROW, CVMAT_COL, testCVMatData, 1);
}
