/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

#include <LLUTMConvLibrary/utm.h>
#include <cstdio>
#include <proj_api.h>
#include <sstream>

static projPJ pj_utm, pj_latlong;
static int pj_utm_zone = -1;

bool need_init(int zone)
{
    if (pj_utm_zone == -1 || pj_utm_zone != zone)
        return true;

    return false;
}

void init_proj(int zone)
{
    std::stringstream ss;
    ss << zone;
    std::string utm_init_cmd = "+proj=utm +zone=" + ss.str() + " +ellps=WGS84";
    if (!(pj_utm = pj_init_plus(utm_init_cmd.c_str())) && zone > 0 && zone <= 60) {
        printf("could not init UTM\n");
        exit(1);
    }
    pj_utm_zone = zone;

    if (!(pj_latlong = pj_init_plus("+proj=latlong +ellps=WGS84"))) {
        printf("could not init lat long\n");
        exit(1);
    }
}

void LatLonToUtmWGS84(double lat, double lon, double& northing, double& easting, int zone)
{
    if (need_init(zone)) {
        init_proj(zone);
    }

    double x = lat * DEG_TO_RAD;
    double y = lon * DEG_TO_RAD;

    pj_transform(pj_latlong, pj_utm, 1, 1, &y, &x, NULL);

    easting = y;
    northing = x;
}

void UtmToLatLonWGS84(double& lat, double& lon, double northing, double easting, int zone)
{
    if (need_init(zone)) {
        init_proj(zone);
    }

    double x = easting;
    double y = northing;

    pj_transform(pj_utm, pj_latlong, 1, 1, &x, &y, NULL);

    lat = y / DEG_TO_RAD;
    lon = x / DEG_TO_RAD;
}

// get UTM zone number from given lattitude and longitude info
int getUTMZone(double lat, double lon)
{
    if (lat >= 56.0 && lat < 64.0 && lon >= 3.0 && lon < 12.0) {
        return 32;
    }

    if (lat > 72.0 && lat < 84.0) {
        if (lon > 0.0 && lon < 9.0)
            return 31;
        else if (lon >= 9.0 && lon < 21.0)
            return 33;
        else if (lon >= 21.0 && lon < 33.0)
            return 35;
        else if (lon >= 33.0 && lon < 42.0)
            return 37;
    }

    return static_cast<int>(floor((lon + 180) / 6) + 1);
}

/**
 * @brief Returns the UTM longitudinal designation from an EPSG code in the form "EPSG:32xxx".
 * Returns 0, an invalid zone, if the code is invalid.
 */
int getUTMZone(const std::string& epsgString)
{
    if (!(epsgString.find("epsg") == 0 || epsgString.find("EPSG") == 0)) {
        printf("\"%s\" may not be a EPSG code\n", epsgString.c_str());
        // If receiving mixed-case "Epsg" or other unexpected prefixes, this message will appear but the CRS code
        // will still be processed.
    }
    auto delimIndex = epsgString.find(":");
    if (delimIndex == std::string::npos) {
        return 0;
    }
    auto epsgNumber = epsgString.substr(delimIndex + 1);
    if (epsgNumber.size() != 5) {
        return 0;
    }
    // EPSG codes for UTM coordinate systems begin with 32
    if (epsgNumber.find("32") != 0) {
        return 0;
    }
    // EPSG codes for UTM coordinate systems have a third digit of 6 or 7 depending on hemisphere
    if (!(epsgNumber[2] == '6' or epsgNumber[2] == '7')) {
        return 0;
    }
    auto utmZone = std::stoi(epsgNumber.substr(3));
    // UTM zones range from 1 to 60 inclusively, EPSG codes outside that range are not UTM
    if (utmZone >= 1 && utmZone <= 60) {
        return utmZone;
    } else {
        return 0;
    }
}
