/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file LatencyLogger.cc
/// @brief Implementation file for LatencyLogger.h

#include "platform/middleware/libs/LatencyLogger/include/LatencyLogger/LatencyLogger.h"
#include <atomic>

namespace av {
namespace {

/// @brief Returns the loggingName found in the config.
///
/// @param config The config section that the logger is being
/// constructed with.
///
/// @return The loggingName.
std::string getLoggingName(ConfigSection& config)
{
    std::string res;
    config.get("loggingName", res, "InvalidLoggingName");
    return res;
}

/// @brief Returns the enabled flag found in the config.
///
/// @param config The config section that the logger is being
/// constructed with.
///
/// @return True iff logging is enabled.
bool getLoggingEnabled(ConfigSection& config)
{
    bool res;
    config.get("enabled", res, false);
    return res;
}

// The two most significant bits identify the type of ID.
constexpr uint64_t kEventTypeShift = 62;
constexpr uint64_t kEventTypeMask = (~UINT64_C(0) << kEventTypeShift);
constexpr uint64_t kEventValueMask = ~kEventTypeMask;

constexpr uint64_t kKeySlotShift = 56;
constexpr uint64_t kKeySlotMask = (~UINT64_C(0) << kKeySlotShift) ^ kEventTypeMask;
constexpr uint64_t kKeyValueMask = ~(kEventTypeMask | kKeySlotMask);

enum class EventType : uint64_t
{
    kRegularEvent = (UINT64_C(0) << kEventTypeShift),
    kKeyEvent = (UINT64_C(1) << kEventTypeShift),
    kCANEvent = (UINT64_C(2) << kEventTypeShift),
};

std::atomic<uint64_t> eventIDCounter{1};

} // namespace

bool LatencyLogger::isInitialized_ = false;
std::mutex LatencyLogger::initializeMutex_{};
std::unique_ptr<latency::DataLoggerWriter> LatencyLogger::dataLoggerWriter_{};

LatencyLogger::LatencyLogger(bool enableLogging, std::string loggingName, Writer writer)
    : enabled_(enableLogging)
    , loggingName_(std::move(loggingName))
    , writer_(std::move(writer))
{}

void LatencyLogger::initialize(ConfigSection& config, const Writer& writer)
{
    std::lock_guard<std::mutex> lock(initializeMutex_);
    if (!isInitialized_) {
        isInitialized_ = true;
        auto enabled = getLoggingEnabled(config);
        auto loggingName = getLoggingName(config);
        getLoggerImpl(enabled, loggingName, writer);
    }
}

void LatencyLogger::initialize(ConfigSection& config)
{
    std::lock_guard<std::mutex> lock(initializeMutex_);
    if (!isInitialized_) {
        isInitialized_ = true;
        auto enabled = getLoggingEnabled(config);
        auto loggingName = getLoggingName(config);
        dataLoggerWriter_ = std::make_unique<latency::DataLoggerWriter>("LatencyLogger/" + loggingName + ".msgpack");
        getLoggerImpl(enabled, loggingName, *dataLoggerWriter_);
    }
}

void LatencyLogger::initialize(bool enableLogging, const std::string& loggingName, const LatencyLogger::Writer& writer)
{
    std::lock_guard<std::mutex> lock(initializeMutex_);
    if (!isInitialized_) {
        isInitialized_ = true;
        getLoggerImpl(enableLogging, loggingName, writer);
    }
}

void LatencyLogger::initialize(bool enableLogging, const std::string& loggingName)
{
    std::lock_guard<std::mutex> lock(initializeMutex_);
    if (!isInitialized_) {
        isInitialized_ = true;
        dataLoggerWriter_ = std::make_unique<latency::DataLoggerWriter>("LatencyLogger/" + loggingName + ".msgpack");
        getLoggerImpl(enableLogging, loggingName, *dataLoggerWriter_);
    }
}

void LatencyLogger::shutdown()
{
    std::lock_guard<std::mutex> lock(initializeMutex_);
    isInitialized_ = false;
}

LatencyLogger& LatencyLogger::getLogger()
{
    std::lock_guard<std::mutex> lock(initializeMutex_);
    if (!isInitialized_) {
        return getDummyLogger();
    } else {
        return getLoggerImpl();
    }
}

latency::EventID LatencyLogger::nextEventID()
{
    return (eventIDCounter.fetch_add(1, std::memory_order::memory_order_relaxed)) & kEventValueMask;
}

latency::EventID LatencyLogger::canEventID(uint32_t canID)
{
    return static_cast<uint64_t>(EventType::kCANEvent) | static_cast<uint64_t>(canID);
}

latency::EventID LatencyLogger::keyEventID(uint8_t slot, uint64_t key)
{
    return (static_cast<uint64_t>(EventType::kKeyEvent) |
            (kKeySlotMask & (static_cast<uint64_t>(slot) << kKeySlotShift)) | (kKeyValueMask & key));
}

latency::EventID LatencyLogger::ForgedEventInternal(const std::string& checkpointName,
                                                    const ottotime::otimePoint& eventTimestamp,
                                                    const std::optional<latency::EventID>& eventIDOpt,
                                                    std::vector<latency::EventID>&& predecessors)
{
    // Either use the given EventID (if present) or generate a new
    // one. We could write eventIDOpt.value_or(nextEventID()),
    // but that would mean computing the new ID even if it isn't
    // needed.
    const latency::EventID eventID = eventIDOpt.has_value() ? eventIDOpt.value() : nextEventID();

    {
        std::lock_guard<std::mutex> lock(writerMutex_);
        // Call the Writer to do the logging. As of this writing, this
        // is the only place writer_ is used.
        writer_(loggingName_, checkpointName, eventTimestamp, eventID, std::move(predecessors));

        latestEventID_[checkpointName] = eventID;
    }

    return eventID;
}

latency::EventID LatencyLogger::SendEventInternal(const std::string& channelName,
                                                  const ottotime::otimePoint& sendTime,
                                                  const std::optional<latency::EventID>& eventIDOpt,
                                                  std::vector<latency::EventID>&& predecessors)
{
    return ForgedEventInternal("send:" + channelName, sendTime, eventIDOpt, std::move(predecessors));
}

latency::EventID LatencyLogger::ReceiveEventInternal(const std::string& channelName,
                                                     const ottotime::otimePoint& sendTime,
                                                     const std::optional<ottotime::otimePoint>& recvTime)
{
    // Predecessors are usually event IDs, but for receive events they are the
    // sent time of the received message (in microseconds since the epoc).
    // Not using value_or such that we lazily create default value.
    const auto& eventTs = recvTime.has_value() ? recvTime.value() : SystemClock::getSystemTime();
    return ForgedEventInternal(
        "recv:" + channelName, eventTs, std::nullopt, {static_cast<uint64_t>(sendTime.toMicrosecondsSinceEpoch())});
}

latency::EventID LatencyLogger::getLatestEventID(const std::string& checkpointName)
{
    std::lock_guard<std::mutex> lock(writerMutex_);
    auto foundIter = latestEventID_.find(checkpointName);
    return (foundIter != latestEventID_.end()) ? foundIter->second : latency::EventID{0};
}

} // namespace av
