## LatencyLogger

# Introduction

The LatencyLogger library was written to help measure latencies across the AV stack. We can use it to log events (and corresponding timestamps) whenever a program reaches certain checkpoints in the code during execution, and postprocess the logs to measure latencies between events. For example, if we have the following code:

    void func()
    {
        foo();
        bar();
    }

One could consider inserting the following checkpoints:

    void func()
    {
        // log event "startFoo"
        foo();
        // log event "endFoo". Its predecessor is the event with name "startFoo"
        bar();
        // log event "endBar". Its predecessor is the event with name "endFoo"
    }

The checkpoints in the above code have names "startFoo", "endFoo" and "endBar", and we log events every time they are reached. Each event is given a unique id (of type `av::latency::EventID`, which is an alias for `uint64_t`), and we also set up associations between events: in this case, "startFoo" is the predecessor to "endFoo", and "endFoo" is the predecessor to "endBar".

One could imagine that func() is called multiple times during execution, and we write logs that contain lists of all events for each checkpoint along with their IDs and timestamps, as well as their predecessors. At the end of each run, we can postprocess the logged events and generate a directed acyclic graph of events. The latency between any two events can be computed by traversing a path through this graph and computing the difference between their corresponding timestamps. We can also compute latency statistics for any valid path of checkpoints by traversing the graph through every set of events corresponding to those checkpoints

The rest of this document is only concerned with the LatencyLogger API and how we can use it in our code to log latency data. Details on postprocessing this data will be linked to in the future.

## Setup

Before we start logging latency data, we must initialize the LatencyLogger. Usually, this is done within the `initialize()` member function of a task, and can be accomplished the following way:

    LatencyLogger::initialize(true, "TaskName");

The first argument specifies if we want to enable logging for this task (if false, no latency log data will be written). The second argument is some identifier string for the log, usually the name of the task. While LatencyLogger can be initialized as shown above, it is most typically initialized the following way:

    ConfigSection cs(getTaskConfig());
    ConfigSection latencyLoggerConfig;
    cs.get("LatencyLogger", latencyLoggerConfig);
    av::LatencyLogger::initialize(latencyLoggerConfig);

For this to work, we must add new fields to the config yaml file associated with the task, under the `parameters` section:

    parameters:
        LatencyLogger:
            enabled: true
            loggingName: LidarSemanticNetworkTask

The parameters under `LatencyLogger` are the same as those explained earlier in this section. Subsequently, if needed, one can always disable logging by editing the config file and not having to recompile the code.

## Cleanup

This step is important because ignoring it could lead to undefined behavior (caused by the destruction of the LatencyLogger singleton object before other objects that make use of it. We hope to address this issue more robustly in the future).

In the `cleanup()` member function of a given task, we must include the following:

    av::LatencyLogger::shutdown();

We are now ready to start using the LatencyLogger to log events.

## Regular Events

Let's dive in with a simple example based on the code mentioned in the Introduction:

    void func()
    {
        auto& logger = LatencyLogger::getLogger();
        auto startFooEventID = logger.RegularEvent("startFoo");
        foo();
        auto endFooEventID = logger.RegularEvent("endFoo", {startFooEventID});
        bar();
        logger.RegularEvent("endBar", {endFooEventID});
    }

Here, `logger` is an instance of the LatencyLogger singleton object. We use it to make all the LatencyLogger calls.

`RegularEvent` is the most common way to log a latency event. Its first parameter is the name of the checkpoint, and it has an optional second parameter, which is a vector of all predecessors. In the above example, the second and third `RegularEvent` calls specify only one preecessor, but we may have situations where an event has more than one predecessor. `RegularEvent` returns an `EventID`, which can be supplied as a predecessor to subsequent calls. In the above example, we don't store the return value for the last call because it's not going to be used anywhere else in the function.

## How do I specify a predecessor `EventID` from a different function?

Consider the following scenario:

    void func()
    {
        foo();
        bar();
    }

    void foo()
    {
        auto& logger = LatencyLogger::getLogger();
        logger.RegularEvent("startFoo");
        doSomething();
    }

    void bar()
    {
        auto& logger = LatencyLogger::getLogger();
        doSomethingElse();
        logger.RegularEvent("endBar", {/*what goes in here?*/});
    }

We would like to pass an event ID from checkpoint "startFoo" as the predecessor to an event at checkpoint "endBar". This can be done the following way:

    void bar()
    {
        auto& logger = LatencyLogger::getLogger();
        auto startFooEventID = logger.getLatestEventID("startFoo");
        doSomethingElse();
        logger.RegularEvent("endBar", {startFooEventID});
    }

`getLatestEventID` takes any checkpoint name as a parameter, and returns the last event ID that was generated at that checkpoint. Hence, we can retrieve predecessor `EventID`s from events logged in other functions.

## Inter-process communication

We have already instrumented the inter-process communications code with LatencyLogger calls, and every Send and Receive event in a given task is logged, as long as LatencyLogger has been initialized for that task.

However, the user must take care to connect LatencyLogger events in a task to these Send and Receive events, so that latency graph can be generated across tasks. For example, in a callback function that's triggered when a message is received, we would want to specify the Receieve event as the predecessor to the first latency event that's logged. And when we publish messages, we want to specify the last latency event as the predecessor to the Send event. The latter was enabled by expanding all publish API calls to accept a predecessor EventID as an optional argument.

Every Receive event has a checkpoint name of the form "recv:<ChannelName>". So in a callback function, we can specify the predecessor in the following way:

    void callback()
    {
        auto& logger = LatencyLogger::getLogger();
        auto recvEventID = logger.getLatestEventID("recv:MergedPointCloud"); // say, we've just received a message on
                                                                             // the MergedPointCloud channel
        auto startFooEventID = logger.RegularEvent("startFoo", {recvEventID});
        foo();
        logger.RegularEvent("endFoo", {startFooEventID});
        bar();
    }

    void bar()
    {
        auto& logger = LatencyLogger::getLogger();
        auto endFooEventID = logger.getLatestEventID("endFoo");
        // populate payload
        writeInterface->publish(payload, endFooEventID); // endFooEventID will be the predecessor to the Send event,
                                                         // with checkpoint name of the form "send:PublishChannel"
    }

## `EventID` maps for each checkpoint

Some tasks such as `LidarPointCloudHandlerTask` and `lidarSemanticNetworkTask` collate multiple messages before processing them, which means that there may be instances when `getLatestEventID` is insufficient to retrieve the correct event IDs. For such scenarios, we have provided the capability to map `uint64_t` keys to `EventID`s. For example, if we receive multiple MergedPointCloud messages, we can store the Receive event's ID in a map using that message's timestamp as the key. We have below a simplified piece of code that illustrates how this may be done:

    void callback()
    {
        // store received MergedPointCloud in `msg`
        auto& logger = LatencyLogger::getLogger();
        auto recvEventID = logger.getLatestEventID("recv:MergedPointCloud");
        // create an EventID map for this checkpoint
        logger.storeEventIDInMap("recv:MergedPointCloud", msg.timestamp, recvEventID);
        foo(msg);
    }

    void foo(MergedPointCloud& msg)
    {
        auto& logger = LatencyLogger::getLogger();
        auto recvEventID = logger.getEventIDFromKey("recv:MergedPointCloud", msg.timestamp);
        logger.clearKey("recv:MergedPointCloud", msg.timestamp)
        doSomething();
        logger.RegularEvent("endFoo", {recvEventID});
    }

`storeEventIDInMap`registers a key-value pair for a specified checkpoint. `getEventIDFromKey` helps retrive an EventID, given the key. And `clearKey` clears a key-value pair that's no longer needed, so our maps don't grow indefinitely.

## Logs

The latency data is written to logs in a binary JSON format that uses [MessagePack](https://msgpack.org/index.html). They are all written to a folder names `LatencyLogger` that resides in the log folder, and have the file extension `.msgpack`. Their names are the same as the `loggingName` specified when initializing LatencyLogger.
