/**
 * \copyright
 * Copyright (c) 2020 Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 *
 * All information contained herein is, and remains the property of
 * Motional and its suppliers, if any.  The intellectual and technical
 * concepts contained herein are proprietary to Motional and its
 * suppliers and may be covered by U.S. and Foreign Patents, patents
 * in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Motional.
 */

/*
 * Design doc:
 *
 * https://confluence.ci.motional.com/confluence/display/SI/Latency+Logging
 */

#pragma once

#include "platform/middleware/libs/LatencyLogger/include/LatencyLogger/DataLoggerWriter.hpp"

#include "platform/middleware/libs/Clock/include/Clock/SystemClock.h"
#include "platform/middleware/libs/TimeStamp/include/TimeStamp/otime.h"
#include "platform/middleware/libs/config/include/config/ConfigSection.h"

#include <mutex>
#include <optional>
#include <vector>

namespace av {

/// @brief A class for logging events, so that the latency between
/// events can be computed.
class LatencyLogger
{
public:
    ~LatencyLogger() = default;
    // delete copy and move constructors, and assignment operators for Singleton
    LatencyLogger(const LatencyLogger&) = delete;
    LatencyLogger& operator=(const LatencyLogger&) = delete;
    LatencyLogger(LatencyLogger&&) = delete;
    LatencyLogger& operator=(LatencyLogger&&) = delete;

    /// @brief The most common function for logging events.  Logs an
    /// event, timestampped with the current time.
    ///
    /// @param checkpointName The name of the checkpoint that the
    /// event represents.
    ///
    /// @param predecessors Zero or more predecessor specifiers.
    ///
    /// @return A generated ID for the event.
    inline latency::EventID RegularEvent(const std::string& checkpointName,
                                         std::vector<latency::EventID>&& predecessors = {});

    /// @brief Logs an event, which uses a special event ID calculated from a
    /// slot and key.
    ///
    /// @param checkpointName The name of the checkpoint that the
    /// event represents.
    ///
    /// @param slot An a priori constant used to deconflict different types of
    /// keys. For a given process and latency log, care must be taken to use
    /// unique slots for different types of keys, or else ID collisions could
    /// occur. Valid range is 0 to 63, larger numbers will be truncated without
    /// warning.
    ///
    /// @param key The key associated with this event. Valid range is 0 to
    /// 1152921504606846975, larger numbers will be truncated without warning.
    ///
    /// @param predecessors Zero or more predecessor specifiers.
    ///
    /// @return A calculated ID for the event.
    inline latency::EventID KeyEvent(const std::string& checkpointName,
                                     uint8_t slot,
                                     uint64_t key,
                                     std::vector<latency::EventID>&& predecessors = {});

    /// @brief Record an event that occurred in a part of the codebase
    /// we don't control.
    ///
    /// The primary examples are acquisition events of sensors or
    /// actuation times of actuators. Since we do not control the
    /// firmware inside said devices, some other portion of our
    /// codebase will have to forge those events that will not be
    /// timestamped under the hood with the creation timestamp, but
    /// will be timestamped with the passed timestamp that needs to be
    /// estimated by the user (for example when receiving an image, we
    /// need to estimate the time when it was acquired)
    ///
    /// We may have to log these event in a different file since some
    /// of these events won't be changed if we re-sim the log.
    ///
    /// @param checkpointName The name of the checkpoint that the
    /// event represents.
    ///
    /// @param eventTimestamp The (estimated) time the event occurred.
    ///
    /// @param predecessors Zero or more predecessor specifiers.
    ///
    /// @return A generated ID for the event
    inline latency::EventID ForgedEvent(const std::string& checkpointName,
                                        const ottotime::otimePoint eventTimestamp,
                                        std::vector<latency::EventID>&& predecessors);

    /// @brief Log an event representing the sending of a message.
    ///
    /// @param channelName The name of the channel over which the
    /// message was sent.
    ///
    /// @param sendTime The time at which the message was sent. Note
    /// that, for a protocol such as SCS or DDS, which transmits the
    /// send time with the message, sendTime must be the same as the
    /// send time sent over the network, so that we can match up the
    /// send time reported by the recipient with the time with which
    /// the send event is logged.
    ///
    /// @param predecessors Zero or more predecessor specifiers.
    ///
    /// @return A generated ID for the event
    inline latency::EventID SendEvent(const std::string& channelName,
                                      const ottotime::otimePoint sendTime,
                                      std::vector<latency::EventID>&& predecessors);

    /// @brief A version of SendEvent() that lets caller specify the event ID.
    ///
    /// This is used when reporting the transmission of a CAN packet;
    /// in that case, the event ID will be generated with
    /// canEventID().
    ///
    /// @param channelName The name of the channel over which the
    /// message was sent.
    ///
    /// @param sendTime The time at which the message was sent. Note
    /// that, for a protocol such as SCS or DDS, which transmits the
    /// send time with the message, sendTime must be the same as the
    /// send time sent over the network, so that we can match up the
    /// send time reported by the recipient with the time with which
    /// the send event is logged.
    ///
    /// @param eventID The ID of the event.
    ///
    /// @param predecessors Zero or more predecessor specifiers.
    ///
    /// @return Either eventID (if logging is enabled) or 0 (otherwise).
    inline latency::EventID SendEvent(const std::string& channelName,
                                      const ottotime::otimePoint& sendTime,
                                      latency::EventID eventID,
                                      std::vector<latency::EventID>&& predecessors);

    /// @brief Log an event representing the receipt of a message.
    ///
    /// Note that ReceiveEvent() does not support specifying
    /// predecessors, because the only predecessor is the send event,
    /// which is known only by the send time.
    ///
    /// @param channelName The name of the channel over which the
    /// message was received.
    ///
    /// @param sendTime The time at which the message was sent.
    ///
    /// @param recvTime The time at which the message was received. Otherwise,
    /// time of calling this method is assumed to be receive time. Optional.
    ///
    /// @return A generated ID for the event
    inline latency::EventID ReceiveEvent(const std::string& channelName,
                                         ottotime::otimePoint sendTime,
                                         const std::optional<ottotime::otimePoint>& recvTime = std::nullopt);

    /// @brief A function which performs the actual logging.
    ///
    /// Note that this is a std::function, which means it can wrap any
    /// callable: a named function, a lambda, or an object with an
    /// call operator.
    ///
    /// @param loggingName Identifies the category of event being logged.
    ///
    /// @param checkpointName The name of the checkpoint that the
    /// event represents.
    ///
    /// @param eventTimestamp The timestamp of the event.
    ///
    /// @param eventID The ID of the event.
    ///
    /// @param predecessors Zero or more predecessor specifiers.
    using Writer = std::function<void(const std::string& loggingName,
                                      const std::string& checkpointName,
                                      const ottotime::otimePoint& eventTimestamp,
                                      const latency::EventID eventID,
                                      std::vector<latency::EventID>&& predecessors)>;

    /// @brief Helper function that allows a user to retrieve the last
    /// generated event ID for a given checkpoint
    latency::EventID getLatestEventID(const std::string& checkpointName);

    /// @brief Initialize LatencyLogger with config parameters and writer
    ///
    /// This function *must* be called before calling getLogger(), else
    /// all subsequent logging calls will result in no-ops. This is true even
    /// if initialize() is called after the first getLogger() is called.
    /// If the ConfigSection does not have valid settings, enabled_ will be set
    /// to false, and latency data will not be logged. enabled_ is set to false
    /// also if writer is a nullptr. If initialize() is called more than once,
    /// any call after the first will not have any effect.
    ///
    /// @param config Config section from which to read the settings
    /// for the logger. The setting "logLatency" is a boolean flag; if
    /// it's false, no logging will be performed. "logLatencyName" is
    /// a string, which identifies the category of events that will be
    /// logged. There can be other settings in this section, depending
    /// on the Writer in use.
    ///
    /// @param writer std::function object of type Writer, it's the actual
    /// function used to log the latency data.
    static void initialize(ConfigSection& config, const Writer& writer);

    /// @brief Initialize LatencyLogger with config parameters
    ///
    /// Here, we use a default writer of type DataLoggerWriter, which writes
    /// the latency data in msgpack format.
    static void initialize(ConfigSection& config);

    /// @brief Initialize LatencyLogger with input parameters and writer
    ///
    /// Same as the other initialize(), but here enabled_ and loggingName_ are
    /// directly set via the constructor arguments.
    static void initialize(bool enableLogging, const std::string& loggingName, const Writer& writer);

    /// @brief Initialize LatencyLogger with input parameters
    ///
    /// Same as the previous initialize(), but we use a default writer of type
    /// DataLoggerWriter, which writes the latency data in msgpack format.
    static void initialize(bool enableLogging, const std::string& loggingName);

    /// @brief Shut down LatencyLogger
    ///
    /// The function can be called at the end of a program to stop logging.
    /// Note: When using LatencyLogger in an av Task, this function *must* be
    /// called within the cleanup() function. Some destructors call comms
    /// functions that call the LatencyLogger after it is destroyed, which
    /// sometimes causes issues. That's mitigated by calling the shutdown()
    /// function first, but we will eventually need a more robust fix.
    static void shutdown();

    /// @brief Get the global logger.
    static LatencyLogger& getLogger();

    /// @brief Check if the logger has been initialized
    static inline bool isInitialized();

    /// @brief Generate an event ID.
    ///
    /// @return Use a monotonically incrementing counter to ensure uniqueness.
    static latency::EventID nextEventID();

    /// @brief Generate an event ID that corresponds to the given CAN packet ID.
    ///
    /// This is used because CAN does not transmit the packet's send
    /// timestamp (unlike SCS and DDS), but it does have a packet ID,
    /// which allows us to match up the send event with the receive event.
    ///
    /// @param canID The ID of the CAN packet.
    ///
    /// @return The event ID of the send event.
    static latency::EventID canEventID(uint32_t canID);

    /// @brief Get an event ID that corresponds to a key for a given slot.
    ///
    /// This type of event is useful when you need to recall an event ID based
    /// on a key, because the event ID can be calculated directly from the
    /// inputs, no lookup required. To avoid collisions between different types
    /// of keys there is a slot number (0 - 63) which is intended to be an a
    /// priori constant.
    ///
    /// @param slot An a priori constant used to deconflict different types of
    /// keys. For a given process and latency log, care must be taken to use
    /// unique slots for different types of keys, or else ID collisions could
    /// occur. Valid range is 0 to 63, larger numbers will be truncated without
    /// warning.
    ///
    /// @param key The key associated with this event. Valid range is 0 to
    /// 1152921504606846975, larger numbers will be truncated without warning.
    ///
    /// @return The event ID of the event.
    static latency::EventID keyEventID(uint8_t slot, uint64_t key);

private:
    /// @brief Constructor
    ///
    /// In this constructor, enabled_ and loggingName_ are set directly using
    /// the constructor arguments, instead of being read from a ConfigSection.
    ///
    /// @param enableLogging Boolean flag to enable logging. If false, no
    /// logging will be performed.
    /// @param loggingName String that identifies the category of events that
    /// will be logged. This is usually the name of the task, or the can be
    /// used to construct name of the file to which the data will be written.
    /// @param writer std::function object of type Writer, that performs the
    /// actual logging of the latency data.
    LatencyLogger(bool enableLogging, std::string loggingName, Writer writer);

    /// @brief Return a dummy logger that does nothing
    ///
    /// This is used when initialize() has not yet been called.
    static inline LatencyLogger& getDummyLogger();

    /// @brief Return an instance of the logger
    static inline LatencyLogger& getLoggerImpl(bool enableLogging = false,
                                               const std::string& loggingName = {},
                                               const Writer& writer = {});

    inline latency::EventID RegularEventInternal(const std::string& checkpointName,
                                                 std::vector<latency::EventID>&& predecessors);

    inline latency::EventID KeyEventInternal(const std::string& checkpointName,
                                             uint8_t slot,
                                             uint64_t key,
                                             std::vector<latency::EventID>&& predecessors);

    latency::EventID ForgedEventInternal(const std::string& checkpointName,
                                         const ottotime::otimePoint& eventTimestamp,
                                         const std::optional<latency::EventID>& eventIDOpt,
                                         std::vector<latency::EventID>&& predecessors);

    latency::EventID SendEventInternal(const std::string& channelName,
                                       const ottotime::otimePoint& sendTime,
                                       const std::optional<latency::EventID>& eventID,
                                       std::vector<latency::EventID>&& predecessors);

    latency::EventID ReceiveEventInternal(const std::string& channelName,
                                          const ottotime::otimePoint& sendTime,
                                          const std::optional<ottotime::otimePoint>& recvTime);

    /// Is logging enabled? If false, the non-inline backing functions
    /// will not be called.
    bool enabled_;

    /// The name of the category of events.
    std::string loggingName_;

    /// Default writer for LatencyLogger should use DataLoggerWriter
    static std::unique_ptr<latency::DataLoggerWriter> dataLoggerWriter_;

    /// The Writer that will be used to log events.
    Writer writer_;

    /// All of the FooEvent() methods take the mutex before logging
    /// anything, to ensure thread safety.
    std::mutex writerMutex_{};

    /// Map to store the last generated event ID for a given event
    std::unordered_map<std::string, latency::EventID> latestEventID_;

    // Set to true if initialize() has been called
    static bool isInitialized_;

    // To ensure that two threads
    static std::mutex initializeMutex_;
};

inline latency::EventID LatencyLogger::RegularEvent(const std::string& checkpointName,
                                                    std::vector<latency::EventID>&& predecessors)
{
    return enabled_ ? RegularEventInternal(checkpointName, std::move(predecessors)) : latency::EventID{0};
}

inline latency::EventID LatencyLogger::KeyEvent(const std::string& checkpointName,
                                                uint8_t slot,
                                                uint64_t key,
                                                std::vector<latency::EventID>&& predecessors)
{
    return enabled_ ? KeyEventInternal(checkpointName, slot, key, std::move(predecessors)) : latency::EventID{0};
}

inline latency::EventID LatencyLogger::ForgedEvent(const std::string& checkpointName,
                                                   const ottotime::otimePoint eventTimestamp,
                                                   std::vector<latency::EventID>&& predecessors)
{
    return enabled_ ? ForgedEventInternal(checkpointName, eventTimestamp, std::nullopt, std::move(predecessors))
                    : latency::EventID{0};
}

inline latency::EventID LatencyLogger::SendEvent(const std::string& channelName,
                                                 const ottotime::otimePoint sendTime,
                                                 std::vector<latency::EventID>&& predecessors)
{
    return enabled_ ? SendEventInternal(channelName, sendTime, std::nullopt, std::move(predecessors))
                    : latency::EventID{0};
}

inline latency::EventID LatencyLogger::SendEvent(const std::string& channelName,
                                                 const ottotime::otimePoint& sendTime,
                                                 latency::EventID eventID,
                                                 std::vector<latency::EventID>&& predecessors)
{
    return enabled_ ? SendEventInternal(channelName, sendTime, eventID, std::move(predecessors)) : latency::EventID{0};
}

inline latency::EventID LatencyLogger::ReceiveEvent(const std::string& channelName,
                                                    ottotime::otimePoint sendTime,
                                                    const std::optional<ottotime::otimePoint>& recvTime)
{
    return enabled_ ? ReceiveEventInternal(channelName, sendTime, recvTime) : av::latency::EventID{0};
}

inline bool LatencyLogger::isInitialized()
{
    return isInitialized_;
}

inline LatencyLogger& LatencyLogger::getDummyLogger()
{
    static LatencyLogger dummyLogger(false, std::string{}, Writer{});
    return dummyLogger;
}

/// @brief Return an instance of the logger
inline LatencyLogger& LatencyLogger::getLoggerImpl(bool enableLogging,
                                                   const std::string& loggingName,
                                                   const Writer& writer)
{
    static LatencyLogger logger(enableLogging, loggingName, writer);
    return logger;
}

inline latency::EventID LatencyLogger::RegularEventInternal(const std::string& checkpointName,
                                                            std::vector<latency::EventID>&& predecessors)
{
    return ForgedEventInternal(checkpointName, SystemClock::getSystemTime(), std::nullopt, std::move(predecessors));
}

inline latency::EventID LatencyLogger::KeyEventInternal(const std::string& checkpointName,
                                                        uint8_t slot,
                                                        uint64_t key,
                                                        std::vector<latency::EventID>&& predecessors)
{
    return ForgedEventInternal(
        checkpointName, SystemClock::getSystemTime(), keyEventID(slot, key), std::move(predecessors));
}

} // namespace av
