/*!
 ***********************************************************************************************************************
 * MOTIONAL PRIVILEGED AND CONFIDENTIAL
 *
 * @file LatencyLoggerTest.cc
 * @brief This file implements tests for the LatencyLogger class.
 *
 * @section License
 *
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#include "platform/middleware/libs/LatencyLogger/test/TestWriter.hpp"
#include "platform/middleware/libs/config/include/config/ConfigParser.h"
#include <gtest/gtest.h>

namespace av {

namespace {

TestWriter testWriter;

class LatencyLoggerTest : public ::testing::Test
{
public:
    LatencyLoggerTest() {}

    void SetUp() override
    {
        ConfigParser parser;
        ConfigSection config;
        parser.parseString("Parameters = {\n"
                           "  loggingName = \"LatencyLoggerTest\";\n"
                           "  enabled = True;\n"
                           "}\n");
        parser.getSection("Parameters", config);

        LatencyLogger::initialize(config, std::ref(testWriter));

        testWriter.clear();
    }
};

TEST_F(LatencyLoggerTest, RegularEventNoPreds)
{
    EXPECT_TRUE(LatencyLogger::isInitialized());

    LatencyLogger& logger = LatencyLogger::getLogger();
    av::latency::EventID event1 = logger.RegularEvent("Liftoff");

    const auto& entries = testWriter.getEntries();
    ASSERT_EQ(1, entries.size());

    EXPECT_EQ(entries[0].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[0].checkpointName, "Liftoff");
    EXPECT_EQ(entries[0].eventID, event1);
    EXPECT_TRUE(entries[0].predecessors.empty());

    EXPECT_EQ(logger.getLatestEventID("Liftoff"), event1);
}

TEST_F(LatencyLoggerTest, RegularEventPred)
{
    LatencyLogger& logger = LatencyLogger::getLogger();

    av::latency::EventID event1 = logger.RegularEvent("Liftoff", {12});

    const auto& entries = testWriter.getEntries();
    ASSERT_EQ(1, entries.size());

    EXPECT_EQ(entries[0].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[0].checkpointName, "Liftoff");
    EXPECT_EQ(entries[0].eventID, event1);
    EXPECT_EQ(entries[0].predecessors.size(), 1);
    EXPECT_EQ(entries[0].predecessors[0], 12);

    EXPECT_EQ(logger.getLatestEventID("Liftoff"), event1);
}

TEST_F(LatencyLoggerTest, TwoRegularEvents)
{
    EXPECT_TRUE(LatencyLogger::isInitialized());

    LatencyLogger& logger = LatencyLogger::getLogger();

    const av::latency::EventID liftoff = logger.RegularEvent("Liftoff");

    av::latency::EventID splashdown = logger.RegularEvent("Splashdown", {liftoff});

    EXPECT_NE(liftoff, splashdown);

    const auto& entries = testWriter.getEntries();
    ASSERT_EQ(2, entries.size());

    EXPECT_EQ(entries[0].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[0].checkpointName, "Liftoff");
    EXPECT_EQ(entries[0].eventID, liftoff);
    EXPECT_TRUE(entries[0].predecessors.empty());

    EXPECT_EQ(entries[1].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[1].checkpointName, "Splashdown");
    EXPECT_EQ(entries[1].eventID, splashdown);
    EXPECT_EQ(entries[1].predecessors.size(), 1);
    EXPECT_EQ(entries[1].predecessors[0], liftoff);

    EXPECT_EQ(logger.getLatestEventID("Liftoff"), liftoff);
    EXPECT_EQ(logger.getLatestEventID("Splashdown"), splashdown);
    EXPECT_EQ(logger.getLatestEventID("InvalidCheckpoint"), av::latency::EventID{0});
}

TEST_F(LatencyLoggerTest, KeyEvent)
{
    constexpr uint8_t kLaunchKeySlot = UINT8_C(10);

    EXPECT_TRUE(LatencyLogger::isInitialized());

    LatencyLogger& logger = LatencyLogger::getLogger();

    // Log the event with calculated event ID
    const av::latency::EventID press = logger.KeyEvent("Press", kLaunchKeySlot, 1234);

    // Recreate the key by calculating it from the slot and key
    const av::latency::EventID recreatedPress = LatencyLogger::keyEventID(kLaunchKeySlot, 1234);
    av::latency::EventID release = logger.RegularEvent("Release", {recreatedPress});

    EXPECT_EQ(press, recreatedPress);
    EXPECT_NE(press, release);

    const auto& entries = testWriter.getEntries();
    ASSERT_EQ(2, entries.size());

    EXPECT_EQ(entries[0].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[0].checkpointName, "Press");
    EXPECT_EQ(entries[0].eventID, press);
    EXPECT_TRUE(entries[0].predecessors.empty());

    EXPECT_EQ(entries[1].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[1].checkpointName, "Release");
    EXPECT_EQ(entries[1].eventID, release);
    EXPECT_EQ(entries[1].predecessors.size(), 1);
    EXPECT_EQ(entries[1].predecessors[0], press);
}

TEST_F(LatencyLoggerTest, ForgedEvent)
{
    LatencyLogger& logger = LatencyLogger::getLogger();

    struct tm timeParts
    {
        7, 28, 13, 12, 3, 121, 1, 101, 0, 0, "GMT"
    };
    const ottotime::otimePoint t(timeParts);
    struct tm predTimeParts
    {
        5, 28, 13, 12, 3, 121, 1, 101, 0, 0, "GMT"
    };
    const ottotime::otimePoint predT(predTimeParts);

    const av::latency::EventID liftoff =
        logger.ForgedEvent("Liftoff", t, {static_cast<av::latency::EventID>(predT.toMicrosecondsSinceEpoch())});

    const auto& entries = testWriter.getEntries();
    ASSERT_EQ(1, entries.size());

    EXPECT_EQ(entries[0].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[0].checkpointName, "Liftoff");
    EXPECT_EQ(entries[0].eventID, liftoff);
    EXPECT_EQ(entries[0].predecessors.size(), 1);
    EXPECT_EQ(entries[0].predecessors[0], predT.toMicrosecondsSinceEpoch());

    EXPECT_EQ(logger.getLatestEventID("Liftoff"), liftoff);
}

TEST_F(LatencyLoggerTest, SendEvent)
{
    LatencyLogger& logger = LatencyLogger::getLogger();

    struct tm timeParts
    {
        7, 28, 13, 12, 3, 121, 1, 101, 0, 0, "GMT"
    };
    const ottotime::otimePoint t(timeParts);

    const av::latency::EventID ansible = logger.SendEvent("Ansible", t, {12});

    const auto& entries = testWriter.getEntries();
    ASSERT_EQ(1, entries.size());

    EXPECT_EQ(entries[0].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[0].checkpointName, "send:Ansible");
    EXPECT_EQ(entries[0].eventID, ansible);
    EXPECT_EQ(entries[0].predecessors.size(), 1);
    EXPECT_EQ(entries[0].predecessors[0], 12);

    EXPECT_EQ(logger.getLatestEventID("send:Ansible"), ansible);
}

TEST_F(LatencyLoggerTest, ReceiveEvent)
{
    LatencyLogger& logger = LatencyLogger::getLogger();

    struct tm timeParts
    {
        7, 28, 13, 12, 3, 121, 1, 101, 0, 0, "GMT"
    };
    const ottotime::otimePoint t(timeParts);

    const av::latency::EventID ansible = logger.ReceiveEvent("Ansible", t);

    const auto& entries = testWriter.getEntries();
    ASSERT_EQ(1, entries.size());

    EXPECT_EQ(entries[0].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[0].checkpointName, "recv:Ansible");
    EXPECT_EQ(entries[0].eventID, ansible);
    EXPECT_EQ(entries[0].predecessors.size(), 1);
    EXPECT_EQ(entries[0].predecessors[0], t.toMicrosecondsSinceEpoch());

    EXPECT_EQ(logger.getLatestEventID("recv:Ansible"), ansible);
}

TEST_F(LatencyLoggerTest, ReceiveEventForged)
{
    LatencyLogger& logger = LatencyLogger::getLogger();

    struct tm timePartsSendEv
    {
        7, 28, 13, 12, 3, 121, 1, 101, 0, 0, "GMT"
    };
    struct tm timePartsForgedRecvEv
    {
        9, 28, 13, 12, 3, 121, 1, 101, 0, 0, "GMT"
    };
    const ottotime::otimePoint tSend(timePartsSendEv);
    const ottotime::otimePoint tRecv(timePartsForgedRecvEv);

    const av::latency::EventID ansible = logger.ReceiveEvent("Ansible", tSend, tRecv);

    const auto& entries = testWriter.getEntries();
    ASSERT_EQ(1, entries.size());

    EXPECT_EQ(entries[0].loggingName, "LatencyLoggerTest");
    EXPECT_EQ(entries[0].checkpointName, "recv:Ansible");
    EXPECT_EQ(entries[0].eventID, ansible);
    EXPECT_EQ(entries[0].predecessors.size(), 1);
    EXPECT_EQ(entries[0].predecessors[0], tSend.toMicrosecondsSinceEpoch());
    EXPECT_EQ(entries[0].timestamp, tRecv);

    EXPECT_EQ(logger.getLatestEventID("recv:Ansible"), ansible);
}

} // namespace
} // namespace av
