/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file LogsCutterUtility.cc
 * @product Common
 * @version TODO
 * @released TODO
 * @updated Jul 11, 2018
 * @brief This file implements a C++ utility to search for all log files of a given directory and generate cut-logs.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 ***********************************************************************************************************************
 */

//=== Includes =========================================================================================================
#include "LogsCutterUtility/LogsCutterUtility.h"

#include <dataLog/utility.h>

#include <sqlite3.h>

#include <ctime>   // size_t
#include <fstream> // ifstream, ofstream
#include <iomanip>
#include <optional>
#include <sstream> // stringstream

#include "infrastructure/logging/log_reader/DataLogCopy.hpp"
#include "infrastructure/logging/log_reader/LogURL.hpp"
#include "infrastructure/logging/log_reader/TimestampRange.hpp"
#include "lib/avlog/avlog.hpp"
#include <TimeStamp/otime.h>
#include <TimeStamp/otimeConversions.h>

namespace filesystem = std::experimental::filesystem;
namespace log_reader = av::logging::log_reader;

//=== Static Constants =================================================================================================

const std::size_t LogsCutterUtility::timeStrLen_(25);

//=== Implementation ===================================================================================================

/*!
 * @brief Default constructor for LogsCutterUtility
 *
 * @return An instance of LogsCutterUtility
 */
LogsCutterUtility::LogsCutterUtility()
    : beforeOffset_(ottotime::oSeconds(0))
    , afterOffset_(ottotime::oSeconds(0))
{}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Destructor for LogsCutterUtility
 */
LogsCutterUtility::~LogsCutterUtility()
{
    // Nothing to do here
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Sets the log directory location to search
 *
 * @param logDir : Full-path description of the logs directory
 * @return void
 */
void LogsCutterUtility::setLogDirectory(const std::string& logDir)
{
    logDir_ = logDir;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Sets the cut-logs destination directory location
 *
 * @param destDir : Full-path description of the destination directory
 * @return void
 */
void LogsCutterUtility::setDestDirectory(const std::string& destDir)
{
    destDir_ = destDir;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Sets the event time of interest
 *
 * @param evtTime : Time of the event
 * @return void
 */
void LogsCutterUtility::setEventTime(const ottotime::otimePoint& evtTime)
{
    eventTime_ = evtTime;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Sets the before-event offset
 *
 * @param beforeEvtOffset : Offset amount before the event
 * @return void
 */
void LogsCutterUtility::setBeforeEventOffset(const ottotime::otimeDuration& beforeEvtOffset)
{
    beforeOffset_ = beforeEvtOffset;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Sets the after-event offset
 *
 * @param afterEvtOffset : Offset amount after the event
 * @return void
 */
void LogsCutterUtility::setAfterEventOffset(const ottotime::otimeDuration& afterEvtOffset)
{
    afterOffset_ = afterEvtOffset;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Returns count of number of files in log directory.
 *
 * @return std::size_t : Count of files in current log directory
 */
std::size_t LogsCutterUtility::howManyFiles()
{
    std::size_t ret = 0;
    filesystem::path logPath = filesystem::system_complete(filesystem::path(logDir_));

    if (!filesystem::exists(logPath)) {
        AVLOG_ERROR("{} does not exist", AVLOG_ARGS("path"), logDir_);
    } else {
        filesystem::directory_iterator endIter;
        for (filesystem::directory_iterator it(logPath); it != endIter; ++it) {
            ++ret;
        }
    }

    return ret;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Executes the cut-logs process
 *
 * Cuts all logs (*.db files) found at logDir_ and stores them in destDir_ (or ~/cutLogs if unset)
 *
 * @return void
 */
bool LogsCutterUtility::cutAllLogs()
{
    bool validArgs{false};

    filesystem::path logPath = filesystem::system_complete(filesystem::path(logDir_));
    ottotime::otimePoint startCut = eventTime_ - beforeOffset_;
    ottotime::otimePoint endCut = eventTime_ + afterOffset_;
    if (validateParameters(logPath, startCut, endCut)) {
        validArgs = true;

        // All predicates are satisfied, so let's do some work!
        filesystem::path cutDir(destDir_);
        prepareTempFiles(cutDir);

        const std::string sqlExt = ".db";
        const std::string pstatBmw = "@bmw";
        const std::string pstatMvp = "@mvp";
        filesystem::directory_iterator endIter;
        for (filesystem::directory_iterator it(logPath); it != endIter; ++it) {
            filesystem::path logCandidate = it->path();
            if (!filesystem::is_directory(logCandidate)) {
                // Cut if it's a SCS log file
                if (logCandidate.extension().native() == sqlExt) {
                    AVLOG_DEBUG("Cutting log file {}", AVLOG_ARGS("path"), logCandidate.native());

                    cutSCSLog(logCandidate.native(), startCut, endCut);
                    filesystem::path tmpLog("temp.db");
                    filesystem::path cutLog = cutDir;
                    cutLog /= logCandidate.filename();

                    filesystem::copy(tmpLog, cutLog);
                    filesystem::remove(tmpLog);
                }

                // Cut if it's a Process Status log file
                else if (logCandidate.native().find(pstatBmw) != std::string::npos ||
                         logCandidate.native().find(pstatMvp) != std::string::npos) {
                    AVLOG_DEBUG("Cutting log file {}", AVLOG_ARGS("path"), logCandidate.native());

                    cutStatusLog(logCandidate.native(), startCut, endCut);
                    filesystem::path tmpLog("temp.txt");
                    filesystem::path cutLog = cutDir;
                    cutLog /= logCandidate.filename();

                    filesystem::copy(tmpLog, cutLog);
                    filesystem::remove(tmpLog);
                } else {
                    AVLOG_DEBUG("Skipping processing of file {}", AVLOG_ARGS("path"), logCandidate.native());
                }
            }
        }

        // Cut AVPL logs
        log_reader::TimestampRange timestampRange{av::timePointFromOtimePoint(startCut),
                                                  av::timePointFromOtimePoint(endCut)};
        log_reader::DataLogCopy::run(log_reader::LogURL{logPath.string()}, cutDir, {}, timestampRange);

        copySupportFiles();
    }

    return validArgs;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Cuts a single SCS log, generating a file named temp.db
 *
 * @param logFile : Name of the log to cut
 * @param beginCut : Earliest time to filter
 * @param endCut : Latest time to filter
 * @return void
 * @sa logSnipSQL.sh
 */
void LogsCutterUtility::cutSCSLog(const std::string& logFile,
                                  const ottotime::otimePoint& beginCut,
                                  const ottotime::otimePoint& endCut)
{
    struct sqlite3* writeDbHandle = nullptr;
    int32_t sqlErrCode = SQLITE_OK;
    char* sqlErrMsg = nullptr;

    // Prepare the database for operations
    sqlErrCode = sqlite3_open_v2("temp.db", &writeDbHandle, SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, NULL);
    if (sqlErrCode != SQLITE_OK) {
        AVLOG_ERROR("Can't open temporary database file!\n\tSqlite3 err msg: {}",
                    AVLOG_ARGS("errMsg"),
                    sqlite3_errmsg(writeDbHandle));
    } else {
        /*
         * File opened ok, so let's try to follow the winning formula
         */
        std::string createStmt;
        std::string attachTempStmt;
        std::stringstream attachOrigStmt;
        std::stringstream cutStmt;

        attachTempStmt = "attach \'temp.db\' as cut_db;";
        attachOrigStmt << "attach \'" << logFile << "\' as ori_db;";
        cutStmt << "INSERT INTO cut_db.OTTOLOG SELECT * from ori_db.OTTOLOG WHERE time >= "
                << beginCut.to_otimeSpec().tv_sec << " and time < " << endCut.to_otimeSpec().tv_sec << ";";

        // Carry the log version forward
        struct sqlite3* readDbHandle = nullptr;
        auto logVersion = dataLog::DATA_LOG_VERSION;
        sqlErrCode = sqlite3_open_v2(logFile.c_str(), &readDbHandle, SQLITE_OPEN_READONLY, NULL);
        if (sqlErrCode != SQLITE_OK) {
            AVLOG_WARN("Can't open {} for version info. Defaulting to latest version", AVLOG_ARGS("path"), logFile);
            logVersion = dataLog::DATA_LOG_VERSION;
        } else {
            sqlErrCode = dataLog::readLogVersion(readDbHandle, logVersion, &sqlErrMsg);
            if (sqlErrCode != SQLITE_OK) {
                AVLOG_WARN("Failed to read log version info from {}. Defaulting to latest version",
                           AVLOG_ARGS("path"),
                           logFile);
                logVersion = dataLog::DATA_LOG_VERSION;
            }
        }

        // Find out how many columns in data
        sqlite3_stmt* pStmt = nullptr;
        const char* stmt = "SELECT * FROM OTTOLOG LIMIT 1";
        sqlite3_prepare_v2(readDbHandle, stmt, static_cast<int>(strlen(stmt)), &pStmt, nullptr);
        int colNum = sqlite3_column_count(pStmt);
        sqlErrCode = sqlite3_finalize(pStmt);
        if (sqlErrCode != SQLITE_OK) {
            AVLOG_WARN("Failed to detect column number in table");
        }
        if (colNum == 4) {
            createStmt = "CREATE TABLE OTTOLOG(time REAL PRIMARY KEY NOT NULL, data BLOB NOT NULL,  publisherId "
                         "INTEGER NOT NULL,  messageCounter INTEGER NOT NULL);";
        } else {
            createStmt = "CREATE TABLE OTTOLOG(time REAL PRIMARY KEY NOT NULL, data BLOB NOT NULL);";
        }

        std::string versionCmd = "PRAGMA user_version=" + std::to_string(logVersion);
        sqlErrCode = sqlite3_exec(writeDbHandle, versionCmd.c_str(), nullptr, nullptr, &sqlErrMsg);

        /*
         * Executing all of the above SQL statements only if the previous one completed successfully
         */
        sqlErrCode = sqlite3_exec(writeDbHandle, createStmt.c_str(), sqliteCallback, NULL, &sqlErrMsg);
        if (sqlErrCode != SQLITE_OK) {
            logSqlStmtError(createStmt, sqlErrMsg);
        } else {
            sqlErrCode = sqlite3_exec(writeDbHandle, attachTempStmt.c_str(), sqliteCallback, NULL, &sqlErrMsg);
            if (sqlErrCode != SQLITE_OK) {
                logSqlStmtError(attachTempStmt, sqlErrMsg);
            } else {
                sqlErrCode =
                    sqlite3_exec(writeDbHandle, attachOrigStmt.str().c_str(), sqliteCallback, NULL, &sqlErrMsg);
                if (sqlErrCode != SQLITE_OK) {
                    logSqlStmtError(attachOrigStmt.str(), sqlErrMsg);
                } else {
                    sqlErrCode = sqlite3_exec(writeDbHandle, cutStmt.str().c_str(), sqliteCallback, NULL, &sqlErrMsg);
                    if (sqlErrCode != SQLITE_OK) {
                        logSqlStmtError(cutStmt.str(), sqlErrMsg);
                    }
                }
            }
        }

        sqlite3_close(readDbHandle);
    }

    // Cleanup!
    sqlite3_close(writeDbHandle);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Cuts a single Process Status log, generating a file named temp.txt
 *
 * @param logFile : Name of the log to cut
 * @param beginCut : Earliest time to filter
 * @param endCut : Latest time to filter
 * @return void
 */
void LogsCutterUtility::cutStatusLog(const std::string& logFile,
                                     const ottotime::otimePoint& beginCut,
                                     const ottotime::otimePoint& endCut)
{
    std::ifstream logIn(logFile.c_str());
    std::ofstream cutOut("temp.txt");

    if (logIn.is_open() && cutOut.is_open()) {
        std::string logLine;
        bool validEntry = false;

        // Find first valid time line
        while (!logIn.eof() && !validEntry) {
            std::getline(logIn, logLine);

            // Check the timestamp if available
            auto entryTime = parseTimestamp(logLine);

            if (!entryTime) {
                continue; // It wasn't a timestamped line, so let's move to the next line
            } else {
                validEntry = *entryTime >= beginCut;
            }
        }

        // Write each line until we find next invalid time line
        if (validEntry) {
            cutOut << logLine << std::endl;
        }

        while (!logIn.eof() && validEntry) {
            std::getline(logIn, logLine);
            cutOut << logLine << std::endl;

            // Check the timestamp if available
            auto entryTime = parseTimestamp(logLine);

            if (!entryTime) {
                continue; // It wasn't a timestamped line, so let's just move on.
            } else {
                validEntry = *entryTime <= endCut;
            }
        }

        // Clean up
        cutOut.close();
        logIn.close();
    } else {
        if (!logIn.is_open()) {
            AVLOG_ERROR("{} failed to open!", AVLOG_ARGS("path"), logFile);
        }
        if (!cutOut.is_open()) {
            AVLOG_ERROR("temp.txt failed to open!");
        }
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Static callback needed by SQLite3 library. This does nothing.
 *
 * @param All four params are not used
 * @return int : Always zero
 * @sa https://www.sqlite.org/quickstart.html
 */
int LogsCutterUtility::sqliteCallback(void*, int, char**, char**)
{
    // This really has no purpose other than to satisfy the sqlite3_exec() interface.
    // Please do not modify this method.
    return 0;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Logs SQLite3 statement errors
 *
 * @param sqlStmt : Statement that was attempted
 * @param sqlErrCstr : Sqlite3's error message
 * @return void
 */
void LogsCutterUtility::logSqlStmtError(const std::string& sqlStmt, char* sqlErrCstr)
{
    AVLOG_ERROR("Can't execute statement \'{}\'\n\tSqlite3 Err Msg: {}",
                AVLOG_ARGS("statement", "errMsg"),
                sqlStmt,
                sqlErrCstr);
    sqlite3_free(sqlErrCstr);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Copies a single support file (DriveInfoSummary*.txt, LocalMap-*.txt, or config.tar.gz*)
 *
 * @param cpyCandidate : Path to support file to be copied
 * @param destDir : Absolute path string of the destination directory
 * @return void
 */
void LogsCutterUtility::copySupportFile(const filesystem::path& cpyCandidate, const std::string& destDir)
{
    const std::string disPrefix = "DriveInfoSummary";
    const std::string lmPrefix = "Local_Map_";
    const std::string disExtent = ".txt";
    const std::string cfgPrefix = "config.tar.gz";
    filesystem::path cutDir(destDir);

    if (((cpyCandidate.filename().native().substr(0, disPrefix.size()) == disPrefix) && // Starts with DriveInfoSummary
         (cpyCandidate.extension().native() == disExtent)) ||                           // Ends with *.txt
        ((cpyCandidate.filename().native().substr(0, lmPrefix.size()) == lmPrefix) &&   // Starts with Local_Map_
         (cpyCandidate.extension().native() == disExtent)) ||                           // Ends with *.txt
        (cpyCandidate.filename().native().substr(0, cfgPrefix.size()) == cfgPrefix))    // Starts with config.tar.gz*
    {
        filesystem::path cpyTarget = cutDir;
        cpyTarget /= cpyCandidate.filename();
        if (filesystem::exists(cpyTarget)) {
            filesystem::remove(cpyTarget);
        }
        filesystem::copy(cpyCandidate, cpyTarget);
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Copies support files for the log (DriveInfoSummary*.txt, Local_Map_*.txt, and config.tar.gz* files)
 *
 * @return void
 */
void LogsCutterUtility::copySupportFiles()
{
    filesystem::path logPath(logDir_);

    filesystem::directory_iterator endIter; // Some sort of boost black magic...
    for (filesystem::directory_iterator it(logPath); it != endIter; ++it) {
        copySupportFile(it->path(), destDir_);
    }
}

/*!
 * @brief Parses the timestamp in UTC from a log line
 *
 * @param logLine a single line from a status log
 *
 * @return otimePoint representing the timestamp or nothing
 */
std::optional<ottotime::otimePoint> LogsCutterUtility::parseTimestamp(const std::string& logLine)
{
    using namespace std;
    using namespace chrono;
    std::stringstream timeParseStream;
    timeParseStream << logLine.substr(0, timeStrLen_);

    const auto timestamp = logger::LogEvent::parseTime(timeParseStream.str());
    if (timestamp == logger::LogEvent::ChronoTimePoint{}) {
        // Failed to parse a timestamp from this line
        return std::nullopt;
    }

    const auto sinceEpoch = timestamp.time_since_epoch();
    ottotime::otimeSpec entryTimeTS{};
    entryTimeTS.tv_sec = duration_cast<seconds>(sinceEpoch).count();
    entryTimeTS.tv_nsec = (duration_cast<nanoseconds>(sinceEpoch) - duration_cast<seconds>(sinceEpoch)).count();
    entryTimeTS.normalize();

    return ottotime::otimePoint{entryTimeTS};
}

void LogsCutterUtility::prepareTempFiles(const filesystem::path& cutDir)
{
    if (filesystem::exists(cutDir)) {
        filesystem::remove_all(cutDir);
    }
    filesystem::create_directories(cutDir);

    const filesystem::path oldTmpScsLog{"temp.db"};
    if (filesystem::exists(oldTmpScsLog)) {
        // Likely left over from some previous failed run
        filesystem::remove(oldTmpScsLog);
    }

    const filesystem::path oldTmpPstatLog{"temp.txt"};
    if (filesystem::exists(oldTmpPstatLog)) {
        // Likely left over from some previous failed run
        filesystem::remove(oldTmpPstatLog);
    }
}

bool LogsCutterUtility::validateParameters(const filesystem::path& logPath,
                                           const ottotime::otimePoint& startCut,
                                           const ottotime::otimePoint& endCut)
{
    bool isValid{false};
    if (!filesystem::exists(logPath)) {
        AVLOG_ERROR("{} does not exist!", AVLOG_ARGS("path"), logPath.native());
    } else if (!filesystem::is_directory(logPath)) {
        AVLOG_ERROR("{} is not a directory!", AVLOG_ARGS("path"), logPath.native());
    } else if (!eventTime_.isTimeValid()) {
        AVLOG_ERROR("Event time is not properly set!");
    } else if (startCut == endCut) {
        AVLOG_WARN("Invalid offsets set; no cutting executed!");
    } else if (destDir_.empty()) {
        AVLOG_WARN("Destination directory wasn't set; setting it to ~/cutLogs");
        destDir_ = "~/cutLogs";
    } else {
        isValid = true;
    }

    return isValid;
}
