/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file LogsCutterUtility.h
 * @product Common
 * @version TODO
 * @released TODO
 * @updated Aug 1, 2018
 * @brief This file describes a C++ utility to search for all log files of a given directory and generate cut-logs.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 ***********************************************************************************************************************
 */

#ifndef LOGSCUTTERUTILITY_H_
#define LOGSCUTTERUTILITY_H_

#include <experimental/filesystem>
#include <log/Logger.h>

#include <optional>
#include <string>

#include <TimeStamp/otime.h>

// Forward Declarations

struct sqlite3;

// Type Declarations

/*!
 * @class LogsCutterUtility
 *
 * @brief A utility to search for all log files of a given directory and generate cut-logs.
 *
 * These cut-logs are contained within the bounds of before-event and after-event offsets.
 *
 * @sa See AHT-779 on Jira
 */
class LogsCutterUtility
{
public:
    /*!
     * @brief Default constructor for LogsCutterUtility
     *
     * @return An instance of LogsCutterUtility
     */
    LogsCutterUtility();

    /*!
     * @brief Destructor for LogsCutterUtility
     */
    virtual ~LogsCutterUtility();

    /*!
     * @brief Sets the log directory location to search
     *
     * @param logDir : Full-path description of the logs directory
     * @return void
     */
    void setLogDirectory(const std::string& logDir);

    /*!
     * @brief Sets the cut-logs destination directory location
     *
     * @param destDir : Full-path description of the destination directory
     * @return void
     */
    void setDestDirectory(const std::string& destDir);

    /*!
     * @brief Sets the event time of interest
     *
     * @param evtTime : Time of the event
     * @return void
     */
    void setEventTime(const ottotime::otimePoint& evtTime);

    /*!
     * @brief Sets the before-event offset
     *
     * @param beforeEvtOffset : Offset amount before the event*
     * @return void
     */
    void setBeforeEventOffset(const ottotime::otimeDuration& beforeEvtOffset);

    /*!
     * @brief Sets the after-event offset
     *
     * @param afterEvtOffset : Offset amount after the event
     * @return void
     */
    void setAfterEventOffset(const ottotime::otimeDuration& afterEvtOffset);

    /*!
     * @brief Returns count of number of files in log directory.
     *
     * @return std::size_t : Count of files in current log directory
     */
    std::size_t howManyFiles();

    /*!
     * @brief Executes the cut-logs process
     *
     * Cuts all logs (*.db files) found at logDir_ and stores them in destDir_ (or ~/cutLogs if unset)
     *
     * @return Whether it successfully cut the log or not.
     */
    bool cutAllLogs();

    /*!
     * @brief Cuts a single SCS log, generating a file named temp.db
     *
     * @param logFile : Name of the log to cut
     * @param beginCut : Earliest time to filter
     * @param endCut : Latest time to filter
     * @return void
     * @sa logSnipSQL.sh
     */
    void cutSCSLog(const std::string& logFile,
                   const ottotime::otimePoint& beginCut,
                   const ottotime::otimePoint& endCut);

    /*!
     * @brief Cuts a single Process Status log, generating a file named temp.txt
     *
     * @param logFile : Name of the log to cut
     * @param beginCut : Earliest time to filter
     * @param endCut : Latest time to filter
     * @return void
     */
    void cutStatusLog(const std::string& logFile,
                      const ottotime::otimePoint& beginCut,
                      const ottotime::otimePoint& endCut);

    /*!
     * @brief Copies a single support file (DriveInfoSummary*.txt, LocalMap-*.txt, or config.tar.gz*)
     *
     * @param cpyCandidate : Path to support file to be copied
     * @param destDir : Absolute path string of the destination directory
     * @return void
     */
    void copySupportFile(const std::experimental::filesystem::path& cpyCandidate, const std::string& destDir);

    /*!
     * @brief Copies support files for the log (DriveInfoSummary*.txt, LocalMap-*.txt, and config.tar.gz* files)
     *
     * @return void
     */
    void copySupportFiles();

    static std::optional<ottotime::otimePoint> parseTimestamp(const std::string& logLine);

private:
    std::string logDir_;                   ///< Location of the log files
    std::string destDir_;                  ///< Destination of the cut log files
    ottotime::otimePoint eventTime_;       ///< Time of the event of interest
    ottotime::otimeDuration beforeOffset_; ///< Offset amount prior to the event to start the cut-log
    ottotime::otimeDuration afterOffset_;  ///< Offset amount in excess of the event to end the cut-log
    logger::Logger logger_;                ///< Logger used by the process
    static const std::size_t timeStrLen_;  ///< Expected length of a timestring in process status log

    /*!
     * @brief Static callback needed by SQLite3 library. This does nothing.
     *
     * @param All four params are not used
     * @return int : Always zero
     * @sa https://www.sqlite.org/quickstart.html
     */
    static int sqliteCallback(void*, int, char**, char**);

    /*!
     * @brief Logs SQLite3 statement errors
     *
     * @param sqlStmt : Statement that was attempted
     * @param sqlErrCstr : Sqlite3's error message
     * @return void
     */
    void logSqlStmtError(const std::string& sqlStmt, char* sqlErrCstr);

    /// @brief Clean the target cutDir and prepare temporary files
    void prepareTempFiles(const std::experimental::filesystem::path& cutDir);

    /// @brief Check that the path contains a log and the start and endpoints are valid.
    /// @param logPath : path to log
    /// @param startCut : beginning of cut
    /// @param endCut : end of cut
    /// @return true if params deemed valid
    bool validateParameters(const std::experimental::filesystem::path& logPath,
                            const ottotime::otimePoint& startCut,
                            const ottotime::otimePoint& endCut);
};

#endif /* LOGSCUTTERUTILITY_H_ */
