/*!
 ***********************************************************************************************************************
 * @file TestLogsCutterUtility.cpp
 * @brief This file implements a concise unit test of the LogsCutterUtility class
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 *
 ***********************************************************************************************************************
 */

// Includes
#include "LogsCutterUtility/LogsCutterUtility.h"
#include "platform/middleware/libs/TestUtils/include/TestUtils/TmpDir.hpp"
#include <log/Logger.h>

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <chrono>
#include <cstdlib>
#include <ctime>
#include <experimental/filesystem>
#include <iomanip>
#include <iostream>
#include <optional>
#include <string>

namespace chr = std::chrono;
namespace fs = std::experimental::filesystem;
using ::testing::HasSubstr;

#ifndef LOGSDIR
#define LOGSDIR "platform/middleware/libs/LogsCutterUtility/test/"
#endif

#ifndef CUTSDIR
#define CUTSDIR "./"
#endif

class LogsCutterUtility_CutFiles : public ::testing::Test
{
protected:
    void SetUp() override
    {
        auto cutDir = tmpDir_.getPath() / "unittest/testcuts1";
        testCutDir_ = cutDir.string();

        cut_.setLogDirectory(testLogDir_);
        cut_.setDestDirectory(testCutDir_);
        cut_.setEventTime(evtTime_);
        cut_.setBeforeEventOffset(beforeTime_);
        cut_.setAfterEventOffset(afterTime_);
    }

    TestUtils::TestTmpDirectory tmpDir_;
    std::string testLogDir_{"platform/middleware/libs/LogsCutterUtility/test/unittest/testlogs1"};
    std::string testCutDir_;
    ottotime::otimePoint evtTime_{ottotime::otimeDuration{ottotime::oSeconds(1517615510)}};
    ottotime::otimeDuration beforeTime_{ottotime::oSeconds(10.0)};
    ottotime::otimeDuration afterTime_{ottotime::oSeconds(10.0)};

    LogsCutterUtility cut_;
};

//=== TESTS ============================================================================================================

/*!
 * @test CutFiles:CutAllDBFiles
 *
 * @brief Searches through unittest/testlogs1, cuts *.db files, and stores them in unittest/testcuts1
 *
 * This will require the user to search the resulting directory to see if the cuts are appropriate sized and other log
 * artifacts are copied, while omitting junk files.
 */
TEST_F(LogsCutterUtility_CutFiles, CutAllDBFiles)
{
    bool result{false};
    EXPECT_NO_THROW(result = cut_.cutAllLogs());
    EXPECT_TRUE(result);

    EXPECT_TRUE(fs::exists(testCutDir_));
    ASSERT_TRUE(fs::is_directory(testCutDir_));

    // Count the files
    size_t avplFiles = 0;
    size_t dbFiles = 0;
    size_t totalFiles = 0;
    for (auto const& entry : fs::directory_iterator(testCutDir_)) {
        EXPECT_TRUE(fs::is_regular_file(entry.status())) << entry.path().string();
        auto const path = entry.path();
        if (path.extension() == ".avpl") {
            ++avplFiles;
        } else if (path.extension() == ".db") {
            ++dbFiles;
        }
        ++totalFiles;
    }
    EXPECT_EQ(avplFiles, 1);
    EXPECT_EQ(dbFiles, 18);
    EXPECT_EQ(totalFiles, 25);
}

/*!
 * @test CutFiles:ErrorLogPath
 *
 * @brief test an error case, which given log path doesn't exist
 *
 * This will require the user to search the resulting directory to see if the cuts are appropriate sized and other log
 * artifacts are copied, while omitting junk files.
 */
TEST_F(LogsCutterUtility_CutFiles, ErrorLogPath)
{
    std::string testLogDir{testLogDir_};
    testLogDir += "unittest/noSuchLogPath";

    cut_.setLogDirectory(testLogDir);

    bool result{true};
    EXPECT_NO_THROW(result = cut_.cutAllLogs());
    EXPECT_FALSE(result);
}

/*!
 * @test CutFiles:LogPathIsNotDirectory
 *
 * @brief test an error case, which given log path is not a directory
 *
 * This will require the user to search the resulting directory to see if the cuts are appropriate sized and other log
 * artifacts are copied, while omitting junk files.
 */
TEST_F(LogsCutterUtility_CutFiles, LogPathIsNotDirectory)
{
    std::string testLogDir(testLogDir_);
    testLogDir += "/HMICommand.db";
    cut_.setLogDirectory(testLogDir);

    bool result{true};
    EXPECT_NO_THROW(result = cut_.cutAllLogs());
    EXPECT_FALSE(result);
}

/*!
 * @test CutFiles:InvalidEventTime
 *
 * @brief test an error case, which the given event time is not valid
 *
 * This will require the user to search the resulting directory to see if the cuts are appropriate sized and other log
 * artifacts are copied, while omitting junk files.
 */
TEST_F(LogsCutterUtility_CutFiles, InvalidEventTime)
{
    ottotime::otimeDuration evtTimeFromEpoch(ottotime::oSeconds(9517615510));
    ottotime::otimePoint evtTime(evtTimeFromEpoch);

    cut_.setEventTime(evtTime);

    bool result{true};
    EXPECT_NO_THROW(result = cut_.cutAllLogs());
    EXPECT_FALSE(result);
}

/*!
 * @test CutFiles:EventTimeZeroOffset
 *
 * @brief test an error case, which the given event time is not valid
 *
 * This will require the user to search the resulting directory to see if the cuts are appropriate sized and other log
 * artifacts are copied, while omitting junk files.
 */
TEST_F(LogsCutterUtility_CutFiles, EventTimeZeroOffset)
{
    ottotime::otimeDuration beforeTime(ottotime::oSeconds(0.0));
    ottotime::otimeDuration afterTime(ottotime::oSeconds(0.0));

    cut_.setBeforeEventOffset(beforeTime);
    cut_.setAfterEventOffset(afterTime);

    bool result{true};
    EXPECT_NO_THROW(result = cut_.cutAllLogs());
    EXPECT_FALSE(result);
}
//----------------------------------------------------------------------------------------------------------------------

class LogsCutterUtility_CountFiles : public LogsCutterUtility_CutFiles
{};

/*!
 * @test CountFiles:GetFileCount
 *
 * @brief Checks to see if we are correctly counting the number of files in a log directory for status-assessment.
 */
TEST_F(LogsCutterUtility_CountFiles, GetFileCount)
{
    std::size_t count = 0;

    EXPECT_NO_THROW(count = cut_.howManyFiles());
    EXPECT_EQ(count, 26u);
}

/*!
 * @test CountFiles:GetZeroFileCount
 *
 * @brief Checks to see if we are correctly counting the number of files in a log directory for status-assessment.
 */
TEST_F(LogsCutterUtility_CountFiles, GetZeroFileCount)
{
    std::string testLogDir{testLogDir_};
    testLogDir += "unittest/noSuchLogPath";
    cut_.setLogDirectory(testLogDir);

    std::size_t count = 0;

    EXPECT_NO_THROW(count = cut_.howManyFiles());
    EXPECT_EQ(count, 0);
}

/*!
 * @test TimestampParser:TestParseSuccess
 *
 * @brief Checks that a line without a valid timestamp is parsed as such
 */
TEST(TestTimestampParser, TestParseSuccess)
{
    const std::string logLine("2001-11-12 18:31:01");
    EXPECT_NE(std::nullopt, LogsCutterUtility::parseTimestamp(logLine));
}

/*!
 * @test TimestampParser:TestParseFail
 *
 * @brief Checks that a line without a valid timestamp is parsed as such
 */
TEST(TestTimestampParser, TestParseFailed)
{
    const std::string logLine("This line does not have a timestamp");
    EXPECT_EQ(std::nullopt, LogsCutterUtility::parseTimestamp(logLine));
}

class TestTimestampParserRanges : public ::testing::TestWithParam<time_t>
{};

/*!
 * @test TimestampParserRanges:TestParser
 *
 * @brief Checks for correct parsing over various ranges of time
 */
TEST_P(TestTimestampParserRanges, TestParser)
{
    time_t timestamp = GetParam();
    std::string logLine(logger::LogEvent::formatTime(chr::system_clock::from_time_t(timestamp)));
    logLine += " | Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore "
               "et dolore magna aliqua.";

    EXPECT_EQ(timestamp, (LogsCutterUtility::parseTimestamp(logLine)->to_time_t()));
}

using namespace std::chrono_literals; // for min/h suffix support

INSTANTIATE_TEST_SUITE_P(ExerciseTimestampParser,
                         TestTimestampParserRanges,
                         ::testing::Range(chr::system_clock::to_time_t(chr::system_clock::now()),
                                          chr::system_clock::to_time_t(chr::system_clock::now() + 1min),
                                          10)); // 1 minute in 10 second increments

INSTANTIATE_TEST_SUITE_P(ExerciseTimestampParserHours,
                         TestTimestampParserRanges,
                         ::testing::Range(chr::system_clock::to_time_t(chr::system_clock::now()),
                                          chr::system_clock::to_time_t(chr::system_clock::now() + 24h),
                                          3600)); // 1 day in 1 hour increments

INSTANTIATE_TEST_SUITE_P(ExerciseTimestampParserDays,
                         TestTimestampParserRanges,
                         ::testing::Range(chr::system_clock::to_time_t(chr::system_clock::now()),
                                          chr::system_clock::to_time_t(chr::system_clock::now() + 168h),
                                          86400)); // 1 week in 1 day increments

INSTANTIATE_TEST_SUITE_P(ExerciseTimestampParserMonths,
                         TestTimestampParserRanges,
                         ::testing::Range(chr::system_clock::to_time_t(chr::system_clock::now()),
                                          chr::system_clock::to_time_t(chr::system_clock::now() + 720h),
                                          2592000)); // 1 year in 1 month increments

INSTANTIATE_TEST_SUITE_P(ExerciseTimestampParserYears,
                         TestTimestampParserRanges,
                         ::testing::Range(chr::system_clock::to_time_t(chr::system_clock::now()),
                                          chr::system_clock::to_time_t(chr::system_clock::now() + 87600h),
                                          31536000)); // 10 years in 1 year increments

//=== MAIN =============================================================================================================

/*!
 * @brief Main executable for LogsCutterTest's unit tests
 *
 * @param argc : Number of command-line arguments
 * @param argv : Array of c-strings
 * @return int : Executable's error code (should be zero)
 */
int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
