/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Definition of the MessageCache templated class.
 */
#pragma once

#include <task/CallbackSet.h>
#include <task/Task.h>

#include <chrono>
#include <functional>
#include <mutex>
#include <stdexcept>
#include <string>

namespace task {

/// @brief Templated class for storing the most recent input from a channel.
/// @note Before using this class, consider whether a callback-driven implementation would be cleaner than polling.
template<class Type>
class MessageCache
{
public:
    typedef std::function<std::chrono::nanoseconds(const Type&)> TimestampExtractor;

public:
    /// @brief Constructor.
    /// @param channel: The true name of the channel to receive input from.
    /// @param callbacks: The task's CallbackSet.
    MessageCache(std::string channel, CallbackSet& callbacks);
    // This class must be non-copyable, because only the original is registered with CallbackSet.
    MessageCache(const MessageCache&) = delete;
    MessageCache& operator=(const MessageCache&) = delete;

    /// @brief Set the threshold for "stale" messages.
    /// @param staleThreshold: How old a message is allowed to be without reporting an error.
    /// @param extractor: Function for extracting a timestamp from a message (if not given, use the current time).
    void setStaleThreshold(av::Duration staleThreshold, TimestampExtractor extractor = TimestampExtractor());

    /// @brief Check if any data has been received.
    /// @return True if any data has been received, whether or not it's stale.
    bool hasData() const;

    /// @brief Check if data has been received within the stale threshold window.
    /// @return True if the data's timestamp is within the stale threshold window (if any).
    bool hasRecentData() const;

    /// @brief Get the most recently received data.
    /// @return The most recent data.
    /// @throw A runtime_error if the data is stale, or if no data has been received.
    const Type& get();

    /// @brief Clear internal state to indicate no data is available.
    void reset();

    /// @brief callback method that sets received message
    void callback(const Type& data);
    inline std::chrono::nanoseconds getReceivedTime() { return m_receivedTime; }

private:
    // The name of the channel.
    const std::string m_channel;
    // How old the data is allowed to be.
    // TODO: Switch to using the new Time library here.
    std::chrono::nanoseconds m_staleThreshold{0};
    // Function for getting the timestamp from a message.
    TimestampExtractor m_timestampExtractor;
    // The timestamp of the most recent data.
    std::chrono::nanoseconds m_receivedTime{0};
    // The most recently received data.
    Type m_data;
    // This object should only ever be used in the callback thread, but just in case:
    mutable std::mutex m_mutex;
};

// Templated functions:

template<class Type>
MessageCache<Type>::MessageCache(std::string channel, CallbackSet& callbacks)
    : m_channel(channel)
{
    callbacks.add(channel, this, &MessageCache<Type>::callback);
    // TODO: Use the new Time library once it's available.
    m_timestampExtractor = [](const Type&) {
        return std::chrono::nanoseconds(task::Task::getSystemTime().toNanosecondsSinceEpoch());
    };
}

template<class Type>
void MessageCache<Type>::setStaleThreshold(av::Duration staleThreshold, TimestampExtractor extractor)
{
    m_staleThreshold = staleThreshold;
    if (extractor) {
        m_timestampExtractor = extractor;
    }
}

template<class Type>
bool MessageCache<Type>::hasData() const
{
    std::unique_lock<std::mutex> lock(m_mutex);
    return (m_receivedTime.count() != 0);
}

template<class Type>
bool MessageCache<Type>::hasRecentData() const
{
    // If no threshold was given, just check if data exists.
    if (!m_staleThreshold.count()) {
        return hasData();
    }
    std::chrono::nanoseconds now{task::Task::getSystemTime().toNanosecondsSinceEpoch()};
    std::unique_lock<std::mutex> lock(m_mutex);
    return (m_receivedTime > now - m_staleThreshold);
}

template<class Type>
const Type& MessageCache<Type>::get()
{
    if (!hasData()) {
        throw std::runtime_error("No messages have been received from channel \"" + m_channel + "\".");
    }
    if (!hasRecentData()) {
        throw std::runtime_error("Most recent message from channel \"" + m_channel + "\" is stale.");
    }
    // Note: Since get() should only be called in the callback thread, this mutex
    // exists only to satisfy the thread sanitizer in unit tests.
    std::unique_lock<std::mutex> lock(m_mutex);
    return m_data;
}

template<class Type>
void MessageCache<Type>::reset()
{
    std::unique_lock<std::mutex> lock(m_mutex);
    m_receivedTime = std::chrono::nanoseconds::zero();
}

template<class Type>
void MessageCache<Type>::callback(const Type& data)
{
    std::unique_lock<std::mutex> lock(m_mutex);
    m_data = data;
    m_receivedTime = m_timestampExtractor(data);
}

} // namespace task
