
#include "platform/middleware/libs/MessageCache/include/MessageCache.hpp"
#include <ODatum/ODatum.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <interfaces/SampleMessage/SampleMessage.h>
#include <task/CallbackSet.h>
#include <task/MessageOutput.h>
task::Task* task::Task::get(void)
{
    return nullptr;
}

class MockCallbackSet : public task::CallbackSet
{
public:
    MockCallbackSet()
        : CallbackSet(framework_)
    {}
    struct Message
    {
        std::string channel;
        std::string blob;
        std::chrono::nanoseconds timestamp;
        MessageKey key;
    };
    /// \brief Expect that sendMsg() will not be called at all in this test.
    void expectNoSendCalls() {}
    /// \brief Get the data that was sent most recently, deserialized into a given data type.
    template<class Type>
    Type getSentData() const
    {
        Type sentData;
        std::string body = data_.substr(sizeof(interfaces::baseTypes::InterfaceHeader));
        aptv::deserializeODatum(sentData, body);
        return sentData;
    }
    std::queue<Message> m_messages;

private:
    av::repro::EventFramework framework_;

    std::string data_;
};

class MessageCacheTest : public ::testing::Test
{
protected:
    void SetUp() override
    {
        payload.value = 0;
        header.headerVersion = htonl(OTTO_INTERFACE_HEADER_VERSION);
        aptv::serializeODatumMessage(std::string(reinterpret_cast<char*>(&header), sizeof(header)), blob, payload);
        m_mockCallbackSet.call("test", blob, {});
    }

    void TearDown() override {}

    MockCallbackSet m_mockCallbackSet{};
    std::string blob;
    av::msg::SampleMessage payload;
    interfaces::baseTypes::InterfaceHeader header;
};

TEST_F(MessageCacheTest, normalTest)
{
    task::MessageCache<av::msg::SampleMessage> test("test", m_mockCallbackSet);
    EXPECT_NO_THROW(test.reset());
    EXPECT_EQ(test.getReceivedTime(), std::chrono::nanoseconds::zero());
}

TEST_F(MessageCacheTest, setsalethresholdandhasDataTest)
{
    av::Duration staleThreshold = av::Duration{100};
    auto anyTimeStamp = std::chrono::nanoseconds{25};
    auto timeStampExtractor = [anyTimeStamp](const av::msg::SampleMessage&) { return anyTimeStamp; };
    task::MessageCache<av::msg::SampleMessage> test("test", m_mockCallbackSet);
    test.setStaleThreshold(staleThreshold, timeStampExtractor);
    EXPECT_EQ(false, test.hasData());
    test.callback(payload);
    EXPECT_EQ(true, test.hasData());
}

TEST_F(MessageCacheTest, hasRecentDataTest)
{
    av::Duration staleThreshold = av::Duration{100};
    auto anyTimeStamp = std::chrono::nanoseconds{25};
    auto timeStampExtractor = [anyTimeStamp](const av::msg::SampleMessage&) { return anyTimeStamp; };
    task::MessageCache<av::msg::SampleMessage> test("test", m_mockCallbackSet);
    test.setStaleThreshold(staleThreshold, timeStampExtractor);
    test.callback(payload);
    EXPECT_EQ(test.hasRecentData(), true);
}

TEST_F(MessageCacheTest, getTest_systemErrorAndPass)
{
    av::Duration staleThreshold = av::Duration{100};
    auto anyTimeStamp = std::chrono::nanoseconds{25};
    auto timeStampExtractor = [anyTimeStamp](const av::msg::SampleMessage&) { return anyTimeStamp; };
    task::MessageCache<av::msg::SampleMessage> test("test", m_mockCallbackSet);
    test.setStaleThreshold(staleThreshold, timeStampExtractor);
    EXPECT_THROW(test.get(), std::runtime_error);
    test.callback(payload);
    EXPECT_NO_THROW(test.get());
}

int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
