/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 *
 * NOTICE: Removing this heading is prohibited.
 *
 * All information contained herein is, and remains the property of
 * Motional, its affiliates, and its suppliers (if any). The
 * intellectual and technical concepts contained herein are
 * proprietary to Motional, its affiliates, and its suppliers (if any)
 * and may be covered by U.S. and Foreign Patents, patents in process,
 * and are protected by trade secrets or copyright law. Dissemination
 * of this information or reproduction of this material is strictly
 * forbidden unless prior written permission is obtained from
 * Motional.
 *
 * \file
 * \brief Implements the ChannelStatusMonitor class, which publishes
 * the status of each channel in the system.
 */
#include "lib/avlog/avlog.hpp"
#include <NetworkMetrics/ChannelStatusMonitor.hpp>
#include <interfaces/ChannelStatusMessage/ChannelStatusMessage.h>

namespace network {
namespace {
using ChannelStatus = av::msg::ChannelStatus;
using ChannelStatusMessage = av::msg::ChannelStatusMessage;

std::shared_ptr<ChannelStatusInterface> makeInterface(const InterfaceBuilder& interfaceBuilder, ConfigSection& config)
{
    std::shared_ptr<ChannelStatusInterface> res{};
    bool disabled{false};

    config.get("channelStatusDisabled", disabled, false);
    if (!disabled) {
        res = interfaceBuilder("ChannelStatusMessageOutput");
    }

    return res;
}
} // namespace

using namespace std::literals::chrono_literals;

ChannelStatusMonitor::ChannelConfig::ChannelConfig()
    : freqLow(0)
    , heartbeatInterval(1min)
{}

ChannelStatusMonitor::ChannelConfig::ChannelConfig(double _freqLow, double heartbeatIntervalSec)
    : freqLow(_freqLow)
    , heartbeatInterval(
          std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::duration<double>(heartbeatIntervalSec)))
{}

void ChannelStatusMonitor::ChannelState::record(av::TimePoint timestamp, bool messageReceived)
{
    const std::pair<av::TimePoint, bool> entry{timestamp, messageReceived};
    if (!timestampStart_) {
        timestamps_[0] = entry;
        bufferSize_ = 1;
        timestampStart_ = 0;
    } else {
        // If we get a tick and we've already gotten a message on this pass,
        // ignore the tick.
        if (!messageReceived) {
            if (gotMsgAfterLastHeartbeat_) {
                return;
            }
        }

        if (bufferSize_ < timestamps_.size()) {
            timestamps_[bufferSize_] = entry;
            ++bufferSize_;
        } else {
            if (timestamps_[*timestampStart_].second) {
                --timestampCount_;
            }
            timestamps_[*timestampStart_] = entry;
            timestampStart_ = ((*timestampStart_ + 1) % timestamps_.size());
        }
    }

    ++numEntries_;
    if (messageReceived) {
        gotMsgAfterLastHeartbeat_ = true;
        ++timestampCount_;
    }
}

std::optional<double> ChannelStatusMonitor::ChannelState::computeFrequency(av::TimePoint now) const
{
    if (numEntries_ < 1 || !timestampStart_) {
        return 0;
    }

    const av::TimePoint from = timestamps_[*timestampStart_].first;
    if (now == from) {
        return std::nullopt;
    }

    const double count{static_cast<double>(timestampCount_)};
    const av::Duration::rep nanoseconds{(now - from) / av::Duration(1)};
    const double seconds{static_cast<double>(nanoseconds) / 1000000000.0};
    return count / seconds;
}

ChannelStatusMonitor::ChannelState ChannelStatusMonitor::makeChannelState(const std::string& channelName)
{
    const auto it = configs_.find(channelName);
    ChannelConfig config = (it == configs_.end() ? ChannelConfig(1.0, 5.0) : it->second);
    return ChannelState(config);
}

ChannelStatusMonitor::ChannelState& ChannelStatusMonitor::getChannelState(const std::string& channelName)
{
    const auto it = monitored_.find(channelName);
    if (it == monitored_.end()) {
        ChannelState state{makeChannelState(channelName)};
        monitored_[channelName] = std::move(state);
        return monitored_[channelName];
    }

    return it->second;
}

void ChannelStatusMonitor::onMessage(const task::stats::MessageReport& report)
{
    if (senderChannelName_ == report.channelName) {
        return;
    }

    {
        const auto dur = report.timestamp.time_since_epoch();
        const auto zero = decltype(dur)::zero();
        if (dur <= zero) {
            return;
        }
    }

    switch (report.eventType) {
        case task::stats::MessageReport::EventType::Sent:
            break;

        case task::stats::MessageReport::EventType::Received: {
            std::unique_lock<std::mutex> lock(monitoredMutex_);
            getChannelState(report.channelName).record(report.timestamp, true);
        } break;

        case task::stats::MessageReport::EventType::Connected: {
            std::unique_lock<std::mutex> lock(monitoredMutex_);
            getChannelState(report.channelName);
        } break;

        case task::stats::MessageReport::EventType::Disconnected:
        default:
            break;
    }
}

ChannelStatusMonitor::ChannelStatusMonitor(const InterfaceBuilder& interfaceBuilder, ConfigSection& config)
    : discardSender_([]([[maybe_unused]] const av::msg::ChannelStatusMessage& msg) {})
    , registration_([this](const task::stats::MessageReport& report) { onMessage(report); })
{
    {
        auto interface = makeInterface(interfaceBuilder, config);
        if (interface && interface->hasChannel()) {
            publishSender_ = [interface](const av::msg::ChannelStatusMessage& msg) { interface->publish(msg); };
            sender_ = &publishSender_;
            senderChannelName_ = interface->channel().getName();
        } else {
            sender_ = &discardSender_;
            senderChannelName_ = {};
        }
    }

    ConfigSection channelStatusConfigs;
    if (config.get("channelStatusConfigs", channelStatusConfigs) == ConfigSection::GS_OK) {

        for (const auto& channelName : channelStatusConfigs.getParameterNames()) {
            ConfigSection channelConfig;
            if (channelStatusConfigs.get(channelName, channelConfig) == ConfigSection::GS_OK) {
                double freq{0};
                double heartbeatIntervalSec{0};
                if (channelConfig.get("freq", freq, 0.1) &&
                    channelConfig.get("heartbeatIntervalSec", heartbeatIntervalSec, 5.0)) {
                    ChannelConfig cfg(freq, heartbeatIntervalSec);
                    configs_[channelName] = cfg;
                }
            }
        }
    }
}

void ChannelStatusMonitor::tick(av::TimePoint now)
{
    std::vector<av::msg::ChannelStatusMessage> msgs;

    {
        std::unique_lock<std::mutex> lock(monitoredMutex_);
        for (auto& entry : monitored_) {
            ChannelState& state = entry.second;

            // Record the tick.
            state.record(now, false);

            const auto freq = state.computeFrequency(now);
            const auto newStatus = freq ? state.getConfig().getStatus(*freq) : av::msg::ChannelStatus::Up;

            if (state.hasStatusChanged(newStatus) || state.needsHeartbeat(now)) {
                av::msg::ChannelStatusMessage msg;
                msg.channelName = entry.first;
                msg.status = newStatus;
                msg.timestamp = now;

                msgs.push_back(msg);
                state.recordHeartbeat(now);
            }
        }
    }

    for (const auto& msg : msgs) {
        (*(sender_.load()))(msg);
    }
}

void ChannelStatusMonitor::stop()
{
    sender_ = &discardSender_;
}

} // namespace network
