#include "interfaces/NetworkMetricsMessage/NetworkMetricsMessage.h"
#include "lib/avlog/avlog.hpp"
#include <NetworkMetrics/NetworkStats.hpp>
#include <TimeStamp/otimeConversions.h>

using namespace std::chrono_literals;

namespace network {

std::optional<ConfigSection> channelMetrics(const ConfigSection& config, const std::string& channelName)
{
    ConfigSection channelMetrics;
    if (config.get(channelName, channelMetrics) == ConfigSection::GS_OK) {
        return channelMetrics;
    }
    return {};
}

template<typename T>
std::optional<T> metric(const std::string& metricName, const ConfigSection& config)
{
    T value{};
    if (config.get(metricName, value) == ConfigSection::GS_OK) {
        return value;
    }
    return {};
}

template<typename T>
T metric(const std::string& metricName, const ConfigSection& config, const T& defaultValue)
{
    auto result = metric<T>(metricName, config);
    if (!result) {
        return defaultValue;
    }
    return *result;
}

/// @brief Constructor with default windows.
NetworkStats::NetworkStats(av::TimePoint startTime, const ConfigSection& config)
    : NetworkStats(startTime, 1s, 1, config)
{}

/// @brief Constructor with specified windows.
NetworkStats::NetworkStats(av::TimePoint startTime, av::Duration intervalSize, size_t numWindows, ConfigSection config)
    : latencyWindows_([this](const RawStats& window,
                             av::TimePoint windowStart) { reportLatency(std::string{""}, window, windowStart, true); },
                      intervalSize,
                      numWindows,
                      startTime)
    , bytesWindows_([this](const RawStats& window,
                           av::TimePoint windowStart) { reportBandwidth(std::string{""}, window, windowStart, true); },
                    intervalSize,
                    numWindows,
                    startTime)
    , callbackRegistration_([this, startTime](const task::stats::MessageReport& mr) {
        switch (mr.eventType) {
            case task::stats::MessageReport::EventType::Sent:
                updateChannelStatsForSent(mr);
                break;

            case task::stats::MessageReport::EventType::Received:
                addLatencySample(mr.channelName, mr.latencyns.value(), mr.timestamp);
                addBandwidthSample(mr.channelName, mr.messageSize, mr.timestamp);
                break;

            case task::stats::MessageReport::EventType::Connected:
                if (channelToStats_.find(mr.channelName) == channelToStats_.end()) {
                    addChannel(mr.channelName, startTime);
                }
                break;

            case task::stats::MessageReport::EventType::Disconnected:
            default:
                break;
        }
    })
    , windowSize_(intervalSize * numWindows)
    , config_(std::move(config))
    , startTime_(startTime)
{}

bool NetworkStats::addChannel(const std::string& name, av::TimePoint startTime)
{
    // Checks that both optional inputs have values and hence can be compared.
    auto comparable = [](const auto& lhs, const auto& rhs) { return (lhs.has_value() && rhs.has_value()); };

    auto updateChannelLatency = [this, name, comparable](const RawStats& window, av::TimePoint windowStart) {
        auto windowMax = window.get<0>().getMax();
        auto windowJitter = window.get<1>().getJitter();

        // Only output a message if the stats violate any specified constraint.
        const auto maxJitter = channelToStats_[name].constraints.maxJitter;
        const auto maxLatency = channelToStats_[name].constraints.maxLatency;

        const auto constraintViolation = ((comparable(windowJitter, maxJitter) && *windowJitter > *maxJitter) ||
                                          (comparable(windowMax, maxLatency) && *windowMax > *maxLatency));

        const auto publish = channelToStats_[name].autoPublish || constraintViolation;
        reportLatency(name, window, windowStart, publish);
    };

    auto updateChannelBandwidth = [this, name, comparable](const RawStats& window, av::TimePoint windowStart) {
        auto windowMin = window.get<0>().getMin();
        auto windowMax = window.get<0>().getMax();

        // Only output a message if the stats violate any specified constraint.
        const auto minBandwidth = channelToStats_[name].constraints.minBandwidth;
        const auto maxBandwidth = channelToStats_[name].constraints.maxBandwidth;

        const auto constraintViolation = ((comparable(windowMin, minBandwidth) && *windowMin < *minBandwidth) ||
                                          (comparable(windowMin, maxBandwidth) && *windowMax > *maxBandwidth));

        const auto publish = channelToStats_[name].autoPublish || constraintViolation;
        reportBandwidth(name, window, windowStart, publish);
    };

    // Global channel configuration options.
    const auto autoPublishChannels = metric<bool>("autoPublish", config_, false);
    const auto channelWindowSizes = metric<size_t>("windowSize", config_, 10);
    const std::chrono::milliseconds channelIntervalSizes{metric<int>("intervalInMilliseconds", config_, 1000)};

    const auto metricsConfig = channelMetrics(config_, name);
    if (metricsConfig || autoPublishChannels) {
        WindowedStats latencyWindow{updateChannelLatency, channelIntervalSizes, channelWindowSizes, startTime};
        WindowedStats bandwidthWindow{updateChannelBandwidth, channelIntervalSizes, channelWindowSizes, startTime};
        ChannelStats stats(std::move(latencyWindow), std::move(bandwidthWindow), autoPublishChannels);
        if (metricsConfig) {
            stats.constraints.minBandwidth = metric<double>("minBandwidth", *metricsConfig);
            stats.constraints.maxBandwidth = metric<double>("maxBandwidth", *metricsConfig);
            stats.constraints.maxJitter = metric<double>("maxJitter", *metricsConfig);
            stats.constraints.maxLatency = metric<double>("maxLatency", *metricsConfig);
            stats.constraints.alertExpectedHz = metric<int>("alertExpectedHz", *metricsConfig);
            stats.autoPublish = stats.autoPublish || metric<bool>("autoPublish", *metricsConfig, false);
        }

        if (stats.constrained()) {
            channelToStats_.insert({name, std::move(stats)});
            AVLOG_DEBUG("Added metrics tracking for channel {} ", name);
            return true;
        }
    }
    return false;
}

/// @brief Add latency datapoint; delegates to latencyWindows_.
void NetworkStats::addLatencySample(const std::string& channel, av::Duration latency, av::TimePoint timestamp)
{
    std::scoped_lock<std::mutex> lock(mutex_);

    if (isTimestampValid(timestamp)) {
        latencyWindows_.addSample(static_cast<double>(latency.count()), timestamp);
        if (channelToStats_.find(channel) != channelToStats_.end()) {
            channelToStats_[channel].latencyWindows.addSample(static_cast<double>(latency.count()), timestamp);
        }
    } else {
        AVLOG_WARN("Received latency datapoint out of order for channel '{}': {} < {}",
                   channel,
                   timestamp.time_since_epoch().count(),
                   prevDatapointTimestamp_->time_since_epoch().count());
    }
}

/// @brief Add bandwidth datapoint; delegates to bytesWindows_.
void NetworkStats::addBandwidthSample(const std::string& channel, size_t nbytes, av::TimePoint timestamp)
{
    std::scoped_lock<std::mutex> lock(mutex_);

    if (isTimestampValid(timestamp)) {
        bytesWindows_.addSample(static_cast<double>(nbytes), timestamp);
        if (channelToStats_.find(channel) != channelToStats_.end()) {
            channelToStats_[channel].bytesWindows.addSample(static_cast<double>(nbytes), timestamp);
        }
    } else {
        AVLOG_WARN("Received bandwidth datapoint out of order for channel '{}': {} < {}",
                   channel,
                   timestamp.time_since_epoch().count(),
                   prevDatapointTimestamp_->time_since_epoch().count());
    }
}

void NetworkStats::updateChannelStatsForSent(const task::stats::MessageReport& mr)
{
    auto it = channelToStats_.find(mr.channelName);
    if ((it != channelToStats_.end()) && (it->second.constraints.alertExpectedHz)) {
        channelToStats_[mr.channelName].mostRecentSentTime = mr.timestamp;
    }
}

void NetworkStats::forEachStaleChannel(const StaleChannelHandler& f)
{
    using namespace std::chrono_literals;
    const auto now = av::timePointFromOtimePoint(SystemClock::getSystemTime());

    auto missedCycles = [](const auto& currentTime, const auto& channelStats) {
        if (!channelStats.constraints.alertExpectedHz.has_value()) {
            return 0;
        }
        const auto cycleDuration = 1000ms / *channelStats.constraints.alertExpectedHz;
        const auto threshold = currentTime - cycleDuration;
        if (channelStats.mostRecentSentTime < threshold) {
            return static_cast<int>((threshold - channelStats.mostRecentSentTime) / cycleDuration);
        }
        return 0;
    };

    for (const auto& [name, stats] : channelToStats_) {
        auto numCyclesMissed = missedCycles(now, stats);
        if (numCyclesMissed > 0 && stats.constraints.alertExpectedHz) {
            f(now, name, *stats.constraints.alertExpectedHz, numCyclesMissed);
        }
    }
}

} // namespace network
