#include <NetworkMetrics/NetworkStatsSender.hpp>

using namespace std::chrono_literals;

namespace network {
NetworkStatsSender::NetworkStatsSender(av::TimePoint startTime,
                                       const InterfaceBuilder& interfaceBuilder,
                                       const AlertInterfaceBuilder& alertInterfaceBuilder,
                                       std::string_view taskName,
                                       std::string_view hostName,
                                       const ConfigSection& taskConfig,
                                       int frequencyHz)
    : NetworkStatsSender(startTime,
                         1s,
                         1,
                         interfaceBuilder,
                         alertInterfaceBuilder,
                         taskName,
                         hostName,
                         taskConfig,
                         frequencyHz)
{}

NetworkStatsSender::NetworkStatsSender(av::TimePoint startTime,
                                       av::Duration intervalSize,
                                       size_t numWindows,
                                       const InterfaceBuilder& interfaceBuilder,
                                       const AlertInterfaceBuilder& alertInterfaceBuilder,
                                       std::string_view taskName,
                                       std::string_view hostName,
                                       const ConfigSection& taskConfig,
                                       int frequencyHz)
    : NetworkStats(startTime, intervalSize, numWindows, taskConfig)
    , interface_(interfaceBuilder("NetworkMetricsMessageOutput"))
    , alertInterface_(alertInterfaceBuilder("NetworkMetricsAlertMessageOutput"))
    , taskName_(taskName)
    , hostName_(hostName)
    , frequencyHz_(frequencyHz)
{}

bool NetworkStatsSender::PendingMetrics::validWindow(const size_t j, const av::TimePoint windowStart)
{
    if (pending_[j].second.latency.valid || pending_[j].second.bytes.valid) {
        av::TimePoint pendingWindowStart = pending_[j].second.windowStart;
        return (pendingWindowStart == windowStart);
    }

    return true;
}

size_t NetworkStatsSender::PendingMetrics::getIndex(const av::TimePoint windowStart)
{
    for (size_t i = 0; i < pendingCount_; ++i) {
        const size_t j = (i + pendingStartIndex_) % pending_.size();
        if (validWindow(j, windowStart)) {
            return j;
        }
    }

    if (pendingCount_ == pending_.size()) {
        pending_[pendingStartIndex_].second.bytes.valid = false;
        pending_[pendingStartIndex_].second.latency.valid = false;

        const size_t res = pendingStartIndex_;
        ++pendingStartIndex_;
        pendingStartIndex_ %= pending_.size();
        return res;
    }

    return (pendingStartIndex_ + pendingCount_++) % pending_.size();
}

std::pair<bool, av::msg::NetworkMetricsMessage>& NetworkStatsSender::PendingMetrics::initPending(
    const size_t index,
    const NetworkStatsSender::RawStats& window,
    const av::TimePoint windowStart,
    av::Duration windowSize)
{
    auto& p = pending_[index];
    p.second.windowStart = windowStart;
    p.second.windowLength = windowSize;
    p.second.count = window.get<1>().getCount();
    return p;
}

void NetworkStatsSender::reportLatency(const std::string& channel,
                                       const NetworkStatsSender::RawStats& window,
                                       const av::TimePoint windowStart,
                                       bool forceReport)
{
    auto& pendingMetrics = pending_[channel];
    const size_t index = pendingMetrics.getIndex(windowStart);

    auto& [report, p] = pendingMetrics.initPending(index, window, windowStart, getWindowSize());
    p.taskName = taskName_;
    p.hostName = hostName_;
    p.channelName = channel;

    p.latency.min = window.get<0>().getMin().value_or(-1);
    p.latency.max = window.get<0>().getMax().value_or(-1);
    p.latency.mean = window.get<1>().getMean().value_or(-1);
    p.latency.variance = window.get<1>().getVariance().value_or(-1);
    p.latency.jitter = window.get<1>().getJitter().value_or(-1);
    p.latency.valid = true;

    report = report || forceReport;
    if (report) {
        report = !reportMessage(p);
    }
}

void NetworkStatsSender::reportBandwidth(const std::string& channel,
                                         const NetworkStatsSender::RawStats& window,
                                         av::TimePoint windowStart,
                                         bool forceReport)
{
    auto& pendingMetrics = pending_[channel];
    const size_t index = pendingMetrics.getIndex(windowStart);

    auto& [report, p] = pendingMetrics.initPending(index, window, windowStart, getWindowSize());
    p.taskName = taskName_;
    p.hostName = hostName_;
    p.channelName = channel;

    p.bytes.min = window.get<0>().getMin().value_or(-1);
    p.bytes.max = window.get<0>().getMax().value_or(-1);
    p.bytes.mean = window.get<1>().getMean().value_or(-1);
    p.bytes.variance = window.get<1>().getVariance().value_or(-1);
    p.bytes.jitter = -1;
    p.bytes.valid = true;

    report = report || forceReport;
    if (report) {
        report = !reportMessage(p);
    }
}

bool NetworkStatsSender::reportMessage(av::msg::NetworkMetricsMessage& message)
{
    if (message.bytes.valid && message.latency.valid) {
        sendStats(message);
        return true;
    }
    return false;
}

void NetworkStatsSender::sendStats(av::msg::NetworkMetricsMessage& message)
{
    if (interface_) {
        interface_->publish(message);
        message.latency.valid = false;
        message.bytes.valid = false;
    }
}

void NetworkStatsSender::reportMissedCycle(av::TimePoint timeDetected,
                                           std::string_view channelName,
                                           int expectedFrequencyHz,
                                           int numCyclesMissed)
{
    if (alertInterface_ != nullptr) {
        av::msg::NetworkMetricsAlertMessage alert;
        alert.taskName = taskName_;
        alert.hostName = hostName_;
        alert.channelName = channelName;
        alert.timeDetected = timeDetected;
        alert.category = av::msg::NetworkMetricsAlertMessageCategory::AlertMessagesNotSent;
        alert.expectedFrequencyHz = expectedFrequencyHz;
        alert.numCyclesMissed = numCyclesMissed;
        alertInterface_->publish(alert);
    }
}

void NetworkStatsSender::onTick()
{
    if (alertInterface_ != nullptr) {
        forEachStaleChannel([this](av::TimePoint timeDetected,
                                   std::string_view channelName,
                                   int expectedFrequencyHz,
                                   int numCyclesMissed) {
            this->reportMissedCycle(timeDetected, channelName, expectedFrequencyHz, numCyclesMissed);
        });
    }
}

} // namespace network
