/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 *
 * NOTICE: Removing this heading is prohibited.
 *
 * All information contained herein is, and remains the property of
 * Motional, its affiliates, and its suppliers (if any). The
 * intellectual and technical concepts contained herein are
 * proprietary to Motional, its affiliates, and its suppliers (if any)
 * and may be covered by U.S. and Foreign Patents, patents in process,
 * and are protected by trade secrets or copyright law. Dissemination
 * of this information or reproduction of this material is strictly
 * forbidden unless prior written permission is obtained from
 * Motional.
 *
 * \file
 * \brief Declares the ChannelStatusMonitor class, which publishes
 * the status of each channel in the system.
 */

#ifndef CHANNEL_STATUS_MONITOR_H
#define CHANNEL_STATUS_MONITOR_H

#include <Comms/Channel.h>
#include <avtime/Time.hpp>
#include <interfaces/ChannelStatusMessage/ChannelStatusMessage.h>
#include <interfaces/Generic/Output/Channel/Impl.h>

#include <mutex>
#include <optional>

namespace network {
using ChannelStatusInterface = GenericOutputChannel<av::msg::ChannelStatusMessage>;
using InterfaceBuilder = std::function<std::shared_ptr<ChannelStatusInterface>(const std::string& interfaceName)>;

class ChannelStatusMonitor
{
public:
    /// @brief Constructor.
    ///
    /// @param interfaceBuilder Given an interface name, returns a
    /// pointer to the interface (in Write mode).
    ///
    /// @param config The section in which to search for the config data.
    ChannelStatusMonitor(const InterfaceBuilder& interfaceBuilder, ConfigSection& config);

    /// @brief Call this regularly to produce the status reports.
    void tick(av::TimePoint now);

    /// @brief Stops all activity; tick() and onMessage() will have no effect.
    void stop();

private:
    /// @brief The callback when a message is received.
    void onMessage(const task::stats::MessageReport& report);

    /// @brief Stores the configuration for a channel; computes current status.
    struct ChannelConfig
    {
        ChannelConfig();
        ChannelConfig(double freqLow, double heartbeatIntervalSec);

        ChannelConfig(const ChannelConfig& other) = default;
        ChannelConfig(ChannelConfig&& other) = default;
        ChannelConfig& operator=(const ChannelConfig& other) = default;
        ChannelConfig& operator=(ChannelConfig&& other) = default;

        /// @brief Given the current frequency of messages, compute status.
        ///
        /// @param curFreq The current frequency of messages.
        /// @return The channel status: up (frequency above the
        /// threshold), down (frequency 0), or degraded (in between).
        av::msg::ChannelStatus getStatus(double curFreq) const
        {
            if (curFreq >= freqLow) {
                return av::msg::ChannelStatus::Up;
            }

            if (curFreq <= 0) {
                return av::msg::ChannelStatus::Down;
            }

            return av::msg::ChannelStatus::Degraded;
        }

        double freqLow;
        av::Duration heartbeatInterval;
    };

    /// @brief Tracks the frequency of messages.
    class ChannelState
    {
    public:
        ChannelState()
            : config_()
            , lastHeartbeatTime_()
            , timestamps_(5)
        {}

        ChannelState(const ChannelConfig& config)
            : config_(config)
            , lastHeartbeatTime_()
            , timestamps_(5)
        {}

        ChannelState(ChannelConfig&& config)
            : config_(config)
            , lastHeartbeatTime_()
            , timestamps_(5)
        {}

        ChannelState(const ChannelState& other) = default;
        ChannelState(ChannelState&& other) = default;
        ChannelState& operator=(const ChannelState& other) = default;
        ChannelState& operator=(ChannelState&& other) = default;

        /// @brief Record that a message was received.
        /// @param timestamp The time the message was received.
        void record(av::TimePoint timestamp, bool messageReceived);

        /// @brief Record that time has passed.
        ///
        /// @param now The current time. If it is greater than the
        /// most recent message-received timestamp, an empty timestamp
        /// will be added to the log.
        void recordTick(av::TimePoint now);

        /// @brief Compute the observed frequency.
        ///
        /// @param now The current time; necessary to know the length
        /// of the time represented by the window.
        ///
        /// @return The message frequency, or std::nullopt if now is equal
        /// to the earliest timestamp (which would be an infinite
        /// frequency).
        std::optional<double> computeFrequency(av::TimePoint now) const;

        /// @return The configuration of the channel.
        const ChannelConfig& getConfig() const { return config_; }

        /// @brief Has the status changed since the last time we asked?
        ///
        /// Updates the "previous status" to newStatus.
        ///
        /// @param newStatus The current status.
        ///
        /// @return True if the new status is different from the
        /// previous one, false otherwise.
        bool hasStatusChanged(av::msg::ChannelStatus newStatus)
        {
            const bool res = (!lastStatus) || ((*lastStatus) != newStatus);
            lastStatus = newStatus;
            return res;
        }

        /// @brief Have we overrun the heartbeat interval?
        ///
        /// @param now The current time.
        ///
        /// @return True if the time between the last heartbeat and
        /// now is at least equal to heartbeatInterval, from the
        /// channels' config.
        bool needsHeartbeat(av::TimePoint now)
        {
            return (needsFirstHeartbeat_ || ((now - lastHeartbeatTime_) >= config_.heartbeatInterval));
        }

        /// @brief Record that a heartbeat was sent.
        ///
        /// @param now The time the heartbeat was sent.
        void recordHeartbeat(av::TimePoint now)
        {
            needsFirstHeartbeat_ = false;
            lastHeartbeatTime_ = now;
            gotMsgAfterLastHeartbeat_ = false;
        }

    private:
        ChannelConfig config_;
        bool needsFirstHeartbeat_{true};
        av::TimePoint lastHeartbeatTime_;
        bool gotMsgAfterLastHeartbeat_{false};

        std::optional<av::msg::ChannelStatus> lastStatus{};

        // A circular buffer, arranged as follows:
        //
        // * timestamps_ is a vector (fixed at 5 elements). Its
        //   elements are a time and a flag; if the flag is false, the
        //   entry is uninitialized. (I suppose this could be
        //   optional<TimePoint>.)
        //
        // * bufferSize_ is the number of elements in timestamps_.
        //
        // * timestampCount_ is the number of elements in timestamps_
        //   which actually represent messages, as opposed to just
        //   ticks.
        //
        // * numEntries_ is the number of ticks we've received, even
        //   if they've been discarded when the circular buffer got
        //   full.
        //
        // * timestampStart_ is the index into timestamps_ of the
        //   earliest entry.
        std::vector<std::pair<av::TimePoint, bool>> timestamps_{};
        size_t bufferSize_{0};
        size_t timestampCount_{0};
        size_t numEntries_{0};
        std::optional<size_t> timestampStart_{};
    };

    ChannelState makeChannelState(const std::string& channelName);

    /// Does not get monitoredMutex_, because it's called from a function
    /// that holds that mutex.
    ChannelState& getChannelState(const std::string& channelName);

    using Sender = std::function<void(const av::msg::ChannelStatusMessage& msg)>;
    std::atomic<Sender*> sender_;
    Sender publishSender_;
    Sender discardSender_;

    std::optional<std::string> senderChannelName_{};

    std::unordered_map<std::string, ChannelState> monitored_;
    std::unordered_map<std::string, ChannelConfig> configs_;
    task::Channel::CallbackRegistration registration_;

    // Mutex to guard everything except sender_.
    std::mutex monitoredMutex_{};
};
} // namespace network

#endif // CHANNEL_STATUS_MONITOR_H
