/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Gathers statistics on network traffic.
 */

#ifndef NETWORK_STATS_HPP
#define NETWORK_STATS_HPP

#include <Comms/Channel.h>
#include <Statistics/MeanVariance.hpp>
#include <Statistics/MinMax.hpp>
#include <Statistics/SeriesStatisticsInterface.hpp>
#include <Statistics/TimeWindowedSeriesStatistics.hpp>
#include <Statistics/TupleStatistics.hpp>
#include <avtime/Time.hpp>

#include <mutex>
#include <unordered_map>

class ConfigSection;

namespace network {

/// @brief Collects statistics on network traffic.
///
/// Data is collected via a TimeWindowedSeriesStatistics of a
/// TupleStatistics, gathering min/max and mean/variance.  Results are
/// reported via two protected pure-virtual methods, reportLatency()
/// and reportBandwidth().
///
/// Data is reported after every interval, size specified by
/// intervalSize. numWindows is the number of intervals whose data is
/// collected. So we're computing data in a sliding window. If (as is
/// the default) intervalSize is 1min, and numWindows is 5, then, every
/// minute, data will be reported about the past 5 minutes.
class NetworkStats
{
public:
    /// @brief Constructor.
    ///
    /// intervalSize defaults to 1min, numWindows defaults to 5.
    ///
    /// @param startTime The start time of the first reporting window.
    /// @param config Metrics configuration section.
    NetworkStats(av::TimePoint startTime, const ConfigSection& config);

    /// @brief Constructor.
    ///
    /// @param startTime The start time of the first reporting window.
    ///
    /// @param intervalSize The length of each reporting interval.
    ///
    /// @param numWindows The number of windows to compute. Windows
    /// overlap; each has a length of intervalSize * numWindows, and
    /// they start every intervalSize.
    NetworkStats(av::TimePoint startTime, av::Duration intervalSize, size_t numWindows, ConfigSection config);

    /// @brief Returns the size of each window.
    ///
    /// @return intervalSize * numWindows
    av::Duration getWindowSize() const { return windowSize_; }

    /// @brief Virtual destructor, because C++ can't figure it out for us.
    virtual ~NetworkStats() = default;

    /// @brief Report the latency that it took a message to arrive.
    ///
    /// Datapoints must be delivered in monotonically nondecreasing
    /// order (i.e., if x comes after y, x >= y). If a datapoint is
    /// received out of order, a warning will be logged, and the
    /// datapoint will be dropped.
    /// @param channelName Channel associated with the latency data, or "" for task latency data.
    /// @param latency The duration between the time the message was
    /// sent and the time it was received.
    /// @param timestamp The time the message was received.
    void addLatencySample(const std::string& channelName, av::Duration latency, av::TimePoint timestamp);

    /// @brief Report the size of a received message.
    ///
    /// Datapoints must be delivered in monotonically nondecreasing
    /// order (i.e., if x comes after y, x >= y). If a datapoint is
    /// received out of order, a warning will be logged, and the
    /// datapoint will be dropped.
    /// @param channelName Channel associated with the bandwidth data, or "" for task bandwidth data.
    /// @param nbytes The size of the message, in bytes.
    ///
    /// @param timestamp The time the message was received.
    void addBandwidthSample(const std::string& channelName, size_t nbytes, av::TimePoint timestamp);

    /// @brief For test purposes: reports data for all remaining windows.
    void callInterimCallbacks()
    {
        std::scoped_lock<std::mutex> lock(mutex_);
        latencyWindows_.callInterimCallbacks();
        bytesWindows_.callInterimCallbacks();
    }

protected:
    /// @brief An underlying SeriesStatisticsInterface for each window.
    using RawStats = statistics::TupleStatistics<statistics::MinMax<double>, statistics::MeanVariance<double>>;

    /// @brief Tracks the statistics for the various windows.
    using WindowedStats = statistics::TimeWindowedSeriesStatistics<double, RawStats, av::AvClock>;

    /// Queue latency data for publishing.
    /// @param channel The channel associated with the data.
    /// @param window The data.
    /// @param windowStart Beginning time point of the above window.
    /// @param forceReport When true, indicates the data shall be published once all necessary data
    /// is gathered (Publishing may be delayed until bandwidth data for the given window is available).
    virtual void reportLatency(const std::string& channel,
                               const RawStats& window,
                               av::TimePoint windowStart,
                               bool forceReport) = 0;

    /// Queue bandwidth data for publishing.
    /// @param channel The channel associated with the data.
    /// @param window The data.
    /// @param windowStart Beginning time point of the above window.
    /// @param forceReport When true, indicates the data shall be published once all necessary data
    /// is gathered (Publishing may be delayed until bandwidth data for the given window is available).
    virtual void reportBandwidth(const std::string& channel,
                                 const RawStats& window,
                                 av::TimePoint windowStart,
                                 bool forceReport) = 0;

    /// Add the given channel to the list of channels to track metrics for.
    /// @param name The channel name.
    /// @param startTime Time point of the beginning of the first window.
    bool addChannel(const std::string& name, av::TimePoint startTime);

private:
    std::mutex mutex_{};

    /// @brief Tracks the latency statistics for each window.
    WindowedStats latencyWindows_;

    /// @brief Tracks the bandwidth statistics for each window.
    WindowedStats bytesWindows_;

    /// @brief Registers this object with the message callback system.
    task::Channel::CallbackRegistration callbackRegistration_;

    std::optional<av::TimePoint> prevDatapointTimestamp_{};

    const av::Duration windowSize_;

    // Assumes mutex_ is locked.
    bool isTimestampValid(av::TimePoint t)
    {
        if (prevDatapointTimestamp_ && (t < (*prevDatapointTimestamp_))) {
            return false;
        } else {
            prevDatapointTimestamp_ = t;
            return true;
        }
    }

    /// @brief Per-channel statistics.
    struct ChannelStats
    {
        // Default constructor. Needed for the map.
        ChannelStats()
            : mostRecentSentTime()
            , latencyWindows([](auto&, auto) {}, av::Duration{}, 0, av::TimePoint{})
            , bytesWindows([](auto&, auto) {}, av::Duration{}, 0, av::TimePoint{})
        {}

        ChannelStats(WindowedStats latencyWindows_, WindowedStats bytesWindows_, bool autoPublish_)
            : autoPublish{autoPublish_}
            , mostRecentSentTime()
            , latencyWindows(std::move(latencyWindows_))
            , bytesWindows(std::move(bytesWindows_))
        {}

        bool autoPublish{false};
        av::TimePoint mostRecentSentTime;
        WindowedStats latencyWindows;
        WindowedStats bytesWindows;

        struct
        {
            std::optional<int> alertExpectedHz;
            std::optional<double> minBandwidth;
            std::optional<double> maxBandwidth;
            std::optional<double> maxJitter;
            std::optional<double> maxLatency;
        } constraints;

        bool constrained() const
        {
            return autoPublish || constraints.alertExpectedHz || constraints.minBandwidth || constraints.maxBandwidth ||
                   constraints.maxJitter || constraints.maxLatency;
        }
    };

    const ConfigSection config_;
    /// @brief Tracks the per-channel statistics.
    std::unordered_map<std::string, ChannelStats> channelToStats_{};

    /// @brief Given a sent message, update channelToStats_.
    void updateChannelStatsForSent(const task::stats::MessageReport& mr);
    const av::TimePoint startTime_;

protected:
    /// @param timeDetected The time at which the problem was
    /// detected.
    ///
    /// @param channelName The name of the stale channel.
    ///
    /// @param expectedFrequencyHz How often the channel is expected
    /// to send.
    ///
    /// @param numCyclesMissed The number of cycles since we've seen a
    /// message sent on this channel.
    using StaleChannelHandler = std::function<
        void(av::TimePoint timeDetected, std::string_view channelName, int expectedFrequencyHz, int numCyclesMissed)>;

    /// @brief Loop over the stale channels, call f() for each one.
    ///
    /// A channel is considered stale if the duration since the latest
    /// sent message is at least 3 cycles, where a cycle length is
    /// 1/taskFrequencyHz.
    ///
    /// @param f A function invoked for each stale channel.
    void forEachStaleChannel(const StaleChannelHandler& f);
};

} // namespace network

#endif // NETWORK_STATS_HPP
