/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief A NetworkStats which publishes its statistics on an interface.
 */

#ifndef NETWORK_STATS_SENDER_HPP
#define NETWORK_STATS_SENDER_HPP

#include <NetworkMetrics/NetworkStats.hpp>
#include <TimeStamp/otimeConversions.h>
#include <interfaces/Generic/Output/Channel/Impl.h>
#include <interfaces/NetworkMetricsAlertMessage/NetworkMetricsAlertMessage.h>
#include <interfaces/NetworkMetricsMessage/NetworkMetricsMessage.h>

#include <functional>

namespace network {

/// @brief Publishes statistics on network traffic.
class NetworkStatsSender : public NetworkStats
{
public:
    using Interface = GenericOutputChannel<av::msg::NetworkMetricsMessage>;
    using InterfaceBuilder = std::function<std::shared_ptr<Interface>(const std::string& interfaceName)>;
    using AlertInterface = GenericOutputChannel<av::msg::NetworkMetricsAlertMessage>;
    using AlertInterfaceBuilder = std::function<std::shared_ptr<AlertInterface>(const std::string& interfaceName)>;

    /// @brief Constructor.
    ///
    /// @param startTime The start time of the first reporting window.
    /// @param interfaceBuilder A function to produce the necessary interface.
    /// @param alertInterfaceBuilder A function to produce the necessary interface.
    /// @param taskName name of the current task (used when reporting statistics).
    /// @param hostName Name of the current host (used when reporting statistics).
    /// @param config The metrics configuration.
    /// @param frequencyHz The execution rate of the current task.
    NetworkStatsSender(ottotime::otimePoint startTime,
                       const InterfaceBuilder& interfaceBuilder,
                       const AlertInterfaceBuilder& alertInterfaceBuilder,
                       std::string_view taskName,
                       std::string_view hostName,
                       const ConfigSection& config,
                       int frequencyHz)
        : NetworkStatsSender(av::timePointFromOtimePoint(startTime),
                             interfaceBuilder,
                             alertInterfaceBuilder,
                             taskName,
                             hostName,
                             config,
                             frequencyHz)
    {}

    /// @brief Constructor.
    ///
    /// @param startTime The start time of the first reporting window.
    /// @param interfaceBuilder A function to produce the necessary interface.
    /// @param alertInterfaceBuilder A function to produce the necessary interface.
    /// @param taskName name of the current task (used when reporting statistics).
    /// @param hostName Name of the current host (used when reporting statistics).
    /// @param config The metrics configuration.
    /// @param frequencyHz The execution rate of the current task.
    NetworkStatsSender(av::TimePoint startTime,
                       const InterfaceBuilder& interfaceBuilder,
                       const AlertInterfaceBuilder& alertInterfaceBuilder,
                       std::string_view taskName,
                       std::string_view hostName,
                       const ConfigSection& config,
                       int frequencyHz);

    /// @brief Constructor.
    ///
    /// @param startTime The start time of the first reporting window.
    ///
    /// @param intervalSize The length of each reporting interval.
    ///
    /// @param numWindows The number of windows to compute. Windows
    /// overlap; each has a length of intervalSize * numWindows, and
    /// they start every intervalSize.
    ///
    /// @param interfaceBuilder A function to produce the necessary interface.
    NetworkStatsSender(av::TimePoint startTime,
                       av::Duration intervalSize,
                       size_t numWindows,
                       const InterfaceBuilder& interfaceBuilder,
                       const AlertInterfaceBuilder& alertInterfaceBuilder,
                       std::string_view taskName,
                       std::string_view hostName,
                       const ConfigSection& config,
                       int frequencyHz);

    void callInterimCallbacks()
    {
        NetworkStats::callInterimCallbacks();
        for (auto& [unused, pendingChannel] : pending_) {
            for (auto& [unused, metrics] : pendingChannel.allPending()) {
                if (metrics.latency.valid || metrics.bytes.valid) {
                    sendStats(metrics);
                }
            }
        }
    }

    /// @brief Performs logic that needs to be done on every tick.
    ///
    /// In the legacy task framework, "on every tick" means every call
    /// to executive(). TODO: FASTER will need to do this differently,
    /// perhaps with a separate thread that calls onTick() every N
    /// ms. If we take that approach, we'll have to make onTick()
    /// thread-safe.
    void onTick();

protected:
    /// @brief Notifies when a window of latency data is available.
    ///
    /// Data will be buffered, so that we send out one
    /// NetworkMetricsMessage per each latency/bandwidth window pair.
    /// @param channel The channel associated with this latency data, or "" for task latency data.
    /// @param window The SeriesStatisticsInterface with the latency stats.
    /// @param windowStart The time at which the window started.
    /// @param forceReport True to publish when corresponding latency data is available, false otherwise.
    void reportLatency(const std::string& channel,
                       const RawStats& window,
                       av::TimePoint windowStart,
                       bool forceReport) override;

    /// @brief Notifies when a window of latency data is available.
    ///
    /// Data will be buffered, so that we send out one
    /// NetworkMetricsMessage per each latency/bandwidth window pair.
    ///
    /// @param window The SeriesStatisticsInterface with the latency stats.
    ///
    /// @param windowStart The time at which the window started.
    bool reportMessage(av::msg::NetworkMetricsMessage& message);

    /// @brief Notifies when a window of bandwidth data is available.
    ///
    /// Data will be buffered, so that we send out one
    /// NetworkMetricsMessage per each latency/bandwidth window pair.
    /// @param channel The channel associated with this bandwidth data, or "" for task bandwidth data.
    /// @param window The SeriesStatisticsInterface with the latency stats.
    /// @param windowStart The time at which the window started.
    /// @param forceReport True to publish when corresponding bandwidth data is available, false otherwise.
    void reportBandwidth(const std::string& channel,
                         const RawStats& window,
                         av::TimePoint windowStart,
                         bool forceReport) override;

    bool validWindow(const size_t j, const av::TimePoint windowStart);
    size_t getIndex(const av::TimePoint windowStart);

    /// Publish the specified @ref NetworkMetricsMessage
    /// @param message Message to publish.
    void sendStats(av::msg::NetworkMetricsMessage& message);

    /// @brief Send an alert that we aren't sending on a channel.
    void reportMissedCycle(av::TimePoint timeDetected,
                           std::string_view channelName,
                           int expectedFrequencyHz,
                           int numCyclesMissed);

    struct PendingMetrics
    {
        PendingMetrics()
            : pending_(100, std::make_pair(false, av::msg::NetworkMetricsMessage()))
        {}

        std::pair<bool, av::msg::NetworkMetricsMessage>& initPending(const size_t index,
                                                                     const NetworkStatsSender::RawStats& window,
                                                                     const av::TimePoint windowStart,
                                                                     av::Duration windowSize);

        bool validWindow(const size_t j, const av::TimePoint windowStart);
        size_t getIndex(const av::TimePoint windowStart);

        std::vector<std::pair<bool, av::msg::NetworkMetricsMessage>>& allPending() { return pending_; }

    private:
        std::vector<std::pair<bool, av::msg::NetworkMetricsMessage>> pending_;
        size_t pendingStartIndex_{0};
        size_t pendingCount_{0};
    };

    // Key: channel name ("" for task level metrics)
    std::unordered_map<std::string, PendingMetrics> pending_;
    std::shared_ptr<Interface> interface_;
    std::shared_ptr<AlertInterface> alertInterface_;
    const std::string taskName_;
    const std::string hostName_;
    const int frequencyHz_;
};

} // namespace network

#endif // NETWORK_STATS_SENDER_HPP
