/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 *
 * NOTICE: Removing this heading is prohibited.
 *
 * All information contained herein is, and remains the property of
 * Motional, its affiliates, and its suppliers (if any). The
 * intellectual and technical concepts contained herein are
 * proprietary to Motional, its affiliates, and its suppliers (if any)
 * and may be covered by U.S. and Foreign Patents, patents in process,
 * and are protected by trade secrets or copyright law. Dissemination
 * of this information or reproduction of this material is strictly
 * forbidden unless prior written permission is obtained from
 * Motional.
 *
 * \file
 * \brief Unit tests for ChannelStatusMonitor.
 */

#include <gtest/gtest.h>

#include "DDSConfiguratorHelper.hpp"
#include "platform/middleware/libs/DDSComms/test/DDSTestPathsHelper.hpp"
#include <Comms/InterfaceUtils.h>
#include <NetworkMetrics/ChannelStatusMonitor.hpp>
#include <TimeStamp/otime.h>
#include <chrono>
#include <dds/dds.hpp>
#include <future>
#include <interfaces/Generic/Input/Channel/Impl.h>
#include <memory>
#include <thread>
#include <vector>

#include "CustomClock.hpp"

namespace {
using namespace std::chrono_literals;
using DomainParticipant = dds::domain::DomainParticipant;
using BytesTopicType = DDSComms::BytesTopicType;
using BytesTopic = dds::topic::Topic<BytesTopicType>;
using Publisher = dds::pub::Publisher;
using Subscriber = dds::sub::Subscriber;
using BytesWriter = dds::pub::DataWriter<BytesTopicType>;
using BytesReader = dds::sub::DataReader<BytesTopicType>;

DomainParticipant dp{0}; /// allows for reuse of DP and faster tests
using testing::CustomClock;

template<typename T>
std::optional<T> getWithTimeout(std::future<T>& f)
{
    const auto status = f.wait_for(10s);
    if (status == std::future_status::ready) {
        return f.get();
    }

    return std::nullopt;
}

class ChannelStatusMonitorTest : public testing::Test
{
public:
    ChannelStatusMonitorTest() {}

    void SetUp() override
    {
        clock_.initCurrentTime();
        SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeCustomClock, &clock_);
        auto configPtr{std::make_shared<Config>()};
        configPtr->readString(testChannelStatusConfig);

        ConfigSection configTest{};
        ASSERT_TRUE(configTest.init("testChannelStatus", configPtr));

        ConfigSection interfaceConfig{};
        ASSERT_TRUE(interfaceConfig.init("monitoring", configPtr));

        ConfigSection monitorInterfaceConfig{};
        ASSERT_EQ(interfaceConfig.get("ChannelStatusMessageOutput", monitorInterfaceConfig), ConfigSection::GS_OK);

        ConfigSection monitorConfig{};
        ASSERT_EQ(monitorInterfaceConfig.get("channel", monitorConfig), ConfigSection::GS_OK);

        std::shared_ptr<network::ChannelStatusInterface> channelStatusMonitorInterface =
            comms::getInterface<network::ChannelStatusInterface>(
                "ChannelStatusMessageOutput", monitorInterfaceConfig, task::Access::Write);
        monitorReadInterface_ = comms::getInterface<GenericInputChannel<av::msg::ChannelStatusMessage>>(
            "ChannelStatusMessageOutput", monitorInterfaceConfig, task::Access::Read);
        readChannel_ = task::ChannelFactory::createChannelManaged(configTest, task::Access::Read);

        network::InterfaceBuilder builder([&channelStatusMonitorInterface](const std::string& interfaceName) {
            if (interfaceName == "ChannelStatusMessageOutput") {
                return channelStatusMonitorInterface;
            } else {
                return std::shared_ptr<network::ChannelStatusInterface>(nullptr);
            }
        });
        channelStatusMonitor_ = std::make_unique<network::ChannelStatusMonitor>(builder, interfaceConfig);

        readChannel_->connect();
        while (bWriter_.publication_matched_status().current_count() == 0) {
            std::this_thread::sleep_for(100ms);
        }
    }

    void TearDown() override { readChannel_->disconnect(); }

protected:
    DomainParticipant dp_{dp};
    DDSComms::DDSConfiguratorType ddsConfig{{DDSComms::testing::configFile}};
    BytesTopic topic_{dp_, "testChannelStatus"};
    Publisher publisher_{dp_};
    BytesWriter bWriter_{publisher_, topic_, ddsConfig.createWriterQos()};

    std::unique_ptr<task::Channel> readChannel_;
    std::shared_ptr<GenericInputChannel<av::msg::ChannelStatusMessage>> monitorReadInterface_;
    std::unique_ptr<network::ChannelStatusMonitor> channelStatusMonitor_;

    const std::string testChannelStatusConfig = fmt::format("testChannelStatus ="
                                                            "{{"
                                                            "name = \"testChannelStatus\";"
                                                            "type = \"DDS\";"
                                                            "ddsConfigFile = \"{}\";"
                                                            "queueSize = 100LL;"
                                                            "domainId = 0LL;"
                                                            "}};"
                                                            "monitoring ="
                                                            "{{"
                                                            "channelStatusConfigs ="
                                                            "{{"
                                                            "testChannelStatus ="
                                                            "{{"
                                                            "freq = 1.0;"
                                                            "heartbeatIntervalSec = 2.0;"
                                                            "}};"
                                                            "}};"
                                                            "ChannelStatusMessageOutput ="
                                                            "{{"
                                                            "interface = \"ChannelStatusMessageOutput\";"
                                                            "channel ="
                                                            "{{"
                                                            "name = \"monitor\";"
                                                            "type = \"DDS\";"
                                                            "ddsConfigFile = \"{}\";"
                                                            "queueSize = 100LL;"
                                                            "domainId = 0LL;"
                                                            "}};"
                                                            "}};"
                                                            "}};",
                                                            DDSComms::testing::configFile,
                                                            DDSComms::testing::configFile);
    static CustomClock clock_;

    void sendRecvMsg(const std::string& sendMsg)
    {
        auto data = DDSComms::createMsg(sendMsg);
        ASSERT_GT(bWriter_.publication_matched_status().current_count(), 0);
        bWriter_.write(data);
        std::this_thread::sleep_for(500ms);

        auto readMsg = std::async(std::launch::async, [this]() {
            std::string msg;
            this->readChannel_->read(msg, true);
            return msg;
        });

        {
            const auto readResult = getWithTimeout(readMsg);
            ASSERT_TRUE(readResult);
            EXPECT_EQ(readResult, sendMsg);
        }
    }

    void advanceTick()
    {
        clock_.advanceCurrentTime(1);
        channelStatusMonitor_->tick(clock_.now());
    }

    void sleepExpectChannelStatus(const av::msg::ChannelStatus& status)
    {
        std::this_thread::sleep_for(500ms);
        {
            av::msg::ChannelStatusMessage statusMsg;
            const bool found = monitorReadInterface_->get(statusMsg, true);
            EXPECT_TRUE(found);
            EXPECT_EQ(statusMsg.channelName, "testChannelStatus");
            EXPECT_TRUE(statusMsg.status == status);
        }
    }

    void advanceTickExpectChannelStatus(const av::msg::ChannelStatus& status)
    {
        advanceTick();
        sleepExpectChannelStatus(status);
    }

    void advanceTickExpectNoMsg()
    {
        advanceTick();
        std::this_thread::sleep_for(500ms);
        {
            av::msg::ChannelStatusMessage statusMsg;
            const bool found = monitorReadInterface_->get(statusMsg);
            EXPECT_FALSE(found);
        }
    }

    void testUpTransitionToDegraded()
    {
        sleepExpectChannelStatus(av::msg::ChannelStatus::Up);
        advanceTickExpectNoMsg();
        advanceTickExpectChannelStatus(av::msg::ChannelStatus::Degraded);
        advanceTickExpectNoMsg();
        advanceTickExpectChannelStatus(av::msg::ChannelStatus::Degraded);
    }
};

CustomClock ChannelStatusMonitorTest::clock_{};

TEST_F(ChannelStatusMonitorTest, WriteAndReadOnce)
{
    sendRecvMsg("Hello World");
    channelStatusMonitor_->tick(clock_.now());
    testUpTransitionToDegraded();
}

TEST_F(ChannelStatusMonitorTest, WriteAndReadTwice)
{
    sendRecvMsg("Hello World");
    advanceTick();
    sendRecvMsg("It's got beets!");
    channelStatusMonitor_->tick(clock_.now());
    testUpTransitionToDegraded();
    advanceTickExpectNoMsg();
    advanceTickExpectChannelStatus(av::msg::ChannelStatus::Degraded);
    advanceTickExpectChannelStatus(av::msg::ChannelStatus::Down);
    advanceTickExpectNoMsg();
}
} // namespace
