#ifndef TEST_DUMMYSTATS_H
#define TEST_DUMMYSTATS_H

#include <NetworkMetrics/NetworkStats.hpp>
#include <chrono>

namespace {
using namespace std::literals::chrono_literals;

/// @brief For testing: A NetworkStats that just stores the reported
/// stats.
class DummyStats : public network::NetworkStats
{
public:
    /// @brief Constructors.
    DummyStats(ConfigSection config = ConfigSection())
        : NetworkStats(av::TimePoint(0s), 1min, 5, config)
    {}

    /// @brief Represents the statistics gathered for one window.
    ///
    /// There are separate collections of WindowResults for latency
    /// and bandwidth.
    struct WindowResults
    {
        /// @brief Constructor, so we can use emplace_back().
        WindowResults(av::TimePoint _start,
                      uint64_t _count,
                      std::optional<double> _min,
                      std::optional<double> _max,
                      std::optional<double> _mean,
                      std::optional<double> _variance)
            : start(_start)
            , count(_count)
            , min(_min)
            , max(_max)
            , mean(_mean)
            , variance(_variance)
        {}

        /// @brief The time the window started.
        av::TimePoint start;

        /// @brief The number of messages received in the window.
        uint64_t count;

        /// @brief The minimum measurement (latency or bandwidth) in the window.
        std::optional<double> min;

        /// @brief The maximum measurement in the window.
        std::optional<double> max;

        /// @brief The mean measurement in the window.
        std::optional<double> mean;

        /// @brief The variance of the measurements in the window.
        std::optional<double> variance;
    };

    /// @brief Get the WindowResults collection for latency stats.
    ///
    /// @return A vector of WindowResults, representing all the
    /// latency stats collected in all the windows that have occurred.
    const std::vector<WindowResults>& latencyResults(const std::string& channelName = "")
    {
        return latencyResults_[channelName];
    }

    /// @brief Get the WindowResults collection for bandwidth stats.
    ///
    /// @return A vector of WindowResults, representing all the
    /// bandwidth stats collected in all the windows that have
    /// occurred.
    const std::vector<WindowResults>& bandwidthResults(const std::string& channelName = "")
    {
        return bandwidthResults_[channelName];
    }

    using NetworkStats::addChannel;
    bool addChannel(const std::string& name) { return this->addChannel(name, av::TimePoint(0s)); }

protected:
    /// @brief Records the latency results from the given window.
    void reportLatency(const std::string& channel,
                       const RawStats& window,
                       av::TimePoint windowStart,
                       bool forceReport) override
    {
        if (forceReport) {
            latencyResults_[channel].emplace_back(windowStart,
                                                  window.get<1>().getCount(),
                                                  window.get<0>().getMin(),
                                                  window.get<0>().getMax(),
                                                  window.get<1>().getMean(),
                                                  window.get<1>().getVariance());
        }
    }

    /// @brief Records the bandwidth results from the given window.
    void reportBandwidth(const std::string& channel,
                         const RawStats& window,
                         av::TimePoint windowStart,
                         bool forceReport) override
    {
        if (forceReport) {
            bandwidthResults_[channel].emplace_back(windowStart,
                                                    window.get<1>().getCount(),
                                                    window.get<0>().getMin(),
                                                    window.get<0>().getMax(),
                                                    window.get<1>().getMean(),
                                                    window.get<1>().getVariance());
        }
    }

private:
    /// @brief The latency results for all windows seen so far.
    std::map<std::string, std::vector<WindowResults>> latencyResults_{};

    /// @brief The bandwidth results for all windows seen so far.
    std::map<std::string, std::vector<WindowResults>> bandwidthResults_{};
};
} // namespace
#endif // TEST_DUMMYSTATS_H
