#include <Comms/InterfaceUtils.h>
#include <NetworkMetrics/NetworkStatsSender.hpp>
#include <TimeStamp/otime.h>
#include <interfaces/Generic/Input/Channel/Impl.h>

#include "platform/middleware/libs/DDSComms/test/DDSTestPathsHelper.hpp"

#include <chrono>
#include <fmt/format.h>
#include <future>
#include <gtest/gtest.h>
#include <memory>
#include <thread>
#include <vector>

#include "CustomClock.hpp"

namespace {

using namespace std::chrono_literals;

class NetworkStatsSenderTest : public testing::Test
{
public:
    NetworkStatsSenderTest() {}

    void SetUp() override
    {
        clock_.initCurrentTime();
        SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeCustomClock, &clock_);
        auto configPtr{std::make_shared<Config>()};
        configPtr->readString(testNetworkStatsConfig);

        ConfigSection configTest{};
        ASSERT_TRUE(configTest.init("testNetworkStats", configPtr));

        ConfigSection interfaceConfig{};
        ASSERT_TRUE(interfaceConfig.init("networkStats", configPtr));

        ConfigSection metricsConfig{};
        ASSERT_TRUE(metricsConfig.init("Metrics", configPtr));

        ConfigSection monitorInterfaceConfig{};
        ASSERT_EQ(interfaceConfig.get("NetworkMetricsMessage", monitorInterfaceConfig), ConfigSection::GS_OK);

        ConfigSection monitorAlertInterfaceConfig{};
        ASSERT_EQ(interfaceConfig.get("NetworkMetricsAlertMessage", monitorAlertInterfaceConfig), ConfigSection::GS_OK);

        std::shared_ptr<network::NetworkStatsSender::Interface> networkStatsMonitorInterface =
            comms::getInterface<network::NetworkStatsSender::Interface>(
                "NetworkMetricsMessageOutput", monitorInterfaceConfig, task::Access::Write);
        statsReadInterface_ = comms::getInterface<GenericInputChannel<av::msg::NetworkMetricsMessage>>(
            "NetworkMetricsMessageOutput", monitorInterfaceConfig, task::Access::Read);
        assert(statsReadInterface_ != nullptr);
        statsAlertReadInterface_ = comms::getInterface<GenericInputChannel<av::msg::NetworkMetricsAlertMessage>>(
            "NetworkMetricsAlertMessageOutput", monitorAlertInterfaceConfig, task::Access::Read);
        assert(statsAlertReadInterface_ != nullptr);

        std::shared_ptr<network::NetworkStatsSender::AlertInterface> networkStatsMonitorAlertInterface =
            comms::getInterface<network::NetworkStatsSender::AlertInterface>(
                "NetworkMetricsAlertMessageOutput", monitorAlertInterfaceConfig, task::Access::Write);
        assert(networkStatsMonitorAlertInterface != nullptr);

        readChannel_ = task::ChannelFactory::createChannelManaged(configTest, task::Access::Read);
        writeChannel_ = task::ChannelFactory::createChannelManaged(configTest, task::Access::Write);

        network::NetworkStatsSender::InterfaceBuilder builder(
            [networkStatsMonitorInterface](const std::string& interfaceName) {
                if (interfaceName == "NetworkMetricsMessageOutput") {
                    return networkStatsMonitorInterface;
                } else {
                    return std::shared_ptr<network::NetworkStatsSender::Interface>(nullptr);
                }
            });
        network::NetworkStatsSender::AlertInterfaceBuilder alertBuilder(
            [networkStatsMonitorAlertInterface](const std::string& interfaceName) {
                assert(networkStatsMonitorAlertInterface != nullptr);
                if (interfaceName == "NetworkMetricsAlertMessageOutput") {
                    return networkStatsMonitorAlertInterface;
                } else {
                    return std::shared_ptr<network::NetworkStatsSender::AlertInterface>(nullptr);
                }
            });
        networkStatsSender_ = std::make_unique<network::NetworkStatsSender>(
            clock_.now(), 1min, 5, builder, alertBuilder, "NetworkStatsSenderTest", "mordor", metricsConfig, 10);

        readChannel_->connect();
        writeChannel_->connect();
    }

    void TearDown() override
    {
        readChannel_->disconnect();
        writeChannel_->disconnect();
        networkStatsSender_.release();
    }

protected:
    std::unique_ptr<task::Channel> readChannel_;
    std::unique_ptr<task::Channel> writeChannel_;
    std::shared_ptr<GenericInputChannel<av::msg::NetworkMetricsMessage>> statsReadInterface_;
    std::shared_ptr<GenericInputChannel<av::msg::NetworkMetricsAlertMessage>> statsAlertReadInterface_;
    std::unique_ptr<network::NetworkStatsSender> networkStatsSender_;

    const std::string testNetworkStatsConfig = fmt::format("testNetworkStats ="
                                                           "{{"
                                                           "name = \"testNetworkStats\";"
                                                           "type = \"DDS\";"
                                                           "ddsConfigFile = \"{}\";"
                                                           "queueSize = 100LL;"
                                                           "domainId = 0LL;"
                                                           "}};"
                                                           "networkStats ="
                                                           "{{"
                                                           "networkStatsConfigs ="
                                                           "{{"
                                                           "testNetworkStats ="
                                                           "{{"
                                                           "freq = 1.0;"
                                                           "heartbeatIntervalSec = 2.0;"
                                                           "}};"
                                                           "}};"
                                                           "NetworkMetricsMessage ="
                                                           "{{"
                                                           "interface = \"NetworkMetricsMessage\";"
                                                           "channel ="
                                                           "{{"
                                                           "name = \"NetworkStats\";"
                                                           "type = \"DDS\";"
                                                           "ddsConfigFile = \"{}\";"
                                                           "queueSize = 100LL;"
                                                           "domainId = 0LL;"
                                                           "}};"
                                                           "}};"
                                                           "NetworkMetricsAlertMessage ="
                                                           "{{"
                                                           "interface = \"NetworkMetricsAlertMessage\";"
                                                           "channel ="
                                                           "{{"
                                                           "name = \"NetworkStatsAlert\";"
                                                           "type = \"DDS\";"
                                                           "ddsConfigFile = \"{}\";"
                                                           "queueSize = 100LL;"
                                                           "domainId = 0LL;"
                                                           "}};"
                                                           "}};"
                                                           "}};"
                                                           "Metrics ="
                                                           "{{"
                                                           "testNetworkStats ="
                                                           "{{"
                                                           "alertExpectedHz = 10LL;"
                                                           "}};"
                                                           "}};",
                                                           DDSComms::testing::configFile,
                                                           DDSComms::testing::configFile,
                                                           DDSComms::testing::configFile);
    static testing::CustomClock clock_;
};

testing::CustomClock NetworkStatsSenderTest::clock_{};

TEST_F(NetworkStatsSenderTest, WriteAndReadOnce)
{
    const std::string sendMsg{"Hello World"};
    const av::TimePoint startTime{clock_.now()};
    writeChannel_->write(sendMsg);
    std::this_thread::sleep_for(500ms);

    clock_.advanceCurrentTime(1);
    const av::TimePoint detectedTime{clock_.now()};

    networkStatsSender_->onTick();

    auto readMsg = std::async(std::launch::async, [this]() {
        std::string msg;
        this->readChannel_->read(msg, true);
        return msg;
    });

    EXPECT_EQ(readMsg.get(), sendMsg);

    networkStatsSender_->callInterimCallbacks();

    std::this_thread::sleep_for(500ms);
    {
        av::msg::NetworkMetricsMessage metricsMsg;
        const bool found = statsReadInterface_->get(metricsMsg, true);
        EXPECT_TRUE(found);
        EXPECT_EQ(metricsMsg.windowStart, startTime);
        EXPECT_EQ(metricsMsg.windowLength, static_cast<av::msg::Duration>(5min));
        EXPECT_EQ(metricsMsg.count, 1);

        EXPECT_TRUE(metricsMsg.latency.valid);
        EXPECT_EQ(metricsMsg.latency.min, 1000000000);
        EXPECT_EQ(metricsMsg.latency.max, 1000000000);
        EXPECT_EQ(metricsMsg.latency.mean, 1000000000);
        EXPECT_EQ(metricsMsg.latency.variance, 0);
        EXPECT_EQ(metricsMsg.latency.jitter, 0);

        const size_t nbytes{sendMsg.size()};

        EXPECT_TRUE(metricsMsg.bytes.valid);
        EXPECT_EQ(metricsMsg.bytes.min, nbytes);
        EXPECT_EQ(metricsMsg.bytes.max, nbytes);
        EXPECT_EQ(metricsMsg.bytes.mean, nbytes);
        EXPECT_EQ(metricsMsg.bytes.variance, 0);

        EXPECT_EQ(metricsMsg.taskName, "NetworkStatsSenderTest");
        EXPECT_EQ(metricsMsg.hostName, "mordor");
    }

    {
        av::msg::NetworkMetricsAlertMessage alert;
        const bool found = statsAlertReadInterface_->get(alert, true);
        EXPECT_TRUE(found);

        EXPECT_EQ(alert.taskName, "NetworkStatsSenderTest");
        EXPECT_EQ(alert.hostName, "mordor");
        EXPECT_EQ(alert.channelName, "testNetworkStats");
        EXPECT_EQ(alert.timeDetected, detectedTime);
        EXPECT_EQ(alert.category, av::msg::NetworkMetricsAlertMessageCategory::AlertMessagesNotSent);
        EXPECT_EQ(alert.expectedFrequencyHz, 10);
        EXPECT_EQ(alert.numCyclesMissed, 9);
    }
}
} // namespace
