#include "DummyStats.hpp"
#include <NetworkMetrics/NetworkStats.hpp>
#include <gtest/gtest.h>

using namespace std::chrono_literals;

using RawStats = statistics::TupleStatistics<statistics::MinMax<double>, statistics::MeanVariance<double>>;
using WindowedStats = statistics::TimeWindowedSeriesStatistics<double, RawStats, av::AvClock>;

auto windowProcessor = [](auto channelName, auto& cache) {
    return [channel = channelName, &cache](const RawStats& window, av::TimePoint startTime) {
        cache.emplace_back(startTime,
                           window.get<1>().getCount(),
                           window.get<0>().getMin(),
                           window.get<0>().getMax(),
                           window.get<1>().getMean(),
                           window.get<1>().getVariance());
    };
};

struct Metrics
{
    explicit Metrics(const std::string& channelName, av::Duration interval = 1min, std::size_t windowSize = 5)
        : latencyCache{}
        , bandwidthCache{}
        , latency(windowProcessor(channelName, latencyCache), interval, windowSize, av::TimePoint{0s})
        , bandwidth(windowProcessor(channelName, bandwidthCache), interval, windowSize, av::TimePoint{0s})
    {}

    std::vector<DummyStats::WindowResults> latencyCache;
    std::vector<DummyStats::WindowResults> bandwidthCache;
    WindowedStats latency;
    WindowedStats bandwidth;
};

void verifyWindows(const DummyStats::WindowResults& lhs, const DummyStats::WindowResults& rhs)
{
    auto optionalIsEqual = [](const auto& lhs_, const auto& rhs_) {
        const auto hasValue = lhs_.has_value() && rhs_.has_value();
        if (hasValue) {
            ASSERT_DOUBLE_EQ(*lhs_, *rhs_);
        }
        EXPECT_EQ(lhs_.has_value(), rhs_.has_value());
    };
    EXPECT_EQ(lhs.start, rhs.start);
    EXPECT_EQ(lhs.count, rhs.count);
    optionalIsEqual(lhs.min, rhs.min);
    optionalIsEqual(lhs.max, rhs.max);
    optionalIsEqual(lhs.mean, rhs.mean);
    optionalIsEqual(lhs.variance, rhs.variance);
}

namespace network {
namespace {
class NetworkStatsTest : public testing::Test
{
public:
    NetworkStatsTest()
        : configPtr_{std::make_shared<Config>()}
    {}

    void setConfig(const std::string& metricsConfig)
    {
        configPtr_->readString(metricsConfig);
        ASSERT_TRUE(config.init("Metrics", configPtr_));
    }

    ConfigSection config;

    // Just a shorthand, so the next line is more legible.
    using Entry = std::pair<av::Duration, size_t>;

    // The (latency, bandwidth) for each message.
    const std::vector<Entry> entries{{100ns, 1000},
                                     {75ns, 500},
                                     {125ns, 250},
                                     {85ns, 1200},
                                     {45ns, 750},
                                     {250ns, 10000},
                                     {25ns, 50},
                                     {35ns, 75},
                                     {15ns, 10},
                                     {150ns, 5000}};

private:
    std::shared_ptr<Config> configPtr_;
};

/// @brief Tests the results of reporting latency for a single message.
TEST_F(NetworkStatsTest, AddLatency1)
{
    DummyStats stats{};

    // Report a 1-microsecond latency, at 65 seconds past the epoch.
    stats.addLatencySample("", 1us, av::TimePoint(65s));

    // Record the statistics for all the windows, even though they
    // haven't all finished yet.
    stats.callInterimCallbacks();

    // Get the latency stats.
    const auto& latencyResults = stats.latencyResults();

    // There should be one entry for each window.
    EXPECT_EQ(latencyResults.size(), 5);

    // Make sure there are entries for the first two windows.
    ASSERT_GE(latencyResults.size(), 2);

    // The first two windows should have stats. Since there's just
    // one datapoint, min/max/mean should all be equal to that
    // datapoint.
    for (size_t i = 0; i < 2; ++i) {
        EXPECT_TRUE(latencyResults[i].min.has_value());
        if (latencyResults[i].min.has_value()) {
            EXPECT_DOUBLE_EQ(latencyResults[i].min.value(), 1000);
        }

        EXPECT_TRUE(latencyResults[i].max.has_value());
        if (latencyResults[i].max.has_value()) {
            EXPECT_DOUBLE_EQ(latencyResults[i].max.value(), 1000);
        }

        EXPECT_TRUE(latencyResults[i].mean.has_value());
        if (latencyResults[i].mean.has_value()) {
            EXPECT_DOUBLE_EQ(latencyResults[i].mean.value(), 1000);
        }

        EXPECT_EQ(latencyResults[i].count, 1);
        EXPECT_EQ(latencyResults[i].start, av::TimePoint(1min * i));

        /// Variance is 0, because the single datapoint doesn't vary.
        EXPECT_TRUE(latencyResults[i].variance.has_value());
        if (latencyResults[i].variance.has_value()) {
            EXPECT_EQ(latencyResults[i].variance.value(), 0);
        }
    }

    // Make sure there are entries for the remaining windows.
    ASSERT_GE(latencyResults.size(), 5);
    for (size_t i = 2; i < 5; ++i) {
        EXPECT_FALSE(latencyResults[i].min.has_value());
        EXPECT_FALSE(latencyResults[i].max.has_value());
        EXPECT_FALSE(latencyResults[i].mean.has_value());

        EXPECT_EQ(latencyResults[i].count, 0);
        EXPECT_EQ(latencyResults[i].start, av::TimePoint(1min * i));
    }

    // Check that there aren't any bandwidth stats. In real use, this
    // wouldn't happen, because we always report both in the
    // callback; but let's make sure the code is behaving as
    // expected.
    const auto& bandwidthResults = stats.bandwidthResults();

    // We should have all the windows, they just won't have any data.
    EXPECT_EQ(bandwidthResults.size(), 5);
    ASSERT_GE(bandwidthResults.size(), 5);

    // Check that all the windows are empty.
    for (size_t i = 0; i < 5; ++i) {
        EXPECT_EQ(bandwidthResults[i].count, 0);
        EXPECT_FALSE(bandwidthResults[i].min.has_value());
        EXPECT_FALSE(bandwidthResults[i].max.has_value());
        EXPECT_FALSE(bandwidthResults[i].mean.has_value());
        EXPECT_FALSE(bandwidthResults[i].variance.has_value());
    }
}

/// @brief Tests the results of reporting bandwidth for a single message.
TEST_F(NetworkStatsTest, AddBandwidth1)
{
    DummyStats stats{};

    // Report a 1000-byte message, at 65 seconds past the epoch.
    stats.addBandwidthSample("", 1000, av::TimePoint(65s));

    // Record the statistics for all the windows, even though they
    // haven't all finished yet.
    stats.callInterimCallbacks();

    // Get the bandwidth stats.
    const auto& bandwidthResults = stats.bandwidthResults();

    // There should be one entry for each window.
    EXPECT_EQ(bandwidthResults.size(), 5);

    // Make sure there are entries for the first two windows.
    ASSERT_GE(bandwidthResults.size(), 2);

    // The first two windows should have stats. Since there's just
    // one datapoint, min/max/mean should all be equal to that
    // datapoint.
    for (size_t i = 0; i < 2; ++i) {
        EXPECT_TRUE(bandwidthResults[i].min.has_value());
        if (bandwidthResults[i].min.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].min.value(), 1000.0);
        }

        EXPECT_TRUE(bandwidthResults[i].max.has_value());
        if (bandwidthResults[i].max.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].max.value(), 1000.0);
        }

        EXPECT_TRUE(bandwidthResults[i].mean.has_value());
        if (bandwidthResults[i].mean.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].mean.value(), 1000.0);
        }

        EXPECT_EQ(bandwidthResults[i].count, 1);
        EXPECT_EQ(bandwidthResults[i].start, av::TimePoint(1min * i));

        /// Variance is 0, because the single datapoint doesn't vary.
        EXPECT_TRUE(bandwidthResults[i].variance.has_value());
        if (bandwidthResults[i].variance.has_value()) {
            EXPECT_EQ(bandwidthResults[i].variance.value(), 0);
        }
    }

    // Make sure there are entries for the remaining windows.
    ASSERT_GE(bandwidthResults.size(), 5);
    for (size_t i = 2; i < 5; ++i) {
        EXPECT_FALSE(bandwidthResults[i].min.has_value());
        EXPECT_FALSE(bandwidthResults[i].max.has_value());
        EXPECT_FALSE(bandwidthResults[i].mean.has_value());
        EXPECT_FALSE(bandwidthResults[i].variance.has_value());
        EXPECT_EQ(bandwidthResults[i].count, 0);
        EXPECT_EQ(bandwidthResults[i].start, av::TimePoint(1min * i));
    }

    // Check that there aren't any latency stats. In real use, this
    // wouldn't happen, because we always report both in the
    // callback; but let's make sure the code is behaving as
    // expected.
    const auto& latencyResults = stats.latencyResults();

    // We should have all the windows, they just won't have any data.
    EXPECT_EQ(latencyResults.size(), 5);
    ASSERT_GE(latencyResults.size(), 5);

    // Check that all the windows are empty.
    for (size_t i = 0; i < 5; ++i) {
        EXPECT_EQ(latencyResults[i].count, 0);
        EXPECT_FALSE(latencyResults[i].min.has_value());
        EXPECT_FALSE(latencyResults[i].max.has_value());
        EXPECT_FALSE(latencyResults[i].mean.has_value());
        EXPECT_FALSE(latencyResults[i].variance.has_value());
    }
}

/// @brief Tests the results of reporting bandwidth for 10 identical messages.
TEST_F(NetworkStatsTest, AddBandwidth10Uniform)
{
    DummyStats stats{};

    // Add a datapoint for each of 10 messages, starting at 65s past
    // the epoch, and continuing every 30s. This will cover 5 windows.
    for (size_t i = 0; i < 10; ++i) {
        stats.addBandwidthSample("", 1000, av::TimePoint(65s + i * 30s));
    }

    // Record the statistics for all the windows, even though they
    // haven't all finished yet.
    stats.callInterimCallbacks();

    // Get the bandwidth stats.
    const auto& bandwidthResults = stats.bandwidthResults();

    // There should be one entry for each window.
    EXPECT_EQ(bandwidthResults.size(), 5);
    ASSERT_GE(bandwidthResults.size(), 5);

    // The message counts for the 5 windows.
    EXPECT_EQ(bandwidthResults[0].count, 8);
    EXPECT_EQ(bandwidthResults[1].count, 10);
    EXPECT_EQ(bandwidthResults[2].count, 8);
    EXPECT_EQ(bandwidthResults[3].count, 6);
    EXPECT_EQ(bandwidthResults[4].count, 4);

    // All 5 windows should have data. Since all datapoints are the
    // same, min/max/mean should be 1000, and variance should be 0.
    for (size_t i = 0; i < 5; ++i) {
        EXPECT_TRUE(bandwidthResults[i].min.has_value());
        if (bandwidthResults[i].min.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].min.value(), 1000.0);
        }

        EXPECT_TRUE(bandwidthResults[i].max.has_value());
        if (bandwidthResults[i].max.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].max.value(), 1000.0);
        }

        EXPECT_TRUE(bandwidthResults[i].mean.has_value());
        if (bandwidthResults[i].mean.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].mean.value(), 1000.0);
        }

        EXPECT_TRUE(bandwidthResults[i].variance.has_value());
        if (bandwidthResults[i].variance.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].variance.value(), 0);
        }
    }

    // Check that there aren't any latency stats. In real use, this
    // wouldn't happen, because we always report both in the
    // callback; but let's make sure the code is behaving as
    // expected.
    const auto& latencyResults = stats.latencyResults();

    // Only 5 windows, since we didn't report any latency data.
    EXPECT_EQ(latencyResults.size(), 5);
    ASSERT_GE(latencyResults.size(), 5);

    // Check that all the windows are empty.
    for (size_t i = 0; i < 5; ++i) {
        EXPECT_EQ(latencyResults[i].count, 0);
        EXPECT_FALSE(latencyResults[i].min.has_value());
        EXPECT_FALSE(latencyResults[i].max.has_value());
        EXPECT_FALSE(latencyResults[i].mean.has_value());
        EXPECT_FALSE(latencyResults[i].variance.has_value());
    }
}

/// @brief Tests the results of reporting bandwidth for 10 different messages.
TEST_F(NetworkStatsTest, AddBandwidth10Nonuniform)
{
    DummyStats stats{};

    // The sizes of the 10 messages.
    const std::vector<size_t> sizes{1000, 75, 500, 250, 1200, 750, 5000, 50, 10000, 10};

    // The minimum size in each of 6 windows.
    const std::vector<int> mins{50, 10, 10, 10, 10, 10};

    // The maximum size in each of 6 windows.
    const std::vector<int> maxes{5000, 10000, 10000, 10000, 10000, 10000};

    // The mean size in each of 6 windows.
    const std::vector<double> means{1103.125, 1883.5, 2220, 2835, 3765, 5005};

    // The variance of the sizes in each of 6 windows. Computed with a
    // separate variance algorithm, independently implemented.
    const std::vector<double> variances{
        2655435.2678571427, 10316389.166666666, 12555771.42857143, 15756350.0, 22767233.333333332, 49900050.0};

    // A sanity check against typos.
    ASSERT_EQ(sizes.size(), 10);
    ASSERT_EQ(mins.size(), 6);
    ASSERT_EQ(maxes.size(), 6);
    ASSERT_EQ(means.size(), 6);
    ASSERT_EQ(variances.size(), 6);

    // Add a datapoint for each of 10 messages, starting at 65s past
    // the epoch, and continuing every 30s. This will cover 5 windows.
    for (size_t i = 0; i < 10; ++i) {
        stats.addBandwidthSample("", sizes[i], av::TimePoint(65s + i * 30s));
    }

    // Record the statistics for all the windows, even though they
    // haven't all finished yet.
    stats.callInterimCallbacks();

    // Get the bandwidth stats.
    const auto& bandwidthResults = stats.bandwidthResults();

    // There should be one entry for each window.
    EXPECT_EQ(bandwidthResults.size(), 5);
    ASSERT_GE(bandwidthResults.size(), 5);

    // The message counts for the 5 windows.
    EXPECT_EQ(bandwidthResults[0].count, 8);
    EXPECT_EQ(bandwidthResults[1].count, 10);
    EXPECT_EQ(bandwidthResults[2].count, 8);
    EXPECT_EQ(bandwidthResults[3].count, 6);
    EXPECT_EQ(bandwidthResults[4].count, 4);

    // All 5 windows should have data, matching the expected values
    // listed above.
    for (size_t i = 0; i < 5; ++i) {
        EXPECT_TRUE(bandwidthResults[i].min.has_value());
        if (bandwidthResults[i].min.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].min.value(), mins[i]);
        }

        EXPECT_TRUE(bandwidthResults[i].max.has_value());
        if (bandwidthResults[i].max.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].max.value(), maxes[i]);
        }

        EXPECT_TRUE(bandwidthResults[i].mean.has_value());
        if (bandwidthResults[i].mean.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].mean.value(), means[i]);
        }

        EXPECT_TRUE(bandwidthResults[i].variance.has_value());
        if (bandwidthResults[i].variance.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].variance.value(), variances[i]);
        }
    }

    // Check that there aren't any latency stats. In real use, this
    // wouldn't happen, because we always report both in the
    // callback; but let's make sure the code is behaving as
    // expected.
    const auto& latencyResults = stats.latencyResults();

    // Only 5 windows, since we didn't report any latency data.
    EXPECT_EQ(latencyResults.size(), 5);
    ASSERT_GE(latencyResults.size(), 5);

    // Check that all the windows are empty.
    for (size_t i = 0; i < 5; ++i) {
        EXPECT_EQ(latencyResults[i].count, 0);
        EXPECT_FALSE(latencyResults[i].min.has_value());
        EXPECT_FALSE(latencyResults[i].max.has_value());
        EXPECT_FALSE(latencyResults[i].mean.has_value());
        EXPECT_FALSE(latencyResults[i].variance.has_value());
    }
}

/// @brief Tests the results of reporting bandwidth and latency for 10
/// different messages.
TEST_F(NetworkStatsTest, AddBoth)
{
    DummyStats stats{};

    // The min (latency, bandwidth) for each window.
    const std::vector<std::pair<double, int>> mins{{25, 50}, {15, 10}, {15, 10}, {15, 10}, {15, 10}, {15, 10}};

    // The max (latency, bandwidth) for each window.
    const std::vector<std::pair<double, int>> maxes{
        {250, 10000}, {250, 10000}, {250, 10000}, {250, 10000}, {150, 5000}, {150, 5000}};

    // The mean (latency, bandwidth) for each window.
    const std::vector<std::pair<double, double>> means{
        {92.5, 1728.125}, {90.5, 1883.5}, {91.25, 2166.875}, {86.66666666, 2647.5}, {56.25, 1283.75}, {82.5, 2505.0}};

    // The variance of the (latency, bandwidth) for each
    // window. Computed independently.
    const std::vector<std::pair<double, double>> variances{{5200.0, 11347399.55357143},
                                                           {5074.722222222223, 10316389.166666666},
                                                           {6476.785714285715, 12787206.69642857},
                                                           {8806.666666666664, 16703037.5},
                                                           {3972.9166666666665, 6138722.916666667},
                                                           {9112.5, 12450050.0}};

    // A sanity check against typos.
    ASSERT_EQ(entries.size(), 10);
    ASSERT_EQ(mins.size(), 6);
    ASSERT_EQ(maxes.size(), 6);
    ASSERT_EQ(means.size(), 6);
    ASSERT_EQ(variances.size(), 6);

    // Add latency and bandwidth datapoints for each of 10 messages,
    // starting at 65s past the epoch, and continuing every 30s. This
    // will cover 5 windows.
    for (size_t i = 0; i < 10; ++i) {
        const av::TimePoint t{65s + i * 30s};
        stats.addLatencySample("", entries[i].first, t);
        stats.addBandwidthSample("", entries[i].second, t);
    }

    // Record the statistics for all the windows, even though they
    // haven't all finished yet.
    stats.callInterimCallbacks();

    // Get the latency stats.
    const auto& latencyResults = stats.latencyResults();

    // There should be one entry for each window.
    EXPECT_EQ(latencyResults.size(), 5);
    ASSERT_GE(latencyResults.size(), 5);

    // The message counts for the 6 windows.
    EXPECT_EQ(latencyResults[0].count, 8);
    EXPECT_EQ(latencyResults[1].count, 10);
    EXPECT_EQ(latencyResults[2].count, 8);
    EXPECT_EQ(latencyResults[3].count, 6);
    EXPECT_EQ(latencyResults[4].count, 4);

    // All 5 windows should have data, matching the expected values
    // listed above.
    for (size_t i = 0; i < 5; ++i) {
        EXPECT_TRUE(latencyResults[i].min.has_value());
        if (latencyResults[i].min.has_value()) {
            EXPECT_DOUBLE_EQ(latencyResults[i].min.value(), mins[i].first);
        }

        EXPECT_TRUE(latencyResults[i].max.has_value());
        if (latencyResults[i].max.has_value()) {
            EXPECT_DOUBLE_EQ(latencyResults[i].max.value(), maxes[i].first);
        }

        EXPECT_TRUE(latencyResults[i].mean.has_value());
        if (latencyResults[i].mean.has_value()) {
            EXPECT_NEAR(latencyResults[i].mean.value(), means[i].first, 1e-8);
        }

        EXPECT_TRUE(latencyResults[i].variance.has_value());
        if (latencyResults[i].variance.has_value()) {
            EXPECT_NEAR(latencyResults[i].variance.value(), variances[i].first, 1e-8);
        }
    }

    // Get the bandwidth stats.
    const auto& bandwidthResults = stats.bandwidthResults();

    // There should be one entry for each window.
    EXPECT_EQ(bandwidthResults.size(), 5);
    ASSERT_GE(bandwidthResults.size(), 5);

    // The message counts for the 5 windows.
    EXPECT_EQ(bandwidthResults[0].count, 8);
    EXPECT_EQ(bandwidthResults[1].count, 10);
    EXPECT_EQ(bandwidthResults[2].count, 8);
    EXPECT_EQ(bandwidthResults[3].count, 6);
    EXPECT_EQ(bandwidthResults[4].count, 4);

    // All 5 windows should have data, matching the expected values
    // listed above.
    for (size_t i = 0; i < 5; ++i) {
        EXPECT_TRUE(bandwidthResults[i].min.has_value());
        if (bandwidthResults[i].min.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].min.value(), mins[i].second);
        }

        EXPECT_TRUE(bandwidthResults[i].max.has_value());
        if (bandwidthResults[i].max.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].max.value(), maxes[i].second);
        }

        EXPECT_TRUE(bandwidthResults[i].mean.has_value());
        if (bandwidthResults[i].mean.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].mean.value(), means[i].second);
        }

        EXPECT_TRUE(bandwidthResults[i].variance.has_value());
        if (bandwidthResults[i].variance.has_value()) {
            EXPECT_DOUBLE_EQ(bandwidthResults[i].variance.value(), variances[i].second);
        }
    }
}

TEST_F(NetworkStatsTest, ChannelAddedOnlyWhenChannelConfigSpecified)
{
    const std::vector<std::string> configsWithNoChannelConstraint = {
        "Metrics = {};",
        "Metrics = { A = { }; };",
    };

    for (const auto& configData : configsWithNoChannelConstraint) {
        AVLOG_INFO("{}", configData);
        setConfig(configData);
        DummyStats stats(this->config);
        EXPECT_EQ(stats.addChannel("A"), false);
    }

    const std::vector<std::string> configsWithChannelConstraint = {
        "Metrics = { autoPublish = true; };",
        "Metrics = { A = { maxJitter = 1000.0; }; };",
        "Metrics = { A = { maxBandwidth = 1000.0; }; };",
        "Metrics = { A = { minBandwidth = 1000.0; }; };",
        "Metrics = { A = { maxLatency = 1000.0; }; };",
        "Metrics = { A = { autoPublish = True; }; };",
        "Metrics = { A = { alertExpectedHz = 1000LL; }; };",
    };

    for (const auto& configData : configsWithChannelConstraint) {
        AVLOG_INFO("{}", configData);
        setConfig(configData);
        DummyStats stats(this->config);
        EXPECT_EQ(stats.addChannel("A"), true);
    }
}

TEST_F(NetworkStatsTest, PerChannelTracking)
{
    auto configData = "Metrics = "
                      "{"
                      "windowSize = 5LL;"
                      "intervalInMilliseconds = 60000LL;"
                      "A ="
                      "{"
                      "autoPublish = True;"
                      "};"
                      "};";

    setConfig(configData);
    DummyStats stats(this->config);
    stats.addChannel("A");

    Metrics channelA{"A"};
    Metrics channelB{"B"};

    // Add latency and bandwidth datapoints for each of 10 messages,
    // starting at 65s past the epoch, and continuing every 30s. This
    // will cover 5 windows.
    for (size_t i = 0; i < 10; ++i) {
        const av::TimePoint t{65s + i * 30s};
        stats.addLatencySample("A", entries[i].first, t);
        channelA.latency.addSample(static_cast<double>(entries[i].first.count()), t);
    }
    stats.callInterimCallbacks();

    EXPECT_EQ(channelA.latencyCache.size(), stats.latencyResults("A").size());
    EXPECT_EQ(channelA.bandwidthCache.size(), stats.bandwidthResults("A").size());

    for (size_t i = 0; i < channelA.latencyCache.size(); i++) {
        verifyWindows(channelA.latencyCache[i], stats.latencyResults("A")[i]);
    }

    EXPECT_EQ(channelB.latencyCache.size(), 0);
    EXPECT_EQ(channelB.bandwidthCache.size(), 0);
}

TEST_F(NetworkStatsTest, ChannelWindowSizeAndIntervalConfigurable)
{
    auto configData = "Metrics = "
                      "{"
                      "autoPublish = True;"
                      "windowSize = 5LL;"
                      "intervalInMilliseconds = 60000LL;"
                      "};";

    setConfig(configData);
    DummyStats stats(this->config);
    stats.addChannel("A");
    Metrics intervalDifferent{"A", 2min, 5};
    Metrics windowSizeDifferent{"A", 1min, 10};
    Metrics same{"A", 1min, 5};

    // Add latency and bandwidth datapoints for each of 10 messages,
    // starting at 65s past the epoch, and continuing every 30s. This
    // will cover 5 windows.
    for (size_t i = 0; i < 10; ++i) {
        const av::TimePoint t{65s + i * 30s};
        stats.addLatencySample("A", entries[i].first, t);
        intervalDifferent.latency.addSample(static_cast<double>(entries[i].first.count()), t);
        same.latency.addSample(static_cast<double>(entries[i].first.count()), t);
        windowSizeDifferent.latency.addSample(static_cast<double>(entries[i].first.count()), t);
    }
    stats.callInterimCallbacks();

    EXPECT_EQ(same.latencyCache.size(), stats.latencyResults("A").size());
    EXPECT_NE(windowSizeDifferent.latencyCache.size(), stats.latencyResults("A").size());
    EXPECT_NE(intervalDifferent.latencyCache.size(), stats.latencyResults("A").size());
}

TEST_F(NetworkStatsTest, OnlyReportChannelMetricsWhenConstraintViolated)
{
    auto configData = "Metrics = "
                      "{"
                      "windowSize = 1LL;"
                      "intervalInMilliseconds = 1LL;"
                      "A ="
                      "{"
                      "minBandwidth = 1.0;"
                      "maxBandwidth = 2.0;"
                      "maxJitter = 1.0;"
                      "maxLatency = 5.0;"
                      "};"
                      "};";

    setConfig(configData);

    const std::vector<av::TimePoint> times = {
        av::TimePoint{0ms},
        av::TimePoint{1ms},
    };

    const auto maxLatency = av::Duration{5};
    const auto aboveMaxLatency = maxLatency + av::Duration{1};
    const auto minBandwidth = 1U;
    const auto belowMinBandwidth = minBandwidth - 1U;
    const auto maxBandwidth = 2U;
    const auto aboveMaxBandwidth = maxBandwidth + 1U;

    auto verifyConstraintReporting =
        [this, &times](auto addSample, auto getResults, const auto& constraint, auto expectedReportings) {
            DummyStats stats(this->config);
            stats.addChannel("A");

            for (const auto& time : times) {
                (stats.*addSample)("A", constraint, time);
            }
            EXPECT_EQ((stats.*getResults)("A").size(), expectedReportings);
        };

    auto verifyLatencyConstraintReporting = [&verifyConstraintReporting](const auto& constraint,
                                                                         auto expectedReportings) {
        return verifyConstraintReporting(
            &DummyStats::addLatencySample, &DummyStats::latencyResults, constraint, expectedReportings);
    };

    auto verifyBandwidthConstraintReporting = [&verifyConstraintReporting](const auto& constraint,
                                                                           auto expectedReportings) {
        return verifyConstraintReporting(
            &DummyStats::addBandwidthSample, &DummyStats::bandwidthResults, constraint, expectedReportings);
    };

    verifyLatencyConstraintReporting(maxLatency, 0);          // Don't violate max latency constraint
    verifyLatencyConstraintReporting(aboveMaxLatency, 1);     // Violate max latency constraint
    verifyBandwidthConstraintReporting(minBandwidth, 0);      // Don't violate min bandwidth constraint
    verifyBandwidthConstraintReporting(belowMinBandwidth, 1); // Violate min bandwidth constraint
    verifyBandwidthConstraintReporting(maxBandwidth, 0);      // Don't violate max bandwidth constraint
    verifyBandwidthConstraintReporting(aboveMaxBandwidth, 1); // Violate min bandwidth constraint

    // Don't violate max jitter constraint
    {
        DummyStats stats(this->config);
        stats.addChannel("A");

        stats.addLatencySample("A", av::Duration{0}, times[0]);
        stats.addLatencySample("A", av::Duration{1}, times[1]);
        EXPECT_EQ(stats.latencyResults("A").size(), 0);
    }

    // Violate max jitter constraint
    {
        DummyStats stats(this->config);
        stats.addChannel("A");

        stats.addLatencySample("A", av::Duration{0}, av::TimePoint{0ms});
        stats.addLatencySample("A", av::Duration{2}, av::TimePoint{500us});
        stats.addLatencySample("A", av::Duration{0}, av::TimePoint{1ms});
        EXPECT_EQ(stats.latencyResults("A").size(), 1);
    }
}

} // namespace
} // namespace network
