/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file Network.cc
 * @product common
 * @version <Release Number>
 * @released 20 June 2018
 * @updated 20 June 2018
 * @brief Common operating system network code
 *
 * This is the header for common code used to interact with the operating system's network
 * functionality.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#include <Networking/Network.h>

#include "lib/avlog/avlog.hpp"
#include <StringUtils/StringUtils.h>

#include <algorithm>
#include <cerrno>
#include <dirent.h>
#include <fstream>
#include <string>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <unistd.h>

namespace {
const net::Status SUCCESS = {net::StatusCode::OK, "Success"};

const std::string PROC_NET_SNMP = "/proc/net/snmp";

net::Status SplitFileLine(const std::string& path, const size_t line, std::vector<std::string>& splits)
{
    std::ifstream in(path, std::ios::in);
    if (in.fail()) {
        return {net::StatusCode::ENV_ERROR, "Failed to open " + path};
    }

    std::string values;
    for (size_t i = 0; i < line; ++i) {
        if (in.eof()) {
            return {net::StatusCode::ENV_ERROR, path + " shorter than expected"};
        }

        std::getline(in, values);
    }

    splits = string_utils::splitString(values, ' ');

    return SUCCESS;
}

/*!
 * @class DirectoryWrapper
 *
 * @brief This class shall be a C++ wrapper to open a directory of a given name with absolute path.
 * \n It provides get method which returns the directory stream to read the directory.It properly
 * \n closes the directory stream.
 * \n
 */
class DirectoryWrapper
{
public:
    /*!
     * @brief Default Constructor
     */
    DirectoryWrapper() = default;
    /*!
     * @brief Parameterized Constructor.It shall open the directory of a given name.
     *
     * @throw runtime_error If directory open fails, throws exception with specific error message.
     * 						Exception shall be handled by the caller.
     *
     * @param[in] path as string : Given directory name with absolute path to open.
     */
    DirectoryWrapper(const std::string& path) { open(path); }
    /*!
     * @brief This method shall open the directory of a given name.
     *
     * @throw runtime_error If directory open fails,throws exception with specific error message.
     * \n                   Exception shall be handled by the caller.
     *
     * @param[in] path as string : Given directory name with absolute path to open.
     *
     * @return void
     */
    void open(const std::string& path)
    {
        dirptr_ = opendir(path.c_str());
        if (dirptr_ == nullptr) {
            // These are the listed errors in the man page
            std::string errorMsg;
            switch (errno) {
                case EACCES:
                    errorMsg = "Permission denied to access " + path;
                    break;
                case EBADF:
                    errorMsg = "Bad file descriptor, can't open " + path;
                    break;
                case EMFILE:
                    errorMsg = "Process can't open more file descriptors";
                    break;
                case ENFILE:
                    errorMsg = "System can't open more file descriptors";
                    break;
                case ENOENT:
                    errorMsg = path + " does not exist";
                    break;
                case ENOMEM:
                    errorMsg = "Not enough memory to open " + path;
                    break;
                case ENOTDIR:
                    errorMsg = path + " is not a directory";
                    break;
                default:
                    errorMsg = "Unknown errno: " + std::to_string(errno);
                    break;
            }

            throw std::runtime_error(errorMsg);
        }
    }
    /*!
     * @brief Virtual destructor. It shall close the directory stream associated with directory stream pointer.
     * \n A warning message shall be logged if unable to close the directory.
     */
    virtual ~DirectoryWrapper()
    {
        if (dirptr_ != nullptr) {

            if (0 != closedir(dirptr_)) {
                AVLOG_WARN("Invalid directory stream descriptor is {}", errno);
            }
        }
    }
    /*!
     * @brief This method shall return directory stream pointer.
     *
     * @return directory stream pointer
     */
    DIR* get() { return dirptr_; }

private:
    DIR* dirptr_{nullptr}; ///< Directory stream pointer to access the directory.
};

/*!
 * @brief This method shall check whether a given directory entry is valid.
 * \n It shall check directory entry against invalid entries list {".",".."}
 *
 * @param [in] name as string : name of the directory entry
 *
 * @return bool :  true, if valid file. otherwise, false.
 */
bool validEntry(const std::string& name)
{
    static std::vector<std::string> invalidEntries = {".", ".."};

    return std::find(std::begin(invalidEntries), std::end(invalidEntries), name) == std::end(invalidEntries);
}
} // namespace

namespace net {

std::string GetHostname()
{
    // The max hostname size is generally 64 characters on Linux, but the POSIX max is 255, so defer
    // to that and add one for the NUL character;
    char buff[256] = {0};
    if (gethostname(static_cast<char*>(buff), static_cast<size_t>(sizeof(buff))) == -1) {
        AVLOG_WARN("gethostname failed with error $ {}", errno);
    }
    return std::string(buff);
}

bool getHostMachineName(std::string& hostName)
{
    bool result{false};

    struct utsname buf; // NOLINT(cppcoreguidelines-pro-type-member-init)
    if (uname(&buf) == 0) {
        hostName = std::string{static_cast<char*>(buf.nodename)};
        result = true;
    }

    return result;
}

std::string getHostMachineName()
{
    std::string hostName{};
    if (!getHostMachineName(hostName)) {
        return "unknown";
    }
    return hostName;
}

Status GetInterfaceStats(const std::string& interface, av::msg::InterfaceStats& stats)
{
    const std::string statsDirPath("/sys/class/net/" + interface + "/statistics");

    bool bContinue(true);
    Status status = SUCCESS;

    DirectoryWrapper dir;
    try {
        dir.open(statsDirPath);
    } catch (const std::runtime_error& e) {
        bContinue = false;
        status = Status{StatusCode::ENV_ERROR, e.what()};
    }

    if (bContinue) {
        stats.name = interface;

        auto entry = readdir(dir.get());
        while (entry != nullptr) {
            const auto statName = std::string(entry->d_name);
            if ((entry->d_type == DT_REG) && validEntry(statName)) {
                std::ifstream in(statsDirPath + "/" + statName, std::ios::in);

                std::string val;
                static_cast<void>(std::getline(in, val));

                av::msg::InterfaceStatPair stat{statName, std::stoll(val)};
                stats.values.push_back(stat);
            }

            entry = readdir(dir.get());
        }
        // Read the interface carrier changes
        const std::string carrierChangesFile("/sys/class/net/" + interface + "/carrier_changes");
        std::string val;
        std::ifstream in(carrierChangesFile);
        if (in.good()) {
            static_cast<void>(std::getline(in, val));

            av::msg::InterfaceStatPair carrierStat{"carrier_changes", std::stoll(val)};
            stats.values.push_back(carrierStat);
        }
    }
    return status;
}

Status GetUDPStats(av::msg::UDPStats& stats)
{
    auto const UDP_LINE_NUM = 10;
    auto const UDP_LINE_NUM_LENGTH = 9;

    std::vector<std::string> words;
    auto status = SplitFileLine(PROC_NET_SNMP, UDP_LINE_NUM, words);
    if (status != SUCCESS) {
        return status;
    }

    if (words.size() < UDP_LINE_NUM_LENGTH) {
        return {StatusCode::ENV_ERROR, "/proc/net/snmp UDP line shorter than expected"};
    }

    stats.received = std::stoull(words[1]);
    stats.unknownPort = std::stoull(words[2]);
    stats.inErrors = std::stoull(words[3]);
    stats.sent = std::stoull(words[4]);
    stats.receiveBufErrors = std::stoull(words[5]);
    stats.sendBufErrors = std::stoull(words[6]);
    stats.checksumErrors = std::stoull(words[7]);
    stats.ignoredMulti = std::stoull(words[8]);

    return SUCCESS;
}

Status GetTCPStats(av::msg::TCPStats& stats)
{
    auto const TCP_LINE_NUM = 8;
    auto const TCP_LINE_NUM_LENGTH = 16;

    std::vector<std::string> words;
    auto status = SplitFileLine(PROC_NET_SNMP, TCP_LINE_NUM, words);
    if (status != SUCCESS) {
        return status;
    }

    if (words.size() < TCP_LINE_NUM_LENGTH) {
        return {StatusCode::ENV_ERROR, "/proc/net/snmp TCP line shorter than expected"};
    }

    stats.retransSegments = std::stoull(words[12]);
    stats.inErrors = std::stoull(words[13]);
    stats.checksumErrors = std::stoull(words[15]);

    return SUCCESS;
}

Status GetInterfaces(std::vector<std::string>& interfaces)
{
    const std::string intfDirPath("/sys/class/net");
    bool bContinue(true);
    Status status = SUCCESS;
    DirectoryWrapper dir;
    try {
        dir.open(intfDirPath);
    } catch (const std::runtime_error& e) {
        bContinue = false;
        status = Status{StatusCode::ENV_ERROR, e.what()};
    }

    if (bContinue) {
        auto entry = readdir(dir.get());
        while (entry != nullptr) {
            const auto interfaceName = std::string(entry->d_name);
            if (entry->d_type == DT_LNK) {
                interfaces.emplace_back(interfaceName);
            }

            entry = readdir(dir.get());
        }
    }
    return status;
}
} // namespace net
