/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file Network.h
 * @product common
 * @version <Release Number>
 * @released 20 June 2018
 * @updated 20 June 2018
 * @brief Common operating system network code
 *
 * This is the header for common code used to interact with the operating system's network
 * functionality.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#ifndef NETWORKING_NETWORK_H
#define NETWORKING_NETWORK_H

#include <interfaces/NetworkStatistics/NetworkStatistics.h>

#include <cereal/access.hpp>
#include <cereal/cereal.hpp>
#include <cereal/types/map.hpp>
#include <cereal/types/string.hpp>

#include <cstdint>
#include <map>
#include <string>
#include <vector>

namespace net {

/*
 * @enum StatusCode
 *
 * @brief Status code returned by functions in this module
 */
enum class StatusCode : uint32_t
{
    OK,
    ENV_ERROR,
};

/*!
 * @struct InterfaceStats
 *
 * @brief This structure shall hold statistics for a specific network interface.For example eth0.
 * \n Statistics for a specific network interface shall be extracted from multiple files for that network interface.
 * \n So, statistics for a network interface shall be represented as name of the network interface and statistics data
 * stored in a container for the interface. \n So, a map container shall be chosen to store the statistics for the
 * interface where, \n Key    - in the map container is a file name. \n Value  - in the map container is statistics data
 * contained in the file.It can hold statistics data of size 8 bytes(int64_t).
 */
struct InterfaceStats
{
    std::string name; ///< name of the network interface.example:eth0

    std::map<std::string, ssize_t>
        values; ///< map contains network interface statistics data shall be extracted from multiple files.
                ///< Key   - in the map container is a file name.
                ///< Value - in the map container is statistics data contained in the file.It can hold statistics data
                ///< of size 8 bytes(int64_t).

    template<class Archive>
    void serialize(Archive& ar, const uint32_t version)
    {
        static_cast<void>(version);
        ar& cereal::make_nvp("name", name);
        ar& cereal::make_nvp("values", values);
    }
};

/*
 * @struct UDPStats
 *
 * @brief Statistics for UDP traffic
 */
struct UDPStats
{
    uint64_t received{0};
    uint64_t unknownPort{0};
    uint64_t inErrors{0};
    uint64_t sent{0};
    uint64_t receiveBufErrors{0};
    uint64_t sendBufErrors{0};
    uint64_t checksumErrors{0};
    uint64_t ignoredMulti{0};

    template<class Archive>
    void serialize(Archive& ar, const unsigned int version)
    {
        ar& CEREAL_NVP(received);
        ar& CEREAL_NVP(unknownPort);
        ar& CEREAL_NVP(inErrors);
        ar& CEREAL_NVP(sent);
        ar& CEREAL_NVP(receiveBufErrors);
        ar& CEREAL_NVP(sendBufErrors);
        ar& CEREAL_NVP(checksumErrors);
        ar& CEREAL_NVP(ignoredMulti);
    }
};

/*
 * @struct TCPStats
 *
 * @brief Statistics for TCP traffic
 */
struct TCPStats
{
    uint64_t retransSegments{0};
    uint64_t inErrors{0};
    uint64_t checksumErrors{0};

    template<class Archive>
    void serialize(Archive& ar, const unsigned int version)
    {
        ar& CEREAL_NVP(retransSegments);
        ar& CEREAL_NVP(inErrors);
        ar& CEREAL_NVP(checksumErrors);
    }
};

/*!
 * @struct Status
 *
 * @brief Status structure returned by methods in networking module
 * \n This includes a code and a human readable message.
 * \n For success, StatusCode shall be filled with "OK" and msg shall be filled with "success"
 * \n Otherwise, StatusCode shall be filled with "ENV_ERROR" and msg shall be specific to the error.
 */
struct Status
{
    StatusCode code; ///< status code used in networking module.
    std::string msg; ///< status message used in networking module.

    /*!
     * @brief Overloaded operator equals to (==) method
     * \n This method shall check if the contents of two Status objects are equal
     *
     * @param[in] other : Status object
     *
     * @return bool : true, if both objects are equal. otherwise, false.
     *
     * \sa net::
     */
    bool operator==(const Status& other) { return (this->code == other.code) && (this->msg == other.msg); }

    /*!
     * @brief Overloaded operator equals to (==) method
     * \n This method shall check if the contents of two Status objects are not equal
     *
     * @param[in] other : Status object
     *
     * @return bool : true, if both objects are not equal. otherwise, false.
     *
     * \sa net::
     */
    bool operator!=(const Status& other) { return !(*this == other); }
};

/*!
 * @brief This method shall return the host name of the current machine. It shall be a wrapper for Linux system call
 * \n for fetching the host name and returns it in a C++ string.The host name cann't be beyond 255 bytes.A warning
 * \n a message shall be logged, if the host name is beyond 255 bytes.
 *
 * @return std::string : Upon successful fetch of the host name, a string containing the host name shall be returned.
 * \n                    Otherwise, empty string shall be returned..
 */
std::string GetHostname();

/// Get the host machine name using uname()
/// @param[out] hostName The returned name
/// @return True if the getting was fine, false, otherwise
bool getHostMachineName(std::string& hostName);

/// Get the host machine name using uname(), returning "unknown" if uname() fails
/// @returns name of host machine. If name retrieval fails, returns "unknown"
std::string getHostMachineName();

/*!
 * @brief This method shall get the interface statistics for a specific interface.
 * \n \b Operational \b Flow:
 * \n This method shall look at the specific directory to scrape interface statistics for a given interface. It shall
 * read \n all the files in interface's statistics directory and checks for file name validness. If the file name is
 * valid, \n extracts the statistics(content) from the file and populates the statistics data in the given
 * interfaceStats object. \n It handles runtime_error exception if directory open fails.
 *
 * \n Note: Invalid file names to check for validation : ".", ".."
 * \n Interface statistics dir : /sys/class/net/(given InterfaceName)/statistics
 *
 * @param[in]  interface as string     : name of the interface to scrape
 * @param[out] stats as InterfaceStats : container in which the statistics are returned
 *
 * @return For successful extraction of statistics, StatusCode "OK" with message "success".
 * \n Otherwise, StatusCode "ENV_ERROR" with error message.
 *
 * \sa net::Status
 * \n net::StatusCode
 */
Status GetInterfaceStats(const std::string& interface, av::msg::InterfaceStats& stats);

/*
 * @brief Gets the UDP stats for the host
 */
Status GetUDPStats(av::msg::UDPStats& stats);

/*
 * @brief Gets the TCP stats for the host
 */
Status GetTCPStats(av::msg::TCPStats& stats);

/*
 * @brief This method shall get all of the available network interfaces on the system.
 * \n \b Operational \b Flow:
 * \n This method shall read the specific directory path and all the files names in the directory are copied to the
 * \n given interfaces container object. It handles runtime_error exception if directory open fails.
 * \n Interface statistics dir path : /sys/class/net/
 *
 * @param[out] interfaces as vector : container in which the interface names shall be  returned
 *
 * @return For successful extraction of interfaces, StatusCode "OK" with message "success".
 * \n Otherwise, StatusCode "ENV_ERROR" with error message.
 *
 * \sa net::Status
 * \n net::StatusCode
 */
Status GetInterfaces(std::vector<std::string>& interfaces);
} // namespace net

CEREAL_CLASS_VERSION(net::UDPStats, 1U)
CEREAL_CLASS_VERSION(net::TCPStats, 1U)
CEREAL_CLASS_VERSION(net::InterfaceStats, 1U)

#endif // NETWORKING_NETWORK_H
