/**
 * \copyright
 * Copyright (c) 2022- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * @file
 * @brief Defines unit tests for Networking which mock out uname(2).
 */

#include "platform/middleware/libs/Networking/include/Networking/Network.h"

#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <sys/utsname.h>

/// Allows uname to be configured for mocking and usable in case tests are run in parallel
struct UnameMock
{
    int returnValue = 0;
    std::string nodeName{};
};
UnameMock unameMock;

int uname(struct utsname* buf)
{
    const auto charsToCopy = std::min(size_t(_UTSNAME_NODENAME_LENGTH - 1), unameMock.nodeName.length());
    unameMock.nodeName.copy(buf->nodename, charsToCopy);
    buf->nodename[charsToCopy] = '\0'; // Uname normally null-terminates all fields in utsname

    return unameMock.returnValue;
}

class NetworkingTestUnameMocked : public ::testing::Test
{
protected:
    void SetUp() override
    {
        // Replace the static unameMock with a fresh one
        unameMock = UnameMock{};
    }
};

TEST_F(NetworkingTestUnameMocked, getHostMachineName_returnsFalse)
{
    unameMock.returnValue = -1;
    std::string hostName{""};
    EXPECT_FALSE(net::getHostMachineName(hostName)) << "Hostname was " << hostName;
}

TEST_F(NetworkingTestUnameMocked, getHostMachineName_unknown)
{
    unameMock.returnValue = -1;

    auto hostName = net::getHostMachineName();
    EXPECT_EQ(hostName, "unknown") << "Hostname was " << hostName;
}

TEST_F(NetworkingTestUnameMocked, getHostMachineName_validValue)
{
    unameMock.returnValue = 0;
    const std::string expectedHostName = "HelloMyNameIsHost";
    unameMock.nodeName = expectedHostName;

    const auto hostName = net::getHostMachineName();
    EXPECT_EQ(hostName, expectedHostName) << "Hostname was " << hostName << " instead of " << expectedHostName;
}

TEST_F(NetworkingTestUnameMocked, getHostMachineName_emptyHostName)
{
    unameMock.returnValue = 0;
    const std::string expectedHostName = "";
    unameMock.nodeName = expectedHostName;

    const auto hostName = net::getHostMachineName();
    EXPECT_EQ(hostName, expectedHostName) << "Hostname was " << hostName << " instead of " << expectedHostName;
}
