#ifndef _ODATUM_FIELD_INFO_H_
#define _ODATUM_FIELD_INFO_H_

namespace aptv {

/*
 * @brief Helper struct for packaging up datum information
 */
struct FieldInfo final
{
    FieldInfo(std::string type_,
              std::string name_,
              std::string value_,
              std::shared_ptr<std::vector<FieldInfoPtr>> children_)
        : type(type_)
        , name(name_)
        , value(value_)
        , children(children_)
    {}

    FieldInfo() = default;
    FieldInfo(const FieldInfo& other) = default;
    FieldInfo(FieldInfo&& other) = default;
    FieldInfo& operator=(const FieldInfo& other) = default;
    FieldInfo& operator=(FieldInfo&& other) = default;

    ~FieldInfo() = default;

    /*
     * @brief C++ type of the field
     */
    std::string type;

    /*
     * @brief Name of the field
     */
    std::string name;

    /*
     * @brief String representation of the field value
     */
    std::string value;

    /*
     * @brief Children elements of the field
     *
     * May be null if the field is not a container.  We use a pointer here to avoid lots of vector
     * resizing for the root FieldInfo vector.
     */
    std::shared_ptr<std::vector<FieldInfoPtr>> children;

    /*
     * @brief Equality operator
     *
     * @param other: FieldInfo to compare
     */
    bool operator==(const FieldInfo& other) const
    {
        if ((children == nullptr) && (other.children != nullptr)) {
            return false;
        }

        if ((children != nullptr) && (other.children == nullptr)) {
            return false;
        }

        // Check the simple comparisons before the vector in case we can skip the vector comparison
        if ((type != other.type) || (name != other.name) || (value != other.value)) {
            return false;
        }

        if ((children != nullptr) && (other.children != nullptr)) {
            // Shortcut with a size check
            if (children->size() != other.children->size()) {
                return false;
            }

            // Finally, recursively compare everything
            auto thisIt = children->cbegin();
            auto thatIt = other.children->cbegin();
            while ((thisIt != children->cend()) && (thatIt != other.children->cend())) {
                if (*(*thisIt) != *(*thatIt)) {
                    return false;
                }

                ++thisIt;
                ++thatIt;
            }
        }

        return true;
    }

    /*
     * @brief Inequality operator
     *
     * @param other: FieldInfo to compare
     */
    bool operator!=(const FieldInfo& other) const { return !(*this == other); }
};

} // namespace aptv

#endif // _ODATUM_FIELD_INFO_H_
