/*!
***********************************************************************************************************************
* APTIV PRIVILEGED AND CONFIDENTIAL
*
* @file IntrospectionArchive.h
* @product Messaging Infrastructure
* @version
* @released
* @updated
* @brief Introspection archive for ottoware base datums
*
* This file implements a cereal-compatible Archive type that allows us to inspect datums at runtime.
* It does this by recursing through a datum's serialize() methods like any other archive, but
* instead it gets the demangled variable type name, the variable name if the user has added it
* using a name-value pair, and the value.
*
* @section License
*
* Copyright (c) 2018- Aptiv. All rights reserved.
* NOTE: Removing this heading is prohibited.
*
* @section Open Source License List
* - Cereal
*
* This is a list of any open-source project used by this file.
***********************************************************************************************************************
*/

#ifndef _ODATUM_INTROSPECTION_ARCHIVE_H_
#define _ODATUM_INTROSPECTION_ARCHIVE_H_

#include <array>
#include <map>
#include <set>
#include <stack>
#include <string>
#include <unordered_map>
#include <unordered_set>
#include <utility>
#include <vector>

#include <cereal/cereal.hpp>

#include <ODatum/IntrospectionForwardDecl.h>

#include "FieldInfo.h"
#include "TypeSupport.h"

namespace cereal {

/*
 * @brief Output only archive that generates a map of datum info
 *
 * Currently this produces a simple FieldInfo tree, but could be upgraded to use
 * JSON or another well-defined format so we can make use of external tools and libraries.
 */
class IntrospectionArchive : public OutputArchive<IntrospectionArchive>
{
public:
    IntrospectionArchive()
        : OutputArchive<IntrospectionArchive>(this)
    {
        // Make the root node
        info_ = std::make_shared<aptv::FieldInfo>();

        // Every datum has children
        info_->children = std::make_shared<std::vector<aptv::FieldInfoPtr>>();

        // Initialize the current node
        current_ = info_;

        nextNameCounterStack_.push(0);
    }

    ~IntrospectionArchive() = default;

    /*
     * @brief Access to the introspected info
     */
    const aptv::FieldInfoPtr getInfo() { return info_->children->operator[](0); }

    /*
     * @brief Saving arithmetic types
     */
    template<typename T>
    typename std::enable_if<std::is_scalar<T>::value>::type saveValue(T& t)
    {
        current_->children->emplace_back(
            std::make_shared<aptv::FieldInfo>(TypeSupport::name(t), getName(), std::to_string(t), nullptr));
    }

    /*
     * @brief Saving a string
     *
     * This is specialized to avoid to_string and to avoid iterating and saving every char
     */
    void saveValue(const std::string& t)
    {
        current_->children->emplace_back(
            std::make_shared<aptv::FieldInfo>(TypeSupport::name(t), getName(), t, nullptr));
    }

    /*
     * @brief Start a class node
     */
    template<typename T>
    void startClass(const T& t)
    {
        // Create the new aptv::FieldInfo vector in the parent
        current_->children->emplace_back(std::make_shared<aptv::FieldInfo>(
            TypeSupport::name(t), getName(), "", std::make_shared<std::vector<aptv::FieldInfoPtr>>()));

        // Cache the parent
        fieldInfoStack_.push(current_);
        nextNameCounterStack_.push(0);

        // Stage the newly created parent
        current_ = current_->children->back();
    }

    /*
     * @brief Restore the previous parent node
     */
    void endClass()
    {
        current_ = fieldInfoStack_.top();
        fieldInfoStack_.pop();
        nextNameCounterStack_.pop();
    }

    /*
     * @brief Cache a name for the next FieldInfo
     */
    void setNextName(const std::string& name) { nextName_ = name; }

private:
    aptv::FieldInfoPtr info_; ///< Root node

    aptv::FieldInfoPtr current_; ///< Current FieldInfo on which we are operating

    std::stack<aptv::FieldInfoPtr>
        fieldInfoStack_; ///< Stack of aptv::FieldInfo pointers so we can recurse into aptv::FieldInfo structs

    std::string nextName_ = ""; ///< Next name to use for a container.

    std::stack<size_t> nextNameCounterStack_; ///< Counters for distinguishing unnamed values. Each startClass will push
                                              ///< zero, so the index applies for each container.

    std::string getName()
    {
        if (nextName_.empty()) {
            return "value" + std::to_string(nextNameCounterStack_.top()++);
        } else {
            auto ret = nextName_;
            nextName_.clear();
            return ret;
        }
    }
};

/*
 * @brief Base type for template specialization of cereal::NameValuePairs
 */
template<typename T>
struct is_nvp : std::false_type
{};

/*
 * @brief Specialization for NVPs
 */
template<typename T>
struct is_nvp<cereal::NameValuePair<T>> : std::true_type
{};

/*
 * @brief Prologue for starting new FieldInfo nodes
 *
 * Containers are new parent nodes in the FieldInfo tree.  This generally just classes, i.e.
 * the user's datum, but we need to explicitly exclude a couple classes that
 * are effectively primitives in this context.
 */
template<class T>
inline
    typename std::enable_if<std::is_class<T>::value && !is_nvp<T>::value && !std::is_same<T, std::string>::value>::type
    prologue(IntrospectionArchive& ar, T const& t)
{
    ar.startClass(t);
}

/*
 * @brief Epilogue for finishing a node
 */
template<class T>
inline
    typename std::enable_if<std::is_class<T>::value && !is_nvp<T>::value && !std::is_same<T, std::string>::value>::type
    epilogue(IntrospectionArchive& ar, T const& t)
{
    ar.endClass();
}

//
// Type introspection
//
// Note that we don't specialize for std::pair.  The default cereal save function works for this
//

/*
 * @brief Arithmetic introspection
 */
template<class T, traits::EnableIf<std::is_arithmetic<T>::value> = traits::sfinae>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, T const& t)
{
    ar.saveValue(t);
}

/*
 * @brief String introspection
 */
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, std::string const& t)
{
    ar.saveValue(t);
}

/*
 * @brief SizeTag introspection
 */
template<typename T>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, SizeTag<T> const& t)
{
    // Skip the SizeTag, we don't need it
}

/*
 * @brief BinaryData introspection
 *
 * Based on:
 * https://codereview.stackexchange.com/questions/78535/converting-array-of-bytes-to-the-hex-string-representation
 */
template<typename T>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, BinaryData<T> const& t)
{
    constexpr char hexLUT[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};
    std::string byteStr;

    // Each byte will be 2 chars for hex
    auto byteStrSize = 2 * t.size;
    byteStr.reserve(byteStrSize);
    byteStr.resize(byteStrSize);

    // BinaryData objects store data as a type erased void*, so convert to something we can access
    auto dataPtr = static_cast<const uint8_t*>(t.data);

    // Build the hex representation
    for (decltype(t.size) i = 0; i < t.size; ++i) {
        byteStr[i * 2] = hexLUT[(dataPtr[i] & 0xF0) >> 4];
        byteStr[(i * 2) + 1] = hexLUT[dataPtr[i] & 0x0F];
    }

    // Finally save it
    ar(byteStr);
}

/*
 * @brief NVP introspection
 */
template<typename T>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, NameValuePair<T> const& t)
{
    ar.setNextName(t.name);
    ar(t.value);
}

/*
 * @brief Vector introspection
 */
template<class T, class Alloc>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, std::vector<T, Alloc> const& vec)
{
    for (auto const& v : vec) {
        ar(v);
    }
}

/*
 * @brief Array introspection
 */
template<class T, size_t N>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, std::array<T, N> const& arr)
{
    for (auto const& element : arr) {
        ar(element);
    }
}

// [Parminder]
// ERROR: cereal.hpp:448:9: error: static_assert failed due to requirement 'traits::detail::count_output_serializers<std::
// __bit_const_reference<std::vector<bool, std::allocator<bool>>>, cereal::IntrospectionArchive>::value != 0' 
// "cereal could not find any output serialization functions for the provided type and archive combination.

/*
 * @brief Vector of bool introspection
 */
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, std::vector<bool> const& vec)
{
    for (auto const& v : vec) {
        ar(static_cast<char>(v));
    }
}

/*
 * @brief Map introspection
 */
template<class K, class V, class Compare, class Alloc>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, std::map<K, V, Compare, Alloc> const& m)
{
    for (auto const& p : m) {
        ar(cereal::make_nvp("key", p.first));
        ar(cereal::make_nvp("value", p.second));
    }
}

/*
 * @brief Unordered map introspection
 */
template<class K, class V, class Compare, class Alloc>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, std::unordered_map<K, V, Compare, Alloc> const& m)
{
    for (auto const& p : m) {
        ar(cereal::make_nvp("key", p.first));
        ar(cereal::make_nvp("value", p.second));
    }
}

/*
 * @brief Set introspection
 */
template<class T, class Compare, class Alloc>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, std::set<T, Compare, Alloc> const& s)
{
    for (auto const& i : s) {
        ar(i);
    }
}

/*
 * @brief Unordered set introspection
 */
template<class T, class Compare, class Alloc>
inline void CEREAL_SAVE_FUNCTION_NAME(IntrospectionArchive& ar, std::unordered_set<T, Compare, Alloc> const& s)
{
    for (auto const& i : s) {
        ar(i);
    }
}

} // namespace cereal

// Register archive for polymorphic support
CEREAL_REGISTER_ARCHIVE(cereal::IntrospectionArchive)

#endif // _ODATUM_INTROSPECTION_ARCHIVE_H_
