/*!
***********************************************************************************************************************
* APTIV PRIVILEGED AND CONFIDENTIAL
*
* @file ODatum.h
* @product Messaging Infrastructure
* @version
* @released
* @updated
* @brief Base OttoWare Datum (ODatum) implementation
*
* This file implements a base datum type from which all datums passed over IPC inherit.
* Serialization is implemented using Cereal, an open source serialization library.  Support is
* implemented for generic serialization into a base pointer for objects not serialized from
* a base pointer.
*
* @section License
*
* Copyright (c) 2018- Aptiv. All rights reserved.
* NOTE: Removing this heading is prohibited.
*
* @section Open Source License List
* - Cereal
*
* This is a list of any open-source project used by this file.
***********************************************************************************************************************
*/

#ifndef _ODATUM_H_
#define _ODATUM_H_

#include <cereal/access.hpp>
#include <cereal/archives/portable_binary.hpp>
#include <cereal/cereal.hpp>
#include <cereal/types/polymorphic.hpp>
#include <cereal/types/string.hpp>

#include <Serialization/SerializationUtilities.h>

#include <memory>
#include <string>

// Register the name of a message without adding support for polymorphism.
#define ODATUM_NAMED_TYPE(T)                                                                                           \
    namespace cereal::detail {                                                                                         \
    template<>                                                                                                         \
    struct binding_name<T>                                                                                             \
    {                                                                                                                  \
        static constexpr char const* name() { return static_cast<char const*>(#T); }                                   \
    };                                                                                                                 \
    }

namespace aptv {

// Forward declare for convenience typedef
struct ODatum;

// Convenience typedef
using ODatumPtr = std::shared_ptr<ODatum>;

// Helper container for the header info
struct ODatumHeader
{
    ODatumHeader() {}
    explicit ODatumHeader(const std::string& typeNameInput)
        : typeName(typeNameInput)
    {}

    // Type name used for polymorphic deserialization lookup
    std::string typeName;

    template<class Archive>
    void serialize(Archive& ar, const unsigned int /* version */)
    {
        ar(typeName);
    }
};

/*
 * @brief Base serializable datum type tag
 */
struct ODatum
{
public:
    ODatum() {}
    virtual ~ODatum() {}

private:
    // This serialization function is never used, but is required due to template instantiation
    friend class cereal::access;
    template<class Archive>
    void serialize(Archive& ar, const unsigned int version)
    {}
};

//
// Generic deserialization support
//

/*
 * @brief Strip the type header from an ODatum blob
 *
 * This helper utility is required to support toCSV
 *
 * @param blob : Blob to strip
 */
inline void stripODatumHeader(std::string& blob)
{
    // Grab the ODatumHeader version, this is actually the registered cereal version
    std::uint32_t headerVersion;
    SerializeUtil::restoreType(headerVersion, blob);
    blob.erase(1, sizeof(headerVersion));

    if (headerVersion >= 1) {
        std::string name;
        SerializeUtil::restoreType(name, blob);

        // Skip the first byte, that's the endianness indicator
        blob.erase(1, sizeof(name.length()) + name.length());
    }
}

/*
 * @brief Add an ODatumHeader to a serialized blob
 *
 * This is required for backwards compatibility with old logs
 */
inline void addODatumHeader(std::string& blob, const ODatumHeader& header)
{
    // Serialize the header info
    std::stringstream buffer(std::stringstream::out | std::stringstream::binary);
    {
        cereal::PortableBinaryOutputArchive oa(buffer);
        oa(header);
    }
    // Erase the little endian bool that prefixes portable binary archive serialized data
    blob.erase(0, 1);

    // Prepend the header
    blob = buffer.str() + blob;
}

/*
 * @brief Serialization helper
 */
template<class T,
         cereal::traits::EnableIf<std::is_base_of<aptv::ODatum, T>::value, !std::is_convertible<T, ODatumPtr>::value> =
             cereal::traits::sfinae>
inline void serializeODatum(std::stringstream& buffer, const T& datum)
{
    // the binary archives will flush their output immediately, but it's better to
    // use cereal archives in an RAII matter all the time
    cereal::PortableBinaryOutputArchive oa(buffer);
    ODatumHeader header{cereal::detail::binding_name<T>::name()};
    oa(header);
    oa(datum);
}

/*
 * @brief Serialize a datum
 *
 * This function also takes care of appending the datum header
 */
template<class T,
         cereal::traits::EnableIf<std::is_base_of<aptv::ODatum, T>::value, !std::is_convertible<T, ODatumPtr>::value> =
             cereal::traits::sfinae>
inline void serializeODatum(std::string& blob, const T& datum)
{
    std::stringstream buffer(std::stringstream::out | std::stringstream::binary);
    serializeODatum(buffer, datum);
    blob = std::move(buffer.str());
}

/*
 * @brief Serialize an entire message
 *
 * This function builds a message compatible with our messaging infrastructure
 */
template<class T,
         cereal::traits::EnableIf<std::is_base_of<aptv::ODatum, T>::value, !std::is_convertible<T, ODatumPtr>::value> =
             cereal::traits::sfinae>
inline void serializeODatumMessage(const std::string& header, std::string& blob, const T& datum)
{
    std::stringstream buffer(std::stringstream::out | std::stringstream::binary);
    buffer << header;
    serializeODatum(buffer, datum);
    blob = std::move(buffer.str());
}

/*
 * @brief Serialize a pointer to a datum
 *
 * This is a helper that simply dereferences to avoid how cereal handles pointers
 */
template<class T, cereal::traits::EnableIf<std::is_convertible<T, ODatumPtr>::value> = cereal::traits::sfinae>
inline void serializeODatum(std::string& blob, const T& datum)
{
    serializeODatum(blob, *datum);
}

/*
 * @brief Deserialize a datum by reference
 */
template<typename T,
         cereal::traits::EnableIf<std::is_base_of<aptv::ODatum, T>::value, !std::is_convertible<T, ODatumPtr>::value> =
             cereal::traits::sfinae>
inline void deserializeODatum(T& value, std::string& blob)
{
    std::stringstream buffer(std::stringstream::in | std::stringstream::binary);
    buffer.str(blob);
    {
        cereal::PortableBinaryInputArchive ia(buffer);
        ODatumHeader header;
        // We deserialize the header here but don't use it.  We have to do this since we never know if a
        // datum will be deserialized into an object by reference or into a shared pointer
        ia(header);
        ia(value);
    }
}

/*
 * @brief Deserialize a datum into a pointer to the base datum
 *
 * We skip the normal stuff here and defer to the base restoreType since that will hit our
 * custom load function for ODatumPtrs
 */
inline void deserializeODatum(aptv::ODatumPtr& value, std::string& blob)
{
    SerializeUtil::restoreType(value, blob);
}

/*
 * @brief Custom load function for ODatums into a base pointer
 *
 * This will be called whenever we attempt to deserialize an object into an ODatumPtr or a class that
 * can be converted to an ODatumPtr
 *
 * Note that the templating is still required for the function to be dispatched correctly, even though
 * the enable_if basically reduces it to a function overload
 *
 * @param ar : Cereal Archive object that holds serialized data
 * @param ptr : Base ODatum pointer that will hold the deserialized data
 */
template<
    class Archive,
    cereal::traits::EnableIf<std::is_same<Archive, cereal::PortableBinaryInputArchive>::value> = cereal::traits::sfinae>
inline void load(Archive& ar, aptv::ODatumPtr& ptr)
{
    ODatumHeader header;
    ar(cereal::make_nvp("header", header));

    // Look up the polymorphic binding. We use a custom input binding map here, see the
    // comments in polymorphic_impl.hpp for additional context.
    auto& bindingMap = cereal::detail::StaticObject<cereal::detail::ODatumInputBindingMap<Archive>>::getInstance().map;
    auto binding = bindingMap.find(header.typeName);
    if (binding == bindingMap.end()) {
        throw cereal::Exception("Trying to deserialize an unregistered ODatum");
    }

    // Now deserialize
    std::shared_ptr<void> result;
    binding->second.untracked_shared_ptr(&ar, result);
    ptr = std::static_pointer_cast<aptv::ODatum>(result);
}

} // namespace aptv

// This is required in order to instantiate polymorphic bindings for ODatum
CEREAL_CLASS_VERSION(aptv::ODatum, 1)

// This is just a normal version registration
CEREAL_CLASS_VERSION(aptv::ODatumHeader, 1)

#endif // _ODATUM_H_
