/*!
 ***********************************************************
 * @file SerializationUtilities.h
 * @brief A file defining utility functions for
 * serialization
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 02/29/2016
 * @remark
 * Serialization library is replaced with cereal
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * SerializationUtilities.h
 * Tugrul Galatali
 * 10/17/2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */

#ifndef SERIALIZATIONUTILITIES_H
#define SERIALIZATIONUTILITIES_H

#include <sstream>
#include <type_traits>

#include <cereal/archives/json.hpp>
#include <cereal/archives/portable_binary.hpp>

namespace SerializeUtil {

/*!
 * @brief This template method shall restore the given serialized binary into the object of Class T type.
 * \n \b Operational \b Flow:
 * \n  Create a string stream with mode as input and binary. It shall be initialized with the given serialized binary
 * message. \n  Create a portable binary input archive and initialize with the string stream which holds the serialized
 * binary message of specific size. \n  Input archive deserializes the message and fills the out parameter 'value' in
 * specified data type T format.
 *
 * @param[out] value : The output to be updated with the deserialized value in specified data type T format.
 *
 * @param[in] blob : The pointer to the serialized message.
 *
 * @param[in] size : The size of the serialized message.
 *
 * @return Void.
 *
 */
template<typename T>
void restoreType(T& value, const void* blob, const uint32_t size)
{
    std::stringstream buffer(std::stringstream::in | std::stringstream::binary);
    buffer.rdbuf()->pubsetbuf(reinterpret_cast<char*>(const_cast<void*>(blob)), size);
    {
        cereal::PortableBinaryInputArchive ia(buffer);
        ia(value);
    }
}

/*!
 * @brief This template method shall restore the given serialized binary into the object of Class T type.
 * \n \b Operational \b Flow:
 * \n Create a string stream with mode as input and binary. It shall be initialized with the given serialized binary
 * message. \n Create a portable binary input archive and initialize with the string stream which holds the serialized
 * binary message. \n Input archive deserializes the message and fills the out parameter 'value' in specified data type
 * T format.
 *
 * @param[out] value : The output to be updated with the deserialize'd value in specified data type T format
 *
 * @param[in] blob : The serialized message in string.
 *
 * @return Void.
 *
 */
template<typename T>
void restoreType(T& value, const std::string& blob)
{
    std::stringstream buffer(std::stringstream::in | std::stringstream::binary);
    buffer.str(blob);
    {
        cereal::PortableBinaryInputArchive ia(buffer);
        ia(value);
    }
}

/*!
 * @brief This template method shall serialize the given value of specific data type T.
 * \n \b Operational \b Flow:
 * \n Create a string stream with mode as output and binary.
 * \n Create a portable binary out archive and initialize with string stream which is going to hold the serialized
 * message. \n Load the given input value of specific data type T into the output archive. \n Output archive serializes
 * and fills the out parameter 'blob' with the serialized binary message.
 *
 * @param[out] blob : The output message in serialized form.
 *
 * @param[in] value : The input value of specific data type T to be serialized.
 *
 * @return Void.
 *
 */
template<typename T>
void serializeType(std::string& blob, const T& value)
{
    std::stringstream buffer(std::stringstream::out | std::stringstream::binary);
    {
        cereal::PortableBinaryOutputArchive oa(buffer);
        oa(value);
    }
    blob = std::move(buffer.str());
}

/*!
 * @brief This template method shall serialize the given value pointed by the specific data type T shared pointer.
 *
 * @param[out] blob : The output message in serialized form.
 *
 * @param[in] value : The input to be serialized.
 *
 * @return Void.
 *
 * @sa serializeType(std::string& blob, const T& value)
 *
 */
template<typename T>
void serializeType(std::string& blob, const std::shared_ptr<T>& value)
{
    serializeType(blob, *value);
}

/*!
 * @brief This template method shall restore the given serialized binary into the Class T format and shall return as
 * \n class T2 format(base class pointer).Below are the traits on types T and T2 for restoreAltType to compile.
 * \n traitReq1 - T2 must be a pointer
 * \n traitReq2 - T is a pointer and
 * \n traitReq3 - T points to a class which inherits from T2.
 *
 * \n \b Operational \b Flow:
 * \n Create a string stream and initialize with given serialized message of specific size.
 * \n Create a portable binary input archive and initialize with the string stream which is holding the serialized
 * message. \n Input archive deserializes the serialized message and loads into class T form. \n The deserialized value
 * of class T form shall be converted into class T2 form (base class pointer) and returned.
 *
 * @param[in] blob : The pointer to the serialized message
 * @param[in] size : The size of the serialized message
 *
 * @return The pointer of type T2 to the de-serialized value.
 *
 */
template<typename T, typename T2>
T2 restoreAltType(const void* blob,
                  const uint32_t size,
                  const typename std::enable_if<std::is_pointer<T2>::value>::type* traitReq1 = nullptr,
                  const typename std::enable_if<std::is_pointer<T>::value>::type* traitReq2 = nullptr,
                  const typename std::enable_if<std::is_base_of<typename std::remove_pointer<T2>::type,
                                                                typename std::remove_pointer<T>::type>::value>::type*
                      traitReq3 = nullptr)
{
    static_cast<void>(traitReq1);
    static_cast<void>(traitReq2);
    static_cast<void>(traitReq3);
    T value;
    {
        const std::string blobStr = std::string(static_cast<const char*>(blob), size);
        std::stringstream buffer(blobStr);
        cereal::PortableBinaryInputArchive ia(buffer);
        ia(value);
    }
    return dynamic_cast<T2>(value);
}

/*!
 * @brief This template method shall restore the serialized binary into the Class T format and shall return as
 * \n class T2 format(base class pointer). Below are the traits on types T and T2 for restoreAltType to compile.
 * \n traitReq1  - T2 must be a pointer
 * \n traitReq2  - T is not a pointer,and
 * \n traitReq3  - T is inherits from T2.
 *
 * \n \b Operational \b Flow:
 * \n Create a string stream and initialize with given serialized message of specific size.
 * \n Create a portable binary input archive and initialize with the string stream which holds the serialized message.
 * \n Input archive deserializes the message and loads into class T form.
 * \n The deserialized value of class T form shall be converted into class T2 form (base class pointer) and returned.
 *
 * @param[in] blob  :  The pointer to the serialized message
 * @param[in] size  :  The size of the serialized message
 *
 * @return The pointer of type T2 to the de-serialized value.
 *
 */
template<typename T, typename T2>
T2 restoreAltType(const void* blob,
                  const uint32_t size,
                  const typename std::enable_if<std::is_pointer<T2>::value>::type* traitReq1 = nullptr,
                  const typename std::enable_if<!(std::is_pointer<T>::value)>::type* traitReq2 = nullptr,
                  const typename std::enable_if<std::is_base_of<typename std::remove_pointer<T2>::type,
                                                                typename std::remove_pointer<T>::type>::value>::type*
                      traitReq3 = nullptr)
{
    static_cast<void>(traitReq1);
    static_cast<void>(traitReq2);
    static_cast<void>(traitReq3);
    T* const value = new T();
    {
        const std::string blobStr = std::string(static_cast<const char*>(blob), size);
        std::stringstream buffer(blobStr);
        cereal::PortableBinaryInputArchive ia(buffer);
        ia(*value);
    }
    return dynamic_cast<T2>(value);
}

/*!
 * @brief This specialized template method shall return nullptr if class T does not inherit from class T2
 * \n and also won't restore the serialized binary.Below are the traits on types T and T2 for restoreAltType to compile.
 * \n traitReq1  - T2 must be a pointer
 * \n traitReq3  - T doesn't inherit from T2.
 *
 * @param[in] blob  :  The pointer to the serialized message
 * @param[in] size  :  The size of the serialized message
 *
 * @return nullptr
 */
template<typename T, typename T2>
T2 restoreAltType(
    const void* blob,
    const uint32_t size,
    typename std::enable_if<std::is_pointer<T2>::value>::type* traitReq1 = nullptr,
    typename std::enable_if<!std::is_base_of<typename std::remove_pointer<T2>::type,
                                             typename std::remove_pointer<T>::type>::value>::type* traitReq3 = nullptr)
{

    static_cast<void>(blob);
    static_cast<void>(size);
    static_cast<void>(traitReq1);
    static_cast<void>(traitReq3);

    return nullptr;
}

/*!
 * @brief Restores the serialized JSON to the Class T format.
 *
 * @param value The output to be updated with the deserialized
 * value as the Class T format.
 *
 * @param blob The serialized message in JSON string.
 *
 * @return Void.
 *
 */
template<typename T>
void restoreTypeJSON(T& value, const std::string& blob)
{
    std::istringstream buffer(blob);
    cereal::JSONInputArchive ia(buffer);
    ia(value);
}

/*!
 * @brief Serializes the given messsage to JSON format.
 *
 * @param blob The output JSON message that is just serialized.
 *
 * @param value The input to be serialized.
 *
 * @return Void.
 *
 */
template<typename T>
void serializeTypeJSON(std::string& blob, const T& value)
{
    std::ostringstream buffer;
    // define a scope so when cereal archive finishes, string buffer gets flushed
    {
        cereal::JSONOutputArchive oa(buffer);
        oa(CEREAL_NVP(value));
    }
    blob = buffer.str();
}

/*!
 * @brief take serialized binary message and serialize it to JSON style
 *
 * @param value : template ottomatika interface data type
 * @param blob  : serialized binary message
 * ...
 * @return std::string: string in JSON format
 * @sa References if necessary
 *
 */
template<typename T>
std::string toJson(T& value, const std::string& blob)
{
    std::string msg;
    restoreType(value, blob);
    serializeTypeJSON(msg, value);
    return msg;
}

/*!
 * @brief take JSON message and serialize it to binary
 *
 * @param value : template ottomatika interface data type
 * @param blob  : serialized JSON message
 * ...
 * @return std::string: string in binary format
 * @sa References if necessary
 *
 */
template<typename T>
std::string toBin(T& value, const std::string& blob)
{
    std::string msg;
    restoreTypeJSON(value, blob);
    serializeType(msg, value);
    return msg;
}

} // namespace SerializeUtil

#endif // SERIALIZATIONUTILITIES_H
