/*!
***********************************************************************************************************************
* APTIV PRIVILEGED AND CONFIDENTIAL
*
* @file IntrospectionUnitTests.cc
* @product Messaging Infrastructure
* @version
* @released
* @updated
* @brief Datum introspection unit tests
*
* Unit tests for introspection utilities into datums
*
* @section License
*
* Copyright (c) 2018- Aptiv. All rights reserved.
* NOTE: Removing this heading is prohibited.
*
* @section Open Source License List
* - Cereal
*
* This is a list of any open-source project used by this file.
***********************************************************************************************************************
*/

#include "TestDatums.h"
#include <ODatum/IntrospectionArchive.h>
#include <gtest/gtest.h>
#include <memory> // make_shared
#include <vector> // vector

TEST(IntrospectionArchiveTests, TestBasic)
{
    auto expected = std::make_shared<aptv::FieldInfo>(
        "StringWrapper", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "5", nullptr));
    expected->children->emplace_back(std::make_shared<aptv::FieldInfo>("string", "data", "test", nullptr));

    StringWrapper sw;
    sw.data = "test";
    cereal::IntrospectionArchive ia;

    ia(sw);

    auto actual = ia.getInfo();

    EXPECT_EQ(*expected, *actual);
}

TEST(IntrospectionArchiveTests, TestNested)
{
    auto expected = std::make_shared<aptv::FieldInfo>(
        "NestedDatum", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "2", nullptr));
    // Add the SizeTWrapper
    auto sizeTChildren = std::make_shared<std::vector<aptv::FieldInfoPtr>>();
    expected->children->emplace_back(std::make_shared<aptv::FieldInfo>("SizeTWrapper", "size", "", sizeTChildren));

    sizeTChildren->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "1", nullptr));
    sizeTChildren->emplace_back(std::make_shared<aptv::FieldInfo>("unsigned long", "data", "5", nullptr));

    // Add the StringWrapper
    auto stringChildren = std::make_shared<std::vector<aptv::FieldInfoPtr>>();
    expected->children->emplace_back(std::make_shared<aptv::FieldInfo>("StringWrapper", "str", "", stringChildren));
    stringChildren->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "5", nullptr));
    stringChildren->emplace_back(std::make_shared<aptv::FieldInfo>("string", "data", "test val", nullptr));

    NestedDatum nd;
    nd.size.data = 5;
    nd.str.data = "test val";

    cereal::IntrospectionArchive ia;
    ia(nd);

    auto actual = ia.getInfo();

    EXPECT_EQ(*expected, *actual);
}

TEST(IntrospectionArchiveTests, TestVector)
{
    auto expected = std::make_shared<aptv::FieldInfo>(
        "VectorDatum", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "1", nullptr));
    auto v =
        std::make_shared<aptv::FieldInfo>("vector", "vec", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->push_back(v);
    for (size_t i = 0; i < 6; ++i) {
        v->children->emplace_back(std::make_shared<aptv::FieldInfo>(
            "unsigned char", "value" + std::to_string(i), std::to_string(i), nullptr));
    }

    VectorDatum vd;
    vd.vec.push_back(0);
    vd.vec.push_back(1);
    vd.vec.push_back(2);
    vd.vec.push_back(3);
    vd.vec.push_back(4);
    vd.vec.push_back(5);

    cereal::IntrospectionArchive ia;
    ia(vd);

    auto actual = ia.getInfo();

    EXPECT_EQ(*expected, *actual);
}

TEST(IntrospectionArchiveTests, TestArray)
{
    auto expected = std::make_shared<aptv::FieldInfo>(
        "ArrayDatum", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "1", nullptr));
    auto v = std::make_shared<aptv::FieldInfo>("array", "arr", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->push_back(v);
    for (size_t i = 0; i < 8; ++i) {
        v->children->emplace_back(std::make_shared<aptv::FieldInfo>(
            "unsigned char", "value" + std::to_string(i), std::to_string(i), nullptr));
    }

    ArrayDatum ad;
    for (uint8_t i = 0; i < ad.arr.size(); ++i) {
        ad.arr[i] = i;
    }

    cereal::IntrospectionArchive ia;
    ia(ad);

    auto actual = ia.getInfo();

    EXPECT_EQ(*expected, *actual);
}

TEST(IntrospectionArchiveTests, TestMap)
{
    auto expected = std::make_shared<aptv::FieldInfo>(
        "MapDatum", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "1", nullptr));
    auto m = std::make_shared<aptv::FieldInfo>("map", "map", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->push_back(m);
    for (size_t i = 1; i <= 6; ++i) {
        m->children->emplace_back(
            std::make_shared<aptv::FieldInfo>("string", "key", "str" + std::to_string(i), nullptr));
        m->children->emplace_back(
            std::make_shared<aptv::FieldInfo>("unsigned char", "value", std::to_string(i), nullptr));
    }

    MapDatum md;
    md.m.emplace(std::pair<std::string, uint8_t>{"str1", 1});
    md.m.emplace(std::pair<std::string, uint8_t>{"str2", 2});
    md.m.emplace(std::pair<std::string, uint8_t>{"str3", 3});
    md.m.emplace(std::pair<std::string, uint8_t>{"str4", 4});
    md.m.emplace(std::pair<std::string, uint8_t>{"str5", 5});
    md.m.emplace(std::pair<std::string, uint8_t>{"str6", 6});

    cereal::IntrospectionArchive ia;
    ia(md);

    auto actual = ia.getInfo();

    EXPECT_EQ(*expected, *actual);
}

TEST(IntrospectionArchiveTests, TestUnorderedMap)
{
    UnorderedMapDatum md;
    md.m.emplace(std::pair<std::string, uint8_t>{"str1", 1});
    md.m.emplace(std::pair<std::string, uint8_t>{"str2", 2});
    md.m.emplace(std::pair<std::string, uint8_t>{"str3", 3});
    md.m.emplace(std::pair<std::string, uint8_t>{"str4", 4});
    md.m.emplace(std::pair<std::string, uint8_t>{"str5", 5});

    cereal::IntrospectionArchive ia;
    ia(md);

    auto actual = ia.getInfo();

    // We can't really test order here, so just test that nothing explodes for now
}

TEST(IntrospectionArchiveTests, TestPair)
{
    auto expected = std::make_shared<aptv::FieldInfo>(
        "PairDatum", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "1", nullptr));
    auto p = std::make_shared<aptv::FieldInfo>("pair", "pair", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->push_back(p);
    p->children->emplace_back(std::make_shared<aptv::FieldInfo>("string", "first", "pair datum", nullptr));
    p->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned char", "second", std::to_string(40), nullptr));

    PairDatum pd;
    pd.p = std::make_pair<std::string, uint8_t>("pair datum", 40);

    cereal::IntrospectionArchive ia;
    ia(pd);

    auto actual = ia.getInfo();

    EXPECT_EQ(*expected, *actual);
}

TEST(IntrospectionArchiveTests, TestBinaryData)
{
    auto expected = std::make_shared<aptv::FieldInfo>(
        "BinaryDataDatum", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "1", nullptr));
    auto b =
        std::make_shared<aptv::FieldInfo>("binary", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->push_back(b);
    b->children->emplace_back(std::make_shared<aptv::FieldInfo>("string", "value0", "DEADBEEF01020304", nullptr));

    BinaryDataDatum bd;
    bd.b = {0xDE, 0xAD, 0xBE, 0xEF, 0x01, 0x02, 0x03, 0x04};

    cereal::IntrospectionArchive ia;
    ia(bd);

    auto actual = ia.getInfo();

    EXPECT_EQ(*expected, *actual);
}

TEST(IntrospectionArchiveTests, TestSet)
{
    auto expected = std::make_shared<aptv::FieldInfo>(
        "SetDatum", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "1", nullptr));
    auto v = std::make_shared<aptv::FieldInfo>("set", "set", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->push_back(v);
    for (size_t i = 0; i < 5; ++i) {
        v->children->emplace_back(std::make_shared<aptv::FieldInfo>(
            "unsigned char", "value" + std::to_string(i), std::to_string(i), nullptr));
    }

    SetDatum sd;
    sd.s.emplace(0);
    sd.s.emplace(1);
    sd.s.emplace(2);
    sd.s.emplace(3);
    sd.s.emplace(4);

    cereal::IntrospectionArchive ia;
    ia(sd);

    auto actual = ia.getInfo();

    EXPECT_EQ(*expected, *actual);
}

TEST(IntrospectionArchiveTests, TestUnorderedSet)
{
    UnorderedSetDatum sd;
    sd.s.emplace(1);
    sd.s.emplace(2);
    sd.s.emplace(3);
    sd.s.emplace(4);
    sd.s.emplace(5);

    cereal::IntrospectionArchive ia;
    ia(sd);

    auto actual = ia.getInfo();

    // We can't really test order here, so just test that nothing explodes for now
}

TEST(IntrospectionArchiveTests, TestNestedContainer)
{
    auto expected = std::make_shared<aptv::FieldInfo>(
        "NestedVectorDatum", "value0", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->emplace_back(
        std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "1", nullptr));
    auto v =
        std::make_shared<aptv::FieldInfo>("vector", "vec", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
    expected->children->push_back(v);
    for (size_t ii = 0; ii < 2; ++ii) {
        auto vectorDatum = std::make_shared<aptv::FieldInfo>(
            "VectorDatum", "value" + std::to_string(ii), "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());

        if (ii == 0) {
            vectorDatum->children->emplace_back(
                std::make_shared<aptv::FieldInfo>("unsigned int", "cereal_class_version", "1", nullptr));
        }
        auto vInside =
            std::make_shared<aptv::FieldInfo>("vector", "vec", "", std::make_shared<std::vector<aptv::FieldInfoPtr>>());
        v->children->push_back(vectorDatum);
        vectorDatum->children->push_back(vInside);
        for (size_t jj = 0; jj < 3; ++jj) {
            vInside->children->emplace_back(std::make_shared<aptv::FieldInfo>(
                "unsigned char", "value" + std::to_string(jj), std::to_string(jj), nullptr));
        }
    }

    VectorDatum vd1;
    vd1.vec.push_back(0);
    vd1.vec.push_back(1);
    vd1.vec.push_back(2);

    VectorDatum vd2;
    vd2.vec.push_back(0);
    vd2.vec.push_back(1);
    vd2.vec.push_back(2);

    NestedVectorDatum nvd;
    nvd.vec.push_back(vd1);
    nvd.vec.push_back(vd2);

    cereal::IntrospectionArchive ia;
    ia(nvd);

    auto actual = ia.getInfo();

    EXPECT_EQ(*expected, *actual);
}
