/*!
***********************************************************************************************************************
* APTIV PRIVILEGED AND CONFIDENTIAL
*
* @file ODatumUnitTests.cc
* @product Messaging Infrastructure
* @version
* @released
* @updated
* @brief Base OttoWare Datum (ODatum) unit tests
*
* Unit tests for our base datum type
*
* @section License
*
* Copyright (c) 2018- Aptiv. All rights reserved.
* NOTE: Removing this heading is prohibited.
*
* @section Open Source License List
* - Cereal
*
* This is a list of any open-source project used by this file.
***********************************************************************************************************************
*/
#include "TestDatums.h"

#include <Serialization/SerializationUtilities.h> // serializeType
#include <cereal/archives/json.hpp>
#include <gtest/gtest.h>
#include <memory> // make_shared

//
// Test that we can serialize an ODatum on the stack and then deserialize into
// a pointer to the base class
//
// This is the primary use case for our extensions to cereal
//
TEST(ODatumTests, StackToSharedBasePointer)
{
    // Set some data so we're not comparing constructor values
    StringWrapper stackD1{"test"};

    std::string blob;
    aptv::serializeODatum(blob, stackD1);

    // Deserialize the stack allocated object into a base pointer
    auto basePtr = std::make_shared<aptv::ODatum>();
    aptv::deserializeODatum(basePtr, blob);

    // Downcast so we can compare
    auto downcast = std::static_pointer_cast<StringWrapper>(basePtr);

    EXPECT_EQ((*downcast), stackD1);
}

//
// Test that we can serialize an ODatum on the stack and then deserialize into
// another stack allocated object.
//
// This is the general use case
//
TEST(ODatumTests, StackToStack)
{
    StringWrapper in;
    StringWrapper out;

    // Set some data so we're not comparing constructor values
    in.data = "test";

    std::string blob;
    aptv::serializeODatum(blob, in);
    aptv::deserializeODatum(out, blob);

    EXPECT_EQ(in, out);
}

//
// Test that we can serialize an ODatum by reference, then eventually output a JSON representation
//
TEST(ODatumTests, StackToJSON)
{
    // this was generated by snagging the output from JSONOutputArchive
    std::string expected = R"V0G0N({
    "value0": {
        "polymorphic_id": 2147483649,
        "polymorphic_name": "StringWrapper",
        "ptr_wrapper": {
            "id": 2147483649,
            "data": {
                "cereal_class_version": 5,
                "data": "test"
            }
        }
    })V0G0N";

    StringWrapper in;
    aptv::ODatumPtr out = std::make_shared<aptv::ODatum>();

    // Set some data so we're not comparing constructor values
    in.data = "test";

    std::string blob;
    aptv::serializeODatum(blob, in);
    aptv::deserializeODatum(out, blob);

    std::stringstream ss;
    cereal::JSONOutputArchive ar(ss);
    ar(out);

    EXPECT_EQ(expected, ss.str());
}

//
// Test that a datum serialized from a shared pointer can also be deserialized into a shared
// pointer to the base type
//
TEST(ODatumTests, SharedPointerToSharedBasePointer)
{
    std::shared_ptr<StringWrapper> in = std::make_shared<StringWrapper>();
    aptv::ODatumPtr out = std::make_shared<aptv::ODatum>();

    in->data = "test";

    std::string blob;
    aptv::serializeODatum(blob, in);
    aptv::deserializeODatum(out, blob);

    std::shared_ptr<StringWrapper> downcast = std::static_pointer_cast<StringWrapper>(out);

    EXPECT_EQ(*in, *downcast);
}

//
// Test that we can append dummy data to a blob and deserialize correctly into an ODatum.
// This is required in order to support backwards compatibility with old logs.
//
TEST(ODatumTests, AppendPolymorphicHeader)
{
    std::string data = "some data";

    OldStringWrapper oldDatum;
    oldDatum.data = data;

    StringWrapper newDatum;
    newDatum.data = data;

    // This is our expected outcome after converting the old data
    std::string newDatumBlob;
    aptv::serializeODatum(newDatumBlob, newDatum);

    aptv::ODatumHeader oheader{"StringWrapper"};

    // Serialize the old data, this is what would be in old logs
    std::string convertedBlob;
    SerializeUtil::serializeType(convertedBlob, oldDatum);

    // Prepend the header
    addODatumHeader(convertedBlob, oheader);

    EXPECT_EQ(newDatumBlob, convertedBlob);
}

//
// Test the ability to strip an ODatum blob of the type header.  This is required
// for backwards compatibility with toCSV
//
TEST(ODatumTests, StripPolymorphicHeader)
{
    std::string data = "some data";

    OldStringWrapper oldDatum;
    oldDatum.data = data;

    StringWrapper newDatum;
    newDatum.data = data;

    // This is what we would have in a new log
    std::string newDatumBlob;
    aptv::serializeODatum(newDatumBlob, newDatum);

    // This is what we want to convert the new datum to
    std::string oldDatumBlob;
    SerializeUtil::serializeType(oldDatumBlob, oldDatum);

    aptv::stripODatumHeader(newDatumBlob);

    EXPECT_EQ(oldDatumBlob, newDatumBlob);
}

TEST(ODatumTests, SerializeODatumMessage)
{
    StringWrapper in;
    StringWrapper out;

    // Set some data so we're not comparing constructor values
    in.data = "test";

    std::string header = "header";

    std::string blob;
    aptv::serializeODatumMessage(header, blob, in);

    std::string expectedHeader = blob.substr(0, header.size());
    EXPECT_EQ(header, expectedHeader);

    blob.erase(0, header.size());

    aptv::deserializeODatum(out, blob);

    EXPECT_EQ(in, out);
}
