/**
 * \copyright
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief Templated abstract class from which all other operational requirements checkers are extended.
 */

#ifndef OPERATIONALREQUIREMENTSCHECKER_HPP_
#define OPERATIONALREQUIREMENTSCHECKER_HPP_

#include <config/ConfigSection.h>
#include <interfaces/DiagMonitorStatus/DiagMonitorStatus.h>
#include <interfaces/Generic/Output/Abstract.h>
#include <log/Logger.h>

/**
 * \brief Interface specification for specialized operational requirements checkers
 *
 * \sa TODO : Add link to ORC documentation
 */
template<typename RequirementsInput>
class OperationalRequirementsChecker
{
public:
    /**
     * \brief Default constructor for OperationalRequirementsChecker.
     *
     * \return An instance of OperationalRequirementsChecker.
     */
    OperationalRequirementsChecker() = default;
    OperationalRequirementsChecker(const OperationalRequirementsChecker&) = delete;
    OperationalRequirementsChecker(OperationalRequirementsChecker&&) = delete;
    OperationalRequirementsChecker& operator=(const OperationalRequirementsChecker&) = delete;
    OperationalRequirementsChecker& operator=(OperationalRequirementsChecker&&) = delete;

    /**
     * \brief Destructor for OperationalRequirementsChecker.
     */
    virtual ~OperationalRequirementsChecker() = default;

    /**
     * \brief Sets the requirements of the checker
     *
     * Note: Argument is non-const reference because getters are non-const
     *
     * \param[in] requirementParams : The configuration parameters specific to diagnostic requirements
     * \return void
     */
    virtual void setRequirements(ConfigSection& requirementParams) = 0;

    /**
     * \brief Executes the check of the operational requirements
     *
     * \param[in] requirementsInput : Single container/type that is the minimum set of input(s) needed to check
     * \return void
     */
    virtual void checkRequirements(const RequirementsInput& requirementsInput) = 0;

    /**
     * \brief Executes the status broadcasting requirements of the operational requirements state
     *
     * Note that this could implement the following strategies:
     * - Periodic publish
     * - Publish only as neccessary
     * - Publish only failed requirement(s) status
     * - Publish current status
     *
     * \return void
     */
    virtual void sendDiagnosticData() = 0;

    /**
     * \brief Sets the logger if needed.
     *
     * \param[in] taskLogger : Logger from task
     * \return void
     */
    virtual void setLogger(const logger::Logger& taskLogger) final { logger_ = taskLogger; }

    /**
     * \brief Sets the Diagnostics Status Output interface
     *
     * \param[in] channelName : Name of the specific channel to publish
     * \return void
     */
    virtual bool setInterface(const std::shared_ptr<GenericOutput<DiagMonitorStatus>>& diagnosticsOutput) final
    {
        m_DiagnosticsOutput = diagnosticsOutput;
        return true;
    }

protected:
    std::shared_ptr<GenericOutput<DiagMonitorStatus>>
        m_DiagnosticsOutput;  ///< SCS Output interface reference to report diagnostics statuses
    logger::Logger logger_{}; ///< Logger for debugging support
};

#endif /* OPERATIONALREQUIREMENTSCHECKER_HPP_ */
