/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file OfflineLogReader.h
 * @product ReSimulation
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated Dec 12, 2018
 * @brief This file defines an interface for reading log files without interacting with the SCS Interface.
 *
 * These log files satisfy the logging data interface format defined by SQLChannel.h:
 *  - A single table called OTTOLOG containing two columns
 *    - time(double) : connoting Time since Epoch (key)
 *    - data(string) : compressed SCS message data
 *
 * @sa src/include/task/SQLLogChannel.h
 * @sa src/common/task/SQLLogChannel.cc
 * @sa https://conf.ptc.delphiauto.net/confluence/display/SI/OfflineLogReader
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * - Sqlite3 Public Domain Licence (https://www.sqlite.org/copyright.html)
 * - 3-Clause BSD Licence (from Cereal - https://opensource.org/licenses/BSD-3-Clause)
 ***********************************************************************************************************************
 */
#ifndef OFFLINELOGREADER_H_
#define OFFLINELOGREADER_H_

//=== Includes =========================================================================================================

#include <string>
#include <vector>

#include <avtime/Time.hpp>

#include <OfflineLogUtils/OfflineLogException.h>
#include <TimeStamp/otimeConversions.h>

#include "infrastructure/logging/log_reader/DataLogReader.hpp"

//=== Class Declaration ================================================================================================

/*!
 * @class OfflineLogReader
 *
 * @brief Provides an offline interface for reading SCS logs
 */
template<typename MsgType>
class OfflineLogReader
{
public:
    /*!
     * @typedef LogEntry
     *
     * @brief Represents row in the database files.
     *
     * This pair represents the two columns of the database entry:
     *    first (otimePoint) - Time of the log entry (may not match the time contained in the data)
     *    second (MsgType)   - The SCS message expansion of the data column
     */
    typedef std::pair<ottotime::otimePoint, MsgType> LogEntry;

    /*!
     * @brief Specialized constructor for OfflineLogReader
     *
     * @param logName : Name of the log file; corresponds to channel name
     * @return An instance of OfflineLogReader
     */
    OfflineLogReader(const std::string& logName);

    /*!
     * @brief Copy constructor for OfflineLogReader
     *
     * @param original : The original copy of the OfflineLogReader
     * @return A deep-copied of OfflineLogReader
     */
    OfflineLogReader(const OfflineLogReader& original);

    /*!
     * @brief Destructor for OfflineLogReader
     */
    ~OfflineLogReader() = default;

    /*!
     * @brief Assignment operator for OfflineLogReader
     *
     * @param original : The original copy of the OfflineLogReader
     * @return A deep-copied assignment of OfflineLogReader
     * @sa
     */
    OfflineLogReader& operator=(const OfflineLogReader& original);

    /*!
     * @brief Retrieves a single message sequentially
     *
     * @return MsgType : Current SCS message in the log
     */
    MsgType get();

    /*!
     * @brief Retrieves all messages in the log
     *
     * @return std::vector<MsgType> : Vector containing all messages
     */
    std::vector<MsgType> getAll();

    /*!
     * @brief Retrieves a single message/time pair sequentially
     *
     * @return LogEntry : Current SCS message/time pair in the log
     */
    LogEntry getWithTime();

    /*!
     * @brief Gets the time of the first (readable) log entry.
     *
     * Note that the times specified are with respect to the LOG times, not the timestamp contained in the message
     *
     * @return ottotime::otimePoint : Time of the first log message
     */
    ottotime::otimePoint getStartTime();

    /*!
     * @brief Gets the time of the last (readable) log entry.
     *
     * Note that the times specified are with respect to the LOG times, not the timestamp contained in the message
     *
     * @return ottotime::otimePoint : Time of the last log message
     */
    ottotime::otimePoint getEndTime();

    /*!
     * @brief Resets the internal database step mechanism to retrieve messages from beginning of log again
     *
     * @return void
     */
    void reset();

    /*!
     * @brief Returns if a log contains more data
     *
     * @return bool : Returns TRUE if there is more data in the log available after the previous get-calls
     */
    bool hasMoreData() const;

    /*!
     * @brief This function is no longer used.
     *
     * It is kept for backward compatibility with scripts using the pybindings
     *
     * @return void
     */
    void finalize(){};

private:
    std::string logName_;   ///< Name of the log file
    uint32_t logVersion_;   ///< Version of the log format
    LogEntry currentEntry_; ///< keeps track of the last logEntry received from the db

    av::logging::log_reader::DataLogReader logReader_;             ///< The log reader
    av::logging::log_reader::DataLogReader::ConstIterator logPos_; /// current reading position in the log
    /*!
     * @brief Extracts an SCS message of MsgType from the Cereal-ized data
     *
     * @param portableData : Cereal-ized MsgType
     * @return MsgType : The decompressed message
     */
    MsgType decerealize(std::string& portableData) const;
};

#include "OfflineLogReader/OfflineLogReader.hxx"

#endif /* OFFLINELOGREADER_H_ */
