/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file OfflineLogReader.hxx
 * @product ReSimulation
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated Dec 12, 2018
 * @brief This file implements an interface for reading log files without interacting with the SCS Interface.
 *
 * @sa src/include/task/SQLLogChannel.h
 * @sa src/common/task/SQLLogChannel.cc
 * @sa https://conf.ptc.delphiauto.net/confluence/display/SI/OfflineLogReader
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 ***********************************************************************************************************************
 */

//=== Includes =========================================================================================================
#include <sstream>

#include <cereal/archives/portable_binary.hpp>

#include <ODatum/ODatum.h>
#include <Serialization/SerializationUtilities.h>
#include <TimeStamp/TimeStamp.h>
#include <dataLog/utility.h>

#include "infrastructure/logging/log_reader/LogURL.hpp"

#include <experimental/filesystem>

namespace av::offline_log_reader::detail {
inline av::logging::log_reader::DataLogReader createOfflineReader(std::string const& logName)
{
    // Convert the log name to a url and channel name
    std::experimental::filesystem::path logFile{logName};
    av::logging::log_reader::LogURL url{logFile.parent_path().string()}; // url is from directory
    std::string channel_name{logFile.filename().stem().string()}; // channel name is the filename minus the extension

    // Read metadata
    auto metadata = av::logging::log_reader::DataLogMetadata::fromURL(std::move(url));
    if (!metadata) {
        try {
            std::rethrow_exception(metadata.error());
        } catch (const std::exception& e) {
            throw OfflineLogException(OfflineLogException::Cause::SQL_ERROR, e.what());
        }
    }

    // Verify the channel exists in the metadata
    auto channels = metadata->channelNames();
    if (std::find(channels.begin(), channels.end(), channel_name) == channels.end()) {
        throw OfflineLogException(OfflineLogException::Cause::SQL_ERROR, "failed to find channel");
    }

    // Create the reader
    auto reader = av::logging::log_reader::DataLogReader::build(
        metadata.value(), [&](std::string_view name) { return name == channel_name; });
    if (!reader) {
        try {
            std::rethrow_exception(reader.error());
        } catch (const std::exception& e) {
            throw OfflineLogException(OfflineLogException::Cause::SQL_ERROR, e.what());
        }
    }

    return std::move(reader).value();
}
} // namespace av::offline_log_reader::detail

//=== Implementation ===================================================================================================

/*!
 * @brief Specialized constructor for OfflineLogReader
 *
 * @param logName : Name of the log file; corresponds to channel name
 * @return An instance of OfflineLogReader
 */
template<typename MsgType>
OfflineLogReader<MsgType>::OfflineLogReader(const std::string& logName)
    : logName_(logName)
    , logReader_(std::move(av::offline_log_reader::detail::createOfflineReader(logName_)))
    , logPos_(logReader_.begin())
{}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Copy constructor for OfflineLogReader
 *
 * @param original : The original copy of the OfflineLogReader
 * @return A deep-copied of OfflineLogReader
 */
template<typename MsgType>
OfflineLogReader<MsgType>::OfflineLogReader(const OfflineLogReader& original)
    : logName_(original.logName_)
    , logReader_(std::move(av::offline_log_reader::detail::createOfflineReader(logName_)))
    , logPos_(logReader_.begin())

{}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Assignment operator for OfflineLogReader
 *
 * @param original : The original copy of the OfflineLogReader
 * @return A deep-copied assignment of OfflineLogReader
 */
template<typename MsgType>
OfflineLogReader<MsgType>& OfflineLogReader<MsgType>::operator=(const OfflineLogReader<MsgType>& original)
{
    if (std::addressof(original) != this) {
        logName_ = original.logName_;
        reset();
    }
    return *this;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Retreives a single message sequentially
 *
 * @return MsgType : Current SCS message in the log
 */
template<typename MsgType>
MsgType OfflineLogReader<MsgType>::get()
{
    LogEntry entry = getWithTime();
    return entry.second;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Retrieves all messages in the log
 *
 * @return std::vector<MsgType> : Vector containing all messages
 */
template<typename MsgType>
std::vector<MsgType> OfflineLogReader<MsgType>::getAll()
{
    std::vector<MsgType> entries;
    std::string blob;

    reset();
    for (auto const& logEntry : logReader_) {
        auto const serializedMsgBytes = logEntry.messageData().msgdata();
        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
        blob.assign(reinterpret_cast<char const*>(serializedMsgBytes.data()), serializedMsgBytes.size());
        entries.push_back(decerealize(blob));
    }
    reset();

    return entries;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Retrieves a single message/time pair sequentially
 *
 * @return LogEntry : Current SCS message/time pair in the log
 */
template<typename MsgType>
std::pair<ottotime::otimePoint, MsgType> OfflineLogReader<MsgType>::getWithTime()
{
    auto const& logEntry = logPos_->messageData(); // read entry at current position

    auto const serializedMsgBytes = logEntry.msgdata();
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    std::string blob(reinterpret_cast<char const*>(serializedMsgBytes.data()), serializedMsgBytes.size());
    auto logTime = av::otimePointFromTimePoint(logEntry.timestamp());

    ++logPos_; // move to next record, this invalidates logEntry

    return {logTime, decerealize(blob)};
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Gets the time of the first (readable) log entry.
 *
 * Note that the times specified are with respect to the LOG times, not the timestamp contained in the message
 *
 * @return ottotime::otimePoint : Time of the first log message
 */
template<typename MsgType>
ottotime::otimePoint OfflineLogReader<MsgType>::getStartTime()
{
    auto start = logReader_.firstMessageTimestamp();
    if (!start) {
        // no messages
        return ottotime::otimePoint{};
    }
    return av::otimePointFromTimePoint(start.value());
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Gets the time of the last (readable) log entry.
 *
 * Note that the times specified are with respect to the LOG times, not the timestamp contained in the message
 *
 * @return ottotime::otimePoint : Time of the last log message
 */
template<typename MsgType>
ottotime::otimePoint OfflineLogReader<MsgType>::getEndTime()
{
    auto last = logReader_.lastMessageTimestamp();
    if (!last) {
        // no messages
        return ottotime::otimePoint{};
    }
    return av::otimePointFromTimePoint(last.value());
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Resets the internal database step mechanism to retrieve messages from beginning of log again
 *
 * @return void
 */
template<typename MsgType>
void OfflineLogReader<MsgType>::reset()
{
    logReader_ = std::move(av::offline_log_reader::detail::createOfflineReader(logName_));
    logPos_ = logReader_.begin();
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Returns if a log contains more data
 *
 * @return bool : Returns TRUE if there is more data in the log available after the previous get-calls
 */
template<typename MsgType>
bool OfflineLogReader<MsgType>::hasMoreData() const
{
    return logPos_ != logReader_.end();
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Extracts an SCS message of type MsgType from the Cereal-ized data
 *
 * @param portableData : Cereal-ized MsgType
 * @return MsgType : The decompressed message
 */
template<typename MsgType>
MsgType OfflineLogReader<MsgType>::decerealize(std::string& portableData) const
{
    MsgType msg;

    try {
        aptv::deserializeODatum(msg, portableData);
    } catch (...) {
        OfflineLogException err;
        std::stringstream eout;
        eout << "Failed to extract message type " << typeid(MsgType).name();
        err.exceptionCause_ = OfflineLogException::Cause::MSG_TYPE_ERROR;
        err.whatStr_ = eout.str();
        throw err;
    }

    return msg;
}
