/*!
***********************************************************************************************************************
* APTIV PRIVILEGED AND CONFIDENTIAL
*
* @file OfflineLogWriter.h
* @product ReSimulation
* @version TODO <Release Number>
* @released TODO <date>
* @updated Jun 29, 2018
* @brief This file defines an interface for writing log files without interacting with the SCS Interface.
*
* These log files satisfy the logging data interface format defined by SQLChannel.h:
*  - A single table called OTTOLOG containing two columns
*    - time(double) : connoting Time since Epoch (key)
*    - data(string) : compressed SCS message data
*
* @sa src/include/task/SQLLogChannel.h
* @sa src/common/task/SQLLogChannel.cc
* @sa https://conf.ptc.delphiauto.net/confluence/display/SI/OfflineLogWriter
*
* @section License
*
* Copyright (c) 2018- Aptiv. All rights reserved.
* NOTE: Removing this heading is prohibited.
*
* @section Open Source License List
*
* - Sqlite3 Public Domain Licence (https://www.sqlite.org/copyright.html)
* - 3-Clause BSD Licence (from Cereal - https://opensource.org/licenses/BSD-3-Clause)
***********************************************************************************************************************
*/
#ifndef OFFLINELOGWRITER_H_
#define OFFLINELOGWRITER_H_

//=== Includes =========================================================================================================

#include <mutex>
#include <sqlite3.h>
#include <string>
#include <vector>

#include <avtime/Time.hpp>

#include <OfflineLogUtils/OfflineLogException.h>
#include <TimeStamp/otimeConversions.h>
#include <dataLog/utility.h>

//=== Class Declaration ================================================================================================

/*
 * @class Open mode for logs
 *
 */
enum class LogOpenMode
{
    MODIFY,
    OVERWRITE,
};

/*!
 * @class OfflineLogWriter
 *
 * @brief Provides an offline interface for writing SCS logs
 */
template<typename MsgType>
class OfflineLogWriter
{
public:
    /*!
     * @brief Specialized constructor for OfflineLogWriter
     *
     * @param logName : Name of the log file; corresponds to channel name
     * @return An instance of OfflineLogWriter
     */
    OfflineLogWriter(const std::string& logName,
                     const LogOpenMode mode = LogOpenMode::MODIFY,
                     const uint32_t logVersion = dataLog::DATA_LOG_VERSION);

    /*!
     * @brief Copy constructor for OfflineLogWriter
     *
     * @param original : The original copy of the OfflineLogWriter
     * @return An deep-copied of OfflineLogWriter
     */
    OfflineLogWriter(const OfflineLogWriter& original);

    /*!
     * @brief Destructor for OfflineLogWriter
     */
    virtual ~OfflineLogWriter();

    /*!
     * @brief Assignment operator for OfflineLogWriter
     *
     * @param original : Original copy of OfflineLogWriter instance
     * @return OfflineLogWriter& : Deep-copy assignment
     */
    virtual OfflineLogWriter& operator=(const OfflineLogWriter& original);

    /*!
     * @brief Publishes the SCS message to the log file at the current system-time.
     *
     * @param msg : The SCS message
     * @return void
     */
    void publish(const MsgType& msg);

    /*!
     * @brief Publishes the SCS message to the log file at a user-defined time.
     *
     * @param msg : The SCS message
     * @param msgTime : User-defined log time
     * @return void
     */
    void publishAtTime(const MsgType& msg, const ottotime::otimePoint& msgTime);
    void publishAtTime(const MsgType& msg, const av::TimePoint& msgTime);

    /*!
     * @brief SQL interface finalizer. Called automatically in destructor. Made public for pybind.
     *
     * @return void
     */
    void finalize();

private:
    struct sqlite3* dbHandle_{nullptr};              ///< Handle to the referenced database
    struct sqlite3_stmt* statement_{nullptr};        ///< Reference to the entry-reading SQL statement
    std::string logName_;                            ///< Name of the log file
    static std::vector<std::string> openedLogFiles;  ///< Ensures that only a single file is ever opened.
    LogOpenMode mode_;                               ///< Mode of the log writer
    uint32_t logVersion_{dataLog::DATA_LOG_VERSION}; ///< Version of the log format

    /*!
     * @brief SQL interface initializer
     *
     * @return void
     */
    void init();

    /*!
     * @brief Initialize the log version
     *
     * @param writeVersion : whether or not the log version should be written to the .db file
     *
     * @return void
     */
    void initLogVersion(const bool writeVersion);

    /*!
     * @brief Drop the OTTOLOG table in a database
     *
     * @return void
     */
    void dropLogTable(struct sqlite3* handle);

    /*!
     * @brief Create the OTTOLOG table in a database
     *
     * @return void
     */
    void createLogTable(struct sqlite3* handle);

    /*!
     * @brief Throws exception for SQL Error propogation
     *
     * @param arg_name : arg_brief
     * @return void : Always throws OfflineLogException, but with prepared error string
     */
    void throwSqlError(const char* devErrMsg);
};

#include "OfflineLogWriter/OfflineLogWriter.hxx"

#endif /* OFFLINELOGWRITER_H_ */
