/*!
***********************************************************************************************************************
* APTIV PRIVILEGED AND CONFIDENTIAL
*
* @file OfflineLogWriter.hxx
* @product ReSimulation
* @version TODO <Release Number>
* @released TODO <date>
* @updated Jun 29, 2018
* This file implements an interface for writing log files without interacting with the SCS Interface.
*
* These log files satisfy the logging data interface format defined by SQLChannel.h:
*  - A single table called OTTOLOG containing two columns
*    - time(double) : connoting Time since Epoch (key)
*    - data(string) : compressed SCS message data
*
* @sa src/include/task/SQLLogChannel.h
* @sa src/common/task/SQLLogChannel.cc
* @sa https://conf.ptc.delphiauto.net/confluence/display/SI/OfflineLogWriter
*
* @section License
*
* Copyright (c) 2018- Aptiv. All rights reserved.
* NOTE: Removing this heading is prohibited.
*
* @section Open Source License List
*
* - Sqlite3 Public Domain Licence (https://www.sqlite.org/copyright.html)
* - 3-Clause BSD Licence (from Cereal - https://opensource.org/licenses/BSD-3-Clause)
***********************************************************************************************************************
*/
#ifndef OFFLINELOGWRITER_HXX_
#define OFFLINELOGWRITER_HXX_

//=== Includes =========================================================================================================
#include <algorithm>
#include <limits>
#include <sstream>

#include <sys/stat.h>

#include <cereal/archives/portable_binary.hpp>
#include <Serialization/SerializationUtilities.h>
#include <TimeStamp/TimeStamp.h>

//=== Static Initialization ============================================================================================

template<typename MsgType>
std::vector<std::string> OfflineLogWriter<MsgType>::openedLogFiles = std::vector<std::string>();

//=== Implementation ===================================================================================================

/*!
 * @brief Specialized constructor for OfflineLogWriter
 *
 * @param logName : Name of the log file; corresponds to channel name
 * @return An instance of OfflineLogWriter
 */
template<typename MsgType>
OfflineLogWriter<MsgType>::OfflineLogWriter(const std::string& logName,
                                            const LogOpenMode mode,
                                            const uint32_t logVersion)
    : logName_(logName)
    , mode_(mode)
    , logVersion_(logVersion)
{
    init();
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Copy constructor for OfflineLogWriter
 *
 * @param original : The original copy of the OfflineLogWriter
 * @return An deep-copied of OfflineLogWriter
 */
template<typename MsgType>
OfflineLogWriter<MsgType>::OfflineLogWriter(const OfflineLogWriter<MsgType>& original)
    : logName_(original.logName_)
    , mode_(original.mode_)
{
    init();
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Destructor for OfflineLogWriter
 */
template<typename MsgType>
OfflineLogWriter<MsgType>::~OfflineLogWriter()
{
    finalize();
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Assignment operator for OfflineLogWriter
 *
 * @param original : Original copy of OfflineLogWriter instance
 * @return OfflineLogWriter& : Deep-copy assignment
 */
template<typename MsgType>
OfflineLogWriter<MsgType>& OfflineLogWriter<MsgType>::operator=(const OfflineLogWriter<MsgType>& original)
{
    finalize();

    logName_ = original.logName_;
    init();
    return *this;
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Publishes the SCS message to the log file at the current system-time.
 *
 * @param msg : The SCS message
 * @return void
 */
template<typename MsgType>
void OfflineLogWriter<MsgType>::publish(const MsgType& msg)
{
    // Changed this to use wall clock time, which is closest to the old otimePoint::getLocalTime().
    // If you want a system-synced time you can use SystemClock, but that's not guaranteed to be
    // synced up with real world time.
    publishAtTime(msg, ottotime::otimePoint::buildCurrentWallClockTime());
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Publishes the SCS message to the log file at a user-defined time.
 *
 * @param msg : The SCS message
 * @param msgTime : User-defined log time
 * @return void
 */
template<typename MsgType>
void OfflineLogWriter<MsgType>::publishAtTime(const MsgType& msg, const ottotime::otimePoint& msgTime)
{
    std::string blob;
    if (logVersion_ == dataLog::PRE_ODATUM_LOG_VERSION)
    {
        SerializeUtil::serializeType(blob, msg);
    }
    else
    {
        aptv::serializeODatum(blob, msg);
    }

    double key = otimeToSeconds(msgTime);

    std::string sqlCmd{""};
    if (logVersion_ >= dataLog::MSG_COUNTER_AND_PUBLISHER_ID_VERSION)
    {
        // version 3
        sqlCmd = "INSERT INTO OTTOLOG(time, data, publisherId, messageCounter) VALUES(?, ?, 0, 0)";
    }
    else
    {
        sqlCmd = "INSERT INTO OTTOLOG(time, data) VALUES(?, ?)";
    }

    struct sqlite3_stmt* sqlCmdStmt = nullptr;
    int sqlErrCode = SQLITE_OK;

    sqlErrCode = sqlite3_prepare_v2(dbHandle_, sqlCmd.c_str(), -1, &sqlCmdStmt, nullptr);
    if (sqlErrCode != SQLITE_OK)
    {
        // Failed to prepare writing statement
        throwSqlError("Failed to prepare database insert statement");
    }

    sqlErrCode = sqlite3_bind_double(sqlCmdStmt, 1, key);
    if (sqlErrCode != SQLITE_OK)
    {
        // Failed to bind/write key (time)
        throwSqlError("Failed to bind time key");
    }

    sqlErrCode = sqlite3_bind_blob(sqlCmdStmt, 2, blob.c_str(), blob.size(), SQLITE_STATIC);
    if (sqlErrCode != SQLITE_OK)
    {
        // Failed to bind/write blob (data)
        throwSqlError("Failed to bind message blob");
    }

    sqlErrCode = sqlite3_step(sqlCmdStmt);
    if (sqlErrCode != SQLITE_DONE)
    {
        // Failed to step the entry
        throwSqlError("Failed to step when writing log");
    }

    sqlite3_finalize(sqlCmdStmt);
}

template<typename MsgType>
void OfflineLogWriter<MsgType>::publishAtTime(const MsgType& msg, const av::TimePoint& msgTime)
{
    publishAtTime(msg, av::otimePointFromTimePoint(msgTime));
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief SQL interface initializer
 *
 * @return void
 */
template<typename MsgType>
void OfflineLogWriter<MsgType>::init()
{
    // Make sure we haven't opened this more than once
    auto findIter = std::find(openedLogFiles.begin(), openedLogFiles.end(), logName_);
    if (findIter != openedLogFiles.end())
    {
        std::stringstream eout;
        eout << "Attempted to open " << logName_ << " more than once for writing!";

        OfflineLogException except;
        except.exceptionCause_ = OfflineLogException::Cause::LOG_MULTIPLE_WRITE_OPEN;
        except.whatStr_ = eout.str();
        throw except;
    }
    else
    {
        openedLogFiles.push_back(logName_);
    }

    // If we're about to create the db, we also need to write the version
    // stat returns 0 on success indicating that the file exists and we have adequate permissions
    struct stat statBuff;
    bool logExists = stat(logName_.c_str(), &statBuff) == 0;
    bool shouldWriteVersion = !logExists;

    int32_t opFlags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE; // Open to write and create if non-existent
    int32_t sqlErrCode = sqlite3_open_v2(logName_.c_str(), &dbHandle_, opFlags, nullptr);
    if (sqlErrCode != SQLITE_OK)
    {
        // Failed to open the file
        throwSqlError("Cannot open log file for creation or writing.");
    }

    initLogVersion(shouldWriteVersion);

    if (mode_ == LogOpenMode::OVERWRITE)
    {
        dropLogTable(dbHandle_);
    }

    if (!logExists || mode_ == LogOpenMode::OVERWRITE)
    {
        createLogTable(dbHandle_);
    }

    // logger will not wait for the write to be over before writing the next data
    sqlite3_exec(dbHandle_, "PRAGMA synchronous = 0;", nullptr, nullptr, nullptr);

    // temporary storage for db is memory
    sqlite3_exec(dbHandle_, "PRAGMA temp_store = 2;", nullptr, nullptr, nullptr);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Initialize the log version
 *
 * @param writeVersion : whether or not the log version should be written to the .db file
 *
 * @return void
 */
template<typename MsgType>
void OfflineLogWriter<MsgType>::initLogVersion(const bool writeVersion)
{
    auto sqlErrCode = 0;
    if (writeVersion)
    {
        char* zErrMsg = nullptr;
        sqlErrCode = dataLog::writeLogVersion(dbHandle_, logVersion_, &zErrMsg);
        if (sqlErrCode != SQLITE_OK)
        {
            sqlite3_free(zErrMsg);
            throwSqlError("Failed to write log version");
        }
    }

    // Read back the written log version or the existing log's version
    char* zErrMsg = 0;
    uint32_t readVersion = std::numeric_limits<uint32_t>::max();
    sqlErrCode = dataLog::readLogVersion(dbHandle_, readVersion, &zErrMsg);
    if (sqlErrCode != SQLITE_OK)
    {
        sqlite3_free(zErrMsg);
        throwSqlError("Failed to read log version");
    }

    // Assign the read version of the log since we don't need to verify
    if (!writeVersion)
    {
        logVersion_ = readVersion;
    }

    // If we wrote the version do a sanity read back
    if (writeVersion && (readVersion != logVersion_))
    {
        OfflineLogException except;
        except.whatStr_ = "Read version " + std::to_string(readVersion) + " does not match written version " +
                          std::to_string(logVersion_);
        throw except;
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Drop the OTTOLOG table in a database
 *
 * @return void
 */
template<typename MsgType>
void OfflineLogWriter<MsgType>::dropLogTable(struct sqlite3* handle)
{
    auto sqlErrCode = sqlite3_exec(dbHandle_, "DROP TABLE IF EXISTS OTTOLOG;", nullptr, nullptr, nullptr);
    if (sqlErrCode != SQLITE_OK)
    {
        // Failed to drop table OTTOLOG
        throwSqlError("Caught error dropping possibly existent OTTOLOG table.");
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Create the OTTOLOG table in a database
 *
 * @return void
 */
template<typename MsgType>
void OfflineLogWriter<MsgType>::createLogTable(struct sqlite3* handle)
{
    std::string sqlCmd{""};
    if (logVersion_ >= dataLog::MSG_COUNTER_AND_PUBLISHER_ID_VERSION)
    {
        // version 3
        sqlCmd = "CREATE TABLE OTTOLOG(time REAL NOT NULL, data BLOB NOT NULL, publisherId INTEGER "
                 "NOT NULL, messageCounter INTEGER NOT NULL);";
    }
    else
    {
        sqlCmd = "CREATE TABLE OTTOLOG(time REAL NOT NULL, data BLOB NOT NULL);";
    }

    auto sqlErrCode = sqlite3_exec(dbHandle_,
                                   sqlCmd.c_str(),
                                   nullptr,
                                   nullptr,
                                   nullptr);
    if (sqlErrCode != SQLITE_OK)
    {
        std::cerr << "error creating" << std::endl;
        // Failed to create table OTTOLOG
        throwSqlError("Caught error creating table in database.");
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief SQL interface finalizer
 *
 * @return void
 */
template<typename MsgType>
void OfflineLogWriter<MsgType>::finalize()
{
    if (statement_ != nullptr)
    {
        sqlite3_finalize(statement_);
        statement_ = nullptr;
    }

    if (dbHandle_ != nullptr)
    {
        sqlite3_close(dbHandle_);
        dbHandle_ = nullptr;
    }

    // Clear the opened list
    auto findIter = std::find(openedLogFiles.begin(), openedLogFiles.end(), logName_);
    if (findIter != openedLogFiles.end())
    {
        openedLogFiles.erase(findIter);
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Throws exception for SQL Error propogation
 *
 * @param arg_name : arg_brief
 * @return void : Always throws OfflineLogException, but with prepared error string
 */
template<typename MsgType>
void OfflineLogWriter<MsgType>::throwSqlError(const char* devErrMsg)
{
    OfflineLogException err;
    std::stringstream sout;
    sout << devErrMsg << "\n";
    sout << "\tSqlite3 Err Msg:\n"
         << "\t\t" << sqlite3_errmsg(dbHandle_) << "\n";
    err.whatStr_ = sout.str();
    err.exceptionCause_ = OfflineLogException::Cause::SQL_ERROR;
    throw err;
}

#endif /* OFFLINELOGWRITER_HXX_ */
