/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file LogReaderTimingTest.cpp
 * @product OfflineLogUtilsTests
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated Mar 11, 2019
 * @brief Runs a bunch of OfflineLogReader operations on various logs and displays their timing results.
 *
 * The point is to only interact with OfflineLogReader via its API, but separate builds of that library may have
 * competing implementations. Thus, the outputs will show the difference between those.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 ***********************************************************************************************************************
 */

//=== Includes =========================================================================================================

#include <OfflineLogUtils/OfflineLogReader.h>
#include <iostream>

#define __ENABLE_OTTO_PROFILER__
#include <OttoProfiler/OttoProfiler.h>

#include <interfaces/LidarScan/LidarScan.h>
#include <interfaces/VehicleState/Type.h>

//=== Local Constants ==================================================================================================

const std::string VehicleStateFile = "VehicleState.db";
const std::string IbeoScanFile = "IBEOLeftScan.db";
const std::string VelodyneScanFile = "LeftMirrorVelodyneScan.db";
const std::string ACCTrackerFile = "ACCTargetTracker.db";

//=== Function Declarations ===========================================================================================

/*!
 * @brief Tests OfflineLogReader construction and destruction timing
 *
 * @return void
 */
void testConstrDestr();
void testConstrDestr_VehicleState();
void testConstrDestr_LidarScan();

/*!
 * @brief Tests OfflineLogReader::getWithTime() timing
 *
 * @return void
 */
void testGetWithTime();
void testGetWithTime_VehicleState();
void testGetWithTime_LidarScan();

/*!
 * @brief Tests OfflineLogReader::getStartTime() timing
 *
 * @return void
 */
void testGetStartTime();
void testGetStartTime_VehicleState();
void testGetStartTime_LidarScan();

/*!
 * @brief Tests OfflineLogReader::getEndTime() timing
 *
 * @return void
 */
void testGetEndTime();
void testGetEndTime_VehicleState();
void testGetEndTime_LidarScan();

//=== MAIN =============================================================================================================

/*!
 * @brief Main process function
 *
 * @param int    : argument count (not used)
 * @param char** : argument c-string array (not used)
 * @return int : Process success status
 */
int main(int, char**)
{
    testConstrDestr();
    testGetWithTime();
    testGetNWithTime();
    testGetAllWithTime();
    testGetUntilWithTime(); // Not complete...
    testGetBetweenWithTime();
    testGetStartTime();
    testGetEndTime();
    __OTTO_PROFILER_DISPLAY__

    return 0;
}

//=== Function Implementations =========================================================================================

/*!
 * @brief Tests OfflineLogReader construction and destruction timing
 *
 * @return void
 */
void testConstrDestr()
{
    std::cout << "Start "
              << "testConstrDstr"
              << "..." << std::endl;
    constexpr int count = 100;
    for (int i = 0; i < count; ++i) {
        testConstrDestr_VehicleState();
        testConstrDestr_LidarScan();
    }
    std::cout << "\tDone!" << std::endl;
}

void testConstrDestr_VehicleState()
{
    __OTTO_PROFILER__("testConstrDestr_VehicleState");
    OfflineLogReader<VehicleState> vstateReader(VehicleStateFile);
}

void testConstrDestr_LidarScan()
{
    __OTTO_PROFILER__("testConstrDestr_LidarScan");
    OfflineLogReader<av::msg::LidarScan> lscanReader(IbeoScanFile);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Tests OfflineLogReader::get() timing
 *
 * @return void
 */
void testGetWithTime()
{
    std::cout << "Start "
              << "testGetWithTime"
              << "..." << std::endl;

    testGetWithTime_VehicleState();
    testGetWithTime_LidarScan();

    std::cout << "\tDone!" << std::endl;
}

void testGetWithTime_VehicleState()
{
    OfflineLogReader<VehicleState> vstateReader(VehicleStateFile);

    while (vstateReader.hasMoreData()) {
        __OTTO_PROFILER__("testGetWithTime_VehicleState");
        auto msg = vstateReader.getWithTime();
    }
}

void testGetWithTime_LidarScan()
{
    OfflineLogReader<av::msg::LidarScan> lscanReader(IbeoScanFile);

    while (lscanReader.hasMoreData()) {
        __OTTO_PROFILER__("testGetWithTime_LidarScan");
        auto msg = lscanReader.getWithTime();
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Tests OfflineLogReader::getStartTime() timing
 *
 * @return void
 */
void testGetStartTime()
{
    std::cout << "Start "
              << "testGetStartTime"
              << "..." << std::endl;

    constexpr int count = 10;
    for (int i = 0; i < count; ++i) {
        testGetStartTime_VehicleState();
        testGetStartTime_LidarScan();
    }

    std::cout << "\tDone!" << std::endl;
}

void testGetStartTime_VehicleState()
{
    OfflineLogReader<VehicleState> vstateReader(VehicleStateFile);
    {
        __OTTO_PROFILER__("testGetStartTime_VehicleState");
        auto t = vstateReader.getStartTime();
    }
}

void testGetStartTime_LidarScan()
{
    OfflineLogReader<av::msg::LidarScan> lscanReader(IbeoScanFile);
    {
        __OTTO_PROFILER__("testGetStartTime_LidarScan");
        auto t = lscanReader.getStartTime();
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @brief Tests OfflineLogReader::getEndTime() timing
 *
 * @return void
 */
void testGetEndTime()
{
    std::cout << "Start "
              << "testGetEndTime"
              << "..." << std::endl;

    constexpr int count = 10;
    for (int i = 0; i < count; ++i) {
        testGetEndTime_VehicleState();
        testGetEndTime_LidarScan();
    }

    std::cout << "\tDone!" << std::endl;
}

void testGetEndTime_VehicleState()
{
    OfflineLogReader<VehicleState> vstateReader(VehicleStateFile);
    {
        __OTTO_PROFILER__("testGetEndTime_VehicleState");
        auto t = vstateReader.getEndTime();
    }
}

void testGetEndTime_LidarScan()
{
    OfflineLogReader<av::msg::LidarScan> lscanReader(IbeoScanFile);
    {
        __OTTO_PROFILER__("testGetEndTime_LidarScan");
        auto t = lscanReader.getEndTime();
    }
}
