/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file TestOfflineLogReader.cpp
 * @product ReSimulation
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated Jun 29, 2018
 * @brief This file implements a unit-test suite for the OfflineLogReader utility
 *
 * @sa https://conf.ptc.delphiauto.net/confluence/display/SI/OfflineLogReader
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 ***********************************************************************************************************************
 */

// Includes
#include <gtest/gtest.h>
#include <interfaces/EventLoggerMessageList/EventLoggerMessageList.h>
#include <interfaces/SystemStatus/SystemStatus.h>
#include <iostream>

#include <OfflineLogUtils/OfflineLogReader.h>

//=== TESTS ============================================================================================================

#ifndef TESTDIR
#define TESTDIR "platform/middleware/libs/OfflineLogUtils/test/"
#endif

/*!
 * @test SimpleUse:GoodConstr
 *
 * @brief Tests that OfflineLogReader can construct correctly
 */
TEST(OfflineLogReader_SimpleUse, GoodConstr)
{
    std::string dataFile(TESTDIR);
    dataFile += "testData/SystemStatus.db";
    ASSERT_NO_THROW(OfflineLogReader<av::msg::SystemStatus> logParser(dataFile));
    OfflineLogReader<av::msg::SystemStatus> logParser2(dataFile);
    EXPECT_TRUE(logParser2.hasMoreData());
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:BadConstr
 *
 * @brief Tests that OfflineLogReader fails to construct when given bad log name
 */
TEST(OfflineLogReader_SimpleUse, BadConstr)
{
    EXPECT_THROW(OfflineLogReader<av::msg::SystemStatus> logParser("bad.db"), OfflineLogException);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:GoodGet
 *
 * @brief Tests getting a single message
 */
TEST(OfflineLogReader_SimpleUse, GoodGet)
{
    std::string dataFile(TESTDIR);
    dataFile += "testData/SystemStatus.db";
    OfflineLogReader<av::msg::SystemStatus> logParser(dataFile);
    ASSERT_NO_THROW(logParser.get());
    av::msg::SystemStatus msg = logParser.get();
    EXPECT_EQ(msg.machineFailInfo.size(), 0u);
    EXPECT_EQ(msg.sensorFailInfo.size(), 0u);
    EXPECT_EQ(msg.actuatorFailInfo.size(), 0u);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:BadGet
 *
 * @brief Tests failing to get a single message because of the wrong message type specified.
 */
TEST(OfflineLogReader_SimpleUse, BadGet)
{
    std::string dataFile(TESTDIR);
    dataFile += "testData/SystemStatus.db";
    OfflineLogReader<av::msg::EventLoggerMessageList> logParser(dataFile);
    ASSERT_THROW(logParser.get(), OfflineLogException);
    try {
        logParser.get();
    } catch (OfflineLogException& e) {
        EXPECT_EQ(e.exceptionCause_, OfflineLogException::Cause::MSG_TYPE_ERROR);
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:GetAll
 *
 * @brief Tests getting all entries
 */
TEST(OfflineLogReader_SimpleUse, GetAll)
{
    std::string dataFile(TESTDIR);
    dataFile += "testData/SystemStatus.db";
    OfflineLogReader<av::msg::SystemStatus> logParser(dataFile);
    auto testFirstEntry = logParser.getWithTime();

    ASSERT_NO_THROW(logParser.getAll());
    EXPECT_EQ(logParser.getAll().size(), 5u);
    EXPECT_TRUE(logParser.hasMoreData());
    EXPECT_EQ(testFirstEntry.first, logParser.getWithTime().first);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:GoodGetTimes
 *
 * @brief Test getting log start/end times under nominal circumstances.
 */
TEST(OfflineLogReader_SimpleUse, GoodGetTimes)
{
    std::string dataFile(TESTDIR);
    dataFile += "testData/SystemStatus.db";
    OfflineLogReader<av::msg::SystemStatus> logParser(dataFile);

    ottotime::otimePoint start;
    ottotime::otimePoint end;
    ASSERT_NO_THROW(start = logParser.getStartTime());
    ASSERT_NO_THROW(end = logParser.getEndTime());
    EXPECT_EQ(start.to_otimeSpec().tv_sec, 1613592771);
    EXPECT_EQ(end.to_otimeSpec().tv_sec, 1613592803);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test FringeUse:ThousandOpens
 *
 * @brief Opening and closing a database file reliably
 */
TEST(OfflineLogReader_FringeUse, ThousandOpens)
{
    std::string dataFile(TESTDIR);
    dataFile += "testData/SystemStatus.db";
    for (uint16_t oCount = 0; oCount < 1000; ++oCount) {
        try {
            OfflineLogReader<av::msg::SystemStatus> logParser(dataFile);
            logParser.getAll();
            logParser.reset();
        } catch (OfflineLogException& e) {
            FAIL() << e.what() << " count=" << oCount;
        } catch (std::exception& e) {
            FAIL() << e.what() << " count=" << oCount;
        } catch (int& e) {
            FAIL() << "Caught integer error code ERRNO[" << e << "]"
                   << " count=" << oCount;
        } catch (...) {
            FAIL() << "Caught unknown exception!"
                   << "count = " << oCount;
        }
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test FringeUse:ConcurrentOpens
 *
 * @brief Testing that we can have multiple logs opened at the same time
 */
TEST(OfflineLogReader_FringeUse, ConcurrentOpens)
{
    std::string dataFile(TESTDIR);
    dataFile += "testData/SystemStatus.db";

    // Assignment test
    {
        OfflineLogReader<av::msg::SystemStatus> logParser1(dataFile);
        ASSERT_NO_THROW(OfflineLogReader<av::msg::SystemStatus> logParser2 = logParser1);
        OfflineLogReader<av::msg::SystemStatus> logParser3 = logParser1;
        ASSERT_NO_THROW(logParser1.getAll());
        EXPECT_TRUE(logParser1.hasMoreData());
        ASSERT_TRUE(logParser3.hasMoreData());
        EXPECT_NO_THROW(logParser3.get());
    }

    // Many opens
    std::vector<OfflineLogReader<av::msg::SystemStatus>> logReaderVect;
    OfflineLogReader<av::msg::SystemStatus> dummyLogReader(dataFile);
    for (std::vector<OfflineLogReader<av::msg::SystemStatus>>::size_type oCount = 0; oCount < 64; ++oCount) {
        logReaderVect.push_back(OfflineLogReader<av::msg::SystemStatus>(dataFile));
        dummyLogReader = *logReaderVect.rbegin();
        ASSERT_TRUE(logReaderVect.at(oCount).hasMoreData());
        ASSERT_NO_THROW(logReaderVect.at(oCount).getAll());
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test FringeUse:MultipleGets
 *
 * @brief Testing that fetching from one does not affect fetching from another
 */
TEST(OfflineLogReader_FringeUse, MultipleGets)
{
    std::string dataFile(TESTDIR);
    dataFile += "testData/SystemStatus.db";

    OfflineLogReader<av::msg::SystemStatus> olr1(dataFile);
    olr1.get();
    av::msg::SystemStatus msg1 = olr1.get();
    OfflineLogReader<av::msg::SystemStatus> olr2 = olr1;
    av::msg::SystemStatus msg2 = olr2.get();
    EXPECT_FALSE(msg1.timeStamp == msg2.timeStamp);
}

//=== MAIN =============================================================================================================

/*!
 * @brief Main executable for OfflineLogReader's unit tests
 *
 * @param argc : Number of command-line arguments
 * @param argv : Array of c-strings
 * @return int : Executable's error code (should be zero)
 */
int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
