/*!
***********************************************************************************************************************
* APTIV PRIVILEGED AND CONFIDENTIAL
*
* @file TestOfflineLogWriter.cpp
* @product ReSimulation
* @version TODO <Release Number>
* @released TODO <date>
* @updated Jul 2, 2018
* @brief This file implements a unit-test suite for the OfflineLogWriter utility
*
* @sa https://conf.ptc.delphiauto.net/confluence/display/SI/OfflineLogWriter
*
* @section License
*
* Copyright (c) 2018- Aptiv. All rights reserved.
* NOTE: Removing this heading is prohibited.
*
* @section Open Source License List
*
* - Sqlite3 Public Domain Licence (https://www.sqlite.org/copyright.html)
* - 3-Clause BSD Licence (from Cereal - https://opensource.org/licenses/BSD-3-Clause)
***********************************************************************************************************************
*/

// Includes
#include <OfflineLogUtils/OfflineLogReader.h>
#include <OfflineLogUtils/OfflineLogWriter.h>
#include <cstdio>
#include <experimental/filesystem>
#include <gtest/gtest.h>
#include <interfaces/EventLoggerMessageList/EventLoggerMessageList.h>
#include <recgeometry/recGeometry.h>

#ifndef TESTDIR
#define TESTDIR "./"
#endif

namespace {
class OfflineLogWriterTest : public ::testing::Test
{
protected:
    std::vector<std::string> testFiles_;

    const std::string testDir_{TESTDIR};

    /*
     * @brief Caching test file name generator
     *
     * We need this so we can clean up after a test is run
     */
    std::string makeTestFileName(const std::string name)
    {
        auto fileName = testDir_ + name;
        testFiles_.push_back(fileName);
        return fileName;
    }

    void TearDown() override
    {
        for (auto& testfile : testFiles_) {
            if (std::experimental::filesystem::exists(testfile)) {
                std::experimental::filesystem::remove(testfile);
            }
        }
    }
};

using OfflineLogWriter_SimpleUse = OfflineLogWriterTest;
using OfflineLogWriter_IntegrityCase = OfflineLogWriterTest;
using OfflineLogWriter_ErantUse = OfflineLogWriterTest;
using OfflineLogWriter_Modes = OfflineLogWriterTest;
} // namespace

//=== TESTS ============================================================================================================

/*!
 * @test SimpleUse:Constr
 *
 * @brief Tests that OfflineLogWriter can construct and destruct cleanly
 */
TEST_F(OfflineLogWriter_SimpleUse, Constr)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_SimpleUse.db");
    ASSERT_NO_THROW(OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(dataFile));
    ASSERT_NO_THROW(
        OfflineLogReader<av::msg::EventLoggerMessage> logReader(dataFile)); // Assuming OfflineLogReader works
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:VersionedConstr
 *
 * @brief Tests that OfflineLogWriter can construct and destruct cleanly with a specified version
 */
TEST_F(OfflineLogWriter_SimpleUse, VersionedConstr)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_VersionedSimpleUse.db");
    ASSERT_NO_THROW(OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(
        dataFile, LogOpenMode::MODIFY, dataLog::PRE_ODATUM_LOG_VERSION));
    ASSERT_NO_THROW(
        OfflineLogReader<av::msg::EventLoggerMessage> logReader(dataFile)); // Assuming OfflineLogReader works
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:WriteDatum
 *
 * @brief Test correctly writing a single message
 */
TEST_F(OfflineLogWriter_SimpleUse, WriteDatum)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_SimpleUse_WriteDatum.db");

    const ottotime::otimePoint msgTime = ottotime::otimePoint::buildCurrentWallClockTime();
    const ottotime::otimeDuration msgDelta(ottotime::oMilliSeconds(500)); // Half a second
    const ottotime::otimeDuration msgStartDelta(ottotime::oSeconds(3));
    RecRadians ang;
    ang.setDegrees(0.0);
    av::msg::EventLoggerMessage msg;
    msg.otimeStamp = av::timePointFromOtimePoint(msgTime - msgDelta);
    msg.location = av::geometry::RecPose3D{515.1985,
                                           1023.1989,
                                           314.1529,
                                           av::geometry::RecRadians{ang},
                                           av::geometry::RecRadians{ang},
                                           av::geometry::RecRadians{ang},
                                           static_cast<av::geometry::EULER_ANGLE_ORDER>(RecPose3D::XYZ)};
    msg.utmPlane = 1;
    msg.vehSpeed = 88.1;
    msg.timeSinceStart = "00:01:23";
    msg.feature = "Test Feature";
    msg.passOrFail = "UNKNOWN";
    msg.passOrFailMode = "Not Appicable";
    msg.comments = "Singular dummy message";
    msg.vehSpeedUnit = "mph";
    msg.locationUnit = "ft'";
    msg.resend = false;
    msg.startTime = av::timePointFromOtimePoint(msgTime - msgStartDelta);
    msg.drivername = "Fred Flintstone";
    msg.passengername = "Barney Ruble";

    // Testing that we can write without failure
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(dataFile);
        ASSERT_NO_THROW(logWriter.publish(msg));
    }

    // Testing data integrity
    OfflineLogReader<av::msg::EventLoggerMessage> logReader(dataFile);
    std::vector<av::msg::EventLoggerMessage> data;
    ASSERT_NO_THROW(data = logReader.getAll());
    ASSERT_EQ(data.size(), 1u);
    av::msg::EventLoggerMessage readMsg = data[0];
    EXPECT_EQ(readMsg.otimeStamp, av::timePointFromOtimePoint(msgTime - msgDelta));
    EXPECT_DOUBLE_EQ(readMsg.location.x, 515.1985);
    EXPECT_DOUBLE_EQ(readMsg.location.y, 1023.1989);
    EXPECT_DOUBLE_EQ(readMsg.location.z, 314.1529);
    EXPECT_EQ(readMsg.location.rot1, av::geometry::RecRadians{ang});
    EXPECT_EQ(readMsg.location.rot2, av::geometry::RecRadians{ang});
    EXPECT_EQ(readMsg.location.rot3, av::geometry::RecRadians{ang});
    EXPECT_EQ(readMsg.location.angleOrder, static_cast<av::geometry::EULER_ANGLE_ORDER>(RecPose3D::XYZ));
    EXPECT_EQ(readMsg.utmPlane, 1);
    EXPECT_DOUBLE_EQ(readMsg.vehSpeed, 88.1);
    EXPECT_EQ(readMsg.timeSinceStart, "00:01:23");
    EXPECT_EQ(readMsg.feature, "Test Feature");
    EXPECT_EQ(readMsg.passOrFail, "UNKNOWN");
    EXPECT_EQ(readMsg.passOrFailMode, "Not Appicable");
    EXPECT_EQ(readMsg.comments, "Singular dummy message");
    EXPECT_EQ(readMsg.vehSpeedUnit, "mph");
    EXPECT_EQ(readMsg.locationUnit, "ft'");
    EXPECT_EQ(readMsg.resend, false);
    EXPECT_EQ(readMsg.startTime, av::timePointFromOtimePoint(msgTime - msgStartDelta));
    EXPECT_EQ(readMsg.drivername, "Fred Flintstone");
    EXPECT_EQ(readMsg.passengername, "Barney Ruble");
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:WritePreODatumVersion
 *
 * @brief Test that we can write to an old log
 *
 * Reading pre-ODATUM logs is no longer supported in the OfflineLogReader
 */
TEST_F(OfflineLogWriter_SimpleUse, DISABLED_WritePreODatumVersion)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_SimpleUse_PreODatum.db");

    av::msg::EventLoggerMessage msg;
    msg.drivername = "Fred Flintstone";

    av::msg::EventLoggerMessage msg2;
    msg2.drivername = "Barney Ruble";

    // Create a log with the old version
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(
            dataFile, LogOpenMode::OVERWRITE, dataLog::PRE_ODATUM_LOG_VERSION);
        ASSERT_NO_THROW(logWriter.publish(msg));
    }

    // Write another message with a new log writer to test the existing log case
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(
            dataFile, LogOpenMode::MODIFY, dataLog::DATA_LOG_VERSION);
        ASSERT_NO_THROW(logWriter.publish(msg2));
    }

    // Now read the messages back.  If the version is wrong then cereal should throw an exception,
    // but we still want to compare on the driver name to be thorough
    OfflineLogReader<av::msg::EventLoggerMessage> logReader(dataFile);
    std::vector<av::msg::EventLoggerMessage> data;
    ASSERT_NO_THROW(data = logReader.getAll());
    ASSERT_EQ(data.size(), 2);
    ASSERT_EQ(data[0].drivername, msg.drivername);
    ASSERT_EQ(data[1].drivername, msg2.drivername);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:TwoWriters
 *
 * @brief Test that we can write to a log with two writers and read back the data
 */
TEST_F(OfflineLogWriter_SimpleUse, TwoWriters)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_SimpleUse_TwoWriters.db");

    av::msg::EventLoggerMessage msg;
    msg.drivername = "Fred Flintstone";

    av::msg::EventLoggerMessage msg2;
    msg2.drivername = "Barney Ruble";

    // Create the log
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(
            dataFile, LogOpenMode::OVERWRITE, dataLog::DATA_LOG_VERSION);
        ASSERT_NO_THROW(logWriter.publish(msg));
    }

    // Write another message
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(
            dataFile, LogOpenMode::MODIFY, dataLog::DATA_LOG_VERSION);
        ASSERT_NO_THROW(logWriter.publish(msg2));
    }

    // Now read the messages back.  If the version is wrong then cereal should throw an exception,
    // but we still want to compare on the driver name to be thorough
    OfflineLogReader<av::msg::EventLoggerMessage> logReader(dataFile);
    std::vector<av::msg::EventLoggerMessage> data;
    ASSERT_NO_THROW(data = logReader.getAll());
    ASSERT_EQ(data.size(), 2);
    ASSERT_EQ(data[0].drivername, msg.drivername);
    ASSERT_EQ(data[1].drivername, msg2.drivername);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test SimpleUse:IgnoreVersionIfExists
 *
 * @brief Test that we don't write different version datums to the same log
 */
TEST_F(OfflineLogWriter_SimpleUse, IgnoreVersionIfExists)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_SimpleUse_IgnoreVersion.db");

    av::msg::EventLoggerMessage msg;
    msg.drivername = "Fred Flintstone";

    av::msg::EventLoggerMessage msg2;
    msg2.drivername = "Barney Ruble";

    // Create a log with the current version
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(
            dataFile, LogOpenMode::OVERWRITE, dataLog::DATA_LOG_VERSION);
        ASSERT_NO_THROW(logWriter.publish(msg));
    }

    // Write another message with a new log writer misconfigured to use an old version
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(
            dataFile, LogOpenMode::MODIFY, dataLog::PRE_ODATUM_LOG_VERSION);
        ASSERT_NO_THROW(logWriter.publish(msg2));
    }

    // Now read the messages back.  If the version is wrong then cereal should throw an exception,
    // but we still want to compare on the driver name to be thorough
    OfflineLogReader<av::msg::EventLoggerMessage> logReader(dataFile);
    std::vector<av::msg::EventLoggerMessage> data;
    ASSERT_NO_THROW(data = logReader.getAll());
    ASSERT_EQ(data.size(), 2);
    ASSERT_EQ(data[0].drivername, msg.drivername);
    ASSERT_EQ(data[1].drivername, msg2.drivername);
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test PerformanceCase:OneThousandWrites
 *
 * @brief Test 1000 non-trivial writes and ensure write accuracy
 */
TEST_F(OfflineLogWriter_IntegrityCase, OneThousandWrites)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_PerformanceCase.db");

    // Make base message for all of the other ones to be based upon
    const ottotime::otimePoint msgTime = ottotime::otimePoint::buildCurrentWallClockTime();
    const ottotime::otimeDuration msgDelta(ottotime::oMilliSeconds(-1));
    const ottotime::otimeDuration msgStartDelta(ottotime::oSeconds(3));
    RecRadians ang;
    ang.setDegrees(0.0);
    double x0 = 515.1985, dx = 1.18;
    double y0 = 1023.1989, dy = 7.93;
    double z0 = 314.1529, dz = -2.16;
    av::msg::EventLoggerMessage baseMsg;
    ottotime::otimePoint tp = ottotime::otimePoint::buildCurrentWallClockTime();
    baseMsg.otimeStamp = av::timePointFromOtimePoint(tp);
    baseMsg.location = av::geometry::RecPose3D{x0,
                                               y0,
                                               z0,
                                               av::geometry::RecRadians{ang},
                                               av::geometry::RecRadians{ang},
                                               av::geometry::RecRadians{ang},
                                               static_cast<av::geometry::EULER_ANGLE_ORDER>(RecPose3D::XYZ)};
    baseMsg.utmPlane = 1;
    baseMsg.vehSpeed = 88.1;
    baseMsg.timeSinceStart = "00:01:23";
    baseMsg.feature = "Test Feature";
    baseMsg.passOrFail = "UNKNOWN";
    baseMsg.passOrFailMode = "Not Appicable";
    baseMsg.comments = "Singular dummy message";
    baseMsg.vehSpeedUnit = "mph";
    baseMsg.locationUnit = "ft'";
    baseMsg.resend = false;
    baseMsg.startTime = av::timePointFromOtimePoint(msgTime - msgStartDelta);
    baseMsg.drivername = "Fred Flintstone";
    baseMsg.passengername = "Barney Ruble";

    // Build all messages from base message
    std::vector<av::msg::EventLoggerMessage> msgVect(1000);
    for (auto msgIter = msgVect.begin(); msgIter != msgVect.end(); ++msgIter) {
        *msgIter = baseMsg;
        tp -= msgDelta;
        baseMsg.otimeStamp = av::timePointFromOtimePoint(tp);
        baseMsg.location.x += dx;
        baseMsg.location.y += dy;
        baseMsg.location.z += dz;
    }

    // TEST: Write integrity
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(dataFile);
        for (auto msg : msgVect) {
            logWriter.publishAtTime(msg, msg.otimeStamp);
        }
    }

    OfflineLogReader<av::msg::EventLoggerMessage> logReader(dataFile);
    std::vector<av::msg::EventLoggerMessage> readVect = logReader.getAll();
    ASSERT_EQ(readVect.size(), 1000u);
    for (std::size_t i = 0; i < 1000; ++i) {
        auto origMsg = msgVect[i];
        auto readMsg = readVect[i];

        // Compare with original message
        ASSERT_EQ(readMsg.otimeStamp, origMsg.otimeStamp);
        ASSERT_DOUBLE_EQ(readMsg.location.x, origMsg.location.x);
        ASSERT_DOUBLE_EQ(readMsg.location.y, origMsg.location.y);
        ASSERT_DOUBLE_EQ(readMsg.location.z, origMsg.location.z);
    }
}

//----------------------------------------------------------------------------------------------------------------------

/*!
 * @test ErantUse:MultipleOpen
 *
 * @brief Ensure that multiple log files are not open to writing
 */
TEST_F(OfflineLogWriter_ErantUse, MultipleOpen)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_MultipleOpen.db");

    // TEST: Two active instances trying to open the same log
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> dummyWriter1(dataFile);
        ASSERT_THROW(OfflineLogWriter<av::msg::EventLoggerMessage> dummyWriter2(dataFile), OfflineLogException);
        // Note: dummyWriter1 closes the log at destruction
    }

    // TEST: Able to open again after previous instance closed
    ASSERT_NO_THROW(OfflineLogWriter<av::msg::EventLoggerMessage> dummyWriter3(dataFile));
}

//----------------------------------------------------------------------------------------------------------------------

TEST_F(OfflineLogWriter_Modes, OverwriteOverwrites)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_Modes_OverwriteOverwrites.db");

    av::msg::EventLoggerMessage msg;
    msg.drivername = "Fred Flintstone";

    av::msg::EventLoggerMessage msg2;
    msg2.drivername = "Barney Ruble";

    // Create a logwith the old version
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(dataFile, LogOpenMode::OVERWRITE);
        ASSERT_NO_THROW(logWriter.publish(msg));
    }

    // Write another message with a new log writer set to overwrite
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(dataFile, LogOpenMode::OVERWRITE);
        ASSERT_NO_THROW(logWriter.publish(msg2));
    }

    // There should only be one message and it should be the second message written
    OfflineLogReader<av::msg::EventLoggerMessage> logReader(dataFile);
    std::vector<av::msg::EventLoggerMessage> data;
    ASSERT_NO_THROW(data = logReader.getAll());
    ASSERT_EQ(data.size(), 1);
    ASSERT_EQ(data[0].drivername, msg2.drivername);
}

//----------------------------------------------------------------------------------------------------------------------

TEST_F(OfflineLogWriter_Modes, ModifyModifies)
{
    auto dataFile = makeTestFileName("OfflineLogWriter_Modes_ModifyModifies.db");

    av::msg::EventLoggerMessage msg;
    msg.drivername = "Fred Flintstone";

    av::msg::EventLoggerMessage msg2;
    msg2.drivername = "Barney Ruble";

    // Create a logwith the old version
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(dataFile, LogOpenMode::OVERWRITE);
        ASSERT_NO_THROW(logWriter.publish(msg));
    }

    // Write another message with a new log writer set to modify
    {
        OfflineLogWriter<av::msg::EventLoggerMessage> logWriter(dataFile, LogOpenMode::MODIFY);
        ASSERT_NO_THROW(logWriter.publish(msg2));
    }

    // Both messages should be present
    OfflineLogReader<av::msg::EventLoggerMessage> logReader(dataFile);
    std::vector<av::msg::EventLoggerMessage> data;
    ASSERT_NO_THROW(data = logReader.getAll());
    ASSERT_EQ(data.size(), 2);
    ASSERT_EQ(data[0].drivername, msg.drivername);
    ASSERT_EQ(data[1].drivername, msg2.drivername);
}

//=== MAIN =============================================================================================================

/*!
 * @brief Main executable for OfflineLogWriter's unit tests
 *
 * @param argc : Number of command-line arguments
 * @param argv : Array of c-strings
 * @return int : Executable's error code (should be zero)
 */
int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
