/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \brief Implements a class for profiling execution time of segments of a program
 */

#ifndef OTTOPROFILER_H
#define OTTOPROFILER_H

#include <chrono>
#include <cmath>
#include <iomanip>
#include <iostream>
#include <map>
#include <vector>

// See example in test OttoProfiler.cc
namespace ottoProfiler {

/**
 * \class   WelfordsAlgorithm
 *
 * \brief   Implements Welford's algorithm to calculate standard deviation
 */
class WelfordsAlgorithm
{
public:
    WelfordsAlgorithm() = default;
    virtual ~WelfordsAlgorithm() = default;

    /**
     * \brief       Updates internal state to reflect adding of a new value to the sample set
     * \param[in]   value: A numerical measurement to add to the population
     * \note        Modifies member variables and doesn't return anything.
     */
    void update(const double value);

    /**
     * \brief       Returns number of values added.
     *
     * \return      A uint64_t count of the number of sampled values
     */
    uint64_t getCount() const { return count_; }

    /**
     * \brief       Return variance as calculated by Welford's Method. If fewer than two samples, return 0.
     *
     * \return      Variance as a double.
     */
    double getVariance() const { return (count_ < 2) ? 0 : (intermediate_ / (count_ - 1)); }

    /**
     * \brief       Return standard deviation of the added samples.
     *
     * \return      Standard deviation as a double.
     */
    double getStandardDeviation() const { return sqrt(getVariance()); }

    /**
     * \brief       Return the mean of all added samples.
     *
     * \return      Mean as a double.
     */
    double getMean() const { return mean_; }

    /**
     * \brief       Return the most recent value added.
     *
     * \return      Most recent value added, as a double.
     *
     * \note        Not part of Welford's Method but it seems cleaner to store all the stats in one place.
     */
    double getLast() const { return last_; }

private:
    uint64_t count_{0};      ///< Number of values already added
    double intermediate_{0}; ///< Intermediate value from Welford's Algorithm. Used to find variance.
    double mean_{0};         ///< Mean of all values added
    double last_{0};         ///< Most recent value added
};

using my_clock = std::chrono::time_point<std::chrono::high_resolution_clock>;
using my_time = std::chrono::duration<double, std::nano>;
using mapType = std::map<std::string, WelfordsAlgorithm>;

// Used as extern variable
// TODO: Can create problem when NameTimeMap is in a shared library
static mapType NameTimeMap;

/**
 * \class   ottoProfiler
 *
 * \brief   Records time on construction, measures and records elapsed time on destruction
 */
class ottoProfiler
{
public:
    /**
     * \brief       Constructor.
     * \param[in]   profiledName: A name for the code section to be profiled
     * \note        Records time of construction to determine how long until destruction.
     */
    ottoProfiler(const std::string& profiledName);

    /**
     * \brief       Destructor.
     * \detail      Calculates elapsed time since constructed, then adds that measurement to the map.
     */
    ~ottoProfiler();

private:
    std::string profiledName_; ///< Name of the profiled section
    my_clock consTime_;        ///< Section start time
};

/**
 * \brief       Print profiled statistics in a readable format.
 */
void displayProfiler();

/**
 * \brief       Print the most recently added value for each profiled section
 */
void showLastRecorded();

/**
 * \brief       Returns a copy of the map containing all profiling data
 *
 * \return      A map with all current profiling data
 */
mapType getTimeMap();

} // namespace ottoProfiler

// Profile a scope
// This macro has to put at the first line of a scope
// TODO: how to make sure the name is unique?
#ifdef __ENABLE_OTTO_PROFILER__
#define __OTTO_PROFILER__(SCOPE_NAME) ottoProfiler::ottoProfiler OTTO_PROFILER_NAME_XYZ(#SCOPE_NAME);
#else
#define __OTTO_PROFILER__(SCOPE_NAME)
#endif

// Profile a scope
// Input is a string
#ifdef __ENABLE_OTTO_PROFILER__
#define __OTTO_PROFILER_STRING_NAME__(STRING) ottoProfiler::ottoProfiler OTTO_PROFILER_NAME_XYZ(STRING);
#else
#define __OTTO_PROFILER_STRING_NAME__(STRING)
#endif

// Display results
#ifdef __ENABLE_OTTO_PROFILER__
#define __OTTO_PROFILER_DISPLAY__                                                                                      \
    extern ottoProfiler::mapType NameTimeMap;                                                                          \
    ottoProfiler::displayProfiler();
#else
#define __OTTO_PROFILER_DISPLAY__
#endif

#endif // OTTOPROFILER_H
