/*!
 ***********************************************************
 * @file processUtil.h
 *
 * @brief file defining ProcessUtil class
 *
 * This class is collection of useful utility functions
 * that are re-used within the process launching system.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 **********************************************************
 */

/**********************************************************
  Copyright (c) 2014- 2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * processUtil.h
 * Jason Ziglar
 * 2007/02/10
 *
 * Junsung Kim
 * 2012/06/12
 * killTaskByName() is added to kill possible zombie tasks
 *  before the scenario is launched.
 *
 * Van Chan Ngo
 * 09/25/2018
 * isAliveTaskByName() is added to support checking backup
 * tasks before a recovery command is sent
 *
 * 09/27/2018
 * killTaskByName() is modified to kill possible zombie tasks
 * based on their instance names
 *
 * killTaskByExecutableName() is added to kill possible zombie tasks
 * based on their executable file names
 *
 * 09/28/2018
 * processesByName() is added to factor isAliveTaskByName(), killTaskByName(),
 * and killTaskByExecutableName()
 *
 * Copyright (C) 2006
 * Carnegie Mellon University
 * Developed under contract for ____________
 * All rights reserved
 *
 * LIMITED RIGHTS:
 * The Government is granted Limited Rights to this Data.
 *            Use, duplication, or disclosure is subject to the
 *            restrictions as stated in Agreement _______
 *            between Carnegie Mellon University and the Government.
 */

#ifndef _PROCESSUTIL_H_
#define _PROCESSUTIL_H_

// C++ headers
#include <dirent.h>
#include <errno.h>
#include <list>
#include <signal.h>
#include <string>
#include <sys/stat.h>
#include <sys/types.h>

// Infrastructure specific headers
#include <TimeStamp/otime.h>

/**
 * @class ProcessUtil
 *
 * @brief Set of useful utility functions that are re-used within the process launching system.
 *
 */
class ProcessUtil
{
public:
    // This class is just a collection of static functions. Never instantiate it.
    ProcessUtil() = delete;

    /**
     * \brief This method executes command on bash shell
     *
     * \param command - this is the command to be executed
     *
     * \return returns true on successful execution, false if it fails
     */
    static bool executeCommand(const std::string& command);

    /**
     * \brief returns timestamp derived from SystemTime
     *
     * \param timeOfDayOnly determines whether we need time of the day only or whole date format inlcuding YYYY/MM/DD
     *
     * \return returns TimeStamp string based on parameter timeOfDayOnly
     */
    static std::string getTimeStamp(const bool timeOfDayOnly);

    /*!
     * \brief returns timestamp dervied from SystemTime. This uses otimePoint as argument to the function.
     *
     * \param timeOfDayOnly determines whether we need time of the day only or whole date format inlcuding YYYY/MM/DD
     *
     * \param time otimePoint which is derived from SystemTime
     *
     * \return TimeStamp string based on parameter timeOfDayOnly
     */
    static std::string getTimeStamp(const bool timeOfDayOnly, ottotime::otimePoint time);

    /*!
     * \brief finds any running process such that the instanceName argument is specified by the given name
     *
     * \param name instanceName argument
     *
     * \return returns a list of process PIDs
     */
    static std::list<pid_t> processesByName(const std::string& name, const bool isFirstOccurrence);

    /*!
     * \brief returns true if there exists any running task(process) by the given instanceName
     *
     * \param name instanceName argument
     *
     * \return returns true if there is any such running task, false otherwise
     */
    static bool isAliveTaskByName(const std::string& name);

    /*!
     * \brief finds any running task(process) by the given instanceName and sends kill signal to it
     *
     * \param name instanceName argument
     *
     * \return returns true on successfully sending kill signal to the process, false otherwise
     */
    static bool killTaskByName(const std::string& name);

    /*!
     * \brief finds any running task(process) by the given name and sends kill signal to it
     *
     * \param name name of the executable file name
     *
     * \return returns true on successfully sending kill signal to the process, false otherwise
     */
    static bool killTaskByExecutableName(const std::string& name);
};
#endif
