/*!
 ***********************************************************
 * @file processUtil.cc
 *
 * @brief file implementing ProcessUtil class
 *
 * This class is collection of useful utility functions
 * that are re-used within the process launching system.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * processUtil.cc
 * Jason Ziglar
 * 2007/02/10
 *
 * Junsung Kim
 * 2012/06/12
 * killTaskByName() is added to kill possible zombie tasks
 *  before the scenario is launched.
 *
 * Van Chan Ngo
 * 09/25/2018
 * isAliveTaskByName() is added to support checking backup
 * tasks before a recovery command is sent
 *
 * 09/27/2018
 * killTaskByName() is modified to kill possible zombie tasks
 * based on their instance names
 *
 * killTaskByExecutableName() is added to kill possible zombie tasks
 * based on their executable file names
 *
 * 09/28/2018
 * processesByName() is added to factor isAliveTaskByName(), killTaskByName(),
 * and killTaskByExecutableName()
 *
 * Copyright (C) 2006
 * Carnegie Mellon University
 * Developed under contract for ____________
 * All rights reserved
 *
 * LIMITED RIGHTS:
 * The Government is granted Limited Rights to this Data.
 *            Use, duplication, or disclosure is subject to the
 *            restrictions as stated in Agreement _______
 *            between Carnegie Mellon University and the Government.
 */

#include <Clock/SystemClock.h>
#include <ProcessLauncher/processUtil.h>
#include <utility/utility.h>

#include "lib/avlog/avlog.hpp"

// for std::getline()
#include <cstring>
#include <fstream>

using namespace std;

bool ProcessUtil::executeCommand(const string& command)
{
    AVLOG_DEBUG("Executing command: {}", command);
    errno = 0;
    bool retValue(true);
    int systemresult(system(command.c_str()));

    if (systemresult != 0 && (systemresult != -1 && errno != ECHILD)) {
        AVLOG_ERROR(
            "Execution failed for the following command: {}, return code: {}", command, WEXITSTATUS(systemresult));
        retValue = false;
    } else {
        AVLOG_DEBUG("Execution of {} reports success.", command);
    }

    return retValue;
}

string ProcessUtil::getTimeStamp(const bool timeOfDayOnly)
{
    ottotime::otimePoint time(ottotime::SystemClock::getSystemTime());
    if (!time.isTimeValid()) {
        time.getCurrentWallClockTime();
    }
    return getTimeStamp(timeOfDayOnly, time);
}

string ProcessUtil::getTimeStamp(const bool timeOfDayOnly, ottotime::otimePoint time)
{
    struct tm today = time.to_tm();

    int year = today.tm_year + 1900;
    int month = today.tm_mon + 1;
    int day = today.tm_mday;
    int hour = today.tm_hour;
    int minute = today.tm_min;
    int second = today.tm_sec;
    constexpr int maxStringSize(64);

    char timestamp[maxStringSize];
    if (timeOfDayOnly) {
        snprintf(timestamp, maxStringSize, "%02u:%02u:%02u", hour, minute, second);
    } else {
        snprintf(timestamp, maxStringSize, "%04u.%02u.%02u.%02u.%02u.%02u", year, month, day, hour, minute, second);
    }

    string timeStamp(timestamp);
    return timeStamp;
}

/*
 * if there is any running process then
 * there exist a folder [PID] in /proc
 * We will examine the folder /proc/[PID]/cmdline
 * to get the task name
 * the command line format is (note that parameters are divided by NULL bytes )
 * executable_path --instanceName task_name
 */
list<pid_t> ProcessUtil::processesByName(const std::string& name, const bool isFirstOccurrence)
{
    list<pid_t> processes;
    DIR* d = opendir("/proc");
    struct dirent* f;
    bool firstOccurrenceFlag = false; // it avoids to use a branching statement in the end of the outer loop

    if (d == nullptr) {
        return processes;
    }

    while ((f = readdir(d)) != nullptr && !firstOccurrenceFlag) {
        // ignore folder name with dot prefix
        if (f->d_name[0] == '.') {
            continue;
        }

        // only consider folder names that are numbers
        string folderName(f->d_name);
        bool hasOnlyDigits = (folderName.find_first_not_of("0123456789") == string::npos);
        if (hasOnlyDigits) {
            string cmdlineFileName = "/proc/" + folderName + "/cmdline";
            ifstream cmdlineFile(cmdlineFileName.c_str());
            if (cmdlineFile.is_open()) {
                // read the command line
                string cmdLine;
                while (std::getline(cmdlineFile, cmdLine, '\0')) {
                    if (cmdLine.compare(name) == 0) {
                        processes.push_back(static_cast<pid_t>(stoi(folderName)));
                        firstOccurrenceFlag = isFirstOccurrence ? true : false;
                        break;
                    }
                }
                // close the cmdline file
                cmdlineFile.close();
            }
            // continue working on other directories
        }
    }
    // close the directory
    closedir(d);
    return processes;
}

bool ProcessUtil::isAliveTaskByName(const std::string& name)
{
    // get the first processe whose instanceName argument is name
    list<pid_t> runningProcesses = processesByName(name, true);

    return (!runningProcesses.empty());
}

bool ProcessUtil::killTaskByName(const std::string& name)
{
    bool result = false;
    // get the list of processes whose instanceName arguments are name
    list<pid_t> runningProcesses = processesByName(name, false);

    for (pid_t proc : runningProcesses) {
        // send SIGKILL (9) signal
        kill(proc, SIGKILL);
        result = true;
    }

    return result;
}

bool ProcessUtil::killTaskByExecutableName(const string& name)
{
    // in current setting, executable path is only the executable name
    return killTaskByName(name);
}
