/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief File declaring the progress::report() API.
 */
#ifndef PROGRESSREPORTER_H
#define PROGRESSREPORTER_H

#include <config/ConfigSection.h>
#include <string>

namespace progress {

/// Identifies the type of progress to report.  If you add any more
/// values to this enum, go to ProgressReporter.cc and update
/// progressTypeName().
enum class ProgressType : uint8_t
{
    OS_BOOT_COMPLETE,           // The OS is booted and we can start the stack
    STACK_BOOT_COMPLETE,        // The stack is running
    TASK_INITIALIZED,           // The Task has been initialized.
    TASK_INITIALIZATION_FAILED, // The Task's initialization failed.
    TASK_STARTED,               // The Task has started running.
    PROCESS_STARTED,            // A process (possibly without any Tasks) has started running.
    SERVICE_AVAILABLE,          // The service offered by this process is now running and available to clients.
    TASK_EXITED,                // The Task has stopped running.
    LATENCY_EVENT               // An event logged via LatencyLogger
};

/// This class doesn't have any instances; it's just a collection of
/// static members. It's almost a namespace, but it needed to be a
/// class to keep the s_enabled flag private.
class Reporter
{
public:
    /// No instances, please.
    Reporter() = delete;

    /// Checks the config section for a boolean entry called
    /// "progressEnabled"; if it's true, reporting will be
    /// enabled. Note that, once enabled, reporting cannot be
    /// disabled; that would require atomic access to the s_enabled
    /// flag, which would increase the cost of checking the flag--and
    /// we want to be able to check the flag, quickly, and not cost
    /// much when reporting is disabled. The result is that
    /// initialize() is a no-op if s_enabled is true, and it's not
    /// meant to be called after program initialization.
    ///
    /// \param config The config section to check.
    static void initialize(ConfigSection& config);

    /// Report some progress. Does nothing unless s_enabled has been
    /// set. Since it's inline, this should be cheap to call when
    /// reporting is disabled.
    ///
    /// \param progressType The type of progress to report.
    /// \param parameter A string to include in the progress report.
    static inline void report(const ProgressType progressType, const std::string& parameter = "")
    {
        if (s_enabled) {
            reportInner(progressType, parameter);
        }
    }

private:
    /// Report some progress. Ignores s_enabled. This is the function
    /// that does the work for report().
    ///
    /// \param progressType The type of progress to report.
    /// \param parameter A string to include in the progress report.
    static void reportInner(ProgressType progressType, const std::string& parameter);

    /// Unless this flag is true, report() will do nothing.
    static bool s_enabled;
};

/// A shorthand for Reporter::report(), so that callers can just say
/// "progress::report".
///
/// \param progressType The type of progress to report.
/// \param parameter A string to include in the progress report.
inline void report(const ProgressType progressType, const std::string& parameter = "")
{
    Reporter::report(progressType, parameter);
}

} // namespace progress

#endif // PROGRESSREPORTER_H
