/*!
 ***********************************************************
 * @file BasePeer.cc
 * @brief A file defining the BasePeer class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim
 * @date 03/09/2016
 * @remark
 * Added the cross-platform support
 *  -- 32-bit and big-endian
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * BasePeer.cc
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#include "BasePeer.h"
#include <SimpleComms/ControlMessages.h>
#include <Threading/PThreadLocker.h>
#include <interfaces/baseTypes/Header.h>

#include <algorithm>
#include <cerrno>
#include <cstdio>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>

#include <SimpleComms/TimeStamp.h>

using namespace std;

namespace SimpleComms {

BasePeer::BasePeer(const Transports::TransportTypes& transport,
                   const bool supportsPartialWrites,
                   const TimeStamp& staleFragmentPruningThreshold)
    : transport_(transport)
    , supportsPartialWrites_(supportsPartialWrites)
    , staleFragmentPruningThreshold_(staleFragmentPruningThreshold)
    , staleThresholdValid_(staleFragmentPruningThreshold != TimeStamp())
    , fd_(-1)
    , serverData_(new ServerData(this))
{
    pthread_mutex_init(&clientsMutex_, NULL);
}

BasePeer::~BasePeer()
{
    pthread_mutex_destroy(&clientsMutex_);
}

bool BasePeer::isSubscribedTo(const string& channel)
{
    PThreadLocker clientsLock(clientsMutex_);

    ChannelData tmp;
    tmp.name = channel;

    std::vector<ChannelData>::iterator iter =
        lower_bound(channels_.begin(), channels_.end(), tmp, ChannelDataSortByName());
    return iter != channels_.end() && iter->name == channel && !iter->subscribers.empty();
}

void BasePeer::getSubscriptions(std::vector<std::string>& channels)
{
    PThreadLocker clientsLock(clientsMutex_);

    for (vector<ChannelData>::iterator i = channels_.begin(); i != channels_.end(); ++i) {
        if (!(i->subscribers.empty())) {
            channels.push_back(i->name);
        }
    }
}

void BasePeer::getPublishedChannels(std::vector<std::string>& channels)
{
    PThreadLocker clientsLock(clientsMutex_);
    for (vector<ChannelData>::iterator i = channels_.begin(); i != channels_.end(); ++i) {
        if (!(i->publishers.empty())) {
            channels.push_back(i->name);
        }
    }
}

const std::string& BasePeer::getName()
{
    return peerName_;
}

bool BasePeer::isConnectedTo(const std::string& clientName)
{
    PThreadLocker clientsLock(clientsMutex_);
    bool isConnected = false;

    ClientList::iterator clientIter = clients_.end();
    if (clientLookup(clientName, clientIter)) {
        if ((*clientIter)->fd != -1) {
            isConnected = true;
        }
    }

    return isConnected;
}

Transports::TransportTypes BasePeer::getTransport()
{
    return transport_;
}

Status::States BasePeer::queueMsg(const MessagePtr& msg, BaseClientList& clients, const std::set<std::string>& skip)
{
    if (fd_ == -1) {
        return Status::NotConnected;
    }

    Status::States queueMsgStatus = Status::Ok;

    static const TimeStamp staleMsgReportPeriod{2, 0};
    static TimeStamp lastStaleReport{0, 0};
    static uint32_t staleMsgTotalCnt{0};
    static uint32_t staleMsgItrCount{0};

    const string channel(msg->channel);
    AVLOG_DEBUG("{} to {}", msg->origin, channel);

    if (channel == SIMPLECOMMS_CONTROL_CHANNEL) {
        PThreadLocker clientsLock(clientsMutex_);
        ClientList::iterator clientIter(clients_.end());

        if (!skip.count(msg->origin) && clientLookup(msg->origin, clientIter)) {
            if (!(*clientIter)->hasDataToWrite()) {
                clients.push_back(*clientIter);
            }

            (*clientIter)->outBox.push_back(msg);
            (*clientIter)->dataPending = true;
        } else {
            return Status::NotConnected;
        }
    } else {
        PThreadLocker clientsLock(clientsMutex_);
        ChannelData tmp;
        tmp.name = channel;

        auto channelIter = lower_bound(channels_.begin(), channels_.end(), tmp, ChannelDataSortByName());
        if (channelIter == channels_.end()) {
            return Status::Ok;
        }

        ChannelData& currentChannel(*channelIter);
        if (currentChannel.name != channel) {
            return Status::Ok;
        }

        TimeStamp now;
        now.setNow();
        for (ClientList::iterator i = currentChannel.subscribers.begin(); i != currentChannel.subscribers.end(); ++i) {
            ClientData& currentClient(**i);

            if (!skip.count(currentClient.name) && currentClient.fd != -1) {
                AVLOG_DEBUG("{} to {} over {}", msg->origin, currentClient.name, channel);

                const bool hadDataToWrite = currentClient.hasDataToWrite();

                currentClient.outBox.push_back(msg);

                while (staleThresholdValid_ && !currentClient.outBox.empty()) {
                    const auto& sendMsgTime = currentClient.outBox.front()->sendMsgTime;
                    const TimeStamp msgTime{ntohls(sendMsgTime.tv_sec), ntohls(sendMsgTime.tv_usec)};

                    if (now - msgTime > staleFragmentPruningThreshold_) {
                        ++staleMsgItrCount;
                        ++staleMsgTotalCnt;
                        AVLOG_DEBUG(
                            "Message has gone stale. src: {} dest: {} channel: {}. sent at {}, discarding at {}",
                            currentClient.outBox.front()->origin,
                            currentClient.name,
                            channel,
                            msgTime.toString(),
                            now.toString());

                        currentClient.outBox.pop_front();
                    } else {
                        break;
                    }
                }

                currentClient.dataPending =
                    !currentClient.outBox.empty() || (currentClient.currentOutboundMessage.get() != NULL);

                if (!hadDataToWrite && currentClient.hasDataToWrite()) {
                    clients.push_back(*i);
                }
            }
        }

        // report stale messages if state messages on this iteration
        if (staleMsgItrCount > 0 && now - lastStaleReport > staleMsgReportPeriod) {
            AVLOG_ERROR("stale msgs this itr: {}; total: {}; ", staleMsgItrCount, staleMsgTotalCnt);
            staleMsgItrCount = 0;
            lastStaleReport.setNow();
        }
    }

    return queueMsgStatus;
}

Status::States BasePeer::receiveMsg(const BaseClientPtr& clientHandle, MessagePtr& msg)
{
    if (fd_ == -1) {
        return Status::NotConnected;
    }

    if (clientHandle != serverData_) {
        return Status::InvalidHandle;
    }

    unsigned int msgLength;
    const int fionreadStatus = ioctl(fd_, FIONREAD, &msgLength);

    if (fionreadStatus == -1) {
        return Status::Error;
    } else if (msgLength < static_cast<int>(sizeof(MessageHeader))) {
        return Status::Error;
    }

    msg.reset(new (msgLength - static_cast<unsigned int>(sizeof(MessageHeader))) MessageStub());
    const ssize_t recvStatus = recv(fd_, msg.get(), msgLength, MSG_DONTWAIT | MSG_NOSIGNAL);

    if ((recvStatus == -1) && (errno == EAGAIN)) {
        return Status::Ok;
    } else if (recvStatus == -1) {
        return Status::Error;
    }

    if (string(msg->channel) == SIMPLECOMMS_CONTROL_CHANNEL) {
        return handleControlMsg(msg);
    }

    return Status::NewData;
}

Status::States BasePeer::getNamedReadHandles(BaseClientList& readHandles)
{
    readHandles.push_back(serverData_);

    return Status::Ok;
}

Status::States BasePeer::getNamedMonitorErrorHandles(BaseClientList& monitorHandles)
{
    PThreadLocker clientsLock(clientsMutex_);

    for (ClientList::iterator i = clients_.begin(); i != clients_.end(); ++i) {
        ClientData& currentClient(**i);

        if (currentClient.outBox.empty() && currentClient.fd != -1) {
            monitorHandles.push_back(*i);
        }
    }

    return Status::Ok;
}

Status::States BasePeer::getNamedWriteHandles(BaseClientList& writeHandles)
{
    PThreadLocker clientsLock(clientsMutex_);

    for (ClientList::iterator i = clients_.begin(); i != clients_.end(); ++i) {
        ClientData& currentClient(**i);

        if (!currentClient.outBox.empty() && currentClient.fd != -1) {
            writeHandles.push_back(*i);
        }
    }

    return Status::Ok;
}

Status::States BasePeer::doWrites(const BaseClientList& writeHandles)
{
    Status::States doWritesStatus = Status::Ok;

    static const TimeStamp msgErrReportPeriod{2, 0};
    static TimeStamp lastMsgErrReport{0, 0};
    static uint32_t msgErrTotalCnt{0};
    static uint32_t msgErrItrCount{0};

    for (BaseClientList::const_iterator i = writeHandles.begin(); i != writeHandles.end(); ++i) {
        const ClientDataPtr namedClient(std::dynamic_pointer_cast<ClientData>(*i));

        if (namedClient.get() == NULL) {
            doWritesStatus = Status::Error;
            continue;
        }

        pthread_mutex_lock(&clientsMutex_);

        ClientData& currentClient(*namedClient);

        AVLOG_DEBUG("doWrites({}) handling {}", transportName(transport_), currentClient.name);

        if ((currentClient.fd == -1) || !currentClient.hasDataToWrite()) {
            pthread_mutex_unlock(&clientsMutex_);
        } else {
            const int fd = currentClient.fd;
            const TimeStamp fdStamp(currentClient.fdStamp);
            const uint32_t bytesWrittenSoFar = currentClient.bytesWrittenSoFar;

            if (currentClient.currentOutboundMessage == nullptr) {
                if (!currentClient.outBox.empty()) {
                    currentClient.currentOutboundMessage = currentClient.outBox.front();
                } else {
                    doWritesStatus = Status::Error;
                    AVLOG_WARN("client '{}' currentOutboundMessage is null & outBox is empty; bytesWrittenSoFar {}",
                               currentClient.getName(),
                               bytesWrittenSoFar);
                    pthread_mutex_unlock(&clientsMutex_);
                    continue;
                }
            }

            const MessagePtr msg(currentClient.currentOutboundMessage);

            AVLOG_DEBUG(
                "{} to {} over {} ({})", msg->origin, currentClient.name, msg->channel, transportName(transport_));

            pthread_mutex_unlock(&clientsMutex_);

            uint32_t sendOffset = 0, sendLength = msg->MessageLength();
            if (supportsPartialWrites_ && bytesWrittenSoFar) {
                sendOffset = bytesWrittenSoFar;
                sendLength = sendLength - bytesWrittenSoFar;
            }

            // TODO: refactor by moving this block into a separate function (BasePeer::doSend)
            const ssize_t sendStatus =
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast,cppcoreguidelines-pro-bounds-pointer-arithmetic)
                send(fd, reinterpret_cast<const char*>(msg.get()) + sendOffset, sendLength, MSG_NOSIGNAL);
            if (sendStatus < 0) {
                const auto errNum = errno;
                ++msgErrItrCount;
                ++msgErrTotalCnt;
                AVLOG_DEBUG("Error in send(): {} - {} [sendLength={}]", errNum, strerror(errNum), sendLength);
                PThreadLocker clientsLock(clientsMutex_);
                if ((errNum == EAGAIN) || (errNum == EWOULDBLOCK) || (errNum == ENOBUFS)) // recoverable error
                {
                    if (fdStamp == currentClient.fdStamp) {
                        currentClient.readyToWrite = false;
                        currentClient.currentOutboundMessage.reset();
                    }
                } else // non-recoverable error
                {
                    AVLOG_ERROR(
                        "Non-recov err in send(): {} - {} [sendLength={}]", errNum, strerror(errNum), sendLength);
                    if (fdStamp == currentClient.fdStamp) {
                        clientReset(namedClient, true);
                    } else {
                        AVLOG_DEBUG("Error in send() is stale");
                    }
                    doWritesStatus = Status::Error;
                }
            } else // sendStatus >= 0
            {
                const auto nbytesSent = static_cast<uint32_t>(sendStatus);
                if (nbytesSent < sendLength) // Partial write
                {
                    if (supportsPartialWrites_) {
                        PThreadLocker clientsLock(clientsMutex_);

                        if (fdStamp == currentClient.fdStamp) {
                            currentClient.readyToWrite = false;
                            currentClient.bytesWrittenSoFar += nbytesSent;
                        }
                    } else {
                        AVLOG_WARN("Partial write ({} < {}) on {}",
                                   nbytesSent,
                                   msg->MessageLength(),
                                   transportName(transport_));
                        doWritesStatus = Status::Error;
                    }
                } else // Full write
                {
                    PThreadLocker clientsLock(clientsMutex_);

                    if (fdStamp == currentClient.fdStamp) {
                        if (!currentClient.outBox.empty() &&
                            (currentClient.outBox.front().get() == currentClient.currentOutboundMessage.get())) {
                            currentClient.outBox.pop_front();
                        }

                        currentClient.currentOutboundMessage.reset();
                        currentClient.bytesWrittenSoFar = 0;

                        currentClient.dataPending = !currentClient.outBox.empty();
                    }
                }
            }
        }
    }

    // report send msg errors, only report if errors this iteration
    if (msgErrItrCount > 0) {
        TimeStamp now;
        now.setNow();
        if (now - lastMsgErrReport > msgErrReportPeriod) {
            AVLOG_INFO("peer msg send errs this itr: {}; total: {}; ", msgErrItrCount, msgErrTotalCnt);
            msgErrItrCount = 0;
            lastMsgErrReport.setNow();
        }
    }

    return doWritesStatus;
}

Status::States BasePeer::handleErrors(BaseClientList& errorHandles)
{
    Status::States status = Status::Ok;
    bool serverError = false;

    for (BaseClientList::const_iterator i(errorHandles.begin()); i != errorHandles.end(); ++i) {
        PThreadLocker clientsLock(clientsMutex_);
        const BaseClientPtr& baseClientPtr(*i);

        if (baseClientPtr == serverData_) {
            serverError = true;
        } else {
            const ClientDataPtr currentClientPtr(std::dynamic_pointer_cast<ClientData>(baseClientPtr));

            if (currentClientPtr.get()) {
                int err = -1;
                socklen_t len = sizeof(err);

                if (getsockopt(currentClientPtr->fd, SOL_SOCKET, SO_ERROR, &err, &len) < 0) {
                    err = -1;
                }

                if (err != 0) {
                    clientReset(currentClientPtr, true);
                }
            } else {
                status = Status::InvalidHandle;
            }
        }
    }

    if (serverError) {
        disconnect();
        status = connect();
    }

    return status;
}

Status::States BasePeer::handleControlMsg(const MessagePtr& msg)
{
    Status::States status = Status::Ok;

    if (peerName_ == msg->origin) {
        return Status::Ok;
    }

    // Read the message type.
    ControlMessages::ControlMessageTypes msgType =
        static_cast<ControlMessages::ControlMessageTypes>(ntohl(*reinterpret_cast<const uint32_t*>(msg->data)));
    // Make a single connection attempt upon receiving a Connect message.
    if (msgType == ControlMessages::Connect) {
        const ConnectionMessage* connMsg = reinterpret_cast<ConnectionMessage*>(msg->data);

        if ((Transports::TransportTypes)ntohl(connMsg->transport) == transport_) {
            PThreadLocker clientsLock(clientsMutex_);
            ClientList::iterator clientIter(clients_.end());

            if (clientLookup(msg->origin, clientIter, true)) {
                if (clientConnect(*clientIter, connMsg)) {
                    (*clientIter)->fdStamp.setNow();
                    status = Status::NewData;
                } else {
                    clients_.erase(clientIter);
                }
            }
        } else {
            AVLOG_WARN("Unexpected transport ({}) specified in connection message from {}",
                       ntohl(connMsg->transport),
                       msg->origin);
        }
    } else if (msgType == ControlMessages::Disconnect) {
        const DisconnectionMessage* connMsg = reinterpret_cast<DisconnectionMessage*>(msg->data);

        if ((Transports::TransportTypes)ntohl(connMsg->transport) == transport_) {
            PThreadLocker clientsLock(clientsMutex_);
            ClientList::iterator clientIter(clients_.end());

            if (clientLookup(msg->origin, clientIter)) {
                ClientDataPtr namedClient(*clientIter);

                clients_.erase(clientIter);

                clientReset(namedClient);
            }
        } else {
            AVLOG_WARN("Unexpected transport specified in disconnection message from {}", msg->origin);
        }
    } else if (msgType == ControlMessages::Declare) {
        PThreadLocker clientsLock(clientsMutex_);

        const DeclareChannelMessage* declareMessage = reinterpret_cast<const DeclareChannelMessage*>(msg->data);
        ClientList::iterator clientIter = clients_.end();

        if (clientLookup(msg->origin, clientIter)) {
            const ClientDataPtr namedClient(*clientIter);

            bool subscribed = false;
            bool advertised = false;

            ChannelData emptyChannel = {declareMessage->channel, {}, {}};
            vector<ChannelData>::iterator i =
                lower_bound(channels_.begin(), channels_.end(), emptyChannel, ChannelDataSortByName());

            const auto channelType = static_cast<ChannelType>(ntohl(declareMessage->channelType));

            if (channelType != ChannelType::Unlisted && (channels_.end() == i || i->name != declareMessage->channel)) {
                i = channels_.insert(i, emptyChannel);
            }

            ChannelData& currentChannel(*i);
            if (channelType == ChannelType::Subscriber || channelType == ChannelType::SubPub) {
                ClientList::iterator j =
                    lower_bound(currentChannel.subscribers.begin(), currentChannel.subscribers.end(), *clientIter);

                if (currentChannel.subscribers.end() == j) {
                    subscribed = true;
                    currentChannel.subscribers.push_back(*clientIter);
                } else if (*j != *clientIter) {
                    subscribed = true;
                    currentChannel.subscribers.insert(j, *clientIter);
                }
            }

            if (channelType == ChannelType::Publisher || channelType == ChannelType::SubPub) {
                ClientList::iterator j =
                    lower_bound(currentChannel.publishers.begin(), currentChannel.publishers.end(), *clientIter);

                if (currentChannel.publishers.end() == j) {
                    advertised = true;
                    currentChannel.publishers.push_back(*clientIter);
                } else if (*j != *clientIter) {
                    advertised = true;
                    currentChannel.publishers.insert(j, *clientIter);
                }
            }

            if (subscribed) {
                AVLOG_INFO("{} subscribed to {} over {}",
                           namedClient->name,
                           declareMessage->channel,
                           transportName(transport_));
            }

            if (advertised) {
                AVLOG_INFO("{} started publishing {} over {}",
                           namedClient->name,
                           declareMessage->channel,
                           transportName(transport_));
            }

            status = Status::NewData;
        }
    } else if (msgType == ControlMessages::Undeclare) {
        PThreadLocker clientsLock(clientsMutex_);

        const DeclareChannelMessage* declareMessage = reinterpret_cast<const DeclareChannelMessage*>(msg->data);
        ClientList::iterator clientIter = clients_.end();

        if (clientLookup(msg->origin, clientIter)) {
            const ClientDataPtr namedClient(*clientIter);
            ChannelData emptyChannel = {declareMessage->channel, {}, {}};
            vector<ChannelData>::iterator i =
                lower_bound(channels_.begin(), channels_.end(), emptyChannel, ChannelDataSortByName());

            const auto channelType = static_cast<ChannelType>(ntohl(declareMessage->channelType));

            if (i != channels_.end() && i->name == declareMessage->channel) {
                ChannelData& currentChannel(*i);
                if (channelType == ChannelType::Subscriber || channelType == ChannelType::SubPub) {
                    ClientList::iterator j =
                        lower_bound(currentChannel.subscribers.begin(), currentChannel.subscribers.end(), *clientIter);

                    if (j != currentChannel.subscribers.end() && *j == *clientIter) {
                        currentChannel.subscribers.erase(j);

                        AVLOG_INFO("{} unsubscribed from {} over {}",
                                   namedClient->name,
                                   declareMessage->channel,
                                   transportName(transport_));
                    }
                }

                if (channelType == ChannelType::Publisher || channelType == ChannelType::SubPub) {
                    ClientList::iterator j =
                        lower_bound(currentChannel.publishers.begin(), currentChannel.publishers.end(), *clientIter);

                    if (j != currentChannel.publishers.end() && *j == *clientIter) {
                        currentChannel.publishers.erase(j);
                        AVLOG_INFO("{} stopped publishing {} over {}",
                                   namedClient->name,
                                   declareMessage->channel,
                                   transportName(transport_));
                    }
                }

                if (currentChannel.subscribers.empty() && currentChannel.publishers.empty()) {
                    channels_.erase(i);
                }
            }

            status = Status::NewData;
        }
    } else if (msgType == ControlMessages::Schedule || msgType == ControlMessages::Executive ||
               msgType == ControlMessages::Reproducible || msgType == ControlMessages::Throttle) {
        // Pass this message on to the MessageControl object, which will handle it.
        status = Status::NewData;
    } else if (msgType == ControlMessages::Drop) {
        const DropMessage& dropMessage = *reinterpret_cast<const DropMessage*>(msg->data);
        AVLOG_WARN("Task {} dropped message {} from {}", msg->origin, ntohl(dropMessage.id), dropMessage.origin);
        status = Status::NewData;
    }

    return status;
}

BasePeer::ClientDataPtr BasePeer::clientCreate(const string& name)
{
    ClientDataPtr tmp(new ClientData(this, name));

    return tmp;
}

/**
 * @brief Find or create the desired client entry in the clientsList_
 *
 * @param name String identifier of the client
 * @param namedClient Iterator to return entry in
 *
 * @return True if an existing entry was found or a blank one was created
 *
 * Assumes clientsMutex_ is held by calling function
 */
bool BasePeer::clientLookup(const string& name, ClientList::iterator& namedClient, const bool create)
{
    ClientDataPtr tmp(clientCreate(name));

    namedClient = lower_bound(clients_.begin(), clients_.end(), tmp, ClientDataSortByName());

    bool exists = true;
    if (namedClient == clients_.end()) {
        exists = false;
    } else if ((*namedClient)->name != name) {
        exists = false;
    }

    if (!exists) {
        if (create) {
            tmp->fd = -1;

            AVLOG_DEBUG("Adding new client {}", name);

            namedClient = clients_.insert(namedClient, tmp);
        } else {
            AVLOG_DEBUG("Client {} not found", name);
            namedClient = clients_.end();
        }
    }

    return (namedClient != clients_.end());
}

/**
 * @brief Deinitialize and remove traces of a client
 *
 * @param namedClient A pointer to the client to reset
 *
 * Assumes clientsMutex_ is held by calling function
 * The system call close may block with clientsMutex_ held
 */
void BasePeer::clientReset(const ClientDataPtr& namedClient, const bool retireEntry)
{
    ClientData& currentClient(*namedClient);

    AVLOG_DEBUG("Reseting client {}", currentClient.name);

    close(currentClient.fd);
    currentClient.fd = -1;

    currentClient.fdStamp = TimeStamp();

    currentClient.outBox.clear();
    currentClient.currentOutboundMessage.reset();

    currentClient.dataPending = false;

    if (retireEntry) {
        ClientList::iterator clientIter(clients_.end());

        if (clientLookup(currentClient.name, clientIter)) {
            clients_.erase(clientIter);
        }
    }

    vector<ChannelData> oldChannels(channels_);
    channels_.clear();
    for (vector<ChannelData>::iterator i = oldChannels.begin(); i != oldChannels.end(); ++i) {
        ChannelData& currentChannel(*i);
        ClientList::iterator j =
            lower_bound(currentChannel.subscribers.begin(), currentChannel.subscribers.end(), namedClient);

        if (j != currentChannel.subscribers.end()) {
            if (*j == namedClient) {
                currentChannel.subscribers.erase(j);
            }
        }

        j = lower_bound(currentChannel.publishers.begin(), currentChannel.publishers.end(), namedClient);

        if (j != currentChannel.publishers.end() && *j == namedClient) {
            currentChannel.publishers.erase(j);
        }

        if (!currentChannel.subscribers.empty() || !currentChannel.publishers.empty()) {
            channels_.push_back(currentChannel);
        }
    }
}

} // namespace SimpleComms
