/*!
 ***********************************************************
 * @file BasePeer.h
 * @brief A file defining the BasePeer class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 03/23/2018
 * @remark
 * removing boost dependency
 * @n
 *
 * @author Junsung Kim
 * @date 03/09/2016
 * @remark
 * Added the cross-platform support
 *  -- 32-bit and big-endian
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * BasePeer.h
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#ifndef _SIMPLECOMMS_BASEPEER_H_
#define _SIMPLECOMMS_BASEPEER_H_

#include <list>
#include <pthread.h>

#include "Peer.h"
#include "lib/avlog/avlog.hpp"
#include <SimpleComms/ControlMessages.h>
#include <SimpleComms/TimeStamp.h>

#include <set>
#include <string>

namespace SimpleComms {

class BasePeer : public Peer
{
public:
    BasePeer(const Transports::TransportTypes& transport,
             const bool supportsPartialWrites,
             const TimeStamp& staleFragmentPruningThreshold);
    virtual ~BasePeer();

    virtual bool isSubscribedTo(const std::string& channel);
    virtual void getSubscriptions(std::vector<std::string>& channels);
    virtual void getPublishedChannels(std::vector<std::string>& channels);

    virtual const std::string& getName();
    virtual Status::States connect() = 0;
    virtual bool isConnectedTo(const std::string& clientName);
    virtual Status::States disconnect() = 0;

    virtual Transports::TransportTypes getTransport();

    virtual Status::States queueMsg(const MessagePtr& msg, BaseClientList& clients, const std::set<std::string>& skip);
    virtual Status::States receiveMsg(const BaseClientPtr& clientHandle, MessagePtr& msg);

    virtual Status::States getNamedReadHandles(BaseClientList& readHandles);
    virtual Status::States getNamedMonitorErrorHandles(BaseClientList& monitorHandles);
    virtual Status::States getNamedWriteHandles(BaseClientList& writeHandles);

    virtual Status::States doWrites(const BaseClientList& writeHandles);

    virtual Status::States handleErrors(BaseClientList& errorHandles);

    virtual Status::States handleControlMsg(const MessagePtr& msg);

protected:
    const Transports::TransportTypes transport_;
    const bool supportsPartialWrites_;
    const TimeStamp staleFragmentPruningThreshold_;
    const bool staleThresholdValid_;
    std::string peerName_;
    int fd_;

    class ServerData : public BaseClientType
    {
    public:
        ServerData(BasePeer* bp)
            : bp_(bp)
        {}
        virtual ~ServerData() {}

        virtual Peer* getParentPeer() { return bp_; }
        virtual std::string getName() { return bp_->peerName_; }
        virtual int getReadHandle() { return bp_->fd_; }
        virtual int getWriteHandle() { return -1; }
        virtual bool isReadyToWrite() { return false; }
        virtual void setReadyToWrite() {}
        virtual bool hasDataToWrite() { return false; }
        virtual void setOnWriteQueue() {}
        virtual void setOffWriteQueue() {}
        virtual bool isOnWriteQueue() { return false; }

    private:
        BasePeer* bp_;
    };
    std::shared_ptr<ServerData> serverData_;
    class ClientData : public BaseClientType
    {
    public:
        ClientData(BasePeer* bp, const std::string& _name)
            : name(_name)
            , fd(-1)
            , readyToWrite(false)
            , bytesWrittenSoFar(0)
            , dataPending(false)
            , onWriteQueue(false)
            , bp_(bp)
        {}
        virtual ~ClientData() {}

        const std::string name;
        int fd;
        volatile bool readyToWrite;
        TimeStamp fdStamp;
        std::list<MessagePtr> outBox;
        MessagePtr currentOutboundMessage;
        uint32_t bytesWrittenSoFar;
        volatile bool dataPending;
        volatile bool onWriteQueue;

        virtual Peer* getParentPeer() { return bp_; }
        virtual std::string getName() { return name; }
        virtual int getReadHandle() { return -1; }
        virtual int getWriteHandle() { return fd; }
        virtual bool isReadyToWrite() { return readyToWrite; }
        virtual void setReadyToWrite() { readyToWrite = true; }
        virtual bool hasDataToWrite() { return dataPending; }
        virtual void setOnWriteQueue() { onWriteQueue = true; }
        virtual void setOffWriteQueue() { onWriteQueue = false; }
        virtual bool isOnWriteQueue() { return onWriteQueue; }

    private:
        BasePeer* bp_;
    };

    typedef std::shared_ptr<ClientData> ClientDataPtr;

    typedef std::vector<ClientDataPtr> ClientList;

    struct ClientDataSortByName : public std::binary_function<ClientDataPtr, ClientDataPtr, bool>
    {
        bool operator()(const ClientDataPtr a, const ClientDataPtr b) { return a->name < b->name; }
    };

    ClientList clients_;
    pthread_mutex_t clientsMutex_;

    struct ChannelData
    {
        std::string name;
        ClientList subscribers;
        ClientList publishers;
    };

    struct ChannelDataSortByName : public std::binary_function<ChannelData, ChannelData, bool>
    {
        bool operator()(const ChannelData& a, const ChannelData& b) { return a.name < b.name; }
    };

    std::vector<ChannelData> channels_;

    virtual ClientDataPtr clientCreate(const std::string& name);
    virtual bool clientLookup(const std::string& name, ClientList::iterator& namedClient, const bool create = false);
    virtual bool clientConnect(const ClientDataPtr& namedClient, const ConnectionMessage* connMsg) = 0;
    virtual void clientReset(const ClientDataPtr& namedClient, const bool retireEntry = false);
};

} // namespace SimpleComms

#endif
