/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief CallbackSet member function definitions.
 */
#include "CentralScheduler.h"

#include <SimpleComms/ControlMessages.h>
#include <interfaces/baseTypes/Header.h>

#include <algorithm>
#include <cstring>

namespace SimpleComms {

namespace {
// Use the steady clock, but with Unix epoch timestamps.
// The system_clock epoch on Linux is the Unix epoch. (This will become
// part of the standard in C++2a, but is an "unofficial" standard already.)
static const std::chrono::nanoseconds EPOCH_OFFSET =
    std::chrono::system_clock::now().time_since_epoch() - std::chrono::steady_clock::now().time_since_epoch();
// This is the Unix epoch as a steady_clock time point.
static const std::chrono::steady_clock::time_point EPOCH = std::chrono::steady_clock::time_point() - EPOCH_OFFSET;

// Get a monotonic timestamp, in nanoseconds since the Unix epoch.
std::chrono::nanoseconds now()
{
    return EPOCH_OFFSET + std::chrono::steady_clock::now().time_since_epoch();
}
} // namespace

const std::chrono::nanoseconds CentralScheduler::MINIMUM_INTERVAL(100000);

CentralScheduler::CentralScheduler(SendFunction send)
    : m_send(send)
{}

CentralScheduler::~CentralScheduler()
{
    // Kill the thread if it's running.
    freeze();
}

void CentralScheduler::reset()
{
    // Initial state: the queue is empty, the thread is not running, and the rolling ID is 0.
    // Don't change the "frozen" flag; the scheduler should stay in the same mode as before.
    {
        std::unique_lock<std::mutex> lock(m_mutex);
        m_queue = Queue();
    }
    if (m_thread.joinable()) {
        m_thread.join();
    }
    m_rollingID = 0;
}

void CentralScheduler::add(const std::string& task, std::chrono::nanoseconds interval)
{
    // If the interval is invalid, don't register this client at all.
    // Don't throw an error, because we can't afford the possibility that SCS
    // can be crashed by an invalid input from a task.
    if (interval.count() <= 0) {
        return;
    }
    // Don't allow the interval to be less than the minimum interval.
    interval = std::max(interval, MINIMUM_INTERVAL);

    if (m_frozen) {
        // We're using "frozen time" controlled by the log player.
        m_queue.emplace(m_now, interval, task);
    } else {
        // We're using current system time.
        bool mustLaunchThread{false};
        {
            std::unique_lock<std::mutex> lock(m_mutex);
            mustLaunchThread = m_queue.empty();
            m_queue.emplace(now(), interval, task);
        }
        m_condition.notify_all();
        // If the thread isn't running, start it.
        if (mustLaunchThread) {
            // It's possible that the thread was launched previously, then all
            // tasks disconnected, then a new task requested scheduling. In that
            // case there's an old thread that must join.
            if (m_thread.joinable()) {
                m_thread.join();
            }
            m_thread = std::thread(&CentralScheduler::loop, this);
        }
    }
}

void CentralScheduler::freeze()
{
    if (!m_frozen) {
        m_frozen = true;
        // It's possible the thread is running. If so, interrupt it.
        if (m_thread.joinable()) {
            m_condition.notify_all();
            m_thread.join();
        }
        // Every entry in the queue should be reset to the next advance() time.
        // Note: this only happens if the log player is connected after tasks start up.
        // If the user is doing things in the right order, that's never the case.
        m_now = std::chrono::nanoseconds(0);
        Queue newQueue;
        while (!m_queue.empty()) {
            Entry entry = m_queue.top();
            m_queue.pop();
            std::get<0>(entry) = m_now;
            newQueue.push(entry);
        }
        newQueue.swap(m_queue);
    }
}

void CentralScheduler::advance(std::chrono::nanoseconds timestamp)
{
    // Loop until there are no more entries to advance.
    while (advanceOne(timestamp)) {
    }
}

std::chrono::nanoseconds CentralScheduler::timestamp() const
{
    // In free-running mode, return the steady clock time (the same one used for executive scheduling).
    // If connected to a log player, return the most recent advance() time.
    return m_frozen ? m_now : now();
}

void CentralScheduler::loop()
{
    std::unique_lock<std::mutex> lock(m_mutex);
    while (!m_queue.empty() && !m_frozen) {
        // Loop through any entries that ought to have been published by now.
        // Each one may be scheduled with a slightly different timestamp,
        // because time is advancing as we send these messages.
        if (!advanceOne(now())) {
            // Wait until it's time to publish another message. By using a condition
            // variable, it can also be interrupted if new tasks are added.
            m_condition.wait_until(lock, EPOCH + std::get<0>(m_queue.top()));
        }
    }
}

bool CentralScheduler::advanceOne(std::chrono::nanoseconds timestamp)
{
    m_now = timestamp;
    // Check whether there is an entry in the queue that needs to be advanced.
    if (m_queue.empty() || std::get<0>(m_queue.top()) > m_now) {
        return false;
    }
    // Get the next executive that must be triggered.
    Entry entry = m_queue.top();
    m_queue.pop();

    // Create a new control message.
    SimpleComms::MessagePtr msg = messageCreate<ExecutiveMessage>(std::get<2>(entry).c_str(), m_rollingID++);
    // Store the timestamp, which must be in network byte order.
    ExecutiveMessage& message = *reinterpret_cast<ExecutiveMessage*>(msg->data);
    message.tv_sec = htonlls(m_now.count() / NSEC_PER_SEC);
    message.tv_nsec = htonlls(m_now.count() % NSEC_PER_SEC);
    // Unless sending fails, queue this task to be scheduled again.
    if (m_send(msg)) {
        // Special case: this queue entry is not initialized, because it was just added:
        if (!std::get<0>(entry).count()) {
            std::get<0>(entry) = m_now;
        }
        // Advance the time and add this task back to the queue.
        std::get<0>(entry) += std::get<1>(entry);
        m_queue.push(entry);
    }
    return true;
}

} // namespace SimpleComms
