/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Definition of the CallbackSet class.
 */
#pragma once

#include <SimpleComms/Message.h>

#include <atomic>
#include <chrono>
#include <condition_variable>
#include <functional>
#include <mutex>
#include <queue>
#include <string>
#include <thread>
#include <tuple>
#include <vector>

namespace SimpleComms {

/// \brief Class that manages executive function triggers for any number of clients.
class CentralScheduler
{
public:
    /// \brief Type of the function for sending messages.
    typedef std::function<bool(const MessagePtr&)> SendFunction;
    /// \brief Minimum allowable interval between executive funtions.
    static const std::chrono::nanoseconds MINIMUM_INTERVAL;

public:
    /// \brief Constructor, linking the scheduler to the server.
    /// \param send: A function to be used for sending executive triggers.
    explicit CentralScheduler(SendFunction send);
    /// \brief Destructor, which shuts down the scheduling thread if it's running.
    ~CentralScheduler();

    /// Reset to the initial condition (no tasks registered for executives).
    void reset();

    /// \brief Register the given task to receive executive function triggers at the given interval.
    /// \param task: The name of the client that requested executive scheduling.
    /// \param interval: The executive function's period, in nanoseconds.
    /// \note If the interval is less than MINIMUM_INTERVAL, it will be changed to MINIMUM_INTERVAL.
    void add(const std::string& task, std::chrono::nanoseconds interval);

    /// \brief Instead of scheduling based on wall time, wait for advance() calls.
    void freeze();

    /// \brief In frozen mode, advance the clock to the given time.
    /// \param timestamp: The time, in nanoseconds since the epoch, to advance the clock to.
    void advance(std::chrono::nanoseconds timestamp);

    /// \brief Get the current timestamp.
    /// \return A timestamp in nanoseconds since the Unix epoch.
    std::chrono::nanoseconds timestamp() const;

private:
    // Thread entry point.
    void loop();
    // Advance one entry in the queue. Unless frozen, this must be called within a mutex-protected scope.
    bool advanceOne(std::chrono::nanoseconds timestamp);

private:
    // Function used to send a message.
    SendFunction m_send;
    // Thread sending the messages.
    std::thread m_thread;
    // If the scheduler is frozen, wall clock time is not used and the thread is not launched.
    std::atomic<bool> m_frozen{false};
    std::chrono::nanoseconds m_now;
    // Rolling message ID.
    uint32_t m_rollingID{0};
    // Queue storing (next run time, interval, task):
    typedef std::tuple<std::chrono::nanoseconds, std::chrono::nanoseconds, std::string> Entry;
    typedef std::priority_queue<Entry, std::vector<Entry>, std::greater<Entry>> Queue;
    Queue m_queue;
    // Mutex protecting access to the queue.
    std::mutex m_mutex;
    // Condition variable used for waiting and notification.
    std::condition_variable m_condition;
};

} // namespace SimpleComms
