/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief MessageControl function definitions.
 */
#include "MessageControl.h"

#include "lib/avlog/avlog.hpp"
#include <ODatum/ODatum.h>
#include <SimpleComms/ControlMessages.h>
#include <config/TaskParser.h>
#include <interfaces/Clock/Playback/PlaybackClockParams.h>
#include <interfaces/ScsRecordingCommand/ScsRecordingCommand.h>
#include <interfaces/baseTypes/Header.h>

#include <unistd.h>

#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <experimental/filesystem>
#include <fstream>
#include <stdexcept>
namespace fs = std::experimental::filesystem;

namespace SimpleComms {

// Helper functions:
namespace {
// Check if this is published on an internal channel
bool isInternalChannel(const MessagePtr& msg);
// Check if this is a control message.
bool isControl(const MessagePtr& msg);
// Get the type of a control message.
ControlMessages::ControlMessageTypes controlType(const MessagePtr& msg);
// Get the base name of a client, minus the random suffix.
std::string clientName(const std::string& str);
// Make sure that the given origin has the given prefix.
bool originMatches(const char* origin, const std::string& prefix);
// Replace the given message's header timestamp with the given value.
void injectTimestamp(const MessagePtr& msg, int64_t nsec);
// Use this string in the client lookup maps to identify the executive scheduler.
static const std::string EXECUTIVE_ORIGIN{"executive"};

const std::string recordingCommandChannel{"ScsRecordingCommand"};
} // namespace

MessageControl::MessageControl(const std::string& logPath,
                               Mode mode,
                               CentralScheduler::SendFunction send,
                               ConfigSection& params)
    : params_(params)
    , m_scheduler(send)
    , m_send(send)
    , m_dynamicLogging(mode == Mode::LIVE || mode == Mode::FROZEN)
    , m_isLogging(mode == Mode::LOG_LIVE || mode == Mode::LOG_FROZEN)
    , m_isReplaying(mode == Mode::REPLAY)
{
    // If asked to replay, load the previous metadata.
    // Load it first in case we're saving over this same log file.
    if (m_isReplaying) {
        if (!m_log.open(logPath, false)) {
            AVLOG_ERROR("Could not open log for reading: {}", logPath);
        }
        // TODO: Limit how far we "read ahead" in the log to limit memory use.
        SQLMessageLog::Row row;
        // Read all rows, including "subscription" events.
        while (m_log.read(row, true)) {
            m_metadata.emplace_back(std::move(row));
        }
    } else if (m_isLogging) {
        if (!m_log.open(logPath, true)) {
            AVLOG_ERROR("Could not open log for writing: {}", logPath);
        }
    }
    // Check if the scheduler should start out in frozen mode.
    if (mode == Mode::FROZEN || mode == Mode::LOG_FROZEN || mode == Mode::REPLAY) {
        m_scheduler.freeze();
    }
}

bool MessageControl::preprocess(const MessagePtr& msg)
{
    if (msg == nullptr) {
        throw std::invalid_argument("MessageControl::preprocess: empty input message");
    }

    const char* const channelPtr{static_cast<char*>(msg->channel)};
    const std::string channel{channelPtr, strnlen(channelPtr, MAX_NULL_TERM_CHANNEL_NAME)};
    // Determine whether this message should pass on to the next stage in the pipeline.
    bool passMessageOn = true;
    if (isControl(msg)) {
        // These control messages don't need to be echoed.
        ControlMessages::ControlMessageTypes msgType = controlType(msg);
        passMessageOn &= (msgType != ControlMessages::Reproducible);
        passMessageOn &= (msgType != ControlMessages::Schedule);
        passMessageOn &= (msgType != ControlMessages::Drop);
        passMessageOn &= (msgType != ControlMessages::Throttle);
        // External executive triggers are only allowed in replay mode.
        passMessageOn &= (msgType != ControlMessages::Executive || m_isReplaying);
        // Throttling works in both logging and replay mode.
        if (msgType == ControlMessages::Throttle) {
            m_throttledClients.insert(msg->origin);
        } else if (msgType == ControlMessages::Reproducible) {
            m_reproducibleClients.insert(msg->origin);
        }
        // Subscriptions should be handled in either mode.
        else if (msgType == ControlMessages::Declare) {
            const DeclareChannelMessage& message = *reinterpret_cast<const DeclareChannelMessage*>(msg->data);
            ChannelType channelType = static_cast<ChannelType>(ntohl(message.channelType));
            if (channelType == ChannelType::Subscriber || channelType == ChannelType::SubPub) {
                subscribe(msg->origin, message.channel);
            }
        } else if (msgType == ControlMessages::Schedule) {
            subscribe(static_cast<char*>(msg->origin), channel);
        }
    }

    // ScsRecordingCommand is between LOCAL safer and LOCAL scs. Don't forward it.
    passMessageOn &= (channel != recordingCommandChannel);

    if (m_isReplaying) {
        // In replay mode, none of the processing below needs to happen.
        // But, any task that connects does need to know we're in replay mode.
        if (isControl(msg) && controlType(msg) == ControlMessages::Connect) {
            m_send(messageCreate<ReplayMessage>(msg->origin));
        }
    } else if (isControl(msg)) {
        ControlMessages::ControlMessageTypes msgType = controlType(msg);
        if (msgType == ControlMessages::Schedule) {
            // Someone has requested scheduling. Pass the request on to the scheduler.
            const ScheduleMessage& scheduleMessage = *reinterpret_cast<const ScheduleMessage*>(msg->data);
            m_scheduler.add(msg->origin, std::chrono::nanoseconds(scheduleMessage.nsecInterval));
        } else if (msgType == ControlMessages::Drop) {
            // Record the dropped message in the log.
            const DropMessage& dropMessage = *reinterpret_cast<const DropMessage*>(msg->data);
            std::string client = clientName(msg->origin);
            // If the origin field in the packet header matches the supposed
            // origin of the dropped message, it's an executive trigger.
            // Look back in the history to find the row ID of this message in the log.
            std::string origin = originMatches(dropMessage.origin, client) ? EXECUTIVE_ORIGIN : dropMessage.origin;
            uint32_t id = ntohl(dropMessage.id);

            // Use mutex protection just in case this call to preprocess() is happening
            // at the same time the scheduling thread calls queueMsg().
            std::unique_lock<std::mutex> lock(m_logMutex);

            const std::map<uint32_t, uint64_t>& history = m_histories[origin];
            auto it = history.find(id);
            if (it != history.end() && m_isLogging) {
                m_log.writeDropped(static_cast<int64_t>(it->second), client);
            }
        }
    } else if (ntohl(msg->pageNumber) == 1) {
        // This is not a control message, and it's the first packet (the one with the timestamp).

        // Special case: recording commands may cause logging to restart or flush.
        if (channel == recordingCommandChannel) {
            handleRecordingCommand(msg);
        }
        // If this is a log playback message, add the origin to the list of active log players.
        if (channel == "log_playback_avt" || channel == "PlaybackClockParams") {
            m_logPlayers.insert(msg->origin);

            // It's possible that this SCS instance is running on one host, and the log player is
            // on another. In that case, we're guaranteed to get PlaybackClock messages from the
            // log player (because all tasks consume them) but we might not get any other messages.
            // In that case the PlaybackClock is the only way the scheduler gets advanced.
            if (channel == "PlaybackClockParams") {
                // Extract the timestamp from the clock message.
                size_t headerSize = sizeof(interfaces::baseTypes::InterfaceHeader);
                const char* data = reinterpret_cast<const char*>(msg->data + headerSize);
                std::string body(data, ntohl(msg->dataLength) - headerSize);
                PlaybackClockParams params;
                aptv::deserializeODatum(params, body);
                // Advance the scheduler.
                m_scheduler.advance(std::chrono::nanoseconds(params.baseTime.toNanosecondsSinceEpoch()));
            }
        } else {
            // This is an ordinary message. Fill in an up-to-date timestamp.
            injectTimestamp(msg, m_scheduler.timestamp().count());
        }
    }
    return passMessageOn;
}

bool MessageControl::add(const MessagePtr& msg)
{
    // Check if this is an ignored channel or we're not in replay mode.
    if (!m_isReplaying || shouldIgnore(msg)) {
        return false;
    }
    // If this message comes from the log player, it can match any logged message
    // from the same channel regardless of origin.
    bool checkOrigin = !isControl(msg) && !m_logPlayers.count(msg->origin);
    // Find the first entry in the queue that could receive this message.
    for (Metadata& it : m_metadata) {
        // Is this an entry from the same channel?
        if (it.channel != msg->channel) {
            continue;
        }
        // Subscription events are handled separately, in the subscribe() function.
        if (!it.nsec) {
            continue;
        }
        // If this message did not come from the log channel, its origin must be
        // the same as the origin that was recorded in the log.
        if (checkOrigin && !originMatches(msg->origin, it.origin)) {
            continue;
        }
        // Has this entry already received the expected number of pages?
        if (!it.pagesRemaining) {
            continue;
        }
        // This message is from the right channel (and origin, if applicable).
        if (it.pagesRemaining < 0) {
            // This is the first page from this message. Store the expected number of pages.
            it.pagesRemaining = static_cast<int32_t>(ntohl(msg->totalPages));
            incrementBacklog();
        }
        // Record the fact that this page has been seen, and enqueue it.
        --it.pagesRemaining;
        it.pages.push(msg);
        return true;
    }
    // If we've reached the end of the log, don't allow any further messages to be published.
    return true;
}

MessagePtr MessageControl::pop()
{
    while (!m_metadata.empty()) {
        Metadata& it = m_metadata.front();
        if (it.pagesRemaining < 0) {
            // If we have received no pages for this message, there's nothing to do right now.
            break;
        } else if (!it.pages.empty()) {
            // This message has a page that's ready to send.
            MessagePtr next = it.pages.front();
            it.pages.pop();
            // The list of dropped clients should be the union of the "unready" clients and the
            // ones that dropped this message the first time around.
            std::set<std::string>& dropped = m_messageToDropped[next];
            for (const std::string& client : it.dropped) {
                dropped.insert(findOrigin(client));
            }
            for (const auto& client : m_clientReady[it.channel]) {
                if (!client.second) {
                    dropped.insert(findOrigin(client.first));
                }
            }
            // If this is an executive message, figure out the new name of its origin.
            if (isControl(next)) {
                const std::string& remappedOrigin = findOrigin(it.origin);
                if (!remappedOrigin.empty()) {

                    // note: The character count and the final index happen to be the same number
                    std::size_t const toBeCopied =
                        std::min(remappedOrigin.size(), static_cast<std::size_t>(MAX_NULL_TERM_CLIENT_NAME - 1));

                    std::strncpy(
                        next->origin, remappedOrigin.c_str(), toBeCopied); // Not guaranteed to append a null-terminator
                    next->origin[toBeCopied] = '\0';
                }
            }
            // Copy the original timestamp into the packet.
            injectTimestamp(next, it.nsec);
            return next;
        } else if (!it.pagesRemaining) {
            // If this is a subscription event, the client is now able to receive messages.
            if (!it.nsec) {
                m_clientReady[it.channel][it.origin] = true;
            }
            // If we've sent out every page of the current message, move on to the next.
            m_metadata.pop_front();
            decrementBacklog();
        } else {
            // If no pages are ready to publish and we haven't published them all, we need to wait.
            break;
        }
    }
    // If we reached the end of the list or no messages are ready, return null to signal that.
    return MessagePtr();
}

void MessageControl::log(const MessagePtr& msg)
{
    // The first packet of each message is the one we care about, because that
    // is where the timestamp is stored.
    if (!m_isLogging || shouldIgnore(msg)) {
        return;
    }
    // (origin, rolling ID) is a message's unique key. But, for executive messages
    // the "origin" field actually stored the message destination. So, in that
    // case replace it with a special origin string for the executive scheduler.
    std::string origin = (isControl(msg) ? EXECUTIVE_ORIGIN : msg->origin);
    uint32_t id = ntohl(msg->rollingMessageID);
    // Check how many packets this message is split into and which one this is.
    // Page numbers begin at 1, not 0.
    uint32_t pageNumber = ntohl(msg->pageNumber);
    uint32_t totalPages = ntohl(msg->totalPages);
    // Make sure the page number is feasible.
    if (totalPages <= 1 && pageNumber != 1) {
        AVLOG_ERROR("Invalid message page number: {} / {}", pageNumber, totalPages);
        return;
    }
    // This timestamp will be set in one of the conditions below.
    int64_t nsec = 0;
    if (pageNumber == 1) {
        // The first packet contains the timestamp.
        int64_t tv_sec{0};
        int64_t tv_nsec{0};
        if (isControl(msg)) {
            // We're ignoring everything other than executive messages, so that
            // must be what this one is.
            const ExecutiveMessage& executiveMessage = *reinterpret_cast<const ExecutiveMessage*>(msg->data);
            tv_sec = executiveMessage.tv_sec;
            tv_nsec = executiveMessage.tv_nsec;
        } else {
            typedef interfaces::baseTypes::InterfaceHeader Header;
            const Header& header = *reinterpret_cast<const Header*>(msg->data);
            tv_sec = header.sendTimeStamp.tv_sec;
            tv_nsec = header.sendTimeStamp.tv_nsec;
        }
        // Convert to host byte order, then calculate nsec since the epoch.
        // A signed 64-bit nanosecond count won't overflow until the year 2260.
        nsec = ntohlls(tv_sec) * NSEC_PER_SEC + ntohlls(tv_nsec);
        // Don't bother adding "pending" entries for single-packet messages.
        if (totalPages > 1) {
            m_pending[origin][id].nsec = nsec;
        }
    }
    if (totalPages > 1) {
        // If the message was split into more than one packet, count how many we
        // have seen and bail out if there are still packets missing.
        Pending& pending = m_pending[origin][id];
        ++pending.pages;
        if (pending.pages < totalPages) {
            return;
        }
        // All packets have been received. Use the timestamp from the first one.
        nsec = pending.nsec;
        // We no longer have to store the pending record for this message.
        m_pending[origin].erase(id);
    }

    // Use mutex protection just in case this call to log() is happening inside the scheduling
    // thread at the same time that Server is preprocessing a message.
    std::unique_lock<std::mutex> lock(m_logMutex);
    // It's possible that m_isLogging got cleared between the start of the function and here.
    if (!m_isLogging) {
        return;
    }

    const auto client = clientName(&(msg->origin[0]));

    // Write this line of the log.
    const auto row = m_log.writeMessage(nsec, client, &(msg->channel[0]));
    if (row) {
        // Remember recently logged lines in case we need to mark one of them as dropped.
        std::map<uint32_t, uint64_t>& history = m_histories[origin];
        // In case we're running for a very long time, limit the history length.
        static const size_t MAX_HISTORY = 8000;
        if (m_historyOrder.size() == MAX_HISTORY) {
            m_historyOrder.front().first.erase(m_historyOrder.front().second);
            m_historyOrder.pop();
        }
        m_historyOrder.emplace(history, id);
        history[id] = static_cast<uint64_t>(*row);
    } else {
        AVLOG_INFO("MessageControl::log(): writeMessage() failed to write to client {} for channel {}",
                   client,
                   &(msg->channel[0]));
    }
}

std::set<std::string> MessageControl::dropped(const MessagePtr& msg)
{
    if (msg == nullptr) {
        throw std::invalid_argument("MessageControl::dropped: empty input message");
    }
    std::set<std::string> dropped;
    auto it = m_messageToDropped.find(msg);
    if (it != m_messageToDropped.end()) {
        dropped = std::move(it->second);
        m_messageToDropped.erase(it);
    }
    return dropped;
}

void MessageControl::subscribe(const std::string& origin, const std::string& channel)
{
    // We don't need to log or reproduce subscriptions from tasks that need not be reproducible.
    if (!m_reproducibleClients.count(origin) || shouldIgnoreChannel(channel)) {
        return;
    }
    // Only count the first time any client subscribes to a given channel.
    const std::string client = clientName(origin);
    std::map<std::string, bool>& ready = m_clientReady[channel];
    if (ready.count(client)) {
        return;
    }
    ready[client] = false;

    if (m_isLogging) {
        // Use mutex protection just in case this call to preprocess() is happening
        // at the same time the scheduling thread calls queueMsg().
        std::unique_lock<std::mutex> lock(m_logMutex);

        // Record the order in which this client subscribed to this channel.
        // It's possible m_isLogging changed in between when we checked it and when we locked the mutex.
        if (m_isLogging) {
            m_log.writeMessage(0, client, channel);
        }
    } else if (m_isReplaying) {
        // Find this message in the queue and mark it as done. This client will be marked ready
        // when we reach that line of the log.
        for (Metadata& it : m_metadata) {
            // Note: above, we made sure that we only log the first subscription from each client.
            // So, at most there is one row of the metadata that matches.
            if (!it.nsec && it.origin == client && it.channel == channel) {
                // Mark this message as received.
                it.pagesRemaining = 0;
                break;
            }
        }
    }
}

bool MessageControl::shouldIgnore(const MessagePtr& msg) const
{
    if (msg == nullptr) {
        return true;
    }
    // Ignore control messages other than executive function triggers.
    if (isControl(msg)) {
        return (controlType(msg) != ControlMessages::Executive);
    }
    return shouldIgnoreChannel(msg->channel);
}

bool MessageControl::shouldIgnoreChannel(const std::string& channel) const
{
    if (m_isReplaying) {
        // If in replay mode, ignore any channels that are not in the log.
        return !m_log.hasChannel(channel);
    } else {
        // Otherwise, ignore certain messages that are only for logging and status reporting.
        static std::set<std::string> ignoredChannels;
        if (ignoredChannels.empty()) {
            std::vector<std::string> channels;
            params_.getArray("ignoredChannels", channels);
            ignoredChannels.insert(channels.begin(), channels.end());
            ignoredChannels.insert("PublishListMessage");
            ignoredChannels.insert(SIMPLECOMMS_STATS_CHANNEL);
        }
        return ignoredChannels.count(channel);
    }
}

void MessageControl::incrementBacklog()
{
    static const int PAUSE_THRESHOLD = 1000;
    if (++m_backlog == PAUSE_THRESHOLD && !m_paused) {
        m_paused = true;
        for (const std::string& client : m_throttledClients) {
            m_send(messageCreate<PauseMessage>(client.c_str()));
        }
    }
}

void MessageControl::decrementBacklog()
{
    static const int RESUME_THRESHOLD = 200;
    if (--m_backlog == RESUME_THRESHOLD && m_paused) {
        m_paused = false;
        for (const std::string& client : m_throttledClients) {
            m_send(messageCreate<ResumeMessage>(client.c_str()));
        }
    }
}

void MessageControl::handleRecordingCommand(const MessagePtr& msg)
{
    // King commands do nothing unless we're in a mode where we need to start a new
    // messages log whenever a new safer scenario is launched.
    if (!m_dynamicLogging) {
        return;
    }
    // Extract the body of the message.
    size_t headerSize = sizeof(interfaces::baseTypes::InterfaceHeader);
    const char* data = reinterpret_cast<const char*>(msg->data + headerSize);
    std::string body(data, ntohl(msg->dataLength) - headerSize);
    // Deserialize it.
    av::msg::ScsRecordingCommand command;
    aptv::deserializeODatum(command, body);

    // If the scenario is stopping, stop message logging and flush to disk.
    // If it's starting, begin logging to a new file.
    if (command.command == av::msg::ScsRecordingCommandType::Stop && m_isLogging) {
        AVLOG_INFO("Stopped logging messages in response to an ScsRecordingCommand Stop command.");
        std::unique_lock<std::mutex> lock(m_logMutex);
        m_log.flush();
        m_isLogging = false;
    } else if (command.command == av::msg::ScsRecordingCommandType::Start) {
        // Get the path to the log directory.
        std::string logDir;
        params_.get("logDataDir", logDir, "/tmp/logs");
        logDir += "/" + command.logName;
        // Get the hostname.
        char hostname[64] = "unknown";
        if (gethostname(hostname, sizeof(hostname)) < 0) {
            AVLOG_ERROR("Unable to read the host name.");
        }
        // Make sure the log directory exists.
        std::error_code ec;
        fs::create_directories(logDir, ec);
        if (ec) {
            AVLOG_ERROR("Unable to create log directory \"{}\".", logDir);
            return;
        }
        // Begin logging to a new messages.db file.
        std::string messagesPath = logDir + "/messages@" + hostname + ".db";
        AVLOG_INFO("Logging messages to \"{}\" in response to an ScsRecordingCommand Start command.", messagesPath);
        std::unique_lock<std::mutex> lock(m_logMutex);
        if (!m_log.open(messagesPath, true)) {
            AVLOG_ERROR("Could not open message log {} for writing", messagesPath);
        }
        // Reset all the objects for tracking status of messages in the log.
        // Assume that all tasks were shut down and must be started up again.
        // (This includes the log player.)
        m_histories.clear();
        m_historyOrder = HistoryQueue();
        m_pending.clear();
        m_backlog = 0;
        m_paused = false;
        m_throttledClients.clear();
        m_logPlayers.clear();
        m_reproducibleClients.clear();
        m_scheduler.reset();
        m_clientReady.clear();
        m_messageToDropped.clear();
        // Begin allowing logging again.
        m_isLogging = true;
    }
}

const std::string& MessageControl::findOrigin(const std::string& prefix) const
{
    auto it = m_reproducibleClients.lower_bound(prefix);
    if (it == m_reproducibleClients.end() || it->compare(0, prefix.size(), prefix)) {
        static const std::string EMPTY;
        return EMPTY;
    }
    return *it;
}

// Helper functions:
namespace {

bool isInternalChannel(const MessagePtr& msg)
{
    // Check that the channel name begins with SIMPLECOMMS_INTERNAL_TAG
    static const std::string internalTag{SIMPLECOMMS_INTERNAL_TAG};
    return !internalTag.compare(0, std::string::npos, msg->channel, internalTag.length());
}

bool isControl(const MessagePtr& msg)
{
    static const std::string CHANNEL = SIMPLECOMMS_CONTROL_CHANNEL;
    return !CHANNEL.compare(msg->channel);
}

ControlMessages::ControlMessageTypes controlType(const MessagePtr& msg)
{
    uint32_t value = *reinterpret_cast<const uint32_t*>(msg->data);
    return static_cast<ControlMessages::ControlMessageTypes>(ntohl(value));
}

std::string clientName(const std::string& str)
{
    // Origin names have the format "ClientName-%%%%%%%%%%", where the "%" are random characters.
    return str.substr(0, str.rfind('-'));
}

bool originMatches(const char* origin, const std::string& prefix)
{
    return !prefix.compare(0, prefix.size(), origin, prefix.size());
}

void injectTimestamp(const MessagePtr& msg, int64_t nsec)
{
    // The timestamp is only on the first page of messages.
    if (ntohl(msg->pageNumber) != 1) {
        return;
    }
    // Make sure this is a valid timestamp.
    if (nsec <= 0) {
        AVLOG_ERROR("Invalid message timestamp: {}", nsec);
        return;
    }
    int64_t tv_sec = htonlls(nsec / NSEC_PER_SEC);
    int64_t tv_nsec = htonlls(nsec % NSEC_PER_SEC);
    if (isControl(msg)) {
        // This is a control message. Check if it's an executive message; if not, ignore it.
        ControlMessages::ControlMessageTypes msgType = controlType(msg);
        if (msgType != ControlMessages::Executive) {
            return;
        }
        ExecutiveMessage& message = *reinterpret_cast<ExecutiveMessage*>(msg->data);
        message.tv_sec = tv_sec;
        message.tv_nsec = tv_nsec;
    } else if (!isInternalChannel(msg)) {
        // Messages published using Interfaces apply an InterfaceHeader, which starts with a timestamp
        // stored as two int64_t values.
        reinterpret_cast<int64_t*>(msg->data)[0] = tv_sec;
        reinterpret_cast<int64_t*>(msg->data)[1] = tv_nsec;
    }
}

} // namespace

} // namespace SimpleComms
