/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Definition of the MessageControl class.
 */
#pragma once

#include "CentralScheduler.h"

#include <SimpleComms/Message.h>
#include <SimpleComms/SQLMessageLog.h>
#include <config/ConfigSection.h>

#include <atomic>
#include <deque>
#include <fstream>
#include <map>
#include <mutex>
#include <queue>
#include <set>
#include <utility>

namespace SimpleComms {

/// \brief Class for logging and controlling the order of messages passing through scs.
class MessageControl
{
public:
    /// \brief Possible operating modes for reproducible resim.
    enum class Mode
    {
        LIVE,       ///< Default mode: no logging and no message order control.
        FROZEN,     ///< Synchronize executive functions with the log player.
        LOG_LIVE,   ///< Log messages, and schedule executive functions based on the steady clock.
        LOG_FROZEN, ///< Log messages, and synchronize executive functions with the log player.
        REPLAY      ///< Read a previous message log and force messages to follow that same order.
    };

public:
    /// \brief Constructor, specifying a log file path and operating mode.
    /// \param logPath: Path to the log file to open for logging or replay.
    /// \param mode: Reproducible mode.
    /// \param send: Function for queing a message to be sent.
    MessageControl(const std::string& logPath, Mode mode, CentralScheduler::SendFunction send, ConfigSection& params);

    /// \brief Preprocess a message to extract timestamp, control messages, etc.
    /// \param msg: A message for this object to examine.
    /// \throw std::invalid_argument on empty message pointer.
    /// \return True if the message should be queued for sending.
    bool preprocess(const MessagePtr& msg);
    /// \brief Add the given message to the replay queue.
    /// \param msg: A message to potentially be queued up internally.
    /// \throw std::invalid_argument on empty message pointer.
    /// \return False if the message was not added to the queue and can be published immediately.
    bool add(const MessagePtr& msg);
    /// \brief Get the next message that should be sent, or null if none are ready.
    /// \return The next message to send, or a null pointer if no unblocked messages are ready.
    MessagePtr pop();

    /// \brief If logging, save metadata for the given message.
    /// \param msg: A message.
    /// \throw Throws std::invalid_argument on empty message pointer.
    void log(const MessagePtr& msg);

    /// \brief Get the names of clients that dropped this message in the original run.
    /// \param msg: A message.
    /// \return A set of strings identifying any clients that should not receive this message.
    std::set<std::string> dropped(const MessagePtr& msg);

private:
    // Handle a client declaring it will receive messages on the given channel.
    void subscribe(const std::string& origin, const std::string& channel);
    // Check if the ordering of the given message does not matter.
    bool shouldIgnore(const MessagePtr& msg) const;
    // Check if the given channel should be ignored.
    bool shouldIgnoreChannel(const std::string& channel) const;
    // Increment the backlog counter, and send the log player a Pause message if necessary.
    void incrementBacklog();
    // Decrement the backlog counter, and send the log player a Resume message if necessary.
    void decrementBacklog();
    // Handle ScsRecordingCommands, which may reset or flush the messages log.
    void handleRecordingCommand(const MessagePtr& msg);
    // Get the full origin name given a client prefix.
    const std::string& findOrigin(const std::string& prefix) const;

private:
    // Configuration.
    ConfigSection& params_;

    // Message log to read or write.
    SQLMessageLog m_log;
    // Mapping of [origin][rolling id] to rows in the database.
    std::map<std::string, std::map<uint32_t, uint64_t>> m_histories;
    using HistoryQueue = std::queue<std::pair<std::map<uint32_t, uint64_t>&, uint32_t>>;
    HistoryQueue m_historyOrder;
    // Mutex protecting access to the above members, because of the scheduling thread calling queueMsg().
    std::mutex m_logMutex;

    // Timestamps and page counts of pending messages to log.
    struct Pending
    {
        int64_t nsec{0};
        uint32_t pages{0};
    };
    std::map<std::string, std::map<uint32_t, Pending>> m_pending;

    // Metadata that must be tracked for each message to control replay order.
    struct Metadata : public SQLMessageLog::Row
    {
        // Allow move-constructing from a SQLMessageLog::Row.
        Metadata() = default;
        explicit Metadata(SQLMessageLog::Row&& row)
            : SQLMessageLog::Row(std::move(row))
        {}
        // Packets from this message that have been received but are not ready to send.
        std::queue<std::shared_ptr<MessageStub>> pages;
        // How many pages have not yet been received, or -1 if we don't know how many, yet.
        int pagesRemaining{-1};
    };
    // Message metadata read from the log.
    std::deque<Metadata> m_metadata;

    // Track how many "blocked" packets are queued up.
    int m_backlog{0};
    // Have clients currently been told to pause?
    bool m_paused{false};
    // Set of clients that are able to be throttled.
    std::set<std::string> m_throttledClients;
    // Set of clients that have identified themselves as log players.
    std::set<std::string> m_logPlayers;
    // Record of which clients are using the reproducible framework.
    std::set<std::string> m_reproducibleClients;

    // Helper object for scheduling the executive function.
    CentralScheduler m_scheduler;
    // Function pointer for sending messages.
    CentralScheduler::SendFunction m_send;

    // For each channel, store which clients have subscribed but should not
    // receive messages yet (because in the original run, they had not yet
    // subscribed at this point). Use a map to bool because the entries can have
    // three possible states: not yet seen, unready, or ready.
    std::map<std::string, std::map<std::string, bool>> m_clientReady;
    // Map storing which clients dropped a given message.
    std::map<MessagePtr, std::set<std::string>> m_messageToDropped;

    // Operating mode.
    const bool m_dynamicLogging{false};
    std::atomic<bool> m_isLogging{false};
    const bool m_isReplaying{false};
};

} // namespace SimpleComms
