/*!
 ***********************************************************
 * @file NetworkIntfUtils.cc
 * @brief A file implementing the network interface utility
 * functions.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 03/23/2018
 * @remark
 * removing boost dependency
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * NetworkIntfUtils.cc
 * Tugrul Galatali
 * 03/10/2007
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */
#include "NetworkIntfUtils.h"

#include <arpa/inet.h>
#include <iostream>
#include <memory>
#include <net/if.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <unistd.h>

namespace NetworkIntfUtils {

bool GetInterfaceAddresses(IntfNameAddrMap& intfs)
{
    int sockfd;

    if (0 > (sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP))) {
        return false;
    }

    struct ifconf ifConf;

    ifConf.ifc_req = NULL;
    ifConf.ifc_len = 0;

    if (ioctl(sockfd, SIOCGIFCONF, &ifConf)) {
        close(sockfd);
        return false;
    }
    std::unique_ptr<unsigned char[]> ifcReqBuf(new unsigned char[ifConf.ifc_len]);
    ifConf.ifc_req = reinterpret_cast<ifreq*>(ifcReqBuf.get());

    if (ioctl(sockfd, SIOCGIFCONF, &ifConf)) {
        close(sockfd);
        return false;
    }

    if (ifConf.ifc_len >= 0) {
        const size_t N = static_cast<size_t>(ifConf.ifc_len) / sizeof(ifreq);
        for (size_t i = 0; i < N; i++) {
            struct ifreq& ifr(ifConf.ifc_req[i]);

            intfs[ifr.ifr_name] = ntohl(reinterpret_cast<struct sockaddr_in*>(&ifr.ifr_addr)->sin_addr.s_addr);
        }
    }

    close(sockfd);

    return true;
}

bool GetInterfaceParameter(const std::string& name, const ParameterTypes& type, unsigned int& value)
{
    unsigned int ioctlNumber;

    switch (type) {
        case Address:
            ioctlNumber = SIOCGIFADDR;
            break;

        case Netmask:
            ioctlNumber = SIOCGIFNETMASK;
            break;

        case Broadcast:
            ioctlNumber = SIOCGIFBRDADDR;
            break;

        default:
            return false;
    }

    int sockfd;

    if (0 > (sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP))) {
        return false;
    }

    struct ifreq ifr;

    strncpy(&(ifr.ifr_name[0]), name.c_str(), IFNAMSIZ - 1);
    ifr.ifr_name[IFNAMSIZ - 1] = 0;

    if (ioctl(sockfd, ioctlNumber, &ifr)) {
        close(sockfd);
        return false;
    }

    value = ntohl(reinterpret_cast<struct sockaddr_in*>(&ifr.ifr_addr)->sin_addr.s_addr);

    close(sockfd);

    return true;
}

int setSocketBufferSize(const int sockfd, const std::size_t bufSize)
{
    int rcvbuf = static_cast<int>(bufSize);
    if (setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &rcvbuf, sizeof(rcvbuf)) != 0) {
        perror("NetworkIntfUtils::setSocketBufferSize setsockopt SO_RCVBUF failed");
        std::cerr << "sockfd: " << sockfd << " bufSize: " << bufSize << std::endl;
        return -1;
    }
    int sndbuf = static_cast<int>(bufSize);
    if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, &sndbuf, sizeof(sndbuf)) != 0) {
        perror("NetworkIntfUtils::setSocketBufferSize setsockopt SO_SNDBUF failed");
        std::cerr << "sockfd: " << sockfd << " bufSize: " << bufSize << std::endl;
        return -1;
    }
    return 0;
}

} // namespace NetworkIntfUtils
