/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Helper class for wrapping the scs message log
 *        See https://confluence.ci.motional.com/confluence/display/AUTRUN/SQL+format+messaging+log
 **/

#include "lib/avlog/avlog.hpp"
#include <SimpleComms/SQLMessageLog.h>

#include <sqlite3.h>
#include <unistd.h>

#include <condition_variable>
#include <cstring>
#include <functional>
#include <future>
#include <mutex>
#include <queue>
#include <thread>
#include <unordered_map>

namespace {

// I am not bothering to set up the foreign key ralationships here because the logs are supposed to be append only

constexpr const char* CREATE_MESSAGES =
    "CREATE TABLE MESSAGES (timestamp INTEGER, originId INTEGER, channelId INTEGER)";
constexpr const char* CREATE_DROPPED = "CREATE TABLE DROPPED (droppedMessageId INTEGER, droppedOriginId INTEGER)";
constexpr const char* CREATE_ORIGINS = "CREATE TABLE ORIGINS (origin TEXT)";
constexpr const char* CREATE_CHANNELS = "CREATE TABLE CHANNELS (channel TEXT)";

constexpr const char* INSERT_MESSAGES = "INSERT INTO MESSAGES (timestamp, originId, channelId) VALUES (?, ?, ?)";
constexpr const char* INSERT_ORIGINS = "INSERT INTO ORIGINS (origin) VALUES (?)";
constexpr const char* INSERT_CHANNELS = "INSERT INTO CHANNELS (channel) VALUES (?)";
constexpr const char* INSERT_DROPPED = "INSERT INTO DROPPED (droppedMessageId, droppedOriginId) VALUES (?, ?)";

constexpr const char* SELECT_ALL_ORIGINS = "SELECT * FROM ORIGINS ORDER BY rowid";
constexpr const char* SELECT_ALL_CHANNELS = "SELECT * FROM CHANNELS ORDER BY rowid";

// We could get rid of these joins but it is plenty fast enough as is
constexpr const char* SELECT_MESSAGES =
    "SELECT MESSAGES.rowid, MESSAGES.timestamp, ORIGINS.origin, CHANNELS.channel, group_concat(DESTORIGINS.origin) "
    "FROM MESSAGES "
    "INNER JOIN ORIGINS ON MESSAGES.originId = ORIGINS.rowid "
    "INNER JOIN CHANNELS ON MESSAGES.channelId = CHANNELS.rowid "
    "LEFT JOIN DROPPED ON MESSAGES.rowid = DROPPED.droppedMessageId "
    "LEFT JOIN ORIGINS AS DESTORIGINS ON DROPPED.droppedOriginId = DESTORIGINS.rowid "
    "GROUP BY MESSAGES.rowid";

} // namespace

namespace SimpleComms {

// helper type to safely store sqlite3_stmts
class StatementDeleter
{
public:
    void operator()(::sqlite3_stmt* x) const { ::sqlite3_finalize(x); }
};

using Statement = std::unique_ptr<sqlite3_stmt, StatementDeleter>;

class SQLMessageLog::Impl
{
private:
    enum class RowType
    {
        Message,
        Dropped
    };
    struct BufferedRow
    {
        RowType type;
        union
        {
            int64_t nsec;
            int64_t messageId;
        };
        int originId;
        int channelId;
    };

public:
    Impl() = default;
    ~Impl();

    // Returns false on error.
    bool open(const std::string& databasePath, bool write, size_t buffer);

    std::optional<int64_t> writeMessage(int64_t nsec, const std::string& origin, const std::string& channel);
    void writeDropped(int64_t messageId, const std::string& client);
    bool read(SQLMessageLog::Row& data);

    // Returns false on failure.
    bool flush(bool wait = true);

    void stopWriteThread();

    bool hasChannel(const std::string& channel) const;

private:
    int getOriginId(const std::string& origin);
    int getChannelId(const std::string& channel);
    Statement prepareStatement(const std::string& statement);

    // Throw an error if the result of an SQL function is an error.
    void trySQL(int result, const std::string& info = "");

    void setupDatabase();
    void readMapping(std::unordered_map<std::string, int>& map, const char* query);
    int getId(const std::string& name, std::unordered_map<std::string, int>& map, const char* insert);

    void writeBuffer(const std::vector<SQLMessageLog::Impl::BufferedRow>& rows);
    void writeThreadFunction();

private:
    std::unordered_map<std::string, int> originToId_;
    std::unordered_map<std::string, int> channelToId_;

    sqlite3* logDb_{nullptr};
    Statement insertMessageStatement_{nullptr};
    Statement insertDroppedStatement_{nullptr};
    Statement readMessagesStatement_{nullptr};
    bool readDone_{false};
    bool writing_{false};

    std::vector<BufferedRow> unflushedRows_;
    size_t maxBufferSize_{500};
    uint64_t nextMessageId_{1};

    std::queue<std::function<void()>> databaseTaskQueue_;
    std::condition_variable taskQueueNotEmpty_;
    std::mutex writeThreadMutex_;
    std::thread writeThread_;
    bool shuttingDown_{false};
};

Statement SQLMessageLog::Impl::prepareStatement(const std::string& statement)
{
    sqlite3_stmt* stmt{nullptr};

    trySQL(::sqlite3_prepare_v2(logDb_, statement.c_str(), -1, &stmt, nullptr), statement);

    return Statement{stmt};
}

void SQLMessageLog::Impl::trySQL(int result, const std::string& info)
{
    if ((result != SQLITE_OK) && (result != SQLITE_DONE) && (result != SQLITE_ROW)) {
        std::string errorMessage{::sqlite3_errstr(result)};
        throw std::runtime_error("SQLMessageLog failed: " + errorMessage + info);
    }
}

bool SQLMessageLog::Impl::open(const std::string& databasePath, bool write, size_t buffer)
{
    // If the database is already open, flush and close it.
    stopWriteThread();
    readMessagesStatement_.reset();
    if (logDb_ != nullptr) {
        ::sqlite3_close_v2(logDb_);
        logDb_ = nullptr;
    }
    // Open the database file for reading or writing.
    writing_ = write;
    originToId_.clear();
    channelToId_.clear();
    int flags{0};
    if (write) {
        // Delete the given file if it already exists.
        unlink(databasePath.c_str());
        flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE;
    } else {
        flags = SQLITE_OPEN_READONLY;
    }

    try {
        trySQL(::sqlite3_open_v2(databasePath.c_str(), &logDb_, flags, nullptr));

        // Some pragmas for performance: don't fsync, and use memory for temporary tables.
        trySQL(sqlite3_exec(logDb_, "PRAGMA synchronous = 0;", nullptr, nullptr, nullptr));
        trySQL(sqlite3_exec(logDb_, "PRAGMA temp_store = 2;", nullptr, nullptr, nullptr));

        if (write) {
            setupDatabase();
            insertMessageStatement_ = prepareStatement(INSERT_MESSAGES);
            insertDroppedStatement_ = prepareStatement(INSERT_DROPPED);
            maxBufferSize_ = buffer;
            unflushedRows_.reserve(maxBufferSize_);
            nextMessageId_ = 1;
        } else {
            // Read-only mode. Read the mappings so we can query which channels and origins exist.
            readMapping(originToId_, SELECT_ALL_ORIGINS);
            readMapping(channelToId_, SELECT_ALL_CHANNELS);
            readMessagesStatement_ = prepareStatement(SELECT_MESSAGES);
            readDone_ = false;
        }
    } catch (const std::runtime_error& e) {
        AVLOG_ERROR("SQLMessageLog::Impl::open(): {}", e.what());
        return false;
    }

    return true;
}

void SQLMessageLog::Impl::setupDatabase()
{
    for (const auto& s : {CREATE_MESSAGES, CREATE_DROPPED, CREATE_ORIGINS, CREATE_CHANNELS}) {
        trySQL(::sqlite3_exec(logDb_, s, nullptr, nullptr, nullptr));
    }
}

void SQLMessageLog::Impl::readMapping(std::unordered_map<std::string, int>& map, const char* query)
{
    // The results of the query are ordered by ROWID, and row IDs are always
    // sequential beginning with 1.
    auto processRows = [](void* instance, int, char** columnData, char**) -> int {
        auto& m = *static_cast<std::unordered_map<std::string, int>*>(instance);
        int id{static_cast<int>(m.size()) + 1};
        m[columnData[0]] = id;
        return 0;
    };
    trySQL(::sqlite3_exec(logDb_, query, processRows, &map, nullptr));
}

int SQLMessageLog::Impl::getId(const std::string& name, std::unordered_map<std::string, int>& map, const char* insert)
{
    const auto it = map.find(name);
    if (it != map.end()) {
        return it->second;
    }

    // This name isn't in the table yet, so we need to insert it.
    Statement statement = prepareStatement(insert);
    trySQL(::sqlite3_bind_text(statement.get(), 1, name.c_str(), -1, SQLITE_TRANSIENT));

    trySQL(::sqlite3_step(statement.get()));
    // The tables are always written from scratch, and rows are never deleted.
    // So, the rowId is equal to the number of items we've inserted into the table.
    int rowId = static_cast<int>(map.size() + 1);
    map[name] = rowId;
    return rowId;
}

bool SQLMessageLog::Impl::hasChannel(const std::string& channel) const
{
    return channelToId_.count(channel);
}

/** \brief getOrignId - looks up the originId for a string. Creates a new record if it doesn't already exist
 **/
int SQLMessageLog::Impl::getOriginId(const std::string& origin)
{
    return getId(origin, originToId_, INSERT_ORIGINS);
}

int SQLMessageLog::Impl::getChannelId(const std::string& channel)
{
    return getId(channel, channelToId_, INSERT_CHANNELS);
}

SQLMessageLog::Impl::~Impl()
{
    if (writing_) {
        stopWriteThread();
    }
    if (logDb_) {
        sqlite3_close_v2(logDb_);
    }
}

std::optional<int64_t> SQLMessageLog::Impl::writeMessage(int64_t nsec,
                                                         const std::string& origin,
                                                         const std::string& channel)
{
    if (!writing_ || !logDb_) {
        AVLOG_ERROR("SQLMessageLog::Impl::writeMessage called when log is not open in write mode.");
        return {};
    }

    try {
        int originId = getOriginId(origin);
        int channelId = getChannelId(channel);

        unflushedRows_.push_back({RowType::Message, nsec, originId, channelId});
    } catch (const std::runtime_error& e) {
        AVLOG_ERROR("SQLMessageLog::Impl::writeMessage(): {}", e.what());
        return {};
    }

    if (unflushedRows_.size() >= maxBufferSize_) {
        flush(false);
    }
    return nextMessageId_++;
}

void SQLMessageLog::Impl::writeDropped(int64_t messageId, const std::string& client)
{
    if (!writing_ || !logDb_) {
        AVLOG_ERROR("SQLMessageLog::Impl::writeDropped called when log is not open in write mode.");
        return;
    }
    int originId = getOriginId(client);
    unflushedRows_.push_back({RowType::Dropped, messageId, originId, 0});

    if (unflushedRows_.size() >= maxBufferSize_) {
        flush(false);
    }
}

/*** Reads the specified number of rows (but may return less rows if it reaches the end of the table)
 * returns false if at "EOF"
 */
bool SQLMessageLog::Impl::read(SQLMessageLog::Row& data)
{
    if (writing_ || !logDb_) {
        AVLOG_ERROR("SQLMessageLog::Impl::read called when log is not open in read mode.");
        return {};
    }
    if (readDone_) {
        return false;
    }

    int result = ::sqlite3_step(readMessagesStatement_.get());
    try {
        trySQL(result);
    } catch (const std::runtime_error& e) {
        AVLOG_ERROR("SQLMessageLog::Impl::read(): {}", e.what());
        return false;
    }
    if (result == SQLITE_ROW) {
        //   "SELECT MESSAGES.rowid, MESSAGES.timestamp, ORIGINS.origin, CHANNELS.channel,
        //   group_concat(DESTORIGINS.origin) "

        int64_t timestamp = ::sqlite3_column_int64(readMessagesStatement_.get(), 1);
        std::string origin(reinterpret_cast<const char*>(::sqlite3_column_text(readMessagesStatement_.get(), 2)));
        std::string channel(reinterpret_cast<const char*>(::sqlite3_column_text(readMessagesStatement_.get(), 3)));

        // column 4 is a comma seperated list of strings (but may be NULL)
        const char* possiblyNullListOfDroppedDestinations =
            reinterpret_cast<const char*>(::sqlite3_column_text(readMessagesStatement_.get(), 4));

        std::set<std::string> dropped;

        if (possiblyNullListOfDroppedDestinations != nullptr) {
            std::string droppedDestinationsCSL{possiblyNullListOfDroppedDestinations};

            size_t pos = 0;
            size_t next = droppedDestinationsCSL.find(",", pos);
            while (next != std::string::npos) {
                dropped.insert(droppedDestinationsCSL.substr(pos, (next - pos)));
                pos = next + 1;
                next = droppedDestinationsCSL.find(",", pos);
            }
            dropped.insert(droppedDestinationsCSL.substr(pos, std::string::npos));
        }

        data.nsec = timestamp;
        data.origin = std::move(origin);
        data.channel = std::move(channel);
        data.dropped = std::move(dropped);
    }
    if (result == SQLITE_DONE) {
        readMessagesStatement_.reset();
        readDone_ = true;
    }
    return (readDone_ == false);
}

void SQLMessageLog::Impl::writeThreadFunction()
{
    std::unique_lock<std::mutex> lock(writeThreadMutex_);
    while (!shuttingDown_ || !databaseTaskQueue_.empty()) {
        while (!databaseTaskQueue_.empty()) {
            auto function = databaseTaskQueue_.front();
            databaseTaskQueue_.pop();
            lock.unlock();
            try {
                function();
            } catch (const std::runtime_error& e) {
                AVLOG_ERROR("SQLMessageLog::Impl::writeThreadFunction(): {}", e.what());
            }
            lock.lock();
        }

        taskQueueNotEmpty_.wait(lock, [this]() { return shuttingDown_ || !databaseTaskQueue_.empty(); });
    }
}

void SQLMessageLog::Impl::writeBuffer(const std::vector<BufferedRow>& rows)
{
    // this is now called internally on the writeThread
    trySQL(::sqlite3_exec(logDb_, "BEGIN TRANSACTION", nullptr, nullptr, nullptr), "BEGIN TRANSACTION");

    for (const auto& r : rows) {
        switch (r.type) {
            case SQLMessageLog::Impl::RowType::Message: {
                trySQL(::sqlite3_bind_int64(insertMessageStatement_.get(), 1, r.nsec));
                trySQL(::sqlite3_bind_int(insertMessageStatement_.get(), 2, r.originId));
                trySQL(::sqlite3_bind_int(insertMessageStatement_.get(), 3, r.channelId));

                trySQL(::sqlite3_step(insertMessageStatement_.get()), "message write statement");
                trySQL(::sqlite3_reset(insertMessageStatement_.get()));
                break;
            }
            case SQLMessageLog::Impl::RowType::Dropped: {
                trySQL(::sqlite3_bind_int64(insertDroppedStatement_.get(), 1, r.messageId));
                trySQL(::sqlite3_bind_int(insertDroppedStatement_.get(), 2, r.originId));
                trySQL(::sqlite3_step(insertDroppedStatement_.get()), "delete write statement");
                trySQL(::sqlite3_reset(insertDroppedStatement_.get()));
                break;
            }

            default:
                // This should be caught at compile time, with a "enum
                // value not covered by switch" error; but the
                // compiler insists we have a default case, which
                // obscures that error.
                AVLOG_ERROR("SQLMessageLog::Impl::writeBuffer(): unknown row type. Shouldn't ever happen. This means "
                            "that somebody has added a row type without updating writeBuffer() to handle it.");
                break;
        }
    }
    trySQL(::sqlite3_exec(logDb_, "END TRANSACTION", nullptr, nullptr, nullptr), "END TRANSACTION");
}

bool SQLMessageLog::Impl::flush(bool wait)
{
    // Do nothing if there is nothing to write and the thread is not running.
    if (!writeThread_.joinable()) {
        if (unflushedRows_.empty()) {
            return true;
        }
        // Something needs to be written, so start the thread.
        writeThread_ = std::thread(&SQLMessageLog::Impl::writeThreadFunction, this);
    }

    // If we use this promise, it has to survive until we call waitFuture.wait().
    std::promise<void> promise;
    std::future<void> waitFuture;
    try {
        std::lock_guard<std::mutex> lock{writeThreadMutex_};

        if (!unflushedRows_.empty()) {
            databaseTaskQueue_.emplace([this, rows{std::move(unflushedRows_)}]() { writeBuffer(rows); });
            unflushedRows_.clear();
        }

        if (wait) {
            waitFuture = std::move(promise.get_future());
            databaseTaskQueue_.emplace([&promise]() { promise.set_value(); });
        }
    } catch (const std::runtime_error& e) {
        AVLOG_ERROR("SQLMessageLog::Impl::flush(): {}", e.what());
        return false;
    }

    taskQueueNotEmpty_.notify_one();

    if (wait) {
        waitFuture.wait();
    }

    return true;
}

void SQLMessageLog::Impl::stopWriteThread()
{
    flush(true);
    if (writeThread_.joinable()) {
        // We have to set shuttingDown_ within the protection of the mutex. Otherwise
        // even if it were atomic, there would be a race condition if it is set in
        // between when wait() checks the predicate and when it waits.
        {
            std::unique_lock<std::mutex> lock(writeThreadMutex_);
            shuttingDown_ = true;
        }
        taskQueueNotEmpty_.notify_one();
        writeThread_.join();
        shuttingDown_ = false;
    }
}

// ********************* SQLMessageLog ***************************** //

SQLMessageLog::SQLMessageLog()
{
    impl_ = std::make_unique<SQLMessageLog::Impl>();
}

// Note: this is only needed so that the unique_ptr<> destructor will work.
SQLMessageLog::~SQLMessageLog() {}

bool SQLMessageLog::open(const std::string& databasePath, bool write, size_t buffer)
{
    return impl_->open(databasePath, write, buffer);
}

std::optional<int64_t> SQLMessageLog::writeMessage(int64_t nsec, const std::string& origin, const std::string& channel)
{
    return impl_->writeMessage(nsec, origin, channel);
}

void SQLMessageLog::writeDropped(int64_t messageId, const std::string& client)
{
    return impl_->writeDropped(messageId, client);
}

bool SQLMessageLog::read(SQLMessageLog::Row& data, bool includeSubscriptions)
{
    // If includeSubscriptions is false, we must skip any subscription
    // rows, which are marked by a timestamp of 0.
    while (impl_->read(data)) {
        if (includeSubscriptions || data.nsec) {
            return true;
        }
    }
    return false;
}

bool SQLMessageLog::flush(bool wait)
{
    return impl_->flush(wait);
}

bool SQLMessageLog::hasChannel(const std::string& channel) const
{
    return impl_->hasChannel(channel);
}

} // namespace SimpleComms
