/*!
 ***********************************************************
 * @file Server.cc
 * @brief A file implementing the Server class.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * Server.cc
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#include "Server.h"
#include "BasePeer.h"
#include "lib/avlog/avlog.hpp"
#include <SimpleComms/ControlMessages.h>
#include <Threading/PThreadLocker.h>

#include <SimpleComms/UnixClient.h>
#include <chrono>
#include <cstdio>
#include <cstdlib>
#include <errno.h>
#include <thread>

using namespace std;

namespace SimpleComms {

Server::Server(const PeerPtr local,
               const std::unordered_set<PeerPtr>& remotes,
               const std::string& logPath,
               MessageControl::Mode mode,
               ConfigSection& params)
    : local_(local)
    , remotes_(remotes)
    , messageControl_(logPath, mode, std::bind(&Server::queueMsg, this, std::placeholders::_1, nullptr), params)
    , writefds_(new pollfd[1])
{
    initialize();
}

Server::~Server()
{
    stop();

    pthread_cond_destroy(&writeQueuesNotEmpty_);
    pthread_mutex_destroy(&writeQueuesMutex_);
    pthread_mutex_destroy(&writeWaitingMutex_);
}

void Server::initialize()
{
    pthread_mutex_init(&writeQueuesMutex_, nullptr);
    pthread_mutex_init(&writeWaitingMutex_, nullptr);
    pthread_cond_init(&writeQueuesNotEmpty_, nullptr);
}

bool Server::start(const uint32_t numberOfOutputWorkers)
{
    numberOfOutputWorkers_ = numberOfOutputWorkers;

    if (!running_) {
        bool inToOutPipeActive = false, inThreadActive = false;
        uint32_t createdThreadsCount = 0;
        running_ = true;

        if (pipe(inToOutPipe_) != 0) {
            AVLOG_WARN("Error creating in to out signalling pipe: {}", strerror(errno));

            running_ = false;
        } else {
            inToOutPipeActive = true;
        }

        if (running_) {
            const int result = pthread_create(&inThread_, nullptr, inThreadMainLauncher, this);

            if (result != 0) {
                AVLOG_WARN("Error creating input thread: {}", strerror(result));

                running_ = false;
            } else {
                inThreadActive = true;
            }
        }

        if (running_) {
            outThreads_.reset(new pthread_t[numberOfOutputWorkers_]);

            for (uint32_t i = 0; (i < numberOfOutputWorkers_) && running_; ++i) {
                const int result = pthread_create(&outThreads_[i], nullptr, outThreadMainLauncher, this);

                if (result != 0) {
                    AVLOG_WARN("Error creating output thread: {}", strerror(result));

                    running_ = false;
                } else {
                    ++createdThreadsCount;
                }
            }
        }

        // Clean up if the process failed. The last item doesn't need cleaning up as either it wasn't attempted at all,
        // or was the reason for the failure, not requiring clean up.
        if (!running_) {
            if (inToOutPipeActive) {
                close(inToOutPipe_[0]);
                close(inToOutPipe_[1]);

                inToOutPipe_[0] = -1;
                inToOutPipe_[1] = -1;
            }

            if (inThreadActive) {
                pthread_join(inThread_, nullptr);
            }

            pthread_cond_broadcast(&writeQueuesNotEmpty_);

            for (uint32_t i = 0; i < createdThreadsCount; ++i) {
                pthread_join(outThreads_[i], nullptr);
            }
        }
    }

    return running_;
}

void Server::stop()
{
    if (running_) {
        running_ = false;

        pthread_cond_broadcast(&writeQueuesNotEmpty_);

        pthread_join(inThread_, nullptr);

        for (uint32_t i = 0; i < numberOfOutputWorkers_; i++) {
            pthread_join(outThreads_[i], nullptr);
        }

        close(inToOutPipe_[0]);
        close(inToOutPipe_[1]);

        inToOutPipe_[0] = -1;
        inToOutPipe_[1] = -1;
    }
}

void* Server::inThreadMainLauncher(void* arg)
{
    reinterpret_cast<Server*>(arg)->inThreadMain();

    return nullptr;
}

void* Server::outThreadMainLauncher(void* arg)
{
    reinterpret_cast<Server*>(arg)->outThreadMain();

    return nullptr;
}

void Server::inThreadMain()
{
    TimeStamp then, now;
    Peer::BaseClientList readHandles;
    std::unique_ptr<pollfd[]> ufds;
    size_t maxHandles = 0;
    while (running_) {
        readHandles.clear();

        MessagePtr msg;

        local_->getNamedReadHandles(readHandles);
        for (auto const& remote : remotes_) {
            remote->getNamedReadHandles(readHandles);
        }

        const auto nfds = readHandles.size();

        if (nfds > maxHandles) {
            maxHandles = nfds;
            readHandles.reserve(maxHandles);
            ufds.reset(new pollfd[maxHandles]);
        }

        for (unsigned int i = 0; i < nfds; i++) {
            ufds[i].fd = readHandles[i]->getReadHandle();
            ufds[i].events = POLLIN;
        }

        const int pollResult = poll(ufds.get(), nfds, 1000);

        if (pollResult > 0) {
            for (unsigned int i = 0; i < nfds; i++) {
                const Peer::BaseClientPtr& handle = readHandles[i];
                Peer* const currentHandlePeer(handle->getParentPeer());
                if (ufds[i].revents & POLLIN) {
                    if (currentHandlePeer->receiveMsg(handle, msg) == Status::NewData) {
                        // Allow MessageControl to check whether the message should be published or not.
                        // Some control messages reach this point only so MessageControl can see them.
                        // MessageControl::add() returns true if we need to query the message log for the replay order.
                        if (messageControl_.preprocess(msg) && !messageControl_.add(msg)) {
                            queueMsg(msg, currentHandlePeer);
                        } else {
                            // This is not an ignored message. Rather than sending it immediately, we must ask the
                            // message log which message(s) should be sent.
                            while (MessagePtr next = messageControl_.pop()) {
                                queueMsg(next, local_.get());
                            }
                        }
                        reads_++;

                        if (currentHandlePeer == local_.get()) {
                            localReadBytes_ += msg->MessageLength();
                        } else {
                            remoteReadBytes_ += msg->MessageLength();
                        }
                    }
                }

                if (ufds[i].revents & (POLLERR | POLLHUP | POLLNVAL)) {
                    Peer::BaseClientList errorHandles(1, handle);

                    currentHandlePeer->handleErrors(errorHandles);
                }
            }
        }

        now.setNow();
        {
            const auto dt = static_cast<double>(now - then);
            if (dt > 1) {
                AVLOG_INFO("reads {} (local {} kB/s, remote {} kB/s); writes {}",
                           reads_ / dt,
                           (localReadBytes_ / 1024.0) / dt,
                           (remoteReadBytes_ / 1024.0) / dt,
                           writes_ / dt);

                reads_ = 0;
                localReadBytes_ = 0;
                remoteReadBytes_ = 0;

                writes_ = 0;
                then = now;
            }
        }
    }
}

bool Server::queueMsg(const MessagePtr& msg, Peer* currentHandlePeer)
{
    // Don't allow multiple threads to queue a message at the same time.
    PThreadLocker lock(writeQueuesMutex_);

    // In logging mode, log the fact that this message was sent.
    messageControl_.log(msg);
    // In replay mode, check if any clients dropped this message originally.
    std::set<std::string> dropped = messageControl_.dropped(msg);

    Peer::BaseClientList newPendingClients;

    bool success = (local_->queueMsg(msg, newPendingClients, dropped) != Status::NotConnected);

    // Don't check for sending messages to remotes if currentHandlePeer is null.
    // In that case, this is a control message only being sent to one local client.
    if (currentHandlePeer) {
        for (auto const& remote : remotes_) {
            if (remote && currentHandlePeer != remote.get()) {
                remote->queueMsg(msg, newPendingClients, dropped);
            }
        }
    }

    // If there are clients requiring polling...
    for (Peer::BaseClientList::iterator i = newPendingClients.begin(); i != newPendingClients.end(); ++i) {
        if (!(*i)->isOnWriteQueue()) {
            AVLOG_DEBUG("WRITEON {}", (*i)->getName());
            (*i)->setOnWriteQueue();

            if ((*i)->isReadyToWrite()) {
                writeReady_.push_back(*i);
            } else {
                writeWaiting_.push_back(*i);
            }

            pthread_cond_signal(&writeQueuesNotEmpty_);
        } else {
            AVLOG_DEBUG("WRITEDUP {}", (*i)->getName());
        }
    }

    if (!writeWaiting_.empty()) {
        // and I can't acquire a lock to the poll routine, wake it up
        if (pthread_mutex_trylock(&writeWaitingMutex_) != 0) {
            const char buf = 0;
            if (write(inToOutPipe_[1], &buf, 1) == -1) {
                AVLOG_ERROR("Failed to write to pipe: {}", strerror(errno));
            }
        }
        // otherwise, never mind
        else {
            pthread_mutex_unlock(&writeWaitingMutex_);
        }
    }

    return success;
}

void Server::outThreadMain()
{
    pthread_mutex_lock(&writeQueuesMutex_);

    while (running_) {
        if (!writeWaiting_.empty() && (pthread_mutex_trylock(&writeWaitingMutex_) == 0)) {
            doPoll();

            pthread_mutex_unlock(&writeWaitingMutex_);
        } else if (!writeReady_.empty()) {
            Peer::BaseClientPtr currentClientPtr = writeReady_.front();
            writeReady_.pop_front();

            pthread_mutex_unlock(&writeQueuesMutex_);

            {
                Peer::BaseClientList writeHandles(1, currentClientPtr);
                currentClientPtr->getParentPeer()->doWrites(writeHandles);
            }

            pthread_mutex_lock(&writeQueuesMutex_);
            writes_++;

            if (currentClientPtr->hasDataToWrite()) {
                if (currentClientPtr->isReadyToWrite()) {
                    AVLOG_DEBUG("WRITEREADY {}", currentClientPtr->getName());
                    writeReady_.push_back(currentClientPtr);
                } else {
                    AVLOG_DEBUG("WRITEWAIT {}", currentClientPtr->getName());
                    writeWaiting_.push_back(currentClientPtr);
                }
            } else {
                AVLOG_DEBUG("WRITEOFF {}", currentClientPtr->getName());
                currentClientPtr->setOffWriteQueue();
            }
        } else {
            pthread_cond_wait(&writeQueuesNotEmpty_, &writeQueuesMutex_);
        }
    }

    pthread_mutex_unlock(&writeQueuesMutex_);
}

void Server::doPoll()
{
    const auto writeWaitingCount = writeWaiting_.size();

    if (writeWaitingCount > maxWriteHandles_) {
        maxWriteHandles_ = static_cast<uint32_t>(writeWaitingCount);
        writefds_.reset(new pollfd[maxWriteHandles_ + 1]);
    }

    {
        uint32_t j = 0;
        for (std::list<Peer::BaseClientPtr>::iterator i = writeWaiting_.begin(); i != writeWaiting_.end(); ++i, ++j) {
            AVLOG_DEBUG("DOPOLL {}", (*i)->getName());
            writefds_[j].fd = (*i)->getWriteHandle();
            writefds_[j].events = POLLOUT;
        }
    }

    pthread_mutex_unlock(&writeQueuesMutex_);

    writefds_[writeWaitingCount].fd = inToOutPipe_[0];
    writefds_[writeWaitingCount].events = POLLIN;

    const int pollResult = poll(writefds_.get(), writeWaitingCount + 1, 1000);
    bool readError = false;
    int errnoCache = 0;

    // Clear the wake up from the input thread, the write polling will be updated when this loop turns over
    if ((pollResult > 0) && (writefds_[writeWaitingCount].revents & POLLIN)) {
        char buf = 0;
        readError = (read(inToOutPipe_[0], &buf, sizeof(buf)) < 0);
        if (readError) {
            errnoCache = errno;
        }
    }

    pthread_mutex_lock(&writeQueuesMutex_);

    if (readError) {
        AVLOG_ERROR("Failed to read from pipe: {}", strerror(errnoCache));
    }

    if (pollResult > 0) {
        std::list<Peer::BaseClientPtr> newWaiting_;

        std::list<Peer::BaseClientPtr>::iterator i(writeWaiting_.begin());
        for (uint32_t j = 0; (j < writeWaitingCount) && (i != writeWaiting_.end()); ++i, ++j) {
            if (writefds_[j].revents & (POLLERR | POLLHUP | POLLNVAL)) {
                AVLOG_DEBUG("POLLERR {}", (*i)->getName());
                Peer::BaseClientList errorHandles(1, *i);

                (*i)->getParentPeer()->handleErrors(errorHandles);
            } else if (writefds_[j].revents & POLLOUT) {
                AVLOG_DEBUG("POLL+++ {}", (*i)->getName());
                (*i)->setReadyToWrite();
                writeReady_.push_back(*i);
            } else {
                AVLOG_DEBUG("POLL--- {}", (*i)->getName());
                newWaiting_.push_back(*i);
            }
        }

        if (i == writeWaiting_.end()) {
            writeWaiting_.swap(newWaiting_);
        } else {
            writeWaiting_.erase(writeWaiting_.begin(), i);
            writeWaiting_.insert(writeWaiting_.end(), newWaiting_.begin(), newWaiting_.end());
        }
    }
}
} // namespace SimpleComms
