/**
 * \copyright
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief Support functionality for the main SimpleCoimms server program
 */

#include "ServerMainSupport.h"
#include "TcpPeer.h"
#include "UdpPeer.h"

#include <iostream>

namespace {
/*
 * @brief Small helper for parsing the interface lists
 */
std::vector<std::string> getAuxInterfaces(ConfigSection& hostConfig)
{
    std::vector<std::string> interfaces{};
    hostConfig.getArray("AuxInterfaces", interfaces);

    return interfaces;
}
} // namespace

namespace SimpleComms {
CLIHelp::CLIHelp()
{
    help_info_.push_back(
        StringPair("advertisedIntf", "Details of this interface will be advertised for remote connections"));
    help_info_.push_back(StringPair("help", "Produce this help message"));
    help_info_.push_back(
        StringPair("maxExitSignalsToCatch", "Number of exit signals to catch during one second before terminating"));
    help_info_.push_back(StringPair("staleFragmentPruningThreshold",
                                    "Fragments sent more than specified seconds ago will be discarded"));
    help_info_.push_back(StringPair("numberOfOutputWorkers", "Number of threads use to dispatch messages, minimum 2"));
    help_info_.push_back(StringPair("tcp", "Use TCP for remote peering"));
    help_info_.push_back(StringPair("tcpPort", "TCP port for peering, 0 for automatic"));
    help_info_.push_back(StringPair("tcpListenBackLog", "listen(2) backlog for TCP peer"));
    help_info_.push_back(StringPair("udp", "Use UDP for remote peering"));
    help_info_.push_back(StringPair("udpPort", "UDP port for peering"));
    help_info_.push_back(StringPair("unixPath", "Local unix domain socket path"));
    help_info_.push_back(StringPair("udpDiscoveryPort", "UDP port for autodiscovery of peers"));
    help_info_.push_back(
        StringPair("logLevel", "Logging level. One of DEBUG, INFO, NOTICE, WARN, ERROR, CRIT, ALERT, FATAL, EMERG"));
    help_info_.push_back(StringPair("log", "Message log file to save"));
    help_info_.push_back(StringPair("replay", "Load and replay the message order from this log"));
    help_info_.push_back(StringPair("frozen", "Base executive timing on log playback rather than clock time"));
}

void CLIHelp::print() const
{
    std::cerr << "SimpleComms Usage:\n";
    for (auto const& info : help_info_) {
        std::cerr << " --" << info.first << "\n\t" << info.second << "\n";
    }
    std::cerr << std::endl;
}

std::unique_ptr<UdpDiscovery> createDiscoveryAgent(const uint16_t port, const std::string& interface)
{
    auto discovery = std::make_unique<UdpDiscovery>(interface, port);

    if (discovery->initialize() != Status::Ok) {
        throw std::runtime_error("Failed to initialize discovery agent: " + interface);
    }

    return discovery;
}

std::vector<std::unique_ptr<UdpDiscovery>> createDiscoveryAgents(
    const std::string& defaultIntf,
    const PeerPtr defaultRemotePeer,
    const uint16_t discoveryPort,
    const PeerPtr localPeer,
    const std::unordered_set<std::string>& defaultIntfBlacklist,
    const std::unordered_map<std::string, std::unordered_set<std::string>>& auxIntfWhitelists,
    const std::unordered_map<std::string, PeerPtr> interfacePeerMap)
{
    std::vector<std::unique_ptr<UdpDiscovery>> agents;
    // Since everything is sent over the default interface by default we exclude channels
    // rather than allow them
    if (defaultRemotePeer != nullptr) {
        auto discovery = createDiscoveryAgent(discoveryPort, defaultIntf);

        discovery->registerLocalPeeringMethod(localPeer);
        discovery->registerRemotePeeringMethod(defaultRemotePeer);
        if (discovery->setBlacklist(defaultIntfBlacklist) != Status::Ok) {
            throw std::runtime_error("Failed to set channel blacklist");
        }

        agents.push_back(std::move(discovery));
    }
    for (auto const& intfPeerPair : auxIntfWhitelists) {
        auto interface = intfPeerPair.first;
        auto whitelist = intfPeerPair.second;

        auto discovery = createDiscoveryAgent(discoveryPort, interface);
        discovery->registerLocalPeeringMethod(localPeer);
        discovery->registerRemotePeeringMethod(interfacePeerMap.at(interface));
        if (discovery->setWhitelist(whitelist) != Status::Ok) {
            throw std::runtime_error("Failed to set channel whitelist for " + interface);
        }

        agents.push_back(std::move(discovery));
    }

    return agents;
}

void parseCLIArgs(std::map<std::string, std::string>& vm, int argc, char** argv)
{
    std::vector<std::string> args;
    for (int ii = 1; ii < argc; ++ii) {
        args.push_back(argv[ii]);
    }

    for (unsigned int ii = 0; ii < args.size(); /**/) {
        if (args[ii].length() < 3) {
            // we've encountered something that isn't a parameter specification, print
            // the help message and exit
            vm["help"] = "true";
            break;
        }
        if (args[ii][0] != '-' || args[ii][1] != '-') {
            // ditto.
            vm["help"] = "true";
            break;
        }

        if (args[ii] == "--help") {
            vm["help"] = "true";
            break;
        } else if (args[ii] == "--version") {
            vm["help"] = "true";
            break;
        } else if (args[ii] == "--tcp") {
            vm["tcp"] = "true";
            ii++;
        } else if (args[ii] == "--udp") {
            vm["udp"] = "true";
            ii++;
        } else if (args[ii] == "--frozen") {
            vm["frozen"] = "true";
            ii++;
        } else {
            // every other argument should have a default value in the table
            // if it's not in the table....
            if (vm.count(args[ii].substr(2)) == 0) {
                AVLOG_ERROR("unsupported parameter: {}", args[ii]);
                vm["help"] = "true";
                break;
            } else {
                if (ii < args.size() - 1) {
                    vm[args[ii].substr(2)] = args[ii + 1];
                    ii += 2;
                } else {
                    AVLOG_ERROR("parameter {} requires an argument", args[ii]);
                    vm["help"] = "true";
                    break;
                }
            }
        }
    }
}

/*
 * @brief Helper function for creating a remote peer
 *
 * @param name Remote peer name
 * @param vm Program options
 *
 * @return Pointer to a constructed remote peer
 */
PeerPtr createRemotePeer(const std::string& name, std::map<std::string, std::string>& vm)
{
    PeerPtr remotePeer;
    if (vm.count("udp") > 0) {
        remotePeer.reset(new UdpPeer(name,
                                     static_cast<uint16_t>(atoi(vm["udpPort"].c_str())),
                                     TimeStamp(atof(vm["staleFragmentPruningThreshold"].c_str()))));
    } else if (vm.count("tcp") > 0) {
        remotePeer.reset(new TcpPeer(name,
                                     static_cast<uint16_t>(atoi(vm["tcpPort"].c_str())),
                                     static_cast<uint16_t>(atoi(vm["tcpListenBackLog"].c_str())),
                                     TimeStamp(atof(vm["staleFragmentPruningThreshold"].c_str()))));
    }

    if (remotePeer) {
        if (remotePeer->connect() != Status::Ok) {
            return nullptr;
        }
    }

    return remotePeer;
}

void parseInterfaceLists(ConfigSection& hostConfig,
                         const std::string& defaultIntf,
                         std::vector<std::string>& auxIntfs,
                         std::unordered_set<std::string>& defaultIntfBlacklist,
                         std::unordered_map<std::string, std::unordered_set<std::string>>& auxIntfWhitelists)
{
    auxIntfs = getAuxInterfaces(hostConfig);

    if (!auxIntfs.empty()) {
        for (auto const& intf : auxIntfs) {
            auxIntfWhitelists.emplace(intf, std::unordered_set<std::string>());
        }

        ConfigSection routes;
        if (hostConfig.get("ChannelRoutes", routes) != ConfigSection::GS_OK) {
            // This may not be an error, we could just be enabling an aux interface on a host
            // for sending
            AVLOG_WARN("No channel routes");
        } else {
            buildDiscoveryFilterLists(routes, defaultIntf, defaultIntfBlacklist, auxIntfWhitelists);
        }
    }
}

void createPeers(std::map<std::string, std::string>& vm,
                 const std::vector<std::string>& auxIntfs,
                 PeerPtr& defaultRemotePeer,
                 std::unordered_set<PeerPtr>& remotePeers,
                 std::unordered_map<std::string, PeerPtr>& interfacePeerMap)
{
    defaultRemotePeer = createRemotePeer(vm["advertisedIntf"], vm);
    if (!defaultRemotePeer) {
        auto msg = "Failed to initialize default remote peer on " + vm["advertisedIntf"];
        throw std::runtime_error(msg);
    }

    remotePeers.insert(defaultRemotePeer);
    interfacePeerMap.emplace(vm["advertisedIntf"], defaultRemotePeer);

    for (auto const& auxIntf : auxIntfs) {
        auto peer = createRemotePeer(auxIntf, vm);
        if (!defaultRemotePeer) {
            auto msg = "Failed to initialize remote peer on " + auxIntf;
            throw std::runtime_error(msg);
        }

        remotePeers.insert(peer);
        interfacePeerMap.emplace(auxIntf, peer);
    }
}
} // namespace SimpleComms
