/**
 * \copyright
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief Support functionality for the main SimpleCoimms server program
 */

#include "Peer.h"
#include "UdpDiscovery.h"

#include <map>
#include <memory>
#include <string>
#include <utility>
#include <vector>

namespace SimpleComms {
class CLIHelp
{
public:
    CLIHelp();
    void print() const;

private:
    using StringPair = std::pair<std::string, std::string>;
    std::vector<StringPair> help_info_;
};

/*
 * @brief Helper function for creating a discovery agent
 *
 * @param port Port to run discovery on
 * @param interface Physical interface name to run discovery on
 *
 * @return Constructed discovery agent
 */
std::unique_ptr<UdpDiscovery> createDiscoveryAgent(const uint16_t port, const std::string& interface);

/*
 * @brief Helper function for creating all of the discovery agents for an scs instance
 *
 * @param defaultIntf default interface name
 * @param defaultRemotePeer default remote peer instance
 * @param discoveryPort port for sending discovery broadcasts
 * @param localPeer local scs peer
 * @param defaultIntfBlacklist channel blacklist for the default interface
 * @param auxIntfWhitelists channel whitelists for the auxiliary interfaces
 * @param interfacePeerMap map of interface-peer key-values
 *
 * @return Initailized list of discovery agent pointers
 */
std::vector<std::unique_ptr<UdpDiscovery>> createDiscoveryAgents(
    const std::string& defaultIntf,
    const PeerPtr defaultRemotePeer,
    const uint16_t discoveryPort,
    const PeerPtr localPeer,
    const std::unordered_set<std::string>& defaultIntfBlacklist,
    const std::unordered_map<std::string, std::unordered_set<std::string>>& auxIntfWhitelists,
    const std::unordered_map<std::string, PeerPtr> interfacePeerMap);

/*
 * @brief Parse the command line arguments
 *
 * @param[out] vm map of default arguments
 * @param[in] argc argument count
 * @param[in] argv argument values
 *
 * @return Map of name-value pairs
 */
void parseCLIArgs(std::map<std::string, std::string>& vm, int argc, char** argv);

/*
 * @brief Helper function for creating a remote peer
 *
 * @param name Remote peer name
 * @param vm Program options
 *
 * @return Pointer to a constructed remote peer
 */
PeerPtr createRemotePeer(const std::string& name, std::map<std::string, std::string>& vm);

/*
 * @brief Parse the interface lists from the simplecomms config
 *
 * @param[in] hostConfig config to parse corresponding to a single host
 * @param[in] defaultIntf Default interface
 * @param[out] auxIntfs List of auxiliary interfaces
 * @param[out] defaultIntfBlacklist Blacklist of channels on the default interface
 * @param[out] auxIntfWhitelists Map of whitelists for each auxiliary interface
 */
void parseInterfaceLists(ConfigSection& hostConfig,
                         const std::string& defaultIntf,
                         std::vector<std::string>& auxIntfs,
                         std::unordered_set<std::string>& defaultIntfBlacklist,
                         std::unordered_map<std::string, std::unordered_set<std::string>>& auxIntfWhitelists);

/*
 * @brief Create all of the peers needed for scs
 *
 * @param[in] vm argument map
 * @param[in] auxIntfs list of auxiliary interfaces
 * @param[out] defaultRemotePeer default remote peer
 * @param[out] remotePeers list of remote peers
 * @param[out] interfacePeerMap convenience map between interface and peer
 */
void createPeers(std::map<std::string, std::string>& vm,
                 const std::vector<std::string>& auxIntfs,
                 PeerPtr& defaultRemotePeer,
                 std::unordered_set<PeerPtr>& remotePeers,
                 std::unordered_map<std::string, PeerPtr>& interfacePeerMap);
} // namespace SimpleComms
