/*!
 ***********************************************************
 * @file UdpDiscovery.cc
 * @brief A file implementing the UdpDiscovery class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim
 * @date 03/09/2016
 * @remark
 * Added the cross-platform support
 *  -- 32-bit and big-endian
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * UdpDiscovery.cc
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#include "UdpDiscovery.h"
#include <SimpleComms/Constants.h>
#include <cerrno>
#include <cstdio>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "NetworkIntfUtils.h"
#include "lib/avlog/avlog.hpp"
#include <SimpleComms/ControlMessages.h>
#include <interfaces/baseTypes/Header.h>

using namespace std;

namespace SimpleComms {

UdpDiscovery::UdpDiscovery(const std::string& advertisedIntf, const uint16_t port)
    : advertisedIntf_(advertisedIntf)
    , port_(port)
    , broadcastAddress_(INADDR_BROADCAST)
    , fd_(-1)
    , rollingMessageID_(0)
    , threadRunning_(0)
{}

UdpDiscovery::~UdpDiscovery()
{
    terminate();
}

Status::States UdpDiscovery::initialize()
{
    if (threadRunning_) {
        return Status::AlreadyConnected;
    }

    {
        char hostNameBuf[HOST_NAME_MAX];

        if (gethostname(hostNameBuf, HOST_NAME_MAX) == -1) {
            return Status::Error;
        }

        hostName_ = hostNameBuf;
    }

    if (!NetworkIntfUtils::GetInterfaceParameter(advertisedIntf_, NetworkIntfUtils::Broadcast, broadcastAddress_)) {
        AVLOG_WARN("Error retrieving broadcast address for {}", advertisedIntf_);

        return Status::Error;
    }

    if ((fd_ = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
        AVLOG_WARN("Error creating socket: {}", strerror(errno));

        return Status::ErrorCreatingSocket;
    }

    {
        int broadcast = 1;
        if (setsockopt(fd_, SOL_SOCKET, SO_BROADCAST, &broadcast, sizeof(int))) {
            AVLOG_WARN("Error setting SO_BROADCAST: {}", strerror(errno));

            terminate();

            return Status::Error;
        }
    }

    struct sockaddr_in servAddr;
    servAddr.sin_family = AF_INET;
    servAddr.sin_port = htons(port_);

    // Normally we would just bind to the broadcast address here, but for some reason this does not
    // work on QNX, whereas binding to INADDR_ANY does
    servAddr.sin_addr.s_addr = htonl(INADDR_ANY);

    if (::bind(fd_, reinterpret_cast<struct sockaddr*>(&servAddr), sizeof(servAddr)) == -1) {
        terminate();

        return Status::ErrorBindingSocket;
    }

    if (fcntl(fd_, F_SETFL, O_NONBLOCK) == -1) {
        terminate();

        return Status::Error;
    }

    // Spawn a thread that sits on fd_ and queues the incoming data
    if (pthread_create(&discoveryThreadHandle_, NULL, discoveryThreadLauncher, this) == 0) {
        // Flag that we successfully launched a thread for later joining
        threadRunning_++;
    } else {
        terminate();

        return Status::Error;
    }

    instanceEpoch_.setNow();

    return Status::Ok;
}

Status::States UdpDiscovery::isActive()
{
    return (fd_ == -1) ? Status::Error : Status::Ok;
}

Status::States UdpDiscovery::terminate()
{
    return terminate(false);
}

Status::States UdpDiscovery::registerLocalPeeringMethod(const PeerPtr localPeer)
{
    localPeeringMethods_[localPeer->getTransport()] = localPeer;

    return Status::Ok;
}

Status::States UdpDiscovery::unregisterLocalPeeringMethod(const PeerPtr localPeer)
{
    PeeringMethodsMap::iterator i = localPeeringMethods_.find(localPeer->getTransport());

    if (i != localPeeringMethods_.end()) {
        localPeeringMethods_.erase(i);
    }

    return Status::Ok;
}

Status::States UdpDiscovery::registerRemotePeeringMethod(const PeerPtr remotePeer)
{
    remotePeeringMethods_[remotePeer->getTransport()] = remotePeer;

    return Status::Ok;
}

Status::States UdpDiscovery::unregisterRemotePeeringMethod(const PeerPtr remotePeer)
{
    PeeringMethodsMap::iterator i = remotePeeringMethods_.find(remotePeer->getTransport());

    if (i != remotePeeringMethods_.end()) {
        remotePeeringMethods_.erase(i);
    }

    return Status::Ok;
}

Status::States UdpDiscovery::setBlacklist(const std::unordered_set<std::string>& list)
{
    // Only allow one filter
    if (filterList_ != nullptr) {
        return Status::Error;
    }

    filterList_ = std::make_unique<FilterList>(FilterList{FilterListType::BLACKLIST, list});

    return Status::Ok;
}

Status::States UdpDiscovery::setWhitelist(const std::unordered_set<std::string>& list)
{
    // Only allow one filter
    if (filterList_ != nullptr) {
        return Status::Error;
    }

    filterList_ = std::make_unique<FilterList>(FilterList{FilterListType::WHITELIST, list});

    return Status::Ok;
}

Status::States UdpDiscovery::terminate(const bool discoveryThread)
{
    if (fd_ != -1) {
        close(fd_);
        fd_ = -1;
    }

    if (!discoveryThread && threadRunning_) {
        threadRunning_--;
        pthread_join(discoveryThreadHandle_, NULL);
    }

    return Status::Ok;
}

void* UdpDiscovery::discoveryThreadLauncher(void* arg)
{
    static_cast<UdpDiscovery*>(arg)->discoveryThreadMain();

    return NULL;
}

void UdpDiscovery::discoveryThreadMain()
{
    TimeStamp nextCycle;
    nextCycle.setNow();
    nextCycle = nextCycle + 1.0;

    while (fd_ != -1) {
        fd_set rfds;

        FD_ZERO(&rfds);
        FD_SET(fd_, &rfds);

        TimeStamp now;
        now.setNow();

        TimeStamp wl = max(nextCycle - now, TimeStamp(0, 0));
        struct timeval waitLimit;
        waitLimit.tv_sec = wl.tv_sec;
        waitLimit.tv_usec = wl.tv_usec;

        const int selectStatus = select(fd_ + 1, &rfds, NULL, NULL, &waitLimit);

        if (selectStatus == -1) {
            switch (errno) {
                case EINTR:
                    // Ignore
                    break;
                default:
                    AVLOG_WARN("Error during select: {}", strerror(errno));

                    terminate(true);

                    return;
            }
        }

        if (fd_ == -1) {
            break;
        }

        if (FD_ISSET(fd_, &rfds)) {
            handleIncomingMsg();
        }

        now.setNow();
        if (now > nextCycle) {
            nextCycle = now + 1.0;

            sendSubscriptions();
            sendRegisteredRemotePeeringMethods();
        }
    }
}

Status::States UdpDiscovery::sendMsg(Message& msg, const uint32_t dataLength)
{
    memcpy(msg.origin, hostName_.c_str(), hostName_.length() + 1);
    memcpy(msg.channel, SIMPLECOMMS_CONTROL_CHANNEL, sizeof(SIMPLECOMMS_CONTROL_CHANNEL));
    msg.sendMsgTime.setNow();
    msg.sendMsgTime.tv_sec = htonls(msg.sendMsgTime.tv_sec);
    msg.sendMsgTime.tv_usec = htonls(msg.sendMsgTime.tv_usec);
    msg.length = htonl(dataLength);
    msg.rollingMessageID = htonl(rollingMessageID_++);
    msg.totalPages = htonl(1);

    msg.pageNumber = htonl(1);
    msg.dataLength = htonl(dataLength);

    messageSetChecksum(msg); // TODO remove this once checksum is ignored on falcon boards

    struct sockaddr_in servAddr;
    servAddr.sin_family = AF_INET;
    servAddr.sin_port = htons(port_);
    servAddr.sin_addr.s_addr = htonl(broadcastAddress_);

    const ssize_t sendStatus = sendto(fd_,
                                      reinterpret_cast<const void*>(&msg),
                                      msg.MessageLength(),
                                      MSG_NOSIGNAL,
                                      reinterpret_cast<sockaddr*>(&servAddr),
                                      sizeof(servAddr));

    if (sendStatus == -1) {
        AVLOG_WARN("Error sending message: {}", strerror(errno));
    }

    return (sendStatus == -1) ? Status::Error : Status::Ok;
}

Status::States UdpDiscovery::sendSubscriptions()
{
    vector<string> locallySubscribedChannels;

    for (PeeringMethodsMap::iterator i(localPeeringMethods_.begin()); i != localPeeringMethods_.end(); ++i) {
        (*i).second->getSubscriptions(locallySubscribedChannels);
    }

    {
        Message msg;
        SubscriptionListMessage* subListMsg = reinterpret_cast<SubscriptionListMessage*>(msg.data);

        subListMsg->type = (ControlMessages::ControlMessageTypes)htonl(ControlMessages::SubscriptionList);
        subListMsg->numChannels = 0;

        for (auto const& channelName : locallySubscribedChannels) {
            if (channelName.length() < sizeof(subListMsg->channel[0])) {
                // Simple filtering mechanism
                if ((filterList_->type == FilterListType::BLACKLIST) &&
                    (filterList_->entries.find(channelName) != filterList_->entries.end())) {
                    // Skip entries in the blacklist
                    continue;
                } else if ((filterList_->type == FilterListType::WHITELIST) &&
                           (filterList_->entries.find(channelName) == filterList_->entries.end())) {
                    // Skip entries if _not_ in the whitelist
                    continue;
                }

                memcpy(subListMsg->channel[subListMsg->numChannels], channelName.c_str(), channelName.length() + 1);
                subListMsg->numChannels++;

                if (subListMsg->numChannels == 999) {
                    break;
                }
            }
        }

        uint16_t numChannels = subListMsg->numChannels;
        subListMsg->numChannels = htons(subListMsg->numChannels);

        return sendMsg(msg,
                       static_cast<uint32_t>(sizeof(subListMsg->type) + sizeof(subListMsg->numChannels) +
                                             sizeof(subListMsg->channel[0]) * numChannels));
    }

    return Status::Ok;
}

Status::States UdpDiscovery::sendRegisteredRemotePeeringMethods()
{
    Status::States returnStatus = Status::Ok;

    if (!remotePeeringMethods_.empty()) {
        for (PeeringMethodsMap::iterator i = remotePeeringMethods_.begin(); i != remotePeeringMethods_.end(); ++i) {
            AVLOG_DEBUG("Sending connection method {}", transportName(i->first));

            Message msg;
            const PeerPtr& remotePeer(i->second);
            const Status::States getConnInfoStatus = remotePeer->getConnectionInfo(msg);

            if (getConnInfoStatus == Status::Ok) {
                ConnectionMessage* connMsg(reinterpret_cast<ConnectionMessage*>(&msg.data));
                connMsg->offerEpoch.tv_sec = htonls(instanceEpoch_.tv_sec);
                connMsg->offerEpoch.tv_usec = htonls(instanceEpoch_.tv_usec);

                const Status::States sendStatus = sendMsg(msg, ntohl(msg.dataLength));

                if (sendStatus != Status::Ok) {
                    returnStatus = sendStatus;
                }
            } else {
                returnStatus = getConnInfoStatus;
            }
        }
    }

    return returnStatus;
}

Status::States UdpDiscovery::handleIncomingMsg()
{
    Status::States returnStatus = Status::Ok;
    Message msg;
    const ssize_t recvStatus = recv(fd_, &msg, sizeof(msg), MSG_DONTWAIT | MSG_NOSIGNAL);

    if (recvStatus == -1) {
        if (errno != EAGAIN) {
            AVLOG_WARN("Error receiving message: {}", strerror(errno));

            terminate(true);

            returnStatus = Status::Error;
        }
    } else {
        returnStatus = Status::NewData;
    }

    if (returnStatus == Status::NewData) {
        if (string(msg.channel) == SIMPLECOMMS_CONTROL_CHANNEL) {
            returnStatus = handleControlMsg(msg);
        } else {
            returnStatus = Status::Error;
            AVLOG_WARN("Non-control message on discovery broadcast");
        }
    }

    return returnStatus;
}

Status::States UdpDiscovery::handleControlMsg(const Message& msg)
{
    const string origin(msg.origin);

    // Ignore messages that originated here
    if (hostName_ != origin) {
        const auto msgType =
            static_cast<uint32_t>((msg.data[3] << 24) | (msg.data[2] << 16) | (msg.data[1] << 8) | (msg.data[0]));

        if (ControlMessages::SubscriptionList == ntohl(msgType)) {
            const SubscriptionListMessage* subListMsg = reinterpret_cast<const SubscriptionListMessage*>(msg.data);
            vector<string> subscribes, unsubscribes;

            uint32_t i = 0;
            ChannelSubscriberMap::iterator j = channelSubscribers_.begin();
            while ((i < ntohs(subListMsg->numChannels)) && (j != channelSubscribers_.end())) {
                const string channel(subListMsg->channel[i]);
                SubscriberSet& subscribers(j->second);
                const int channelCompareResult = channel.compare(j->first);

                if (channelCompareResult > 0) {
                    if (subscribers.erase(origin) > 0) {
                        unsubscribes.push_back(j->first);
                    }

                    ++j;
                } else if (channelCompareResult == 0) {
                    const pair<SubscriberSet::iterator, bool> result = subscribers.insert(origin);

                    if (result.second) {
                        subscribes.push_back(channel);
                    }

                    ++j;
                    ++i;
                } else if (channelCompareResult < 0) {
                    SubscriberSet tmpSubscriberSet;
                    tmpSubscriberSet.insert(origin);

                    channelSubscribers_.insert(j, make_pair(subListMsg->channel[i], tmpSubscriberSet));

                    subscribes.push_back(channel);

                    ++i;
                }
            }

            while (i < ntohs(subListMsg->numChannels)) {
                SubscriberSet tmpSubscriberSet;
                tmpSubscriberSet.insert(origin);

                channelSubscribers_.insert(j, make_pair(subListMsg->channel[i], tmpSubscriberSet));

                subscribes.push_back(subListMsg->channel[i]);

                ++i;
            }

            while (j != channelSubscribers_.end()) {
                if (j->second.erase(origin) > 0) {
                    unsubscribes.push_back(j->first);
                }

                ++j;
            }

            for (const auto& pmm : remotePeeringMethods_) {
                updateSubscriptions(pmm.second, origin, subscribes, unsubscribes);
            }

            j = channelSubscribers_.end();
            for (auto subIter = channelSubscribers_.begin(); subIter != channelSubscribers_.end(); j = subIter++) {
                if (channelSubscribers_.end() != j) {
                    if (j->second.empty()) {
                        channelSubscribers_.erase(j);
                    }
                }
            }
        } else if (ControlMessages::Connect == ntohl(msgType)) {
            const ConnectionMessage* connMsg = reinterpret_cast<const ConnectionMessage*>(msg.data);
            MessagePtr msgPtr(new (ntohl(msg.dataLength)) MessageStub(msg));
            bool newOffer = false;

            AVLOG_DEBUG("Got connection offer from {} over {}",
                        origin,
                        transportName((Transports::TransportTypes)ntohl(connMsg->transport)));

            {
                const PeerTransportPair connInfoKey(origin, (Transports::TransportTypes)ntohl(connMsg->transport));
                ConnectionInfoMap::iterator connInfoIter(connectionOffers_.find(connInfoKey));
                if (connInfoIter == connectionOffers_.end()) {
                    newOffer = true;
                    connectionOffers_[PeerTransportPair(
                        origin, (Transports::TransportTypes)ntohl(connMsg->transport))] = msgPtr;
                } else {
                    const MessagePtr& prevMsg((*connInfoIter).second);
                    const ConnectionMessage* prevConnMsg(reinterpret_cast<ConnectionMessage*>(prevMsg->data));
                    if (TimeStamp(prevConnMsg->offerEpoch) != TimeStamp(connMsg->offerEpoch)) {
                        newOffer = true;
                    }

                    (*connInfoIter).second = msgPtr;
                }
            }

            PeeringMethodsMap::iterator i =
                remotePeeringMethods_.find((Transports::TransportTypes)ntohl(connMsg->transport));
            if (i != remotePeeringMethods_.end()) {
                const PeerPtr& remotePeer(i->second);

                if (newOffer || !remotePeer->isConnectedTo(origin)) {
                    if (remotePeer->handleControlMsg(msgPtr) == Status::NewData) {
                        vector<string> subscribes;

                        for (ChannelSubscriberMap::iterator j = channelSubscribers_.begin();
                             j != channelSubscribers_.end();
                             ++j) {
                            if (j->second.find(origin) != j->second.end()) {
                                subscribes.push_back(j->first);
                            }
                        }

                        updateSubscriptions(remotePeer, origin, subscribes, vector<string>());
                    }
                }
            }
        }
    }

    return Status::Ok;
}

Status::States UdpDiscovery::updateSubscriptions(const PeerPtr remotePeer,
                                                 const string& origin,
                                                 const vector<string>& subscribes,
                                                 const vector<string>& unsubscribes)
{
    MessagePtr tmp(new (sizeof(DeclareChannelMessage)) MessageStub());

    memcpy(tmp->origin, origin.c_str(), origin.length() + 1);
    memcpy(tmp->channel, SIMPLECOMMS_CONTROL_CHANNEL, sizeof(SIMPLECOMMS_CONTROL_CHANNEL));
    tmp->length = tmp->dataLength = htonl(sizeof(DeclareChannelMessage));

    for (vector<string>::const_iterator i = subscribes.begin(); i != subscribes.end(); ++i) {
        DeclareChannelMessage* declareMessage(reinterpret_cast<DeclareChannelMessage*>(tmp->data));

        declareMessage->channelType = static_cast<ChannelType>(htonl(ChannelType::Subscriber));
        declareMessage->type = static_cast<ControlMessages::ControlMessageTypes>(htonl(ControlMessages::Declare));
        memcpy(&declareMessage->channel, i->c_str(), i->length() + 1);
        messageSetChecksum(*tmp); // TODO remove this once checksum is ignored on falcon boards

        remotePeer->handleControlMsg(tmp);
    }

    for (vector<string>::const_iterator i = unsubscribes.begin(); i != unsubscribes.end(); ++i) {
        DeclareChannelMessage* declareMessage(reinterpret_cast<DeclareChannelMessage*>(tmp->data));

        declareMessage->channelType = static_cast<ChannelType>(htonl(ChannelType::Subscriber));
        declareMessage->type = static_cast<ControlMessages::ControlMessageTypes>(htonl(ControlMessages::Undeclare));
        memcpy(&declareMessage->channel, i->c_str(), i->length() + 1);
        messageSetChecksum(*tmp); // TODO remove this once checksum is ignored on falcon boards

        remotePeer->handleControlMsg(tmp);
    }

    return Status::Ok;
}

} // namespace SimpleComms
