/*!
 ***********************************************************
 * @file UnixClient.cc
 * @brief A file implementing the UnixClient class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim
 * @date 03/09/2016
 * @remark
 * Added the cross-platform support
 *  -- 32-bit and big-endian
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * UnixClient.cc
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#include <SimpleComms/UnixClient.h>

#include "NetworkIntfUtils.h"
#include "lib/avlog/avlog.hpp"
#include "platform/middleware/libs/Comms/include/Comms/Channel.h"
#include <SimpleComms/ControlMessages.h>
#include <Threading/PThreadLocker.h>
#include <TimeStamp/otimeConversions.h>
#include <interfaces/baseTypes/Header.h>

#include <cerrno>

#include <algorithm>
#include <cstdint>
#include <cstring>
#include <fcntl.h>
#include <limits>
#include <random>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <unistd.h>

using namespace std;

namespace SimpleComms {

/**
 * @brief UnixClient constructor with client and local server path information
 *
 * @param serverPath The path of the local server
 * @param clientPath The path this client will register itself at
 */
UnixClient::UnixClient(const string& serverPath,
                       const string& clientPath,
                       Client::Callback callback,
                       const double& staleBufferPruningThreshold_s)
    : Client(callback)
    , serverPath_(serverPath)
    , clientPath_(clientPath)
    , fd_(-1)
    , serverfd_(-1)
    , rollingMessageID_(0)
    , staleBufferPruningThreshold_s_(staleBufferPruningThreshold_s)
    , threadRunning_(0)
{}

/*!
 * @brief This method shall make a unique client name for connecting to scs.
 * \n This method shall add given string as a prefix to the generated unique name.
 *
 * @param[in] prefix as string : The client name prefix. Adding given prefix to the generated client name does not
 * impact uniqueness.
 *
 * @return string: Unique client name
 */
std::string UnixClient::makeUniqueClientName(const std::string& prefix)
{
    std::string randName{"0123456789abcdefghijklmnopqrstuvwxyz0123456789"};
    std::mt19937 generator{std::random_device{}()};
    std::shuffle(randName.begin(), randName.end(), generator);
    // max 50 chars from prefix + 10 chars from randName
    return prefix.substr(0, 50) + "-" + randName.substr(0, 10);
}

/*!
 * @brief This method shall get the default SimpleComms server path.
 * \n The path is for connecting Unix Domain Sockets correctly.
 * \n By default, the path is $SCS_PATH if defined. Otherwise, "scs".
 *
 * @return string : Default path
 */
std::string UnixClient::getDefaultScsPath()
{
    const char* serverPath = getenv("SCS_PATH");
    if (serverPath == nullptr) {
        serverPath = "scs";
    }
    return std::string(serverPath);
}

/**
 * @brief UnixClient class destructor
 */
UnixClient::~UnixClient()
{
    // TODO: Call to virtual method 'UnixClient::disconnect' during destruction bypasses virtual dispatch
    disconnect();
}

Status::States UnixClient::connect()
{
    {
        // We plan on working with the file descriptors
        std::unique_lock<std::recursive_mutex> fdLock(fdMutex_);

        if (fd_ != -1) {
            AVLOG_WARN("Already connected");

            return Status::AlreadyConnected;
        }
        // If the receiverThread encountered an error and terminated the connection, we may have to join it now
        else if (threadRunning_) {
            // Temporarily release the file descriptors so any active threads can spot the -1 and terminate
            fdLock.unlock();

            threadRunning_--;
            pthread_join(receiverThreadHandle_, NULL);

            fdLock.lock();
        }

        // Tie fd_ to clientPath_ to listen for messages from the server
        {
            sockaddr_un servAddr{};

            const int maxPathLength = min(MAX_NULL_TERM_CLIENT_NAME, static_cast<int>(sizeof(servAddr.sun_path)) - 1);
            if (static_cast<int>(clientPath_.length() + 1) > maxPathLength) {
                AVLOG_WARN("Specified path exceeds maximum length ({} bytes): {}", maxPathLength, clientPath_);

                return Status::SocketPathTooLong;
            }
            if (clientPath_.empty()) {
                AVLOG_WARN("Specified path is empty");

                return Status::SocketPathTooLong;
            }

            if ((fd_ = socket(PF_UNIX, SOCK_DGRAM, 0)) < 0) {
                AVLOG_WARN("Error during socket(PF_UNIX, SOCK_DGRAM, 0): {}", strerror(errno));

                disconnect();

                return Status::ErrorCreatingSocket;
            }

            if (NetworkIntfUtils::setSocketBufferSize(fd_) != 0) {
                AVLOG_WARN("Error during setSocketBufferSize: {}", strerror(errno));

                disconnect();

                return Status::ErrorCreatingSocket;
            }

            memset(reinterpret_cast<void*>(&servAddr), 0, sizeof(servAddr));
            servAddr.sun_family = AF_UNIX;

            // Abstract namespace triggered via leading null
            // Available sun_path size is subtracted by 2 because is starts and ends with null
            strncpy(&servAddr.sun_path[1], clientPath_.c_str(), sizeof(servAddr.sun_path) - 2);

            if (::bind(fd_, reinterpret_cast<struct sockaddr*>(&servAddr), sizeof(servAddr)) < 0) {
                AVLOG_WARN("Error binding socket: {}", strerror(errno));

                disconnect();

                return Status::ErrorBindingSocket;
            }

            if (fcntl(fd_, F_SETFD, FD_CLOEXEC) < 0) {
                AVLOG_WARN("Error setting FD_CLOEXEC: {}", strerror(errno));

                disconnect();

                return Status::Error;
            }
        }

        // Tie serverfd_ to serverPath_ to send messages to the server
        {
            struct sockaddr_un servAddr;
            if (serverPath_.length() + 1 > sizeof(servAddr.sun_path) - 1) {
                AVLOG_WARN("Specified server path exceeds maximum length ({} bytes)", sizeof(servAddr.sun_path) - 1);

                disconnect();

                return Status::SocketPathTooLong;
            }

            memset(reinterpret_cast<void*>(&servAddr), 0, sizeof(servAddr));
            servAddr.sun_family = AF_UNIX;

            // Abstract path_space triggered via leading null
            strncpy(&servAddr.sun_path[1], serverPath_.c_str(), sizeof(servAddr.sun_path) - 2);

            if ((serverfd_ = socket(PF_UNIX, SOCK_DGRAM, 0)) < 0) {
                AVLOG_WARN("Error during socket(PF_UNIX, SOCK_DGRAM, 0): {}", strerror(errno));

                disconnect();

                return Status::ErrorCreatingSocket;
            }

            if (NetworkIntfUtils::setSocketBufferSize(serverfd_) != 0) {
                AVLOG_WARN("Error during setSocketBufferSize: {}", strerror(errno));

                disconnect();

                return Status::ErrorCreatingSocket;
            }

            if (::connect(serverfd_, reinterpret_cast<struct sockaddr*>(&servAddr), sizeof(servAddr)) < 0) {
                AVLOG_WARN("Error during connect: {}", strerror(errno));

                disconnect();

                return Status::ErrorConnectingSocket;
            }

            if (fcntl(serverfd_, F_SETFD, FD_CLOEXEC) < 0) {
                AVLOG_WARN("Error setting FD_CLOEXEC: {}", strerror(errno));

                disconnect();

                return Status::Error;
            }
        }

        // Spawn a thread that sits on fd_ and queues the incoming data
        if (pthread_create(&receiverThreadHandle_, NULL, receiverThreadLauncher, this) == 0) {
            // Flag that we successfully launched a thread for later joining
            threadRunning_++;
        } else {
            AVLOG_WARN("Error initializing receiver thread: {}", strerror(errno));

            disconnect();

            return Status::Error;
        }

    } // We are done messing around with the file descriptors

    // Simple exchange to make the the connection is established
    {
        // Reset the connectionAcknowledged promise.
        connectionAcknowledged_ = std::promise<void>();
        std::future<void> connectionFuture = connectionAcknowledged_.get_future();

        // Send the local server a notice that we want to connect
        ConnectionMessage connectionMessage;

        connectionMessage.type = (ControlMessages::ControlMessageTypes)htonl(ControlMessages::Connect);
        connectionMessage.transport = (Transports::TransportTypes)htonl(Transports::UnixDomainSocket);

        const std::string controlChannel{SIMPLECOMMS_CONTROL_CHANNEL};
        const std::string connectionMessageStr{reinterpret_cast<const char*>(&connectionMessage),
                                               sizeof(ConnectionMessage)};
        const Status::States sendMsgStatus = sendMsg(controlChannel, connectionMessageStr);

        if (sendMsgStatus != Status::Ok) {
            AVLOG_WARN("Error sending connection message: {}", strerror(errno));

            return sendMsgStatus;
        }

        // Wait for local server to echo our notice
        if (connectionFuture.wait_for(std::chrono::seconds(10)) == std::future_status::timeout) {
            AVLOG_WARN("Connection not acknowledged");

            disconnect();

            return Status::Error;
        }
    }

    {
        std::scoped_lock<std::mutex> channelsLock(channelsMutex_);

        for (ChannelMap::iterator i = channels_.begin(); i != channels_.end(); ++i) {
            declare((*i).first, (*i).second.type, (*i).second.maxReceiveDepth);
        }
    }
    return Status::Ok;
}

Status::States UnixClient::disconnect()
{
    return closeConnection(false);
}

Status::States UnixClient::declare(const string& channel, const ChannelType type, const uint32_t receiveCapacity)
{
    Status::States status = Status::Ok;

    if (channel.length() < MAX_NULL_TERM_CHANNEL_NAME) {
        bool sendRequest = false;

        {
            std::scoped_lock<std::mutex> channelsLock(channelsMutex_);
            const ChannelMap::iterator i(channels_.find(channel));

            if (i == channels_.end()) {
                sendRequest = true;
                channels_.insert(make_pair(channel, ChannelData(receiveCapacity, type)));
            } else {
                ChannelData& channelData((*i).second);

                ChannelType newType = static_cast<ChannelType>(type | channelData.type);

                sendRequest = channelData.type != newType;
                channelData.type = newType;

                // Never decrease queue depth - could cause problems. Convert 0 ("unbounded") to uint32_t max
                const uint32_t actualCapacity{(receiveCapacity == 0) ? std::numeric_limits<uint32_t>::max()
                                                                     : receiveCapacity};
                const uint32_t greaterValue{std::max(channelData.maxReceiveDepth, actualCapacity)};

                AVLOG_INFO("Channel {} already declared with queue depth {}. Will now use queue depth {}",
                           channel,
                           channelData.maxReceiveDepth,
                           greaterValue);
                channelData.maxReceiveDepth = greaterValue;

                if (!channelData.acknowledged) {
                    sendRequest = true;
                }
            }
        }

        if (sendRequest && type != ChannelType::Unlisted) {
            DeclareChannelMessage declareMessage;

            declareMessage.type = static_cast<ControlMessages::ControlMessageTypes>(htonl(ControlMessages::Declare));
            declareMessage.channelType = static_cast<ChannelType>(htonl(type));
            memcpy(&declareMessage.channel, channel.c_str(), channel.length() + 1);

            status = sendMsg(SIMPLECOMMS_CONTROL_CHANNEL,
                             string(reinterpret_cast<const char*>(&declareMessage), sizeof(DeclareChannelMessage)));
        }
    } else {
        status = Status::ChannelNameTooLong;
    }

    return status;
}

Status::States UnixClient::undeclare(const string& channel, const ChannelType type)
{
    Status::States status = Status::Ok;
    bool sendRequest = false;

    {
        std::scoped_lock<std::mutex> channelsLock(channelsMutex_);
        const ChannelMap::iterator i(channels_.find(channel));

        if (i != channels_.end()) {
            if ((i->second.type & type) != 0) {
                sendRequest = true;
            }

            i->second.type = static_cast<ChannelType>(i->second.type & ~type);

            if (i->second.type == 0) {
                channels_.erase(i);
            }
        }
    }

    if (sendRequest) {
        DeclareChannelMessage undeclareMessage;

        undeclareMessage.type = static_cast<ControlMessages::ControlMessageTypes>(htonl(ControlMessages::Undeclare));
        undeclareMessage.channelType = static_cast<ChannelType>(htonl(type));
        memcpy(&undeclareMessage.channel, channel.c_str(), channel.length() + 1);

        status = sendMsg(SIMPLECOMMS_CONTROL_CHANNEL,
                         string(reinterpret_cast<const char*>(&undeclareMessage), sizeof(DeclareChannelMessage)));
    }

    return status;
}

/**
 * @brief Retrieve totals for channel/origin
 *
 * @param statsKey channel/origin to fetch totals for
 *
 * @return ChannelOriginStatsMap::iterator to either the existing entry or newly created entry for channel/origin
 */
UnixClient::ChannelOriginStatsMap::iterator UnixClient::fetchTotals(const ChannelOriginKey& statsKey)
{
    ChannelOriginStatsMap::iterator statsIter(channelOriginStats_.find(statsKey));

    if (statsIter == channelOriginStats_.end()) {
        const pair<ChannelOriginKey, ChannelOriginTotals> newStats(statsKey, ChannelOriginTotals());
        const pair<ChannelOriginStatsMap::iterator, bool> insertionResult(channelOriginStats_.insert(newStats));

        statsIter = insertionResult.first;

        memcpy((*statsIter).second.channel, statsKey.first.c_str(), statsKey.first.length() + 1);
        memcpy((*statsIter).second.origin, statsKey.second.c_str(), statsKey.second.length() + 1);
    }

    return statsIter;
}

Status::States UnixClient::sendMsgV(const string& channel,
                                    const string& data,
                                    std::vector<av::latency::Pred>&& predecessors)
{
    MessageBase msg{};
    if (channel.length() + 1 > sizeof(msg.channel)) {
        return Status::ChannelNameTooLong;
    }

    memcpy(&msg.origin, clientPath_.c_str(), clientPath_.length() + 1);
    memcpy(&msg.channel, channel.c_str(), channel.length() + 1);
    msg.sendMsgTime.setNow();

    // Do this before running htonl() on tv_sec and tv_usec.
    {
        const ottotime::otimeSpec sendTimeSpec{msg.sendMsgTime.tv_sec, msg.sendMsgTime.tv_usec * 1000};
        const ottotime::otimePoint sendTime{sendTimeSpec};
        av::LatencyLogger::getLogger().SendEvent(channel, sendTime, std::move(predecessors));
    }

    msg.sendMsgTime.tv_sec = htonls(msg.sendMsgTime.tv_sec);
    msg.sendMsgTime.tv_usec = htonls(msg.sendMsgTime.tv_usec);
    msg.rollingMessageID = htonl(rollingMessageID_++);
    const auto length = static_cast<uint32_t>(data.length());
    msg.length = htonl(length);
    uint32_t totalPages = length / MAX_FRAGMENT_PAYLOAD + ((length % MAX_FRAGMENT_PAYLOAD) > 0);
    msg.totalPages = htonl(totalPages);

    Status::States status = Status::Ok;
    // report send event right as we start sending the data
    task::Channel::reportMessage(channel, data.size());
    for (uint32_t i = 0; (i < totalPages) & (status == Status::Ok); ++i) {
        msg.pageNumber = htonl(i + 1);
        msg.dataLength = htonl(min(length - MAX_FRAGMENT_PAYLOAD * i, MAX_FRAGMENT_PAYLOAD));
        msg.data = reinterpret_cast<unsigned char*>(const_cast<char*>(data.c_str())) + MAX_FRAGMENT_PAYLOAD * i;

        status = sendMsgM(msg);
    }

    return status;
}

bool UnixClient::hasMsg() const
{
    return (messagesAvailable_ > 0);
}

Status::States UnixClient::receiveMsg(const string& channel, string& data, MessageInfo& msgInfo)
{
    std::scoped_lock<std::mutex> channelsLock(channelsMutex_);
    Status::States returnStatus = Status::Ok;

    const ChannelMap::iterator channelIter(channels_.find(channel));
    if (channelIter != channels_.end()) {
        ChannelData& channelData((*channelIter).second);

        if (!channelData.completeBuffers.empty()) {
            {
                FragmentBuffer& fragBuf(*channelData.completeBuffers.front());

                msgInfo.origin = fragBuf.lastHeader.origin;
                msgInfo.sendMsgTime.tv_sec = ntohls(fragBuf.lastHeader.sendMsgTime.tv_sec);
                msgInfo.sendMsgTime.tv_usec = ntohls(fragBuf.lastHeader.sendMsgTime.tv_usec);
                msgInfo.msgReadyTime = fragBuf.lastMessageAdded;

                data.swap(fragBuf.data);
            }

            channelData.completeBuffers.pop();
            --messagesAvailable_;

            // we don't need to reportMessage here because the only caller of this method should be
            // SCChannel, which calls it itself.
            returnStatus = Status::NewData;
        } else {
            if (fd_ == -1) {
                returnStatus = Status::NotConnected;
            }
        }
    }
    // note: any calls to LatencyLogger should performed by caller
    return returnStatus;
}

void UnixClient::allowThrottling(std::function<void(bool)> callback)
{
    pauseCallback_ = callback;
    ThrottleMessage throttleMessage;
    sendMsg(SIMPLECOMMS_CONTROL_CHANNEL,
            string(reinterpret_cast<const char*>(&throttleMessage), sizeof(throttleMessage)));
}

/**
 * @brief Internal function to send a message fragment
 *
 * @param msg Fragment to send
 *
 * @return XXX
 */
Status::States UnixClient::sendMsgM(const Message& msg)
{
    const ssize_t sendStatus = send(serverfd_, &msg, msg.MessageLength(), MSG_NOSIGNAL);

    if (sendStatus == -1) {
        // XXX
        return Status::Error;
    }

    return Status::Ok;
}

Status::States UnixClient::sendMsgM(const MessageBase& msg)
{
    struct iovec messageIovec[2];

    messageIovec[0].iov_base = const_cast<MessageBase*>(&msg);
    messageIovec[0].iov_len = sizeof(MessageHeader);
    messageIovec[1].iov_base = const_cast<unsigned char*>(msg.data);
    messageIovec[1].iov_len = ntohl(msg.dataLength);

    struct msghdr messageHeader;

    messageHeader.msg_name = NULL;
    messageHeader.msg_namelen = 0;
    messageHeader.msg_iov = messageIovec;
    messageHeader.msg_iovlen = 2;
    messageHeader.msg_control = NULL;
    messageHeader.msg_controllen = 0;
    messageHeader.msg_flags = 0;

    const ssize_t sendStatus = sendmsg(serverfd_, &messageHeader, MSG_NOSIGNAL);

    if (sendStatus == -1) {
        // XXX
        return Status::Error;
    }

    return Status::Ok;
}

Status::States UnixClient::handleIncomingMsg()
{
    Status::States status(Status::Ok);
    MessageBase msgBase;
    const ssize_t recvStatus = recv(fd_, &msgBase, sizeof(MessageHeader), MSG_DONTWAIT | MSG_PEEK);

    if ((recvStatus == -1) && (errno == EAGAIN)) {
        return Status::Ok;
    } else if (recvStatus == -1) {
        closeConnection(true);

        return Status::Error;
    }

    totalFragmentsReceived_++;

    if (string(msgBase.channel) == SIMPLECOMMS_CONTROL_CHANNEL) {
        Message fullMsg;
        recv(fd_, reinterpret_cast<void*>(&fullMsg), sizeof(Message), 0);

        fdMutex_.unlock();
        status = handleControlMsg(fullMsg);
        fdMutex_.lock();
    } else {
        const MessageKey fragmentBufferKey(msgBase.origin, ntohl(msgBase.rollingMessageID));
        FragmentBufferMap::iterator fbIter(fragmentBuffers_.find(fragmentBufferKey));

        if (fbIter == fragmentBuffers_.end()) {
            const pair<MessageKey, FragmentBufferPtr> newBuffer(
                fragmentBufferKey, FragmentBufferPtr(new FragmentBuffer(fragmentBuffersLRU_.end())));
            const pair<FragmentBufferMap::iterator, bool> insertionResult(fragmentBuffers_.insert(newBuffer));

            fbIter = insertionResult.first;
        }

        FragmentBuffer& fragBuf(*(*fbIter).second);

        // If we are on the recently used list, pull us off for now
        if (fragBuf.lruEntry != fragmentBuffersLRU_.end()) {
            fragmentBuffersLRU_.erase(fragBuf.lruEntry);
        }

        if (fragBuf.receivedCount == 0) {
            fragBuf.data.resize(ntohl(msgBase.length), 0);
            fragBuf.received.resize(ntohl(msgBase.totalPages));
        }

        {
            msgBase.data = reinterpret_cast<unsigned char*>(
                const_cast<char*>(&fragBuf.data[(ntohl(msgBase.pageNumber) - 1) * MAX_FRAGMENT_PAYLOAD]));

            struct iovec messageIovec[2];

            messageIovec[0].iov_base = &fragBuf.lastHeader;
            messageIovec[0].iov_len = sizeof(MessageHeader);
            messageIovec[1].iov_base = msgBase.data;
            messageIovec[1].iov_len = ntohl(msgBase.dataLength);

            struct msghdr messageHeader;

            messageHeader.msg_name = NULL;
            messageHeader.msg_namelen = 0;
            messageHeader.msg_iov = messageIovec;
            messageHeader.msg_iovlen = 2;
            messageHeader.msg_control = NULL;
            messageHeader.msg_controllen = 0;
            messageHeader.msg_flags = 0;

            recvmsg(fd_, &messageHeader, 0);
        }

        fdMutex_.unlock();

        fragBuf.lastMessageAdded.setNow();
        fragBuf.received[ntohl(msgBase.pageNumber) - 1] = true;
        fragBuf.receivedCount++;
        fragBuf.receivedBytes += ntohl(msgBase.dataLength);

        // Update stats
        ChannelOriginTotals& stats((*fetchTotals(ChannelOriginKey(msgBase.channel, msgBase.origin))).second);

        stats.totalFragments++;
        stats.totalBytes += ntohl(msgBase.dataLength);
        struct TimeStamp sendMsgTime;
        sendMsgTime.tv_sec = ntohls(msgBase.sendMsgTime.tv_sec);
        sendMsgTime.tv_usec = ntohls(msgBase.sendMsgTime.tv_usec);
        stats.totalDelay += static_cast<double>(fragBuf.lastMessageAdded - sendMsgTime);

        // Handle message completion
        if (fragBuf.receivedCount == ntohl(msgBase.totalPages)) {
            totalMessagesReceived_++;
            stats.completedMessages++;
            stats.completedFragments += fragBuf.receivedCount;
            stats.completedBytes += ntohl(msgBase.length);

            if (callback_) {
                {
                    // If a callback has been registered, calling callback_ is where received messages
                    // are provided to the consumer. Currently, that means added to the EventFramework's
                    // input queue, ready to be consumed by customer.
                    const ottotime::otimeSpec sendTimeSpec{sendMsgTime.tv_sec, sendMsgTime.tv_usec * 1000};
                    const ottotime::otimePoint sendTime{sendTimeSpec};
                    // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-array-to-pointer-decay)
                    av::LatencyLogger::getLogger().ReceiveEvent(msgBase.channel, sendTime);
                    task::Channel::reportMessage(static_cast<const char*>(msgBase.channel),
                                                 stats.completedBytes,
                                                 av::timePointFromOtimePoint(sendTime));
                }
                callback_(msgBase.channel, fbIter->second->data, fragmentBufferKey);
            }

            {
                std::scoped_lock<std::mutex> channelsLock(channelsMutex_);
                ChannelMap::iterator channelIter(channels_.find(msgBase.channel));

                if (channelIter != channels_.end()) {
                    ChannelData& channelData((*channelIter).second);

                    channelData.completeBuffers.push((*fbIter).second);

                    // If we've buffered more than the maximum number of messages, pop one.
                    // Otherwise, increment the count of available messages.
                    if ((channelData.maxReceiveDepth > 0) &&
                        (channelData.completeBuffers.size() > channelData.maxReceiveDepth)) {
                        channelData.completeBuffers.pop();
                    } else {
                        ++messagesAvailable_;
                    }
                }
            }

            fragmentBuffers_.erase(fbIter);
        } else {
            // Incomplete fragBuf still in play, reintroduce to LRU
            fragBuf.lruEntry = fragmentBuffersLRU_.insert(fragmentBuffersLRU_.end(), fbIter);
        }

        fdMutex_.lock();
    }

    return status;
}

Status::States UnixClient::handleControlMsg(const Message& msg)
{
    uint32_t msgType = ntohl(*reinterpret_cast<const uint32_t*>(msg.data));
    AVLOG_DEBUG("UnixClient::handleControlMsg ({})", msgType);
    if (callback_) {
        MessageKey key(msg.origin, ntohl(msg.rollingMessageID));
        std::string blob(reinterpret_cast<const char*>(msg.data), ntohl(msg.length));
        callback_(msg.channel, blob, key);
    }

    switch (msgType) {
        case ControlMessages::Connect: {
            connectionAcknowledged_.set_value();
            break;
        }

        case ControlMessages::Declare: {
            const DeclareChannelMessage* declareMessage = reinterpret_cast<const DeclareChannelMessage*>(msg.data);
            std::scoped_lock<std::mutex> channelsLock(channelsMutex_);
            const ChannelMap::iterator i(channels_.find(declareMessage->channel));

            if (i != channels_.end()) {
                (*i).second.acknowledged = true;
            }

            break;
        }

        case ControlMessages::Pause:
        case ControlMessages::Resume: {
            if (pauseCallback_) {
                pauseCallback_(msgType == ControlMessages::Pause);
            }
        }
    }

    return Status::Ok;
}

Status::States UnixClient::closeConnection(const bool receiverThread)
{
    {
        std::scoped_lock<std::recursive_mutex> fdLock(fdMutex_);

        if (fd_ != -1) {
            close(fd_);
            fd_ = -1;
        }

        if (serverfd_ != -1) {
            {
                DisconnectionMessage disconnectionMessage;

                disconnectionMessage.type = (ControlMessages::ControlMessageTypes)htonl(ControlMessages::Disconnect);
                disconnectionMessage.transport = (Transports::TransportTypes)htonl(Transports::UnixDomainSocket);

                sendMsg(SIMPLECOMMS_CONTROL_CHANNEL,
                        string(reinterpret_cast<const char*>(&disconnectionMessage), sizeof(disconnectionMessage)));
            }

            close(serverfd_);
            serverfd_ = -1;
        }
    }

    if (!receiverThread && threadRunning_) {
        threadRunning_--;
        pthread_join(receiverThreadHandle_, NULL);
    }

    {
        std::scoped_lock<std::mutex> channelsLock(channelsMutex_);

        for (ChannelMap::iterator i = channels_.begin(); i != channels_.end(); ++i) {
            (*i).second.acknowledged = false;
        }
    }

    return Status::Ok;
}

/**
 * @brief Internal function to cull partial fragment buffers that have been around past a threshold
 */
void UnixClient::pruneStalePartialBuffers()
{
    TimeStamp now;
    now.setNow();

    while (!fragmentBuffersLRU_.empty()) {
        FragmentBuffer& fragBuf(*(*fragmentBuffersLRU_.front()).second);

        if (static_cast<double>(now - fragBuf.lastMessageAdded) > staleBufferPruningThreshold_s_) {
            uint32_t totalPages = fragBuf.lastHeader.totalPages;
            AVLOG_WARN("Pruning stale message from {} on channel {} with {} out of {} parts",
                       fragBuf.lastHeader.origin,
                       fragBuf.lastHeader.channel,
                       fragBuf.receivedCount,
                       totalPages);

            { // Update stats
                const ChannelOriginKey statsKey(fragBuf.lastHeader.channel, fragBuf.lastHeader.origin);
                ChannelOriginTotals& stats((*fetchTotals(statsKey)).second);

                stats.incompleteMessages++;
                stats.expectedFragments += fragBuf.lastHeader.totalPages;
                stats.receivedFragments += fragBuf.receivedCount;
                stats.receivedBytes += fragBuf.receivedBytes;
                stats.expectedBytes += fragBuf.lastHeader.length;
            }

            fragmentBuffers_.erase(fragmentBuffersLRU_.front());
            fragmentBuffersLRU_.pop_front();
        } else {
            break;
        }
    }
}

void* UnixClient::receiverThreadLauncher(void* arg)
{
    static_cast<UnixClient*>(arg)->receiverThread();

    return NULL;
}

void UnixClient::receiverThread()
{
    lastReport_.setNow();
    totalFragmentsReceived_ = 0;
    totalMessagesReceived_ = 0;

    std::unique_lock<std::recursive_mutex> fdLock(fdMutex_);
    while (fd_ != -1) {
        fd_set rfds;

        {
            const int fd = fd_;
            fdLock.unlock();

            FD_ZERO(&rfds);
            FD_SET(fd, &rfds);

            struct timeval waitLimit;
            waitLimit.tv_sec = 1;
            waitLimit.tv_usec = 0;

            const int selectStatus = select(fd + 1, &rfds, NULL, NULL, &waitLimit);

            if (selectStatus == -1) {
                switch (errno) {
                    case EINTR:
                        // Ignore
                        break;
                    default:
                        AVLOG_WARN("Error during select: {}", strerror(errno));

                        closeConnection(true);

                        return;
                }
            }
        }

        pruneStalePartialBuffers();

        fdLock.lock();

        if (fd_ == -1) {
            break;
        }

        if (FD_ISSET(fd_, &rfds)) {
            if (handleIncomingMsg() != Status::Ok) {
                break;
            }
        }

        {
            TimeStamp now;
            now.setNow();
            const auto delta = static_cast<double>(now - lastReport_);

            if (delta > 1) {
                AVLOG_DEBUG("Total Fragments: {}; Total Messages: {}",
                            totalFragmentsReceived_ / delta,
                            totalMessagesReceived_ / delta);

                totalFragmentsReceived_ = 0;
                totalMessagesReceived_ = 0;

                { // Publish stats
                    string statsBuffer;
                    for (ChannelOriginStatsMap::iterator i(channelOriginStats_.begin()); i != channelOriginStats_.end();
                         ++i) {
                        const ChannelOriginTotals& stats((*i).second);

                        statsBuffer.append(reinterpret_cast<const char*>(&stats), sizeof(stats));
                    }

                    sendMsg(SIMPLECOMMS_STATS_CHANNEL, statsBuffer);

                    channelOriginStats_.clear();
                }

                lastReport_ = now;
            }
        }
    }
}

std::vector<std::string> UnixClient::getPublishListMessage()
{
    std::lock_guard<std::mutex> guard(channelsMutex_);
    std::vector<std::string> publishedChannels;

    for (const auto& channel : channels_) {
        if (channel.second.type == Publisher || channel.second.type == SubPub) {
            publishedChannels.push_back(channel.first);
        }
    }
    return publishedChannels;
}

} // namespace SimpleComms
