/*!
 ***********************************************************
 * @file UnixPeer.cc
 * @brief A file implementing the UnixPeer class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim
 * @date 03/09/2016
 * @remark
 * Added the cross-platform support
 *  -- 32-bit and big-endian
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * UnixPeer.cc
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#include "UnixPeer.h"
#include "NetworkIntfUtils.h"
#include "lib/avlog/avlog.hpp"
#include <SimpleComms/ControlMessages.h>

#include <cerrno>
#include <cstdio>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <unistd.h>

using namespace std;

namespace SimpleComms {

UnixPeer::UnixPeer(const string& path, const TimeStamp& staleFragmentPruningThreshold)
    : BasePeer(Transports::UnixDomainSocket, false, staleFragmentPruningThreshold)
{
    peerName_ = path;
}

UnixPeer::~UnixPeer() {}

Status::States UnixPeer::connect()
{
    if (fd_ != -1) {
        AVLOG_INFO("UnixPeer::connect -> Status::AlreadyConnected");
        return Status::AlreadyConnected;
    }

    struct sockaddr_un servAddr;

    if (peerName_.length() > sizeof(servAddr.sun_path) - 2) {
        AVLOG_ERROR("UnixPeer::connect -> Status::SocketPathTooLong");
        return Status::SocketPathTooLong;
    }

    if ((fd_ = socket(PF_UNIX, SOCK_DGRAM | SOCK_NONBLOCK, 0)) < 0) {
        AVLOG_ERROR("UnixPeer::connect -> Status::ErrorCreatingSocket");
        return Status::ErrorCreatingSocket;
    }

    if (NetworkIntfUtils::setSocketBufferSize(fd_) != 0) {
        close(fd_);
        fd_ = -1;

        return Status::ErrorCreatingSocket;
    }

    memset(reinterpret_cast<void*>(&servAddr), 0, sizeof(servAddr));
    servAddr.sun_family = AF_UNIX;

    // Abstract namespace triggered via leading null
    strncpy(&servAddr.sun_path[1], peerName_.c_str(), sizeof(servAddr.sun_path) - 2);

    if (::bind(fd_, reinterpret_cast<struct sockaddr*>(&servAddr), sizeof(servAddr)) < 0) {
        AVLOG_ERROR("UnixPeer::connect -> Status::ErrorBindingSocket: {}", strerror(errno));
        close(fd_);
        fd_ = -1;

        return Status::ErrorBindingSocket;
    }

    return Status::Ok;
}

Status::States UnixPeer::disconnect()
{
    close(fd_);
    fd_ = -1;

    return Status::Ok;
}

Status::States UnixPeer::getConnectionInfo(Message& msg)
{
    ConnectionMessage* connMsg = reinterpret_cast<ConnectionMessage*>(msg.data);

    connMsg->type = (ControlMessages::ControlMessageTypes)htonl(ControlMessages::Connect);
    connMsg->transport = (Transports::TransportTypes)htonl(Transports::UnixDomainSocket);

    return Status::Ok;
}

/**
 * @brief Connect to the given client entry using the given connection information
 *
 * @param namedClient Pointer to the client entry to connect to
 *
 * Assumes clientsMutex_ is held by calling function
 * System calls are made while holding clientsMutex_
 */
bool UnixPeer::clientConnect(const ClientDataPtr& namedClient, const ConnectionMessage*)
{
    ClientData& currentClient(*namedClient);

    if (currentClient.fd != -1) {
        clientReset(namedClient);
    }

    AVLOG_DEBUG("Connecting to client {}", currentClient.name);

    struct sockaddr_un servAddr;
    if (currentClient.name.length() > sizeof(servAddr.sun_path) - 2) {
        return false;
    }

    memset(reinterpret_cast<void*>(&servAddr), 0, sizeof(servAddr));
    servAddr.sun_family = AF_UNIX;

    // Abstract namespace triggered via leading null
    strncpy(&servAddr.sun_path[1], currentClient.name.c_str(), sizeof(servAddr.sun_path) - 2);

    if ((currentClient.fd = socket(AF_UNIX, SOCK_DGRAM | SOCK_NONBLOCK, 0)) < 0) {
        AVLOG_WARN("Error creating socket: {}", strerror(errno));

        return false;
    }

    if (NetworkIntfUtils::setSocketBufferSize(currentClient.fd) != 0) {
        AVLOG_WARN("Error setting socket buffer size: {}", strerror(errno));

        clientReset(namedClient);

        return false;
    }

    if (::connect(currentClient.fd, reinterpret_cast<struct sockaddr*>(&servAddr), sizeof(servAddr)) < 0) {
        AVLOG_WARN("Error connecting socket: {}", strerror(errno));

        clientReset(namedClient);

        return false;
    }

    currentClient.readyToWrite = true;

    return true;
}

} // namespace SimpleComms
