/*!
 ***********************************************************
 * @file Client.h
 * @brief A file defining the Client class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * Client.h
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#ifndef _SIMPLECOMMS_CLIENT_H_
#define _SIMPLECOMMS_CLIENT_H_

#include "BaseTypes.h"
#include "MessageInfo.h"
#include "Status.h"
#include "platform/middleware/libs/LatencyLogger/include/LatencyLogger/LatencyLogger.h"

#include <optional>
#include <string>
#include <vector>

namespace SimpleComms {

class Client
{

public:
    /// Origin and rollingMessageID are unique over significant periods making it a suitable message key.
    using MessageKey = std::pair<std::string, uint32_t>;
    /// Callback function interface to notify the client code when each complete message is received.
    using Callback = std::function<void(const std::string&, const std::string&, const MessageKey&)>;

    /// Constructor.
    Client(Callback callback = Callback());
    // Disable copying.
    Client(const Client&) = delete;
    Client& operator=(const Client&) = delete;
    /// Virtual destructor.
    virtual ~Client() {}

    /// Connect to the messaging system.
    /// @return Status::Ok if successful.
    virtual Status::States connect() = 0;
    /// Disconnect from the messaging system.
    /// @return Status::Ok if successful.
    virtual Status::States disconnect() = 0;

    /// Declare that we are subscribing or publishing to a channel.
    /// @param channel The channel name.
    /// @param type The type (subscribing, publishing, or both).
    /// @param receiveCapacity The receive capacity for the reader being declared. 0 for unlimited depth.
    /// @return Status::Ok if successful.
    virtual Status::States declare(const std::string& channel, ChannelType type, uint32_t receiveCapacity) = 0;
    /// Stop subscribing or publishing to a channel.
    /// @param channel The channel name.
    /// @param type The type (subscribing, publishing, or both).
    /// @return Status::Ok if successful.
    virtual Status::States undeclare(const std::string& channel, ChannelType type) = 0;

    /// Send a message.
    /// @param channel Name of the channel to send the message to.
    /// @param data The data to send to named channel.
    /// @param predecessors Specifiers for this event's predecessors (the events that triggered this one).
    /// @return Status::Ok if successful.
    Status::States sendMsg(const std::string& channel,
                           const std::string& data,
                           std::vector<av::latency::Pred>&& predecessors = {});

    /// Send a message.
    /// @param channel Name of the channel to send the message to.
    /// @param data The data to send to named channel.
    /// @param predecessors Specifiers for this event's predecessors (the events that triggered this one).
    /// @return Status::Ok if successful.
    virtual Status::States sendMsgV(const std::string& channel,
                                    const std::string& data,
                                    std::vector<av::latency::Pred>&& predecessors) = 0;

    /// Check if a message is ready to receive.
    /// @return True if receiveMsg() will return new data.
    virtual bool hasMsg() const = 0;

    /// Receive a message.
    /// @param channel Name of the channel to send the message to.
    /// @param[out] data A reference to a string where the received data will be placed.
    /// @param[out] msgInfo A reference to an object to store message info.
    /// @return Status::NewData if data was received; Status::Ok if no error occurred but no data was received.
    virtual Status::States receiveMsg(const std::string& channel, std::string& data, MessageInfo& msgInfo) = 0;

    /// \brief notify that this client responds to requests to pause its output. Must be called after connecting.
    virtual void allowThrottling(std::function<void(bool)> callback) = 0;

    /// \brief get all published channels from client
    virtual std::vector<std::string> getPublishListMessage() = 0;

protected:
    const Callback callback_; ///< Callback for when complete messages have been received.
};

inline Status::States Client::sendMsg(const std::string& channel,
                                      const std::string& data,
                                      std::vector<av::latency::Pred>&& predecessors)
{
    return sendMsgV(channel, data, std::move(predecessors));
}

} // namespace SimpleComms

#endif
