/*!
 ***********************************************************
 * @file ControlMessages.h
 * @brief A file defining Control Messages.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim
 * @date 03/09/2016
 * @remark
 * Added the cross-platform support
 *  -- 32-bit and big-endian
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * ControlMessages.h
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#ifndef _SIMPLECOMMS_CONTROLMESSAGES_H_
#define _SIMPLECOMMS_CONTROLMESSAGES_H_

#include <stdint.h>

#include "BaseTypes.h"
#include "Constants.h"
#include "TimeStamp.h"

#include <arpa/inet.h>

#include <string>
#include <vector>

namespace SimpleComms {

namespace ControlMessages {
enum ControlMessageTypes
{
    Connect,
    Disconnect,
    Declare,
    Undeclare,
    SubscriptionList,
    PublishList,
    Reproducible, // Informs SCS that a client uses the reproducible framework
    Schedule,     // Asks SCS to schedule a client's executive function
    Executive,    // An executive function trigger from SCS
    Replay,       // Informs a client that SCS is running in "replay" mode
    Drop,         // Informs SCS that a client dropped a message
    Throttle,     // Informs SCS that a client is capable of throttling
    Pause,        // Asks a client to stop sending messages
    Resume        // Asks a client to resume sending messages
};
}

namespace Transports {
enum TransportTypes
{
    UnixDomainSocket,
    UDP,
    TCP
};
}

const char* transportName(const Transports::TransportTypes& transport);

struct ConnectionMessage
{
    ControlMessages::ControlMessageTypes type;
    Transports::TransportTypes transport;
    struct TimeStamp offerEpoch;
} __attribute__((packed));

struct ConnectionMessageIPV4
{
    ControlMessages::ControlMessageTypes type;
    Transports::TransportTypes transport;
    struct TimeStamp offerEpoch;
    uint32_t inetAddr;
    uint16_t inetPort;
} __attribute__((packed));

struct DisconnectionMessage
{
    ControlMessages::ControlMessageTypes type;
    Transports::TransportTypes transport;
} __attribute__((packed));

struct DeclareChannelMessage
{
    ControlMessages::ControlMessageTypes type;
    ChannelType channelType;
    char channel[MAX_NULL_TERM_CHANNEL_NAME];
} __attribute__((packed));

struct SubscriptionListMessage
{
    ControlMessages::ControlMessageTypes type;
    uint16_t numChannels;
    char channel[999][MAX_NULL_TERM_CHANNEL_NAME];
} __attribute__((packed));

// Helper template for a control message with an auto-initialized type field.
template<ControlMessages::ControlMessageTypes Type>
struct TypedMessage
{
    ControlMessages::ControlMessageTypes type;
    TypedMessage() { type = static_cast<ControlMessages::ControlMessageTypes>(htonl(Type)); }
} __attribute__((packed));

struct ReproducibleMessage : public TypedMessage<ControlMessages::Reproducible>
{
} __attribute__((packed));

struct ScheduleMessage : public TypedMessage<ControlMessages::Schedule>
{
    int64_t nsecInterval{0};
} __attribute__((packed));

struct ExecutiveMessage : public TypedMessage<ControlMessages::Executive>
{
    int64_t tv_sec{-1};
    int64_t tv_nsec{-1};
} __attribute__((packed));

struct ReplayMessage : public TypedMessage<ControlMessages::Replay>
{
} __attribute__((packed));

struct DropMessage : public TypedMessage<ControlMessages::Drop>
{
    char origin[MAX_NULL_TERM_CLIENT_NAME];
    uint32_t id{0};
} __attribute__((packed));

struct ThrottleMessage : public TypedMessage<ControlMessages::Throttle>
{
} __attribute__((packed));

struct PauseMessage : public TypedMessage<ControlMessages::Pause>
{
} __attribute__((packed));

struct ResumeMessage : public TypedMessage<ControlMessages::Resume>
{ } __attribute__((packed)); } // namespace SimpleComms

#endif
