/*!
 ***********************************************************
 * @file Message.def.h
 * @brief A file implementing general messages.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim
 * @date 03/09/2016
 * @remark
 * Added the cross-platform support
 *  -- 32-bit and big-endian
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * Message.def.h
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */

// TODO remove this once checksum is ignored on falcon boards
template<typename T>
bool messageComputeChecksum(T& msg,
                            uint32_t& newHeaderChecksum,
                            uint32_t& newDataChecksum,
                            const bool headerOnly = false)
{
    uint32_t dataLength = ntohl(msg.dataLength);

    if (dataLength <= MAX_FRAGMENT_PAYLOAD) {
        const uint32_t oldHeaderChecksum = msg.headerChecksum;
        const uint32_t oldDataChecksum = msg.dataChecksum;

        msg.headerChecksum = 0;

        if (!headerOnly) {
            // Adler is faster and good enough for our purposes with enough data
            // http://www.zlib.net/zlib_tech.html
            if (dataLength > 1024) {
                newDataChecksum = adler32(0, reinterpret_cast<const Bytef*>(msg.data), dataLength);
            } else {
                newDataChecksum = crc32(0, reinterpret_cast<const Bytef*>(msg.data), dataLength);
            }

            newDataChecksum = htonl(newDataChecksum);
            msg.dataChecksum = newDataChecksum;
        }

        newHeaderChecksum = crc32(0, reinterpret_cast<const Bytef*>(&msg), sizeof(MessageHeader));
        newHeaderChecksum = htonl(newHeaderChecksum);

        msg.headerChecksum = oldHeaderChecksum;
        msg.dataChecksum = oldDataChecksum;

        return true;
    }

    return false;
}

// TODO remove this once checksum is ignored on falcon boards
template<typename T>
bool messageSetChecksum(T& msg)
{
    uint32_t newHeaderChecksum = 0, newDataChecksum = 0;
    const bool status = messageComputeChecksum(msg, newHeaderChecksum, newDataChecksum);

    if (status) {
        msg.headerChecksum = newHeaderChecksum;
        msg.dataChecksum = newDataChecksum;
    }

    return status;
}

// Create a message with a payload big enough for the given type, and default-initialize it.
template<typename T>
MessagePtr messageCreate(const char* origin, uint32_t id = 0, const char* channel = SIMPLECOMMS_CONTROL_CHANNEL)
{
    MessagePtr msg(new (sizeof(T)) MessageStub());
    // Fill in the message header.
    strcpy(msg->origin, origin);
    strcpy(msg->channel, channel);
    msg->sendMsgTime.setNow();
    msg->sendMsgTime.tv_sec = htonl(msg->sendMsgTime.tv_sec);
    msg->sendMsgTime.tv_usec = htonl(msg->sendMsgTime.tv_usec);
    msg->rollingMessageID = htonl(id);
    msg->length = htonl(sizeof(T));
    msg->totalPages = htonl(1);
    msg->pageNumber = htonl(1);
    msg->dataLength = msg->length;
    // Default-initialize the message data.
    *reinterpret_cast<T*>(msg->data) = T();
    return msg;
}
