/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Helper class for wrapping the scs message log
 *        See https://confluence.ci.motional.com/confluence/display/AUTRUN/SQL+format+messaging+log
 *
 * Writes are batched (set the buffer size in the constructor). The db is opened in read/write mode
 * and both operations are supported (reads force a flush).
 *
 * Unflushed data is flushed on read or on in the destructor
 *
 */

#pragma once

#include <memory>
#include <optional>
#include <set>
#include <string>

namespace SimpleComms {

/// \brief A log file recording the order in which messages were published through SCS, and information about them.
class SQLMessageLog
{
public:
    /// \brief Constructor.
    explicit SQLMessageLog();
    /// \brief Destructor.
    ~SQLMessageLog();

    SQLMessageLog(const SQLMessageLog&) = delete;
    SQLMessageLog& operator=(const SQLMessageLog&) = delete;

    /// \brief Information stored for each row of the database.
    struct Row
    {
        int64_t nsec;                  ///< Message timestamp, in nanoseconds.
        std::string origin;            ///< Origin of the message (or, for executive triggers, the destination).
        std::string channel;           ///< Channel the message was sent over.
        std::set<std::string> dropped; ///< List of clients that dropped this message.
    };

    /// \brief Open a database file for reading or writing.
    /// \param databasePath: The path to the SQLite db file.
    /// \param write: If true, create a new file and write to it; otherwise, read from the given file.
    /// \param buffer: In write mode, the number of lines to buffer before writing a batch to the database.
    [[gnu::warn_unused_result]] bool open(const std::string& databasePath, bool write = false, size_t buffer = 500);

    /// \brief Write a line to the database.
    /// \param nsec: The message's timestamp, in nanoseconds (or 0, to record a subscription).
    /// \param origin: The message's origin.
    /// \param channel: The channel the message was sent over.
    /// \return An ID to be used to refer to this message if it is
    /// later dropped, or std::nullopt on failure.
    std::optional<int64_t> writeMessage(int64_t nsec, const std::string& origin, const std::string& channel);

    /// \brief Record the fact that a message was dropped.
    /// \param messageId: The ID that was returned when this message was first added to the database.
    /// \param client: The name of the subscriber that dropped this message.
    void writeDropped(int64_t messageId, const std::string& client);

    /// \brief Read a row from the database.
    /// \param[out] data: The Row object to store the information in.
    /// \param includeSubscriptions: Whether to include database lines marking a client subscribed to a channel.
    /// \return True if data was returned; false if all data has been read already.
    bool read(SQLMessageLog::Row& data, bool includeSubscriptions = false);

    /// \brief Flush unwritten writes to the database. Returns false on error.
    bool flush(bool wait = true);

    /// \brief Check whether the given channel is mentioned in the log.
    /// \param channel: A channel name.
    /// \return True if messages from the given channel exist in the log.
    bool hasChannel(const std::string& channel) const;

private:
    class Impl;
    std::unique_ptr<SQLMessageLog::Impl> impl_;
};

} // namespace SimpleComms
