/*!
 ***********************************************************
 * @file UnixClient.h
 * @brief A file defining the UnixClient class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 03/23/2018
 * @remark
 * removing boost dependency
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * UnixClient.h
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#ifndef _SIMPLECOMMS_UNIXCLIENT_H_
#define _SIMPLECOMMS_UNIXCLIENT_H_

#include "Client.h"
#include "ControlMessages.h"
#include "Message.h"
#include "StatsMessages.h"

#include <atomic>
#include <functional>
#include <future>
#include <list>
#include <map>
#include <memory>
#include <mutex>
#include <queue>
#include <string>
#include <vector>

namespace SimpleComms {

class MockUnixClient;

class UnixClient : public Client
{
public:
    UnixClient(const std::string& serverPath,
               const std::string& clientPath,
               Client::Callback callback = Client::Callback(),
               const double& staleBufferPruningThreshold_s = 1.0);

    ~UnixClient() override;

    static std::string makeUniqueClientName(const std::string& prefix = "");
    static std::string getDefaultScsPath();

    Status::States connect() override;
    Status::States disconnect() override;

    Status::States declare(const std::string& channel, ChannelType type, uint32_t receiveCapacity) override;
    Status::States undeclare(const std::string& channel, ChannelType type) override;

    Status::States sendMsgV(const std::string& channel,
                            const std::string& data,
                            std::vector<av::latency::Pred>&& predecessors) override;

    bool hasMsg() const override;
    Status::States receiveMsg(const std::string& channel, std::string& data, MessageInfo& msgInfo) override;

    /// \brief Tell SCS that this client responds to requests to pause its output. Must be called after connecting.
    void allowThrottling(std::function<void(bool)> callback) override;

    std::vector<std::string> getPublishListMessage() override;

private:
    const std::string serverPath_; ///< Path to local server
    const std::string clientPath_; ///< Path this client will register at
    std::recursive_mutex fdMutex_; ///< Mutex to protect operations to fd_/serverfd_/threadRunning_
    int fd_;                       ///< The socket this client is listening on, protected by fdMutex_
    int serverfd_;                 ///< The socket the server is listening on, protected by fdMutex_

    //! A promise to be set when a connection request is acknowledged.
    std::promise<void> connectionAcknowledged_;

    std::function<void(bool)> pauseCallback_; ///< Callback for pausing or resuming this task's output.

    std::atomic<uint32_t> rollingMessageID_; ///< A count that increments on every sendMsg

    //! Break circular reference between FragmentBuffer and leastRecentlyUpdatedFragmentBuffers_
    struct FragmentBuffer;
    typedef std::shared_ptr<FragmentBuffer> FragmentBufferPtr;

    //! MessageKey to FragmentBuffer map
    typedef std::map<MessageKey, FragmentBufferPtr> FragmentBufferMap;
    //! Least recently used FragmentBuffers list
    typedef std::list<FragmentBufferMap::iterator> FragmentBufferLRU;

    /**
     * @brief A workspace for reassembly of message fragments
     */
    struct FragmentBuffer
    {
        FragmentBuffer(const FragmentBufferLRU::iterator& i)
            : receivedCount(0)
            , receivedBytes(0)
            , lruEntry(i)
        {}

        TimeStamp lastMessageAdded;           ///< Time stamp of the last message added, for tracking stale buffers
        MessageHeader lastHeader;             ///< Header of the latest page
        std::string data;                     ///< Assembly area of received fragments
        std::vector<bool> received;           ///< Flag to indicate whether a fragment was received
        uint32_t receivedCount;               ///< Number of received fragments
        uint32_t receivedBytes;               ///< Number of received bytes
        FragmentBufferLRU::iterator lruEntry; //< Least recently used list entry
    };

    //! Fragment buffers associated to messages to provide for reassembly
    FragmentBufferMap fragmentBuffers_;
    //! Least recently updated fragment buffer list
    FragmentBufferLRU fragmentBuffersLRU_;
    //! Pruning threshold
    double staleBufferPruningThreshold_s_;
    void pruneStalePartialBuffers();

    struct ChannelData
    {
        ChannelData(const uint32_t inMaxReceiveDepth)
            : acknowledged(false)
            , maxReceiveDepth(inMaxReceiveDepth)
            , type(ChannelType::Subscriber)
        {}
        ChannelData(const uint32_t inMaxReceiveDepth, const ChannelType inType)
            : acknowledged(false)
            , maxReceiveDepth(inMaxReceiveDepth)
            , type(inType)
        {}

        bool acknowledged;
        uint32_t maxReceiveDepth;
        ChannelType type;
        std::queue<FragmentBufferPtr> completeBuffers;
    };

    typedef std::map<std::string, ChannelData> ChannelMap;

    //! A mapping of channels to lists of completed fragment buffers, protected by completeBuffersPerChannelMutex_
    ChannelMap channels_;
    //! Mutex to protect completeBuffersPerChannel_ between the receiver thread and the client
    std::mutex channelsMutex_;
    //! A count of how many messages are waiting to be received.
    std::atomic<int> messagesAvailable_{0};

    typedef std::pair<std::string, std::string> ChannelOriginKey;
    typedef std::map<ChannelOriginKey, ChannelOriginTotals> ChannelOriginStatsMap;
    ChannelOriginStatsMap channelOriginStats_;
    ChannelOriginStatsMap::iterator fetchTotals(const ChannelOriginKey& statsKey);

    Status::States sendMsgM(const Message& msg);
    Status::States sendMsgM(const MessageBase& msg);
    Status::States handleIncomingMsg();
    Status::States handleControlMsg(const Message& msg);
    Status::States closeConnection(const bool receiverThread);

    uint32_t threadRunning_;         ///< A counter tracking pthread_(create|join), protected by fdMutex_
    pthread_t receiverThreadHandle_; ///< Handle for the receiver thread
    static void* receiverThreadLauncher(void* arg);
    void receiverThread();

    TimeStamp lastReport_;
    uint32_t totalFragmentsReceived_, totalMessagesReceived_;

    friend class MockUnixClient;
};

} // namespace SimpleComms

#endif
