/*!
 ***********************************************************
 * @file MockUnixClient.h
 * @brief A file defining a testing Mock of the UnixClient class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 03/23/2018
 * @remark
 * removing boost dependency
 * @n
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Simple Comms
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * UnixClient.h
 * Tugrul Galatali
 * 02/10/2007
 *
 * Copyright (C) 2007
 * Carnegie Mellon University
 * All rights reserved
 */
#ifndef _SIMPLECOMMS_MOCKUNIXCLIENT_H_
#define _SIMPLECOMMS_MOCKUNIXCLIENT_H_

#include <list>
#include <map>
#include <queue>
#include <string>
#include <vector>

#include <memory>

#include <SimpleComms/Client.h>
#include <SimpleComms/ControlMessages.h>
#include <SimpleComms/Message.h>
#include <SimpleComms/StatsMessages.h>
#include <SimpleComms/UnixClient.h>
#include <gmock/gmock.h>

namespace SimpleComms {
class MockUnixClient : public UnixClient
{
public:
    using UnixClient::hasMsg;

    MockUnixClient(const std::string& serverPath = "testscs",
                   const std::string& clientPath = "testclient",
                   const double& staleBufferPruningThreshold_s = 1.0)
        : UnixClient(serverPath, clientPath, nullptr, staleBufferPruningThreshold_s)
    {}

    MOCK_METHOD0(connect, Status::States());
    MOCK_METHOD0(disconnect, Status::States());

    MOCK_METHOD3(declare, Status::States(const std::string&, ChannelType, uint32_t));
    MOCK_METHOD2(undeclare, Status::States(const std::string&, ChannelType));

    MOCK_METHOD3(sendMsgV, Status::States(const std::string&, const std::string&, std::vector<av::latency::Pred>&&));
    MOCK_METHOD0(hasMsg, bool());
    MOCK_METHOD3(receiveMsg, Status::States(const std::string&, std::string&, MessageInfo&));
    MOCK_METHOD1(allowThrottling, void(std::function<void(bool)> callback));

    Status::States UnixClientDeclare(const std::string& channel, const ChannelType type, const uint32_t maxReceiveDepth)
    {
        return UnixClient::declare(channel, type, maxReceiveDepth);
    }

    Status::States UnixClientUndeclare(const std::string& channel, const ChannelType type)
    {
        return UnixClient::undeclare(channel, type);
    }

    Status::States ConfirmChannelMap(const std::string& /* channel */,
                                     const std::string& data,
                                     const std::vector<av::latency::Pred>& /* preds */)
    {
        const DeclareChannelMessage* declareMessage = reinterpret_cast<const DeclareChannelMessage*>(data.data());
        const ChannelMap::iterator i(channels_.find(declareMessage->channel));

        if (i != channels_.end()) {
            if (declareMessage->type == ControlMessages::Declare) {
                if ((ntohl(declareMessage->channelType) & ChannelType::Subscriber) &&
                    !((*i).second.type & ChannelType::Subscriber)) {
                    return Status::States::Error;
                }
                if ((ntohl(declareMessage->channelType) & ChannelType::Publisher) &&
                    !((*i).second.type & ChannelType::Publisher)) {
                    return Status::States::Error;
                }
            }

            if (declareMessage->type == ControlMessages::Declare) {
                if ((ntohl(declareMessage->channelType) & ChannelType::Subscriber) &&
                    ((*i).second.type & ChannelType::Subscriber)) {
                    return Status::States::Error;
                }
                if ((ntohl(declareMessage->channelType) & ChannelType::Publisher) &&
                    ((*i).second.type & ChannelType::Publisher)) {
                    return Status::States::Error;
                }
            }
        }
        return Status::States::Ok;
    }

    Status::States checkChannelMapAndBlobEquivalence(const std::string& channel,
                                                     const std::string& data,
                                                     const std::vector<av::latency::Pred>& preds)
    {
        return static_cast<Status::States>(CheckBlobEquivalence(channel, data) |
                                           ConfirmChannelMap(channel, data, preds));
    }

    Status::States AcknowledgeChannel(const std::string& /* channel */,
                                      const std::string& data,
                                      const std::vector<av::latency::Pred>& /* preds */)
    {
        const DeclareChannelMessage* declareMessage = reinterpret_cast<const DeclareChannelMessage*>(data.data());
        const ChannelMap::iterator i(channels_.find(declareMessage->channel));

        if (i != channels_.end()) {
            (*i).second.acknowledged = true;
            return Status::States::Ok;
        }
        return Status::States::Error;
    }

    Status::States CheckBlobEquivalence(const std::string& /* channel */, const std::string& data)
    {
        const DeclareChannelMessage* dcmExisting = reinterpret_cast<const DeclareChannelMessage*>(blob.data());
        const DeclareChannelMessage* dcmData = reinterpret_cast<const DeclareChannelMessage*>(data.data());

        bool eq = (dcmExisting->type == dcmData->type) && (dcmExisting->channelType == dcmData->channelType) &&
                  (std::string(dcmExisting->channel) == std::string(dcmData->channel));

        return (eq) ? Status::States::Ok : Status::States::Error;
    }

    bool CheckChannelType(const std::string& channel, ChannelType type)
    {
        const ChannelMap::iterator i(channels_.find(channel));

        if (i != channels_.end()) {
            return (*i).second.type == type;
        }
        return false;
    }

    ChannelType getChannelType(const std::string& channel)
    {
        const ChannelMap::iterator i(channels_.find(channel));

        return (i == channels_.end()) ? ChannelType::Unlisted : (*i).second.type;
    }

    size_t getMaxReceiveDepth(const std::string& channel)
    {
        const ChannelMap::iterator i(channels_.find(channel));

        return (i == channels_.end()) ? 0 : (*i).second.maxReceiveDepth;
    }

    std::string blob;
};

} // namespace SimpleComms

#endif
