/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file BasePeerTests.cc
 * @brief This file implements tests for the BasePeer class.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#include <chrono>
#include <memory>
#include <string>
#include <thread>
#include <vector>

#include <gtest/gtest-param-test.h>
#include <gtest/gtest.h>

#include "TestingBasePeer.h"

using namespace ::testing;
using namespace SimpleComms;

class BPTests
    : public ::testing::TestWithParam<bool>
    , public TestingBasePeer
{
protected:
    std::unique_ptr<TestingBasePeer> testPeer_;

    TestingBasePeer::ChannelData no_sub_no_pub, sub_pub, sub_no_pub, no_sub_pub, sub_pub_2;
    void SetUp() override
    {
        testPeer_ = std::make_unique<TestingBasePeer>(GetParam());
        std::vector<TestingBasePeer::ChannelData> channels;

        no_sub_no_pub.name = "NoSubNoPub";
        channels.push_back(no_sub_no_pub);

        TestingBasePeer::ClientList::iterator pub1;
        EXPECT_TRUE(testPeer_->clientLookupPassthrough("Publisher1", pub1, true));
        no_sub_pub.publishers.push_back(*pub1);
        no_sub_pub.name = "NoSubPub";
        channels.push_back(no_sub_pub);

        sub_no_pub.name = "SubNoPub";

        TestingBasePeer::ClientList::iterator sub2;
        EXPECT_TRUE(testPeer_->clientLookupPassthrough("Subscriber1", sub2, true));

        sub_no_pub.subscribers.push_back(*sub2);
        channels.push_back(sub_no_pub);

        sub_pub.name = "SubPub";
        TestingBasePeer::ClientList::iterator sub1;
        TestingBasePeer::ClientList::iterator pub2;
        EXPECT_TRUE(testPeer_->clientLookupPassthrough("Subscriber2", sub1, true));
        sub_pub.subscribers.push_back(*sub1);
        EXPECT_TRUE(testPeer_->clientLookupPassthrough("Publisher2", pub2, true));
        sub_pub.publishers.push_back(*pub2);
        channels.push_back(sub_pub);

        sub_pub_2.name = "SubPub2";
        TestingBasePeer::ClientList::iterator subpub;
        EXPECT_TRUE(testPeer_->clientLookupPassthrough("SubPub", subpub, true));

        sub_pub_2.subscribers.push_back(*subpub);
        sub_pub_2.publishers.push_back(*subpub);
        channels.push_back(sub_pub_2);

        testPeer_->setChannels(channels);
    }

    void TearDown() override {}
};

TEST_P(BPTests, testHandleErrorsClient)
{
    TestingBasePeer::ClientList::iterator pub1;
    TestingBasePeer::ClientList::iterator pub2;

    ASSERT_TRUE(testPeer_->clientLookupPassthrough("Publisher1", pub1, true));
    ASSERT_TRUE(testPeer_->clientLookupPassthrough("Publisher2", pub2, true));

    Peer::BaseClientList clients;
    clients.push_back(*pub1);
    clients.push_back(*pub2);
    EXPECT_EQ(testPeer_->handleErrors(clients), Status::Ok);
}

TEST_P(BPTests, testIsSubscribedTo)
{
    EXPECT_TRUE(testPeer_->isSubscribedTo("SubPub"));
    EXPECT_TRUE(testPeer_->isSubscribedTo("SubNoPub"));
    EXPECT_TRUE(testPeer_->isSubscribedTo("SubPub2"));
    EXPECT_FALSE(testPeer_->isSubscribedTo("NoSubNoPub"));
    EXPECT_FALSE(testPeer_->isSubscribedTo("NoSubPub"));
    EXPECT_FALSE(testPeer_->isSubscribedTo("NoEntry"));

    EXPECT_EQ(testPeer_->numChannels(), 5);
}

TEST_P(BPTests, testGetSubscriptions)
{
    std::vector<std::string> correct_answer, answer;
    correct_answer = {"SubNoPub", "SubPub", "SubPub2"};

    testPeer_->getSubscriptions(answer);
    EXPECT_EQ(correct_answer, answer);

    EXPECT_EQ(testPeer_->numChannels(), 5);
}

TEST_P(BPTests, testGetPublishedChannels)
{
    std::vector<std::string> correct_answer, answer;
    correct_answer = {"NoSubPub", "SubPub", "SubPub2"};

    testPeer_->getPublishedChannels(answer);
    EXPECT_EQ(correct_answer, answer);

    EXPECT_EQ(testPeer_->numChannels(), 5);
}

TEST_P(BPTests, testResetClientPublisherSubscriberStays)
{

    std::vector<std::string> correct_pub, correct_sub, answer_pub, answer_sub;

    correct_pub = {"NoSubPub", "SubPub2"};

    correct_sub = {"SubNoPub", "SubPub", "SubPub2"};

    TestingBasePeer::ClientList::iterator i;
    EXPECT_TRUE(testPeer_->clientLookupPassthrough("Publisher2", i, false));

    testPeer_->resetPassthrough(*i, false);

    testPeer_->getPublishedChannels(answer_pub);
    testPeer_->getSubscriptions(answer_sub);

    EXPECT_EQ(answer_pub, correct_pub);
    EXPECT_EQ(answer_sub, correct_sub);

    EXPECT_EQ(testPeer_->numChannels(), 4);
}

TEST_P(BPTests, testResetClientRecentSubscriberPublisherStays)
{

    std::vector<std::string> correct_pub, correct_sub, answer_pub, answer_sub;

    correct_pub = {"NoSubPub", "SubPub", "SubPub2"};
    correct_sub = {"SubNoPub", "SubPub2"};

    TestingBasePeer::ClientList::iterator i;
    EXPECT_TRUE(testPeer_->clientLookupPassthrough("Subscriber2", i, false));

    testPeer_->resetPassthrough(*i, false);

    testPeer_->getPublishedChannels(answer_pub);
    testPeer_->getSubscriptions(answer_sub);

    EXPECT_EQ(answer_pub, correct_pub);
    EXPECT_EQ(answer_sub, correct_sub);

    EXPECT_EQ(testPeer_->numChannels(), 4);
}

TEST_P(BPTests, testResetClientRecentBothSeparately)
{

    std::vector<std::string> correct_pub, correct_sub, intermediate_sub, intermediate_pub, answer_pub, answer_sub;

    intermediate_pub = {"NoSubPub", "SubPub", "SubPub2"};
    intermediate_sub = {"SubNoPub", "SubPub2"};

    correct_pub = {"NoSubPub", "SubPub2"};
    correct_sub = {"SubNoPub", "SubPub2"};

    TestingBasePeer::ClientList::iterator i;
    EXPECT_TRUE(testPeer_->clientLookupPassthrough("Subscriber2", i, false));
    testPeer_->resetPassthrough(*i, false);

    testPeer_->getPublishedChannels(answer_pub);
    testPeer_->getSubscriptions(answer_sub);

    EXPECT_EQ(answer_pub, intermediate_pub);
    EXPECT_EQ(answer_sub, intermediate_sub);
    answer_pub.clear();
    answer_sub.clear();

    EXPECT_TRUE(testPeer_->clientLookupPassthrough("Publisher2", i, false));
    testPeer_->resetPassthrough(*i, false);

    testPeer_->getPublishedChannels(answer_pub);
    testPeer_->getSubscriptions(answer_sub);

    EXPECT_EQ(answer_pub, correct_pub);
    EXPECT_EQ(answer_sub, correct_sub);

    EXPECT_EQ(testPeer_->numChannels(), 3);
}

TEST_P(BPTests, testResetClientRecentBothSeparatelyInverse)
{

    std::vector<std::string> correct_pub, correct_sub, intermediate_sub, intermediate_pub, answer_pub, answer_sub;

    intermediate_pub = {"NoSubPub", "SubPub2"};
    intermediate_sub = {"SubNoPub", "SubPub", "SubPub2"};
    correct_pub = {"NoSubPub", "SubPub2"};
    correct_sub = {"SubNoPub", "SubPub2"};

    TestingBasePeer::ClientList::iterator i;
    EXPECT_TRUE(testPeer_->clientLookupPassthrough("Publisher2", i, false));
    testPeer_->resetPassthrough(*i, false);

    testPeer_->getPublishedChannels(answer_pub);
    testPeer_->getSubscriptions(answer_sub);
    EXPECT_EQ(answer_pub, intermediate_pub);
    EXPECT_EQ(answer_sub, intermediate_sub);
    answer_pub.clear();
    answer_sub.clear();

    EXPECT_TRUE(testPeer_->clientLookupPassthrough("Subscriber2", i, false));
    testPeer_->resetPassthrough(*i, false);

    testPeer_->getPublishedChannels(answer_pub);
    testPeer_->getSubscriptions(answer_sub);

    EXPECT_EQ(answer_pub, correct_pub);
    EXPECT_EQ(answer_sub, correct_sub);

    EXPECT_EQ(testPeer_->numChannels(), 3);
}

TEST_P(BPTests, testResetClientRecentBothTogether)
{

    std::vector<std::string> correct_pub, correct_sub, answer_pub, answer_sub;

    correct_pub = {"NoSubPub", "SubPub"};
    correct_sub = {"SubNoPub", "SubPub"};

    TestingBasePeer::ClientList::iterator i;
    EXPECT_TRUE(testPeer_->clientLookupPassthrough("SubPub", i, false));

    testPeer_->resetPassthrough(*i, false);

    testPeer_->getPublishedChannels(answer_pub);
    testPeer_->getSubscriptions(answer_sub);

    EXPECT_EQ(answer_pub, correct_pub);
    EXPECT_EQ(answer_sub, correct_sub);

    EXPECT_EQ(testPeer_->numChannels(), 3);
}

INSTANTIATE_TEST_SUITE_P(BPTestsPNP, BPTests, ::testing::Bool());

class BPTests5Statuses
    : public ::testing::TestWithParam<int>
    , public TestingBasePeer
{
protected:
    SimpleComms::ChannelType clientAChannelA_;
    SimpleComms::ChannelType clientAChannelB_;
    SimpleComms::ChannelType clientBChannelA_;
    SimpleComms::ChannelType clientBChannelB_;
    SimpleComms::ChannelType clientADeclaration_;
    std::unique_ptr<TestingBasePeer> testPeer_;

    TestingBasePeer::ChannelData channelA_;
    TestingBasePeer::ChannelData channelB_;

    std::vector<TestingBasePeer::ChannelData> startingChannels_;

    std::vector<SimpleComms::ChannelType> ctypes_ = {SimpleComms::ChannelType::Unlisted,
                                                     SimpleComms::ChannelType::Subscriber,
                                                     SimpleComms::ChannelType::Publisher,
                                                     SimpleComms::ChannelType::SubPub};
    void SetUp() override
    {
        const auto param = static_cast<size_t>(GetParam());
        EXPECT_TRUE(param < 1024);
        clientAChannelA_ = ctypes_[param % 4];
        clientAChannelB_ = ctypes_[(param >> 2) % 4];
        clientBChannelA_ = ctypes_[(param >> 4) % 4];
        clientBChannelB_ = ctypes_[(param >> 6) % 4];
        clientADeclaration_ = ctypes_[(param >> 8) % 4];

        testPeer_ = std::make_unique<TestingBasePeer>();
        channelA_.name = "ChannelA";
        channelB_.name = "ChannelB";

        TestingBasePeer::ClientList::iterator clientIter;
        EXPECT_TRUE(testPeer_->clientLookupPassthrough("ClientA", clientIter, true));
        TestingBasePeer::ClientDataPtr clientA = *clientIter;

        EXPECT_TRUE(testPeer_->clientLookupPassthrough("ClientB", clientIter, true));

        TestingBasePeer::ClientDataPtr clientB = *clientIter;

        if (clientAChannelA_ == SubPub || clientAChannelA_ == Subscriber) {
            channelA_.subscribers.push_back(clientA);
        }
        if (clientAChannelA_ == SubPub || clientAChannelA_ == Publisher) {
            channelA_.publishers.push_back(clientA);
        }
        if (clientAChannelB_ == SubPub || clientAChannelB_ == Subscriber) {
            channelB_.subscribers.push_back(clientA);
        }
        if (clientAChannelB_ == SubPub || clientAChannelB_ == Publisher) {
            channelB_.publishers.push_back(clientA);
        }

        if (clientBChannelA_ == SubPub || clientBChannelA_ == Subscriber) {
            channelA_.subscribers.push_back(clientB);
        }
        if (clientBChannelA_ == SubPub || clientBChannelA_ == Publisher) {
            channelA_.publishers.push_back(clientB);
        }
        if (clientBChannelB_ == SubPub || clientBChannelB_ == Subscriber) {
            channelB_.subscribers.push_back(clientB);
        }
        if (clientBChannelB_ == SubPub || clientBChannelB_ == Publisher) {
            channelB_.publishers.push_back(clientB);
        }

        std::sort(channelA_.publishers.begin(), channelA_.publishers.end());
        std::sort(channelA_.subscribers.begin(), channelA_.subscribers.end());
        std::sort(channelB_.publishers.begin(), channelB_.publishers.end());
        std::sort(channelB_.subscribers.begin(), channelB_.subscribers.end());

        if (!(channelA_.publishers.empty() && channelA_.subscribers.empty())) {
            startingChannels_.push_back(channelA_);
        }

        if (!(channelB_.publishers.empty() && channelB_.subscribers.empty())) {
            startingChannels_.push_back(channelB_);
        }

        testPeer_->setChannels(startingChannels_);
    }
};

TEST_P(BPTests5Statuses, testDeclareAllScenarios)
{
    ChannelData correctAChannelData = channelA_;
    ChannelData correctBChannelData = channelB_;
    std::vector<ChannelData> correctChannels;

    if (!(clientAChannelA_ == SubPub || clientAChannelA_ == Subscriber) &&
        (clientADeclaration_ == SubPub || clientADeclaration_ == Subscriber)) {
        // Need to add Subscriber permission to Client A Channel A

        TestingBasePeer::ClientList::iterator client;
        EXPECT_TRUE(testPeer_->clientLookupPassthrough("ClientA", client, false));

        correctAChannelData.subscribers.push_back(*client);
    }

    if (!(clientAChannelA_ == SubPub || clientAChannelA_ == Publisher) &&
        (clientADeclaration_ == SubPub || clientADeclaration_ == Publisher)) {
        // Need to add Publisher permission to Client A Channel A

        TestingBasePeer::ClientList::iterator client;
        EXPECT_TRUE(testPeer_->clientLookupPassthrough("ClientA", client, false));

        correctAChannelData.publishers.push_back(*client);
    }

    std::sort(correctAChannelData.publishers.begin(), correctAChannelData.publishers.end());
    std::sort(correctAChannelData.subscribers.begin(), correctAChannelData.subscribers.end());
    std::sort(correctBChannelData.publishers.begin(), correctBChannelData.publishers.end());
    std::sort(correctBChannelData.subscribers.begin(), correctBChannelData.subscribers.end());

    if (!(correctAChannelData.publishers.empty() && correctAChannelData.subscribers.empty())) {
        correctChannels.push_back(correctAChannelData);
    }

    if (!(correctBChannelData.publishers.empty() && correctBChannelData.subscribers.empty())) {
        correctChannels.push_back(correctBChannelData);
    }

    SimpleComms::MessagePtr msg(new (sizeof(SimpleComms::DeclareChannelMessage)) SimpleComms::MessageStub());
    std::string origin = "ClientA";
    memcpy(msg->origin, origin.c_str(), origin.length() + 1);

    SimpleComms::DeclareChannelMessage* dcm = reinterpret_cast<SimpleComms::DeclareChannelMessage*>(msg->data);
    dcm->type = static_cast<SimpleComms::ControlMessages::ControlMessageTypes>(
        htonl(SimpleComms::ControlMessages::ControlMessageTypes::Declare));
    std::string channel = "ChannelA";
    memcpy(dcm->channel, channel.c_str(), channel.length() + 1);
    dcm->channelType = static_cast<SimpleComms::ChannelType>(htonl(clientADeclaration_));

    EXPECT_EQ(testPeer_->handleControlMsg(msg), SimpleComms::Status::NewData);

    std::vector<ChannelData> answerChannels = testPeer_->getChannels();

    EXPECT_EQ(correctChannels.size(), answerChannels.size());
    for (size_t i = 0; i < std::min(correctChannels.size(), answerChannels.size()); ++i) {
        EXPECT_EQ(correctChannels[i].name, answerChannels[i].name);
        EXPECT_EQ(correctChannels[i].publishers.size(), answerChannels[i].publishers.size());
        EXPECT_EQ(correctChannels[i].subscribers.size(), answerChannels[i].subscribers.size());

        for (size_t j = 0; j < std::min(correctChannels[i].publishers.size(), answerChannels[i].publishers.size());
             ++j) {
            EXPECT_EQ(correctChannels[i].publishers[j], answerChannels[i].publishers[j]);
        }
        for (size_t j = 0; j < std::min(correctChannels[i].subscribers.size(), answerChannels[i].subscribers.size());
             ++j) {
            EXPECT_EQ(correctChannels[i].subscribers[j], answerChannels[i].subscribers[j]);
        }
    }
}

TEST_P(BPTests5Statuses, testUndeclareAllScenarios)
{
    ChannelData correctAChannelData = channelA_;
    ChannelData correctBChannelData = channelB_;
    std::vector<ChannelData> correctChannels;

    TestingBasePeer::ClientList::iterator client;
    EXPECT_TRUE(testPeer_->clientLookupPassthrough("ClientA", client, false));
    TestingBasePeer::ClientDataPtr clientPtr = *client;

    if ((clientAChannelA_ == SubPub || clientAChannelA_ == Subscriber) &&
        (clientADeclaration_ == SubPub || clientADeclaration_ == Subscriber)) {
        // Need to add Subscriber permission to Client A Channel A
        TestingBasePeer::ClientList::iterator i =
            std::find(correctAChannelData.subscribers.begin(), correctAChannelData.subscribers.end(), clientPtr);
        if (i != correctAChannelData.subscribers.end()) {
            correctAChannelData.subscribers.erase(i);
        }
    }

    if ((clientAChannelA_ == SubPub || clientAChannelA_ == Publisher) &&
        (clientADeclaration_ == SubPub || clientADeclaration_ == Publisher)) {
        // Need to add Publisher permission to Client A Channel A
        TestingBasePeer::ClientList::iterator i =
            std::find(correctAChannelData.publishers.begin(), correctAChannelData.publishers.end(), clientPtr);
        if (i != correctAChannelData.publishers.end()) {
            correctAChannelData.publishers.erase(i);
        }
    }

    if (!(correctAChannelData.publishers.empty() && correctAChannelData.subscribers.empty())) {
        correctChannels.push_back(correctAChannelData);
    }

    if (!(correctBChannelData.publishers.empty() && correctBChannelData.subscribers.empty())) {
        correctChannels.push_back(correctBChannelData);
    }

    SimpleComms::MessagePtr msg(new (sizeof(SimpleComms::DeclareChannelMessage)) SimpleComms::MessageStub());
    std::string origin = "ClientA";
    memcpy(msg->origin, origin.c_str(), origin.length() + 1);

    SimpleComms::DeclareChannelMessage* dcm = reinterpret_cast<SimpleComms::DeclareChannelMessage*>(msg->data);
    dcm->type = static_cast<SimpleComms::ControlMessages::ControlMessageTypes>(
        htonl(SimpleComms::ControlMessages::ControlMessageTypes::Undeclare));
    std::string channel = "ChannelA";
    memcpy(dcm->channel, channel.c_str(), channel.length() + 1);
    dcm->channelType = static_cast<SimpleComms::ChannelType>(htonl(clientADeclaration_));

    EXPECT_EQ(testPeer_->handleControlMsg(msg), SimpleComms::Status::NewData);

    std::vector<ChannelData> answerChannels = testPeer_->getChannels();
    EXPECT_EQ(correctChannels.size(), answerChannels.size());
    for (size_t i = 0; i < std::min(correctChannels.size(), answerChannels.size()); ++i) {
        EXPECT_EQ(correctChannels[i].name, answerChannels[i].name);
        EXPECT_EQ(correctChannels[i].publishers.size(), answerChannels[i].publishers.size());
        EXPECT_EQ(correctChannels[i].subscribers.size(), answerChannels[i].subscribers.size());

        for (size_t j = 0; j < std::min(correctChannels[i].publishers.size(), answerChannels[i].publishers.size());
             ++j) {
            EXPECT_EQ(correctChannels[i].publishers[j], answerChannels[i].publishers[j]);
        }
        for (size_t j = 0; j < std::min(correctChannels[i].subscribers.size(), answerChannels[i].subscribers.size());
             ++j) {
            EXPECT_EQ(correctChannels[i].subscribers[j], answerChannels[i].subscribers[j]);
        }
    }
}

INSTANTIATE_TEST_SUITE_P(HCMTests, BPTests5Statuses, ::testing::Range(0, 1024));

GTEST_API_ int main(int argc, char** argv)
{
    InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
