#include <gtest/gtest-param-test.h>
#include <gtest/gtest.h>

#include "CentralScheduler.h"

#include <chrono>
#include <functional>
#include <memory>
#include <string>

namespace {
bool kMessageSent{true};
bool SendFunction(const SimpleComms::MessagePtr&)
{
    kMessageSent = true;
    return true;
}
bool NoSendFunction(const SimpleComms::MessagePtr&)
{
    return false;
}
} // unnamed namespace

TEST(CentralSchedulerTest, testConstructorDestructorNoThrow)
{

    std::shared_ptr<SimpleComms::CentralScheduler> cs;
    EXPECT_NO_THROW(cs = std::make_shared<SimpleComms::CentralScheduler>(SendFunction));
    EXPECT_NO_THROW(cs.reset());
}

TEST(CentralSchedulerTest, testFreezeNoThrow)
{

    SimpleComms::CentralScheduler cs(SendFunction);
    ASSERT_NO_THROW(cs.freeze());

    // Call freeze again on the same central scheduler for test coverage.
    // The scheduler is already frozen and freeze should return as no-op.
    // Further, registering a task after the same scheduler is frozen should
    // emplace that task into the queue without throwing (coverage)
    std::string task{"task"};
    std::chrono::nanoseconds ns{1234};
    EXPECT_NO_THROW(cs.add(task, ns));
    EXPECT_NO_THROW(cs.freeze());

    // Create a new scheduler, add a task and then call freeze to reset the queue
    // to the next time
    SimpleComms::CentralScheduler csAux(SendFunction);
    EXPECT_NO_THROW(csAux.add(task, ns));
    EXPECT_NO_THROW(csAux.freeze());
}

TEST(CentralSchedulerTest, testAddNoThrow)
{

    SimpleComms::CentralScheduler cs(SendFunction);

    std::string task{"task"};
    std::chrono::nanoseconds ns{};
    EXPECT_NO_THROW(cs.add(task, ns));

    std::chrono::nanoseconds ns10{10};
    EXPECT_NO_THROW(cs.add(task, ns10));

    ASSERT_NO_THROW(cs.freeze());
    EXPECT_NO_THROW(cs.add(task, ns10));
}

TEST(CentralSchedulerTest, testAdvanceNoThrow)
{

    SimpleComms::CentralScheduler cs(SendFunction);
    SimpleComms::CentralScheduler csNeg(NoSendFunction);

    // Test advance with empty queue
    std::chrono::nanoseconds ns{};
    ASSERT_NO_THROW(cs.advance(ns));

    // Add entry into the queue and call freeze agains
    std::string task{"task"};
    std::chrono::nanoseconds ns10{10};
    EXPECT_NO_THROW(cs.add(task, ns));
    EXPECT_NO_THROW(csNeg.add(task, ns));
    ASSERT_NO_THROW(cs.freeze());
    ASSERT_NO_THROW(csNeg.freeze());

    // Test advance with non empty queue, timestamp lower than queue entry
    ASSERT_NO_THROW(cs.advance(ns));

    // Test advance with non empty queue, timestamp higher than queue entry
    std::chrono::nanoseconds ns100{100};
    ASSERT_NO_THROW(cs.advance(ns100));
    ASSERT_NO_THROW(csNeg.advance(ns100));
}

TEST(CentralSchedulerTest, testAdvanceOnFreezedScheduler)
{
    SimpleComms::CentralScheduler cs(SendFunction);

    // Reset sent message flag
    kMessageSent = false;

    // Freeze the scheduler to avoid asynch advancing the queue
    ASSERT_NO_THROW(cs.freeze());

    // Add task into the queue
    std::string task{"task"};
    std::chrono::nanoseconds ns{1};
    EXPECT_NO_THROW(cs.add(task, ns));

    // Test advance with non-empty queue
    ASSERT_NO_THROW(cs.advance(ns));
    ASSERT_TRUE(kMessageSent) << "The message should have been sent";
}

TEST(CentralSchedulerTest, testTimestamp)
{

    SimpleComms::CentralScheduler cs(SendFunction);
    SimpleComms::CentralScheduler csAux(SendFunction);
    std::chrono::nanoseconds ns10{10};
    ASSERT_NO_THROW(cs.advance(ns10));
    ASSERT_NO_THROW(cs.freeze());

    std::chrono::nanoseconds ns{0};
    EXPECT_EQ(cs.timestamp(), ns);
    EXPECT_GT(csAux.timestamp(), ns);
}
