#include <gtest/gtest-param-test.h>
#include <gtest/gtest.h>

#include "CentralScheduler.h"
#include "MessageControl.h"
#include "platform/middleware/libs/TestUtils/include/TestUtils/TmpDir.hpp"
#include <ODatum/ODatum.h>
#include <SimpleComms/Constants.h>
#include <SimpleComms/ControlMessages.h>
#include <SimpleComms/Message.h>
#include <config/ConfigParser.h>
#include <interfaces/ScsRecordingCommand/ScsRecordingCommand.h>
#include <interfaces/baseTypes/Header.h>

#include <algorithm> // for std::min
#include <exception> // for std::runtime_error
#include <experimental/filesystem>
#include <functional>
#include <set>
#include <stdexcept>
#include <string>
#include <system_error> // for std::error_code
#include <vector>

using namespace SimpleComms;

namespace {

bool kGlobalFlag{false};
ConfigSection params;

bool SendFunction(const SimpleComms::MessagePtr&)
{
    kGlobalFlag = true;
    return true;
}

struct ConnectMessage : public SimpleComms::TypedMessage<SimpleComms::ControlMessages::Connect>
{
} __attribute__((packed));

// Use the steady clock, but with Unix epoch timestamps.
// The system_clock epoch on Linux is the Unix epoch. (This will become
// part of the standard in C++2a, but is an "unofficial" standard already.)
static const std::chrono::nanoseconds EPOCH_OFFSET =
    std::chrono::system_clock::now().time_since_epoch() - std::chrono::steady_clock::now().time_since_epoch();
// This is the Unix epoch as a steady_clock time point.
static const std::chrono::steady_clock::time_point EPOCH = std::chrono::steady_clock::time_point() - EPOCH_OFFSET;

// Get a monotonic timestamp, in nanoseconds since the Unix epoch.
std::chrono::nanoseconds now()
{
    return EPOCH_OFFSET + std::chrono::steady_clock::now().time_since_epoch();
}

} // unnamed namespace

MessagePtr makeDeclare(const std::string& origin, const std::string& channel, SimpleComms::ChannelType type)
{
    MessagePtr msg = SimpleComms::messageCreate<SimpleComms::DeclareChannelMessage>(origin.c_str());

    DeclareChannelMessage& message = *reinterpret_cast<DeclareChannelMessage*>(msg->data);
    message.type = static_cast<ControlMessages::ControlMessageTypes>(htonl(SimpleComms::ControlMessages::Declare));
    strcpy(message.channel, channel.c_str());
    message.channelType = static_cast<SimpleComms::ChannelType>(htonl(type));

    return msg;
}

MessagePtr makeData(const std::string& origin, const std::string& channel, uint32_t id)
{
    typedef interfaces::baseTypes::InterfaceHeader Header;
    return SimpleComms::messageCreate<Header>(origin.c_str(), id, channel.c_str());
}

// Make a message reporting that "client" dropped message "id" from "origin".
MessagePtr makeDrop(const std::string& client, const std::string& origin, uint32_t id)
{
    MessagePtr msg = SimpleComms::messageCreate<SimpleComms::DropMessage>(client.c_str());

    DropMessage& message = *reinterpret_cast<DropMessage*>(msg->data);
    strcpy(message.origin, origin.c_str());
    message.id = htonl(id);

    return msg;
}

int64_t readTimestamp(const MessagePtr& msg)
{
    typedef interfaces::baseTypes::InterfaceHeader Header;
    const Header& header = *reinterpret_cast<const Header*>(msg->data);
    return ntohlls(header.sendTimeStamp.tv_sec) * NSEC_PER_SEC + ntohlls(header.sendTimeStamp.tv_nsec);
}

namespace fs = std::experimental::filesystem;

class MessageControlWithDBTests : public ::testing::Test
{
protected:
    void SetUp() override
    {
        TestUtils::TestTmpDirectory tmpDir{};
        fs::path dbPath = tmpDir.getPath() / "SQLMessageLogTest.db";
        std::error_code e;
        fs::remove(dbPath, e);

        m_sqlMsgLog.reset(new SimpleComms::SQLMessageLog());
        ASSERT_TRUE(m_sqlMsgLog->open(dbPath, true));

        // Setup global flag
        kGlobalFlag = false;
    }

    fs::path getDBPath() const noexcept
    {
        TestUtils::TestTmpDirectory tmpDir{};
        fs::path dbPath = tmpDir.getPath() / "SQLMessageLogTest.db";
        return dbPath;
    }

private:
    std::unique_ptr<SimpleComms::SQLMessageLog> m_sqlMsgLog;
};

TEST(MessageControlTest, testConstructorNoThrow)
{
    TestUtils::TestTmpDirectory tmpDir{};
    fs::path dbPath = tmpDir.getPath() / "SQLMessageLogTest.db";
    std::error_code e;
    fs::remove(dbPath, e);
    EXPECT_NO_THROW(SimpleComms::MessageControl("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params));
    EXPECT_NO_THROW(
        SimpleComms::MessageControl(dbPath, SimpleComms::MessageControl::Mode::LOG_LIVE, SendFunction, params));
    EXPECT_NO_THROW(
        SimpleComms::MessageControl(dbPath, SimpleComms::MessageControl::Mode::LOG_FROZEN, SendFunction, params));
    EXPECT_NO_THROW(
        SimpleComms::MessageControl(dbPath, SimpleComms::MessageControl::Mode::REPLAY, SendFunction, params));
}

TEST(MessageControlTest, testEmptyMsgPtrThrows)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);
    EXPECT_THROW(mc.preprocess(nullptr), std::invalid_argument);
}

TEST(MessageControlTest, testPreprocessRandomMsgPtrNoThrow)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);

    // This message is just a non-empty pointer, the idea is that
    // the "preprocess" method should not throw
    auto msg = std::make_shared<SimpleComms::MessageStub>();

    bool res{false};
    ASSERT_NO_THROW(res = mc.preprocess(msg));
    EXPECT_TRUE(res) << "the message should be queued for sending";
}

TEST(MessageControlTest, testPreprocessControlMsgNoReplayNoThrow)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);

    // Create new messages.
    // Note: by default these are on CONTROL channel
    const std::string origin{"origin"};
    auto scheduleMsg = SimpleComms::messageCreate<SimpleComms::ScheduleMessage>(origin.c_str());
    auto dropMsg = SimpleComms::messageCreate<SimpleComms::DropMessage>(origin.c_str());

    bool res{false};
    ASSERT_NO_THROW(res = mc.preprocess(scheduleMsg));
    EXPECT_FALSE(res) << "the message should NOT be queued for sending";

    ASSERT_NO_THROW(res = mc.preprocess(dropMsg));
    EXPECT_FALSE(res) << "the message should NOT be queued for sending";
}

TEST_F(MessageControlWithDBTests, testPreprocessReplayNoThrow)
{
    SimpleComms::MessageControl mc(getDBPath(), SimpleComms::MessageControl::Mode::REPLAY, SendFunction, params);

    // Test replaying no control channel
    const std::string origin{"origin"};
    auto noControlMsg = SimpleComms::messageCreate<SimpleComms::ScheduleMessage>(origin.c_str(), 0, origin.c_str());
    ASSERT_NO_THROW(mc.preprocess(noControlMsg));
    EXPECT_FALSE(kGlobalFlag);

    // Test replaying control no connect
    auto controlNoConnectMsg = SimpleComms::messageCreate<SimpleComms::ScheduleMessage>(origin.c_str());
    ASSERT_NO_THROW(mc.preprocess(controlNoConnectMsg));
    EXPECT_FALSE(kGlobalFlag);

    // Test replaying control and connect
    auto controlAndConnectMsg = SimpleComms::messageCreate<ConnectMessage>(origin.c_str());
    ASSERT_NO_THROW(mc.preprocess(controlAndConnectMsg));
    EXPECT_TRUE(kGlobalFlag);
}

TEST_F(MessageControlWithDBTests, testPreprocessExternalTimestampsSet)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);

    MessagePtr msg = makeData("origin", "ExternalChannel", 1);
    memset(msg->data, 0, sizeof(interfaces::baseTypes::InterfaceHeader));

    ASSERT_NO_THROW(mc.preprocess(msg));

    // We expect this to be within 10 seconds of real time to not be overdemanding of Jenkins
    EXPECT_LT(std::abs(now().count() - readTimestamp(msg)), 10000000000);
}

TEST_F(MessageControlWithDBTests, testPreprocessInternalTimestampsNotSet)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);

    MessagePtr msg = makeData("origin", SIMPLECOMMS_INTERNAL_TAG "InternalChannel", 1);
    memset(msg->data, 0, sizeof(interfaces::baseTypes::InterfaceHeader));

    ASSERT_NO_THROW(mc.preprocess(msg));

    EXPECT_EQ(readTimestamp(msg), 0);
}

TEST(MessageControlTest, testAddEmptyMsgPtrNoReplayNoThrow)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);

    bool res{false};
    ASSERT_NO_THROW(res = mc.add(nullptr));
    EXPECT_FALSE(res);
}

TEST_F(MessageControlWithDBTests, testAddEmptyMsgPtrReplayNoThrow)
{
    SimpleComms::MessageControl mc(getDBPath(), SimpleComms::MessageControl::Mode::REPLAY, SendFunction, params);

    bool res{false};
    ASSERT_NO_THROW(res = mc.add(nullptr));
    EXPECT_FALSE(res);
}

TEST_F(MessageControlWithDBTests, testAddNonEmptyMsgPtrReplayNoThrow)
{
    SimpleComms::MessageControl mc(getDBPath(), SimpleComms::MessageControl::Mode::REPLAY, SendFunction, params);

    bool res{false};
    const std::string origin{"origin"};
    auto connectMsg = SimpleComms::messageCreate<SimpleComms::ScheduleMessage>(origin.c_str());
    ASSERT_NO_THROW(res = mc.add(connectMsg));
    EXPECT_FALSE(res) << "Non executive messages should be ignored";

    auto exeMsg = SimpleComms::messageCreate<SimpleComms::ExecutiveMessage>(origin.c_str());
    ASSERT_NO_THROW(res = mc.add(exeMsg));
    EXPECT_TRUE(res) << "Executive messages should NOT be ignored";
}

TEST(MessageControlTest, testPopNoThrow)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);
    SimpleComms::MessagePtr ptr;
    ASSERT_NO_THROW(ptr = mc.pop());
    EXPECT_EQ(ptr, nullptr);
}

TEST(MessageControlTest, testLogNoLoggingNoThrow)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);
    SimpleComms::MessagePtr ptr;
    ASSERT_NO_THROW(mc.log(ptr));
}

TEST_F(MessageControlWithDBTests, testLogEmptyMsgNoThrow)
{
    SimpleComms::MessageControl mc(getDBPath(), SimpleComms::MessageControl::Mode::LOG_LIVE, SendFunction, params);
    SimpleComms::MessagePtr ptr;
    ASSERT_NO_THROW(mc.log(ptr));
}

TEST_F(MessageControlWithDBTests, testLogExecutiveMsgNoThrow)
{
    SimpleComms::MessageControl mc(getDBPath(), SimpleComms::MessageControl::Mode::LOG_LIVE, SendFunction, params);

    const std::string origin{"origin"};
    auto exeMsg = SimpleComms::messageCreate<SimpleComms::ExecutiveMessage>(origin.c_str());
    ASSERT_NO_THROW(mc.log(exeMsg));
}

TEST(MessageControlTest, testDroppedEmptyMsgThrows)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);
    SimpleComms::MessagePtr ptr;
    ASSERT_THROW(mc.dropped(ptr), std::invalid_argument);
}

TEST(MessageControlTest, testDroppedNonEmptyMsgNoThrow)
{
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, params);
    const std::string origin{"origin"};
    auto exeMsg = SimpleComms::messageCreate<SimpleComms::ExecutiveMessage>(origin.c_str());

    std::set<std::string> setDropped;
    ASSERT_NO_THROW(setDropped = mc.dropped(exeMsg));
    EXPECT_TRUE(setDropped.empty());
}

TEST(MessageControlTest, testRecordingCommands)
{
    std::string tempDir = getenv("TEST_TMPDIR");

    ConfigParser configParser;
    std::string config = "Parameters = {\nlogDataDir = \"" + tempDir + "\";\n}\n";
    configParser.parseString(config);
    ConfigSection section;
    configParser.getSection("Parameters", section);
    SimpleComms::MessageControl mc("", SimpleComms::MessageControl::Mode::LIVE, SendFunction, section);

    av::msg::ScsRecordingCommand scsRecordCmd;
    scsRecordCmd.command = av::msg::ScsRecordingCommandType::Start;
    scsRecordCmd.logName = "test_logname";

    interfaces::baseTypes::InterfaceHeader header;
    interfaces::baseTypes::setHeaderVersion(header, OTTO_INTERFACE_HEADER_VERSION);

    std::string blob;
    aptv::serializeODatumMessage(std::string(reinterpret_cast<char*>(&header), sizeof(header)), blob, scsRecordCmd);

    MessagePtr msg(new (static_cast<unsigned int>(blob.size())) MessageStub());
    strcpy(msg->origin, "origin");
    strcpy(msg->channel, "ScsRecordingCommand");
    msg->length = htonl(static_cast<uint32_t>(blob.size()));
    msg->totalPages = htonl(1);
    msg->pageNumber = htonl(1);
    msg->dataLength = msg->length;
    memcpy(msg->data, blob.data(), blob.size());

    char hostname[64];
    gethostname(hostname, 64);
    std::string expectedFile = tempDir + "/test_logname/messages@" + hostname + ".db";
    ASSERT_FALSE(fs::exists(expectedFile));

    bool preprocessResult{true};
    ASSERT_NO_THROW(preprocessResult = mc.preprocess(msg));
    EXPECT_FALSE(preprocessResult);

    scsRecordCmd.command = av::msg::ScsRecordingCommandType::Stop;
    aptv::serializeODatumMessage(std::string(reinterpret_cast<char*>(&header), sizeof(header)), blob, scsRecordCmd);
    memcpy(msg->data, blob.data(), blob.size());

    ASSERT_NO_THROW(preprocessResult = mc.preprocess(msg));
    EXPECT_FALSE(preprocessResult);

    ASSERT_TRUE(fs::exists(expectedFile));
    fs::remove(expectedFile);
}

TEST(MessageControlTest, LogAndReplay)
{
    // Scenario: a pipeline of three tasks: TaskA --Channel1--> TaskB --Channel2--> TaskC.
    // TaskA starts publishing before TaskB is ready to receive.
    // TaskC starts up before any of them but receives no data until TaskB is running.
    // TaskC drops every fourth message (e.g. it's taking too long to process data).

    // Generate a set of messages to be sent.
    std::vector<MessagePtr> messages;
    // Simulate a "rolling message ID" for each task.
    uint32_t idA = 0, idB = 0;
    // TaskC is the first to start up.
    messages.push_back(SimpleComms::messageCreate<SimpleComms::ReproducibleMessage>("TaskC"));
    messages.push_back(makeDeclare("TaskC", "Channel2", SimpleComms::ChannelType::Subscriber));
    // TaskA publishes some number of messages before anyone is ready for them.
    messages.push_back(SimpleComms::messageCreate<SimpleComms::ReproducibleMessage>("TaskA"));
    messages.push_back(makeDeclare("TaskA", "Channel1", SimpleComms::ChannelType::Publisher));
    const int PUBLISHED_BEFORE_B = 10;
    for (int i = 0; i < PUBLISHED_BEFORE_B; ++i) {
        messages.push_back(makeData("TaskA", "Channel1", idA++));
    }
    // Now, TaskB connects.
    messages.push_back(SimpleComms::messageCreate<SimpleComms::ReproducibleMessage>("TaskB"));
    messages.push_back(makeDeclare("TaskB", "Channel1", SimpleComms::ChannelType::Subscriber));
    messages.push_back(makeDeclare("TaskB", "Channel2", SimpleComms::ChannelType::Publisher));
    // The whole pipeline starts processing.
    const int PUBLISHED = 50;
    for (int i = PUBLISHED_BEFORE_B; i < PUBLISHED; ++i) {
        messages.push_back(makeData("TaskA", "Channel1", idA++));
        messages.push_back(makeData("TaskB", "Channel2", idB));
        if (!(i % 4)) {
            messages.push_back(makeDrop("TaskC", "TaskB", static_cast<uint32_t>(idB)));
        }
        ++idB;
    }
    // Log everything to this file.
    TestUtils::TestTmpDirectory tmpDir{};
    fs::path dbPath = tmpDir.getPath() / "MessageControlTest.db";

    // First, play and log this sequence of events.
    {
        SimpleComms::MessageControl mc(dbPath, SimpleComms::MessageControl::Mode::LOG_LIVE, SendFunction, params);
        for (const MessagePtr& msg : messages) {
            if (mc.preprocess(msg) && !mc.add(msg)) {
                mc.log(msg);
            }
        }
    }

    // Now, read the SQL file and make sure it contains the entries we expect.
    // Also copy the timestamps, so we can verify they get injected during replay.
    std::vector<int64_t> timestamps;
    {
        SimpleComms::SQLMessageLog log;
        ASSERT_TRUE(log.open(dbPath));
        SimpleComms::SQLMessageLog::Row row;
        // First line: TaskC subscribing.
        ASSERT_TRUE(log.read(row, true));
        EXPECT_EQ(row.nsec, 0);
        EXPECT_EQ(row.origin, "TaskC");
        EXPECT_EQ(row.channel, "Channel2");
        // Next, TaskA publishing messages with no one to receive them.
        for (int i = 0; i < PUBLISHED_BEFORE_B; ++i) {
            ASSERT_TRUE(log.read(row, true));
            timestamps.push_back(row.nsec);
            EXPECT_EQ(row.origin, "TaskA");
            EXPECT_EQ(row.channel, "Channel1");
            EXPECT_TRUE(row.dropped.empty());
        }
        // Now, TaskB subscribes to that channel.
        ASSERT_TRUE(log.read(row, true));
        EXPECT_EQ(row.nsec, 0);
        EXPECT_EQ(row.origin, "TaskB");
        EXPECT_EQ(row.channel, "Channel1");
        // And, the whole pipeline begins running.
        for (int i = PUBLISHED_BEFORE_B; i < PUBLISHED; ++i) {
            ASSERT_TRUE(log.read(row, true));
            timestamps.push_back(row.nsec);
            EXPECT_EQ(row.origin, "TaskA");
            EXPECT_EQ(row.channel, "Channel1");
            EXPECT_TRUE(row.dropped.empty());

            ASSERT_TRUE(log.read(row, true));
            timestamps.push_back(row.nsec);
            EXPECT_EQ(row.origin, "TaskB");
            EXPECT_EQ(row.channel, "Channel2");
            EXPECT_EQ(row.dropped.empty(), i % 4 != 0);
        }
        // Verify that the timestamps are monotonically increasing.
        EXPECT_TRUE(std::is_sorted(timestamps.begin(), timestamps.end()));
    }

    // Now, replay everything, but this time around:
    // The tasks all connect before any data gets published.
    // Then TaskA publishes everything, then TaskB.
    // Make sure the messages get re-ordered properly to match the original order.
    {
        SimpleComms::MessageControl mc(dbPath, SimpleComms::MessageControl::Mode::REPLAY, SendFunction, params);
        // "Reproducible" messages don't need to get passed on; they are only to
        // update the internal state of MessageControl itself.
        ASSERT_FALSE(mc.preprocess(SimpleComms::messageCreate<SimpleComms::ReproducibleMessage>("TaskA")));
        ASSERT_FALSE(mc.preprocess(SimpleComms::messageCreate<SimpleComms::ReproducibleMessage>("TaskB")));
        ASSERT_FALSE(mc.preprocess(SimpleComms::messageCreate<SimpleComms::ReproducibleMessage>("TaskC")));
        // "Declare" messages get echoed back to the origin.
        std::vector<MessagePtr> declares = {makeDeclare("TaskA", "Channel1", SimpleComms::ChannelType::Publisher),
                                            makeDeclare("TaskB", "Channel1", SimpleComms::ChannelType::Subscriber),
                                            makeDeclare("TaskB", "Channel2", SimpleComms::ChannelType::Publisher),
                                            makeDeclare("TaskC", "Channel2", SimpleComms::ChannelType::Subscriber)};
        for (const MessagePtr& msg : declares) {
            ASSERT_TRUE(mc.preprocess(msg));
            ASSERT_FALSE(mc.add(msg));
        }
        // Ordinary messages get added to the queue of messages to publish.
        std::vector<MessagePtr> channel1;
        idA = 0;
        for (int i = 0; i < PUBLISHED; ++i) {
            MessagePtr msg = makeData("TaskA", "Channel1", idA++);
            ASSERT_TRUE(mc.preprocess(msg));
            ASSERT_TRUE(mc.add(msg));
            channel1.push_back(msg);
        }
        std::vector<MessagePtr> channel2;
        idB = 0;
        for (int i = PUBLISHED_BEFORE_B; i < PUBLISHED; ++i) {
            MessagePtr msg = makeData("TaskB", "Channel2", idB++);
            ASSERT_TRUE(mc.preprocess(msg));
            ASSERT_TRUE(mc.add(msg));
            channel2.push_back(msg);
        }
        auto it1 = channel1.begin();
        auto it2 = channel2.begin();
        auto timeIt = timestamps.begin();
        // Now, pop messages from the queue and verify their order.
        for (int i = 0; i < PUBLISHED; ++i) {
            // Expect a message from Channel1 first.
            MessagePtr msg = mc.pop();
            ASSERT_TRUE(msg);
            EXPECT_EQ(msg, *it1++);
            EXPECT_EQ(readTimestamp(msg), *timeIt++);
            // If TaskB wasn't running yet, expect it to drop this message.
            auto dropped = mc.dropped(msg);
            if (i < PUBLISHED_BEFORE_B) {
                EXPECT_TRUE(dropped.count("TaskB"));
                EXPECT_TRUE(dropped.size() == 1);
                continue;
            } else {
                EXPECT_TRUE(dropped.empty());
            }
            // Now, expect a message from Channel2.
            msg = mc.pop();
            ASSERT_TRUE(msg);
            EXPECT_EQ(msg, *it2++);
            EXPECT_EQ(readTimestamp(msg), *timeIt++);
            // Check if the message was dropped.
            dropped = mc.dropped(msg);
            if (i % 4) {
                EXPECT_TRUE(dropped.empty());
            } else {
                EXPECT_TRUE(dropped.count("TaskC"));
                EXPECT_TRUE(dropped.size() == 1);
            }
        }
    }
}
