/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 **/

#include <array>
#include <chrono>
#include <experimental/filesystem>
#include <gtest/gtest.h>
#include <iostream>
#include <sstream>

#include <SimpleComms/SQLMessageLog.h>

#include "platform/middleware/libs/TestUtils/include/TestUtils/TmpDir.hpp"

namespace fs = std::experimental::filesystem;

using namespace SimpleComms;

class Histogram
{
public:
    Histogram(size_t lower, size_t upper, size_t numBuckets)
        : m_lower(lower)
        , m_upper(upper)
        , m_buckets(numBuckets + 2)
    {}

    ~Histogram() {}

    void insert(size_t result)
    {
        size_t bucket;
        if (result < m_lower) {
            bucket = 0;
        } else if (result > m_upper) {
            bucket = m_buckets.size() - 1;
        } else {
            const size_t rangeLen = m_buckets.size() - 2;
            bucket = ((result - m_lower) * rangeLen) / (m_upper - m_lower) + 1;
        }
        m_buckets[bucket]++;
    }

    friend std::ostream& operator<<(std::ostream& os, const Histogram& h);

private:
    size_t m_lower;
    size_t m_upper;
    std::vector<size_t> m_buckets;
};

std::ostream& operator<<(std::ostream& os, const Histogram& h)
{
    for (const auto& i : h.m_buckets) {
        os << i << ", ";
    }
    return os;
}

TEST(SQLMessageLogTests, BasicDatabaseOperations)
{
    // basic test to make sure no exceptions are thrown
    TestUtils::TestTmpDirectory tmpDir{};
    fs::path dbPath = tmpDir.getPath() / "SQLMessageLogTest.db";
    std::error_code e;
    fs::remove(dbPath, e);

    Histogram h(0, 500000, 50);

    SQLMessageLog database; // should not throw
    EXPECT_TRUE(database.open(dbPath, true, 500));

    constexpr std::array<const char*, 3> origins{"origin1", "origin2", "origin3"};
    constexpr std::array<const char*, 6> channels{
        "channel1", "channel2", "channel3", "channel4", "channel5", "channel6"};
    constexpr size_t numberOfRows = 1000350;

    auto begining = std::chrono::steady_clock::now();

    for (size_t i = 0; i < numberOfRows; ++i) {
        int64_t timestamp = std::chrono::steady_clock::now().time_since_epoch().count();
        auto before = std::chrono::steady_clock::now();
        database.writeMessage(timestamp, origins[i % origins.size()], channels[i % channels.size()]);
        auto after = std::chrono::steady_clock::now();
        h.insert(static_cast<size_t>(std::chrono::duration_cast<std::chrono::nanoseconds>(after - before).count()));
    }
    database.flush();
    auto end = std::chrono::steady_clock::now();

    std::cout << "Writing rows to database took "
              << std::chrono::duration_cast<std::chrono::milliseconds>(end - begining).count() << "ms" << std::endl;

    EXPECT_TRUE(database.open(dbPath, false, 500));

    std::cout << h << std::endl;

    begining = std::chrono::steady_clock::now();

    SQLMessageLog::Row t;
    size_t r{0};
    int64_t lastTimestamp{0};
    while (database.read(t)) {
        EXPECT_LE(lastTimestamp, t.nsec);
        lastTimestamp = t.nsec;
        EXPECT_EQ(t.origin, origins[r % origins.size()]);
        EXPECT_EQ(t.channel, channels[r % channels.size()]);
        r++;
    }
    end = std::chrono::steady_clock::now();
    std::cout << "Reading rows from database took "
              << std::chrono::duration_cast<std::chrono::milliseconds>(end - begining).count() << "ms" << std::endl;
    EXPECT_EQ(r, numberOfRows);

    EXPECT_FALSE(database.read(t)); // additional reads should return false
}

TEST(SQLMessageLogTests, ReopenDatabase)
{
    // test reopening the database a second time an continueing to write
    TestUtils::TestTmpDirectory tmpDir{};
    fs::path dbPath = tmpDir.getPath() / "SQLMessageLogTest.db";
    std::error_code e;
    fs::remove(dbPath, e);

    SQLMessageLog database; // should not throw

    // Each time the database is opened in "write" mode, it should be cleared and start from scratch.
    constexpr std::array<const char*, 3> origins{"origin1", "origin2", "origin3"};
    constexpr std::array<const char*, 4> channels{"channel1", "channel2", "channel3", "channel4"};
    const int64_t REPS = 50;
    for (int i = 0; i < 2; ++i) {
        EXPECT_TRUE(database.open(dbPath, true, 500));

        // add some data
        for (uint64_t j = 0; j < REPS; ++j) {
            const auto rowId = database.writeMessage(std::chrono::steady_clock::now().time_since_epoch().count(),
                                                     origins[j % origins.size()],
                                                     channels[j % channels.size()]);
            EXPECT_EQ(rowId, j + 1);
        }
    }
    database.flush();

    // reopen the database
    EXPECT_TRUE(database.open(dbPath, false));

    SQLMessageLog::Row row;
    for (size_t i = 0; i < REPS; ++i) {
        ASSERT_TRUE(database.read(row));
        EXPECT_EQ(row.origin, origins[i % origins.size()]);
        EXPECT_EQ(row.channel, channels[i % channels.size()]);
    }
    EXPECT_FALSE(database.read(row));
}

TEST(SQLMessageLogTests, DroppedMessages)
{
    TestUtils::TestTmpDirectory tmpDir{};
    fs::path dbPath = tmpDir.getPath() / "SQLMessageLogTest.db";
    std::error_code e;
    fs::remove(dbPath, e);

    SQLMessageLog database; // should not throw
    EXPECT_TRUE(database.open(dbPath, true, 500));

    // TODO : properly test dropping messages
    constexpr std::array<const char*, 3> origins{"origin1", "origin2", "origin3"};
    constexpr std::array<const char*, 6> channels{
        "channel1", "channel2", "channel3", "channel4", "channel5", "channel6"};
    constexpr std::array<const char*, 6> destinations{"dest1", "dest2", "dest3", "dest4"};

    constexpr size_t numberOfRows = 1000000;

    auto begining = std::chrono::steady_clock::now();
    for (size_t i = 1; i <= numberOfRows; ++i) {
        int64_t timestamp = std::chrono::steady_clock::now().time_since_epoch().count();
        auto id = database.writeMessage(timestamp, origins[i % origins.size()], channels[i % channels.size()]);
        ASSERT_TRUE(id);

        if ((i % 3) == 0) {
            database.writeDropped(*id, destinations[0]);
        }
        if ((i % 5) == 0) {
            database.writeDropped(*id, destinations[1]);
        }
        if ((i % 7) == 0) {
            database.writeDropped(*id, destinations[2]);
        }
        if ((i % 8) == 0) {
            database.writeDropped(*id, destinations[3]);
        }
    }
    EXPECT_TRUE(database.flush());
    auto end = std::chrono::steady_clock::now();

    std::cout << "Writing rows to database took "
              << std::chrono::duration_cast<std::chrono::milliseconds>(end - begining).count() << "ms" << std::endl;

    begining = std::chrono::steady_clock::now();
    EXPECT_TRUE(database.open(dbPath, false));

    SQLMessageLog::Row r;
    int64_t id = 1;
    while (database.read(r)) {
        std::set<std::string> expectedDropped;
        if ((id % 3) == 0) {
            expectedDropped.insert(destinations[0]);
        }
        if ((id % 5) == 0) {
            expectedDropped.insert(destinations[1]);
        }
        if ((id % 7) == 0) {
            expectedDropped.insert(destinations[2]);
        }
        if ((id % 8) == 0) {
            expectedDropped.insert(destinations[3]);
        }
        ++id;

        EXPECT_EQ(r.dropped, expectedDropped);
    }
    end = std::chrono::steady_clock::now();

    std::cout << "Reading rows with dropped messages took "
              << std::chrono::duration_cast<std::chrono::milliseconds>(end - begining).count() << "ms" << std::endl;
}

TEST(SQLMessageLogTests, InsertSimilarMessage)
{
    TestUtils::TestTmpDirectory tmpDir{};
    fs::path dbPath = tmpDir.getPath() / "SQLMessageLogTest.db";
    std::error_code e;
    fs::remove(dbPath, e);

    SQLMessageLog database; // should not throw
    EXPECT_TRUE(database.open(dbPath, true, 500));
    EXPECT_EQ(1L, database.writeMessage(12345678, "myorigin", "mychannel"));
    EXPECT_EQ(2L, database.writeMessage(12345678, "myorigin", "mychannel"));
    EXPECT_TRUE(database.flush());
}

TEST(SQLMessageLogTests, OpenReadonlyNonexistent)
{
    TestUtils::TestTmpDirectory tmpDir{};
    fs::path dbPath = tmpDir.getPath() / "SQLNonexistent.db";
    std::error_code e;
    fs::remove(dbPath, e);

    SQLMessageLog database;
    EXPECT_FALSE(database.open(dbPath, false, 500));
}

TEST(SQLMessageLogTests, CreateThenOpenReadonly)
{
    TestUtils::TestTmpDirectory tmpDir{};
    fs::path dbPath = tmpDir.getPath() / "SQLMessageLogTest.db";
    std::error_code e;
    fs::remove(dbPath, e);

    {
        SQLMessageLog database; // should not throw
        EXPECT_TRUE(database.open(dbPath, true, 500));
        EXPECT_EQ(1L, database.writeMessage(12345678, "myorigin", "mychannel"));
        EXPECT_EQ(2L, database.writeMessage(12345678, "myorigin", "mychannel"));
        {
            SQLMessageLog::Row data;
            EXPECT_FALSE(database.read(data, false));
        }
        EXPECT_TRUE(database.flush());
    }

    {
        SQLMessageLog database;
        EXPECT_TRUE(database.open(dbPath, false, 500));
        EXPECT_FALSE(database.writeMessage(12345678, "myorigin", "mychannel"));
    }
}
