/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file UnixClientTests.cc
 * @brief This file implements tests for the TimeStamp class.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#include <SimpleComms/TimeStamp.h>
#include <gtest/gtest.h>

using namespace ::testing;

TEST(TimeStampTest, defaultConstructor)
{
    const TimeStamp t{};

    EXPECT_EQ(0, t.tv_sec);
    EXPECT_EQ(0, t.tv_usec);
}

TEST(TimeStampTest, secAndUsecConstructor)
{
    const TimeStamp t{5, 50};

    EXPECT_EQ(5, t.tv_sec);
    EXPECT_EQ(50, t.tv_usec);
}

TEST(TimeStampTest, doubleConstructor)
{
    const TimeStamp t{2.5};

    EXPECT_EQ(2, t.tv_sec);
    EXPECT_EQ(500000, t.tv_usec);
}

TEST(TimeStampTest, timevalConstructor)
{
    const struct timeval tv
    {
        2, 500
    };
    EXPECT_EQ(2, tv.tv_sec);
    EXPECT_EQ(500, tv.tv_usec);

    const TimeStamp t{tv};
    EXPECT_EQ(2, t.tv_sec);
    EXPECT_EQ(500, t.tv_usec);
}

TEST(TimeStampTest, copyConstructor)
{
    const TimeStamp t{2, 500};
    EXPECT_EQ(2, t.tv_sec);
    EXPECT_EQ(500, t.tv_usec);

    const TimeStamp t2{t};
    EXPECT_EQ(2, t2.tv_sec);
    EXPECT_EQ(500, t2.tv_usec);
}

TEST(TimeStampTest, asDouble)
{
    const TimeStamp t{5, 500000};

    EXPECT_EQ(5.5, static_cast<double>(t));
}

TEST(TimeStampTest, addTimeStamps)
{
    const TimeStamp t1{5, 500000};
    const TimeStamp t2{1, 250000};

    const TimeStamp t = t1 + t2;

    EXPECT_EQ(6, t.tv_sec);
    EXPECT_EQ(750000, t.tv_usec);
}

TEST(TimeStampTest, addDouble)
{
    const TimeStamp t1{5, 500000};
    const TimeStamp t2{1.25};

    const TimeStamp t = t1 + t2;

    EXPECT_EQ(6, t.tv_sec);
    EXPECT_EQ(750000, t.tv_usec);
}

TEST(TimeStampTest, subtractTimeStamps)
{
    const TimeStamp t1{5, 500000};
    const TimeStamp t2{1, 250000};

    const TimeStamp t = t1 - t2;

    EXPECT_EQ(4, t.tv_sec);
    EXPECT_EQ(250000, t.tv_usec);
}

TEST(TimeStampTest, subtractDouble)
{
    const TimeStamp t1{5, 500000};
    const TimeStamp t2{1.25};

    const TimeStamp t = t1 - t2;

    EXPECT_EQ(4, t.tv_sec);
    EXPECT_EQ(250000, t.tv_usec);
}

TEST(TimeStampTest, equal)
{
    const TimeStamp t1{5, 500000};
    const TimeStamp t2{5.5};
    const TimeStamp t3{2, 0};

    EXPECT_EQ(t1, t2);
    EXPECT_NE(t1, t3);
    EXPECT_NE(t2, t3);
    EXPECT_EQ(t2, t1);
    EXPECT_NE(t3, t1);
    EXPECT_NE(t3, t2);
}

TEST(TimeStampTest, lt)
{
    const TimeStamp t1{5, 500000};
    const TimeStamp t2{5.5};
    const TimeStamp t3{2, 0};

    EXPECT_FALSE(t1 < t2);
    EXPECT_FALSE(t1 < t3);
    EXPECT_FALSE(t2 < t1);
    EXPECT_TRUE(t3 < t1);
}

TEST(TimeStampTest, le)
{
    const TimeStamp t1{5, 500000};
    const TimeStamp t2{5.5};
    const TimeStamp t3{2, 0};

    EXPECT_TRUE(t1 <= t2);
    EXPECT_FALSE(t1 <= t3);
    EXPECT_TRUE(t2 <= t1);
    EXPECT_TRUE(t3 <= t1);
}

TEST(TimeStampTest, gt)
{
    const TimeStamp t1{5, 500000};
    const TimeStamp t2{5.5};
    const TimeStamp t3{2, 0};

    EXPECT_FALSE(t1 > t2);
    EXPECT_TRUE(t1 > t3);
    EXPECT_FALSE(t2 > t1);
    EXPECT_FALSE(t3 > t1);
}

TEST(TimeStampTest, ge)
{
    const TimeStamp t1{5, 500000};
    const TimeStamp t2{5.5};
    const TimeStamp t3{2, 0};

    EXPECT_TRUE(t1 >= t2);
    EXPECT_TRUE(t1 >= t3);
    EXPECT_TRUE(t2 >= t1);
    EXPECT_FALSE(t3 >= t1);
}

TEST(TimeStampTest, now)
{
    const TimeStamp zero{};

    TimeStamp t;
    t.setNow();

    EXPECT_GT(t, zero);
}

TEST(TimeStampTest, normalize)
{
    {
        TimeStamp t{5, 500000};
        t.normalize();
        EXPECT_EQ(5, t.tv_sec);
        EXPECT_EQ(500000, t.tv_usec);
    }

    {
        TimeStamp t{5, 1600000};
        t.normalize();
        EXPECT_EQ(6, t.tv_sec);
        EXPECT_EQ(600000, t.tv_usec);
    }

    {
        TimeStamp t{5, -700000};
        t.normalize();
        EXPECT_EQ(4, t.tv_sec);
        EXPECT_EQ(300000, t.tv_usec);
    }
}

TEST(TimeStampTest, toString)
{
    {
        const TimeStamp t{2, 50};
        EXPECT_EQ(std::string("{2s, 50us}"), t.toString());
    }

    {
        const TimeStamp t{-2, 50};
        EXPECT_EQ(std::string("{-2s, 50us}"), t.toString());
    }

    {
        const TimeStamp t{2, 0};
        EXPECT_EQ(std::string("{2s, 0us}"), t.toString());
    }

    {
        const TimeStamp t{2.5};
        EXPECT_EQ(std::string("{2s, 500000us}"), t.toString());
    }

    {
        const TimeStamp t{2.0};
        EXPECT_EQ(std::string("{2s, 0us}"), t.toString());
    }
}
