/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file UnixClientTests.cc
 * @brief This file implements tests for the UnixClient class.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#include <chrono>
#include <memory>
#include <string>
#include <thread>
#include <vector>

#include <SimpleComms/test/MockUnixClient.h>
#include <gmock/gmock.h>
#include <gtest/gtest-param-test.h>
#include <gtest/gtest.h>

using namespace ::testing;
using ::testing::_;
using ::testing::DefaultValue;
using ::testing::Invoke;

TEST(UnixClientTests, TestMakeUniqueClientPrefix)
{
    auto name = SimpleComms::UnixClient::makeUniqueClientName("prefix");
    auto pos = name.find("prefix");
    EXPECT_EQ(0, pos);
}

/*
 * @brief Test the uniqueness of the names made with makeUniqueClientName
 *
 * Truly testing uniqueness is difficult, so we do a simple validation
 */
TEST(UnixClientTests, TestMakeUniqueClientUniqueness)
{
    // Currently this is testing the underlying UUID implementation, but if that
    // should change we still want some kind of test
    //
    // On a Thinkpad P50 this takes 25ms to run
    std::set<std::string> names;
    names.insert(SimpleComms::UnixClient::makeUniqueClientName());
    for (size_t i = 0; i < 1000; ++i) {
        auto name = SimpleComms::UnixClient::makeUniqueClientName();
        EXPECT_EQ(names.find(name), names.end());
        names.insert(name);
    }
}

/*
 * @brief Test the expected default path for scs
 */
TEST(UnixClientTests, TestGetDefaultPathNoEnvVar)
{
    // the tests are already run in a different session than the calling process, so this is okay
    unsetenv("SCS_PATH");
    EXPECT_EQ(SimpleComms::UnixClient::getDefaultScsPath(), "scs");
}

/*
 * @brief Test the path for scs if environmentally configured
 */
TEST(UnixClientTests, TestGetDefaultPathWithEnvVar)
{
    // the tests are already run in a different session than the calling process, so this is okay
    // Overwrite to true
    setenv("SCS_PATH", "great_path", true);
    EXPECT_EQ(SimpleComms::UnixClient::getDefaultScsPath(), "great_path");
}

class UCTests : public ::testing::Test
{
protected:
    NiceMock<SimpleComms::MockUnixClient> testClient_;

    void SetUp() override
    {
        ON_CALL(testClient_, declare(_, _, _))
            .WillByDefault(Invoke(&testClient_, &SimpleComms::MockUnixClient::UnixClientDeclare));

        ON_CALL(testClient_, undeclare(_, _))
            .WillByDefault(Invoke(&testClient_, &SimpleComms::MockUnixClient::UnixClientUndeclare));
    }

    void TearDown() override {}
};

TEST_F(UCTests, DeclarePublisher)
{
    // Confirm that the declare function forms a properly-formatted message, and that all versions of declare are
    // equivalent
    std::string blob;
    std::string channelName = "Channel";
    SimpleComms::ChannelType ctype = SimpleComms::ChannelType::Publisher;

    SimpleComms::DeclareChannelMessage expectedResult;
    expectedResult.type =
        static_cast<SimpleComms::ControlMessages::ControlMessageTypes>(htonl(SimpleComms::ControlMessages::Declare));
    expectedResult.channelType = static_cast<SimpleComms::ChannelType>(htonl(ctype));
    memcpy(expectedResult.channel, channelName.c_str(), channelName.length() + 1);

    testClient_.blob =
        std::string(reinterpret_cast<const char*>(&expectedResult), sizeof(SimpleComms::DeclareChannelMessage));

    EXPECT_CALL(testClient_, declare(channelName, ctype, 0)).Times(1);

    EXPECT_CALL(testClient_, sendMsgV(_, _, _))
        .Times(1)
        .WillRepeatedly(Invoke(&testClient_, &SimpleComms::MockUnixClient::checkChannelMapAndBlobEquivalence));

    EXPECT_EQ(testClient_.declare(channelName, ctype, 0), SimpleComms::Status::States::Ok);
}

TEST_F(UCTests, DeclareSubscriber)
{
    // Confirm that the declare function forms a properly-formatted message, and that all versions of declare are
    // equivalent
    std::string blob;
    std::string channelName = "Channel";
    SimpleComms::ChannelType ctype = SimpleComms::ChannelType::Subscriber;

    SimpleComms::DeclareChannelMessage expectedResult;
    expectedResult.type =
        static_cast<SimpleComms::ControlMessages::ControlMessageTypes>(htonl(SimpleComms::ControlMessages::Declare));
    expectedResult.channelType = static_cast<SimpleComms::ChannelType>(htonl(ctype));
    memcpy(expectedResult.channel, channelName.c_str(), channelName.length() + 1);

    testClient_.blob =
        std::string(reinterpret_cast<const char*>(&expectedResult), sizeof(SimpleComms::DeclareChannelMessage));

    EXPECT_CALL(testClient_, declare(channelName, ctype, 10));

    EXPECT_CALL(testClient_, sendMsgV(_, _, _))
        .Times(1)
        .WillRepeatedly(Invoke(&testClient_, &SimpleComms::MockUnixClient::checkChannelMapAndBlobEquivalence));

    EXPECT_EQ(testClient_.declare(channelName, ctype, 10), SimpleComms::Status::States::Ok);
}

TEST_F(UCTests, DeclareSubscriberQueueSizeChanges)
{
    // Confirm that the declare function forms a properly-formatted message, and that all versions of declare are
    // equivalent
    std::string blob;
    std::string channelName = "Channel";
    SimpleComms::ChannelType ctype = SimpleComms::ChannelType::Subscriber;

    SimpleComms::DeclareChannelMessage expectedResult;
    expectedResult.type =
        static_cast<SimpleComms::ControlMessages::ControlMessageTypes>(htonl(SimpleComms::ControlMessages::Declare));
    expectedResult.channelType = static_cast<SimpleComms::ChannelType>(htonl(ctype));
    memcpy(expectedResult.channel, channelName.c_str(), channelName.length() + 1);

    testClient_.blob =
        std::string(reinterpret_cast<const char*>(&expectedResult), sizeof(SimpleComms::DeclareChannelMessage));

    EXPECT_CALL(testClient_, declare(channelName, ctype, 10));
    EXPECT_CALL(testClient_, declare(channelName, ctype, 1));
    EXPECT_CALL(testClient_, declare(channelName, ctype, 11));

    EXPECT_CALL(testClient_, sendMsgV(_, _, _))
        .Times(3)
        .WillRepeatedly(Invoke(&testClient_, &SimpleComms::MockUnixClient::checkChannelMapAndBlobEquivalence));

    EXPECT_EQ(testClient_.declare(channelName, ctype, 10), SimpleComms::Status::States::Ok);

    EXPECT_EQ(testClient_.getMaxReceiveDepth(channelName), 10);

    EXPECT_EQ(testClient_.declare(channelName, ctype, 1), SimpleComms::Status::States::Ok);

    EXPECT_EQ(testClient_.getMaxReceiveDepth(channelName), 10);

    EXPECT_EQ(testClient_.declare(channelName, ctype, 11), SimpleComms::Status::States::Ok);

    EXPECT_EQ(testClient_.getMaxReceiveDepth(channelName), 11);
}

TEST_F(UCTests, DeclareSubPub)
{
    // Confirm that the declare function forms a properly-formatted message, and that all versions of declare are
    // equivalent
    std::string blob;
    std::string channelName = "Channel";
    SimpleComms::ChannelType ctype = SimpleComms::ChannelType::SubPub;

    SimpleComms::DeclareChannelMessage expectedResult;
    expectedResult.type =
        static_cast<SimpleComms::ControlMessages::ControlMessageTypes>(htonl(SimpleComms::ControlMessages::Declare));
    expectedResult.channelType = static_cast<SimpleComms::ChannelType>(htonl(ctype));
    memcpy(expectedResult.channel, channelName.c_str(), channelName.length() + 1);

    testClient_.blob =
        std::string(reinterpret_cast<const char*>(&expectedResult), sizeof(SimpleComms::DeclareChannelMessage));

    EXPECT_CALL(testClient_, declare(channelName, ctype, 10));

    EXPECT_CALL(testClient_, sendMsgV(_, _, _))
        .Times(1)
        .WillRepeatedly(Invoke(&testClient_, &SimpleComms::MockUnixClient::checkChannelMapAndBlobEquivalence));

    EXPECT_EQ(testClient_.declare(channelName, ctype, 10), SimpleComms::Status::States::Ok);
}

TEST_F(UCTests, DeclareUnlisted)
{
    // Confirm that the declare function forms a properly-formatted message, and that all versions of declare are
    // equivalent
    std::string blob;
    std::string channelName = "Channel";
    SimpleComms::ChannelType ctype = SimpleComms::ChannelType::Unlisted;

    SimpleComms::DeclareChannelMessage expectedResult;
    expectedResult.type =
        static_cast<SimpleComms::ControlMessages::ControlMessageTypes>(htonl(SimpleComms::ControlMessages::Declare));
    expectedResult.channelType = static_cast<SimpleComms::ChannelType>(htonl(ctype));
    memcpy(expectedResult.channel, channelName.c_str(), channelName.length() + 1);

    testClient_.blob =
        std::string(reinterpret_cast<const char*>(&expectedResult), sizeof(SimpleComms::DeclareChannelMessage));

    EXPECT_CALL(testClient_, declare(channelName, ctype, 10));

    EXPECT_CALL(testClient_, sendMsgV(_, _, _)).Times(0);

    EXPECT_EQ(testClient_.declare(channelName, ctype, 10), SimpleComms::Status::States::Ok);
}

TEST_F(UCTests, GetPublishedMsg)
{
    auto stringVec = testClient_.getPublishListMessage();
    EXPECT_EQ(stringVec.size(), 0);
}

class UCTestsTypeParam : public ::testing::TestWithParam<SimpleComms::ChannelType>
{
protected:
    NiceMock<SimpleComms::MockUnixClient> testClient_;
    SimpleComms::ChannelType ctype_;

    void SetUp() override
    {
        ctype_ = GetParam();
        ON_CALL(testClient_, declare(_, _, _))
            .WillByDefault(Invoke(&testClient_, &SimpleComms::MockUnixClient::UnixClientDeclare));
        ON_CALL(testClient_, undeclare(_, _))
            .WillByDefault(Invoke(&testClient_, &SimpleComms::MockUnixClient::UnixClientUndeclare));
    }

    void TearDown() override {}
};

TEST_P(UCTestsTypeParam, NoRedeclareMessage)
{
    std::string channelName = "channel";

    InSequence is;
    EXPECT_CALL(testClient_, declare(_, _, _)).RetiresOnSaturation();

    EXPECT_CALL(testClient_, sendMsgV(_, _, _))
        .WillOnce(DoAll(IgnoreResult(Invoke(&testClient_, &SimpleComms::MockUnixClient::AcknowledgeChannel)),
                        Invoke(&testClient_, &SimpleComms::MockUnixClient::ConfirmChannelMap)))
        .RetiresOnSaturation();

    EXPECT_CALL(testClient_, declare(_, _, _)).RetiresOnSaturation();

    EXPECT_CALL(testClient_, undeclare(_, _)).RetiresOnSaturation();

    EXPECT_CALL(testClient_, sendMsgV(_, _, _)).RetiresOnSaturation();

    EXPECT_CALL(testClient_, undeclare(_, _)).RetiresOnSaturation();

    EXPECT_CALL(testClient_, declare(_, _, _)).RetiresOnSaturation();

    EXPECT_CALL(testClient_, sendMsgV(_, _, _))
        .WillOnce(DoAll(IgnoreResult(Invoke(&testClient_, &SimpleComms::MockUnixClient::ConfirmChannelMap)),
                        Invoke(&testClient_, &SimpleComms::MockUnixClient::ConfirmChannelMap)));

    EXPECT_EQ(testClient_.declare(channelName, ctype_, 10), SimpleComms::Status::States::Ok);
    EXPECT_EQ(testClient_.declare(channelName, ctype_, 10), SimpleComms::Status::States::Ok);
    EXPECT_EQ(testClient_.undeclare(channelName, ctype_), SimpleComms::Status::States::Ok);
    EXPECT_EQ(testClient_.undeclare(channelName, ctype_), SimpleComms::Status::States::Ok);
    EXPECT_EQ(testClient_.declare(channelName, ctype_, 10), SimpleComms::Status::States::Ok);
}

TEST_P(UCTestsTypeParam, NoRedeclareMismatch)
{
    std::string channelName = "channel";

    InSequence is;
    EXPECT_CALL(testClient_, declare(_, _, _)).RetiresOnSaturation();
    EXPECT_CALL(testClient_, sendMsgV(_, _, _))
        .WillOnce(DoAll(IgnoreResult(Invoke(&testClient_, &SimpleComms::MockUnixClient::AcknowledgeChannel)),
                        Invoke(&testClient_, &SimpleComms::MockUnixClient::ConfirmChannelMap)))
        .RetiresOnSaturation();

    EXPECT_CALL(testClient_, declare(_, _, _));
    EXPECT_CALL(testClient_, sendMsgV(_, _, _)).Times(0);

    EXPECT_EQ(testClient_.declare(channelName, SimpleComms::ChannelType::SubPub, 10), SimpleComms::Status::States::Ok);
    EXPECT_EQ(testClient_.declare(channelName, ctype_, 10), SimpleComms::Status::States::Ok);
}

class UCTestsType2Params : public ::testing::TestWithParam<size_t>
{
protected:
    NiceMock<SimpleComms::MockUnixClient> testClient_;
    SimpleComms::ChannelType ctype1_;
    SimpleComms::ChannelType ctype2_;

    std::vector<SimpleComms::ChannelType> ctypes_ = {SimpleComms::ChannelType::Unlisted,
                                                     SimpleComms::ChannelType::Subscriber,
                                                     SimpleComms::ChannelType::Publisher,
                                                     SimpleComms::ChannelType::SubPub};
    void SetUp() override
    {
        size_t param = GetParam();

        EXPECT_TRUE(param < 8);

        ctype1_ = ctypes_[param % 4];
        ctype2_ = ctypes_[param / 4];

        ON_CALL(testClient_, declare(_, _, _))
            .WillByDefault(Invoke(&testClient_, &SimpleComms::MockUnixClient::UnixClientDeclare));
        ON_CALL(testClient_, undeclare(_, _))
            .WillByDefault(Invoke(&testClient_, &SimpleComms::MockUnixClient::UnixClientUndeclare));
    }

    void TearDown() override {}
};

TEST_P(UCTestsType2Params, AddPermissionsUniversal)
{
    std::string channelName = "channel";

    InSequence is;

    // Declare original state
    EXPECT_CALL(testClient_, declare(_, ctype1_, _)).RetiresOnSaturation();
    if (ctype1_ != SimpleComms::ChannelType::Unlisted) {
        EXPECT_CALL(testClient_, sendMsgV(_, _, _))
            .WillOnce(DoAll(IgnoreResult(Invoke(&testClient_, &SimpleComms::MockUnixClient::AcknowledgeChannel)),
                            Invoke(&testClient_, &SimpleComms::MockUnixClient::ConfirmChannelMap)))
            .RetiresOnSaturation();
    }

    // Declare new state
    EXPECT_CALL(testClient_, declare(_, ctype2_, _)).RetiresOnSaturation();

    // There exist permissions to declare in ctype2_ that were not already declared in ctype1_
    if ((ctype1_ | ctype2_) != ctype1_) {
        EXPECT_CALL(testClient_, sendMsgV(_, _, _))
            .WillOnce(DoAll(IgnoreResult(Invoke(&testClient_, &SimpleComms::MockUnixClient::AcknowledgeChannel)),
                            Invoke(&testClient_, &SimpleComms::MockUnixClient::ConfirmChannelMap)))
            .RetiresOnSaturation();
    }

    // Declare original state
    EXPECT_EQ(testClient_.declare(channelName, ctype1_, 10), SimpleComms::Status::States::Ok);
    EXPECT_EQ(ctype1_, testClient_.getChannelType(channelName));

    // Declare the new state
    EXPECT_EQ(testClient_.declare(channelName, ctype2_, 10), SimpleComms::Status::States::Ok);
    EXPECT_EQ(ctype1_ | ctype2_, testClient_.getChannelType(channelName));
}

TEST_P(UCTestsType2Params, RemovePermissionsUniversal)
{
    std::string channelName = "channel";

    InSequence is;

    // Declare original state
    EXPECT_CALL(testClient_, declare(_, ctype1_, _)).RetiresOnSaturation();
    if (ctype1_ != SimpleComms::ChannelType::Unlisted) {
        EXPECT_CALL(testClient_, sendMsgV(_, _, _))
            .WillOnce(DoAll(IgnoreResult(Invoke(&testClient_, &SimpleComms::MockUnixClient::AcknowledgeChannel)),
                            Invoke(&testClient_, &SimpleComms::MockUnixClient::ConfirmChannelMap)))
            .RetiresOnSaturation();
    }

    // Undeclare the new state
    EXPECT_CALL(testClient_, undeclare(_, ctype2_)).RetiresOnSaturation();

    // There exist permissions that need to be undeclared.
    if ((ctype1_ & ~ctype2_) != ctype1_) {
        EXPECT_CALL(testClient_, sendMsgV(_, _, _)).RetiresOnSaturation();
    }

    // Declare original state
    EXPECT_EQ(testClient_.declare(channelName, ctype1_, 10), SimpleComms::Status::States::Ok);
    EXPECT_EQ(ctype1_, testClient_.getChannelType(channelName));

    // Declare the new state
    EXPECT_EQ(testClient_.undeclare(channelName, ctype2_), SimpleComms::Status::States::Ok);
    EXPECT_EQ(ctype1_ & ~ctype2_, testClient_.getChannelType(channelName));
}

INSTANTIATE_TEST_SUITE_P(RedeclareNoRedeclare,
                         UCTestsTypeParam,
                         ::testing::Values(SimpleComms::ChannelType::Subscriber,
                                           SimpleComms::ChannelType::Publisher,
                                           SimpleComms::ChannelType::SubPub));

INSTANTIATE_TEST_SUITE_P(AddRemoveUniversal, UCTestsType2Params, ::testing::Range<size_t>(0, 8));

GTEST_API_ int main(int argc, char** argv)
{
    ::testing::InitGoogleMock(&argc, argv);
    return RUN_ALL_TESTS();
}
