#include <SimpleComms/Client.h>
#include <SimpleComms/ControlMessages.h>
#include <SimpleComms/UnixClient.h>

#include <atomic>
#include <chrono>
#include <cmath>
#include <condition_variable>
#include <fstream>
#include <iostream>
#include <memory>
#include <mutex>
#include <sstream>
#include <string>
#include <thread>
#include <unordered_map>
#include <vector>

struct receiveDescriptor
{
    std::string name;
    std::string channel;
    size_t queueSize;
};

struct receivePair
{
    std::shared_ptr<SimpleComms::UnixClient> receiver;
    std::string channel;
    std::string name;
};

struct statistics
{
    std::chrono::steady_clock::time_point start;
    std::chrono::steady_clock::time_point end;

    std::unordered_map<std::string, double> receiveCounts;
};

struct cpuTime_t
{
    size_t userTime_ticks;
    size_t kernelTime_ticks;
    size_t childrenWaitUserTime_ticks;
    size_t childrenWaitKernelTime_ticks;
    size_t vmem_bytes;
    size_t rss_pages;
};

namespace {
std::optional<size_t> parseSizeT(const std::string& s)
{
    try {
        return std::stoi(s);
    } catch (std::invalid_argument&) {
        return {};
    } catch (std::out_of_range&) {
        return {};
    }
}
} // namespace

cpuTime_t getProcCpuTimes(int pid)
{
    std::ifstream infile("/proc/" + std::to_string(pid) + "/stat");

    // Get the raw information
    std::string statStr;
    std::getline(infile, statStr);

    std::stringstream ss(statStr);

    std::string token;
    std::vector<std::string> statTokens;
    while (std::getline(ss, token, ' ')) {
        statTokens.push_back(token);
    }

    cpuTime_t cpuTimes{};
    cpuTimes.userTime_ticks = parseSizeT(statTokens[13]).value_or(0);
    cpuTimes.kernelTime_ticks = parseSizeT(statTokens[14]).value_or(0);
    cpuTimes.vmem_bytes = parseSizeT(statTokens[22]).value_or(0);
    cpuTimes.rss_pages = parseSizeT(statTokens[23]).value_or(0);

    return cpuTimes;
}

double procCpuUsage(const int pid, const cpuTime_t& prev, const double duration)
{
    const double hertz = 100;
    auto const curr = getProcCpuTimes(pid);

    // How many ticks did the process use since the last sampling?
    size_t intervalTicks =
        (curr.userTime_ticks - prev.userTime_ticks) + (curr.kernelTime_ticks - prev.kernelTime_ticks);

    // Convert to percentage
    return static_cast<double>(100 * intervalTicks) / (hertz * duration);
}

int main(int argc, char** argv)
{
    if (argc != 3) {
        std::cerr << "Usge: scsRx [path_to_config_file] [scs pid]" << std::endl;
        return 1;
    }

    const int pid = std::stoi(argv[2]);

    std::vector<receiveDescriptor> receiverDescs;
    std::vector<size_t> latencies;
    latencies.reserve(100000000);
    std::mutex latenciesLock;

    std::ifstream infile(argv[1]);
    std::string line;
    while (std::getline(infile, line)) {
        std::istringstream ss(line);

        std::string name;
        std::getline(ss, name, ',');
        std::string channel;
        std::getline(ss, channel, ',');
        std::string depth;
        std::getline(ss, depth, ',');

        receiveDescriptor rd = {name, channel, static_cast<size_t>(atoi(depth.c_str()))};
        receiverDescs.push_back(rd);
    }

    statistics stats;
    std::mutex statsLock;

    // Build the receivers
    std::vector<receivePair> receivePairs;
    for (auto const& desc : receiverDescs) {
        receivePair rp;

        rp.receiver =
            std::make_shared<SimpleComms::UnixClient>(SimpleComms::UnixClient::getDefaultScsPath(), desc.name);
        rp.channel = desc.channel;
        rp.name = desc.name;

        stats.receiveCounts[rp.name] = 0;

        if (rp.receiver->connect() != SimpleComms::Status::Ok) {
            std::cerr << "Failed to connect receiver to scs!" << std::endl;
            std::cerr << "\nname: " << desc.name << "\nchannel: " << desc.channel << std::endl;

            for (auto& pair : receivePairs) {
                pair.receiver->disconnect();
            }

            return 1;
        }

        rp.receiver->declare(desc.channel, SimpleComms::ChannelType::Subscriber, static_cast<uint32_t>(desc.queueSize));

        receivePairs.push_back(rp);
    }

    // Shared variables across threads
    std::atomic<bool> done(false);
    std::atomic<bool> recording(false);

    // And the command channel
    SimpleComms::UnixClient command(SimpleComms::UnixClient::getDefaultScsPath(), "command-rx");
    if (command.connect() != SimpleComms::Status::Ok) {
        std::cerr << "Failed to connect command to scs!" << std::endl;
        return 1;
    }

    command.declare("command", SimpleComms::ChannelType::Subscriber, 10);

    std::thread commandThread([&stats, &done, &recording, &command]() {
        while (!done) {
            if (command.hasMsg()) {
                std::string msg;
                MessageInfo msgInfo;
                if (command.receiveMsg("command", msg, msgInfo)) {
                    std::cout << "Received command message..." << std::endl;
                    if (msg == "start") {
                        recording = true;
                        stats.start = std::chrono::steady_clock::now();
                    }
                    if (msg == "stop") {
                        recording = false;
                        done = true;

                        stats.end = std::chrono::steady_clock::now();
                    }
                }
            }
        }
    });

    std::vector<double> cpu_usages;
    std::vector<size_t> vmem_usages;
    std::vector<size_t> rss_usages;

    // launch the monitoring thread
    std::thread monitorThread([&cpu_usages, &vmem_usages, &rss_usages, &pid, &done]() {
        while (!done) {
            auto const times = getProcCpuTimes(pid);

            std::this_thread::sleep_for(std::chrono::seconds(1));

            auto const usage = procCpuUsage(pid, times, 1.0);

            cpu_usages.push_back(usage);
            vmem_usages.push_back(times.vmem_bytes);
            rss_usages.push_back(times.rss_pages * 4096);
        }
    });

    std::vector<std::thread> receiveThreads;
    for (auto& receivePair : receivePairs) {
        receiveThreads.push_back(
            std::thread([&statsLock, &latenciesLock, &latencies, &done, &recording, &receivePair, &stats]() {
                std::string::size_type sz;
                std::string msg;
                MessageInfo msgInfo;

                std::vector<size_t> localLatencies;

                size_t byteCount = 0;

                while (!done) {
                    if (receivePair.receiver->hasMsg()) {
                        while (receivePair.receiver->receiveMsg(receivePair.channel, msg, msgInfo)) {
                            if (recording) {
                                auto now = std::chrono::steady_clock::now();
                                auto now_ns = std::chrono::time_point_cast<std::chrono::nanoseconds>(now)
                                                  .time_since_epoch()
                                                  .count();
                                byteCount += msg.size();

                                auto send_now_str = msg.substr(0, 20);
                                send_now_str.erase(0, send_now_str.find_first_not_of('0'));
                                auto send_now_ns = std::stol(send_now_str, &sz);

                                localLatencies.emplace_back(now_ns - send_now_ns);
                            }
                        }
                    }
                }

                {
                    std::lock_guard<std::mutex> lock(statsLock);
                    stats.receiveCounts[receivePair.name] = static_cast<double>(byteCount);
                }
                {
                    std::lock_guard<std::mutex> lock(latenciesLock);
                    for (auto& i : localLatencies) {
                        latencies.push_back(i);
                    }
                }
            }));
    }

    for (auto& t : receiveThreads) {
        t.join();
    }

    // Clean up
    commandThread.join();
    monitorThread.join();

    for (auto& receivePair : receivePairs) {
        receivePair.receiver->disconnect();
    }

    // Crunch the numbers
    double total_received = 0;
    for (auto const count : stats.receiveCounts) {
        std::cout << count.first << ": " << count.second << " bytes" << std::endl;
        total_received += count.second;
    }

    std::chrono::duration<double> total_dur = stats.end - stats.start;
    double total_time = total_dur.count();
    double receiveRate = total_received / total_time;
    std::cout << total_received << "\n" << total_time << "\n" << receiveRate / 1000000.0 << std::endl;

    // Get basic latency figures
    double latency_sum = 0;
    size_t max_latency = 0;
    size_t min_latency = std::numeric_limits<size_t>::max();
    for (auto const x : latencies) {
        if (x > max_latency) {
            max_latency = x;
        }

        if (x < min_latency) {
            min_latency = x;
        }

        latency_sum += static_cast<double>(x);
    }

    double latency_ns_avg = latency_sum / static_cast<double>(latencies.size());

    // Calculate standard deviation
    double stddev_sum = 0;
    for (auto const x : latencies) {
        double diff = static_cast<double>(x) - latency_ns_avg;
        stddev_sum += diff * diff;
    }

    double stddev_ns = std::sqrt(stddev_sum / static_cast<double>(latencies.size()));

    double latency_ms_avg = latency_ns_avg / 1000000;

    std::cout << "avg:    " << latency_ms_avg << std::endl;
    std::cout << "stddev: " << stddev_ns / 1000000.0 << std::endl;
    std::cout << "min:    " << (static_cast<double>(min_latency) / 1000000.0) << std::endl;
    std::cout << "max:    " << (static_cast<double>(max_latency) / 1000000.0) << std::endl;

    std::string csvOutStr = std::to_string(latency_ms_avg) + "," + std::to_string(stddev_ns / 1000000.0) + "," +
                            std::to_string(static_cast<double>(min_latency) / 1000000.0) + "," +
                            std::to_string(static_cast<double>(max_latency) / 1000000.0) + "," +
                            std::to_string(receiveRate / 1000000.0) + ",";

    // Dump info in csv
    std::cout << latency_ms_avg << "," << stddev_ns / 1000000.0 << "," << static_cast<double>(min_latency) / 1000000.0
              << "," << static_cast<double>(max_latency) / 1000000.0 << "," << receiveRate / 1000000.0 << ",";

    // Simple buckets
    std::vector<size_t> buckets = {0, 0, 0, 0, 0, 0, 0, 0, 0};
    for (auto lat : latencies) {
        double x = static_cast<double>(lat) / 1000000.0;
        if (x < 0.3)
            ++buckets[0];
        else if (x < 0.5)
            ++buckets[1];
        else if (x < 1)
            ++buckets[2];
        else if (x < 2)
            ++buckets[3];
        else if (x < 5)
            ++buckets[4];
        else if (x < 10)
            ++buckets[5];
        else if (x < 25)
            ++buckets[6];
        else if (x < 50)
            ++buckets[7];
        else
            ++buckets[8];
    }

    for (auto const b : buckets) {
        std::cout << b << ",";
        csvOutStr += std::to_string(b) + ",";
    }

    // And now the cpu/mem
    double avg_cpu = 0;
    double avg_vmem = 0;
    double avg_rss = 0;

    for (auto x : cpu_usages) {
        avg_cpu += x;
    }
    avg_cpu /= static_cast<double>(cpu_usages.size());

    for (auto x : vmem_usages) {
        avg_vmem += static_cast<double>(x);
    }
    avg_vmem /= static_cast<double>(vmem_usages.size());
    avg_vmem /= 1000000.0;

    for (auto x : rss_usages) {
        avg_rss += static_cast<double>(x);
    }
    avg_rss /= static_cast<double>(rss_usages.size());
    avg_rss /= 1000000.0;

    std::cout << avg_cpu << "," << avg_vmem << "," << avg_rss;
    csvOutStr += std::to_string(avg_cpu) + "," + std::to_string(avg_vmem) + "," + std::to_string(avg_rss);

    std::cout << std::endl;

    return 0;
}
