#include <SimpleComms/Client.h>
#include <SimpleComms/UnixClient.h>

#include <algorithm>
#include <atomic>
#include <chrono>
#include <condition_variable>
#include <fstream>
#include <iostream>
#include <memory>
#include <mutex>
#include <sstream>
#include <string>
#include <thread>
#include <vector>

struct sendDescriptor
{
    std::string name;
    std::string channel;
};

struct sendPair
{
    std::shared_ptr<SimpleComms::UnixClient> sender;
    std::string channel;
};

std::string nowString()
{
    auto now = std::chrono::steady_clock::now();
    auto now_ns = std::chrono::time_point_cast<std::chrono::nanoseconds>(now);
    auto now_ns_str = std::to_string(now_ns.time_since_epoch().count());
    size_t padLength = 20 - now_ns_str.length();

    return std::string(padLength, '0').append(now_ns_str);
}

int main(int argc, char** argv)
{
    if (argc != 2) {
        std::cerr << "Usge: scsRx [path_to_config_file]" << std::endl;
        return 1;
    }

    std::vector<sendDescriptor> senderDescs;
    std::ifstream infile(argv[1]);
    std::string line;
    std::getline(infile, line);
    int _msgSize = atoi(line.c_str());
    if (_msgSize < 0) {
        std::cerr << "Message size must be a nonnegative integer";
        return 1;
    }

    size_t msgSize = static_cast<size_t>(_msgSize);

    std::getline(infile, line);
    auto duration_s = std::chrono::duration<double>(atoi(line.c_str()));

    // Now get the channel configs
    while (std::getline(infile, line)) {
        std::istringstream ss(line);

        std::string name;
        std::getline(ss, name, ',');
        std::string channel;
        std::getline(ss, channel, ',');

        sendDescriptor sd = {name, channel};
        senderDescs.push_back(sd);
    }

    // Build the senders
    std::vector<sendPair> sendPairs;
    for (auto const& desc : senderDescs) {
        sendPair sp;

        sp.sender = std::make_shared<SimpleComms::UnixClient>(SimpleComms::UnixClient::getDefaultScsPath(), desc.name);
        sp.channel = desc.channel;
        sendPairs.push_back(sp);

        if (sendPairs.back().sender->connect() != SimpleComms::Status::Ok) {
            std::cerr << "Failed to connect to scs!" << std::endl;
            return 1;
        }
    }

    // And the command channel
    SimpleComms::UnixClient command(SimpleComms::UnixClient::getDefaultScsPath(), "command-tx");
    if (command.connect() != SimpleComms::Status::Ok) {
        std::cerr << "Failed to connect to scs!" << std::endl;
        return 1;
    }

    // Generate some test data
    std::string data = {};
    for (size_t i = 0; i < msgSize; ++i) {
        data += 'a';
    }

    auto start = std::chrono::steady_clock::now();

    // Signal to the receiving end to start recording information
    if (command.sendMsg("command", "start") != SimpleComms::Status::Ok) {
        std::cerr << "Failed to send message..." << std::endl;
        return 1;
    }

    // Actually push out all of the data
    std::vector<std::thread> sendThreads;
    for (auto& sendPair : sendPairs) {
        sendThreads.push_back(std::thread([start, duration_s, &data, &sendPair]() {
            while ((std::chrono::steady_clock::now() - start) < duration_s) {
                auto nowstr = nowString();
                auto timestampped_data = nowstr + data;
                if (sendPair.sender->sendMsg(sendPair.channel, timestampped_data) != SimpleComms::Status::Ok) {
                    std::cerr << "Failed to send message..." << std::endl;
                }

                std::this_thread::sleep_for(std::chrono::milliseconds(10));
            }
        }));
    }

    for (auto& t : sendThreads) {
        t.join();
    }

    // Spam this a few times to make sure it really stops in case we're saturating the link
    for (size_t i = 0; i < 20; ++i) {
        // Signal to the receiving end to stop recording information
        if (command.sendMsg("command", "stop") != SimpleComms::Status::Ok) {
            std::cerr << "Failed to send message..." << std::endl;
            return 1;
        }

        std::this_thread::sleep_for(std::chrono::milliseconds(50));
    }

    for (auto& sendPair : sendPairs) {
        sendPair.sender->disconnect();
    }

    return 0;
}
