#include <SimpleComms/Client.h>
#include <SimpleComms/ControlMessages.h>
#include <SimpleComms/UnixClient.h>

#include <atomic>
#include <chrono>
#include <cmath>
#include <condition_variable>
#include <fstream>
#include <iostream>
#include <memory>
#include <mutex>
#include <sstream>
#include <string>
#include <thread>
#include <unordered_map>
#include <vector>

struct receiveDescriptor
{
    std::string name;
    std::string channel;
    size_t queueSize;
};

struct receivePair
{
    std::shared_ptr<SimpleComms::UnixClient> receiver;
    std::string channel;
    std::string name;
};

int main(int argc, char** argv)
{
    if (argc != 2) {
        std::cerr << "Usge: scsRx [path_to_config_file]" << std::endl;
        return 1;
    }

    std::vector<receiveDescriptor> receiverDescs;
    std::ifstream infile(argv[1]);
    std::string line;
    while (std::getline(infile, line)) {
        std::istringstream ss(line);

        std::string name;
        std::getline(ss, name, ',');
        std::string channel;
        std::getline(ss, channel, ',');
        std::string depth;
        std::getline(ss, depth, ',');

        receiveDescriptor rd = {name, channel, static_cast<size_t>(atoi(depth.c_str()))};
        receiverDescs.push_back(rd);
    }

    // Build the receivers
    std::vector<receivePair> receivePairs;
    for (auto const& desc : receiverDescs) {
        receivePair rp;

        rp.receiver =
            std::make_shared<SimpleComms::UnixClient>(SimpleComms::UnixClient::getDefaultScsPath(), desc.name);
        rp.channel = desc.channel;
        rp.name = desc.name;

        rp.receiver->declare(desc.channel, SimpleComms::ChannelType::Subscriber, static_cast<uint32_t>(desc.queueSize));

        receivePairs.push_back(rp);

        if (receivePairs.back().receiver->connect() != SimpleComms::Status::Ok) {
            std::cerr << "Failed to connect to scs!" << std::endl;
            return 1;
        }
    }

    // Shared variables across threads
    std::atomic<bool> done(false);

    // And the command channel
    SimpleComms::UnixClient command(SimpleComms::UnixClient::getDefaultScsPath(), "command-rx");
    if (command.connect() != SimpleComms::Status::Ok) {
        std::cerr << "Failed to connect to scs!" << std::endl;
        return 1;
    }

    command.declare("command", SimpleComms::ChannelType::Subscriber, 10);

    std::thread commandThread([&done, &command]() {
        while (!done) {
            if (command.hasMsg()) {
                std::string msg;
                MessageInfo msgInfo;
                if (command.receiveMsg("command", msg, msgInfo)) {
                    std::cout << "Relaying command message..." << std::endl;
                    command.sendMsg("command-relay", msg);
                }
            }
        }
    });

    std::vector<std::thread> receiveThreads;
    for (auto& receivePair : receivePairs) {
        receiveThreads.push_back(std::thread([&done, &receivePair]() {
            std::string msg;
            MessageInfo msgInfo;
            while (!done) {
                if (receivePair.receiver->hasMsg()) {
                    while (receivePair.receiver->receiveMsg(receivePair.channel, msg, msgInfo)) {
                        // Bounce back
                        receivePair.receiver->sendMsg(receivePair.channel + "-relay", msg);
                    }
                }
            }
        }));
    }

    for (auto& t : receiveThreads) {
        t.join();
    }

    // Clean up
    commandThread.join();

    for (auto& receivePair : receivePairs) {
        receivePair.receiver->disconnect();
    }

    return 0;
}
