/*!
***********************************************************
* @file SparseMap.cc
* @brief Source file of Sparse Map.
*
* Space-efficient way to store feature map data, like
* radar map, lidar map, and lidar intensity map.
*
* @section License
*
* Copyright (c) 2017- Delphi Automotive. All rights
* reserved.
*
* @section changeLogs Change Logs
*
* @author Sida Wang (sida.wang@delphi.com)
* @date 08/31/2017
* @remark
* Initial implementation.
*
*
***********************************************************
*/

#include <SparseMap/SparseMap.h>

#include "gridMapCore/GridMap.hpp"
#include "lib/avlog/avlog.hpp"

#include <algorithm>
#include <cereal/archives/json.hpp>
#include <fstream>
#include <nlohmann/json.hpp>
#include <set>
#include <sstream> // stringstream

using namespace std;

using av::CoordinateFrameVariant;

namespace grid_map {
const double thres = 0.00001;

vector<string> getPairs(string s)
{
    vector<string> pair_list;
    string s2;
    stringstream ss;
    ss.str(s);
    while (std::getline(ss, s2, '&')) {
        pair_list.emplace_back(s2);
    }
    return pair_list;
}

bool notSpaceLine(string s)
{
    for (const auto& c : s) {
        if (!isspace(c)) {
            return true;
        }
    }
    return false;
}

int getRowNum(string& s)
{
    string::size_type pos = s.find('#');
    if (pos != std::string::npos) {
        string ans = s.substr(0, pos);
        s = s.substr(pos + 1);
        return stoi(ans);
    } else {
        return stoi(s);
    }
}

/* Searches the vector "a" for the value "target".
   The vector "a" must be sorted and must not contain duplicate elements.
   If target is in a, returns its index;
   Otherwise, returns the index of the smallest element in a that is larger than target.
   Specially, if target is larger than the largest element in a, returns a.size(). */

int binarySearch(const vector<int>& a, int target)
{
    size_t lo = 0, hi = a.size();
    while (hi > lo) {
        auto mi = lo + (hi - lo) / 2;
        if (a[mi] == target)
            return static_cast<int>(mi);
        if (a[mi] < target)
            lo = mi + 1;
        else
            hi = mi;
    }
    return static_cast<int>(lo);
}

SparseMapParams readSparseMapConfig(const std::string& configFilePath)
{
    // check EPSG
    std::ifstream config_stream(configFilePath);
    nlohmann::json config;
    config_stream >> config;
    int epsg = -1;
    if (config.find("EPSG") != config.end()) {
        epsg = config["EPSG"];
    } else {
        AVLOG_WARN("Could not find EPSG");
    }
    struct SparseMapParams sparseMapPara(epsg);
    std::ifstream SparseMapStructure_json_ifstream(configFilePath);
    cereal::JSONInputArchive archive(SparseMapStructure_json_ifstream);
    archive(sparseMapPara.centerX_m_,
            sparseMapPara.centerY_m_,
            sparseMapPara.mapSizeX_m_,
            sparseMapPara.mapSizeY_m_,
            sparseMapPara.mapCellSize_m_);
    AVLOG_INFO("EPSG parameter is: {}", sparseMapPara.epsg_);
    if (epsg > 0) {
        // flip to NED
        std::swap(sparseMapPara.centerX_m_, sparseMapPara.centerY_m_);
        std::swap(sparseMapPara.mapSizeX_m_, sparseMapPara.mapSizeY_m_);
        sparseMapPara.coord_system_ = CoordinateFrameVariant::NED;
    }
    return sparseMapPara;
}

void writeSparseMapConfig(const SparseMapParams& spmParams, const std::string& configFile)
{
    std::ofstream os(configFile.c_str());
    cereal::JSONOutputArchive archive(os);
    archive(CEREAL_NVP(spmParams.centerX_m_),
            CEREAL_NVP(spmParams.centerY_m_),
            CEREAL_NVP(spmParams.mapSizeX_m_),
            CEREAL_NVP(spmParams.mapSizeY_m_),
            CEREAL_NVP(spmParams.mapCellSize_m_));
}

bool SparseMap::checkBoundary(const RecPoint2D& inputPt)
{
    double inputPtX = inputPt.x, inputPtY = inputPt.y;
    double globalMapMinX = spmPara_.centerX_m_ - spmPara_.mapSizeX_m_ / 2;
    double globalMapMaxX = spmPara_.centerX_m_ + spmPara_.mapSizeX_m_ / 2;
    double globalMapMinY = spmPara_.centerY_m_ - spmPara_.mapSizeY_m_ / 2;
    double globalMapMaxY = spmPara_.centerY_m_ + spmPara_.mapSizeY_m_ / 2;
    return inputPtX >= globalMapMinX && inputPtX <= globalMapMaxX && inputPtY >= globalMapMinY &&
           inputPtY <= globalMapMaxY;
}

bool SparseMap::checkBoundary(const Index& index)
{
    double globalMapCellSize = spmPara_.mapCellSize_m_;

    int size_X = int(spmPara_.mapSizeX_m_ / globalMapCellSize);
    int size_Y = int(spmPara_.mapSizeY_m_ / globalMapCellSize);

    return index(0) >= 0 && index(0) < size_X && index(1) >= 0 && index(1) < size_Y;
}

CoordinateFrameVariant SparseMap::getPlaneCoordinate() const
{
    return spmPara_.coord_system_;
}

bool SparseMap::getVal(const RecPoint2D& inputPt, float& val)
{
    if (!checkBoundary(inputPt)) {
        return false;
    }
    double inputPtX = inputPt.x, inputPtY = inputPt.y;

    double globalMapCellSize = spmPara_.mapCellSize_m_;

    int globalMapInputPtIndX = static_cast<int>(spmPara_.mapSizeX_m_ / spmPara_.mapCellSize_m_ / 2);
    int globalMapInputPtIndY = static_cast<int>(spmPara_.mapSizeY_m_ / spmPara_.mapCellSize_m_ / 2);

    int subMapInputPtIndX =
        globalMapInputPtIndX + static_cast<int>((spmPara_.centerX_m_ - inputPtX) / globalMapCellSize);
    int subMapInputPtIndY =
        globalMapInputPtIndY + static_cast<int>((spmPara_.centerY_m_ - inputPtY) / globalMapCellSize);

    return getVal(Index(subMapInputPtIndX, subMapInputPtIndY), val);
}

bool checkSubGridMapBound(double globalMapMinX,
                          double globalMapMaxX,
                          double globalMapMinY,
                          double globalMapMaxY,
                          double subMapCenterPosX,
                          double subMapCenterPosY)
{
    return subMapCenterPosX <= globalMapMaxX && subMapCenterPosX >= globalMapMinX &&
           subMapCenterPosY <= globalMapMaxY && subMapCenterPosY >= globalMapMinY;
}

bool SparseMap::getSubMap(Matrix& mat, const Position& position, const Length& length, Index& refIndex)
{
    /* Get map x and y direction information in meter */
    double globalMapMinX = spmPara_.centerX_m_ - spmPara_.mapSizeX_m_ / 2;
    double globalMapMaxX = spmPara_.centerX_m_ + spmPara_.mapSizeX_m_ / 2;
    double globalMapMinY = spmPara_.centerY_m_ - spmPara_.mapSizeY_m_ / 2;
    double globalMapMaxY = spmPara_.centerY_m_ + spmPara_.mapSizeY_m_ / 2;

    double subMapCenterPosX = position.x(), subMapCenterPosY = position.y();
    double subMapSizeX = length.x(), subMapSizeY = length.y();

    /* Check if the requested part map area is out of bound or not */
    if (!checkSubGridMapBound(
            globalMapMinX, globalMapMaxX, globalMapMinY, globalMapMaxY, subMapCenterPosX, subMapCenterPosY)) {
        return false;
    }

    double globalMapCellSize = spmPara_.mapCellSize_m_;

    /* Get part map index information */
    int subMapCellNumX = int(subMapSizeX / globalMapCellSize) + 1;
    int subMapCellNumY = int(subMapSizeY / globalMapCellSize) + 1;

    int size_X = int(spmPara_.mapSizeX_m_ / globalMapCellSize);
    int size_Y = int(spmPara_.mapSizeY_m_ / globalMapCellSize);

    int globalMapCenterIndX = static_cast<int>(spmPara_.mapSizeX_m_ / spmPara_.mapCellSize_m_ / 2);
    int globalMapCenterIndY = static_cast<int>(spmPara_.mapSizeY_m_ / spmPara_.mapCellSize_m_ / 2);

    int subMapCenterIndX =
        globalMapCenterIndX + static_cast<int>((spmPara_.centerX_m_ - subMapCenterPosX) / globalMapCellSize);
    int subMapCenterIndY =
        globalMapCenterIndY + static_cast<int>((spmPara_.centerY_m_ - subMapCenterPosY) / globalMapCellSize);

    int submapIndMinX = 0, submapIndMinY = 0;

    /* Check the cell number of the part map is odd or even */
    auto fl_num = static_cast<float>(abs(subMapCenterPosX - spmPara_.centerX_m_) / globalMapCellSize);
    int int_num = static_cast<int>(fl_num);
    float rem = fl_num - static_cast<float>(int_num);

    if (abs(rem) < 0.00001f) {
        submapIndMinX = subMapCenterIndX - static_cast<int>(subMapCellNumX / 2);
    } else {
        if (subMapCellNumX % 2 == 0) {
            submapIndMinX = subMapCenterIndX - static_cast<int>(subMapCellNumX / 2) + 1;
        } else {
            submapIndMinX = subMapCenterIndX - static_cast<int>(subMapCellNumX / 2);
        }
    }
    fl_num = static_cast<float>(abs(subMapCenterPosY - spmPara_.centerY_m_) / globalMapCellSize);
    int_num = static_cast<int>(fl_num);
    rem = fl_num - static_cast<float>(int_num);

    if (abs(rem) < 0.00001f) {
        submapIndMinY = subMapCenterIndY - static_cast<int>(subMapCellNumY / 2);
    } else {
        if (subMapCellNumY % 2 == 0) {
            submapIndMinY = subMapCenterIndY - static_cast<int>(subMapCellNumY / 2) + 1;
        } else {
            submapIndMinY = subMapCenterIndY - static_cast<int>(subMapCellNumY / 2);
        }
    }

    int submapIndMaxX = submapIndMinX + subMapCellNumX - 1, submapIndMaxY = submapIndMinY + subMapCellNumY - 1;

    /* Get the sub map index in the global map */
    submapIndMinX = max(submapIndMinX, 0);
    submapIndMinY = max(submapIndMinY, 0);
    submapIndMaxX = min(submapIndMaxX, size_X - 1);
    submapIndMaxY = min(submapIndMaxY, size_Y - 1);

    refIndex = Index(submapIndMinX, submapIndMinY);

    subMapCellNumX = submapIndMaxX - submapIndMinX + 1;
    subMapCellNumY = submapIndMaxY - submapIndMinY + 1;

    mat.resize(subMapCellNumX, subMapCellNumY);
    mat.setConstant(0.0);
    auto& data_v = globalSparseMap_.SparseMapData;
    auto& index_v = globalSparseMap_.SparseMapIndex;
    auto& rowInd = globalSparseMap_.rowIndex;

    int x_lo = binarySearch(rowInd, submapIndMinX);
    int x_hi = binarySearch(rowInd, submapIndMaxX + 1);

    for (int xx = x_lo; xx < x_hi; ++xx) {
        int x = rowInd[static_cast<size_t>(xx)];

        auto& data_vec = data_v[x];
        auto& index_vec = index_v[x];

        int ind_lo = binarySearch(index_vec, submapIndMinY);
        int ind_hi = binarySearch(index_vec, submapIndMaxY + 1);

        for (int i = ind_lo; i < ind_hi; ++i) {
            mat(x - submapIndMinX, index_vec[static_cast<size_t>(i)] - submapIndMinY) =
                data_vec[static_cast<size_t>(i)];
        }
    }

    /* Deal with the situation when size of part map is 0 */
    if (mat.rows() == 0 || mat.cols() == 0) {
        return false;
    }
    return true;
}

bool SparseMap::addPoint(const RecPoint2D& inputPt, const float val)
{
    if (!checkBoundary(inputPt)) {
        return false;
    }
    double inputPtX = inputPt.x, inputPtY = inputPt.y;

    double globalMapCellSize = spmPara_.mapCellSize_m_;

    auto globalMapInputPtIndX = spmPara_.mapSizeX_m_ / spmPara_.mapCellSize_m_ / 2;
    auto globalMapInputPtIndY = spmPara_.mapSizeY_m_ / spmPara_.mapCellSize_m_ / 2;

    auto subMapInputPtIndX = globalMapInputPtIndX + (spmPara_.centerX_m_ - inputPtX) / globalMapCellSize;
    auto subMapInputPtIndY = globalMapInputPtIndY + (spmPara_.centerY_m_ - inputPtY) / globalMapCellSize;

    return addPoint(Index(subMapInputPtIndX, subMapInputPtIndY), val);
}

bool SparseMap::addPoint(const Index& index, const float val)
{
    if (!checkBoundary(index)) {
        return false;
    }

    int x = index(0), y = index(1);
    vector<float>& data_vec = (globalSparseMap_.SparseMapData)[x];
    vector<int>& ind_vec = (globalSparseMap_.SparseMapIndex)[x];
    vector<int>& row_ind = globalSparseMap_.rowIndex;

    if (!binary_search(row_ind.begin(), row_ind.end(), x)) {
        int ind = binarySearch(row_ind, x);
        row_ind.insert(row_ind.begin() + ind, x);
    }
    int ind = binarySearch(ind_vec, y);
    if (!binary_search(ind_vec.begin(), ind_vec.end(), y)) {
        ind_vec.insert(ind_vec.begin() + ind, y);
        data_vec.insert(data_vec.begin() + ind, val);
    } else {
        data_vec[static_cast<size_t>(ind)] = val;
    }

    return true;
}

bool SparseMap::getSubGridMap(GridMap& gridMap,
                              const string& layer,
                              const Position& position,
                              const Length& length,
                              Index& refIndex)
{
    gridMap.setPlaneCoordinate(getPlaneCoordinate());
    Matrix mat;
    /* Get map x and y direction information in meter */
    double globalMapMinX = spmPara_.centerX_m_ - spmPara_.mapSizeX_m_ / 2;
    double globalMapMaxX = spmPara_.centerX_m_ + spmPara_.mapSizeX_m_ / 2;
    double globalMapMinY = spmPara_.centerY_m_ - spmPara_.mapSizeY_m_ / 2;
    double globalMapMaxY = spmPara_.centerY_m_ + spmPara_.mapSizeY_m_ / 2;

    double subMapCenterPosX = position.x(), subMapCenterPosY = position.y();
    double subMapSizeX = length.x(), subMapSizeY = length.y();

    /* Check if the requested part map area is out of bound or not */
    if (!checkSubGridMapBound(
            globalMapMinX, globalMapMaxX, globalMapMinY, globalMapMaxY, subMapCenterPosX, subMapCenterPosY)) {
        gridMap = GridMap(gridMap.getLayers());
        return false;
    }

    double globalMapCellSize = spmPara_.mapCellSize_m_;

    /* Get part map index information */
    int subMapCellNumX = static_cast<int>(subMapSizeX / globalMapCellSize) + 1;
    int subMapCellNumY = static_cast<int>(subMapSizeY / globalMapCellSize) + 1;

    int size_X = static_cast<int>(spmPara_.mapSizeX_m_ / globalMapCellSize);
    int size_Y = static_cast<int>(spmPara_.mapSizeY_m_ / globalMapCellSize);

    int globalMapCenterIndX = static_cast<int>(spmPara_.mapSizeX_m_ / spmPara_.mapCellSize_m_ / 2);
    int globalMapCenterIndY = static_cast<int>(spmPara_.mapSizeY_m_ / spmPara_.mapCellSize_m_ / 2);

    int subMapCenterIndX =
        globalMapCenterIndX + static_cast<int>((spmPara_.centerX_m_ - subMapCenterPosX) / globalMapCellSize);
    int subMapCenterIndY =
        globalMapCenterIndY + static_cast<int>((spmPara_.centerY_m_ - subMapCenterPosY) / globalMapCellSize);

    int submapIndMinX = 0, submapIndMinY = 0;

    /* Check the cell number of the part map is odd or even */
    auto fl_num = static_cast<float>(abs(subMapCenterPosX - spmPara_.centerX_m_) / globalMapCellSize);
    int int_num = static_cast<int>(fl_num);
    float rem = fl_num - static_cast<float>(int_num);

    if (abs(rem) < 0.00001f) {
        submapIndMinX = subMapCenterIndX - static_cast<int>(subMapCellNumX / 2);
    } else {
        if (subMapCellNumX % 2 == 0) {
            submapIndMinX = subMapCenterIndX - static_cast<int>(subMapCellNumX / 2) + 1;
        } else {
            submapIndMinX = subMapCenterIndX - static_cast<int>(subMapCellNumX / 2);
        }
    }
    fl_num = static_cast<float>(abs(subMapCenterPosY - spmPara_.centerY_m_) / globalMapCellSize);
    int_num = static_cast<int>(fl_num);
    rem = fl_num - static_cast<float>(int_num);

    if (abs(rem) < 0.00001f) {
        submapIndMinY = subMapCenterIndY - static_cast<int>(subMapCellNumY / 2);
    } else {
        if (subMapCellNumY % 2 == 0) {
            submapIndMinY = subMapCenterIndY - static_cast<int>(subMapCellNumY / 2) + 1;
        } else {
            submapIndMinY = subMapCenterIndY - static_cast<int>(subMapCellNumY / 2);
        }
    }

    int submapIndMaxX = submapIndMinX + subMapCellNumX - 1, submapIndMaxY = submapIndMinY + subMapCellNumY - 1;

    /* Get the sub map index in the global map */
    submapIndMinX = max(submapIndMinX, 0);
    submapIndMinY = max(submapIndMinY, 0);
    submapIndMaxX = min(submapIndMaxX, size_X - 1);
    submapIndMaxY = min(submapIndMaxY, size_Y - 1);

    refIndex = Index(submapIndMinX, submapIndMinY);

    subMapCellNumX = submapIndMaxX - submapIndMinX + 1;
    subMapCellNumY = submapIndMaxY - submapIndMinY + 1;

    mat.resize(subMapCellNumX, subMapCellNumY);
    mat.setConstant(0.0);
    auto& data_v = globalSparseMap_.SparseMapData;
    auto& index_v = globalSparseMap_.SparseMapIndex;
    auto& rowInd = globalSparseMap_.rowIndex;

    int x_lo = binarySearch(rowInd, submapIndMinX);
    int x_hi = binarySearch(rowInd, submapIndMaxX + 1);

    for (int xx = x_lo; xx < x_hi; ++xx) {
        int x = rowInd[static_cast<size_t>(xx)];

        auto& data_vec = data_v[x];
        auto& index_vec = index_v[x];

        int ind_lo = binarySearch(index_vec, submapIndMinY);
        int ind_hi = binarySearch(index_vec, submapIndMaxY + 1);

        for (int i = ind_lo; i < ind_hi; ++i) {
            mat(x - submapIndMinX, index_vec[static_cast<size_t>(i)] - submapIndMinY) =
                data_vec[static_cast<size_t>(i)];
        }
    }

    /* Deal with the situation when size of part map is 0 */
    if (mat.rows() == 0 || mat.cols() == 0) {
        vector<string> layers;
        layers.emplace_back(layer);
        gridMap = GridMap(gridMap.getLayers());
        return false;
    }

    GridMap helper;
    helper.setGeometry(Length(spmPara_.mapSizeX_m_, spmPara_.mapSizeY_m_),
                       spmPara_.mapCellSize_m_,
                       Position(spmPara_.centerX_m_, spmPara_.centerY_m_));
    bool isSuccess;
    helper = helper.getSubmap(position, Length(1, 1), isSuccess);

    gridMap.setGeometry(Length(static_cast<double>(mat.rows()) * spmPara_.mapCellSize_m_,
                               static_cast<double>(mat.cols()) * spmPara_.mapCellSize_m_),
                        spmPara_.mapCellSize_m_,
                        Position(helper.getPosition().x(), helper.getPosition().y()));
    gridMap.add(layer, mat);

    return true;
}

bool SparseMap::getVal(const Index& index, float& val)
{
    if (!checkBoundary(index)) {
        return false;
    }

    int x = index(0), y = index(1);
    auto& data_v = (globalSparseMap_.SparseMapData)[x];
    auto& index_v = (globalSparseMap_.SparseMapIndex)[x];
    if (!binary_search(index_v.begin(), index_v.end(), y)) {
        val = 0;
    } else {
        int ind = binarySearch(index_v, y);
        val = data_v[static_cast<size_t>(ind)];
    }
    return true;
}

bool SparseMap::getIndex(const RecPoint2D& inputPt, Index& index)
{
    if (!checkBoundary(inputPt)) {
        return false;
    }
    double inputPtX = inputPt.x, inputPtY = inputPt.y;

    double globalMapCellSize = spmPara_.mapCellSize_m_;

    auto globalMapInputPtIndX = spmPara_.mapSizeX_m_ / spmPara_.mapCellSize_m_ / 2;
    auto globalMapInputPtIndY = spmPara_.mapSizeY_m_ / spmPara_.mapCellSize_m_ / 2;

    auto subMapInputPtIndX = globalMapInputPtIndX + (spmPara_.centerX_m_ - inputPtX) / globalMapCellSize;
    auto subMapInputPtIndY = globalMapInputPtIndY + (spmPara_.centerY_m_ - inputPtY) / globalMapCellSize;

    index = Index(subMapInputPtIndX, subMapInputPtIndY);
    return true;
}

bool SparseMap::getPosition(const Index& index, RecPoint2D& outputPt)
{
    if (!checkBoundary(index)) {
        return false;
    }

    int indX = index(0);
    int indY = index(1);

    auto globalMapInputPtIndX = spmPara_.mapSizeX_m_ / spmPara_.mapCellSize_m_ / 2;
    auto globalMapInputPtIndY = spmPara_.mapSizeY_m_ / spmPara_.mapCellSize_m_ / 2;

    double ptLocX = -((indX - globalMapInputPtIndX) * spmPara_.mapCellSize_m_ - spmPara_.centerX_m_);
    double ptLocY = -((indY - globalMapInputPtIndY) * spmPara_.mapCellSize_m_ - spmPara_.centerY_m_);

    outputPt.x = ptLocX;
    outputPt.y = ptLocY;

    return true;
}

/*
 * Construct an empty sparse map.
 */
SparseMap::SparseMap(const SparseMapParams& spmParas)
{
    setSparseMapConfig(spmParas);
}

pair<int, float> SparseMap::parsePair(const string& s)
{
    string::size_type pos = s.find(',');
    return make_pair(stoi(s.substr(0, pos)), stod(s.substr(pos + 1)));
}

void SparseMap::verifyCoordinate(const CoordinateFrameVariant& target_coordinate)
{
    if (target_coordinate != spmPara_.coord_system_) {
        std::stringstream ss;
        ss << "Coordinate Dismatch! \n SparseMap Coordinate is " << static_cast<int>(spmPara_.coord_system_)
           << " target is " << static_cast<int>(target_coordinate);
        throw std::runtime_error(ss.str());
    }
}

void SparseMap::setSparseMapConfig(SparseMapParams spmParas)
{
    spmPara_ = spmParas;
    globalSparseMap_.coord_system_ = spmParas.coord_system_;
}

SparseMapParams SparseMap::getSparseMapConfig()
{
    return spmPara_;
}

SparseMapStructure SparseMap::getGlobalSparseMap()
{
    return globalSparseMap_;
}

void SparseMap::setGlobalSparseMap(SparseMapStructure sparseMap)
{
    verifyCoordinate(sparseMap.coord_system_);
    globalSparseMap_ = sparseMap;
}

SparseMap::SparseMap(GridMap& gridMap, const string& layer, const SparseMapParams& spmParas)
{
    Matrix mat = gridMap[layer];
    setSparseMapConfig(spmParas);
    auto row = mat.rows(), col = mat.cols();
    for (int i = 0; i < row; ++i) {
        for (int j = 0; j < col; ++j) {
            if (static_cast<double>(mat(i, j)) > thres) {
                (globalSparseMap_.SparseMapData)[i].push_back(mat(i, j)); // [i].push_back(make_pair(j, mat(i, j))) ;
                (globalSparseMap_.SparseMapIndex)[i].push_back(j);        // [i].push_back(make_pair(j, mat(i, j))) ;
                if ((globalSparseMap_.rowIndex).empty() || (globalSparseMap_.rowIndex).back() != i)
                    (globalSparseMap_.rowIndex).push_back(i);
            }
        }
    }
}

SparseMap::SparseMap(const string& sparseMapFile, const SparseMapParams& spmParas)
{
    setSparseMapConfig(spmParas);
    string tempSparseMapFilePath = sparseMapFile;
    try {
        std::ifstream mapStream(tempSparseMapFilePath.c_str());
        if (mapStream.fail()) {
            AVLOG_ERROR("Unable to open {}", tempSparseMapFilePath);
            exit(1);
        }
        string gridmapString;
        getline(mapStream, gridmapString, '\0');
        stringstream ss(gridmapString);
        string line;
        if (spmParas.epsg_ < 0) {
            while (getline(ss, line, '\n')) {
                if (notSpaceLine(line)) {
                    int rowNum = getRowNum(line);
                    vector<string> pairs = getPairs(line);
                    for (const auto& ele : pairs) {
                        pair<int, float> p = parsePair(ele);
                        // globalSparseMap_[rowNum].push_back(make_pair(p.first, p.second)); // [p.first] = p.second;
                        (globalSparseMap_.SparseMapData)[rowNum].push_back(p.second);
                        (globalSparseMap_.SparseMapIndex)[rowNum].push_back(p.first);
                        if ((globalSparseMap_.rowIndex).empty() || (globalSparseMap_.rowIndex).back() != rowNum)
                            (globalSparseMap_.rowIndex).push_back(rowNum);
                    }
                }
            }
        } else {
            // flip x,y to NED
            std::set<int> rowIndexSet;
            while (getline(ss, line, '\n')) {
                if (notSpaceLine(line)) {
                    int colNum = getRowNum(line);
                    vector<string> pairs = getPairs(line); // row:weight
                    for (const auto& ele : pairs) {
                        pair<int, float> p = parsePair(ele);
                        int rowNum = p.first;
                        (globalSparseMap_.SparseMapData)[rowNum].push_back(p.second);
                        (globalSparseMap_.SparseMapIndex)[rowNum].push_back(colNum);
                        rowIndexSet.insert(rowNum);
                    }
                }
            }
            for (int r : rowIndexSet) {
                globalSparseMap_.rowIndex.push_back(r);
            }
            globalSparseMap_.coord_system_ = CoordinateFrameVariant::NED;
        }
    } catch (const std::runtime_error& e) {
        AVLOG_ERROR("Unable to open {}. Error Message:{}", tempSparseMapFilePath, e.what());
        exit(1);
    }
}

void SparseMap::writeSparseMapToFile(const string& sparseMapFile, const string& configFile)
{
    try {
        std::ofstream mapStream;
        mapStream.open(sparseMapFile);
        for (const auto& iter1 : globalSparseMap_.SparseMapData) {
            mapStream << to_string(iter1.first) << "#";
            auto& data_v = iter1.second;
            auto& index_v = (globalSparseMap_.SparseMapIndex)[iter1.first];
            for (unsigned i = 0; i < data_v.size(); ++i) {
                mapStream << to_string(index_v[i]) << "," << to_string(data_v[i]) << "&";
            }
            mapStream << "\n";
        }
        mapStream.close();
    } catch (const std::runtime_error& e) {
        AVLOG_ERROR("Unable to save map to {}. Error Message:{}", sparseMapFile, e.what());
        exit(1);
    }

    writeSparseMapConfig(spmPara_, configFile);
}

void SparseMap::getGridMap(GridMap& gridMap, const string& layer)
{
    Matrix mat;
    double mapSizeX = spmPara_.mapSizeX_m_, mapSizeY = spmPara_.mapSizeY_m_;
    double globalMapCellSize = spmPara_.mapCellSize_m_;
    mat.resize(int(mapSizeX / globalMapCellSize), static_cast<int>(mapSizeY / globalMapCellSize));
    mat.setConstant(0.0);
    for (const auto& iter1 : globalSparseMap_.SparseMapData) {
        auto& data_v = iter1.second;
        auto& ind_v = (globalSparseMap_.SparseMapIndex)[iter1.first];
        for (unsigned i = 0; i < data_v.size(); ++i) {
            mat(iter1.first, ind_v[i]) = data_v[i];
        }
    }
    if (mat.rows() == 0 || mat.cols() == 0) {
        gridMap = GridMap(gridMap.getLayers());
        return;
    }
    gridMap.setGeometry(Length(spmPara_.mapSizeX_m_, spmPara_.mapSizeY_m_),
                        spmPara_.mapCellSize_m_,
                        Position(spmPara_.centerX_m_, spmPara_.centerY_m_));
    gridMap.add(layer, mat);
}
} // namespace grid_map
