/*!
***********************************************************
* @file SparseMap.h
* @brief Header file of Sparse Map.
*
* Space-efficient way to store feature map data, like
* radar map, lidar map, and lidar intensity map.
*
* @section License
*
* Copyright (c) 2017- Delphi Automotive. All rights
* reserved.
*
* @section changeLogs Change Logs
*
* @author Sida Wang (sida.wang@delphi.com)
* @date 07/31/2017
* @remark
* Initial implementation.
*
*
***********************************************************
*/

#ifndef _SPARSEMAP_H_
#define _SPARSEMAP_H_

#include "gridMapCore/TypeDefs.hpp"
#include <avutils/CoordinateFrameVariant.hpp>

#include <recgeometry/recGeometry.h>
#include <string>
#include <utility>
#include <vector>

namespace grid_map {
// Forward decl
class GridMap;

struct SparseMapStructure
{
    SparseMapStructure() {}
    std::map<int, std::vector<float>> SparseMapData;
    std::map<int, std::vector<int>> SparseMapIndex;
    std::vector<int> rowIndex;
    av::CoordinateFrameVariant coord_system_ = av::CoordinateFrameVariant::NED;
};

/*!
 * @struct SparseMapParams
 *
 * @brief Parameters to define a sparse map.
 *
 * The following parameters are used to define a global
 * sparse map: mapSizeX_m_, mapSizeY_m_, mapCellSize_m_,
 * centerX_m_, centerY_m_.
 * epsg_ coordinate reference system http://www.epsg.org/. For SG, epsg is 3414; Default value is -1, means UTM
 * Coordinate System;
 */
struct SparseMapParams
{
    SparseMapParams(const int epsg = -1)
        : mapSizeX_m_(0)
        , mapSizeY_m_(0)
        , mapCellSize_m_(0)
        , centerX_m_(0)
        , centerY_m_(0)
        , epsg_{epsg}
    {
        coord_system_ = (epsg_ > 0) ? av::CoordinateFrameVariant::NED : av::CoordinateFrameVariant::ENU;
    }
    SparseMapParams(double mapSizeX_m,
                    double mapSizeY_m,
                    double mapCellSize_m,
                    double centerX_m,
                    double centerY_m,
                    const int epsg = -1)
        : mapSizeX_m_(mapSizeX_m)
        , mapSizeY_m_(mapSizeY_m)
        , mapCellSize_m_(mapCellSize_m)
        , centerX_m_(centerX_m)
        , centerY_m_(centerY_m)
        , epsg_{epsg}
    {
        coord_system_ = (epsg_ > 0) ? av::CoordinateFrameVariant::NED : av::CoordinateFrameVariant::ENU;
    }
    SparseMapParams(const SparseMapParams& src)
    {
        mapSizeX_m_ = src.mapSizeX_m_;
        mapSizeY_m_ = src.mapSizeY_m_;
        mapCellSize_m_ = src.mapCellSize_m_;
        centerX_m_ = src.centerX_m_;
        centerY_m_ = src.centerY_m_;
        epsg_ = src.epsg_;
        coord_system_ = src.coord_system_;
    }
    SparseMapParams& operator=(const SparseMapParams& that)
    {
        mapSizeX_m_ = that.mapSizeX_m_;
        mapSizeY_m_ = that.mapSizeY_m_;
        mapCellSize_m_ = that.mapCellSize_m_;
        centerX_m_ = that.centerX_m_;
        centerY_m_ = that.centerY_m_;
        epsg_ = that.epsg_;
        coord_system_ = that.coord_system_;

        return *this;
    }
    double mapSizeX_m_;
    double mapSizeY_m_;
    double mapCellSize_m_;
    double centerX_m_;
    double centerY_m_;
    int epsg_ = -1;
    av::CoordinateFrameVariant coord_system_ = av::CoordinateFrameVariant::NED;
};

SparseMapParams readSparseMapConfig(const std::string& configFilePath);
void writeSparseMapConfig(const SparseMapParams& spmParams, const std::string& configFile);

/*!
 * @struct SparseMap
 *
 * @brief SparseMap class
 *
 * SparseMap class is a new data structure based on
 * hashmap to be more space efficient to store the grid map
 * based feature map data. globalSparseMap_ is pre-computed
 * static global map, and main functionality of the
 * SparseMap class is getSubGridMap function, which returns
 * submap of the grid map data structure.
 * Furthermore, the contents of sparse map will be stored in
 * an txt file. For each line, before # is the x direction
 * index, and followed by pairs of y direction index as well
 * as the value of the point, separated by ,. And the end of
 * the line is &. For example, if the contents in the sparse
 * map txt file is:
 * 0#2,2.460000&7,1.270000&8,1.320000&
 * 2#2,1.200000&4,5.600000&
 * which represents the points of:
 * (0, 2): 2.460000, (0, 7): 1.270000, (0, 8): 1.320000
 * (2, 2): 1.200000, (2, 4): 5.60000
 *
 */

class SparseMap
{
public:
    SparseMap(){};

    SparseMap(const SparseMapParams& spmPara);
    /*!
     * @brief Constructor with the path of sparse map text file
     * and sparse map parameter struct.
     *
     * @param sparseMapFile The path of the sparse map text file.
     * @param spmPara is the parameters to define a sparse map.
     *
     */
    SparseMap(const std::string& sparseMapFile, const SparseMapParams& spmPara);

    /*!
     * @brief Constructor with grid map, layer name, and sparse map parameter struct.
     *
     * @param gridMap Input grid map.
     * @param layer Layer of the grid map to be compressed.
     * @param spmPara Parameters to define a sparse map.
     *
     */
    SparseMap(GridMap& gridMap, const std::string& layer, const SparseMapParams& spmPara);

    /*!
     * @brief Get the whole grid map.
     *
     * @param gridMap Input grid map.
     * @param layer Layer of the grid map to be compressed.
     *
     */

    void setSparseMapConfig(SparseMapParams spmPara);

    /*!
     * @brief Get sparse map configuration.
     *
     * @return spmPara Sparse map parameters used to config the
     * sparse map.
     *
     */
    SparseMapParams getSparseMapConfig();

    /*!
     * @brief Set sparse map data with given data.
     *
     * @param sparseMap Sparse map data used to set the
     * data of the sparse map.
     *
     */
    void setGlobalSparseMap(SparseMapStructure sparseMap);

    /*!
     * @brief Get sparse map data.
     *
     * @return SparseMapSructure Global sparse map data.
     *
     */
    SparseMapStructure getGlobalSparseMap();
    void writeSparseMapToFile(const std::string& sparseMapFile, const std::string& configFile);
    void getGridMap(GridMap& gridMap, const std::string& layer);
    bool getSubGridMap(GridMap& gridMap,
                       const std::string& layer,
                       const Position& position,
                       const Length& length,
                       Index& refIndex);
    bool getSubMap(Matrix& mat, const Position& position, const Length& length, Index& refIndex);
    bool getIndex(const RecPoint2D& inputPt, Index& index);
    bool getPosition(const Index& index, RecPoint2D& outputPt);
    bool getVal(const RecPoint2D& input, float& val);
    bool addPoint(const Index& index, const float val);
    bool addPoint(const RecPoint2D& inputPt, const float val);

    /*!
     * @brief Get the value of the given point.
     *
     * @param index the index of the sparse map.
     * @param val the required value of the given point.
     *
     * @return true if index in the boundary of sparse map.
     */
    bool getVal(const Index& index, float& val);

    bool checkBoundary(const Index& index);

    /*!
     * @brief check the boundary of given point.
     *
     * @param inputPt input point to be checked.
     *
     * @return true if input point is in the boundary of sparse map.
     */
    bool checkBoundary(const RecPoint2D& inputPt);

    av::CoordinateFrameVariant getPlaneCoordinate() const;

private:
    std::pair<int, float> parsePair(const std::string& s);
    void verifyCoordinate(const av::CoordinateFrameVariant& target_coordinate);

    SparseMapStructure globalSparseMap_; /// global sparse map data
    SparseMapParams spmPara_;            /// global sparse map configuration parameters
};
} // namespace grid_map
#endif // _SPARSEMAP_H_
