/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Implements Welford's algorithm as a SeriesStatisticsInterface
 */

#ifndef COMBINATION_STATISTIC_HPP
#define COMBINATION_STATISTIC_HPP

#include <Statistics/SeriesStatisticsInterface.hpp>
#include <memory>
#include <vector>

namespace statistics {

/**
 * \brief   Allows one object to collect multiple types of statistics on the same series of data
 */

template<typename T>
class CombinationStatistic : public SeriesStatisticsInterface<T>
{
public:
    /**
     * \brief Default constructor
     */
    CombinationStatistic() = default;

    /**
     * \brief Add a SeriesStatisticsInterface to the collector set
     *
     * \param[in] collector: A unique_ptr to a SeriesStatisticsInterface to include in the results
     */
    void addCollector(std::unique_ptr<SeriesStatisticsInterface<T>> collector)
    {
        m_collectors.emplace_back(std::move(collector));
    }

    /**
     * \brief Updates internal state based on the addition of a new value.
     * \param[in] value: Value to add.
     * \return True if all collectors report success, else False.
     */
    bool addSample(const T& value)
    {
        bool ret{true};

        for (auto& collector : m_collectors) {
            ret &= collector->addSample(value);
        }

        return ret;
    }

    uint64_t getCount() const override
    {
        if (m_collectors.size() > 0) {
            return m_collectors[0]->getCount();
        } else {
            return 0;
        }
    }

    /**
     * \brief Returns JSON representation of the statistics
     * \note Standard deviation is not given, as it can be calculated from the variance.
     *
     * \return Statistics for each collector added to the object.
     */
    nlohmann::json getJSON() const
    {
        nlohmann::json result{};

        for (auto& collector : m_collectors) {
            result.push_back(collector->getJSON());
        }

        return result;
    }

    virtual ~CombinationStatistic() = default;

protected:
    std::vector<std::unique_ptr<SeriesStatisticsInterface<T>>> m_collectors{};
};
#endif
} // namespace statistics
