/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Implements Welford's algorithm as a SeriesStatisticsInterface for determining mean and variance
 */

#ifndef MEAN_VARIANCE_HPP
#define MEAN_VARIANCE_HPP

#include <Statistics/SeriesStatisticsInterface.hpp>

#include <optional>

namespace statistics {

/**
 * \brief   Calculates the mean and variance of the samples added using Welford's Algorithm. Also keeps count of the
 * samples.
 */

template<typename T>
class MeanVariance : public SeriesStatisticsInterface<T>
{
public:
    static_assert(std::is_floating_point<T>::value, "MeanVariance<T>: T must be floating point :-(");

    /**
     * \brief Default constructor
     */
    MeanVariance() = default;

    /**
     * \brief Updates internal state based on the addition of a new value.
     * \param[in] value: Value to add.
     * \return true
     */
    bool addSample(const T& value)
    {
        ++m_count;
        const T oldmean{m_mean};
        m_mean += (value - m_mean) / m_count;
        m_intermediate += ((value - m_mean) * (value - oldmean));

        if (m_previous.has_value()) {
            const T gap{value - (*m_previous)};
            m_mean_gap += (gap - m_mean_gap) / (m_count - 1);
        } else {
            m_mean_gap = 0;
        }
        m_previous = value;

        return true;
    }

    /**
     * \brief Returns number of samples added
     * \return Number of samples added
     */
    uint64_t getCount() const override { return m_count; }

    /**
     * \brief Returns variance of samples added
     * \return Optional variance of data
     */
    std::optional<T> getVariance() const
    {
        std::optional<T> ret{};
        if (m_count == 1) {
            ret = 0;
        } else if (m_count > 1) {
            ret = (m_intermediate / (m_count - 1));
        }

        return ret;
    }

    std::optional<T> getJitter() const
    {
        if (m_previous.has_value()) {
            return m_mean_gap;
        } else {
            return std::nullopt;
        }
    }

    /**
     * \brief Returns standard deviation of samples added (sqrt of variance)
     * \return Optional standard deviation of samples added
     */
    std::optional<T> getStandardDeviation() const
    {
        std::optional<T> ret{getVariance()};

        if (ret.has_value()) {
            ret = sqrt(*ret);
        }

        return ret;
    }

    /**
     * \brief Returns mean of samples added
     * \return optional mean of data
     */
    std::optional<T> getMean() const { return (m_count == 0) ? std::optional<T>{} : m_mean; }

    /**
     * \brief Returns JSON representation of the statistics
     * \note Standard deviation is not given, as it can be calculated from the variance.
     *
     * \return Count, mean, variance in a JSON object.
     */
    nlohmann::json getJSON() const
    {
        nlohmann::json result{};
        result["meanvariance"]["count"] = m_count;

        auto mean = getMean();
        auto var = getVariance();

        result["meanvariance"]["mean"] = mean.has_value() ? nlohmann::json(*mean) : nlohmann::json{};
        result["meanvariance"]["variance"] = var.has_value() ? nlohmann::json(*var) : nlohmann::json{};

        return result;
    }

protected:
    uint64_t m_count{0};           ///< Number of data samples added
    T m_intermediate{0};           ///< Intermediate number for Welford's calculation
    T m_mean{0};                   ///< Mean of data added
    std::optional<T> m_previous{}; /// Latest sample.
    T m_mean_gap{0};               /// Avg. diff. between samples.
};
} // namespace statistics
#endif
