/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Implements a SeriesStatisticsInterface identifying the highest and lowest values seen
 */

#ifndef MIN_MAX_HPP
#define MIN_MAX_HPP

#include <Statistics/SeriesStatisticsInterface.hpp>

#include <limits>
#include <optional>

namespace statistics {

/**
 * \brief Tracks Min/Max values seen over time
 */

template<typename T>
class MinMax : public SeriesStatisticsInterface<T>
{
public:
    /**
     * \brief Default construction
     */
    MinMax() = default;

    /**
     * \brief Updates internal state based on the addition of a new value.
     * \param[in] value: Value to add.
     * \return true
     */
    bool addSample(const T& value)
    {
        m_min = std::min(m_min, value);
        m_max = std::max(m_max, value);
        ++m_count;
        return true;
    }

    uint64_t getCount() const override { return m_count; }

    /**
     * \brief Returns max of samples added
     * \return max of data
     */
    std::optional<T> getMax() const { return hasData() ? m_max : std::optional<T>{}; }

    /**
     * \brief Returns max of samples added
     * \return max of data
     */
    std::optional<T> getMin() const { return hasData() ? m_min : std::optional<T>{}; }

    /**
     * \brief Returns JSON representation of the statistics
     * \note Standard deviation is not given, as it can be calculated from the variance.
     *
     * \return Count, mean, variance in a JSON object.
     */
    nlohmann::json getJSON() const
    {
        nlohmann::json result{};
        result["minmax"]["min"] = hasData() ? nlohmann::json(m_min) : nlohmann::json();
        result["minmax"]["max"] = hasData() ? nlohmann::json(m_max) : nlohmann::json();

        return result;
    }

protected:
    /**
     * \brief Returns true if we have seen data
     *
     * \return true if valid data has been passed
     */
    inline bool hasData() const
    {
        return (m_min != std::numeric_limits<T>::max() || m_max != std::numeric_limits<T>::lowest());
    }

    T m_min{std::numeric_limits<T>::max()};    ///< Lowest value observed
    T m_max{std::numeric_limits<T>::lowest()}; ///< Highest value observed
    uint64_t m_count;
};
} // namespace statistics
#endif
