/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Defines a class for collecting windowed statistics.
 */
#pragma once

#include <iostream>

#include <Statistics/SeriesStatisticsInterface.hpp>
#include <avtime/AvClock.hpp>

#include <chrono>
#include <deque>
#include <memory>

namespace statistics {

/// @brief A template class which computes windowed statistics for a
/// given SeriesStatisticsInterface subclass.
///
/// This class is not thread-safe; it is intended to be used by a thread-
/// safe caller.
///
/// @param T The type of datapoints and statistic results.
/// @param S A subclass of SeriesStatisticsInterface<T>.
/// @param C The clock used.
template<typename T,
         typename S,
         typename C,
         // Make sure S is a subclass of SeriesStatisticsInterface<T>.
         std::enable_if_t<std::is_base_of<SeriesStatisticsInterface<T>, S>::value, bool> = false>
class TimeWindowedSeriesStatistics
{
public:
    /// Convenience names.
    using Time = typename C::time_point;
    using Duration = typename C::duration;
    using Callback = std::function<void(const S&, Time)>;

    /// @brief A window of stats. Uses its own S.
    class Window
    {
    public:
        /// @brief Constructor.
        ///
        /// @param start The start time of the window.
        Window(Time start)
            : series_(new S())
            , start_(start)
        {}

        /// Copy/move constructors/assignment operators.
        Window(const Window&) = default;
        Window(Window&&) = default;
        Window& operator=(const Window&) = default;
        Window& operator=(Window&&) = default;

        /// @brief Get the underlying S.
        S& getSeries() { return *series_; }

        /// @brief Get the underlying S.
        const S& getSeries() const { return *series_; }

        /// @brief Get the start time of the window.
        Time getStart() const { return start_; }

    private:
        /// @brief The underlying S.
        std::unique_ptr<S> series_;

        /// @brief The start time of the window.
        Time start_;
    };

    /// @brief Constructor.
    ///
    /// @param callback Invoked whenever a window is complete.
    ///
    /// @param intervalSize The interval between the start of each window.
    ///
    /// @param numWindows The number of windows. Each window is
    /// intervalSize * numWindows long.
    ///
    /// @param startTime The start time of the first window.
    TimeWindowedSeriesStatistics(Callback callback, Duration intervalSize, size_t numWindows, av::TimePoint startTime)
        : callback_(callback)
        , intervalSize_(intervalSize)
        , numWindows_(numWindows)
        , totalWindowSize_(intervalSize_ * numWindows_)
        , startTime_(startTime)
    {
        for (size_t i = 0; i < numWindows_; ++i) {
            windows_.emplace_back(startTime_ + i * intervalSize);
        }
    }

    /// @brief Add the given datapoint to the active windows.
    void addSample(const T& sample, Time t)
    {
        if (updateWindows(t)) {
            for (Window& window : windows_) {
                if ((window.getStart() <= t) && (t < window.getStart() + totalWindowSize_)) {
                    window.getSeries().addSample(sample);
                }
            }
        }
    }

    /// @brief For testing: invoke the callback for each active window.
    void callInterimCallbacks()
    {
        for (const auto& w : windows_) {
            callback_(w.getSeries(), w.getStart());
        }
    }

private:
    /// @brief Check which windows have fallen out of range; invoke
    /// the callback for them.
    ///
    /// @param t The timestamp of an incoming datapoint.
    bool updateWindows(Time t)
    {
        // If the timestamp is earlier than the start of the first
        // active window, do nothing, and tell the caller not to
        // bother updating any of the windows.
        if (t < windows_.front().getStart()) {
            return false;
        }

        // Drop windows (and invoke their callbacks, if they're
        // nonempty) for which t is later than the end of the window.
        if (t >= (windows_.front().getStart() + totalWindowSize_)) {
            size_t numDropped{0};
            for (const auto& w : windows_) {
                if (t >= w.getStart() + totalWindowSize_) {
                    if (w.getSeries().getCount() > 0) {
                        callback_(w.getSeries(), w.getStart());
                    }
                    ++numDropped;
                }
            }

            if (numDropped > 0) {
                windows_.erase(windows_.begin(), windows_.begin() + numDropped);
            }
        }

        if (windows_.empty() || (t >= (windows_.back().getStart() + totalWindowSize_))) {
            const Duration timeSinceStart = t - startTime_;
            uint64_t windowsSinceStart = timeSinceStart / intervalSize_;
            if (windowsSinceStart * intervalSize_ < timeSinceStart) {
                ++windowsSinceStart;
            }

            const Time windowStart{startTime_ + windowsSinceStart * intervalSize_};
            windows_.emplace_back(windowStart);
        }

        return true;
    }

    /// @brief The callback to invoke when a window is ready.
    const Callback callback_;

    /// @brief The interval between the start of each window.
    const Duration intervalSize_;

    /// @brief The number of windows.
    const size_t numWindows_;

    /// @brief The total duration of active windows: intervalSize_ *
    /// numWindows_.
    const Duration totalWindowSize_;

    /// @brief The start of the first window.
    const Time startTime_;

    /// @brief A queue of windows.
    std::deque<Window> windows_{};
};
} // namespace statistics
