/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief A type-safe alternative to CombinationStatistic.
 */

#ifndef TUPLE_STATISTICS_HPP
#define TUPLE_STATISTICS_HPP

#include <Statistics/SeriesStatisticsInterface.hpp>
#include <memory>
#include <tuple>

namespace statistics {

/// @brief A SeriesStatisticsInterface that aggregates stats, type-safely.
///
/// @param SeriesPlural The series (series-es, if that were a word)
/// collecting stats. All must have the same underlying T. The T of
/// the first series in the list is taken as the T for this series.
template<typename... SeriesPlural>
class TupleStatistics
    : public SeriesStatisticsInterface<typename std::tuple_element<0, std::tuple<SeriesPlural...>>::type::value_type>
{
public:
    /// @brief Constructor.
    TupleStatistics() = default;

    /// @brief Add a sample to each underlying series.
    ///
    /// @param True iff every underlying addSample() call returns true.
    bool addSample(const typename std::tuple_element<0, std::tuple<SeriesPlural...>>::type::value_type& value) override
    {
        return addSampleHelper<0>(value);
    }

    uint64_t getCount() const override { return get<0>().getCount(); }

    /// @brief Fill out a JSON structure aggregating the JSON for the
    /// underlying series.
    ///
    /// @return The resulting JSON.
    nlohmann::json getJSON() const override
    {
        nlohmann::json res;
        appendJSON<0>(res);
        return res;
    }

    /// @brief Get the ith underlying series.
    ///
    /// @return A reference to the ith underlying series.
    template<size_t i>
    typename std::tuple_element<i, std::tuple<SeriesPlural...>>::type& get()
    {
        return std::get<i>(series_);
    }

    /// @brief Get the ith underlying series.
    ///
    /// @return A const reference to the ith underlying series.
    template<size_t i>
    const typename std::tuple_element<i, std::tuple<SeriesPlural...>>::type& get() const
    {
        return std::get<i>(series_);
    }

private:
    /// @brief Helper template method that adds the sample to each
    /// series, starting with the ith.
    ///
    /// @value The datapoint being added.
    template<size_t i>
    bool addSampleHelper(const typename std::tuple_element<0, std::tuple<SeriesPlural...>>::type::value_type& value)
    {
        /// Constexpr so that we can expand into a finite number of calls.
        if constexpr (i < std::tuple_size<decltype(series_)>{}) {
            if (!get<i>().addSample(value)) {
                return false;
            }
            return addSampleHelper<i + 1>(value);
        } else {
            return true;
        }
    }

    /// @brief Helper template method that stores the JSON for each
    /// series, starting with the ith, into res.
    ///
    /// @param res The JSON that's being filled out.
    template<size_t i>
    void appendJSON(nlohmann::json& res) const
    {
        if constexpr (i < std::tuple_size<decltype(series_)>{}) {
            nlohmann::json j = get<i>().getJSON();
            res.push_back(j);
            appendJSON<i + 1>(res);
        }
    }

    /// The underlying series.
    std::tuple<SeriesPlural...> series_;
};

} // namespace statistics
#endif // TUPLE_STATISTICS_HPP
