/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Tests for TupleStatistics.
 */

#include <Statistics/MeanVariance.hpp>
#include <Statistics/MinMax.hpp>
#include <Statistics/TupleStatistics.hpp>
#include <gtest/gtest.h>

namespace statistics {
namespace {

template<typename T>
class Total : public SeriesStatisticsInterface<T>
{
public:
    Total() = default;

    bool addSample(const T& value)
    {
        total_ += value;
        return true;
    }

    T getTotal() const { return total_; }

    uint64_t getCount() const { return count_; }

    nlohmann::json getJSON() const
    {
        nlohmann::json result{};
        result["total"]["total"] = total_;
        return result;
    }

protected:
    T total_{0};
    uint64_t count_{0};
};

using TupleStat1 = TupleStatistics<MeanVariance<double>>;
using TupleStat2 = TupleStatistics<MeanVariance<double>, MinMax<double>>;
using TupleStat3 = TupleStatistics<MeanVariance<double>, MinMax<double>, Total<double>>;

#define EXPECT_OPTIONAL_DOUBLE_EQ(opt, val)                                                                            \
    {                                                                                                                  \
        const auto optCache{opt};                                                                                      \
        EXPECT_TRUE(optCache);                                                                                         \
        if (optCache) {                                                                                                \
            EXPECT_DOUBLE_EQ(*optCache, (val));                                                                        \
        }                                                                                                              \
    }

TEST(Tuple1Tests, TestInitialConstruction)
{
    TupleStat1 ts;

    EXPECT_EQ(ts.get<0>().getCount(), 0);
    EXPECT_FALSE(ts.get<0>().getMean().has_value());
    EXPECT_FALSE(ts.get<0>().getVariance().has_value());

    const nlohmann::json j = ts.getJSON();

    const nlohmann::json meanvar = j[0].at("meanvariance");
    EXPECT_EQ(meanvar.at("count"), 0);
    EXPECT_TRUE(meanvar.at("mean").is_null());
    EXPECT_TRUE(meanvar.at("variance").is_null());
}

TEST(Tuple1Tests, TestSingleUpdate)
{
    TupleStat1 ts;

    EXPECT_TRUE(ts.addSample(1.5));

    EXPECT_EQ(ts.get<0>().getCount(), 1);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getMean(), 1.5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getVariance(), 0);

    nlohmann::json j = ts.getJSON();
    const nlohmann::json meanvar = j[0].at("meanvariance");
    EXPECT_EQ(meanvar.at("count"), 1);
    EXPECT_DOUBLE_EQ(meanvar.at("mean"), 1.5);
    EXPECT_EQ(meanvar.at("variance"), 0);
}

TEST(Tuple1Tests, TestMoreUpdates)
{
    TupleStat1 ts;

    EXPECT_TRUE(ts.addSample(1.0));
    EXPECT_TRUE(ts.addSample(2.0));
    EXPECT_TRUE(ts.addSample(3.0));
    EXPECT_TRUE(ts.addSample(5.0));
    EXPECT_TRUE(ts.addSample(4.0));

    EXPECT_EQ(ts.get<0>().getCount(), 5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getMean(), 3);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getStandardDeviation(), sqrt(2.5));
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getVariance(), 2.5);

    nlohmann::json j = ts.getJSON();
    const nlohmann::json meanvar = j[0].at("meanvariance");
    EXPECT_EQ(meanvar.at("count"), 5);
    EXPECT_EQ(meanvar.at("mean"), 3);
    EXPECT_DOUBLE_EQ(meanvar.at("variance"), 2.5);
}

TEST(Tuple2Tests, TestInitialConstruction)
{
    TupleStat2 ts;

    EXPECT_EQ(ts.get<0>().getCount(), 0);
    EXPECT_FALSE(ts.get<0>().getMean().has_value());
    EXPECT_FALSE(ts.get<0>().getVariance().has_value());
    EXPECT_FALSE(ts.get<1>().getMin().has_value());
    EXPECT_FALSE(ts.get<1>().getMax().has_value());

    const nlohmann::json j = ts.getJSON();

    const nlohmann::json meanvar = j[0].at("meanvariance");
    EXPECT_EQ(meanvar.at("count"), 0);
    EXPECT_TRUE(meanvar.at("mean").is_null());
    EXPECT_TRUE(meanvar.at("variance").is_null());

    const nlohmann::json minmax = j[1].at("minmax");
    EXPECT_TRUE(minmax.at("min").is_null());
    EXPECT_TRUE(minmax.at("max").is_null());
}

TEST(Tuple2Tests, TestSingleUpdate)
{
    TupleStat2 ts;

    EXPECT_TRUE(ts.addSample(1.5));

    EXPECT_EQ(ts.get<0>().getCount(), 1);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getMean(), 1.5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getVariance(), 0);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<1>().getMin(), 1.5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<1>().getMax(), 1.5);

    nlohmann::json j = ts.getJSON();
    const nlohmann::json meanvar = j[0].at("meanvariance");
    EXPECT_EQ(meanvar.at("count"), 1);
    EXPECT_DOUBLE_EQ(meanvar.at("mean"), 1.5);
    EXPECT_EQ(meanvar.at("variance"), 0);

    const nlohmann::json minmax = j[1].at("minmax");
    EXPECT_DOUBLE_EQ(minmax.at("min"), 1.5);
    EXPECT_DOUBLE_EQ(minmax.at("max"), 1.5);
}

TEST(Tuple2Tests, TestMoreUpdates)
{
    TupleStat2 ts;

    EXPECT_TRUE(ts.addSample(1.0));
    EXPECT_TRUE(ts.addSample(2.0));
    EXPECT_TRUE(ts.addSample(3.0));
    EXPECT_TRUE(ts.addSample(5.0));
    EXPECT_TRUE(ts.addSample(4.0));

    EXPECT_EQ(ts.get<0>().getCount(), 5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getMean(), 3);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getStandardDeviation(), sqrt(2.5));
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getVariance(), 2.5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<1>().getMin(), 1);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<1>().getMax(), 5);

    nlohmann::json j = ts.getJSON();
    const nlohmann::json meanvar = j[0].at("meanvariance");
    EXPECT_EQ(meanvar.at("count"), 5);
    EXPECT_EQ(meanvar.at("mean"), 3);
    EXPECT_DOUBLE_EQ(meanvar.at("variance"), 2.5);

    const nlohmann::json minmax = j[1].at("minmax");
    EXPECT_DOUBLE_EQ(minmax.at("min"), 1);
    EXPECT_DOUBLE_EQ(minmax.at("max"), 5);
}

TEST(Tuple3Tests, TestInitialConstruction)
{
    TupleStat3 ts;

    EXPECT_EQ(ts.get<0>().getCount(), 0);
    EXPECT_FALSE(ts.get<0>().getMean().has_value());
    EXPECT_FALSE(ts.get<0>().getVariance().has_value());
    EXPECT_FALSE(ts.get<1>().getMin().has_value());
    EXPECT_FALSE(ts.get<1>().getMax().has_value());
    EXPECT_EQ(ts.get<2>().getTotal(), 0);

    const nlohmann::json j = ts.getJSON();

    const nlohmann::json meanvar = j[0].at("meanvariance");
    EXPECT_EQ(meanvar.at("count"), 0);
    EXPECT_TRUE(meanvar.at("mean").is_null());
    EXPECT_TRUE(meanvar.at("variance").is_null());

    const nlohmann::json minmax = j[1].at("minmax");
    EXPECT_TRUE(minmax.at("min").is_null());
    EXPECT_TRUE(minmax.at("max").is_null());

    const nlohmann::json total = j[2].at("total");
    EXPECT_EQ(total.at("total"), 0);
}

TEST(Tuple3Tests, TestSingleUpdate)
{
    TupleStat3 ts;

    EXPECT_TRUE(ts.addSample(1.5));

    EXPECT_EQ(ts.get<0>().getCount(), 1);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getMean(), 1.5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getVariance(), 0);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<1>().getMin(), 1.5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<1>().getMax(), 1.5);
    EXPECT_DOUBLE_EQ(ts.get<2>().getTotal(), 1.5);

    nlohmann::json j = ts.getJSON();
    const nlohmann::json meanvar = j[0].at("meanvariance");
    EXPECT_EQ(meanvar.at("count"), 1);
    EXPECT_DOUBLE_EQ(meanvar.at("mean"), 1.5);
    EXPECT_EQ(meanvar.at("variance"), 0);

    const nlohmann::json minmax = j[1].at("minmax");
    EXPECT_DOUBLE_EQ(minmax.at("min"), 1.5);
    EXPECT_DOUBLE_EQ(minmax.at("max"), 1.5);

    const nlohmann::json total = j[2].at("total");
    EXPECT_EQ(total.at("total"), 1.5);
}

TEST(Tuple3Tests, TestMoreUpdates)
{
    TupleStat3 ts;

    EXPECT_TRUE(ts.addSample(1.0));
    EXPECT_TRUE(ts.addSample(2.0));
    EXPECT_TRUE(ts.addSample(3.0));
    EXPECT_TRUE(ts.addSample(5.0));
    EXPECT_TRUE(ts.addSample(4.0));

    EXPECT_EQ(ts.get<0>().getCount(), 5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getMean(), 3);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getStandardDeviation(), sqrt(2.5));
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<0>().getVariance(), 2.5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<1>().getMin(), 1);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<1>().getMax(), 5);
    EXPECT_OPTIONAL_DOUBLE_EQ(ts.get<1>().getMin(), 1);
    EXPECT_DOUBLE_EQ(ts.get<2>().getTotal(), 15);

    nlohmann::json j = ts.getJSON();
    const nlohmann::json meanvar = j[0].at("meanvariance");
    EXPECT_EQ(meanvar.at("count"), 5);
    EXPECT_EQ(meanvar.at("mean"), 3);
    EXPECT_DOUBLE_EQ(meanvar.at("variance"), 2.5);

    const nlohmann::json minmax = j[1].at("minmax");
    EXPECT_DOUBLE_EQ(minmax.at("min"), 1);
    EXPECT_DOUBLE_EQ(minmax.at("max"), 5);

    const nlohmann::json total = j[2].at("total");
    EXPECT_EQ(total.at("total"), 15);
}

} // namespace
} // namespace statistics
