
/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief unit tests for StringUtils related classes
 */

#include <StringUtils/StringUtils.h>
#include <gtest/gtest.h>

#include <cstdint>
#include <limits>
#include <stdexcept>
#include <string>
#include <vector>

namespace {
namespace StringUtilsTest {
namespace SplitStringTests {
using std::string;
using std::vector;
using string_utils::splitString;

TEST(StringUtils_splitString, EmptyString)
{
    auto expect = vector<string>{};
    auto result = splitString("", '\n');
    EXPECT_EQ(expect, result);
}

TEST(StringUtils_splitString, SingleCharNoDelimiter)
{
    auto expect = vector<string>{"a"};
    auto result = splitString("a", '\n');
    EXPECT_EQ(expect, result);
}

TEST(StringUtils_splitString, LonginputNoDelimiter)
{
    auto expect = vector<string>{"mary had a little lamb"};
    auto result = splitString("mary had a little lamb", '\n');
    EXPECT_EQ(expect, result);
}

TEST(StringUtils_splitString, SingleCharOnlyDelimiter)
{
    auto expect = vector<string>{"", ""};
    auto result = splitString("\n", '\n');
    EXPECT_EQ(expect, result);
}

TEST(StringUtils_splitString, MultiCharOnlyDelimiter)
{
    auto expect = vector<string>{"", "", "", ""};
    auto result = splitString("\n\n\n", '\n');
    EXPECT_EQ(expect, result);
}

TEST(StringUtils_splitString, LongInputWithDelimiters)
{
    auto expect = vector<string>{"mary", "had", "a", "little", "lamb"};
    auto result = splitString("mary had a little lamb", ' ');
    EXPECT_EQ(expect, result);
}

TEST(StringUtils_splitString, LongInputWithAltDelimiters)
{

    auto expect = vector<string>{"m", "ry h", "d ", " little l", "mb"};
    auto result = splitString("mary had a little lamb", 'a');
    EXPECT_EQ(expect, result);
}

TEST(StringUtils_splitString, BeginsWithDelimiter)
{
    auto expect = vector<string>{"", "abcd"};
    auto result = splitString("\nabcd", '\n');
    EXPECT_EQ(expect, result);
}

TEST(StringUtils_splitString, EndsWithDelimiter)
{
    auto expect = vector<string>{"abcd", ""};
    auto result = splitString("abcd\n", '\n');
    EXPECT_EQ(expect, result);
}
} // namespace SplitStringTests

namespace toShortTests {
using std::invalid_argument;
using std::numeric_limits;
using std::out_of_range;
using std::range_error;
using std::to_string;
using std::uint16_t;
using string_utils::toShort;

TEST(StringUtils_toShort, EmptyString)
{
    EXPECT_THROW(toShort(""), invalid_argument);
}

TEST(StringUtils_toShort, NonNumercic)
{
    EXPECT_THROW(toShort("abcd"), invalid_argument);
}

TEST(StringUtils_toShort, OutOfRangeTooLarge)
{
    auto val = 1ll + static_cast<long long>(numeric_limits<int>::max());
    EXPECT_THROW(toShort(to_string(val)), out_of_range);
}

TEST(StringUtils_toShort, OutOfRangeTooSmall)
{
    auto val = -1ll + static_cast<long long>(numeric_limits<int>::min());
    EXPECT_THROW(toShort(to_string(val)), out_of_range);
}

TEST(StringUtils_toShort, InRangeButNegativeValue)
{
    EXPECT_THROW(toShort("-1"), range_error);
}

TEST(StringUtils_toShort, InRangeButTooLargeForUINT16)
{
    auto val = 1 + static_cast<int>(numeric_limits<uint16_t>::max());
    EXPECT_THROW(toShort(to_string(val)), range_error);
}

TEST(StringUtils_toShort, InRangeSmallestValue)
{
    EXPECT_EQ(toShort("0"), numeric_limits<uint16_t>::min());
}

TEST(StringUtils_toShort, InRangeLargestValue)
{
    auto val = numeric_limits<uint16_t>::max();
    EXPECT_EQ(toShort(to_string(val)), numeric_limits<uint16_t>::max());
}

TEST(StringUtils_toShort, InRangeValidValue)
{
    uint16_t val = 100;
    EXPECT_EQ(toShort(to_string(val)), val);
}
} // namespace toShortTests
} // namespace StringUtilsTest
} // namespace
