/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Copy interface files from build tree for use in tests
 */

#ifndef TASK_FRAMEWORK_TESTUTILS_COPYINTERFACES_H
#define TASK_FRAMEWORK_TESTUTILS_COPYINTERFACES_H

#include <TestUtils/Common.hpp>
#include <TestUtils/EnvBackup.hpp>
#include <cstdlib>
#include <string>

namespace TestUtils {
/* \brief A class that provides a convenient interface to copy interface files needed for testing and configure the
 * environment to use them.
 */
class CopyInterfaces
{
public:
    /**
     * \brief Constructor. Will configure OTTO_DIR environment variable to point to the search path.
     *
     * \param[in] testname as string: Name of the test, directory to store libraries in
     */
    explicit CopyInterfaces(const std::string& testname)
        : intfSearchingPath{"./middleware/TestUtils/TestInterfaceLibs/" + testname}
    {}

    /**
     * \brief Default destructor
     */
    ~CopyInterfaces() = default;

    /**
     * \brief Move an interface from the build tree to an interfaces directory that matches the structure of the
     * installed interfaces directory.
     *
     * \param[in] source as string: The source directory of the interface, relative to
     * ".../aptivware-core/src/interfaces" \param[in] dest as string: The destination directory, relative to the
     * ".../interfaces" directory \param[in] intfLibName as string: The name of the shared library. Typically
     * "libChannel.so"
     *
     * \return True if successful copy.
     */
    bool copyInterface(const std::string& source,
                       const std::string& dest,
                       const std::string intfLibName = "libChannel.so")
    {
        const std::string expandedSource{intfLibPath + source};
        const std::string expandedDest{intfLibDestPath + dest};
        return copyFile(getTestingDataPath(expandedSource, true), getTestingDataPath(expandedDest, false), intfLibName);
    }

private:
    const static std::string intfLibPath; ///< The common path to all interface lib files
    const std::string intfSearchingPath;  ///< The path for seachring interface lib files
    const std::string intfLibDestPath{intfSearchingPath +
                                      "/lib/interfaces"};                ///< Common path to copied interface lib files
    EnvBackup libEnv{"OTTO_DIR", getTestingDataPath(intfSearchingPath)}; ///< Temporarily set OTTO_DIR
};

const std::string CopyInterfaces::intfLibPath{"./src/interfaces"};
} // namespace TestUtils

#endif // TASK_FRAMEWORK_TESTUTILS_COPYINTERFACES_H
