/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Temporarily configure environment variables for testing
 */

#ifndef TASK_FRAMEWORK_TESTUTILS_ENVBACKUP_H
#define TASK_FRAMEWORK_TESTUTILS_ENVBACKUP_H

#include <cstdlib>
#include <string>

namespace TestUtils {
/* \brief RII class to remove an environment variable during a scope.
 *
 * In some unit tests, we need the OTTO_CONFIG_DIR and
 * ACTIVE_CONFIG_DIR environment variables to be absent; when running
 * locally, they're generally set. So this class's constructor saves
 * and unsets the environment variable, if any, and its destructor
 * restores it.
 */
class EnvBackup
{
public:
    EnvBackup(const char* name, const std::string& value)
        : m_name(name)
    {
        const char* const oldValue = getenv(name);
        m_oldWasPresent = (oldValue != nullptr);
        m_oldValue = (oldValue == nullptr) ? "" : oldValue;

        if (value.empty()) {
            if (m_oldWasPresent) {
                unsetenv(m_name.c_str());
            }
        } else {
            setenv(m_name.c_str(), value.c_str(), true);
        }
    }

    explicit EnvBackup(const char* name)
        : EnvBackup(name, "")
    {}

    void update(const std::string& value) { setenv(m_name.c_str(), value.c_str(), true); }

    void unset() { unsetenv(m_name.c_str()); }

    ~EnvBackup()
    {
        if (m_oldWasPresent) {
            setenv(m_name.c_str(), m_oldValue.c_str(), true);
        } else {
            unsetenv(m_name.c_str());
        }
    }

private:
    std::string m_name;
    std::string m_oldValue;
    bool m_oldWasPresent;
};
} // namespace TestUtils

#endif // TASK_FRAMEWORK_TESTUTILS_ENVBACKUP_H
