/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional
 */

#pragma once

#include <experimental/filesystem>
#include <string>
#include <uuid/uuid.h>

namespace TestUtils {

class TestTmpDirectory
{
public:
    /**
     * @brief Establish a directory we can use for a unit test
     *
     * @note If Bazel has set a TEST_TMPDIR, use that. Otherwise, create a UUID-based dir in /tmp.
     */
    TestTmpDirectory()
    {
        char* envTestTmpPath = getenv("TEST_TMPDIR");
        if (envTestTmpPath == nullptr || envTestTmpPath[0] == '\0') {
            path = std::experimental::filesystem::path("/tmp/test_" + getUuid());
            std::experimental::filesystem::create_directory(path);
        } else {
            path = std::experimental::filesystem::path(envTestTmpPath);
            bazelManaged = true;
        }
    }

    /**
     * @brief If we had to create a directory, clean it up when we're done.
     */
    ~TestTmpDirectory()
    {
        if (!bazelManaged) {
            std::experimental::filesystem::remove_all(path);
        }
    }

    const std::experimental::filesystem::path& getPath() { return path; }

private:
    // 36 bytes for the UUID string representation, see the man page for uuid_unparse for more info
    static constexpr uint32_t TMP_PATH_UUID_SIZE = 36;

    /**
     * @brief Fetch a UUID string
     *
     * @return A new UUID
     */
    std::string getUuid() noexcept
    {
        uuid_t uuid;
        uuid_generate(uuid);

        // +1 for the NUL termination
        char uuid_cstr[TMP_PATH_UUID_SIZE + 1];
        uuid_unparse(uuid, uuid_cstr);

        return std::string(uuid_cstr);
    }

    bool bazelManaged{false};
    std::experimental::filesystem::path path;
};
} // namespace TestUtils