/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief unit tests for EnvBackup.
 */

#include <TestUtils/EnvBackup.hpp>
#include <gtest/gtest.h>
#include <unistd.h>

using TestUtils::EnvBackup;

TEST(EnvBackup, backUpExisting)
{
    const char name[] = "EnvBackupTest_backUpExisting";
    const char value1[] = "Keyboard";
    const char value2[] = "Mouse";

    setenv(name, value1, /*overwrite=*/false);

    // Make sure we actually set it to what we wanted to.
    {
        const char* const actual = getenv(name);
        ASSERT_NE(nullptr, actual);

        const std::string actualStr{actual};
        EXPECT_EQ(value1, actualStr);
    }

    // Override it and make sure it now has the right value.
    {
        EnvBackup eb(name, value2);
        const char* const actual = getenv(name);
        ASSERT_NE(nullptr, actual);

        const std::string actualStr{actual};
        EXPECT_EQ(value2, actualStr);
    }

    // Make sure that it was restored to the correct value.
    {
        const char* const actual = getenv(name);
        ASSERT_NE(nullptr, actual);

        const std::string actualStr{actual};
        EXPECT_EQ(value1, actualStr);
    }
}

TEST(EnvBackup, backUpNonexistent)
{
    const char name[] = "EnvBackupTest_backUpNonexistent";
    const char value[] = "Mouse";

    // Make sure it doesn't exist yet.
    {
        const char* const actual = getenv(name);
        ASSERT_EQ(nullptr, actual);
    }

    // Override it and make sure it now has the right value.
    {
        EnvBackup eb(name, value);
        const char* const actual = getenv(name);
        ASSERT_NE(nullptr, actual);

        const std::string actualStr{actual};
        EXPECT_EQ(value, actualStr);
    }

    // Make sure it doesn't exist now that it's been restored.
    {
        const char* const actual = getenv(name);
        ASSERT_EQ(nullptr, actual);
    }
}

TEST(EnvBackup, backUpErased)
{
    const char name[] = "EnvBackupTest_backUpErased";
    const char value[] = "Keyboard";

    setenv(name, value, /*overwrite=*/false);

    // Make sure we actually set it to what we wanted to.
    {
        const char* const actual = getenv(name);
        ASSERT_NE(nullptr, actual);

        const std::string actualStr{actual};
        EXPECT_EQ(value, actualStr);
    }

    // Override it and make sure it now doesn't exist.
    {
        EnvBackup eb(name, "");
        const char* const actual = getenv(name);
        ASSERT_EQ(nullptr, actual);
    }

    // Make sure that it was restored to the correct value.
    {
        const char* const actual = getenv(name);
        ASSERT_NE(nullptr, actual);

        const std::string actualStr{actual};
        EXPECT_EQ(value, actualStr);
    }
}

TEST(EnvBackup, backUpErasedNonexistent)
{
    const char name[] = "EnvBackupTest_backUpErasedNonexistent";

    // Make sure nobody else has set it yet.
    {
        const char* const actual = getenv(name);
        ASSERT_EQ(nullptr, actual);
    }

    // Override it and make sure it now doesn't exist.
    {
        EnvBackup eb(name, "");
        const char* const actual = getenv(name);
        ASSERT_EQ(nullptr, actual);
    }

    // Make sure it doesn't exist now that it's been restored.
    {
        const char* const actual = getenv(name);
        ASSERT_EQ(nullptr, actual);
    }
}

TEST(EnvBackup, backupUnsetUpdate)
{
    const char name[] = "EnvBackupTest_backUpErasedNonexistent";

    // Make sure nobody else has set it yet.
    {
        const char* const actual = getenv(name);
        ASSERT_EQ(nullptr, actual);
    }

    // Override it and make sure it now doesn't exist.
    {
        EnvBackup eb(name, "");
        const char* actual = getenv(name);
        ASSERT_EQ(nullptr, actual);

        eb.update("val");

        actual = getenv(name);
        ASSERT_STREQ(actual, "val");

        eb.unset();

        actual = getenv(name);
        ASSERT_EQ(nullptr, actual);
    }

    // Make sure it doesn't exist now that it's been restored.
    {
        const char* const actual = getenv(name);
        ASSERT_EQ(nullptr, actual);
    }
}
