/*!
 ***********************************************************
 * @file PThreadLocker.h
 * @brief A file defining the PThreadLocker class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * PThreadLocker.h
 * Tugrul Galatali
 * 08/24/2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */
#ifndef PTHREADLOCKER_H
#define PTHREADLOCKER_H

#include <pthread.h>

/*!
 * @class PThreadLocker
 *
 * @brief This class shall provide "automatic" unlocking of mutex object on loss of scope.
 * \n On construction, it locks the mutex object. On destruction, it unlocks the mutex object.
 * \n It guarantees that mutex object is properly unlocked.Specially useful for blocks with
 * \n multiple exit points and exception is thrown.
 *
 * \n It shall be modelled on QT's QMutexLocker, declare it in a scope where exclusive access of a resource is needed.
 * \n Upon exit from that scope, the variable is descoped. So the destructor is called and the mutex is unlocked.
 **/

class PThreadLocker
{
public:
    /*!
     * @brief Parameterized constructor
     * \n It shall lock a mutex.
     *
     * @param[in] mLock constant mutex pointer
     */
    explicit PThreadLocker(pthread_mutex_t& mutex)
        : mutex_(mutex)
    {
        pthread_mutex_lock(&mutex_);
    }

    PThreadLocker() = delete;
    PThreadLocker(const PThreadLocker&) = delete;
    PThreadLocker(const PThreadLocker&&) = delete;
    PThreadLocker& operator=(PThreadLocker const&) = delete;
    PThreadLocker& operator=(const PThreadLocker&&) = delete;

    /*!
     * @brief Destructor. It shall unlock the mutex
     *
     */
    virtual ~PThreadLocker() { pthread_mutex_unlock(&mutex_); }

private:
    pthread_mutex_t& mutex_;
};

#endif // PTHREADLOCKER_H
