/*!
 ***********************************************************
 * @file Stopwatch.h
 * @brief Implements a resettable stopwatch for measuring local time.
 *
 * @section License
 *
 * Copyright (c) 2020- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * * @author Matthew Ivester
 * @date 01/31/2020
 * @remark
 * Added.
 *
 ***********************************************************
 */

#ifndef OTIME_STOPWATCH_H
#define OTIME_STOPWATCH_H

#include <TimeStamp/otime.h>

namespace ottotime {
/*!
 * @class Stopwatch
 * @brief This class shall implement resettable stopwatch for measuring local time.
 * \n The underlying clock it uses can be steady clock(default) or high resolution clock.
 */
class Stopwatch final
{
public:
    /*!
     * @brief The default destructor
     */
    ~Stopwatch() = default;

    /*!
     * \enum UnderlyingClock
     * @brief enum representing the type of the clock
     */
    enum class UnderlyingClock : uint32_t
    {
        SteadyTimer,         ///< steady clock
        HighResolutionTimer, ///< high resolution clock
    };

    /*!
     * @brief Parameterized constructor.
     * \n It shall initialize the clock with specific clock type and starts the Stopwatch.
     *
     * @param[in] clock :  Which underlying clock to use.
     * \n Default is SteadyTimer
     * \n HighResolutionTimer, may have higher precision, but is not guaranteed to be steady over long periods of time
     */
    Stopwatch(UnderlyingClock clock = UnderlyingClock::SteadyTimer)
        : clockType_(clock)
    {
        startTime_ = getCurrentTime();
    }

    /*!
     * @brief This method shall return the start time of the stopwatch, as reported by the underlying clock.
     *
     * @return otimePoint : The time when Stopwatch is started (or) The last call to reset() method where Stopwatch's
     * start time is updated.
     */
    otimePoint getStartTime() const { return startTime_; }

    /*!
     * @brief This method shall return the elapsed time as measured by the underlying clock.
     *
     * @return otimeDuration : The amount of time elapsed since the Stopwatch is started (or) the last call to reset.
     */
    otimeDuration getElapsedTime() const { return getCurrentTime() - startTime_; }

    /*!
     * @brief This method shall reset the starting time of the stopwatch.
     *
     * @return void
     */
    void reset(void) { startTime_ = getCurrentTime(); }

private:
    /*!
     * @brief  This method shall be helper to get the current time as measured by the underlying clock.
     * \n IF UnderlyingClock is SteadyTimer THEN
     * \n &emsp; get the current time using the steady clock
     * \n ELSE
     * \n &emsp; get the current time using the high resolution clock
     * \n END IF
     * @return otimePoint object
     */
    otimePoint getCurrentTime() const
    {
        if (clockType_ == UnderlyingClock::SteadyTimer) {
            return otimePoint::buildCurrentSteadyTime();
        } else {
            return otimePoint::buildCurrentHighResolutionTime();
        }
    }

    otimePoint startTime_{}; ///< Start time of the Stopwatch
                             ///< It gets updated with current time when the Stopwatch instance is created.
    const UnderlyingClock clockType_{
        UnderlyingClock::SteadyTimer}; ///< Represents type of the clock that Stopwatch uses.
                                       ///< The clock type shall be steady clock/high resolution clock.
};                                     // end class Stopwatch

} // namespace ottotime

#endif // OTIME_STOPWATCH_H
