/*!
 ***********************************************************
 * @file TimeStamp.h
 * @brief A file implementing several functions using otime.
 *
 * @section License
 *
 * Copyright (c) 2016- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * * @author Matthew Ivester
 * @date 01/31/2020
 * These should largely be deprecated now in favor of the
 * methods on otimePoint and otimeDuration.
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 02/11/2016
 * Helper functions to use otime.
 * @n
 *
 ***********************************************************
 */

#ifndef TIMESTAMP_H
#define TIMESTAMP_H

#include <TimeStamp/otime.h>

using namespace ottotime;

/*!
 * @brief This method shall calculate the elapsed time between two time points say t1 and t2.
 * \n If t2 occurs  after t1, the result shall be  positive.
 *
 * @param[in] t1 : otimePoint object
 * @param[in] t2 : otimePoint object
 *
 * @return long double: the delta between t1 and t2,in fractional seconds with the calculation performed to microsecond
 * resolution.
 */
inline long double elapsedTimeInSeconds(const otimePoint& t1, const otimePoint& t2)
{

    return static_cast<long double>((t2 - t1).totalSecondsDouble());
}

/*!
 * @brief This method shall convert a given time  otimePoint to a double representation of the seconds,with nanosecond
 * accuracy
 *
 * @param[in] t1 : otimePoint object
 *
 * @return double representation of seconds
 */
inline long double otimeToSeconds(const otimePoint& t1)
{

    return static_cast<long double>(t1.toSecondsSinceEpochDouble());
}

/*!
 * @brief This method shall convert a given time as long double to otimePoint.
 *
 * @param[in] t : given time as long double
 *
 * @return otimePoint object
 */
inline otimePoint convertTootime(const long double t)
{
    return otimePoint::fromSecondsSinceEpochDouble(static_cast<double>(t));
}

/*!
 * @brief This method shall return otimeDuration object from given value seconds
 *
 * @param[in] secs : seconds as double
 *
 * @return otimeDuration object
 *
 */
inline otimeDuration secondsTooduration(const double secs)
{

    return otimeDuration{oSeconds::fromDouble(secs)};
}

/*!
 * @brief This method shall convert a given time in otimeDuration to a double representation of the seconds, with
 * nanosecond accuracy.
 *
 * @param[in] td : otimeDuration object
 *
 * @return seconds as double
 */
inline double odurationToSeconds(const otimeDuration& td)
{
    return td.totalSecondsDouble();
}

#endif // ifndef TIMESTAMP_H
