/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file otime.h
 * @product OttoKit
 * @version TODO
 * @released TODO
 * @updated Jul 11, 2018
 * @brief A file defining several classes representing time
 *
 * @section License
 *
 * Copyright (c) 2016- Delphi Automotive. All rights reserved.
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 ***********************************************************************************************************************
 */

#ifndef OTTOTIME_H
#define OTTOTIME_H

#include <cereal/access.hpp>
#include <cereal/cereal.hpp>
#include <cstdint>
#include <ctime>
#include <iostream>
#include <sys/time.h>

const long double OTIME_MAX = 9223372036.854775807;
const uint64_t OTIME_SECOND_MAX = 9223372036U;
const uint32_t MSEC_PER_SEC = 1000U;
const uint32_t USEC_PER_MSEC = 1000U;
const uint32_t NSEC_PER_USEC = 1000U;
const uint32_t NSEC_PER_MSEC = 1000000U;
const uint32_t USEC_PER_SEC = 1000000U;
const uint32_t NSEC_PER_SEC = 1000000000U;

namespace ottotime {

// TODO [Matt Ivester 2/12/2020]: ugly forward declares so I can keep ugly ABI compatibility...
class oSeconds;
class oMilliSeconds;
class oMicroSeconds;
class oNanoSeconds;

/*!
 * \struct otimeSpec
 * @brief This structure shall represent the duration of time using two int64_t values.
 * \n one for seconds and another for nanoseconds.
 */
struct otimeSpec
{
    /*!
     * @brief Default constructor
     * \n It shall initialize the seconds and nanoseconds of otimeSpec with -1.
     */
    otimeSpec()
        : tv_sec(-1)
        , tv_nsec(-1)
    {}

    /*!
     * @brief Default copy constructor
     *
     * @param[in] other : The otimeSpec object to copy
     */
    otimeSpec(const otimeSpec& other) = default;

    /*!
     * @brief Default move constructor
     *
     * @param[in] other : The otimeSpec object.
     */
    otimeSpec(otimeSpec&& other) = default;

    // TODO [Matt Ivester 2/11/2020]:
    // I'd *also* like this to be explicit, but I found some places doing:
    // ottotime::otimeSpec = {blah blah};
    // ...which isn't really intended but worked because the implicit
    // default ctor allowed it.  Once these are cleaned up we should
    // make this ctor explicit.

    /*!
     * @brief Parameterized constructor
     *
     * @param[in] secs as int64_t :  seconds
     * @param[in] nsecs int64_t :  nano seconds
     */
    otimeSpec(int64_t secs, int64_t nsecs)
        : tv_sec(secs)
        , tv_nsec(nsecs)
    {}

    /*!
     * @brief This method shall return time with seconds and nano seconds of otimeSpec initialized with 0.
     *
     * @return otimeSpec object
     */
    static otimeSpec getEpochTimeSpec() { return otimeSpec{0, 0}; }

    int64_t tv_sec;  ///< represents seconds part of otimeSpec
    int64_t tv_nsec; ///< represents nanoseconds part of otimeSpec

    /*!
     * @brief This method shall return true, if the time is valid.
     *
     * @return bool : true, if seconds and nanoseconds of otimeSpec not equals to -1. Otherwise, false.
     */
    bool isValidAsTimestamp(void) const { return tv_sec != -1 && tv_nsec != -1; }

    /*!
     * @brief This method shall return true, if the time is valid.
     *
     * @return bool : true, if seconds and nanoseconds of otimeSpec not equals to -1. Otherwise, false.
     */
    void invalidateAsTimestamp(void)
    {
        tv_sec = -1;
        tv_nsec = -1;
    }

    /*!
     * @brief This method shall normalize the time to make sure that nanoseconds stays within a second.
     * \n It shall convert nanoseconds of otimeSpec to equivalent seconds and adds to seconds of otimeSpec.
     * \n Ensures no negative nanoseconds in otimeSpec.
     *
     * @return bool : true, if seconds of otimeSpec is positive.otherwise, false
     *
     */
    bool normalize(void)
    {
        int64_t excess = tv_nsec / NSEC_PER_SEC;
        tv_nsec -= excess * NSEC_PER_SEC;
        tv_sec += excess;
        if (tv_nsec < 0) {
            tv_nsec += static_cast<int64_t>(NSEC_PER_SEC);
            --tv_sec;
        }

        return tv_sec >= 0;
    }

    /*!
     * @brief Default assignment operator (=) method
     * \n This method shall assign the values of source otimeSpec object 'other' into destination otimeSpec object.
     *
     * @param[in] other The reference of otimeSpec object
     *
     * @return otimeSpec object
     */
    otimeSpec& operator=(const otimeSpec& other) = default;

    /*!
     * @brief Default move assignment operator (=) method
     *
     * @param[in] other : rvalue reference of otimeSpec object
     *
     * @return otimeSpec object
     */
    otimeSpec& operator=(otimeSpec&& other) = default;

    /*!
     * @brief Overloaded is equals to (==) operator method
     * \n This method shall check if both lhs and rhs are equal.
     *
     * @param[in] lhs : The otimeSpec object
     * @param[in] rhs : The otimeSpec object
     *
     * @return bool : true, if both lhs and rhs are equal. false, otherwise
     */
    friend bool operator==(const otimeSpec& lhs, const otimeSpec& rhs);

    /*!
     * @brief Overloaded not equal to operator (!=) method.
     * \n This method shall check if both lhs and rhs are not equal.
     *
     * @param[in] lhs : The otimeSpec object
     * @param[in] rhs : The otimeSpec object
     *
     * @return bool : true, both lhs and rhs are not equal.false, otherwise.
     *
     */
    friend bool operator!=(const otimeSpec& lhs, const otimeSpec& rhs);

    /*!
     * @brief Overloaded less than operator (<) method.
     * \n This method shall check if lhs time is less than rhs time.
     * \n \b Operational \b Flow:
     * \n IF seconds of both lhs and rhs is same THEN
     * \n &emsp; if seconds of lhs is positive and nanoseconds of lhs is less than rhs then return true.otherwise false
     * \n &emsp; if seconds of lhs is negative and nanoseconds of lhs is greater than rhs return true.otherwise false
     * \n ELSE
     * \n &emsp; return true if seconds of lhs is less than rhs.otherwise false
     * \n END IF
     *
     * @param[in] lhs : The otimeSpec object
     * @param[in] rhs : The otimeSpec object
     *
     * @return bool : true, if lhs time is less than the rhs time.otherwise, false.
     */
    friend bool operator<(const otimeSpec& lhs, const otimeSpec& rhs);

    /*!
     * @brief Overloaded greater than operator (>) method.
     * \n This method shall check if lhs time is greater than rhs time.
     *
     * @param[in] lhs The otimeSpec object
     * @param[in] rhs The otimeSpec object
     *
     * @return bool : true, if lhs time is greater than the rhs time.otherwise, false.
     */
    friend bool operator>(const otimeSpec& lhs, const otimeSpec& rhs);

    /*!
     * @brief Overloaded less than or equals to operator (<=) method.
     * \n This method shall check if lhs time is less than than or equals to rhs time.
     *
     * @param[in] lhs : The otimeSpec object
     * @param[in] rhs : The otimeSpec object
     *
     * @return bool : true, if lhs time is less than or equals to the rhs time.otherwise, false.
     */
    friend bool operator<=(const otimeSpec& lhs, const otimeSpec& rhs);

    /*!
     * @brief Overloaded greater than or equals to operator (>=) method.
     * \n This method shall check if lhs time is greater than than or equals to rhs time.
     *
     * @param[in] lhs : The otimeSpec object
     * @param[in] rhs : The otimeSpec object
     *
     * @return bool : true, if lhs time is greater than or equals to the rhs time.otherwise, false
     */
    friend bool operator>=(const otimeSpec& lhs, const otimeSpec& rhs);

    /*!
     * @brief This method shall add given 'other' otimeSpec object (rhs) to the lhs otimeSpec object.
     * \n It shall apply normalization on lhs.
     *
     * @param[in] other : The otimeSpec object
     *
     * @return otimeSpec object
     */
    otimeSpec& operator+=(const otimeSpec& other);

    /*!
     * @brief  This method shall subtract given 'other' otimeSpec object (rhs) from the lhs otimeSpec object.
     * \n It shall apply normalization on lhs..
     *
     * @param[in] other : The otimeSpec object
     *
     * @return otimeSpec object
     */
    otimeSpec& operator-=(const otimeSpec& other);

    /*!
     * @brief  This method shall add time in otimeSpec(rhs) to the lhs time in otimeSpec.
     *
     * @param[in] lhs : The otimeSpec object
     * @param[in] rhs : The otimeSpec object
     *
     * @return otimeSpec object
     */
    friend otimeSpec operator+(const otimeSpec& lhs, const otimeSpec& rhs);

    /*!
     * @brief  This method shall subtract time in otimeSpec(rhs) from the lhs time in otimeSpec.
     *
     * @param[in] lhs : Time in otimeSpec
     * @param[in] rhs : Time in otimeSpec
     *
     * @return otimeSpec object
     */
    friend otimeSpec operator-(const otimeSpec& lhs, const otimeSpec& rhs);

    /*!
     * @brief Overloaded operator -
     * \n This method shall make otimeSpec as negative.
     * \n It shall make seconds and nanoseconds of otimeSpec as negative values.
     *
     * @return otimeSpec object
     */
    otimeSpec operator-(void) const;
};

inline bool operator==(const otimeSpec& lhs, const otimeSpec& rhs)
{
    return ((lhs.tv_sec == rhs.tv_sec) && (lhs.tv_nsec == rhs.tv_nsec));
}

inline bool operator!=(const otimeSpec& lhs, const otimeSpec& rhs)
{
    return !operator==(lhs, rhs);
}

inline bool operator<(const struct otimeSpec& lhs, const struct otimeSpec& rhs)
{
    if (lhs.tv_sec == rhs.tv_sec) {
        if (lhs.tv_sec >= 0) {
            return lhs.tv_nsec < rhs.tv_nsec;
        } else {
            return !(lhs.tv_nsec < rhs.tv_nsec);
        }
    } else {
        return lhs.tv_sec < rhs.tv_sec;
    }
}

inline bool operator>(const struct otimeSpec& lhs, const struct otimeSpec& rhs)
{
    return operator<(rhs, lhs);
}

inline bool operator<=(const struct otimeSpec& lhs, const struct otimeSpec& rhs)
{
    return !operator>(lhs, rhs);
}

inline bool operator>=(const struct otimeSpec& lhs, const struct otimeSpec& rhs)
{
    return !operator<(lhs, rhs);
}

inline otimeSpec& otimeSpec::operator+=(const struct otimeSpec& other)
{
    tv_sec += other.tv_sec;
    tv_nsec += other.tv_nsec;
    static_cast<void>(normalize());
    return *this;
}

inline otimeSpec& otimeSpec::operator-=(const struct otimeSpec& other)
{
    tv_sec -= other.tv_sec;
    tv_nsec -= other.tv_nsec;
    static_cast<void>(normalize());
    return *this;
}

inline otimeSpec operator+(const struct otimeSpec& lhs, const struct otimeSpec& rhs)
{
    struct otimeSpec ots(lhs);
    ots += rhs;
    return ots;
}

inline otimeSpec operator-(const struct otimeSpec& lhs, const struct otimeSpec& rhs)
{
    struct otimeSpec ots(lhs);
    ots -= rhs;
    return ots;
}

inline otimeSpec otimeSpec::operator-(void) const
{
    otimeSpec ots;
    ots.tv_sec = -this->tv_sec;
    ots.tv_nsec = -this->tv_nsec;
    return ots;
}

class otimePoint;
class otimeDuration;

/*!
 * @class otimeDuration
 *
 * @brief This class shall represent a duration of time using the otimeSpec struct.
 * \n It represents time using two int64_t values: one for seconds and another for nanoseconds.
 * \n It has several member functions to manipulate the class object.
 * \n This class can be serialized using the cereal library.
 *
 */
class otimeDuration
{
public:
    // constructors

    /*!
     * @brief Default constructor
     * \n It shall initialize the seconds and nanoseconds of otimeSpec with 0.
     */
    otimeDuration()
        : time_(0, 0)
    {}

    /*!
     * @brief Default copy constructor
     *
     * @param[in] other : The otimeDuration object to copy
     */
    otimeDuration(const otimeDuration& other) = default;

    /*!
     * @brief Default move constructor
     *
     * @param[in] other : The otimeDuration object
     */
    otimeDuration(otimeDuration&& other) = default;

    // TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
    /*!
     * @brief Parameterized constructor for otimeDuration takes given duration in the form of otimePoint.
     * \n This shall normalize otimeDuration after initialization  with otimePoint
     *
     * @param[in] t : The given duration in otimePoint.
     *
     * \sa otimeDuration::normalize
     */
    explicit otimeDuration(otimePoint& t);

    /*!
     * @brief Parameterized constructor for otimeDuration using given duration in the form of otimePoint.
     * \n This shall normalize otimeDuration after initialization  with otimePoint
     *
     * @param[in] t : const otimePoint object.
     *
     * \sa otimeDuration::normalize
     */
    explicit otimeDuration(const otimePoint& t);
    // explicit otimeDuration(otimePoint&& t);

    // TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... used to have
    // these clunky (and non-const-reference-taking) ctors instead of just allowing
    // copy construction as an otimeDuration...
    /*!
     * @brief Parameterized constructor for otimeDuration takes given duration in the form of oSeconds.
     *
     * @param[in] t : The given duration oSeconds object .
     */
    otimeDuration(const oSeconds& t);

    /*!
     * @brief Parameterized constructor for otimeDuration takes given duration in the form of oMilliSeconds.
     *
     * @param[in] t : The given duration oMilliSeconds object .
     */
    otimeDuration(const oMilliSeconds& t);

    /*!
     * @brief Parameterized constructor for otimeDuration takes given duration in the form of oMicroSeconds.
     *
     * @param[in] t : The given duration oMicroSeconds object .
     */
    otimeDuration(const oMicroSeconds& t);

    /*!
     * @brief Parameterized constructor for otimeDuration takes given duration in the form of oNanoSeconds.
     *
     * @param[in] t : The given duration oNanoSeconds object .
     */
    otimeDuration(const oNanoSeconds& t);

    // utility functions

    /*!
     * @brief This method shall check if the seconds of otimeDuration is negative.
     *
     * @return bool : true if seconds of otimeDuration is negative.otherwise, false.
     *
     */
    bool isNegative() const;

    /*!
     * @brief This method shall return the total duration in seconds.
     * \n It shall convert seconds and nanoseconds of otimeSpec to seconds.
     * \n And the resultant total seconds shall be returned.
     *
     * @return int64_t : Duration in seconds.
     *
     */
    int64_t totalSeconds() const;

    /*!
     * @brief This method shall return the total duration in seconds.
     * \n It shall convert seconds and nanoseconds of otimeSpec to seconds.
     * \n And the resultant total seconds shall be returned as a double value.
     *
     * @return double : Duration in seconds.
     *
     */
    double totalSecondsDouble() const;

    /*!
     * @brief This method shall return the total duration in milliseconds.
     * \n It shall convert seconds and nanoseconds of otimeSpec to milliseconds.
     * \n And the resultant total milliseconds shall be returned.
     *
     * @return int64_t :  Duration in milliseconds.
     *
     */
    int64_t totalMilliSeconds() const;

    /*!
     * @brief This method shall return the total duration in microseconds.
     * \n It shall convert seconds and nanoseconds of otimeSpec to microseconds
     * \n And the resultant total microseconds shall be returned.
     *
     * @return  int64_t : Duration in microseconds
     *
     */
    int64_t totalMicroSeconds() const;

    /*!
     * @brief This method shall return the total duration in nanoseconds.
     * \n It shall convert seconds and nanoseconds of otimeSpec to nanoseconds
     * \n And the resultant total nanoseconds shall be returned
     *
     * @return int64_t : Duration in nanoseconds.
     *
     */
    int64_t totalNanoSeconds() const;

    /*!
     * @brief This method shall convert the otimeDuration object to time structure tm format.
     *
     * @return  An instance of tm
     *
     */
    tm to_tm() const;

    /*!
     * @brief This method shall convert the otimeDuration object to system time type time_t.
     * \n It shall convert seconds otimeSpec to time_t and ignores the nanoseconds of otimeSpec.
     *
     * @return An instance of time_t
     *
     */
    time_t to_time_t() const { return static_cast<time_t>(time_.tv_sec); }

    /*!
     * @brief This method shall convert the otimeDuration object to elapsed time structure timeval.
     *
     * @return An instance of timeval
     *
     */
    timeval to_timeval() const;

    // operator definitions
    // TODO [Matt Ivester 2/13/2020]: the operator=s should be defaulted... keeping ABI compatibility
    /*!
     * @brief Overloaded assignment operator (=) method
     * \n This method shall assign the values of source otimeDuration object(rhs) into destination otimeDuration
     * object(lhs)
     *
     * @param[in] rhs : otimeDuration object
     *
     * @return otimeDuration object (lhs)
     */
    otimeDuration& operator=(const otimeDuration& rhs);

    /*!
     * @brief move assignment operator (=) method
     * \n This method shall transfer source otimeDuration object(rhs) into destination otimeDuration object(lhs)
     *
     * @param[in] rhs : rvalue reference of otimeDuration object
     *
     * @return otimeDuration object (lhs)
     */
    otimeDuration& operator=(otimeDuration&& rhs);

    /*!
     * @brief Overloaded assignment operator (=) method
     * \n This method shall assign the values of source otimePoint object(rhs) into destination otimeDuration
     * object(lhs)
     *
     * @param[in] rhs : otimePoint object
     *
     * @return otimeDuration object (lhs)
     */
    otimeDuration& operator=(const otimePoint& rhs);

    /*!
     * @brief move assignment operator (=) method
     * \n This method shall transfer source otimePoint object(rhs) into destination otimeDuration object(lhs)
     *
     * @param[in] rhs : rvalue reference of otimePoint object
     *
     * @return otimeDuration object (lhs)
     */
    otimeDuration& operator=(otimePoint&& rhs);

    /*!
     * @brief Overloaded Is equals to (==) operator method
     * \n This method shall check if both this object and rhs are equal.
     *
     * @param[in] rhs : The otimeDuration object
     *
     * @return bool : true, if lhs and rhs are equal.otherwise, false.
     */
    bool operator==(const otimeDuration& rhs) const;

    /*!
     * @brief Overloaded Not equal to operator (!=) method.
     * \n This method shall check if both this object and rhs are not equal.
     *
     * @param[in] rhs : The otimeDuration object
     *
     * @return bool : true, if lhs and rhs are not equal.otherwise, false.
     *
     */
    bool operator!=(const otimeDuration& rhs) const { return !operator==(rhs); }

    /*!
     * @brief Overloaded less than operator (<) method.
     * \n This method shall check if this object time is less than rhs time.
     *
     * @param[in] rhs : The otimeDuration object
     *
     * @return bool : true, if lhs time is less than rhs time.otherwise, false.
     */
    bool operator<(const otimeDuration& rhs) const;

    /*!
     * @brief Overloaded greater than operator (>) method.
     * \n This method shall check if this object time is greater than rhs time.
     *
     * @param[in] rhs : The otimeDuration object
     *
     * @return bool : true, if lhs time is greater than rhs time.otherwise, false.
     */
    bool operator>(const otimeDuration& rhs) const;

    /*!
     * @brief Overloaded less than or equals to operator (<=) method.
     * \n This method shall check if this object time is less than or equals to rhs time.
     *
     * @param[in] lhs : The otimeDuration object
     * @param[in] rhs : The otimeDuration object
     *
     * @return bool : true, if lhs time is less than or equals to rhs time.otherwise, false.
     */
    bool operator<=(const otimeDuration& rhs) const { return !operator>(rhs); }

    /*!
     * @brief Overloaded greater than or equals to operator (>=) method.
     * \n This method shall check if this object time is greater than or equals to rhs time.
     *
     * @param[in] lhs : The otimeDuration object
     * @param[in] rhs : The otimeDuration object
     *
     * @return bool : true, if lhs time is greater than or equals to rhs time.otherwise, false.
     */
    bool operator>=(const otimeDuration& rhs) const { return !operator<(rhs); }

    /*!
     * @brief Overloaded operator -
     * \n This method shall make otimeDuration as negative.
     * \n It shall make seconds and nanoseconds of otimeSpec as negative values.
     *
     * @return otimeSpec object
     */
    otimeDuration operator-(void) const;

    /*!
     * @brief Overloaded operator * method
     * \n This method shall multiply the time in otimeDuration(lhs) with given numeric value rhs and returns the
     * resultant time. \n \b Operational \b Flow: \n nanoseconds of otimeDuration(lhs) converted to seconds and added to
     * seconds of otimeDuration. \n Now total seconds of otimeDuration which is a whole number multiplied with given
     * numeric value. \n IF the resultant seconds value is less than or equals to OTIME_MAX( A unit representing max
     * time) THEN \n &emsp; oNanoSeconds object is created with resultant seconds and returned as otimeDuration object
     * \n ELSE IF resultant seconds value is greater than OTIME_MAX but less than or equals to the value of (OTIME_MAX *
     * NSEC_PER_SEC) \n &emsp; where NSEC_PER_SEC represent nanoseconds per sec THEN \n &emsp; oMicroSeconds object is
     * created with resultant seconds and returned as otimeDuration object \n ELSE \n &emsp; oMilliSeconds object is
     * created with resultant seconds and returned as otimeDuration object \n END IF
     *
     * @param[in] rhs : given numeric value
     *
     * @return otimeDuration object
     */
    otimeDuration operator*(const int rhs) const;

    /*!
     * @brief Overloaded operator / method
     * \n This method shall divide the time in otimeDuration(lhs) with given numeric value rhs and returns the resultant
     * time. \n \b Operational \b Flow: \n nanoseconds of otimeDuration(lhs) converted to seconds and added to seconds
     * of otimeDuration. \n Now total seconds of otimeDuration which is a whole number divided with give numeric value.
     * \n IF the resultant seconds value is less than or equals to OTIME_MAX( A unit representing max time) THEN
     * \n &emsp; oNanoSeconds object is created with resultant seconds and returned as otimeDuration object
     * \n ELSE IF resultant seconds value is greater than OTIME_MAX but less than or equals to the value of (OTIME_MAX *
     * NSEC_PER_SEC) \n &emsp; where NSEC_PER_SEC represent nanoseconds per sec THEN \n &emsp; oMicroSeconds object is
     * created with resultant seconds and returned as otimeDuration object \n ELSE \n &emsp; oMilliSeconds object is
     * created with resultant seconds and returned as otimeDuration object \n END IF
     *
     * @param[in] rhs : given numeric value
     *
     * @return otimeDuration object
     */
    otimeDuration operator/(const int rhs) const;

    /*!
     * @brief  This method shall add time in otimePoint(rhs) to the lhs time in otimeDuration.
     *
     * @param[in] rhs : The otimePoint object
     *
     * @return otimeDuration object
     */
    otimeDuration& operator+=(const otimePoint& rhs);

    /*!
     * @brief  This method shall subtract time in otimePoint(rhs) from the lhs time in otimeDuration.
     *
     * @param[in] rhs : The otimePoint object
     *
     * @return otimeDuration object
     */
    otimeDuration& operator-=(const otimePoint& rhs);

    /*!
     * @brief  This method shall add time in otimeDuration(rhs) to the lhs time in otimeDuration.
     *
     * @param[in] rhs : The otimeDuration object
     *
     * @return otimeDuration object(lhs)
     */
    otimeDuration& operator+=(const otimeDuration& rhs);

    /*!
     * @brief  This method shall subtract time in otimeDuration(rhs) from the lhs time in otimeDuration.
     *
     * @param[in] rhs : The otimeDuration object
     *
     * @return otimeDuration object(lhs)
     */
    otimeDuration& operator-=(const otimeDuration& rhs);

protected:
    // no normalization here, assume subclasses are handling it!
    explicit otimeDuration(int64_t secs, int64_t nsecs)
        : time_(secs, nsecs)
    {}

    explicit otimeDuration(const otimeSpec& t)
        : time_(t)
    {}

    explicit otimeDuration(otimeSpec&& t)
        : time_(std::move(t))
    {}

    struct otimeSpec time_; ///< Represents the time duration as otimeSpec structure
                            ///< otimeSpec has two int64_t values: one for seconds and one for nanoseconds.

private:
    friend class cereal::access;
    friend class otimePoint;
    // TODO [Matt Ivester 2/19/2020]: these were also defined wrong, lhs parameter
    // should be a 'const otimePoint&'.  Keeping for ABI compatibility.
    /*!
     * @brief This method shall add given two otimePoint objects and returns resultant time
     *
     * @param[in] lhs : otimePoint object
     * @param[in] rhs : otimePoint object
     *
     * @return otimeDuration object
     */
    friend otimeDuration operator+(otimePoint lhs, const otimePoint& rhs);

    /*!
     * @brief This method shall subtract given rhs otimePoint object from the lhs otimePoint object and returns
     * resultant time.
     *
     * @param[in] lhs : otimePoint object
     * @param[in] rhs : otimePoint object
     *
     * @return otimeDuration object
     */
    friend otimeDuration operator-(otimePoint lhs, const otimePoint& rhs);

    /*!
     * @brief This method shall add given lhs otimePoint and lhs otimeDuration objects.It returns the resultant time.
     *
     * @param[in] lhs : time in otimePoint
     * @param[in] rhs : time in otimeDuration
     *
     * @return otimePoint object
     */
    friend otimePoint operator+(otimePoint lhs, const otimeDuration& rhs);

    /*!
     * @brief  This method shall subtract given rhs otimeDuration object from the given lhs otimePoint object and
     * returns resultant time.
     *
     * @param[in] lhs : otimePoint object
     * @param[in] rhs : otimeDuration object
     *
     * @return otimePoint object
     */
    friend otimePoint operator-(otimePoint lhs, const otimeDuration& rhs);

    /*!
     * @brief  This method shall add given two otimeDuration objects and returns resultant time
     *
     * @param[in] lhs : otimeDuration object
     * @param[in] rhs : otimeDuration object
     *
     * @return otimeDuration object
     */
    friend otimeDuration operator+(const otimeDuration& lhs, const otimeDuration& rhs);

    /*!
     * @brief  This method shall subtract given rhs otimeDuration object from the given lhs otimeDuration object and
     * returns resultant time.
     *
     * @param[in] lhs : otimeDuration object
     * @param[in] rhs : otimeDuration object
     *
     * @return otimeDuration object
     */
    friend otimeDuration operator-(const otimeDuration& lhs, const otimeDuration& rhs);

    // utility functions
    /*!
     * @brief This method shall apply normalization on otimeDuration.
     * \n i.e No negative nanoseconds in otimeDuration and nanoseconds stays with in seconds of otimeDuration.
     *
     * @return void
     *
     * \sa otimeSpec::normalize()
     */
    void normalize();

    /*!
     * @brief This method shall serialize otimeSpec type value
     *
     * @param[in] ar : cereal Archive object
     * @param[in] version : current data version
     *
     * @return void
     *
     * \sa otimeSpec::serialize()
     */
    template<class Archive>
    void serialize(Archive& ar, unsigned int version)
    {
        static_cast<void>(version);
        ar& CEREAL_NVP(time_.tv_sec) & CEREAL_NVP(time_.tv_nsec);
    }
}; // end class otimeDuration

/*!
 * @class otimePoint
 *
 * @brief The class represents a time point.
 * \n \b Detailed \b Description:
 * \n This class shall represent a time using the otimeSpec struct.
 * \n It represents time using two int64_t values: one for seconds and one for nanoseconds.
 * \n It has several methods to manipulate the class object.
 * \n This class can be serialized using the cereal library.
 */
class otimePoint
{
public:
    // constructors

    /*!
     * @brief Default constructor
     *
     */
    otimePoint() {}

    /*!
     * @brief Default copy constructor
     *
     * @param[in] other : The otimePoint object to copy
     */
    otimePoint(const otimePoint& other) = default;

    /*!
     * @brief Default move constructor
     *
     * @param[in] other : The otimePoint object
     */
    otimePoint(otimePoint&& other) = default;

    // TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
    /*!
     * @brief Parameterized constructor for otimePoint taking time structure timespec
     * \n \b Operational \b Flow:
     * \n It shall initialize the otimePoint with given time in timespec.
     * \n IF seconds part of timespec is greater than or equal to OTIME_SECOND_MAX ( A value representing the max number
     * of seconds ) THEN \n &emsp; invalidate the otimePoint object by setting with value -1 \n ELSE \n &emsp; apply
     * normalization on otimePoint i.e no negative nanoseconds in otimePoint and nanoseconds stays with in seconds of
     * otimePoint. \n &emsp; IF normalization fails THEN \n &emsp; &emsp; invalidate the otimePoint object by setting
     * with value -1 \n &emsp; END IF \n END IF
     *
     * @param[in] t  timespec object.
     * \sa otimePoint::normalize otimePoint::invalidate
     */
    otimePoint(struct timespec t);
    // otimePoint(const struct timespec& t);

    // TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
    /*!
     * @brief Parameterized constructor for otimePoint taking time structure tm.
     * \n \b Operational \b Flow:
     * \n It shall initialize the otimePoint with given time in tm.
     * \n converts the given time tm to timespec format
     * \n IF seconds of timespec is greater than or equal to OTIME_SECOND_MAX ( A value representing the max number of
     * seconds) THEN \n &emap; invalidate the otimePoint object by setting with value -1 \n ELSE \n &emsp; initialize
     * the otimePoint with timespec \n &emsp; apply normalization on otimePoint i.e no negative nanoseconds in
     * otimePoint and nanoseconds stays with in seconds of otimePoint. \n &emsp; IF normalization fails THEN \n &emsp;
     * &emsp; invalidate the otimePoint object by setting with value -1 \n &emsp; END IF \n END IF
     *
     * @param[in] t : The given time in tm structure
     *
     * \sa otimePoint::normalize
     * \n otimePoint::invalidate
     */
    otimePoint(struct tm t);
    // otimePoint(const struct tm& t);

    // TODO [Matt Ivester 2/11/2020] I'd like these to be explicit,
    // but found some dependent packages that were doing things like:
    // (someoTimePoint - otimeSpec{blah blah}).<someoDurationFunction>()
    // ...which won't compile if the otimePoint(otimeSpec) ctor is explicit.
    // TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
    /*!
     * @brief Parameterized constructor for otimePoint takes time structure otimeSpec
     * \n \b Operational \b Flow:
     * \n It shall initialize the otimePoint with given time in otimeSpec.
     * \n IF seconds of otimeSpec is greater than or equal to OTIME_SECOND_MAX( A value representing the max number of
     * seconds) THEN \n &emsp; invalidate the otimePoint object by setting with value -1 \n ELSE \n &emsp; initialize
     * the otimePoint with given time otimeSpec \n &emsp; apply normalization on otimePoint i.e no negative nanoseconds
     * in otimePoint and nanoseconds stays with in seconds of otimePoint \n &emsp; IF normalization fails THEN \n &emsp;
     * &emsp; invalidate the otimePoint object by setting with value -1 \n &emsp; END IF \n END IF
     *
     * @param[in] t : The given time in otimeSpec.
     *
     * \sa otimePoint::normalize otimePoint::invalidate
     *
     */
    otimePoint(struct otimeSpec t);
    // otimePoint(const struct otimeSpec& t);
    // otimePoint(struct otimeSpec&& t);

    /*!
     * @brief Constructor using a given time.
     *
     * @param t The given time in otimeDuration.
     *
     */
    // TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
    /*!
     * @brief Parameterized constructor for otimePoint takes otimeDuration
     * \n \b Operational \b Flow:
     * \n It shall initialize the otimePoint with given time in otimeDuration.
     * \n IF seconds of otimeDuration is greater than or equal to OTIME_SECOND_MAX( A value representing the max number
     * of seconds) THEN \n &emap; invalidate the otimePoint object by setting with value -1 \n ELSE \n &emsp; initialize
     * the otimePoint with given time otimeDuration \n &emsp; apply normalization on otimePoint i.e no negative
     * nanoseconds in otimePoint and nanoseconds stays with in seconds of otimePoint. \n &emsp; IF normalization fails
     * THEN \n &emsp; &emsp; invalidate the otimePoint object by setting with value -1 \n &emsp; END IF \n END IF
     *
     * @param[in] t : The given time in timeDuration object
     *
     * \sa otimePoint::normalize otimePoint::invalidate
     *
     */
    explicit otimePoint(otimeDuration& t);

    /*!
     * @brief Parameterized constructor for otimePoint takes otimeDuration
     * \n \b Operational \b Flow:
     * \n It shall initialize the otimePoint with given time in otimeDuration.
     * \n IF seconds of otimeDuration is greater than or equal to OTIME_SECOND_MAX( A value representing the max number
     * of seconds) THEN \n &emap; invalidate the otimePoint object by setting with value -1 \n ELSE \n &emsp; initialize
     * the otimePoint with given time otimeDuration \n &emsp; apply normalization on otimePoint i.e no negative
     * nanoseconds in otimePoint and nanoseconds stays with in seconds of otimePoint. \n &emsp; IF normalization fails
     * THEN \n &emsp; &emsp; invalidate the otimePoint object by setting with value -1 \n &emsp; END IF \n END IF
     *
     * @param[in] t : The given time in const otimeDuration object.
     *
     * \sa otimePoint::normalize otimePoint::invalidate
     */
    explicit otimePoint(const otimeDuration& t);
    // explicit otimePoint(otimeDuration&& t);

    // utility functions

    /*!
     * @brief This method shall check if the time is valid.
     * \n It returns true if seconds and nanoseconds of otimePoint is not -1.

     * @return bool :  true, if valid time.otherwise, false
     */
    bool isTimeValid() const;

    /*!
     * @brief This method shall normalizes the time to make sure that nanoseconds stays within a second and
     * \n  there is  no negative nanoseconds.
     * \n \b Operational \b Flow:
     * \n IF normalization fails THEN
     * \n &emsp; invalidate otimePoint object by setting with a value -1 and returns false.
     * \n END IF
     *
     * @return bool : true, if seconds of otimePoint is positive. otherwise, false
     *
     * \sa otimeSpec::normalize
     */
    bool normalize();

    /*!
     * @brief This method shall invalidate the time by setting seconds and nanoseconds of otimePoint as -1.
     *
     * @return Void.
     *
     */
    void invalidate();

    /*!
     * @brief This method shall convert otimePoint to time structure tm .
     *
     * @return Time in tm.
     */
    tm to_tm() const;

    /*!
     * @brief This method shall convert otimePoint to time structure time_t
     *
     * @return Time in time_t.
     */
    time_t to_time_t() const { return static_cast<time_t>(time_.tv_sec); }

    /*!
     * @brief This method shall return otimeSpec which otimePoint holds.
     *
     * @return Time in otimeSpec.
     */
    otimeSpec to_otimeSpec() const { return time_; }

    /*!
     * @brief This method shall convert otimePoint to time structure  timeval
     *
     * @return Time in timeval.
     */
    timeval to_timeval() const;

    /*!
     * @brief This method shall return an otimePoint representing the epoch time.
     *
     * @return the epoch time as an otimePoint.
     *
     */
    static otimePoint getEpochTime()
    {
        static const otimePoint eTime{otimeSpec::getEpochTimeSpec()};
        return eTime;
    }

    /*!
     * @brief This method shall return the amount of time from the epoch to this otimePoint, as an otimeDuration.
     *        Result is indeterminate if the otimePoint is invalid.
     *
     * @return otimeDuration object.The elapsed time since the epoch. This may be negative for dates far in the past.
     *
     */
    otimeDuration timeSinceEpoch() const { return *this - getEpochTime(); }

    /*!
     * @brief This method shall return the number of seconds since the epoch.
     *        Result is indeterminate if the otimePoint is invalid.
     *
     * @return int64_t : Number of seconds since the epoch, as an integer. Note that this will be negative for times way
     * in the past.
     *
     */
    int64_t toSecondsSinceEpochInt() const;
    /*!
     * @brief This method shall convert number of seconds from the epoch into an otimePoint.
     *
     * @param[in] secs as int64_t: Number of seconds from the epoch.
     *
     * @return the otimePoint representing the epoch time.
     *
     */
    static otimePoint fromSecondsSinceEpochInt(int64_t secs);

    /*!
     * @brief This method shall return the number of seconds since the epoch.
     *        Result is indeterminate if the otimePoint is invalid.
     *
     * @return double : Number of seconds since the epoch. Note that this will be negative for times way in the past.
     *
     */
    double toSecondsSinceEpochDouble() const;
    /*!
     * @brief This method shall convert number of seconds from the epoch into an otimePoint.
     *        Result is indeterminate if the otimePoint is invalid.
     *
     * @param[in] secs as double : Number of seconds.
     *
     * @return the otimePoint representing the epoch time.
     *
     */
    static otimePoint fromSecondsSinceEpochDouble(double secs);

    /*!
     * @brief This method shall convert the number of milliseconds since the epoch.
     *        Result is indeterminate if the otimePoint is invalid.
     *
     * @return int64_t : Number of milliseconds since the epoch.  Note that this will be negative for times way in the
     * past.
     *
     */
    int64_t toMillisecondsSinceEpoch() const;
    /*!
     * @brief This method shall convert the number of milliseconds from the epoch into an otimePoint.
     *
     * @param[in] millis as int64_t : Number of milliseconds.
     *
     * @return the otimePoint representing the epoch time.
     *
     */
    static otimePoint fromMillisecondsSinceEpoch(int64_t millis);

    /*!
     * @brief This method shall convert the number of microseconds since the epoch.
     *        Result is indeterminate if the otimePoint is invalid.
     *
     * @return int64_t : Number of microseconds since the epoch, as an integer.  Note that this will be negative for
     * times way in the past.
     *
     */
    int64_t toMicrosecondsSinceEpoch() const;
    /*!
     * @brief This method shall convert number of microseconds from the epoch into an otimePoint.
     *
     * @param[in] micros as int64_t : Number of microseconds.
     *
     * @return the otimePoint representing the epoch time.
     *
     */
    static otimePoint fromMicrosecondsSinceEpoch(int64_t micros);

    /*!
     * @brief This method shall convert number of nanoseconds since the epoch
     *        Result is indeterminate if the otimePoint is invalid.
     *
     * @return int64_t : Number of nanoseconds since the epoch.  Note that this will be negative for times way in the
     * past.
     *
     */
    int64_t toNanosecondsSinceEpoch() const;
    /*!
     * @brief This method shall convert number of nanoseconds from the epoch into an otimePoint.
     *
     * @param[in] nanos as int64_t : Number of nanoseconds.
     *
     * @return the otimePoint representing the epoch time.
     *
     */
    static otimePoint fromNanosecondsSinceEpoch(int64_t nanos);

    /*!
     * @brief DEPRECATED -- sets the internal otimeSpec to
     *        an internal clock time.  Currently this is equivalent
     *        to calling getCurrentWallClockTime().  Uses of this
     *        should be changed to explicitly request one of
     *        the other clocks.
     *
     * @return void
     */
    // TODO [Matt Ivester 2/11/2020]: should be deprecated but this breaks ottomatika build-missing builds.
    // [[deprecated("Use getCurrentWallClockTime()/getCurrentSteadyTime()/getCurrentHighResolutionTime()")]]
    void getLocalTime(void);

    /*!
     * @brief This method shall set the otimePoint's internal time structure( otimeSpec ) with the current
     * \n wall clock time as reported by the system.  Note that this time is NOT guaranteed to be steady and may jump
     * \n or move backwards in some cases.This is most useful for things like timestamps in a log.
     *
     * \n If you need a time that is synchronized across the system, use SystemClock instead.
     * \n (Note that SystemClock also has a method that returns formatted wall clock strings for logging.)
     *
     * @return Void.
     *
     */
    void getCurrentWallClockTime();
    /*!
     * @brief This method shall return the current wall clock time in otimePoint form.
     * \n The current wall clock time is reported by the system which uses the clock type as system_clock.
     *
     * @return otimePoint
     *
     */
    static otimePoint buildCurrentWallClockTime();

    /*!
     * @brief This method shall set the otimePoint's internal time structure( otimeSpec ) with a steady clock time
     * \n as returned by the system.  This clock is guaranteed to  be monotonic and steady,
     * \n but may not correlate to real world time.It is usable to track durations of events, even very long ones.
     *
     * \n If you need a steady time that is synchronized across the system, use SystemClock instead.
     *
     * @return Void.
     *
     */
    void getCurrentSteadyTime();
    /*!
     * @brief This method shall return the current clock time in otimePoint form.
     * \n The current clock time is reported by the system which uses the clock type as steady_clock.
     *
     * @return otimePoint
     *
     */
    static otimePoint buildCurrentSteadyTime();

    /*!
     * @brief This method shall set the otimePoint's internal time structure(otimeSpec)
     * \n using the available system clock with the most accurate time.  This clock
     * \n may not correlate with any other clocks.  This clock is NOT guaranteed to be
     * \n monotonic or steady, but may be better in terms of internal event timing.
     * \n It is primarily useful for timing extremely short duration local events.
     *
     * \n If you are timing events that take more than a few seconds, or you
     * \n do not need 1ms precision or better, use getCurrentSteadyTime() instead.
     * \n If you need to time events across the system, use SystemClock.
     *
     * \n It uses high resolution clock.
     *
     * @return Void.
     *
     */
    void getCurrentHighResolutionTime();
    /*!
     * @brief This method shall return current time in otimePoint form.
     * \n The current time is reported by the system which uses the clock type as high resolution clock.
     *
     * @return Void.
     *
     */
    static otimePoint buildCurrentHighResolutionTime();

    // operator definitions

    // TODO [Matt Ivester 2/13/2020]: the operator=s should be defaulted, keeping ABI compatiblity...
    /*!
     * @brief overloaded assignment operator (=) method
     * \n This method shall assign the values of source otimePoint object(rhs) into destination otimePoint object(lhs)
     *
     * @param[in] rhs : const otimePoint object
     *
     * @return otimePoint object(this object)
     */
    otimePoint& operator=(const otimePoint& rhs);

    /*!
     * @brief overloaded assignment operator (=) method
     * \n This method shall assign the values of source otimePoint object(rhs) into destination otimePoint object(lhs)
     *
     * @param[in] rhs : otimePoint object
     *
     * @return otimePoint object(this object)
     */
    otimePoint& operator=(otimePoint&& rhs);

    /*!
     * @brief overloaded Is equals to (==) operator method
     * \n This method shall check if both this object and rhs are equal.
     *
     * @param[in] rhs : otimePoint object
     *
     * @return bool : true, if both this object and rhs are equal.otherwise, false
     */
    bool operator==(const otimePoint& rhs) const;

    /*!
     * @brief overloaded Not equal to operator (!=) method.
     * \n This method shall check if both this object and rhs are not equal
     *
     * @param[in] rhs : otimePoint object
     *
     * @return bool : true, if this object and rhs are not equal.otherwise, false.
     */
    bool operator!=(const otimePoint& rhs) const { return !operator==(rhs); }

    /*!
     * @brief Overloaded less than operator (<) method.
     * \n This method shall check if this object time is less than rhs time.
     *
     * @param[in] rhs : otimePoint object
     *
     * @return bool : true, if this object time is less than rhs time.otherwise, false.
     *
     * \sa otimeSpec::operator<
     *
     */
    bool operator<(const otimePoint& rhs) const;

    /*!
     * @brief Overloaded greater than operator (>) method.
     * \n This method shall check if this object time is greater than rhs time.
     *
     * @param[in] rhs : otimePoint object
     *
     * @return bool : true, if this object time is greater than rhs time.otherwise, false.
     *
     * \sa otimeSpec::operator>
     *
     */
    bool operator>(const otimePoint& rhs) const;

    /*!
     * @brief Overloaded less than or equals to operator (<=) method.
     * \n This method shall check if this object time is less than or equals to rhs time.
     *
     * @param[in] rhs : otimePoint object
     *
     * @return bool : true, if this object time is less than or equals to rhs time.otherwise, false.
     */
    bool operator<=(const otimePoint& rhs) const { return !operator>(rhs); }

    /*!
     * @brief Overloaded greater than or equals to operator (>=) method.
     * \n This method shall check if this object time is greater than or equals to rhs time.
     *
     * @param[in] rhs : otimePoint object
     *
     * @return bool : true, if this object time is greater than or equals to rhs time.otherwise, false.
     */
    bool operator>=(const otimePoint& rhs) const { return !operator<(rhs); }

    /*!
     * @brief This method shall add given rhs otimeDuration object (rhs) to the lhs otimePoint object.
     * \n It shall apply normalization on lhs.
     *
     * @param[in] rhs : otimePoint object
     *
     *
     * @return otimePoint object(lhs)
     */
    otimePoint& operator+=(const otimeDuration& rhs);

    /*!
     * @brief  This method shall subtract time in otimeDuration(rhs) from the lhs time in otimePoint.
     *
     * @param[in] rhs : otimeDuration object
     *
     *
     * @return otimePoint object(this object)
     */
    otimePoint& operator-=(const otimeDuration& rhs);

    /*!
     * @brief This method shall serialize otimePoint type value
     *
     * @param[in] ar : cereal Archive object
     * @param[in] version : current data version
     *
     * @return void
     *
     * \sa otimePoint::serialize()
     */
    template<class Archive>
    void serialize(Archive& ar, unsigned int version)
    {
        static_cast<void>(version);
        ar& CEREAL_NVP(time_.tv_sec) & CEREAL_NVP(time_.tv_nsec);
    }

private:
    friend class otimeDuration;
    // TODO [Matt Ivester 2/19/2020]: these were also defined wrong, lhs parameter
    // should be a 'const otimePoint&'.  Keeping for ABI compatibility.
    friend otimeDuration operator+(const otimePoint lhs, const otimePoint& rhs);
    friend otimeDuration operator-(const otimePoint lhs, const otimePoint& rhs);
    friend otimePoint operator+(otimePoint lhs, const otimeDuration& rhs);
    friend otimePoint operator-(otimePoint lhs, const otimeDuration& rhs);
    friend otimeDuration operator+(const otimeDuration& lhs, const otimeDuration& rhs);
    friend otimeDuration operator-(const otimeDuration& lhs, const otimeDuration& rhs);

    struct otimeSpec time_; ///< Represents the time as otimeSpec structure
                            ///< otimeSpec has two int64_t values: one for seconds and one for nanoseconds.
};                          // end class otimePoint

/*!
 * @class oSeconds
 *
 * @brief A wrapper class representing time in second.
 *
 */
class oSeconds : public otimeDuration
{
public:
    /*!
     * @brief Parameterized constructor
     * \n It shall initialize the oSeconds with given seconds.
     * \n It shall initialize seconds otimeDuration with given seconds
     * \n and nanoseconds otimeDuration with 0.
     *
     * @param[in] t : time in int64_t
     */
    explicit oSeconds(int64_t t)
        : otimeDuration(t, 0)
    {}

    // Would have just made this an alternative constructor, but then the compiler complains
    // about ambiguous casting if you do something like oSeconds(1).
    /*!
     * @brief This method shall convert given number of seconds to oSeconds.
     *
     * @param[in] secs : Number of seconds as double
     *
     * @return oSeconds object
     */
    static oSeconds fromDouble(double secs)
    {
        const int64_t secsInt = static_cast<int64_t>(secs);
        const double remainder = secs - static_cast<double>(secsInt);
        const int64_t nsecsInt = static_cast<int64_t>(remainder * 1000000000.);

        oSeconds temp{secsInt};
        temp.time_.tv_nsec = nsecsInt;
        return temp;
    }

    /*!
     * @brief Default virtual destructor
     */
    virtual ~oSeconds() {}
};

/*!
 * @class oMilliSeconds
 *
 * @brief A wrapper class representing time in millisecond.
 *
 */
class oMilliSeconds : public otimeDuration
{
public:
    /*!
     * @brief Parameterized constructor
     * \n It shall initialize the oMilliSeconds with given value split into seconds and milliseconds
     *
     * @Note explicit is commented out because some downstream code uses oMilliSeconds ojb = 3000
     *
     * @param[in] t : time in int64_t
     */
    /*explicit*/ oMilliSeconds(int64_t t)
        : otimeDuration(t / 1000, (t % 1000) * 1000000)
    {}

    /*!
     *  @brief Default virtual destructor
     */
    virtual ~oMilliSeconds() {}
};

/*!
 * @class oMicroSeconds
 *
 * @brief A wrapper class representing time in microsecond.
 *
 */
class oMicroSeconds : public otimeDuration
{
public:
    /*!
     * @brief Parameterized constructor
     * \n It shall initialize the oMicroSeconds with given value split into seconds and microseconds
     *
     * @param[in] t :  time in int64_t
     */
    explicit oMicroSeconds(int64_t t)
        : otimeDuration(t / 1000000, (t % 1000000) * 1000)
    {}

    /*!
     * @brief Default virtual destructor
     */
    virtual ~oMicroSeconds() {}
};

/*!
 * @class oNanoSeconds
 *
 * @brief A wrapper class representing time in nanosecond.
 *
 */
class oNanoSeconds : public otimeDuration
{
public:
    /*!
     *  @brief Parameterized constructor
     *  \n It shall initialize the oNanoSeconds with given value  split into seconds and nanoseconds.
     *
     *  @param[in] t : time in int64_t
     */
    explicit oNanoSeconds(int64_t t)
        : otimeDuration(t / 1000000000, t % 1000000000)
    {}

    /*!
     * Default virtual destructor
     */
    virtual ~oNanoSeconds() {}
};

/*!
 * @brief Overloaded insertion operator(<<) to output otimePoint
 *
 * @param[in] out : reference to ostream object
 * @param[in] obj : reference to otimePoint object
 *
 * @return reference to ostream
 */
std::ostream& operator<<(std::ostream& out, const otimePoint& obj);

/*!
 * @brief overloaded insertion operator(<<) to output otimeDuration
 *
 * @param[in] out : reference to ostream object
 * @param[in] obj : reference to otimeDuration object
 *
 * @return reference to ostream
 */
std::ostream& operator<<(std::ostream& out, const otimeDuration& obj);

inline int64_t otimePoint::toSecondsSinceEpochInt() const
{
    return time_.tv_sec;
}
/* static */ inline otimePoint otimePoint::fromSecondsSinceEpochInt(int64_t secs)
{
    return otimePoint{otimeDuration{oSeconds{secs}}};
}

inline double otimePoint::toSecondsSinceEpochDouble() const
{
    return timeSinceEpoch().totalSecondsDouble();
}
/* static */ inline otimePoint otimePoint::fromSecondsSinceEpochDouble(double secs)
{
    return otimePoint{otimeDuration{oSeconds::fromDouble(secs)}};
}

inline int64_t otimePoint::toMillisecondsSinceEpoch() const
{
    return timeSinceEpoch().totalMilliSeconds();
}
/* static */ inline otimePoint otimePoint::fromMillisecondsSinceEpoch(int64_t millis)
{
    return otimePoint{otimeDuration{oMilliSeconds{millis}}};
}

inline int64_t otimePoint::toMicrosecondsSinceEpoch() const
{
    return timeSinceEpoch().totalMicroSeconds();
}
/* static */ inline otimePoint otimePoint::fromMicrosecondsSinceEpoch(int64_t micros)
{
    return otimePoint{otimeDuration{oMicroSeconds{micros}}};
}

inline int64_t otimePoint::toNanosecondsSinceEpoch() const
{
    return timeSinceEpoch().totalNanoSeconds();
}
/* static */ inline otimePoint otimePoint::fromNanosecondsSinceEpoch(int64_t nanos)
{
    return otimePoint{otimeDuration{oNanoSeconds{nanos}}};
}

} // namespace ottotime

#endif // OTTOTIME_H
