/*!
 ***********************************************************
 * @file otime.cc
 * @brief A file implementing several classes representing
 * time
 *
 * @section License
 *
 * Copyright (c) 2016- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/12/2016
 * @remark
 * Adding a limiter for otimePoint. otimePoint can support
 * only up to April 2262.
 * @n
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 02/11/2016
 * A new way of representing time.
 * @n
 *
 ***********************************************************
 */

#ifndef OTTOTIME_CC
#define OTTOTIME_CC

#include <TimeStamp/otime.h>
#include <cmath>

// TEMP [Matt Ivester 2/11/2020]:
// here so we can log a warning if code is still using otimePoint::getLocalTime(),
// which should be replaced by one of the clocks with guaranteed behavior.
#include "lib/avlog/avlog.hpp"

namespace ottotime {

// TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
// otimePoint::otimePoint(const struct timespec& t)
otimePoint::otimePoint(struct timespec t)
{
    if (static_cast<int64_t>(t.tv_sec) >= static_cast<int64_t>(OTIME_SECOND_MAX)) {
        this->invalidate();
        return;
    }

    time_.tv_sec = static_cast<int64_t>(t.tv_sec);
    time_.tv_nsec = static_cast<int64_t>(t.tv_nsec);

    if (!this->normalize()) {
        this->invalidate();
    }
}

/*!
 * @brief Specialized Constructor for otimePoint using tm
 *
 * @param t : A tm structure
 *
 * @return otimePoint : An instance of otimePoint.
 */
// TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
// otimePoint::otimePoint(const struct tm& t)
otimePoint::otimePoint(struct tm t)
{
    struct timespec ts;
    struct tm temp = t;
    ts.tv_sec = mktime(&temp);

    if (static_cast<int64_t>(ts.tv_sec) >= static_cast<int64_t>(OTIME_SECOND_MAX)) {
        this->invalidate();
        return;
    }

    time_.tv_sec = static_cast<int64_t>(ts.tv_sec);
    time_.tv_nsec = 0;

    if (!this->normalize()) {
        this->invalidate();
    }
}

// TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
// otimePoint::otimePoint(const struct otimeSpec& t)
otimePoint::otimePoint(struct otimeSpec t)
    : time_{t}
{
    if (static_cast<int64_t>(t.tv_sec) >= static_cast<int64_t>(OTIME_SECOND_MAX)) {
        this->invalidate();
        return;
    }

    if (!this->normalize()) {
        this->invalidate();
    }
}
// otimePoint::otimePoint(struct otimeSpec&& t)
//     : time_{std::move(t)}
// {
//     if (static_cast<int64_t>(t.tv_sec) >= static_cast<int64_t>(OTIME_SECOND_MAX))
//     {
//         this->invalidate();
//         return;
//     }

//     if (!this->normalize())
//     {
//         this->invalidate();
//     }
// }

// TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
otimePoint::otimePoint(otimeDuration& t)
    : time_{t.time_}
{
    if (static_cast<int64_t>(t.time_.tv_sec) >= static_cast<int64_t>(OTIME_SECOND_MAX)) {
        this->invalidate();
        return;
    }

    if (!this->normalize()) {
        this->invalidate();
    }
}

otimePoint::otimePoint(const otimeDuration& t)
    : time_{t.time_}
{
    if (static_cast<int64_t>(t.time_.tv_sec) >= static_cast<int64_t>(OTIME_SECOND_MAX)) {
        this->invalidate();
        return;
    }

    if (!this->normalize()) {
        this->invalidate();
    }
}

// otimePoint::otimePoint(otimeDuration&& t)
//     : time_{std::move(t.time_)}
// {
//     if ((int64_t)t.time_.tv_sec >= (int64_t)OTIME_SECOND_MAX)
//     {
//         this->invalidate();
//         return;
//     }

//     if (!this->normalize())
//     {
//         this->invalidate();
//     }
// }

void otimePoint::getLocalTime(void)
{
    // TEMP [Matt Ivester 2/11/2020]: can be removed once we get rid of this API.
    static bool havePrintedWarning{false};

    if (!havePrintedWarning) {
        AVLOG_WARN("Something called otimePoint::getLocalTime(), which is being deprecated.  This will only warn "
                   "once per run of the application.");
        havePrintedWarning = true;
    }

    getCurrentWallClockTime();
}

tm otimePoint::to_tm(void) const
{
    struct tm* timeInfo;

    if (!isTimeValid()) {
        // Time is invalid, make sure we *don't* apply timezone adjustment.
        // We (unfortunately) found some old code that was blindly calling to_tm()
        // and broke if the timezone adjustment pushed the reported time forward
        // into 1970.  So now we try to do this consistently.
        const time_t invalidRawTime = static_cast<time_t>(-1);
        timeInfo = gmtime(&invalidRawTime);
    } else {
        // Time is valid, hand it off to localtime() for conversion+timezone adjustment.
        const time_t rawTime = to_time_t();
        timeInfo = localtime(&rawTime);
    }

    tm tmTemp{};
    if (timeInfo != nullptr) {
        tmTemp = *timeInfo;
    }
    return tmTemp;
}

timeval otimePoint::to_timeval(void) const
{
    timeval tv;

    tv.tv_sec = this->time_.tv_sec;
    tv.tv_usec = this->time_.tv_nsec / NSEC_PER_USEC;

    return tv;
}

// TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... the ctor was defined like this earlier.
otimeDuration::otimeDuration(otimePoint& t)
    : time_{t.time_}
{
    this->normalize();
}
otimeDuration::otimeDuration(const otimePoint& t)
    : time_{t.time_}
{
    this->normalize();
}
// otimeDuration::otimeDuration(otimePoint&& t)
//     : time_{std::move(t.time_)}
// {
//     this->normalize();
// }

int64_t otimeDuration::totalSeconds(void) const
{
    double convertedTime;
    convertedTime =
        static_cast<double>(time_.tv_sec) + (static_cast<double>(time_.tv_nsec) / static_cast<double>(NSEC_PER_SEC));
    return static_cast<int64_t>(convertedTime);
}

double otimeDuration::totalSecondsDouble(void) const
{
    return static_cast<double>(time_.tv_sec) + (static_cast<double>(time_.tv_nsec) / static_cast<double>(NSEC_PER_SEC));
}

int64_t otimeDuration::totalMilliSeconds(void) const
{
    double convertedTime;
    convertedTime = (static_cast<double>(time_.tv_sec) * static_cast<double>(MSEC_PER_SEC)) +
                    (static_cast<double>(time_.tv_nsec) / static_cast<double>(NSEC_PER_MSEC));
    return static_cast<int64_t>(convertedTime);
}

int64_t otimeDuration::totalMicroSeconds(void) const
{
    double convertedTime;
    convertedTime = (static_cast<double>(time_.tv_sec) * static_cast<double>(USEC_PER_SEC)) +
                    (static_cast<double>(time_.tv_nsec) / static_cast<double>(NSEC_PER_USEC));
    return static_cast<int64_t>(convertedTime);
}

int64_t otimeDuration::totalNanoSeconds(void) const
{
    int64_t convertedTime;

    convertedTime = (time_.tv_sec * static_cast<int64_t>(NSEC_PER_SEC)) + time_.tv_nsec;
    return static_cast<int64_t>(convertedTime);
}

tm otimeDuration::to_tm(void) const
{
    struct tm* timeInfo;
    const otimeDuration rawTimeDuration(*this);
    time_t rawTime = rawTimeDuration.to_time_t();

    timeInfo = gmtime(&rawTime);

    tm tmTemp{};
    if (timeInfo != nullptr) {
        tmTemp = *timeInfo;
    }

    return tmTemp;
}

timeval otimeDuration::to_timeval(void) const
{
    timeval tv;

    tv.tv_sec = this->time_.tv_sec;
    tv.tv_usec = this->time_.tv_nsec / NSEC_PER_USEC;

    return tv;
}

namespace {
otimeDuration secondsToDuration(const long double seconds)
{
    if (fabsl(seconds) <= OTIME_MAX) {
        return oNanoSeconds(static_cast<int64_t>((seconds * static_cast<double>(NSEC_PER_SEC))));
    }

    if ((fabsl(seconds) <= (OTIME_MAX * static_cast<double>(NSEC_PER_USEC))) && (fabsl(seconds) > OTIME_MAX)) {
        return oMicroSeconds(static_cast<int64_t>((seconds * static_cast<double>(USEC_PER_SEC))));
    }

    return oMilliSeconds(static_cast<int64_t>((seconds * static_cast<double>(MSEC_PER_SEC))));
}
} // namespace

otimeDuration otimeDuration::operator*(const int rhs) const
{
    long double seconds{0.0};
    seconds += static_cast<long double>(this->time_.tv_sec);
    seconds += (static_cast<long double>(this->time_.tv_nsec) / static_cast<long double>(NSEC_PER_SEC));
    seconds *= static_cast<double>(rhs);

    return secondsToDuration(seconds);
}

otimeDuration otimeDuration::operator/(const int rhs) const
{
    long double seconds(0.0);
    seconds += static_cast<long double>(this->time_.tv_sec);
    seconds += (static_cast<long double>(this->time_.tv_nsec) / static_cast<double>(NSEC_PER_SEC));
    seconds /= static_cast<double>(rhs);

    return secondsToDuration(seconds);
}

std::ostream& operator<<(std::ostream& out, const otimePoint& obj)
{
    char buf[120];
    const otimePoint tmp(obj);
    const timeval timeOp = tmp.to_timeval();

    // Discard return value
    static_cast<void>(ctime_r(&timeOp.tv_sec, static_cast<char*>(buf)));

    std::string output = buf;
    try {
        // Discard return value
        static_cast<void>(snprintf(static_cast<char*>(buf),
                                   120U,
                                   "%s.%06d %s",
                                   output.substr(0, 19U).c_str(),
                                   static_cast<int32_t>(timeOp.tv_usec),
                                   output.substr(20U).c_str()));
        output = buf;
        out << output;
    } catch (std::out_of_range& ex) {
        AVLOG_WARN("substr-out of range error");
    }

    return out;
}

std::ostream& operator<<(std::ostream& out, const otimeDuration& obj)
{
    char buf[120];
    const otimeDuration tmp(obj);
    const struct tm timeOd = tmp.to_tm();
    otimeDuration fracSec(obj);
    fracSec -= oSeconds(tmp.totalSeconds());

    // Discard return value
    static_cast<void>(snprintf(static_cast<char*>(buf),
                               120U,
                               "%d:%02d:%02d.%06d",
                               timeOd.tm_hour,
                               timeOd.tm_min,
                               timeOd.tm_sec,
                               static_cast<int32_t>(fracSec.totalMicroSeconds())));
    std::string output = buf;
    out << output;

    return out;
}

// TODO [Matt Ivester 2/13/2010]: all of this should be able to go in the header but some of it breaks ABI
// compatibility. :(
bool otimePoint::operator==(const otimePoint& rhs) const
{
    return this->time_ == rhs.time_;
}

bool otimePoint::operator<(const otimePoint& rhs) const
{
    return this->time_ < rhs.time_;
}

bool otimePoint::operator>(const otimePoint& rhs) const
{
    return this->time_ > rhs.time_;
}

otimePoint& otimePoint::operator+=(const otimeDuration& rhs)
{
    this->time_ += rhs.time_;
    static_cast<void>(normalize()); // have to explicitly normalize so we invalidate if it's bad
    return *this;
}

otimePoint& otimePoint::operator-=(const otimeDuration& rhs)
{
    this->time_ -= rhs.time_;
    static_cast<void>(normalize()); // have to explicitly normalize so we invalidate if it's bad
    return *this;
}

bool otimeDuration::isNegative(void) const
{
    return time_.tv_sec < 0;
}

otimeDuration& otimeDuration::operator=(const otimePoint& timePoint)
{
    time_ = timePoint.time_;
    return *this;
}

otimeDuration& otimeDuration::operator=(otimePoint&& timePoint)
{
    time_ = std::move(timePoint.time_);
    return *this;
}

bool otimeDuration::operator==(const otimeDuration& rhs) const
{
    return this->time_ == rhs.time_;
}

bool otimeDuration::operator<(const otimeDuration& rhs) const
{
    return this->time_ < rhs.time_;
}

bool otimeDuration::operator>(const otimeDuration& rhs) const
{
    return this->time_ > rhs.time_;
}

otimeDuration otimeDuration::operator-(void) const
{
    otimeDuration otd;
    otd.time_ = -this->time_;
    return otd;
}

otimeDuration& otimeDuration::operator+=(const otimePoint& rhs)
{
    this->time_ += rhs.time_;
    return *this;
}

otimeDuration& otimeDuration::operator-=(const otimePoint& rhs)
{
    this->time_ -= rhs.time_;
    return *this;
}

otimeDuration& otimeDuration::operator+=(const otimeDuration& rhs)
{
    this->time_ += rhs.time_;
    return *this;
}

otimeDuration& otimeDuration::operator-=(const otimeDuration& rhs)
{
    this->time_ -= rhs.time_;
    return *this;
}

// TODO [Matt Ivester 2/19/2020]: these were also defined wrong, lhs parameter
// should be a 'const otimePoint&'.  Keeping for ABI compatibility.
otimeDuration operator+(otimePoint lhs, const otimePoint& rhs)
{
    otimeDuration otd(lhs);
    otd += rhs;
    return otd;
}

otimeDuration operator-(otimePoint lhs, const otimePoint& rhs)
{
    otimeDuration otd(lhs);
    otd -= rhs;
    return otd;
}

otimePoint operator+(otimePoint lhs, const otimeDuration& rhs)
{
    otimePoint otp(lhs);
    otp += rhs;
    return otp;
}

otimePoint operator-(otimePoint lhs, const otimeDuration& rhs)
{
    otimePoint otp(lhs);
    otp -= rhs;
    return otp;
}

otimeDuration operator+(const otimeDuration& lhs, const otimeDuration& rhs)
{
    otimeDuration otd(lhs);
    otd += rhs;
    return otd;
}

otimeDuration operator-(const otimeDuration& lhs, const otimeDuration& rhs)
{
    otimeDuration otd(lhs);
    otd -= rhs;
    return otd;
}

bool otimePoint::isTimeValid(void) const
{
    return time_.isValidAsTimestamp();
}

void otimePoint::invalidate(void)
{
    time_.invalidateAsTimestamp();
}

bool otimePoint::normalize(void)
{
    const bool normResult = time_.normalize();
    if (!normResult) {
        invalidate();
    }
    return normResult;
}

void otimeDuration::normalize(void)
{
    static_cast<void>(time_.normalize());
}

// TODO [Matt Ivester 2/12/2020]: trying to keep ABI compatibility... used to have
// these clunky (and non-const-reference-taking) ctors instead of just allowing
// copy construction as an otimeDuration...
otimeDuration::otimeDuration(const oSeconds& t)
    : time_{t.time_}
{}
otimeDuration::otimeDuration(const oMilliSeconds& t)
    : time_{t.time_}
{}
otimeDuration::otimeDuration(const oMicroSeconds& t)
    : time_{t.time_}
{}
otimeDuration::otimeDuration(const oNanoSeconds& t)
    : time_{t.time_}
{}

// TODO [Matt Ivester 2/13/2020]: these should be defaulted...
otimeDuration& otimeDuration::operator=(const otimeDuration& rhs)
{
    this->time_ = rhs.time_;
    return *this;
}

otimeDuration& otimeDuration::operator=(otimeDuration&& rhs)
{
    this->time_ = rhs.time_;
    return *this;
}

otimePoint& otimePoint::operator=(const otimePoint& rhs)
{
    this->time_ = rhs.time_;
    return *this;
}

otimePoint& otimePoint::operator=(otimePoint&& rhs)
{
    this->time_ = rhs.time_;
    return *this;
}

} // namespace ottotime

#endif // OTTOTIME_CC
