/*!
 *****************************************************************************
 * @file      testTimeStamp.cc
 * @author    Junsung Kim
 * @date      02/16/2016
 * @attention Unit test for otime
 *
 * @attention Copyright (c) 2016- Delphi Automotive. All rights reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Yunpeng Xu (yunpeng.xu@aptiv.com)
 * @date 03/29/2018
 * @remark port TimeStamp part to QNX.
 *         - replace high_resolution_clock with system_clock to fix build
 *           issue for QNX
 *
 *****************************************************************************
 */

#include <TimeStamp/otime.h>

#include <chrono>
#include <ctime>
#include <stdlib.h>

#include <Clock/SystemClockUtils.h>
#include <TimeStamp/Stopwatch.h>
#include <TimeStamp/TimeStamp.h>
#include <TimeStamp/otimeConversions.h>

#include <gtest/gtest.h>

using namespace ottotime;

constexpr auto maxNanoseconds = av::Duration::max();

TEST(TestTimeStamp, durationFromOtimeDuration)
{
    EXPECT_EQ(av::Duration{}, av::durationFromOtimeDuration(otimeDuration{}));
    EXPECT_EQ(av::Duration{1}, av::durationFromOtimeDuration(oNanoSeconds(1)));
    EXPECT_EQ(maxNanoseconds, av::durationFromOtimeDuration(oNanoSeconds(maxNanoseconds.count())));
}

TEST(TestTimeStamp, timePointFromOtimePoint)
{
    EXPECT_EQ(av::TimePoint{}, av::timePointFromOtimePoint(otimePoint{oNanoSeconds(0)}));
    EXPECT_EQ(av::TimePoint{av::Duration{1}}, av::timePointFromOtimePoint(otimePoint{oNanoSeconds(1)}));
    EXPECT_EQ(maxNanoseconds.count(), oNanoSeconds(maxNanoseconds.count()).totalNanoSeconds());
}

TEST(TestTimeStamp, timePointFromOtimeSpec)
{
    EXPECT_EQ(av::TimePoint{}, av::timePointFromOtimeSpec(otimeSpec{0, 0}));
    EXPECT_EQ(av::TimePoint{av::Duration{1}}, av::timePointFromOtimeSpec(otimeSpec{0, 1}));
    otimeSpec otsMax{0, maxNanoseconds.count()};
    otsMax.normalize();
    EXPECT_EQ(av::TimePoint{maxNanoseconds}, av::timePointFromOtimeSpec(otsMax));
}

TEST(TestTimeStamp, otimeDurationFromDuration)
{
    EXPECT_EQ(otimeDuration{}, av::otimeDurationFromDuration(av::Duration{}));
    EXPECT_EQ(oNanoSeconds(1), av::otimeDurationFromDuration(av::Duration{1}));
    EXPECT_EQ(oNanoSeconds(maxNanoseconds.count()), av::otimeDurationFromDuration(maxNanoseconds));
}

TEST(TestTimeStamp, otimePointFromTimePoint)
{
    EXPECT_EQ(otimePoint{oNanoSeconds(0)}, av::otimePointFromTimePoint(av::TimePoint{}));
    EXPECT_EQ(otimePoint{oNanoSeconds(1)}, av::otimePointFromTimePoint(av::TimePoint{av::Duration{1}}));
    EXPECT_EQ(otimePoint{oNanoSeconds(maxNanoseconds.count())},
              av::otimePointFromTimePoint(av::TimePoint{maxNanoseconds}));
}

TEST(TestTimeStamp, otimeSpecFromTimePoint)
{
    EXPECT_EQ((otimeSpec{0, 0}), av::otimeSpecFromTimePoint(av::TimePoint{}));
    EXPECT_EQ((otimeSpec{0, 1}), av::otimeSpecFromTimePoint(av::TimePoint{av::Duration{1}}));
    otimeSpec otsMax{0, maxNanoseconds.count()};
    otsMax.normalize();
    EXPECT_EQ(otsMax, av::otimeSpecFromTimePoint(av::TimePoint{maxNanoseconds}));
}

// disabled due to timezone information in the docker container not set properly
TEST(TestTimeStamp, DISABLED_Original_Test)
{
    using std::chrono::system_clock;

    long timeChange = 220;

    /**
     * From ISO C++ standard:
     * system_clock represent wall clock time from the system-wide realtime clock.
     * high_resolution_clock represent clocks with the shortest tick period.
     * high_resolution_clock may be a synonym for system_clock or steady_clock
     *
     * From gcc 5.3.0:
     * high_resolution_clock is an alias for system_clock
     *
     * Conclusion: right now there is really no difference between those clocks
     * in implementation of popular compilers
     *
     * To fix the compilation issue on QNX platform, use system_clock instead of high_resolution_clock
     */
    std::chrono::system_clock::time_point today = system_clock::now();

    std::chrono::nanoseconds ns = std::chrono::duration_cast<std::chrono::nanoseconds>(today.time_since_epoch());
    std::chrono::seconds s = std::chrono::duration_cast<std::chrono::seconds>(ns);
    std::chrono::microseconds ms = std::chrono::microseconds(timeChange);

    std::time_t t = s.count();
    long fractional_seconds = ns.count() % 1000000000;

    std::time_t tt;

    tt = system_clock::to_time_t(today);
    std::cout << "Today is: " << ctime(&tt);

    struct timespec ts1, ts2;
    ts1.tv_sec = t;
    ts1.tv_nsec = fractional_seconds;

    ns += ms;
    ts2.tv_sec = std::chrono::duration_cast<std::chrono::seconds>(ns).count();
    ts2.tv_nsec = ns.count() % 1000000000;

    // test otimePoint
    // test constructors
    otimePoint otp1(ts1);
    std::cout << "otp1: " << otp1;

    // test utility functions
    otimePoint otp2(ts2);
    std::cout << "otp2: " << otp2;

    otimePoint otp3;
    otp3 = otp1;
    if (otp1 == otp3)
        std::cout << "otp1 and otp3 are the same." << std::endl;

    if (otp3.isTimeValid())
        std::cout << "otp3 is valid." << std::endl;
    otp3.invalidate();
    if (!otp3.isTimeValid())
        std::cout << "otp3 is not valid." << std::endl;
    otp3 = otp1;

    if (otp1 < otp2)
        std::cout << "otp2 is later than otp1." << std::endl;
    if (otp2 > otp1)
        std::cout << "otp2 is later than otp1." << std::endl;

    oMicroSeconds oms(timeChange);
    otimeDuration otd1(oms);

    otp3 = otp2 - otd1;

    std::cout << "otp3: " << otp3;

    if (otp1 == otp3)
        std::cout << "otp1 and otp3 are the same." << std::endl;

    otp2 -= oSeconds(1);
    otimePoint otp4 = std::min(otp1, otp2);
    std::cout << "otp1: " << otp1;
    std::cout << "otp2: " << otp2;
    std::cout << "otp4: " << otp4;

    otimeDuration otd2;
    otd2 = otp1 + otp2;
    std::cout << "otd2: " << otd2 << std::endl;

    otd2 = otp2 - otp3;
    std::cout << "otd2: " << otd2 << std::endl;

    if (otd1 == otd2)
        std::cout << "otd1 and otd2 are the same." << std::endl;

    std::cout << otd2.totalSeconds() << std::endl;
    std::cout << otd2.totalMilliSeconds() << std::endl;
    std::cout << otd2.totalMicroSeconds() << std::endl;
    std::cout << (otp1 - otp3).totalNanoSeconds() << std::endl;
    std::cout << (-otd2).totalNanoSeconds() << std::endl;

    otimeDuration otd3, otd4;
    otd3 = oMilliSeconds(-500);
    otd4 = oMilliSeconds(-600);

    if (otd3 > otd4)
        std::cout << "otd3 is later than otd4." << std::endl;
    if (otd4 < otd3)
        std::cout << "otd3 is later than otd4." << std::endl;

    otd3 += oMilliSeconds(700);
    otd4 += oMilliSeconds(700);
    if (otd3 > otd4)
        std::cout << "otd3 is later than otd4." << std::endl;
    if (otd4 < otd3)
        std::cout << "otd3 is later than otd4." << std::endl;

    otimeDuration otd5 = std::min(otd3, otd4);
    std::cout << "otd4: " << otd4 << std::endl;
    std::cout << "otd5: " << otd5 << std::endl;

    std::cout << otd3 << std::endl;
    std::cout << otd4 << std::endl;
    // test operators

    // test otimeDuration
    // test constructors
    // test utility functions

    // test operators

    std::cout << convertTootime(18446744071500566528.) << std::endl;
}

TEST(TestTimeStamp, totalNanoSecondsTest)
{
    constexpr uint64_t utc_time_in_millisecond = 1574783392171; // utc time in 11/26/2019, 10:49:52 AM
    ottotime::otimeDuration duration{ottotime::oMilliSeconds(utc_time_in_millisecond)};
    ottotime::otimePoint sensorTimeStamp(duration);

    struct timespec expected_timespec = {1574783392, 171000000};
    ottotime::otimePoint expected_sensor_timestamp = ottotime::otimePoint{expected_timespec};

    EXPECT_EQ(expected_sensor_timestamp, sensorTimeStamp);

    constexpr uint64_t expected_total_nano_seconds = utc_time_in_millisecond * NSEC_PER_MSEC;
    EXPECT_EQ(expected_total_nano_seconds, duration.totalNanoSeconds());

    const long double expected_utc_time_in_second = 1574783392.171000000;
    EXPECT_DOUBLE_EQ(expected_utc_time_in_second, static_cast<double>(otimeToSeconds(sensorTimeStamp)));
}

TEST(TestTimeStamp, moreTotalNanoSecondsTest)
{
    uint64_t utc_time_in_millisecond = 0;

    for (int i = 0; i < 1000000; ++i) {
        utc_time_in_millisecond += 1;
        ottotime::otimeDuration duration{ottotime::oMilliSeconds(static_cast<int64_t>(utc_time_in_millisecond))};
        ottotime::otimePoint sensorTimeStamp(duration);

        const uint64_t expected_total_nano_seconds = utc_time_in_millisecond * NSEC_PER_MSEC;
        EXPECT_EQ(expected_total_nano_seconds, duration.totalNanoSeconds());

        const double expected_utc_time_in_second = static_cast<double>(utc_time_in_millisecond) / MSEC_PER_SEC;
        EXPECT_DOUBLE_EQ(expected_utc_time_in_second, static_cast<double>(otimeToSeconds(sensorTimeStamp)));
    }
}

TEST(TestTimeStamp, consistentInvalidTimeToTM)
{
    const otimePoint invalidTimePoint{};
    const struct tm asTM = invalidTimePoint.to_tm();

    EXPECT_EQ(asTM.tm_year, 69);
    EXPECT_EQ(asTM.tm_yday, 364);
    EXPECT_EQ(asTM.tm_mon, 11);
    EXPECT_EQ(asTM.tm_mday, 31);
    EXPECT_EQ(asTM.tm_hour, 23);
    EXPECT_EQ(asTM.tm_min, 59);
    EXPECT_EQ(asTM.tm_sec, 59);
}

TEST(TestTimeStamp, normalize)
{
    otimeSpec overflowed;
    overflowed.tv_sec = 0;
    overflowed.tv_nsec = 1234567890123456789LL;

    otimeSpec underflowed;
    underflowed.tv_sec = 0;
    underflowed.tv_nsec = -1234567890123456789LL;

    otimePoint time = overflowed;
    EXPECT_EQ(time.to_otimeSpec().tv_sec, 1234567890LL);
    EXPECT_EQ(time.to_otimeSpec().tv_nsec, 123456789LL);
    time = underflowed;
    EXPECT_EQ(time.to_otimeSpec().tv_sec, -1);
    EXPECT_EQ(time.to_otimeSpec().tv_nsec, -1);

    overflowed.normalize();
    underflowed.normalize();

    EXPECT_EQ(overflowed.tv_sec, 1234567890LL);
    EXPECT_EQ(overflowed.tv_nsec, 123456789LL);
    EXPECT_EQ(underflowed.tv_sec, -1234567890LL - 1LL);
    EXPECT_EQ(underflowed.tv_nsec, 1000000000LL - 123456789LL);
}

TEST(TestChronoConversion, testEpochIsUNIX)
{
    // test that our epoch time is 1/1/1970.
    const auto epochChronoPoint = std::chrono::system_clock::time_point{};
    const time_t epochChronoAsTimeT = std::chrono::system_clock::to_time_t(epochChronoPoint);

    struct tm* epochTM = gmtime(&epochChronoAsTimeT);
    EXPECT_TRUE(epochTM);

    EXPECT_EQ(epochTM->tm_year, 70);
    EXPECT_EQ(epochTM->tm_yday, 0);
    EXPECT_EQ(epochTM->tm_mon, 0);
    EXPECT_EQ(epochTM->tm_mday, 1);
    EXPECT_EQ(epochTM->tm_hour, 0);
    EXPECT_EQ(epochTM->tm_min, 0);
    EXPECT_EQ(epochTM->tm_sec, 0);
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock and sleep() behavior.
//                   This could behave inconsistently on, e.g. heavily loaded Jenkins instances.
TEST(TestStopwatch, DISABLED_basicStopwatch)
{
    // test that we can create a basic stopwatch and it works
    const ottotime::otimePoint testStart = ottotime::otimePoint::buildCurrentSteadyTime();
    ottotime::Stopwatch testWatch{ottotime::Stopwatch::UnderlyingClock::SteadyTimer};
    sleep(1);
    const ottotime::otimeDuration watchDuration = testWatch.getElapsedTime();
    const int64_t durationMillis = watchDuration.totalMilliSeconds();
    const ottotime::otimePoint steadyTime = ottotime::otimePoint::buildCurrentSteadyTime();

    const ottotime::otimePoint watchStartTime = testWatch.getStartTime();
    std::cout << "tried to sleep for 1s, reported sleep time: " << durationMillis << "ms" << std::endl;
    std::cout << "reported duration: " << watchDuration << std::endl;
    EXPECT_GE(durationMillis, 900);
    EXPECT_LE(durationMillis, 1100);

    const ottotime::otimeDuration watchStartDuration = steadyTime - watchStartTime;
    const int64_t watchStartMillis = watchStartDuration.totalMilliSeconds();
    std::cout << "reported time since start: " << watchStartMillis << "ms" << std::endl;
    std::cout << "reported duration since start: " << watchStartDuration << std::endl;
    EXPECT_GE(watchStartMillis, 900);
    EXPECT_LE(watchStartMillis, 1100);

    const ottotime::otimeDuration manualDuration = steadyTime - testStart;
    const int64_t manualMillis = manualDuration.totalMilliSeconds();
    std::cout << "manual time: " << manualMillis << "ms" << std::endl;
    std::cout << "manual duration: " << manualDuration << std::endl;
    EXPECT_GE(manualMillis, 900);
    EXPECT_LE(manualMillis, 1100);
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock and sleep() behavior.
//                   This could behave inconsistently on, e.g. heavily loaded Jenkins instances.
TEST(TestStopwatch, DISABLED_HRStopwatch)
{
    // test that we can create a high-resolution stopwatch and it works
    const ottotime::otimePoint testStart = ottotime::otimePoint::buildCurrentHighResolutionTime();
    ottotime::Stopwatch testWatch{ottotime::Stopwatch::UnderlyingClock::HighResolutionTimer};
    sleep(1);
    const ottotime::otimeDuration watchDuration = testWatch.getElapsedTime();
    const int64_t durationMillis = watchDuration.totalMilliSeconds();
    const ottotime::otimePoint hrTime = ottotime::otimePoint::buildCurrentHighResolutionTime();

    const ottotime::otimePoint watchStartTime = testWatch.getStartTime();
    std::cout << "tried to sleep for 1s, reported sleep time: " << durationMillis << "ms" << std::endl;
    std::cout << "reported duration: " << watchDuration << std::endl;
    EXPECT_GE(durationMillis, 900);
    EXPECT_LE(durationMillis, 1100);

    const ottotime::otimeDuration watchStartDuration = hrTime - watchStartTime;
    const int64_t watchStartMillis = watchStartDuration.totalMilliSeconds();
    std::cout << "reported time since start: " << watchStartMillis << "ms" << std::endl;
    std::cout << "reported duration since start: " << watchStartDuration << std::endl;
    EXPECT_GE(watchStartMillis, 900);
    EXPECT_LE(watchStartMillis, 1100);

    const ottotime::otimeDuration manualDuration = hrTime - testStart;
    const int64_t manualMillis = manualDuration.totalMilliSeconds();
    std::cout << "manual time: " << manualMillis << "ms" << std::endl;
    std::cout << "manual duration: " << manualDuration << std::endl;
    EXPECT_GE(manualMillis, 900);
    EXPECT_LE(manualMillis, 1100);
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock and sleep() behavior.
//                   This could behave inconsistently on, e.g. heavily loaded Jenkins instances.
TEST(TestStopwatch, DISABLED_resetBasicStopwatch)
{
    // test that Stopwatch::reset() works with a basic stopwatch
    const ottotime::otimePoint testStart = ottotime::otimePoint::buildCurrentSteadyTime();
    ottotime::Stopwatch testWatch{ottotime::Stopwatch::UnderlyingClock::SteadyTimer};
    sleep(1);
    testWatch.reset();
    sleep(1);
    const ottotime::otimeDuration watchDuration = testWatch.getElapsedTime();
    const int64_t durationMillis = watchDuration.totalMilliSeconds();
    const ottotime::otimePoint steadyTime = ottotime::otimePoint::buildCurrentSteadyTime();

    const ottotime::otimePoint watchStartTime = testWatch.getStartTime();
    std::cout << "tried to sleep for 1s after reset, reported sleep time: " << durationMillis << "ms" << std::endl;
    std::cout << "reported duration: " << watchDuration << std::endl;
    EXPECT_GE(durationMillis, 900);
    EXPECT_LE(durationMillis, 1100);

    const ottotime::otimeDuration watchStartDuration = steadyTime - watchStartTime;
    const int64_t watchStartMillis = watchStartDuration.totalMilliSeconds();
    std::cout << "reported time since start: " << watchStartMillis << "ms" << std::endl;
    std::cout << "reported duration since start: " << watchStartDuration << std::endl;
    EXPECT_GE(watchStartMillis, 900);
    EXPECT_LE(watchStartMillis, 1100);

    const ottotime::otimeDuration manualDuration = steadyTime - testStart;
    const int64_t manualMillis = manualDuration.totalMilliSeconds();
    std::cout << "manual time: " << manualMillis << "ms" << std::endl;
    std::cout << "manual duration: " << manualDuration << std::endl;
    EXPECT_GE(manualMillis, 1900);
    EXPECT_LE(manualMillis, 2100);
}

// TODO (SI-7313) -- we should fake/mock out the underlying otime clock and sleep() behavior.
//                   This could behave inconsistently on, e.g. heavily loaded Jenkins instances.
TEST(TestStopwatch, DISABLED_resetHRStopwatch)
{
    // test that Stopwatch::reset() works with a high-resolution stopwatch
    const ottotime::otimePoint testStart = ottotime::otimePoint::buildCurrentHighResolutionTime();
    ottotime::Stopwatch testWatch{ottotime::Stopwatch::UnderlyingClock::HighResolutionTimer};
    sleep(1);
    testWatch.reset();
    sleep(1);
    const ottotime::otimeDuration watchDuration = testWatch.getElapsedTime();
    const int64_t durationMillis = watchDuration.totalMilliSeconds();
    const ottotime::otimePoint hrTime = ottotime::otimePoint::buildCurrentHighResolutionTime();

    const ottotime::otimePoint watchStartTime = testWatch.getStartTime();
    std::cout << "tried to sleep for 1s after reset, reported sleep time: " << durationMillis << "ms" << std::endl;
    std::cout << "reported duration: " << watchDuration << std::endl;
    EXPECT_GE(durationMillis, 900);
    EXPECT_LE(durationMillis, 1100);

    const ottotime::otimeDuration watchStartDuration = hrTime - watchStartTime;
    const int64_t watchStartMillis = watchStartDuration.totalMilliSeconds();
    std::cout << "reported time since start: " << watchStartMillis << "ms" << std::endl;
    std::cout << "reported duration since start: " << watchStartDuration << std::endl;
    EXPECT_GE(watchStartMillis, 900);
    EXPECT_LE(watchStartMillis, 1100);

    const ottotime::otimeDuration manualDuration = hrTime - testStart;
    const int64_t manualMillis = manualDuration.totalMilliSeconds();
    std::cout << "manual time: " << manualMillis << "ms" << std::endl;
    std::cout << "manual duration: " << manualDuration << std::endl;
    EXPECT_GE(manualMillis, 1900);
    EXPECT_LE(manualMillis, 2100);
}

TEST(TestImplicitCtor, implicitTimeSpectoOTimePoint)
{
    // TODO [Matt Ivester 2/11/2020]: see comments in otime.h.
    // I'd like this to be illegal (should have to be "timePoint - ottotime::otimePoint(ottotime::otimeSpec{0,0})")
    // to discourage constructing otimeSpecs by hand rather than using the proper functions.
    // But we need to clean up all the places doing stuff like this first.
    const ottotime::otimePoint timePoint = ottotime::otimePoint::buildCurrentWallClockTime();
    const ottotime::otimeDuration timeDuration = timePoint - ottotime::otimeSpec{0, 0};
    const int64_t timePointMicros = timePoint.toMicrosecondsSinceEpoch();
    const int64_t timeDurationMicros = timeDuration.totalMicroSeconds();

    EXPECT_EQ(timePointMicros, timeDurationMicros);
}

TEST(TestImplicitCtor, implicitTimeSpecFromArgList)
{
    // TODO [Matt Ivester 2/11/2020]: see comments in otime.h.
    // I'd like this to be illegal (should have to be "ottotime::otimeSpec epochSpec{0, 0};"
    // or "ottotime::otimeSpec epochSpec = ottotime::otimeSpec{0,0};")
    // because this is extremely vague syntax.
    // But we need to clean up all the places doing stuff like this first.
    const ottotime::otimeSpec epochSpec = {0, 0};
    const ottotime::otimePoint epochTime{epochSpec};
    const ottotime::otimePoint nowTime = ottotime::otimePoint::buildCurrentWallClockTime();
    const ottotime::otimeDuration timeDuration = nowTime - epochTime;

    EXPECT_EQ(timeDuration.totalMicroSeconds(), nowTime.toMicrosecondsSinceEpoch());
}

// TODO: test the new functions to convert otimePoint to/from epoch time.

int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
