/*!
 ***********************************************************
 * @file ConfigParser.cc
 * @brief A file implementing the ConfigParser class
 *
 * The defined class is used to parse ruby configuration files
 * used by a Task.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Yunpeng Xu (yunpeng.xu@aptiv.com)
 * @date 03/16/2018
 * @remark port config part to qnx.
 *         - change script dependency from ruby to python
 *         - add split function
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 09/19/2017
 * @remark replacing ruby functions with libconfig++.
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * ConfigParser.cc
 * Chris Urmson (curmson@ri.cmu.edu)
 * August 22, 2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */

#ifndef CONFIGPARSER_CC
#define CONFIGPARSER_CC

#include <config/ConfigParser.h>

#include <array>
#include <libconfig.h++>
#include <vector>

#include "lib/avlog/avlog.hpp"

ConfigParser::ConfigParser(void)
{
    configInitialized_ = false;
    inputConfig_ = std::make_shared<libconfig::Config>();
    inputConfig_->setAutoConvert(true); // be more forgiving when e.g. saving an int to a float or vice-versa
}

ConfigParser::~ConfigParser() {}

std::string ConfigParser::generateLibConfigFile(const std::string& pyFile)
{
    std::string libConfigFile{""};
    const char* ottoConfigPtr{getenv("OTTO_CONFIG_DIR")};
    std::string ottoDir{""};
    if (ottoConfigPtr != nullptr) {
        ottoDir = std::string(ottoConfigPtr);
    }
    const std::string command("PYTHONPATH=" + ottoDir + " python3 -B " + pyFile);
    AVLOG_DEBUG("Executing Python config: {}", command);

    std::unique_ptr<std::FILE, decltype(&pclose)> pipe(popen(command.c_str(), "r"), pclose);
    if (pipe == nullptr) {
        AVLOG_ERROR("Failed to run configuration script: {}", command);
    } else {
        const size_t BUFFER_SIZE = 512U;
        std::array<char, BUFFER_SIZE> buffer{0};
        std::string result{""};
        // Gather the output of the script to get the config file path
        while (fgets(buffer.data(), static_cast<int32_t>(BUFFER_SIZE), pipe.get()) != nullptr) {
            result += std::string(buffer.data());
        }
        // Split the output into lines and grab the end
        std::stringstream ss{result};
        std::vector<std::string> lines{};
        std::string line{""};
        while (std::getline(ss, line)) {
            if (!line.empty()) {
                lines.push_back(line);
            }
        }
        if (lines.empty()) {
            AVLOG_ERROR("Got empty output when running configuration script command: {}", command);
        } else {
            const std::string lastLine = lines.back();

            // Extract the config file name from the output line
            // (line should look like "INFO:CONFIG:Wrote to configfile: /path/to/file.cfg")
            const std::string prefix("INFO:CONFIG:Wrote to config file: ");
            if (lastLine.find(prefix) == 0) {
                libConfigFile = lastLine.substr(prefix.length(), std::string::npos);
            } else {
                AVLOG_ERROR("Failed to parse config file path from \"{}\"", lastLine);
            }
        }
    }
    return libConfigFile;
}

bool ConfigParser::parseFile(const std::string& filename)
{
    bool retStatus{true};
    std::string configFile{""};

    if (fileExists(filename)) {
        if (filename.rfind(".py") == (filename.length() - 3U)) {
            configFile = generateLibConfigFile(filename);
        } else {
            configFile = filename;
        }
        try {
            inputConfig_->readFile(configFile.c_str());
            configurationFile_ = configFile;
            configInitialized_ = true;
        } catch (const libconfig::FileIOException& fioex) {
            static_cast<void>(fioex);
            AVLOG_WARN("I/O error while reading file: {}", configFile);
            retStatus = false;
        } catch (const libconfig::ParseException& pex) {
            AVLOG_WARN("Parse error at {} : {} - {}", pex.getFile(), pex.getLine(), pex.getError());
            retStatus = false;
        }
    } else {
        AVLOG_WARN("unable to find configuration file: {}", filename);
        retStatus = false;
    }
    return retStatus;
}

bool ConfigParser::parseString(const std::string& str)
{
    bool retStatus{true};
    try {
        inputConfig_->readString(str);
        configurationFile_ = "!!string_data!!";
        configInitialized_ = true;
    } catch (const libconfig::ParseException& pex) {
        AVLOG_WARN("Parse error at {} : {} - {}",
                   pex.getFile() ? pex.getFile() : "<no input file>",
                   pex.getLine(),
                   pex.getError());
        retStatus = false;
    }
    return retStatus;
}

bool ConfigParser::getSection(const std::string& name, ConfigSection& section)
{
    bool retStatus{false};
    if (!configInitialized_) {
        AVLOG_WARN("ConfigParser::inputConfig_ is not initialized");
    } else {
        libconfig::Setting& root = inputConfig_->getRoot();

        if (!root.exists(name.c_str())) {
            AVLOG_WARN("Section {} is undefined", name);
        } else {
            retStatus = section.init(name, inputConfig_);
        }
    }
    return retStatus;
}

bool ConfigParser::hasSection(const std::string& name)
{
    bool retStatus{false};
    if (!configInitialized_) {
        AVLOG_WARN("ConfigParser::inputConfig_ is not initialized");
    } else {
        libconfig::Setting& root = inputConfig_->getRoot();
        retStatus = root.exists(name.c_str());
    }
    return retStatus;
}

bool ConfigParser::isInterfacePlugin(const ConfigSection& section)
{
    return (section.hasParameter("implementation"));
}

void ConfigParser::getAllSectionsWithParamName(std::list<ConfigSection>& sections,
                                               ConfigSection& curSection,
                                               const std::string& paramName,
                                               const int32_t depth)
{
    const ConfigSection empty;
    std::list<ConfigSection> nextArray;
    std::list<ConfigSection>::iterator nextIt;

    const std::list<std::string> paramNames = curSection.getParameterNames();
    for (auto const& name : paramNames) {
        ConfigSection next;
        // try to get the given parameter as a configSection
        curSection.displayWarnings(false);

        // first try to get a single element....
        if (curSection.get(name, next, empty) == true) {
            if (next.hasParameter(paramName)) {
                // this is a section that has the param name, add it to the list
                sections.push_back(std::move(next));
            }

            // we've got a section, recurse if we aren't at depth 0
            if (depth) {
                // we only decrement the depth if the section we're evaluating is an interface
                getAllSectionsWithParamName(
                    sections, next, paramName, depth - static_cast<int32_t>(isInterfacePlugin(next)));
            }
        } else if (curSection.getArray(name, nextArray)) {
            // we've got an array of config sections...
            for (nextIt = nextArray.begin(); nextIt != nextArray.end(); ++nextIt) {
                getAllSectionsWithParamName(sections, *nextIt, paramName, depth);
                // check to see if the array element has the paramater we're interested in
                if (nextIt->hasParameter(paramName)) {
                    sections.push_back(*nextIt);
                }
            } // for
        }     // array of configSections
        else {
            // No actions
        }
    }
    curSection.displayWarnings(true);
}

#endif // CONFIGPARSER_CC
