/*!
 ***********************************************************
 * @file ConfigSection.cc
 * @brief A file implementing the ConfigSection class
 *
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 09/19/2017
 * @remark replacing ruby functions with libconfig++.
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * ConfigSection.cc
 * Chris Urmson (curmson@ri.cmu.edu)
 * August 22, 2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */
#ifndef CONFIGSECTION_CC
#define CONFIGSECTION_CC

#include <config/ConfigSection.h>

#include "limits"
#include "type_traits"

ConfigSection::ConfigSection()
{
    // Make a default empty config
    inputConfig_ = std::make_shared<libconfig::Config>();
}

std::string ConfigSection::getName() const
{
    return sectionName_;
}

void ConfigSection::displayWarnings(const bool state)
{
    silent_ = !state;
}

std::list<std::string> ConfigSection::getParameterNames() const
{
    std::list<std::string> keys;
    for (auto const& it : accessedElements_) {
        keys.push_back(it.first);
    }
    return keys;
}

bool ConfigSection::checkAllParametersUsed()
{
    bool allUsed(true);

    for (auto const& it : accessedElements_) {
        if (!it.second) {
            allUsed = false;
            AVLOG_WARN("parameter {} is unused in section {}", (it.first), sectionName_);
        }
    }
    return allUsed;
}

bool ConfigSection::hasParameter(const std::string& name) const
{
    return accessedElements_.find(name) != accessedElements_.end();
}

bool ConfigSection::init(const std::string& name, const ConfigPtr& cfg)
{
    bool status(true);
    // now we setup our access table
    accessedElements_.clear();

    // Never allow a null config
    if (!cfg) {
        return false;
    }

    std::list<std::string> keys;

    inputConfig_ = cfg;

    try {
        libconfig::Setting& child = inputConfig_->lookup(name.c_str());
        sectionName_ = name;
        settingPath_ = child.getPath();

        // The return value of getLength is int ; hence the len type is modified int32_t type
        // Refer libconfig.h++
        const int32_t len = child.getLength();

        for (int32_t ii = 0; ii < len; ii++) {
            if (child[ii].getName() != nullptr) {
                const std::string key = child[ii].getName();
                keys.push_back(key);
            }
        }
    } catch (const libconfig::SettingNotFoundException& setex) {
        AVLOG_WARN("{} not found", setex.getPath());
        status = false;
    }
    if (status) {
        // initialize all of the parameters in our ConfigSection as unaccessed
        std::list<std::string>::iterator keyIt;
        for (keyIt = keys.begin(); keyIt != keys.end(); ++keyIt) {
            accessedElements_[*keyIt] = false;
        }
    }
    return status;
}

std::ostream& operator<<(std::ostream& out, const ConfigSection& cs)
{
    out << "Section (" << cs.sectionName_ << ")";
    return out;
}

/*
 * @brief utility function to wrap exists with exception catching and access tracking
 *
 */
libconfig::Setting& ConfigSection::getSetting(const std::string& name, GetStatus& result) const
{
    // record that we've tried to access the parameter name
    accessedElements_[name] = true;
    result = GS_OK;
    libconfig::Setting& val = inputConfig_->lookup(settingPath_);
    try {
        if (!val.exists(name)) {
            if (!silent_) {
                AVLOG_INFO("{}[\"{}\"] is undefined", sectionName_, name);
            }
            result = GS_Undefined;
        }
    } catch (libconfig::SettingNotFoundException& snfex) {
        AVLOG_WARN("Can't find {}", snfex.getPath());
        result = GS_Undefined;
    }
    return val;
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, ConfigSection& val) const
{
    GetStatus status;
    libconfig::Setting& setting = getSetting(name, status);

    if (status == GS_OK) {
        if (setting.exists(name.c_str())) {
            // Discard return value
            static_cast<void>(val.init(settingPath_ + '.' + name, inputConfig_));
        } else {
            // check to make sure that the element isn't null
            if (!silent_) {
                AVLOG_INFO("{}[\"{}\"] is undefined", sectionName_, name);
            }
            status = GS_Undefined;
        }
    }
    return status;
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, bool& val) const
{
    GetStatus status;
    libconfig::Setting& setting = getSetting(name, status);

    if (status == GS_OK) {
        if (!setting.lookupValue(name.c_str(), val)) {
            // We know the setting exists, so it must not be the right type
            AVLOG_WARN("{}[\"{}\"] is the wrong type", sectionName_, name);
            status = GS_TypeMismatch;
        }
    }
    return status;
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, int8_t& val) const
{
    return getTemplated(name, val);
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, int16_t& val) const
{
    return getTemplated(name, val);
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, int32_t& val) const
{
    return getTemplated(name, val);
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, int64_t& val) const
{
    return getTemplated(name, val);
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, uint8_t& val) const
{
    return getTemplated(name, val);
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, uint16_t& val) const
{
    return getTemplated(name, val);
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, uint32_t& val) const
{
    return getTemplated(name, val);
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, uint64_t& val) const
{
    //  The templated version cannot handle the upper end check on this type.
    GetStatus status;
    libconfig::Setting& setting = getSetting(name, status);

    if (status == GS_OK) {
        // with autoConvert set to true in libconfig, negative values are truncated to 0 and values in
        // the top half of the range are truncated down to 0x7FFFFFFFFFFFFFFF
        // reading with an unsigned long long int.  This approach will stop silent truncations.
        // addressed by https://jira.ptc.delphiauto.net/atljira/browse/SI-3300
        long long temp; // signed temp allows detection of negative parameters
        if (!setting.lookupValue(name.c_str(), temp)) {
            // We know the setting exists, so it must not be the right type
            AVLOG_WARN("{}[\"{}\"] is the wrong type", sectionName_, name);
            status = GS_TypeMismatch;
        }
        // this would never be true if libconfig didn't round up to zero reading unsigned long long ints.
        else if (temp < 0) {
            const std::string message = sectionName_ + '[' + name + ']' + "is out of range at:" + std::to_string(temp);

            AVLOG_ERROR("{}", message);
            throw(std::out_of_range(message));
        } else {
            val = static_cast<uint64_t>(temp);
        }
    }
    return status;
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, float& val) const
{
    return getTemplated(name, val);
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, double& val) const
{
    return getTemplated(name, val);
}

ConfigSection::GetStatus ConfigSection::get(const std::string& name, std::string& val) const
{
    GetStatus status;
    libconfig::Setting& setting = getSetting(name, status);

    if (status == GS_OK) {
        if (!setting.lookupValue(name.c_str(), val)) {
            // We know the setting exists, so it must not be the right type
            AVLOG_WARN("{}[\"{}\"] is the wrong type", sectionName_, name);
            status = GS_TypeMismatch;
        }
    }
    return status;
}
// specialization for Type std::vector<uint16_t>.
int32_t ConfigSection::getArray(const std::string& name, std::vector<uint16_t>& dest) const
{
    int32_t retNumElems(0);
    bool bContinue(true);
    // record that we've tried to access the parameter name
    accessedElements_[name] = true;
    libconfig::Setting& setting = inputConfig_->lookup(settingPath_);

    if (setting.exists(name.c_str())) {
        libconfig::Setting& arrSetting = setting[name.c_str()];
        if (!arrSetting.isArray()) {
            if (!silent_) {
                AVLOG_WARN("{}[\"{}\"] is not defined as an array of values", sectionName_, name);
            }
            bContinue = false;
        }
    } else {
        if (!silent_) {
            AVLOG_WARN("{}[\"{}\"] is undefined", sectionName_, name);
        }
        bContinue = false;
    }

    if (bContinue) {
        dest.clear();
        std::back_insert_iterator<std::vector<uint16_t>> outit(dest);
        libconfig::Setting& arraySetting = setting[name.c_str()];
        const int32_t size = arraySetting.getLength();
        bool ok(true);
        int32_t numElems(0);

        for (numElems = 0; ok && (numElems < size); numElems++, ++outit) {
            uint16_t element(0);
            try {
                uint32_t temp(0);
                temp = arraySetting[numElems];
                element = static_cast<uint16_t>(temp);
            } catch (libconfig::SettingTypeException& setex) {
                static_cast<void>(setex);
                const std::string path = arraySetting.getPath() + "[" + std::to_string(numElems) + "]";
                AVLOG_WARN("{} is undefined", path);
                ok = false;
            }

            if (ok) {
                *outit = element;
            }
        }

        if (ok) {
            retNumElems = numElems;
        }
    }
    return retNumElems;
}

#endif // ifndef CONFIGSECTION_CC
