/*!
 ***********************************************************
 * @file ConfigParser.h
 * @brief ConfigParser class is used to parse ruby
 * configuration files and extract sections from that file.
 * Sections contain configuration values.
 *
 * @see ConfigSection
 *
 * The ConfigParser class is built on top of the Ruby/C api.
 * All configuration variables in a file must be stored in
 * hash tables in the file.  Hash tables are referred to as
 * Sections in the configuration system.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 09/19/2017
 * @remark replacing ruby functions with libconfig++.
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * ConfigParser.h
 * Chris Urmson (curmson@ri.cmu.edu)
 * Aug 22, 2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */
#ifndef CONFIGPARSER_H
#define CONFIGPARSER_H

#include "ConfigSection.h"
#include <iostream>
#include <list>
#include <string>
#include <unistd.h>

/*!
 * @class ConfigParser
 * @brief This class shall help in parsing the configuration file. The configuration file contains config sections which
 * \n shall be referred as hash tables.It provides methods to get the config sections from the configuration file.
 */
class ConfigParser
{

public:
    /*!
     * @brief Default Constructor. This shall initialize the parser. It shall create libConfig::Config object which will
     * be \n initialized with configuration file while parsing the file. It shall enable auto conversion.i.e Any
     * conversion involves \n int to float (or) float to int won't throw exception while parsing.
     */
    ConfigParser();

    /*!
     * @brief Virtual destructor. It's content in the configuration file becomes in accessible after the parser goes out
     * of scope.
     */
    virtual ~ConfigParser();

    /*!
     * @brief This method shall read and parse the configuration data from the given named file.
     * \n It shall initialize the config parser with the given file name.
     * \n \b Operational \b Flow:
     * \n IF the input file is python configuration file THEN
     * \n &emsp; Execute the python file and extract the configuration file name(.cfg)
     * \n &emsp; from the output obtained as a result of python file execution.
     * \n ELSE
     * \n &emsp; Use the given file as configuration file.
     * \n ENDIF
     * \n Read the configuration file and mark as config file initialization is successful.
     *
     * \n It handles the following two exceptions
     * \n FileIOException - In case of error while reading the file
     * \n ParseException  - In case of error while parsing the file
     *
     * @param[in] filename : The name of the configuration file to be loaded.The file must be somewhere in the search
     * path.
     *
     * @return bool : If the file exists and successfully reads the configuration, true. Otherwise, false.
     *
     * @sa ConfigParser::generateLibConfigFile
     *
     */
    bool parseFile(const std::string& filename);

    /*!
     * @brief This method shall read and parse the configuration data from the given string.
     * \n It handles ParseException if parser error.
     *
     * @param[in] str : A string containing configuration data, such as the contents of a .cfg file that have been read
     * into memory.
     *
     * @return bool : If the string contains valid code and successfully reads the configuration, true. Otherwise,
     * false.
     */
    bool parseString(const std::string& str);

    /*!
     * @brief This method shall initialize a config section for a given name.ConfigSections will be represented as
     * HashTables \n in the configuration file \n \b Operational \b Flow: \n IF the ConfigParser is initialized THEN \n
     * &emsp; Look up for the given name in the configuration. \n &emsp; IF the name doesn't exist in the configuration
     * THEN \n &emsp; &emsp; Log a warning message that section is undefined. \n &emsp; ELSE \n &emsp;&emsp;  Given
     * config section object shall be initialized. \n &emsp; END IF \n ELSE \n &emsp; Log a warning message that
     * ConfigParser is not initialized. \n END IF \n Example: config section looks like below in the configuration file
     * \n 	'test_section_1' = {"Bool_Parameter_1": True, "Uint_Parameter": 567 }
     * \n 	'test_section_2' = {"Bool_Parameter_1": True, "Uint_Parameter": 567 }
     * \n 	input   -  name : 'test_section_1'
     * \n 	output  -  config section object shall be initialized with the details of config section  'test_section_1'
     *
     * @param[in]  name : The name of the hash table in the file which we call a Configuration Section
     * @param[out] section : The structure to be initialized from the configuration file
     *
     * @return bool : If the named configuration section hash exists in the config file, true. Otherwise false.
     *
     * @sa ConfigSection::init
     *
     */
    bool getSection(const std::string& name, ConfigSection& section);

    /*!
     * @brief This method shall check if a ConfigSection given by name exists in the config file provided that
     * \n ConfigPareser is initialized. It shall log a warning message if the  ConfigParser is not initialized.
     *
     * @param[in] name : section name to look up.
     *
     * @return bool : True, If the section name exists in the config file. False, otherwise.
     *
     * \n Note: the same functionality can be generated by using getSection, but this call
     * \n will not generate warnings in the log stream if the name is undefined
     */
    bool hasSection(const std::string& name);

    /**
     * @brief This method shall return true if the passed config section represents an interface plugin and loadable by
     * the InterfaceDb. \n This could either be a messaging Interface or an OttoView widget. \n The following is the
     * sample config section for the interface in configuration file. \n  LeftEBoxSickInterface = { \n 	"interface" =>
     * "LineScanner", \n 	"implementation" => "SickLMS", \n 	"channel" => LeftEBoxSickIpcChannel, \n 	"port" =>
     * "/dev/ttyS2", \n 	"baud" => "115200", \n 	"flow_control" => "none", \n 	"clippingDistance_m" => 15 \n } \n
     * For the above example, true shall be returned as config section has a parameter named "implementation"
     *
     * @param[in] section : config section name
     *
     * @return bool : If section is an interface, true. Otherwise false.
     */
    static bool isInterfacePlugin(const ConfigSection& section);

    /**
     * @brief This method shall return all config sections within a given config section 'curSection' that contain a
     * given parameter name \n 'paramName'  up to some depth. \n \b Operational \b Flow: \n First,it shall get list of
     * all the config section name parameters at root level within a given config section 'curSection' \n It retrieves
     * 'config section' for each name in the list \n IF the 'config section' is single element THEN \n &emsp; Add
     * 'config section' to return list if it has a given parameter. \n &emsp; Until depth becomes zero, \n &emsp;
     * Traverse 'config section' recursively  by calling the method getAllSectionsWithParamName \n &emsp; with
     * 'sections','config section','paramName','depth' with value decremented by 1 if the  'config section' is
     * represents an interface. \n &emsp;\n \n ELSE IF config section is a collection (array) THEN \n &emsp; For each
     * config section object  in collection(array) \n &emsp;&emsp; Traverse recursively getAllSectionsWithParamName with
     * 'sections','config section','paramName','depth'. \n &emsp;&emsp; Add config section to return list if it has a
     * given parameter. \n &emsp; END FOR \n END IF
     *
     * @param[out] sections as list container: The config sections list containing the passed parameter name are
     * returned in
     *
     * @param[in] curSection as ConfigSection object:  - The section that we are considering parameters within
     *
     * @param[in] paramName as string:- The paramName the sections must contain
     *
     * @param[in] depth as integer: The maximum depth relative to the given current section to which we'll look for the
     * parameters.
     *
     * \n If depth is negative, the whole tree is recursed
     *
     * @return void
     */
    void getAllSectionsWithParamName(std::list<ConfigSection>& sections,
                                     ConfigSection& curSection,
                                     const std::string& paramName,
                                     const int32_t depth);

protected:
    friend class ConfigSection; ///< @see ConfigSection

    /*!
     * @brief This methods shall return configuration file name.This file shall be read and parsed by the ConfigParser.
     *
     * @return string : Name of the configuration file.
     */
    std::string getConfigFile() const { return configurationFile_; }

    /*!
     * @brief This method shall extract libconfig file name from a given Python file.
     * \n \b Operational \b Flow:
     * \n It shall execute the given Python file and extracts the configuration file from the output obtained
     * \n as a result of python file execution.
     * \n The output obtained as a result of Python file execution will have content like   .. "INFO:CONFIG:Wrote to
     * config file:  /path/to/file.cfg" \n It scans config file and returns lib config file as "/path/to/file.cfg"
     *
     * @param[in] pyFile : Python config file
     *
     * @return string : libconfig file with complete path  or empty string on error
     */
    std::string generateLibConfigFile(const std::string& pyFile);

    bool configInitialized_; ///< This variable allows to determine if the config parser is initialized.
                             ///< It becomes true, if the configuration represented as a file or string is parsed.
    ConfigPtr inputConfig_;  ///< libconfig::Config pointer shall be used to read the config file.
    std::string configurationFile_; ///< Name of the configuration file shall be provided with full path to
                                    ///< ConfigParser. Configuration data shall be read from this file.

private:
    /*!
     * @brief This methods shall check if the the given file exists or not
     *
     * @param[in] name : file name in string format
     *
     * @return bool : If file exists,true.Otherwise,false
     */
    inline bool fileExists(const std::string& name) const { return (access(name.c_str(), F_OK) != -1); }
};

#endif // CONFIGPARSER_H
