
/*!
 ***********************************************************
 * @file ConfigSection.def.h
 *
 * @brief ConfigSection is the interface to config sections
 * (hash tables) stored in a ruby configuration file.
 *
 * Configuration section is used to read standard types and
 * sub sections from a ruby configuration file.  There are
 * interfaces to read individual values, arrays of
 * values and STL containers of values from the files.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 09/19/2017
 * @remark replacing ruby functions with libconfig++.
 *
 ***********************************************************
 */
/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/
/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

 This source code was developed in part with support from
 sponsors including General Motors, the National Science
 Foundation and the US Department of Transportation.
 Use of this software is strictly governed by agreements
 between Carnegie Mellon University and various sponsors.
 Users of this software must also be fully aware and agree
 that Carnegie Mellon does not guarantee the correct
 functioning of this software in any system. Carnegie
 Mellon or any of its affiliates will not be liable for any
 damage and/or penalties resulting from the use of this
 software. Any user of this software takes complete
 responsibility for the use of any software or design.
 **********************************************************/
/*
 * ConfigSection.def.h
 * Chris Urmson (curmson@ri.cmu.edu)
 *
 * Copyright (c) 2007
 * Carnegie Mellon University
 * All rights reserved.
 */
#ifndef CONFIGSECTION_DEF_H
#define CONFIGSECTION_DEF_H

#ifndef CONFIGSECTION_H
#error "This file must be included from ConfigSection.h"
#endif

#include <stdexcept>

template<>
struct fmt::formatter<ConfigSection>
{
    template<typename ParseContext>
    constexpr auto parse(ParseContext& ctx)
    {
        return ctx.begin();
    }

    template<typename FormatContext>
    auto format(ConfigSection const& cs, FormatContext& ctx)
    {
        return fmt::format_to(ctx.out(), "Section ({})", cs.getName());
    }
};

template<class Type>
typename std::enable_if<std::is_integral<Type>::value, ConfigSection::GetStatus>::type ConfigSection::getTemplated(
    const std::string& name,
    Type& val) const
{
    GetStatus status;
    Setting& setting = getSetting(name, status);

    if (status == GS_OK) {
        long long int temp; // make sure a truncation or overflow is not missed.
        if (setting.lookupValue(name.c_str(), temp) == false) {
            // We know the setting exists, so it must not be the right type
            AVLOG_WARN("{}[\"{}\"] is the wrong type", sectionName_, name);
            status = GS_TypeMismatch;
        } else if ((temp > std::numeric_limits<Type>::max()) || (temp < std::numeric_limits<Type>::min())) {
            const std::string message = sectionName_ + '[' + name + ']' + "is out of range at:" + std::to_string(temp);
            AVLOG_ERROR("{}", message);
            throw(std::out_of_range(message));
            status = GS_RangeError;
        } else {
            val = temp;
        }
    }
    return status;
}

template<class Type>
typename std::enable_if<std::is_floating_point<Type>::value, ConfigSection::GetStatus>::type
ConfigSection::getTemplated(const std::string& name, Type& val) const
{
    GetStatus status;
    Setting& setting = getSetting(name, status);

    if (status == GS_OK) {
        double temp{0.0};
        if (setting.lookupValue(name.c_str(), temp) == false) {
            // We know the setting exists, so it must not be the right type
            AVLOG_WARN("{}[\"{}\"] is the wrong type", sectionName_, name);
            status = GS_TypeMismatch;
        } else if ((temp > std::numeric_limits<Type>::max()) || (temp < std::numeric_limits<Type>::lowest())) {
            const std::string message = sectionName_ + '[' + name + ']' + "is out of range at:" + std::to_string(temp);
            AVLOG_ERROR("{}", message);
            throw(std::out_of_range(message));
            status = GS_RangeError;
        } else {
            val = temp;
        }
    }
    return status;
}

template<class Type, class DefType>
bool ConfigSection::get(const std::string& name, Type& val, const DefType& def) const
{
    const GetStatus status = get(name, val);

    if (status != GS_OK) {
        if (!silent_) {
            AVLOG_INFO("{} [\"{}\"] using default {}", sectionName_, name, def);
        }
        val = static_cast<Type>(def);
    }

    return ((status == GS_OK) || (status == GS_Undefined));
}

// specialization for Container having <std::string>.
template<class Container>
typename std::enable_if<std::is_same<std::string, typename Container::value_type>::value, int32_t>::type
ConfigSection::getArray(const std::string& name, Container& dest) const
{
    int32_t retNumElems{0};
    bool bContinue{true};
    // record that we've tried to access the parameter name
    accessedElements_[name] = true;

    Setting& setting = inputConfig_->lookup(settingPath_);
    if (setting.exists(name.c_str()) == true) {
        Setting& arraySetting = setting[name.c_str()];
        if (arraySetting.isArray() == false) {
            if (!silent_) {
                AVLOG_WARN("{}[\"{}\"] is not defined as an array of values", sectionName_, name);
            }
            bContinue = false;
        }
    } else {
        if (!silent_) {
            AVLOG_WARN("{}[\"{}\"] is undefined", sectionName_, name);
        }
        bContinue = false;
    }
    if (bContinue) {
        dest.clear();
        std::back_insert_iterator<Container> outit(dest);
        Setting& arraySetting = setting[name.c_str()];
        const int32_t size = arraySetting.getLength();
        bool ok{true};
        int32_t numElems{0};

        for (numElems = 0; ok && (numElems < size); numElems++, ++outit) {
            std::string element{""};

            try {
                element = arraySetting[numElems].c_str();
            } catch (SettingTypeException& setex) {
                static_cast<void>(setex);
                const std::string path = arraySetting.getPath() + "[" + std::to_string(numElems) + "]";
                AVLOG_WARN("{} is undefined", path);
                ok = false;
            }

            if (ok) {
                *outit = element;
            }
        }

        if (ok) {
            retNumElems = numElems;
        }
    }
    return retNumElems;
}

template<class T>
typename std::enable_if<std::is_integral<T>::value || std::is_floating_point<T>::value,
                        std::pair<std::size_t, std::size_t>>::type
ConfigSection::getMatrix(const std::string& name, std::vector<std::vector<T>>& dest) const
{
    std::pair<std::size_t, std::size_t> retPair(0, 0);

    dest.clear();

    ConfigSection matrix_cs;
    if (get(name, matrix_cs) == GS_OK) {
        accessedElements_.at(name) = true;

        std::size_t nrows;
        std::size_t ncols;
        if ((matrix_cs.get("rows", nrows) == GS_OK) && (matrix_cs.get("cols", ncols) == GS_OK)) {
            std::vector<T> flat_dest;
            matrix_cs.getArray("data", flat_dest);
            if (flat_dest.size() == (nrows * ncols)) {
                dest.resize(nrows);
                auto iter = flat_dest.cbegin();
                for (std::size_t r = 0; r < nrows; ++r) {
                    std::copy(iter, iter + static_cast<int>(ncols), std::back_inserter(dest.at(r)));
                    iter += static_cast<int>(ncols);
                }

                retPair = std::pair<std::size_t, std::size_t>(nrows, ncols);
            }
        }
    }
    return retPair;
}

// Array of config sections
template<class Container>
typename std::enable_if<std::is_same<ConfigSection, typename Container::value_type>::value, int32_t>::type
ConfigSection::getArray(const std::string& name, Container& dest) const
{
    int32_t retNumElems{0};
    bool bContinue{true};
    // record that we've tried to access the parameter name
    accessedElements_[name] = true;

    Setting& setting = inputConfig_->lookup(settingPath_);
    if (setting.exists(name.c_str()) == true) {
        Setting& arraySetting = setting[name.c_str()];
        if (arraySetting.isAggregate() == false) {
            if (!silent_) {
                AVLOG_WARN("{}[\"{}\"] is not defined as an array of values", sectionName_, name);
            }
            bContinue = false;
        }
    } else {
        if (!silent_) {
            AVLOG_WARN("{}[\"{}\"] is undefined", sectionName_, name);
        }
        bContinue = false;
    }
    if (bContinue) {
        dest.clear();
        std::back_insert_iterator<Container> outit(dest);
        Setting& arraySetting = setting[name.c_str()];
        const int32_t size = arraySetting.getLength();
        bool ok{true};
        int32_t numElems{0};

        for (numElems = 0; ok && (numElems < size); numElems++, ++outit) {
            ConfigSection section;

            try {
                std::string src(setting.getSourceFile());
                static_cast<void>(section.init(arraySetting[numElems].getPath(), inputConfig_));
            } catch (SettingTypeException& setex) {
                static_cast<void>(setex);
                AVLOG_WARN("{}[{}] is undefined", arraySetting.getPath(), numElems);
                ok = false;
            }

            if (ok) {
                *outit = section;
            }
        }

        if (ok) {
            retNumElems = numElems;
        }
    }
    return retNumElems;
}

template<class Container>
typename std::enable_if<std::is_integral<typename Container::value_type>::value ||
                            std::is_floating_point<typename Container::value_type>::value,
                        int32_t>::type
ConfigSection::getArray(const std::string& name, Container& dest) const
{
    int32_t retNumElems{0};
    bool bContinue{true};
    // record that we've tried to access the parameter name
    accessedElements_[name] = true;

    Setting& setting = inputConfig_->lookup(settingPath_);
    if (setting.exists(name.c_str()) == true) {
        Setting& arraySetting = setting[name.c_str()];
        if (arraySetting.isAggregate() == false) {
            if (!silent_) {
                AVLOG_WARN("{}[\"{}\"] is not defined as an array of values", sectionName_, name);
            }
            bContinue = false;
        }
    } else {
        if (!silent_) {
            AVLOG_WARN("{}[\"{}\"] is undefined", sectionName_, name);
        }
        bContinue = false;
    }
    if (bContinue) {
        dest.clear();
        std::back_insert_iterator<Container> outit(dest);
        Setting& arraySetting = setting[name.c_str()];
        const int32_t size = arraySetting.getLength();
        bool ok{true};
        int32_t numElems{0};

        for (numElems = 0; ok && (numElems < size); numElems++, ++outit) {
            typename Container::value_type element;

            try {
                element = arraySetting[numElems];
            } catch (SettingTypeException& setex) {
                static_cast<void>(setex);
                AVLOG_WARN("{}[{}] is undefined", arraySetting.getPath(), numElems);
                ok = false;
            }

            if (ok) {
                *outit = element;
            }
        }

        if (ok) {
            retNumElems = numElems;
        }
    }
    return retNumElems;
}

#endif // CONFIGSECTION_DEF_H
