/*!
 ***********************************************************
 * @file ConfigSection.h
 *
 * @brief ConfigSection is the interface to config sections
 * (hash tables) stored in a ruby configuration file.
 *
 * Configuration section is used to read standard types and
 * sub sections from a ruby configuration file.  There are
 * interfaces to read individual values, arrays of
 * values and STL containers of values from the files.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Vasudeva Pai (vasudeva.pai@delphi.com)
 * @date 09/19/2017
 * @remark replacing ruby functions with libconfig++.
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * ConfigSection.h
 * Chris Urmson (curmson@ri.cmu.edu)
 * August 22, 2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */
#ifndef CONFIGSECTION_H
#define CONFIGSECTION_H

#include "lib/avlog/avlog.hpp"

#include <algorithm>
#include <iterator>
#include <libconfig.h++>
#include <limits>
#include <list>
#include <map>
#include <sstream>
#include <string>
#include <type_traits>
#include <typeinfo>
#include <vector>

using namespace libconfig;

// Forward declarations
class ConfigParser;
class ConfigSection;

// Convenience typedefs
using ConfigPtr = std::shared_ptr<Config>;
using ConfigSectionList = std::list<ConfigSection>;
using ConfigSectionListIter = ConfigSectionList::iterator;

/*!
 * @class ConfigSection
 * @brief ConfigSection is the interface to config sections (hash tables) stored in a configuration file.
 * \n ConfigParser class shall be used to read the ConfigSection from configuration file. The following is the
 * \n sample configuration file:
 * \n ReadWriteInterfaces = {
 * \n "leftSick" => Interfaces::LeftEBoxSickInterface,
 * \n "rightSick" => Interfaces::RightEBoxSickInterface
 * \n }
 * \n In the above example config section object represents "ReadWriteInterfaces" section.
 */
class ConfigSection
{
public:
    /*!
     * @brief Default constructor.It shall initializes the ConfigSection to be empty.
     */
    explicit ConfigSection();

    /*!
     * @brief Copy Constructor. It shall construct the ConfigSection object by using the contents of the given
     * ConfigSection object.
     *
     * @param[in] copy : Reference of the ConfigSection object
     */
    ConfigSection(const ConfigSection& copy) = default;

    /*!
     * @brief Overloaded assignment operator method. This method shall assign the values of the given ConfigSection
     * object \n into the destination ConfigSection object(LHS).
     *
     * @param[in] copy reference of the ConfigSection object
     *
     * @return ConfigSection object after  assignment.
     */
    ConfigSection& operator=(const ConfigSection& copy) = default;

    // default move ctor and assignment
    ConfigSection(ConfigSection&&) = default;
    ConfigSection& operator=(ConfigSection&&) = default;

    /*!
     * @brief virtual destructor
     */
    virtual ~ConfigSection() = default;

    /*!
     * @enum GetStatus
     *
     * @brief enum for status of reading values from config
     */
    enum GetStatus
    {
        GS_OK,
        GS_Undefined,
        GS_TypeMismatch,
        GS_RangeError // value in config file cannot be represented by the provided variable type.
    };

    /*!
     * This method shall initialize the config section object for a given section name if it exists in the current
     * config section. \n \b Operational \b Flow: \n First, it shall get the current config section. \n IF the given
     * section name exists in the current config section(parent) THEN \n &emsp; Given config object shall be
     * initialized. \n ELSE \n &emsp; Log a trace message that the config object is undefined. \n END IF
     *
     * @param[in] name : the configuration section element
     * @param[out] val : config object to be loaded
     *
     * @return If the config section name exists, GS_OK . Otherwise, GS_Undefined
     *
     * @sa ConfigSection::init
     * \n	ConfigSection::getSetting
     */
    virtual GetStatus get(const std::string& name, ConfigSection& val) const;

    /*!
     * @brief This method shall read the value for a given element in config section.
     * \n templates disallow virtual and block Mocking.
     *
     * @param[in] name : The configuration section element we're trying to access
     * @param[in] val  : The value we're going to fill from the configuration file
     *
     * @return If the value could be read, GS_OK.
     *         If the value is undefined, GS_Undefined.
     *         If the value is defined as some other type in the configuration file, GS_TypeMismatch.
     *
     * @throw std::out_of_range If the value from configuration is out side the range of the given type.
     *
     * @sa ConfigSection::getTemplated where traits on template type is integral(
     * int8_t,int16_t,int32_t,int64_t,uint8_t,uint16_t,uint16_t) \n  ConfigSection::getTemplated	where traits on
     * template type is floating point( float,double)
     */
    virtual GetStatus get(const std::string& name, bool& val) const;
    virtual GetStatus get(const std::string& name, int8_t& val) const;
    virtual GetStatus get(const std::string& name, int16_t& val) const;
    virtual GetStatus get(const std::string& name, int32_t& val) const;
    virtual GetStatus get(const std::string& name, int64_t& val) const;
    virtual GetStatus get(const std::string& name, uint8_t& val) const;
    virtual GetStatus get(const std::string& name, uint16_t& val) const;
    virtual GetStatus get(const std::string& name, uint32_t& val) const;

    /*!
     * @brief This method shall read the value for a given element in config section.
     *
     * @param[in] name : The configuration section element we're trying to access
     * @param[in] val  : The value we're going to fill from the configuration file
     *
     * @return If the value could be read, GS_OK.
     *         If the value is undefined, GS_Undefined.
     *         If the value is defined as some other type in the configuration file, GS_TypeMismatch.
     *
     * @throw std::out_of_range If the value from configuration is out side the range of the given type.
     *
     * @sa ConfigSection::getSetting
     *
     */
    virtual GetStatus get(const std::string& name, uint64_t& val) const;
    virtual GetStatus get(const std::string& name, float& val) const;
    virtual GetStatus get(const std::string& name, double& val) const;

    /*!
     * @brief This method shall read the value for a given configuration element name.
     * \n It handles SettingNotFoundException if settings doesn't exist for a given name.
     *
     * @param[in] name : The configuration section element we're trying to access
     * @param[in] val  : The value we're going to fill from the configuration file
     *
     * @return If the value could be read, GS_OK.
     *         If the value is undefined, GS_Undefined.
     *         If the value is defined as some other type in the configuration file, GS_TypeMismatch.
     *
     * @sa ConfigSection::getSetting
     */
    virtual GetStatus get(const std::string& name, std::string& val) const;

    /*!
     * @brief This templatized method shall be used to read a primitive type from a configuration section.
     * \n Warnings shall be published if the default value is used.
     *
     * @param[in] name : The configuration section element we're trying to access
     * @param[out] val : The value we're going to fill from the configuration file
     * @param[in] def  : The default value provided if val is undefined in the configuration file
     *
     * @throw std::out_of_range If the value from configuration is out side the range of the given type.
     *
     * @return bool : If the value could be set (either via the default or reading it from the file), true.
     * \n &emsp;&emsp; If the value is defined as some other type in the configuration file, false.
     *
     * @sa ConfigSection::getTemplated
     */
    template<class Type, class DefType>
    bool get(const std::string& name, Type& val, const DefType& def) const;

    // string container has known problem with libconfig.
    /*!
     * @brief This method shall read array of string values for a given config section into passed container which holds
     * strings. \n \b Operational \b Flow: \n First, it shall fetch the current config section \n IF the given config
     * section name exists in the current config section (parent) THEN \n &emsp; Fetch the config section for a given
     * section name(child) \n &emsp; Log a warning message if the config section is not a collection. \n &emsp; IF the
     * config section is a collection (array) THEN \n &emsp;&emsp; Get the number of elements in the config
     * section(length) \n &emsp;&emsp; FOR each element in config section \n &emsp;&emsp;&emsp; Load the element  into
     * container \n &emsp;&emsp; END FOR \n &emsp; END IF \n ELSE \n &emsp; Log a warning message like "given config
     * section doesn't exist" \n END IF
     *
     * \n Traits Req on type: Container type must be string
     *
     * \n Example:
     * \n level_1  = { level_2 = ["one","two"] }
     * \n
     * \n @param[in] name : The name of the array in the configuration section.
     * \n @param[out] v   : To be filled from the configuration section
     * \n
     * \n @return integer : Number of elements written to passed container.
     * \n                   If there is some error reading the values, 0.
     *
     * @sa ConfigSection::init
     */
    template<class Container>
    typename std::enable_if<std::is_same<std::string, typename Container::value_type>::value, int32_t>::type getArray(
        const std::string& name,
        Container& dest) const;

    // for Container = ConfigSection
    /*!
     * @brief This method shall read array of config sections for a given config section name into passed container.
     * \n \b Operational \b Flow:
     * \n First, it shall fetch the current config section
     * \n IF the given config section name exists in the current config section (parent)
     * \n &emsp; Fetch the config section for a given section name(child)
     * \n &emsp; Log a warning message if the config section is not a collection.
     * \n &emsp; IF the config section is a collection (array) THEN
     * \n &emsp;&emsp; Get the number of elements in the config section(length)
     * \n &emsp;&emsp; FOR each element in config section
     * \n &emsp;&emsp;&emsp; Initialize the element(config section) and load into container.
     * \n &emsp;&emsp; END FOR
     * \n &emsp; END IF
     * \n ELSE
     * \n &emsp; Log a warning message like "given config section doesn't exist"
     * \n END IF
     * \n
     * \n Traits Req on type: Container type must be ConfigSection
     *
     * \n Example:
     * \n level_1  = { level_2_1 = [1,2,3,4],
     *                 level_2_2 = [5,6,7,8]
     *               }
     * \n
     * \n @param[in] name : The name of the array in the configuration section
     * \n @param[out] v   : To be filled from the configuration section
     * \n
     * \n @return  integer : Number of elements written to passed container.
     * \n                    If there is some error reading the values, 0.
     *
     * @sa ConfigSection::init
     *
     */
    template<class Container>
    typename std::enable_if<std::is_same<ConfigSection, typename Container::value_type>::value, int32_t>::type getArray(
        const std::string& name,
        Container& dest) const;

    // for std::vector<uint16_t>. libconfig does not supoort short int.
    /*!
     * @brief This method shall read array of short integer values into passed container for a given config section.
     * \n \b Operational \b Flow:
     * \n First, it shall fetch the current config section
     * \n IF the config section for a given section name exists in the current config section THEN
     * \n &emsp; Fetch the config section for a given section name(child)
     * \n &emsp; Log a warning message if the config section is not a collection
     * \n &emsp; IF given config section is a collection type(array) THEN
     * \n &emsp;&emsp; Get the number of elements in config section(length)
     * \n &emsp;&emsp; FOR each element in config section
     * \n &emsp;&emsp;&emsp; Load element into container
     * \n &emsp;&emsp; END FOR
     * \n &emsp; END IF
     * \n ELSE
     * \n &emsp; Log a warning message like "given config section doesn't exist"
     * \n END IF
     * \n
     * \n Example:
     * \n level_1  = { level_2 = [1,2,3,4] }
     *
     * \n @param[in] name : The name of the array in the configuration section
     * \n @param[out] v   : To be filled from the configuration section
     * \n
     * \n @return integer :  Number of elements in written to passed container.
     *                       If there is some error reading the values, 0.
     *
     * \n \b Notes:Used std::vector<uint16_t> as libconfig does not supoort short int.
     */
    int32_t getArray(const std::string& name, std::vector<uint16_t>& dest) const;

    /*!
     * @brief This method shall read an array which is a collection of elements of type integral or floating point into
     * the passed container from the configuration file \n \b Operational \b Flow: \n First, it shall fetch the current
     * config section \n IF the config section for a given section name  exists in the current config section THEN \n
     * &emsp; Fetch the config section for a given section name. \n &emsp; Log a warning message if the config section
     * is not a collection \n &emsp; IF the given config section is a collection (array) THEN \n &emsp;&emsp; Get the
     * number of elements in config section(length) \n &emsp;&emsp; FOR each element in config section \n
     * &emsp;&emsp;&emsp; Load element into container \n &emsp;&emsp; END FOR \n &emsp; END IF \n ELSE \n &emsp; Log a
     * warning message like "given config section doesn't exist" \n END IF \n Traits Req on type: Container type must be
     * int / float
     *
     *
     * \n Example:
     * \n level_1  = { level_2 = [1.9,2.4,3.6,4.9] }
     *
     * @param[in] name  : The name of the array in the configuration section
     * @param[out] dest : The container to be filled from the configuration section
     *
     * @return integer  : Number of elements written to passed container
     * \n                 If there is some error reading the values, 0.
     *
     * \n \b Notes: The container is cleared so any values previously stored in it before the call to get are removed.
     * Container \n must provide a push_back method.
     */
    template<class Container>
    typename std::enable_if<std::is_integral<typename Container::value_type>::value ||
                                std::is_floating_point<typename Container::value_type>::value,
                            int32_t>::type
    getArray(const std::string& name, Container& dest) const;

    /*!
     * @brief This method shall read a matrix from the configuration file into the passed container
     * \n \b Operational \b Flow:
     * \n The config section needs to be in the following format:
     * \n 'sampleMatrix': {
     * \n 'rows': 2,
     * \n 'cols': 2,
     * \n 'data': [1.0, -2.0,
     * \n          0.0, -0.0]
     * \n }
     *
     * \n To get the configuration (in matrix form) into config object, call : get("sampleMatrix",matrix_config_object)
     * \n To get rows from config object, call get("rows", nrows_int)
     * \n To get columns from config object, call get("cols", ncols_int)
     * \n To get data from config object, matrix_config_object.getArray("data", container_object);
     *
     * @param[in] name  : The name of the matrix in the configuration section
     * @param[out] dest : The container to be filled from the configuration section
     *
     * @return a pair of integers which indicate the number of rows and columns.
     * \n 			return <0, 0> if we failed to get number of rows, columns, data or the rows
     * \n 			(or) columns does not agree with the provided data.
     *
     */
    template<class T>
    typename std::enable_if<std::is_integral<T>::value || std::is_floating_point<T>::value,
                            std::pair<std::size_t, std::size_t>>::type
    getMatrix(const std::string& name, std::vector<std::vector<T>>& dest) const;

    /*!
     * @brief This method shall return the list of parameter names available keys in a section
     *
     * @return A list of parameter names that can be used to access elements in this section.
     */
    std::list<std::string> getParameterNames() const;

    /*!
     * @brief This method shall log a warning message if not all parameters in a ConfigSection have been accessed
     *
     * @return bool : If all of the parameters in a section have been accessed, true. Otherwise false..
     *
     * \n Note: This test only verifies that there has been an attempt to access the parameters, not that they were
     * accessed successfully
     */
    bool checkAllParametersUsed();

    /*!
     * @brief This method shall report if a given parameter name is present in this config section.
     *
     * @param[in] name : the parameter name to be tested for
     *
     * @return bool : If the parameter is defined in this ConfigSection,true.Otherwise,false.
     *
     */
    bool hasParameter(const std::string& name) const;

    /*!
     * @brief This method shall turn on or off warning output.
     *
     * @param[in] state : warning output on/off
     *
     * @return bool : If true, warning text will be displayed. If false, it will not be.
     */
    void displayWarnings(bool state);

    /*!
     * @brief This method shall return name of the Config section.
     *
     * @return string : Name of the Config Section.
     */
    std::string getName() const;

    /*!
     * @brief This method shall initialize given configuration section.
     * \n \b Operational \b Flow:
     * \n It shall initializes the lib config pointer to point to the given config section.
     * \n It shall set up the access table (hash map) for all the elements listed in  a given config section(name).
     * \n It shall set the path of the config section object with respect to root config section.
     * \n It handles the SettingNotFoundException if the given section name doesn't exist
     *
     * @param[in] cfg  : Config pointer to ConfigSection
     * @param[in] name : The ConfigSection name
     *
     * @return bool : If the configSection is properly initialized,true. Otherwise,false.
     *
     */
    bool init(const std::string& name, const ConfigPtr& cfg);

    /*!
     * @brief This method shall write a configuration section to an output stream
     *
     * @param[in] reference of ostream
     * @param[in] ConfigSection
     */
    friend std::ostream& operator<<(std::ostream&, const ConfigSection& cs);

protected:
    /*!
     * @brief This template method shall read the value for a given element in config section.
     *
     * \n Traits req on Type: It must be integral type
     *
     * @param[in] name : The configuration section element
     * @param[in] val  : The value going to be filled from the configuration file
     *
     * @throw std::out_of_range If the value from configuration is out side the range of the given type.
     *
     * @return  GetStatus : If config section name exists, GS_OK. Otherwise, GS_Undefined
     */
    template<class Type>
    typename std::enable_if<std::is_integral<Type>::value, GetStatus>::type getTemplated(const std::string& name,
                                                                                         Type& val) const;

    /*!
     * @brief This template method shall read the value for a given element in config section.
     *
     * \n Traits req on Type: It must be floating type
     *
     * @param[in] name  : The configuration section element
     * @param[in] val   : The value going to be filled from the configuration file
     *
     * @throw std::out_of_range If the value from configuration is out side the range of the given type.
     *
     * @return  If config section name exists, GS_OK. Otherwise, GS_Undefined
     */
    template<class Type>
    typename std::enable_if<std::is_floating_point<Type>::value, GetStatus>::type getTemplated(const std::string& name,
                                                                                               Type& val) const;
    /*!
     * @brief This method shall locate and returns the configuration setting for a given path.
     * \n It handles exception SettingNotFoundException if it doesn't find the setting for a given path.
     *
     * @param[in]  name as string: configuration setting name
     * @param[out] result as GetStatus enum:  On success, GS_OK.
     * \n 									  Otherwise, GS_Undefined.
     *
     * @return configuration setting
     */
    virtual Setting& getSetting(const std::string& name, GetStatus& result) const;

    friend class ConfigParser; ///< friend class @see ConfigParser

    ConfigPtr inputConfig_;   ///< libconfig::Config pointer shall be used to read the config file
    std::string settingPath_; ///< Complete dot-separated path of current config section with respect to root level
                              ///< settings. Example: root.ReadWriteInterfaces.leftSick
    std::string sectionName_; ///< The fully qualified name of this config section, just used for debugging output.
                              ///< Example.root.ReadWriteInterfaces.leftSick

    using AccessedMap = std::map<std::string, bool>;      ///< typedef hash map
    using AccessedMapIter = AccessedMap::iterator;        ///< map iterator
    using AccessedMapCIter = AccessedMap::const_iterator; ///< const_iterator for the map

    mutable AccessedMap
        accessedElements_; ///< Config Section names and it's accesses status stored in map container
                           ///< where key - in map is name of the Config Section,
                           ///< Value - in map is Config Section access status.
                           ///< true if Config Section is accessed.false If Config Section is not accessed.

    bool silent_{false}; ///< This variable shall be used to determine whether to display warning message
};

#include "ConfigSection.def.h"

#endif // CONFIGSECTION_H
