/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file InterfaceParser.h
 * @product config
 * @version
 * @released
 * @updated 02/05/2019
 * @brief Interface configuration parser
 *
 * @section License
 *
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#ifndef INTERFACE_PARSER_H
#define INTERFACE_PARSER_H

#include <config/ConfigParser.h>

#include <list>

/*!
 * @brief This class shall help to get Read, Write and ReadWrite interfaces from configuration file and also checks the
 * \n interfaces presence in configuration file.It shall inherit the ConfigParser to read and parse the configuration
 * file \n created for interfaces. The following is the sample configuration file.
 *
 * \n
 * \n ReadWriteInterfaces = {
 * \n 	"leftSick" => Interfaces::LeftEBoxSickInterface,
 * \n 	"rightSick" => Interfaces::RightEBoxSickInterface
 * \n }
 * \n WriteOnlyInterfaces = {
 * \n 	"outputMap" => Interfaces::CostMapInterface
 * \n }
 * \n LeftEBoxSickInterface = {
 * \n 	"interface" => "LineScanner",
 * \n 	"implementation" => "SickLMS",
 * \n 	"channel" => LeftEBoxSickIpcChannel,
 * \n 	"port" => "/dev/ttyS2",
 * \n 	"baud" => "115200",
 * \n 	"flow_control" => "none",
 * \n 	"clippingDistance_m" => 15
 * \n }
 * \n RightEBoxSickInterface = {
 * \n 	"interface" => "LineScanner",
 * \n 	"implementation" => "SickLMS_v2",
 * \n 	"channel" => RightEBoxSickIpcChannel,
 * \n 	"port" => "/dev/ttyS3",
 * \n 	"baud" => "115200",
 * \n 	"flow_control" => "none",
 * \n 	"clippingDistance_m" => 15
 * \n }
 * \n CostMapInterface = {
 * \n 	"interface" => "BaseMap",
 * \n 	"implementation" => "NavMap",
 * \n 	"channel" => CostMapIpcChannel,
 * \n 	"defaultCost" => 255,
 * \n 	"maxMsgSize" => 1024768
 * \n }
 *
 */
class InterfaceParser : public ConfigParser
{
public:
    /*!
     * @brief Default constructor
     */
    InterfaceParser() = default;

    /*!
     * @brief Parameterized constructor. It shall initialize the interface parser.
     *
     * @param[in]  ConfigPtr represents the loaded configuration file
     */
    InterfaceParser(const ConfigPtr configPtr)
        : ConfigParser()
    {
        this->inputConfig_ = configPtr;
        configInitialized_ = true;
    }

    /*!
     * @brief Default desstructor
     */
    virtual ~InterfaceParser() = default;

    /*!
     * @brief This method shall report if the config file has a ReadOnlyInterfaces section. The following is the sample
     * configuration file.
     *
     * \n ReadWriteInterfaces = {
     * \n 	"leftSick" => Interfaces::LeftEBoxSickInterface,
     * \n 	"rightSick" => Interfaces::RightEBoxSickInterface
     * \n }
     * \n WriteOnlyInterfaces = {
     * \n 	"outputMap" => Interfaces::CostMapInterface
     * \n }
     * \n Returns false as above example doesn't have ReadOnlyInterfaces.
     *
     * @return bool : On success, true. If not found,false.
     *
     * @sa ConfigParser::hasSection
     */
    bool hasReadOnlyInterfaces();

    /**
     * @brief This method shall report if the config file has a read-only interface section with the specified name.
     *
     * @param[in] interfaceName : Name of the interface to be searched for
     *
     * @return bool : On success, true. If not found, false.
     *
     * @sa InterfaceParser::hasInterface
     */
    bool hasReadOnlyInterface(const std::string& interfaceName);

    /*!
     * @brief This method shall report if the config file has a WriteOnlyInterfaces section. Below is the sample
     * configuration file
     *
     * \n ReadWriteInterfaces = {
     * \n 	"leftSick" => Interfaces::LeftEBoxSickInterface,
     * \n 	"rightSick" => Interfaces::RightEBoxSickInterface
     * \n }
     * \n WriteOnlyInterfaces = {
     * \n 	"outputMap" => Interfaces::CostMapInterface
     * \n }
     * \n Returns true as above example has WriteOnlyInterfaces
     *
     * @return bool : On success, true. If not found, false.
     *
     * @sa ConfigParser::hasSection
     */
    bool hasWriteOnlyInterfaces();

    /*!
     * @brief This method shall report if the config file has a write-only interface section with the specified name.
     *
     * @param[in] interfaceName : name of the interface to be searched for
     *
     * @return bool : On success, true. If not found, false.
     *
     * @sa InterfaceParser::hasInterface
     */
    bool hasWriteOnlyInterface(const std::string& interfaceName);

    /*!
     * @brief This method shall report if the config file has a ReadWriteInterfaces section. Below is the sample
     * configuration file
     *
     * \n ReadWriteInterfaces = {
     * \n 	"leftSick" => Interfaces::LeftEBoxSickInterface,
     * \n 	"rightSick" => Interfaces::RightEBoxSickInterface
     * \n }
     * \n WriteOnlyInterfaces = {
     * \n 	"outputMap" => Interfaces::CostMapInterface
     * \n }
     * \n Returns true as above example has ReadWriteInterfaces
     *
     * @return bool : On success, true. If not found, false
     *
     * @sa ConfigParser::hasSection
     */
    bool hasReadWriteInterfaces();

    /*!
     * @brief This method shall report if the config file has a read-write interface section with the specified name.
     *
     * @param[in] interfaceName : Name of the interface to be searched for.
     *
     * @return On success, true. If not found,false.
     *
     * @sa InterfaceParser::hasInterface
     */
    bool hasReadWriteInterface(const std::string& interfaceName);

    /*!
     * @brief This method shall get ReadOnlyInterfacesSection from config file.
     *
     * @param[out] roIntf : This object is initialized with ReadOnlyInterfacesSection
     *
     * @return bool : On success, true. If not found, false.
     *
     * @sa ConfigParser::getSection
     */
    bool getReadOnlyInterfacesSection(ConfigSection& roIntf);

    /*!
     * @brief This method shall get WriteOnlyInterfacesSection from config file.
     *
     * @param[out] woIntf : ConfigSection object is initialized with WriteOnlyInterfacesSection
     *
     * @return bool : On success, true. If not found, false.
     *
     * @sa ConfigParser::getSection
     */
    bool getWriteOnlyInterfacesSection(ConfigSection& woIntf);

    /**
     * @brief This method shall get ReadWriteInterfacesSection from config file.
     *
     * @param[out] rwIntf : This object is initialized with ReadWriteInterfacesSection
     *
     * @return bool : On success, true. If not found, false.
     *
     * @sa ConfigParser::getSection
     */
    bool getReadWriteInterfacesSection(ConfigSection& rwIntf);

    /*!
     * @brief This method shall get all of the ReadOnly interfaces, in the order they must be constructed
     * \n \b Operational \b Flow:
     * \n It shall get list of all 'ReadOnly interfaces section' from configuration file.
     * \n It shall check each of the interface section from the list, if any interface config section has a parameter
     * \n named "implementation". If so, interface config section shall be added to return list.
     *
     * @return list of config sections for ReadOnly interfaces. empty list if ReadOnly interfaces section doesn't exist
     * in config file.
     *
     * @sa ConfigParser::getReadOnlyInterfacesSection
     * \n  ConfigParser::getAllSectionsWithParamName
     */
    std::list<ConfigSection> getAllReadOnlyInterfaces();

    /*!
     * @brief This method shall get single ReadOnly interface config section for a given interface name.
     * \n \b Operational \b Flow:
     * \n It shall get the list of ReadOnly interfaces.
     * \n FOR each interface in ReadOnly interfaces list
     * \n &emsp; IF the interface name matches with the given interface name THEN
     * \n &emsp; &emsp; Return config section
     * \n &emsp; ELSE
     * \n &emsp; &emsp; Return empty config section
     * \n &emsp; END IF
     * \n END FOR
     *
     * @param[in] interfaceName : Name of the interface to fetch config section
     *
     * @return ConfigSection object for ReadOnly interface with given name, if it exists. otherwise, empty ConfigSection
     * object.
     *
     * @sa InterfaceParser::getAllReadOnlyInterfaces
     */
    ConfigSection getReadOnlyInterface(const std::string& interfaceName);

    /*!
     * @brief This method shall get all of the WriteOnly interfaces, in the order they must be constructed
     * \n \b Operational \b Flow:
     * \n It shall get list of all 'WriteOnly interfaces section' from configuration file.
     * \n It shall check each of the interface section from the list, if any interface config section has a parameter
     * \n named "implementation". If so, interface config section shall be added to return list.
     *
     * @return list of config sections for WriteOnly interfaces.empty list if WriteOnly interfaces section doesn't exist
     * in config file
     *
     * @sa InterfaceParser::getWriteOnlyInterfacesSection
     * \n  ConfigParser::getAllSectionsWithParamName
     */
    std::list<ConfigSection> getAllWriteOnlyInterfaces();

    /*!
     * @brief This method shall get config section for single WriteOnly interface config section for a given interface
     * name. \n \b Operational \b Flow: \n It shall get list of WriteOnly interfaces \n FOR each interface in WriteOnly
     * interfaces list \n &emsp; IF the interface name matches with the given interface name THEN \n &emsp; &emsp;
     * return config section \n &emsp; ELSE \n &emsp; &emsp; return empty config section \n &emsp; END IF \n END FOR
     *
     * @param[in] interfaceName : Name of the interface to fetch config section
     *
     * @return ConfigSection object for WriteOnly interface with given name, if it exists. otherwise, empty
     * ConfigSection object.
     *
     * @sa InterfaceParser::getAllWriteOnlyInterfaces
     */
    ConfigSection getWriteOnlyInterface(const std::string& interfaceName);

    /*!
     * @brief This method shall get all of the ReadWrite interfaces, in the order they must be constructed
     * \n \b Operational \b Flow:
     * \n It shall get list of all 'ReadWrite interfaces section' from configuration file.
     * \n It shall check each of the interface section from the list, if any interface config section has a parameter
     * \n named "implementation". If so, interface config section shall be added to return list.
     *
     * @return list of config sections for ReadWrite interfaces.empty list if ReadWrite interfaces section doesn't exist
     * in config file
     *
     * @sa InterfaceParser::getReadWriteInterfacesSection
     * \n  ConfigParser::getAllSectionsWithParamName
     */
    std::list<ConfigSection> getAllReadWriteInterfaces();

    /*!
     * @brief This method shall get config section for single ReadWrite interface config section for a given interface
     * name. \n \b Operational \b Flow: \n It shall get list of ReadWrite interfaces \n FOR each interface in ReadWrite
     * interfaces list \n &emsp; IF the interface name matches with the given interface name THEN \n &emsp; &emsp;
     * return config  section \n &emsp; ELSE \n &emsp; &emsp; return empty config section \n &emsp; END IF \n ENDFOR
     *
     * @param[in] interfaceName : Name of the interface to fetch config section
     *
     * @return ConfigSection object for ReadWrite interface with given name, if it exists. otherwise, empty
     * ConfigSection object.
     *
     * @sa InterfaceParser::getAllReadWriteInterfaces
     */
    ConfigSection getReadWriteInterface(const std::string& interfaceName);

private:
    /*!
     * @brief This method shall check if the given interface name is in the given  list of interfaces
     * \n Name is in the form "interfaceType.interfaceName"
     *
     * @param[in] interfaceType :  READ_ONLY_INTERFACES/WRITE_ONLY_INTERFACES/READ_WRITE_INTERFACES
     * @param[in] interfaceName :  Name of the interface to lookup
     * @param[in] interfaces    :  List of interface config sections used to search within for a given interface name.
     *
     * @return bool : If interfaceName is found in the list of interfaces, true. Otherwise, false.
     */
    static inline bool hasInterface(const std::string& interfaceType,
                                    const std::string& interfaceName,
                                    const std::list<ConfigSection>& interfaces)
    {
        const auto it = std::find_if(
            interfaces.begin(), interfaces.end(), [interfaceType, interfaceName](const ConfigSection& c) -> bool {
                return ((interfaceType + "." + interfaceName) == c.getName());
            });
        return (it != interfaces.end());
    }

    static const std::string
        READ_ONLY_INTERFACES; ///< Name of the config section for ReadOnly Interfaces in configuration file.
    static const std::string
        WRITE_ONLY_INTERFACES; ///< Name of the config section for WriteOnly interfaces in configuration file.
    static const std::string
        READ_WRITE_INTERFACES; ///< Name of the config section for ReadWriteOnly interfaces in configuration file.
};

#endif // INTERFACE_PARSER_H
