/*!
 ***********************************************************
 * @file Parameter.h
 * @brief this file defines Parameter template, helps to
 * have convenient access to config variables.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * Parameter.h
 * Michael Darms (michael.darms@us.contiautomotive.com)
 *
 * Copyright (c) 2007
 * Carnegie Mellon University
 * All rights reserved.
 */
#ifndef PARAMETER_H
#define PARAMETER_H

#include <config/ConfigSection.h>
#include <list>

#define PARAMETER(name, configSection, value) name(configSection, #name, value)
#define PARAMETERDEF(name, value) name(#name, value)

/*!
 * @class Parameter
 * @brief This template class shall help to have convenient access to configuration variables
 *
 * \n Usage:
 * \n 1) Define a Parameter inside a class Foo:
 * \n   Parameter<type> FOO_PARAMETER_A;
 * \n 2) Initialize the parameters inside the constructor Foo::Foo(ConfigSection configSection)
 * \n   FOO_PARAMETER_A(configSection, "ParameterString", value)
 * \n   Note: "ParameterString" is the string used in the configSection.
 * \n 3) Use the defined parameters like constants anywhere in the code.
 *
 * \n Note: std::string has to be used with a type cast: std::string(FOO_STRING_PARAMETER)
 *
 * \n Using the macro PARAMETER it is ensured that the parameter and the ConfigSting are identical:
 * \n PARAMETER(FOO_PARAMETER_A, configSection, defaultValue)
 */
template<typename ParameterType>
class Parameter
{
public:
    /*!
     * \brief Parameterized constructor. It shall be initialized with given config name and default value.
     *
     * \param[in] configName - name of the parameter read from configSection
     * \param[in] defaultValue - default value with which parameter should be assigned
     */
    Parameter(const std::string& configName, const ParameterType& defaultValue)
    {
        parameter_ = defaultValue;
        configName_ = configName;
        defaultValue_ = defaultValue;
    }

    /*!
     * \brief Parameterized constructor. It shall be initialized with given config section object.
     *
     * \param[in] cs - name of configSection
     * \param[in] configName - name of the parameter as shown in configSection
     * \param[in] defaultValue - default value with which parameter should be assigned
     */
    Parameter(ConfigSection& cs, const std::string& configName, const ParameterType& defaultValue)
    {
        parameter_ = defaultValue;
        configName_ = configName;
        defaultValue_ = defaultValue;
        readParameter(cs);
    }

    /*!
     * default destructor
     */
    virtual ~Parameter() = default;

    /*!
     * \brief Overloaded function call operator()  method. This method shall return current value of the parameter which
     * shall be \n retrieved from given config section for a particular parameter name.
     *
     * @return The current value of the parameter.
     */
    operator ParameterType() const { return parameter_; }

    /*!
     * \brief This method shall read the value for given parameter name from given config section and stores the value.
     *
     * \param[in] cs :  Given config section in which parameter is defined
     *
     * \return bool :  If the value could be set (either via the default or reading it from the file), true.
     *                 If the value is defined as some other type in the configuration file, false.
     */
    bool readParameter(ConfigSection& cs) { return cs.get(configName_, parameter_, defaultValue_); }

private:
    ParameterType parameter_;    ///< The parameter which will be assigned value after reading from configSection for a
                                 ///< particular config element. It's type is generic ( template type).
    ParameterType defaultValue_; ///< The default value for parameter is considered for a particular config element if
                                 ///< no value exists in the configSection It's type is generic ( template type).
    std::string configName_{""}; ///< The name of the parameter(config element) as defined in configSection.
};

#endif // idndef PARAMETER_H
