/*!
 ***********************************************************
 * @file TaskParser.h
 *
 * @brief file defining TaskParser class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 * TaskParser.h
 * Chris Urmson (curmson@ri.cmu.edu)
 * August 23, 2006
 *
 * Copyright (c) 2006
 * Carnegie Mellon University
 * All rights reserved.
 */
#ifndef TASKPARSER_H
#define TASKPARSER_H
#include "ConfigParser.h"
#include "InterfaceParser.h"
#include <string>

/**
 * @class TaskParser
 * @brief This class shall help to get the task parameters from configuration file. It shall inherit from ConfigParser
 * to read and parse \n the configuration file created for a task. The following is the sample task configuration file.
 * \n Execution = {
 * \n "executableName" => "task1",
 * \n  "requiredTasks" => [ "Task2" ],
 * \n  "preCommands" => "echo $SOMETHING",
 * \n  "postCommands" => "",
 * \n  "defaultMachine" => "default"
 * \n }
 */
class TaskParser : public ConfigParser
{
public:
    /**
     * @brief Default constructor
     */
    TaskParser();

    /*!
     * @brief This method shall set the directory path for the task configuration parameters.
     * \n It takes directory name as input and appends "/" to form the directory path
     *
     * @param[in] dir  : Name of the directory
     *
     * @return void
     */
    void setTaskParamDirectory(const std::string& dir);

    /**
     * @brief This method shall read and parse the configuration of the passed taskName.
     * \n \b Operational \b Flow:
     * \n config files can be libconfig (.cfg) or python files(.py)
     * \n IF .py (python) configuration file exists THEN
     * \n &emsp; parse and load the configuration file
     * \n END IF
     * \n IF parse failed THEN
     * \n &emsp; look for .cfg (libconfig) configuration file
     * \n &emsp; parse and load the configuration file
     * \n END IF
     * \n It initializes the interface parser to access the interfaces configured for the task.
     *
     * @param[in] taskName : The taskName to be loaded.
     *
     * @return If the task configuration could be loaded, true. If load failed or task configuration parameters
     * directory doesn't exit, false.
     *
     * @sa ConfigParser::parseFile
     */
    bool parseTask(const std::string& taskName);

    /**
     * @brief This method shall report the task parameters ConfigSection stored in the configuration file.
     * \n \b Operational \b Flow:
     * \n It checks the configuration file has a section with name "Parameters".
     * \n if so, it loads the task parameters.
     * \n sample ConfigSection for task parameters in tasks configuration file
     * \n Parameters = {
     * \n 	"cycleRate_hz" => 0.5,
     * \n 	"someParameter" => 500
     * \n }
     *
     * @param[out] taskParams : The ConfigSection object to be filled
     *
     * @return If the task has a Parameters section, true. Otherwise,false.
     *
     * @sa ConfigParser::getSection
     */
    inline bool getTaskParameters(ConfigSection& taskParams) { return this->getSection("Parameters", taskParams); }

    /**
     * @brief This method shall report the execution parameters ConfigSection stored in the configuration file.
     * \n \b Operational \b Flow:
     * \n It checks the configuration file has a section with name "Execution".
     * \n If so, it loads the tasks execution parameters.
     * \n sample config section for task execution parameters in tasks configuration file
     * \n Execution = {
     * \n 	"binary" => "SomeTask",
     * \n 	"requiredTasks" => [ "Task2" ],
     * \n 	"preCommands" => "echo $SOMETHING",
     * \n 	"postCommands" => "",
     * \n 	"defaultMachine" => "default"
     * \n }
     *
     * @param[out] taskParams : The ConfigSection to be filled
     *
     * @return If the task has a Execution section, true. Otherwise, false.
     *
     * @sa ConfigParser::getSection
     *
     */
    inline bool getExecutionParameters(ConfigSection& execParams) { return this->getSection("Execution", execParams); }

    /**
     * @brief This method shall report the metrics parameters ConfigSection stored in the configuration file.
     * \n \b Operational \b Flow:
     * \n It checks the configuration file has a section with name "Metrics".
     * \n If so, it loads the tasks metrics parameters.
     * \n sample config section for task execution parameters in tasks configuration file
     * \n Metrics = {
     * \n ...
     * \n }
     *
     * @param[out] metricsParams : The ConfigSection to be filled
     *
     * @return If the task has a Metrics section, true. Otherwise, false.
     */
    inline bool getMetricsParameters(ConfigSection& metricsParams)
    {
        return this->getSection("Metrics", metricsParams);
    }

    /**
     * @brief This method shall report if the config file has a ReadOnlyInterfaces section.
     * \n Below is the sample configuration file
     * \n ReadOnlyInterfaces = {
     * \n 	"leftSick" => Interfaces::LeftEBoxSickInterface,
     * \n 	"rightSick" => Interfaces::RightEBoxSickInterface
     * \n }
     *
     * @return bool : On success, true. If not found, false.
     */
    inline bool hasReadOnlyInterfaces() { return interfaceParser_.hasReadOnlyInterfaces(); }

    /**
     * @brief This method shall report if the config file has a read-only interface with the specified name
     *
     * @return bool : true, on success. false, if not found
     */
    inline bool hasReadOnlyInterface(const std::string& interfaceName)
    {
        return interfaceParser_.hasReadOnlyInterface(interfaceName);
    }

    /**
     * @brief This method shall report if the config file has a WriteOnlyInterfaces section
     *
     * @return bool : true, on success. false, if not found
     */
    inline bool hasWriteOnlyInterfaces() { return interfaceParser_.hasWriteOnlyInterfaces(); }

    /**
     * @brief This method shall report if the config file has write-only interface with the specified name.
     *
     * @return bool : true, on success. false, if not found
     */
    inline bool hasWriteOnlyInterface(const std::string& interfaceName)
    {
        return interfaceParser_.hasWriteOnlyInterface(interfaceName);
    }

    /**
     * @brief This method shall report if the config file has a ReadWriteInterfaces section.
     *
     * @return bool : On success, true. If not found, false.
     */
    inline bool hasReadWriteInterfaces() { return interfaceParser_.hasReadWriteInterfaces(); }

    /**
     * @brief This method shall report if the config file has a ReadWrite interface with the specified name
     *
     * @return bool : On success, true. If not found, false.
     */
    inline bool hasReadWriteInterface(const std::string& interfaceName)
    {
        return interfaceParser_.hasReadWriteInterface(interfaceName);
    }

    /**
     * @brief This method shall get ReadOnlyInterfacesSection from config file.
     *
     * @param[out]  interfaces : ConfigSection object is initialized with ReadOnlyInterfacesSection.
     *
     * @return bool : On success, true. If not found, false.
     */
    inline bool getReadOnlyInterfacesSection(ConfigSection& interfaces)
    {
        return interfaceParser_.getReadOnlyInterfacesSection(interfaces);
    }

    /**
     * @brief This method shall get WriteOnlyInterfacesSection from config file.
     *
     * @param[out] interfaces  : ConfigSection object is initialized with WriteOnlyInterfacesSection
     *
     * @return bool : On success, true. On error, false.
     */
    inline bool getWriteOnlyInterfacesSection(ConfigSection& interfaces)
    {
        return interfaceParser_.getWriteOnlyInterfacesSection(interfaces);
    }

    /**
     * @brief This method shall get ReadWriteInterfacesSection from config file.
     * \n Below is the sample configuration file
     * \n ReadWriteInterfaces = {
     * \n 	"leftSick" => Interfaces::LeftEBoxSickInterface,
     * \n 	"rightSick" => Interfaces::RightEBoxSickInterface
     * \n }
     *
     * @param[out] interfaces : ConfigSection object is initialized with ReadWriteInterfacesSection
     *
     * @return bool : On success, true. On error, false.
     */
    inline bool getReadWriteInterfacesSection(ConfigSection& interfaces)
    {
        return interfaceParser_.getReadWriteInterfacesSection(interfaces);
    }

    /**
     * @brief This method shall get all of the ReadOnly interfaces, in the order they must be constructed.
     * \n \b Operational \b Flow:
     * \n First, it shall get the read only interfaces section.
     * \n It shall traverse whole tree for every interface in read only interfaces section and add to return list,
     * \n if interface section has a parameter named "implementation"
     *
     * @return list of config section objects for ReadOnly interfaces
     *
     * @sa InterfaceParser::getAllReadOnlyInterfaces
     */
    inline std::list<ConfigSection> getAllReadOnlyInterfaces() { return interfaceParser_.getAllReadOnlyInterfaces(); }

    /**
     * @brief This method shall get a single ReadOnly interface ConfigSection for a given interface name.
     *
     * @param[in] interfaceName : Name of the interface
     *
     * @return config section for ReadOnly interface with given name, if it exists.otherwise,empty config section
     *
     * @sa InterfaceParser::getReadOnlyInterface
     */
    inline ConfigSection getReadOnlyInterface(const std::string& interfaceName)
    {
        return interfaceParser_.getReadOnlyInterface(interfaceName);
    }

    /**
     * @brief This method shall get all the WriteOnly interfaces, in the order they constructed in configuration.
     * \n \b Operational \b Flow:
     * \n First, it shall get the write only interfaces section.
     * \n It shall traverse whole tree for every interface in write only interfaces section and
     * \n add to return list, if interface section has parameter named "implementation".
     *
     * @return list of config section objects for WriteOnly interfaces.
     *
     * @sa InterfaceParser::getAllWriteOnlyInterfaces
     */
    inline std::list<ConfigSection> getAllWriteOnlyInterfaces() { return interfaceParser_.getAllWriteOnlyInterfaces(); }

    /**
     * @brief This method shall get a single WritelOnly interface ConfigSection for a given interface name.
     *
     * @param[in] interfaceName  : Name of the interface
     *
     * @return config section for WriteOnly interface with given name, if it exists.otherwise,empty config section
     *
     * @sa InterfaceParser::getWriteOnlyInterface
     */
    inline ConfigSection getWriteOnlyInterface(const std::string& interfaceName)
    {
        return interfaceParser_.getWriteOnlyInterface(interfaceName);
    }

    /**
     * @brief This method shall get all the ReadWrite interfaces, in the order they constructed in configuration.
     * \n \b Operational \b Flow:
     * \n First, it shall get the readwrite only interfaces section.
     * \n It shall traverse whole tree for every interface in readwrite only interfaces section and
     * \n add to return list, if interface section has parameter named "implementation".
     *
     * @return list of config section objects for ReadWrite interfaces
     *
     * @sa InterfaceParser::getAllReadWriteInterfaces
     */
    inline std::list<ConfigSection> getAllReadWriteInterfaces() { return interfaceParser_.getAllReadWriteInterfaces(); }

    /**
     * @brief This method shall get a single ReadWrite interface ConfigSection for a given interface name.
     *
     * @param[in] interfaceName name of the interface
     *
     * @return config section for ReadWrite interface with given name, if it exists.otherwise,empty config section
     *
     * @sa InterfaceParser::getReadWriteInterface
     */
    inline ConfigSection getReadWriteInterface(const std::string& interfaceName)
    {
        return interfaceParser_.getReadWriteInterface(interfaceName);
    }

protected:
    InterfaceParser interfaceParser_; ///< To retrieve interfaces configuration.
    std::string taskName_;            ///< Name of the task.
    std::string taskParamDir_;        ///< Directory where task related configuration is stored.
};

#endif // ifndef TASKPARSER_H
