/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Unit tests for @ref ConfigSection

#include "ConfigFixture.hpp"
#include <config/ConfigParser.h>
#include <config/ConfigSection.h>
#include <gtest/gtest.h>

using namespace std;
using GS = ConfigSection::GetStatus;

const std::string TESTDIR = std::getenv("TESTDIR");

TEST(ConfigSection, init_gracefullyHandlesNullConfig)
{
    std::shared_ptr<Config> config{};
    EXPECT_EQ(false, ConfigSection().init("", config));
}

TEST(ConfigSection, init_returnsTrueOnlyWhenSectionIsFound)
{
    ConfigSection objectUnderTest;
    auto config = std::make_shared<Config>();
    config->readString("section:{ setting = \"value\";};");

    EXPECT_EQ(false, objectUnderTest.init("non existent section", config));
    EXPECT_EQ(true, objectUnderTest.init("section", config));
}

TEST(ConfigSection, ctor)
{
    EXPECT_NO_THROW(ConfigSection());
}

TEST(ConfigSection, copyConstructorCopiesSectionToNewObject)
{
    std::string setting;
    ConfigSection objectToCopy{};

    auto config = std::make_shared<Config>();
    config->readString("section:{ setting = \"value\";};");
    EXPECT_NO_THROW(objectToCopy.init("section", config));

    ConfigSection objectUnderTest(objectToCopy);
    EXPECT_EQ(true, objectUnderTest.hasParameter("setting"));
    EXPECT_EQ(GS::GS_OK, objectUnderTest.get("setting", setting));
    EXPECT_EQ(setting, "value");
}

TEST(ConfigSection, assignmentOperatorReplacesSectionWithNewObjectsSection)
{
    std::string setting;
    ConfigSection objectToAssign{};

    auto config = std::make_shared<Config>();
    config->readString("section:{ setting = \"value\";};");
    EXPECT_NO_THROW(objectToAssign.init("section", config));

    ConfigSection objectUnderTest;
    objectUnderTest = objectToAssign;
    EXPECT_EQ(true, objectUnderTest.hasParameter("setting"));
    EXPECT_EQ(GS::GS_OK, objectUnderTest.get("setting", setting));
    EXPECT_EQ(setting, "value");

    // Verify non-empty ConfigSection has data overwritten on assignment
    auto otherConfig = std::make_shared<Config>();
    otherConfig->readString("otherSection:{ otherSetting = \"otherValue\";};");
    EXPECT_NO_THROW(objectToAssign.init("otherSection", otherConfig));

    objectUnderTest = objectToAssign;
    EXPECT_FALSE(objectUnderTest.hasParameter("setting"));
    EXPECT_TRUE(objectUnderTest.hasParameter("otherSetting"));
    EXPECT_EQ(GS::GS_OK, objectUnderTest.get("otherSetting", setting));
    EXPECT_EQ(setting, "otherValue");
}

TEST(ConfigSection, checkAllParametersUsed)
{
    std::string setting;
    ConfigSection objectUnderTest;
    auto config = std::make_shared<Config>();
    config->readString("section:{ setting = \"value\";};");
    objectUnderTest.init("section", config);

    EXPECT_EQ(false, objectUnderTest.checkAllParametersUsed());

    objectUnderTest.get("setting", setting);
    EXPECT_EQ(true, objectUnderTest.checkAllParametersUsed());
}

TEST(ConfigSection, init_resetsUsedParameters)
{
    std::string setting;
    ConfigSection objectUnderTest;
    auto config = std::make_shared<Config>();
    config->readString("section:{ setting = \"value\";};");

    objectUnderTest.init("section", config);
    objectUnderTest.get("setting", setting);
    EXPECT_EQ(true, objectUnderTest.checkAllParametersUsed());
    objectUnderTest.init("section", config);
    EXPECT_EQ(false, objectUnderTest.checkAllParametersUsed());

    // Verify used parameters are reset even for failed initialization
    std::shared_ptr<Config> nullConfig{};
    objectUnderTest.init("", nullConfig);
    // No accessable elements, so all parameters used.
    EXPECT_EQ(true, objectUnderTest.checkAllParametersUsed());

    objectUnderTest.init("section", config);
    EXPECT_EQ(false, objectUnderTest.checkAllParametersUsed());
    objectUnderTest.init("invalid section", config);
    // No accessable elements, so all parameters used.
    EXPECT_EQ(true, objectUnderTest.checkAllParametersUsed());
}

TEST(ConfigSection, insertionOperator)
{
    std::stringstream out;
    ConfigSection section;
    auto config = std::make_shared<Config>();
    config->readString("section:{ setting = \"value\";};");
    section.init("section", config);

    // Verify operator << prints "Section (<name>)"
    out << section;
    EXPECT_EQ("Section (section)", out.str());
}

class SimpleConfigTest : public ConfigFixture
{
public:
    ConfigParser parser_;
    std::string fileName_;
    std::string sectionName_;
    std::string paramName_;
    ConfigSection section_;

    virtual void SetUp() override
    {
        fileName_ = TESTDIR + "/ConfigParserTest1.cfg";
        sectionName_ = "Section2";
        ASSERT_TRUE(parser_.parseFile(fileName_)) << "Failed to parse a valid config file.";
        ASSERT_TRUE(parser_.hasSection(sectionName_)) << "Failed to find section_: " << sectionName_;
        ASSERT_TRUE(parser_.getSection(sectionName_, section_)) << "Failed to read section data from section_: ";
    }
};

/*
 * Look for deliberate type mismatches for all the types
 */
TEST_F(SimpleConfigTest, typeErrorsBool)
{
    paramName_ = "Bool_Parameter";
    ASSERT_TRUE(section_.hasParameter(paramName_));
    bool boolParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, boolParam));

    int8_t int8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, int8_tParam));

    int16_t int16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, int16_tParam));

    int32_t int32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, int32_tParam));

    int64_t int64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, int64_tParam));

    uint8_t uint8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, uint8_tParam));

    uint16_t uint16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, uint16_tParam));

    uint32_t uint32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, uint32_tParam));

    uint64_t uint64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, uint64_tParam));

    float float_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, float_tParam));

    double double_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, double_tParam));

    std::string string_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, string_tParam));
}

TEST_F(SimpleConfigTest, typeErrorsFloat)
{
    paramName_ = "Float_Parameter";

    ASSERT_TRUE(section_.hasParameter(paramName_));
    bool boolParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, boolParam));

    int8_t int8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, int8_tParam));

    int16_t int16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, int16_tParam));

    int32_t int32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, int32_tParam));

    int64_t int64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, int64_tParam));

    uint8_t uint8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, uint8_tParam));

    uint16_t uint16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, uint16_tParam));

    uint32_t uint32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, uint32_tParam));

    uint64_t uint64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, uint64_tParam));

    float float_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, float_tParam));

    double double_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get(paramName_, double_tParam));

    std::string string_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get(paramName_, string_tParam));
}

TEST_F(SimpleConfigTest, typeErrorsString)
{
    ASSERT_TRUE(section_.hasParameter("String_Parameter"));
    bool boolParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", boolParam));

    int8_t int8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", int8_tParam));

    int16_t int16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", int16_tParam));

    int32_t int32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", int32_tParam));

    int64_t int64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", int64_tParam));

    uint8_t uint8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", uint8_tParam));

    uint16_t uint16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", uint16_tParam));

    uint32_t uint32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", uint32_tParam));

    uint64_t uint64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", uint64_tParam));

    float float_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", float_tParam));

    double double_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("String_Parameter", double_tParam));

    std::string string_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("String_Parameter", string_tParam));
}

TEST_F(SimpleConfigTest, typeErrorsInt)
{
    ASSERT_TRUE(section_.hasParameter("Int_Parameter"));
    const long int expInt = -123;
    bool boolParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Int_Parameter", boolParam));

    int8_t int8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Int_Parameter", int8_tParam));
    EXPECT_EQ(expInt, int8_tParam);

    int16_t int16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Int_Parameter", int16_tParam));
    EXPECT_EQ(expInt, int16_tParam);

    int32_t int32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Int_Parameter", int32_tParam));
    EXPECT_EQ(expInt, int32_tParam);

    int64_t int64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Int_Parameter", int64_tParam));
    EXPECT_EQ(expInt, int64_tParam);

    uint8_t uint8_tParam;
    EXPECT_THROW(section_.get("Int_Parameter", uint8_tParam), std::out_of_range);
    EXPECT_NE(expInt, uint8_tParam);

    uint16_t uint16_tParam;
    EXPECT_THROW(section_.get("Int_Parameter", uint16_tParam), std::out_of_range);
    EXPECT_NE(expInt, uint16_tParam);

    uint32_t uint32_tParam;
    EXPECT_THROW(section_.get("Int_Parameter", uint32_tParam), std::out_of_range);
    EXPECT_NE(expInt, uint32_tParam);

    uint64_t uint64_tParam;
    EXPECT_THROW(section_.get("Int_Parameter", uint64_tParam), std::out_of_range);
    EXPECT_NE(expInt, uint64_tParam);

    float float_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Int_Parameter", float_tParam));
    EXPECT_EQ(expInt, float_tParam);

    double double_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Int_Parameter", double_tParam));
    EXPECT_EQ(expInt, double_tParam);

    std::string string_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Int_Parameter", string_tParam));
}

TEST_F(SimpleConfigTest, typeErrorsUint)
{
    unsigned long long expUint = 234;
    ASSERT_TRUE(section_.hasParameter("Uint_Parameter"));
    bool boolParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Uint_Parameter", boolParam));

    int8_t int8_tParam;
    EXPECT_THROW(section_.get("Uint_Parameter", int8_tParam), std::out_of_range);
    EXPECT_NE(expUint, int8_tParam);

    int16_t int16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", int16_tParam));
    EXPECT_EQ(expUint, int16_tParam);

    int32_t int32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", int32_tParam));
    EXPECT_EQ(expUint, int32_tParam);

    int64_t int64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", int64_tParam));
    EXPECT_EQ(expUint, int64_tParam);

    uint8_t uint8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", uint8_tParam));
    EXPECT_EQ(expUint, uint8_tParam);

    uint16_t uint16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", uint16_tParam));
    EXPECT_EQ(expUint, uint16_tParam);

    uint32_t uint32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", uint32_tParam));
    EXPECT_EQ(expUint, uint32_tParam);

    uint64_t uint64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", uint64_tParam));
    EXPECT_EQ(expUint, uint64_tParam);

    float float_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", float_tParam));
    EXPECT_EQ(expUint, float_tParam);

    double double_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", double_tParam));
    EXPECT_EQ(expUint, double_tParam);

    std::string string_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Uint_Parameter", string_tParam));
}

TEST_F(SimpleConfigTest, typeErrorsSection)
{
    ASSERT_TRUE(section_.hasParameter("Level_2"));

    bool boolParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", boolParam));

    int8_t int8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", int8_tParam));

    int16_t int16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", int16_tParam));

    int32_t int32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", int32_tParam));

    int64_t int64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", int64_tParam));

    uint8_t uint8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", uint8_tParam));

    uint16_t uint16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", uint16_tParam));

    uint32_t uint32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", uint32_tParam));

    uint64_t uint64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", uint64_tParam));

    float float_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", float_tParam));

    double double_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", double_tParam));

    std::string string_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_TypeMismatch, section_.get("Level_2", string_tParam));

    ConfigSection localSection;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Level_2", localSection));
}
/*
 * Look for deliberate range errors
 */
TEST_F(SimpleConfigTest, rangeErrorsSection)
{
    int8_t int8_tParam;
    EXPECT_THROW(section_.get("Uint_Parameter", int8_tParam), std::out_of_range);

    int16_t int16_tParam;
    EXPECT_THROW(section_.get("Long_IntParameter", int16_tParam), std::out_of_range);

    int32_t int32_tParam;
    EXPECT_THROW(section_.get("Long_IntParameter", int32_tParam), std::out_of_range);

    int64_t int64_tParam;
    EXPECT_NO_THROW(section_.get("Long_IntParameter", int64_tParam));

    uint8_t uint8_tParam;
    EXPECT_THROW(section_.get("Int16_Parameter", uint8_tParam), std::out_of_range);

    uint16_t uint16_tParam;
    EXPECT_THROW(section_.get("Long_IntParameter", uint16_tParam), std::out_of_range);

    uint32_t uint32_tParam;
    EXPECT_THROW(section_.get("Long_IntParameter", uint32_tParam), std::out_of_range);

    uint64_t uint64_tParam;
    EXPECT_NO_THROW(section_.get("Long_IntParameter", uint64_tParam));

    // long long ints seem to vex lib config.
    EXPECT_EQ(ConfigSection::GS_OK, section_.get("Long_Long_IntParameter", uint64_tParam))
        << "Reading value too large for long long int";
    EXPECT_EQ(0x7FFFFFFFFFFFFFFFLLU, uint64_tParam);

    EXPECT_EQ(ConfigSection::GS_OK, section_.get("Large_Long_IntParameter", uint64_tParam))
        << "Reading value too large for long long int: " << uint64_tParam;
    EXPECT_EQ(0x7FFFFFFFFFFFFFFFLLU, uint64_tParam);

    EXPECT_EQ(ConfigSection::GS_OK, section_.get("VLarge_Long_IntParameter", uint64_tParam))
        << "Reading value way too large for long long int: " << uint64_tParam;
    EXPECT_EQ(0x7FFFFFFFFFFFFFFFLLU, uint64_tParam);

    EXPECT_THROW(section_.get("Neg_Large_Long_IntParameter", uint64_tParam), std::out_of_range)
        << "Reading neg value for long long int";

    EXPECT_THROW(section_.get("Neg_Small_Long_IntParameter", uint64_tParam), std::out_of_range)
        << "Reading neg value for long long int";

    float float_tParam;
    EXPECT_THROW(section_.get("Long_Float_Parameter", float_tParam), std::out_of_range);

    double double_tParam;
    EXPECT_NO_THROW(section_.get("Long_IntParameter", double_tParam));
}

/*
 * Look for deliberate undefined parameter errors
 */
TEST_F(SimpleConfigTest, definedErrorsSection)
{
    ASSERT_FALSE(section_.hasParameter("NoSuchElement"));

    bool boolParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", boolParam));

    int8_t int8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", int8_tParam));

    int16_t int16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", int16_tParam));

    int32_t int32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", int32_tParam));

    int64_t int64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", int64_tParam));

    uint8_t uint8_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", uint8_tParam));

    uint16_t uint16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", uint16_tParam));

    uint32_t uint32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", uint32_tParam));

    uint64_t uint64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", uint64_tParam));

    float float_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", float_tParam));

    double double_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", double_tParam));

    std::string string_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", string_tParam));

    ConfigSection localSection;
    EXPECT_EQ(ConfigSection::GetStatus::GS_Undefined, section_.get("NoSuchElement", localSection));
}

/*
 * Test loading of a section and that all values read ok.
RWMService['test_section_2'] = {"Bool_Parameter": True,
                                "Float_Parameter": 3.14159265359,
                                "String_Parameter": 'This is a string',
                                "Int_Parameter": -(123) ,
                                "Uint_Parameter": 234,
                                "Int16_Parameter": 32000,
                                "Long_IntParameter": 22147483647,#extra leading 2
                                "Long_Long_IntParameter":   0x8000000000000000,#max signed long long
                                "Large_Long_IntParameter":  0x8000000000000002,#Too large for long long int
                                "VLarge_Long_IntParameter": 0xFFFFFFFFFFFFFFFF,#Too large for long long int
                                "Neg_Small_Long_IntParameter": -1,
                                "Neg_Large_Long_IntParameter": -(0x7FFFFFFFFFFFFFFF),
                                "Long_Float_Parameter": 3.14e256,
                                "Underflow_Float": 3.14e-256,
 */
TEST_F(SimpleConfigTest, sectionRead)
{
    bool boolParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Bool_Parameter", boolParam));
    EXPECT_EQ(true, boolParam);

    float float_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Float_Parameter", float_tParam));
    EXPECT_NEAR(3.14159265359, float_tParam, 0.00001);

    double double_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Float_Parameter", double_tParam));
    EXPECT_NEAR(3.14159265359, float_tParam, 9e-8);

    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Long_Float_Parameter", double_tParam));
    EXPECT_NEAR(3.14e256, double_tParam, 1e252);

    std::string string_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("String_Parameter", string_tParam));
    EXPECT_EQ("This is a string", string_tParam);

    int32_t int32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Int_Parameter", int32_tParam));
    EXPECT_EQ(-123, int32_tParam);

    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Int16_Parameter", int32_tParam));
    EXPECT_EQ(32000, int32_tParam);

    uint16_t uint16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Uint_Parameter", uint16_tParam));
    EXPECT_EQ(234, uint16_tParam);

    int64_t int64_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Long_IntParameter", int64_tParam));
    EXPECT_EQ(22147483647, int64_tParam);

    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Neg_Small_Long_IntParameter", int64_tParam));
    EXPECT_EQ(-1, int64_tParam);

    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Neg_Large_Long_IntParameter", int64_tParam));
    EXPECT_EQ(-(0x7FFFFFFFFFFFFFFF), int64_tParam);
}

/*
 * Test loading of a sub section_
 */
TEST_F(SimpleConfigTest, subSectionRead)
{
    ConfigSection localSection;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, section_.get("Level_2", localSection));

    bool boolParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, localSection.get("Bool_Parameter", boolParam));

    float float_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, localSection.get("Float_Parameter", float_tParam));
    EXPECT_NEAR(3.1415926535914159265359141592653591415926535914159265359, float_tParam, 0.00001);

    std::string string_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, localSection.get("String_Parameter", string_tParam));
    EXPECT_EQ("This is a level 2 string", string_tParam);

    int32_t int32_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, localSection.get("Int_Parameter", int32_tParam));
    EXPECT_EQ(-234, int32_tParam);

    uint16_t uint16_tParam;
    EXPECT_EQ(ConfigSection::GetStatus::GS_OK, localSection.get("Uint_Parameter", uint16_tParam));
    EXPECT_EQ(567, uint16_tParam);
}

TEST(ConfigSection, getArray_returnsZeroWhenArrayNotFound)
{
    ConfigSection objectUnderTest;
    auto config = std::make_shared<Config>();
    config->readString("section:{ setting = [ 1, 2 ]; };");
    EXPECT_EQ(true, objectUnderTest.init("section", config));

    // When array is NOT found, verify:
    //    Input is untouched (any pre-existing data is unaltered).
    //    0 is returned.
    std::vector<std::uint16_t> array{std::numeric_limits<std::uint16_t>::max()};
    EXPECT_EQ(0, objectUnderTest.getArray("invalid setting name", array));
    EXPECT_EQ(1, array.size());
    EXPECT_EQ(std::numeric_limits<std::uint16_t>::max(), array.front());
}

TEST(ConfigSection, getArray)
{
    ConfigSection objectUnderTest;
    auto config = std::make_shared<Config>();
    config->readString("section:{ setting = [ 1, 2 ]; };");
    EXPECT_EQ(true, objectUnderTest.init("section", config));

    // When array is found, verify:
    //    Input only contains data in array (any pre-existing data is removed).
    //    Number of elements in array is returned.
    std::vector<std::uint16_t> array{6, 5, 4, 3};
    EXPECT_EQ(2, objectUnderTest.getArray("setting", array));
    EXPECT_EQ(2, array.size());
    EXPECT_EQ(1, array.front());
    EXPECT_EQ(2, array.back());
}
