/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Unit tests for @ref InterfaceParser

#include "gtest/gtest.h"
#include <config/InterfaceParser.h>

TEST(InterfaceParser, hasWriteOnlyInterfaces_detectsWriteOnlyInterfacesSection)
{
    InterfaceParser objectUnderTest;
    EXPECT_FALSE(objectUnderTest.hasWriteOnlyInterfaces());
    objectUnderTest.parseString("WriteOnlyInterfaces:{}");
    EXPECT_TRUE(objectUnderTest.hasWriteOnlyInterfaces());
}

TEST(InterfaceParser, getAllWriteOnlyInterfaces_retrievesAllWriteOnlyInterfacesContainingImplementationParam)
{
    // Verify InterfaceParser::getAllWriteOnlyInterfaces retrieves all sections within 'WriteOnlyInterfaces' section
    // that contain the parameter 'implementation'
    InterfaceParser objectUnderTest;
    EXPECT_TRUE(objectUnderTest.parseString(
        R"(WriteOnlyInterfaces:{ WithImplementationParam: { implementation: ""; }; WithoutImplementationParam: {}; };)"));
    const auto actual = objectUnderTest.getAllWriteOnlyInterfaces();
    EXPECT_EQ(1, actual.size());
    EXPECT_EQ("WriteOnlyInterfaces.WithImplementationParam", actual.front().getName());
}

TEST(InterfaceParser, hasWriteOnlyInterface)
{
    InterfaceParser objectUnderTest;
    EXPECT_TRUE(objectUnderTest.parseString(R"(WriteOnlyInterfaces:{ TestInterface: { implementation: ""; }; };)"));
    EXPECT_FALSE(objectUnderTest.hasWriteOnlyInterface("MissingInterface"));
    EXPECT_TRUE(objectUnderTest.hasWriteOnlyInterface("TestInterface"));
}

TEST(InterfaceParser, getWriteOnlyInterface)
{
    InterfaceParser objectUnderTest;
    objectUnderTest.parseString(R"(WriteOnlyInterfaces:{ TestInterface: { implementation: ""; }; };)");
    EXPECT_EQ(ConfigSection().getName(), objectUnderTest.getWriteOnlyInterface("MissingInterface").getName());
    EXPECT_EQ("WriteOnlyInterfaces.TestInterface", objectUnderTest.getWriteOnlyInterface("TestInterface").getName());
}

TEST(InterfaceParser, getWriteOnlyInterfacesSection)
{
    ConfigSection actual;
    InterfaceParser objectUnderTest;

    EXPECT_FALSE(objectUnderTest.getWriteOnlyInterfacesSection(actual));
    objectUnderTest.parseString(R"(WriteOnlyInterfaces:{ TestInterface: { implementation: ""; }; };)");
    EXPECT_TRUE(objectUnderTest.getWriteOnlyInterfacesSection(actual));
    EXPECT_EQ("WriteOnlyInterfaces", actual.getName());
}

TEST(InterfaceParser, hasReadOnlyInterfaces)
{
    InterfaceParser objectUnderTest;
    EXPECT_FALSE(objectUnderTest.hasReadOnlyInterfaces());
    objectUnderTest.parseString("ReadOnlyInterfaces:{}");
    EXPECT_TRUE(objectUnderTest.hasReadOnlyInterfaces());
}

TEST(InterfaceParser, hasReadOnlyInterface)
{
    InterfaceParser objectUnderTest;
    objectUnderTest.parseString(R"(ReadOnlyInterfaces:{ TestInterface: { implementation: ""; }; };)");
    EXPECT_FALSE(objectUnderTest.hasReadOnlyInterface("MissingInterface"));
    EXPECT_TRUE(objectUnderTest.hasReadOnlyInterface("TestInterface"));
}

TEST(InterfaceParser, getReadOnlyInterface)
{
    InterfaceParser objectUnderTest;
    objectUnderTest.parseString(R"(ReadOnlyInterfaces:{ TestInterface: { implementation: ""; }; };)");
    EXPECT_EQ(ConfigSection().getName(), objectUnderTest.getReadOnlyInterface("MissingInterface").getName());
    EXPECT_EQ("ReadOnlyInterfaces.TestInterface", objectUnderTest.getReadOnlyInterface("TestInterface").getName());
}

TEST(InterfaceParser, hasReadWriteInterfaces)
{
    InterfaceParser objectUnderTest;
    EXPECT_FALSE(objectUnderTest.hasReadWriteInterfaces());
    objectUnderTest.parseString("ReadWriteInterfaces:{}");
    EXPECT_TRUE(objectUnderTest.hasReadWriteInterfaces());
}

TEST(InterfaceParser, hasReadWriteInterface)
{
    InterfaceParser objectUnderTest;
    EXPECT_TRUE(objectUnderTest.parseString(R"(ReadWriteInterfaces:{ TestInterface: { implementation: ""; }; };)"));
    EXPECT_FALSE(objectUnderTest.hasReadWriteInterface("MissingInterface"));
    EXPECT_TRUE(objectUnderTest.hasReadWriteInterface("TestInterface"));
}

TEST(InterfaceParser, getReadWriteInterface)
{
    InterfaceParser objectUnderTest;
    objectUnderTest.parseString(R"(ReadWriteInterfaces:{ TestInterface: { implementation: ""; }; };)");
    EXPECT_EQ(ConfigSection().getName(), objectUnderTest.getReadWriteInterface("MissingInterface").getName());
    EXPECT_EQ("ReadWriteInterfaces.TestInterface", objectUnderTest.getReadWriteInterface("TestInterface").getName());
}

TEST(InterfaceParser, getAllReadOnlyInterfaces_retrievesAllReadOnlyInterfacesContainingImplementationParam)
{
    // Verify InterfaceParser::getAllReadOnlyInterfaces retrieves all sections within 'ReadOnlyInterfaces' section
    // that contain the parameter 'implementation'
    InterfaceParser objectUnderTest;
    EXPECT_TRUE(objectUnderTest.parseString(
        R"(ReadOnlyInterfaces:{ WithImplementationParam: { implementation: ""; }; WithoutImplementationParam: {}; };)"));
    const auto actual = objectUnderTest.getAllReadOnlyInterfaces();
    EXPECT_EQ(1, actual.size());
    EXPECT_EQ("ReadOnlyInterfaces.WithImplementationParam", actual.front().getName());
}

TEST(InterfaceParser, getReadWriteInterfacesSection)
{
    ConfigSection actual;
    InterfaceParser objectUnderTest;

    EXPECT_FALSE(objectUnderTest.getReadWriteInterfacesSection(actual));
    objectUnderTest.parseString(R"(ReadWriteInterfaces:{ TestInterface: { implementation: ""; }; };)");
    EXPECT_TRUE(objectUnderTest.getReadWriteInterfacesSection(actual));
    EXPECT_EQ("ReadWriteInterfaces", actual.getName());
}

TEST(InterfaceParser, getAllReadWriteInterfaces_retrievesAllReadWriteInterfacesContainingImplementationParam)
{
    // Verify InterfaceParser::getAllReadWriteInterfaces retrieves all sections within 'ReadWriteInterfaces' section
    // that contain the parameter 'implementation'
    InterfaceParser objectUnderTest;
    EXPECT_TRUE(objectUnderTest.parseString(
        R"(ReadWriteInterfaces:{ WithImplementationParam: { implementation: ""; }; WithoutImplementationParam: {}; };)"));
    const auto actual = objectUnderTest.getAllReadWriteInterfaces();
    EXPECT_EQ(1, actual.size());
    EXPECT_EQ("ReadWriteInterfaces.WithImplementationParam", actual.front().getName());
}
