/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file outputConfig.h
 *
 * @brief utilities for handling config files
 * @author Yimu Wang (yimu.wang@aptiv.com)
 * @author Bing-Jui Ho (bing-jui.ho@aptiv.com)
 * @date 01/12/2018
 * @remark
 * Initial implementation.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#ifndef CONFIGUTILS_OUTPUTCONFIG_H
#define CONFIGUTILS_OUTPUTCONFIG_H

#include <assert.h>
#include <cxxabi.h>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

namespace config {

const auto DICT_START = "{";

const auto DICT_END = "}";

const auto END_WORD = ",";

const auto STRING_SEP = "\"";

const auto VEC_START = "[";

const auto VEC_END = "]";

class OutputConfig
{
public:
    OutputConfig()
    {
        // set default precision to double
        setPrecision(std::numeric_limits<double>::digits10 + 1);
    }

    OutputConfig(size_t indent)
        : indent_(indent)
    {
        // set default precision to double
        setPrecision(std::numeric_limits<double>::digits10 + 1);
    }

    void setPrecision(const int precision) { os_ << std::setprecision(precision); }

    // Calling order
    // 1. try overload functions.
    // 2. try template function (SFINAE).
    template<typename T>
    void operator()(const std::string& name, T t, const std::string& comments = "")
    {
        handleComments(comments, t);
        handleEntry(name, t);

        constexpr bool EXTRA_LINE = true;
        endEntry(EXTRA_LINE);
    }

    template<typename T>
    void handleComments(const std::string& comments, const T& t)
    {
        if (comments.length() > 0) {
            os_ << std::string(indent_, ' ') << "# " << comments << std::endl;
        }

        // TODO: Compiler optimization may remove full name
        constexpr bool SHOW_TYPE = true;
        if (SHOW_TYPE) {
            handleType(t);
        }
    }

    template<typename T>
    void handleType(const T& t)
    {
        int status;
        char* realname = abi::__cxa_demangle(typeid(t).name(), 0, 0, &status);
        os_ << std::string(indent_, ' ') << "# Type in C++: " << realname << std::endl;
        free(realname);
    }

    // Default string name is ugly
    void handleType(const std::string& /* t */)
    {
        os_ << std::string(indent_, ' ') << "# Type in C++: string" << std::endl;
    }

    // Default name for vector of strings is ugly
    void handleType(const std::vector<std::string>& /* t */)
    {
        os_ << std::string(indent_, ' ') << "# Type in C++: vector<string>" << std::endl;
    }

    template<typename T>
    typename std::enable_if<!std::is_fundamental<T>::value, void>::type handleEntry(const std::string& name, T t)
    {
        dictFirstPart(name);
        OutputConfig ar(indent_);
        t.serializeConfig(ar);
        os_ << ar.str();
    }

    template<typename T>
    typename std::enable_if<std::is_fundamental<T>::value, void>::type handleEntry(const std::string& name, T t)
    {
        // Python dict
        dictFirstPart(name);
        os_ << t << END_WORD;
    }

    // // overload for string
    void handleEntry(const std::string& name, std::string t)
    {
        // Python dict
        dictFirstPart(name);
        os_ << STRING_SEP << t << STRING_SEP << END_WORD;
    }

    // overload for bool
    void handleEntry(const std::string& name, bool t)
    {
        dictFirstPart(name);
        if (t) {
            os_ << "True" << END_WORD;
        } else {
            os_ << "False" << END_WORD;
        }
    }

    // overload for vector
    template<typename T>
    void handleEntry(const std::string& name, std::vector<T> vec)
    {
        // Python dict
        dictFirstPart(name);

        // handle the case when vector is empty
        if (vec.empty()) {
            os_ << VEC_START << VEC_END << END_WORD;
            return;
        }

        T last = vec.back();
        vec.pop_back();

        os_ << VEC_START;
        for (const auto& v : vec) {
            os_ << v << ",";
        }
        os_ << last << VEC_END << END_WORD;
    }

    // overload for vector of strings
    void handleEntry(const std::string& name, std::vector<std::string> vec)
    {
        // Python dict
        dictFirstPart(name);

        // handle the case when vector is empty
        if (vec.empty()) {
            os_ << VEC_START << VEC_END << END_WORD;
            return;
        }

        std::string last = vec.back();
        vec.pop_back();

        os_ << VEC_START;
        for (const auto& v : vec) {
            os_ << STRING_SEP << v << STRING_SEP << ",";
        }
        os_ << STRING_SEP << last << STRING_SEP << VEC_END << END_WORD;
    }

    void start()
    {
        // meaning: "name" : {
        constexpr size_t START_INDENT = 0;
        os_ << std::string(START_INDENT, ' ') << std::endl << std::string(indent_, ' ') << DICT_START << std::endl;
        indent_ += 4;
    }

    void end()
    {
        size_t indent = indent_ - 4;
        // assert(indent_ >= 0);
        if (indent_ <= indent) {
            throw std::underflow_error("indent size underflow");
        }

        indent_ = indent;
        os_ << std::string(indent_, ' ') << DICT_END << END_WORD;
    }

    std::string str() const { return os_.str(); }

    void dictFirstPart(const std::string& name)
    {
        os_ << std::string(indent_, ' ') << STRING_SEP << name << STRING_SEP << ": ";
    }

    void endEntry(const bool& extraline)
    {
        if (extraline) {
            os_ << std::endl << std::endl;
        } else {
            os_ << std::endl;
        }
    }

private:
    std::ostringstream os_;
    size_t indent_ = 0;
};

} // namespace config

#endif
