/*!
***********************************************************************************************************************
* APTIV PRIVILEGED AND CONFIDENTIAL
*
* @file ConfigUtilsUnitTest
* @product Tests
* @version 1 <Release Number>
* @released TODO <date>
* @updated 2019-04-02 <date>
* @brief This file implements a unit-test suite for Configuration Utilities
*
* Copyright (c) 2019 - Aptiv. All rights reserved.
* NOTE: Removing this heading is prohibited.
***********************************************************************************************************************
*/

#include <gtest/gtest.h>

#include <config/ConfigParser.h>
#include <configUtils/configIo.h>

#include "ConfigUtilsTestStruct.h"

#ifndef TESTDIR
#define TESTDIR "platform/middleware/libs/configUtils/test"
#endif

class InputConfigTest : public ::testing::Test
{
protected:
    /*
     * cfg is the intermediate data format class Task uses to store data when loading from python config file
     * please look into scripts/dump_to_cfg to see how to generate cfg from python config file
     */
    void SetUp() override
    {
        // parse cfg file
        ConfigParser parser;
        ASSERT_TRUE(parser.parseFile(std::string(TESTDIR) + "/data/test_config_file.cfg"))
            << "error parsing the python config file";

        // initialize inputConfig
        ASSERT_TRUE(parser.getSection("root", config_.getConfigSection())) << "error getting section";
    }

    void TearDown() override {}

    config::InputConfig config_;
};

TEST(OutputConfigTest, Start)
{
    config::OutputConfig config;
    std::ostringstream oss;

    // Test Case 1: test start() behaves as expected
    config.start();
    oss << '\n' << config::DICT_START << '\n';
    EXPECT_EQ(oss.str(), config.str());
}

TEST(OutputConfigTest, End)
{
    config::OutputConfig config;
    std::ostringstream oss;

    // Test Case 1: indent size underflow
    ASSERT_THROW(config.end(), std::underflow_error);

    // Test Case 2: test end() behaves as expected
    config.start();
    config.end();
    oss << '\n' << config::DICT_START << '\n' << config::DICT_END << config::END_WORD;
    EXPECT_EQ(oss.str(), config.str());
}

TEST(OutputConfigTest, HandleComments)
{
    config::OutputConfig config;
    std::ostringstream oss;

    // Test Case 1: length of comment is zero and type of variable is boolean
    config.handleComments("", true);
    oss << "# Type in C++: bool\n";
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 2: length of comment is not zero and type of variable is double
    config.handleComments("Test Case 2", 2.0);
    oss << "# Test Case 2\n";
    oss << "# Type in C++: double\n";
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 3: length of comment is not zero and type of variable is integer
    config.handleComments("Test Case 3", 3);
    oss << "# Test Case 3\n";
    oss << "# Type in C++: int\n";
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 4: length of comment is not zero and type of variable is string
    config.handleComments("Test Case 4", std::string("It is a string"));
    oss << "# Test Case 4\n";
    oss << "# Type in C++: string\n";
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 5: length of comment is not zero and type of variable is vecotr<string>
    config.handleComments("Test Case 5", std::vector<std::string>{"string #1", "string #2", "string #3"});
    oss << "# Test Case 5\n";
    oss << "# Type in C++: vector<string>\n";
    EXPECT_EQ(oss.str(), config.str());
}

TEST(OutputConfigTest, HandleEntry)
{
    config::OutputConfig config;
    std::ostringstream oss;

    // Test Case 1: bool NameOfTheBool = true
    config.handleEntry("NameOfTheBool", true);
    // Expected: "NameOfTheBool": "True",
    oss << config::STRING_SEP << "NameOfTheBool" << config::STRING_SEP << ": ";
    oss << "True" << config::END_WORD;
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 2: bool NameOfTheBool = false
    config.handleEntry("NameOfTheBool", false);
    // Expected: "NameOfTheBool": "False",
    oss << config::STRING_SEP << "NameOfTheBool" << config::STRING_SEP << ": ";
    oss << "False" << config::END_WORD;
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 3: bool NameOfTheDouble = 3.0
    config.handleEntry("NameOfTheDouble", 3.0);
    // Expected: "NameOfTheDouble": 3.0,
    oss << config::STRING_SEP << "NameOfTheDouble" << config::STRING_SEP << ": ";
    oss << 3.0 << config::END_WORD;
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 4: int NameOfTheInteger = 4
    config.handleEntry("NameOfTheInteger", 4);
    // Expected: "NameOfTheInteger": 4,
    oss << config::STRING_SEP << "NameOfTheInteger" << config::STRING_SEP << ": ";
    oss << 4 << config::END_WORD;
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 5: std::string NameOfTheString = std::string("It is a string")
    config.handleEntry("NameOfTheString", std::string("It is a string"));
    // Expected: "NameOfTheString": "It is a string",
    oss << config::STRING_SEP << "NameOfTheString" << config::STRING_SEP << ": ";
    oss << config::STRING_SEP << std::string("It is a string") << config::STRING_SEP << config::END_WORD;
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 6: std::vector<int> NameOfTheVectorOfIntegers = std::vector<int>{1,2,3}
    config.handleEntry("NameOfTheVectorOfIntegers", std::vector<int>{1, 2, 3});
    // Expected: "NameOfTheVectorOfIntegers": [1,2,3,],
    oss << config::STRING_SEP << "NameOfTheVectorOfIntegers" << config::STRING_SEP << ": ";
    oss << config::VEC_START << "1,2,3" << config::VEC_END << config::END_WORD;
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 7: std::vector<std::string> NameOfTheVectorOfStrings = std::vector<std::string>{...}
    config.handleEntry("NameOfTheVectorOfStrings", std::vector<std::string>{"string #1", "string #2", "string #3"});
    // Expected: "NameOfTheVectorOfStrings": ["string #1","string #2","string #3",],
    oss << config::STRING_SEP << "NameOfTheVectorOfStrings" << config::STRING_SEP << ": ";
    oss << config::VEC_START << "\"string #1\",\"string #2\",\"string #3\"" << config::VEC_END << config::END_WORD;
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 8: OutputConfigSimpleTestParams NameOfTheTestParams;
    config.handleEntry("NameOfTheTestParams", OutputConfigSimpleTestParams());
    // EXPECTED: "NameOfTheTestParams": \n{\n    # Type in C++: bool\n    \"testParam1\": False,\n\n},
    oss << config::STRING_SEP << "NameOfTheTestParams" << config::STRING_SEP << ": ";
    oss << "\n{\n    # Type in C++: bool\n    \"testParam1\": False,\n\n},";
    EXPECT_EQ(oss.str(), config.str());

    // Test Case 9: size_t NameOfTheSizeT = 100
    config.handleEntry("NameOfTheSizeT", static_cast<size_t>(100));
    // Expected: "NameOfTheSizeT": 100,
    oss << config::STRING_SEP << "NameOfTheSizeT" << config::STRING_SEP << ": ";
    oss << 100 << config::END_WORD;
    EXPECT_EQ(oss.str(), config.str());
}

TEST(OutputConfigTest, StructToPyConfig)
{
    // store the data into file
    std::ifstream ifs(std::string(TESTDIR) + "/data/test_config_file.txt");
    std::string expected((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));
    // remove the extra newline read - this might not be an ideal solution but work for now
    expected.pop_back();
    ifs.close();

    // construct C++ struct
    TestParams testParameters;
    std::cout << config::toPythonDict(testParameters, "testParameters") << std::endl;
    std::cout << expected << std::endl;
    EXPECT_EQ(expected, config::toPythonDict(testParameters, "testParameters"));
}

TEST_F(InputConfigTest, CheckConfigExist)
{
    // Test Case 1: config exists
    EXPECT_NO_THROW(config_.checkConfigExist("testParameters"));
    // Test Case 2: config does not exist
    ASSERT_THROW(config_.checkConfigExist("testParameter"), std::runtime_error);
    ASSERT_THROW(config_.checkConfigExist("testParams"), std::runtime_error);
}

TEST_F(InputConfigTest, CfgToStruct)
{
    TestParams testParams1, testParams2;
    testParams1.aaa = 10000;
    testParams1.s = "";

    // Test Case 1: make sure the values are changed
    EXPECT_NE(testParams1.aaa, testParams2.aaa);
    EXPECT_NE(testParams1.s, testParams2.s);

    // load cfg data to struct testParams
    config_("testParameters", testParams1);

    // Test Case 2: make sure the values are loaded correctly
    EXPECT_EQ(testParams1.aaa, testParams2.aaa);
    EXPECT_EQ(testParams1.s, testParams2.s);

    // Test Case 3: load a parameter of type string
    EXPECT_EQ(testParams1.testClass.s, testParams2.testClass.s);

    // Test Case 4: load a parameter of type bool
    EXPECT_EQ(testParams1.testStruct.testClass.d, testParams2.testStruct.testClass.d);

    // Test Case 5: load a parameter of type unsigned int
    EXPECT_EQ(testParams1.testStruct.a, testParams2.testStruct.a);

    // Test Case 6: load a parameter of type double
    EXPECT_EQ(testParams1.testClass.bbbb, testParams2.testClass.bbbb);

    // Test Case 7: load a parameter of type vector<int>
    for (unsigned i = 0; i < testParams1.testClass.vec.size(); i++) {
        EXPECT_EQ(testParams1.testClass.vec[i], testParams2.testClass.vec[i]);
    }
}

//=== MAIN =============================================================================================================

/*!
 * @brief Main executable for ConfigUtils' unit tests
 *
 * @param argc : Number of command-line arguments
 * @param argv : Array of c-strings
 * @return int : Executable's error code (should be zero)
 */
int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
