/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file scs/conversions.cpp
 * @product coordinate-transform
 * @version 0.1
 * @released 05-Nov-2018
 * @updated 17-Nov-2018
 * @brief Conversions for different transforms backends to be used alongside SimpleComms.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 ***********************************************************************************************************************
 */

#include "IdlHelper/recgeometryHelper.h"
#include <TimeStamp/otimeConversions.h>
#include <coordinate-transform/conversions.hpp>

namespace aptv {
namespace ct {
namespace scs {
namespace bullet {

btTransform fromMsg(const av::msg::TransformStamped& msg)
{
    return btTransform(btQuaternion(msg.rotation[0], msg.rotation[1], msg.rotation[2], msg.rotation[3]),
                       btVector3(msg.translation[0], msg.translation[1], msg.translation[2]));
}

av::msg::TransformStamped toMsg(const btTransform& bt,
                                const otimePoint& stamp,
                                const std::string& parentFrame,
                                const std::string& childFrame)
{
    btVector3 translation(bt.getOrigin());
    btQuaternion rotation(bt.getRotation());
    av::msg::TransformStamped msg;
    msg.timeStamp = av::timePointFromOtimePoint(stamp);
    msg.parentFrame = parentFrame;
    msg.childFrame = childFrame;
    msg.translation[0] = translation.getX();
    msg.translation[1] = translation.getY();
    msg.translation[2] = translation.getZ();
    msg.rotation[0] = rotation.getX();
    msg.rotation[1] = rotation.getY();
    msg.rotation[2] = rotation.getZ();
    msg.rotation[3] = rotation.getW();
    return msg;
}
} // namespace bullet

namespace storage {
TransformStorage fromMsg(const av::msg::TransformStamped& msg)
{
    TransformStorage ts;
    // convert otime to microseconds
    otimePoint otp(av::otimePointFromTimePoint(msg.timeStamp));
    ottotime::otimeDuration otd(otp);
    ts.stamp = Time(Duration(otd.totalMicroSeconds()));
    ts.parentFrame = msg.parentFrame;
    ts.childFrame = msg.childFrame;
    ts.translation = btVector3(msg.translation[0], msg.translation[1], msg.translation[2]);
    ts.rotation = btQuaternion(msg.rotation[0], msg.rotation[1], msg.rotation[2], msg.rotation[3]);
    ts.isStatic = msg.isStatic;
    return ts;
}

av::msg::TransformStamped toMsg(const TransformStorage& ts)
{
    av::msg::TransformStamped msg;
    // convert microseconds to otime
    auto stamp_us = timeToMicroseconds(ts.stamp);
    ottotime::otimeDuration otd{ottotime::oMicroSeconds(stamp_us)};
    msg.timeStamp = av::timePointFromOtimePoint(ottotime::otimePoint(otd));
    msg.parentFrame = ts.parentFrame;
    msg.childFrame = ts.childFrame;
    msg.translation[0] = ts.translation.getX();
    msg.translation[1] = ts.translation.getY();
    msg.translation[2] = ts.translation.getZ();
    msg.rotation[0] = ts.rotation.getX();
    msg.rotation[1] = ts.rotation.getY();
    msg.rotation[2] = ts.rotation.getZ();
    msg.rotation[3] = ts.rotation.getW();
    return msg;
}
} // namespace storage

av::msg::TransformStampedList toTransformList(const av::msg::SmoothedLocalizationPose& smoothedPose)
{
    av::msg::TransformStampedList transformList;
    RecQuaternion quatENU;
    quatENU.quat[0] = smoothedPose.poseENU.rotation.x;
    quatENU.quat[1] = smoothedPose.poseENU.rotation.y;
    quatENU.quat[2] = smoothedPose.poseENU.rotation.z;
    quatENU.quat[3] = smoothedPose.poseENU.rotation.w;
    RecTransform3D transformENU(quatENU);
    transformENU.translate(
        RecVector3D(smoothedPose.poseENU.position.x, smoothedPose.poseENU.position.y, smoothedPose.poseENU.position.z));
    const RecPose3D poseENU(transformENU);

    {
        av::msg::TransformStamped ts;
        aptv::ct::scs::fillInTransAndRot(poseENU, ts);
        ts.timeStamp = smoothedPose.timeStamp;
        ts.parentFrame = "earth";
        ts.childFrame = "odom";
        transformList.transforms.push_back(ts);
    }

    {
        av::msg::TransformStamped ts;
        aptv::ct::scs::fillInTransAndRot(poseENU, ts);
        ts.timeStamp = smoothedPose.timeStamp;
        ts.parentFrame = "earth";
        ts.childFrame = "base_link";
        transformList.transforms.push_back(ts);
    }

    {
        av::msg::TransformStamped ts;
        aptv::ct::scs::fillInTransAndRot(poseENU, ts);
        ts.timeStamp = smoothedPose.timeStamp;
        ts.parentFrame = "earth";
        ts.childFrame = "base_footprint";
        transformList.transforms.push_back(ts);
    }

    {
        av::msg::TransformStamped ts;
        auto map2D = poseENU;
        map2D.rot2 = 0.0;
        map2D.rot3 = 0.0;
        aptv::ct::scs::fillInTransAndRot(map2D, ts);
        ts.timeStamp = smoothedPose.timeStamp;
        ts.parentFrame = "earth";
        ts.childFrame = "base_footprint_upright";
        transformList.transforms.push_back(ts);
    }

    {
        av::msg::TransformStamped ts;
        RecTransform3D earthMap;
        earthMap.identity();
        earthMap.translate(RecVector3D(0, 0, poseENU.z));
        aptv::ct::scs::fillInTransAndRot(earthMap, ts);
        ts.timeStamp = smoothedPose.timeStamp;
        ts.parentFrame = "earth";
        ts.childFrame = "map";
        transformList.transforms.push_back(ts);
    }
    return transformList;
}

RecPose3D getPoseInENU(const RecPose3D& transform)
{
    RecTransform3D NEDToENU;
    NEDToENU.identity();
    NEDToENU.matrixPP[0][0] = 0;
    NEDToENU.matrixPP[0][1] = 1;
    NEDToENU.matrixPP[1][0] = 1;
    NEDToENU.matrixPP[1][1] = 0;
    NEDToENU.matrixPP[2][2] = -1;
    const RecTransform3D poseNED = transform;
    RecTransform3D bodyFrame;
    bodyFrame.identity();
    bodyFrame.rotateAboutXAxis(M_PI);
    return NEDToENU * poseNED * bodyFrame;
}

av::geometry::RecPose3D getPoseInENU(const av::geometry::RecPose3D& transformMsg)
{
    RecPose3D poseInENU = getPoseInENU(idlhelper::getRecPose3D(transformMsg));
    return idlhelper::getRecPose3DMsg(poseInENU);
}

/******************************************
 * V' = R*V*R_t
 * When R is NEDToENU, V' is basically just a re-orientated V.
 * i.e. V = [x,y,z], V' = [y,x,z]
 * When R is OttoLocal2NuLocal,V'=V
 ******************************************/
RecPose3D getSTDInENU(const RecPose3D& transform)
{
    RecPose3D poseSTDInENU;
    poseSTDInENU.x = transform.y;
    poseSTDInENU.y = transform.x;
    poseSTDInENU.z = transform.z;
    poseSTDInENU.rot1 = transform.rot1;
    poseSTDInENU.rot2 = transform.rot2;
    poseSTDInENU.rot3 = transform.rot3;
    return poseSTDInENU;
}

av::geometry::RecPose3D getSTDInENU(const av::geometry::RecPose3D& transformMsg)
{
    RecPose3D STDInENU = getSTDInENU(idlhelper::getRecPose3D(transformMsg));
    return idlhelper::getRecPose3DMsg(STDInENU);
}

/******************************************
 * Acceleration is in world frame. Therefore,
 * only world frame transfrom needs to be taken
 * into account.
 ******************************************/
RecPoint3D getAccelerationInENU(const RecPoint3D& acc)
{
    RecTransform3D NEDToENU;
    NEDToENU.identity();
    NEDToENU.matrixPP[0][0] = 0;
    NEDToENU.matrixPP[0][1] = 1;
    NEDToENU.matrixPP[1][0] = 1;
    NEDToENU.matrixPP[1][1] = 0;
    NEDToENU.matrixPP[2][2] = -1;

    return NEDToENU * acc;
}

av::geometry::RecPoint3D getAccelerationInENU(const av::geometry::RecPoint3D& accMsg)
{
    RecPoint3D acc = getAccelerationInENU(idlhelper::getRecPoint3D(accMsg));
    return av::geometry::RecPoint3D(acc.x, acc.y, acc.z);
}

/******************************************
 * We sepeate the transformation of linear velocity and rotational velocity
 * because they are defined with respect to different frames.
 ******************************************/

RecDifferentialPose3D getVelocityInENU(const RecDifferentialPose3D& velocity)
{
    RecPoint3D linearVelocityInWorldNEDFrame(velocity.x, velocity.y, velocity.z);

    /* Angular Rates NED Local frame */
    double wx = velocity.rot3;
    double wy = velocity.rot2;
    double wz = velocity.rot1;
    RecPoint3D rotationVectorInLocalFrame(wx, wy, wz);

    RecTransform3D NEDToENU;
    NEDToENU.identity();
    NEDToENU.matrixPP[0][0] = 0;
    NEDToENU.matrixPP[0][1] = 1;
    NEDToENU.matrixPP[1][0] = 1;
    NEDToENU.matrixPP[1][1] = 0;
    NEDToENU.matrixPP[2][2] = -1;
    const RecPoint3D linearVelocityENU = NEDToENU * linearVelocityInWorldNEDFrame;

    RecTransform3D bodyFrame;
    bodyFrame.identity();
    bodyFrame.rotateAboutXAxis(M_PI);
    RecTransform3D rotationalVelocity;
    const RecPoint3D rotationalVelocityNuLocal = bodyFrame * rotationVectorInLocalFrame;

    /* rot1 -> yaw_rate -> rotationalVelocityNuLocal.z
     * rot2 -> pitch_rate -> rotationalVelocityNuLocal.y
     * rot3 -> roll_rate -> rotationalVelocityNuLocal.x
     */
    RecDifferentialPose3D translatedVelocity(linearVelocityENU.x,
                                             linearVelocityENU.y,
                                             linearVelocityENU.z,
                                             rotationalVelocityNuLocal.z,
                                             rotationalVelocityNuLocal.y,
                                             rotationalVelocityNuLocal.x);
    return translatedVelocity;
}

av::geometry::RecDifferentialPose3D getVelocityInENU(const av::geometry::RecDifferentialPose3D& velocityMsg)
{
    RecDifferentialPose3D velocity = getVelocityInENU(idlhelper::getRecDifferentialPose3D(velocityMsg));
    return idlhelper::getRecDifferentialPose3DMsg(velocity);
}

/******************************************
 * V' = R*V*R_t
 * When R is NEDToENU, V' is basically just a re-orientated V.
 * i.e. V = [x,y,z], V' = [y,x,z]
 * When R is OttoLocal2NuLocal,V'=V
 ******************************************/
RecDifferentialPose3D getVelocitySTDInENU(const RecDifferentialPose3D& velocitySTD)
{
    RecDifferentialPose3D velocitySTDInENU;
    // v is in world frame
    velocitySTDInENU.x = velocitySTD.y;
    velocitySTDInENU.y = velocitySTD.x;
    velocitySTDInENU.z = velocitySTD.z;
    // w is in local frame
    velocitySTDInENU.rot1 = velocitySTD.rot1;
    velocitySTDInENU.rot2 = velocitySTD.rot2;
    velocitySTDInENU.rot3 = velocitySTD.rot3;
    return velocitySTDInENU;
}

av::geometry::RecDifferentialPose3D getVelocitySTDInENU(const av::geometry::RecDifferentialPose3D& velocitySTDMsg)
{
    RecDifferentialPose3D velocitySTD = getVelocitySTDInENU(idlhelper::getRecDifferentialPose3D(velocitySTDMsg));
    return idlhelper::getRecDifferentialPose3DMsg(velocitySTD);
}

void fillInTransAndRot(const RecPose3D& pENU, av::msg::TransformStamped& tfMsg)
{
    tfMsg.translation[0] = pENU.x;
    tfMsg.translation[1] = pENU.y;
    tfMsg.translation[2] = pENU.z;

    btQuaternion quatFromEuler(pENU.rot1, pENU.rot2, pENU.rot3); // [yaw, pitch, roll]
    tfMsg.rotation[0] = quatFromEuler.getX();
    tfMsg.rotation[1] = quatFromEuler.getY();
    tfMsg.rotation[2] = quatFromEuler.getZ();
    tfMsg.rotation[3] = quatFromEuler.getW();
}

} // namespace scs

RecPose3D getPoseFromTransform(const TransformStorage& tf)
{
    RecQuaternion q;
    const auto& r = tf.rotation;
    q.quat[0] = r.x();
    q.quat[1] = r.y();
    q.quat[2] = r.z();
    q.quat[3] = r.w();
    RecTransform3D pose(q);
    const auto& o = tf.translation;
    pose.translate(RecVector3D(o.x(), o.y(), o.z()));
    return pose;
}

} // namespace ct
} // namespace aptv
