/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file scs/listener.hpp
 * @product coordinate-transform
 * @version 0.1
 * @released 05-Nov-2018
 * @updated 17-Nov-2018
 * @brief Header for ScsTransformListener.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 ***********************************************************************************************************************
 */

#ifndef __CT_SCS_LISTENER_H_
#define __CT_SCS_LISTENER_H_

#include "conversions.hpp"
#include "ct/transformer.hpp"
// SimpleComms
#include <interfaces/Generic/Input/Abstract.h>
#include <interfaces/SmoothedLocalizationPoseList/SmoothedLocalizationPoseList.h>
#include <interfaces/TransformStampedList/TransformStampedList.h>
// Threads
#include <atomic>
#include <thread>

namespace aptv {
namespace ct {

/*!
   \brief This class provides an easy way to request and receive coordinate frame transform information through
   \b SimpleComms.
*/
class ScsTransformListener
{
public:
    /*!
       \brief default Constructor
       \param transformer object to be updated every time we receive a new coordinate frame transform
       \param spinThread if true, the listener starts a new thread to automatically call ScsTransformListener::update.
    */
    ScsTransformListener(Transformer& transformer, bool threaded = true);

    /*!
       \brief Destructor
    */
    ~ScsTransformListener();

    /*!
       \brief Connect to SimpleComms using the provided interface
       \param dynamicTransformsInterface interface for the dynamic Transforms
       \param taskParameters the parameters of the task instantiating the transform listener
       \param smoothedLocalizationPoseInterface interface for the smoothed transforms list
       \return true if the connection is successful, false otherwise.
    */
    bool connect(const std::shared_ptr<GenericInput<av::msg::TransformStampedList>>& dynamicTransformsInterface,
                 ConfigSection& taskParameters,
                 const std::shared_ptr<GenericInput<av::msg::SmoothedLocalizationPoseList>>&
                     smoothedLocalizationPoseInterface = nullptr);

    /*!
       \brief disconnect from SimpleComms
       \return true if the disconnection is successful, false otherwise.
    */
    bool disconnect();

    /*!
       \brief Non-blocking method for dynamic transforms that tries to read from the SimpleComms channel.
       If new information is available, it'll updates the ScsTransformListener::transformer_ object.
       \param blocking if true, the update call blocks until a new message is received
    */
    void dynamicTransformsUpdate(const bool blocking);

    /// \brief Update the transforms with the given message.
    void updateTransforms(const av::msg::TransformStampedList& transformMessage);

    static void loadStaticTransform(Transformer& transformer, ConfigSection& taskParameters);

private:
    bool usingDedicatedThread_{true}; /*!< Flag to indicate if we are using a dedicated thread */
    // Transformer
    Transformer& transformer_; /*!< object that stores the coordinate transform frames */

    /// The input interface to receive TransformStamped messages for dynamic transforms
    std::shared_ptr<GenericInput<av::msg::TransformStampedList>> dynamicTransformsInterface_{nullptr};
    // The input interface to receive SmoothedLocalizationPoseList messages for dynamic transforms update
    std::shared_ptr<GenericInput<av::msg::SmoothedLocalizationPoseList>> smoothedLocalizationPoseInterface_{nullptr};
    // Dedicated thread
    std::atomic<bool> doneFlag_{false}; /*!< Flag used to stop running the dedicated thread */
    std::thread ctThreadDynamic_;       /*!< Thread that will call the update function for dynamic transforms */

    /*!
       \brief Callback that is executed in the dedicated thread for dynamic transforms
    */
    void dynamicTransformsListenerCallback();
};
using ScsTransformListenerPtr = std::shared_ptr<ScsTransformListener>;

} // namespace ct
} // namespace aptv

#endif
