/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file scs/listener.cpp
 * @product coordinate-transform
 * @version 0.1
 * @released 05-Nov-2018
 * @updated 17-Nov-2018
 * @brief Source code for ScsTransformListener.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 ***********************************************************************************************************************
 */

#include <coordinate-transform/listener.hpp>
#include <libconfig.h++>

#include "lib/avlog/avlog.hpp"

namespace aptv {
namespace ct {

ScsTransformListener::ScsTransformListener(Transformer& transformer, bool threaded)
    : usingDedicatedThread_{threaded}
    , transformer_{transformer}
{}

ScsTransformListener::~ScsTransformListener()
{
    disconnect();
}

bool ScsTransformListener::connect(
    const std::shared_ptr<GenericInput<av::msg::TransformStampedList>>& dynamicTransformsInterface,
    ConfigSection& taskParameters,
    const std::shared_ptr<GenericInput<av::msg::SmoothedLocalizationPoseList>>& smoothedLocalizationPoseInterface)
{
    if (dynamicTransformsInterface == nullptr) {
        return false;
    }
    dynamicTransformsInterface_ = dynamicTransformsInterface;
    if (smoothedLocalizationPoseInterface) {
        smoothedLocalizationPoseInterface_ = smoothedLocalizationPoseInterface;
    }
    loadStaticTransform(transformer_, taskParameters);

    if (usingDedicatedThread_) {
        ctThreadDynamic_ = std::thread(&ScsTransformListener::dynamicTransformsListenerCallback, this);
    }
    return true;
}

bool ScsTransformListener::disconnect()
{
    doneFlag_.store(true);
    // If connect() has been called and the callback thread is running, unblock
    // the interface so get() returns immediately, and wait for the thread to join.
    if (usingDedicatedThread_ && smoothedLocalizationPoseInterface_) {
        smoothedLocalizationPoseInterface_->unblock();
    }
    if (usingDedicatedThread_ && dynamicTransformsInterface_) {
        dynamicTransformsInterface_->unblock();
        ctThreadDynamic_.join();
    }
    return doneFlag_;
}

void ScsTransformListener::dynamicTransformsListenerCallback()
{
    while (!doneFlag_.load()) {
        dynamicTransformsUpdate(true); // This is a blocking call
    }
}

void ScsTransformListener::dynamicTransformsUpdate(const bool blocking)
{
    if (smoothedLocalizationPoseInterface_) {
        av::msg::SmoothedLocalizationPoseList smoothedPoses;
        if (smoothedLocalizationPoseInterface_->get(smoothedPoses, blocking)) {
            for (const auto& pose : smoothedPoses.poses) {
                av::msg::TransformStampedList transformMessage = scs::toTransformList(pose);
                updateTransforms(transformMessage);
            }
        }
    } else if (dynamicTransformsInterface_) {
        av::msg::TransformStampedList transformMessage;
        if (dynamicTransformsInterface_->get(transformMessage, blocking)) // This could be a blocking call
        {
            updateTransforms(transformMessage);
        }
    }
}

void ScsTransformListener::updateTransforms(const av::msg::TransformStampedList& transformMessage)
{
    for (const auto& tf : transformMessage.transforms) {
        TransformStorage ts = scs::storage::fromMsg(tf);
        if (ts.isStatic) {
            continue;
        }
        auto result = transformer_.setTransform(ts);
        if (!result.isOK()) {
            AVLOG_WARN("TransformListener failed to set transform: {} <- {}, error code: {}",
                       ts.parentFrame,
                       ts.childFrame,
                       static_cast<int>(result.code));
        }
    }
}

void ScsTransformListener::loadStaticTransform(Transformer& transformer, ConfigSection& taskParameters)
{
    // Get all the transforms from calibration.cfg
    size_t numTransforms{0};
    std::string activeConfigDir;
    taskParameters.get("ActiveConfigDirectory", activeConfigDir, "/tmp/active");
    std::string cfgPath = activeConfigDir + "/calibration.cfg";
    libconfig::Config cfg;

    AVLOG_INFO("Loading StaticTransform from {}", cfgPath);
    try {
        cfg.readFile(cfgPath.c_str());
        const libconfig::Setting& root = cfg.getRoot();
        const libconfig::Setting& transformsCfg = root.lookup("Calibration.Transforms");
        numTransforms = static_cast<size_t>(transformsCfg.getLength());
        transformer.loadStaticTransformsFromConfig(transformsCfg);
    } catch (const libconfig::FileIOException& ex) {
        static_cast<void>(ex);
        AVLOG_ERROR("I/O error while reading file: {}", cfgPath);
    } catch (const libconfig::ParseException& ex) {
        AVLOG_ERROR("Parse error at {} : {} - {}", ex.getFile(), ex.getLine(), ex.getError());
    } catch (const libconfig::SettingNotFoundException& ex) {
        static_cast<void>(ex);
        AVLOG_ERROR("Failed to find 'Calibration.Transforms': {}", cfgPath);
    }

    const size_t numFrames = transformer.getFrameNames().size();
    AVLOG_INFO("Number of transforms in the config file: {}", numTransforms);
    AVLOG_INFO("Number of frames added to TransformListener: {}", numFrames);
    if ((numFrames - 1) != numTransforms) {
        AVLOG_WARN("TransformListener loaded an inconsistent number of transforms: {}", cfgPath);
    }
}
} // namespace ct
} // namespace aptv
