#include <gtest/gtest.h>
#include <thread>

#include <interfaces/TransformStamped/TransformStamped.h>
#include <interfaces/TransformStampedList/TransformStampedList.h>

#include <coordinate-transform/broadcaster.hpp>
#include <coordinate-transform/conversions.hpp>
#include <coordinate-transform/listener.hpp>

#include <TimeStamp/otimeConversions.h>

using namespace std;

av::msg::TransformStamped CreateTransformStamped(double t0,
                                                 double t1,
                                                 double t2,
                                                 double r0,
                                                 double r1,
                                                 double r2,
                                                 double r3,
                                                 av::TimePoint ts = av::TimePoint{},
                                                 string parentFrame = "",
                                                 string childFrame = "",
                                                 bool isStatic = false)
{
    av::msg::TransformStamped out;
    out.timeStamp = ts;

    out.parentFrame = parentFrame;
    out.childFrame = childFrame;

    out.translation[0] = t0;
    out.translation[1] = t1;
    out.translation[2] = t2;

    out.rotation[0] = r0;
    out.rotation[1] = r1;
    out.rotation[2] = r2;
    out.rotation[3] = r3;

    out.isStatic = isStatic;

    return out;
}

TEST(ConversionTest, TransformBullet)
{
    av::msg::TransformStamped msg_tf = CreateTransformStamped(1.2, 3.4, 5.6, sqrt(2) / 2, 0, 0, sqrt(2) / 2);

    btTransform msg_bt;
    msg_bt = aptv::ct::scs::bullet::fromMsg(msg_tf);

    av::msg::TransformStamped msg_tfTest;
    ottotime::otimePoint stampTime;
    msg_tfTest = aptv::ct::scs::bullet::toMsg(msg_bt, stampTime, "parentFrame_tf", "childFrame_tf");

    EXPECT_EQ(msg_tfTest.timeStamp, av::timePointFromOtimePoint(stampTime));
    EXPECT_EQ(msg_tfTest.parentFrame, "parentFrame_tf");
    EXPECT_EQ(msg_tfTest.childFrame, "childFrame_tf");

    EXPECT_EQ(msg_tfTest.translation[0], msg_tf.translation[0]);
    EXPECT_EQ(msg_tfTest.translation[1], msg_tf.translation[1]);
    EXPECT_EQ(msg_tfTest.translation[2], msg_tf.translation[2]);

    EXPECT_DOUBLE_EQ(msg_tfTest.rotation[0], msg_tf.rotation[0]);
    EXPECT_DOUBLE_EQ(msg_tfTest.rotation[1], msg_tf.rotation[1]);
    EXPECT_DOUBLE_EQ(msg_tfTest.rotation[2], msg_tf.rotation[2]);
    EXPECT_DOUBLE_EQ(msg_tfTest.rotation[3], msg_tf.rotation[3]);
}

TEST(ConversionTest, TransformStrorage)
{
    av::msg::TransformStamped msg_tf =
        CreateTransformStamped(1.2,
                               3.4,
                               5.6,
                               0,
                               sqrt(2) / 2,
                               sqrt(2) / 2,
                               0,
                               av::timePointFromOtimePoint(otimePoint::buildCurrentSteadyTime()),
                               "parentFrame_tf",
                               "childFrame_tf",
                               false);

    aptv::ct::TransformStorage msg_ts;
    msg_ts = aptv::ct::scs::storage::fromMsg(msg_tf);

    av::msg::TransformStamped msg_tfTest;
    msg_tfTest = aptv::ct::scs::storage::toMsg(msg_ts);

    double tfTest_ts_ = av::otimePointFromTimePoint(msg_tfTest.timeStamp).toSecondsSinceEpochDouble();
    double tf_ts_ = av::otimePointFromTimePoint(msg_tf.timeStamp).toSecondsSinceEpochDouble();
    double delta_ts_ = (tf_ts_ - tfTest_ts_);
    EXPECT_NEAR(tfTest_ts_, tf_ts_, 1e-06) << " Time difference (s): " << (delta_ts_);
    EXPECT_EQ(msg_tfTest.parentFrame, "parentFrame_tf");
    EXPECT_EQ(msg_tfTest.childFrame, "childFrame_tf");

    EXPECT_EQ(msg_tfTest.translation[0], msg_tf.translation[0]);
    EXPECT_EQ(msg_tfTest.translation[1], msg_tf.translation[1]);
    EXPECT_EQ(msg_tfTest.translation[2], msg_tf.translation[2]);

    EXPECT_EQ(msg_tfTest.rotation[0], msg_tf.rotation[0]);
    EXPECT_EQ(msg_tfTest.rotation[1], msg_tf.rotation[1]);
    EXPECT_EQ(msg_tfTest.rotation[2], msg_tf.rotation[2]);
    EXPECT_EQ(msg_tfTest.rotation[3], msg_tf.rotation[3]);
}
TEST(ConversionTest, toTransformListTest)
{
    av::msg::SmoothedLocalizationPose smoothedPose;
    smoothedPose.timeStamp = av::AvClock::now();
    smoothedPose.poseENU.position.x = 1.1;
    smoothedPose.poseENU.position.y = 2.2;
    smoothedPose.poseENU.position.z = 3.3;
    smoothedPose.poseENU.rotation.x = 0.5709415;
    smoothedPose.poseENU.rotation.y = 0.1675188;
    smoothedPose.poseENU.rotation.z = 0.5709415;
    smoothedPose.poseENU.rotation.w = 0.5656758;
    av::msg::TransformStampedList transformList = aptv::ct::scs::toTransformList(smoothedPose);
    EXPECT_EQ(transformList.transforms.size(), 5);
    for (std::size_t i = 0; i < 3; ++i) {
        EXPECT_TRUE(transformList.transforms[i].timeStamp == smoothedPose.timeStamp);
        EXPECT_NEAR(transformList.transforms[i].translation[0], smoothedPose.poseENU.position.x, 1e-7);
        EXPECT_NEAR(transformList.transforms[i].translation[1], smoothedPose.poseENU.position.y, 1e-7);
        EXPECT_NEAR(transformList.transforms[i].translation[2], smoothedPose.poseENU.position.z, 1e-7);
        EXPECT_NEAR(transformList.transforms[i].rotation[0], smoothedPose.poseENU.rotation.x, 1e-7);
        EXPECT_NEAR(transformList.transforms[i].rotation[1], smoothedPose.poseENU.rotation.y, 1e-7);
        EXPECT_NEAR(transformList.transforms[i].rotation[2], smoothedPose.poseENU.rotation.z, 1e-7);
        EXPECT_NEAR(transformList.transforms[i].rotation[3], smoothedPose.poseENU.rotation.w, 1e-7);
    }
}

TEST(ConversionTest, getPoseInENUTest)
{
    RecPose3D transform;
    transform.x = 1.2;
    transform.y = 5.6;
    transform.z = 3.4;
    transform.rot1 = 0.0;        // x
    transform.rot2 = (M_PI / 2); // y
    transform.rot3 = 0.0;        // z

    RecPose3D transform_result = aptv::ct::scs::getPoseInENU(transform);
    EXPECT_DOUBLE_EQ(transform_result.x, 5.6);
    EXPECT_DOUBLE_EQ(transform_result.y, 1.2);
    EXPECT_DOUBLE_EQ(transform_result.z, -3.4);

    EXPECT_NEAR(double(transform_result.rot1), (M_PI / 2), 1e-10);      // z
    EXPECT_NEAR(double(transform_result.rot2), (-1 * M_PI / 2), 1e-10); // y
    EXPECT_NEAR(double(transform_result.rot3), 0.0, 1e-10);             // x
}

TEST(ConversionTest, getPoseInENUMessageTest)
{
    av::geometry::RecPose3D transformMsg;
    transformMsg.x = 1.2;
    transformMsg.y = 5.6;
    transformMsg.z = 3.4;
    transformMsg.rot1.rad = 0.0;        // x
    transformMsg.rot2.rad = (M_PI / 2); // y
    transformMsg.rot3.rad = 0.0;        // z

    av::geometry::RecPose3D transform_result = aptv::ct::scs::getPoseInENU(transformMsg);
    EXPECT_DOUBLE_EQ(transform_result.x, 5.6);
    EXPECT_DOUBLE_EQ(transform_result.y, 1.2);
    EXPECT_DOUBLE_EQ(transform_result.z, -3.4);

    EXPECT_NEAR(double(transform_result.rot1.rad), (M_PI / 2), 1e-10);      // z
    EXPECT_NEAR(double(transform_result.rot2.rad), (-1 * M_PI / 2), 1e-10); // y
    EXPECT_NEAR(double(transform_result.rot3.rad), 0.0, 1e-10);             // x
}

TEST(ConversionTest, getSTDInENUTest)
{
    RecPose3D transform;
    transform.x = 1.2;
    transform.y = 5.6;
    transform.z = 3.4;
    transform.rot1 = 0.0;
    transform.rot2 = (M_PI / 2);
    transform.rot3 = (-1 * M_PI / 2);

    RecPose3D transform_result = aptv::ct::scs::getSTDInENU(transform);
    EXPECT_DOUBLE_EQ(transform_result.x, 5.6);
    EXPECT_DOUBLE_EQ(transform_result.y, 1.2);
    EXPECT_DOUBLE_EQ(transform_result.z, 3.4);

    EXPECT_NEAR(double(transform_result.rot1), 0.0, 1e-10);
    EXPECT_NEAR(double(transform_result.rot2), (M_PI / 2), 1e-10);
    EXPECT_NEAR(double(transform_result.rot3), (-1 * M_PI / 2), 1e-10);
}

TEST(ConversionTest, getSTDInENUMessageTest)
{
    av::geometry::RecPose3D transformMsg;
    transformMsg.x = 1.2;
    transformMsg.y = 5.6;
    transformMsg.z = 3.4;
    transformMsg.rot1.rad = 0.0;
    transformMsg.rot2.rad = (M_PI / 2);
    transformMsg.rot3.rad = (-1 * M_PI / 2);

    av::geometry::RecPose3D transform_result = aptv::ct::scs::getSTDInENU(transformMsg);
    EXPECT_DOUBLE_EQ(transform_result.x, 5.6);
    EXPECT_DOUBLE_EQ(transform_result.y, 1.2);
    EXPECT_DOUBLE_EQ(transform_result.z, 3.4);

    EXPECT_NEAR(double(transform_result.rot1.rad), 0.0, 1e-10);
    EXPECT_NEAR(double(transform_result.rot2.rad), (M_PI / 2), 1e-10);
    EXPECT_NEAR(double(transform_result.rot3.rad), (-1 * M_PI / 2), 1e-10);
}

TEST(ConversionTest, getAccelerationInENUTest)
{
    RecPoint3D acc;
    acc.x = 1.2;
    acc.y = 3.4;
    acc.z = 5.6;

    RecPoint3D acc_result = aptv::ct::scs::getAccelerationInENU(acc);
    EXPECT_DOUBLE_EQ(acc_result.x, 3.4);
    EXPECT_DOUBLE_EQ(acc_result.y, 1.2);
    EXPECT_DOUBLE_EQ(acc_result.z, -5.6);
}

TEST(ConversionTest, getAccelerationInENUMessageTest)
{
    av::geometry::RecPoint3D accMsg;
    accMsg.x = 1.2;
    accMsg.y = 3.4;
    accMsg.z = 5.6;

    av::geometry::RecPoint3D acc_result = aptv::ct::scs::getAccelerationInENU(accMsg);
    EXPECT_DOUBLE_EQ(acc_result.x, 3.4);
    EXPECT_DOUBLE_EQ(acc_result.y, 1.2);
    EXPECT_DOUBLE_EQ(acc_result.z, -5.6);
}

TEST(ConversionTest, getVelocitySTDInENUTest)
{
    RecDifferentialPose3D velocity;
    velocity.x = 1.2;
    velocity.y = 3.4;
    velocity.z = 5.6;
    velocity.rot1 = 0.0;
    velocity.rot2 = (M_PI / 2);
    velocity.rot3 = 0.0;

    RecDifferentialPose3D velocity_result = aptv::ct::scs::getVelocitySTDInENU(velocity);
    EXPECT_DOUBLE_EQ(velocity_result.x, 3.4);
    EXPECT_DOUBLE_EQ(velocity_result.y, 1.2);
    EXPECT_DOUBLE_EQ(velocity_result.z, 5.6);

    EXPECT_NEAR(double(velocity_result.rot1), 0.0, 1e-10);
    EXPECT_NEAR(double(velocity_result.rot2), (M_PI / 2), 1e-10);
    EXPECT_NEAR(double(velocity_result.rot3), 0.0, 1e-10);
}

TEST(ConversionTest, getVelocitySTDInENUMessageTest)
{
    av::geometry::RecDifferentialPose3D velocityMsg;
    velocityMsg.x = 1.2;
    velocityMsg.y = 3.4;
    velocityMsg.z = 5.6;
    velocityMsg.rot1.rad = 0.0;
    velocityMsg.rot2.rad = (M_PI / 2);
    velocityMsg.rot3.rad = 0.0;

    av::geometry::RecDifferentialPose3D velocity_result = aptv::ct::scs::getVelocitySTDInENU(velocityMsg);
    EXPECT_DOUBLE_EQ(velocity_result.x, 3.4);
    EXPECT_DOUBLE_EQ(velocity_result.y, 1.2);
    EXPECT_DOUBLE_EQ(velocity_result.z, 5.6);

    EXPECT_NEAR(double(velocity_result.rot1.rad), 0.0, 1e-10);
    EXPECT_NEAR(double(velocity_result.rot2.rad), (M_PI / 2), 1e-10);
    EXPECT_NEAR(double(velocity_result.rot3.rad), 0.0, 1e-10);
}

TEST(ConversionTest, fillInTransAndRotTest)
{
    RecPose3D pENU;
    pENU.x = 1.2;
    pENU.y = 3.4;
    pENU.z = 5.6;
    pENU.rot1 = 0.0;        // yaw
    pENU.rot2 = (M_PI / 2); // pitch
    pENU.rot3 = 0.0;        // roll

    av::msg::TransformStamped tfMsg;
    aptv::ct::scs::fillInTransAndRot(pENU, tfMsg);
    EXPECT_DOUBLE_EQ(tfMsg.translation[0], 1.2);
    EXPECT_DOUBLE_EQ(tfMsg.translation[1], 3.4);
    EXPECT_DOUBLE_EQ(tfMsg.translation[2], 5.6);

    // Convert Euler Angles to Quaternion
    EXPECT_DOUBLE_EQ(tfMsg.rotation[0], 0.0);         // X-axis
    EXPECT_DOUBLE_EQ(tfMsg.rotation[1], sqrt(2) / 2); // Y-axis
    EXPECT_DOUBLE_EQ(tfMsg.rotation[2], 0.0);         // Z-axis
    EXPECT_DOUBLE_EQ(tfMsg.rotation[3], sqrt(2) / 2); // W-axis
}

// NOTE:: FAILED!
// From above test, Pith = Pi resulted in Euler Angles (0, sqrt(2)/2, 0, sqrt(2)/2)
// For this test, should Euler Angles (0, sqrt(2)/2, 0, sqrt(2)/2) resulted in Pith = Pi as well?
TEST(ConversionTest, getPoseFromTransformTest)
{
    aptv::ct::TransformStorage tf;
    tf.translation.setX(1.2);
    tf.translation.setY(3.4);
    tf.translation.setZ(5.6);

    tf.rotation.setX(0.0);
    tf.rotation.setY(sqrt(2) / 2);
    tf.rotation.setZ(0.0);
    tf.rotation.setW(sqrt(2) / 2);

    // Convert Quaternion to Euler Angles
    RecPose3D pose_result = aptv::ct::getPoseFromTransform(tf);

    EXPECT_DOUBLE_EQ(pose_result.x, 1.2);
    EXPECT_DOUBLE_EQ(pose_result.y, 3.4);
    EXPECT_DOUBLE_EQ(pose_result.z, 5.6);

    EXPECT_DOUBLE_EQ(pose_result.rot1, M_PI);
    EXPECT_DOUBLE_EQ(pose_result.rot2, (M_PI / 2));
    EXPECT_DOUBLE_EQ(pose_result.rot3, M_PI);
}

// This test should be fail because `dynamicTransformsInterface` is `nullptr`
TEST(TransformListenerTest, ListenerConnectFail)
{
    aptv::ct::Transformer transformer;
    bool threaded = false;
    aptv::ct::ScsTransformListener transformListener(transformer, threaded);

    std::shared_ptr<GenericInput<av::msg::TransformStampedList>> dynamicTransformsInterface{nullptr};
    std::shared_ptr<GenericInput<av::msg::SmoothedLocalizationPoseList>> smoothedLocalizationPoseInterface{nullptr};
    ConfigSection taskParameters;
    EXPECT_FALSE(transformListener.connect(dynamicTransformsInterface, taskParameters));
    EXPECT_FALSE(
        transformListener.connect(dynamicTransformsInterface, taskParameters, smoothedLocalizationPoseInterface));
}

TEST(TransformListenerTest, dynamicTransformsUpdate)
{
    aptv::ct::Transformer transformer;
    bool threaded = false;
    aptv::ct::ScsTransformListener transformListener(transformer, threaded);

    bool blocking(false);
    transformListener.dynamicTransformsUpdate(blocking);
}

TEST(TransformListenerTest, updateTransformsTest)
{
    aptv::ct::Transformer transformer;
    bool threaded = false;
    aptv::ct::ScsTransformListener transformListener(transformer, threaded);

    av::msg::TransformStamped tfStamped =
        CreateTransformStamped(1,
                               2,
                               3,
                               0.1,
                               0.2,
                               0.3,
                               0.4,
                               av::timePointFromOtimePoint(otimePoint::buildCurrentSteadyTime()),
                               "parentFrame_tf",
                               "childFrame_tf",
                               false);

    av::msg::TransformStampedList transformMessage;
    transformMessage.transforms.push_back(tfStamped);
    transformListener.updateTransforms(transformMessage);
}

TEST(TransformBroadcasterTest, ConnectFail)
{
    std::shared_ptr<GenericOutput<av::msg::TransformStampedList>> interface = nullptr;

    aptv::ct::ScsTransformBroadcaster transformBroadcaster;
    EXPECT_FALSE(transformBroadcaster.connect(interface));
}

TEST(TransformBroadcasterTest, sendTransformTestFail)
{
    aptv::ct::ScsTransformBroadcaster transformBroadcaster;

    av::msg::TransformStamped tfStamped =
        CreateTransformStamped(1,
                               2,
                               3,
                               0.1,
                               0.2,
                               0.3,
                               0.4,
                               av::timePointFromOtimePoint(otimePoint::buildCurrentSteadyTime()),
                               "parentFrame_tf",
                               "childFrame_tf",
                               false);

    av::msg::TransformStampedList transformMessage;
    transformMessage.transforms.push_back(tfStamped);
    ASSERT_FALSE(transformBroadcaster.sendTransform(transformMessage));
}
