/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file utility.h
 * @product infrastructure
 * @version
 * @released
 * @updated
 * @brief Data log utility library
 *
 * This header declares data log utility functions and classes
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 ***********************************************************************************************************************
 */

#ifndef _DATALOG_UTILITY_H_
#define _DATALOG_UTILITY_H_

#include <cstdint>
#include <cstring>
#include <sqlite3.h>
#include <string>

namespace {
/*
 * @brief Callback for getting the sqlite3 version
 *
 * sqlite3_exec takes an optional callback and that's the only way to extract data from a call
 *
 * @param version : pointer to a type-erased uint32_t
 * @param argc : Argument count
 * @param argv : Argument variables
 * @param colnames : unused
 */
int getVersionCallback(void* version, int argc, char** argv, char** colnames)
{
    std::string::size_type sz;

    uint32_t* versionPtr = (uint32_t*)version;
    for (auto i = 0; i < argc; ++i) {
        *versionPtr = std::stoi(argv[i], &sz);
    }

    return SQLITE_OK;
}
} // namespace

//
// Ideally the functions here wouldn't be implemented in the header, but the OfflineLogReader/Writer
// uses them and that is a templated class
//
namespace dataLog {
/*
 * @brief The version of the data log structure
 *
 * This is different than the version of the individual datums.  The log version corresponds to the
 * overall structure of the log.  For example, modifying or adding a common header to every datum or
 * adding a new column to the log would necessitate a version bump.
 */
constexpr uint32_t DATA_LOG_VERSION = 3;

/*
 * @brief Named aliases for log versions
 */
constexpr uint32_t PRE_ODATUM_LOG_VERSION = 0;
constexpr uint32_t INITIAL_ODATUM_LOG_VERSION = 1;
constexpr uint32_t NO_TIMESTAMP_INDEX_VERSION = 2;
constexpr uint32_t MSG_COUNTER_AND_PUBLISHER_ID_VERSION = 3;

/*
 * @brief Writes the specified log version to a database
 *
 * This function expects that the db handle is already connected
 *
 * @param db : handle to a sqlite3 database
 *
 * @return sqlite3 error code
 */
inline int writeLogVersion(sqlite3* db, const uint32_t version, char** errMsg)
{
    std::string versionCmd = "PRAGMA user_version=" + std::to_string(version);
    return sqlite3_exec(db, versionCmd.c_str(), nullptr, nullptr, errMsg);
}

/*
 * @brief Writes the default log version to a database
 *
 * This function expects that the db handle is already connected
 *
 * @param db : handle to a sqlite3 database
 *
 * @return sqlite3 error code
 */
inline int writeLogVersion(sqlite3* db, char** errMsg)
{
    return writeLogVersion(db, DATA_LOG_VERSION, errMsg);
}

/*
 * @brief Reads the log version from the database
 *
 * This function expects that the db handle is already connected
 *
 * @param db : handle to a sqlite3 database
 * @param version : variable to read the version into
 *
 * @return sqlite3 error code
 */
inline int readLogVersion(sqlite3* db, uint32_t& version, char** errMsg)
{
    return sqlite3_exec(db, "PRAGMA user_version;", getVersionCallback, &version, errMsg);
}
} // namespace dataLog

#endif
