/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**
 * @file    CanFdPort/PCAN/Impl.cc
 * @author  Junsung Kim
 * @date    10/15/2011
 *
 * @attention Copyright (c) 2011
 * @attention Carnegie Mellon University
 * @attention All rights reserved.
 */

#include <Threading/PThreadLocker.h>
#include <algorithm>
#include <interfaces/CanFdPort/PCAN/Impl.h>

#include "lib/avlog/avlog.hpp"

using namespace std;

CanFdPortPCAN::CanFdPortPCAN(ConfigSection& cs, const task::Access& acc)
    : CanFdPort(cs, acc)
    , _fd(-1)
    , _connected(false)
{
    pthread_mutex_init(&_in_queue_mutex, NULL);
    pthread_cond_init(&_packet_available_cond, NULL);
    pthread_cond_init(&_write_complete_cond, NULL);
}

CanFdPortPCAN::~CanFdPortPCAN()
{
    cleanup();
    pthread_mutex_destroy(&_in_queue_mutex);
    pthread_cond_destroy(&_packet_available_cond);
    pthread_cond_destroy(&_write_complete_cond);
}

bool CanFdPortPCAN::initialize()
{
    interfaceParams_.get("device", _device, "can0");

    return connect();
}

void CanFdPortPCAN::cleanup()
{
    if (isConnected()) {
        disconnect();
    }
}

bool CanFdPortPCAN::connect()
{

    if (_connected)
        return true;

    _fd = socket(PF_CAN, SOCK_RAW, CAN_RAW);

    if (_fd == -1) {
        AVLOG_CRITICAL("Error opening {}: {}", _device, strerror(errno));
        return false;
    }

    _addr.can_family = AF_CAN;

    // note: The character count and the final index happen to be the same number
    std::size_t const toBeCopied = std::min(_device.size(), static_cast<std::size_t>(IFNAMSIZ - 1));
    std::strncpy(_ifr.ifr_name, _device.c_str(), toBeCopied); // Not guaranteed to append a null-terminator
    _ifr.ifr_name[toBeCopied] = '\0';

    // check that the interface supports CAN FD MTUs
    auto caddr_ptr = reinterpret_cast<caddr_t>(&_ifr); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
    if (ioctl(_fd, SIOCGIFMTU, caddr_ptr) < 0) {       // NOLINT(cppcoreguidelines-pro-type-vararg)
        AVLOG_CRITICAL("Unable to read MTUs for {}: {}", _device, strerror(errno));
        return false;
    }
    if (static_cast<size_t>(_ifr.ifr_mtu) != CANFD_MTU) {
        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
        AVLOG_CRITICAL("{} is not set up for CAN FD, MTU is set to {}.", _device, _ifr.ifr_mtu);
        return false;
    } else {
        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-union-access)
        AVLOG_DEBUG("{} MTU is properly set to {} for CAN FD", _device, _ifr.ifr_mtu);
    }

    if (ioctl(_fd, SIOCGIFINDEX, &_ifr) < 0) {
        AVLOG_CRITICAL("Error initializing {}: {}", _device, strerror(errno));
        return false;
    }

    static const int canfd_on = 1;
    if (setsockopt(_fd, SOL_CAN_RAW, CAN_RAW_FD_FRAMES, &canfd_on, sizeof(canfd_on)) < 0) {
        AVLOG_CRITICAL("Error setting CAN FD Frame option for {}: {}", _device, strerror(errno));
        return false;
    }

    _addr.can_ifindex = _ifr.ifr_ifindex;

    auto sockaddr_ptr = reinterpret_cast<sockaddr*>(&_addr); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
    if (bind(_fd, sockaddr_ptr, sizeof(_addr)) < 0) {
        AVLOG_CRITICAL("Error binding {}: {}", _device, strerror(errno));

        return false;
    }

    _connected = true;
    auto rc = pthread_create(&_io_thread, NULL, start_io_thread, this);
    if (rc != 0) {
        AVLOG_CRITICAL("Error create io_thread: {}", strerror(rc));

        return false;
    }

    return true;
}

bool CanFdPortPCAN::disconnect()
{
    if (!_connected)
        return true;

    _connected = false;
    pthread_join(_io_thread, NULL);

    if (close(_fd) == -1)
        return false;

    return true;
}

bool CanFdPortPCAN::isConnected()
{
    return _connected;
}

bool CanFdPortPCAN::read_packet(CanFdPacket& pkt, bool blocking)
{
    PThreadLocker lock(_in_queue_mutex);

    bool packet_available = false;

    if (blocking && _connected) {
        if (_in_queue.empty())
            pthread_cond_wait(&_packet_available_cond, &_in_queue_mutex);
    }

    if (!_in_queue.empty()) {
        packet_available = true;

        pkt = _in_queue.front();
        _in_queue.pop_front();
    }

    return packet_available;
}

bool CanFdPortPCAN::write_packet(CanFdPacket pkt)
{

    ssize_t bytes_written = 0;
    if (_connected) {
        {
            struct canfd_frame MsgTx;

            MsgTx.len = pkt.length;
            MsgTx.can_id = pkt.id;
            MsgTx.flags = pkt.flags;
            memcpy(&MsgTx.data, &pkt.data, pkt.length);

            bytes_written = write(_fd, &MsgTx, sizeof(MsgTx));

            if (bytes_written == -1) {
                AVLOG_WARN("Error writing {}: {}", _device, strerror(errno));

                /*
                TODO: What is a smart way of handling this?
                        if (errno != ETIME)
                        {
                            disconnect();
                        }
                */
            }
        }
    }

    pthread_cond_broadcast(&_write_complete_cond);

    return (static_cast<size_t>(bytes_written) == sizeof(canfd_frame));
}

bool CanFdPortPCAN::write_packet(const std::vector<CanFdPacket>& pktArray)
{
    (void)pktArray;
    return false; // not used for PCAN
}

bool CanFdPortPCAN::set_filter(std::vector<uint32_t> ids)
{

    ids_.clear();
    ids_ = ids;

    std::vector<can_filter> rfilter(ids_.size());

    for (unsigned int i = 0; i < ids_.size(); i++) {
        rfilter[i].can_id = ids_.at(i);
        rfilter[i].can_mask = (CAN_EFF_FLAG | CAN_RTR_FLAG | CAN_SFF_MASK);
    }

    return 0 == setsockopt(_fd,
                           SOL_CAN_RAW,
                           CAN_RAW_FILTER,
                           rfilter.data(),
                           static_cast<socklen_t>(sizeof(can_filter) * rfilter.size()));
}

bool CanFdPortPCAN::add_filter(std::vector<uint32_t> ids)
{

    ids_.insert(ids_.end(), ids.begin(), ids.end());

    std::vector<can_filter> rfilter(ids_.size());

    for (unsigned int i = 0; i < ids_.size(); i++) {
        rfilter[i].can_id = ids_.at(i);
        rfilter[i].can_mask = (CAN_EFF_FLAG | CAN_RTR_FLAG | CAN_SFF_MASK);
    }

    return 0 == setsockopt(_fd,
                           SOL_CAN_RAW,
                           CAN_RAW_FILTER,
                           rfilter.data(),
                           static_cast<socklen_t>(sizeof(can_filter) * rfilter.size()));
}

void* CanFdPortPCAN::io_thread()
{

    while (_connected) {
        fd_set rfds;

        FD_ZERO(&rfds);
        FD_SET(_fd, &rfds);

        struct timeval waitLimit;
        waitLimit.tv_sec = 1;
        waitLimit.tv_usec = 0;

        // XXX Are there any errors from select that I care about?
        // TODO
        const int selectStatus = select(_fd + 1, &rfds, NULL, NULL, &waitLimit);

        if (selectStatus == -1) {
            AVLOG_WARN("Error while selecting on {}: {}", _device, strerror(errno));

            break;
        }

        if (FD_ISSET(_fd, &rfds)) {
            struct canfd_frame MsgRx;
            socklen_t len = sizeof(_addr);
            auto sockaddr_ptr =
                reinterpret_cast<sockaddr*>(&_addr); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
            auto readStatus = recvfrom(_fd, &MsgRx, sizeof(MsgRx), 0, sockaddr_ptr, &len);

            if (readStatus > 0) {
                CanFdPacket new_pkt;

                struct timeval tv;

                if (ioctl(_fd, SIOCGSTAMP, &tv) < 0) {
                    new_pkt.timeStamp.getCurrentWallClockTime();
                    AVLOG_WARN("Failed to read timestamp using ioctl Error: {}", strerror(errno));
                } else {
                    struct timespec ts;
                    ts.tv_sec = tv.tv_sec;
                    ts.tv_nsec = tv.tv_usec * 1000;
                    new_pkt.timeStamp = ottotime::otimePoint(ts);
                }

                new_pkt.id = MsgRx.can_id;
                new_pkt.length = MsgRx.len;
                if (MsgRx.len > 11) {
                    new_pkt.extended = true;
                }
                new_pkt.flags = MsgRx.flags;
                memcpy(&new_pkt.data, &MsgRx.data, new_pkt.length);

                {
                    PThreadLocker lock(_in_queue_mutex);
                    _in_queue.push_back(new_pkt);
                }

                pthread_cond_broadcast(&_packet_available_cond);
            } else if (readStatus == -1) {
                AVLOG_WARN("Error reading {}: {}", _device, strerror(errno));

                /*
                TODO: What is a smart way of handling this?
                        disconnect();
                */
            }
        }

    } // <--- end of the loop

    return NULL;
}

void CanFdPortPCAN::unblock()
{
    pthread_cond_broadcast(&_packet_available_cond);
}

void CanFdPortPCAN::clear()
{
    PThreadLocker lock(_in_queue_mutex);

    while (!_in_queue.empty()) {
        _in_queue.pop_front();
    }
}
