/*!
 **********************************************************
 * @file      interfaces/CanPort/Channel/Impl.cc
 * @author    Junsung Kim
 * @date      06/09/2016
 * @attention using a timestamp marked when the data is
 * @attention sent for the log playback support
 *
 * @attention Copyright (c) 2016- Delphi Automotive.
 * @attention All rights reserved.
 **********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

#include <Clock/SystemClock.h>

#include <interfaces/CanPort/Channel/Impl.h>

using namespace std;

CanPortChannel::CanPortChannel(ConfigSection& cs, const task::Access& acc)
    : CanPort(cs, acc)
    , connected_(false)
{
    headerVersion_ = OTTO_INTERFACE_HEADER_VERSION;
}

CanPortChannel::~CanPortChannel()
{
    cleanup();
}

bool CanPortChannel::initialize()
{
    return connect();
}

void CanPortChannel::cleanup()
{
    if (isConnected()) {
        disconnect();
    }
}

bool CanPortChannel::connect()
{
    connected_ = (task::Interface<CanPacket>::channel_->connect() == task::CS_Success);

    return connected_;
}

bool CanPortChannel::disconnect()
{
    task::Interface<CanPacket>::channel_->disconnect();
    connected_ = false;

    return true;
}

bool CanPortChannel::isConnected()
{
    return connected_;
}

bool CanPortChannel::read_packet(CanPacket& pkt, bool blocking)
{
    bool ret = false;
    interfaces::baseTypes::InterfaceHeader hdr;
    std::size_t hdrSize = interfaces::baseTypes::getInterfaceHeaderSize(headerVersion_);

    auto hdr_ptr = reinterpret_cast<char*>(&hdr); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
    ret = task::Interface<CanPacket>::channel_->read(hdr_ptr, hdrSize, pkt, blocking) == task::CS_NewData;

    uint32_t hdrVer;
    getHeaderVersion(hdr, hdrVer);
    if (ret && (hdrVer != headerVersion_)) {
        headerVersion_ = hdrVer;
        ret = false;
    }

    return ret;
}

bool CanPortChannel::write_packet(CanPacket pkt)
{
    interfaces::baseTypes::InterfaceHeader hdr;

    // TODO: this time may be non-steady in a simulation/resim environment -- is this desired?
    struct otimeSpec sendTime = SystemClock::getSystemTime().to_otimeSpec();

    interfaces::baseTypes::setHeaderVersion(hdr, OTTO_INTERFACE_HEADER_VERSION);
    interfaces::baseTypes::setSendTimeStamp(hdr, sendTime);
    std::size_t hdrSize = interfaces::baseTypes::getInterfaceHeaderSize(headerVersion_);
    auto hdr_ptr = reinterpret_cast<char*>(&hdr); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
    return task::Interface<CanPacket>::channel_->write(hdr_ptr, hdrSize, pkt) == task::CS_Success;
}

bool CanPortChannel::write_packet(const std::vector<CanPacket>& pktArray)
{
    interfaces::baseTypes::InterfaceHeader hdr;
    struct otimeSpec sendTime = SystemClock::getSystemTime().to_otimeSpec();
    interfaces::baseTypes::setHeaderVersion(hdr, OTTO_INTERFACE_HEADER_VERSION);
    interfaces::baseTypes::setSendTimeStamp(hdr, sendTime);
    std::size_t hdrSize = interfaces::baseTypes::getInterfaceHeaderSize(headerVersion_);
    bool result{true};
    auto hdr_ptr = reinterpret_cast<char*>(&hdr); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
    for (const auto& pkt : pktArray) {
        result &= task::Interface<CanPacket>::channel_->write(hdr_ptr, hdrSize, pkt) == task::CS_Success;
    }
    return result;
}

bool CanPortChannel::set_filter(std::vector<uint32_t> /*ids*/)
{
    return false;
}

bool CanPortChannel::add_filter(std::vector<uint32_t> /*ids*/)
{
    return false;
}

void CanPortChannel::unblock()
{
    if (task::Interface<CanPacket>::channel_) {
        task::Interface<CanPacket>::channel_->unblock();
    }
}

void CanPortChannel::clear()
{
    if (task::Interface<CanPacket>::channel_) {
        task::Interface<CanPacket>::channel_->clear();
    }
}
