/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**
 * @file    CanPort/PCAN/Impl.cc
 * @author  Junsung Kim
 * @date    10/15/2011
 *
 * @attention Copyright (c) 2011
 * @attention Carnegie Mellon University
 * @attention All rights reserved.
 */

#include <Threading/PThreadLocker.h>
#include <interfaces/CanPort/PCAN/Impl.h>

#include "lib/avlog/avlog.hpp"

using namespace std;

CanPortPCAN::CanPortPCAN(ConfigSection& cs, const task::Access& acc)
    : CanPort(cs, acc)
    , _fd(-1)
    , _connected(false)
{
    pthread_mutex_init(&_in_queue_mutex, NULL);
    pthread_cond_init(&_packet_available_cond, NULL);
    pthread_cond_init(&_write_complete_cond, NULL);
}

CanPortPCAN::~CanPortPCAN()
{
    cleanup();
    pthread_mutex_destroy(&_in_queue_mutex);
    pthread_cond_destroy(&_packet_available_cond);
    pthread_cond_destroy(&_write_complete_cond);
}

bool CanPortPCAN::initialize()
{
    interfaceParams_.get("device", _device, "can0");

    return connect();
}

void CanPortPCAN::cleanup()
{
    if (isConnected()) {
        disconnect();
    }
}

bool CanPortPCAN::connect()
{

    if (_connected)
        return true;

    _fd = socket(PF_CAN, SOCK_RAW, CAN_RAW);

    if (_fd == -1) {
        AVLOG_CRITICAL("Error opening {}: {}", _device, strerror(errno));

        return false;
    }

    _addr.can_family = AF_CAN;

    // note: The character count and the final index happen to be the same number
    std::size_t const toBeCopied = std::min(_device.size(), static_cast<std::size_t>(IFNAMSIZ - 1));
    std::strncpy(_ifr.ifr_name, _device.c_str(), toBeCopied); // Not guaranteed to append a null-terminator
    _ifr.ifr_name[toBeCopied] = '\0';

    if (ioctl(_fd, SIOCGIFINDEX, &_ifr) < 0) {
        AVLOG_CRITICAL("Error initializing {}: {}", _device, strerror(errno));

        return false;
    }

    if (setsockopt(_fd, SOL_CAN_RAW, CAN_RAW_FILTER, NULL, 0) < 0) {
        AVLOG_CRITICAL("Error disabling CAN channel {}: {}", _device, strerror(errno));

        return false;
    }

    _addr.can_ifindex = _ifr.ifr_ifindex;
    auto sockaddr_ptr = reinterpret_cast<sockaddr*>(&_addr); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
    if (bind(_fd, sockaddr_ptr, sizeof(_addr)) < 0) {
        AVLOG_CRITICAL("Error binding {}: {}", _device, strerror(errno));

        return false;
    }

    _connected = true;
    auto rc = pthread_create(&_io_thread, NULL, start_io_thread, this);
    if (rc != 0) {
        AVLOG_CRITICAL("Error create io_thread: {}", strerror(rc));

        return false;
    }

    return true;
}

bool CanPortPCAN::disconnect()
{
    if (!_connected)
        return true;

    _connected = false;
    pthread_join(_io_thread, NULL);

    if (close(_fd) == -1)
        return false;

    return true;
}

bool CanPortPCAN::isConnected()
{
    return _connected;
}

bool CanPortPCAN::read_packet(CanPacket& pkt, bool blocking)
{
    PThreadLocker lock(_in_queue_mutex);

    bool packet_available = false;

    if (blocking && _connected) {
        if (_in_queue.empty())
            pthread_cond_wait(&_packet_available_cond, &_in_queue_mutex);
    }

    if (!_in_queue.empty()) {
        packet_available = true;

        pkt = _in_queue.front();
        _in_queue.pop_front();
    }

    return packet_available;
}

bool CanPortPCAN::write_packet(CanPacket pkt)
{

    ssize_t bytes_written = 0;
    if (_connected) {
        {
            struct can_frame MsgTx;

            // pkt.extended not used
            MsgTx.can_dlc = pkt.length;
            MsgTx.can_id = pkt.id;
            memcpy(&MsgTx.data, &pkt.data, 8);

            bytes_written = write(_fd, &MsgTx, sizeof(MsgTx));

            if (bytes_written == -1) {
                AVLOG_WARN("Error writing {}: {}", _device, strerror(errno));

                /*
                TODO: What is a smart way of handling this?
                        if (errno != ETIME)
                        {
                            disconnect();
                        }
                */
            }
        }
    }

    pthread_cond_broadcast(&_write_complete_cond);

    return (static_cast<size_t>(bytes_written) == sizeof(can_frame));
}

bool CanPortPCAN::write_packet(const std::vector<CanPacket>& pktArray)
{
    (void)pktArray;
    return false; // not used for PCAN
}

bool CanPortPCAN::set_filter(std::vector<uint32_t> ids)
{
    ids_.clear();
    ids_ = ids;

    std::vector<can_filter> rfilter(ids_.size());

    for (unsigned int i = 0; i < ids_.size(); i++) {
        rfilter[i].can_id = ids_.at(i);
        rfilter[i].can_mask = (CAN_EFF_FLAG | CAN_RTR_FLAG | CAN_SFF_MASK);
    }

    return 0 == setsockopt(_fd,
                           SOL_CAN_RAW,
                           CAN_RAW_FILTER,
                           rfilter.data(),
                           static_cast<socklen_t>(sizeof(can_filter) * rfilter.size()));
}

bool CanPortPCAN::add_filter(std::vector<uint32_t> ids)
{
    ids_.insert(ids_.end(), ids.begin(), ids.end());

    std::vector<can_filter> rfilter(ids_.size());

    for (unsigned int i = 0; i < ids_.size(); i++) {
        rfilter[i].can_id = ids_.at(i);
        rfilter[i].can_mask = (CAN_EFF_FLAG | CAN_RTR_FLAG | CAN_SFF_MASK);
    }

    return 0 == setsockopt(_fd,
                           SOL_CAN_RAW,
                           CAN_RAW_FILTER,
                           rfilter.data(),
                           static_cast<socklen_t>(sizeof(can_filter) * rfilter.size()));
}

void* CanPortPCAN::io_thread()
{

    while (_connected) {
        fd_set rfds;

        FD_ZERO(&rfds);
        FD_SET(_fd, &rfds);

        struct timeval waitLimit;
        waitLimit.tv_sec = 1;
        waitLimit.tv_usec = 0;

        // XXX Are there any errors from select that I care about?
        // TODO
        const int selectStatus = select(_fd + 1, &rfds, NULL, NULL, &waitLimit);

        if (selectStatus == -1) {
            AVLOG_WARN("Error while selecting on {}: {}", _device, strerror(errno));

            break;
        }

        if (FD_ISSET(_fd, &rfds)) {
            struct can_frame MsgRx;
            socklen_t len = sizeof(_addr);
            auto sockaddr_ptr =
                reinterpret_cast<sockaddr*>(&_addr); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
            auto readStatus = recvfrom(_fd, &MsgRx, sizeof(MsgRx), 0, sockaddr_ptr, &len);

            if (readStatus > 0) {
                CanPacket new_pkt;

                struct timeval tv;

                if (ioctl(_fd, SIOCGSTAMP, &tv) < 0) {
                    AVLOG_WARN("Failed to read timestamp using ioctl Error: {}", strerror(errno));
                    new_pkt.timeStamp.getCurrentWallClockTime();
                } else {
                    struct timespec ts;
                    ts.tv_sec = tv.tv_sec;
                    ts.tv_nsec = tv.tv_usec * 1000;
                    new_pkt.timeStamp = ottotime::otimePoint(ts);
                }

                new_pkt.id = MsgRx.can_id;
                new_pkt.extended = 0;
                new_pkt.length = MsgRx.can_dlc;
                memcpy(&new_pkt.data, &MsgRx.data, 8);

                {
                    PThreadLocker lock(_in_queue_mutex);
                    _in_queue.push_back(new_pkt);
                }

                pthread_cond_broadcast(&_packet_available_cond);
            } else if (readStatus == -1) {
                AVLOG_WARN("Error reading {}: {}", _device, strerror(errno));

                /*
                TODO: What is a smart way of handling this?
                        disconnect();
                */
            }
        }

    } // <--- end of the loop

    return NULL;
}

void CanPortPCAN::unblock()
{
    pthread_cond_broadcast(&_packet_available_cond);
}

void CanPortPCAN::clear()
{
    PThreadLocker lock(_in_queue_mutex);

    while (!_in_queue.empty()) {
        _in_queue.pop_front();
    }
}
