/**
 * \copyright
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief definition for CANPortUDP class. Supports reading CAN messages from UDP port
 */

#include <Threading/PThreadLocker.h>
#include <interfaces/CanPort/UDP/Impl.h>

#include "lib/avlog/avlog.hpp"

using namespace ceg_udp_data;

CanPortUDP::CanPortUDP(ConfigSection& cs, const task::Access& acc)
    : CanPort(cs, acc)
{
    pthread_mutex_init(&_in_queue_mutex, nullptr);
    pthread_cond_init(&_packet_available_cond, nullptr);
}

CanPortUDP::~CanPortUDP()
{
    cleanup();
    pthread_mutex_destroy(&_in_queue_mutex);
    pthread_cond_destroy(&_packet_available_cond);
}

bool CanPortUDP::initialize()
{
    // rx
    uint16_t port_in{0};
    interfaceParams_.get("port_in", port_in, 50110);
    memset(&_addr_in, 0, sizeof(_addr_in));
    _addr_in.sin_port = htons(port_in);
    _addr_in.sin_addr.s_addr = htonl(INADDR_ANY); // local socket
    _addr_in.sin_family = AF_INET;

    // multicast group
    std::string ip_in_str;
    std::string multicast_addr_str;
    interfaceParams_.get("multicast_addr", multicast_addr_str, "239.255.76.67");
    interfaceParams_.get("ip_in", ip_in_str, "192.168.17.255");

    struct in_addr ip_in
    {};
    if (ip_in_str == "INADDR_ANY") {
        ip_in.s_addr = htonl(INADDR_ANY);
    } else {
        if (inet_pton(AF_INET, ip_in_str.c_str(), &ip_in) <= 0) {
            AVLOG_CRITICAL("Error in setting parameter ip_in for the CanPortUDP interface");
        }
    }

    struct in_addr multicast_addr_in
    {};
    if (inet_pton(AF_INET, multicast_addr_str.c_str(), &multicast_addr_in) <= 0) {
        AVLOG_CRITICAL("Error in setting multicast address for the CanPortUDP interface");
    }

    _multicast_group.imr_interface = ip_in;
    _multicast_group.imr_multiaddr = multicast_addr_in;

    // tx
    uint16_t port_out;

    interfaceParams_.get("port_out", port_out, 50111);
    memset(&_addr_out, 0, sizeof(_addr_out));
    _addr_out.sin_port = htons(port_out);

    std::string ip_out;
    interfaceParams_.get("ip_out", ip_out, "192.168.17.44");
    if (ip_out == "INADDR_ANY") {
        _addr_out.sin_addr.s_addr = htonl(INADDR_ANY);
    } else {
        if (inet_pton(AF_INET, ip_out.c_str(), &(_addr_out.sin_addr)) <= 0) {
            AVLOG_CRITICAL("Error in setting parameter ip_out for the CanPortUDP interface");
        }
    }

    char str[INET_ADDRSTRLEN];
    inet_ntop(AF_INET, &(_multicast_group.imr_interface), str, INET_ADDRSTRLEN);
    AVLOG_INFO("IP/Port Input: {}/{}", str, ntohs(_addr_in.sin_port));
    inet_ntop(AF_INET, &(_multicast_group.imr_multiaddr), str, INET_ADDRSTRLEN);
    AVLOG_INFO("Multicast address: {}", str);
    inet_ntop(AF_INET, &(_addr_out.sin_addr), str, INET_ADDRSTRLEN);
    AVLOG_INFO("IP/Port Output: {}/{}", str, ntohs(_addr_out.sin_port));

    // CAN bus channel
    interfaceParams_.get("channel_number", _channel_num, 0);

    return connect();
}

void CanPortUDP::cleanup()
{
    if (isConnected()) {
        disconnect();
    }
}

bool CanPortUDP::connect()
{

    if (_connected)
        return true;

    // inbound socket
    {
        _fd_in = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if (_fd_in == -1) {
            AVLOG_CRITICAL("Error opening socket at port {}: {}", _addr_in.sin_port, strerror(errno));

            return false;
        }

        /* application to receive copies of the multicast datagrams. */
        {
            int reuse = 1;

            if (setsockopt(_fd_in, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) < 0) {
                AVLOG_CRITICAL("Setting SO_REUSEADDR error");
                close(_fd_in);
                return false;
            } else {
                AVLOG_INFO("Setting SO_REUSEADDR...OK.\n");
            }
        }

        auto sockaddr_ptr =
            reinterpret_cast<sockaddr*>(&_addr_in); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
        if (bind(_fd_in, sockaddr_ptr, sizeof(_addr_in)) < 0) {
            AVLOG_CRITICAL("Error binding to socket at port {}: {}", _addr_in.sin_port, strerror(errno));
            close(_fd_in);
            return false;
        }

        if (setsockopt(_fd_in, IPPROTO_IP, IP_ADD_MEMBERSHIP, &_multicast_group, sizeof(_multicast_group)) < 0) {
            AVLOG_CRITICAL("Adding multicast group error");
            close(_fd_in);
            return false;
        } else {
            AVLOG_INFO("Adding multicast group...OK.\n");
        }
    }

    // outbound socket
    {
        _fd_out = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

        if (_fd_out == -1) {
            AVLOG_CRITICAL("Error opening socket at port {}: {}", _addr_out.sin_port, strerror(errno));

            return false;
        }

        _addr_out.sin_family = AF_INET;
    }

    _connected = true;
    // io_thread for inbound socket should be start after set _connected
    auto rc = pthread_create(&_io_thread, nullptr, start_io_thread, this);
    if (rc != 0) {
        AVLOG_CRITICAL("Error create io_thread: {}", strerror(rc));

        return false;
    }

    return true;
}

bool CanPortUDP::disconnect()
{
    if (!_connected)
        return true;

    _connected = false;
    bool ret{true};
    if (pthread_join(_io_thread, nullptr) != 0) {
        AVLOG_INFO("Error joining io_thread");
        ret = false;
    }

    if (close(_fd_in) == -1) {
        AVLOG_INFO("Error closing inbound socket");
        ret = false;
    }

    if (close(_fd_out) == -1) {
        AVLOG_INFO("Error closing outbound socket");
        ret = false;
    }

    return ret;
}

bool CanPortUDP::isConnected()
{
    return _connected;
}

bool CanPortUDP::read_packet(CanPacket& pkt, bool blocking)
{
    PThreadLocker lock(_in_queue_mutex);

    bool packet_available = false;

    if (blocking && _connected) {
        if (_in_queue.empty())
            pthread_cond_wait(&_packet_available_cond, &_in_queue_mutex);
    }

    if (!_in_queue.empty()) {
        packet_available = true;

        pkt = _in_queue.front();
        _in_queue.pop_front();
    }

    return packet_available;
}

bool CanPortUDP::write_packet(CanPacket pkt)
{
    ssize_t bytes_written = 0;
    aptiv_ceg_udp_send_frame MsgTx;

    if (_connected) {
        // Fill the header as per CEG requirements doc
        MsgTx.ceg_header.header = K_CEG_HEADER_BYTE;
        MsgTx.ceg_header.length = static_cast<uint16_t>(sizeof(aptiv_ceg_udp_send_frame));
        MsgTx.ceg_header.sequenceCount = _txSequenceCounter;

        // fill in the CAN frame data, wrap the CAN according CEG format
        aptiv_ceg_udp_data temp_ceg_data;
        temp_ceg_data.msgType = K_UDP_CAN_MSG_TYPE;
        temp_ceg_data.canDlc = CAN_DATA_LEN;
        temp_ceg_data.canID = pkt.id;
        temp_ceg_data.channelNumber = _channel_num;
        std::copy_n(pkt.data, temp_ceg_data.canDlc, temp_ceg_data.canData);
        MsgTx.ceg_data[0] = temp_ceg_data;

        // send using UDP
        socklen_t len = sizeof(_addr_out);
        auto sockaddr_ptr =
            reinterpret_cast<sockaddr*>(&_addr_out); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
        bytes_written = sendto(_fd_out, &MsgTx, sizeof(MsgTx), 0, sockaddr_ptr, len);
        _txSequenceCounter = (_txSequenceCounter + 1) % std::numeric_limits<uint32_t>::max();
    }

    return (static_cast<size_t>(bytes_written) == sizeof(MsgTx));
}

bool CanPortUDP::write_packet(const std::vector<CanPacket>& pktArray)
{
    if (pktArray.size() > NUM_SEND_CAN_MSG) {
        AVLOG_CRITICAL("Max number of packets to be send on CanUDP = {}", NUM_SEND_CAN_MSG);
        return false;
    }

    if (_connected) {
        aptiv_ceg_udp_send_frame MsgTx{};

        // Fill the header as per CEG requirements doc
        MsgTx.ceg_header.header = K_CEG_HEADER_BYTE;
        MsgTx.ceg_header.length = static_cast<uint16_t>(sizeof(aptiv_ceg_udp_send_frame));
        MsgTx.ceg_header.sequenceCount = _txSequenceCounter;

        auto dataSize = sizeof(MsgTx.ceg_header);

        // fill in the CAN frame data, wrap the CAN according CEG format
        uint8_t count = 0;
        for (const CanPacket& pkt : pktArray) {
            auto& ceg_data = MsgTx.ceg_data[count]; // count will not increase beyond NUM_SEND_CAN_MSG
            ceg_data.msgType = K_UDP_CAN_MSG_TYPE;
            ceg_data.canDlc = CAN_DATA_LEN;
            ceg_data.canID = pkt.id;
            ceg_data.channelNumber = _channel_num;
            std::copy_n(pkt.data, ceg_data.canDlc, ceg_data.canData);

            dataSize += sizeof(ceg_data);
            count++;
        }

        // send using UDP
        socklen_t len = sizeof(_addr_out);
        auto sockaddr_ptr =
            reinterpret_cast<sockaddr*>(&_addr_out); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
        auto bytes_written = sendto(_fd_out, &MsgTx, dataSize, 0, sockaddr_ptr, len);
        _txSequenceCounter = (_txSequenceCounter + 1) % std::numeric_limits<uint32_t>::max();

        return (static_cast<size_t>(bytes_written) == dataSize);
    }

    return false;
}

bool CanPortUDP::set_filter(std::vector<uint32_t> ids)
{
    _canIDFilter.clear();
    for (const auto id : ids) {
        _canIDFilter.insert(id);
    }

    return true;
}

bool CanPortUDP::add_filter(std::vector<uint32_t> ids)
{
    for (const auto id : ids) {
        _canIDFilter.insert(id);
    }

    return true;
}

void* CanPortUDP::io_thread()
{

    while (_connected) {
        fd_set rfds;

        FD_ZERO(&rfds);
        FD_SET(_fd_in, &rfds);

        struct timeval waitLimit;
        waitLimit.tv_sec = 1;
        waitLimit.tv_usec = 0;

        const int selectStatus = select(_fd_in + 1, &rfds, nullptr, nullptr, &waitLimit);

        if (selectStatus == -1) {
            AVLOG_WARN("Error while selecting on socket at port {}: {}", _addr_in.sin_port, strerror(errno));

            break;
        }
        if (selectStatus == 0) {
            AVLOG_WARN("select() function timeout");
            continue;
        }

        if (FD_ISSET(_fd_in, &rfds)) {
            aptiv_ceg_udp_recv_frame MsgRx;
            socklen_t len = sizeof(_addr_in);

            auto sockaddr_ptr =
                reinterpret_cast<sockaddr*>(&_addr_in); // NOLINT(cppcoreguidelines-pro-type-reinterpret-cast)
            auto readStatus = recvfrom(_fd_in, &MsgRx, sizeof(MsgRx), 0, sockaddr_ptr, &len);

            if ((readStatus > 0) && (static_cast<size_t>(readStatus) == sizeof(MsgRx))) {
                // conform data header/format checks
                if (MsgRx.ceg_header.header != K_CEG_HEADER_BYTE || MsgRx.ceg_header.length != CEG_TOTAL_LEN) {
                    AVLOG_ERROR("CEG header bytes do not match to the specification!");
                    continue; // skip this packet
                }

                // check the sequence counters for continuity
                // prevSequenceCount should be always equal to 1 less than current sequenceCount
                // TODO: what do to when we receive out of sequence packets?
                if (!_isRxFirstTime && (_rxPrevSequenceCount != MsgRx.ceg_header.sequenceCount - 1)) {
                    AVLOG_WARN("Out of sequence packet received.. Exepecting {} but receieved {}",
                               static_cast<unsigned long>(_rxPrevSequenceCount + 1),
                               static_cast<unsigned long>(MsgRx.ceg_header.sequenceCount));
                }

                _isRxFirstTime = false; // we are already got the data, so _isRxFirstTime will be false

                // once conforms all the above checks, go through array of data fill the queue
                for (const aptiv_ceg_udp_data& msg : MsgRx.ceg_data) {
                    // we skip the message if it is not in our filter or it is from the expected channel number
                    if ((_canIDFilter.find(msg.canID) == _canIDFilter.end()) || (_channel_num != msg.channelNumber)) {
                        continue;
                    }

                    CanPacket new_pkt;

                    struct timeval tv;
                    struct timespec ts;
                    if (ioctl(_fd_in, SIOCGSTAMP, &tv) < 0) {
                        new_pkt.timeStamp.getLocalTime();
                        AVLOG_WARN("Failed to read timestamp using ioctl Error: {}", strerror(errno));
                    } else {
                        ts.tv_sec = tv.tv_sec;
                        ts.tv_nsec = tv.tv_usec * 1000;
                        new_pkt.timeStamp = ottotime::otimePoint(ts);
                    }

                    new_pkt.id = msg.canID;
                    new_pkt.extended = 0;
                    new_pkt.length = msg.canDlc;
                    std::copy_n(msg.canData, CAN_DATA_LEN, new_pkt.data);

                    PThreadLocker lock(_in_queue_mutex);
                    _in_queue.push_back(new_pkt);
                }
                // save the sequenceCount for comparing
                _rxPrevSequenceCount = MsgRx.ceg_header.sequenceCount;
                // TODO: this condition should be inside for loop or outside?
                pthread_cond_broadcast(&_packet_available_cond);
            } else if (readStatus == -1) {
                AVLOG_ERROR("Error reading socket at port {}: {}", _addr_out.sin_port, strerror(errno));
            }
        }

    } // <--- end of the loop

    return nullptr;
}

void CanPortUDP::unblock()
{
    pthread_cond_broadcast(&_packet_available_cond);
}

void CanPortUDP::clear()
{
    PThreadLocker lock(_in_queue_mutex);

    while (!_in_queue.empty()) {
        _in_queue.pop_front();
    }
}
