/**
 * \copyright
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief CanPortUDP for receiving CAN message from UDP port
 */

#ifndef _CANPORT_UDP_IMPL_H_
#define _CANPORT_UDP_IMPL_H_

#include <RingBuffer/RingBuffer.hpp>
#include <arpa/inet.h>
#include <atomic>
#include <interfaces/CanPort/Abstract.h>
#include <linux/sockios.h>
#include <net/if.h>
#include <netinet/in.h>
#include <pthread.h>
#include <queue>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <unistd.h>
#include <unordered_set>
#include <vector>

#include "lib/avlog/avlog.hpp"

/**
 * @ref Following udp structures follow the design doc -
 * https://conf.ptc.delphiauto.net/confluence/display/SYS/CEG+UDP+Data+Structure
 */

namespace ceg_udp_data {
// data structure size
constexpr uint16_t NUM_RECV_CAN_MSG = 64;
constexpr uint16_t NUM_SEND_CAN_MSG = 12;
constexpr uint16_t CAN_DATA_LEN = 8;
constexpr uint16_t CEG_CAN_MSG_HEADER_LEN = 8; // MsgType + Len + CHL
constexpr uint16_t CEG_CAN_MESSAGE_LEN = CEG_CAN_MSG_HEADER_LEN + CAN_DATA_LEN;
constexpr uint16_t CEG_HEADER_LEN = 24;
constexpr uint16_t CEG_TOTAL_LEN = (CEG_HEADER_LEN + CEG_CAN_MESSAGE_LEN * NUM_RECV_CAN_MSG);

// magic number definition
constexpr uint16_t K_CEG_HEADER_BYTE = 0xFF24;
constexpr uint16_t K_UDP_CAN_MSG_TYPE = 0xFF80;

struct aptiv_ceg_udp_header
{
    uint16_t header{0};
    uint16_t length{0};
    uint32_t sequenceCount{0};
    uint64_t ntp{0};
    uint64_t customArea{0};
};

struct aptiv_ceg_udp_data
{
    uint16_t msgType{0};
    uint8_t canDlc{0};
    uint8_t channelNumber{0};
    uint32_t canID{0};
    uint8_t canData[CAN_DATA_LEN]{};
};

struct aptiv_ceg_udp_recv_frame
{
    aptiv_ceg_udp_header ceg_header{};
    aptiv_ceg_udp_data ceg_data[NUM_RECV_CAN_MSG]{};
};

struct aptiv_ceg_udp_send_frame
{
    aptiv_ceg_udp_header ceg_header{};
    aptiv_ceg_udp_data ceg_data[NUM_SEND_CAN_MSG]{};
};
} // namespace ceg_udp_data

class CanPortUDP : public CanPort
{
public:
    CanPortUDP(ConfigSection& cs, const task::Access& acc);
    virtual ~CanPortUDP();

    virtual bool initialize();
    virtual void cleanup();

    virtual bool connect();
    virtual bool disconnect();

    virtual bool isConnected();

    virtual bool read_packet(CanPacket& pkt, bool blocking = false) override;
    virtual bool write_packet(CanPacket pkt) override;
    virtual bool write_packet(const std::vector<CanPacket>& pktArray) override;
    virtual bool set_filter(std::vector<uint32_t> ids) override;
    virtual bool add_filter(std::vector<uint32_t> ids) override;
    virtual void unblock() override;
    virtual void clear() override;

protected:
    struct sockaddr_in _addr_in
    {};
    struct ip_mreq _multicast_group
    {};
    struct sockaddr_in _addr_out
    {};
    int _fd_in{-1};
    int _fd_out{-1};
    uint8_t _channel_num{0};
    static constexpr size_t inboundMessageQueueCapacity{1200};
    av::FCRingBuffer<CanPacket, inboundMessageQueueCapacity> _in_queue;
    pthread_t _io_thread;
    pthread_mutex_t _in_queue_mutex;
    pthread_cond_t _packet_available_cond;
    std::atomic<bool> _connected{false};
    virtual void* io_thread();
    static void* start_io_thread(void* arg) { return static_cast<CanPortUDP*>(arg)->io_thread(); }
    uint32_t _txSequenceCounter{0};   ///< sequence count for TX UDP packets
    bool _isRxFirstTime{true};        ///< true if we have never received any UDP packets
    uint32_t _rxPrevSequenceCount{0}; ///< sequence count for previous RX UDP packets
    std::unordered_set<uint32_t> _canIDFilter;
};

#endif
