/*!
 **********************************************************
 * @file      Clock/Playback/PlaybackClockParams.h
 * @author    Junsung Kim
 * @date      02/18/2016
 * @attention The clock format is otime instead of boost
 *
 * @attention Copyright (c) 2016- Delphi Automotive.
 * @attention All rights reserved.
 **********************************************************
 */

/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**
 * @file    Clock/Playback/PlaybackClockParams.h
 * @author  Tugrul Galatali
 * @date    11/06/2006
 *
 * @attention Copyright (c) 2006
 * @attention Carnegie Mellon University
 * @attention All rights reserved.
 */
#ifndef _CLOCK_PLAYBACK_PLAYBACKCLOCKPARAMS_H_
#define _CLOCK_PLAYBACK_PLAYBACKCLOCKPARAMS_H_

#include <TimeStamp/otime.h>
#include <algorithm>
#include <chrono>

#include <ODatum/ODatum.h>

using namespace ottotime;

/**
 * @brief Parameters for transforming one time base into another
 */
struct PlaybackClockParams : public aptv::ODatum
{
    PlaybackClockParams()
    {
        // The other fields have proper constructors.
        scaleFactor = 1.0;
    }
    PlaybackClockParams(const PlaybackClockParams& other) = default;
    PlaybackClockParams& operator=(const PlaybackClockParams& other) = default;

    //! desiredClock = givenClock + offset
    otimeDuration offset{};
    //! scaledClock = (desiredClock - baseTime) * scaleFactor + baseTime
    otimePoint baseTime{};
    double scaleFactor{1.0};
    //! stoppedClock = min(scaledClock, stopTime)
    otimePoint stopTime{};

    // The scale factor shouldn't be negative and shouldn't be *extremely* small or large.
    static constexpr double minScaleFactor = 0.0;
    static constexpr double minPositiveScaleFactor = 0.0001;
    static constexpr double maxScaleFactor = 1000.0;

    otimePoint apply(const otimePoint& given)
    {
        otimePoint toReturn{};

        if (baseTimeValid()) {
            // desriedClock is the time in the reference frame (could be log start
            // time + 10 sec for instance)
            const otimePoint desiredClock = given + offset;

            // baseRelative is the number of seconds since the start of the file
            const otimeDuration baseRelative = desiredClock - baseTime;
            const otimePoint scaledClock = baseTime + (baseRelative * static_cast<int>(scaleFactor * 1000)) / 1000;

            // stoppedClock is set to either the scaledClock or the latest time
            // allowed (whichever is smaller)
            const otimePoint stoppedClock = stopTime.isTimeValid() ? std::min(scaledClock, stopTime) : scaledClock;

            toReturn = stoppedClock;
        }

        return toReturn;
    }

    bool baseTimeValid() const { return baseTime.isTimeValid(); }

    bool scaleFactorValid() const
    {
        bool retVal{true};

        if (scaleFactor < 0.0) {
            retVal = false;
        } else if (scaleFactor > 0.0) {
            if ((scaleFactor < minPositiveScaleFactor) || (scaleFactor > maxScaleFactor)) {
                retVal = false;
            }
        }

        return retVal;
    }

    bool isValid() const { return baseTimeValid() && scaleFactorValid(); }

    void normalize()
    {
        // Make sure the scale factor isn't negative and that it isn't extremely large or an extremely small positive
        // number.
        if (scaleFactor < 0.0) {
            scaleFactor = 0.0;
        }

        if (scaleFactor > 0.0) {
            if (scaleFactor < minPositiveScaleFactor) {
                scaleFactor = minPositiveScaleFactor;
            } else if (scaleFactor > maxScaleFactor) {
                scaleFactor = maxScaleFactor;
            }
        }
    }

    friend bool operator==(const PlaybackClockParams& a, const PlaybackClockParams& b)
    {
        return (a.offset == b.offset) && (a.baseTime == b.baseTime) && (a.scaleFactor == b.scaleFactor) &&
               (a.stopTime == b.stopTime);
    }

    friend bool operator!=(const PlaybackClockParams& a, const PlaybackClockParams& b) { return !(a == b); }

    template<class Archive>
    void serialize(Archive& ar, [[maybe_unused]] const unsigned int version)
    {
        ar& CEREAL_NVP(offset);
        ar& CEREAL_NVP(baseTime);
        ar& CEREAL_NVP(scaleFactor);
        ar& CEREAL_NVP(stopTime);
    }
};

CEREAL_CLASS_VERSION(PlaybackClockParams, 1)
CEREAL_REGISTER_TYPE(PlaybackClockParams)

#endif
