/*!
 **********************************************************
 * @file      Generic/Output/Channel/Impl.def.h
 * @author    Junsung Kim
 * @date      06/09/2016
 * @attention publish() is updated to use a timestamp marked
 * @attention when the data is sent for the log playback
 *
 * @attention Copyright (c) 2016- Delphi Automotive.
 * @attention All rights reserved.
 **********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**
 * @file    Generic/Output/Channel/Impl.def.h
 * @author  Tugrul Galatali
 * @date    09/13/2006
 *
 * @author  Junsung Kim
 * @date    12/16/2012
 * @attention A function chmod() is added to support fault-tolerance.
 *
 * @attention Copyright (c) 2006
 * @attention Carnegie Mellon University
 * @attention All rights reserved.
 */
#ifndef _GENERIC_OUTPUT_CHANNEL_IMPL_H_
#error Must be included from Generic/Output/Channel/Impl.h
#endif

#include <Clock/SystemClock.h>
#include <utility/hash.h>

#include "lib/avlog/avlog.hpp"

#include <sys/types.h>
#include <unistd.h>

template<typename T>
GenericOutputChannel<T>::GenericOutputChannel(ConfigSection& cs, const task::Access& acc)
    : GenericOutput<T>(cs, acc)
{
    std::string channelName = "unused";
    ConfigSection channelConfig;
    if (cs.get("channel", channelConfig) == ConfigSection::GS_OK) {
        channelConfig.get("name", channelName);
    }

    publisherId_ = generatePublisherId(channelName);
    AVLOG_INFO("PublisherId {} created for {}", publisherId_, channelName);
}

template<typename T>
GenericOutputChannel<T>::~GenericOutputChannel()
{
    GenericOutputChannel<T>::cleanup();
}

template<typename T>
uint32_t GenericOutputChannel<T>::generatePublisherId(const std::string& channelName)
{
    std::size_t seed{0};
    utility::hash::HashCombine(seed, channelName);

    char buf[256];
    gethostname(buf, sizeof(buf));
    utility::hash::HashCombine(seed, std::string(buf));

    pid_t pid = getpid();
    utility::hash::HashCombine(seed, pid);

    otimePoint now;
    now.getLocalTime();
    utility::hash::HashCombine(seed, now.to_time_t());

    uint32_t ret = (sizeof(std::size_t) >= sizeof(uint64_t)) ? utility::hash::Hash32From64(static_cast<uint64_t>(seed))
                                                             : static_cast<uint32_t>(seed);

    return ret;
}

template<typename T>
bool GenericOutputChannel<T>::publishV(const T& val, std::vector<av::latency::Pred>&& predecessors)
{
    interfaces::baseTypes::InterfaceHeader hdr;
    struct otimeSpec sendTime = ottotime::SystemClock::getSystemTime().to_otimeSpec();
    interfaces::baseTypes::setHeaderVersion(hdr, OTTO_INTERFACE_HEADER_VERSION);
    interfaces::baseTypes::setSendTimeStamp(hdr, sendTime);
    interfaces::baseTypes::setHeaderPublisherId(hdr, publisherId_);
    interfaces::baseTypes::setHeaderMessageCounter(hdr, ++messageCounter_);
    std::size_t hdrSize = interfaces::baseTypes::getInterfaceHeaderSize(OTTO_INTERFACE_HEADER_VERSION);

    auto* ch = task::Interface<T>::channel_;
    if (ch == nullptr) {
        return false;
    }

    return ch->writeV(reinterpret_cast<char*>(&hdr), hdrSize, val, std::move(predecessors)) == task::CS_Success;
}

template<typename T>
bool GenericOutputChannel<T>::initialize()
{
    if (!task::Interface<T>::channel_) {
        return false;
    }

    // TODO: for now, return true, regardless of whether we connected to
    // IPC, need reconnect logic instead, at some point...
    return (task::Interface<T>::channel_->connect() == task::CS_Success);
}

template<typename T>
void GenericOutputChannel<T>::cleanup()
{
    if (task::Interface<T>::channel_) {
        task::Interface<T>::channel_->disconnect();
    }
}

template<typename T>
void GenericOutputChannel<T>::chmod(const task::Access& acc)
{
    if (task::Interface<T>::channel_) {
        task::Interface<T>::channel_->chmod(acc);
    }
}
