/*!
 **********************************************************
 * @file      Base/Header.h
 * @author    Junsung Kim
 * @date      05/27/2016
 * @attention header info for Interface
 *
 * @attention Copyright (c) 2016- Delphi Automotive.
 * @attention All rights reserved.
 **********************************************************
 */

#ifndef _INTERFACES_BASETYPES_HEADER_H_
#define _INTERFACES_BASETYPES_HEADER_H_
#include <TimeStamp/TimeStamp.h>
#include <arpa/inet.h>

// In the future we expect to have additional header versions.
// For the time being, uses of this code appear to do nothing
// but exist to ensure it will be easy to add new header
// versions later
const uint32_t OTTO_INTERFACE_HEADER_VERSION_0 = 0;
const uint32_t OTTO_INTERFACE_HEADER_VERSION_1 = 1;
const uint32_t OTTO_INTERFACE_HEADER_VERSION = OTTO_INTERFACE_HEADER_VERSION_1;

static inline uint64_t htonll(uint64_t x)
{
    if (1 == htonl(1)) {
        return x;
    }
    const uint64_t high = static_cast<uint64_t>(htonl(x & 0xFFFFFFFFULL)) << 32;
    const uint64_t low = htonl(static_cast<uint32_t>(x >> 32));
    return high | low;
}

static inline uint64_t ntohll(uint64_t x)
{
    if (1 == ntohl(1)) {
        return x;
    }
    const uint64_t high = static_cast<uint64_t>(ntohl(x & 0xFFFFFFFFULL)) << 32;
    const uint64_t low = ntohl(static_cast<uint32_t>(x >> 32));
    return high | low;
}

static inline int32_t htonls(int32_t i)
{
    return static_cast<int32_t>(htonl(static_cast<uint32_t>(i)));
}

static inline int32_t ntohls(int32_t i)
{
    return static_cast<int32_t>(ntohl(static_cast<uint32_t>(i)));
}

static inline int64_t htonlls(int64_t i)
{
    return static_cast<int64_t>(htonll(static_cast<uint64_t>(i)));
}

static inline int64_t ntohlls(int64_t i)
{
    return static_cast<int64_t>(ntohll(static_cast<uint64_t>(i)));
}

namespace interfaces {
namespace baseTypes {

// In InterfaceHeader, we will preserve
// the sequence of headerVersion and sendTimeStamp
// even after new fields are introduced
#pragma pack(push, 8)
// the header is 8-byte aligned as we use 64-bit machines
// make sure that the total size of the struct is multiple
// of 8 bytes
struct InterfaceHeader
{
    otimeSpec sendTimeStamp{};  // 128 bit: time stamp when header is sent
    uint32_t headerVersion{0};  //  32 bit: header version information
    uint32_t reserved{0};       //  32 bit: reserved for cross-platform support
    uint32_t publisherId{0};    //  32 bit: publisher identification
    uint32_t messageCounter{0}; //  32 bit: the number of messsages are being sent
};
#pragma pack(pop)

inline std::size_t getInterfaceHeaderSize(uint32_t headerVersion)
{
    // If adding a new header version, ensure that changes to this code
    // do not cause old versions of the header to return the incorrect size
    if (headerVersion == OTTO_INTERFACE_HEADER_VERSION_1) {
        // Currently returns 32, this will need to return 32 as a constant
        // instead of measuring the struct if a new header version is added
        return sizeof(InterfaceHeader);
    }
    if (headerVersion == OTTO_INTERFACE_HEADER_VERSION_0) {
        return 24;
    }
    return 0;
}

inline void setSendTimeStamp(struct InterfaceHeader& hdr, struct otimeSpec time)
{
// disable sign conversion warnings, till we move to C++20 and get std::bit_cast
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wsign-conversion"
    hdr.sendTimeStamp.tv_sec = htonll(time.tv_sec);
    hdr.sendTimeStamp.tv_nsec = htonll(time.tv_nsec);
#pragma GCC diagnostic pop
}

inline void getSendTimeStamp(struct InterfaceHeader hdr, struct otimeSpec& time)
{
// disable sign conversion warnings, till we move to C++20 and get std::bit_cast
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wsign-conversion"
    time.tv_sec = ntohll(hdr.sendTimeStamp.tv_sec);
    time.tv_nsec = ntohll(hdr.sendTimeStamp.tv_nsec);
#pragma GCC diagnostic pop
}

inline void setHeaderVersion(struct InterfaceHeader& hdr, uint32_t version)
{
    hdr.headerVersion = htonl(version);
}

inline void getHeaderVersion(struct InterfaceHeader hdr, uint32_t& version)
{
    version = ntohl(hdr.headerVersion);
}

inline void setHeaderPublisherId(InterfaceHeader& hdr, const uint32_t id)
{
    hdr.publisherId = htonl(id);
}

inline void getHeaderPublisherId(const InterfaceHeader& hdr, uint32_t& id)
{
    id = ntohl(hdr.publisherId);
}

inline void setHeaderMessageCounter(InterfaceHeader& hdr, const uint32_t value)
{
    hdr.messageCounter = htonl(value);
}

inline void getHeaderMessageCounter(const InterfaceHeader& hdr, uint32_t& counter)
{
    counter = ntohl(hdr.messageCounter);
}

} // namespace baseTypes
} // namespace interfaces

#endif
