/*!
 ***********************************************************
 * @file LogEvent.cc
 *
 * @brief file implementing LogEvent class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 ***********************************************************
 * LogEvent.cc
 * Bryan Salesky
 * 02/28/2006
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 ***********************************************************
 */

#include <chrono>
#include <iomanip>
#include <log/LogEvent.h>
#include <sstream>

using namespace std;
using namespace logger;
using namespace chrono;

static const string TIMEFORMAT("%F %T"); // YYYY-MM-DD HH:MM:SS
const char* const DELIM = " | ";

LogEvent::LogEvent(const Priority::PriorityLevel& pri, const std::string& message)
    : pri_(pri)
    , message_(message)
{}

LogEvent::LogEvent(const char* file, int line, const Priority::PriorityLevel& pri, const std::string& message)
    : file_(file)
    , line_(line)
    , pri_(pri)
    , message_(message)
{}

string LogEvent::formatTime(const LogEvent::ChronoTimePoint& timePoint)
{
    ostringstream dateTime;

    tm convertedTime;

    // Generate date/time, excluding milliseconds.
    const time_t time = system_clock::to_time_t(timePoint);
    dateTime << put_time(gmtime_r(&time, &convertedTime), TIMEFORMAT.c_str());

    // Append millisecond precision.
    const auto milliseconds =
        time_point_cast<chrono::milliseconds>(timePoint) - time_point_cast<chrono::seconds>(timePoint);
    dateTime << "." << setw(3) << setfill('0') << milliseconds.count();

    return dateTime.str();
}

LogEvent::ChronoTimePoint LogEvent::parseTime(const string& dateTime)
{
    ChronoTimePoint timePoint{};
    tm timeComponents = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    // Parse date/time, excluding milliseconds
    const char* endOfDateTime = strptime(dateTime.c_str(), TIMEFORMAT.c_str(), &timeComponents);

    if (endOfDateTime != nullptr) {
        timePoint = system_clock::from_time_t(mktime(&timeComponents));

        // Parse milliseconds
        if (endOfDateTime[0] == '.') {
            char* end = nullptr;
            const long ms = strtol(endOfDateTime + 1, &end, 10);
            timePoint += milliseconds(ms);
        }
    }
    return timePoint;
}

void LogEvent::toStream(ostream& out) const
{
    out << formatTime(timestamp_) << DELIM << setw(6) << Priority::priorityLevelAsString(pri_) << DELIM << setw(15)
        << file_ << ":" << line_ << DELIM << message_ << endl;
}

std::string LogEvent::toString() const
{
    std::string tempString;
    intoString(tempString, false);
    return tempString;
}

void LogEvent::intoString(std::string& buffer, bool append /* default: false */) const
{
    std::ostringstream outStream{std::ios_base::app | std::ios_base::ate};
    if (append) {
        outStream.str(buffer);
    }
    toStream(outStream);
    buffer = outStream.str();
}
