/*!
 ***********************************************************
 * @file Logger.cc
 *
 * @brief file implementing Logger class
 *
 *  A logger that sends LogEvents to zero or more writers.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author    Junsung Kim
 * @date      07/26/2016
 * @remark stop using the Priority class to improve
 *  the text logging performance
 * @n
 *
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 ***********************************************************
 * Logger.cc
 * Bryan Salesky
 * 02/28/2006
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 ***********************************************************
 */

#include "va_str.h"
#include <iostream>
#include <log/AVLogWriter.h>
#include <log/Logger.h>

using namespace std;
using namespace logger;

Priority::PriorityLevel Logger::globalPriorityThreshold_(Priority::debug);

/* static */ void Logger::setGlobalPriorityThreshold(Priority::PriorityLevel threshold)
{
    globalPriorityThreshold_ = threshold;
    avlog::setGlobalLogLevel(AVLogWriter::translateLogPriorityLevel(threshold));
}

// ++
// Generate the log methods for each priority.
// --
#define LOG_PRIORITY_IMPL(priority)                                                                                    \
    void Logger::logxx_##priority(const char* file, int linenum, const char* stringFormat, ...)                        \
    {                                                                                                                  \
        if (Priority::priority <= pri_) {                                                                              \
            va_list va;                                                                                                \
            va_start(va, stringFormat);                                                                                \
            sendToWriters(LogEvent(file, linenum, Priority::priority, va_str(stringFormat, va)));                      \
            va_end(va);                                                                                                \
        }                                                                                                              \
    }                                                                                                                  \
    void Logger::logxx_##priority(const char* file, int linenum, const std::string& message)                           \
    {                                                                                                                  \
        if (Priority::priority <= pri_) {                                                                              \
            sendToWriters(LogEvent(file, linenum, Priority::priority, message));                                       \
        }                                                                                                              \
    }

LOG_PRIORITY_IMPL(emerg)
LOG_PRIORITY_IMPL(fatal)
LOG_PRIORITY_IMPL(alert)
LOG_PRIORITY_IMPL(crit)
LOG_PRIORITY_IMPL(error)
LOG_PRIORITY_IMPL(warn)
LOG_PRIORITY_IMPL(notice)
LOG_PRIORITY_IMPL(info)
LOG_PRIORITY_IMPL(debug)

Logger::Logger()
    : pri_(globalPriorityThreshold_)
    , writers_()
{
    // Default to forwarding logs to avlog.
    addWriter(std::make_shared<AVLogWriter>(globalPriorityThreshold_));
}

Logger::Logger(LogWriterPtr writer)
    : pri_(writer->getThreshold())
    , writers_()
{
    addWriter(writer);
}

Logger::Logger(const Logger& copy)
    : pri_(copy.pri_)
    , writers_(copy.writers_)
{}

Logger& Logger::operator=(const Logger& copy)
{
    if (this != &copy) {
        pri_ = copy.pri_;
        writers_ = copy.writers_;
    }
    return *this;
}

Logger::~Logger() {}

bool Logger::addWriter(LogWriterPtr writer)
{
    return writers_.insert(writer).second;
}

bool Logger::removeWriter(LogWriterPtr writer)
{
    bool found(false);
    LogWriterSetIter iter = writers_.find(writer);
    if (iter != writers_.end()) {
        found = true;
        writers_.erase(iter);
    }
    return found;
}

void Logger::sendToWriters(const LogEvent& event)
{
    LogWriterPtr ww;
    for (LogWriterSetIter iter = writers_.begin(); iter != writers_.end(); ++iter) {
        ww = *iter;
        ww->write(event);
    }
}

void Logger::setThreshold(Priority::PriorityLevel threshold)
{
    pri_ = threshold;
    LogWriterPtr ww;
    for (LogWriterSetIter iter = writers_.begin(); iter != writers_.end(); ++iter) {
        ww = *iter;
        ww->setThreshold(threshold);
    }
}
