/*!
 ***********************************************************
 * @file LogEvent.h
 *
 * @brief file defining LogEvent class
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 ***********************************************************
 * LogEvent.h
 * Bryan Salesky
 * 02/28/2006
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 ***********************************************************
 */
#ifndef _LOG_EVENT_H_
#define _LOG_EVENT_H_

#include "LogEvent.h"
#include "Priority.h"
#include <chrono>
#include <iostream>
#include <string>
#include <sys/time.h>

namespace logger {

/*!
 *****************************************************************************
 * @class LogEvent
 * @brief This class defines a discrete log event that can be sent to any particular
 * LogWriter to be logged.
 *****************************************************************************
 */
class LogEvent
{
public:
    using ChronoTimePoint = std::chrono::system_clock::time_point;

    /**
     * @brief The default constructor
     */
    LogEvent() = default;

    /**
     * @brief Construct this LogEvent with the specified parameters.
     *
     * @param pri     The priority of this event.
     *
     * @param message The message for this event.
     */
    LogEvent(const Priority::PriorityLevel& pri, const std::string& message);

    /*!
     * @brief Construct this LogEvent with the specified parameters.
     *
     * @param file    The source file for this event.
     *
     * @param line    The line within the source file for this event.
     *
     * @param pri     The priority of this event.
     *
     * @param message The message for this event.
     */
    LogEvent(const char* file, int line, const Priority::PriorityLevel& pri, const std::string& message);

    /**
     * @brief The copy constructor.
     *
     * @param copy The object to copy.
     */
    LogEvent(const LogEvent&) = default;

    /**
     * @brief The assignment operator
     *
     * @param copy The object to copy
     *
     * @return Side-Effect.
     */
    LogEvent& operator=(const LogEvent&) = default;

    /**
     * @brief The virtual destructor
     */
    virtual ~LogEvent() = default;

    /*!
     * @brief Write this event to the provided output stream.
     *
     * @param out The output stream to write the event to.
     */
    virtual void toStream(std::ostream& out) const;

    /*!
     * @brief Write the event into a new std::string.
     *
     * @return a std::string containing the event's formatted output
     */
    virtual std::string toString() const;

    /*!
     * @brief Write the event into the provided std::string.
     *
     * @param buffer The std::string to write into.
     *
     * @param append if true, appends to the end of buffer.  Otherwise, overwrites its contents.
     */
    virtual void intoString(std::string& buffer, bool append = false) const;

    /**
     * @brief Get the priority level of the event.(Inline function)
     *
     * @return  The priority.
     */
    Priority::PriorityLevel getPriorityLevel() const { return pri_; }

    /**
     * @brief Get the name of the file where the event was logged.
     *
     * @return  The file name.
     */
    const char* getFile() const { return file_; }

    /**
     * @brief Get the line number where the event was logged.
     *
     * @return  The line number.
     */
    const int& getLine() const { return line_; }

    /**
     * @brief Get the message string.
     *
     * @return  The message string (not including the time, level, file, etc.).
     */
    const std::string& message() const { return message_; }

    /**
     * @brief Formats the specified time point to "YYYY-MM-DD HH:MM:SS.mmm"
     * @param timePoint
     * @return formatted date/time.
     */
    static std::string formatTime(const ChronoTimePoint& timePoint);

    /**
     * @brief Parses a logged timestamp with the format "YYYY-MM-DD HH:MM:SS.mmm"
     * into a \ref ChronoTimePoint
     * @param dateTime
     * @return On success, a \ref ChronoTimePoint for the specified time.  Otherwise,
     * a default constructed \ref ChronoTimePoint.
     * @note Times of the format "YYYY-MM-DD HH:MM:SS" shall also be successfully parsed.
     */
    static ChronoTimePoint parseTime(const std::string& dateTime);

private:
    // This is a raw pointer to the actual __LINE__ value, so it will be valid
    // for the whole lifetime of the program.
    const char* file_{""}; //! A source file that the log event was generated from.

    int line_{-1}; //! The line in the source file that the log event was generated from.

    Priority::PriorityLevel pri_{Priority::none}; //! The priority of the event.

    std::string message_; //! The message to be logged.

    ChronoTimePoint timestamp_{std::chrono::system_clock::now()}; //! The time at which the event was generated.
};

} // namespace logger

#endif
