/*!
 ***********************************************************
 * @file LogWriter.h
 *
 * @brief file defining LogWriter class
 *
 * A base class LogWriter.
 *
 * This class is inherited by all derived LogWriter objects.
 * It is an abstract class that required derived LogWriter's
 * to implement the write_() method.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 ***********************************************************
 * LogWriter.h
 * Bryan Salesky
 * 02/28/2006
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 ***********************************************************
 */
#ifndef _LOG_WRITER_H_
#define _LOG_WRITER_H_

#include "LogEvent.h"
#include <set>

namespace logger {

/*!
 * @class LogWriter
 * @brief Base class inherited by all the LogWriter objects.
 */
class LogWriter
{
public:
    /*!
     * @brief The default constructor for a LogWriter.
     *
     * The threshold is set to Debug by default.
     */
    LogWriter();

    /*!
     * @brief Construct a LogWriter with the supplied threshold.
     *
     * @param threshold The threshold to set the LogWriter to.
     */
    LogWriter(Priority::PriorityLevel threshold);

    /*!
     * \brief The virtual destructor
     */
    virtual ~LogWriter();

    /**
     * @brief Write a LogEvent if its priority is at or above the threshold.
     *
     * @param event The LogEvent to write.
     */
    virtual void write(const LogEvent& event);

    /*!
     * @brief Retrieve the threshold for this LogWriter. (Inline function)
     *
     * @return The threshold.
     */
    Priority::PriorityLevel getThreshold(void) const;

    /*!
     * @brief Set the threshold for this LogWriter. (Inline function)
     *
     * @param level The threshold to set.
     */
    void setThreshold(Priority::PriorityLevel level);

protected:
    /**
     * @brief virtual function which will be implemented in inherited class to write the log event to stream
     *
     * @param event LogEvent to write
     */
    virtual void write_(const LogEvent& event) = 0;

private:
    Priority::PriorityLevel threshold_; ///< threshold associated with the LogWriter, below which it ignores the event

    /**
     * @brief Copy constructor
     */
    LogWriter(const LogWriter&);

    /**
     * @brief The assignment operator
     */
    void operator=(const LogWriter&);
};

inline Priority::PriorityLevel LogWriter::getThreshold(void) const
{
    return threshold_;
}

inline void LogWriter::setThreshold(const Priority::PriorityLevel level)
{
    threshold_ = level;
}

} // namespace logger

#endif
