/*!
 ***********************************************************
 * @file Logger.h
 *
 * @brief file defining Logger class
 *
 * A logger that sends LogEvents to zero or more writers.
 *
 * A Logger is the fundamental object used by a user of this library to
 * log events of a particular priority to zero or more LogWriter objects.
 *
 * If a Logger is constructed using the default constructor, an OStreamWriter
 * is added by default.  Writers can be added or removed at will, anything
 * from zero to N writers are supported.  Of course, the more writers, the
 * more inefficient the logger becomes.
 *
 * A Logger contains a Priority object that can be modified by the user.
 * A Logger constructed using the default constructor, will have all
 * priorities enabled by default.  Priorities can be enabled and disabled
 * individually to suit the user's needs.  Note that each LogWriter has a
 * threshold at which it filters out events.  So if an event is not appearing
 * as it should, you must consult two places: a) be sure the threshold isn't
 * set too high for the particular LogWriter such that the event is ignored,
 * and b) make sure the priority for the event is enabled in the Logger's
 * Priority object.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author    Junsung Kim
 * @date      07/26/2016
 * @remark stop using the Priority class to improve
 *  the text logging performance
 * @n
 *
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 ***********************************************************
 * Logger.h
 * Bryan Salesky
 * 02/28/2006
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 ***********************************************************
 */
#ifndef _LOGGER_H_
#define _LOGGER_H_

#include "NamedVars.h"
#include "OStreamWriter.h"
#include "Priority.h"
#include <memory>
#include <sstream>
#include <string>
#include <type_traits>

#pragma message                                                                                                        \
    "The avlog library is now available.  Logger instances default to sending events there.  Please switch over to gain proper file/line/function reporting and the ability to use structured logging.  See: https://confluence.ci.motional.com/confluence/display/SI/Converting+to+avlog"

namespace logger {

// ++
// Generate 3 methods to log a message for each priority,
// first takes a variable number of char* arguments much like printf(...),
// second simply logs a string.
// third takes a variable number of arbitrary arguments (operator<< must be defined for all of them.
//                                                       at least one arg must be a class to distinguish from first
//                                                       overload)
// --
#define LOG_PRIORITY_DECL(priority)                                                                                    \
    void logxx_##priority(const char* file, int linenum, const char* stringFormat, ...);                               \
    void logxx_##priority(const char* file, int linenum, const std::string& message);                                  \
    template<typename... TArg>                                                                                         \
    auto logxx_##priority(const char* file, int linenum, const TArg&... arg)                                           \
        ->decltype(enable_if_at_least_one_is_class<TArg...>(), EXPAND_PACK(std::cout << arg), void())                  \
    {                                                                                                                  \
        if (Priority::priority > pri_)                                                                                 \
            return;                                                                                                    \
        std::ostringstream oss;                                                                                        \
        EXPAND_PACK(oss << arg);                                                                                       \
        logxx_##priority(file, linenum, oss.str());                                                                    \
    }

namespace pre_cxx17 {
// this template is defined starting c++17, may be removed here once available in <type_traits>
template<class...>
struct disjunction : std::false_type
{};
template<class B1>
struct disjunction<B1> : B1
{};
template<class B1, class... Bn>
struct disjunction<B1, Bn...> : std::conditional<bool(B1::value), B1, disjunction<Bn...>>::type
{};
} // namespace pre_cxx17

// shortcut for combination
template<typename... T>
using enable_if_at_least_one_is_class =
    typename std::enable_if<pre_cxx17::disjunction<std::is_class<T>...>::value>::type;

// TODO: can be removed once we have SHORT__FILE__ everywhere.
#ifndef SHORT__FILE__
#define SHORT__FILE__ __FILE__
#endif

// ++
// The various log methods for each priority
// --
#define log_emerg(args...) logxx_emerg(SHORT__FILE__, __LINE__, args)
#define log_fatal(args...) logxx_fatal(SHORT__FILE__, __LINE__, args)
#define log_alert(args...) logxx_alert(SHORT__FILE__, __LINE__, args)
#define log_crit(args...) logxx_crit(SHORT__FILE__, __LINE__, args)
#define log_error(args...) logxx_error(SHORT__FILE__, __LINE__, args)
#define log_warn(args...) logxx_warn(SHORT__FILE__, __LINE__, args)
#define log_notice(args...) logxx_notice(SHORT__FILE__, __LINE__, args)
#define log_info(args...) logxx_info(SHORT__FILE__, __LINE__, args)
#define log_debug(args...) logxx_debug(SHORT__FILE__, __LINE__, args)

// ++
// A LogWriterPtr is a std::shared_ptr object.  This allows the user to
// dynamically allocate a LogWriter which is shared across multiple
// LogWriter apps, without the burden of having to manage the memory
// allocation of the LogWriter.  Once all the Loggers that use a
// particular LogWriter are deleted, the LogWriter itself will be deleted.
// --

typedef std::shared_ptr<LogWriter> LogWriterPtr;

/*!
 * @class Logger
 * @brief A logger that sends LogEvents to zero or more writers.
 *
 */
class Logger
{
public:
    /**
     * @brief The default constructor.
     *
     * By default, the Logger comes with the global OStreamWriter, and is
     * configured with all priorities enabled.
     */
    Logger();

    /**
     * @brief Construct a Logger with the provided LogWriter.
     *
     * @param writer The LogWriter to add, wrapped into a LogWriterPtr.
     */
    Logger(LogWriterPtr writer);

    /**
     * @brief The virtual destructor
     */
    virtual ~Logger();

    /**
     * @brief Copy constructor.
     *
     * @param copy The logger to copy.
     **/
    Logger(const Logger& copy);

    /**
     * @brief Assignment operator.
     *
     * @param copy The logger to copy.
     **/
    Logger& operator=(const Logger& copy);

    /* Check macros defined in the beginning of this file */
    LOG_PRIORITY_DECL(emerg)
    LOG_PRIORITY_DECL(fatal)
    LOG_PRIORITY_DECL(alert)
    LOG_PRIORITY_DECL(crit)
    LOG_PRIORITY_DECL(error)
    LOG_PRIORITY_DECL(warn)
    LOG_PRIORITY_DECL(notice)
    LOG_PRIORITY_DECL(info)
    LOG_PRIORITY_DECL(debug)

    /*!
     * @brief Add a LogWriter to the Logger.
     *
     * @param writer The LogWriter to add.
     *
     * @return True if the LogWriter was added.
     */
    bool addWriter(LogWriterPtr writer);

    /*!
     * @brief Remove the specified LogWriter.
     *
     * This method will remove the LogWriter from this Logger's list of writers.
     * Note that this method does not delete the writer object itself, it just
     * removes the LogWriter pointer from this Logger's list of writers.
     *
     * @param writer The writer to be removed.
     *
     * @return True if the writer was removed, False otherwise.
     */
    bool removeWriter(LogWriterPtr writer);

    /*!
     * \brief Inline method returning PriorityLevel or Threshold of the logger.
     *
     * \return PriorityLevel of the logger.
     */
    Priority::PriorityLevel getThreshold() { return pri_; }

    /*!
     * @brief Set the threshold on all LogWriters that are part of this Logger.
     *
     * @param event The threshold to set.
     */
    void setThreshold(Priority::PriorityLevel level);

    /*!
     * \brief Inline method for setting global priorty for all the LogWriters
     *  associated with this logger. This is also default threshold level in the beginning.
     *
     * \param threshold global threshold level
     */
    static void setGlobalPriorityThreshold(Priority::PriorityLevel threshold);

protected:
    /*!
     * @brief Send a particular event to all writers that are part of this Logger.
     *
     * @param event The event to log.
     */
    void sendToWriters(const LogEvent& event);

private:
    static Priority::PriorityLevel
        globalPriorityThreshold_; ///< Default Threshold level for logger, below this level it ignores the event

    Priority::PriorityLevel pri_; ///< Threshold level associated with the logger, below this level it ignores the event

    typedef std::set<LogWriterPtr> LogWriterSet;            ///< Defining Set of LogWriter objects
    typedef LogWriterSet::iterator LogWriterSetIter;        ///< LogWriter object iterator
    typedef LogWriterSet::const_iterator LogWriterSetCIter; ///< LogWriter object constant iterator

    LogWriterSet writers_; ///< Set of LogWriter objects
};

} // namespace logger

#endif
