/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file NamedVars.h
 * @product <Subsystem name>
 * @version <Release Number>
 * @released 12/17/2018
 * @updated <date>
 * @brief NAMED_VARS macro to print variables along with their names
 *
 * This is a detailed description of the file that spans several lines and characterizes the basic functionality
 * of the class.
 *
 * @sa Optional Reference
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */
#ifndef NAMEDVARS_H
#define NAMEDVARS_H

#include <iostream>
#include <memory>
#include <sstream>
#include <tuple>
#include <utility>

///////////////////////////////////////////////////////////////////////////////
// some type_traits helpers
namespace aptv {
namespace type_traits {
// helper to determine container type - anything with begin/end defined (both stl and c-arrays)
template<typename>
auto has_begin_end(...) -> std::false_type;
template<typename T>
auto has_begin_end(int) -> decltype(std::begin(std::declval<T&>()), std::end(std::declval<T&>()), std::true_type());
template<typename T>
struct is_container : decltype(has_begin_end<T>(0))
{};

// helper to determine string-like type
template<typename T>
using is_string = std::integral_constant<
    bool,
    std::is_same<std::remove_cv_t<T>, std::string>::value || std::is_same<std::decay_t<T>, char const*>::value ||
        std::is_same<std::decay_t<T>, char*>::value || std::is_same<std::decay_t<T>, wchar_t const*>::value ||
        std::is_same<std::decay_t<T>, wchar_t*>::value>;
} // namespace type_traits
} // namespace aptv

///////////////////////////////////////////////////////////////////////////////
// forward declarations
// pretty-print a pair
template<typename T1, typename T2>
std::ostream& operator<<(std::ostream& os, const std::pair<T1, T2>& p);
// pretty-print a tuple
template<class... TArgs>
std::ostream& operator<<(std::ostream& os, const std::tuple<TArgs...>& t);
// pretty-print a container (both stl and c-arrays, except strings)
template<typename TCont,
         typename std::enable_if<aptv::type_traits::is_container<TCont>::value &&
                                 !aptv::type_traits::is_string<TCont>::value>::type* = nullptr>
std::ostream& operator<<(std::ostream& os, const TCont& cont);

///////////////////////////////////////////////////////////////////////////////
// some type_traits helpers
namespace aptv {
namespace type_traits {
// we have to do this so that ADL work properly from inside a namespace
using ::operator<<;

// helpers to determine if  `ostream << val` is defined for the type
template<typename T>
auto has_ostream_defined_f(...) -> std::false_type;
template<typename T>
auto has_ostream_defined_f(int) -> decltype(std::cout << std::declval<T>(), std::true_type());
template<typename T>
struct has_ostream_defined : decltype(has_ostream_defined_f<T>(0))
{};

// helpers to determine pointer-like type
template<typename>
struct is_shared_ptr : std::false_type
{};
template<typename T>
struct is_shared_ptr<std::shared_ptr<T>> : std::true_type
{};

template<typename>
struct is_unique_ptr : std::false_type
{};
template<typename T>
struct is_unique_ptr<std::unique_ptr<T>> : std::true_type
{};

template<typename T>
struct is_any_pointer
    : std::integral_constant<bool,
                             (std::is_pointer<T>::value || is_shared_ptr<T>::value || is_unique_ptr<T>::value) &&
                                 !is_string<T>::value>
{};
} // namespace type_traits
} // namespace aptv

///////////////////////////////////////////////////////////////////////////////
// NAMED_VARS macro creates a tuple of name-value pairs which can be printed later (each used type should support
// operator<<) all variable names are extracted at compile-time, no runtime performance penalty all variable values are
// passed efficiently - by const reference for all lvalues (by value for rvalues)
// Also supports invocation without any arguments - may be useful as a base for other macros
// EXAMPLE: int aaaa{1}, bbbb{2}; double
// cd{std::pow(2, 5)}; string s{"test"}; const int q{777}; RecPoint2D point{4483857.75, 595717.12}; RNDFId id{1, 2, 3};
// otimePoint now{timespec{time(nullptr)}};
//
// Logger().log_error("Upps: ", NAMED_VARS(q, aaaa, now, 1+1.0/2));
// cout << NAMED_VARS(s, 41 + 1, bbbb, cd, point, id) << endl;

// starting with gcc 6.1 this can be simplified to direct call (but check asm code - that it still yields compile-time
// expr) it's anonymous lambda call for now to guarantee compile-time names evaluation
#define NAMED_VARS(args...)                                                                                            \
    ([&] {                                                                                                             \
        constexpr auto names = aptv::tokenizeNames<NUM_ARGS(args)>(aptv::StackString<sizeof(#args)>{{#args}});         \
        return aptv::makeNameValuePairs(names, std::make_index_sequence<NUM_ARGS(args)>{}, aptv::makeTuple(args));     \
    })()

///////////////////////////////////////////////////////////////////////////////
// convenient macros
#define NUM_ARGS(args...) std::tuple_size<decltype(std::make_tuple(args))>::value
#define FIRST_ARG(N, ...) N
// applies provided expression to each parameter in a variadic template parameter pack
#define EXPAND_PACK(expr)                                                                                              \
    std::string { ((void)(expr), '0')... }

///////////////////////////////////////////////////////////////////////////////
// implementation
namespace aptv {
// can be replaced with std::array once C++17 is available (need constexpr operator[])
template<typename T, size_t N>
struct StackArray
{
    T arr[N];

    friend std::ostream& operator<<(std::ostream& os, const StackArray& stackArray)
    {
        using ::operator<<;
        return os << stackArray.arr;
    }
};

template<size_t LEN>
using StackString = StackArray<char, LEN>;

// extracts names at compile-time: "a, b, c" => StackArray{"a", "b", "c"}
template<size_t N, size_t LEN>
constexpr StackArray<StackString<LEN>, N> tokenizeNames(const StackString<LEN>& names)
{
    StackArray<StackString<LEN>, N> out = {};
    for (size_t i{0}, row{0}, col{0}; names.arr[i] != 0 && i < LEN; ++i) {
        if (names.arr[i] == ',') {
            ++row;
            col = 0;
        } else if (names.arr[i] != ' ') {
            out.arr[row].arr[col] = names.arr[i];
            ++col;
        }
    }
    return out;
}

// Wrapper class is used to manage lvalues and rvalues differently
// lvalues are stored by const reference and rvalues by value

// template for rvalues
template<typename T>
struct Wrapper
{
    static_assert(!std::is_reference<T>::value, "Do not use a reference type");
    constexpr explicit Wrapper(T&& t)
        : value(std::move(t))
    {}
    T value;
};

// specialization for lvalues
template<typename T>
struct Wrapper<T&>
{
    constexpr explicit Wrapper(const T& t)
        : value(t)
    {}
    const T& value;
};

template<typename T>
std::ostream& operator<<(std::ostream& os, const Wrapper<T>& wrapper)
{
    using ::operator<<;
    return os << wrapper.value;
}

template<typename T>
constexpr Wrapper<T> wrap(T&& t)
{
    return Wrapper<T>(std::forward<T>(t));
}

template<size_t LEN, typename... TVars, std::size_t... Is>
constexpr std::tuple<std::pair<StackString<LEN>, Wrapper<TVars>>...> makeNameValuePairs(
    const StackArray<StackString<LEN>, sizeof...(Is)>& names,
    std::index_sequence<Is...>,
    std::tuple<Wrapper<TVars>...>&& vars)
{
    return std::tuple<std::pair<StackString<LEN>, Wrapper<TVars>>...>{{names.arr[Is], std::get<Is>(vars)}...};
}

// make it a separate function just to support empty NAMED_VARS() call
template<typename... TVars>
constexpr std::tuple<Wrapper<TVars>...> makeTuple(TVars&&... vars)
{
    return std::tuple<Wrapper<TVars>...>{wrap(std::forward<TVars>(vars))...};
}

// similar as previous - creates a map of "enum_value -> enum_name"
template<typename TEnum, size_t LEN, size_t N>
constexpr StackArray<std::pair<TEnum, StackString<LEN>>, N> makeEnumMap(const std::array<TEnum, N>& vars,
                                                                        const StackString<LEN>& namesStr)
{
    auto names = tokenizeNames<N>(namesStr);
    StackArray<std::pair<TEnum, StackString<LEN>>, N> val2str;
    for (size_t i = 0; i < N; ++i) {
        val2str.arr[i].first = vars[i];
        val2str.arr[i].second = names.arr[i];
    }
    return val2str;
}

template<class Tuple, std::size_t... Is>
void printTupleImpl(std::ostream& os, const Tuple& t, std::index_sequence<Is...>)
{
    using ::operator<<;
    EXPAND_PACK(os << (Is == 0 ? "" : ", ") << std::get<Is>(t));
}
} // namespace aptv

///////////////////////////////////////////////////////////////////////////////
// some pretty-printers
// if you need something different - simply define more specialized version

// pretty-print a pair
template<typename T1, typename T2>
std::ostream& operator<<(std::ostream& os, const std::pair<T1, T2>& p)
{
    using ::operator<<;
    return os << p.first << ": " << p.second;
}

// special case to print plain text, like NAMED_VARS("Error message text", var1, var2)
template<unsigned long LEN1, int LEN2>
std::ostream& operator<<(std::ostream& os,
                         const std::pair<aptv::StackString<LEN1>, aptv::Wrapper<const char (&)[LEN2]>>& p)
{
    using ::operator<<;
    return os << p.second.value;
}

// pretty-print a tuple
template<class... TArgs>
std::ostream& operator<<(std::ostream& os, const std::tuple<TArgs...>& t)
{
    if (sizeof...(TArgs) == 0) {
        return os;
    }
    os << "{";
    aptv::printTupleImpl(os, t, std::index_sequence_for<TArgs...>{});
    return os << "}";
}

// pretty-print a container (except strings)
template<typename TCont,
         typename std::enable_if<aptv::type_traits::is_container<TCont>::value &&
                                 !aptv::type_traits::is_string<TCont>::value>::type*>
std::ostream& operator<<(std::ostream& os, const TCont& cont)
{
    os << "[";
    for (auto it = std::begin(cont); it != std::end(cont); ++it) {
        if (it != std::begin(cont))
            os << ", ";
        os << *it;
    }
    return os << "]";
}

// pretty-print a pointer (pointed object or null)
// if you need to print an address itself - just do (void*)ptr
template<typename T,
         typename std::enable_if<aptv::type_traits::is_any_pointer<T>::value &&
                                 aptv::type_traits::has_ostream_defined<decltype(*std::declval<T&>())>::value>::type* =
             nullptr>
std::ostream& operator<<(std::ostream& os, const T& ptr)
{
    return ptr ? (os << *ptr) : (os << "null");
}

namespace aptv {
// simple helper to convert anything with "operator<<" defined to a string (including structs, enums etc)
template<typename T, typename std::enable_if<type_traits::has_ostream_defined<T>::value>::type* = nullptr>
std::string toString(const T& value)
{
    using ::operator<<;
    std::ostringstream ostr;
    ostr << value;
    return ostr.str();
}
} // namespace aptv

///////////////////////////////////////////////////////////////////////////////
// DEFINE_OSTREAM_XXX  macros are used to define "operator<<(...)" for a class/struct or enum (see testDebugPrint.cc)
// maybe useful for usual cases to output all fields
// EXAMPLE:
//    struct Name
//    {
//        string first;
//        string last;
//        DEFINE_OSTREAM(Name, first, last);
//    };
//    enum PhoneType
//    {
//        mobile = 2,
//        home = 4,
//        work = 8
//    };
//    DEFINE_ENUM_OSTREAM(mobile, home, work);
//
//    struct Employee
//    {
//        enum Type
//        {
//            developer,
//            manager
//        };
//        friend DEFINE_ENUM_OSTREAM(developer, manager);
//
//        Name name;
//        Type type;
//        double salary;
//        map<PhoneType, string> phones;
//        DEFINE_OSTREAM(Employee, name, type, salary, phones);
//    };
//
// usage:
//    map<int, Employee> database =
//    {
//        {1, {{"John", "Smith"}, Employee::developer, 50000, {{mobile,"+1 202 303 404"},{home, "+1 505 606 707"}}}},
//        {2, {{"Rahul", "Ahluwalia"}, Employee::manager, 500000, {{home,"+1 800 BINGO"},{work, "+1 234 567 890"}}}}
//    };
//    cout << database << endl;
//
// output:
//    [1: Employee{name: Name{first: John, last: Smith}, type: developer(0), salary: 50000, phones: [mobile(2): +1 202
//    303 404, home(4): +1 505 606 707]}, 2: Employee{name: Name{first: Rahul, last: Ahluwalia}, type: manager(1),
//    salary: 500000, phones: [home(4): +1 800 BINGO, work(8): +1 234 567 890]}]

///////////////////////////////////////////////////////////////////////////////
// defines "operator<<(...)" to print an enum
// can be used both outside(as "DEFINE_ENUM_OSTREAM(a,b,c)") and inside a class (as "friend DEFINE_ENUM_OSTREAM(a,b,c)")
#define DEFINE_ENUM_OSTREAM(ENUM_VALS...)                                                                              \
    inline std::ostream& operator<<(std::ostream& os, decltype(FIRST_ARG(ENUM_VALS)) element)                          \
    {                                                                                                                  \
        using TEnum = decltype(FIRST_ARG(ENUM_VALS));                                                                  \
        constexpr size_t N = NUM_ARGS(ENUM_VALS);                                                                      \
        constexpr size_t LEN = sizeof(#ENUM_VALS);                                                                     \
        static constexpr auto val2str =                                                                                \
            aptv::makeEnumMap<TEnum, LEN, N>(std::array<TEnum, N>{ENUM_VALS}, aptv::StackString<LEN>{{#ENUM_VALS}});   \
        for (size_t i = 0; i < N; ++i)                                                                                 \
            if (val2str.arr[i].first == element)                                                                       \
                return os << val2str.arr[i].second.arr << "(" << (int)element << ")";                                  \
        return os << "_UNKNOWN_STR_(" << int(element) << ")";                                                          \
    }

///////////////////////////////////////////////////////////////////////////////
// used as DEFINE_OSTREAM(ClassName, field1, field2) - prints "{field1: val1, field2: val2}"
#define DEFINE_OSTREAM(CLASS_NAME, fields...)                                                                          \
    DEFINE_SHIFT_FRIEND_OPERATOR(CLASS_NAME)                                                                           \
    DEFINE_PRINT_FUNCTION_SIGNATURE                                                                                    \
    DEFINE_PRINT_FUNCTION_BODY(CLASS_NAME, fields)

// used as DEFINE_OSTREAM(ClassName, BaseClassName, field1, field2) - prints "{field1: val1, field2: val2} : {base1:
// val3}"
#define DEFINE_OSTREAM_WITH_BASE(CLASS_NAME, BASE_CLASS_NAME, fields...)                                               \
    DEFINE_SHIFT_FRIEND_OPERATOR(CLASS_NAME)                                                                           \
    DEFINE_PRINT_FUNCTION_SIGNATURE                                                                                    \
    DEFINE_PRINT_WITH_BASE_FUNCTION_BODY(CLASS_NAME, BASE_CLASS_NAME, fields)

// same as DEFINE_OSTREAM but defines "virtual" print functions so that object can be printed through base pointer
// use only if the polymorphic behavior is needed, otherwise DEFINE_OSTREAM is preferred
#define DEFINE_VIRT_OSTREAM(CLASS_NAME, fields...)                                                                     \
    DEFINE_SHIFT_FRIEND_OPERATOR(CLASS_NAME)                                                                           \
    virtual DEFINE_PRINT_FUNCTION_SIGNATURE DEFINE_PRINT_FUNCTION_BODY(CLASS_NAME, fields)

// same as DEFINE_OSTREAM_WITH_BASE but defines "virtual" print functions so that object can be printed through base
// pointer use only if the polymorphic behavior is needed, otherwise DEFINE_OSTREAM_WITH_BASE is preferred
#define DEFINE_VIRT_OSTREAM_WITH_BASE(CLASS_NAME, BASE_CLASS_NAME, fields...)                                          \
    DEFINE_SHIFT_FRIEND_OPERATOR(CLASS_NAME)                                                                           \
    DEFINE_PRINT_FUNCTION_SIGNATURE override DEFINE_PRINT_WITH_BASE_FUNCTION_BODY(CLASS_NAME, BASE_CLASS_NAME, fields)

///////////////////////////////////////////////////////////////////////////////
// implementation
#define DEFINE_SHIFT_FRIEND_OPERATOR(CLASS_NAME)                                                                       \
    friend std::ostream& operator<<(std::ostream& out, const CLASS_NAME& object)                                       \
    {                                                                                                                  \
        object.print(out, nullptr);                                                                                    \
        return out;                                                                                                    \
    }

#define DEFINE_PRINT_FUNCTION_SIGNATURE void print(std::ostream& out, const char* prefix) const

#define DEFINE_PRINT_FUNCTION_BODY(CLASS_NAME, fields...)                                                              \
    {                                                                                                                  \
        using ::operator<<;                                                                                            \
        out << (prefix != nullptr ? prefix : #CLASS_NAME) << NAMED_VARS(fields);                                       \
    }

#define DEFINE_PRINT_WITH_BASE_FUNCTION_BODY(CLASS_NAME, BASE_CLASS_NAME, fields...)                                   \
    {                                                                                                                  \
        using ::operator<<;                                                                                            \
        out << (prefix != nullptr ? prefix : #CLASS_NAME) << NAMED_VARS(fields) << " : ";                              \
        BASE_CLASS_NAME::print(out, #BASE_CLASS_NAME);                                                                 \
    }

#endif
