/*!
 ***********************************************************
 * @file Priority.h
 *
 * @brief file defining Priority class
 *
 * This class defines the allowable priorities.
 *
 * Each event has a particular PriorityLevel associated with
 * it. This class defines these priorities, and also contains
 * mask that helps to keep track of which priorities are
 * 'enabled' or 'disabled'.
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 ***********************************************************
 */

/**********************************************************
  Copyright (c) 2014-2015 Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/*
 ***********************************************************
 * Priority.h
 * Bryan Salesky
 * 02/28/2006
 *
 * Copyright (C) 2006 CMU. All rights reserved.
 ***********************************************************
 */
#ifndef _PRIORITY_H_
#define _PRIORITY_H_

#include <string>

namespace logger {

/*!
 * @class Priority
 * @brief This class defines the allowable priorities.
 *
 */
class Priority
{

public:
    //! The available priorities for a log event.
    typedef enum
    {
        emerg = 0x0001,  // 0000000001
        fatal = 0x0002,  // 0000000010
        alert = 0x0004,  // 0000000100
        crit = 0x0008,   // 0000001000
        error = 0x0010,  // 0000010000
        warn = 0x0020,   // 0000100000
        notice = 0x0040, // 0001000000
        info = 0x0080,   // 0010000000
        debug = 0x0100,  // 0100000000
        none = 0x0200    // 1000000000
    } PriorityLevel;

    /**
     * @brief The default constructor.  Emerg is always enabled by default.
     */
    Priority();

    /**
     * @brief The copy constructor.
     *
     * @param copy  The object to copy.
     */
    Priority(const Priority& copy);

    /**
     * @brief The assignment operator.
     *
     * @param copy  The object to copy.
     *
     * @return Side-Effect.
     **/
    Priority& operator=(const Priority& copy);

    /**
     * @brief The destructor
     */
    ~Priority();

    /*!
     * @brief Inline function that enable the particular priority. (set the bit in the pri mask).
     *
     * @param type The priority level to enable.
     */
    void enable(PriorityLevel type);

    /*!
     * @brief Inline function that disable the paritcular priority. (clear the bit in the pri mask).
     *
     * @param type The priority level to disable.
     */
    void disable(PriorityLevel type);

    /*!
     * @brief Inline function that enable the particular priority. (set the bit in the pri mask).
     *
     * @param typeStr The string equivalent of the priority level to enable.
     */
    void enable(std::string typeString);

    /*!
     * @brief Inline function that disable the paritcular priority. (clear the bit in the pri mask).
     *
     * @param typeStr The string equivalent of the priority level to disable.
     */
    void disable(std::string typeString);

    /*!
     * @brief Inline function that tests whether a priority level is enabled (if bit is set).
     *
     * @param type The priority level to test for.
     *
     * @return True if enabled, False otherwise.
     */
    bool enabled(PriorityLevel type);

    /*!
     * @brief Inline function that enable all priorities (set all the bits in the mask).
     *
     * This method will enable all the priorities if the 'from' parameter is
     * not supplied.  Or if 'from' is supplied, all the priorities at or
     * above that level are enabled.
     *
     * @param from  Enable all priorities from this level upward.
     */
    void enableAll(PriorityLevel from = none);

    /*!
     * @brief Inline function that disable all priorities (clear all the bits in the mask).
     *
     * This method will disable all the priorities if the 'from' parameter is
     * not supplied.  Or if 'from' is supplied, all the priorities at or
     * above that disabled are enabled.
     *
     * @param from  Disable all priorities from this level upward.
     */
    void disableAll(PriorityLevel from = none);

    /*!
     * @brief Translate a PriorityLevel to its string equivalent.
     *
     * @param level The PriorityLevel to return as a string.
     *
     * @return The string equivalent of the PriorityLevel, 6 characters or less.
     */
    static const char* priorityLevelAsString(const PriorityLevel& pri);

    /*!
     * @brief Translate a string to its PriorityLevel equivalent.
     *
     * @param typeString The string interpretation of a priority level.
     *
     * @return The priority level type associated with the string (defaults to none)
     */
    static PriorityLevel stringAsPriorityLevel(const std::string& typeString);

protected:
private:
    unsigned short pri_mask_; ///< mask variable used for enable/disable for Priority Level(s)
};

inline void Priority::enable(PriorityLevel type)
{
    pri_mask_ |= type;
}

inline void Priority::disable(PriorityLevel type)
{
    pri_mask_ &= ~type;
}

inline void Priority::enable(std::string typeStr)
{
    PriorityLevel type = stringAsPriorityLevel(typeStr);
    pri_mask_ |= type;
}

inline void Priority::disable(std::string typeStr)
{
    PriorityLevel type = stringAsPriorityLevel(typeStr);
    pri_mask_ &= ~type;
}

inline bool Priority::enabled(PriorityLevel type)
{
    return ((pri_mask_ & type) > 0);
}

inline void Priority::enableAll(PriorityLevel from)
{
    while (from != 0) {
        enable(from);
        from = (PriorityLevel)(((short)from) >> 1);
    }
}

inline void Priority::disableAll(PriorityLevel from)
{
    while (from != 0) {
        disable(from);
        from = (PriorityLevel)(((short)from) >> 1);
    }
}

} // namespace logger

#endif
