/*!
 ***********************************************************
 * @file oprofile.h
 * @brief A file defining several profiling classes
 *
 * @section License
 *
 * Copyright (c) 2016- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/07/2017
 * The first version of profiling classes.
 * @n
 *
 ***********************************************************
 */

#ifndef OTTOPROFILE_H
#define OTTOPROFILE_H

#include <TimeStamp/TimeStamp.h>
#include <map>
#include <mutex>
#include <string>

namespace ottoprofile {

/*!
 * @class ProfileManager
 *
 * @brief This class shall store and clear the data points collected by ProfileCollector. It shall store the data
 * points collected \n by the ProfileCollector for a given profile in a map container. Access to this container shall be
 * synchronized as it shall be \n accessed/updated by profile collectors and profile dumpers. Data point (CPU profiling
 * data) for a profile consists of the \n following. \n     1. Profile name for which cpu profiling data to collect \n
 * 2. Start time is the time when profiling is started \n     3. Duration is the elapsed time between profile start time
 * and end time with nanosecond accuracy. \n ProfileManager shall be created when ProfileCollector starts collecting the
 * data points. The ProfileManager shall exist \n as long as application/task process is running. The ProfileDumper
 * shall get the CPU profiling data from ProfileManager and \n dumps to a file.The CPU profiling data shall be cleared
 * once data is written to a file. ProfileDumper shall create ProfileManager \n if the ProfileManager is not available.
 *
 */
class ProfileManager
{
public:
    /*!
     * @brief Default constructor.
     * \n Default constructor. It shall clear the container which it uses to store CPU profiling data.
     */
    ProfileManager();

    /*!
     * @brief Virtual destructor.It shall clear the container which it uses to store CPU profiling data.
     *
     */
    virtual ~ProfileManager();

    template<class T>
    using ProfileData = std::vector<std::pair<ottotime::otimePoint, T>>;

    using CPUMeasures = std::map<std::string, ProfileData<float>>; ///< A type definition for map container.
                                                                   ///< Profile data stored for a given profile name as
                                                                   ///< (key,value) pairs.

    using CPUMeasuresIter =
        CPUMeasures::iterator; ///< A type definition for CPUMeasures container iterator
                               ///< The iterator CPUMeasuresIter access the contents of the CPUMeasures container.
    /*!
     * @brief This method shall add a given measurement(CPU profiling data) to a container for a given profile name.
     * \n Access to this container shall be synchronized as it is a shared container. The CPU profiling data shall be
     * \n supplied by ProfileCollector.
     *
     * @param[in] profileName : Name of the profile for which profile data collected
     * @param[in] measurement : The profiling data (start time,elapsed time) of the given profile name to be stored in
     * CPU measurements container.
     *
     * @return bool : true always.
     *
     */
    bool addCPUMeasure(const std::string& profileName, const std::pair<ottotime::otimePoint, float>& measurement);

    /*!
     * @brief This method shall clear the  measurements(CPU profiling data) from a container for a given profile.
     * \n Access to this container shall be synchronized as it is a shared container. The measurements data shall
     * \n be cleared after the ProfileDumper writes the profiling data to a file.
     *
     * @param[in] profileName : Name of the profile for which profile data to be deleted from CPU measurements
     * container.
     *
     * @return bool : true, if profile data exists in CPU measurements container and cleared.false, if profile data
     * doesn't exist in container.
     *
     */
    bool clearCPUMeasure(const std::string& profileName);

    /*!
     * @brief This method shall get the  CPU profiling data of the given profile stored in CPU measurements container.
     * \n This measurements data shall be used by the ProfileDumper to write to a file.
     *
     * @param[in] profileName : Name of the profile for which profile data to be returned..
     *
     * @return measurements data as vector. vector holds profile data as pair <start time,elapsed time> objects
     *
     */
    ProfileData<float>& getCPUMeasures(const std::string& profileName) { return cpuMeasures_[profileName]; }

private:
    CPUMeasures cpuMeasures_{}; ///< The CPU measurements map container to store CPU measurements.
                                ///< Key - in map is a profile name for which profile data is collected.
                                ///< Value - in map is a vector of profile data for the corresponding profile.
                                ///< profile data vector contains elements as pair<start time,elapsed time> objects

    std::mutex cpuMeasuresLock_{}; ///< The mutex lock for CPU measurements container cpuMeasures_.
                                   ///< The CPU measurements container is a shared container between ProfileCollector
                                   ///< and ProfileDumper.
};

/*!
 * @class ProfileCollector
 *
 * @brief This class shall collect the CPU profiling data points.Data point (CPU profiling data) for a profile consists
 * of the \n following. \n		1. Profile name for which cpu profiling data to collect. \n		2. Start time is the
 * time when profiling is started. \n		3. Duration is the elapsed time between profile start time and end time with
 * nanosecond accuracy.
 *
 * \n The ProfileCollector shall store the start time when it starts the CPU profiling. It shall calculate the elapsed
 * time when it \n stops the profiling. The ProfileCollector shall create a shared ProfileManager when it starts
 * collecting the profiling data. \n The collected CPU profiling data shall be given to ProfileManager which stores the
 * data in a CPU measurements map container \n where Key is the profile name and value is a vector of profile data which
 * holds elements as pair <start time,elapsed time> \n objects. This class shall allow to set the ProfileManager on
 * ProfileCollector. The profiling functionality shall be controlled \n (enable/disable) via configuration.
 *
 * \n The following is the example to collect the CPU profiling data when a particular method named 'executive' is
 * called for a process \n  named "someTask". \n     START profiling "someTask"   --- Profile collector starts the time
 * \n     run execute() method
 * \n     STOP profiling     -- Profile collector shall calculate the elapsed time to complete the execute() method.
 * \n                           This data shall be stored in CPU profiling container.
 *
 */
class ProfileCollector
{
public:
    /*!
     * @brief Parameterized constructor using a profile name. It shall set the profile name and stores the CPU profiling
     * startTime.
     *
     * @param[in] profileName as string.Name of the profile for which profile data to be collected.
     *
     */
    ProfileCollector(const std::string& profileName);

    /*!
     * @brief Default destructor. This method shall compute the elapsedTime. i.e  endTime - startTime. It shall create a
     * shared \n ProfileManger if doesn't exist and the collected profiling data shall be given to ProfileManger which
     * stores the data \n in a CPU measurements map container where Key is the profile name and value is a vector of
     * profile data which holds \n elements as pair <start time,elapsed time> objects.
     *
     */
    virtual ~ProfileCollector();

    /*!
     * @brief This method shall return a pointer to the ProfileManager. The callers of this method shall use
     * ProfileManager \n to store or clear the CPU profiling data  from CPU measurements.
     *
     * @return The ProfileManager pointer.
     *
     */
    static ProfileManager* getProfileManager() { return profileManager_; }

    /*!
     * @brief This method shall set the ProfileManager. The ProfileDumper shall create ProfileManger if not available
     * and \n supply to ProfileCollector.
     *
     * @param[in] ProfileManager pointer
     *
     * @return void
     *
     */
    static void setProfileManager(ProfileManager* const profileManager) { profileManager_ = profileManager; }

    /*!
     * @brief This method shall return  the mutex lock associated with the shared ProfileManager.
     * \n The profile manager shall be used by profile collectors and dumpers..
     *
     * @return The reference of the mutex lock of profileManager_.
     *
     */
    static std::mutex& getProfileManagerLock() { return profileManagerLock_; }

private:
    std::string profileName_{""}; ///< Name of the profile for which profile data to be collected
    ottotime::otimePoint
        startTime_{}; ///< The profile start time recorded when ProfileCollector starts collecting profile data.

    static ProfileManager* profileManager_; ///< The static pointer to the ProfileManager.
                                            ///< ProfileCollector uses ProfileManager to save the collected profiling
                                            ///< data in CPU measurements map container

    static std::mutex profileManagerLock_; ///< The static mutex lock for profileManager_ .
                                           ///< ProfileManager is shared between ProfileCollector and ProfileDumper.
};

/*!
 * @class ProfileDumper
 *
 * @brief This class shall get the CPU profiling data for a profile from the ProfileManager and dump/publish data to a
 * file. \n The ProfileManager shall clear the profile data once it is published to a file. The ProfileManager shall
 * collect the CPU \n profiling data from ProfileCollector and stores in CPU profiling map container where key in map is
 * a profile name and \n value in map is a vector of profile data which holds elements as pair <start time,elapsed time>
 * objects. The CPU profiling \n data for a profile consists of the following. \n	1. Profile name for which cpu
 * profiling data to collect. \n	2. Start time is the time when profiling is started. \n	3. Duration is the elapsed
 * time between profile start time and end time with nanosecond accuracy.
 *
 * \n This class shall create ProfileManager if not available and shall supply to the ProfileCollector as ProfileManager
 * is a shared \n object among profile collectors and dumpers.
 */
class ProfileDumper
{
public:
    /*!
     * @brief Parameterized Constructor. It shall set the profile name for which CPU profiling data to be collected.
     * \n It shall get the ProfileManager from ProfileCollector If available. Otherwise, It shall create ProfileManager
     * \n and set the ProfileManager on ProfileCollector. ProfileManager shall be a shared object among profile
     * collectors \n and dumpers.
     *
     * @param[in] profileName as string.Name of the profile for which profile data to be published.
     *
     */
    ProfileDumper(const std::string& profileName);

    /*!
     * @brief Default virtual destructor.
     *
     */
    virtual ~ProfileDumper();

    /*!
     * @brief This method shall fetch the CPU profiling data from ProfileManager and writes into a file created with
     * \n profile name (profileName.txt). The profiling data shall be cleared once the data to a file is published.
     *
     * @return void
     *
     */
    void writeToFile();

    /*!
     * @brief This method shall get the latest CPU profiling data for a particular profile. First, it shall get the
     * profile \n data in a vector container for a given profile from ProfileManager. It shall fetch the latest CPU
     * measurement \n (elapsed time) from the element which is the last one pushed in container.
     *
     * @return  The latest CPU profiling data of the given profile as a float.
     * \n       It returns a -1 as a float when there is no valid value.
     *
     */
    float getLastMeasurement();

private:
    std::string profileName_{""}; ///< Name of the profile for which profile data to be dumped to file.

    ProfileManager* profileManager_{
        nullptr}; ///< The pointer to the profile manager.
                  ///< The CPU measurements map container is accessed via ProfileManager to get profile data.
};

} // namespace ottoprofile

#endif // OTTOPROFILE_H
