/*!
 ***********************************************************
 * @file oprofile.cc
 * @brief A file implementing several profiling classes
 *
 * @section License
 *
 * Copyright (c) 2016- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Junsung Kim (junsung.kim@delphi.com)
 * @date 07/07/2017
 * The initial implementation of profiling classes
 * @n
 *
 ***********************************************************
 */

#ifndef OTTOPROFILE_CC
#define OTTOPROFILE_CC

#include <fstream>
#include <iomanip>
#include <profile/oprofile.h>

namespace ottoprofile {

ProfileManager::ProfileManager()
{
    std::lock_guard<std::mutex> lock(cpuMeasuresLock_);
    cpuMeasures_.clear();
}

ProfileManager::~ProfileManager()
{
    std::lock_guard<std::mutex> lock(cpuMeasuresLock_);
    cpuMeasures_.clear();
}

bool ProfileManager::addCPUMeasure(const std::string& profileName,
                                   const std::pair<ottotime::otimePoint, float>& measurement)
{
    std::lock_guard<std::mutex> lock(cpuMeasuresLock_);

    const CPUMeasuresIter it = cpuMeasures_.find(profileName);

    if (it != cpuMeasures_.end()) {
        cpuMeasures_[profileName].push_back(measurement);
    } else {
        ProfileData<float> dummyMeasures;
        dummyMeasures.clear();
        dummyMeasures.push_back(measurement);
        cpuMeasures_[profileName] = dummyMeasures;
    }

    return true;
}

bool ProfileManager::clearCPUMeasure(const std::string& profileName)
{
    bool ret(false);

    std::lock_guard<std::mutex> lock(cpuMeasuresLock_);

    const CPUMeasuresIter it = cpuMeasures_.find(profileName);

    if (it != cpuMeasures_.end()) {
        cpuMeasures_[profileName].clear();
        ret = true;
    }

    return ret;
}

ProfileManager* ProfileCollector::profileManager_ = nullptr;
std::mutex ProfileCollector::profileManagerLock_;

ProfileCollector::ProfileCollector(const std::string& profileName)
{
    profileName_ = profileName;
    startTime_.getCurrentSteadyTime();
}

ProfileCollector::~ProfileCollector()
{
    const ottotime::otimePoint endTime = ottotime::otimePoint::buildCurrentSteadyTime();

    {
        std::lock_guard<std::mutex> lock(profileManagerLock_);
        if (profileManager_ == nullptr) {
            profileManager_ = new ProfileManager();
        }
    }

    float duration = static_cast<float>((endTime - startTime_).totalNanoSeconds()) / 1000000000.0F;
    static_cast<void>(profileManager_->addCPUMeasure(profileName_, std::make_pair(startTime_, duration)));
}

ProfileDumper::ProfileDumper(const std::string& profileName)
{
    profileName_ = profileName;
    profileManager_ = ProfileCollector::getProfileManager();

    if (profileManager_ == nullptr) {
        std::lock_guard<std::mutex> lock(ProfileCollector::getProfileManagerLock());
        profileManager_ = new ProfileManager();
        ProfileCollector::setProfileManager(profileManager_);
    }
}

ProfileDumper::~ProfileDumper() {}

void ProfileDumper::writeToFile()
{
    std::ofstream outFile;
    const std::string fileName = profileName_ + ".txt";
    outFile.open(fileName, std::ios_base::app);
    outFile << std::setprecision(9) << std::fixed;

    ProfileManager::ProfileData<float>& cpuProfileData = profileManager_->getCPUMeasures(profileName_);

    for (std::vector<uint32_t>::size_type ii = 0; ii < cpuProfileData.size(); ++ii) {
        outFile << otimeToSeconds(cpuProfileData.at(ii).first);
        outFile << ", " << cpuProfileData.at(ii).second << std::endl;
    }

    outFile.close();

    static_cast<void>(profileManager_->clearCPUMeasure(profileName_));
}

float ProfileDumper::getLastMeasurement()
{
    float measurement(-1.0F);
    ProfileManager::ProfileData<float>& cpuProfileData = profileManager_->getCPUMeasures(profileName_);

    if (cpuProfileData.size() > 0) {
        measurement = cpuProfileData.back().second;
    }

    return measurement;
}
} // namespace ottoprofile

#endif // OTTOPROFILE_CC
