#include <recgeometry/EuclideanCluster.h>

#include "lib/avlog/avlog.hpp"

EuclideanCluster::EuclideanCluster()
{
    minRadius_ = 0.6;
    maxRadius_ = 1.0;
    // resolutionAng_ = 0.125;
    resolutionAng_ = 0.25; // Constant angular resolution, updated on 10/9/2015
    minLineAng_ = 5.0;
    rate_ = sin(resolutionAng_ * M_PI / 180) / sin((minLineAng_ - resolutionAng_) * M_PI / 180);
}

bool EuclideanCluster::cluster(std::vector<std::vector<RecPoint2D>>& segments, const std::vector<RecPoint2D>& lidarPts)
{
    AVLOG_DEBUG("--------------------start euclideanCluster------------------------");

    // build kd-tree
    RecKDTree2D<size_t> kdTree;
    for (size_t i = 0; i < lidarPts.size(); ++i) {
        kdTree.addElement(lidarPts[i], i);
    }
    kdTree.buildTree();

    // start clustering

    // prepare for adaptive radius
    // get point number
    auto ptNum = lidarPts.size();

    std::vector<bool> checked(ptNum);
    std::vector<size_t> Q;
    std::vector<size_t> neighbors;
    AVLOG_INFO("ptNum {}", ptNum);
    for (size_t i = 0; i < ptNum; ++i) {
        // pull current point from list of all Raw points
        RecPoint2D currentPoint = lidarPts[i];

        // setup parameters for neighbor-getting
        // adapt the radius with the range
        double dist_thres = currentPoint.distance(RecPoint2D());
        radius_ = dist_thres * rate_;
        radius_ = std::max(radius_, minRadius_);
        radius_ = std::min(radius_, maxRadius_);

        // check if the current point has been processed
        if (checked[i] == true)
            continue;

        // recursively find all points that each point has at least one point within radius
        Q.push_back(i);
        AVLOG_DEBUG("Start a cluster: index {}, radius {:.2f}", i, radius_);
        size_t j = 0;
        while (j < Q.size()) {
            if (!checked[Q[j]]) {
                // start checking
                AVLOG_DEBUG("Checking Q[j] {}", Q[j]);
                checked[Q[j]] = true;
                neighbors.clear();
                kdTree.getElementsWithin(lidarPts[Q[j]], radius_, neighbors);
                for (size_t k = 0; k < neighbors.size(); ++k) {
                    auto poses_idx = neighbors[k];
                    if (!checked[poses_idx] && std::find(Q.begin(), Q.end(), poses_idx) == Q.end()) {
                        Q.push_back(poses_idx);
                        AVLOG_DEBUG("Add index {} into Q", poses_idx);
                    }
                }
            }
            j++;
        }
        // save Q as a cluster
        AVLOG_DEBUG("Finish a cluster, point num: {}", Q.size());
        if (Q.size() >= 1) {
            std::vector<RecPoint2D> C(Q.size());
            for (size_t l = 0; l < Q.size(); ++l) {
                C[l] = lidarPts[Q[l]];
            }
            segments.push_back(C);
        }
        Q.clear();
    }
    AVLOG_INFO("Segment Num: {}", segments.size());
    AVLOG_DEBUG("--------------------end euclideanCluster------------------------");

    return true;
}
