/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**
 * @file RecKDTree2D.h
 * @author: Christopher Baker (tallbaker@cmu.edu)
 * @date:
 *
 * @attention Copyright (c) 2007
 * @attention Carnegie Mellon University
 * @attention All rights reserved.
 */
#ifndef _RECKDTREE2D_H_
#define _RECKDTREE2D_H_

#include "recgeometry/recGeometry.h"
#include <algorithm>
#include <list>
#include <math.h>
#include <vector>

/**\brief Provide a generic KD-Tree implementation for mapping RecPoint2D's to some generic payload
 * \ingroup recGeometryGroup
 *
 * Notes:
 * - PAYLOAD must be able to live in a vector, which means unit constructor and assignment
 *     operator at least
 * - For large PAYLOAD classes, it may be advantageous to store the payloads separate from
 *     the tree elements and to let TreeElement::payload be a pointer, since the TreeElements
 *     are continuously swapped as part of sort operations and large PAYLOAD, or those with complex
 *     assignment operators may seriously degrade performance
 * - An additional O(lg(n)) speedup may be had by caching the O(n lg(n)) sort results and reducing the
 *     per-depth re-sorts to an O(n) operation.  Initial experiments point toward a speedup on the
 *     order of 20% for trees up to ~1M elements, but that the extra memory cost associated with this
 *     dynamic programming shifts the limiting agent to cache misses and actualy degrades performance
 *     above 10M elements
 */
template<class PAYLOAD>
class RecKDTree2D final
{
public:
    // Handy Typedefs
    typedef std::vector<PAYLOAD> PayloadVector;
    typedef std::vector<std::pair<RecPoint2D, PAYLOAD>> PointPayloadPairVector;

    // building the tree
    bool reserve(unsigned long size);
    bool addElement(const RecPoint2D& location, const PAYLOAD& payload);
    bool addElements(const PointPayloadPairVector& pairs);
    bool buildTree();
    bool clearTree();

    // using the tree
    bool getClosestElement(const RecPoint2D& queryPoint,
                           PAYLOAD& out,
                           double maxDistSq = INFINITY,
                           RecPoint2D* ploc = nullptr) const;
    bool getElementsWithin(const RecPoint2D& queryPoint, double radius, PayloadVector& out) const;

    void getElements(PointPayloadPairVector& pairs) const;

private:
    /**\brief Used to be more complicated, now is basically a key-value pair */
    struct TreeElement
    {
        RecPoint2D loc;
        PAYLOAD payload;
    };

    /**\brief This is how we store things */
    typedef std::vector<TreeElement> ElemVector;

    // predicates for sorting
    struct LessX
    {
        bool operator()(const TreeElement& a, const TreeElement& b) const { return a.loc.x < b.loc.x; };
    };

    struct LessY
    {
        bool operator()(const TreeElement& a, const TreeElement& b) const { return a.loc.y < b.loc.y; };
    };

    /**\brief Similar to an iterator, this encapsulates the tree-in-an-array indexing stuff */
    class TreeWalker
    {
    public:
        TreeWalker(const ElemVector& elements);
        TreeWalker() = default;
        bool isValid() const { return from_ != std::numeric_limits<unsigned long>::max(); };

        // element accessors
        inline unsigned long getElementIndex() const { return ((from_ + to_) >> 1); };

        // walking
        bool getSmaller(TreeWalker& out) const;
        bool getLarger(TreeWalker& out) const;

    private:
        // The bounds of our walk
        unsigned long from_{std::numeric_limits<unsigned long>::max()};
        unsigned long to_{std::numeric_limits<unsigned long>::max()};
    };

    void buildTree(ElemVector& vec, unsigned int begin, unsigned int end, unsigned int depth);

    double findNearest(const TreeWalker& walker,
                       const RecPoint2D& pt,
                       double maxDistSq,
                       const TreeElement*& nearest,
                       unsigned int depth) const;

    void findWithin(const TreeWalker& tree,
                    const RecPoint2D& pt,
                    double radiusSq,
                    PayloadVector& cluster,
                    unsigned int depth) const;

private:
    const TreeElement& getElement(const TreeWalker& walker) const { return elements_[walker.getElementIndex()]; }

    bool treeBuilt_{false}; ///< Boolean indication of whether the tree has been built
    ElemVector elements_;   ///< This is where we store everything
};

#include "RecKDTree2D.impl.h"

#endif // ifndef _RECKDTREE2D_H_
