/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**
 * @file RecKDTree2D.cc
 * @author: Chris Urmson (curmson@ri.cmu.edu)
 * @date:
 *
 * @attention Copyright (c) 2007
 * @attention Carnegie Mellon University
 * @attention All rights reserved.
 */
#ifndef _RECKDTREE2D_IMPL_H_
#define _RECKDTREE2D_IMPL_H_

#ifndef _RECKDTREE2D_H_
#error Must only be included from RecKDTree2D.h
#endif

#include <list>

/**
 * @brief Analagous to vector::reserve to hint at the number of element on the way
 */
template<class PAYLOAD>
bool RecKDTree2D<PAYLOAD>::reserve(unsigned long size)
{
    if (treeBuilt_) {
        // tree is already built
        return false;
    }
    elements_.reserve(size);
    return true;
}

/**\brief Add a single key/value pair to the elements to me KD-Tree'd
 * \param location The key
 * \param payload The value
 * \return boolean success (false implies the tree has already been build and you should clearTree() first
 */
template<class PAYLOAD>
bool RecKDTree2D<PAYLOAD>::addElement(const RecPoint2D& location, const PAYLOAD& payload)
{
    if (treeBuilt_) {
        // tree is already built, adds are invalid
        return false;
    }
    elements_.push_back(TreeElement{location, payload});
    return true;
}

/**\brief Add a bunch of key/value pairs to the elements to me KD-Tree'd
 * \param pairs The datum
 * \return boolean success (false implies the tree has already been build and you should clearTree() first
 */
template<class PAYLOAD>
bool RecKDTree2D<PAYLOAD>::addElements(const PointPayloadPairVector& pairs)
{
    if (treeBuilt_) {
        // tree is already built, adds are invalid
        return false;
    }

    // iterate and push back
    elements_.reserve(elements_.size() + pairs.size());
    for (const auto& pair : pairs) {
        elements_.push_back(TreeElement{pair.first, pair.second});
    }
    return true;
}

/**\brief Actuallly run the sort/build process
 * \return boolean success (false implies tree already built)
 */
template<class PAYLOAD>
bool RecKDTree2D<PAYLOAD>::buildTree()
{
    if (treeBuilt_ || elements_.empty()) {
        // tree is already built
        return false;
    }
    // now recursively build the tree
    if (elements_.size() > 1) {
        buildTree(elements_, 0, elements_.size(), 0);
    }
    treeBuilt_ = true;
    return true;
}

/**\brief Nuke the tree in-place
 * \return boolean success (always true for now)
 */
template<class PAYLOAD>
bool RecKDTree2D<PAYLOAD>::clearTree()
{
    elements_.clear();
    treeBuilt_ = false;
    return true;
}

/**
 *@brief builds the KD-tree.  The split plane goes through points, and switches every depth change
 *@param vec the vector of TreeElements we're building the tree from
 *@param begin index into the vector this tree starts at
 *@param end index into the vector just after this tree ends
 *@param depth current depth we're building this tree at
 */
template<class PAYLOAD>
void RecKDTree2D<PAYLOAD>::buildTree(ElemVector& vec, unsigned int begin, unsigned int end, unsigned int depth)
{
    // split on the median... build the subtrees
    unsigned int median = (end + begin) >> 1;

    // arrr, there be sorting to do
    if ((depth & 1) == 0) {
        // split along x
        std::sort(&vec[begin], &vec[end], LessX());
    } else {
        // split along y
        std::sort(&vec[begin], &vec[end], LessY());
    }

    if ((begin + 1) < median) {
        // there are at least two elements in the left half, sort some more
        buildTree(vec, begin, median, depth + 1);
    }

    if ((median + 2) < end) {
        // there are at least two elements in the right half, sort some more
        buildTree(vec, median + 1, end, depth + 1);
    }
}

/**\brief Public closest lookup
 * \param pt The query point
 * \param out The payload output (by reference)
 * \return boolean success (falses implies unbuilt/empty tree)
 */
template<class PAYLOAD>
bool RecKDTree2D<PAYLOAD>::getClosestElement(const RecPoint2D& queryPoint,
                                             PAYLOAD& out,
                                             double maxDistSq,
                                             RecPoint2D* ploc) const
{
    if (!treeBuilt_) {
        return false;
    }

    const TreeElement* nearest{nullptr};
    TreeWalker walker(elements_);
    const double distance = findNearest(walker, queryPoint, maxDistSq, nearest, 0);

    if (distance == INFINITY || nearest == nullptr)
        return false;

    out = nearest->payload;
    if (ploc)
        *ploc = nearest->loc;
    return true;
}

/**\brief Public blob query
 * \param pt The query point
 * \param radius The query radius
 * \param out A vector to receive the points in the blob
 * \return boolean success (false implies unbuilt/empty tree or no elements withing radius of pt)
 */
template<class PAYLOAD>
bool RecKDTree2D<PAYLOAD>::getElementsWithin(const RecPoint2D& queryPoint, double radius, PayloadVector& out) const
{
    if (!treeBuilt_)
        return false;

    const auto oldSize = out.size();
    double radiusSq = radius * radius;
    TreeWalker walker(elements_);
    findWithin(walker, queryPoint, radiusSq, out, 0);

    return oldSize != out.size();
}

/**
 *@brief the recursive search that looks for tree elements that are within the given radius
 *@param tree the span we're looking at
 *@param pt the point of interest we're looking for things near
 *@param radiusSq square of the radius of interest
 *@param cluster the returned list of nearby tree elements
 *@param depth the current depth of the search
 */
template<class PAYLOAD>
void RecKDTree2D<PAYLOAD>::findWithin(const RecKDTree2D<PAYLOAD>::TreeWalker& tree,
                                      const RecPoint2D& pt,
                                      double radiusSq,
                                      PayloadVector& cluster,
                                      unsigned int depth) const
{
    bool smallerInRange, largerInRange;
    double hpDis;

    // figure out where the query point is relative to the hinge in our dimension
    if ((depth & 1) == 0) {
        hpDis = pt.x - getElement(tree).loc.x;
    } else {
        hpDis = pt.y - getElement(tree).loc.y;
    }

    double hpDisSq = hpDis * hpDis;
    if (hpDisSq < radiusSq) {
        // close enough to the hinge to require both be searched
        smallerInRange = true;
        largerInRange = true;
    } else if (hpDis < 0) {
        // in the smaller range only
        smallerInRange = true;
        largerInRange = false;
    } else {
        // in larger range only
        smallerInRange = false;
        largerInRange = true;
    }

    if (smallerInRange && largerInRange) {
        // this hinge is a candidate
        double distSq = getElement(tree).loc.distanceSq(pt);
        if (distSq < radiusSq) {
            cluster.push_back(getElement(tree).payload);
        }
    }

    TreeWalker newWalker(tree);

    if (smallerInRange && tree.getSmaller(newWalker)) {
        findWithin(newWalker, pt, radiusSq, cluster, depth + 1);
    }
    if (largerInRange && tree.getLarger(newWalker)) {
        findWithin(newWalker, pt, radiusSq, cluster, depth + 1);
    }
}

/**
 *@brief the recursive nearest neighbor search algorithm from Andrew Moore's thesis
 *@param tree the span we're searching
 *@param pt the point we're trying to find nearest to
 *@param maxDistSq the maximum possible distance a better solution can be
 *@param nearest the current nearest tree element
 *@param depth the current search depth
 *@return the squared distance between the nearest point and the query point
 */
template<class PAYLOAD>
double RecKDTree2D<PAYLOAD>::findNearest(const RecKDTree2D<PAYLOAD>::TreeWalker& walker,
                                         const RecPoint2D& pt,
                                         double maxDistSq,
                                         const RecKDTree2D<PAYLOAD>::TreeElement*& nearest,
                                         const unsigned int depth) const
{
    // depending on depth, we change axis for the split
    const bool evenLayer = (depth & 1) == 0;
    const double delta = evenLayer ? pt.x - getElement(walker).loc.x : pt.y - getElement(walker).loc.y;

    const bool inSmaller = (delta < 0);

    TreeWalker nearWalker;
    TreeWalker farWalker;

    // now find out which half of the tree the node is more likely in....
    if (inSmaller) {
        walker.getSmaller(nearWalker);
        walker.getLarger(farWalker);
    } else {
        walker.getLarger(nearWalker);
        walker.getSmaller(farWalker);
    }

    double distSq{INFINITY};

    // now look there.
    if (nearWalker.isValid())
        distSq = findNearest(nearWalker, pt, maxDistSq, nearest, depth + 1);

    maxDistSq = std::min(maxDistSq, distSq);

    // next we check to see if the other half of the space here, is within range of our test point...
    const double hpDisSq = delta * delta;

    if (hpDisSq < maxDistSq) {
        // it is, oh joy.

        // now we're going to check to see if the pivot point is closer than what we've found so far.... if it is, we
        // have to check the other half
        const double pivotDistSq = getElement(walker).loc.distanceSq(pt);

        // check to see if the pivot point is closer...
        if (pivotDistSq < distSq) {
            nearest = &getElement(walker);
            distSq = pivotDistSq;
            maxDistSq = pivotDistSq;
        }

        // now check out the other tree....
        if (farWalker.isValid()) {
            const TreeElement* tmpElem{nullptr};
            const double tmpDistSq = findNearest(farWalker, pt, maxDistSq, tmpElem, depth + 1);
            if (tmpDistSq < distSq) {
                // it's closer, so update the returned distance squared and nearest...
                nearest = tmpElem;
                distSq = tmpDistSq;
            }
        }
    }

    return distSq;
}

template<class PAYLOAD>
void RecKDTree2D<PAYLOAD>::getElements(PointPayloadPairVector& pairs) const
{
    for (const auto& element : elements_) {
        pairs.push_back(std::make_pair(element.loc, element.payload));
    }
}

// TREEWALKER Implementation

/**\brief Initialization constructor, creates a TreeWalker that spans the whole of elements */
template<class PAYLOAD>
RecKDTree2D<PAYLOAD>::TreeWalker::TreeWalker(const RecKDTree2D<PAYLOAD>::ElemVector& elements)
    : from_(0)
    , to_(elements.size())
{}

/**\brief Populate a TreeWalker representing the left (smaller) half
 * \param out The new TreeWalker (out by reference)
 * \return boolean success (false implies no left half)
 */
template<class PAYLOAD>
bool RecKDTree2D<PAYLOAD>::TreeWalker::getSmaller(TreeWalker& out) const
{
    if ((from_ + 1) >= to_) {
        // we need a separation bigger than 1 to have a smaller branch
        return false;
    }
    out.from_ = from_;
    out.to_ = (from_ + to_) >> 1;
    return true;
}

/**\brief Populate a TreeWalker representing the right (larger) half
 * \param out The new TreeWalker (out by reference)
 * \return boolean success (false implies no right half)
 */
template<class PAYLOAD>
bool RecKDTree2D<PAYLOAD>::TreeWalker::getLarger(TreeWalker& out) const
{
    if ((from_ + 2) >= to_) {
        // we need a separation bigger than 2 to have a larger branch
        return false;
    }
    out.from_ = 1 + ((from_ + to_) >> 1);
    out.to_ = to_;
    return true;
}

#endif // ifndef _RECKDTREE2D_IMPL_H_
