
/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file recClipper.h
 * @product Common / RecGeometry
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated May 28, 2018
 * @brief Contains functions for manipulating 2D polygons.
 *
 * Supported functions include: boolean intersection, boolean union, offset (buffering) and sorting vertices.
 * There is detailed documentation on the Wiki:
 * Confluence --> Perception --> Perception Modules --> RecClipper
 *
 * This wraps a 3rd-party library called Clipper version 6.4.2
 * http://www.angusj.com
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 * - Boost Software License (http://www.boost.org/LICENSE_1_0.txt)
 *
 ***********************************************************************************************************************
 */

/**********************************************************
  Copyright (c) 2015-    Ottomatika, Inc.
 **********************************************************/

#ifndef _RECCLIPPER_H_
#define _RECCLIPPER_H_

#include <list>
#include <recgeometry/recGeometry.h>
#include <vector>

#include <Clipper/clipper.hpp>

// Scaling factor, because Clipper uses integer math.
// Use a double type, so when we multiply the UTM
// co-ordinates, we don't lose precision.
const double SCALING_FACTOR = 100000000;

// ToDo: Add function to sort vertices of a non-convex 2D polygon e.g. outline of a building.

class RecClipper
{
public:
    //----------------------------------------------------------------------------------------------------------------//

    // Functions to sort the winding order of vertices:

    /*!
     * @brief Sort the vertices of a 2D polygon to have a CW (Clockwise) winding order.
     *
     * The algorithm used is as follows: The visual center of the 2D polygon
     * is found. The angle of each vertex with respect to the center point
     * is computed. The vertices are sorted by the angles. The first vertex
     * of the sorted output set of vertices may be different to the first
     * vertex of the input.
     *
     * Note: This only works with convex polygons.
     */
    static RecPolygon2D sortPolygonCW(const RecPolygon2D& polygon);

    /*!
     * @brief Sort the vertices of a 2D polygon to have CCW (Counter-Clockwise) winding order.
     *
     * The algorithm used is as follows: The visual center of the 2D polygon
     * is found. The angle of each vertex with respect to the center point
     * is computed. The vertices are sorted by the angles. The first vertex
     * of the sorted output set of vertices may be different to the first
     * vertex of the input.
     *
     * Note: This only works with convex polygons.
     */
    static RecPolygon2D sortPolygonCCW(const RecPolygon2D& polygon);

    /*!
     * @brief Sort a set of 2D points (representing the vertices of a 2D polygon)
     * to have a CW (Clockwise) winding order.
     *
     * The algorithm used is as follows: The visual center of the 2D polygon
     * is found. The angle of each vertex with respect to the center point
     * is computed. The vertices are sorted by the angles. The first vertex
     * of the sorted output set of vertices may be different to the first
     * vertex of the input.
     *
     * Note: This only works with convex polygons.
     */
    static std::vector<RecPoint2D> sortVerticesCW(const std::vector<RecPoint2D>& vertices);

    /*!
     * @brief Sort a set of 2D points (representing the vertices of a 2D polygon)
     * to have CCW (Counter-Clockwise) winding order.
     *
     * The algorithm used is as follows: The visual center of the 2D polygon
     * is found. The angle of each vertex with respect to the center point
     * is computed. The vertices are sorted by the angles. The first vertex
     * of the sorted output set of vertices may be different to the first
     * vertex of the input.
     *
     * Note: This only works with convex polygons.
     */
    static std::vector<RecPoint2D> sortVerticesCCW(const std::vector<RecPoint2D>& vertices);

    /*!
     * @brief Reverse the winding order of a set of vertices of a 2D polygon.
     */
    static RecPolygon2D reverseVertices(const RecPolygon2D& polygon);

    //----------------------------------------------------------------------------------------------------------------//

    // Functions to check the winding order of vertices:

    /*!
     * @brief Check if the vertices of a 2D polygon have a CW (Clockwise) winding order.
     *
     * This uses an algorithm based on the Shoelace formula. We compute the
     * signed area of the 2D polygon. The +/- sign determines the winding
     * order of the vertices with respect to the center of the polygon.
     */
    static bool polygonCW(const RecPolygon2D& polygon);

    /*!
     * @brief Check if the vertices of a 2D polygon have a CCW (Counter-Clockwise) winding order.
     *
     * This uses an algorithm based on the Shoelace formula. We compute the
     * signed area of the 2D polygon. The +/- sign determines the winding
     * order of the vertices with respect to the center of the polygon.
     */
    static bool polygonCCW(const RecPolygon2D& polygon);

    /*!
     * @brief Check if a set of 2D points (representing the vertices of a 2D polygon)
     * have a CW (Clockwise) winding order.
     *
     * This uses an algorithm based on the Shoelace formula. We compute the
     * signed area of the 2D polygon. The +/- sign determines the winding
     * order of the vertices with respect to the center of the polygon.
     */
    static bool verticesCW(const std::vector<RecPoint2D>& vertices);

    /*!
     * @brief Check if a set of 2D points (representing the vertices of a 2D polygon)
     * have a CCW (Counter-Clockwise) winding order.
     *
     * This uses an algorithm based on the Shoelace formula. We compute the
     * signed area of the 2D polygon. The +/- sign determines the winding
     * order of the vertices with respect to the center of the polygon.
     */
    static bool verticesCCW(const std::vector<RecPoint2D>& vertices);

    //----------------------------------------------------------------------------------------------------------------//

    // Boolean operations on polygons:

    /*!
     * @brief Compute the boolean intersection of two 2D polygons.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeIntersection(const RecPolygon2D& first,
                                    const RecPolygon2D& second,
                                    std::vector<RecPolygon2D>& out);

    /*!
     * @brief Compute the boolean intersection of two 2D polygons.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::list<RecPolygon2D> : One or more polygons.
     */
    static bool computeIntersection(const RecPolygon2D& first,
                                    const RecPolygon2D& second,
                                    std::list<RecPolygon2D>& out);

    /*!
     * @brief Compute the boolean intersection of the set of 2D polygons
     * in 'first' with the set of 2D polygons in 'second'.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeIntersection(const std::vector<RecPolygon2D>& first,
                                    const std::vector<RecPolygon2D>& second,
                                    std::vector<RecPolygon2D>& out);

    /*!
     * @brief Compute the boolean intersection of the set of 2D polygons
     * in 'first' with the set of 2D polygons in 'second'.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::vector<std::vector<RecPolygon2D>> : One or more polygons, each
     * defined by a set of RecPoint2D.
     */
    static bool computeIntersection(const std::vector<RecPoint2D>& first,
                                    const std::vector<RecPoint2D>& second,
                                    std::vector<std::vector<RecPoint2D>>& out);

    /*!
     * @brief Compute the boolean union of two or more 2D polygons.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeUnion(const std::vector<RecPolygon2D>& in, std::vector<RecPolygon2D>& out);

    /*!
     * @brief Compute the boolean union of two 2D polygons.
     *
     * @return boundaries : One or more polygons that have CW winding order.
     * @return holes : One or more polygons that have CCW winding order.
     */
    static bool computeUnion(const RecPolygon2D& first,
                             const RecPolygon2D& second,
                             std::vector<RecPolygon2D>& boundaries,
                             std::vector<RecPolygon2D>& holes);

    /*!
     * @brief Compute the boolean union of two 2D polygons.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeUnion(const RecPolygon2D& first, const RecPolygon2D& second, std::vector<RecPolygon2D>& out);

    /*!
     * @brief Compute the boolean union of two 2D polygons.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::list<RecPolygon2D> : One or more polygons.
     */
    static bool computeUnion(const RecPolygon2D& first, const RecPolygon2D& second, std::list<RecPolygon2D>& out);

    /*!
     * @brief Compute the boolean union of the set of 2D polygons
     * in 'first' with the set of 2D polygons in 'second'.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeUnion(const std::vector<RecPolygon2D>& first,
                             const std::vector<RecPolygon2D>& second,
                             std::vector<RecPolygon2D>& out);

    /*!
     * @brief (Experimental) Compute the union of the set of individual vertices
     * in two or more 2D polygons.
     *
     * @return RecPolygon2D : One 2D polygon, consisting of all the vertices
     * that have been re-sorted.
     */
    static void computeVerticesUnion(const std::vector<RecPolygon2D>& in, RecPolygon2D& out);

    /*!
     * @brief Compute the boolean difference of two 2D polygons. Or in other words,
     * subtract the area of the second polygon from the area of the first.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::list<RecPolygon2D> : One or more polygons.
     */
    static bool computeDifference(const RecPolygon2D& first, const RecPolygon2D& second, std::list<RecPolygon2D>& out);

    /*!
     * @brief Compute the boolean difference of two 2D polygons. Or in other words,
     * subtract the area of the second polygon from the area of the first.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeDifference(const RecPolygon2D& first,
                                  const RecPolygon2D& second,
                                  std::vector<RecPolygon2D>& out);

    /*!
     * @brief Compute the boolean difference of the set of 2D polygons in 'first' and
     * the set of 2D polygons in 'second'.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeDifference(const std::vector<RecPolygon2D>& first,
                                  const std::vector<RecPolygon2D>& second,
                                  std::vector<RecPolygon2D>& out);

    /*!
     * @brief Compute the boolean XOR (Exclusive OR) of two 2D polygons.
     *
     * Note: The output polygon(s) will have CW order if they represent a boundary,
     * or CCW order if they represent a hole. You may like to use the convenience
     * function that returns these both separately.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeXOR(const RecPolygon2D& first, const RecPolygon2D& second, std::vector<RecPolygon2D>& out);

    //----------------------------------------------------------------------------------------------------------------//

    // Functions to offset a polygon:
    // (also called buffering or inflation/deflation)

    /*!
     * @brief Offset the boundary of a 2D polygon.
     *
     * This function can produce either mitered (45 degree) corners or it can
     * produce pointed sharp corners, which is the visually-pleasing result.
     *
     * @param offset : A positive or negative value in meters.
     * @param miter_limit : The default value of 2.0 means the output polygon will
     * have 45 degree corners, whereas a value of 3.0 will create pointed sharp corners.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeOffsetMiter(const RecPolygon2D& polygon,
                                   const double offset,
                                   const double miter_limit,
                                   std::vector<RecPolygon2D>& out);

    /*!
     * @brief Offset the boundary of a 2D polygon.
     *
     * This function can produce either mitered (45 degree) corners or it can
     * produce pointed sharp corners, which is the visually-pleasing result.
     *
     * @param offset : A positive or negative value in meters.
     * @param miter_limit : The default value of 2.0 means the output polygon will
     * have 45 degree corners, whereas a value of 3.0 will create pointed sharp corners.
     *
     * @return std::vector<std::vector<RecPoint2D>> : One or more polygons, each
     * defined by a set of RecPoint2D.
     */
    static bool computeOffsetMiter(const std::vector<RecPoint2D>& vertices,
                                   const double offset,
                                   const double miter_limit,
                                   std::vector<std::vector<RecPoint2D>>& out);

    /*!
     * @brief Offset the boundary of a 2D polygon.
     *
     * Despite the name, this function creates mitered (45 degree) corners.
     *
     * @param offset : A positive or negative value in meters.
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */
    static bool computeOffsetSquare(const RecPolygon2D& polygon, const double offset, std::vector<RecPolygon2D>& out);

    /*!
     * @brief Offset the boundary of a 2D polygon.
     *
     * This produces rounded corners which is the mathematically correct
     * result of offsetting a 2D polygon by a fixed distance.
     *
     * @param offset : A positive or negative value in meters.
     * @param arc_tolerance : A value of 0.01 will make the rounded corners
     * accurate to 1cm. The default tolerance is 0.25 (25cm).
     *
     * @return std::vector<RecPolygon2D> : One or more polygons.
     */

    static bool computeOffsetRound(const RecPolygon2D& polygon,
                                   const double offset,
                                   const double arc_tolerance,
                                   std::vector<RecPolygon2D>& out);

    //----------------------------------------------------------------------------------------------------------------//

protected:
    // Functions to convert between RecGeometry and Clipper types:

    // Convert a 2D polygon (with vertices defined by a set of RecPoint2D) to a Clipper Path.
    static void RecPoint2DToClipperPath(const std::vector<RecPoint2D>& vertices, ClipperLib::Path& out);

    // Convert from a single RecPolygon2D to the Clipper Path type.
    static void RecPolygon2DToClipperPath(const RecPolygon2D& recPoly, ClipperLib::Path& out);

    // Convert a set of RecPolygon2D to a set of Clipper Paths.
    static void RecPolygon2DVectorToClipperPaths(const std::vector<RecPolygon2D>& in, ClipperLib::Paths& out);

    // Convert a Clipper Path to a 2D polygon with vertices defined by a set of RecPoint2D.
    static void ClipperPathToRecPoint2D(const ClipperLib::Path& path, std::vector<RecPoint2D>& vertices_out);

    static void ClipperPathsToRecPoint2DVector(const ClipperLib::Paths& paths,
                                               std::vector<std::vector<RecPoint2D>>& out);

    // Convert a Clipper Path type to a 2D polygon.
    static void ClipperPathToRecPolygon2D(const ClipperLib::Path& clipperPath, RecPolygon2D& out);

    // Convert a set of Clipper Paths type to a set of 2D polygons.
    static bool ClipperPathsToRecPolygon2DVector(const ClipperLib::Paths& clipperPaths, std::vector<RecPolygon2D>& out);

    //----------------------------------------------------------------------------------------------------------------//

    // Functions to perform Clipper operations:

    static bool clipperPolygonClip(const ClipperLib::ClipType operation,
                                   const std::vector<RecPoint2D>& first,
                                   const std::vector<RecPoint2D>& second,
                                   std::vector<std::vector<RecPoint2D>>& out);

    static bool clipperPolygonClip(const ClipperLib::ClipType operation,
                                   const RecPolygon2D& first,
                                   const RecPolygon2D& second,
                                   std::list<RecPolygon2D>& out);

    static bool clipperPolygonClip(const ClipperLib::ClipType operation,
                                   const RecPolygon2D& first,
                                   const RecPolygon2D& second,
                                   std::vector<RecPolygon2D>& out);

    static bool clipperPolygonClip(const ClipperLib::ClipType operation,
                                   const RecPolygon2D& first,
                                   const RecPolygon2D& second,
                                   std::vector<RecPolygon2D>& boundaries,
                                   std::vector<RecPolygon2D>& holes);

    static bool clipperPolygonClip(const ClipperLib::ClipType operation,
                                   const std::vector<RecPolygon2D>& in,
                                   std::vector<RecPolygon2D>& out);

    static bool clipperPolygonClip(const ClipperLib::ClipType operation,
                                   const std::vector<RecPolygon2D>& first,
                                   const std::vector<RecPolygon2D>& second,
                                   std::vector<RecPolygon2D>& out);

    static bool clipperPolygonOffset(const std::vector<RecPoint2D>& vertices_in,
                                     const ClipperLib::JoinType jointype,
                                     const double offset,
                                     const double parameter, // miterLimit or arcTolerance
                                     std::vector<std::vector<RecPoint2D>>& polygons_out);

    //----------------------------------------------------------------------------------------------------------------//
};

#endif // _RECCLIPPER_H_
