
/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file recClipper_helper.h
 * @product Common / RecGeometry
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated May 28, 2018
 * @brief Helper functions for RecClipper.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 ***********************************************************************************************************************
 */

#ifndef RECCLIPPER_HELPER_H
#define RECCLIPPER_HELPER_H

#include <algorithm>
#include <cmath> // atan2
#include <limits>

#include <recgeometry/recGeometry.h>

/*!
 * @brief Calculate the bearing angle of a 2D point w.r.t. another point and the x-axis.
 */
static double getAngleFromPoint(const RecPoint2D& origin_point, const RecPoint2D& point)
{
    return std::atan2((point.y - origin_point.y), (point.x - origin_point.x));
}

/*!
 * @brief Compute the visual center of a 2D polygon.
 * This works even if the vertices are not sorted.
 */
static const RecPoint2D computePolygonVisualCenter(const std::vector<RecPoint2D>& vertices)
{
    double min_x = std::numeric_limits<double>::max();
    double min_y = std::numeric_limits<double>::max();
    double max_x = -std::numeric_limits<double>::max();
    double max_y = -std::numeric_limits<double>::max();

    for (const auto& vertex : vertices) {
        if (vertex.x < min_x)
            min_x = vertex.x;
        if (vertex.x > max_x)
            max_x = vertex.x;
        if (vertex.y < min_y)
            min_y = vertex.y;
        if (vertex.y > max_y)
            max_y = vertex.y;
    }

    return RecPoint2D((min_x + max_x) / 2.0, (min_y + max_y) / 2.0);
}

/*!
 * @brief Sort the vertices of a polygon with CW or CCW winding order.
 *
 * This method is faster than the method in RecPolygon2D, but its also designed to
 * work specifically with RecClipper where we use a flipped y-axis to suit the
 * coordinate frame used in OttoView.
 */
static const std::vector<RecPoint2D> sortVertices(const std::vector<RecPoint2D>& vertices, const bool clockwise)
{
    const RecPoint2D center = computePolygonVisualCenter(vertices);

    std::vector<double> angles(vertices.size());
    for (std::vector<RecPoint2D>::size_type i = 0; i < vertices.size(); ++i) {
        angles.at(i) = getAngleFromPoint(center, vertices.at(i));
    }

    std::vector<RecPoint2D> output_vertices = vertices;

    for (std::vector<double>::size_type i = 0; i < angles.size(); i++) {
        double minAngleValue = angles.at(i);
        std::vector<double>::size_type minAngleIndex = i;

        // Find the min angle from i to the end of vector
        for (std::vector<double>::size_type j = i + 1; j < angles.size(); j++) {
            if (clockwise && (angles.at(j) > minAngleValue)) {
                minAngleValue = angles.at(j);
                minAngleIndex = j;
            } else if (!clockwise && (angles.at(j) < minAngleValue)) {
                minAngleValue = angles.at(j);
                minAngleIndex = j;
            }
        }

        // Swap the data from minAngleIndex to i
        if (minAngleIndex != i) {
            double temp = angles.at(i);
            angles.at(i) = angles.at(minAngleIndex);
            angles.at(minAngleIndex) = temp;

            RecPoint2D tempPoint = output_vertices.at(i);
            output_vertices.at(i) = output_vertices.at(minAngleIndex);
            output_vertices.at(minAngleIndex) = tempPoint;
        }
    }

    return output_vertices;
}

#endif // RECCLIPPER_HELPER_H
