/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file recGeometry.inline.h
 * @product <Subsystem name>
 * @version <Release Number>
 * @released 01/31/2019
 * @updated <date>
 * @brief This file contains inlined RecGeometry functions
 *
 * @sa Optional Reference
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

/*****************************************************************************
 * $Revision: 1.1 $
 *     $Date: 2005/02/07 17:59:21 $
 *   $Author: ekratzer $
 *                         ******************************
 *                         *      Copyright (C) 2002    *
 *                         * Carnegie Mellon University *
 *                         *      All Rights Reserved   *
 *                         ******************************
 *
 *     PROJECT:	Blitz
 *        FILE:	recVector.cpp
 * DESCRIPTION: contains implementation of the RecVector* classes
 *     CREATED: 2003/02/07
 *
 * HISTORY:
 *
 * $Log: recVector2D.cpp,v $
 * Revision 1.1  2005/02/07 17:59:21  ekratzer
 * Adding "recGeometry" to repository
 *
 * Revision 1.1  2004/06/14 17:09:34  colmstea
 * Adding recGeometry.
 *
 * Revision 1.3  2003/02/18 17:13:34  bode
 * Added access featurs to recRadians and recVector classes.  Added comments
 * to recPose classes.  Fixed serializable vector3D bug.
 *
 * Revision 1.2  2003/02/14 22:39:48  bode
 * Added additional comments
 *
 * Revision 1.1  2003/02/14 20:30:10  bode
 * Limited Initial Release.  Only classes RecRadians, RecPoint3D,
 * RecPoint3D, RecVector2D, RecVector3D, RecTransform3D and RecPose3D have
 * been well tested.
 *
 *
 *****************************************************************************/

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*****************************************************************************
 * RecPoint2D PUBLIC MEMBER FUNCTIONS
 *****************************************************************************/

/******************************************************************************
 *      METHOD: RecPoint2D
 * DESCRIPTION: two argument constructor, sets x and y to passed x and y values
 *****************************************************************************/
inline RecPoint2D::RecPoint2D(double inx, double iny)
    : x(inx)
    , y(iny)
{}

/******************************************************************************
 *      METHOD: RecPoint2D
 * DESCRIPTION: copy constructor, copys the fields of the passed point
 *****************************************************************************/
inline RecPoint2D::RecPoint2D(const RecPoint2D& pt2D)
    : x(pt2D.x)
    , y(pt2D.y)
{}

/******************************************************************************
 *      METHOD: +=operator
 * DESCRIPTION: increments the fields of the point by the fields of the
 *              passed point
 *****************************************************************************/
inline void RecPoint2D::operator+=(const RecPoint2D& b)
{
    x += b.x;
    y += b.y;
}

inline void RecPoint2D::operator+=(const RecVector2D& v)
{
    x += v.x;
    y += v.y;
}

/******************************************************************************
 *      METHOD: -=operator
 * DESCRIPTION: decrements the fields of the point by the fields of the
 *              passed point
 *****************************************************************************/
inline void RecPoint2D::operator-=(const RecPoint2D& b)
{
    x -= b.x;
    y -= b.y;
}

inline void RecPoint2D::operator-=(const RecVector2D& v)
{
    x -= v.x;
    y -= v.y;
}

/******************************************************************************
 *      METHOD: *=operator
 * DESCRIPTION: multiplies the fields of the point by the passed double.
 *****************************************************************************/
inline void RecPoint2D::operator*=(double b)
{
    x *= b;
    y *= b;
}

/******************************************************************************
 *      METHOD: /=operator
 * DESCRIPTION: divides the fields of the point by the passed double.  Does
 *              nothing if the passed value is zero
 *****************************************************************************/
inline void RecPoint2D::operator/=(double b)
{
    if (b == 0.0) {
        std::cerr << "Attempting to divide by zero in 2D point division." << std::endl;
        return;
    }
    x /= b;
    y /= b;
}

/******************************************************************************
 *      METHOD: =operator
 * DESCRIPTION: sets the fields of the point to those of the passed point.
 *****************************************************************************/
inline void RecPoint2D::operator=(const RecPoint2D& b)
{
    x = b.x;
    y = b.y;
}

/******************************************************************************
 *      METHOD: +operator
 * DESCRIPTION: adds the fields of the passed points and returns the result
 *              in a new point structure.
 *****************************************************************************/
inline RecPoint2D operator+(const RecPoint2D& a, const RecPoint2D& b)
{
    RecPoint2D c(a);
    c += b;
    return c;
}

inline RecPoint2D operator+(const RecPoint2D& a, const RecVector2D& v)
{
    RecPoint2D c(a);
    c += v;
    return c;
}

/******************************************************************************
 *      METHOD: -operator
 * DESCRIPTION: subtracts the fields of the second point from the fields of the
 *              first point and returns the result in a new point structure.
 *****************************************************************************/
inline RecPoint2D operator-(const RecPoint2D& a, const RecPoint2D& b)
{
    RecPoint2D c(a);
    c -= b;
    return c;
}

inline RecPoint2D operator-(const RecPoint2D& a, const RecVector2D& v)
{
    RecPoint2D c(a);
    c -= v;
    return c;
}

/******************************************************************************
 *      METHOD: *operator
 * DESCRIPTION: multiplies the fields of the passed point by the passed double
 *              and returns the result in a new point structure.
 *****************************************************************************/
inline RecPoint2D operator*(const RecPoint2D& a, double b)
{
    RecPoint2D c(a);
    c *= b;
    return c;
}

/******************************************************************************
 *      METHOD: *operator
 * DESCRIPTION: multiplies the fields of the passed point by the passed double
 *              and returns the result in a new point structure.
 *****************************************************************************/
inline RecPoint2D operator*(double b, const RecPoint2D& a)
{
    return a * b;
}

/******************************************************************************
 *      METHOD: /operator
 * DESCRIPTION: divides the fields of the passed point by the passed double and
 *              returns the result in a new point structure.  Returns a copy
 *              of the passed point if the passed double is zero.
 *****************************************************************************/
inline RecPoint2D operator/(const RecPoint2D& a, double b)
{
    RecPoint2D c(a);
    c /= b;
    return c;
}

/******************************************************************************
 *      METHOD: ==operator
 * DESCRIPTION: compares two points and returns true if all the fields of the
 *              first point are within the tolerance of the second point
 *****************************************************************************/
inline bool operator==(const RecPoint2D& a, const RecPoint2D& b)
{
    return ((fabs(a.x - b.x) < RecPoint2D::tolerance) && (fabs(a.y - b.y) < RecPoint2D::tolerance));
}

/******************************************************************************
 *      METHOD: !=operator
 * DESCRIPTION: compares two points and returns true if all the fields of the
 *              first point are not within the tolerance of the second point
 *****************************************************************************/
inline bool operator!=(const RecPoint2D& a, const RecPoint2D& b)
{
    return !(a == b);
}

/******************************************************************************
 *      METHOD: distanceSq
 * DESCRIPTION: returns the squared distance between two points
 *****************************************************************************/
inline double RecPoint2D::distanceSq(const RecPoint2D& a) const
{
    return ((x - a.x) * (x - a.x) + (y - a.y) * (y - a.y));
}

/******************************************************************************
 *      METHOD: distance
 * DESCRIPTION: returns the distance between two points
 *****************************************************************************/
inline double RecPoint2D::distance(const RecPoint2D& a) const
{
    return (sqrt(distanceSq(a)));
}

/******************************************************************************
 *      METHOD: operator<<
 * DESCRIPTION: puts the point with nice formatting to the passed output stream
 *****************************************************************************/
inline std::ostream& operator<<(std::ostream& out, const RecPoint2D& a)
{
    return out << std::fixed << "( " << a.x << ", " << a.y << " )";
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*****************************************************************************
 * RecVector2D PUBLIC MEMBER FUNCTIONS
 *****************************************************************************/

/******************************************************************************
 *      METHOD: RecVector2D
 * DESCRIPTION: two argument constructor, sets x and y to passed x and y values
 *****************************************************************************/
inline RecVector2D::RecVector2D(double inx, double iny)
    : x(inx)
    , y(iny)
{}

/******************************************************************************
 *      METHOD: RecPoint2D
 * DESCRIPTION: copy constructor, copys the fields of the passed vector
 *****************************************************************************/
inline RecVector2D::RecVector2D(const RecVector2D& v2D)
    : x(v2D.x)
    , y(v2D.y)
{}
inline RecVector2D::RecVector2D(const RecPoint2D& p2D)
    : x(p2D.x)
    , y(p2D.y)
{}

inline void RecVector2D::operator+=(const RecVector2D& b)
{
    x += b.x;
    y += b.y;
}

inline void RecVector2D::operator-=(const RecVector2D& b)
{
    x -= b.x;
    y -= b.y;
}

inline void RecVector2D::operator*=(double b)
{
    x *= b;
    y *= b;
}

inline void RecVector2D::operator/=(double b)
{
    x /= b;
    y /= b;
}

/******************************************************************************
 *      METHOD: =operator
 * DESCRIPTION: sets the fields of the vector to those of the passed vector.
 *****************************************************************************/
inline void RecVector2D::operator=(const RecVector2D& b)
{
    x = b.x;
    y = b.y;
}

/******************************************************************************
 *      METHOD: +operator
 * DESCRIPTION: adds the fields of the passed vectors and returns the result
 *              in a new vector structure.
 *****************************************************************************/
inline RecVector2D operator+(const RecVector2D& a, const RecVector2D& b)
{
    RecVector2D c(a);
    c += b;
    return c;
}

/******************************************************************************
 *      METHOD: -operator
 * DESCRIPTION: subtracts the fields of the second vector from the fields of
 *              the first vector and returns the result in a new vector.
 *****************************************************************************/
inline RecVector2D operator-(const RecVector2D& a, const RecVector2D& b)
{
    RecVector2D c(a);
    c -= b;
    return c;
}

/******************************************************************************
 *      METHOD: *operator
 * DESCRIPTION: multiplies the fields of the passed vector by the passed double
 *              and returns the result in a new vector structure.
 *****************************************************************************/
inline RecVector2D operator*(const RecVector2D& a, double b)
{
    RecVector2D c(a);
    c *= b;
    return c;
}

/******************************************************************************
 *      METHOD: *operator
 * DESCRIPTION: multiplies the fields of the passed vector by the passed double
 *              and returns the result in a new vector structure.
 *****************************************************************************/
inline RecVector2D operator*(double b, const RecVector2D& a)
{
    return a * b;
}

/******************************************************************************
 *      METHOD: /operator
 * DESCRIPTION: divides the fields of the passed vector by the passed double
 *              and returns the result in a new vector structure.  Returns a
 *              copy of the passed vector if the passed double is zero.
 *****************************************************************************/
inline RecVector2D operator/(const RecVector2D& a, double b)
{
    RecVector2D c(a);
    // if (b!=0)		//MAT 10-5-06 - added divide by zero protection
    //{                 //MAT 10-9-06 - removed divide by zero protection
    c /= b;
    //}
    return c;
}

/******************************************************************************
 *      METHOD: operator*
 * DESCRIPTION: computes and returns the dot product of the two passed vectors
 *****************************************************************************/
inline double operator*(const RecVector2D& a, const RecVector2D& b)
{
    return (a.x * b.x + a.y * b.y);
}

inline bool operator==(const RecVector2D& a, const RecVector2D& b)
{
    return ((fabs(a.x - b.x) < RecPoint2D::tolerance) && (fabs(a.y - b.y) < RecPoint2D::tolerance));
}

inline bool operator!=(const RecVector2D& a, const RecVector2D& b)
{
    return !(a == b);
}
/******************************************************************************
 *      METHOD: dot
 * DESCRIPTION: computes and returns the dot product of this vector with
 *              the passed vector
 *****************************************************************************/
inline double RecVector2D::dot(const RecVector2D& a) const
{
    return ((*this) * a);
}

/******************************************************************************
 *      METHOD: xprod
 * DESCRIPTION: computes and returns the cross product of this vector with
 *              the passed vector
 *****************************************************************************/
inline double RecVector2D::xprod(const RecVector2D& a) const
{
    return (x * a.y) - (y * a.x);
}

/******************************************************************************
 *      METHOD: lengthSq
 * DESCRIPTION: computes and returns the square of the length of the vector
 *****************************************************************************/
inline double RecVector2D::lengthSq(void) const
{
    return (x * x + y * y);
}

/******************************************************************************
 *      METHOD: length
 * DESCRIPTION: computes and returns the length of the vector
 *****************************************************************************/
inline double RecVector2D::length(void) const
{
    return (sqrt(lengthSq()));
}

/******************************************************************************
 *      METHOD: normalize
 * DESCRIPTION: makes the vector a unit vector that has the same direction
 *              as the old vector
 *****************************************************************************/
inline void RecVector2D::normalize(void)
{
    // double newlength = length();
    // if (newlength!=0)		//MAT 10-5-06 - added divide by zero protection
    //{
    //	(*this) /= newlength;
    //}

    (*this) /= length(); // MAT 10-9-06 - removed divide by zero protection
}

inline RecVector2D RecVector2D::normal(void) const
{
    return RecVector2D(-y, x);
}

inline std::ostream& operator<<(std::ostream& out, const RecVector2D& a)
{
    return out << "( " << a.x << ", " << a.y << " )";
}

inline RecVector2D RecVector2D::getUnitVector(const RecRadians& radians)
{
    return RecVector2D(cos(radians), sin(radians));
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*****************************************************************************
 * RecPose2D PUBLIC MEMBER FUNCTIONS
 *****************************************************************************/

/******************************************************************************
 *      METHOD: RecPose2D
 * DESCRIPTION: three argument constructor, sets x, y and rotZ to passed values
 *****************************************************************************/
inline RecPose2D::RecPose2D(double ix, double iy, RecRadians rZ)
    : x(ix)
    , y(iy)
    , rotZ(rZ)
{}

/******************************************************************************
 *      METHOD: RecPose2D
 * DESCRIPTION: copy constructor, copys the fields of the passed pose
 *****************************************************************************/
inline RecPose2D::RecPose2D(const RecPose2D& b)
    : x(b.x)
    , y(b.y)
    , rotZ(b.rotZ)
{}

/******************************************************************************
 *      METHOD: RecPose2D
 * DESCRIPTION: transform constructor, converts a 2D Transform to a 2D pose
 *****************************************************************************/
inline RecPose2D::RecPose2D(const RecTransform2D& trans)
    : x(trans.matrixPP[0][2])
    , y(trans.matrixPP[1][2])
    , rotZ(atan2(trans.matrixPP[1][0], trans.matrixPP[1][1]))
{}

/******************************************************************************
 *      METHOD: inverse
 * DESCRIPTION: computes and returns the inverse of a pose by converting the
 *              pose to a transform, taking the inverse of this transform and
 *              converting the inverted transform into a 2D pose.
 *****************************************************************************/
inline RecPose2D RecPose2D::inverse(void) const
{
    RecTransform2D trans(*this);
    RecPose2D poseInv(trans.inverse());
    return poseInv;
}

inline RecPoint2D RecPose2D::getPoint() const
{
    return RecPoint2D(x, y);
}

/******************************************************************************
 *      METHOD: operator+=
 * DESCRIPTION: adds the passed differential pose to the pose
 *****************************************************************************/
inline void RecPose2D::operator+=(const RecDifferentialPose2D& b)
{
    x += b.x;
    y += b.y;
    rotZ += b.rotZ;
}

/******************************************************************************
 *      METHOD: operator-=
 * DESCRIPTION: subtracts the passed differential pose from the pose
 *****************************************************************************/
inline void RecPose2D::operator-=(const RecDifferentialPose2D& b)
{
    x -= b.x;
    y -= b.y;
    rotZ -= b.rotZ;
}

/******************************************************************************
 *      METHOD: operator=
 * DESCRIPTION: assigns one pose to another by copying all fields
 *****************************************************************************/
inline void RecPose2D::operator=(const RecPose2D& b)
{
    x = b.x;
    y = b.y;
    rotZ = b.rotZ;
}

/******************************************************************************
 *      METHOD: operator*
 * DESCRIPTION: returns the pose that is the concatenation of the two passed
 *              poses
 *****************************************************************************/
inline RecPose2D operator*(const RecPose2D& b, const RecPose2D& c)
{
    RecTransform2D btrans(b);
    RecTransform2D ctrans(c);
    RecTransform2D atrans = btrans * ctrans;
    RecPose2D a(atrans);
    return a;
}

/******************************************************************************
 *      METHOD: operator+
 * DESCRIPTION: adds the passed differential pose to the passed pose and
 *              returns the result
 *****************************************************************************/
inline RecPose2D operator+(const RecPose2D& a, const RecDifferentialPose2D& b)
{
    RecPose2D c(a);
    c += b;
    return c;
}

/******************************************************************************
 *      METHOD: operator+
 * DESCRIPTION: adds the passed differential pose to the passed pose and
 *              returns the result
 *****************************************************************************/
inline RecPose2D operator+(const RecDifferentialPose2D& a, const RecPose2D& b)
{
    return b + a;
}

/******************************************************************************
 *      METHOD: operator-
 * DESCRIPTION: subtracts the passed differential pose from the passed pose and
 *              returns the result
 *****************************************************************************/
inline RecPose2D operator-(const RecPose2D& a, const RecDifferentialPose2D& b)
{
    RecPose2D c(a);
    c -= b;
    return c;
}

/******************************************************************************
 *      METHOD: operator==
 * DESCRIPTION: converts the two poses to transforms and returns the result
 *              of transform comparison
 *****************************************************************************/
inline bool operator==(const RecPose2D& a, const RecPose2D& b)
{
    RecTransform2D at(a);
    RecTransform2D bt(b);
    return (at == bt);
}

/******************************************************************************
 *      METHOD: operator!=
 * DESCRIPTION: returns the complement of the == operator
 *****************************************************************************/
inline bool operator!=(const RecPose2D& a, const RecPose2D& b)
{
    return (!(a == b));
}

/******************************************************************************
 *      METHOD: operator<<
 * DESCRIPTION: puts the pose with nice formatting to the passed output stream
 *****************************************************************************/
inline std::ostream& operator<<(std::ostream& out, const RecPose2D& a)
{
    return out << "(x, y, w) = ( " << a.x << ", " << a.y << ", " << a.rotZ << " )";
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*****************************************************************************
 * RecDifferentialPose2D PUBLIC MEMBER FUNCTIONS
 *****************************************************************************/

/******************************************************************************
 *      METHOD: RecDifferentialPose2D
 * DESCRIPTION: three argument constructor, sets x, y and rotZ to passed values
 *****************************************************************************/
inline RecDifferentialPose2D::RecDifferentialPose2D(double inx, double iny, RecRadians inr)
    : RecPose2D(inx, iny, inr)
{}
/******************************************************************************
 *      METHOD: RecDifferentialPose2D
 * DESCRIPTION: copy constructor, copys the fields of the passed pose
 *****************************************************************************/
inline RecDifferentialPose2D::RecDifferentialPose2D(const RecDifferentialPose2D& b)
    : RecPose2D(b)
{}
