/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**
 * @file recSerialize.h
 * @author: Chris Urmson (curmson@ri.cmu.edu)
 * @date: 09/10/2006
 *
 * @attention Copyright (c) 2006
 * @attention Carnegie Mellon University
 * @attention All rights reserved.
 */
#ifndef _RECSERIALIZE_H_
#define _RECSERIALIZE_H_

#include "recgeometry/recGeometry.h"
#include <cereal/access.hpp>
#include <cereal/cereal.hpp>
#include <cereal/types/polymorphic.hpp>
#include <cereal/types/vector.hpp>

namespace cereal {

/**
 * @brief functions to serialize point2d
 */
template<class Archive>
void serialize(Archive& ar, RecPoint2D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.x), CEREAL_NVP(r.y));
}

/**
 * @brief functions to serialize point3d
 */
template<class Archive>
void serialize(Archive& ar, RecPoint3D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.x), CEREAL_NVP(r.y), CEREAL_NVP(r.z));
}

/**
 * @brief functions to serialize vector2d
 */
template<class Archive>
void serialize(Archive& ar, RecVector2D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.x), CEREAL_NVP(r.y));
}

/**
 * @brief functions to serialize vector3d
 */
template<class Archive>
void serialize(Archive& ar, RecVector3D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.x), CEREAL_NVP(r.y), CEREAL_NVP(r.z));
}

/**
 * @brief functions to serialize pose2d
 */
template<class Archive>
void serialize(Archive& ar, RecPose2D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.x), CEREAL_NVP(r.y), CEREAL_NVP(r.rotZ));
}

/**
 * @brief functions to serialize differentialpose2d
 */
template<class Archive>
void serialize(Archive& ar, RecDifferentialPose2D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.x), CEREAL_NVP(r.y), CEREAL_NVP(r.rotZ));
}

/**
 * @brief functions to serialize pose3d
 */
template<class Archive>
void serialize(Archive& ar, RecPose3D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.x), CEREAL_NVP(r.y), CEREAL_NVP(r.z));
    ar(CEREAL_NVP(r.rot1), CEREAL_NVP(r.rot2), CEREAL_NVP(r.rot3));
    ar(CEREAL_NVP(r.angleOrder));
}

/**
 * @brief functions to serialize differentialPose3d
 */
template<class Archive>
void serialize(Archive& ar, RecDifferentialPose3D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.x), CEREAL_NVP(r.y), CEREAL_NVP(r.z));
    ar(CEREAL_NVP(r.rot1), CEREAL_NVP(r.rot2), CEREAL_NVP(r.rot3));
}

/**
 * @brief functions to serialize transform2d
 */
template<class Archive>
void serialize(Archive& ar, RecTransform2D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.matrixPP));
}

/**
 * @brief functions to serialize transform3d
 */
template<class Archive>
void serialize(Archive& ar, RecTransform3D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.matrixPP));
}

/**
 * @brief functions to serialize quaternion
 */
template<class Archive>
void serialize(Archive& ar, RecQuaternion& r, unsigned int version)
{
    ar(CEREAL_NVP(r.quat));
}

/**
 * @brief functions to serialize linesegment3d
 */
template<class Archive>
void serialize(Archive& ar, RecLineSegment3D& r, unsigned int version)
{
    ar(CEREAL_NVP(r.p1), CEREAL_NVP(r.p2));
}

template<class Archive>
void serialize(Archive& ar, RecBox2D& r, unsigned int version)
{}

/**
 * @brief functions to serialize axisAlignedBox2D
 */
template<class Archive>
void serialize(Archive& ar, RecAxisAlignedBox2D& r, unsigned int version)
{
    RecPoint2D min, max;
    min = r.getMinPoint();
    max = r.getMaxPoint();
    ar& CEREAL_NVP(min) & CEREAL_NVP(max);
    r = RecAxisAlignedBox2D(min, max);
}

/**
 * @brief functions to serialize axisAlignedBox3D
 */
template<class Archive>
void serialize(Archive& ar, RecAxisAlignedBox3D& r, unsigned int version)
{
    RecPoint3D min, max;
    min = r.getMin();
    max = r.getMax();
    ar& CEREAL_NVP(min) & CEREAL_NVP(max);
    r = RecAxisAlignedBox3D(min, max);
}

} // namespace cereal

CEREAL_REGISTER_TYPE(RecAxisAlignedBox2D)
CEREAL_REGISTER_TYPE(RecAxisAlignedBox3D)

#endif // ifndef _RECSERIALIZE_H_
