
/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file recClipper.h
 * @product Common / RecGeometry
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated May 28, 2018
 * @brief Contains functions for manipulating 2D polygons.
 *
 * Supported functions include: boolean intersection, boolean union, offset (buffering) and sorting vertices.
 * There is detailed documentation on the Wiki:
 * Confluence --> Perception --> Perception Modules --> RecClipper
 *
 * This wraps a 3rd-party library called Clipper version 6.4.2
 * http://www.angusj.com
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 * @section Open Source License List
 * - Boost Software License (http://www.boost.org/LICENSE_1_0.txt)
 *
 ***********************************************************************************************************************
 */

/**********************************************************
  Copyright (c) 2015-    Ottomatika, Inc.
 **********************************************************/

#include <recgeometry/recClipper.h>
#include <recgeometry/recClipper_helper.h> // sortVertices

//--------------------------------------------------------------------------------------------------------------------//

RecPolygon2D RecClipper::sortPolygonCW(const RecPolygon2D& polygon)
{
    std::vector<RecPoint2D> vertices = polygon.getVertices();
    vertices.pop_back(); // The last vertex is a duplicate of the first

    return RecPolygon2D(sortVertices(vertices, true));
}

RecPolygon2D RecClipper::sortPolygonCCW(const RecPolygon2D& polygon)
{
    std::vector<RecPoint2D> vertices = polygon.getVertices();
    vertices.pop_back(); // The last vertex is a duplicate of the first

    return RecPolygon2D(sortVertices(vertices, false));
}

std::vector<RecPoint2D> RecClipper::sortVerticesCW(const std::vector<RecPoint2D>& vertices)
{
    return sortVertices(vertices, true);
}

std::vector<RecPoint2D> RecClipper::sortVerticesCCW(const std::vector<RecPoint2D>& vertices)
{
    return sortVertices(vertices, false);
}

RecPolygon2D RecClipper::reverseVertices(const RecPolygon2D& polygon)
{
    std::vector<RecPoint2D> vertices = polygon.getVertices();
    vertices.pop_back(); // The last vertex is a duplicate of the first
    std::reverse(vertices.begin(), vertices.end());
    return RecPolygon2D(vertices);
}

bool RecClipper::polygonCW(const RecPolygon2D& polygon)
{
    std::vector<RecPoint2D> vertices = polygon.getVertices();
    vertices.pop_back(); // The last vertex is a duplicate of the first

    return verticesCW(vertices);
}

bool RecClipper::polygonCCW(const RecPolygon2D& polygon)
{
    return !polygonCW(polygon);
}

bool RecClipper::verticesCW(const std::vector<RecPoint2D>& vertices)
{
    std::vector<RecPoint2D> temp_vertices = vertices;

    // A 2D polygon has a ClockWise winding-order
    // if it has a positive area.
    // The formula for computing area is the Shoelace formula
    // but we don't divide by 2 because we don't need the
    // actual numerical value.
    //
    // Note: The area formula only works in all cases if you
    // connect the last vertex back the first vertex!
    //
    // More info:
    // http://web.archive.org/web/20170603063247/blog.element84.com/polygon-winding.html

    if (vertices.at(0) == vertices.at(vertices.size() - 1)) {
        temp_vertices.pop_back();
    }

    double sum = 0.0;
    for (std::vector<RecPoint2D>::size_type i = 0; i < temp_vertices.size() - 1; ++i) {
        sum +=
            (temp_vertices.at(i + 1).x - temp_vertices.at(i).x) * (temp_vertices.at(i + 1).y + temp_vertices.at(i).y);
    }
    sum += (temp_vertices.at(0).x - temp_vertices.at(temp_vertices.size() - 1).x) *
           (temp_vertices.at(0).y + temp_vertices.at(temp_vertices.size() - 1).y);

    return (sum > 0.0); // positive = CW
}

bool RecClipper::verticesCCW(const std::vector<RecPoint2D>& vertices)
{
    return !verticesCW(vertices);
}

//--------------------------------------------------------------------------------------------------------------------//

bool RecClipper::computeIntersection(const std::vector<RecPoint2D>& first,
                                     const std::vector<RecPoint2D>& second,
                                     std::vector<std::vector<RecPoint2D>>& out)
{
    return clipperPolygonClip(ClipperLib::ctIntersection, first, second, out);
}

bool RecClipper::computeIntersection(const RecPolygon2D& first,
                                     const RecPolygon2D& second,
                                     std::list<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctIntersection, first, second, out);
}

bool RecClipper::computeIntersection(const RecPolygon2D& first,
                                     const RecPolygon2D& second,
                                     std::vector<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctIntersection, first, second, out);
}

bool RecClipper::computeIntersection(const std::vector<RecPolygon2D>& first,
                                     const std::vector<RecPolygon2D>& second,
                                     std::vector<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctIntersection, first, second, out);
}

bool RecClipper::computeUnion(const RecPolygon2D& first, const RecPolygon2D& second, std::list<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctUnion, first, second, out);
}

bool RecClipper::computeUnion(const RecPolygon2D& first, const RecPolygon2D& second, std::vector<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctUnion, first, second, out);
}

bool RecClipper::computeUnion(const RecPolygon2D& first,
                              const RecPolygon2D& second,
                              std::vector<RecPolygon2D>& boundaries,
                              std::vector<RecPolygon2D>& holes)
{
    return clipperPolygonClip(ClipperLib::ctUnion, first, second, boundaries, holes);
}

bool RecClipper::computeUnion(const std::vector<RecPolygon2D>& in, std::vector<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctUnion, in, out);
}

bool RecClipper::computeUnion(const std::vector<RecPolygon2D>& first,
                              const std::vector<RecPolygon2D>& second,
                              std::vector<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctUnion, first, second, out);
}

void RecClipper::computeVerticesUnion(const std::vector<RecPolygon2D>& in, RecPolygon2D& out)
{
    // Combine the vertices from all polygons
    std::vector<RecPoint2D> vertices;
    for (const auto& polygon : in) {
        // Skip the last vertex, which is a duplicate of the first vertex.
        for (std::vector<RecPoint2D>::size_type i = 0; i < (polygon.getVertices().size() - 1); ++i) {
            vertices.push_back(polygon.getVertices().at(i));
        }
    }

    out = RecPolygon2D(sortVerticesCW(vertices));
}

bool RecClipper::computeDifference(const RecPolygon2D& first, const RecPolygon2D& second, std::list<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctDifference, first, second, out);
}

bool RecClipper::computeDifference(const RecPolygon2D& first,
                                   const RecPolygon2D& second,
                                   std::vector<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctDifference, first, second, out);
}

bool RecClipper::computeDifference(const std::vector<RecPolygon2D>& first,
                                   const std::vector<RecPolygon2D>& second,
                                   std::vector<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctDifference, first, second, out);
}

bool RecClipper::computeXOR(const RecPolygon2D& first, const RecPolygon2D& second, std::vector<RecPolygon2D>& out)
{
    return clipperPolygonClip(ClipperLib::ctXor, first, second, out);
}

//--------------------------------------------------------------------------------------------------------------------//

bool RecClipper::computeOffsetMiter(const RecPolygon2D& polygon,
                                    const double offset,
                                    const double miter_limit,
                                    std::vector<RecPolygon2D>& out)
{
    std::vector<RecPoint2D> vertices = polygon.getVertices();
    vertices.pop_back(); // In RecPolygon2D, the last vertex is a duplicate of the first

    std::vector<std::vector<RecPoint2D>> offset_vertices;
    const bool res = clipperPolygonOffset(vertices, ClipperLib::jtMiter, offset, miter_limit, offset_vertices);

    for (const auto& polygonVertices : offset_vertices) {
        out.emplace_back(polygonVertices);
    }

    return res;
}

bool RecClipper::computeOffsetMiter(const std::vector<RecPoint2D>& vertices,
                                    const double offset,
                                    const double miter_limit,
                                    std::vector<std::vector<RecPoint2D>>& out)
{
    std::vector<std::vector<RecPoint2D>> offset_vertices;
    const bool res = clipperPolygonOffset(vertices, ClipperLib::jtMiter, offset, miter_limit, offset_vertices);

    for (const auto& polygonVertices : offset_vertices) {
        out.push_back(polygonVertices);
    }

    return res;
}

bool RecClipper::computeOffsetSquare(const RecPolygon2D& polygon, const double offset, std::vector<RecPolygon2D>& out)
{
    std::vector<RecPoint2D> vertices = polygon.getVertices();
    vertices.pop_back(); // In RecPolygon2D, the last vertex is a duplicate of the first

    std::vector<std::vector<RecPoint2D>> offset_vertices;
    const bool res = clipperPolygonOffset(vertices,
                                          ClipperLib::jtSquare,
                                          offset,
                                          0, // Not used
                                          offset_vertices);

    for (const auto& polygonVertices : offset_vertices) {
        out.emplace_back(polygonVertices);
    }

    return res;
}

bool RecClipper::computeOffsetRound(const RecPolygon2D& polygon,
                                    const double offset,
                                    const double arc_tolerance,
                                    std::vector<RecPolygon2D>& out)

{
    std::vector<RecPoint2D> vertices = polygon.getVertices();
    vertices.pop_back(); // In RecPolygon2D, the last vertex is a duplicate of the first

    std::vector<std::vector<RecPoint2D>> offset_vertices;
    const bool res = clipperPolygonOffset(vertices, ClipperLib::jtRound, offset, arc_tolerance, offset_vertices);

    for (const auto& polygonVertices : offset_vertices) {
        out.emplace_back(polygonVertices);
    }

    return res;
}

//--------------------------------------------------------------------------------------------------------------------//

void RecClipper::RecPoint2DToClipperPath(const std::vector<RecPoint2D>& vertices, ClipperLib::Path& out)
{
    for (const auto& vertex : vertices) {
        // ToDo: Check size of double Vs. int before casting?

        // Vertices are in UTM co-ordinates like (3993069.36, 666924.84)
        // therefore we multiply by scaling factor first, so we
        // don't lose precision.
        const ClipperLib::cInt x = static_cast<ClipperLib::cInt>(vertex.x * SCALING_FACTOR);

        // Invert y-axis, because Clipper expects a y-down coordinate frame
        const ClipperLib::cInt y = -static_cast<ClipperLib::cInt>(vertex.y * SCALING_FACTOR);

        out << ClipperLib::IntPoint(x, y);
    }
}

void RecClipper::RecPolygon2DToClipperPath(const RecPolygon2D& recPoly, ClipperLib::Path& out)
{
    if (recPoly.getVertices().size() < 3) {
        return;
    }

    // Skip the last vertex, which is a duplicate of the first vertex.
    for (std::vector<RecPoint2D>::size_type i = 0; i < (recPoly.getVertices().size() - 1); ++i) {
        // ToDo: Check size of double Vs. int before casting?

        // Vertices are in UTM co-ordinates like (3993069.36, 666924.84)
        // therefore we multiply by scaling factor first, so we
        // don't lose precision.
        const ClipperLib::cInt x = static_cast<ClipperLib::cInt>(recPoly.getVertices().at(i).x * SCALING_FACTOR);

        // Invert y-axis, because Clipper expects a y-down coordinate frame
        const ClipperLib::cInt y = -static_cast<ClipperLib::cInt>(recPoly.getVertices().at(i).y * SCALING_FACTOR);

        out << ClipperLib::IntPoint(x, y);
    }
}

void RecClipper::RecPolygon2DVectorToClipperPaths(const std::vector<RecPolygon2D>& in, ClipperLib::Paths& out)
{
    for (const auto& polygon : in) {
        ClipperLib::Path path;
        RecPolygon2DToClipperPath(polygon, path);
        out << path;
    }
}

void RecClipper::ClipperPathToRecPoint2D(const ClipperLib::Path& path, std::vector<RecPoint2D>& out)
{
    out.resize(path.size());

    for (std::vector<RecPoint2D>::size_type i = 0; i < path.size(); ++i) {
        out.at(i).x = static_cast<double>(path.at(i).X) / SCALING_FACTOR;

        // Invert y-axis, because Clipper expects a y-down coordinate frame
        out.at(i).y = -static_cast<double>(path.at(i).Y) / SCALING_FACTOR;
    }
}

void RecClipper::ClipperPathsToRecPoint2DVector(const ClipperLib::Paths& paths,
                                                std::vector<std::vector<RecPoint2D>>& out)
{
    for (const auto& path : paths) {
        std::vector<RecPoint2D> vertices;
        ClipperPathToRecPoint2D(path, vertices);
        out.push_back(vertices);
    }
}

void RecClipper::ClipperPathToRecPolygon2D(const ClipperLib::Path& clipperPath, RecPolygon2D& out)
{
    std::vector<RecPoint2D> recVertices;
    recVertices.resize(clipperPath.size());

    for (std::vector<RecPoint2D>::size_type i = 0; i < recVertices.size(); ++i) {
        recVertices.at(i).x = static_cast<double>(clipperPath.at(i).X) / SCALING_FACTOR;

        // Invert y-axis, because Clipper expects a y-down coordinate frame
        recVertices.at(i).y = -static_cast<double>(clipperPath.at(i).Y) / SCALING_FACTOR;
    }

    out = RecPolygon2D(recVertices);
}

bool RecClipper::ClipperPathsToRecPolygon2DVector(const ClipperLib::Paths& clipperPolygon,
                                                  std::vector<RecPolygon2D>& out)
{
    bool changed = false;

    // Return polygons in reverse order
    // Note: ClipperLib::Paths has no reverse_iterator
    for (auto i = static_cast<int64_t>(clipperPolygon.size() - 1); i >= 0; --i) {
        RecPolygon2D tmpPoly;
        ClipperPathToRecPolygon2D(clipperPolygon.at(static_cast<size_t>(i)), tmpPoly);
        out.push_back(std::move(tmpPoly));
        changed = true;
    }

    return changed;
}

//--------------------------------------------------------------------------------------------------------------------//

bool RecClipper::clipperPolygonClip(const ClipperLib::ClipType operation,
                                    const std::vector<RecPoint2D>& first,
                                    const std::vector<RecPoint2D>& second,
                                    std::vector<std::vector<RecPoint2D>>& out)
{
    ClipperLib::Path firstPath;
    RecPoint2DToClipperPath(first, firstPath);

    ClipperLib::Path secondPath;
    RecPoint2DToClipperPath(second, secondPath);

    ClipperLib::Clipper c;
    const bool closed_path = true;
    c.AddPath(firstPath, ClipperLib::ptSubject, closed_path);
    c.AddPath(secondPath, ClipperLib::ptClip, closed_path);

    // pftNonZero = Use the NonZero filling rule for polygons, which means they all
    // need to have the same orientation (vertex winding order), except for holes
    // which have the opposite orientation.
    ClipperLib::Paths output_paths;
    c.Execute(operation, output_paths, ClipperLib::pftNonZero, ClipperLib::pftNonZero);

    ClipperPathsToRecPoint2DVector(output_paths, out);

    if (!out.empty()) {
        return true;
    }

    return false;
}

bool RecClipper::clipperPolygonClip(const ClipperLib::ClipType operation,
                                    const RecPolygon2D& first,
                                    const RecPolygon2D& second,
                                    std::list<RecPolygon2D>& out)
{
    ClipperLib::Path firstPath;
    RecPolygon2DToClipperPath(first, firstPath);

    ClipperLib::Path secondPath;
    RecPolygon2DToClipperPath(second, secondPath);

    ClipperLib::Clipper c;
    const bool closed_path = true;
    c.AddPath(firstPath, ClipperLib::ptSubject, closed_path);
    c.AddPath(secondPath, ClipperLib::ptClip, closed_path);

    // pftNonZero = Use the NonZero filling rule for polygons, which means they all
    // need to have the same orientation (vertex winding order), except for holes
    // which have the opposite orientation.
    ClipperLib::Paths outputPoly;
    c.Execute(operation, outputPoly, ClipperLib::pftNonZero, ClipperLib::pftNonZero);

    std::vector<RecPolygon2D> vOut;
    ClipperPathsToRecPolygon2DVector(outputPoly, vOut);

    bool changed = false;
    if (!vOut.empty()) {
        changed = true;
        out.insert(out.begin(), vOut.begin(), vOut.end());
    }

    return changed;
}

bool RecClipper::clipperPolygonClip(const ClipperLib::ClipType operation,
                                    const RecPolygon2D& first,
                                    const RecPolygon2D& second,
                                    std::vector<RecPolygon2D>& out)
{
    ClipperLib::Path firstPath;
    RecPolygon2DToClipperPath(first, firstPath);

    ClipperLib::Path secondPath;
    RecPolygon2DToClipperPath(second, secondPath);

    ClipperLib::Clipper c;
    const bool closed_path = true;
    c.AddPath(firstPath, ClipperLib::ptSubject, closed_path);
    c.AddPath(secondPath, ClipperLib::ptClip, closed_path);

    // pftNonZero = Use the NonZero filling rule for polygons, which means they all
    // need to have the same orientation (vertex winding order), except for holes
    // which have the opposite orientation.
    ClipperLib::Paths output_paths;
    c.Execute(operation, output_paths, ClipperLib::pftNonZero, ClipperLib::pftNonZero);

    const bool result = ClipperPathsToRecPolygon2DVector(output_paths, out);

    return result;
}

bool RecClipper::clipperPolygonClip(const ClipperLib::ClipType operation,
                                    const RecPolygon2D& first,
                                    const RecPolygon2D& second,
                                    std::vector<RecPolygon2D>& boundaries,
                                    std::vector<RecPolygon2D>& holes)
{
    ClipperLib::Path firstPath;
    RecPolygon2DToClipperPath(first, firstPath);

    ClipperLib::Path secondPath;
    RecPolygon2DToClipperPath(second, secondPath);

    ClipperLib::Clipper c;
    const bool closed_path = true;
    c.AddPath(firstPath, ClipperLib::ptSubject, closed_path);
    c.AddPath(secondPath, ClipperLib::ptClip, closed_path);

    // pftNonZero = Use the NonZero filling rule for polygons, which means they all
    // need to have the same orientation (vertex winding order), except for holes
    // which have the opposite orientation.
    ClipperLib::Paths output_paths;
    c.Execute(operation, output_paths, ClipperLib::pftNonZero, ClipperLib::pftNonZero);

    std::vector<RecPolygon2D> polygons;
    const bool changed = ClipperPathsToRecPolygon2DVector(output_paths, polygons);
    for (const auto& polygon : polygons) {
        if (polygonCW(polygon)) {
            // Polygons with CW winding-order represent boundaries
            boundaries.push_back(polygon);
        } else {
            // Polygons with CCW winding-order represent holes
            holes.push_back(polygon);
        }
    }

    return changed;
}

bool RecClipper::clipperPolygonClip(const ClipperLib::ClipType operation,
                                    const std::vector<RecPolygon2D>& in,
                                    std::vector<RecPolygon2D>& out)
{
    // Multiple 'subject' polygons
    ClipperLib::Paths paths_in;
    RecPolygon2DVectorToClipperPaths(in, paths_in);

    ClipperLib::Clipper c;
    const bool closed_path = true;
    c.AddPaths(paths_in, ClipperLib::ptSubject, closed_path);

    // pftNonZero = Use the NonZero filling rule for polygons, which means they all
    // need to have the same orientation (vertex winding order), except for holes
    ClipperLib::Paths outputPoly;
    c.Execute(operation, outputPoly, ClipperLib::pftNonZero, ClipperLib::pftNonZero);

    const bool changed = ClipperPathsToRecPolygon2DVector(outputPoly, out);

    return changed;
}

bool RecClipper::clipperPolygonClip(const ClipperLib::ClipType operation,
                                    const std::vector<RecPolygon2D>& first,
                                    const std::vector<RecPolygon2D>& second,
                                    std::vector<RecPolygon2D>& out)
{
    // Multiple 'subject' polygons
    ClipperLib::Paths firstPath;
    RecPolygon2DVectorToClipperPaths(first, firstPath);

    // Multiple 'clipping' polygons
    ClipperLib::Paths secondPath;
    RecPolygon2DVectorToClipperPaths(second, secondPath);

    ClipperLib::Clipper c;
    const bool closed_path = true;
    c.AddPaths(firstPath, ClipperLib::ptSubject, closed_path);
    c.AddPaths(secondPath, ClipperLib::ptClip, closed_path);

    // pftNonZero = Use the NonZero filling rule for polygons, which means they all
    // need to have the same orientation (vertex winding order), except for holes
    ClipperLib::Paths outputPoly;
    c.Execute(operation, outputPoly, ClipperLib::pftNonZero, ClipperLib::pftNonZero);

    const bool changed = ClipperPathsToRecPolygon2DVector(outputPoly, out);

    return changed;
}

// This has only been tested with one input polygon, so the interface reflects that.
bool RecClipper::clipperPolygonOffset(const std::vector<RecPoint2D>& vertices_in,
                                      const ClipperLib::JoinType jointype,
                                      const double offset,
                                      const double parameter, // miterLimit or arcTolerance
                                      std::vector<std::vector<RecPoint2D>>& out)
{
    ClipperLib::Path path;
    RecPoint2DToClipperPath(vertices_in, path);

    double miterLimit = 2.0; // Default value, means 2 x delta. Also 2 is the minimum limit.
    if (jointype == ClipperLib::jtMiter) {
        miterLimit = parameter;
    }

    double arcTolerance = 0.25;
    if (jointype == ClipperLib::jtRound) {
        arcTolerance = parameter;
    }

    ClipperLib::ClipperOffset co(miterLimit * SCALING_FACTOR, arcTolerance * SCALING_FACTOR);
    co.AddPath(path,
               // JoinType: jtSquare, jtRound, jtMiter
               jointype,
               // EndType: etClosedPolygon (the path vertices define a closed polygon)
               //          etClosedLine (the path is a closed loop)
               //          etOpenButt, etOpenSquare, etOpenRound (the path is a line)
               ClipperLib::EndType::etClosedPolygon);
    ClipperLib::Paths offset_paths;
    co.Execute(offset_paths, offset * SCALING_FACTOR);

    ClipperPathsToRecPoint2DVector(offset_paths, out);

    if (!out.empty()) {
        return true;
    }

    return false;
}

//--------------------------------------------------------------------------------------------------------------------//
