/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**
 * @file RecKDTree2DUnitTest.cc
 * @author: Christopher Baker (tallbaker@cmu.edu)
 * @date:
 *
 * @attention Copyright (c) 2007
 * @attention Carnegie Mellon University
 * @attention All rights reserved.
 */

#include <gtest/gtest.h>

#include <TimeStamp/Stopwatch.h>
#include <TimeStamp/TimeStamp.h>
#include <recgeometry/RecKDTree2D.h>

using namespace ::testing;
using namespace std;

typedef std::vector<std::pair<RecPoint2D, double>> RawElementVector;

#define TIME(SOMETHING)                                                                                                \
    {                                                                                                                  \
        ottotime::Stopwatch _timer;                                                                                    \
        SOMETHING;                                                                                                     \
        printf(" - %s: %.3lfs\n", #SOMETHING, _timer.getElapsedTime().totalSecondsDouble());                           \
    }

// these functions should not do dumb things if no tree is built
TEST(RecGeometryRecKDTree2D, nullProtectionTest)
{
    RecKDTree2D<double> doubleTree;
    double dd;
    RecKDTree2D<double>::PayloadVector payloadVector;

    RecPoint2D queryPoint(0, 0);

    bool getClosestElement = doubleTree.getClosestElement(queryPoint, dd);
    EXPECT_FALSE(getClosestElement);
    bool getElementsWithin = doubleTree.getElementsWithin(queryPoint, 1.0, payloadVector);
    EXPECT_FALSE(getElementsWithin);
    EXPECT_EQ(payloadVector.size(), 0);
}

// build a simple tree and run some simple tests
TEST(RecGeometryRecKDTree2D, DISABLED_simpleConstructionTest)
{
    const double interval = 1.0;
    const int count = 15;
    RecKDTree2D<double> doubleTree;

    // make a tree
    doubleTree.reserve(count);
    for (int ii = 0; ii < count; ++ii) {
        const double dd = interval * ii;
        const double expectedValue = dd * dd;
        bool addElement = doubleTree.addElement(RecPoint2D(-dd, dd), expectedValue);
        EXPECT_TRUE(addElement);
    }

    bool buildTree = doubleTree.buildTree();
    EXPECT_TRUE(buildTree);

    // query the tree for specific points
    for (int ii = 0; ii < count; ++ii) {
        const double dd = interval * ii;
        const double expectedValue = dd * dd;
        double treeValue = 0;
        bool getClosestElement = doubleTree.getClosestElement(RecPoint2D(-dd, dd), treeValue);
        EXPECT_TRUE(getClosestElement);
        EXPECT_EQ(expectedValue, treeValue);
    }

    // query the tree for groups small groups of points
    for (int ii = 1; ii < count - 1; ++ii) {
        std::vector<double> expectedValues;
        for (int jj = ii - 1; jj <= ii + 1; ++jj) {
            const double dd = interval * jj;
            expectedValues.push_back(dd * dd);
        }

        const double dd = interval * ii;
        std::vector<double> treeValues;
        bool getElementsWithin = doubleTree.getElementsWithin(RecPoint2D(-dd, dd), 2.0 * interval, treeValues);
        EXPECT_TRUE(getElementsWithin);
        EXPECT_EQ(expectedValues.size(), treeValues.size());

        for (std::vector<double>::const_iterator vv = expectedValues.begin(); vv != expectedValues.end(); ++vv) {
            bool found = std::find(treeValues.begin(), treeValues.end(), *vv) != treeValues.end();
            EXPECT_TRUE(found);
        }
    }

    // make sure the various addElements do not work once the tree is built
    bool addElement = doubleTree.addElement(RecPoint2D(0, 0), 0);
    EXPECT_FALSE(addElement);
    bool addElements = doubleTree.addElements(RawElementVector());
    EXPECT_FALSE(addElements);

    // try clearing the tree
    bool clearTree = doubleTree.clearTree();
    EXPECT_TRUE(clearTree);

    // make sure addElement works again
    addElement = doubleTree.addElement(RecPoint2D(0, 0), 0);
    EXPECT_TRUE(addElement);
}

bool buildLinearElements(const double& interval, const int count, RawElementVector& rawElements)
{
    rawElements.reserve(static_cast<size_t>(count));
    for (int ii = 0; ii < count; ++ii) {
        const double dd = interval * ii;
        const double expectedValue = dd * dd;
        rawElements.push_back(std::make_pair(RecPoint2D(-dd, dd), expectedValue));
    }
    return true;
}

bool naiveGetClosestElement(const RecPoint2D& queryPoint, const RawElementVector& rawElements, double& out)
{
    if (rawElements.size() > 0) {
        double closestDistanceSq = INFINITY;
        for (RawElementVector::const_iterator ii = rawElements.begin(); ii != rawElements.end(); ++ii) {
            double dd = ii->first.distanceSq(queryPoint);
            if (dd < closestDistanceSq) {
                closestDistanceSq = dd;
                out = ii->second;
            }
        }
        return true;
    } else {
        return false;
    }
}

bool naiveGetElementsWithin(const RecPoint2D& queryPoint,
                            const RawElementVector& rawElements,
                            const double& radius,
                            std::vector<double>& out)
{
    bool changed = false;
    const double radiusSq = radius * radius;
    for (RawElementVector::const_iterator ii = rawElements.begin(); ii != rawElements.end(); ++ii) {
        double dd = ii->first.distanceSq(queryPoint);
        if (dd <= radiusSq) {
            out.push_back(ii->second);
            changed = true;
        }
    }
    return changed;
}

TEST(RecGeometryRecKDTree2D, DISABLED_orderedInputTimingTest)
{
    printf("\nRunning orderedInputTimingTest\n");
    for (unsigned int count = 100; count <= 10000; count *= 10) {
        RecKDTree2D<double> doubleTree;
        RawElementVector rawElements;
        buildLinearElements(1.0, static_cast<int>(count), rawElements);
        printf("Count = %d\n", count);
        TIME(doubleTree.addElements(rawElements));
        TIME(doubleTree.buildTree());
        ottotime::Stopwatch timer;
        for (RawElementVector::const_iterator ii = rawElements.begin(); ii != rawElements.end(); ++ii) {
            double naiveExpected = ii->second;
            double naiveLookup = 0;
            bool foundNaive = naiveGetClosestElement(ii->first, rawElements, naiveLookup);
            EXPECT_TRUE(foundNaive);
            EXPECT_EQ(naiveLookup, naiveExpected);
        }
        printf(" - Naive getClosest: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
        timer.reset();
        for (RawElementVector::const_iterator ii = rawElements.begin(); ii != rawElements.end(); ++ii) {
            double treeExpected = ii->second;
            double treeLookup = 0;
            bool foundTree = doubleTree.getClosestElement(ii->first, treeLookup);
            EXPECT_TRUE(foundTree);
            EXPECT_EQ(treeLookup, treeExpected);
        }
        printf(" - Tree  getClosest: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
        timer.reset();
        for (RawElementVector::const_iterator ii = rawElements.begin(); ii != rawElements.end(); ++ii) {
            std::vector<double> naiveLookup;
            bool foundNaive = naiveGetElementsWithin(ii->first, rawElements, 1.5, naiveLookup);
            EXPECT_TRUE(foundNaive);
            unsigned int expectedSize = 3;
            if (ii == rawElements.begin() || (ii + 1) == rawElements.end())
                expectedSize = 2;
            EXPECT_EQ(naiveLookup.size(), expectedSize);
        }
        printf(" - Naive getWithin: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
        timer.reset();
        for (RawElementVector::const_iterator ii = rawElements.begin(); ii != rawElements.end(); ++ii) {
            std::vector<double> treeLookup;
            bool foundTree = doubleTree.getElementsWithin(ii->first, 1.5, treeLookup);
            EXPECT_TRUE(foundTree);
            unsigned int expectedSize = 3;
            if (ii == rawElements.begin() || (ii + 1) == rawElements.end())
                expectedSize = 2;
            EXPECT_EQ(treeLookup.size(), expectedSize);
        }
        printf(" - Tree  getWithin: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
    }
}

bool buildRandomizedElements(const int count, RawElementVector& rawElements)
{
    rawElements.reserve(static_cast<size_t>(count));
    for (int ii = 0; ii < count; ++ii) {
        RecPoint2D pt(rand(), rand());
        const double expectedValue = pt.x * pt.y;
        rawElements.push_back(std::make_pair(pt, expectedValue));
    }
    return true;
}

TEST(RecGeometryRecKDTree2D, DISABLED_randomInputTimingTest)
{
    printf("\nRunning randomInputTimingTest\n");
    for (unsigned int count = 100; count <= 10000; count *= 10) {
        RecKDTree2D<double> doubleTree;
        RawElementVector rawElements;
        srand(0xDEADBEEF);
        buildRandomizedElements(static_cast<int>(count), rawElements);
        // and we want to query with some other random bunch
        RawElementVector queryElements;
        buildRandomizedElements(static_cast<int>(count), queryElements);
        printf("Count = %d\n", count);
        TIME(doubleTree.addElements(rawElements));
        TIME(doubleTree.buildTree());
        ottotime::Stopwatch timer;
        for (RawElementVector::const_iterator ii = queryElements.begin(); ii != queryElements.end(); ++ii) {
            double naiveLookup = 0;
            bool foundNaive = naiveGetClosestElement(ii->first, rawElements, naiveLookup);
            EXPECT_TRUE(foundNaive);
        }
        printf(" - Naive getClosest: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
        timer.reset();
        for (RawElementVector::const_iterator ii = queryElements.begin(); ii != queryElements.end(); ++ii) {
            double treeLookup = 0;
            bool foundTree = doubleTree.getClosestElement(ii->first, treeLookup);
            EXPECT_TRUE(foundTree);
        }
        printf(" - Tree  getClosest: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
        timer.reset();
        for (RawElementVector::const_iterator ii = queryElements.begin(); ii != queryElements.end(); ++ii) {
            std::vector<double> naiveLookup;
            naiveGetElementsWithin(ii->first, rawElements, 1.5, naiveLookup);
        }
        printf(" - Naive getWithin: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
        timer.reset();
        for (RawElementVector::const_iterator ii = queryElements.begin(); ii != queryElements.end(); ++ii) {
            std::vector<double> treeLookup;
            doubleTree.getElementsWithin(ii->first, 1.5, treeLookup);
        }
        printf(" - Tree  getWithin: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
    }
}

TEST(RecGeometryRecKDTree2D, DISABLED_randomInputEquivalenceTest)
{
    printf("\nRunning randomInputEquivalenceTest\n");
    RecKDTree2D<double> doubleTree;
    RawElementVector rawElements;
    srand(0xDEADBEEF);
    buildRandomizedElements(1000, rawElements);
    // and we want to query with some other random bunch
    RawElementVector queryElements;
    buildRandomizedElements(1000, queryElements);
    TIME(doubleTree.addElements(rawElements));
    TIME(doubleTree.buildTree());
    for (RawElementVector::const_iterator ii = queryElements.begin(); ii != queryElements.end(); ++ii) {
        double naiveLookup = 0;
        bool foundNaive = naiveGetClosestElement(ii->first, rawElements, naiveLookup);
        EXPECT_TRUE(foundNaive);
        double treeLookup = 0;
        bool foundTree = doubleTree.getClosestElement(ii->first, treeLookup);
        EXPECT_TRUE(foundTree);
        EXPECT_EQ(naiveLookup, treeLookup);
    }
    for (RawElementVector::const_iterator ii = queryElements.begin(); ii != queryElements.end(); ++ii) {
        std::vector<double> naiveLookupVec;
        naiveGetElementsWithin(ii->first, rawElements, 1.5, naiveLookupVec);
        std::vector<double> treeLookupVec;
        doubleTree.getElementsWithin(ii->first, 1.5, treeLookupVec);
        EXPECT_EQ(naiveLookupVec.size(), treeLookupVec.size());
        if (naiveLookupVec.size() == treeLookupVec.size()) {
            for (unsigned int jj = 0; jj < naiveLookupVec.size(); ++jj) {
                EXPECT_EQ(naiveLookupVec[jj], treeLookupVec[jj]);
            }
        }
    }
}

TEST(RecGeometryRecKDTree2D, DISABLED_enormousTreeTest)
{
    printf("\nRunning enormousTreeTest\n");
    for (unsigned int count = 100000; count <= 10000000; count *= 10) {
        RecKDTree2D<double> doubleTree;
        RawElementVector rawElements;
        srand(0xDEADBEEF);
        buildRandomizedElements(static_cast<int>(count), rawElements);
        // and we want to query with a smaller set of some other random bunch
        RawElementVector queryElements;
        buildRandomizedElements(static_cast<int>(count / 100), queryElements);
        printf("Count = %d\n", count);
        TIME(doubleTree.addElements(rawElements));
        TIME(doubleTree.buildTree());
        ottotime::Stopwatch timer;
        for (RawElementVector::const_iterator ii = queryElements.begin(); ii != queryElements.end(); ++ii) {
            double treeLookup = 0;
            bool foundTree = doubleTree.getClosestElement(ii->first, treeLookup);
            EXPECT_TRUE(foundTree);
        }
        printf(" - Tree  getClosest: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
        timer.reset();
        for (RawElementVector::const_iterator ii = queryElements.begin(); ii != queryElements.end(); ++ii) {
            std::vector<double> treeLookup;
            doubleTree.getElementsWithin(ii->first, 1.5, treeLookup);
        }
        printf(" - Tree  getWithin: %.3lfs\n", timer.getElapsedTime().totalSecondsDouble());
    }
}

GTEST_API_ int main(int argc, char** argv)
{
    InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
