
/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file RoadLaneTestUtil.h
 * @product Common / RecGeometry
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated May 28, 2018
 * @brief Utility functions for testing RoadLane with RecClipper.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 ***********************************************************************************************************************
 */

#include <RoadWorldModel/Builder.h>
#include <RoadWorldModel/RoadWorldModel.h>
#include <log/Logger.h>
#include <recgeometry/recGeometry.h>

/*!
 * @class RoadLaneTestUtil
 * @brief Class that defines utility functions for RoadWorldModel tests.
 *
 * These functions are called from tests to create certain objects
 * or get data structures to a certain state before taking the action
 * (i.e. calling the function) that is being tested. The class is derived
 * from Builder so elements can be added to the RoadWorldModel using protected
 * functions. Some of the functions in this class return a pointer to a new
 * object. If the object has also been added to the RoadWorldModelImpl WorldDb,
 * deleting the RoadWorldModelImpl will delete the memory held by the object.
 */
class RoadLaneTestUtil : public roadWorldModel::Builder
{
public:
    RoadLaneTestUtil(){};
    ~RoadLaneTestUtil(){};

    /*!
     * @brief build() is pure virtual in Builder, but is not implemented here.
     *
     * @param rwmImpl The RoadWorldModelImpl.
     * @return true.
     */
    bool build(roadWorldModel::RoadWorldModelImpl& rwmImpl) override { return true; }

    /*!
     * @brief Creates a new test Waypoint, and adds it to the RoadWorldModel.
     *
     * The RoadWorldModel has ownership of the Waypoint.
     *
     * @param rwm A reference to the RoadWorldModel
     * @param segWId WorldId of the RoadSegment
     * @param laneWId WorldId of the RoadLane
     * @param segNum Segment number
     * @param laneNum Lane number
     * @param wpNum Waypoint number
     * @param latitude Latitude of the Waypoint
     * @param longitude Longitude of the Waypoint
     * @return Pointer to new Waypoint
     */
    roadWorldModel::Waypoint* createTestWaypoint(roadWorldModel::RoadWorldModel& rwm,
                                                 const roadWorldModel::WorldId& segWId,
                                                 const roadWorldModel::WorldId& laneWId,
                                                 int segNum,
                                                 int laneNum,
                                                 int wpNum,
                                                 double latitude,
                                                 double longitude)
    {
        RNDFId rwayptid(segNum, laneNum, wpNum);
        RecPoint2D location(latitude, longitude);
        double leftOffset = 0.0;
        double rightOffset = 0.0;

        roadWorldModel::Waypoint* wp = new roadWorldModel::Waypoint(rwayptid,
                                                                    segWId,
                                                                    laneWId,
                                                                    roadWorldModel::Waypoint::RoadSegment,
                                                                    location,
                                                                    roadWorldModel::RoadLane::BrokenWhite,
                                                                    roadWorldModel::RoadLane::BrokenWhite,
                                                                    leftOffset,
                                                                    rightOffset);
        addElement(*roadWorldModel::RoadWorldModelImpl::getInstance(), wp);
        AVLOG_DEBUG(
            "Created test Waypoint: segWId = {}, laneWId = {}, wpNum = {}", segWId.getId(), laneWId.getId(), wpNum);
        return wp;
    }

    /*!
     * @brief Creates a new test RoadLane, and adds it to the RoadWorldModel.
     *
     * The RoadWorldModel has ownership of the RoadLane.
     *
     * @param rwm The RoadWorldModel
     * @param segWId WorldId of the RoadSegment
     * @param laneWId WorldId of the RoadLane
     * @param waypointIds Vector of Waypoints in the lane
     * @param direction +1 or -1
     * @return Pointer to new RoadLane
     */
    roadWorldModel::RoadLane* createTestRoadLane(roadWorldModel::RoadWorldModel& rwm,
                                                 const roadWorldModel::WorldId& segWId,
                                                 const roadWorldModel::WorldId& laneWId,
                                                 const roadWorldModel::WorldIdVector& waypointIds,
                                                 short direction)
    {
        std::vector<roadWorldModel::StopLine> stopLines;
        std::vector<RecPose2D> poses;
        for (const roadWorldModel::WorldId& wayptId : waypointIds) {
            roadWorldModel::Waypoint* wp = dynamic_cast<roadWorldModel::Waypoint*>(rwm.getWorldElement(wayptId));
            poses.push_back(RecPose2D(wp->getLocation().x, wp->getLocation().y, 0));
        }
        std::vector<RecPose2D> poseSTD;
        RecPose2D defaultPoseSTD(0.2, 0.2, 0.1);
        poseSTD.push_back(defaultPoseSTD);
        uint32_t width = 3.0; // meters.
        double widthSTD = 0.5;
        RoadWorldModelParams params;

        roadWorldModel::RoadLane* roadLane = new roadWorldModel::RoadLane(laneWId,
                                                                          segWId,
                                                                          waypointIds,
                                                                          stopLines,
                                                                          poses,
                                                                          poseSTD,
                                                                          width,
                                                                          widthSTD,
                                                                          roadWorldModel::RoadLane::BrokenWhite,
                                                                          roadWorldModel::RoadLane::BrokenWhite,
                                                                          direction,
                                                                          params);
        addElement(*roadWorldModel::RoadWorldModelImpl::getInstance(), roadLane);
        AVLOG_DEBUG("Created test RoadLane: segWId = {}, laneWId = {}", segWId.getId(), laneWId.getId());
        return roadLane;
    }
};
