
/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file TestRecClipper.h
 * @product Common / RecGeometry
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated May 28, 2018
 * @brief Unit Tests for RecClipper (our interface to the polygon Clipper library).
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 ***********************************************************************************************************************
 */

#include "test_helper.h"

#include <iostream> // cout
#include <list>
#include <vector>

#include <gtest/gtest.h>

#include <TimeStamp/Stopwatch.h>
#include <TimeStamp/TimeStamp.h>
#include <recgeometry/recClipper.h>
#include <recgeometry/recGeometry.h>

using namespace ::testing;
using namespace std;

static RecPolygon2D unitSquare;
static RecPolygon2D unitCircle;
static RecPolygon2D offsetSquare;
static RecPolygon2D largeSquare;
static RecPolygon2D rotatedSquare;

static std::vector<RecPolygon2D> twoSquares1, twoSquares2;

void createTestPolygons()
{
    std::vector<RecPoint2D> vertices;

    // unit square
    vertices.resize(4);
    vertices[0] = RecPoint2D(0, 0);
    vertices[1] = RecPoint2D(1, 0);
    vertices[2] = RecPoint2D(1, 1);
    vertices[3] = RecPoint2D(0, 1);

    unitSquare = RecPolygon2D(vertices);

    rotatedSquare = unitSquare.transform(RecTransform2D(RecPose2D(0, 0, M_PI / 4)));

    // offset square
    for (unsigned int ii = 0; ii < 4; ++ii) {
        vertices[ii].x += 0.5;
        vertices[ii].y += 0.5;
    }
    offsetSquare = RecPolygon2D(vertices);

    vertices[0] = RecPoint2D(0, 0);
    vertices[1] = RecPoint2D(2, 0);
    vertices[2] = RecPoint2D(2, 2);
    vertices[3] = RecPoint2D(0, 2);

    largeSquare = RecPolygon2D(vertices);

    // unit circle: 360 pts
    vertices.resize(360);
    RecRadians angle_rad;
    for (unsigned int ii = 0; ii < 360; ++ii) {
        angle_rad.setDegrees(ii);
        RecVector2D unitVec = RecVector2D::getUnitVector(angle_rad);
        vertices[ii].x = unitVec.x;
        vertices[ii].y = unitVec.y;
    }

    unitCircle = RecPolygon2D(vertices);

    vertices.resize(4);
    vertices[0] = RecPoint2D(0, 0);
    vertices[1] = RecPoint2D(1, 0);
    vertices[2] = RecPoint2D(1, 1);
    vertices[3] = RecPoint2D(0, 1);

    RecPolygon2D tmp(vertices);

    twoSquares1.push_back(tmp);
    for (unsigned int ii = 0; ii < 4; ii++)
        vertices[ii].x += 4;
    RecPolygon2D tmp2(vertices);

    twoSquares1.push_back(tmp2);

    twoSquares2.push_back(tmp2);
    for (unsigned int ii = 0; ii < 4; ii++)
        vertices[ii].x += 4;
    RecPolygon2D tmp3(vertices);
    twoSquares2.push_back(tmp3);
}

void printRecPoly(const RecPolygon2D& poly, const char* prefix)
{
    for (std::vector<RecPoint2D>::const_iterator jj = poly.getVertices().begin(); jj != poly.getVertices().end();
         ++jj) {
        printf("%s%.6lf, %.6lf\n", prefix, jj->x, jj->y);
    }
}

void printRecPolyList(const std::list<RecPolygon2D>& polyList, const char* prefix)
{
    unsigned int ii = 0;
    for (std::list<RecPolygon2D>::const_iterator pp = polyList.begin(); pp != polyList.end(); ++pp) {
        printf("%s %d (%lu pts):\n", prefix, ii++, pp->getVertices().size());
        printRecPoly(*pp, " ");
    }
}

void printRecPolyVector(const std::vector<RecPolygon2D>& polyList, const char* prefix)
{
    unsigned int ii = 0;
    for (std::vector<RecPolygon2D>::const_iterator pp = polyList.begin(); pp != polyList.end(); ++pp) {
        printf("%s %d (%lu pts):\n", prefix, ii++, pp->getVertices().size());
        printRecPoly(*pp, " ");
    }
}

// ToDo: Add tests for other utility functions in Clipper
// e.g. ClipperPathToRecPoint2D()

class RecClipperTest : public testing::Test
{
protected:
    virtual void SetUp() {}
};

// self-intersection test
TEST(RecGeometryRecClipper, selfIntersectionTest)
{
    std::list<RecPolygon2D> polyList;
    EXPECT_TRUE(RecClipper::computeIntersection(unitSquare, unitSquare, polyList));
    printRecPolyList(polyList, "Unit Square Self-Intersection");
    polyList.clear();
    EXPECT_TRUE(RecClipper::computeIntersection(unitCircle, unitCircle, polyList));
    printRecPolyList(polyList, "Unit Circle Self-Intersection");
}

// self-intersection test
TEST(RecGeometryRecClipper, selfUnionTest)
{
    std::list<RecPolygon2D> polyList;
    EXPECT_TRUE(RecClipper::computeUnion(unitSquare, unitSquare, polyList));
    printRecPolyList(polyList, "Unit Square Self-Union");
    polyList.clear();
    EXPECT_TRUE(RecClipper::computeUnion(unitCircle, unitCircle, polyList));
    printRecPolyList(polyList, "Unit Circle Self-Union");
}

// intersect and union circles and squares
TEST(RecGeometryRecClipper, circleSquareTest)
{
    std::list<RecPolygon2D> polyList;
    EXPECT_TRUE(RecClipper::computeUnion(unitSquare, unitCircle, polyList));
    printRecPolyList(polyList, "Unit Square/Circle Union");
    polyList.clear();
    EXPECT_TRUE(RecClipper::computeIntersection(unitSquare, unitCircle, polyList));
    printRecPolyList(polyList, "Unit Square/Circle Intersection");
}

TEST(RecGeometryRecClipper, differenceTest)
{
    std::list<RecPolygon2D> polyList;
    EXPECT_FALSE(RecClipper::computeDifference(unitSquare, unitSquare, polyList));
    printRecPolyList(polyList, "Unit Square/Square Diff");

    EXPECT_FALSE(RecClipper::computeDifference(unitSquare, largeSquare, polyList));
    printRecPolyList(polyList, "Unit Square/largeSquare Diff");

    EXPECT_TRUE(RecClipper::computeDifference(unitSquare, rotatedSquare, polyList));
    printRecPolyList(polyList, "Unit Square/rotatedSquare Diff");

    polyList.clear();
    EXPECT_TRUE(RecClipper::computeDifference(unitSquare, offsetSquare, polyList));
    printRecPolyList(polyList, "Unit Square/OffsetSquare Diff");
}

TEST(RecGeometryRecClipper, vectorTest)
{
    std::vector<RecPolygon2D> result;
    EXPECT_FALSE(RecClipper::computeDifference(twoSquares1, twoSquares1, result));
    printRecPolyVector(result, "two squares Diff");
    result.clear();

    EXPECT_TRUE(RecClipper::computeUnion(twoSquares1, twoSquares2, result));
    printRecPolyVector(result, "two squares Union");
    result.clear();

    EXPECT_TRUE(RecClipper::computeDifference(twoSquares1, twoSquares2, result));
    printRecPolyVector(result, "two squares diff 2");
    result.clear();
}

// Test functions that check the
// winding order of polygons:

TEST_F(RecClipperTest, VerticesInClockwiseOrder)
{
    std::vector<RecPoint2D> vertices; // CW order
    vertices.push_back(RecPoint2D(1.0, 1.0));
    vertices.push_back(RecPoint2D(1.1, 2.0));
    vertices.push_back(RecPoint2D(2.0, 2.0));
    vertices.push_back(RecPoint2D(3.0, 1.0));
    const bool vertices_cw = RecClipper::verticesCW(vertices);
    EXPECT_TRUE(vertices_cw);
}

TEST_F(RecClipperTest, VerticesInCounterClockwiseOrder)
{
    std::vector<RecPoint2D> vertices; // CCW order
    vertices.push_back(RecPoint2D(1.0, 1.0));
    vertices.push_back(RecPoint2D(3.0, 1.0));
    vertices.push_back(RecPoint2D(2.0, 2.0));
    vertices.push_back(RecPoint2D(1.1, 2.0));
    const bool vertices_ccw = RecClipper::verticesCCW(vertices);
    EXPECT_TRUE(vertices_ccw);
}

TEST_F(RecClipperTest, PolygonIsClockwiseOrder)
{
    std::vector<RecPoint2D> vertices; // CW order
    vertices.push_back(RecPoint2D(1.0, 1.0));
    vertices.push_back(RecPoint2D(1.1, 2.0));
    vertices.push_back(RecPoint2D(2.0, 2.0));
    vertices.push_back(RecPoint2D(3.0, 1.0));
    RecPolygon2D polygon(vertices);
    const bool polygon_cw = RecClipper::polygonCW(polygon);
    EXPECT_TRUE(polygon_cw);
}

TEST_F(RecClipperTest, PolygonIsCounterClockwiseOrder)
{
    std::vector<RecPoint2D> vertices; // CCW order
    vertices.push_back(RecPoint2D(1.0, 1.0));
    vertices.push_back(RecPoint2D(3.0, 1.0));
    vertices.push_back(RecPoint2D(2.0, 2.0));
    vertices.push_back(RecPoint2D(1.1, 2.0));
    RecPolygon2D polygon(vertices);
    const bool polygon_ccw = RecClipper::polygonCCW(polygon);
    EXPECT_TRUE(polygon_ccw);
}

// Test functions that sort the
// vertices of a polygon:

TEST_F(RecClipperTest, SortVerticesInClockwiseOrder)
{
    std::vector<RecPoint2D> vertices;
    vertices.push_back(RecPoint2D(1, 1));
    vertices.push_back(RecPoint2D(3, 2));
    vertices.push_back(RecPoint2D(2, 3));
    vertices.push_back(RecPoint2D(1, 2));
    vertices.push_back(RecPoint2D(3, 1));
    std::vector<RecPoint2D> vertices_CW = RecClipper::sortVerticesCW(vertices);
    std::vector<RecPoint2D> expected_vertices = {
        RecPoint2D(1, 2), RecPoint2D(2, 3), RecPoint2D(3, 2), RecPoint2D(3, 1), RecPoint2D(1, 1)};
    EXPECT_CIRCULAR_EQ(vertices_CW, expected_vertices);
}

TEST_F(RecClipperTest, SortVerticesInCounterClockwiseOrder)
{
    std::vector<RecPoint2D> vertices;
    vertices.push_back(RecPoint2D(1, 1));
    vertices.push_back(RecPoint2D(3, 2));
    vertices.push_back(RecPoint2D(2, 3));
    vertices.push_back(RecPoint2D(1, 2));
    vertices.push_back(RecPoint2D(3, 1));
    std::vector<RecPoint2D> vertices_CCW = RecClipper::sortVerticesCCW(vertices);
    std::vector<RecPoint2D> expected_vertices = {
        RecPoint2D(1, 1), RecPoint2D(3, 1), RecPoint2D(3, 2), RecPoint2D(2, 3), RecPoint2D(1, 2)};
    EXPECT_CIRCULAR_EQ(vertices_CCW, expected_vertices);
}

TEST_F(RecClipperTest, SortPolygonInClockwiseOrder)
{
    std::vector<RecPoint2D> vertices;
    vertices.push_back(RecPoint2D(1, 1));
    vertices.push_back(RecPoint2D(3, 2));
    vertices.push_back(RecPoint2D(2, 3));
    vertices.push_back(RecPoint2D(1, 2));
    vertices.push_back(RecPoint2D(3, 1));
    RecPolygon2D polygon(vertices);

    RecPolygon2D polygon_CW = RecClipper::sortPolygonCW(polygon);

    std::vector<RecPoint2D> expected_vertices = {
        RecPoint2D(1, 2), RecPoint2D(2, 3), RecPoint2D(3, 2), RecPoint2D(3, 1), RecPoint2D(1, 1)};
    RecPolygon2D expected_polygon(expected_vertices);

    EXPECT_CIRCULAR_EQ(polygon_CW, expected_polygon);
}

TEST_F(RecClipperTest, SortPolygonInCounterClockwiseOrder)
{
    std::vector<RecPoint2D> vertices;
    vertices.push_back(RecPoint2D(1, 1));
    vertices.push_back(RecPoint2D(3, 2));
    vertices.push_back(RecPoint2D(2, 3));
    vertices.push_back(RecPoint2D(1, 2));
    vertices.push_back(RecPoint2D(3, 1));
    RecPolygon2D polygon(vertices);

    RecPolygon2D polygon_CCW = RecClipper::sortPolygonCCW(polygon);

    std::vector<RecPoint2D> expected_vertices = {
        RecPoint2D(1, 1), RecPoint2D(3, 1), RecPoint2D(3, 2), RecPoint2D(2, 3), RecPoint2D(1, 2)};
    RecPolygon2D expected_polygon(expected_vertices);

    EXPECT_CIRCULAR_EQ(polygon_CCW, expected_polygon);
}

TEST_F(RecClipperTest, FiveSidedPolygonCompareTwoSortFunctions)
{
    // Sort some vertices to CCW order using the
    // function from RecClipper and compare
    // the result to the function in RecPolygon2D.

    // These vertices are from a Railbot,
    // re-ordered to make an overlapping polygon.
    std::vector<RecPoint2D> vertices;
    vertices.push_back(RecPoint2D(4483955.7957665343, 595706.40326852584));
    vertices.push_back(RecPoint2D(4483951.3991223602, 595706.57508239383));
    vertices.push_back(RecPoint2D(4483950.3438909994, 595707.74217783718));
    vertices.push_back(RecPoint2D(4483955.8836258994, 595708.65155247855));
    vertices.push_back(RecPoint2D(4483951.4869817253, 595708.82336634654));

    std::vector<RecPoint2D> method1_vertices_CCW = RecClipper::sortVerticesCCW(vertices);

    RecPolygon2D poly2(vertices);
    poly2.sortVerticesCCW();
    std::vector<RecPoint2D> method2_vertices_CCW = poly2.getVertices();

    EXPECT_CIRCULAR_EQ(method1_vertices_CCW, method2_vertices_CCW);
}

TEST_F(RecClipperTest, FiveSidedPolygonSortVerticesInClockwiseOrder)
{
    // These vertices are from a Railbot,
    // re-ordered to make an overlapping polygon.
    std::vector<RecPoint2D> vertices;
    vertices.push_back(RecPoint2D(4483955.7957665343, 595706.40326852584));
    vertices.push_back(RecPoint2D(4483951.3991223602, 595706.57508239383));
    vertices.push_back(RecPoint2D(4483950.3438909994, 595707.74217783718));
    vertices.push_back(RecPoint2D(4483955.8836258994, 595708.65155247855));
    vertices.push_back(RecPoint2D(4483951.4869817253, 595708.82336634654));

    std::vector<RecPoint2D> vertices_CCW = RecClipper::sortVerticesCW(vertices);

    std::vector<RecPoint2D> expected_vertices_CCW = {RecPoint2D(4483950.3438909994, 595707.74217783718),
                                                     RecPoint2D(4483951.4869817253, 595708.82336634654),
                                                     RecPoint2D(4483955.8836258994, 595708.65155247855),
                                                     RecPoint2D(4483955.7957665343, 595706.40326852584),
                                                     RecPoint2D(4483951.3991223602, 595706.57508239383)};
    EXPECT_CIRCULAR_EQ(vertices_CCW, expected_vertices_CCW);
}

TEST_F(RecClipperTest, FiveSidedPolygonSortVerticesInCounterClockwiseOrder)
{
    // These vertices are from a Railbot,
    // re-ordered to make an overlapping polygon.
    std::vector<RecPoint2D> vertices;
    vertices.push_back(RecPoint2D(4483955.7957665343, 595706.40326852584));
    vertices.push_back(RecPoint2D(4483951.3991223602, 595706.57508239383));
    vertices.push_back(RecPoint2D(4483950.3438909994, 595707.74217783718));
    vertices.push_back(RecPoint2D(4483955.8836258994, 595708.65155247855));
    vertices.push_back(RecPoint2D(4483951.4869817253, 595708.82336634654));

    std::vector<RecPoint2D> vertices_CCW = RecClipper::sortVerticesCCW(vertices);

    std::vector<RecPoint2D> expected_vertices_CCW = {RecPoint2D(4483951.3991223602, 595706.57508239383),
                                                     RecPoint2D(4483955.7957665343, 595706.40326852584),
                                                     RecPoint2D(4483955.8836258994, 595708.65155247855),
                                                     RecPoint2D(4483951.4869817253, 595708.82336634654),
                                                     RecPoint2D(4483950.3438909994, 595707.74217783718)};
    EXPECT_CIRCULAR_EQ(vertices_CCW, expected_vertices_CCW);
}

// Test Clipper binary operations:

TEST_F(RecClipperTest, BinaryIntersection1_VectorOfRecPoint2D)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(100, 100));
    vertices1.push_back(RecPoint2D(120, 200));
    vertices1.push_back(RecPoint2D(330, 220));
    vertices1.push_back(RecPoint2D(200, 150));
    vertices1.push_back(RecPoint2D(300, 100));

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(240, 50));
    vertices2.push_back(RecPoint2D(260, 150));
    vertices2.push_back(RecPoint2D(350, 140));
    vertices2.push_back(RecPoint2D(350, 50));

    // The intersection of these 2 polygons with CW winding order
    // should output 1 polygon with CW order:

    std::vector<std::vector<RecPoint2D>> output_vertices;
    RecClipper::computeIntersection(vertices1, vertices2, output_vertices);
    EXPECT_EQ(output_vertices.size(), 1u);

    std::vector<RecPoint2D> output_poly1 = output_vertices.at(0);

    const bool is_poly_cw = RecClipper::polygonCW(output_poly1);
    EXPECT_EQ(is_poly_cw, true);

    std::vector<RecPoint2D> expected_vertices = {
        RecPoint2D(300.0, 100.0), RecPoint2D(250.0, 100.0), RecPoint2D(254.54545454999999, 122.72727273)};
    RecPolygon2D expected_polygon(expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, output_poly1);
}

TEST_F(RecClipperTest, BinaryIntersection2_ReturnListOfOnePolygon)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(100, 100));
    vertices1.push_back(RecPoint2D(120, 200));
    vertices1.push_back(RecPoint2D(330, 220));
    vertices1.push_back(RecPoint2D(200, 150));
    vertices1.push_back(RecPoint2D(300, 100));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(240, 50));
    vertices2.push_back(RecPoint2D(260, 150));
    vertices2.push_back(RecPoint2D(350, 140));
    vertices2.push_back(RecPoint2D(350, 50));
    RecPolygon2D poly2(vertices2);

    // The intersection of these 2 polygons with CW winding order
    // should output 1 polygon with CW order:

    std::list<RecPolygon2D> output_polygons;
    RecClipper::computeIntersection(poly1, poly2, output_polygons); // test output of std::list
    EXPECT_EQ(output_polygons.size(), 1u);

    RecPolygon2D output_poly1 = output_polygons.front();
    // printVertices(output_poly1);
    // std::cout << "output poly is "; printPolygonOrientation(output_poly1);
    const bool is_poly_cw = RecClipper::polygonCW(output_poly1);
    EXPECT_EQ(is_poly_cw, true);

    std::vector<RecPoint2D> expected_vertices = {
        RecPoint2D(300.0, 100.0), RecPoint2D(250.0, 100.0), RecPoint2D(254.54545454999999, 122.72727273)};
    RecPolygon2D expected_polygon(expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon);
}

TEST_F(RecClipperTest, BinaryIntersection3_ReturnVectorOfTwoPolygons)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(100, 100));
    vertices1.push_back(RecPoint2D(120, 200));
    vertices1.push_back(RecPoint2D(330, 220));
    vertices1.push_back(RecPoint2D(200, 150));
    vertices1.push_back(RecPoint2D(300, 100));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(240, 50));
    vertices2.push_back(RecPoint2D(280, 270));
    vertices2.push_back(RecPoint2D(350, 260));
    vertices2.push_back(RecPoint2D(350, 50));
    RecPolygon2D poly2(vertices2);

    // The intersection of these 2 polygons with CW winding order
    // should output 2 polygons with CW order:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeIntersection(poly1, poly2, output_polygons); // test output of std::vector
    EXPECT_EQ(output_polygons.size(), 2u);

    RecPolygon2D output_poly1 = output_polygons.at(0);
    const bool is_poly1_cw = RecClipper::polygonCW(output_poly1);
    EXPECT_EQ(is_poly1_cw, true);

    RecPolygon2D output_poly2 = output_polygons.at(1);
    const bool is_poly2_cw = RecClipper::polygonCW(output_poly2);
    EXPECT_EQ(is_poly2_cw, true);

    std::vector<RecPoint2D> poly1_expected_vertices = {RecPoint2D(264.49612402999998, 184.72868217000001),
                                                       RecPoint2D(269.86784140999998, 214.27312774999999),
                                                       RecPoint2D(330.0, 220.0)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);

    std::vector<RecPoint2D> poly2_expected_vertices = {
        RecPoint2D(300.0, 100.0), RecPoint2D(249.09090909, 100.0), RecPoint2D(253.33333332999999, 123.33333333)};
    RecPolygon2D expected_polygon2(poly2_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly2, expected_polygon2);
}

TEST_F(RecClipperTest, BinaryIntersection4_InputTwoVectorsOfPolygons)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(1, 1));
    vertices1.push_back(RecPoint2D(1, 3));
    vertices1.push_back(RecPoint2D(3, 3));
    vertices1.push_back(RecPoint2D(3, 1));
    RecPolygon2D poly1(vertices1);
    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(1, 5));
    vertices2.push_back(RecPoint2D(1, 7));
    vertices2.push_back(RecPoint2D(3, 7));
    vertices2.push_back(RecPoint2D(3, 5));
    RecPolygon2D poly2(vertices2);
    std::vector<RecPoint2D> vertices3; // CW order
    vertices3.push_back(RecPoint2D(5, 5));
    vertices3.push_back(RecPoint2D(5, 7));
    vertices3.push_back(RecPoint2D(7, 7));
    vertices3.push_back(RecPoint2D(7, 5));
    RecPolygon2D poly3(vertices3);
    std::vector<RecPoint2D> vertices4; // CW order
    vertices4.push_back(RecPoint2D(5, 1));
    vertices4.push_back(RecPoint2D(5, 3));
    vertices4.push_back(RecPoint2D(7, 5));
    vertices4.push_back(RecPoint2D(7, 1));
    RecPolygon2D poly4(vertices4);
    std::vector<RecPolygon2D> input_polygons1 = {poly1, poly2, poly3, poly4};

    std::vector<RecPoint2D> vertices5; // CW order
    vertices5.push_back(RecPoint2D(2, 2));
    vertices5.push_back(RecPoint2D(2, 6));
    vertices5.push_back(RecPoint2D(6, 6));
    vertices5.push_back(RecPoint2D(6, 2));
    RecPolygon2D poly5(vertices5);
    std::vector<RecPolygon2D> input_polygons2 = {poly5};

    // The intersection of 4 polygons with 1 polygon
    // should output 4 polygons:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeIntersection(input_polygons1, input_polygons2, output_polygons);

    EXPECT_EQ(output_polygons.size(), 4u);

    EXPECT_EQ(RecClipper::polygonCW(output_polygons.at(0)), true);
    EXPECT_EQ(RecClipper::polygonCW(output_polygons.at(1)), true);
    EXPECT_EQ(RecClipper::polygonCW(output_polygons.at(2)), true);
    EXPECT_EQ(RecClipper::polygonCW(output_polygons.at(3)), true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(3, 5),
        RecPoint2D(2, 5),
        RecPoint2D(2, 6),
        RecPoint2D(3, 6),
    };
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    std::vector<RecPoint2D> poly2_expected_vertices = {
        RecPoint2D(6, 5),
        RecPoint2D(5, 5),
        RecPoint2D(5, 6),
        RecPoint2D(6, 6),
    };
    RecPolygon2D expected_polygon2(poly2_expected_vertices);
    std::vector<RecPoint2D> poly3_expected_vertices = {
        RecPoint2D(6, 2),
        RecPoint2D(5, 2),
        RecPoint2D(5, 3),
        RecPoint2D(6, 4),
    };
    RecPolygon2D expected_polygon3(poly3_expected_vertices);
    std::vector<RecPoint2D> poly4_expected_vertices = {
        RecPoint2D(3, 2),
        RecPoint2D(2, 2),
        RecPoint2D(2, 3),
        RecPoint2D(3, 3),
    };
    RecPolygon2D expected_polygon4(poly4_expected_vertices);
    std::vector<RecPolygon2D> expected_output_polygons = {
        expected_polygon1, expected_polygon2, expected_polygon3, expected_polygon4};

    EXPECT_CIRCULAR_EQ(output_polygons, expected_output_polygons);
}

TEST_F(RecClipperTest, BinaryIntersection5_TwoAdjacentPolygons)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(10, 10));
    vertices1.push_back(RecPoint2D(10, 20));
    vertices1.push_back(RecPoint2D(20, 20));
    vertices1.push_back(RecPoint2D(20, 10));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(20, 10));
    vertices2.push_back(RecPoint2D(20, 20));
    vertices2.push_back(RecPoint2D(30, 20));
    vertices2.push_back(RecPoint2D(30, 10));
    RecPolygon2D poly2(vertices2);

    // The intersection of two adjacent (but not overlapping) polygons
    // should output 0 polygons.

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeIntersection(poly1, poly2, output_polygons); // test output of std::vector
    // std::cout << "Intersection output_vertices.size() = " << output_polygons.size() << std::endl;

    EXPECT_EQ(output_polygons.size(), 0u);
}

TEST_F(RecClipperTest, BinaryIntersection6_TwoPolygonsNotTouching)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(10, 10));
    vertices1.push_back(RecPoint2D(10, 20));
    vertices1.push_back(RecPoint2D(20, 20));
    vertices1.push_back(RecPoint2D(20, 10));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(21, 10));
    vertices2.push_back(RecPoint2D(21, 20));
    vertices2.push_back(RecPoint2D(30, 20));
    vertices2.push_back(RecPoint2D(30, 10));
    RecPolygon2D poly2(vertices2);

    // The intersection of two non-adjacent and disjoint polygons
    // should output 0 polygons.

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeIntersection(poly1, poly2, output_polygons); // test output of std::vector
    // std::cout << "Intersection output_vertices.size() = " << output_polygons.size() << std::endl;

    EXPECT_EQ(output_polygons.size(), 0u);
}

TEST_F(RecClipperTest, BinaryUnion1_ReturnVectorOfOnePolygon)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(100, 100));
    vertices1.push_back(RecPoint2D(120, 200));
    vertices1.push_back(RecPoint2D(330, 220));
    vertices1.push_back(RecPoint2D(200, 150));
    vertices1.push_back(RecPoint2D(300, 100));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(240, 50));
    vertices2.push_back(RecPoint2D(260, 150));
    vertices2.push_back(RecPoint2D(350, 140));
    vertices2.push_back(RecPoint2D(350, 50));
    RecPolygon2D poly2(vertices2);

    // The union of these 2 polygons with CW winding order
    // should output 1 polygon with CW order:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeUnion(poly1, poly2, output_polygons); // test output of std::vector
    EXPECT_EQ(output_polygons.size(), 1u);

    RecPolygon2D output_poly1 = output_polygons.at(0);
    EXPECT_EQ(RecClipper::polygonCW(output_poly1), true);

    std::vector<RecPoint2D> expected_vertices = {RecPoint2D(200, 150),
                                                 RecPoint2D(254.54545454999999, 122.72727273),
                                                 RecPoint2D(260, 150),
                                                 RecPoint2D(350, 140),
                                                 RecPoint2D(350, 50),
                                                 RecPoint2D(240, 50),
                                                 RecPoint2D(250, 100),
                                                 RecPoint2D(100, 100),
                                                 RecPoint2D(120, 200),
                                                 RecPoint2D(330, 220)};
    RecPolygon2D expected_polygon(expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon);
}

TEST_F(RecClipperTest, BinaryUnion2_ReturnVectorOfOnePolygonAndOneHole)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(100, 100));
    vertices1.push_back(RecPoint2D(120, 200));
    vertices1.push_back(RecPoint2D(330, 220));
    vertices1.push_back(RecPoint2D(200, 150));
    vertices1.push_back(RecPoint2D(300, 100));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(240, 50));
    vertices2.push_back(RecPoint2D(280, 270));
    vertices2.push_back(RecPoint2D(350, 260));
    vertices2.push_back(RecPoint2D(350, 50));
    RecPolygon2D poly2(vertices2);

    // The union of these 2 polygons
    // should output 1 polygon with a hole in it:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeUnion(poly1, poly2, output_polygons); // test output of std::vector
    EXPECT_EQ(output_polygons.size(), 2u);                   // 1 polygon (CW) + 1 polygon (CCW)

    // When computeUnion() returns a std::list or std::vector of polygons,
    // the polygons are returned in reverse order with the holes first.

    RecPolygon2D output_poly1 = output_polygons.at(0);
    EXPECT_EQ(RecClipper::polygonCCW(output_poly1), true); // CCW = a hole

    RecPolygon2D output_poly2 = output_polygons.at(1);
    EXPECT_EQ(RecClipper::polygonCW(output_poly2), true); // CW = a boundary

    // Expected vertices of the hole
    std::vector<RecPoint2D> poly1_expected_vertices = {RecPoint2D(200, 150),
                                                       RecPoint2D(253.33333332999999, 123.33333333),
                                                       RecPoint2D(264.49612402999998, 184.72868217000001),
                                                       RecPoint2D(200, 150)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);

    // Expected vertices of the boundary
    std::vector<RecPoint2D> poly2_expected_vertices = {RecPoint2D(350, 260),
                                                       RecPoint2D(350, 50),
                                                       RecPoint2D(240, 50),
                                                       RecPoint2D(249.09090909, 100),
                                                       RecPoint2D(100, 100),
                                                       RecPoint2D(120, 200),
                                                       RecPoint2D(269.86784140999998, 214.27312774999999),
                                                       RecPoint2D(280, 270),
                                                       RecPoint2D(350, 260)};
    RecPolygon2D expected_polygon2(poly2_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly2, expected_polygon2);
}

TEST_F(RecClipperTest, BinaryUnion3_ReturnOnePolygonAndTwoHoles)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(100, 100));
    vertices1.push_back(RecPoint2D(120, 200));
    vertices1.push_back(RecPoint2D(100, 300));
    vertices1.push_back(RecPoint2D(300, 300));
    vertices1.push_back(RecPoint2D(200, 250));
    vertices1.push_back(RecPoint2D(330, 220));
    vertices1.push_back(RecPoint2D(200, 150));
    vertices1.push_back(RecPoint2D(300, 100));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(240, 50));
    vertices2.push_back(RecPoint2D(280, 370));
    vertices2.push_back(RecPoint2D(350, 360));
    vertices2.push_back(RecPoint2D(350, 50));
    RecPolygon2D poly2(vertices2);

    // The union of these 2 polygons
    // should output 1 polygon with 2 holes in it:

    std::vector<RecPolygon2D> boundaries;
    std::vector<RecPolygon2D> holes;
    RecClipper::computeUnion(poly1, poly2, boundaries, holes); // test output of std::vector
    EXPECT_EQ(boundaries.size(), 1u);                          // 1 polygon (CW)
    EXPECT_EQ(holes.size(), 2u);                               // 2 polygon (CCW)

    RecPolygon2D boundary1 = boundaries.at(0);
    EXPECT_EQ(RecClipper::polygonCW(boundary1), true); // CW = a boundary

    RecPolygon2D hole1 = holes.at(0);
    EXPECT_EQ(RecClipper::polygonCCW(hole1), true);
    RecPolygon2D hole2 = holes.at(1);
    EXPECT_EQ(RecClipper::polygonCCW(hole2), true); // CCW = a hole

    // Expected vertices of the boundary
    std::vector<RecPoint2D> boundary1_expected_vertices = {RecPoint2D(350, 360),
                                                           RecPoint2D(350, 50),
                                                           RecPoint2D(240, 50),
                                                           RecPoint2D(246.25, 100),
                                                           RecPoint2D(100, 100),
                                                           RecPoint2D(120, 200),
                                                           RecPoint2D(100, 300),
                                                           RecPoint2D(271.25, 300),
                                                           RecPoint2D(280, 370)};
    RecPolygon2D expected_boundary1(boundary1_expected_vertices);
    EXPECT_CIRCULAR_EQ(boundary1, expected_boundary1);

    // Expected vertices of the holes
    std::vector<RecPoint2D> hole1_expected_vertices = {RecPoint2D(200, 250),
                                                       RecPoint2D(263.17757009000002, 235.42056074999999),
                                                       RecPoint2D(269.33333333000002, 284.66666666999998)};
    RecPolygon2D expected_hole1(hole1_expected_vertices);
    EXPECT_CIRCULAR_EQ(hole1, expected_hole1);

    std::vector<RecPoint2D> hole2_expected_vertices = {
        RecPoint2D(200, 150),
        RecPoint2D(249.41176471, 125.29411765),
        RecPoint2D(256.28865979, 180.30927835),
    };
    RecPolygon2D expected_hole2(hole2_expected_vertices);
    EXPECT_CIRCULAR_EQ(hole2, expected_hole2);
}

TEST_F(RecClipperTest, BinaryUnion4_TwoAdjacentPolygons)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(10, 10));
    vertices1.push_back(RecPoint2D(10, 20));
    vertices1.push_back(RecPoint2D(20, 20));
    vertices1.push_back(RecPoint2D(20, 10));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(20, 10));
    vertices2.push_back(RecPoint2D(20, 20));
    vertices2.push_back(RecPoint2D(30, 20));
    vertices2.push_back(RecPoint2D(30, 10));
    RecPolygon2D poly2(vertices2);

    // The union of two adjacent (but not overlapping) polygons
    // should output 1 polygon with CW order:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeUnion(poly1, poly2, output_polygons);

    EXPECT_EQ(output_polygons.size(), 1u);
    // std::cout << "output: " << std::endl; printVertices(output_polygons.at(0));

    RecPolygon2D output_poly1 = output_polygons.at(0);
    const bool is_poly1_cw = RecClipper::polygonCW(output_poly1);
    EXPECT_EQ(is_poly1_cw, true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(10, 20), RecPoint2D(30, 20), RecPoint2D(30, 10), RecPoint2D(10, 10)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, BinaryUnion5_MultipleAdjacentPolygons1)
{
    // This tests computeUnion() overload that takes 2 input vectors

    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(10, 10));
    vertices1.push_back(RecPoint2D(10, 20));
    vertices1.push_back(RecPoint2D(20, 20));
    vertices1.push_back(RecPoint2D(20, 10));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(20, 10));
    vertices2.push_back(RecPoint2D(20, 20));
    vertices2.push_back(RecPoint2D(30, 20));
    vertices2.push_back(RecPoint2D(30, 10));
    RecPolygon2D poly2(vertices2);

    std::vector<RecPoint2D> vertices3; // CW order
    vertices3.push_back(RecPoint2D(30, 10));
    vertices3.push_back(RecPoint2D(30, 20));
    vertices3.push_back(RecPoint2D(40, 20));
    vertices3.push_back(RecPoint2D(40, 10));
    RecPolygon2D poly3(vertices3);

    std::vector<RecPoint2D> vertices4; // CW order
    vertices4.push_back(RecPoint2D(40, 10));
    vertices4.push_back(RecPoint2D(40, 20));
    vertices4.push_back(RecPoint2D(50, 20));
    vertices4.push_back(RecPoint2D(50, 10));
    RecPolygon2D poly4(vertices4);

    std::vector<RecPolygon2D> input_polygons = {poly1, poly2, poly3, poly4};

    // The union of multiple adjacent (but not overlapping) polygons
    // should output 1 polygon with CW order:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeUnion(input_polygons,              // Vector of 'subject' polygons
                             std::vector<RecPolygon2D>(), // Empty vector of 'clipping' polygons
                             output_polygons);

    EXPECT_EQ(output_polygons.size(), 1u);
    // std::cout << "output_polygons.size() = " << output_polygons.size() << std::endl;
    // std::cout << "output: " << std::endl; printVertices(output_polygons.at(0));

    RecPolygon2D output_poly1 = output_polygons.at(0);
    const bool is_poly1_cw = RecClipper::polygonCW(output_poly1);
    EXPECT_EQ(is_poly1_cw, true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(10, 20), RecPoint2D(50, 20), RecPoint2D(50, 10), RecPoint2D(10, 10)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, BinaryUnion6_MultipleAdjacentPolygons2)
{
    // This tests computeUnion() overload that takes 1 input vector

    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(10, 10));
    vertices1.push_back(RecPoint2D(10, 20));
    vertices1.push_back(RecPoint2D(20, 20));
    vertices1.push_back(RecPoint2D(20, 10));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(20, 10));
    vertices2.push_back(RecPoint2D(20, 20));
    vertices2.push_back(RecPoint2D(30, 20));
    vertices2.push_back(RecPoint2D(30, 10));
    RecPolygon2D poly2(vertices2);

    std::vector<RecPoint2D> vertices3; // CW order
    vertices3.push_back(RecPoint2D(30, 10));
    vertices3.push_back(RecPoint2D(30, 20));
    vertices3.push_back(RecPoint2D(40, 20));
    vertices3.push_back(RecPoint2D(40, 10));
    RecPolygon2D poly3(vertices3);

    std::vector<RecPoint2D> vertices4; // CW order
    vertices4.push_back(RecPoint2D(40, 10));
    vertices4.push_back(RecPoint2D(40, 20));
    vertices4.push_back(RecPoint2D(50, 20));
    vertices4.push_back(RecPoint2D(50, 10));
    RecPolygon2D poly4(vertices4);

    std::vector<RecPolygon2D> input_polygons = {poly1, poly2, poly3, poly4};

    // The union of multiple adjacent (but not overlapping) polygons
    // should output 1 polygon with CW order:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeUnion(input_polygons, // Vector of 'subject' polygons
                             output_polygons);

    EXPECT_EQ(output_polygons.size(), 1u);
    // std::cout << "output_polygons.size() = " << output_polygons.size() << std::endl;
    // std::cout << "output: " << std::endl; printVertices(output_polygons.at(0));

    RecPolygon2D output_poly1 = output_polygons.at(0);
    const bool is_poly1_cw = RecClipper::polygonCW(output_poly1);
    EXPECT_EQ(is_poly1_cw, true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(10, 20), RecPoint2D(50, 20), RecPoint2D(50, 10), RecPoint2D(10, 10)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, BinaryUnion7_MultipleAdjacentAndOverlappingPolygons)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(10, 10));
    vertices1.push_back(RecPoint2D(10, 20));
    vertices1.push_back(RecPoint2D(20, 20));
    vertices1.push_back(RecPoint2D(20, 10));
    RecPolygon2D poly1(vertices1);

    // This polygon is adjacent to the first
    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(20, 10));
    vertices2.push_back(RecPoint2D(20, 20));
    vertices2.push_back(RecPoint2D(30, 20));
    vertices2.push_back(RecPoint2D(30, 10));
    RecPolygon2D poly2(vertices2);

    // This polygon overlaps the second
    std::vector<RecPoint2D> vertices3; // CW order
    vertices3.push_back(RecPoint2D(30, 10));
    vertices3.push_back(RecPoint2D(25, 15));
    vertices3.push_back(RecPoint2D(30, 20));
    vertices3.push_back(RecPoint2D(40, 20));
    vertices3.push_back(RecPoint2D(40, 10));
    RecPolygon2D poly3(vertices3);

    std::vector<RecPolygon2D> input_polygons = {poly1, poly2, poly3};

    // The union of multiple adjacent or overlapping polygons
    // should output 1 polygon with CW order:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeUnion(input_polygons, // Vector of 'subject' polygons
                             output_polygons);

    EXPECT_EQ(output_polygons.size(), 1u);
    // std::cout << "output_polygons.size() = " << output_polygons.size() << std::endl;
    // std::cout << "output: " << std::endl; printVertices(output_polygons.at(0));

    RecPolygon2D output_poly1 = output_polygons.at(0);
    const bool is_poly1_cw = RecClipper::polygonCW(output_poly1);
    EXPECT_EQ(is_poly1_cw, true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(10, 20), RecPoint2D(40, 20), RecPoint2D(40, 10), RecPoint2D(10, 10)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, BinaryUnion8_TwoPolygonsNotTouching)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(10, 10));
    vertices1.push_back(RecPoint2D(10, 20));
    vertices1.push_back(RecPoint2D(20, 20));
    vertices1.push_back(RecPoint2D(20, 10));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(21, 10));
    vertices2.push_back(RecPoint2D(21, 20));
    vertices2.push_back(RecPoint2D(30, 20));
    vertices2.push_back(RecPoint2D(30, 10));
    RecPolygon2D poly2(vertices2);

    // The union of two non-adjacent and disjoint polygons
    // returns the same two disconnected polygons!
    // Note: This computeUnion() method returns them in
    // reverse order, with poly2 first.

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeUnion(poly1, poly2, output_polygons);
    // std::cout << "output_polygons.size() = " << output_polygons.size() << std::endl;;

    EXPECT_EQ(output_polygons.size(), 2u);
    // std::cout << "output 1: " << std::endl; printVertices(output_polygons.at(0));
    // std::cout << "output 2: " << std::endl; printVertices(output_polygons.at(1));

    EXPECT_EQ(RecClipper::polygonCW(output_polygons.at(0)), true);
    EXPECT_EQ(RecClipper::polygonCW(output_polygons.at(1)), true);

    EXPECT_CIRCULAR_EQ(output_polygons.at(0), poly2); // output poly1 = input poly2
    EXPECT_CIRCULAR_EQ(output_polygons.at(1), poly1); // output poly2 = input poly1
}

TEST_F(RecClipperTest, VerticesUnion1_TwoPolygonsNotTouching)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(10, 10));
    vertices1.push_back(RecPoint2D(10, 20));
    vertices1.push_back(RecPoint2D(20, 20));
    vertices1.push_back(RecPoint2D(20, 10));
    RecPolygon2D poly1(vertices1);

    // This polygon overlaps the second
    std::vector<RecPoint2D> vertices3; // CW order
    vertices3.push_back(RecPoint2D(30, 10));
    vertices3.push_back(RecPoint2D(25, 15));
    vertices3.push_back(RecPoint2D(30, 20));
    vertices3.push_back(RecPoint2D(40, 20));
    vertices3.push_back(RecPoint2D(40, 10));
    RecPolygon2D poly3(vertices3);

    std::vector<RecPolygon2D> input_polygons = {poly1, poly3};

    // The union of the vertices of multiple polygons
    // (dis-joint, adjacent or overlapping)
    // should output 1 polygon with CW order:

    RecPolygon2D output_polygon;
    RecClipper::computeVerticesUnion(input_polygons, // Vector of 'subject' polygons
                                     output_polygon);

    const bool is_poly1_cw = RecClipper::polygonCW(output_polygon);
    EXPECT_EQ(is_poly1_cw, true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(10, 20),
        RecPoint2D(20, 20),
        RecPoint2D(30, 20),
        RecPoint2D(40, 20),
        RecPoint2D(25, 15),
        RecPoint2D(40, 10),
        RecPoint2D(30, 10),
        RecPoint2D(20, 10),
        RecPoint2D(10, 10),
        RecPoint2D(10, 20),
        RecPoint2D(),

    };
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_polygon, expected_polygon1);
}

TEST_F(RecClipperTest, BinaryDifference1_ReturnListOfOnePolygon)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(100, 100));
    vertices1.push_back(RecPoint2D(120, 200));
    vertices1.push_back(RecPoint2D(330, 220));
    vertices1.push_back(RecPoint2D(200, 150));
    vertices1.push_back(RecPoint2D(300, 100));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(240, 50));
    vertices2.push_back(RecPoint2D(260, 150));
    vertices2.push_back(RecPoint2D(350, 140));
    vertices2.push_back(RecPoint2D(350, 50));
    RecPolygon2D poly2(vertices2);

    // The difference of these 2 polygons
    // should output 1 polygon.
    // Polygon1 subtract Polygon2:

    std::list<RecPolygon2D> output_polygons;
    RecClipper::computeDifference(poly1, poly2, output_polygons); // test output of std::list
    EXPECT_EQ(output_polygons.size(), 1u);

    RecPolygon2D output_poly1 = output_polygons.front();

    // ToDo: add this back in later pull request
    // const bool is_poly1_cw = RecClipper::polygonCW(output_poly1);
    // EXPECT_EQ(is_poly1_cw, true);

    std::vector<RecPoint2D> poly1_expected_vertices = {RecPoint2D(200, 150),
                                                       RecPoint2D(254.54545454999999, 122.72727273),
                                                       RecPoint2D(250, 100),
                                                       RecPoint2D(100, 100),
                                                       RecPoint2D(120, 200),
                                                       RecPoint2D(330, 220)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, BinaryDifference2_ReturnVectorOfOnePolygon)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(100, 100));
    vertices1.push_back(RecPoint2D(120, 200));
    vertices1.push_back(RecPoint2D(330, 220));
    vertices1.push_back(RecPoint2D(200, 150));
    vertices1.push_back(RecPoint2D(300, 100));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(240, 50));
    vertices2.push_back(RecPoint2D(260, 150));
    vertices2.push_back(RecPoint2D(350, 140));
    vertices2.push_back(RecPoint2D(350, 50));
    RecPolygon2D poly2(vertices2);

    std::vector<RecPolygon2D> input_polygons1 = {poly1};
    std::vector<RecPolygon2D> input_polygons2 = {poly2};

    // The difference of these 2 polygons
    // should output 1 polygon.
    // Polygon1 subtract Polygon2:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeDifference(input_polygons1, input_polygons2, output_polygons); // test output of std::vector
    EXPECT_EQ(output_polygons.size(), 1u);

    RecPolygon2D output_poly1 = output_polygons.at(0);

    std::vector<RecPoint2D> poly1_expected_vertices = {RecPoint2D(200, 150),
                                                       RecPoint2D(254.54545454999999, 122.72727273),
                                                       RecPoint2D(250, 100),
                                                       RecPoint2D(100, 100),
                                                       RecPoint2D(120, 200),
                                                       RecPoint2D(330, 220)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, BinaryXOR1_ReturnVectorOfTwoPolygons)
{
    std::vector<RecPoint2D> vertices1; // CW order
    vertices1.push_back(RecPoint2D(100, 100));
    vertices1.push_back(RecPoint2D(120, 200));
    vertices1.push_back(RecPoint2D(330, 220));
    vertices1.push_back(RecPoint2D(200, 150));
    vertices1.push_back(RecPoint2D(300, 100));
    RecPolygon2D poly1(vertices1);

    std::vector<RecPoint2D> vertices2; // CW order
    vertices2.push_back(RecPoint2D(240, 50));
    vertices2.push_back(RecPoint2D(260, 150));
    vertices2.push_back(RecPoint2D(350, 140));
    vertices2.push_back(RecPoint2D(350, 50));
    RecPolygon2D poly2(vertices2);

    // The difference of these 2 polygons
    // should output 1 polygon.
    // Polygon1 subtract Polygon2:

    std::vector<RecPolygon2D> output_polygons;
    RecClipper::computeXOR(poly1, poly2, output_polygons); // test output of std::vector

    EXPECT_EQ(output_polygons.size(), 2u);

    // std::cout << "poly1: " << std::endl; printVertices(output_polygons.at(0));
    // std::cout << "poly1: " << std::endl; printVertices(output_polygons.at(1));

    RecPolygon2D output_poly1 = output_polygons.at(0);
    EXPECT_EQ(RecClipper::polygonCW(output_poly1), true);

    RecPolygon2D output_poly2 = output_polygons.at(1);
    EXPECT_EQ(RecClipper::polygonCW(output_poly2), true);

    std::vector<RecPoint2D> poly1_expected_vertices = {RecPoint2D(200, 150),
                                                       RecPoint2D(254.54545454999999, 122.72727273),
                                                       RecPoint2D(250, 100),
                                                       RecPoint2D(100, 100),
                                                       RecPoint2D(120, 200),
                                                       RecPoint2D(330, 220)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);

    std::vector<RecPoint2D> poly2_expected_vertices = {RecPoint2D(350, 140),
                                                       RecPoint2D(350, 50),
                                                       RecPoint2D(240, 50),
                                                       RecPoint2D(250, 100),
                                                       RecPoint2D(300, 100),
                                                       RecPoint2D(254.54545454999999, 122.72727273),
                                                       RecPoint2D(260, 150)};
    RecPolygon2D expected_polygon2(poly2_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly2, expected_polygon2);
}

// Test Clipper offset operations:

TEST_F(RecClipperTest, OffsetPositive_Miter1_VectorOfRecPoint2D)
{
    // Apply a positive "miter" offset to a square shape

    std::vector<RecPoint2D> vertices; // CW order
    vertices.push_back(RecPoint2D(10, 10));
    vertices.push_back(RecPoint2D(10, 20));
    vertices.push_back(RecPoint2D(20, 20));
    vertices.push_back(RecPoint2D(20, 10));

    // When offsetting a polygon, a miter with a limit
    // of 3 will produce sharper pointed corners.

    const double offset = 0.5;
    const double miter_limit = 3.0; // default = 2.0
    std::vector<std::vector<RecPoint2D>> offset_vertices;
    RecClipper::computeOffsetMiter(vertices, offset, miter_limit, offset_vertices);
    // std::cout << "Output: " << std::endl; printVertices(offset_vertices.at(0));

    EXPECT_EQ(offset_vertices.size(), 1u);

    std::vector<RecPoint2D> output_poly1 = offset_vertices.at(0);
    EXPECT_EQ(RecClipper::polygonCW(output_poly1), true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(20.5, 9.5), RecPoint2D(9.5, 9.5), RecPoint2D(9.5, 20.5), RecPoint2D(20.5, 20.5)};
    EXPECT_CIRCULAR_EQ(output_poly1, poly1_expected_vertices);
}

TEST_F(RecClipperTest, OffsetPositive_Miter2_RecPolygon2D)
{
    // Apply a positive "miter" offset to a square shaped polygon

    std::vector<RecPoint2D> vertices; // CW order
    vertices.push_back(RecPoint2D(10, 10));
    vertices.push_back(RecPoint2D(10, 20));
    vertices.push_back(RecPoint2D(20, 20));
    vertices.push_back(RecPoint2D(20, 10));
    RecPolygon2D polygon(vertices);

    // When offsetting a polygon, a miter with a limit
    // of 3 will produce sharper pointed corners.

    const double offset = 0.5;
    const double miter_limit = 3.0; // default = 2.0
    std::vector<RecPolygon2D> offset_polygons;
    RecClipper::computeOffsetMiter(polygon, offset, miter_limit, offset_polygons);

    EXPECT_EQ(offset_polygons.size(), 1u);

    RecPolygon2D output_poly1 = offset_polygons.at(0);
    EXPECT_EQ(RecClipper::polygonCW(output_poly1), true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(20.5, 9.5), RecPoint2D(9.5, 9.5), RecPoint2D(9.5, 20.5), RecPoint2D(20.5, 20.5)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, OffsetPositive_Square1_RecPolygon2D)
{
    // Apply a positive "square" offset to a square shaped polygon

    std::vector<RecPoint2D> vertices; // CW order
    vertices.push_back(RecPoint2D(10, 10));
    vertices.push_back(RecPoint2D(10, 20));
    vertices.push_back(RecPoint2D(20, 20));
    vertices.push_back(RecPoint2D(20, 10));
    RecPolygon2D polygon(vertices);

    const double offset = 0.5;
    std::vector<RecPolygon2D> offset_polygons;
    RecClipper::computeOffsetSquare(polygon, offset, offset_polygons);
    // std::cout << "Output: " << std::endl; printVertices(offset_polygons.at(0));

    EXPECT_EQ(offset_polygons.size(), 1u);

    RecPolygon2D output_poly1 = offset_polygons.at(0);
    EXPECT_EQ(RecClipper::polygonCW(output_poly1), true);

    std::vector<RecPoint2D> poly1_expected_vertices = {RecPoint2D(20.5, 20.20710678),
                                                       RecPoint2D(20.5, 9.79289322),
                                                       RecPoint2D(20.20710678, 9.5),
                                                       RecPoint2D(9.79289322, 9.5),
                                                       RecPoint2D(9.5, 9.79289322),
                                                       RecPoint2D(9.5, 20.20710678),
                                                       RecPoint2D(9.79289322, 20.5),
                                                       RecPoint2D(20.20710678, 20.5)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, OffsetPositive_Round1_RecPolygon2D)
{
    // Apply a positive "round" offset to a square shaped polygon

    std::vector<RecPoint2D> vertices; // CW order
    vertices.push_back(RecPoint2D(10, 10));
    vertices.push_back(RecPoint2D(10, 20));
    vertices.push_back(RecPoint2D(20, 20));
    vertices.push_back(RecPoint2D(20, 10));
    RecPolygon2D polygon(vertices);

    // When offsetting a polygon, a "round" with a
    // tolerance of 0.01 will make the rounded corners
    // accurate to 1cm.

    const double offset = 0.5;
    const double arc_tolerance = 0.01; // default = 0.25
    std::vector<RecPolygon2D> offset_polygons;
    RecClipper::computeOffsetRound(polygon, offset, arc_tolerance, offset_polygons);
    // std::cout << "Output: " << std::endl; printVertices(offset_polygons.at(0));

    EXPECT_EQ(offset_polygons.size(), 1u);

    RecPolygon2D output_poly1 = offset_polygons.at(0);
    EXPECT_EQ(RecClipper::polygonCW(output_poly1), true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(20.19501754, 20.4604),
        RecPoint2D(20.3591443, 20.34787264),
        RecPoint2D(20.4663826, 20.18024225),
        RecPoint2D(20.5, 20),
        RecPoint2D(20.5, 10),
        RecPoint2D(20.4604, 9.80498246),
        RecPoint2D(20.34787264, 9.6408557),
        RecPoint2D(20.18024225, 9.5336174),
        RecPoint2D(20, 9.5),
        RecPoint2D(10, 9.5),
        RecPoint2D(9.80498246, 9.5396),
        RecPoint2D(9.6408557, 9.65212736),
        RecPoint2D(9.5336174, 9.81975775),
        RecPoint2D(9.5, 10),
        RecPoint2D(9.5, 20),
        RecPoint2D(9.5396, 20.19501754),
        RecPoint2D(9.65212736, 20.3591443),
        RecPoint2D(9.81975775, 20.4663826),
        RecPoint2D(10, 20.5),
        RecPoint2D(20, 20.5),
    };
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, OffsetPositive_Round2_RecPolygon2D)
{
    // Apply a positive "round" offset to the example 2D polygon
    // from the Clipper manual:
    // http://www.angusj.com/delphi/clipper/documentation/Docs/Units/ClipperLib/Classes/ClipperOffset/_Body.htm

    std::vector<RecPoint2D> vertices; // CW order
    vertices.push_back(RecPoint2D(348, 257));
    vertices.push_back(RecPoint2D(364, 148));
    vertices.push_back(RecPoint2D(362, 148));
    vertices.push_back(RecPoint2D(326, 241));
    vertices.push_back(RecPoint2D(295, 219));
    vertices.push_back(RecPoint2D(258, 88));
    vertices.push_back(RecPoint2D(440, 129));
    vertices.push_back(RecPoint2D(370, 196));
    vertices.push_back(RecPoint2D(372, 275));
    RecPolygon2D polygon(vertices);

    const double offset = 7.0;
    const double arc_tolerance = 0.25; // default = 0.25
    std::vector<RecPolygon2D> offset_polygons;
    RecClipper::computeOffsetRound(polygon, offset, arc_tolerance, offset_polygons);
    // std::cout << "Output: " << std::endl; printVertices(offset_polygons.at(0));

    EXPECT_EQ(offset_polygons.size(), 1u);

    RecPolygon2D output_poly1 = offset_polygons.at(0);
    EXPECT_EQ(RecClipper::polygonCW(output_poly1), true);

    std::vector<RecPoint2D> poly1_expected_vertices = {
        RecPoint2D(374.91014869, 281.36639887), RecPoint2D(377.75385461, 278.98662227),
        RecPoint2D(378.99775785, 274.82284157), RecPoint2D(377.07608740, 198.91685872),
        RecPoint2D(444.84019850, 134.05692381), RecPoint2D(446.74421420, 130.87498663),
        RecPoint2D(446.75572080, 127.16690524), RecPoint2D(444.87148943, 123.97321268),
        RecPoint2D(441.53837098, 122.17113371), RecPoint2D(259.53837098, 81.17113371),
        RecPoint2D(255.83427772, 81.34345082),  RecPoint2D(252.73791266, 83.38367713),
        RecPoint2D(251.11815374, 86.71930016),  RecPoint2D(251.26354170, 89.9026638),
        RecPoint2D(288.26354170, 220.9026638),  RecPoint2D(290.94877096, 224.70855001),
        RecPoint2D(321.94877096, 246.70855001), RecPoint2D(325.43316529, 247.97701214),
        RecPoint2D(329.07662038, 247.28763923), RecPoint2D(332.52797669, 243.52695872),
        RecPoint2D(349.23993966, 200.35438773), RecPoint2D(341.07421729, 255.98337134),
        RecPoint2D(341.52664334, 259.66376679), RecPoint2D(343.80000000, 262.6),
        RecPoint2D(367.80000000, 280.6),        RecPoint2D(371.24981942, 281.959686)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);
}

TEST_F(RecClipperTest, OffsetNegative_Round3_RecPolygon2D)
{
    // Apply a negative "round" offset to the example 2D polygon
    // from the Clipper manual:
    // http://www.angusj.com/delphi/clipper/documentation/Docs/Units/ClipperLib/Classes/ClipperOffset/_Body.htm
    //
    // Because part of it is so thin, the resulting
    // output is actually 2 polygons.

    std::vector<RecPoint2D> vertices; // CW order
    vertices.push_back(RecPoint2D(348, 257));
    vertices.push_back(RecPoint2D(364, 148));
    vertices.push_back(RecPoint2D(362, 148));
    vertices.push_back(RecPoint2D(326, 241));
    vertices.push_back(RecPoint2D(295, 219));
    vertices.push_back(RecPoint2D(258, 88));
    vertices.push_back(RecPoint2D(440, 129));
    vertices.push_back(RecPoint2D(370, 196));
    vertices.push_back(RecPoint2D(372, 275));
    RecPolygon2D polygon(vertices);

    const double offset = -7.0;
    const double arc_tolerance = 0.25; // default = 0.25
    std::vector<RecPolygon2D> offset_polygons;
    RecClipper::computeOffsetRound(polygon, offset, arc_tolerance, offset_polygons);
    // std::cout << "Output poly1: " << std::endl; printVertices(offset_polygons.at(0));
    // std::cout << "Output poly2: " << std::endl; printVertices(offset_polygons.at(1));

    EXPECT_EQ(offset_polygons.size(), 2u);

    RecPolygon2D output_poly1 = offset_polygons.at(0);
    EXPECT_EQ(RecClipper::polygonCW(output_poly1), true);

    RecPolygon2D output_poly2 = offset_polygons.at(1);
    EXPECT_EQ(RecClipper::polygonCW(output_poly2), true);

    std::vector<RecPoint2D> poly1_expected_vertices = {RecPoint2D(355.47202331, 145.47304128),
                                                       RecPoint2D(357.67873315, 142.49303597),
                                                       RecPoint2D(362, 141),
                                                       RecPoint2D(364, 141),
                                                       RecPoint2D(367.57566713, 141.98214286),
                                                       RecPoint2D(370.14795828, 144.65297012),
                                                       RecPoint2D(370.92578271, 149.01662866),
                                                       RecPoint2D(364.68063764, 191.56167944),
                                                       RecPoint2D(365.15980150, 190.94307619),
                                                       RecPoint2D(425.73675842, 132.96227456),
                                                       RecPoint2D(267.93247351, 97.41295763),
                                                       RecPoint2D(301.06520026, 214.72071989),
                                                       RecPoint2D(322.71811496, 230.08730451)};
    RecPolygon2D expected_polygon1(poly1_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly1, expected_polygon1);

    std::vector<RecPoint2D> poly2_expected_vertices = {RecPoint2D(363.14947217, 201.99274421),
                                                       RecPoint2D(355.53038326, 253.89778744),
                                                       RecPoint2D(364.63642339, 260.72731755)};
    RecPolygon2D expected_polygon2(poly2_expected_vertices);
    EXPECT_CIRCULAR_EQ(output_poly2, expected_polygon2);
}

int main(int argc, char** argv)
{
    createTestPolygons();
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
