/**********************************************************
  Copyright (c) 2014-    Ottomatika, Inc.
 **********************************************************/

/**********************************************************
  Copyright (c) 2010-2013 Carnegie Mellon University,
  All rights reserved.

  This source code was developed in part with support from
  sponsors including General Motors, the National Science
  Foundation and the US Department of Transportation.

  Use of this software is strictly governed by agreements
  between Carnegie Mellon University and various sponsors.

  Users of this software must also be fully aware and agree
  that Carnegie Mellon does not guarantee the correct
  functioning of this software in any system. Carnegie
  Mellon or any of its affiliates will not be liable for any
  damage and/or penalties resulting from the use of this
  software. Any user of this software takes complete
  responsibility for the use of any software or design.
 **********************************************************/

/**
 * @file serializeUnitTest.cc
 * @author: Chris Urmson (curmson@ri.cmu.edu)
 * @date: 09/10/06
 *
 * @attention Copyright (c) 2006
 * @attention Carnegie Mellon University
 * @attention All rights reserved.
 */
#ifndef _SERIALIZEUNITTEST_CC_
#define _SERIALIZEUNITTEST_CC_

#include <gtest/gtest.h>

#include <cereal/archives/xml.hpp>
#include <fstream>
#include <iostream>
#include <recgeometry/recGeometry.h>
#include <vector>

using namespace ::testing;
using namespace cereal;
using namespace std;

template<class T>
void save(const string& name, const T& to_save)
{
    ofstream ofs(name.c_str());
    XMLOutputArchive oa(ofs);
    oa(to_save);
}

template<class T>
void load(const string& name, T& to_load)
{
    ifstream ifs(name.c_str());
    XMLInputArchive ia(ifs);
    ia(to_load);
}

#define TEST_RW(out, in)                                                                                               \
    save("test.xml", out);                                                                                             \
    load("test.xml", in);                                                                                              \
    EXPECT_EQ(in, out);

TEST(RecGeometrySerialization, testRecRadiansSerialization)
{
    cout << "Testing recRadians" << endl;
    RecRadians r(0), r1;

    TEST_RW(r, r1);

    r = 7.6;
    TEST_RW(r, r1);

    r = 2 * M_PI;
    TEST_RW(r, r1);

    r = -2.2 * M_PI;
    TEST_RW(r, r1);
}

TEST(RecGeometrySerialization, testRecPointVectors)
{
    cout << "Testing points/vectors" << endl;
    RecPoint2D p2o(0.1, -6.8), p2i;
    TEST_RW(p2o, p2i);

    RecPoint3D p3o(-9.4, 3.6, 2.1), p3i;
    TEST_RW(p2o, p2i);

    RecVector2D v2o(p2o), v2i;
    TEST_RW(v2o, v2i);

    RecVector3D v3o(p3o), v3i;
    TEST_RW(v3o, v3i);
}

TEST(RecGeometrySerialization, testPose)
{
    cout << "Testing pose/transform types" << endl;
    RecPose2D p2o(0.3, -0.6, M_PI / 2), p2i;
    TEST_RW(p2o, p2i);

    RecDifferentialPose2D dp2o(-0.2, -0.34, M_PI / 3), dp2i;
    TEST_RW(dp2o, dp2i);

    //  enum RecPose3D::EulerAngleOrder order =
    //    enum RecPose3D::EulerAngleOrder::YXY;

    RecPose3D p3o(-0.36, +923942, 9.3, M_PI / 3, -M_PI / 2, RecPose3D::YXY), p3i;

    TEST_RW(p3o, p3i);

    RecDifferentialPose3D dp3o(-0.36, +923942, 9.3, M_PI / 3, -M_PI / 2, RecPose3D::XZX), dp3i;
    TEST_RW(dp3o, dp3i);

    RecTransform2D t2o(p2o), t2i;
    TEST_RW(t2o, t2i);

    RecTransform3D t3o(p3o), t3i;
    TEST_RW(t3o, t3i);

    // RecQuaternion doesn't specify an operator==
    // Regardless of what quaternion equality is, here we just want to know that the member variables are set the same
    RecQuaternion qo(t3o), qi;
    save("test.txt", qo);
    load("test.txt", qi);
    for (int i = 0; i < 4; i++)
        EXPECT_EQ(qo.quat[i], qi.quat[i]);
}

TEST(RecGeometrySerialization, testLinesAndBoxes)
{
    cout << "Test lines and boxes" << endl;
    RecPoint2D a(5, -6), b(3, 9);
    RecPoint3D c(5, 6, 7), d(10, 23, 34);

    RecLineSegment2D l2o(a, b), l2i;
    save("test.txt", l2o);
    load("test.txt", l2i);
    EXPECT_EQ(l2o.p1(), l2i.p1());
    EXPECT_EQ(l2o.p2(), l2i.p2());

    RecLineSegment3D l3o(c, d), l3i;
    save("test.txt", l3o);
    load("test.txt", l3i);
    EXPECT_EQ(l3o.p1, l3i.p1);
    EXPECT_EQ(l3o.p2, l3i.p2);

    RecAxisAlignedBox2D b2o(a, b), b2i;
    save("test.txt", b2o);
    load("test.txt", b2i);
    EXPECT_EQ(b2o.getMinPoint(), b2i.getMinPoint());
    EXPECT_EQ(b2o.getMaxPoint(), b2i.getMaxPoint());

    RecAxisAlignedBox3D b3o(c, d), b3i;
    save("test.txt", b3o);
    load("test.txt", b3i);
    EXPECT_EQ(b3o.getMin(), b3i.getMin());
    EXPECT_EQ(b3o.getMax(), b3i.getMax());
}

TEST(RecGeometrySerialization, testPolygon)
{
    cout << "Testing polygon" << endl;
    RecPolygon2D polyOut, polyIn;
    RecPoint2D p(0, 0);
    polyOut.addVertex(p);
    p.x = 10;
    p.y = 10;
    polyOut.addVertex(p);
    p.x = -10;
    p.y = 10;
    polyOut.addVertex(p);

    save("test.txt", polyOut);
    load("test.txt", polyIn);

    RecPoint2D a, b;
    polyOut.getCenter(a);
    polyIn.getCenter(b);
    cout << "center: " << a << " " << b << endl;
    EXPECT_EQ(a, b);

    polyOut.getVertex(0, a);
    polyIn.getVertex(0, b);
    cout << "v: " << a << " " << b << endl;
    EXPECT_EQ(a, b);
    polyOut.getVertex(1, a);
    polyIn.getVertex(1, b);
    cout << "v: " << a << " " << b << endl;
    EXPECT_EQ(a, b);
    polyOut.getVertex(2, a);
    polyIn.getVertex(2, b);
    cout << "v: " << a << " " << b << endl;
    EXPECT_EQ(a, b);
}

#include <TimeStamp/Stopwatch.h>
#include <sys/time.h>
#include <unistd.h>

void buildUnitCircle(RecPolygon2D& out, const double& step, const RecVector2D& offset)
{
    std::vector<RecVector2D> pointVec;
    pointVec.reserve(static_cast<size_t>(1 + 360.0 / step));
    for (double angle = 0.0; angle < 360.0; angle += step) {
        RecRadians angle_rad;
        angle_rad.setDegrees(angle);
        pointVec.push_back(offset + RecVector2D::getUnitVector(angle_rad));
    }
    out = RecPolygon2D(pointVec);
}

TEST(RecGeometrySerialization, testPolygonStress)
{
    // make a big, dense circle
    RecPolygon2D bigCircle;
    RecPolygon2D otherCircle;
    ottotime::Stopwatch timer;
    for (int ii = 0; ii < 10; ++ii) {
        buildUnitCircle(bigCircle, 0.125, RecVector2D(ii, ii));
        save("test.txt", bigCircle);
        timer.reset();
        load("test.txt", otherCircle);
        printf(" - Deserialize %d: %.3lfs\n", ii + 1, timer.getElapsedTime().totalSecondsDouble());
        EXPECT_EQ(bigCircle.getNumVertices(), otherCircle.getNumVertices());
    }
}

GTEST_API_ int main(int argc, char** argv)
{
    InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}

#endif // ifndef _SERIALIZEUNITTEST_CC_
