
/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file test_helper.h
 * @product Common / RecGeometry
 * @version TODO <Release Number>
 * @released TODO <date>
 * @updated May 28, 2018
 * @brief Helper functions for testing the RecClipper interface.
 *
 * @section License
 *
 * Copyright (c) 2018- Aptiv. All rights reserved.
 * NOTE: Removing this heading is prohibited.
 *
 ***********************************************************************************************************************
 */

#include <algorithm> // copy_n, search
#include <vector>

#include <gtest/gtest.h>

#include <recgeometry/recGeometry.h>

/**\brief Check if two vector of vertices are equal.
 *        This treats each vector as if it was a circular buffer
 *        e.g. [2,3,4,1] = [1,2,3,4]
 *        It is expected that the vertices are unique.
 */
bool checkCircularVectorsAreEqual(std::vector<RecPoint2D>& A, std::vector<RecPoint2D>& B)
{
    // Concatenate a copy of 'B' to itself
    const std::vector<RecPoint2D>::size_type old_size = B.size();
    B.reserve(2 * old_size);
    std::copy_n(B.begin(), old_size, std::back_inserter(B));
    // Check if 'A' is a sub-vector of 'B'
    auto it = std::search(std::begin(B), std::end(B), std::begin(A), std::end(A));

    return (it != std::end(B));
}

/**\brief Assert that two vectors of vertices represent
 *        the same polygon, with the same winding order.
 */
void ASSERT_CIRCULAR_EQ(std::vector<RecPoint2D>& vertices1, std::vector<RecPoint2D>& vertices2)
{
    const bool result = checkCircularVectorsAreEqual(vertices1, vertices2);
    ASSERT_TRUE(result);
}

/**\brief Assert that two polygons are the same
 *        polygon, with the same winding order.
 */
void ASSERT_CIRCULAR_EQ(RecPolygon2D& polygon1, RecPolygon2D& polygon2)
{

    std::vector<RecPoint2D> vertices1 = polygon1.getVertices();
    vertices1.pop_back();

    std::vector<RecPoint2D> vertices2 = polygon2.getVertices();
    vertices2.pop_back();

    ASSERT_CIRCULAR_EQ(vertices1, vertices2);
}

/**\brief Assert that two vectors of polygons present the
 *        same polygons, each with the same winding order.
 */
void ASSERT_CIRCULAR_EQ(std::vector<RecPolygon2D>& polygons1, std::vector<RecPolygon2D>& polygons2)
{
    ASSERT_EQ(polygons1.size(), polygons2.size());

    const std::vector<RecPolygon2D>::size_type num_polygons = polygons1.size();
    unsigned int num_polygons_matched = 0;
    for (std::vector<RecPolygon2D>::size_type i = 0; i < num_polygons; ++i) {
        for (std::vector<RecPolygon2D>::size_type j = 0; j < num_polygons; ++j) {
            std::vector<RecPoint2D> vertices1 = polygons1.at(i).getVertices();
            vertices1.pop_back();

            std::vector<RecPoint2D> vertices2 = polygons2.at(j).getVertices();
            vertices2.pop_back();

            const bool same = checkCircularVectorsAreEqual(vertices1, vertices2);
            if (same) {
                num_polygons_matched++;
                break;
            }
        }
    }

    ASSERT_EQ(num_polygons_matched, static_cast<unsigned int>(num_polygons));
}

/**\brief Expect that two vectors of vertices represent
 *        the same polygon, with the same winding order.
 */
void EXPECT_CIRCULAR_EQ(std::vector<RecPoint2D>& vertices1, std::vector<RecPoint2D>& vertices2)
{
    const bool result = checkCircularVectorsAreEqual(vertices1, vertices2);
    EXPECT_TRUE(result);
}

/**\brief Expect that two polygons are the same
 *        polygon, with the same winding order.
 */
void EXPECT_CIRCULAR_EQ(RecPolygon2D& polygon1, RecPolygon2D& polygon2)
{

    std::vector<RecPoint2D> vertices1 = polygon1.getVertices();
    vertices1.pop_back();

    std::vector<RecPoint2D> vertices2 = polygon2.getVertices();
    vertices2.pop_back();

    EXPECT_CIRCULAR_EQ(vertices1, vertices2);
}

/**\brief Expect that two vectors of polygons present the
 *        same polygons, each with the same winding order.
 */
void EXPECT_CIRCULAR_EQ(std::vector<RecPolygon2D>& polygons1, std::vector<RecPolygon2D>& polygons2)
{
    EXPECT_EQ(polygons1.size(), polygons2.size());

    const std::vector<RecPolygon2D>::size_type num_polygons = polygons1.size();
    unsigned int num_polygons_matched = 0;
    for (std::vector<RecPolygon2D>::size_type i = 0; i < num_polygons; ++i) {
        for (std::vector<RecPolygon2D>::size_type j = 0; j < num_polygons; ++j) {
            std::vector<RecPoint2D> vertices1 = polygons1.at(i).getVertices();
            vertices1.pop_back();

            std::vector<RecPoint2D> vertices2 = polygons2.at(j).getVertices();
            vertices2.pop_back();

            const bool same = checkCircularVectorsAreEqual(vertices1, vertices2);
            if (same) {
                num_polygons_matched++;
                break;
            }
        }
    }

    EXPECT_EQ(num_polygons_matched, static_cast<unsigned int>(num_polygons));
}

/**\brief Print the vertices of a 2D polygon.
 */
void printVertices(const std::vector<RecPoint2D>& vertices)
{
    const std::streamsize default_precision = std::cout.precision();
    for (auto v : vertices) {
        std::cout << "  " << std::setprecision(17) << v << std::setprecision(static_cast<int>(default_precision))
                  << std::endl;
    }
}

/**\brief Print the vertices of a 2D polygon.
 */
void printVertices(const RecPolygon2D& polygon)
{
    printVertices(polygon.getVertices());
}
