/**
 * @copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 */

#include <fstream>

#include <avtime/ConversionsUsec.hpp>
#include <spdlog/sinks/ostream_sink.h>

#include "sms/SmsLogger.hpp"
#include "sms/TraceEvent.hpp"

namespace av {
namespace sms {

static std::string createFileName(const std::string& name)
{
    const auto pid = getpid();
    return name + "-" + std::to_string(pid) + ".sms.msgpack";
}

AvlogSmsLogger::AvlogSmsLogger(const std::string& name, bool async)
    : writer_(createFileName(name), async)
{
    UniformMetadata uniformMetadata;
    uniformMetadata.processID = static_cast<uint32_t>(getpid());
    uniformMetadata.processName = name;
    nlohmann::json msg = createUniformMetadataEvent(uniformMetadata);
    writer_.log(msg);
}

void AvlogSmsLogger::log(const av::sms::TraceEvent& message)
{
    thread_local nlohmann::json traceEventMsg(TraceEventMinimumEntryCount, nlohmann::json{});

    TraceEventInternCollection ids = memoizeTraceEventData(message);
    populateTraceEvent(traceEventMsg, message, ids);
    writer_.log(traceEventMsg);
}

void AvlogSmsLogger::warn(const std::string& message)
{
    nlohmann::json dst = createMessageEvent(message, "warning");
    writer_.log(dst);
}

void AvlogSmsLogger::info(const std::string& message)
{
    nlohmann::json dst = createMessageEvent(message, "info");
    writer_.log(dst);
}

TraceEventInternCollection AvlogSmsLogger::memoizeTraceEventData(const av::sms::TraceEvent& message)
{
    thread_local nlohmann::json internStringMsg{{0, 0}}; // Array of size 2
    TraceEventInternCollection ids;

    bool writeFileName{false};
    bool writeFunctionName{false};
    bool writeCategory{false};
    bool writeLabel{false};
    bool writeName{false};

    // Logic is divided here in an unintuitive manner as to explicitly scope the block of code where the intern lock
    // is held. This is to help mitigate contention.
    {
        std::lock_guard lock{tableLock_};

        // Wrap up state into lambda to avoid boilerplate
        // Returns the intern ID for the string, adding into the table if necessary.
        // Also returns if the string requires an `InternEvent` to be written to disk (i.e., was just added to table)
        auto findInternID = [this](const std::string& value) -> auto
        {
            decltype(internTable_)::iterator iter;
            bool needsInternEvent;
            std::tie(iter, needsInternEvent) = internTable_.try_emplace(value, nextInternId_);
            if (needsInternEvent) {
                nextInternId_++;
            }
            return std::make_tuple(iter->second, needsInternEvent);
        };

        std::tie(ids.fileNameID, writeFileName) = findInternID(message.fileName);
        std::tie(ids.functionNameID, writeFunctionName) = findInternID(message.functionName);
        std::tie(ids.categoryID, writeCategory) = findInternID(message.category);
        std::tie(ids.labelID, writeLabel) = findInternID(message.label);
        std::tie(ids.nameID, writeName) = findInternID(message.name);
    }

    // Sink the intern string and ID if this is the first time it was encountered.
    // Note: syscall may or may not be invoked in this lambda, which is a bad thing if we're holding the lock from above
    auto sinkInternIfNeeded = [this](bool needsWrite, uint32_t internID, const std::string& val) {
        if (needsWrite) {
            populateInternEvent(internStringMsg, internID, val);
            writer_.log(internStringMsg);
        }
    };

    sinkInternIfNeeded(writeFileName, ids.fileNameID, message.fileName);
    sinkInternIfNeeded(writeFunctionName, ids.functionNameID, message.functionName);
    sinkInternIfNeeded(writeCategory, ids.categoryID, message.category);
    sinkInternIfNeeded(writeLabel, ids.labelID, message.label);
    sinkInternIfNeeded(writeName, ids.nameID, message.name);

    return ids;
}

} // namespace sms
} // namespace av
