/**
 * @copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 */

#include <fstream>

#include <avtime/ConversionsUsec.hpp>
#include <spdlog/sinks/ostream_sink.h>

#include "sms/SmsLogger.hpp"
#include "sms/TraceEvent.hpp"

namespace av {
namespace sms {

static av::TimePoint floorToStaggeredWindow(const av::TimePoint& time, int64_t windowSize)
{
    using namespace std::chrono;
    auto floorSeconds = floor<seconds>(time);

    // Turn the time (in [ms]) into 1000/windowSize buckets
    int64_t bucket = floor<milliseconds>(time).time_since_epoch().count() / windowSize;
    return floorSeconds + milliseconds(bucket * windowSize);
}

bool SmsLogger::checkRateLimit(av::TimePoint timePoint)
{
    if (rateLimit_ == 0) {
        // zero means unlimited
        return true;
    }
    if (disabled_.load(std::memory_order_acquire)) {
        // fuse already popped
        // remove to throttle logging for this second only
        return false;
    }

    // Determine the limit of messages allowed in the window by finding how many windows can be partitioned
    // into any 1 second time frame. This is necessary as the limit rate is in units of `messages/sec`.
    const uint32_t windowPartitionsPerSecond = 1000 / windowSizeMs();
    uint32_t limitPerWindow = rateLimit_ / windowPartitionsPerSecond;

    const av::TimePoint timeBucket = floorToStaggeredWindow(timePoint, static_cast<int64_t>(windowSizeMs()));

    std::scoped_lock lock(mutex_);
    if (lastTimeBucket_ == timeBucket) {
        ++logCounter_;
        if (logCounter_ >= limitPerWindow) {
            disabled_ = true; // pop the fuse
            if (logCounter_ == limitPerWindow) {
                warn("Rate limit exceeded. AV_TRACE logging disabled.");
                return false;
            }
        }
    } else {
        lastTimeBucket_ = timeBucket;
        logCounter_ = 0;
    }
    return true;
}

} // namespace sms
} // namespace av
