#include <avtime/ConversionsUsec.hpp>
#include <sms/SmsSchema.hpp>

#include <stdexcept>

namespace av::sms {

nlohmann::json createUniformMetadataEvent(const UniformMetadata& src)
{
    nlohmann::json dst;
    dst["metadata"] = {
        {"processID", src.processID}, {"processName", src.processName}, {"schemaVersion", src.schemaVersion}};
    return dst;
}

nlohmann::json createMessageEvent(const std::string& message, const std::string& severity)
{
    return {{"message", message}, {"severity", severity}};
}

void populateInternEvent(nlohmann::json& dst, uint32_t internID, const std::string& value)
{
    dst[0] = internID;
    dst[1] = value;
}

void populateTraceEvent(nlohmann::json& dst, const TraceEvent& src, const TraceEventInternCollection& ids)
{
    if (dst.size() < TraceEventMinimumEntryCount) {
        throw std::invalid_argument("Trace event array was too small");
    }

    const auto startTime = usecFromTimePoint(src.startTime);
    const auto endTime = usecFromTimePoint(src.endTime);

    dst[StartTimeIndex] = startTime;
    dst[EndTimeIndex] = endTime - startTime;

    // If threadID == processID, assign to 0 to save bytes.
    if (src.processId != src.threadId) {
        dst[ThreadIdIndex] = src.threadId;
    } else {
        dst[ThreadIdIndex] = 0;
    }

    dst[FileNameIndex] = ids.fileNameID;
    dst[FunctionNameIndex] = ids.functionNameID;
    dst[LineNumberIndex] = src.lineNumber;
    dst[CategoryIndex] = ids.categoryID;
    dst[LabelIndex] = ids.labelID;
    dst[CycleCountIndex] = src.cycleCount;
    dst[NameIndex] = ids.nameID;
    dst[ValueTypeIndex] = static_cast<uint8_t>(src.valueType);

    const auto writeValue = [&](const auto& value) { dst[ValueIndex] = value; };
    switch (src.valueType) {
        case ValueType::Timing:
            writeValue(0);
            break;
        case ValueType::Int64:
            writeValue(src.intValue);
            break;
        case ValueType::Uint64:
            writeValue(src.uIntValue);
            break;
        case ValueType::Float32:
            writeValue(src.floatValue);
            break;
        case ValueType::Float64:
            writeValue(src.doubleValue);
            break;
        case ValueType::String:
            writeValue(src.stringValue);
            break;
        case ValueType::Boolean:
            writeValue(src.boolValue);
            break;
    }
}

UniformMetadata extractUniformMetadataEvent(const nlohmann::json& src)
{
    UniformMetadata dst;
    // Expect an object with at least 3 keys within a `metadata` key
    const auto& metadata = src["metadata"];
    dst.processID = metadata["processID"];
    dst.processName = metadata["processName"];
    dst.schemaVersion = metadata["schemaVersion"];
    return dst;
}

InternString extractInternEvent(const nlohmann::json& src)
{
    // Expect an array with exactly 2 elements
    if (src.is_array() && src.size() == 2) {
        InternString dst;
        dst.id = src[0];
        dst.value = src[1];
        return dst;
    }

    throw std::invalid_argument("Invalid intern event input");
}

void extractTraceEvent(nlohmann::json& dst,
                       const nlohmann::json& src,
                       const std::unordered_map<uint32_t, std::string>& internTable,
                       const UniformMetadata& metadata)
{
    // Expect an array with at least N elements
    if (!src.is_array()) {
        throw std::invalid_argument("Expected array input for trace event");
    }

    if (src.size() < TraceEventMinimumEntryCount) {
        throw std::invalid_argument("Trace event array was too small");
    }

    // Ensure all interned strings have an entry in the table
    uint32_t fileNameID = src[FileNameIndex];
    uint32_t functionNameID = src[FunctionNameIndex];
    uint32_t categoryID = src[CategoryIndex];
    uint32_t labelID = src[LabelIndex];
    uint32_t nameID = src[NameIndex];

    uint64_t startTime = src[StartTimeIndex];
    uint64_t endTime = src[EndTimeIndex];

    dst["startTime"] = startTime;
    dst["endTime"] = startTime + endTime; // endTime encoded as relative offset
    dst["fileName"] = internTable.at(fileNameID);
    dst["functionName"] = internTable.at(functionNameID);
    dst["lineNumber"] = src[LineNumberIndex];
    dst["category"] = internTable.at(categoryID);
    dst["label"] = internTable.at(labelID);
    dst["cycleCount"] = src[CycleCountIndex];
    dst["name"] = internTable.at(nameID);
    dst["processId"] = metadata.processID;
    dst["processName"] = metadata.processName;

    if (src[ThreadIdIndex] != 0) {
        dst["threadId"] = src[ThreadIdIndex];
    } else {
        dst["threadId"] = metadata.processID;
    }

    auto valueType = static_cast<ValueType>(src[ValueTypeIndex]);
    dst["valueType"] = valueType;

    // Future improvement: see if DP team can live without all these fields set unconditionally..
    dst["intValue"] = 0;
    dst["uIntValue"] = 0;
    dst["floatValue"] = 0.f;
    dst["doubleValue"] = 0.0;
    dst["stringValue"] = "";
    dst["boolValue"] = false;

    switch (valueType) {
        case ValueType::Timing:
            // Nothing
            break;
        case ValueType::Int64:
            dst["intValue"] = static_cast<int64_t>(src[ValueIndex]);
            break;
        case ValueType::Uint64:
            dst["uIntValue"] = static_cast<uint64_t>(src[ValueIndex]);
            break;
        case ValueType::Float32:
            dst["floatValue"] = static_cast<float>(src[ValueIndex]);
            break;
        case ValueType::Float64:
            dst["doubleValue"] = static_cast<double>(src[ValueIndex]);
            break;
        case ValueType::String:
            dst["stringValue"] = static_cast<std::string>(src[ValueIndex]);
            break;
        case ValueType::Boolean:
            dst["boolValue"] = static_cast<bool>(src[ValueIndex]);
            break;
        default: {
            throw std::invalid_argument("Invalid valueType enum encountered for trace event");
        }
    }
}

} // namespace av::sms
