/**
 * @copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 */

#ifndef SMS_HPP
#define SMS_HPP

#include <sms/TraceClient.hpp>

///
/// @brief AV_TRACE_SCOPED_PERF Logs a @ref av::sms::TraceEvent message at the end of the current scope, populated
/// with the duration of time to the end of the scope (see @ref TraceEvent::startTime and @ref TraceEvent::endTime).
/// @param NAME User definable name to populate in @ref TraceEvent::name.
/// @param ENUM_CATEGORY Enumeration value representing the type of the event being logged.
/// @param USER_LABEL User defined label assigned to the event being logged.
/// @note The enumeration requires an @ref base::EnumTrait specialization.
///
#define AV_TRACE_SCOPED_PERF(NAME, ENUM_CATEGORY, USER_LABEL)                                                          \
    AV_TRACE_SCOPED_PERF_UNIQUE(ENUM_CATEGORY, USER_LABEL, NAME, -1, __COUNTER__)

///
/// @brief AV_TRACE_SCOPED_PERF_CYCLE Same as @ref AV_TRACE_SCOPED_PERF, but with an additional CYCLE option.
/// @param NAME User definable name to populate in @ref TraceEvent::name.
/// @param ENUM_CATEGORY Enumeration value representing the type of the event being logged.
/// @param USER_LABEL User defined label assigned to the event being logged.
/// @note The enumeration requires an @ref base::EnumTrait specialization.
/// @param CYCLE Defines the current cycle within the task's execution frequency.  Populated
/// in @ref TraceEvent::cycleCount of the logged message.
///
#define AV_TRACE_SCOPED_PERF_CYCLE(NAME, ENUM_CATEGORY, USER_LABEL, CYCLE)                                             \
    AV_TRACE_SCOPED_PERF_UNIQUE(ENUM_CATEGORY, USER_LABEL, NAME, CYCLE, __COUNTER__)

///
/// @brief AV_TRACE_BEGIN Logs a @ref av::sms::TraceEvent message either:
///   - At the end of the current scope
///   - Upon call to @ref AV_TRACE_END
/// Message contains the duration of time to the end of the scope
/// (see @ref TraceEvent::startTime and @ref TraceEvent::endTime).
/// @param NAME A unique identifier for the specified trace, required by @ref AV_TRACE_END.
/// @param ENUM_CATEGORY Enumeration value representing the type of the event being logged.
/// @param USER_LABEL User defined label assigned to the event being logged.
/// @note The enumeration requires an @ref base::EnumTrait specialization.
///
#define AV_TRACE_BEGIN(NAME, ENUM_CATEGORY, USER_LABEL)                                                                \
    AV_TRACE_SCOPED_PERF_UNIQUE(ENUM_CATEGORY, USER_LABEL, NAME, -1, NAME)

///
/// @brief AV_TRACE_BEGIN Same as @ref AV_TRACE_BEGIN, but with an additional CYCLE option.
/// @param NAME A unique identifier for the specified trace, required by @ref AV_TRACE_END.
/// @param ENUM_CATEGORY Enumeration value representing the type of the event being logged.
/// @param USER_LABEL User defined label assigned to the event being logged.
/// @note The enumeration requires an @ref base::EnumTrait specialization.
/// @param CYCLE Defines the current cycle within the task's execution frequency.  Populated
/// in @ref TraceEvent::cycleCount of the logged message.
///
#define AV_TRACE_BEGIN_CYCLE(NAME, ENUM_CATEGORY, USER_LABEL, CYCLE)                                                   \
    AV_TRACE_SCOPED_PERF_UNIQUE(ENUM_CATEGORY, USER_LABEL, NAME, CYCLE, NAME)

///
/// @brief AV_TRACE_END Immediately logs a @ref av::sms::TraceEvent message for the specified trace, without waiting
/// for the end of the current scope.
/// @param NAME The unique identifier specified when starting the trace with @ref AV_TRACE_BEGIN[_CYCLE]
///
#define AV_TRACE_END(NAME) client##NAME.end()

///
/// @brief AV_TRACE_LOG_INSTANT Immediately logs a @ref av::sms::TraceEvent message, populating the message
/// with the specified value.
/// @param NAME User definable name to populate in @ref TraceEvent::name.
/// @param ENUM_CATEGORY Enumeration value representing the type of the event being logged.
/// @param USER_LABEL User defined label assigned to the event being logged.
/// @note The enumeration requires an @ref base::EnumTrait specialization.
/// @param VALUE the value to log.  See @ref av::sms::ValueType for the list of loggable value types.
///
#define AV_TRACE_LOG_INSTANT(NAME, ENUM_CATEGORY, USER_LABEL, VALUE)                                                   \
    AV_TRACE_SCOPED_PERF_UNIQUE(ENUM_CATEGORY, USER_LABEL, NAME, -1, NAME);                                            \
    client##NAME.setValue(VALUE);                                                                                      \
    client##NAME.end(false)

///
/// @brief AV_TRACE_LOG_INSTANT_CYCLE Same as @ref AV_TRACE_LOG_INSTANT, but with an additional CYCLE option.
/// @param NAME User definable name to populate in @ref TraceEvent::name.
/// @param ENUM_CATEGORY Enumeration value representing the type of the event being logged.
/// @param USER_LABEL User defined label assigned to the event being logged.
/// @note The enumeration requires an @ref base::EnumTrait specialization.
/// @param CYCLE Defines the current cycle within the task's execution frequency.  Populated
/// in @ref TraceEvent::cycleCount of the logged message.
/// @param VALUE the value to log.  See @ref av::sms::ValueType for the list of loggable value types.
///
#define AV_TRACE_LOG_INSTANT_CYCLE(NAME, ENUM_CATEGORY, USER_LABEL, CYCLE, VALUE)                                      \
    AV_TRACE_SCOPED_PERF_UNIQUE(ENUM_CATEGORY, USER_LABEL, NAME, CYCLE, NAME);                                         \
    client##NAME.setValue(VALUE);                                                                                      \
    client##NAME.end(false)

#endif // SMS_HPP
