/**
 * @copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 */

#ifndef SMS_LOGGER_H
#define SMS_LOGGER_H

#include "lib/avlog/DataLogger.hpp"
#include <avtime/AvClock.hpp>
#include <sms/SmsSchema.hpp>

#include <atomic>
#include <fstream>
#include <memory>
#include <mutex>

namespace av::sms {
class TraceEvent;
}

namespace av {
namespace sms {

/// @brief SmsLogger class
class SmsLogger
{
public:
    virtual ~SmsLogger() = default;

    /// @brief log Shall output the specified message.
    /// @param message Populated @ref TraceEvent message.
    virtual void log(const av::sms::TraceEvent& message) = 0;

    /// @brief log Shall output the specified message.
    /// @param message string to log as a warning.
    virtual void warn(const std::string& message) = 0;

    /// @brief Log a general message
    /// @param message string to log as info.
    virtual void info(const std::string& message) = 0;

    /// @brief Sets the value for the `processName` attribute in each trace event
    void setProcessName(std::string name) { processName_ = std::move(name); }

    /// @brief Retrieves the `processName` attribute of this logger
    const std::string& getProcessName() const { return processName_; }

    /// @brief Sets the limit of how many trace events are allowed to be logged per second. In units of `events/sec`
    void setRateLimit(uint32_t limit) { rateLimit_ = limit; }

    /// @brief Retrieves the rate limit for trace events. In units of `events/sec`
    uint32_t getRateLimit() const { return rateLimit_; }

    /// @brief returns true if a trace event can be logged while respecting the imposed rate limit policy.
    bool checkRateLimit(av::TimePoint timePoint);

    /// @brief returns the window size used to sample the rate at which logging occurs
    static constexpr uint64_t windowSizeMs() { return WindowSizeMs; }

private:
    static constexpr uint64_t WindowSizeMs = 250;

    std::string processName_{};
    uint32_t rateLimit_;             ///< rate limit [events/sec]
    std::atomic_bool disabled_{};    ///< rate limit exceeded
    std::mutex mutex_;               ///< mutex to protect rate limiter data
    av::TimePoint lastTimeBucket_{}; ///< time bucket used for rate limit
    uint32_t logCounter_{};          ///< num of events counted in current time bucket
};

/// @brief AvlogSmsLogger class
class AvlogSmsLogger : public SmsLogger
{
public:
    /// @brief AvlogSmsLogger Initializes an AV logger to write @ref TraceEvent messages to.
    /// The initialized logger writes to "name-<pid>.sms.msgpack".
    /// @param name Task/process name.
    /// @param async true if the logging should be aynschronous
    explicit AvlogSmsLogger(const std::string& name, bool async = true);

    /// @brief log writes the specified message to AV log.
    /// @param message The @ref TraceEvent message to write.
    void log(const av::sms::TraceEvent& message) override;

    /// @brief log writes the specified message to AV log.
    /// @param message string to log as a warning.
    void warn(const std::string& message) override;

    /// @brief Writes the specified message to AV log
    /// @param message string to log as info.
    void info(const std::string& message) override;

private:
    /// Retrieve the intern IDs for the strings within the given TraceEvent.
    /// Sinks `InternEvent` messages and populates them within the intern table
    /// if a new string is encountered.
    TraceEventInternCollection memoizeTraceEventData(const av::sms::TraceEvent& message);

    avlog::DataLogger writer_;
    std::mutex tableLock_;
    uint32_t nextInternId_{0};
    std::unordered_map<std::string, uint32_t> internTable_;
};

} // namespace sms
} // namespace av

#endif // SMS_LOGGER_H
