#pragma once

#include <sms/TraceClient.hpp>

#include <nlohmann/json.hpp>

#include <unordered_map>

namespace av::sms {

// Field offsets within TraceEvent message
// `TraceEvent` is serialized as an array as to reduce
// the serialized size of the message.
inline constexpr unsigned int StartTimeIndex = 0U;
inline constexpr unsigned int EndTimeIndex = 1U;
inline constexpr unsigned int ThreadIdIndex = 2U;
inline constexpr unsigned int FileNameIndex = 3U;
inline constexpr unsigned int FunctionNameIndex = 4U;
inline constexpr unsigned int LineNumberIndex = 5U;
inline constexpr unsigned int CategoryIndex = 6U;
inline constexpr unsigned int LabelIndex = 7U;
inline constexpr unsigned int CycleCountIndex = 8U;
inline constexpr unsigned int NameIndex = 9U;
inline constexpr unsigned int ValueTypeIndex = 10U;
inline constexpr unsigned int ValueIndex = 11U;
// The minimum number of elements composing the serialized array for a `TraceEvent`
inline constexpr unsigned int TraceEventMinimumEntryCount = 12U;
inline constexpr unsigned int SchemaVersion = 1U;

struct UniformMetadata
{
    uint32_t processID{};
    std::string processName{};
    uint32_t schemaVersion{};
};

struct InternString
{
    uint32_t id{};
    std::string value{};
};

// Collection of TraceEvent intern ID markers
struct TraceEventInternCollection
{
    uint32_t fileNameID{};
    uint32_t functionNameID{};
    uint32_t categoryID{};
    uint32_t labelID{};
    uint32_t nameID{};
};

/// @brief Convert a UniformMetadata to the json object representation expected by `DataLogger`.
/// @param src Metdata to be converted to JSON form.
nlohmann::json createUniformMetadataEvent(const UniformMetadata& src);

/// @brief Create a JSON representation of the provided message and severity
/// @param message Textual representation of the message
/// @param severity Severity level, in textual form
nlohmann::json createMessageEvent(const std::string& message, const std::string& severity);

/// @brief Convert an intern'd string to the json array representation expected by `DataLogger`.
/// @param dst The output json array to be processed by `DataLogger`. Expected to be an array of size 2.
/// @param uint32_t Unsigned and unique ID to be written to the log entry
/// @param value The string to be represented by the ID
void populateInternEvent(nlohmann::json& dst, uint32_t internID, const std::string& value);

/// @brief Convert a TraceEvent message to the compressed JSON array representation expected by `DataLogger`.
/// @param dst The output json array to be processed by `DataLogger`. Expected to be an array of at least size
/// `TraceEventMinimumEntryCount`.
/// @param src TraceEvent message populated by user data
/// @param ids Collection of IDs which represent strings placed within the intern table in the log file
void populateTraceEvent(nlohmann::json& dst, const TraceEvent& src, const TraceEventInternCollection& ids);

/// @brief Convert the JSON representation of metadata into its structural form
/// @param src A JSON object with the schema
UniformMetadata extractUniformMetadataEvent(const nlohmann::json& src);

/// @brief Convert the JSON array representation of an intern'd string into its structured form.
/// @param src A JSON array of exactly size 2
InternString extractInternEvent(const nlohmann::json& src);

/// @brief Convert the compressed JSON array representation of the `TraceEvent` into something human+tooling friendly.
/// @param dst Human and tooling friendly JSON representation of the TraceEvent
/// @param src Compresed JSON array representation of the underlying TraceEvent. Matches the schema from the function
/// populateTraceEvent.
/// @param internTable Snapshot of the intern table at the point in time this TraceEvent was encountered within the log
/// file. All intern IDs from `src` are expected to be present in this parameter.
/// @param metadata Metadata for the current trace event
void extractTraceEvent(nlohmann::json& dst,
                       const nlohmann::json& src,
                       const std::unordered_map<uint32_t, std::string>& internTable,
                       const UniformMetadata& metadata);

} // namespace av::sms
