/**
 * @copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 */

#ifndef TRACE_CLIENT_HPP
#define TRACE_CLIENT_HPP

#include <atomic>
#include <memory>
#include <mutex>
#include <sys/syscall.h>
#include <sys/types.h>
#include <unistd.h>

#include <avtime/AvClock.hpp>

#include "lib/base/enum/Utility.hpp"
#include "lib/base/typeInfo/TypeName.hpp"
#include "lib/base/typeTraits/TypeTraits.hpp"
#include "lib/base/utility/Cast.hpp"
#include "platform/middleware/libs/sms/TraceEventValue.hpp"
#include "sms/TraceEvent.hpp"

namespace av {
namespace sms {

class SmsLogger;

/// @brief TraceClient class Writes @ref TraceEvent messages to a user specified sink.  @ref TraceClient
/// can be initialized to either write @ref TraceEvent messages to a log file via @ref TraceClient::initializeWithAvLog
/// or publish @ref TraceEvent messages over SCS via @ref TraceClient::initializeWithScs.  Beyond initialization,
/// this class is not intended to be used directly.  Refer to the AV_TRACE_* macros in Sms.h for usage details.
class TraceClient
{
public:
    /// @brief initialize Initializes all @ref TraceClient objects to publish @ref TraceEvent messages to
    /// the specified publisher.
    /// @param publisher Used to publish @ref TraceEvent messages.
    /// @param name Name of the process or task running.
    /// @param limit Number of events per second allowed. Default = 1000
    static void initialize(std::unique_ptr<SmsLogger> publisher, const std::string& name, uint32_t limit = 1000);

    /// @brief TraceClient Not to be used directly.  See AV_TRACE_* macros in Sms.h
    explicit TraceClient(TraceEvent& message);

    /// @brief TraceClient Not to be used directly. For advanced use cases where a custom logger is required.
    TraceClient(TraceEvent& message, SmsLogger* publisher);

    /// @brief ~TraceClient If not already published, compiles final metrics and publishes result.
    ~TraceClient();

    /// @brief setValue Sets the appropriate @ref TraceEvent member data for the specified value and type.
    /// See @ref av::sms::ValueTypes for the list of valid types allowed.
    /// @param value Value to populate in @ref TraceEvent message.
    template<typename T>
    void setValue(const T& value)
    {
        TraceEventValue<T>::set(message_, value);
    }

    /// @brief end Compiles final metrics (end time, final value) and publishes/logs result.
    /// @note multiple calls are ignored (only first call is published/logs)
    /// @param trackEndTime Specifies if the end time should be unique (not equal to the start time)
    void end(bool trackEndTime = true);

    /// @brief returns rateLimit value set in initialize()
    static uint32_t rateLimit();

private:
    void initializeMessage();

    static constexpr uint64_t WindowSizeMs = 250;

    TraceEvent& message_;
    SmsLogger* logger_;
    bool ended_{false};
    static std::unique_ptr<SmsLogger> s_defaultPublisher;
};

template<typename T>
inline std::string getEnumCategory(const T enumCategory)
{
    auto result = std::string{base::typeName<base::RemoveCVRefT<decltype(enumCategory)>>()} + "::";
    if (auto const maybeName = base::enumeratorName(enumCategory); maybeName) {
        result.append(std::string{*maybeName});
    } else {
        result.append(std::string("Error[") + std::to_string(base::toUnderlying(enumCategory)) + "]");
    }
    return result;
}

#define AV_TRACE_SCOPED_PERF_UNIQUE_IMPL(ENUM_CATEGORY, USER_LABEL, VAR_NAME, CYCLE, NAME)                             \
    av::sms::TraceEvent e##NAME{::av::TimePoint{},                                                                     \
                                ::av::TimePoint{},                                                                     \
                                "",                                                                                    \
                                0 /* pid */,                                                                           \
                                0 /* tid */,                                                                           \
                                __FILE__,                                                                              \
                                static_cast<const char*>(__PRETTY_FUNCTION__),                                         \
                                __LINE__,                                                                              \
                                ::av::sms::getEnumCategory(ENUM_CATEGORY),                                             \
                                USER_LABEL,                                                                            \
                                CYCLE,                                                                                 \
                                ::av::sms::ValueType::Timing,                                                          \
                                #VAR_NAME,                                                                             \
                                0,                                                                                     \
                                0U,                                                                                    \
                                0.0F,                                                                                  \
                                0.0,                                                                                   \
                                "",                                                                                    \
                                false};                                                                                \
    ::av::sms::TraceClient client##NAME(e##NAME)

#define AV_TRACE_SCOPED_PERF_UNIQUE(ENUM_CATEGORY, USER_LABEL, VAR_NAME, CYCLE, NAME)                                  \
    AV_TRACE_SCOPED_PERF_UNIQUE_IMPL(ENUM_CATEGORY, USER_LABEL, VAR_NAME, CYCLE, NAME)

} // namespace sms
} // namespace av

#endif // TRACE_CLIENT_HPP
