#include "SmsIngestionReader.hpp"

#include <iostream>

static void printHelp();
static void dumpJson(const nlohmann::json& events, std::ostream& output, bool prettyPrint);

int main(int argc, char** argv)
{

    std::string inFileName{};
    std::string outFileName{};
    bool loggerMessageOutput{false};
    bool prettyPrint{false};

    int c;
    while ((c = getopt(argc, argv, "i:mo:p")) != -1) {
        switch (c) {
            case 'i':
                inFileName = optarg;
                break;
            case 'm':
                loggerMessageOutput = true;
                break;
            case 'o':
                outFileName = optarg;
                break;
            case 'p':
                prettyPrint = true;
                break;
            case '?':
            default:
                printHelp();
                return 1;
        }
    }

    av::sms::SmsIngestionReader reader(inFileName, loggerMessageOutput);
    auto msgs = reader.read();

    if (outFileName.empty()) {
        dumpJson(msgs, std::cout, prettyPrint);
    } else {
        std::ofstream outFileStream{outFileName};
        if (!outFileStream.is_open()) {
            std::cerr << "Failed to open file " << outFileName << std::endl;
            return 2;
        }

        dumpJson(msgs, outFileStream, prettyPrint);
    }

    return 0;
}

static void printHelp()
{
    std::cout << "sms-ingest: converts SMS log files to JSON." << std::endl;
    std::cout << "  Returns 0 if the conversion was successful and had no errors, nonzero otherwise." << std::endl;
    std::cout << "  If some messages are unable to convert it may produce partial output; please check the return code "
                 "to verify that no errors occurred."
              << std::endl;
    std::cout << "  Error messages will print on stderr." << std::endl;
    std::cout << "  JSON will be printed to standard output or file (if specified)." << std::endl;
    std::cout << "Usage:" << std::endl;
    std::cout << "  -i <filename>" << std::endl;
    std::cout << "    Specifies input file name" << std::endl;
    std::cout << "  -m" << std::endl;
    std::cout << "    Output info/warning messages from SMS (Disabled by default)" << std::endl;
    std::cout << "  -o <filename>" << std::endl;
    std::cout << "    Specifies output file name. If not specified, prints to standard output." << std::endl;
    std::cout << "  -p" << std::endl;
    std::cout << "    Pretty print JSON (Disabled by default)" << std::endl;
}

static void dumpJson(const nlohmann::json& events, std::ostream& output, bool prettyPrint)
{

    if (prettyPrint) {
        if (events.empty()) {
            // Short circuit out to simplify logic below
            output << "[]\n";
        }

        // Adhere to JSON spec which does not allow trailing comma at end of array.
        // Python JSON module refuses to parse JSON with a trailing comma.
        output << "[\n";

        size_t len = events.size();
        for (size_t i = 0; i < len - 1; i++) {
            output << events[i].dump(2) << ",\n";
        }
        if (!events.empty()) {
            output << events[len - 1].dump(2) << "\n";
        }

        output << "]\n";
    } else {
        output << events;
    }
}
