#include "lib/base/enum/Trait.hpp"

#include <sms/Sms.hpp>
#include <sms/SmsLogger.hpp>

#include <iostream>
#include <string>
#include <thread>
#include <vector>

struct ProgramArgs
{
    size_t entryCount;
    size_t threadCount;
};

enum class CustomEnumLabels
{
    Foo,
};

template<>
struct base::EnumTrait<CustomEnumLabels>
{
    static constexpr std::array namedEnumerators = {
        base::Enumerator{CustomEnumLabels::Foo, "Foo"},
    };
};

static void initSms()
{
    av::sms::TraceClient::initialize(
        std::make_unique<av::sms::AvlogSmsLogger>("avlogSmsBenchmark"), "avlogSmsBenchmark", 0);
}

static ProgramArgs parseArgs(int argc, const char** argv)
{
    size_t entryCount{0};
    size_t threadCount{0};

    if (argc != 3) {
        throw std::invalid_argument("3 arguments must be provided");
    }

    entryCount = std::stoull(argv[1]);
    threadCount = std::stoull(argv[2]);

    if (entryCount < 1 or threadCount < 1) {
        throw std::invalid_argument("entry count and thread count must be greater than 0");
    }

    if (threadCount > entryCount) {
        throw std::invalid_argument("entry count must be greater than thread count");
    }

    return {entryCount, threadCount};
}

static void worker(size_t count)
{
    size_t counter{1};
    while (counter < count) {
        AV_TRACE_LOG_INSTANT(Instant1, CustomEnumLabels::Foo, "Label", static_cast<int64_t>(counter));
        counter++;
    }
}

static auto run(const ProgramArgs& args)
{
    std::vector<std::thread> threads;
    threads.reserve(args.threadCount);

    const size_t threadWorkerCount{args.entryCount / args.threadCount};
    std::chrono::steady_clock::time_point start = std::chrono::steady_clock::now();

    for (size_t i = 0; i < args.threadCount - 1; i++) {
        threads.emplace_back([=]() { worker(threadWorkerCount); });
    }

    // In case entry count is not divisible by thread count, add the remainder onto the last thread.
    threads.emplace_back([=]() { worker(threadWorkerCount + args.entryCount % args.threadCount); });

    for (auto& thread : threads) {
        thread.join();
    }

    return std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() - start).count();
}

static void printUsage(const char* cmdName)
{
    std::cerr << "Usage:\n"
              << cmdName << " entryCount threadCount\n"
              << "entryCount: the number of trace events to capture\n"
              << "threadCount: Number of threads on which to evenly distribute the entryCount trace events"
              << std::endl;
}

int main(int argc, const char** argv)
{
    initSms();
    ProgramArgs args;

    try {
        args = parseArgs(argc, argv);
    } catch (const std::invalid_argument& e) {
        printUsage(argv[0]);
        return 1;
    } catch (const std::exception& e) {
        std::cerr << e.what() << std::endl;
        return 1;
    }

    auto ms = run(args);
    std::cout << ms << std::endl;

    return 0;
}
