#!/usr/bin/env python3

import os
import shutil
import subprocess

MILLION = 1000000


def get_results_file(working_dir, pid):
    result_file = os.path.join(working_dir, f"avlogSmsBenchmark-{pid}.sms.msgpack")
    if os.path.isfile(result_file):
        return result_file, os.path.getsize(result_file)
    raise RuntimeError(f"Expected results file at {result_file}")


def run_benchmark(entry_count, thread_count, cwd, exec_command):
    args = [exec_command, str(entry_count), str(thread_count)]
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, cwd=cwd)

    outs, errs = proc.communicate()
    proc_results = str(outs.decode("utf8"))
    returncode = proc.returncode
    pid = proc.pid

    if returncode != 0:
        raise RuntimeError(f"process {args} returned with code {returncode}")

    result_file, result_file_size = get_results_file(cwd, pid)
    result_time = int(proc_results)  # [ms]

    return {
        "file": result_file,
        "fileSize": result_file_size,
        "duration": result_time,
        "entryCount": entry_count,
        "threadCount": thread_count,
    }


class BenchmarkSet:
    def __init__(self, runs, cwd, exec_command):
        self.__runs = runs
        self.__cwd = cwd
        self.__exec_command = exec_command

    def run(self):
        results = list()
        cwd = self.__cwd
        exec_command = self.__exec_command

        if shutil.which(exec_command) is None:
            raise RuntimeError(f"Command {exec_command} not in path")

        for run in self.__runs:
            entry_count = run[0]
            thread_count = run[1]
            results.append(run_benchmark(entry_count, thread_count, cwd, exec_command))

        return results


def display_results(results):
    # Display in tabular form

    print("fileSize [KB] | duration [ms] | entry count | thread count")
    for entry in results:
        file_size = entry["fileSize"] // 1000
        duration = entry["duration"]
        entry_count = entry["entryCount"]
        thread_count = entry["threadCount"]
        print(f"{file_size:<15} {duration:<15} {entry_count:<13} {thread_count:<13}")


runs = [
    (MILLION, 1),
    (MILLION, 2),
    (MILLION, 4),
    (MILLION, 8),
    (MILLION, 16),
    (MILLION, 24),
    (MILLION, 32),
]

cwd = os.getcwd()
exec_command = "avlogSmsBenchmark"

benchmark = BenchmarkSet(runs, cwd, exec_command)

try:
    display_results(benchmark.run())
except Exception as e:
    import sys

    sys.stderr.write(str(e) + "\n")
    sys.exit(1)
