/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 */

#include <functional>
#include <utility>
#include <vector>

#include <sms/Sms.hpp>
#include <sms/SmsLogger.hpp>

#include <avtime/AvClock.hpp>

#include <gmock/gmock.h>

using namespace av::sms;
using ::testing::_;
using ::testing::Invoke;

enum class FakeCategories : int32_t
{
    Speed = 0,
    Location
};

struct MockSmsLogger : av::sms::SmsLogger
{
    virtual ~MockSmsLogger() = default;
    MOCK_METHOD(void, log, (const av::sms::TraceEvent&));
    MOCK_METHOD(void, warn, (const std::string&));
    MOCK_METHOD(void, info, (const std::string&));
};

template<>
struct base::EnumTrait<FakeCategories>
{
    static constexpr std::array namedEnumerators = {
        base::Enumerator{FakeCategories::Speed, "Speed"},
        base::Enumerator{FakeCategories::Location, "Location"},
    };
};

template<typename T>
T anyValue = std::numeric_limits<T>::max();

static const std::string AnyName = "AnyName";

struct SmsFixture : public testing::Test
{
    void SetUp()
    {
        av::sms::TraceClient::initialize(std::unique_ptr<av::sms::SmsLogger>(mockPublisher), "TestTask", 100);
    }

    void TearDown() { av::sms::TraceClient::initialize(nullptr, "TestTask"); }

    ///
    /// \brief captureMessage Used in GMOCK calls to \ref MockGenericOutput::publish to capture the
    /// \ref av::TraceEvent message being published.  messages are pushed into capturedMessages member.
    ///
    bool captureMessage(const TraceEvent& message)
    {
        capturedMessages.push_back(message);
        return true;
    };

    bool captureInfoMessage(const std::string& message)
    {
        capturedInfoMessages.push_back(message);
        return true;
    };

    MockSmsLogger* mockPublisher = new MockSmsLogger;

    std::vector<TraceEvent> capturedMessages;
    std::vector<std::string> capturedInfoMessages;
};

TEST_F(SmsFixture, smsLogger_setterGetters)
{
    constexpr auto rateLimit = 42;
    const std::string processName = "NewTaskName";

    mockPublisher->setRateLimit(rateLimit);
    EXPECT_EQ(mockPublisher->getRateLimit(), rateLimit);

    mockPublisher->setProcessName(processName);
    EXPECT_EQ(mockPublisher->getProcessName(), processName);
}

TEST_F(SmsFixture, scopedPerf_publishesAtEndOfScope)
{
    {
        EXPECT_CALL(*mockPublisher, log(_)).Times(1);
        AV_TRACE_SCOPED_PERF(AnyName, FakeCategories::Speed, "MyLabel");
    }
    EXPECT_CALL(*mockPublisher, log(_)).Times(1);
    AV_TRACE_SCOPED_PERF(AnyName, FakeCategories::Speed, "MyLabel");
}

TEST_F(SmsFixture, dataCorrectlyPopulated)
{
    using Publisher = std::function<std::pair<int, const char*>()>;
    Publisher scopedPerf = [this]() -> std::pair<int, const char*> {
        EXPECT_CALL(*mockPublisher, log(_)).WillOnce(Invoke(this, &SmsFixture::captureMessage));
        AV_TRACE_SCOPED_PERF(AnyName, FakeCategories::Speed, "MyLabel");
        return {__LINE__ - 1, __PRETTY_FUNCTION__};
    };
    Publisher traceBeginEnd = [this]() -> std::pair<int, const char*> {
        EXPECT_CALL(*mockPublisher, log(_)).WillOnce(Invoke(this, &SmsFixture::captureMessage));
        AV_TRACE_BEGIN(AnyName, FakeCategories::Speed, "MyLabel");
        AV_TRACE_END(AnyName);
        return {__LINE__ - 2, __PRETTY_FUNCTION__};
    };

    for (auto& publish : {scopedPerf, traceBeginEnd}) {
        auto [expectedLineNumber, expectedFunctionName] = publish();
        EXPECT_EQ(__FILE__, capturedMessages.back().fileName);
        EXPECT_EQ(expectedFunctionName, capturedMessages.back().functionName);
        EXPECT_EQ(expectedLineNumber, capturedMessages.back().lineNumber);
        std::string category =
            std::string(base::typeName<FakeCategories>()) + "::" +
            std::string(
                base::EnumTrait<FakeCategories>::namedEnumerators[static_cast<int>(FakeCategories::Speed)].name);
        EXPECT_EQ(category, capturedMessages.back().category);
        EXPECT_EQ("MyLabel", capturedMessages.back().label);
        EXPECT_EQ(ValueType::Timing, capturedMessages.back().valueType);
        EXPECT_EQ("TestTask", capturedMessages.back().processName);
        EXPECT_EQ(AnyName, capturedMessages.back().name);
    }
}

TEST_F(SmsFixture, correctlyHandlesConstRefEnum)
{
    using Publisher = std::function<std::pair<int, const char*>()>;
    Publisher scopedPerf = [this]() -> std::pair<int, const char*> {
        EXPECT_CALL(*mockPublisher, log(_)).WillOnce(Invoke(this, &SmsFixture::captureMessage));
        auto const& speed = FakeCategories::Speed;
        AV_TRACE_SCOPED_PERF(AnyName, speed, "MyLabel");
        return {__LINE__ - 1, __PRETTY_FUNCTION__};
    };

    auto [expectedLineNumber, expectedFunctionName] = scopedPerf();
    EXPECT_EQ(__FILE__, capturedMessages.back().fileName);
    EXPECT_EQ(expectedFunctionName, capturedMessages.back().functionName);
    EXPECT_EQ(expectedLineNumber, capturedMessages.back().lineNumber);
    std::string category =
        std::string(base::typeName<FakeCategories>()) + "::" +
        std::string(base::EnumTrait<FakeCategories>::namedEnumerators[static_cast<int>(FakeCategories::Speed)].name);
    EXPECT_EQ(category, capturedMessages.back().category);
    EXPECT_EQ("MyLabel", capturedMessages.back().label);
    EXPECT_EQ(ValueType::Timing, capturedMessages.back().valueType);
    EXPECT_EQ("TestTask", capturedMessages.back().processName);
    EXPECT_EQ(AnyName, capturedMessages.back().name);
}

TEST_F(SmsFixture, handlesNonEnumeratedValue)
{
    using Publisher = std::function<std::pair<int, const char*>()>;
    Publisher scopedPerf = [this]() -> std::pair<int, const char*> {
        EXPECT_CALL(*mockPublisher, log(_)).WillOnce(Invoke(this, &SmsFixture::captureMessage));
        AV_TRACE_SCOPED_PERF(AnyName, FakeCategories{-1}, "MyLabel");
        return {__LINE__ - 1, __PRETTY_FUNCTION__};
    };

    auto [expectedLineNumber, expectedFunctionName] = scopedPerf();
    EXPECT_EQ(__FILE__, capturedMessages.back().fileName);
    EXPECT_EQ(expectedFunctionName, capturedMessages.back().functionName);
    EXPECT_EQ(expectedLineNumber, capturedMessages.back().lineNumber);
    std::string category = std::string(base::typeName<FakeCategories>()) + "::Error[-1]";
    EXPECT_EQ(category, capturedMessages.back().category);
    EXPECT_EQ("MyLabel", capturedMessages.back().label);
    EXPECT_EQ(ValueType::Timing, capturedMessages.back().valueType);
    EXPECT_EQ("TestTask", capturedMessages.back().processName);
    EXPECT_EQ(AnyName, capturedMessages.back().name);
}

TEST_F(SmsFixture, traceBegin_autoPublishesAtEndOfScope)
{
    EXPECT_CALL(*mockPublisher, log(_)).Times(1);
    AV_TRACE_BEGIN(AnyName, FakeCategories::Speed, "MyLabel");
}

TEST_F(SmsFixture, traceBeginEnd_onlyPublishesFirstEndRequest)
{
    // Verify only message from first AV_TRACE_END is published, even though two end (plus out of scope) occur
    EXPECT_CALL(*mockPublisher, log(_)).WillOnce(Invoke(this, &SmsFixture::captureMessage));
    AV_TRACE_BEGIN(AnyName, FakeCategories::Speed, "MyLabel");
    AV_TRACE_END(AnyName);
    AV_TRACE_END(AnyName);
}

TEST_F(SmsFixture, scopedPerf_multipleTracesAllowedInSameScope)
{
    int expectedLineNumber1 = 0, expectedLineNumber2 = 0;
    {
        EXPECT_CALL(*mockPublisher, log(_)).Times(2).WillRepeatedly(Invoke(this, &SmsFixture::captureMessage));
        expectedLineNumber1 = __LINE__ + 1;
        AV_TRACE_SCOPED_PERF(AnyName, FakeCategories::Speed, "MyLabel");
        expectedLineNumber2 = __LINE__ + 1;
        AV_TRACE_SCOPED_PERF(AnyName, FakeCategories::Location, "MyLabel");
    }

    EXPECT_EQ(capturedMessages.size(), 2U);
    EXPECT_EQ(capturedMessages[0].fileName, capturedMessages[1].fileName);
    EXPECT_EQ(capturedMessages[0].functionName, capturedMessages[1].functionName);
    EXPECT_EQ(expectedLineNumber1, capturedMessages[1].lineNumber);
    EXPECT_EQ(expectedLineNumber2, capturedMessages[0].lineNumber);
    EXPECT_EQ("FakeCategories::Speed", capturedMessages[1].category);
    EXPECT_EQ("FakeCategories::Location", capturedMessages[0].category);
    EXPECT_EQ(capturedMessages[0].label, capturedMessages[1].label);
    EXPECT_EQ(capturedMessages[0].valueType, capturedMessages[1].valueType);
}

TEST_F(SmsFixture, cycleCountDefaultsToNegativeOne)
{
    EXPECT_CALL(*mockPublisher, log(_)).Times(2).WillRepeatedly(Invoke(this, &SmsFixture::captureMessage));
    {
        AV_TRACE_SCOPED_PERF(AnyName, FakeCategories::Speed, "MyLabel");
    }
    AV_TRACE_BEGIN(AnyName, FakeCategories::Location, "MyLabel");
    AV_TRACE_END(AnyName);

    EXPECT_EQ(-1, capturedMessages[0].cycleCount);
    EXPECT_EQ(-1, capturedMessages[1].cycleCount);
}

TEST_F(SmsFixture, cycleCountSetToDefinedValue)
{
    using CycleCount = decltype(TraceEvent::cycleCount);
    auto publishTrace = [this](CycleCount cycleCount) -> void {
        EXPECT_CALL(*mockPublisher, log(_)).Times(2).WillRepeatedly(Invoke(this, &SmsFixture::captureMessage));
        AV_TRACE_SCOPED_PERF_CYCLE(AnyName, FakeCategories::Speed, "MyLabel", cycleCount);
        AV_TRACE_BEGIN_CYCLE(AnyName, FakeCategories::Location, "MyLabel", cycleCount);
        AV_TRACE_END(AnyName);
    };

    for (const CycleCount expectedCycleCount :
         {std::numeric_limits<CycleCount>::lowest(), std::numeric_limits<CycleCount>::max()}) {
        publishTrace(expectedCycleCount);
        EXPECT_EQ(expectedCycleCount, capturedMessages.front().cycleCount);
        EXPECT_EQ(expectedCycleCount, capturedMessages.back().cycleCount);
        capturedMessages.clear();
    }
}

TEST_F(SmsFixture, traceBeginEnd_multipleTracesAllowedInSameScope)
{
    EXPECT_CALL(*mockPublisher, log(_)).Times(2).WillRepeatedly(Invoke(this, &SmsFixture::captureMessage));
    const auto expectedLineNumber1 = __LINE__ + 1;
    AV_TRACE_BEGIN(First, FakeCategories::Speed, "MyLabel");
    const auto expectedLineNumber2 = __LINE__ + 1;
    AV_TRACE_BEGIN(Second, FakeCategories::Location, "MyLabel");
    AV_TRACE_END(Second);
    AV_TRACE_END(First);

    EXPECT_EQ(capturedMessages.size(), 2U);
    EXPECT_EQ(capturedMessages[0].fileName, capturedMessages[1].fileName);
    EXPECT_EQ(capturedMessages[0].functionName, capturedMessages[1].functionName);
    EXPECT_EQ(expectedLineNumber1, capturedMessages[1].lineNumber);
    EXPECT_EQ(expectedLineNumber2, capturedMessages[0].lineNumber);
    EXPECT_EQ("FakeCategories::Speed", capturedMessages[1].category);
    EXPECT_EQ("FakeCategories::Location", capturedMessages[0].category);
    EXPECT_EQ(capturedMessages[0].label, capturedMessages[1].label);
    EXPECT_EQ(capturedMessages[0].valueType, capturedMessages[1].valueType);
}

TEST_F(SmsFixture, logInstantSetsNameAndValue)
{
    EXPECT_CALL(*mockPublisher, log(_)).WillRepeatedly(Invoke(this, &SmsFixture::captureMessage));
    AV_TRACE_LOG_INSTANT(LongValue, FakeCategories::Speed, "MyLabel", anyValue<std::int64_t>);
    AV_TRACE_LOG_INSTANT(UnsignedLongValue, FakeCategories::Speed, "MyLabel", anyValue<std::uint64_t>);
    AV_TRACE_LOG_INSTANT(FloatValue, FakeCategories::Speed, "MyLabel", anyValue<float>);
    AV_TRACE_LOG_INSTANT(DoubleValue, FakeCategories::Speed, "MyLabel", anyValue<double>);
    AV_TRACE_LOG_INSTANT(StringValue, FakeCategories::Speed, "MyLabel", std::string{"LOG_INSTANT Test"});
    AV_TRACE_LOG_INSTANT(BoolValue, FakeCategories::Speed, "MyLabel", true);
    EXPECT_EQ(6, capturedMessages.size());

    EXPECT_EQ(ValueType::Int64, capturedMessages[0].valueType);
    EXPECT_EQ(ValueType::Uint64, capturedMessages[1].valueType);
    EXPECT_EQ(ValueType::Float32, capturedMessages[2].valueType);
    EXPECT_EQ(ValueType::Float64, capturedMessages[3].valueType);
    EXPECT_EQ(ValueType::String, capturedMessages[4].valueType);
    EXPECT_EQ(ValueType::Boolean, capturedMessages[5].valueType);

    EXPECT_EQ("LongValue", capturedMessages[0].name);
    EXPECT_EQ("UnsignedLongValue", capturedMessages[1].name);
    EXPECT_EQ("FloatValue", capturedMessages[2].name);
    EXPECT_EQ("DoubleValue", capturedMessages[3].name);
    EXPECT_EQ("StringValue", capturedMessages[4].name);
    EXPECT_EQ("BoolValue", capturedMessages[5].name);

    EXPECT_EQ(anyValue<std::int64_t>, capturedMessages[0].intValue);
    EXPECT_EQ(anyValue<std::uint64_t>, capturedMessages[1].uIntValue);
    EXPECT_EQ(anyValue<float>, capturedMessages[2].floatValue);
    EXPECT_EQ(anyValue<double>, capturedMessages[3].doubleValue);
    EXPECT_EQ("LOG_INSTANT Test", capturedMessages[4].stringValue);
    EXPECT_EQ(true, capturedMessages[5].boolValue);
}

TEST_F(SmsFixture, logInstant_cycleCountSetToDefinedValue)
{
    using CycleCount = decltype(TraceEvent::cycleCount);
    auto publishTrace = [this](CycleCount cycleCount) -> void {
        EXPECT_CALL(*mockPublisher, log(_)).WillOnce(Invoke(this, &SmsFixture::captureMessage));
        AV_TRACE_LOG_INSTANT_CYCLE(AnyName, FakeCategories::Speed, "MyLabel", cycleCount, 25L);
    };

    for (const CycleCount expectedCycleCount :
         {std::numeric_limits<CycleCount>::lowest(), std::numeric_limits<CycleCount>::max()}) {
        publishTrace(expectedCycleCount);
        EXPECT_EQ(expectedCycleCount, capturedMessages.front().cycleCount);
        capturedMessages.clear();
    }
}

TEST_F(SmsFixture, rateLimit)
{
    auto cycleCount = av::sms::TraceClient::rateLimit() / (1000 / av::sms::SmsLogger::windowSizeMs());

    EXPECT_CALL(*mockPublisher, log(_)).Times(static_cast<int>(cycleCount));
    EXPECT_CALL(*mockPublisher, warn(_)).Times(1);

    for (uint32_t i = 0; i < cycleCount + 10; ++i) {
        AV_TRACE_SCOPED_PERF(AnyName, FakeCategories::Speed, "MyLabel");
    }
}

TEST_F(SmsFixture, infoMessage)
{
    EXPECT_CALL(*mockPublisher, info(_)).WillRepeatedly(Invoke(this, &SmsFixture::captureInfoMessage));

    mockPublisher->info("A Message");
    EXPECT_EQ("A Message", capturedInfoMessages[0]);
}

TEST_F(SmsFixture, logInstant_startAndEndTimeIdentical)
{
    EXPECT_CALL(*mockPublisher, log(_)).WillRepeatedly(Invoke(this, &SmsFixture::captureMessage));
    AV_TRACE_LOG_INSTANT(Something1, FakeCategories::Speed, "MyLabel", anyValue<std::int64_t>);
    AV_TRACE_LOG_INSTANT(Something2, FakeCategories::Speed, "MyLabel", anyValue<std::int64_t>);

    EXPECT_EQ(capturedMessages.size(), 2U);
    EXPECT_EQ(capturedMessages[0].startTime, capturedMessages[0].endTime);
    EXPECT_EQ(capturedMessages[1].startTime, capturedMessages[1].endTime);
}
