/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 */

#include <sms/Sms.hpp>
#include <sms/SmsLogger.hpp>
#include <sms/sms-ingest/SmsIngestionReader.hpp>

#include <gtest/gtest.h>

#include <experimental/filesystem>

using namespace av::sms;
namespace fs = std::experimental::filesystem;

// Return the absolute path for the test file, checking if it exists
static std::string getTestFilePath(const char* fileName)
{
    const std::string basePath = std::getenv("TESTDIR");
    const std::string expectedFileName = basePath + "/" + fileName;

    if (!fs::exists(expectedFileName) or (fs::file_size(expectedFileName) == 0)) {
        throw std::runtime_error("File not found: " + expectedFileName);
    }

    return expectedFileName;
}

TEST(Sms, IngestionDataValidate)
{
    const std::string expectedFileName = getTestFilePath("smsIngest-test1.sms.msgpack");
    const std::string FileName = "platform/middleware/libs/sms/tests/testSmsIngest.cpp";
    const std::string FunctionName = "virtual void Sms_Ingestion_Test::TestBody()";

    SmsIngestionReader reader{expectedFileName, false};
    auto msgs = reader.read();

    EXPECT_EQ(msgs.size(), 3);

    auto& msg1 = msgs[0];
    EXPECT_EQ(msg1["fileName"], FileName);
    EXPECT_EQ(msg1["functionName"], FunctionName);
    EXPECT_EQ(msg1["lineNumber"], 74);
    EXPECT_EQ(msg1["label"], "Integer");
    EXPECT_EQ(msg1["valueType"], static_cast<int>(ValueType::Int64));
    EXPECT_EQ(msg1["intValue"], 42);
    EXPECT_EQ(msg1["cycleCount"], -1);
    EXPECT_EQ(msg1["processId"], 242399);
    EXPECT_EQ(msg1["name"], "Instant1");

    auto& msg2 = msgs[1];
    EXPECT_EQ(msg2["fileName"], FileName);
    EXPECT_EQ(msg2["functionName"], FunctionName);
    EXPECT_EQ(msg2["lineNumber"], 75);
    EXPECT_EQ(msg2["label"], "Float");
    EXPECT_EQ(msg2["valueType"], static_cast<int>(ValueType::Float32));
    EXPECT_EQ(msg2["floatValue"], 42.f);
    EXPECT_EQ(msg2["cycleCount"], -1);
    EXPECT_EQ(msg2["processId"], 242399);
    EXPECT_EQ(msg2["name"], "Instant2");

    auto& msg3 = msgs[2];
    EXPECT_EQ(msg3["fileName"], FileName);
    EXPECT_EQ(msg3["functionName"], FunctionName);
    EXPECT_EQ(msg3["lineNumber"], 76);
    EXPECT_EQ(msg3["label"], "Scoped");
    EXPECT_EQ(msg3["valueType"], static_cast<int>(ValueType::Timing));
    EXPECT_EQ(msg3["cycleCount"], -1);
    EXPECT_EQ(msg3["processId"], 242399);
    EXPECT_EQ(msg3["name"], "Scoped1");
}

// Test on same file, ensuring `messageOutput` parameter to `SmsIngestionReader`
// reads expected number of messages
// [
//   UniformMetadata(...)
//   Message(...),
//   Message(...)
// ]
TEST(Sms, IngestionValidateCanReadMessages)
{
    const std::string expectedFileName = getTestFilePath("smsIngest-test2.sms.msgpack");

    SmsIngestionReader reader{expectedFileName, true};
    auto msgs = reader.read();
    // No trace events exist in file; expect only 2 messages and no other data
    EXPECT_EQ(msgs.size(), 2);

    EXPECT_NE(msgs[0].find("message"), msgs[0].end());
    EXPECT_NE(msgs[1].find("message"), msgs[1].end());
}

// Validate file with more than one intern ID with the same ID will throw an exception
// input file:
// [
//   UniformMetadata(...)
//   InternString(id:1, value:"foo"),
//   InternString(id:1, value:"bar")
// ]
TEST(Sms, IngestionValidateDuplicateInternIDs)
{
    const std::string expectedFileName = getTestFilePath("smsIngest-test3.sms.msgpack");

    SmsIngestionReader reader{expectedFileName, true};

    EXPECT_THROW({ auto msgs = reader.read(); }, std::invalid_argument);
}

// Validate file with more than one UniformMetadata entry will throw an exception
// input file:
// [
//   UniformMetadata(processId:42, processName:"foo"),
//   UniformMetadata(processId:43, processName:"bar"),
// ]
TEST(Sms, IngestionValidateDuplicateUniformMetadataMessages)
{
    const std::string expectedFileName = getTestFilePath("smsIngest-test4.sms.msgpack");

    SmsIngestionReader reader{expectedFileName, true};

    EXPECT_THROW({ auto msgs = reader.read(); }, std::invalid_argument);
}

// Validate file with invalid msgpack messages will throw exception
// Input file:
// [
//   42,
//   42.0,
//   {"a": "A", "b": "B"}
// ]
TEST(Sms, IngestionValidateInvalidFile)
{
    const std::string expectedFileName = getTestFilePath("smsIngest-test5.sms.msgpack");

    SmsIngestionReader reader{expectedFileName, true};

    EXPECT_THROW({ auto msgs = reader.read(); }, std::invalid_argument);
}

// Validate file with input which used to crash due to `InternString` messages appearing after a `TraceEvent`
// dependent on it. See https://jira.ci.motional.com/atljira/browse/IR-587 for background info.
// Input file:
// [
//   UniformMetadata(processId:42, processName:"foo", schemaVersion:1),
//   InternString(id:0, value:"platform/middleware/libs/sms/tests/testSmsIngest.cpp"),
//   InternString(id:1, value:"virtual void Sms_Ingestion_Test::TestBody()"),
//   InternString(id:2, value:"CustomEnumLabels::One"),
//   InternString(id:3, value:"Integer"),
//   TraceEvent(...),
//   InternString(id:4, value:"Instant1"),
// ]
TEST(Sms, Ingestion_Regression_IR587)
{
    const std::string expectedFileName = getTestFilePath("smsIngest-test6.sms.msgpack");
    const std::string FileName = "platform/middleware/libs/sms/tests/testSmsIngest.cpp";
    const std::string FunctionName = "virtual void Sms_Ingestion_Test::TestBody()";

    SmsIngestionReader reader{expectedFileName, false};
    auto msgs = reader.read();

    EXPECT_EQ(msgs.size(), 1);

    auto& msg1 = msgs[0];
    EXPECT_EQ(msg1["fileName"], FileName);
    EXPECT_EQ(msg1["functionName"], FunctionName);
    EXPECT_EQ(msg1["lineNumber"], 74);
    EXPECT_EQ(msg1["category"], "CustomEnumLabels::One");
    EXPECT_EQ(msg1["label"], "Integer");
    EXPECT_EQ(msg1["valueType"], static_cast<int>(ValueType::Int64));
    EXPECT_EQ(msg1["intValue"], 42);
    EXPECT_EQ(msg1["cycleCount"], -1);
    EXPECT_EQ(msg1["processId"], 42);
    EXPECT_EQ(msg1["name"], "Instant1");
}
