/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 */

#include <avtime/AvClock.hpp>
#include <avtime/ConversionsUsec.hpp>
#include <sms/SmsSchema.hpp>

#include <gmock/gmock.h>

using namespace av::sms;

struct SmsFixture : public testing::Test
{
    void SetUp() {}

    void TearDown() {}
};

TEST_F(SmsFixture, createUniformMetadataEvent)
{
    UniformMetadata src;
    src.processID = 42;
    src.processName = "Forty Two";
    src.schemaVersion = SchemaVersion;

    nlohmann::json dst = createUniformMetadataEvent(src);

    const auto& metadata = dst["metadata"];
    // dst.keys() == {"metadata"}
    // metadata.keys() == {"processID", "schemaVersion", "processName"}
    EXPECT_EQ(metadata["processID"], src.processID);
    EXPECT_EQ(metadata["processName"], src.processName);
    EXPECT_EQ(metadata["schemaVersion"], SchemaVersion);
}

TEST_F(SmsFixture, createMessageEvent)
{
    const std::string message = "A Message";
    const std::string severity = "Info";

    nlohmann::json dst = createMessageEvent(message, severity);

    // dst.keys() == {"message", "severity"}
    EXPECT_EQ(dst["message"], message);
    EXPECT_EQ(dst["severity"], severity);
}

TEST_F(SmsFixture, populateInternEvent)
{
    nlohmann::json dst;
    uint32_t internID = 42;
    const std::string& value = "Forty Two";

    populateInternEvent(dst, internID, value);

    EXPECT_EQ(dst.size(), 2);
    EXPECT_EQ(dst[0], internID);
    EXPECT_EQ(dst[1], value);
}

TEST_F(SmsFixture, populateTraceEvent)
{
    nlohmann::json dst = nlohmann::json(TraceEventMinimumEntryCount, nlohmann::json{});
    TraceEvent traceEvent;

    // 1: Test trace event with value.
    //    `threadID` != `processID`
    //    `startTime` != `endTime`

    auto startTime = 12345;
    auto endTime = startTime + 1;
    auto processId = 42;
    auto threadId = 43;
    auto cycleCount = 100;
    auto lineNumber = 12;
    auto valueType = ValueType::Int64;
    auto value = 600;

    traceEvent.startTime = av::timePointFromUsec<av::AvClock>(startTime);
    traceEvent.endTime = av::timePointFromUsec<av::AvClock>(endTime);
    traceEvent.processId = processId;
    traceEvent.threadId = threadId;
    traceEvent.cycleCount = cycleCount;
    traceEvent.lineNumber = lineNumber;
    traceEvent.valueType = valueType;
    traceEvent.intValue = value;

    // Don't worry about populating following fields:
    // * processName: Is placed within `UniformMetadata` message
    // * fileName: intern'd below
    // * functionName: ditto
    // * category: ditto
    // * label: ditto
    // * name: ditto
    // * All other Xvalue types (they are not touched as value type is int64)

    TraceEventInternCollection ids;
    ids.fileNameID = 1;
    ids.functionNameID = 2;
    ids.categoryID = 3;
    ids.labelID = 4;
    ids.nameID = 5;

    populateTraceEvent(dst, traceEvent, ids);

    EXPECT_TRUE(dst.is_array());
    EXPECT_EQ(dst[StartTimeIndex], startTime);
    EXPECT_EQ(dst[EndTimeIndex], endTime - startTime);
    EXPECT_EQ(dst[ThreadIdIndex], threadId);
    EXPECT_EQ(dst[FileNameIndex], ids.fileNameID);
    EXPECT_EQ(dst[FunctionNameIndex], ids.functionNameID);
    EXPECT_EQ(dst[LineNumberIndex], lineNumber);
    EXPECT_EQ(dst[CategoryIndex], ids.categoryID);
    EXPECT_EQ(dst[LabelIndex], ids.labelID);
    EXPECT_EQ(dst[CycleCountIndex], cycleCount);
    EXPECT_EQ(dst[NameIndex], ids.nameID);
    EXPECT_EQ(dst[ValueTypeIndex], static_cast<int>(valueType));
    EXPECT_EQ(dst[ValueIndex], value);

    // 2: Modify message above to have:
    // * `threadID == processID`
    // * `startTime == endTime`

    threadId = processId;
    traceEvent.threadId = threadId;

    endTime = startTime;
    traceEvent.endTime = av::timePointFromUsec<av::AvClock>(endTime);

    populateTraceEvent(dst, traceEvent, ids);

    // Fields should be set to 0 to save space in serialization
    EXPECT_EQ(dst[EndTimeIndex], 0);
    EXPECT_EQ(dst[ThreadIdIndex], 0);

    // 3: Expect to throw is `dst` is too small
    dst = nlohmann::json::array({0, 1, 2});
    EXPECT_THROW({ populateTraceEvent(dst, traceEvent, ids); }, std::invalid_argument);
}

TEST_F(SmsFixture, extractUniformMetadataEvent)
{
    const auto processID = 42;
    const std::string processName{"SmsUnitTests"};

    nlohmann::json src;
    src["metadata"] =
        nlohmann::json{{"processID", processID}, {"schemaVersion", SchemaVersion}, {"processName", processName}};

    UniformMetadata dst = extractUniformMetadataEvent(src);

    EXPECT_EQ(dst.processID, processID);
    EXPECT_EQ(dst.processName, processName);
}

TEST_F(SmsFixture, extractInternEvent)
{
    const auto id = 42;
    const std::string value{"A String"};

    nlohmann::json src = nlohmann::json::array({id, value});

    InternString dst = extractInternEvent(src);

    EXPECT_EQ(dst.id, id);
    EXPECT_EQ(dst.value, value);

    // Expect to throw if `src` is not an array of size 2
    src = nlohmann::json::array({0, 1, 2});
    EXPECT_THROW({ InternString ignore = extractInternEvent(src); }, std::invalid_argument);

    src = nlohmann::json{{"foo", "bar"}};
    EXPECT_THROW({ InternString ignore = extractInternEvent(src); }, std::invalid_argument);
}

TEST_F(SmsFixture, extractTraceEvent)
{
    TraceEventInternCollection ids;
    ids.fileNameID = 1;
    ids.functionNameID = 2;
    ids.categoryID = 3;
    ids.labelID = 4;
    ids.nameID = 5;

    std::unordered_map<uint32_t, std::string> internTable;
    internTable[0] = "zero";
    internTable[1] = "one";
    internTable[2] = "two";
    internTable[3] = "three";
    internTable[4] = "four";
    internTable[5] = "five";
    internTable[6] = "six";

    // 1: Test trace event with value.
    //    `threadID` != `processID`

    auto startTime = 12345;
    auto endTime = startTime + 1;
    auto processId = 42;
    auto threadId = 43;
    auto cycleCount = 100;
    auto lineNumber = 12;
    auto valueType = ValueType::Int64;
    auto value = 600;

    UniformMetadata metadata;
    metadata.processID = static_cast<uint32_t>(processId);
    metadata.processName = "Forty Two";

    // Compressed representation which is an array
    nlohmann::json src = nlohmann::json(TraceEventMinimumEntryCount, nlohmann::json{});
    nlohmann::json dst;
    TraceEvent traceEvent;
    traceEvent.startTime = av::timePointFromUsec<av::AvClock>(startTime);
    traceEvent.endTime = av::timePointFromUsec<av::AvClock>(endTime);
    traceEvent.processId = processId;
    traceEvent.threadId = threadId;
    traceEvent.cycleCount = cycleCount;
    traceEvent.lineNumber = lineNumber;
    traceEvent.valueType = valueType;
    traceEvent.intValue = value;

    // Create serialized message from trace event for simplicity
    populateTraceEvent(src, traceEvent, ids);
    extractTraceEvent(dst, src, internTable, metadata);

    EXPECT_EQ(dst["startTime"], startTime);
    EXPECT_EQ(dst["endTime"], endTime);
    EXPECT_EQ(dst["threadId"], threadId);
    EXPECT_EQ(dst["fileName"], internTable.at(ids.fileNameID));
    EXPECT_EQ(dst["functionName"], internTable.at(ids.functionNameID));
    EXPECT_EQ(dst["lineNumber"], lineNumber);
    EXPECT_EQ(dst["category"], internTable.at(ids.categoryID));
    EXPECT_EQ(dst["label"], internTable.at(ids.labelID));
    EXPECT_EQ(dst["cycleCount"], cycleCount);
    EXPECT_EQ(dst["name"], internTable.at(ids.nameID));
    EXPECT_EQ(static_cast<ValueType>(dst["valueType"]), valueType);
    EXPECT_EQ(dst["intValue"], value);

    // 2: Test with various data types
    // (To make SonarQube happy)
    valueType = ValueType::Uint64;
    traceEvent.valueType = valueType;
    traceEvent.uIntValue = static_cast<uint64_t>(value);
    populateTraceEvent(src, traceEvent, ids);
    extractTraceEvent(dst, src, internTable, metadata);
    EXPECT_EQ(static_cast<ValueType>(dst["valueType"]), valueType);
    EXPECT_EQ(dst["uIntValue"], static_cast<uint64_t>(value));

    valueType = ValueType::Float32;
    traceEvent.valueType = valueType;
    traceEvent.floatValue = static_cast<float>(value);
    populateTraceEvent(src, traceEvent, ids);
    extractTraceEvent(dst, src, internTable, metadata);
    EXPECT_EQ(static_cast<ValueType>(dst["valueType"]), valueType);
    EXPECT_EQ(dst["floatValue"], static_cast<float>(value));

    valueType = ValueType::Float64;
    traceEvent.valueType = valueType;
    traceEvent.doubleValue = static_cast<double>(value);
    populateTraceEvent(src, traceEvent, ids);
    extractTraceEvent(dst, src, internTable, metadata);
    EXPECT_EQ(static_cast<ValueType>(dst["valueType"]), valueType);
    EXPECT_EQ(dst["doubleValue"], static_cast<double>(value));

    valueType = ValueType::String;
    traceEvent.valueType = valueType;
    traceEvent.stringValue = "A string";
    populateTraceEvent(src, traceEvent, ids);
    extractTraceEvent(dst, src, internTable, metadata);
    EXPECT_EQ(static_cast<ValueType>(dst["valueType"]), valueType);
    EXPECT_EQ(dst["stringValue"], std::string{"A string"});

    valueType = ValueType::Boolean;
    traceEvent.valueType = valueType;
    traceEvent.boolValue = false;
    populateTraceEvent(src, traceEvent, ids);
    extractTraceEvent(dst, src, internTable, metadata);
    EXPECT_EQ(static_cast<ValueType>(dst["valueType"]), valueType);
    EXPECT_EQ(dst["boolValue"], false);

    // 3: Test trace event with value.
    //    `threadID` == `processID`
    threadId = processId;
    traceEvent.threadId = threadId;
    populateTraceEvent(src, traceEvent, ids);
    extractTraceEvent(dst, src, internTable, metadata);

    EXPECT_EQ(dst["threadId"], threadId);

    // 4: Expect to throw if `src` is not an array of size >= TraceEventMinimumEntryCount
    populateTraceEvent(src, traceEvent, ids);

    src = nlohmann::json::array({0, 1, 2});
    EXPECT_THROW({ extractTraceEvent(dst, src, internTable, metadata); }, std::invalid_argument);

    src = nlohmann::json{{"foo", "bar"}};
    EXPECT_THROW({ extractTraceEvent(dst, src, internTable, metadata); }, std::invalid_argument);

    // 5: Expect to throw if `valueType` is an invalid value
    src = nlohmann::json(TraceEventMinimumEntryCount, nlohmann::json{});
    traceEvent.valueType = static_cast<ValueType>(100);
    populateTraceEvent(src, traceEvent, ids);
    EXPECT_THROW({ extractTraceEvent(dst, src, internTable, metadata); }, std::invalid_argument);
}
