/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 */

#include <sms/Sms.hpp>
#include <sms/SmsLogger.hpp>
#include <sms/sms-ingest/SmsIngestionReader.hpp>

#include <gtest/gtest.h>

#include <exception>
#include <experimental/filesystem>
#include <thread>

using namespace av::sms;
namespace fs = std::experimental::filesystem;

enum class CustomEnumLabels
{
    One,
    Two,
    Three
};

template<>
struct base::EnumTrait<CustomEnumLabels>
{
    static constexpr std::array namedEnumerators = {
        base::Enumerator{CustomEnumLabels::One, "One"},
        base::Enumerator{CustomEnumLabels::Two, "Two"},
        base::Enumerator{CustomEnumLabels::Three, "Three"},
    };
};

struct Metadata
{
    int lineNumber;
    std::string fileName;
    std::string functionName;
};

Metadata getMetadata(std::string functionName,
                     int lineNumber = __builtin_LINE(),
                     std::string fileName = __builtin_FILE())
{
    return Metadata{lineNumber, std::move(fileName), std::move(functionName)};
}

TEST(Sms, IngestionDataValidate)
{
    // testing::internal::CaptureStdout();
    const std::string oldPath = fs::current_path();
    const std::string newPath = std::getenv("TEST_TMPDIR");

    const int32_t pid = getpid();
    const int32_t tid = static_cast<int>(syscall(SYS_gettid));
    const std::string smsName = "testSmsWriter";
    const std::string expectedFileName = smsName + "-" + std::to_string(pid) + ".sms.msgpack";

    try {
        fs::current_path(newPath);

        int32_t backgroundThread{0};

        // Write entries
        TraceClient::initialize(std::make_unique<AvlogSmsLogger>(smsName, false), smsName, 0);

        // clang-format off
        Metadata md4{};
        // NOTE: Metadata must be on the same line to capture line number correctly
        AV_TRACE_LOG_INSTANT(Instant1, CustomEnumLabels::One, "Integer", static_cast<int64_t>(42)); Metadata md1 = getMetadata(__PRETTY_FUNCTION__);
        AV_TRACE_LOG_INSTANT(Instant2, CustomEnumLabels::Two, "Float", 42.f); Metadata md2 = getMetadata(__PRETTY_FUNCTION__);
        AV_TRACE_BEGIN(Scoped1, CustomEnumLabels::Two, "Scoped"); Metadata md3 = getMetadata(__PRETTY_FUNCTION__);
        AV_TRACE_END(Scoped1);

        {
            // Spawn background thread to test threadID field is appropriately set
            std::thread bgThread{[&]() {
                backgroundThread = static_cast<int>(syscall(SYS_gettid)); // NOLINT(cppcoreguidelines-pro-type-vararg)
                AV_TRACE_LOG_INSTANT(Background1, CustomEnumLabels::Three, "BackgroundThread", static_cast<int64_t>(3)); md4 = getMetadata(__PRETTY_FUNCTION__);
            }};
            bgThread.join();
        }
        // clang-format on

        // Flush to disk by re-initializing
        TraceClient::initialize(nullptr, smsName, 0);
        // Read entries

        SmsIngestionReader reader{expectedFileName, false};
        auto msgs = reader.read();

        EXPECT_EQ(msgs.size(), 4);

        auto& msg1 = msgs[0];
        EXPECT_EQ(msg1["fileName"], md1.fileName);
        EXPECT_EQ(msg1["functionName"], md1.functionName);
        EXPECT_EQ(msg1["lineNumber"], md1.lineNumber);
        EXPECT_EQ(msg1["label"], "Integer");
        EXPECT_EQ(msg1["valueType"], static_cast<int>(ValueType::Int64));
        EXPECT_EQ(msg1["intValue"], 42);
        EXPECT_EQ(msg1["cycleCount"], -1);
        EXPECT_EQ(msg1["processId"], pid);
        EXPECT_EQ(msg1["threadId"], tid);
        EXPECT_EQ(msg1["name"], "Instant1");

        auto& msg2 = msgs[1];
        EXPECT_EQ(msg2["fileName"], md2.fileName);
        EXPECT_EQ(msg2["functionName"], md2.functionName);
        EXPECT_EQ(msg2["lineNumber"], md2.lineNumber);
        EXPECT_EQ(msg2["label"], "Float");
        EXPECT_EQ(msg2["valueType"], static_cast<int>(ValueType::Float32));
        EXPECT_EQ(msg2["floatValue"], 42.f);
        EXPECT_EQ(msg2["cycleCount"], -1);
        EXPECT_EQ(msg2["processId"], pid);
        EXPECT_EQ(msg2["threadId"], tid);
        EXPECT_EQ(msg2["name"], "Instant2");

        auto& msg3 = msgs[2];
        EXPECT_EQ(msg3["fileName"], md3.fileName);
        EXPECT_EQ(msg3["functionName"], md3.functionName);
        EXPECT_EQ(msg3["lineNumber"], md3.lineNumber);
        EXPECT_EQ(msg3["label"], "Scoped");
        EXPECT_EQ(msg3["valueType"], static_cast<int>(ValueType::Timing));
        EXPECT_EQ(msg3["cycleCount"], -1);
        EXPECT_EQ(msg3["processId"], pid);
        EXPECT_EQ(msg3["threadId"], tid);
        EXPECT_EQ(msg3["name"], "Scoped1");

        auto& msg4 = msgs[3];
        EXPECT_EQ(msg4["fileName"], md4.fileName);
        EXPECT_EQ(msg4["functionName"], md4.functionName);
        EXPECT_EQ(msg4["lineNumber"], md4.lineNumber);
        EXPECT_EQ(msg4["label"], "BackgroundThread");
        EXPECT_EQ(msg4["valueType"], static_cast<int>(ValueType::Int64));
        EXPECT_EQ(msg4["intValue"], 3);
        EXPECT_EQ(msg4["cycleCount"], -1);
        EXPECT_EQ(msg4["processId"], pid);
        EXPECT_EQ(msg4["threadId"], backgroundThread);
        EXPECT_EQ(msg4["name"], "Background1");

    } catch (...) {
        // Ensure path is reset
        std::error_code ignore;
        fs::current_path(oldPath, ignore);

        // Propagate after cleaning up resources
        std::rethrow_exception(std::current_exception());
    }
}

TEST(Sms, IngestionDataValidateMessages)
{
    // testing::internal::CaptureStdout();
    const std::string oldPath = fs::current_path();
    const std::string newPath = std::getenv("TEST_TMPDIR");

    const int32_t pid = getpid();
    const std::string smsName = "testSmsWriter";
    const std::string expectedFileName = smsName + "-" + std::to_string(pid) + ".sms.msgpack";

    try {
        fs::current_path(newPath);

        auto loggerObj = std::make_unique<AvlogSmsLogger>(smsName, false);
        auto logger = loggerObj.get();

        // Write entries
        TraceClient::initialize(std::move(loggerObj), smsName, 0);

        logger->info("A Message");

        // Flush to disk by re-initializing
        TraceClient::initialize(nullptr, smsName, 0);
        // Read entries

        SmsIngestionReader reader{expectedFileName, true};
        auto msgs = reader.read();

        EXPECT_EQ(msgs.size(), 1);

        auto& msg1 = msgs[0];
        EXPECT_EQ(msg1["message"], "A Message");
        EXPECT_EQ(msg1["severity"], "info");

    } catch (...) {
        // Ensure path is reset
        std::error_code ignore;
        fs::current_path(oldPath, ignore);

        // Propagate after cleaning up resources
        std::rethrow_exception(std::current_exception());
    }
}
