/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines the standard backup recovery handler
 */

#include "TaskUtil.h"
#include <Clock/SystemClock.h>
#include <interfaces/Generic/Input/Abstract.h>
#include <interfaces/Generic/Output/Abstract.h>
#include <interfaces/SaferCommand/SaferCommand.h>
#include <interfaces/SaferResponse/SaferResponse.h>
#include <task/BackupRecoveryHandler.h>

#include "lib/avlog/avlog.hpp"

#include <sstream>

namespace {
const std::string vassalName{"VassalDaemon"};     ///< The default vassal name
const std::string kingDaemonName{"KingDaemon"};   ///< The default king daemon name
const std::string saferDaemonName{"SaferDaemon"}; ///< The safer daemon name
} // namespace

namespace task {

BackupRecoveryHandler::BackupRecoveryHandler(std::shared_ptr<GenericInputIntf<av::msg::SaferCommand>> saferCmdInput,
                                             std::shared_ptr<GenericOutputIntf<av::msg::SaferResponse>> saferRespOutput,
                                             std::string taskName,
                                             std::string hostName)
    : saferCmdInput_(std::move(saferCmdInput))
    , saferResponseOutput_(std::move(saferRespOutput))
    , taskName_(std::move(taskName))
    , hostName_(std::move(hostName))
{}

bool BackupRecoveryHandler::initialize()
{
    bool result{false};

    result = (saferCmdInput_ != nullptr);
    result &= (saferResponseOutput_ != nullptr);
    return result;
}

bool BackupRecoveryHandler::handleBackupRecovery(ConfigSectionList& writeOnlyInterfaces,
                                                 ConfigSectionList& readWriteInterfaces,
                                                 BackupRecoveryParameter& handlingParams,
                                                 std::map<std::string, IDbObject*>& rawPtrInterfaces,
                                                 std::map<std::string, std::shared_ptr<IDbObject>>& sharedPtrInterfaces)
{
    bool result{saferCmdInput_ != nullptr};
    if (result && (taskName_ != vassalName) && (taskName_ != kingDaemonName) && (taskName_ != saferDaemonName)) {
        av::msg::SaferCommand saferCmd{};
        // Get all commands from safer
        while (saferCmdInput_->get(saferCmd)) {
            const bool needRecovery = needRecoveryWithBackup(saferCmd.taskName_, taskName_);
            switch (saferCmd.messageType_) {
                // If the primary is hung
                case av::msg::MessageId::TASKHUNG: {
                    if (needRecovery) {
                        result &= handleHungTask(saferCmd,
                                                 writeOnlyInterfaces,
                                                 readWriteInterfaces,
                                                 handlingParams,
                                                 rawPtrInterfaces,
                                                 sharedPtrInterfaces);
                    }
                    break;
                }
                // If the primary rejoins
                case av::msg::MessageId::TASKHEALTHY: {
                    if (needRecovery) {
                        result &= handleRejoinedTask(saferCmd,
                                                     writeOnlyInterfaces,
                                                     readWriteInterfaces,
                                                     handlingParams,
                                                     rawPtrInterfaces,
                                                     sharedPtrInterfaces);
                    }
                    break;
                }

                case av::msg::MessageId::UNDEFINED:
                case av::msg::MessageId::PAUSE:
                case av::msg::MessageId::LAUNCH:
                case av::msg::MessageId::STOP:
                case av::msg::MessageId::STOPALL:
                case av::msg::MessageId::LOAD:
                case av::msg::MessageId::STATUS:
                case av::msg::MessageId::LOADERROR:
                case av::msg::MessageId::LOADCOMPLETE:
                case av::msg::MessageId::GETCONFIGURATION:
                case av::msg::MessageId::NORMAL:
                case av::msg::MessageId::PAUSESTATUS:
                case av::msg::MessageId::USERRESTART:
                case av::msg::MessageId::CONFIGREMIND:
                case av::msg::MessageId::LAUNCHREMIND:
                case av::msg::MessageId::CHECKPOINTCHANGE:
                default:
                    break;
            }
        }
    }
    return result;
}

bool BackupRecoveryHandler::sendSaferResponse(const std::string& reportMsg)
{
    bool result{false};
    if (saferResponseOutput_ != nullptr) {
        av::msg::SaferResponse saferRes{};
        saferRes.messageType_ = av::msg::MessageId::STATUS;
        saferRes.responseType_ = av::msg::SaferResponseId::REPORT;
        saferRes.message_ = reportMsg;
        saferRes.sender_ = taskName_ + "@" + hostName_;
        result = saferResponseOutput_->publish(saferRes);
    }
    return result;
}

bool BackupRecoveryHandler::handleHungTask(const av::msg::SaferCommand& saferCmd,
                                           ConfigSectionList& writeOnlyInterfaces,
                                           ConfigSectionList& readWriteInterfaces,
                                           BackupRecoveryParameter& handlingParams,
                                           std::map<std::string, IDbObject*>& rawPtrInterfaces,
                                           std::map<std::string, std::shared_ptr<IDbObject>>& sharedPtrInterfaces)
{
    bool result{true};
    auto currentTime =
        handlingParams.isSteadyClkRequired ? SystemClock::getSynchronizedSteadyTime() : SystemClock::getSystemTime();
    AVLOG_WARN("Received Hung Task name {}", AVLOG_ARGS("taskName"), saferCmd.taskName_);
    // Need to check the numReplicas first to avoid the case the primary task was already recovered where numReplicas ==
    // 0, but still logging that the primary task is now recovered
    if (handlingParams.numReplicas == 1) {
        // Make invisible output interfaces visible
        if (!writeOnlyInterfaces.empty()) {
            result &= chmodTaskInterfaces(writeOnlyInterfaces, Access::Write, rawPtrInterfaces, sharedPtrInterfaces);
        }
        if (!readWriteInterfaces.empty()) {
            result &=
                chmodTaskInterfaces(readWriteInterfaces, Access::ReadWrite, rawPtrInterfaces, sharedPtrInterfaces);
        }
        // If the task is a cold standby, change to hot one
        if (handlingParams.coldStandby == 1) {
            handlingParams.coldStandby = 0;
        }
        // Mark the primary hung task is recovered by assign numReplicas = 0
        handlingParams.numReplicas = 0;
        // Log if the hung task has been recovered
        currentTime = handlingParams.isSteadyClkRequired ? SystemClock::getSynchronizedSteadyTime()
                                                         : SystemClock::getSystemTime();
        AVLOG_WARN("RECOVERED Hung Task name {}", AVLOG_ARGS("taskName"), saferCmd.taskName_);
    }
    return result;
}

bool BackupRecoveryHandler::handleRejoinedTask(const av::msg::SaferCommand& saferCmd,
                                               ConfigSectionList& writeOnlyInterfaces,
                                               ConfigSectionList& readWriteInterfaces,
                                               BackupRecoveryParameter& handlingParams,
                                               std::map<std::string, IDbObject*>& rawPtrInterfaces,
                                               std::map<std::string, std::shared_ptr<IDbObject>>& sharedPtrInterfaces)
{
    bool result{true};
    AVLOG_WARN("Received Healthy Task name {} @{} {}",
               AVLOG_ARGS("cmdTaskName", "taskName", "recipient"),
               saferCmd.taskName_,
               taskName_,
               saferCmd.recipient_);
    // There is a case that a task does not have any write nor readwrite interfaces, thus need to check numReplicas ==
    // 0, first
    if (handlingParams.numReplicas == 0) {
        // Make all visible output interfaces invisible
        if (!writeOnlyInterfaces.empty()) {
            result &= chmodTaskInterfaces(writeOnlyInterfaces, Access::Standby, rawPtrInterfaces, sharedPtrInterfaces);
        }
        if (!readWriteInterfaces.empty()) {
            result &= chmodTaskInterfaces(readWriteInterfaces, Access::Standby, rawPtrInterfaces, sharedPtrInterfaces);
        }
        // Mark the primary task rejoined
        handlingParams.numReplicas = 1;
    }
    return result;
}

} // namespace task
