/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief CallbackSet member function definitions.
 */
#include <task/CallbackSet.h>

#include "lib/avlog/avlog.hpp"
#include <SimpleComms/Constants.h>
#include <SimpleComms/ControlMessages.h>
#include <repro/blobUtils.hpp>

#include <algorithm>
#include <chrono>
#include <cmath>
#include <iomanip>
#include <iostream>
#include <map>
#include <stdexcept>

namespace {
// Helper class for storing timing stats, in milliseconds:
class Timing
{
public:
    void add(std::chrono::steady_clock::time_point begin, std::chrono::steady_clock::time_point end)
    {
        // Store the time of the first and last call, to measure frequency.
        if (m_count == 0) {
            m_start = begin;
        } else {
            m_seconds = begin - m_start;
        }
        double ms = std::chrono::duration<double>(end - begin).count() * 1000.;
        m_count++;
        m_sum += ms;
        m_squared += ms * ms;
        m_max = std::max(m_max, ms);
    }

    int count() const { return m_count; }
    double mean() const { return m_sum / m_count; }
    double sigma() const { return std::sqrt(m_squared * m_count - m_sum * m_sum) / m_count; }
    double max() const { return m_max; }
    double frequency() const { return (m_count - 1) / m_seconds.count(); }

private:
    int m_count{0};
    double m_sum{0.};
    double m_squared{0.};
    double m_max{0.};

    std::chrono::steady_clock::time_point m_start;
    std::chrono::duration<double> m_seconds{0.};
};
// Helper function to get the timestamp from a blob:
std::chrono::nanoseconds messageTimestamp(const std::string& blob)
{
    // TODO: Does this need to handle previous header versions?
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    const auto& header = *reinterpret_cast<const interfaces::baseTypes::InterfaceHeader*>(blob.data());
    otimeSpec time;
    interfaces::baseTypes::getSendTimeStamp(header, time);
    return std::chrono::nanoseconds(time.tv_sec * NSEC_PER_SEC + time.tv_nsec);
}
} // namespace

namespace task {

CallbackSet::CallbackSet(av::repro::EventFramework& framework)
    : framework_(framework)
{}

CallbackSet::~CallbackSet()
{
    stop();
}

void CallbackSet::init(bool isReproducible, const std::string& logPath)
{
    m_isReproducible = isReproducible;
    // If a path was given, begin saving the client-side log.
    if (m_isReproducible && !logPath.empty()) {
        log_.open(logPath, "time", "outID", "topic", "origin", "firstID", "lastID");
    }
}

void CallbackSet::stop()
{
    {
        std::scoped_lock lock(m_messageMutex);
        m_interrupt = true;
    }
    m_messageCondition.notify_all();
    if (m_callbackThread.joinable()) {
        m_callbackThread.join();
    }
}

void CallbackSet::setStaleThreshold(double seconds)
{
    // This ought to be done before we start receiving messages, but just in case:
    std::scoped_lock lock(m_messageMutex);
    m_staleThreshold = std::chrono::nanoseconds(static_cast<int64_t>(seconds * NSEC_PER_SEC));
}

void CallbackSet::setExecutiveBacklogLimit(int count)
{
    // This ought to be done before we start receiving messages, but just in case:
    std::scoped_lock lock(m_messageMutex);
    m_executiveBacklogLimit = count;
}

void CallbackSet::call(const std::string& channel, const std::string& blob, const MessageKey& key)
{
    if (channel == SIMPLECOMMS_CONTROL_CHANNEL) {
        // This is a control message. Check if it's one we care about.
        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast,cppcoreguidelines-pro-type-vararg)
        uint32_t msgType = ntohl(*reinterpret_cast<const uint32_t*>(blob.data()));
        if (msgType == SimpleComms::ControlMessages::Replay) {
            std::scoped_lock lock(m_messageMutex);
            m_isReplaying = true;
        } else if (msgType == SimpleComms::ControlMessages::Executive) {
            // Increment the count of executive functions in the queue.
            {
                std::scoped_lock lock(m_messageMutex);
                ++m_executiveBacklog;
            }

            // Convert this message into a blob with the standard header and no payload.
            // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
            const SimpleComms::ExecutiveMessage& message =
                *reinterpret_cast<const SimpleComms::ExecutiveMessage*>(blob.data());
            interfaces::baseTypes::InterfaceHeader header;
            header.sendTimeStamp.tv_sec = message.tv_sec;
            header.sendTimeStamp.tv_nsec = message.tv_nsec;
            header.headerVersion = htonl(OTTO_INTERFACE_HEADER_VERSION); // NOLINT(cppcoreguidelines-pro-type-vararg)
            header.messageCounter = htonl(key.second);
            // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
            std::string placeholder(reinterpret_cast<const char*>(&header), sizeof(header));

            // An empty channel name represents the executive function.
            call("", placeholder, key);
        }
        return;
    }
    if (has(channel)) {
        {
            const std::chrono::nanoseconds sendTime = messageTimestamp(blob);

            std::scoped_lock lock(m_messageMutex);
            // If the callback thread has been stopped, ignore all messages from here on.
            if (m_interrupt) {
                return;
            }
            // Update the variables used for dropping messages. The message timestamp
            // is from the steady clock in SCS, so this is always the latest time.
            m_latestTimestamp = sendTime;
            // Enqueue the message.
            m_messages.push({channel, blob, m_latestTimestamp, key});
            // Make sure the callback thread is running. Call this inside the mutex so we
            // never spuriously launch the thread if the interrupt flag is set right after
            // the check above.
            launchThread();
        }
        m_messageCondition.notify_all();
    }
}

av::repro::EventFramework& CallbackSet::framework() const
{
    return framework_;
}

const otimePoint& CallbackSet::timestamp() const
{
    std::scoped_lock lock(m_messageMutex);
    return m_timestamp;
}

bool CallbackSet::isReplaying() const
{
    std::scoped_lock lock(m_messageMutex);
    return m_isReplaying;
}

bool CallbackSet::has(const std::string& channel) const
{
    std::shared_lock lock(m_callbackMutex);
    return m_callbacks.count(channel) != 0;
}

void CallbackSet::launchThread()
{
    if (!m_callbackThread.joinable()) {
        m_callbackThread = std::thread(&CallbackSet::callbackThread, this);
    }
}

void CallbackSet::callbackThread()
{
    // Collect timing statistics for each channel.
    std::map<std::string, Timing> timing;
    while (true) {
        Message message;
        bool shouldDrop = false;
        {
            // Wait until either a message is received or this thread is interrupted.
            std::unique_lock lock(m_messageMutex);
            m_messageCondition.wait(lock, [this] { return !m_messages.empty() || m_interrupt; });
            if (m_interrupt) {
                break;
            }
            // Copy the blob, then pop the message and free the mutex.
            message = std::move(m_messages.front());
            m_messages.pop();
            // Store this message's timestamp.
            int64_t nsec = message.timestamp.count();
            m_timestamp = otimeSpec(nsec / NSEC_PER_SEC, nsec % NSEC_PER_SEC);
            // Check if this message should be dropped. When replaying, never drop messages.
            if (!m_isReplaying) {
                shouldDrop = (m_staleThreshold.count() > 0 && m_latestTimestamp - message.timestamp > m_staleThreshold);
                if (message.channel.empty()) {
                    shouldDrop |= (m_executiveBacklogLimit > 0 && m_executiveBacklog > m_executiveBacklogLimit);
                    --m_executiveBacklog;
                }
            }
        }
        if (shouldDrop) {
            if (m_isReproducible && framework_.client()) {
                // Tell SCS that we dropped this message.
                SimpleComms::DropMessage dropMessage;

                std::strncpy(&dropMessage.origin[0],
                             message.key.first.c_str(),
                             MAX_NULL_TERM_CLIENT_NAME); // Not guaranteed to append a null-terminator
                dropMessage.origin[MAX_NULL_TERM_CLIENT_NAME - 1] = '\0';

                dropMessage.id = htonl(message.key.second); // NOLINT(cppcoreguidelines-pro-type-vararg)

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
                std::string blob(reinterpret_cast<const char*>(&dropMessage), sizeof(dropMessage));
                framework_.client()->sendMsg(SIMPLECOMMS_CONTROL_CHANNEL, blob);
            } else {
                AVLOG_WARN("Dropping message {} {} ", message.key.first, message.key.second);
            }
            continue;
        }
        if (log_) {
            // If this is an executive trigger, use the new client-side topic and origin names.
            bool isExecutive = message.channel.empty();
            std::string topic = isExecutive ? "!periodic" : message.channel;
            // Strip the "-$$$$" random suffix from the origin name.
            std::string origin = isExecutive ? "scheduler" : message.key.first.substr(0, message.key.first.rfind('-'));
            uint32_t id = av::repro::parseID(message.blob);
            log_.write(message.timestamp.count(), framework_.outputID(), topic, origin, id, id);
        }
        // Other messages can be enqueued while we're processing this one.
        // Make sure the callbacks don't change while we're calling them.
        std::shared_lock lock(m_callbackMutex);
        // Collect timing statistics.
        auto begin = std::chrono::steady_clock::now();
        // Note: In general, the callback set isn't changing once task setup is
        // done, so there's no reason to worry about how long we hold onto this
        // mutex for.
        auto range = m_callbacks.equal_range(message.channel);
        for (auto it = range.first; it != range.second; ++it) {
            (*it->second)(message.blob);
        }
        timing[message.channel].add(begin, std::chrono::steady_clock::now());
    }

    // Print timing statistics.
    for (const auto& it : timing) {
        double frequency = it.second.count() > 1 ? it.second.frequency() : 0.0;
        const std::string callbackName{it.first.empty() ? "Executive" : it.first};

        AVLOG_INFO("{} handled {} times ({})",
                   AVLOG_ARGS("callback", "calls", "frequency"),
                   callbackName,
                   it.second.count(),
                   frequency);

        if (it.second.count() > 0) {
            AVLOG_INFO("    averaging {} ms (s.d. {}, max {})",
                       AVLOG_ARGS("mean", "stdev", "max"),
                       it.second.mean(),
                       it.second.sigma(),
                       it.second.max());
        }
    }
}

} // namespace task
