/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an implementation of the interface for handling closing signals
 */

#include "lib/avlog/avlog.hpp"
#include <task/ClosingSignalHandler.h>
#include <utility/utility.h>

#include <algorithm>
#include <cerrno>
#include <chrono>
#include <cstring>
#include <fstream>
#include <iostream>
#include <set>
#include <sys/types.h>

namespace {
// The maximum number of caught signals before exiting the whole process.
constexpr int maxSignalCaught{3};
// These signals are not valid closing signals as they cannot be blocked.
const std::set<int> illegalClosingSignals{SIGKILL, SIGSTOP};
} // namespace

namespace task {

ClosingSignalHandler::ClosingSignalHandler()
{
    // Clear the signal mask.
    sigemptyset(&signalSet_);
}

ClosingSignalHandler::~ClosingSignalHandler()
{
    if (handlingThread_.joinable()) {
        joinFlag_ = true;
        pthread_kill(handlingThread_.native_handle(), SIGINT);
        handlingThread_.join();
    }
}

bool ClosingSignalHandler::shouldExit() const
{
    return exitFlag_;
}

void ClosingSignalHandler::wait(std::optional<uint64_t> timeoutMS) const
{
    std::unique_lock<std::mutex> lock(thrd_mtx_);

    if (timeoutMS.has_value()) {
        const std::chrono::milliseconds timeoutDur(timeoutMS.value());
        exitCondition_.wait_for(lock, timeoutDur, [this] { return exitFlag_.load(); });
    } else {
        // TODO: C++20 alows waiting directly on an atomic variable, avoiding this condition variable.
        exitCondition_.wait(lock, [this] { return exitFlag_.load(); });
    }
}

void ClosingSignalHandler::quit()
{
    {
        std::unique_lock<std::mutex> lock(thrd_mtx_);
        exitFlag_ = true;
    }
    exitCondition_.notify_all();
}

bool ClosingSignalHandler::setSignals(const std::vector<int>& closingSignals)
{
    AVLOG_INFO("ClosingSignalHandler::setSignals([{}])", middleware::utility::join(closingSignals, ","));

    // Note: the only reason any of these functions can fail is because of an invalid
    // input, and we always use the same set of signal numbers that are known to be
    // valid, so this function should never fail. Just in case, check for errors:
    if (closingSignals.empty()) {
        return false;
    }
    for (int signalNumber : closingSignals) {
        if (illegalClosingSignals.count(signalNumber) > 0) {
            AVLOG_ERROR("Could not add signal {} to signalSet_: it's an illegal closing signal", signalNumber);
            return false;
        }
        if (sigaddset(&signalSet_, signalNumber) != 0) {
            int tmp = errno;
            AVLOG_ERROR("Could not add signal {} to signalSet_: {}", signalNumber, strerror(tmp));
            return false;
        }
    }
    if (pthread_sigmask(SIG_BLOCK, &signalSet_, nullptr) != 0) {
        AVLOG_ERROR("Could not set signal mask: {}", strerror(errno));
        return false;
    }
    return true;
}

bool ClosingSignalHandler::handleClosingSignals()
{
    if (!handlingThread_.joinable()) {
        // Spawn a thread to handle the closing signals if no one is running
        handlingThread_ = std::thread(&ClosingSignalHandler::threadEntry, this);
    } else {
        AVLOG_WARN("The closing signal handling thread is running.");
    }

    return handlingThread_.joinable();
}

std::string pidName(pid_t pid)
{
    using namespace std::string_literals;
    const std::string path = "/proc/"s + std::to_string(pid) + "/cmdline";
    std::ifstream f(path);

    if (!f.is_open()) {
        return "??";
    }

    std::string firstWord{};
    f >> firstWord;

    if (firstWord.empty()) {
        return "??";
    }
    return firstWord;
}

void ClosingSignalHandler::threadEntry()
{
    while (!joinFlag_) {
        siginfo_t siginfo;
        const int caughtSigNum = sigwaitinfo(&signalSet_, &siginfo);
        if (caughtSigNum < 0) {
            const std::string message = std::strerror(errno);
            AVLOG_ERROR("Failed to wait for the closing signals with error {}", message);
            continue;
        }
        if (joinFlag_) {
            // this signal is notifying us to join with main thread.
            AVLOG_INFO("Caught signal while joinFlag_ is true.");
            break;
        }
        // Increase the total number of caught signals
        numberSignalsCaught_++;

        const pid_t pid = siginfo.si_pid;
        const std::string processName = pidName(pid);

        // If caught more than maximum number of signals, exit
        if (numberSignalsCaught_ >= maxSignalCaught) {
            AVLOG_WARN("Caught signal {} ({}) from PID {} (\"{}\"), setting exit_flag to true",
                       caughtSigNum,
                       strsignal(caughtSigNum),
                       pid,
                       processName);
            AVLOG_WARN(
                "Caught signal {} ({}) from PID {} (\"{}\"), not exiting gracefully after {} closing signals, will "
                "forcefully exit",
                caughtSigNum,
                strsignal(caughtSigNum),
                pid,
                processName,
                maxSignalCaught);
            // Exit the task process
            exit(1);
        } else {
            AVLOG_WARN("Caught signal {} ({}) from PID {} (\"{}\"), setting exit_flag to true",
                       caughtSigNum,
                       strsignal(caughtSigNum),
                       pid,
                       processName);
            // Set the exit flag. This must be done under protection of the mutex in order
            // to avoid a race that could deadlock the condition variable.
            {
                std::lock_guard<std::mutex> lk(thrd_mtx_);
                exitFlag_ = true;
            }
            exitCondition_.notify_all();
        }
    }
}

} // namespace task
