/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an implementation for processing command line options
 */

#include <task/CmdLineOptProcessor.h>

#include <cstdlib>
#include <iostream>

#include "lib/avlog/avlog.hpp"

namespace task {

CmdLineOptProcessor::CmdLineOptProcessor(std::shared_ptr<VersionPrinterInterface> versionPrinter)
    : versionPrinter_(std::move(versionPrinter))
{
    // All options are not set at creating
    options_.fill({nullptr, 0, nullptr, 0});
}

CmdLineOptMap CmdLineOptProcessor::parseCmdLine(int argc, char** argv)
{
    CmdLineOptMap cmdMap{};
    if (numberOfOptions_ > 0) {
        // Reset the parsing
        optind = 0;
        // Parse the input command line arguments w.r.t the current options
        auto opt = getopt_long(argc, argv, shortOptions_.c_str(), &options_[0], nullptr);
        while (opt != -1) {
            // Process the parsed option
            if (!processParsedOpt(opt, cmdMap)) {
                AVLOG_ERROR("Failed to process the option {}", AVLOG_ARGS("option"), argv[optind - 1]);
            }
            // Parse the next argument
            opt = getopt_long(argc, argv, shortOptions_.c_str(), &options_[0], nullptr);
        }
    }

    return cmdMap;
}

bool CmdLineOptProcessor::addOption(const option& opt, const std::string& usageMessage)
{
    bool result{false};
    // Add only if the opt.val is in range [-128, 127] of char type
    if ((opt.val >= -128) && (opt.val <= 127)) {
        // Add only if the it does not reach the maximum number of options
        if (numberOfOptions_ < options_.max_size()) {
            std::string shortOpt{1, static_cast<char>(opt.val)};
            result = true;
            // Check the option type
            switch (opt.has_arg) {
                case no_argument:
                    break;
                case optional_argument:
                    shortOpt += "::";
                    break;
                case required_argument:
                    shortOpt += ":";
                    break;
                default:
                    AVLOG_ERROR("Failed to add because of unknown option type");
                    result = false;
                    break;
            }
            if (result) {
                // Add the given option into the current list
                options_.at(numberOfOptions_) = opt;
                // Update the short option string
                shortOptions_ += shortOpt;
                // Update the index map
                optionIndexMap_[opt.val] = std::make_pair(numberOfOptions_, usageMessage);
                // Update the number of options
                numberOfOptions_++;
            }
        } else {
            // Out of the maximum number of options
            AVLOG_ERROR("Failed to add because the number of options is already maximum that is {}",
                        AVLOG_ARGS("maximum"),
                        options_.max_size());
        }
    } else {
        AVLOG_ERROR("Failed to add because the short option is not a char");
    }
    return result;
}

bool CmdLineOptProcessor::addDefaultOptions()
{
    bool result{true};
    for (const auto& opt : defaultOpts) {
        result &= addOption(opt.first, opt.second);
    }

    return result;
}

void CmdLineOptProcessor::clearOptions()
{
    options_.fill({nullptr, 0, nullptr, 0});
    // Reset the number of options
    numberOfOptions_ = 0;
    shortOptions_ = "";
    // Clear the index map
    optionIndexMap_.clear();
}

void CmdLineOptProcessor::setExitOnRequestFlag(const bool exitOnRequestFlag)
{
    exitOnRequest_ = exitOnRequestFlag;
}

bool CmdLineOptProcessor::processParsedOpt(const int parsedShortOpt, CmdLineOptMap& cmdMap)
{
    bool result{false};
    // Check to see if the short option is valid
    auto optIndex = optionIndexMap_.find(parsedShortOpt);
    if (optIndex != optionIndexMap_.end()) {
        const auto opt = options_.at(optIndex->second.first);
        const auto longOptionName = std::string{opt.name};
        const auto optValue = optarg != nullptr ? std::string{optarg} : std::string{""};
        cmdMap.setOptValue(longOptionName, optValue);

        // If it is help argument, print the usage messages
        if (longOptionName == "help") {
            printHelpMessage();
            if (exitOnRequest_) {
                std::exit(0);
            }
        }
        // If it is version printing
        else if (longOptionName == "version") {
            if (versionPrinter_ != nullptr) {
                auto buildNumbers = versionPrinter_->getAVStackBuildNumber(optValue);
                buildNumbers.printBuildNumbers();
            }
            if (exitOnRequest_) {
                std::exit(0);
            }
        }

        result = true;
    }

    return result;
}

void CmdLineOptProcessor::printHelpMessage()
{
    std::cout << "Command Line Options" << std::endl;
    for (const auto& optIndex : optionIndexMap_) {
        const auto& opt = options_.at(optIndex.second.first);
        std::cout << opt.name << "\n\t" << optIndex.second.second << std::endl;
    }
}

} // namespace task
