/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an implementation for sending heartbeats at a frequency defined in the task's configuration file
 */

#include "lib/avlog/avlog.hpp"
#include <Clock/SystemClock.h>
#include <TimeStamp/otimeConversions.h>
#include <avtime/Time.hpp>
#include <task/HeartbeatSender.h>
#include <unistd.h>

namespace task {

HeartbeatSender::HeartbeatSender(std::unique_ptr<HeartbeatPublisher> heartbeatPublisher,
                                 const double cycleTime_s,
                                 const double taskCycleRate_hz)
    : cycleTime_s_(cycleTime_s)
    , heartbeatPublisher_(std::move(heartbeatPublisher))
{
    lastUpdateTime_ = SystemClock::getSynchronizedSteadyTime();
    cycleEndTime_ = lastUpdateTime_;

    if (taskCycleRate_hz > 0.0) {
        // clang-tidy 6.0 complains about slicing the ottotime objects here, because oNanoSeconds is derived from
        // otimeDuration. Because the derived class holds no data and is just a convenience ctor, this doesn't cause any
        // issues and can safely be ignored.
        //
        // NOLINTNEXTLINE(cppcoreguidelines-slicing)
        taskCycleDuration =
            ottotime::oNanoSeconds{static_cast<int64_t>(static_cast<double>(NSEC_PER_SEC) / taskCycleRate_hz)};
    }

    // Setup the heartbeat message
    heartbeatMessage_.source = unInitializedSender;
    heartbeatMessage_.status = unInitializedSender;
    heartbeatMessage_.cycleRate_hz = -1.0;
    heartbeatMessage_.intendedCycleRate_hz = taskCycleRate_hz;
    heartbeatMessage_.host = unInitializedSender;
}

bool HeartbeatSender::sendHeartbeat(bool terminating)
{
    if (terminating) {
        heartbeatMessage_.status = "terminating";
    }

    bool result{false};
    // Update the time stamp
    heartbeatMessage_.timeStamp = av::timePointFromOtimePoint(SystemClock::getSynchronizedSteadyTime());
    // Send the current heartbeat message
    if (heartbeatPublisher_ != nullptr) {
        AVLOG_DEBUG("Sending a heartbeat where actual task frequency is {}",
                    AVLOG_ARGS("frequency"),
                    heartbeatMessage_.cycleRate_hz);
        result = heartbeatPublisher_->publishMessage(heartbeatMessage_);
    }
    if (!result) {
        AVLOG_WARN("Failed to send a heartbeat");
    }

    return result;
}

bool HeartbeatSender::needToSendHeartbeat()
{
    bool result{false};
    // Increase the number of task cycles
    numCycles_++;
    const ottotime::otimePoint currentTime = SystemClock::getSynchronizedSteadyTime();

    // This gets called once per task execution cycle, so to ensure the heartbeat is never late if it would be
    // due before the next task execution send it now.
    //
    // This avoids the heartbeat being sent in the cycle after it expires, ie 1 cycle late, which for very low
    // frequency tasks (eg something that runs at 1 Hz) can be late enough to trigger warnings in safer.
    if (currentTime + taskCycleDuration > cycleEndTime_) {
        // Mark the next heartbeat cycle time

        // clang-tidy 6.0 complains about slicing the ottotime objects here, because oMicroSeconds is derived from
        // otimeDuration. Because the derived class holds no data and is just a convenience ctor, this doesn't cause any
        // issues and can safely be ignored.
        //
        // NOLINTNEXTLINE(cppcoreguidelines-slicing)
        cycleEndTime_ = currentTime + ottotime::otimeDuration{
                                          oMicroSeconds{static_cast<int64_t>(cycleTime_s_ * hbSenderSecToMicroSecs)}};
        // Compute the duration between the last
        const auto difference = (currentTime - lastUpdateTime_).totalMicroSeconds();
        // Task actual rate as average value w.r.t the above duration
        if (difference > 0) {
            heartbeatMessage_.cycleRate_hz =
                (static_cast<double>(numCycles_) * hbSenderSecToMicroSecs) / static_cast<double>(difference);
        }
        // Replace the last update time with the current time
        lastUpdateTime_ = currentTime;
        // Reset the count of task cycles
        numCycles_ = 0;
        result = true;
    }

    return result;
}

} // namespace task
