/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an implementation of MonitoringStatusSenderInterface for sending AV tasks internal status to the
 * monitoring system
 */

#include "lib/avlog/avlog.hpp"
#include <algorithm>
#include <task/MonitoringStatusSender.h>

using namespace av::msg;

namespace task {

namespace {
/**
 * \brief Second to micro seconds
 */
constexpr double msSenderSecToMicroSecs{1000000.0};

/**
 * \brief Default sending frequency
 */
constexpr double defaultFreq_Hz{1.0};
} // namespace

MonitoringStatusSender::MonitoringStatusSender(std::unique_ptr<AvTaskStatusPublisher> avTaskStatusPublisher)
    : m_avTaskStatusPublisher(std::move(avTaskStatusPublisher))
{}

bool MonitoringStatusSender::initialize(ConfigSection& taskParams)
{
    if (!m_isInitialized) {
        // Parse the sending frequency
        double sendingFreqParam_Hz{0};
        m_isInitialized = taskParams.get("monitoringStatusSendingRate_hz", sendingFreqParam_Hz, 0);
        // Get the task frequency
        double taskFreq_Hz{0};
        m_isInitialized &= taskParams.get("cycleRate_hz", taskFreq_Hz, 0);
        if (m_isInitialized) {
            setSendingFrequency(sendingFreqParam_Hz, taskFreq_Hz);
            m_cycleTime_us = static_cast<int64_t>(msSenderSecToMicroSecs / m_frequency_Hz);
            // Initialize the default state
            m_lastStatus.state = TaskState::UP;
        } else {
            AVLOG_WARN("Failed to initialize the monitoring status sender");
        }
    } else {
        AVLOG_INFO("The monitoring status sender is already initialized");
    }
    return m_isInitialized;
}

bool MonitoringStatusSender::sendStatus(const AvTaskStatus& status)
{
    bool result{false};
    // Send the current status message if it is needed
    if ((m_avTaskStatusPublisher != nullptr) && m_isInitialized) {
        if (needToSend(status)) {
            AVLOG_DEBUG("Sending the current status");
            result = m_avTaskStatusPublisher->publishMessage(m_lastStatus);
        } else {
            AVLOG_DEBUG("The current status does not need to be sent");
            result = true;
        }
    }
    if (!result) {
        AVLOG_WARN("Failed to send the current status");
    }
    return result;
}

void MonitoringStatusSender::setSendingFrequency(const double freqParam, const double taskFreq)
{
    if (freqParam > 0) {
        // Set the sending frequency is positive minimum value of user-defined and task frequencies
        m_frequency_Hz = (taskFreq > 0) ? std::min(taskFreq, freqParam) : freqParam;
    } else {
        // Set the sending frequency is positive minimum value of default and task frequencies
        m_frequency_Hz = (taskFreq > 0) ? std::min(taskFreq, defaultFreq_Hz) : defaultFreq_Hz;
    }
    AVLOG_INFO("The monitoring status sending frequency: {} Hz", AVLOG_ARGS("frequency"), m_frequency_Hz);
}

bool MonitoringStatusSender::needToSend(const AvTaskStatus& currentStatus)
{
    const auto duration_us = currentStatus.timestamp - m_lastStatus.timestamp;
    // The current status is sent if this is the first sending time, or the sending cycle is expired, or the state has
    // been changed
    const auto result = m_isFirstTime || (duration_us >= m_cycleTime_us) || (m_lastStatus.state != currentStatus.state);
    if (result) {
        // Mark it is not the first time sending
        m_isFirstTime = false;
        // Update the last sent status
        m_lastStatus = currentStatus;
    }
    return result;
}

} // namespace task
