/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines a basic building block of any application. All applications start with deriving a task class, and
 * implementing the initialize, executive, and cleanup methods. On initialization, this class will look to load a
 * configuration file with the same name as the taskName with a '.cfg' suffix in the config file directory. A
 * task can be configured to loop as fast as possible by providing a cycleRate of zero (0) in the configuration file.
 * Otherwise, a value in seconds is specified as the frequency at which to loop. The default is to loop as fast as
 * possible (no delay. This is useful if the application blocks in the executive loop, and is event driven.
 */

#include "TaskUtil.h"
#ifdef MONITOR_MALLOC_TIMES
#include "alloc.h"
#endif
#include "lib/avlog/avlog.hpp"
#include "platform/middleware/libs/Networking/include/Networking/Network.h"
#include <Clock/SystemClock.h>
#include <Comms/ChannelFactory.h>
#include <ProgressReporter/ProgressReporter.h>
#include <SimpleComms/Client.h>
#include <SimpleComms/ControlMessages.h>
#include <SimpleComms/UnixClient.h>
#include <TimeStamp/otimeConversions.h>
#include <avtime/ConversionsTimespec.hpp>
#include <interfaces/Generic/Input/Channel/Impl.h>
#include <interfaces/Generic/Output/Channel/Impl.h>
#include <interfaces/ProcessHealthMessage/ProcessHealthMessage.h>
#include <task/BackupRecoveryHandler.h>
#include <task/HeartbeatSender.h>
#include <task/MonitoringStatusSender.h>
#include <task/Task.h>
#include <task/TaskTime.hpp>

#if !defined(__aarch64__)
#include <DDSComms/DDSClient.hpp>
#endif

#include <algorithm>
#include <optional>
#include <vector>

using namespace task;
using namespace logger;
using namespace ottotime;
using namespace std::placeholders;
using namespace av::msg;
using namespace avlog;

/**
 * \brief Internal constants of the task framework
 */
namespace {
const std::string configDirEnv{"OTTO_CONFIG_DIR"}; ///< The configuration directory environment variable
const std::string ottoDirEnv{"OTTO_DIR"};          ///< The AV stack base directory environment variable
const std::vector<std::string> commonInterfaceNames{"ProcessHealthOutput",
                                                    "SaferCommandInput",
                                                    "SaferResponseOutput"}; ///< Common interfaces used by all AV tasks
const std::string defaultStatusMsg{
    "Task has not registered a status provider callback"}; ///< The default message for the monitoring status
const std::string defaultParamName{
    "status_provider_callback_provided"}; ///< The default parameter name for the monitoring status
} // namespace

Task::Task(std::string taskName)
    : taskName_(std::move(taskName))
    , callbacks_(framework_)
{
    time::TaskTimeProvider::setTimeSource([this]() { return getSystemTime_(); });
}

void Task::run(ClosingSignalHandlerInterface& closingSignal)
{
    closingSignal_ = &closingSignal;
    // Start the event framework and the scheduler.
    framework_.start();
    closingSignal.wait();
    // Stop the event framework.
    framework_.quit();
}

bool Task::_initialize()
{
    bool status{false};

    // Initialize parameters from the config file
    status = initParameters(getConfigFileDirectory());

    std::string commSystem;
    taskParams_.get("CommSystem", commSystem, "SimpleComms");

    // Check if this task should be included in the reproducible message log.
    std::string clockType;
    taskParams_.get("clockType", clockType, "system");
    // force override to system clock if we are using DDS
    if (commSystem == "DDS") {
        clockType = "system";
    }
    // The "reproducible" config parameter controls whether this task saves a
    // client-side event log for reproducible resim.
    bool configuredReproducible = false;
    taskParams_.get("reproducible", configuredReproducible, false);
    std::string eventLogPath;
    if (configuredReproducible) {
        eventLogPath = taskName_ + ".events";
    }
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-vararg)
    const char* reproducible = getenv("REPRODUCIBLE");
    if (reproducible != nullptr) {
        // If the environment variable is set, the task may be placed in live
        // or frozen mode. If the task itself is not reproducible, it will
        // never be placed in replay mode, because there is no log to replay.
        // During replay, non-reproducible tasks should run in frozen mode.
        if (reproducible == std::string("live")) {
            reproMode_ = av::repro::EventFramework::Mode::live;
        } else if (reproducible == std::string("frozen") || !configuredReproducible) {
            reproMode_ = av::repro::EventFramework::Mode::frozen;
        } else {
            reproMode_ = av::repro::EventFramework::Mode::replay;
            eventLogPath = reproducible + ("/" + eventLogPath);
        }
    }
    // If the clock type is set to "central," use central scheduling unless
    // we are running in replay mode.
    if (clockType == "central" && reproMode_ != av::repro::EventFramework::Mode::replay) {
        reproMode_ = av::repro::EventFramework::Mode::central;
    }
    // Check if we need to subscribe to playback clock messages.
    const char* clockTypeEnv = getenv("SYSTEM_CLOCK_TYPE");
    bool useExternalClock = (clockTypeEnv != nullptr && clockTypeEnv == std::string("external"));
    if (reproMode_ == av::repro::EventFramework::Mode::frozen || useExternalClock) {
        framework_.declare("PlaybackClockParams", SimpleComms::ChannelType::Subscriber);
    }

    // If task must be run in live mode, then override previous settings
    // For instance: this forces safer to run in live mode even when REPRODUCIBLE env var is set
    bool forceLiveMode = false;
    taskParams_.get("forceLiveMode", forceLiveMode, false);
    if (forceLiveMode) {
        reproMode_ = av::repro::EventFramework::Mode::live;
    }

    taskParams_.get("cycleRate_hz", frequencyHz_, 10.);
    if (status) {
        ConfigSection metricsParameters;
        tp_.getMetricsParameters(metricsParameters);
        networkStatsSender_ = std::make_unique<network::NetworkStatsSender>(
            getCurrentTaskTime(),
            [this](const std::string& interfaceName) {
                return getWriteOnlyInterface<network::NetworkStatsSender::Interface>(interfaceName);
            },
            [this](const std::string& interfaceName) {
                return getWriteOnlyInterface<network::NetworkStatsSender::AlertInterface>(interfaceName);
            },
            getTaskName(),
            getHostName(),
            metricsParameters,
            static_cast<int>(frequencyHz_));
    }

    // Connect to IPC
    if (status) {
        std::string modname{taskName_ + "-" + getHostName()};
        AVLOG_INFO("Connecting to IPC with module name: {}", modname);
        // Construct the comms client.
        SimpleComms::Client::Callback messageHandler;
        if (reproMode_ == av::repro::EventFramework::Mode::central) {
            // In central mode, bypass the EventFramework and send messages to CallbackSet,
            // so that it can handle message dropping with its own logic.
            // NOLINTNEXTLINE(modernize-avoid-bind)
            messageHandler = std::bind(&CallbackSet::call, &callbacks_, _1, _2, _3);
        } else {
            // Otherwise, messages go to EventFramework.
            using Key = CallbackSet::MessageKey;
            messageHandler = [this](const std::string& topic, const std::string& blob, const Key& key) {
                framework_.receive(blob, topic, key.first);
            };
        }
        std::shared_ptr<SimpleComms::Client> client;
        if (commSystem == "SimpleComms") {
            // Construct the comms client.
            // TODO: In DDS mode, construct a DDS client instead.
            auto scsPath = SimpleComms::UnixClient::getDefaultScsPath();
            auto clientName = SimpleComms::UnixClient::makeUniqueClientName(taskName_);
            client = std::make_shared<SimpleComms::UnixClient>(scsPath, clientName, messageHandler);
            ChannelFactory::useClient(client);

        } else if (commSystem == "DDS") {
#if !defined(__aarch64__)
            DDSComms::DDSOptionsType options = initDdsOptions(); // inital settings from base task
            populateDdsOptions(options); // allow individual tasks to override via template method
            auto ddsConfigurator = std::make_shared<DDSComms::DDSConfiguratorType>(options);
            client = std::make_shared<DDSComms::DDSClient>(messageHandler, 0, ddsConfigurator);
#else
            throw std::runtime_error("aarch64 Task cannot use DDS.");
#endif
            // ChannelFactory does not need a client when using DDS.
        } else {
            throw std::runtime_error("Task cannot tell either to use SimpleComms or DDS.");
        }
        framework_.setClient(client);
        // Initialize the CallbackSet.
        callbacks_.init(reproMode_ == av::repro::EventFramework::Mode::central, eventLogPath);
        // Initialize the EventFramework and open the client connection.
        framework_.init(reproMode_, eventLogPath);
    }
    // Initialize the system clock. But, note that this clock should never be
    // used; user code should call Task::getSystemTime() instead.
    if (status) {
        SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeSteady);
    }
    // Init other objects such as heartbeat sender, failure recovery handler
    if (status) {
        // No error here will stop the initialization
        initHeartbeatSender();
        initBackupRecoveryHandler();
        initMonitoringStatusSender();
        initDefaultStatusMessage();
    }
    if (status) {
        channelStatusMonitor_ = std::make_unique<network::ChannelStatusMonitor>(
            [this](const std::string& interfaceName) {
                return getWriteOnlyInterface<network::ChannelStatusInterface>(interfaceName);
            },
            taskParams_);
    }
    // If everything is okay so far, call the derived class's init method
    if (status) {
        unsigned int initTimeoutConfig{0};
        taskParams_.get("initTimeout", initTimeoutConfig, 60u);
        const auto initTimeout = std::clamp(initTimeoutConfig, 1u, 300u);
        if (initTimeout != initTimeoutConfig) {
            // if your task takes more than 5 minutes to initialize it needs to be fixed.
            // do not wait for scs inputs to be available in task::initialize. check them in executive().
            // it blocks the monitoring system from sending the task status.
            AVLOG_WARN("initTimeout ({}) must be in the range 1-300, setting to {}", initTimeoutConfig, initTimeout);
        }

        std::mutex mutex;
        std::condition_variable cv;
        bool retValue{false};
        bool initDone{false};

        std::thread initThread([this, &mutex, &cv, &retValue, &initDone]() {
            retValue = initialize();
            std::unique_lock<std::mutex> lock(mutex);
            initDone = true;
            cv.notify_one();
        });

        std::unique_lock<std::mutex> lock(mutex);
        cv.wait_for(lock, std::chrono::seconds(initTimeout), [&initDone] { return initDone; });
        if (initDone) {
            status = retValue;
            AVLOG_INFO("Task::initialize() done. returned: {}", status);
        } else {
            status = false;
            AVLOG_ERROR("Task::initialize() timeout after {} seconds. Terminating...", initTimeout);
            if (pthread_cancel(initThread.native_handle()) != 0) {
                AVLOG_ERROR("Failed to kill Task::initialize() thread.");
            }
        }
        initThread.join();
    }

    if (status) {
        initExecutive();
    }

    if (status) {
        progress::report(progress::ProgressType::TASK_INITIALIZED, taskName_);
    } else {
        progress::report(progress::ProgressType::TASK_INITIALIZATION_FAILED, taskName_);
    }
    // Send an initial AvTaskStatus message to inform others that this task is fully initialized.
    sendMonitoringStatus();

    if (status) {
        m_publishChannelsThread = std::thread([this] {
            std::shared_ptr<GenericOutput<av::msg::PublishListMessage>> publishedChannelInterface =
                getWriteOnlyInterface<GenericOutputChannel<av::msg::PublishListMessage>>("PublishListMessageOutput");
            if (publishedChannelInterface == nullptr) {
                AVLOG_ERROR("_initialize(): Failed to create the PublishListMessageOutput interface");
            } else {
                while (executionRet.load() && !interrupt.load()) {
                    std::this_thread::sleep_for(std::chrono::seconds(1));
                    publishedChannelInterface->publish(getPublishListMessage());
                }
            }
        });
    }
    return status;
}

Task::HeartbeatSenderPtr Task::createHeartbeatSender()
{
    Task::HeartbeatSenderPtr hbSenderPtr{nullptr};
    // Create a default standard heartbeat sender with type HeartbeatSender
    ConfigSection heartbeatInterfaceConfig{};
    if (ConfigSection::GS_OK == taskParams_.get("ProcessHealthOutput", heartbeatInterfaceConfig)) {
        // Create the heartbeat interface and its publisher
        auto hbOutputInterface = getInterface<GenericOutputChannel<ProcessHealthMessage>>(
            "ProcessHealthOutput", heartbeatInterfaceConfig, Access::Write);
        if (hbOutputInterface != nullptr) {
            // Get the sending rate in seconds
            double heartbeatRate_s{2.0};
            taskParams_.get("HeartbeatRate", heartbeatRate_s, 2.0);

            // Get task intended rate
            double taskIntendedFreq_hz{0.0};
            taskParams_.get("cycleRate_hz", taskIntendedFreq_hz, 0.0);

            hbSenderPtr = std::make_unique<HeartbeatSender>(
                std::make_unique<HeartbeatPublisher>(hbOutputInterface, logger_), heartbeatRate_s, taskIntendedFreq_hz);
        }
    } else {
        AVLOG_ERROR("Failed to fetch the ProcessHealthOutput config section from taskParams. Are the "
                    "CommonTaskParams::Parameters included properly?");
    }
    return hbSenderPtr;
}

bool Task::initHeartbeatSender()
{
    bool result{false};
    heartbeatSender_ = createHeartbeatSender();
    if (heartbeatSender_ != nullptr) {
        // Set the heartbeat message content
        ProcessHealthMessage hbMessage{};
        hbMessage.processID = getpid();
        hbMessage.host = getHostName();
        hbMessage.source = getTaskName();
        hbMessage.status = "alive";
        double taskIntendedFreq_hz{-1.0};
        taskParams_.get("cycleRate_hz", taskIntendedFreq_hz, 0);
        hbMessage.intendedCycleRate_hz = taskIntendedFreq_hz;
        heartbeatSender_->setHeartbeatMessage(hbMessage);
        result = true;
    } else {
        AVLOG_ERROR("Failed to create the heartbeat sender");
    }
    return result;
}

bool Task::chmodInterfaces(ConfigSectionList ifclist, const Access& acc)
{
    bool result = chmodTaskInterfaces(ifclist, acc, interfaces_, interfacesManaged_);
    if (!result) {
        AVLOG_ERROR("Failed to change the mode for task interfaces");
    }
    return result;
}

bool Task::_executive()
{
    if (numExecutions_ == 0) {
        progress::report(progress::ProgressType::TASK_STARTED, taskName_);
    }

    ++numExecutions_;

    executionRet.store(true);

    if (networkStatsSender_) {
        networkStatsSender_->onTick();
    }

    if (executionRet.load()) {
        if (handlingParams_.coldStandby == 0) {
            if (profileEnabled_) {
                ottoprofile::ProfileCollector pc(taskName_);
                executionRet.store(this->executive());
            } else {
                executionRet.store(this->executive());
            }
        }
    }

    if (executionRet.load()) {
        sendMonitoringStatus();
        // Send a heartbeat
        if ((heartbeatSender_ != nullptr) && heartbeatSender_->needToSendHeartbeat()) {
            heartbeatSender_->sendHeartbeat();
        }
        // Handling command from Safer daemon for recovering failure using replicas
        if (backupRecoveryHandler_ != nullptr) {
            handlingParams_.isSteadyClkRequired = true;
            auto writeOnlyInterfaces =
                tp_.hasWriteOnlyInterfaces() ? tp_.getAllWriteOnlyInterfaces() : ConfigSectionList{};
            auto readWriteInterfaces =
                tp_.hasReadWriteInterfaces() ? tp_.getAllReadWriteInterfaces() : ConfigSectionList{};
            backupRecoveryHandler_->handleBackupRecovery(
                writeOnlyInterfaces, readWriteInterfaces, handlingParams_, interfaces_, interfacesManaged_);
        }
    }

    // There's at least one test where apparently _initialize() never
    // gets called, or is mocked, and channelStatusMonitor_ is left
    // null. So we check it before calling.
    if (channelStatusMonitor_) {
        channelStatusMonitor_->tick(av::timePointFromOtimePoint(getSystemTime()));
    }

    return executionRet.load();
}

void Task::_cleanup()
{
    interrupt.store(true);

    // Send one last heartbeat, with terminating status
    if (heartbeatSender_ != nullptr) {
        heartbeatSender_->sendHeartbeat(true);
    }

    if (profileEnabled_) {
        ottoprofile::ProfileDumper pd(taskName_);
        pd.writeToFile();
    }

    // Wait for the callback thread to stop, if it's running.
    callbacks_.stop();

    SystemClock::freeze();

    if (m_publishChannelsThread.joinable()) {
        m_publishChannelsThread.join();
    }

    // Now that there's no callback thread to call CSM::tick(), we can
    // close down the CSM.
    if (channelStatusMonitor_) {
        channelStatusMonitor_->stop();
    }
    channelStatusMonitor_.reset();

    // invoke derived class' cleanup method
    cleanup();

    // Purge the interfaces
    for (auto& interfaceKeyValue : interfacesManaged_) {
        if (interfaceKeyValue.second != nullptr) {
            interfaceKeyValue.second->cleanup();
        }
    }
    for (auto& commonIntf : taskCommonInterfaces_) {
        if (commonIntf.second != nullptr) {
            commonIntf.second->cleanup();
        }
    }

    progress::report(progress::ProgressType::TASK_EXITED, taskName_ + "," + std::to_string(numExecutions_));

    ChannelFactory::cleanup();
    if (framework_.client()) {
        framework_.client()->disconnect();
    }

#ifdef MONITOR_MALLOC_TIMES
    {
        AllocStats stats;
        alloc_fetch_stats(&stats);
        AVLOG_INFO("Memory alloc stats: {} mallocs, {} frees, {} seconds",
                   stats.numMallocs,
                   stats.numFrees,
                   static_cast<double>(stats.total_dsec) + static_cast<double>(stats.total_dusec) / 1000000);
    }
#endif

    AVLOG_INFO("Task::_cleanup(): Done cleanup");
}

ottotime::otimePoint Task::getCurrentTaskTime()
{
    // If bypassing the event framework, the time comes from CallbackSet.
    if (reproMode_ == av::repro::EventFramework::Mode::central) {
        return callbacks_.timestamp();
    }
    return av::timespecFromDuration<ottotime::otimeSpec>(framework_.frozenTime());
}

ottotime::otimePoint Task::getSystemTime_()
{
    ottotime::otimePoint systemTime{};
    auto currentTask = Task::get();
    if (currentTask != nullptr) {
        systemTime = currentTask->getCurrentTaskTime();
    } else {
        systemTime = SystemClock::getSystemTime();
    }
    if (!systemTime.isTimeValid()) {
        AVLOG_WARN("Invalid time returned from getSystemTime");
    }
    return systemTime;
}

ottotime::otimePoint Task::getSystemTime()
{
    return time::TaskTimeProvider::getSystemOTime();
}

std::string Task::getSystemErrorCodeText(const int errnum)
{
    return systemErrorCodeToString(errnum);
}

bool Task::parseLogLevelParams()
{
    bool result{false};
    // First thing: get the logger threshold so that things, quiet down if we don't want to be verbose...
    std::string verbosity{"debug"};
    result = taskParams_.get("loggerThreshold", verbosity, "debug");
    if (globalLogLevel_.empty()) {
        result &= setGlobalLogLevel(verbosity);
    } else {
        AVLOG_INFO(
            "Ignoring configured log level '{}' in favor of command-line level '{}'", verbosity, globalLogLevel_);
    }

    // This tells the event reporting system to configure
    // itself. Unless the configuration explicitly enables event
    // reporting, it'll be disabled.
    progress::Reporter::initialize(taskParams_);

    return result;
}

bool Task::parseRecoveryParams()
{
    bool result{true};
    /**
     *  In current setting, the assumption is that each task can has at most 1 replicas which is hot or cold one
     * coldStandby == 1: it is cold one
     * coldStandby == 0: it is hot one
     */
    taskParams_.get("numReplicas", handlingParams_.numReplicas, 0);
    AVLOG_INFO("Task has {} replicas.", handlingParams_.numReplicas);
    taskParams_.get("coldStandby", handlingParams_.coldStandby, 0);
    AVLOG_INFO("Task is cold: {}.", handlingParams_.coldStandby);
    const auto taskNameDelim = taskName_.find('_');
    /**
     * If this is primary or backup task, check the consistency of replicas configuration
     * numReplicas == 0 if it is the primary, otherwise, numReplicas == 1
     */
    if (taskNameDelim != std::string::npos) {
        const auto postfixTaskName = taskName_.substr(taskNameDelim + 1);
        if (postfixTaskName == "primary") {
            if (handlingParams_.numReplicas != 0) {
                result = false;
                AVLOG_ERROR("Number of replicas must be 0 in a primary task");
            } else if ((handlingParams_.numReplicas == 0) && (handlingParams_.coldStandby != 0)) {
                result = false;
                AVLOG_ERROR("Primary task cannot be a cold standby");
            }
        } else if (postfixTaskName == "backup") {
            if (handlingParams_.numReplicas != 1) {
                result = false;
                AVLOG_ERROR("Number of replicas must be 1 in a backup task");
            } else if ((handlingParams_.numReplicas == 1) && (handlingParams_.coldStandby > 1)) {
                result = false;
                AVLOG_ERROR("coldStandby must be 0 or 1 in a backup task");
            }
        }
    }
    return result;
}

bool Task::initParameters(const std::string& configDirPath)
{
    bool result{false};
    tp_.setTaskParamDirectory(configDirPath);
    if (tp_.parseTask(taskName_)) {
        if (tp_.getTaskParameters(taskParams_)) {
            // Parse log level, no error here will stop the initialization
            parseLogLevelParams();
            // Parse profiling param
            taskParams_.get("profile", profileEnabled_, false);
            AVLOG_INFO("Profile is {}.", profileEnabled_);
            // Parse backup recovery params
            result = parseRecoveryParams();
        } else {
            AVLOG_ERROR("Task's \"Parameters\" section is not defined");
        }
    } else {
        AVLOG_ERROR(
            "Error parsing task config file. If {}.cfg is not found, run config_generator to generate the cfg "
            "file from the yaml task config file before running the task. "
            "Refer to https://confluence.ci.motional.com/confluence/display/SI/AV+Configuration+System+Overview "
            "for more info about config_generator and the new AV Config System. ",
            taskName_);
    }
    return result;
}

void Task::initExecutive()
{
    // Get the requested frequency.
    taskParams_.get("cycleRate_hz", frequencyHz_, 10.);
    // Don't allow the executive to run faster than 10 kHz.
    auto nsecPeriod = std::max(100000L, static_cast<int64_t>(frequencyHz_ == 0. ? 0. : 1e9 / frequencyHz_));
    // If we're using a central scheduler, activate it now that the task is ready to run.
    if (reproMode_ == av::repro::EventFramework::Mode::central) {
        // An empty channel name represents the executive function.
        // Begin listening for it before requesting it, so we don't drop any.
        callbacks_.add("", this, &Task::executiveCallback);

        SimpleComms::ScheduleMessage scheduleMessage;
        scheduleMessage.nsecInterval = nsecPeriod;
        // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
        std::string blob(reinterpret_cast<const char*>(&scheduleMessage), sizeof(scheduleMessage));
        framework_.client()->sendMsg(SIMPLECOMMS_CONTROL_CHANNEL, blob);
    } else {
        // Schedule the executive through EventFramework.
        framework_.setTimeoutHandler("!periodic", this, &Task::executiveCallback, std::chrono::nanoseconds(nsecPeriod));
    }
}

bool Task::setGlobalLogLevel(const std::string& level)
{
    bool result{false};
    const Priority::PriorityLevel ll = Priority::stringAsPriorityLevel(level);
    if (ll != Priority::none) {
        globalLogLevel_ = level;
        avlog::LogLevel levelAvlog = avlog::utils::stringToLevel(level);
        logger::Logger::setGlobalPriorityThreshold(ll);
        avlog::setGlobalLogLevel(levelAvlog);
        // also adjust member loggers
        logger_.setThreshold(ll);
        result = true;
    } else {
        AVLOG_WARN("Ignoring none/unkown log level '{}'", level);
    }
    return result;
}

std::string Task::getConfigFileDirectory()
{
    const std::optional<std::string> dir = getEnvVariableValue(configDirEnv);
    std::string result{};
    if (dir) {
        result = *dir;
        AVLOG_INFO("Using Config Directory: {}", result);
    } else {
        AVLOG_INFO("Defaulting to current directory for config files, set {} to override the default", configDirEnv);
        result = std::string{"."};
    }
    return result;
}

std::string Task::getUCDirectory()
{
    const std::optional<std::string> dir = getEnvVariableValue(ottoDirEnv);
    std::string result{};
    if (dir) {
        result = *dir;
        AVLOG_INFO("Using top-level UC directory: {}", result);
    } else {
        AVLOG_INFO("Defaulting to current directory for {}, set it to override the default", ottoDirEnv);
        result = std::string{"."};
    }
    return result;
}

std::string Task::getHostName() const
{
    std::string hostName;
    if (!net::getHostMachineName(hostName)) {
        AVLOG_WARN("Failed to get the host machine name");
    }
    return hostName;
}

bool Task::sendStatusReport(const std::string& message)
{
    bool result{false};
    if (backupRecoveryHandler_ != nullptr) {
        result = backupRecoveryHandler_->sendSaferResponse(message);
    }
    return result;
}

Task::BackupRecoveryHandlerPtr Task::createBackupRecoveryHandler()
{
    Task::BackupRecoveryHandlerPtr returnPtr{nullptr};

    std::shared_ptr<GenericInputChannel<av::msg::SaferCommand>> inputIntfc{nullptr};
    std::shared_ptr<GenericOutputChannel<av::msg::SaferResponse>> outputIntfc{nullptr};

    ConfigSection intfCfgInp{};
    if (ConfigSection::GS_OK == taskParams_.get("SaferCommandInput", intfCfgInp)) {
        inputIntfc =
            getInterface<GenericInputChannel<av::msg::SaferCommand>>("SaferCommandInput", intfCfgInp, Access::Read);
    }

    ConfigSection intfCfgOut{};
    if (ConfigSection::GS_OK == taskParams_.get("SaferResponseOutput", intfCfgOut)) {
        outputIntfc = getInterface<GenericOutputChannel<av::msg::SaferResponse>>(
            "SaferResponseOutput", intfCfgOut, Access::Write);
    }

    // create handler if at least one safer interface is available
    if ((inputIntfc != nullptr) || (outputIntfc != nullptr)) {
        returnPtr = std::make_unique<BackupRecoveryHandler>(inputIntfc, outputIntfc, taskName_, getHostName());
    }

    if (returnPtr == nullptr) {
        AVLOG_ERROR("Failed to create the backup recovery handler");
    }
    return returnPtr;
}

bool Task::initBackupRecoveryHandler()
{
    bool result{false};
    backupRecoveryHandler_ = createBackupRecoveryHandler();
    if (backupRecoveryHandler_ != nullptr) {
        // Init the handler
        if (backupRecoveryHandler_->initialize()) {
            result = true;
        } else {
            AVLOG_ERROR("Failed to set up the backup recovery handler");
        }
    }
    return result;
}

bool Task::isCommonInterface(const std::string& interfaceName)
{
    const auto result = (std::find(commonInterfaceNames.begin(), commonInterfaceNames.end(), interfaceName) !=
                         commonInterfaceNames.end());
    if (result) {
        AVLOG_DEBUG("{} is a common interface used by all AV tasks", interfaceName);
    }
    return result;
}

Task::StatusSenderPtr Task::createMonitoringStatusSender()
{
    Task::StatusSenderPtr msSenderPtr{nullptr};
    // Create a default standard AvTaskStatus sender
    ConfigSection avTaskStatusConfig{};
    if (taskParams_.get("AvTaskStatusOutput", avTaskStatusConfig) == ConfigSection::GS_OK) {
        auto senderOutputInterface = getInterface<GenericOutputChannel<av::msg::AvTaskStatus>>(
            "AvTaskStatusOutput", avTaskStatusConfig, Access::Write);
        if (senderOutputInterface != nullptr) {
            msSenderPtr = std::make_unique<MonitoringStatusSender>(
                std::make_unique<AvTaskStatusPublisher>(senderOutputInterface, logger_));
        } else {
            AVLOG_WARN("Failed to create AvTaskStatusOutput interface for sending the monitoring status");
        }
    } else {
        AVLOG_WARN("Failed to fetch the AvTaskStatusOutput from CommonTaskParams.Parameters");
    }
    return msSenderPtr;
}

bool Task::initMonitoringStatusSender()
{
    bool result{false};
    m_monitoringStatusSender = createMonitoringStatusSender();
    if (m_monitoringStatusSender != nullptr) {
        // Initialize the sender
        result = m_monitoringStatusSender->initialize(taskParams_);
    } else {
        AVLOG_WARN("Failed to create the monitoring status sender");
    }
    return result;
}

void Task::initDefaultStatusMessage()
{
    m_defaultAvTaskStatus.taskName = taskName_;
    m_defaultAvTaskStatus.state = TaskState::DEGRADED;
    m_defaultAvTaskStatus.source = FaultSource::INTERNAL;
    m_defaultAvTaskStatus.message = defaultStatusMsg;
    m_defaultAvTaskStatus.type = ParameterType::BOOLEAN;
    m_defaultAvTaskStatus.parameter_name = defaultParamName;
    m_defaultAvTaskStatus.measured_value = "false";
    m_defaultAvTaskStatus.threshold_value = "true";
}

void Task::sendMonitoringStatus()
{
    // Send the monitoring status
    if (m_monitoringStatusSender) {
        // Update the timestamp
        m_defaultAvTaskStatus.timestamp = getCurrentTaskTime().toMicrosecondsSinceEpoch();
        m_monitoringStatusSender->sendStatus(m_monitoringStatusProvider ? m_monitoringStatusProvider->getStatus()
                                                                        : m_defaultAvTaskStatus);
    }
}

void Task::executiveCallback()
{
    // Don't call the executive again if it's already said we should quit.
    if (closingSignal_ != nullptr && closingSignal_->shouldExit()) {
        return;
    }
    if (!_executive() && closingSignal_ != nullptr) {
        closingSignal_->quit();
    }
}

av::msg::PublishListMessage Task::getPublishListMessage()
{
    if (framework_.client()) {
        std::lock_guard<std::mutex> guard(publishedChannelMutex_);
        for (const auto& channelName : framework_.client()->getPublishListMessage()) {
            {
                if (std::find(publishedChannels_.channels.begin(), publishedChannels_.channels.end(), channelName) ==
                    publishedChannels_.channels.end()) {
                    publishedChannels_.channels.emplace_back(channelName);
                }
            }
        }
    }
    return publishedChannels_;
}

DDSComms::DDSOptionsType Task::initDdsOptions()
{
    DDSComms::DDSOptionsType ddsOptions;

    // All PandarQT128C2X and VLS128 Lidar UDP Channel readers should be best effort
    ddsOptions.readChannelProfileMap.emplace("PandarQT128C2X_F_UdpPacketList", "AVQosLib::Generic.BestEffort");
    ddsOptions.readChannelProfileMap.emplace("PandarQT128C2X_L_UdpPacketList", "AVQosLib::Generic.BestEffort");
    ddsOptions.readChannelProfileMap.emplace("PandarQT128C2X_R_UdpPacketList", "AVQosLib::Generic.BestEffort");
    ddsOptions.readChannelProfileMap.emplace("PandarQT128C2X_B_UdpPacketList", "AVQosLib::Generic.BestEffort");
    ddsOptions.readChannelProfileMap.emplace("VLS128_T_UdpPacketList", "AVQosLib::Generic.BestEffort");

    // All PandarQT128C2X and VLS128 Lidar UDP Channel readers should be best effort
    ddsOptions.writeChannelProfileMap.emplace("PandarQT128C2X_F_UdpPacketList", "AVQosLib::Generic.BestEffort");
    ddsOptions.writeChannelProfileMap.emplace("PandarQT128C2X_L_UdpPacketList", "AVQosLib::Generic.BestEffort");
    ddsOptions.writeChannelProfileMap.emplace("PandarQT128C2X_R_UdpPacketList", "AVQosLib::Generic.BestEffort");
    ddsOptions.writeChannelProfileMap.emplace("PandarQT128C2X_B_UdpPacketList", "AVQosLib::Generic.BestEffort");
    ddsOptions.writeChannelProfileMap.emplace("VLS128_T_UdpPacketList", "AVQosLib::Generic.BestEffort");

    // MPC is 20hz, so after about 5 missed messages, we need to move on. Allow sender more history.
    ddsOptions.writeChannelProfileMap.emplace("MergedPointCloud", "AVQosLib::Generic.KeepLast10ReliableAsyncWrite");
    ddsOptions.readChannelProfileMap.emplace("MergedPointCloud", "AVQosLib::Generic.KeepLast5Reliable");

    // these are relatively high bandwidth.
    ddsOptions.writeChannelProfileMap.emplace("PnSDebug", "AVQosLib::Generic.KeepLast25ReliableAsyncWrite");
    ddsOptions.writeChannelProfileMap.emplace("HomotopyOverPlan0", "AVQosLib::Generic.KeepLast25ReliableAsyncWrite");

    return ddsOptions;
}
