/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines the standard AV task application
 */

#include "TaskApplication.h"
#include <task/ClosingSignalHandler.h>
#include <task/CmdLineOptProcessor.h>
#include <task/VersionPrinter.h>

#include "lib/avlog/avlog.hpp"

#include <csignal>
#include <cstdio>

#ifdef __linux__
// The headers in this conditional block are used to set process security bits
// on Linux to restrict root process rights. These bits are an extension of
// POSIX capabilities that only exist on Linux.
#include <sys/prctl.h>
#include <unistd.h>
#endif

/**
 * \brief Internal constants
 */
namespace {
/** This bit pattern is a combination of SECURE_NOROOT, SECURE_NO_ROOT_LOCKED, SECURE_NO_SETUID_FIXUP,
 *  SECURE_NO_SETUID_FIXUP_LOCKED, and SECURE_KEEP_CAPS_LOCKED. At the time of this writing, those
 *  constants are not defined in a Linux system header in a usable form. Collectively these ensure
 *  that tasks run as root do not get any special privileges (NOROOT) and simplify setresuid if it's used
 *  (NO_SETUID_FIXUP). The _LOCKED bits ensure the relevant bit cannot be changed again (i.e., in this
 *  case we ensure that none of the currently supported security bits can be changed after we set them
 *  initially).
 */
const unsigned long SECURE_BIT_LOCK = 0x2F;
} // namespace

namespace task {

TaskApplication::TaskApplication(std::unique_ptr<ClosingSignalHandlerInterface> closingSignalHandler,
                                 std::unique_ptr<CmdLineOptProcessorInterface> cmdLineOptProcessor,
                                 std::shared_ptr<VersionPrinterInterface> versionPrinter,
                                 Task* task)
    : closingSignalHandler_(std::move(closingSignalHandler))
    , cmdLineOptProcessor_(std::move(cmdLineOptProcessor))
    , versionPrinter_(std::move(versionPrinter))
    , task_(task)
{}

bool TaskApplication::initTask(int argc, char** argv)
{
    if (!isInitialized_) {
        isInitialized_ = (closingSignalHandler_ != nullptr) && (cmdLineOptProcessor_ != nullptr) &&
                         (versionPrinter_ != nullptr) && (task_ != nullptr);
        if (isInitialized_) {
            // Handle the closing signals
            isInitialized_ &= closingSignalHandler_->handleClosingSignals();
            // Parse the command line argument into a CmdLineOptMap object
            if (cmdLineOptProcessor_->addDefaultOptions()) {
                cmdLineOptMap_ = cmdLineOptProcessor_->parseCmdLine(argc, argv);
            } else {
                isInitialized_ = false;
            }
            // Set the secure bits, ignore the error
            setSecureBits();
            // Configure the task
            isInitialized_ &= configureTask();
            // Print the vehicle/platform name
            versionPrinter_->getVehicleName(defaultVehicleConfigFile, true);
            // Print build numbers
            versionPrinter_->getAVStackBuildNumber(task_->getTaskName(), defaultAVBuildTags, true);
        }
    } else {
        AVLOG_INFO("Task application is already initialized");
    }
    return isInitialized_;
}

bool TaskApplication::runTask()
{
    bool result{true};
    if (isInitialized_ && !isStarted_) {
        if (task_->_initialize()) {
            const auto configTest = cmdLineOptMap_.isOptSet("configTest");
            if (!configTest) {
                task_->run(*closingSignalHandler_);
            } else {
                AVLOG_INFO("Config Test Only: Skipping executive");
            }
            isStarted_ = true;
        }
        // Cleanup before exiting
        task_->_cleanup();
        result = !(closingSignalHandler_->shouldExit());
    } else if (isStarted_) {
        AVLOG_INFO("Task application is already running");
    } else {
        AVLOG_WARN("Task application is not initialized");
    }
    return result;
}

bool TaskApplication::setSecureBits()
{
    bool result{true};
#ifdef __linux__
    /** This call should fail if we are not root. However, we attempt to perform the operation anyway
     * in case a task is run as another uid that has enough privileges to set process security bits. We
     * do this because tasks should not require root privileges. If a task is run as root accidentally
     * or because an attacker elevated the process's privileges, we want to ensure the process does not
     * have the full power of root.
     */
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-vararg)
    const auto secureBitsRet = prctl(PR_SET_SECUREBITS, SECURE_BIT_LOCK);
    const auto uid = getuid();
    if (uid == 0) {
        AVLOG_WARN("Task run as a root! SECURE_NOROOT status: {}", secureBitsRet);
    } else if (secureBitsRet == 0) {
        AVLOG_WARN("Task was not run as root, but it has enough privileges to set security bits. This "
                   "requires root or the CAP_SETPCAP capability. SECURE_NOROOT is set.");
    }
    result = (secureBitsRet != -1);
#endif
    return result;
}

bool TaskApplication::configureTask()
{
    bool result{true};
    // Set log level defined by the command line argument
    std::string cmdLineLogLevel;
    if (cmdLineOptMap_.getOptValue("logLevel", cmdLineLogLevel)) {
        result &= task_->setGlobalLogLevel(cmdLineLogLevel);
    }
    // Set task name defined by the command line argument
    std::string cmdLineTaskName;
    if (cmdLineOptMap_.getOptValue("instanceName", cmdLineTaskName)) {
        task_->setTaskName(cmdLineTaskName);
    }
    return result;
}

} // namespace task
