/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines the standard AV task application
 */

#ifndef TASK_APPLICATION_H
#define TASK_APPLICATION_H

#include <task/ClosingSignalHandlerInterface.h>
#include <task/CmdLineOptProcessorInterface.h>
#include <task/Task.h>
#include <task/TaskApplicationInterface.h>
#include <task/VersionPrinterInterface.h>

#include <memory>
#include <string>

namespace task {

/**
 * \brief Defines an interface for AV task application
 */
class TaskApplication : public TaskApplicationInterface
{
public:
    /**
     * \brief Default, copy constructors and assignment are deleted
     */
    TaskApplication() = delete;
    TaskApplication(const TaskApplication&) = delete;
    TaskApplication& operator=(const TaskApplication&) = delete;
    TaskApplication(const TaskApplication&&) = delete;
    TaskApplication& operator=(const TaskApplication&&) = delete;

    /**
     * \brief User-defined constructor
     *
     * \param closingSignalHandler The closing signal handler
     * \param cmdLineOptProcessor The command line options processor
     * \param versionPrinter The version printer
     * \param task The task implementation
     */
    explicit TaskApplication(std::unique_ptr<ClosingSignalHandlerInterface> closingSignalHandler,
                             std::unique_ptr<CmdLineOptProcessorInterface> cmdLineOptProcessor,
                             std::shared_ptr<VersionPrinterInterface> versionPrinter,
                             Task* task);

    /**
     * \brief Default destructor
     */
    virtual ~TaskApplication() = default;

    /**
     * \brief Initialize the task application
     *
     * \param argc The number of arguments
     * \param argv The list of arguments
     *
     * \return True if the creation is fine, false, otherwise
     */
    bool initTask(int argc, char** argv) override;

    /**
     * \brief Run the task application
     *
     * \param Void
     *
     * \return True if the running is fine, false, otherwise
     */
    bool runTask() override;

private:
    /**
     * \brief Set the secure bits w.r.t security policy
     *
     * \note Should call this function only when isCreated_ is true
     *
     * \param Void
     *
     * \return True of the setting is fine, false, otherwise
     */
    bool setSecureBits();

    /**
     * \brief Configure the task application w.r.t the command line options
     *
     * \note Should call this function only when isCreated_ is true
     *
     * \param Void
     *
     * \return True if the configuration is fine, false. otherwise
     */
    bool configureTask();

    std::unique_ptr<ClosingSignalHandlerInterface> closingSignalHandler_{nullptr}; ///< The closing signal hander
    std::unique_ptr<CmdLineOptProcessorInterface> cmdLineOptProcessor_{nullptr}; ///< The command line option processor
    std::shared_ptr<VersionPrinterInterface> versionPrinter_{nullptr};           ///< The version printer
    Task* task_{nullptr};           ///< The AV task implementation, raw pointer is used due to the legacy code
    CmdLineOptMap cmdLineOptMap_{}; ///< The command line options
    bool isInitialized_{false};     ///< True if the task application is initialized
    bool isStarted_{false};         ///< True if the task application is started
};

} // namespace task

#endif // TASK_APPLICATION_H
