/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines a set of helper functions for the AV task framework
 */
#include "TaskUtil.h"

#include <features.h>
#include <sys/utsname.h>
#include <unistd.h>

#include <algorithm>
#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <optional>
#include <sstream>

namespace task {

long double elapsedTimeInmilliSeconds(const ottotime::otimePoint& startTime, const ottotime::otimePoint& endTime)
{
    ottotime::otimeDuration otDuration = endTime - startTime;
    return static_cast<long double>(otDuration.totalMilliSeconds());
}

long double convertTomilliSeconds(const ottotime::otimePoint& aTime)
{
    ottotime::otimePoint otp{aTime};
    ottotime::otimeDuration otd(otp);
    return static_cast<long double>(otd.totalMilliSeconds());
}

std::string systemErrorCodeToString(const int errnum)
{
    std::string result;
    char buf[1024];
    /**
     * GNU extension version of strerror_r, where the returned char * pointer may not be the buffer provided
     * It is a pain in the ass to enable the POSIX version, especially in C++ where the GNU STL seems to rely on GNU C
     * functionality
     */
    const char* const errorStringPtr = strerror_r(errnum, static_cast<char*>(buf), 1024);
    if (errorStringPtr != nullptr) {
        result = std::string(errorStringPtr);
    } else {
        std::stringstream unknown;
        unknown << "Unknown errno #" << errnum;
        result = unknown.str();
    }
    return result;
}

bool validateInterfaceConfig(ConfigSection& interfaceConfig,
                             std::string& instanceName,
                             std::string& interfaceName,
                             std::string& implName)
{
    bool result{false};
    // Get the class name that we should instantiate
    interfaceConfig.get("interface", interfaceName, "");
    interfaceConfig.get("implementation", implName, "");
    // Get the name we should use as the key for looking up the object that will be created
    instanceName = interfaceConfig.getName();
    // Strip the quotes from the name, since it is really annoying to have to escape the quotes when doing lookups
    instanceName.erase(std::remove(instanceName.begin(), instanceName.end(), '\"'), instanceName.end());
    result = (!interfaceName.empty()) && (!implName.empty()) && (!instanceName.empty());
    return result;
}

bool needRecoveryWithBackup(const std::string& receivedTaskName, const std::string& backupTaskName)
{
    bool result{false};
    // Check this task and primary task are valid
    const auto primaryDelim = receivedTaskName.find('_');
    const auto backupDelim = backupTaskName.find('_');
    if ((primaryDelim != std::string::npos) && (backupDelim != std::string::npos)) {
        const auto prefixPrimary = receivedTaskName.substr(0, primaryDelim);
        const auto prefixBackup = backupTaskName.substr(0, backupDelim);
        const auto postfixPrimary = receivedTaskName.substr(primaryDelim + 1);
        const auto postfixBackup = backupTaskName.substr(backupDelim + 1);
        result = (prefixPrimary == prefixBackup) && (postfixBackup == "backup") && (postfixPrimary == "primary");
    }
    return result;
}

std::optional<std::string> getEnvVariableValue(const std::string& envVarName)
{
    std::optional<std::string> result{};
    const auto envVarVal = getenv(envVarName.c_str());
    if (envVarVal != nullptr) {
        result = std::string(envVarVal);
    }
    return result;
}

bool chmodTaskInterfaces(ConfigSectionList& ifclist,
                         const Access& acc,
                         std::map<std::string, IDbObject*>& rawPtrInterfaces,
                         std::map<std::string, std::shared_ptr<IDbObject>>& sharedPtrInterfaces)
{
    bool result{true};
    for (auto& interfaceConfig : ifclist) {
        std::string instanceName;
        std::string interfaceName;
        std::string implName;
        if (validateInterfaceConfig(interfaceConfig, instanceName, interfaceName, implName)) {
            // Chmod interfaces
            auto rawPtrIntfIter = rawPtrInterfaces.find(instanceName);
            if ((rawPtrIntfIter != rawPtrInterfaces.end()) && (rawPtrIntfIter->second != nullptr)) {
                rawPtrIntfIter->second->chmod(acc);
            }
            auto sharedPtrIntfIter = sharedPtrInterfaces.find(instanceName);
            if ((sharedPtrIntfIter != sharedPtrInterfaces.end()) && (sharedPtrIntfIter->second != nullptr)) {
                sharedPtrIntfIter->second->chmod(acc);
            }
            result &= true;
        } else {
            result = false;
        }
    }
    return result;
}

} // namespace task
