/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines a set of helper functions for the AV task framework
 */

#ifndef TASK_FRAMEWORK_UTIL_H
#define TASK_FRAMEWORK_UTIL_H

#include <Comms/Access.h>
#include <Comms/InterfaceDb.h>
#include <TimeStamp/otime.h>
#include <config/ConfigSection.h>

#include <map>
#include <memory>
#include <optional>
#include <string>

namespace task {

/**
 * \brief Compute the time duration in milliseconds between two given time points
 *
 * \param startTime The starting time
 * \param endTime The ending time
 *
 * \return The time duration
 */
long double elapsedTimeInmilliSeconds(const ottotime::otimePoint& startTime, const ottotime::otimePoint& endTime);

/**
 * \brief Convert the given time point into milliseconds from epoch
 *
 * \param aTime The time point
 *
 * \return The time duration
 */
long double convertTomilliSeconds(const ottotime::otimePoint& aTime);

/**
 * \brief Wrap the strerror() call which retrieves the text corresponding to an error
 *
 * \param errnum The system error of interest
 *
 * \return std::string containing some reasonable interpretation of given errnum
 */
std::string systemErrorCodeToString(int errnum);

/**
 * \brief Check an interface configuration is valid. The interface instance name is returned when it is valid
 *
 * \param interfaceConfig The interface configuration. It cannot be const because its get() method is not const
 * \param[out] instanceName The returned instance name
 * \param[out] interfaceName The returned interface name
 * \param[out] implName The interface implementation name
 *
 * \return True if the interface is valid
 */
bool validateInterfaceConfig(ConfigSection& interfaceConfig,
                             std::string& instanceName,
                             std::string& interfaceName,
                             std::string& implName);

/**
 * \brief Check that the recovery with backup is needed. If the received task name is a primary task and this task is
 * its backup then return true
 *
 * \param receivedTaskName The task name from received message
 * \param backupTaskName The current back up task name
 *
 * \return True the recovery is needed
 */
bool needRecoveryWithBackup(const std::string& receivedTaskName, const std::string& backupTaskName);

/**
 * \brief Gets the value pointed by the given environment variable as a string. Return original input string value if
 * the environment variable is not defined. For example, it is used to get the path to the config directory defined by
 * the environment variable OTTO_CONFIG_DIR
 *
 * \param envVarName The environment variable name
 *
 * \return An optional containing the value defined by the variable,
 * if present, or an empty optional if not.
 */
std::optional<std::string> getEnvVariableValue(const std::string& envVarName);

/**
 * \brief Changes the mode of all interfaces given a list of config sections
 *
 * \param ifclist The interface list for looking up in the given map
 * \param acc The access mode (R/W/RW/Standby) to be set
 * \param rawPtrInterfaces The raw pointer interfaces to be changed mode
 * \param sharedPtrInterfaces The shared pointer interfaces to be changed mode
 *
 * \return True if the interface modes were successfully changed
 */
bool chmodTaskInterfaces(ConfigSectionList& ifclist,
                         const Access& acc,
                         std::map<std::string, IDbObject*>& rawPtrInterfaces,
                         std::map<std::string, std::shared_ptr<IDbObject>>& sharedPtrInterfaces);

} // namespace task

#endif // TASK_FRAMEWORK_UTIL_H
