/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an implementation for printing vehicle name, si-middleware version, and build number at av-stack
 */

#include <task/VersionPrinter.h>

#include <cstdlib>
#include <fstream>

#include "lib/avlog/avlog.hpp"

namespace task {

std::string VersionPrinter::getVehicleName(const std::string& configFile, const bool logMode)
{
    std::string vehicleName{"N/A"};
    // Check if the env ACTIVE_CONFIG_DIR is defined
    const auto activeConfigDir = getenv("ACTIVE_CONFIG_DIR");
    if (activeConfigDir != nullptr) {
        const std::string configFilePath = std::string(activeConfigDir) + "/" + configFile;
        // Get the vehicle/platform name as the first line of the file
        std::ifstream fileHandle{configFilePath};
        if (fileHandle.fail()) {
            AVLOG_ERROR("Failed to open the file {} for getting the vehicle name", AVLOG_ARGS("path"), configFilePath);
        } else {
            std::string firstLine;
            if (std::getline(fileHandle, firstLine) && !firstLine.empty()) {
                vehicleName = firstLine;
            }
            fileHandle.close();
        }
    }
    // Log the vehicle name if needed
    if (logMode) {
        AVLOG_INFO("Vehicle name: {}", AVLOG_ARGS("vehicleName"), vehicleName);
    }

    return vehicleName;
}

AVStackBuildNumber VersionPrinter::getAVStackBuildNumber(const std::string& mode,
                                                         const std::string& buildTagFile,
                                                         const bool logMode)
{
    AVStackBuildNumber buildNumbers{};
    // Get av-stack build number from the build tags file
    buildNumbers.avStackNumber = getAVStackBuildNumberFromTag(buildTagFile);
    // The playback mode and comparision are kept unchanaged for code compability
    if (mode == "OttoView_playback") {
        // Get the av-stack build number from logs
        buildNumbers.avStackNumberFromLog = getAVStackBuildNumberFromLogs();
        buildNumbers.playbackMode = true;
    }
    // Log the build numbers if needed
    if (logMode) {
        AVLOG_INFO("AV-Stack build number: {}", AVLOG_ARGS("buildNumber"), buildNumbers.avStackNumber);
        if (buildNumbers.playbackMode) {
            AVLOG_INFO(
                "AV-Stack build number from logs: {}", AVLOG_ARGS("logBuildNumber"), buildNumbers.avStackNumberFromLog);
        }
    }

    return buildNumbers;
}

std::string VersionPrinter::getAVStackBuildNumberFromTag(const std::string& buildTagFile)
{
    std::string buildNumber{"N/A"};
    // Check if the env OTTO_CONFIG_DIR is defined
    const auto configDir = getenv("OTTO_CONFIG_DIR");
    if (configDir != nullptr) {
        const std::string buildTagFilePath = std::string(configDir) + "/" + buildTagFile;
        // Get the av-stack build number
        std::ifstream fileHandle{buildTagFilePath};
        if (fileHandle.fail()) {
            AVLOG_ERROR("Failed to open the file {} for getting the av-stack build number from build tags",
                        AVLOG_ARGS("path"),
                        buildTagFilePath);
        } else {
            const std::string buildNumberSubStr{"build_number="};
            std::string readLine;
            while (std::getline(fileHandle, readLine)) {
                // Read until the line that contains the buildNumberSubStr
                if (readLine.find(buildNumberSubStr) != std::string::npos) {
                    // Extract the build number
                    buildNumber = readLine.substr(buildNumberSubStr.size(), readLine.size() - buildNumberSubStr.size());
                    break;
                }
            }
            fileHandle.close();
        }
    }

    return buildNumber;
}

std::string VersionPrinter::getAVStackBuildNumberFromLogs()
{
    std::string buildNumber{"N/A"};
    const std::string driveInfoSummaryFilePath = std::string("./") + driveInfoSummaryFile;
    std::ifstream fileHandle{driveInfoSummaryFilePath};
    if (fileHandle.fail()) {
        AVLOG_ERROR("Failed to open the file {} for getting the av-stack build number from logs",
                    AVLOG_ARGS("path"),
                    driveInfoSummaryFilePath);
    } else {
        const std::string buildNumberSubStr{"Software version : "};
        std::string readLine;
        while (std::getline(fileHandle, readLine)) {
            // Read until the line that contains the buildNumberSubStr
            if (readLine.find(buildNumberSubStr) != std::string::npos) {
                // Extract the build number
                buildNumber = readLine.substr(buildNumberSubStr.size(), readLine.size() - buildNumberSubStr.size());
                break;
            }
        }
        fileHandle.close();
    }

    return buildNumber;
}

} // namespace task
