/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 *
 * All information contained herein is, and remains the property of
 * Motional and its suppliers, if any.  The intellectual and technical
 * concepts contained herein are proprietary to Motional and its
 * suppliers and may be covered by U.S. and Foreign Patents, patents
 * in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Motional.
 *
 * \brief When this module is linked in, it will measure the time
 * spent on malloc() and free().
 */

/**
 * This module is written in C instead of C++ because C++ can allocate
 * in unexpected places, which makes it difficult to make a malloc()
 * that doesn't reenter.
 *
 * This module doesn't uses mutexes to protect its stats on time used,
 * because it's not really necessary. The goal of IR-473 is to get us
 * some numbers on how much we're spending on allocation, so that we
 * can make the case for moving to a zero-allocation design. The lack
 * of mutexes may result in an undercount, but what we need to worry
 * about is an overcount (which could lead us to spend a lot of effort
 * on the new design based on flawed data). An undercount should be
 * fine.
 */

// Needed for RTLD_NEXT
#define _GNU_SOURCE

#include <assert.h>
#include <dlfcn.h>
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/time.h>

#include "platform/middleware/libs/task/alloc.h"

// Note that this total will max out somewhere around 2 kiloseconds, which is
// about half an hour. If allocation is as much as 5% of elapsed time (seems
// high), this means we'd overflow after about ten hours of testing. That
// should be OK, for code that's only meant to run in a test whose purpose
// is to measure allocation time.
static long total_dusec = 0;

static void timestamp_before(struct timespec* timestamp)
{
    clock_gettime(CLOCK_MONOTONIC, timestamp);
}

static void timestamp_after(struct timespec* timestamp)
{
    struct timespec now;

    clock_gettime(CLOCK_MONOTONIC, &now);

    const long dusec = (now.tv_nsec - timestamp->tv_nsec) / 1000;
    const long dsec = now.tv_sec - timestamp->tv_sec;

    const long dusec_combined = (dsec * 1000000) + dusec;

    total_dusec += dusec_combined;
}

typedef void* (*MallocPtr)(size_t);
typedef void (*FreePtr)(void*);

static MallocPtr mallocCache = NULL;
static FreePtr freeCache = NULL;

static pthread_once_t lookup_once_control = PTHREAD_ONCE_INIT;

static void initFuncPtrs(void)
{
    *(void**)(&mallocCache) = dlsym(RTLD_NEXT, "malloc");
    if (mallocCache == NULL) {
        exit(2);
    }
    *(void**)(&freeCache) = dlsym(RTLD_NEXT, "free");
    if (freeCache == NULL) {
        exit(2);
    }
}

static MallocPtr getMalloc(void)
{
    pthread_once(&lookup_once_control, initFuncPtrs);
    return mallocCache;
}

static FreePtr getFree(void)
{
    pthread_once(&lookup_once_control, initFuncPtrs);
    return freeCache;
}

static unsigned long numMallocs = 0;
static unsigned long numFrees = 0;

void alloc_write_report()
{
    fprintf(stderr, "Time: %ld.%03ld\n", total_dusec / 1000000, total_dusec % 1000000);
    fprintf(stderr, "# mallocs: %ld\n", numMallocs);
    fprintf(stderr, "# frees: %ld\n", numFrees);
}

void alloc_fetch_stats(AllocStats* stats)
{
    stats->total_dsec = total_dusec / 1000000;
    stats->total_dusec = total_dusec % 1000000;
    stats->numMallocs = numMallocs;
    stats->numFrees = numFrees;
}

void* malloc(size_t size)
{
    ++numMallocs;
    MallocPtr m = getMalloc();
    if (m == NULL) {
        return NULL;
    }

    struct timespec timestamp;
    timestamp_before(&timestamp);

    void* res = m(size);
    timestamp_after(&timestamp);
    return res;
}

void free(void* p)
{
    ++numFrees;
    FreePtr f = getFree();
    if (f != NULL) {
        struct timespec timestamp;
        timestamp_before(&timestamp);

        f(p);
        timestamp_after(&timestamp);
    }
}
