/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines the standard backup recovery handler
 */

#ifndef BACKUP_RECOVERY_HANDLER_H
#define BACKUP_RECOVERY_HANDLER_H

#include "BackupRecoveryHandlerInterface.h"
#include <interfaces/SaferCommand/SaferCommand.h>
#include <interfaces/SaferResponse/SaferResponse.h>

#include <string>

namespace task {

/**
 * \brief Defines the standard backup recovery handler
 */
class BackupRecoveryHandler : public BackupRecoveryHandlerInterface
{
public:
    /**
     * \brief Deleted default constructor
     */
    BackupRecoveryHandler() = delete;

    /**
     * \brief User-defined constructor
     * \param[in] saferCmdInput Shared pointer to Generic Input Interface with message
     *   of type SaferCommand
     * \param[in] saferCmdInput Shared pointer to Generic Output Interface with message
     *   of type SaferResponse
     * \param[in] taskName The task name that contains this handler
     * \param[in] hostName The machine name that runs the task
     */
    BackupRecoveryHandler(std::shared_ptr<GenericInputIntf<av::msg::SaferCommand>> saferCmdInput,
                          std::shared_ptr<GenericOutputIntf<av::msg::SaferResponse>> saferRespOutput,
                          std::string taskName,
                          std::string hostName);

    /**
     * \brief Default destructor
     * \details This class does not clean up interface pointers as they are
     * assumed to be externally managed by InterfaceDb.
     */
    virtual ~BackupRecoveryHandler() = default;

    /**
     * \brief Initialize the handler
     *
     * \return True if the initialization is fine, false, otherwise
     */
    bool initialize() override;

    /**
     * \brief Handler the backup recovery requested from Safer
     *
     * \param[in] writeOnlyInterfaces The task's list of write only interface configs
     * \param[in] readWriteInterfaces The task's list of read/write interface configs
     * \param[in,out] handlingParams The recovery handling parameters
     * \param[in,out] rawPtrInterfaces Raw pointer interfaces map
     * \param[in,out] sharedPtrInterfaces Shared pointer interfaces map
     *
     * \return True if the handling is fine, false, otherwise
     */
    bool handleBackupRecovery(ConfigSectionList& writeOnlyInterfaces,
                              ConfigSectionList& readWriteInterfaces,
                              BackupRecoveryParameter& handlingParams,
                              std::map<std::string, IDbObject*>& rawPtrInterfaces,
                              std::map<std::string, std::shared_ptr<IDbObject>>& sharedPtrInterfaces) override;

    /**
     * \brief Sends the given message to the Safer daemon
     *
     * \param reportMsg The message to be sent
     *
     * \return True if the message was sent, false, otherwise
     */
    bool sendSaferResponse(const std::string& reportMsg) override;

private:
    /**
     * \brief Handle a hung task requested by safer
     *
     * \param saferCmd The safer command
     * \param writeOnlyInterfaces The task's list of write only interface configs
     * \param readWriteInterfaces The task's list of read/write interface configs
     * \param handlingParams The recovery handling parameters
     * \param rawPtrInterfaces Raw pointer interfaces map
     * \param sharedPtrInterfaces Shared pointer interfaces map
     *
     * \return True if the handling is fine, false, otherwise
     */
    bool handleHungTask(const av::msg::SaferCommand& saferCmd,
                        ConfigSectionList& writeOnlyInterfaces,
                        ConfigSectionList& readWriteInterfaces,
                        BackupRecoveryParameter& handlingParams,
                        std::map<std::string, IDbObject*>& rawPtrInterfaces,
                        std::map<std::string, std::shared_ptr<IDbObject>>& sharedPtrInterfaces);

    /**
     * \brief Handle a rejoined task requested by safer
     *
     * \param saferCmd The safer command
     * \param writeOnlyInterfaces The task's list of write only interface configs
     * \param readWriteInterfaces The task's list of read/write interface configs
     * \param handlingParams The recovery handling parameters
     * \param rawPtrInterfaces Raw pointer interfaces map
     * \param sharedPtrInterfaces Shared pointer interfaces map
     *
     * \return True if the handling is fine, false, otherwise
     */
    bool handleRejoinedTask(const av::msg::SaferCommand& saferCmd,
                            ConfigSectionList& writeOnlyInterfaces,
                            ConfigSectionList& readWriteInterfaces,
                            BackupRecoveryParameter& handlingParams,
                            std::map<std::string, IDbObject*>& rawPtrInterfaces,
                            std::map<std::string, std::shared_ptr<IDbObject>>& sharedPtrInterfaces);

    std::shared_ptr<GenericInputIntf<av::msg::SaferCommand>> saferCmdInput_{
        nullptr}; ///< The pointer to the SaferCommand Input interface
    std::shared_ptr<GenericOutputIntf<av::msg::SaferResponse>> saferResponseOutput_{
        nullptr}; ///< The pointer to the SaferResponse Output interface

    std::string taskName_{""}; ///< The task name that contains this handler
    std::string hostName_{""}; ///< The machine name that runs the task
};

} // namespace task

#endif // BACKUP_RECOVERY_HANDLER_H
