/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an interface for backup recovery handler
 */

#ifndef BACKUP_RECOVERY_HANDLER_INTERFACE_H
#define BACKUP_RECOVERY_HANDLER_INTERFACE_H

#include <Comms/Access.h>
#include <Comms/InterfaceDb.h>
#include <config/TaskParser.h>

#include <map>
#include <memory>
#include <string>

namespace task {

/**
 * \brief Backup recovery parameters. In current setting, the assumptions are:
 * - Each task can has at most 1 replicas. Valid values are 0 and 1
 *      + replicas == 0: for primary task
 *      + replicas == 1: for backup task
 *
 * - A backup (or replicas) can be hot or cold. Valid values are 0 and 1
 *      + coldStandby == 1: it is cold one
 *      + coldStandby == 0: it is hot one
 */
struct BackupRecoveryParameter
{
    bool isSteadyClkRequired{false}; ///< True if the steady clock is required by the task under handling
    unsigned int coldStandby{0};     ///< Indicate the backup (or replicas) is cold or hot standby
    unsigned int numReplicas{0};     ///< The number of standby, cold and hot ones
};

/**
 * \brief Defines an interface for backup recovery handler
 */
class BackupRecoveryHandlerInterface
{
public:
    /**
     * \brief Default constructor
     */
    BackupRecoveryHandlerInterface() = default;

    /**
     * \brief Default destructor
     */
    virtual ~BackupRecoveryHandlerInterface() = default;

    /**
     * \brief Initialize the handler such as the interfaces initialization
     *
     * \return True if the initialization is fine, false, otherwise
     */
    virtual bool initialize() = 0;

    /**
     * \brief Handler the backup recovery requested from Safer
     *
     * \param[in] writeOnlyInterfaces The task's list of write only interface configs
     * \param[in] readWriteInterfaces The task's list of read/write interface configs
     * \param[in,out] handlingParams The recovery handling parameters
     * \param[in,out] rawPtrInterfaces Raw pointer interfaces map
     * \param[in,out] sharedPtrInterfaces Shared pointer interfaces map
     *
     * \return True if the handling is fine, false, otherwise
     */
    virtual bool handleBackupRecovery(ConfigSectionList& writeOnlyInterfaces,
                                      ConfigSectionList& readWriteInterfaces,
                                      BackupRecoveryParameter& handlingParams,
                                      std::map<std::string, IDbObject*>& rawPtrInterfaces,
                                      std::map<std::string, std::shared_ptr<IDbObject>>& sharedPtrInterfaces) = 0;

    /**
     * \brief Sends the given message to the Safer daemon
     *
     * \param reportMsg The message to be sent
     *
     * \return True if the message was sent, false, otherwise
     */
    virtual bool sendSaferResponse(const std::string& reportMsg) = 0;
};

} // namespace task

#endif // BACKUP_RECOVERY_HANDLER_INTERFACE_H
