/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Definition of the CallbackSet class.
 */
#pragma once

#include "MessageCallback.h"

#include <SimpleComms/Client.h>
#include <TimeStamp/otime.h>
#include <avdb/SQLWriter.hpp>
#include <repro/EventFramework.hpp>

#include <chrono>
#include <condition_variable>
#include <map>
#include <memory>
#include <mutex>
#include <queue>
#include <shared_mutex>
#include <string>
#include <thread>
#include <utility>
#include <vector>

namespace task {

/// \brief A collection of message callbacks, also storing a queue of incoming messages.
/// \note Channels that are managed through CallbackSet do not need to be listed in the configuration file.
/// \note All channels are referred to by their "true" names, not any aliases defined in the configuration file.
class CallbackSet
{
public:
    /// \brief Unique message identifier: the origin name and a rolling message ID.
    typedef std::pair<std::string, uint32_t> MessageKey;

public:
    /// \brief Default constructor.
    CallbackSet(av::repro::EventFramework& framework);
    /// \brief Destructor, which must stop the callback thread if it is running.
    ~CallbackSet();

    /// \brief Initialize the client connection.
    /// \param isReproducible: True if this client should be included in SCS's message log.
    void init(bool isReproducible, const std::string& logPath = "");
    /// \brief Stop the callback thread (because we're asking this task to quit).
    void stop();

    /// \brief Set the threshold for dropping "stale" messages.
    /// \param seconds: A maximum message age, in seconds.
    void setStaleThreshold(double seconds);
    /// \brief Set the number of executive functions that are allowed to build up in the queue.
    /// \param count: The maximum count; if more executive triggers arrive, the oldest will be dropped.
    void setExecutiveBacklogLimit(int count);

    /// \brief Register an input callback on the given channel.
    /// \param channel: The name of the channel.
    /// \param receiver: A pointer to the object receiving notification.
    /// \param callback: The member function that should be called on that object.
    /// \note: The receiver must be an object that will exist up until stop() is called on this object.
    template<class Receiver, class Type>
    void add(const std::string& channel, Receiver* receiver, void (Receiver::*callback)(const Type&));
    template<class Receiver>
    void add(const std::string& channel, Receiver* receiver, void (Receiver::*callback)());

    /// \brief Call the callback(s) for the given input.
    /// \param channel: The channel the message was received from.
    /// \param blob: The message payload, as a serialized blob (including an InterfaceHeader).
    /// \param key: The message's unique identifier.
    void call(const std::string& channel, const std::string& blob, const MessageKey& key);

    /// Get the pointer to the event framework.
    av::repro::EventFramework& framework() const;

    /// \brief Get the timestamp associated with the most recent message.
    /// \return A timestamp as an otimePoint.
    /// \note If this is called within a callback, it is always the timestamp for that message.
    const otimePoint& timestamp() const;

    /// \brief Check if we're in "replay" mode.
    /// \bool True if SCS is replaying a previous run.
    bool isReplaying() const;

private:
    // Check if an input callback is registered for the given channel.
    bool has(const std::string& channel) const;
    // If the callback thread is not running, launch it.
    void launchThread();
    // Callback thread entry point.
    void callbackThread();

private:
    // Patching legacy use of CallbackSet with the new client-side EventFramework:
    av::repro::EventFramework& framework_;
    // For declaring interfaces, sending Drop messages, etc:
    std::shared_ptr<SimpleComms::Client> m_client;
    // Remember whether we're connected to SCS yet or not.
    bool m_connected{false};
    // Input/output channel declarations made before the connection was established.
    std::vector<std::pair<std::string, SimpleComms::ChannelType>> m_toDeclare;

    // The registered callbacks:
    std::multimap<std::string, std::unique_ptr<MessageCallback>> m_callbacks;
    // TODO: With C++17, this could be a shared_mutex.
    mutable std::shared_timed_mutex m_callbackMutex;

    // The message queue:
    struct Message
    {
        std::string channel;
        std::string blob;
        std::chrono::nanoseconds timestamp;
        MessageKey key;
    };
    std::queue<Message> m_messages;
    mutable std::mutex m_messageMutex;
    std::condition_variable m_messageCondition;
    // Variables for deciding when to drop messages:
    std::chrono::nanoseconds m_latestTimestamp{0};
    std::chrono::nanoseconds m_staleThreshold{250000000};
    int m_executiveBacklog{0};
    int m_executiveBacklogLimit{0};
    // The callback thread, and a variable for interrupting it:
    std::thread m_callbackThread;
    otimePoint m_timestamp;
    bool m_interrupt{false};
    // Store whether we're replaying:
    bool m_isReplaying{false};
    bool m_isReproducible{false};
    // Client-side reproducibility log:
    using Log = av::db::SQLWriter<int64_t, uint32_t, std::string_view, std::string_view, uint32_t, uint32_t>;
    Log log_;
};

// Templated function specializations:

template<class Receiver, class Type>
void CallbackSet::add(const std::string& channel, Receiver* receiver, void (Receiver::*callback)(const Type&))
{
    // The CallbackSet's collection of callbacks is used only in "central" mode.
    // Once everything is switched to DDS, we will phase this out.
    using Function = void (Receiver::*)(const Type&);
    {
        std::scoped_lock lock(m_callbackMutex);
        m_callbacks.emplace(channel, std::make_unique<MessageCallbackT<Receiver, Function>>(receiver, callback));
    }
    // If we have a client-side event framework, register a callback with it.
    // Whether that callback or the one above gets called will depend on
    // whether the task is in "central" mode or not.
    av::repro::MessageHandler<Type> handler = [receiver, callback](const av::repro::Inputs<Type>& inputs) {
        // Call the callback function separately on each input.
        for (const av::repro::Message<Type>& it : inputs) {
            (receiver->*callback)(*it);
        }
    };
    framework_.setMessageHandler(channel, handler);
}

template<class Receiver>
void CallbackSet::add(const std::string& channel, Receiver* receiver, void (Receiver::*callback)())
{
    using Function = void (Receiver::*)();
    {
        std::scoped_lock lock(m_callbackMutex);
        m_callbacks.emplace(channel, std::make_unique<MessageCallbackT<Receiver, Function>>(receiver, callback));
    }
    framework_.declare(channel, SimpleComms::ChannelType::Subscriber);
}

} // namespace task
