/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an implementation for processing command line options using getopt_long()
 */

#ifndef CMD_LINE_OPT_PROCESSOR_H
#define CMD_LINE_OPT_PROCESSOR_H

#include <task/CmdLineOptProcessorInterface.h>
#include <task/VersionPrinterInterface.h>

#include <array>
#include <memory>
#include <unordered_map>
#include <utility>

namespace task {

/**
 * \brief The maximum number of options supported
 */
constexpr std::size_t MaxNumberOpts = 256;

/**
 * \brief The list of default options
 */
const std::array<std::pair<option, std::string>, 5> defaultOpts{
    {{{"help", no_argument, nullptr, 'h'}, "produce this help message"},
     {{"instanceName", required_argument, nullptr, 'i'}, "the task's instance name"},
     {{"configTest", required_argument, nullptr, 'c'},
      "run a configuration check only (config, initialize, then exit)"},
     {{"logLevel", required_argument, nullptr, 'l'}, "run at an alternate log level"},
     {{"version", no_argument, nullptr, 'v'}, "the version from the repository this binary was built from"}}};

/**
 * \brief The map type from short options to their indexes in options_ and usage messages
 */
using ShortOptIndexMap = std::unordered_map<int, std::pair<std::size_t, std::string>>;

/**
 * \brief This class defines an implementation for processing command line options using getopt_long()
 */
class CmdLineOptProcessor : public CmdLineOptProcessorInterface
{
public:
    /**
     * \brief The default constructor
     */
    CmdLineOptProcessor() = delete;

    /**
     * \brief The user-defined constructor
     *
     * \param versionPrinter The pointer to a version printer object
     */
    CmdLineOptProcessor(std::shared_ptr<VersionPrinterInterface> versionPrinter);

    /**
     * \brief The default destructor
     */
    virtual ~CmdLineOptProcessor() = default;

    /**
     * \brief Parse the given command line options
     *
     * \param argc The number of options
     * \param argv The list of options and values
     *
     * \return The parsed options as a CmdLineOptMap value
     */
    CmdLineOptMap parseCmdLine(int argc, char** argv) override;

    /**
     * \brief Add an option for parsing the command line arguments. The added option must have type struct option to be
     * compatible with getopt_long()
     *
     * \param opt The option to be added
     * \param usageMessage The option usage message
     *
     * \return True if the adding is fine, false otherwise
     */
    bool addOption(const option& opt, const std::string& usageMessage = std::string{""}) override;

    /**
     * \brief Add the default options for parsing the command line arguments
     *
     * \param Void
     *
     * \return True if the adding is fine, false otherwise
     */
    bool addDefaultOptions() override;

    /**
     * \brief Clear the current options for parsing the command line arguments
     *
     * \param Void
     *
     * \return Void
     */
    void clearOptions() override;

    /**
     * \brief Specify whether to invoke `std::exit` upon encountering help or version flags
     *
     * \param exitOnRequestFlag Whether to enable or disable this functionality
     */
    void setExitOnRequestFlag(const bool exitOnRequestFlag);

private:
    /**
     * \brief Process a parsed command line argument returned by getopt_long
     *
     * \param parsedShortOpt The parsed option in the short version
     * \param[out] The command line map to be updated
     *
     * \return true if the option is valid, false otherwise
     */
    bool processParsedOpt(const int parsedShortOpt, CmdLineOptMap& cmdMap);

    /**
     * \brief Print the help message
     *
     * \param Void
     *
     * \return Void
     */
    void printHelpMessage();

    std::array<option, MaxNumberOpts> options_{}; ///< The list of options used for parsing
    std::string shortOptions_{""};                ///< The list of short options as a string used by getopt_long()
    std::size_t numberOfOptions_{0};              ///< The current number of options
    ShortOptIndexMap optionIndexMap_{}; ///< A map from short options to their indexes in options_ and usage messages
    std::shared_ptr<VersionPrinterInterface> versionPrinter_{nullptr}; ///< The pointer to a version printer object
    bool exitOnRequest_{true}; ///< Specifies whether `std::exit` should be invoked upon encountering
                               ///< help or version flags
};

} // namespace task

#endif // CMD_LINE_OPT_PROCESSOR_H
