/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an interface for processing command line options using getopt_long()
 */

#ifndef CMD_LINE_OPT_PROCESSOR_INTERFACE_H
#define CMD_LINE_OPT_PROCESSOR_INTERFACE_H

#include <getopt.h>
#include <string>
#include <unordered_map>

namespace task {

/**
 * \brief Data structure represents command line options
 */
struct CmdLineOptMap
{
    std::unordered_map<std::string, std::string> optMap{}; ///< A map from option names to their values

    /**
     * \brief Set the given option name and its value. If the option is already set the value will be updated
     *
     * \param optName The option name
     * \param optValue The option value
     *
     * \return Void
     */
    void setOptValue(const std::string& optName, const std::string& optValue) { optMap[optName] = optValue; }

    /**
     * \brief Get the value of the given command line option
     *
     * \param optName The option name
     * \param[out] optValue The returned value
     *
     * \return True if the option has a value, otherwise false
     */
    bool getOptValue(const std::string& optName, std::string& optValue) const
    {
        bool result{false};
        if (optMap.count(optName) > 0) {
            optValue = optMap.at(optName);
            result = true;
        }

        return result;
    }

    /**
     * \brief Check that the given command line option is set
     *
     * \param optName The option name
     *
     * \return True if the option is set, otherwise false
     */
    bool isOptSet(const std::string& optName) const { return (optMap.count(optName) > 0); }
};

/**
 * \brief This class defines an interface for processing command line options using getopt_long()
 */
class CmdLineOptProcessorInterface
{
public:
    /**
     * \brief The default constructor
     */
    CmdLineOptProcessorInterface() = default;

    /**
     * \brief The default destructor
     */
    virtual ~CmdLineOptProcessorInterface() = default;

    /**
     * \brief Parse the given command line options using getopt_long()
     *
     * \param argc The number of options
     * \param argv The list of options and values
     *
     * \return The parsed options as a CmdLineOptMap value
     */
    virtual CmdLineOptMap parseCmdLine(int argc, char** argv) = 0;

    /**
     * \brief Add an option for parsing the command line arguments. The added option must have type struct option to be
     * compatible with getopt_long()
     *
     * \param opt The option to be added
     * \param usageMessage The option usage message
     *
     * \return True if the adding is fine, false otherwise
     */
    virtual bool addOption(const option& opt, const std::string& usageMessage = std::string{""}) = 0;

    /**
     * \brief Add the default options for parsing the command line arguments
     *
     * \param Void
     *
     * \return True if the adding is fine, false otherwise
     */
    virtual bool addDefaultOptions() = 0;

    /**
     * \brief Clear the current options for parsing the command line arguments
     *
     * \param Void
     *
     * \return Void
     */
    virtual void clearOptions() = 0;
};

} // namespace task

#endif // CMD_LINE_OPT_PROCESSOR_INTERFACE_H
