/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines a generic message publisher
 */

#ifndef GENERIC_MESSAGE_PUBLISHER_H
#define GENERIC_MESSAGE_PUBLISHER_H

#include <log/Logger.h>

#include <memory>
#include <string>

namespace task {

/**
 * \brief This class defines a publisher for publishing messages of type MessageType via the output interface of type
 * OutputInterfaceType
 *
 * Note that it is templatized in order to enable unit tests by mocking the output interface and messages
 */
template<class OutputInterfaceType, class MessageType>
class GenericMessagePublisher
{
public:
    /**
     * \brief The default constructor
     */
    GenericMessagePublisher() = delete;

    /**
     * \brief The user-defined constructor
     *
     * \param outputInterface The interface for publishing messages
     * \param logger The logger from the AV task
     */
    explicit GenericMessagePublisher(std::shared_ptr<OutputInterfaceType> outputInterface,
                                     const logger::Logger& logger);

    /**
     * \brief The default destructor
     */
    virtual ~GenericMessagePublisher() = default;

    /**
     * \brief Publish the given message
     *
     * \param  message Message to be published
     *
     * \return True if the message was published
     */
    virtual bool publishMessage(const MessageType& message);

private:
    std::shared_ptr<OutputInterfaceType> m_outputInterface{nullptr}; ///< Output interface for publishing messages
    logger::Logger m_logger{};                                       ///< The text logger
};

} // namespace task

// The implementation
#include "GenericMessagePublisher.def.h"

#endif // GENERIC_MESSAGE_PUBLISHER_H
