/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines a generic message subscriber
 */

#ifndef GENERIC_MESSAGE_SUBSCRIBER_H
#define GENERIC_MESSAGE_SUBSCRIBER_H

#include <log/Logger.h>

#include <memory>
#include <string>

namespace task {

/**
 * \brief This class defines a subscriber for receiving messages of type MessageType via the input interface of type
 * InputInterfaceType
 *
 * Note that it is templatized in order to enable unit tests by mocking the input interface and messages
 */
template<class InputInterfaceType, class MessageType>
class GenericMessageSubscriber
{
public:
    /**
     * \brief The default constructor
     */
    GenericMessageSubscriber() = delete;

    /**
     * \brief The user-defined constructor
     *
     * \param outputInterface The interface for publishing messages
     * \param logger The logger from the AV task
     */
    explicit GenericMessageSubscriber(std::shared_ptr<InputInterfaceType> inputInterface, const logger::Logger& logger);

    /**
     * \brief The default destructor
     */
    virtual ~GenericMessageSubscriber() = default;

    /**
     * \brief Receive a message
     *
     * \param message Message to be received
     * \param blocking Blocking mode
     *
     * \return True if the message was received
     */
    virtual bool receiveMessage(MessageType& message, const bool blocking = false);

private:
    std::shared_ptr<InputInterfaceType> m_inputInterface{nullptr}; ///< Input interface for publishing messages
    logger::Logger m_logger{};                                     ///< The text logger
};

} // namespace task

// The implementation
#include "GenericMessageSubscriber.def.h"

#endif // GENERIC_MESSAGE_SUBSCRIBER_H
