/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an implementation for sending heartbeats at a frequency defined in the task's configuration file
 */

#ifndef HEARTBEAT_SENDER_H
#define HEARTBEAT_SENDER_H

#include <Comms/InterfaceDb.h>
#include <TimeStamp/otime.h>
#include <interfaces/Generic/Output/Abstract.h>
#include <interfaces/ProcessHealthMessage/ProcessHealthMessage.h>
#include <task/GenericMessagePublisher.h>
#include <task/HeartbeatSenderInterface.h>

#include <cstdint>
#include <string>

namespace task {

/**
 * \brief Second to micro seconds
 */
constexpr double hbSenderSecToMicroSecs{1000000.0};

/**
 * \brief Uninitialized senser string
 */
const std::string unInitializedSender{"Sender Uninitialized"};

/**
 * \brief Alias of HeartbeatMsgPublisher
 */
using HeartbeatPublisher =
    GenericMessagePublisher<GenericOutput<av::msg::ProcessHealthMessage>, av::msg::ProcessHealthMessage>;

/**
 * \brief Defines an implementation for sending heartbeats at a frequency defined in the task's configuration file
 */
class HeartbeatSender : public HeartbeatSenderInterface
{
public:
    /**
     * \brief Default constructor
     */
    HeartbeatSender() = delete;

    /**
     * \brief User-defined constructor
     *
     * \param heartbeatPublisher The publisher for sending messages to Safer
     * \param cycleTime_s The heartbeat sender cycle time in seconds
     * \param taskCycleRate_hz The task cycle rate
     */
    explicit HeartbeatSender(std::unique_ptr<HeartbeatPublisher> heartbeatPublisher,
                             const double cycleTime_s,
                             const double taskCycleRate_hz);

    /**
     * \brief Default destructor
     */
    virtual ~HeartbeatSender() = default;

    /**
     * \brief Set the heartbeat with the given message
     *
     * \param hbMessage The message to be set
     *
     * \return Void
     */
    inline void setHeartbeatMessage(const av::msg::ProcessHealthMessage& hbMessage) override
    {
        heartbeatMessage_ = hbMessage;
    }

    /**
     * \brief Get the current heartbeat message
     *
     * \param Void
     *
     * \return The current heartbeat content
     */
    inline const av::msg::ProcessHealthMessage& getHeartbeatMessage() const override { return heartbeatMessage_; }

    /**
     * \brief Send a heartbeat with current message including the actual AV task process frequency calculated by the
     * heartbeat sender
     *
     * \param True if this is to be the last heartbeat sent.
     *
     * \return True if the heartbeat was successfully sent
     */
    bool sendHeartbeat(bool terminating = false) override;

    /**
     * \brief Check if a heartbeat needs to be sent, for example, the heartbeat sending cycle is expired. By doing that,
     * it also calculates the actual frequency of the AV task process every cycle time of the heartbeat sender in
     * seconds
     *
     * \param Void
     *
     * \return True if the heartbeat needs to be sent
     */
    bool needToSendHeartbeat() override;

private:
    uint32_t numCycles_{0}; ///< A variable representing the number of the AV task cycles during the sender cycle time
                            ///< It is used to calculate the task process actual frequency
    ottotime::otimePoint lastUpdateTime_{};            ///< The last timestamp that cycleUpdate() is called.
    double cycleTime_s_{2.0};                          ///< The heartbeat sender cycle time in seconds
    ottotime::otimeDuration taskCycleDuration{};       ///< The task cycle duration
    ottotime::otimePoint cycleEndTime_{};              ///< The time when the current heartbeat period ends
    av::msg::ProcessHealthMessage heartbeatMessage_{}; ///< Memory allocation for ProcessHealthMessage
    std::unique_ptr<HeartbeatPublisher> heartbeatPublisher_{nullptr}; ///< The heartbeat message publisher
};

} // namespace task

#endif // HEARTBEAT_SENDER_H
