/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Definition of the MessageCallback base class and MessageCallbackT templated class.
 */
#pragma once

#include <ODatum/ODatum.h>
#include <interfaces/baseTypes/Header.h>

#include <stdexcept>

namespace task {

/// \brief Interface for storing a function to be called when a SimpleComms message is received.
class MessageCallback
{
public:
    virtual ~MessageCallback() = default;
    /// \brief Deserialize the given blob and call the callback with the deserialized data.
    /// \param blob: A message payload as a serialized blob (including an InterfaceHeader).
    virtual void operator()(const std::string& blob) const = 0;
};

/// \brief Template for creating a callback for a specific channel and data type.
/// \note The function must be a Receiver member function taking 0 or 1 arguments and returning void.
template<class Receiver, typename Function>
class MessageCallbackT : public MessageCallback
{
public:
    /// \brief Constructor, taking a member function and the object to call it on.
    /// \param receiver: The object that will receive the callbacks.
    /// \param callback: The member function of that object to call.
    /// \note The object must be one that will exist as long as the task is running.
    MessageCallbackT(Receiver* receiver, Function callback);
};

// Template specializations:

// For callbacks taking a deserialized data type as their argument:
template<class Receiver, class Type>
class MessageCallbackT<Receiver, void (Receiver::*)(const Type&)> : public MessageCallback
{
public:
    MessageCallbackT(Receiver* receiver, void (Receiver::*callback)(const Type&))
        : m_receiver(receiver)
        , m_callback(callback)
    {}
    virtual void operator()(const std::string& blob) const override
    {
        // Do the best we can to check that this blob really is a serialized ODatum.
        // The header version is the only "magic number" we can check.
        if (blob.size() < sizeof(interfaces::baseTypes::InterfaceHeader)) {
            throw std::invalid_argument("MessageCallback: given argument is not a serialized ODatum.");
        }
        const auto& header = *reinterpret_cast<const interfaces::baseTypes::InterfaceHeader*>(blob.data());
        if (ntohl(header.headerVersion) != OTTO_INTERFACE_HEADER_VERSION) {
            throw std::invalid_argument("MessageCallback: given blob does not have the expected interface version.");
        }
        // TODO: If there are multiple callbacks registered for a given channel, it
        // would be more efficient to only deserialize once, e.g. by deserializing
        // into a boost::any and then passing it to each callback. But for now,
        // since multiple callbacks will be rare, just do this:
        // TODO: C++17 would let us use string_view rather than copying this substring.
        std::string body = blob.substr(sizeof(interfaces::baseTypes::InterfaceHeader));
        // TODO: deserializeODatum takes a non-const string as input, even though
        // the string is not modified. Switch to const reference?
        Type data;
        aptv::deserializeODatum(data, body);
        // Call the function.
        (m_receiver->*m_callback)(data);
    }

private:
    Receiver* m_receiver{nullptr};
    void (Receiver::*m_callback)(const Type&){nullptr};
};

// For callbacks taking no arguments:
template<class Receiver>
class MessageCallbackT<Receiver, void (Receiver::*)()> : public MessageCallback
{
public:
    MessageCallbackT(Receiver* receiver, void (Receiver::*callback)())
        : m_receiver(receiver)
        , m_callback(callback)
    {}
    virtual void operator()(const std::string&) const override { (m_receiver->*m_callback)(); }

private:
    Receiver* m_receiver{nullptr};
    void (Receiver::*m_callback)(){nullptr};
};

} // namespace task
