/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Definition of the MessageQueue templated class.
 */
#pragma once

#ifndef EVENT_FRAMEWORK_LIGHT
#include "CallbackSet.h"
#endif

#include "platform/middleware/libs/Comms/include/Comms/Channel.h"
#include <ODatum/ODatum.h>
#include <SimpleComms/Client.h>
#include <interfaces/baseTypes/Header.h>
#include <repro/EventFramework.hpp>
#include <task/TaskTime.hpp>

#include <random>
#include <string>

namespace task {

/// \brief Templated class for outputting messages to a particular channel.
template<class Type>
class MessageOutput
{
public:
    /// Construct an output object.
    /// @param channel: The true name of the channel (not any of the aliases in the config).
    /// @param framework: The task's EventFramework.
    MessageOutput(const std::string& channel, av::repro::EventFramework& framework);

#ifndef EVENT_FRAMEWORK_LIGHT
    /// Construct an output object.
    /// @param channel: The true name of the channel (not any of the aliases in the config).
    /// @param callbacks: The task's CallbackSet.
    MessageOutput(const std::string& channel, CallbackSet& callbacks);
#endif

    /// Default destructor.
    ~MessageOutput() = default;

    /// Publish a message, listing individual predecessors.
    template<class... Preds>
    bool publish(const Type& value, Preds... predecessors);
    /// Publish a message.
    /// @param value: The object to publish.
    bool publish(const Type& value, std::vector<av::latency::Pred>&& predecessors);
    /// Publish a message, with specific messageCounter value, and listing individual predecessors.
    template<class... Preds>
    bool publish(uint32_t messageCounter, const Type& value, Preds... predecessors);
    /// Publish a message.
    /// @param messageCounter: set messageCounter in the message header with specific value
    /// @param value: The object to publish.
    bool publish(uint32_t messageCounter, const Type& value, std::vector<av::latency::Pred>&& predecessors);

private:
    av::repro::EventFramework& framework_;
    std::string channel_;
    uint32_t publisherID_{0};
};

// Templated function definitions:

template<class Type>
MessageOutput<Type>::MessageOutput(const std::string& channel, av::repro::EventFramework& framework)
    : framework_(framework)
    , channel_(channel)
{
    // Declare this channel as an output.
    framework_.declare(channel_, SimpleComms::ChannelType::Publisher);

    // TODO: Figure out if anyone actually uses the publisher ID.
    // The code in GenericOutputChannel (Impl.def.h) calculates a hash from
    // the channel name, host name, process ID, and local time. But since
    // the goal is just a random unique integer, this is just as good:
    std::random_device gen;
    std::uniform_int_distribution<uint32_t> dist;
    publisherID_ = dist(gen);
}

#ifndef EVENT_FRAMEWORK_LIGHT
template<class Type>
MessageOutput<Type>::MessageOutput(const std::string& channel, CallbackSet& callbacks)
    : MessageOutput<Type>(channel, callbacks.framework())
{}
#endif

template<class Type>
template<class... Preds>
bool MessageOutput<Type>::publish(const Type& value, Preds... predecessors)
{
    return publish(framework_.nextOutputID(), value, std::vector<av::latency::Pred>{predecessors...});
}

template<class Type>
bool MessageOutput<Type>::publish(const Type& value, std::vector<av::latency::Pred>&& predecessors)
{
    return publish(framework_.nextOutputID(), value, predecessors);
}

template<class Type>
template<class... Preds>
bool MessageOutput<Type>::publish(uint32_t messageCounter, const Type& value, Preds... predecessors)
{
    return publish(messageCounter, value, std::vector<av::latency::Pred>{predecessors...});
}

template<class Type>
bool MessageOutput<Type>::publish(uint32_t messageCounter,
                                  const Type& value,
                                  std::vector<av::latency::Pred>&& predecessors)
{
    // Do nothing if the framework pointer is not valid or there is no client object.
    if (!framework_.client()) {
        return false;
    }
    interfaces::baseTypes::InterfaceHeader header;
    interfaces::baseTypes::setHeaderVersion(header, OTTO_INTERFACE_HEADER_VERSION);
    interfaces::baseTypes::setSendTimeStamp(header, task::time::TaskTimeProvider::getSystemOTime().to_otimeSpec());
    interfaces::baseTypes::setHeaderPublisherId(header, publisherID_);
    interfaces::baseTypes::setHeaderMessageCounter(header, messageCounter);

    std::string blob;
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    aptv::serializeODatumMessage(std::string(reinterpret_cast<char*>(&header), sizeof(header)), blob, value);
    return (framework_.client()->sendMsg(channel_, blob, std::move(predecessors)) == SimpleComms::Status::Ok);
}

} // namespace task
