/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Definition of the MessageQueue templated class.
 */
#pragma once

#ifndef EVENT_FRAMEWORK_LIGHT
#include "CallbackSet.h"
#endif

#include <repro/EventFramework.hpp>

#include <cstddef>
#include <mutex>
#include <queue>
#include <string>

namespace task {

/// \brief Templated class for queueing inputs from a channel.
/// \note Before using this class, consider whether a callback-driven implementation would be cleaner than polling.
template<class Type>
class MessageQueue
{
public:
#ifndef EVENT_FRAMEWORK_LIGHT
    /// Constructor.
    /// @param channel: The true name of the channel to receive input from.
    /// @param callbacks: The task's CallbackSet.
    /// @param size: If nonzero, throw out the oldest message once this size is reached.
    MessageQueue(std::string channel, CallbackSet& callbacks, size_t size = 0);
#endif
    /// Constructor.
    /// @param channel: The true name of the channel to receive input from.
    /// @param framework: The event framework.
    /// @param size: If nonzero, throw out the oldest message once this size is reached.
    MessageQueue(std::string channel, av::repro::EventFramework& framework, size_t size = 0);
    // This class must be non-copyable, because only the original is registered with CallbackSet.
    MessageQueue(const MessageQueue&) = delete;
    MessageQueue& operator=(const MessageQueue&) = delete;

    /// Check if any messages are queued.
    /// @return True if no messages are queued.
    bool empty() const;

    /// Get the next message from the queue.
    /// @return A copy of the data from the next message in the queue.
    /// @pre The queue must contain a message - check empty() before calling this.
    Type pop();

    /// Callback adding a message to the queue.
    void callback(const Type& data);

private:
    size_t m_size{0};
    std::queue<Type> m_queue;
    // This object should only ever be used in the callback thread, but just in case:
    mutable std::mutex m_mutex;
};

// Templated functions:

#ifndef EVENT_FRAMEWORK_LIGHT
template<class Type>
MessageQueue<Type>::MessageQueue(std::string channel, CallbackSet& callbacks, size_t size)
    : m_size(size)
{
    callbacks.add(channel, this, &MessageQueue<Type>::callback);
}
#endif

template<class Type>
MessageQueue<Type>::MessageQueue(std::string channel, av::repro::EventFramework& framework, size_t size)
    : m_size(size)
{
    auto fun = [this](const av::repro::Inputs<Type>& inputs) {
        for (const auto& it : inputs) {
            callback(*it);
        }
    };
    framework.setMessageHandler(channel, av::repro::MessageHandler<Type>(fun));
}

template<class Type>
bool MessageQueue<Type>::empty() const
{
    std::unique_lock<std::mutex> lock{m_mutex};
    return m_queue.empty();
}

template<class Type>
Type MessageQueue<Type>::pop()
{
    std::unique_lock<std::mutex> lock{m_mutex};
    Type data = std::move(m_queue.front());
    m_queue.pop();
    return data;
}

template<class Type>
void MessageQueue<Type>::callback(const Type& data)
{
    std::unique_lock<std::mutex> lock{m_mutex};
    // Note: if a message gets dropped here, it's being dropped because of its order
    // in the message stream, so the choice to drop it is fully deterministic.
    if (m_size && m_queue.size() == m_size) {
        m_queue.pop();
    }
    m_queue.push(data);
}

} // namespace task
