/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 */

#pragma once

#include <functional>
#include <interfaces/AvTaskStatus/AvTaskStatus.h>
#include <mutex>
#include <string>
#include <task/MonitoringStatusProviderInterface.h>

/**
 * @class MonitoringStatusProvider
 * @brief This class implements MonitoringStatusProviderInterface and provides some additional methods.
 * This can be used as a default MonitoringStatusProvider if the task owner does not want to create their
 * own derivation of MonitoringStatusProvider.
 *
 * @note please check SimpleAvTaskStatusPublisher for example usage of this class.
 */

class MonitoringStatusProvider : public task::MonitoringStatusProviderInterface
{
public:
    /**
     * @brief The default constructor for the class MonitoringStatusProvider.
     * @param taskName name of the task registering this MonitoringStatusProvider
     */
    explicit MonitoringStatusProvider(const std::string& taskName)
    {
        avTaskStatus_.state = av::msg::TaskState::UNINITIALIZED; // default state
        avTaskStatus_.taskName = taskName;
    }

    /**
     * @brief overriding the getStatus() with the specific implementation for this task
     * @return AvTaskStatus containing task status for the given executive cycle
     */
    av::msg::AvTaskStatus getStatus() override;

    /**
     * @brief custom setStatus() method to set the AvTaskStatus \
     * Use this method to set first status during executive() or to reset the AvTaskStatus if needed
     * @param timestamp current timestamp in the form of time since epoch in microseconds
     * @param state overall state of the task caused by this status
     * @param source enum for defining if the fault is caused internally or due to dependencies
     * @param message string message explaining the details of the degradation
     * @param type enum defining parameter type to convert the values appropriately (if any, default INT)
     * @param parameter_name name of the parameter causing this degradation (if any, default "")
     * @param measured_value measured value of this parameter (if any, default "")
     * @param threshold_value threshold value for this parameter (if any, default "")
     */
    void setStatus(int64_t timestamp,
                   av::msg::TaskState state,
                   av::msg::FaultSource source,
                   const std::string& message,
                   av::msg::ParameterType type = av::msg::ParameterType::INT,
                   const std::string& parameter_name = "",
                   const std::string& measured_value = "",
                   const std::string& threshold_value = "");

    void setStatus(av::msg::AvTaskStatus status);

    /**
     * @brief to check if the current AvTaskStatus is more severe than intendedState
     * @param intendedState the state we are comparing against current TaskState in avTaskStatus_
     * @return true if intendedState is more severe than the current TaskState in avTaskStatus_, false otherwise
     */
    bool isCurrentStateMoreSevereThan(av::msg::TaskState intendedState);

    /**
     * @brief this method combines setStatus(..) and isCurrentStateMoreSevereThan(..) methods. So that one can call
     * just this method lieu of calling multiple methods everytime to update the AvTaskStatus.
     * Note: this method is only going to update the status if it is worse than the current one.
     * If you are recovering from more severe state, please use setStatus() method.
     * @param intendendState intended state to be updated in the AvTaskStatus
     * @param source enum for defining if the fault is caused internally or due to dependencies
     * @param message string message explaining the details of the degradation
     * @param type enum defining parameter type to convert the values appropriately (if any, default INT)
     * @param parameter_name name of the parameter causing this degradation (if any, default "")
     * @param measured_value measured value of this parameter (if any, default "")
     * @param threshold_value threshold value for this parameter (if any, default "")
     * @return true if intended state is more severe than current state and AvTaskStatus is updated, false otherwise.
     */
    bool updateStatus(int64_t timestamp,
                      av::msg::TaskState intendedState,
                      av::msg::FaultSource source,
                      const std::string& message,
                      av::msg::ParameterType type = av::msg::ParameterType::INT,
                      const std::string& parameter_name = "",
                      const std::string& measured_value = "",
                      const std::string& threshold_value = "");

private:
    av::msg::AvTaskStatus avTaskStatus_{}; ///< AvTaskStatus representing state of the task
    std::mutex mutex_{};                   ///< mutex for making updates to AvTaskStatus thread-safe
};
