/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an implementation of MonitoringStatusSenderInterface for sending AV tasks internal status to the
 * monitoring system
 */

#ifndef MONITORING_STATUS_SENDER_H
#define MONITORING_STATUS_SENDER_H

#include <task/MonitoringStatusSenderInterface.h>

#include <interfaces/AvTaskStatus/AvTaskStatus.h>
#include <interfaces/Generic/Output/Abstract.h>
#include <task/GenericMessagePublisher.h>

namespace task {

/**
 * \brief Alias of monitoring status publisher
 */
using AvTaskStatusPublisher = GenericMessagePublisher<GenericOutput<av::msg::AvTaskStatus>, av::msg::AvTaskStatus>;

/**
 * \brief Defines an implementation of MonitoringStatusSenderInterface for sending AV tasks internal status to the
 * monitoring system
 */
class MonitoringStatusSender : public MonitoringStatusSenderInterface
{
public:
    /**
     * \brief Deleted default constructor
     */
    MonitoringStatusSender() = delete;

    /**
     * \brief User-defined constructor
     *
     * \param avTaskStatusPublisher The publisher for sending messages to the collector
     */
    explicit MonitoringStatusSender(std::unique_ptr<AvTaskStatusPublisher> avTaskStatusPublisher);

    /**
     * \brief Default destructor
     */
    virtual ~MonitoringStatusSender() = default;

    /**
     * \brief Initialize the sender
     *
     * \param taskParams The task parameters
     *
     * \return True if the initialization is fine, false otherwise
     */
    bool initialize(ConfigSection& taskParams) override;

    /**
     * \brief Send the given AvTaskStatus status
     *
     * \param status The status to be sent
     *
     * \return True if the sending is fine, false otherwise
     */
    bool sendStatus(const av::msg::AvTaskStatus& status) override;

    /**
     * \brief Get the sending frequency. For example, it is used for testing
     *
     * \param Void
     *
     * \return The sending frequency in Hz
     */
    inline double getSendingFrequency() { return m_frequency_Hz; }

private:
    /**
     * \brief Set the AvTaskStatus sending frequency based on the parameter and the task frequency
     *
     * \param freqParam The frequency parameter value
     * \param taskFreq The task frequency
     *
     * \return Void
     */
    void setSendingFrequency(const double freqParam, const double taskFreq);

    /**
     * \brief Check if a status message needs to be sent, for example, the sending cycle is expired or the current
     * status is different from the last status
     *
     * \param currentStatus The current status
     *
     * \return True if the sending is needed, false, otherwise
     */
    bool needToSend(const av::msg::AvTaskStatus& currentStatus);

    bool m_isInitialized{false};          ///< True if the sender is initialized
    double m_frequency_Hz{1.0};           ///< The sending frequency in Hz
    int64_t m_cycleTime_us{0};            ///< The cyle time in microseconds
    bool m_isFirstTime{true};             ///< True if the first status is sent
    av::msg::AvTaskStatus m_lastStatus{}; ///< The last sent status
    std::unique_ptr<AvTaskStatusPublisher> m_avTaskStatusPublisher{
        nullptr}; ///< The diagmonitor status message publisher
};

} // namespace task

#endif // MONITORING_STATUS_SENDER_H
