/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines a basic building block of any application. All applications start with deriving a task class, and
 * implementing the initialize, executive, and cleanup methods. On initialization, this class will look to load a
 * configuration file with the same name as the taskName with a '.py' or '.cfg' suffix in the config file directory. A
 * task can be configured to loop as fast as possible by providing a cycleRate of zero (0) in the configuration file.
 * Otherwise, a value in seconds is specified as the frequency at which to loop. The default is to loop as fast as
 * possible (no delay. This is useful if the application blocks in the executive loop, and is event driven.
 */

#ifndef TASK_FRAMEWORK_H
#define TASK_FRAMEWORK_H

#include <cstdlib>
#include <string>
#include <unistd.h>

#include "BackupRecoveryHandlerInterface.h"
#include "CallbackSet.h"
#include "ClosingSignalHandlerInterface.h"
#include "HeartbeatSenderInterface.h"
#include "MonitoringStatusProviderInterface.h"
#include "MonitoringStatusSenderInterface.h"

#include <Clock/ClockSource.h>
#include <Comms/Access.h>
#include <Comms/InterfaceDb.h>
#include <Comms/InterfaceUtils.h>
#include <DDSComms/DDSConfiguratorOptions.hpp>
#include <NetworkMetrics/ChannelStatusMonitor.hpp>
#include <NetworkMetrics/NetworkStatsSender.hpp>
#include <SimpleComms/Client.h>
#include <TimeStamp/otime.h>
#include <config/TaskParser.h>
#include <interfaces/AvTaskStatus/AvTaskStatus.h>
#include <interfaces/PublishListMessage/PublishListMessage.h>
#include <interfaces/SaferCommand/SaferCommand.h>
#include <interfaces/SaferResponse/SaferResponse.h>
#include <log/Logger.h>
#include <profile/oprofile.h>
#include <repro/EventFramework.hpp>

#include "lib/avlog/avlog.hpp"

namespace task {

/*!
 * @class Task
 *
 * @brief The basic building block of any application.
 *
 * All applications start with deriving a task class, and
 * implementing the initialize, executive, and cleanup
 * methods.
 *
 * On initialization, this class will look to load a
 * configuration file with the same name as the taskName
 * with a '.py' or '.cfg' suffix in the config file directory.
 *
 * A task can be configured to loop as fast as possible
 * by providing a cycleRate of zero (0) in the
 * configuration file. Otherwise, a value in seconds is
 * specified as the frequency at which to loop. The default
 * is to loop as fast as possible (no delay).  This is
 * useful if the application blocks in the executive loop,
 * and is event driven.
 */
class Task
{
public:
    /**
     * Alias types for AV task framework
     */
    using ConfigSectionList = std::list<ConfigSection>;
    using ConfigSectionListIter = ConfigSectionList::iterator;
    using HeartbeatSenderPtr = std::unique_ptr<HeartbeatSenderInterface>;
    using BackupRecoveryHandlerPtr = std::unique_ptr<BackupRecoveryHandlerInterface>;
    using StatusProviderPtr = std::shared_ptr<MonitoringStatusProviderInterface>;
    using StatusSenderPtr = std::unique_ptr<MonitoringStatusSenderInterface>;
    using IDObjPtrMap = std::map<std::string, IDbObject*>;
    using IDObjSmartPtrMap = std::map<std::string, std::shared_ptr<IDbObject>>;

    /**
     * \brief Default constructor, copy and assignment constructors are deleted. A task thread is unique for each task
     * whenever it is created. Thus, a task thread object cannot be copied or moved
     */
    Task() = delete;
    Task(const Task&) = delete;
    Task& operator=(const Task&) = delete;
    Task(const Task&&) = delete;
    Task& operator=(const Task&&) = delete;

    /**
     * \brief This function is implemented by a derived class to instantiate a singleton object of Task class
     *
     * \param Void
     *
     * \return A pointer to the singleton object
     */
    static Task* get();

    /// \brief Run the task's main loop, and return when it is interrupted or the executive quits.
    void run(ClosingSignalHandlerInterface& closingSignal);

    /*!
     * @brief The initialize method called by main.
     *
     * This method invokes the derived class' initialize()
     * method if a task config file is found and parsed.
     *
     * @return True if initialization should continue,
     * false to exit.
     *
     */
    virtual bool _initialize();

    /*!
     * @brief The executive loop called by main.
     *
     * This method invokes the derived class' executive()
     * method at the defined cycle rate.
     *
     * @return True if the application should continue
     * executing, False to exit.
     *
     */
    virtual bool _executive();

    /*!
     * @brief The cleanup method called by main.
     *
     * This method invokes the derived class' cleanup()
     * method, then it proceeds to cleanup the base class'
     * work
     *
     */
    virtual void _cleanup();

    /*!
     * @brief initialize() called by _initialize().
     *
     * This function is implemented by each application
     * inherited from the Task class.
     *
     * @return True for successful initialization.
     *
     */
    virtual bool initialize() = 0;

    /*!
     * @brief executive() called by _executive().
     *
     * This function is implemented by each application
     * inherited from the Task class.
     *
     * @return True if the application should continue
     * executing, False to exit.
     *
     */
    virtual bool executive() = 0;

    /*!
     * @brief cleanup() called by _cleanup()
     *
     * This function is implemented by each application
     * inherited from the Task class.
     *
     */
    virtual void cleanup() = 0;

    /*!
     * @brief Gets the name of the task in string.
     *
     * @return The name of this task.
     *
     */
    inline const std::string& getTaskName() const { return taskName_; }

    /*!
     * @brief Sets the name of the task in string.
     *
     * @param taskName The name of the task to set.
     *
     * @return Void.
     *
     */
    inline void setTaskName(std::string taskName)
    {
        if (!taskName.empty()) {
            taskName_ = std::move(taskName);
        }
    }

    /*!
     * @brief Allows command-line override of configured
     * global log level
     *
     * @param level The level to set
     *
     * @return True if the setting was fine, false, otherwise
     *
     */
    bool setGlobalLogLevel(const std::string& level);

    /*!
     * @brief Gets the path to $OTTO_CONFIG_DIR.
     *
     * @return The location of $OTTO_CONFIG_DIR.
     *
     */
    std::string getConfigFileDirectory();

    /*!
     * @brief Gets the path to $OTTO_DIR.
     *
     * @return The location of $OTTO_DIR.
     *
     */
    std::string getUCDirectory();

    /*!
     * @brief  Gets the name of the current machine.
     *
     * @return The name of the current machine.
     *
     */
    std::string getHostName() const;

    /*!
     * @brief Gets the reference of the Logger object.
     *
     * @return The logger object.
     *
     */
    inline logger::Logger& getLogger() const { return logger_; }

    /*!
     * @brief Gets the current time for this Task.
     *
     * If the task is specified to require a steady clock,
     * this returns a steady system-synchronized time
     * (SystemClock::getSynchronizedSteadyTime()).
     *
     * Otherwise, it returns SystemClock::getSystemTime().
     * In a real vehicle this will be a steady clock, but
     * in a simulation environment this may be a clock
     * that can pause or run faster or slower than real time.
     * In resim/playback this will be locked to the time
     * of the data playback.
     *
     * @return The current time in otimePoint.
     *
     */
    ottotime::otimePoint getCurrentTaskTime();

    /*!
     * @brief Returns the time from the Task
     *        returned by Task::get().  If that
     *        method returns nullptr, falls back
     *        to SystemClock::getSystemTime().
     *
     *        equivalent to:
     *        (Task::get() ?
     *         Task::get()->getSystemTime() :
     *         SystemClock::getSystemTime())
     * @note This method is actualized by registering the task instance's
     * getSystemTime_ function with TaskTimeProvider::getSystemOtime, then
     * calling TaskTimeProvider::getSystemOtime here.
     * @return The current time in otimePoint.
     *
     */
    static ottotime::otimePoint getSystemTime();

    /*!
     * @brief Static method to sanely wrap the strerror()
     * call which retrieves the text corresponding to an
     * error.
     *
     * @param errnum The system error of interest.
     *
     * @return std::string containing some reasonable
     * interpretation of given errnum.
     *
     * @note This method is kept to avoid API breaking
     */
    static std::string getSystemErrorCodeText(const int errnum);

    /*!
     * @brief Gets the task configuration.
     *
     * @return The object of ConfigSection.
     *
     */
    inline ConfigSection& getTaskConfig() { return taskParams_; }

    // Not valid before initExecutive() is called.
    [[nodiscard]] double getFrequencyHz() const { return frequencyHz_; }

    /*!
     * @brief Sends a message to the SAFER daemon to
     * report it to OttoView and notes.txt in the
     * logging directory. Tasks can call this to send
     * the message to be reported and recorded.
     *
     * @param message The message to be sent.
     *
     * @return True if the message was sent, false, otherwise
     *
     */
    bool sendStatusReport(const std::string& message);

    /**
     * \brief Create a heartbeat sender. This method creates a standard heartbeat sender for all AV tasks. It can be
     * overrode by a sub-class to have a specific sender. Thus, this follows the design pattern Factory Method
     *
     * \param Void
     *
     * \return A pointer to the created sender object
     */
    virtual HeartbeatSenderPtr createHeartbeatSender();

    /**
     * \brief Create a backup recovery handler. This method creates a standard handler for all AV tasks. It can be
     * overrode by a sub-class to have a specific handler. Thus, this follows the design pattern Factory Method
     *
     * \param Void
     *
     * \return A pointer to the created handler object
     */
    virtual BackupRecoveryHandlerPtr createBackupRecoveryHandler();

    /**
     * \brief Register a monitoring status provider. In general, it should be called by an AV task implementaton during
     * its initializaton in order to provide a status provider
     *
     * \param monitoringStatusProvider The provider to be registered
     *
     * \return Void
     */
    inline void registerMonitoringStatusProvider(StatusProviderPtr monitoringStatusProvider)
    {
        m_monitoringStatusProvider = monitoringStatusProvider;
    }

    /**
     * \brief Create a monitoring status sender. This method creates a standard sender for all AV tasks. It can be
     * overrode by a sub-class to have a specific sender. Thus, this follows the design pattern Factory Method
     *
     * \param Void
     *
     * \return A pointer to the created sender object
     */
    virtual StatusSenderPtr createMonitoringStatusSender();

protected:
    /*!
     * @brief The Task class constructor.
     *
     * @param taskName The name of the task.
     *
     */
    Task(std::string taskName);

    /*!
     * @brief The Task class destructor.
     *
     */
    virtual ~Task() = default;

    /*!
     * @brief Changes the mode of all interfaces given
     * a list of config sections
     *
     * @param ifclist The interface list.
     *
     * @param acc The access mode (R/W/RW/Standby).
     *
     * @return True if the interface modes were
     * successfully changed.
     *
     */
    bool chmodInterfaces(ConfigSectionList ifclist, const Access& acc);

    /*!
     * @brief Get interface, given the interface name, config and access
     *
     * InterfaceType expects a type that inherits from the 'Interface' class, or any type that
     * provides 'InterfaceType(ConfigSection&, const Access&)' and 'bool initialize()'.
     * Generally in the form of ChannelType<MessageType>, with the most common uses being
     * GenericInputChannel<MyMessageType> and GenericOutputChannel<MyMessageType>.
     * It also accepts specialized non-templated interfaces (e.g. VehicleStateInterpolated).
     *
     * @param[in] interfaceName The name of the interface.
     * @param[in] config The interface config section.
     * @param[in] access The access mode (R/W/RW).
     * @param[out] interfaceMap The interface map storage
     *
     * @return Shared pointer to interface on successful retrieval, nullptr otherwise.
     *
     */
    template<typename InterfaceType>
    std::shared_ptr<InterfaceType> getInterfaceFromMap(const std::string& interfaceName,
                                                       ConfigSection& config,
                                                       Access access,
                                                       IDObjSmartPtrMap& interfaceMap)
    {
        std::shared_ptr<InterfaceType> interfaceInstance{nullptr};
        auto instanceName{config.getName()};
        // Check if the interface instance is already created
        IDObjSmartPtrMap::const_iterator instanceIt = interfaceMap.find(instanceName);
        if (instanceIt != interfaceMap.end()) {
            AVLOG_INFO("{} was already created, getting the stored one", AVLOG_ARGS("interfaceName"), interfaceName);
            interfaceInstance = std::dynamic_pointer_cast<InterfaceType>(instanceIt->second);
        }
        if (interfaceInstance == nullptr) {
            interfaceInstance = comms::getInterface<InterfaceType>(interfaceName, config, access);
            if (interfaceInstance != nullptr) {
                // Store the interface (allows chmod for failover/recovery)
                interfaceMap[instanceName] = interfaceInstance;
            } else {
                AVLOG_ERROR("Failed to create and initialize interface {}", AVLOG_ARGS("interfaceName"), interfaceName);
            }
        }
        return interfaceInstance;
    }

    /*!
     * @brief Get interface, given the interface name, config and access
     *
     * InterfaceType expects a type that inherits from the 'Interface' class, or any type that
     * provides 'InterfaceType(ConfigSection&, const Access&)' and 'bool initialize()'.
     * Generally in the form of ChannelType<MessageType>, with the most common uses being
     * GenericInputChannel<MyMessageType> and GenericOutputChannel<MyMessageType>.
     * It also accepts specialized non-templated interfaces (e.g. VehicleStateInterpolated).
     *
     * @param[in] interfaceName The name of the interface.
     * @param[in] config The interface config section.
     * @param[in] access The access mode (R/W/RW).
     *
     * @return Shared pointer to interface on successful retrieval, nullptr otherwise.
     *
     */
    template<typename InterfaceType>
    std::shared_ptr<InterfaceType> getInterface(const std::string& interfaceName, ConfigSection& config, Access access)
    {
        auto& interfaceMap = isCommonInterface(interfaceName) ? taskCommonInterfaces_ : interfacesManaged_;
        if (access == Access::Write || access == Access::ReadWrite) {
            std::string::size_type pos = interfaceName.find("Output");
            if (pos != std::string::npos) {
                std::lock_guard<std::mutex> guard(publishedChannelMutex_);
                publishedChannels_.channels.emplace_back(interfaceName.substr(0, pos));
            }
        }
        return (getInterfaceFromMap<InterfaceType>(interfaceName, config, access, interfaceMap));
    }

    /*!
     * @brief Get read-only interface, given the interface name
     *
     * InterfaceType expects a type that inherits from the 'Interface' class, or any type that
     * provides 'InterfaceType(ConfigSection&, const Access&)' and 'bool initialize()'.
     * Generally in the form of ChannelType<MessageType>, with the most common uses being
     * GenericInputChannel<MyMessageType> and GenericOutputChannel<MyMessageType>.
     * It also accepts specialized non-templated interfaces (e.g. VehicleStateInterpolated).
     *
     * @param[in] interfaceName The name of the interface.
     *
     * @return Shared pointer to interface on successful retrieval, nullptr otherwise.
     *
     */
    template<typename InterfaceType>
    std::shared_ptr<InterfaceType> getReadOnlyInterface(const std::string& interfaceName)
    {
        std::shared_ptr<InterfaceType> interfaceInstance{nullptr};

        if (tp_.hasReadOnlyInterface(interfaceName)) {
            auto interfaceConfig = tp_.getReadOnlyInterface(interfaceName);
            interfaceInstance = getInterface<InterfaceType>(interfaceName, interfaceConfig, Access::Read);
        } else {
            AVLOG_ERROR("No read-only interface {} defined", AVLOG_ARGS("interfaceName"), interfaceName);
        }

        return interfaceInstance;
    }

    /*!
     * @brief Get write-only interface, given the interface name
     *
     * InterfaceType expects a type that inherits from the 'Interface' class, or any type that
     * provides 'InterfaceType(ConfigSection&, const Access&)' and 'bool initialize()'.
     * Generally in the form of ChannelType<MessageType>, with the most common uses being
     * GenericInputChannel<MyMessageType> and GenericOutputChannel<MyMessageType>.
     * It also accepts specialized non-templated interfaces (e.g. VehicleStateInterpolated).
     *
     * @param[in] interfaceName The name of the interface.
     *
     * @return Shared pointer to interface on successful retrieval, nullptr otherwise.
     *
     */
    template<typename InterfaceType>
    std::shared_ptr<InterfaceType> getWriteOnlyInterface(const std::string& interfaceName)
    {
        std::shared_ptr<InterfaceType> interfaceInstance{nullptr};

        if (tp_.hasWriteOnlyInterface(interfaceName)) {
            auto interfaceConfig = tp_.getWriteOnlyInterface(interfaceName);
            // Note that backup task, numReplicas == 1, only can read, cannot write
            auto accessMode = (handlingParams_.numReplicas == 1) ? Access::Standby : Access::Write;
            interfaceInstance = getInterface<InterfaceType>(interfaceName, interfaceConfig, accessMode);
        } else {
            AVLOG_ERROR("Task {} has no write-only interface {} defined",
                        AVLOG_ARGS("taskName", "interfaceName"),
                        taskName_,
                        interfaceName);
        }

        return interfaceInstance;
    }

    /*!
     * @brief Get read-write interface, given the interface name
     *
     * InterfaceType expects a type that inherits from the 'Interface' class, or any type that
     * provides 'InterfaceType(ConfigSection&, const Access&)' and 'bool initialize()'.
     * Generally in the form of ChannelType<MessageType>, with the most common uses being
     * GenericInputChannel<MyMessageType> and GenericOutputChannel<MyMessageType>.
     * It also accepts specialized non-templated interfaces (e.g. VehicleStateInterpolated).
     *
     * @param[in] interfaceName The name of the interface.
     *
     * @return Shared pointer to interface on successful retrieval, nullptr otherwise.
     *
     */
    template<typename InterfaceType>
    std::shared_ptr<InterfaceType> getReadWriteInterface(const std::string& interfaceName)
    {
        std::shared_ptr<InterfaceType> interfaceInstance{nullptr};

        if (tp_.hasReadWriteInterface(interfaceName)) {
            auto interfaceConfig = tp_.getReadWriteInterface(interfaceName);
            // Note that backup task, numReplicas == 1, only can read, cannot write
            auto accessMode = (handlingParams_.numReplicas == 1) ? Access::Read : Access::ReadWrite;
            interfaceInstance = getInterface<InterfaceType>(interfaceName, interfaceConfig, accessMode);
        } else {
            AVLOG_ERROR("No read-write interface {} defined", AVLOG_ARGS("interfaceName"), interfaceName);
        }

        return interfaceInstance;
    }

    TaskParser tp_{};                 ///< An object of the TaskParser class.
    ConfigSection taskParams_{};      ///< An object containing all parameters.
    mutable logger::Logger logger_{}; ///< An object of the Logger class.
    std::string taskName_{""};        ///< The task name in string.
    bool autoInitInterfaces_{true};   ///< Register tasks automatically in InterfacesDB.

    av::repro::EventFramework framework_;
    CallbackSet callbacks_;
    av::repro::EventFramework::Mode reproMode_{av::repro::EventFramework::Mode::live};

    /**
     * \brief Template method that fills in the task's DDS Options for EventFramework Comms
     * \note Comms::DDSChannels are entirely driven by the relevant libconfig ConfigSections
     * \param The DDSOptions for your task to edit
     */
    virtual void populateDdsOptions([[maybe_unused]] DDSComms::DDSOptionsType& ddsOptions){};

    /**
     * \brief Initialize the parameters from the task configuration file
     *
     * \param configDirPath The path to the directory of the configuration file
     *
     * \return True if the initialization was fine, otherwise, false
     */
    bool initParameters(const std::string& configDirPath);

    /// Initialize the executive scheduler.
    void initExecutive();

    /// Collect output channel names of this task
    av::msg::PublishListMessage getPublishListMessage();

private:
    /*!
     * @brief the function which returns task time.
     * See docs for the static method getSystemTime.
     * @return The current time in otimePoint.
     */
    static ottotime::otimePoint getSystemTime_();

    /**
     * \brief Initializes the heartbeat sender
     *
     * \param Void
     *
     * \return True if the initialization was fine, false, otherwise
     */
    bool initHeartbeatSender();

    /**
     * \brief Parse the log level parameter from the configuration file
     *
     * \param Void
     *
     * \return True if the parsing was fine, false otherwise
     */
    bool parseLogLevelParams();

    /**
     * \brief Parse the backup recovery parameter from the configuration file
     *
     * \param Void
     *
     * \return True if the parsing was fine, false otherwise
     */
    bool parseRecoveryParams();

    /*!
     * @brief Helper function for _executive() to handle Safer replica commands.
     */
    void handleExecutiveSaferReplicaCommand();

    /**
     * \brief Initializes the backup recovery handler
     *
     * \param Void
     *
     * \return True if the initialization was fine, false, otherwise
     */
    bool initBackupRecoveryHandler();

    /**
     * \brief Check that the given interface name is one of the common interfaces shared by all tasks
     *
     * \param interfaceName The interface name
     *
     * \return True if it is common interface, false otherwise
     */
    bool isCommonInterface(const std::string& interfaceName);

    /**
     * \brief Initializes the monitoring status sender
     *
     * \param Void
     *
     * \return True if the initialization was fine, false, otherwise
     */
    bool initMonitoringStatusSender();

    /**
     * \brief Initializes the default monitoring status messge
     *
     * \param Void
     *
     * \return Void
     */
    void initDefaultStatusMessage();

    /// Send a monitoring status message, if necessary.
    void sendMonitoringStatus();

    /// \brief Callback function for the executive function.
    void executiveCallback();

    /**
     * \brief Method that inits task's DDS Options for EventFramework Comms with base options
     * \note Comms::DDSChannels are entirely driven by the relevant libconfig ConfigSections
     * \return The DDSOptions for your task
     */
    DDSComms::DDSOptionsType initDdsOptions();

    /*!
     * Managed and unmanaged interfaces are created using different methods.
     *
     * interfaces_ are created within InterfaceDb by dynamically loading the interface shared libary
     * interfacesManaged_ are created within the task::getInterface() function directly
     *
     * Given they use different memory management techniques (raw pointers vs shared pointers)
     * they are stored separately until a time the raw pointer approach has been phased out.
     */
    IDObjPtrMap interfaces_{};                ///< The list of interfaces
    IDObjSmartPtrMap interfacesManaged_{};    ///< The list of interfaces managed by shared ptr
    IDObjSmartPtrMap taskCommonInterfaces_{}; ///< The list of common interfaces for all tasks

    av::msg::PublishListMessage publishedChannels_{}; /// The list of all output channel name of this task
    std::mutex publishedChannelMutex_;
    std::thread m_publishChannelsThread;
    std::atomic<bool> executionRet{true};
    std::atomic<bool> interrupt{false};

    bool profileEnabled_{false};                              ///< True if profiling is on.
    std::string globalLogLevel_{""};                          ///< The global log threshold.
    HeartbeatSenderPtr heartbeatSender_{nullptr};             ///< A unique pointer to an object of heartbeat sender
    BackupRecoveryParameter handlingParams_{};                ///< The backup recovery results
    BackupRecoveryHandlerPtr backupRecoveryHandler_{nullptr}; ///< A unique pointer to an object of recovery handler
    ClosingSignalHandlerInterface* closingSignal_{nullptr};   ///< The interface for checking closing signals.

    StatusProviderPtr m_monitoringStatusProvider{nullptr}; ///< The monitoring status provider
    StatusSenderPtr m_monitoringStatusSender{nullptr};     ///< The monitoring status sender
    av::msg::AvTaskStatus m_defaultAvTaskStatus{};         ///< The default monitoring status message

    /// Count of number of cycles which were performed. Used for tracking
    /// percentage of cycles which were skipped. Also, in
    /// _executive(), if numExecutions_ is 0 (i.e., it's the first
    /// cycle), we report that the task has started.
    int numExecutions_{0};

    std::unique_ptr<network::NetworkStatsSender> networkStatsSender_{};
    std::unique_ptr<network::ChannelStatusMonitor> channelStatusMonitor_{};

    double frequencyHz_{0.};
};

} // namespace task

#endif // TASK_FRAMEWORK_H
