/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional, its affiliates, and its suppliers (if
 * any). The intellectual and technical concepts contained herein are proprietary to Motional, its affiliates, and its
 * suppliers (if any) and may be covered by U.S. and Foreign Patents, patents in process, and are protected by trade
 * secrets or copyright law. Dissemination of this information or reproduction of this material is strictly forbidden
 * unless prior written permission is obtained from Motional.
 *
 * \file
 * \brief Definition of the TaskTimeProvider class.
 */
#pragma once

#include <TimeStamp/otime.h>
#include <array>
#include <atomic>
#include <avtime/Time.hpp>
#include <functional>
#include <mutex>

namespace task::time {

namespace detail {

/// @brief Functor that returns current AvClock time.
class AvTimeProvider
{
public:
    ottotime::otimePoint operator()();
};

/// @brief Functor that returns current SystemClock time. Used as a default.
class SystemClockProvider
{
public:
    ottotime::otimePoint operator()();
};

} // namespace detail

/// @brief Class to set and access a task time source via function object
/// An instance-hiding singleton that provides system time via a settable
/// time source function. Exists primarily to allow a Task to dictate time
/// source while in the context of a proper Task framework task, but allow
/// another time source when the EventFramework is used independently.
/// The base Task will always set the time itself. Otherwise, the default
/// provider is the SystemClock.
class TaskTimeProvider
{
public:
    using OTimeProvider = std::function<ottotime::otimePoint()>;
    using DefaultTimeProvider = detail::SystemClockProvider;

    /// Set the function which shall provide task time
    /// The function is using otime because the current task framework
    /// still primarily uses otime constructs.
    /// When in the context of task framework task, this should return the
    /// task's concept of time, as the base task will set the provider.
    /// If the provider has not been set, the default is the SystemClock,
    /// @note Not thread safe. Should be done early during initialization.
    /// @param[in] otimeProvider: A function which produces a time when called.
    static void setTimeSource(const OTimeProvider& otimeProvider);

    /// Get the current task time, sourced from previously set provider function
    /// @return the current task time as an otimePoint
    [[nodiscard]] static ottotime::otimePoint getSystemOTime();

    /// Get the current task time, sourced from previously set provider function
    /// @return the current task time as an av::Duration
    [[nodiscard]] static av::Duration getSystemTime();

private:
    TaskTimeProvider() = default;
    ~TaskTimeProvider() = default;
    TaskTimeProvider(const TaskTimeProvider&) = delete;
    TaskTimeProvider& operator=(const TaskTimeProvider&) = delete;

    /// Get the singleton TaskTimeProvider instance
    static TaskTimeProvider& get();
    /// set the clock source
    void setTimeSource_(OTimeProvider otimeProvider);
    /// get system time
    ottotime::otimePoint getSystemOTime_() const;
    // active time provider function
    OTimeProvider oTimeProvider_{DefaultTimeProvider{}};
};

} // namespace task::time
