/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an implementation for printing vehicle name, si-middleware version, and build number at av-stack
 */

#ifndef VERSION_PRINTER_H
#define VERSION_PRINTER_H

#include <task/VersionPrinterInterface.h>

namespace task {

/**
 * \brief The text in the working directory that contains av-stack build number from the logs
 */
const std::string driveInfoSummaryFile{"DriveInfoSummary000.txt"};

/**
 * \brief This class defines an implementation for printing vehicle name, si-middleware version, and build number at
 * av-stack
 */
class VersionPrinter : public VersionPrinterInterface
{
public:
    /**
     * \brief The default constructor
     */
    VersionPrinter() = default;

    /**
     * \brief The default destructor
     */
    virtual ~VersionPrinter() = default;

    /**
     * \brief Get the vehicle name in the file config from the active config directory. If the logMode is true then
     * print the vehicle name in the log file
     *
     * \param configFile The name of the file containing the vehicle name
     * \param logMode The log mode
     *
     * \return The vehicle name as a string
     */
    std::string getVehicleName(const std::string& configFile = defaultVehicleConfigFile,
                               const bool logMode = false) override;

    /**
     * \brief Get the si-middleware, av-stack, and aptiview (if the mode is OttoView_playback which seems to be used by
     * simulation code-base) build numbers. If the logMode is true then print the build number in the log file
     *
     * \param mode The playback mode for aptiview build version
     * \param buildTagFile The name of the av-stack build tag file
     * \param logMode The log mode
     *
     * \return The si-middleware, av-stack, aptiview build number as a value of AVStackBuildNumber
     */
    AVStackBuildNumber getAVStackBuildNumber(const std::string& mode = std::string{""},
                                             const std::string& buildTagFile = defaultAVBuildTags,
                                             const bool logMode = false) override;

private:
    /**
     * \brief Get the av-stack build number from the build tags
     *
     * \param buildTagFile The name of the av-stack build tag file
     *
     * \return The build number as a string
     */
    std::string getAVStackBuildNumberFromTag(const std::string& buildTagFile);

    /**
     * \brief Get the av-stack build number from the logs
     *
     * \param Void
     *
     * \return The build number as a string
     */
    std::string getAVStackBuildNumberFromLogs();
};

} // namespace task

#endif // VERSION_PRINTER_H
