/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an interface for printing vehicle name, si-middleware version, and build number at av-stack
 */

#ifndef VERSION_PRINTER_INTERFACE_H
#define VERSION_PRINTER_INTERFACE_H

#include <iostream>
#include <string>

namespace task {

/**
 * \brief AV stack build numbers
 */
struct AVStackBuildNumber
{
    std::string avStackNumber{"N/A"};        ///< The build number of av-stack from the build_tags
    std::string avStackNumberFromLog{"N/A"}; ///< The build number of av-stack from the log data
    bool playbackMode{false};                ///< True if the av-stack build number got from log data

    /**
     * \brief Print the build numbers on standard output
     *
     * \param Void
     *
     * \return Void
     */
    void printBuildNumbers() const
    {
        std::cout << "AV-Stack build number: " << avStackNumber << std::endl;
        if (playbackMode) {
            std::cout << "AV-Stack build number from logs: " << avStackNumberFromLog << std::endl;
        }
    }
};

/**
 * \brief The default vehicle config file name
 */
const std::string defaultVehicleConfigFile{"config"};

/**
 * \brief The default av-stack build tags file name
 */
const std::string defaultAVBuildTags{"build_tags"};

/**
 * \brief This class defines an interface for printing vehicle name, si-middleware version, and build number at av-stack
 */
class VersionPrinterInterface
{
public:
    /**
     * \brief The default constructor
     */
    VersionPrinterInterface() = default;

    /**
     * \brief The default destructor
     */
    virtual ~VersionPrinterInterface() = default;

    /**
     * \brief Get the vehicle name in the file config from the active config directory. If the logMode is true then
     * print the vehicle name in the log file
     *
     * \param configFile The name of the file containing the vehicle name
     * \param logMode The log mode
     *
     * \return The vehicle name as a string
     */
    virtual std::string getVehicleName(const std::string& configFile = defaultVehicleConfigFile,
                                       const bool logMode = false) = 0;

    /**
     * \brief Get the si-middleware, av-stack, and aptiview (if the mode is OttoView_playback which seems to be used by
     * simulation code-base) build numbers. If the logMode is true then print the build number in the log file
     *
     * \param mode The playback mode for aptiview build version
     * \param buildTagFile The name of the av-stack build tag file
     * \param logMode The log mode
     *
     * \return The si-middleware, av-stack, aptiview build number as a value of AVStackBuildNumber
     */
    virtual AVStackBuildNumber getAVStackBuildNumber(const std::string& mode = std::string{""},
                                                     const std::string& buildTagFile = defaultAVBuildTags,
                                                     const bool logMode = false) = 0;
};

} // namespace task

#endif // VERSION_PRINTER_INTERFACE_H
