/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines an entry point for AV task application
 */

#ifndef MAIN_FUNCTION_DEF_H
#define MAIN_FUNCTION_DEF_H

#include <cstdio>
#include <memory>
#include <string>

#include "lib/avlog/avlog.hpp"
#include <task/ClosingSignalHandler.h>
#include <task/CmdLineOptProcessor.h>
#include <task/Task.h>
#include <task/VersionPrinter.h>

#include "TaskApplication.h"

using namespace task;

/**
 * \brief Create a TaskApplication object and run it
 *
 * \param argc  Number of command line arguments
 * \param argv  The command line arguments
 *
 * \return Zero on success, non-zero on error
 */
template<typename ClosingSignalHandlerType,
         typename CmdLineOptProcessorType,
         typename VersionPrinterType,
         typename TaskType,
         typename TaskApplicationType>
int executeTask(int argc, char** argv)
{
    int retVal{EXIT_SUCCESS};

    // Set the signal mask before starting avlog.
    auto closingSignalHandler = std::make_unique<ClosingSignalHandlerType>();
    static const std::vector<int> closingSignals = {SIGINT, SIGTERM, SIGQUIT};
    bool sigmaskSuccess = closingSignalHandler->setSignals(closingSignals);

    // Start the global logging service
    // ('log' now forwards to this by default)
    avlog::AvLogOptions logOptions;
    logOptions.logToConsole = true;
    // Get the instance name. Ideally we would do this using the CmdLineOptProcessor to
    // avoid duplicated work, but CmdLineOptProcessor uses avlog.
    // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-pointer-arithmetic)
    for (char** it = argv + 1; *it != nullptr; ++it) {
        std::string arg = *it;
        static const std::string expected = "--instanceName=";
        if (arg == "--instanceName" || arg == "-i") {
            // Read the next argument and bail out if it's null.
            // NOLINTNEXTLINE(cppcoreguidelines-pro-bounds-pointer-arithmetic)
            ++it;
            if (*it == nullptr) {
                break;
            }
            logOptions.processName = *it;
        } else if (arg.size() > expected.size() && arg.find(expected) == 0) {
            logOptions.processName = arg.substr(expected.size());
        }
    }
    avlog::startup(logOptions);

    // This message can't be logged until after avlog has started.
    // But, because the signal numbers are hard-coded valid values, it will never happen.
    if (!sigmaskSuccess) {
        AVLOG_ERROR("Failed to set the closing signal mask.");
        return 1;
    }

    AVLOG_INFO("Main::executeTask(): Starting Task");

    // Create a standar task application, customized one can be used here
    auto taskPtr = TaskType::get();
    if (taskPtr == nullptr) {
        AVLOG_ERROR("Main::executeTask(): Task::get() is nullptr");
        retVal = EXIT_FAILURE;
    } else {
        auto versionPrinter = std::make_shared<VersionPrinterType>();
        auto taskApplication =
            std::make_unique<TaskApplicationType>(std::move(closingSignalHandler),
                                                  std::make_unique<CmdLineOptProcessorType>(versionPrinter),
                                                  versionPrinter,
                                                  taskPtr);
        if (taskApplication != nullptr) {
            if (!taskApplication->initTask(argc, argv)) {
                AVLOG_ERROR("Main::executeTask(): Failed to initialize");
                retVal = EXIT_FAILURE;
            } else if (!taskApplication->runTask()) {
                AVLOG_INFO("Main::executeTask(): Terminated by receiving closing signals");
                retVal = EXIT_FAILURE;
            }
            taskApplication.reset();
        }
    }

    avlog::shutdown(); // shut down the global logger and make sure everything is flushed to disk
    return retVal;
}

#endif // MAIN_FUNCTION_DEF_H
