/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Definition of the Deserializer class.
#pragma once

#include <repro/DeserializerBase.hpp>

#include <repro/Event.hpp>
#include <repro/Message.hpp>
#include <repro/blobUtils.hpp>
#include <repro/handlerTypes.hpp>

#include "lib/avlog/avlog.hpp"
#include <ODatum/ODatum.h>
#include <interfaces/baseTypes/Header.h>

#include <cstddef>
#include <deque>
#include <functional>
#include <memory>
#include <mutex>
#include <string>
#include <string_view>
#include <vector>

namespace av::repro {

/// Class handling message buffers and deserialization for a single input topic.
template<typename Type>
class Deserializer : public DeserializerBase
{
public:
    /// Function to be called when an event runs.
    using Handler = MessageHandler<Type>;

public:
    /// Constructor, optionally specifying a fixed number of message buffers.
    Deserializer(Enqueuer enqueuer, std::string name, Handler handler, size_t buffers = 0);
    ~Deserializer() noexcept override = default;

    /// Receive a message.
    /// @note This function must only be called on one thread at a time.
    void receive(const std::string& blob, const std::string_view& origin) override;

private:
    // Run the handler, batching messages from the same origin.
    // @return the ID of the last message in the batch.
    uint32_t run(const std::string_view& origin, uint32_t firstID, uint32_t lastID);
    // Get a buffer to deserialize into.
    // Returns null if no buffers are available and no stale pending buffer can be recycled.
    std::shared_ptr<Type> getBuffer();

private:
    // This function should be called once a message has been converted to an event.
    Enqueuer enqueuer_;
    // Name of the topic.
    std::string name_;
    // Function to be called to process messages of this type.
    Handler handler_;

    // Message buffers, to be recycled throughout the life of the task.
    std::vector<std::shared_ptr<Type>> buffers_;
    // Store whether we're allowed to allocate additional buffers.
    bool unlimitedBuffers_{false};

    // Everything below here is in the scope of this mutex.
    std::mutex pendingMutex_;
    // Keep track of which events are currently queued up waiting to be processed.
    std::deque<Message<Type>> pending_;
};

// Templated function definitons:
template<typename Type>
Deserializer<Type>::Deserializer(Enqueuer enqueuer, std::string name, Handler handler, size_t buffers)
    : enqueuer_(std::move(enqueuer))
    , name_(std::move(name))
    , handler_(std::move(handler))
    , buffers_(buffers == 0 ? 2 : buffers)
    , unlimitedBuffers_(buffers == 0)
{
    // If a fixed number of buffers was requested, allocate them all up front.
    // Otherwise, allocate two, which is enough for most use cases.
    for (std::shared_ptr<Type>& buffer : buffers_) {
        buffer = std::make_shared<Type>();
    }
}

template<typename Type>
void Deserializer<Type>::receive(const std::string& blob, const std::string_view& origin)
{
    // Do the best we can to check that this blob really is a serialized ODatum.
    // The header version is the only "magic number" we can check.
    if (blob.size() < sizeof(interfaces::baseTypes::InterfaceHeader)) {
        AVLOG_ERROR("Deserializer {}: given argument is not a serialized ODatum.", name_);
        return;
    }
    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    const auto& header = *reinterpret_cast<const interfaces::baseTypes::InterfaceHeader*>(blob.data());
    if (ntohl(header.headerVersion) != OTTO_INTERFACE_HEADER_VERSION) {
        AVLOG_ERROR("Deserializer {}: blob does not have the expected interface version.", name_);
        return;
    }
    // Get a buffer to deserialize into.
    std::shared_ptr<Type> buffer = getBuffer();
    if (!buffer) {
        AVLOG_ERROR("Deserializer {}: no available buffers.", name_);
        return;
    }
    // TODO: C++17 lets us use string_view rather than copying this substring.
    // But, the ODatum interface would need to be updated to support it.
    std::string body = blob.substr(sizeof(interfaces::baseTypes::InterfaceHeader));
    // TODO: deserializeODatum takes a non-const string as input, even though
    // the string is not modified. Switch to const reference?
    aptv::deserializeODatum(*buffer, body);
    uint32_t firstID = parseID(blob);
    EventInfo info(parseTimestamp(blob), name_, origin, firstID);
    // Pack the buffer and the associated info into a Message.
    {
        std::scoped_lock lock(pendingMutex_);
        pending_.emplace_back(info, std::move(buffer));
    }
    // Push an event that will call run() on that message.
    // Note: it's important that this event is not holding onto the message
    // itself, because we may need to recycle it without processing it.
    enqueuer_(Event(info, [origin, firstID, this](uint32_t lastID) { return run(origin, firstID, lastID); }));
}

template<typename Type>
uint32_t Deserializer<Type>::run(const std::string_view& origin, uint32_t firstID, uint32_t lastID)
{
    // Filter out all pending messages that match this one's origin and the ID range.
    std::vector<Message<Type>> inputs;
    {
        std::scoped_lock lock(pendingMutex_);
        std::deque<Message<Type>> remnant;
        for (const Message<Type>& it : pending_) {
            // If lastID is 0, there is no limit on how many messages we process per event.
            if (it.origin() == origin && (lastID == 0 || it.id() <= lastID)) {
                // Check whether the target message was already handed in a previous batch.
                if (inputs.empty() && it.id() != firstID) {
                    return 0;
                }
                inputs.push_back(it);
            } else {
                remnant.push_back(it);
            }
        }
        // Store the remaining messages back in the queue.
        remnant.swap(pending_);
    }
    // If no messages matched (e.g. because we had to recycle them), there's no
    // need to log this event.
    if (inputs.empty()) {
        return 0;
    }
    handler_(inputs);
    // Return the ID of the last message we processed.
    return inputs.back().id();
}

template<typename Type>
std::shared_ptr<Type> Deserializer<Type>::getBuffer()
{
    // This function should never be called concurrently, and there is no other
    // way for any part of the code to get a reference to a buffer that is not
    // currently in use. So, in this context, if a buffer's use_count is 1, we
    // can trust that no one else is going to start using it.
    for (const std::shared_ptr<Type>& it : buffers_) {
        if (it.use_count() == 1) {
            return it;
        }
    }
    // All buffers are in use. If the number of buffers is unlimited, add one.
    if (unlimitedBuffers_) {
        buffers_.push_back(std::make_shared<Type>());
        return buffers_.back();
    }
    // Otherwise, recycle the oldest pending buffer.
    std::scoped_lock lock(pendingMutex_);
    if (!pending_.empty()) {
        // Don't bother finding the pointer in buffers_; just make a copy of
        // the shared_ptr portion of the message:
        auto buffer = static_cast<std::shared_ptr<Type>&>(pending_.front());
        pending_.pop_front();
        return buffer;
    }
    // Otherwise, drop this message because we have no buffers for it.
    return std::shared_ptr<Type>();
}

} // namespace av::repro
