/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Function definitions for the EventFramework class.
#include <repro/EventFramework.hpp>

#include "lib/avlog/avlog.hpp"
#include <Comms/Channel.h>
#include <SimpleComms/Constants.h>
#include <SimpleComms/ControlMessages.h>

namespace av::repro {

EventFramework::EventFramework()
    : queue_(scheduler_, outputID_)
    , replayer_(scheduler_, outputID_)
{}

void EventFramework::setClient(const std::shared_ptr<SimpleComms::Client>& client)
{
    client_ = client;
}

const std::shared_ptr<SimpleComms::Client>& EventFramework::client() const
{
    return client_;
}

void EventFramework::init(Mode mode, const std::string& logPath)
{
    mode_ = mode;

    // Open the log files before any messages start getting sent.
    if (mode_ == Mode::replay) {
        if (!replayer_.init(logPath)) {
            // If we're in replay mode but loading the log for this task failed,
            // switch over to running in frozen mode instead.
            AVLOG_ERROR("Unable to load events database for reproducible replay.");
            mode_ = Mode::frozen;
        }
    }
    if (mode_ == Mode::live || mode_ == Mode::frozen) {
        queue_.init(logPath);
    }

    if (client_) {
        // Loop until the client has been able to connect.
        while (client_->connect() != SimpleComms::Status::Ok) {
            AVLOG_WARN("Failed to connect to client. Trying again after 1s...");
            std::this_thread::sleep_for(std::chrono::seconds(1));
        }
        AVLOG_INFO("Client connected");

        isConnected_ = true;
        if (mode_ == Mode::central) {
            // Tell SCS that this client should be included in the message log.
            // Note: this message must be sent before the task declares any topics.
            SimpleComms::ReproducibleMessage message;
            // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
            std::string blob(reinterpret_cast<const char*>(&message), sizeof(message));
            client_->sendMsg(SIMPLECOMMS_CONTROL_CHANNEL, blob);
        }
        // Declare any channels that were set up before this.
        for (const std::pair<std::string, SimpleComms::ChannelType>& it : toDeclare_) {
            declare(it.first, it.second);
        }
        toDeclare_.clear();
    }
}

void EventFramework::start()
{
    if (mode_ == Mode::replay) {
        AVLOG_INFO("Starting task in replay mode.");
        replayer_.start();
    }
    // In central mode, scheduling is controlled by SCS and message handling
    // and event logging is handled by CallbackSet.
    else if (mode_ == Mode::live || mode_ == Mode::frozen) {
        bool frozen = (mode_ != Mode::live);
        AVLOG_INFO("Starting task in {} mode.", frozen ? "frozen" : "live");
        // Initialize the scheduler. Its thread should only run in "live" mode.
        // In "frozen" mode, the scheduler advances in response to message timestamps.
        // And, in "replay" mode, the triggers come from the client-side event log.
        scheduler_.start([this](Event&& event) { addEvent(std::move(event)); }, frozen);
        // In live processing mode, read the system clock time before running each event.
        queue_.start();
    }
}

void EventFramework::quit()
{
    if (mode_ == Mode::replay) {
        replayer_.quit();
    } else if (mode_ == Mode::live || mode_ == Mode::frozen) {
        scheduler_.quit();
        queue_.quit();
    }

    // Report disconnections. The previous code was only reporting disconnections from inputs,
    // so I have maintained the same behavior here.
    for (const auto& it : deserializers_) {
        task::Channel::reportDisconnected(it.first);
    }
}

EventFramework::Mode EventFramework::mode() const
{
    return mode_;
}

Duration EventFramework::frozenTime() const
{
    return (mode_ != Mode::replay ? queue_.frozenTime() : replayer_.frozenTime());
}

void EventFramework::declare(const std::string& topic, SimpleComms::ChannelType type)
{
    // Legacy special case: in server-side mode an empty topic name means the executive trigger.
    if (!topic.empty()) {
        if (isConnected_) {
            // Specify a queue depth of 1. Depth 0 means unlimited, which would be a
            // problem because the "legacy" queue is never read for these channels.
            client_->declare(topic, type, 1);
            task::Channel::reportConnected(topic);
        } else {
            toDeclare_.emplace_back(topic, type);
        }
    }
}

uint32_t EventFramework::outputID()
{
    return outputID_;
}

uint32_t EventFramework::nextOutputID()
{
    return outputID_++;
}

void EventFramework::setTimeoutHandler(const std::string& name, const TimeoutHandler& handler, Duration repeat)
{
    scheduler_.setHandler(name, handler, repeat);
}

void EventFramework::schedule(const std::string& name, Duration timeout, Duration repeat)
{
    scheduler_.schedule(name, timeout, repeat);
}

void EventFramework::unschedule(const std::string& name)
{
    scheduler_.unschedule(name);
}

void EventFramework::receive(const std::string& blob, const std::string& topic, const std::string& origin)
{
    // Always ignore control messages.
    if (topic == SIMPLECOMMS_CONTROL_CHANNEL) {
        return;
    }
    // Strip the "-$$$$" random suffix from the origin name.
    std::string baseOrigin = origin.substr(0, origin.rfind('-'));
    // If we're in replay mode, check if this is a message we should ignore.
    // If so, there's no need to deserialize it.
    if (mode_ == Mode::replay && replayer_.ignore(topic, baseOrigin, parseID(blob))) {
        return;
    }
    // In frozen mode, use the playback clock messages to advance the scheduler.
    if (mode_ == Mode::frozen && topic == "PlaybackClockParams") {
        scheduler_.advance(parsePlaybackClock(blob));
        return;
    }
    std::shared_ptr<DeserializerBase> deserializer;
    std::string_view originFlyweight;
    {
        std::scoped_lock lock(deserializersMutex_);
        auto it = deserializers_.find(topic);
        if (it == deserializers_.end()) {
            return;
        }
        deserializer = it->second;
        originFlyweight = *origins_.emplace(baseOrigin).first;
    }
    // Rather than copying strings, all EventInfo objects will refer to a
    // flyweight origin string stored in EventFramework:
    deserializer->receive(blob, originFlyweight);
}

void EventFramework::addEvent(Event&& event)
{
    if (mode_ != Mode::replay) {
        queue_.add(std::move(event));
    } else {
        replayer_.add(std::move(event));
    }
}

void EventFramework::setDeserializer(const std::string& topic, const std::shared_ptr<DeserializerBase>& deserializer)
{
    // Note: this functionality was moved into the .cpp file because apparently using
    // AVLOG in a templated function can cause ASAN link errors.
    {
        std::scoped_lock lock(deserializersMutex_);
        auto& entry = deserializers_[topic];
        if (entry) {
            AVLOG_ERROR("Multiple handlers set for topic \"{}\".", topic);
            return;
        }
        entry = deserializer;
    }
    declare(topic, SimpleComms::ChannelType::Subscriber);
}

} // namespace av::repro
