/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Definition of the EventFramework class.
#pragma once

#include <repro/Deserializer.hpp>
#include <repro/DeserializerBase.hpp>
#include <repro/Event.hpp>
#include <repro/EventQueue.hpp>
#include <repro/Replayer.hpp>
#include <repro/Scheduler.hpp>
#include <repro/handlerTypes.hpp>

#include <SimpleComms/Client.h>
#include <avtime/AvClock.hpp>

#include <functional>
#include <map>
#include <memory>
#include <mutex>
#include <set>
#include <string>
#include <utility>
#include <vector>

namespace av::repro {

class EventFramework
{
public:
    /// Operating modes.
    enum class Mode
    {
        live,   ///< Scheduling events, processing messages as they arrive, and optionally logging.
        frozen, ///< Clock is controlled by the timestamps of incoming messages.
        replay, ///< Replaying events (message batches and timeouts) from a log file.
        central ///< Using "central scheduling" from SCS, and never batching messages. (deprecated)
    };

public:
    /// Default constructor.
    EventFramework();

    /// Set the comms client.
    void setClient(const std::shared_ptr<SimpleComms::Client>& client);
    /// Get the comms client.
    const std::shared_ptr<SimpleComms::Client>& client() const;

    /// Connect to comms using the given client.
    void init(Mode mode, const std::string& logPath = "");
    /// Initialize the framework in the given mode and begin processing events, optionally saving an events log.
    void start();
    /// Stop event handling.
    /// @note Event handling is also automatically interrupted in this object's destructor.
    void quit();

    /// Get the operating mode.
    Mode mode() const;

    /// Get the current (frozen) time, as a duration since the Unix epoch.
    Duration frozenTime() const;

    /// Declare an input or output.
    void declare(const std::string& topic, SimpleComms::ChannelType type);
    /// Get the ID that will be assigned to the next output, without incrementing it.
    uint32_t outputID();
    /// Get the ID that should be assigned to the next output, and increment the ID.
    /// @note This should only be called within the callback thread.
    uint32_t nextOutputID();

    /// Set the handler for a timeout event.
    /// @param name: The name of the timeout.
    /// @param handler: The function to be called for each timeout event of this type.
    void setTimeoutHandler(const std::string& name, const TimeoutHandler& handler, Duration repeat = Duration{0});
    template<typename Receiver>
    void setTimeoutHandler(const std::string& name,
                           Receiver* receiver,
                           void (Receiver::*callback)(),
                           Duration repeat = Duration{0});
    /// Set the handler for a given topic.
    /// @param topic: The name of the topic.
    /// @param handler: The function to be called for each message event from this topic.
    /// @param buffers: The maximum number of buffers to allocate for this topic, or 0 if there should be no limit.
    /// @note Only one callback per topic is supported. If called a second time, this does nothing.
    template<typename Type>
    void setMessageHandler(const std::string& topic, const MessageHandler<Type>& handler, size_t buffers = 0);
    template<typename Receiver, typename Type>
    void setMessageHandler(const std::string& topic,
                           Receiver* receiver,
                           void (Receiver::*callback)(const Inputs<Type>&),
                           size_t buffers = 0);

    /// Schedule a timeout, which will optionally repeat automatically.
    void schedule(const std::string& name, Duration timeout = Duration{0}, Duration repeat = Duration{0});
    /// Cancel whatever timeout is scheduled in the given channel.
    void unschedule(const std::string& name);
    /// This function should be called by the Comms framework whenever a message arrives.
    /// It will deserialize the message and convert it into an event, then add it to the queue.
    void receive(const std::string& blob, const std::string& topic, const std::string& origin);

private:
    /// Add an event (a deserialized message or a timeout) to the queue.
    void addEvent(Event&& event);
    /// Set the deserializer for the given topic, reporting an error if one is already set.
    void setDeserializer(const std::string& topic, const std::shared_ptr<DeserializerBase>& deserializer);

private:
    // Pointer to the comms client.
    std::shared_ptr<SimpleComms::Client> client_;
    bool isConnected_{false};
    // Interfaces to declare as soon as we're connected to comms.
    std::vector<std::pair<std::string, SimpleComms::ChannelType>> toDeclare_;
    // Remember the operating mode.
    Mode mode_{Mode::live};
    // Mapping of input topic names to the objects responsible for deserializing and handling messages of that type.
    // This map should already have grown to its full size before any messages are received, but just to be on the
    // safe side use a map instead of an unordered map so references held by event functions are never invalidated.
    std::mutex deserializersMutex_;
    std::map<std::string, std::shared_ptr<DeserializerBase>> deserializers_;
    // Flyweights for origin names. Use a set so that references are never invalidated.
    std::set<std::string> origins_;
    // Scheduler object.
    Scheduler scheduler_;
    // Event queue, which processes events in its own thread.
    EventQueue queue_;
    // Replayer, which takes the place of the Scheduler and EventQueue.
    Replayer replayer_;
    // Rolling output ID, for uniquely identifying messages.
    uint32_t outputID_{1};
};

// Templated functions:

template<typename Receiver>
void EventFramework::setTimeoutHandler(const std::string& name,
                                       Receiver* receiver,
                                       void (Receiver::*callback)(),
                                       Duration repeat)
{
    setTimeoutHandler(name, std::bind(callback, receiver), repeat);
}

template<typename Type>
void EventFramework::setMessageHandler(const std::string& topic, const MessageHandler<Type>& handler, size_t buffers)
{
    auto enqueuer = [this](Event&& event) { addEvent(std::move(event)); };
    auto deserializer = std::make_shared<Deserializer<Type>>(enqueuer, topic, handler, buffers);
    setDeserializer(topic, deserializer);
}

template<typename Receiver, typename Type>
void EventFramework::setMessageHandler(const std::string& topic,
                                       Receiver* receiver,
                                       void (Receiver::*callback)(const Inputs<Type>&),
                                       size_t buffers)
{
    setMessageHandler(topic, MessageHandler<Type>(std::bind(callback, receiver, std::placeholders::_1)), buffers);
}

} // namespace av::repro
