/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Function definitions for the EventQueue class.
#include <repro/EventQueue.hpp>

#include <repro/Event.hpp>
#include <repro/Scheduler.hpp>

namespace av::repro {

EventQueue::EventQueue(Scheduler& scheduler, const uint32_t& outputID)
    : scheduler_(scheduler)
    , outputID_(outputID)
{}

EventQueue::~EventQueue()
{
    // Make sure the thread quits.
    quit();
}

void EventQueue::init(const std::string& logPath)
{
    if (!logPath.empty()) {
        log_.open(logPath, "time", "outID", "topic", "origin", "firstID", "lastID");
    }
}

void EventQueue::start()
{
    thread_ = std::thread(&EventQueue::loop, this);
}

void EventQueue::quit()
{
    if (thread_.joinable()) {
        {
            std::scoped_lock lock(mutex_);
            interrupted_ = true;
        }
        condition_.notify_all();
        thread_.join();
    }
    // Make sure the log is flushed.
    log_.close();
}

void EventQueue::add(Event&& event)
{
    {
        std::scoped_lock lock(mutex_);
        events_.emplace(std::move(event));
    }
    condition_.notify_all();
}

Duration EventQueue::frozenTime() const noexcept
{
    // Only the handler thread should access this, so no mutex is needed.
    return frozenTime_;
}

void EventQueue::loop()
{
    while (true) {
        Event event;
        {
            std::unique_lock lock(mutex_);
            condition_.wait(lock, [this] { return !events_.empty() || interrupted_; });
            if (interrupted_) {
                return;
            }
            event = events_.front();
            events_.pop();
        }
        // Update the frozen clock, either using the timestamp of this message or
        // the current system time depending on the operating mode.
        frozenTime_ = scheduler_.now();
        // Remember what the output ID was before the event ran.
        uint32_t outID = outputID_;
        // Run the event handler, telling it that it can batch as needed.
        uint32_t lastID = event(0);
        // This is the only place we write the log, so it doesn't need mutex protection.
        if (lastID != 0 && log_) {
            log_.write(frozenTime_.count(), outID, event.topic(), event.origin(), event.id(), lastID);
        }
    }
}

} // namespace av::repro
