/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Definition of the EventQueue class.
#pragma once

#include <avdb/SQLWriter.hpp>
#include <avtime/AvClock.hpp>

#include <condition_variable>
#include <cstdint>
#include <memory>
#include <mutex>
#include <queue>
#include <string_view>
#include <thread>

namespace av::repro {

// Forward declarations.
class Event;
class Scheduler;

/// An event queue, which runs events in a worker thread until an interrupt is received.
class EventQueue
{
public:
    /// Constructor, taking a reference to the Scheduler and the next output ID (so it can be logged).
    EventQueue(Scheduler& scheduler, const uint32_t& outputID);
    /// Destructor, which makes sure the processing thread quits.
    ~EventQueue();
    // Don't allow moving or copying.
    EventQueue(const EventQueue&) = delete;
    EventQueue(EventQueue&&) = delete;

    EventQueue& operator=(const EventQueue&) = delete;
    EventQueue& operator=(EventQueue&&) = delete;

    /// Begin writing the event database.
    void init(const std::string& logPath);
    /// Launch the thread.
    void start();
    /// Wait for the thread to finish an item, then quit.
    /// @note This does not wait for all items in the queue to be processed.
    void quit();

    /// Add an event to the queue.
    void add(Event&& event);

    /// Get the current "frozen" time.
    /// @note If no events have run yet this will return an invalid time (prior to the Unix epoch).
    Duration frozenTime() const noexcept;

private:
    void loop();

private:
    // Collection of events.
    std::queue<Event> events_;
    // Thread servicing that collection.
    std::thread thread_;
    // Allow interrupting the thread.
    std::mutex mutex_;
    std::condition_variable condition_;
    bool interrupted_{false};
    // "Frozen" time, to be returned whenever a handler checks the current time.
    // Some code depends on the timestamp being before the Unix epoch if it's not set yet.
    // TODO: It would be cleaner if that code (e.g. in ResimPubMissionGoalTask) considered
    // a value of 0 nsec since the epoch to be invalid, too.
    Duration frozenTime_{-1};
    // Event log: (time, outID, channel, origin, firstID, lastID)
    using Log = db::SQLWriter<int64_t, uint32_t, std::string_view, std::string_view, uint32_t, uint32_t>;
    Log log_;
    // Reference to the scheduler, in order to query the current time.
    Scheduler& scheduler_;
    // Reference to the next output ID.
    const uint32_t& outputID_;
};

} // namespace av::repro
