## Making Tasks Reproducible

This document describes how to use the reproducible framework for handling messages and scheduling periodic functions and timeouts. A task is reproducible if every time it is presented with the same events in the same order, it produces the exact same output. To ensure reproducibility, we must avoid:

* race conditions in reading inputs (including messages and the clock)
* randomized algorithms (unless the random seed is the same every time)
* behavior that depends on CPU load and timing

The “events” processed by a task fall into two different categories:

* messages received by the comms framework
* scheduled events (timeouts, periodic functions, etc.)

The reproducible framework provides new interfaces for handling both these event types.

## Message handling

Messages are received from the comms framework, which may either be SCS or DDS depending on the system configuration.

### Limitations of the legacy interfaces

The legacy comms framework is built around the `GenericInput` and `GenericOutput` interfaces. Most legacy tasks were originally designed to poll inputs to check when messages arrived, and to process those messages in an “executive function” that runs periodically. That introduced several problems:

* If a task polls the interface very frequently, its main thread keeps waking up just to find that there are no messages available. This wastes CPU time.
* If a task polls the interface infrequently, there will be a delay between when the message arrives and when it gets handled. (The delay, on average, will be half the executive frequency.) This increases overall latencies.
* Some tasks used to create a separate thread just to make a blocking call to read an input, so they could process that input as soon as it arrived. This introduces a race condition and also requires manually managing those threads.
* Both GenericInput and GenericOutput internally use message queues that are regularly polled by a separate thread. This is another race condition, and also a waste of CPU resources.

New task code should not use these interfaces.

### Reproducible interfaces for message handling

The reproducible framework provides a pollable input interface (`MessageQueue`) but it also provides an alternative that often leads to simpler and cleaner code: callbacks. A task can register a function to be called any time an input arrives. All callback handlers (including the executive function) run in the same thread, so they can access the task’s state without any need for mutex protection. Unless another handler is currently running, messages callbacks will run as soon as a message is received, ensuring minimal latency.

Sometimes a task is unable to process messages fast enough to keep up with the inputs. For example, if the LSN task is receiving a point cloud every 50 ms. but it’s averaging 60 ms. to process them, it will continue to get further and further behind. Previously we handled that by just dropping old messages if the queue of pending messages fills up. But, the FASTER proposal includes a better plan for handling that situation: every time a message handler is called, it is given access to _all_ the pending messages for a given topic. (We're calling this “batching” of inputs.) If there is more than one pending message, the task knows that it is falling behind and can intelligently choose how to catch up (e.g. by ignoring all but the most recent message).

Callback handlers can be any function object that takes a `const Inputs<MessageType>&` argument and returns void. That includes lambdas and member functions, such as the following:

    void MyTask::handleMessageType(const Inputs<MessageType>& inputs) { ... }

The function does not need to be a member of your `MyTask` class. For example, if you have classes representing a processing pipeline it might be helpful to register the first object in the pipeline to receive the messages.

The `Inputs<MessageType>` object can be treated as an iterable container of shared pointers to `MessageType`. (In the current implementation it’s a `std::vector`, but when we switch to FASTER the same typedef will probably refer to a different container type.) In addition to holding the messages themselves, the elements in the container also inherit from the `EventInfo` class, so they provide the following functions:

* `time()`: the time when the message was sent
* `topic()`: the name of the topic where this message was published
* `origin()`: the sender of the message
* `id()`: a unique rolling ID assigned to messages from a given origin

Most of this information is only useful to the internal framework code, but there are a few ways it might be useful in task code. For example, the timestamp can be checked to recognize communication delays. And, if multiple topics are registered to use the same handler, the `topic()` can be used to tell the difference between them. (For example, a pipeline merging inputs from multiple lidars or cameras could pass them all to the same callback instead of having to register a different callback for each topic.)

An object can register a handler as follows, using the protected `framework_` variable in `Task`:

    framework_.setMessageHandler("TopicName", this, &MyTask::handleMessageType, bufferCount);

The `bufferCount` is an optional parameter limiting the number of message buffers to be allocated for this topic. If it is 0 (the default) there is no limit and old messages are never discarded. Otherwise, all buffers are allocated when the handler is initialized and old messages are discarded if all the buffers fill up.

The “server-side” reproducibility framework also provides the capability to register a handler through the `CallbackSet` object. These handlers receive a single message at a time, i.e. they have the following signature:

    void MyTask::handleMessageType(const MessageType& message) { ... }

If using `CallbackSet`, dropping stale messages is handled internally and you have less control over it, so for new code (_especially_ if it's possible the task could get behind on processing messages) you should use `EventFramework` instead. Using the newer “batched” inputs will also make it easier to port your task to FASTER in the future.

## Scheduled events

There are several use cases for scheduled events:

* Periodic events, such as the executive function or the functions that publish status messages (`AvTaskStatus`, etc.)
* Input watchdog timers (reporting an error if no input is received in a given topic for some amount of time)
* Output watchdog timers (publishing a “dummy” output if no valid output was produced for some amount of time)
* Triggered events (“step” functions), e.g. a handler that should be scheduled to run whenever any one of several message handlers determine that all the necessary inputs are ready.

In the legacy task framework the only one of these functions that is scheduled automatically is the executive function; all the others need to be simulated by checking elapsed times (e.g. publishing `AvTaskStatus` if it’s been more than a second since the last time it was published). The reproducible framework, on the other hand, allows you to schedule any of these events using the same interface. A timeout handler can be any function (including a lambda or a member function) that takes no arguments and returns void. Handlers are 

First, a timeout handler function must be registered by name, e.g.:

    framework_.setTimeoutHandler("TimeoutName", this, &MyTask::handleTimeoutName, repeatInterval);

The `repeatInterval` is an optional `Duration` argument. If it is given, the function will be called periodically at that interval. Otherwise the timeout has to be scheduled manually. For example:

**Periodic status messages:** Suppose we want to send a status message every second to report that this task is still alive. We can do that by scheduling a timeout that triggers immediately and then repeats every second:

    // In the initialization code:
    framework_.schedule("!periodic", Duration(0), std::chrono::seconds(1));

(Although it is not required, starting timeout handler names with a special character like `!` is a good practice to avoid collisions with channel names.)

**Watchdog timers:** Suppose we want to print a warning message if it has been a second since the last time an input was received. We could do that by resetting the timeout every time we receive that input:

    // In the input handler:
    handleInput(myInput);
    framework_.schedule("!watchdog", std::chrono::milliseconds(100));

Suppose that 80 milliseconds later, the input handler gets called again. It will reset the timeout, clearing the old scheduled timeout. That is, rather than triggering 20 milliseconds later the timeout handler will only trigger if another 100 milliseconds elapse before the next input is received. If there is always less than 100 milliseconds in between inputs, the watchdog will never trigger.

**Publishing dummy data:** Suppose that if it has been too long since we produced an output, we want to produce dummy data (e.g. an empty message or a repeat of the previous message) for the sake of downstream tasks. We could schedule a periodic timeout when the task initializes:

    // In the initialization code:
    framework_.schedule("!dummy", std::chrono::milliseconds(100), std::chrono::milliseconds(100));

Then, every time we publish data on that channel we reset the timeout:

    // In the code that publishes the message:
    publish(myOutput);
    framework_.schedule("!dummy", std::chrono::milliseconds(100), std::chrono::milliseconds(100));

As with the watchdog timer example, if the real message is getting published at least every 100 milliseconds, the dummy handler will never run. Note that if a timeout is scheduled during task initialization, its "clock" does not begin running until initialization is done. For example, if you schedule a watchdog to trigger after one second, then spend five seconds loading a map file, then finish initializing, the watchdog will not trigger until a second later.

**Step functions:** Suppose you are receiving inputs in two different handlers and only want to process them when you have matching data from both. A clean way to do that would be for either handler to schedule a processing function to run if it detects that the data matches:

    // In both event handlers:
    if (messagesAreAligned()) {
        framework_.schedule("!step");
    }

The handler will run as soon as possible, after any other pending events in the queue.

## Clock time

If a task reads the system clock directly, that introduces two problems:

* If reprocessing, the clock time will be the current wall clock time instead of the time when the log was recorded
* If the clock time affects calculations, there's no way to reproduce those calculations unless we reproduce the original clock timestamps

To avoid these issues, the reproducible framework reads the system clock once right before it calls each handler. That time can be read by calling `framework_.frozenTime()`; it is also returned by `task::Task::getSystemTime()`. This does mean that if a calculation takes a long time, the clock time may be a bit out of date by the time that calculation finishes, but in practice this does not appear to be a serious limitation in our current use cases. (If the clock is only being used for benchmarking and its values will not affect the outputs, it's fine to use something like `AvClock` or `std::chrono::steady_clock` instead.)
