/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Function definitions for the Replayer class.
#include <repro/Replayer.hpp>

#include <repro/Event.hpp>
#include <repro/Scheduler.hpp>

#include "lib/avlog/avlog.hpp"

namespace av::repro {

namespace {
/// Check if the given origin string is a possible name of the log player or vehicle simulator task.
bool isLogPlayer(const std::string_view& origin)
{
    static const std::string PREFIX1 = "logplayer";
    static const std::string PREFIX2 = "VehicleDynamicsSim";
    return ((origin.compare(0, PREFIX1.size(), PREFIX1) == 0) || (origin.compare(0, PREFIX2.size(), PREFIX2) == 0));
}
} // anonymous namespace

Replayer::Replayer(Scheduler& scheduler, uint32_t& outputID)
    : scheduler_(scheduler)
    , outputID_(outputID)
{}

Replayer::~Replayer()
{
    // Make sure the thread quits.
    quit();
}

bool Replayer::init(const std::string& logPath)
{
    log_.open(logPath);
    return log_;
}

void Replayer::start()
{
    thread_ = std::thread(&Replayer::loop, this);
}

void Replayer::quit()
{
    if (thread_.joinable()) {
        {
            std::scoped_lock lock(mutex_);
            interrupted_ = true;
        }
        condition_.notify_all();
        thread_.join();
    }
}

bool Replayer::ignore(const std::string& topic, const std::string& origin, uint32_t id)
{
    // Messages from the log player can match previously logged messages from other tasks.
    bool isFromLogPlayer = isLogPlayer(origin);
    // Keep track of whether we need to notify the thread to run.
    bool ready = false;
    bool ignore = false;
    {
        std::scoped_lock lock(mutex_);
        // Find the first event in the upcoming list that matches this one's topic and origin.
        auto it = upcoming_.begin();
        while (true) {
            if (it == upcoming_.end()) {
                if (!log_) {
                    AVLOG_INFO("Unmatched message {} in topic {}.", id, topic);
                    ignore = true;
                    break;
                }
                it = addRow();
                ready |= (it == upcoming_.begin() && it->ready);
            }
            // Ignore rows that already contain all the messages they expect.
            if (!it->ready && it->topic == topic && (isFromLogPlayer || it->origin == origin)) {
                // If the new event was supposed to happen before this one, it must be an event
                // that happened before this task started up (or before this cut log segment).
                // In that case, it should simply be ignored:
                if (id < it->firstID) {
                    AVLOG_INFO("Ignoring message {} in topic {}.", id, topic);
                    ignore = true;
                    break;
                }
                // If the new event comes after all the events in this entry, keep going to see
                // if a later entry will match it. Otherwise, we've found a match.
                if (id <= it->lastID) {
                    break;
                }
                // If this event is not ready to run but we are already seeing messages that
                // match later events, this message will never be matched and must be deleted.
                AVLOG_WARN("Missing messages {}-{} in topic {}.", it->firstID, it->lastID, topic);
                it = upcoming_.erase(it);
                ready |= (it == upcoming_.begin() && it->ready);
                continue;
            }
            ++it;
        }
    }
    if (ready) {
        // Notify the event thread if we added a timeout event to the front of the queue.
        condition_.notify_all();
    }
    return ignore;
}

void Replayer::add(Event&& event)
{
    // Messages from the log player can match previously logged messages from other tasks.
    bool isFromLogPlayer = isLogPlayer(event.origin());
    bool ready = false;
    {
        std::scoped_lock lock(mutex_);
        // If we get here, ignore() returned false for this message, so it must match something.
        for (Row& row : upcoming_) {
            if (row.topic == event.topic() && (isFromLogPlayer || row.origin == event.origin())) {
                if (row.firstID <= event.id() && event.id() <= row.lastID) {
                    // If this is the last event we expected for a row, we can run it.
                    if (event.id() == row.lastID) {
                        row.ready = true;
                        // We only need to notify the event thread if this is the first upcoming event,
                        // meaning that up until now the thread was blocked waiting for it.
                        ready = (&row == &upcoming_.front());
                    }
                    // If this is the first event we've seen for this row, store the event pointer.
                    if (!row.event) {
                        if (event.id() != row.firstID) {
                            AVLOG_WARN("Missing message {} in topic {}.", row.firstID, row.topic);
                        }
                        row.event = std::make_shared<Event>(std::move(event));
                    }
                    break;
                }
            }
        }
    }
    if (ready) {
        // Notify the event thread if the message event in the front of the queue just became ready.
        condition_.notify_all();
    }
}

Duration Replayer::frozenTime() const
{
    return frozenTime_;
}

void Replayer::loop()
{
    while (true) {
        Row row;
        {
            std::unique_lock lock(mutex_);
            condition_.wait(lock, [this] { return interrupted_ || upcoming_.empty() || upcoming_.front().ready; });
            if (interrupted_) {
                return;
            }
            if (upcoming_.empty()) {
                if (!log_ || addRow() == upcoming_.end()) {
                    // We've reached the end of the log.
                    return;
                }
                if (!upcoming_.front().ready) {
                    continue;
                }
            }
            // If we're here, the next upcoming event is ready to run.
            row = upcoming_.front();
            upcoming_.pop_front();
        }
        // Update the frozen clock and the output message ID.
        frozenTime_ = Duration(row.time);
        outputID_ = row.outID;
        if (row.event) {
            // Replay a message event, consuming only up to the last ID.
            (*row.event)(row.lastID);
        } else {
            // Replay a timeout event. Because heterogeneous lookup is not enabled,
            // we need to convert the topic name into a string manually.
            scheduler_.replay(std::string(row.topic));
        }
    }
}

std::deque<Replayer::Row>::iterator Replayer::addRow()
{
    Row row;
    std::tie(row.time, row.outID, row.topic, row.origin, row.firstID, row.lastID) = log_.read();
    row.ready = (row.origin == Scheduler::origin());
    return upcoming_.insert(upcoming_.end(), std::move(row));
}

} // namespace av::repro
