/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Definition of the Replayer class.
#pragma once

#include <avdb/SQLReader.hpp>
#include <avtime/AvClock.hpp>

#include <condition_variable>
#include <cstdint>
#include <deque>
#include <memory>
#include <mutex>
#include <string>
#include <string_view>
#include <thread>

namespace av::repro {

// Forward declarations.
class Event;
class Scheduler;

/// Class for managing the event order during client-side reproducible resim.
class Replayer
{
public:
    /// Constructor. The Replayer is allowed to change the next output ID.
    Replayer(Scheduler& scheduler, uint32_t& outputID);
    /// Destructor, which makes sure the processing thread quits.
    ~Replayer();
    // Don't allow moving or copying.
    Replayer(const Replayer&) = delete;
    Replayer(Replayer&&) = delete;

    Replayer& operator=(const Replayer&) = delete;
    Replayer& operator=(Replayer&&) = delete;

    /// Begin reading the events database.
    /// @return False if reading the database file fails.
    bool init(const std::string& logPath);
    /// Start the event-processing thread.
    void start();
    /// Wait for the thread to finish an item, then quit.
    /// @note This does not wait for all items in the queue to be processed.
    void quit();

    /// Check if the given message should be ignored.
    bool ignore(const std::string& topic, const std::string& origin, uint32_t id);
    /// Add an event to the queue.
    void add(Event&& event);

    /// Get the current "frozen" time.
    Duration frozenTime() const;

private:
    // Information we must track for each upcoming event:
    struct Row
    {
        int64_t time{0};
        uint32_t outID{0};
        std::string_view topic;
        std::string_view origin;
        uint32_t firstID{0};
        uint32_t lastID{0};
        // The event to run when this row is ready. This is the first event in the batch.
        std::shared_ptr<Event> event;
        // This will be set when the event is ready to run.
        bool ready{false};
    };

private:
    void loop();
    // Read one more row of the upcoming events.
    std::deque<Row>::iterator addRow();

private:
    // Reference to the scheduler object, to be used to replay timeout events.
    Scheduler& scheduler_;
    // Reference to the next output ID, to be set based on the reproducible log.
    uint32_t& outputID_;
    // Record of upcoming expected events:
    std::deque<Row> upcoming_;
    // Thread for running the events.
    std::thread thread_;
    // Allow interrupting the thread or notifying it when new events are ready.
    std::mutex mutex_;
    std::condition_variable condition_;
    bool interrupted_{false};
    // "Frozen" time, to be returned whenever a handler checks the current time.
    // Some code depends on the timestamp being before the Unix epoch if it's not set yet.
    Duration frozenTime_{-1};
    // Event log: (time, outID, channel, origin, firstID, lastID)
    using Log = db::SQLReader<int64_t, uint32_t, std::string_view, std::string_view, uint32_t, uint32_t>;
    Log log_;
};

} // namespace av::repro
