/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Function definitions for the Scheduler class.
#include <repro/Scheduler.hpp>

#include <repro/Event.hpp>
#include <repro/EventInfo.hpp>

#include "lib/avlog/avlog.hpp"
#include <avtime/AvClock.hpp>

namespace av::repro {

namespace {
// This string is used for the "origin" to identify scheduler events.
const std::string ORIGIN = "scheduler";
} // namespace

Scheduler::~Scheduler()
{
    quit();
}

void Scheduler::start(const Enqueuer& enqueuer, bool frozen)
{
    std::scoped_lock lock(mutex_);
    enqueuer_ = enqueuer;
    // When this function is called, the thread is not running yet,
    // so we don't need mutex protection.
    frozen_ = frozen;
    if (!frozen) {
        start(protectedNow());
        thread_ = std::thread(&Scheduler::loop, this);
    }
}

void Scheduler::quit()
{
    if (thread_.joinable()) {
        {
            std::scoped_lock lock(mutex_);
            interrupted_ = true;
        }
        condition_.notify_all();
        thread_.join();
    }
}

Duration Scheduler::now() const
{
    // If we're in live mode, there's no need for mutex protection.
    if (!frozen_) {
        return AvClock::now().time_since_epoch();
    }
    std::scoped_lock lock(mutex_);
    return now_;
}

void Scheduler::advance(Duration now)
{
    std::scoped_lock lock(mutex_);
    // Don't advance the clock if start() has not been called yet.
    // In that case we don't even have a link to the EventQueue yet.
    if (!enqueuer_) {
        return;
    }
    // If the time is before the current clock time by more than a tiny amount,
    // assume that the log player has rewound. Reschedule all events relative
    // to the new "now." Also reschedule events if the clock is not started yet,
    // or if the clock moves forward a very large amount.
    bool rewind = (now < now_ - std::chrono::milliseconds(100));
    // Assume every task is receiving messages at least once a second, so if
    // more than a second elapses it must be fast-forwarding.
    bool fastForward = (now > now_ + std::chrono::milliseconds(1000));
    if (now_ == Duration{0} || rewind || fastForward) {
        start(now);
    }
    // Trigger any events that were scheduled for before this time.
    while (!schedule_.empty() && advanceOne(now)) {
    }
}

void Scheduler::setHandler(const std::string& name, const TimeoutHandler& handler, Duration repeat)
{
    {
        std::scoped_lock lock(mutex_);
        auto result = timeouts_.emplace(name, Timeout({name, handler, Duration{0}, repeat}));
        if (!result.second) {
            AVLOG_WARN("Multiple handlers set for timeout {}.", name);
        } else if (repeat != Duration{0}) {
            reschedule(result.first->second, now_, repeat);
        }
    }
    condition_.notify_all();
}

void Scheduler::schedule(const std::string& name, Duration timeout, Duration repeat)
{
    {
        std::scoped_lock lock(mutex_);
        auto it = timeouts_.find(name);
        if (it != timeouts_.end()) {
            reschedule(it->second, now_ + timeout, repeat);
        }
    }
    condition_.notify_all();
}

void Scheduler::unschedule(const std::string& name)
{
    {
        std::scoped_lock lock(mutex_);
        auto it = timeouts_.find(name);
        if (it != timeouts_.end()) {
            it->second.time = Duration{0};
        }
    }
    condition_.notify_all();
}

void Scheduler::replay(const std::string& name)
{
    std::scoped_lock lock(mutex_);
    auto it = timeouts_.find(name);
    if (it != timeouts_.end()) {
        it->second.handler();
    } else {
        AVLOG_WARN("Unable to replay timeout {}; no handler registered.", name);
    }
}

const std::string& Scheduler::origin()
{
    return ORIGIN;
}

Scheduler::ScheduledTimeout::ScheduledTimeout(Timeout& timeout)
    : time(timeout.time)
    , current(&timeout)
{}

bool Scheduler::ScheduledTimeout::operator<(const ScheduledTimeout& other) const
{
    // Since std::priority_queue returns the maximum element by default,
    // use > instead of < to get the smallest element.
    return time > other.time;
}

Scheduler::ScheduledTimeout::operator bool() const
{
    // If this scheduled timeout has been cancelled or superseded, the time
    // stored here won't match the time in the timeout object.
    return (time == current->time);
}

void Scheduler::start(Duration now)
{
    // If the log player rewound, it's possible that this function is actually shifting
    // the schedule instead of initializing it for the first time.
    Duration offset = now - now_;
    now_ = now;
    // Convert all times from durations to time since epoch.
    for (auto& it : timeouts_) {
        it.second.time += offset;
    }
    // Also reset times in the queue.
    std::priority_queue<ScheduledTimeout> rescheduled;
    while (!schedule_.empty()) {
        ScheduledTimeout timeout = schedule_.top();
        schedule_.pop();
        // Convert the time, and add the timeout back to the queue unless it has been superseded.
        timeout.time += offset;
        if (timeout) {
            rescheduled.push(timeout);
        }
    }
    schedule_.swap(rescheduled);
}

void Scheduler::loop()
{
    std::unique_lock lock(mutex_);
    while (!interrupted_) {
        // First, clear any cancelled events from the queue.
        while (!schedule_.empty() && !schedule_.top()) {
            schedule_.pop();
        }
        // If no events are scheduled, sleep until the condition variable is notified.
        if (schedule_.empty()) {
            condition_.wait(lock);
        } else if (!advanceOne(protectedNow())) {
            // There is an event scheduled, but it's not time for it yet.
            // Wait until either we reach that time or the schedule changes.
            condition_.wait_until(lock, TimePoint() + schedule_.top().time);
        }
    }
}

bool Scheduler::advanceOne(Duration now)
{
    now_ = now;
    // Assume this function is only called when the queue is non-empty.
    ScheduledTimeout it = schedule_.top();
    if (it.time > now) {
        return false;
    }
    schedule_.pop();
    // Make sure this timeout has not been cancelled.
    if (it) {
        EventInfo info(now, it.current->name, ORIGIN, it.current->count);
        enqueuer_(Event(info, [it, this](uint32_t /*lastID*/) {
            Timeout timeout;
            {
                std::scoped_lock lock(mutex_);
                if (!it) {
                    return 0u;
                }
                timeout = *it.current;
            }
            timeout.handler();
            // Don't schedule the next iteration of this function until the current one is done,
            // to make sure we first process all messages received up to this point.
            if (timeout.repeat != Duration{0}) {
                {
                    std::scoped_lock lock(mutex_);
                    // If the handler took longer than the repeat interval, schedule it
                    // to repeat immediately. Otherwise repeat at that interval.
                    Duration currentTime = protectedNow();
                    Duration nextTime = it.time + timeout.repeat;
                    if (nextTime < currentTime) {
                        nextTime = currentTime;
                        AVLOG_WARN("Repeating event \"{}\" took longer than its requested period ({} > {} ms).",
                                   it.current->name,
                                   std::chrono::duration_cast<std::chrono::milliseconds>(currentTime - it.time).count(),
                                   std::chrono::duration_cast<std::chrono::milliseconds>(timeout.repeat).count());
                    }
                    reschedule(*it.current, nextTime, timeout.repeat);
                    // In frozen mode, we need to make sure the clock did not already
                    // advance past when this event was supposed to run.
                    if (frozen_) {
                        advanceOne(now_);
                    }
                }
                // In live mode, wake up the scheduler thread to check the new event.
                if (!frozen_) {
                    condition_.notify_all();
                }
            }
            return timeout.count;
        }));
    }
    return true;
}

void Scheduler::reschedule(Timeout& entry, Duration time, Duration repeat)
{
    // Assume the mutex is locked already.
    entry.time = time;
    entry.repeat = repeat;
    ++entry.count;
    schedule_.emplace(entry);
}

Duration Scheduler::protectedNow() const
{
    // Assume the mutex is locked already.
    return frozen_ ? now_ : AvClock::now().time_since_epoch();
}

} // namespace av::repro
