/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Definition of the Scheduler class.
#pragma once

#include <repro/handlerTypes.hpp>

#include <avtime/AvClock.hpp>

#include <condition_variable>
#include <functional>
#include <map>
#include <mutex>
#include <queue>
#include <string>
#include <thread>
#include <vector>

namespace av::repro {

class Scheduler
{
public:
    Scheduler() = default;
    /// Destructor, which makes sure the scheduling thread has quit.
    ~Scheduler();
    // Don't allow moving or copying.
    Scheduler(const Scheduler&) = delete;
    Scheduler(Scheduler&&) = delete;

    Scheduler& operator=(const Scheduler&) = delete;
    Scheduler& operator=(Scheduler&&) = delete;

    /// Initialize, and start the scheduling thread unless frozen.
    void start(const Enqueuer& enqueuer, bool frozen);
    /// Stop the scheduling thread.
    void quit();

    /// Get the current time, in duration since the epoch.
    Duration now() const;
    /// In "frozen" mode, advance the scheduler to this point in time.
    void advance(Duration now);

    /// Set the handler function for a given name, and optionally make it a periodic function.
    /// @note Only one handler can be set per name.
    void setHandler(const std::string& name, const TimeoutHandler& handler, Duration repeat = Duration{0});

    /// Schedule the given timeout handler to run, unless a later schedule() call supersedes it.
    /// @param name: The name of the timeout (registered with a previous call to setHandler()).
    /// @param timeout: How long to wait before running the handler (or 0 to run immediately).
    /// @param repeat: If nonzero, the handler will automatically be rescheduled with this timeout after it runs.
    void schedule(const std::string& name, Duration timeout = Duration{0}, Duration repeat = Duration{0});
    /// Cancel a previously scheduled timeout handler.
    /// @note This is not necessary if you are scheduling a new timeout; the old one is cancelled automatically.
    void unschedule(const std::string& name);

    /// For replay mode, call the given timeout handler.
    void replay(const std::string& name);
    /// Get the special string used to represent the scheduler as an event origin.
    static const std::string& origin();

private:
    /// Information we must store for each timeout.
    /// @note Times are stored as duration since epoch since different clocks may be in use.
    struct Timeout
    {
        std::string name;
        TimeoutHandler handler;
        Duration time{0};
        Duration repeat{0};
        uint32_t count{0};
    };
    /// An entry in the schedule queue.
    struct ScheduledTimeout
    {
        Duration time{0};
        Timeout* current{nullptr};

        /// Constructor, setting the time based on the timeout object.
        ScheduledTimeout(Timeout& timeout);
        /// Sort operator for the priority queue.
        bool operator<(const ScheduledTimeout& other) const;
        /// Check if this timeout has been cancelled.
        operator bool() const;
    };

private:
    /// Start the clock, converting durations in the queue into time points.
    void start(Duration now);
    /// Entry point for the scheduler thread.
    void loop();
    /// If the next item on the queue occurs before the given time, pop it and generate an event.
    /// @return True if an entry was popped (whether or not it generated an event).
    bool advanceOne(Duration now);
    /// Reschedule the given timeout.
    /// @note Call this from a context where the mutex is already locked.
    void reschedule(Timeout& entry, Duration time, Duration repeat);
    /// Read the clock, in a context where we can trust the mutex is locked.
    Duration protectedNow() const;

private:
    /// Function to be called to enqueue events.
    Enqueuer enqueuer_;
    /// Information for each timeout. Use a map so that pointers are never invalidated.
    std::map<std::string, Timeout> timeouts_;
    /// Priority queue of timeouts to run.
    /// @note Cancelled entries are not removed until they reach the front, so one timeout may appear several times.
    std::priority_queue<ScheduledTimeout> schedule_;
    /// In frozen mode, events are generated only when advance() is called.
    bool frozen_{false};
    /// Variable to set when the scheduler is quitting.
    bool interrupted_{false};
    /// The current time. If 0, the clock is not yet started.
    Duration now_{0};
    /// Scheduling thread, which only runs if we're not in "frozen" mode.
    std::thread thread_;
    /// Mutex protecting access to the queue.
    mutable std::mutex mutex_;
    /// Condition variable used for waiting and notification.
    std::condition_variable condition_;
};

} // namespace av::repro
