/// @copyright Copyright (c) 2022- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Function definitions for a lightweight comms framework.
#include <repro/commsLite.hpp>

#include <SimpleComms/UnixClient.h>
#include <task/TaskTime.hpp>

#if !defined(__aarch64__)
#include <DDSComms/DDSClient.hpp>
#endif

#include <cstdlib>
#include <memory>
#include <utility>

namespace av::comms {

bool init(const std::string& taskName, const std::string& commsSystem)
{
    // Don't use the SystemClock. Use AvClock instead.
    task::time::TaskTimeProvider::setTimeSource(task::time::detail::AvTimeProvider());

    // Create a Client of the proper type.
    auto handler = [](const std::string& topic, const std::string& blob, const std::pair<std::string, uint32_t>& key) {
        framework().receive(blob, topic, key.first);
    };
    std::shared_ptr<SimpleComms::Client> client;
    if (commsSystem == "SCS") {
        const auto scsPath = SimpleComms::UnixClient::getDefaultScsPath();
        const auto clientName = SimpleComms::UnixClient::makeUniqueClientName(taskName);
        client = std::make_shared<SimpleComms::UnixClient>(scsPath, clientName, handler);
    }
#if !defined(__aarch64__)
    else if (commsSystem == "DDS") {
        client = std::make_shared<DDSComms::DDSClient>(handler);
    }
#endif
    else {
        return false;
    }

    // Figure out what mode the framework is running in.
    // TODO: Figure out whether support for reproducible replay is needed.
    auto mode = av::repro::EventFramework::Mode::live;
    const char* clockTypeEnv = getenv("SYSTEM_CLOCK_TYPE");
    if (clockTypeEnv != nullptr && clockTypeEnv == std::string("external")) {
        mode = av::repro::EventFramework::Mode::frozen;
    }
    const char* reproducible = getenv("REPRODUCIBLE");
    if (reproducible != nullptr && reproducible == std::string("frozen")) {
        mode = av::repro::EventFramework::Mode::frozen;
    }
    if (mode == av::repro::EventFramework::Mode::frozen) {
        framework().declare("PlaybackClockParams", SimpleComms::ChannelType::Subscriber);
    }

    // Initialize the framework.
    framework().setClient(client);
    framework().init(mode);
    framework().start();

    return true;
}

repro::EventFramework& framework()
{
    static repro::EventFramework instance;
    return instance;
}

void shutdown()
{
    framework().quit();
    if (framework().client()) {
        framework().client()->disconnect();
    }
}

} // namespace av::comms
