# Client-Side Reproducible Framework

## Introduction

The classes in this folder implement “client-side” reproducibility: logging messages and events inside the framework for each task and allowing them to later be replayed in the same order to reproduce the original results. A task can register handler functions for particular input topics and scheduled events (periodic functions, timeouts, etc.); these functions are called by the framework in a single handler thread to avoid race conditions.

In keeping with the proposed design for the FASTER framework, message handlers receive a “batch” of messages from a single topic and origin, instead of being called with one message at a time. The only time a batch will contain more than one message is if the task is taking too long to process events. In that case, the user code can note that messages are getting batched and select a faster way of processing them. (For example, the LSN task might choose to update the fused point cloud with every merged point cloud in a batch, but only to apply the pipeline to the final state of the fused point cloud.)

The framework can operate in one of several modes, described in detail below.

### Live mode

This is the default operating mode. Serialized messages are received from the comms framework (SimpleComms or DDS) and sent to the appropriate `Deserializer` object. The deserialized message is then packaged in a callable `Event` object, which is added to the `EventQueue`. On a separate thread, events in the queue are processed in the order in which they were added, and the `EventQueue` saves a log of those events:

![](figures/comms-live-mode.svg)

Similarly, timeout events (periodic functions, step functions, watchdog timers, etc.) are produced by a Scheduler object based on the current system clock time. They are placed in the `EventQueue` along with the input message events:

![](figures/scheduler-live-mode.svg)

### Frozen mode

Frozen mode is for reprocessing offline using the log player. Instead of using the system clock the task takes the current time from the timestamp of each message as it is received. This means that the timestamps will be approximately the same as they were when the log was originally recorded, except that the communications latency will appear to be 0 (when processing each message, the task's clock will be set to the same time as when the message was sent).

In this mode, the `Scheduler` does not run on its own. Instead, its clock advances whenever a message arrives:

![](figures/comms-frozen-mode.svg)

Although the diagram above shows the `Scheduler` generating a single event, any number of events may be generated each time the clock advances.

### Replay mode

In replay mode a `Replayer` object takes the place of both the `Scheduler` and the `EventQueue`. It reads the log of the original event order and makes sure that events fire in that same order during replay. When it gets to a line of the log that represents an event from the `Scheduler`, it can replay that event immediately. But, events from the `Deserializer` can only be played back once all the messages that were originally included in a given batch have arrived. For example, an event log might look like this:

- Processed messages 1-3 from the “Sample” topic.
- Ran the executive function.
- Processed message 4 from the “Sample” topic.
- Processed messages 5-6 from the “Sample” topic.
- Ran the executive function.
- ...

The `Replayer` must wait until messages 1-3 have been received. It can then fire that message event, as well as the following executive event. Then it must wait for message 4 to arrive, and so on.

If the task sending a particular input originally started up before the task receiving that input, the receiver will not receive some of the messages. The Replayer must compare message IDs to determine which messages need to be ignored because they were not received during the original run:

![](figures/comms-replay-mode.svg)

Note that for reproducible replay to work properly, all the tasks should be started up before log playback begins. Otherwise a message that _was_ received in the original run may be lost during replay.

### Central mode

To continue supporting reproducible resim prior to switching over to DDS, if a task requests the “central” `clockType` and if the `REPRODUCIBLE` environment variable is not set, that task will continue to receive executive function triggers from SCS, bypassing the client-side framework entirely.

## Architecture

The key classes in the reproducible architecture are shown below.

![](figures/architecture-overview.svg)

As shown by the colors in the diagram, these classes can be split into three categories: scheduling, deserializing, and running events. To avoid circular dependencies and to simplify unit testing, none of the classes in these three sub-modules make use of the `EventFramework` class directly. Instead, if they need to call one of `EventFramework`’s functions they are given a `std::function<>` to call.

### Event info

The `EventInfo` base class provides the following information about a message or timeout event:

- **time**: when the event occurred. For messages, this is the timestamp in the header; for timeouts, it's the clock time when the timeout triggered. Note that to query the current time, task code should use `task::Task::getSystemTime()` or `framework_.frozenTime()`; this timestamp should only be used for checking latencies.
- **topic**: the name of the comms topic or the name of the timeout.
- **origin**: the task that sent a message, or “scheduler” if this is a timeout event.
- **id**: a rolling ID that counts the number of messages or timeouts sent from a single origin.

Two classes derive from EventInfo:

- `Message<Type>` stores a pointer to a message buffer.
- `Event` stores a function pointer to be called when an event is processed.

### Scheduling timeouts

The FASTER design includes the concepts of periodic functions (what we used to call the executive function), step functions (called when an “aligner” determines all inputs are ready), and watchdog functions (called if no input is received in a given topic in some time). Rather that treating those handlers separately, the client-side reproducibility framework groups them all into a generic category of “timeout handlers.”

To support these use cases, all the scheduler needs to track for each timeout “topic” is:

- The handler to be called when it triggers.
- The current time when it is scheduled to run.
- If the timeout should automatically repeat, the repeat period.
- A count of how many times the timeout has been scheduled.

The count is not technically needed, but since we assign “rolling IDs” to messages it's been added to the timeout records as well. In theory it might be helpful, for example, to be able to see at a glance how many times the executive function ran in a given events log.

**In “live” mode,** the `Scheduler` uses `AvClock` to get steady clock times relative to the Unix epoch. The `Scheduler` has its own scheduling thread that repeatedly sleeps until either a new timeout event is added or until it's time to fire the next event in the queue.

**In “frozen” mode,** the `Scheduler`’s clock advances only when the `EventFramework` receives a new message and calls `Scheduler::advance()` with the timestamp from that message. In that mode, the scheduling thread does not run.

**In “replay” mode,** the `Scheduler` is never even initialized; `Replayer` takes its place.

### Processing input messages

In the server-side `CallbackSet` architecture, messages were deserialized in the callback thread, the same thread used for all the “work” that the task performs. That is suboptimal, because it means the time taken for deserialization takes away from the time available to the task.

In the client-side framework, messages are instead deserialized as soon as they arrive. This also allows the `Deserializer` object to maintain a collection of “pinnable” `Message<Type>` objects for each topic, and also allows it to reuse those message buffers once the task code is no longer holding onto a reference to them.

When the handler is set for a given topic, optionally the `EventFramework` can be given a number of buffers to allocate for that topic. If so, all buffers are allocated at initialization time and if the deserializer runs out of free buffers:

- If possible, the oldest “pending” buffer (i.e. messages that have been received but not yet handled) will be discarded and recycled to use for the new message.
- If no buffers are “pending” (i.e. they are all pinned by the task code) the new input is discarded. Tasks should request enough buffers to avoid the risk of this happening.

The `EventFramework` uses a polymorphic container to store the deserializers for all the registered topics. The `DeserializerBase` interface defines an abstract function that receives a serialized “blob”. When that function is called, the `Deserializer` deserializes the message and then produces a generic `Event` object that is not dependent on the message type. So, the only place messages are typed is inside `Deserializer` and in the handler functions in the user's code.

### Running events

Both the `EventQueue` and the `Replayer` have a dedicated thread that runs the task’s event handlers. The `EventQueue` saves a .events log as it runs those events. The `Replayer` reads a previous .events file and ensures that the events happen in the same order, and in the same batches as the original run.

When a message event runs, it is possible that the `Deserializer` will decide to batch multiple messages. In order to determine the batch size, the `std::function<>` stored in the `Event` object returns a value, the ID of the last event in the batch. If the return value is 0, that means that no messages were processed because the `Event` refers to a message that was already processed in a previous batch. For example, suppose the `EventQueue` has built up the following backlog of messages from two different topics, **A** and **B**:

![](figures/batch-0.svg)

The `EventQueue` repeatedly pulls events off the right side of the queue and runs them. In this case, message **A₁** is handled, and it gets batched with messages **A₂** and **A₃**. Then, **A₁** is popped from the queue. The other events are still in the queue, because there's no way for `Deserializer` to pull them out. But, they're now no-ops: when they run, `Deserializer` will recognize that they represent a message that was processed in a previous batch and will ignore them:

![](figures/batch-1.svg)

Now, **A₂** gets popped from the queue and ignored, and the queue advances to **B₁**. Suppose two new messages from **A** arrived in the meantime. The queue might now look like this:

![](figures/batch-2.svg)

The message **B₁** from the front of the queue is popped and batched with **B₂**. Next in the queue is **A₃**, which will be ignored because it was already handled (even though there are now new messages from **A** in the queue that have not been handled yet). This order of processing ensures that we always handle the topic with the oldest pending message at a given time, so we don't have to worry that a topic with more frequent messages could completely “starve” another.

Right before each handler is called, the `EventQueue` “freezes” the current time, setting it equal to the system time (in live mode) or the message’s timestamp (in frozen mode). The timestamp is then logged. This ensures that during replay, the current time will be exactly the same as it was in the original run.

### Replaying events

Rather than allowing `Deserializer` to determine the batch size, the `Replayer` tracks which messages have been received that match a given event from the log and only schedules that event to run when all the messages are available. When it calls the handler, it gives it the `Deserializer` the ID of the last message it wants to include in the batch.

The `Replayer` has to handle several situations:

**Messages missing in the original run:** Suppose a task did not finish starting up until the task that publishes one of its inputs had already been up and running for a few seconds. In that case, some messages were lost at the start, and the `Replayer` has to ignore those same messages. To do this, we use the rolling message IDs, which are guaranteed to be unique and monotonically increasing in the original run. (The same rolling ID is used for all of a task’s outputs, so the IDs for one particular output topic may not be sequential.)

**Partially missing batches:** During reproducible resim, it is possible that the receiver task was started up too late to receive all the events it expected. In that case, there may be events for which it received some but not all of the expected messages, and the best `Replayer` can do is to call the handler with whatever messages _were_ received.

**Messages missing during replay:** Suppose a log was cut in a place where a first task does not receive the executive trigger that causes it to produce a message, but a second task is expecting to receive that message. The `Replayer` will notice that has happened when it receives the first message in that channel with an ID _later_ than the range it is waiting for. When that happens, it discards the unmatched row of the event log.

**Inputs not included in the log:** During replay, any input from a topic that does not appear in the events log will be ignored.
