# Lightweight Comms Framework

The `commsLite` framework ([commsLite.hpp](../commsLite.hpp)) provides access to comms without the overhead of `libtask`. By default it supports both SCS and DDS; on aarch64 only SCS is currently supported.

For examples of how to use the framework, see the [LiteSender](../../../../tasks/task/exampleTask/LiteSender.cpp) and [LiteReceiver](../../../../tasks/task/exampleTask/LiteReceiver.cpp) example tasks. To run those tasks, run the following commands in three different terminals, in this order:

```bash
    $ bazel-bin/platform/middleware/tasks/SimpleComms/SimpleComms/scs
    $ bazel-bin/platform/middleware/tasks/task/exampleTask/LiteReceiver
    $ bazel-bin/platform/middleware/tasks/task/exampleTask/LiteSender
```

## Initialization

The framework should be initialized once at the start of your program by passing it the name of your task and, optionally, which comms system to connect to (SCS or DDS). For example:

```c++
    av::comms::init("MyTask", "SCS");
```

At the end of your program, call the shutdown function:

```c++
    av::comms::shutdown();
```

## Inputs

Both polling and event-driven inputs (via callback functions) are supported. A `MessageQueue<>` object can be constructed for a given channel as follows, where the final argument gives the maximum number of buffers:

```c++
    task::MessageQueue<MyMessageType> queue("MyChannelName", av::comms::framework(), 4);
```

It's safe to construct the `MessageQueue<>` before you initialize comms. For example, if your task class has `MessageQueue<>`s as member variables, they can be initialized directly in the constructor's initializer list.

Messages can be received from the queue as follows:

```c++
    while (!queue.empty()) {
        processMessage(queue.pop());
    }
```

In many cases, rather than polling for messages you want to process them as soon as they arrive. In that case you can use callbacks instead. For example:

```c++
    av::repro::MessageHandler<MyMessageType> callback =
        [&](const av::repro::Inputs<MyMessageType>& inputs) {
            for (const std::shared_ptr<MyMessageType>& input : inputs) {
                processMessage(*input);
            }
        };
    av::comms::framework().setMessageHandler("MyChannelName", callback);
```

More often, the callback function will be a member of a class. For example:

```c++
void MyTask::handleMyMessage(const av::repro::Inputs<MyMessageType>& inputs)
{
    //...
}
```

In this case, there is a shortcut for registering the callback:

```c++
    av::comms::framework().setMessageHandler("MyChannelName", &myTask, &MyTask::handleMyMessage);
```

The callback functions will all be called in the same thread, so as long as you do all your operations within that thread you don't have to worry about mutex protection. Executives and other scheduled functions are also called in that same thread.

## Outputs

To publish outputs, construct a `MessageOutput<>` object:

```c++
    task::MessageOutput<MyMessageType> output("MyChannelName", av::comms::framework());
```

It's safe to construct the `MessageOutput<>` before you initialize comms. If you call `publish()` on it before comms are initialized, the message will be ignored.

To send a message, use the `publish()` method:

```c++
    output.publish(myMessage);
```

## Scheduled events

As described in the [README](../README.md) for the reproducible event framework, the `EventFramework` class can also be used to schedule executive functions, watchdog timers, etc. These will all run in the same thread as the message callbacks, so again any work done in them is perfectly thread-safe.

For example:
```c++
    framework_.setTimeoutHandler("!periodic", this, &MyTask::executive, 50ms);
```

## Example

Consider a simple class that consumes two inputs and produces one output. Every time it receives an input in `"ChannelA"`, it performs some calculations using the most recent input from `"ChannelB"` and then produces a result in `"ChannelC"`. The class definition might look like this:

```c++
class MyTask
{
public:
    MyTask();
    ~MyTask();

    void handleMessageA(const av::repro::Inputs<MessageA>& inputs);

private:
    MessageB bCached_;
    task::MessageQueue<MessageB> bInput_;
    task::MessageOutput<MessageC> cOutput_;
};
```

The class implementation might look like this:

```c++
MyTask::MyTask()
    : bInput("ChannelB", av::comms::framework(), 1)
    , cOutput("ChannelC", av::comms::framework())
{
    av::comms::framework().setMessageHandler("ChannelA", this, &MyTask::handleMessageA);
    av::comms::init("MyTask");
}

MyTask::~MyTask()
{
    av::comms::shutdown();
}

void MyTask::handleMessageA(const av::repro::Inputs<MessageA>& inputs)
{
    // First, get the most recent data from channel B, if any has arrived.
    // The queue has a max size of 1, so it only contains the latest message.
    if (!bInput.empty()) {
        bCached_ = bInput.pop();
    }
    // Now, process each input message.
    for (const std::shared_ptr<MyMessageType>& it : inputs) {
        MessageC result = doSomeSortOfProcessing(*it, bCached_);
        cOutput_.publish(result);
    }
}
```