/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Tests for the EventFramework class.
#include <repro/EventFramework.hpp>

#include <repro/handlerTypes.hpp>

#include <ODatum/ODatum.h>
#include <SimpleComms/test/MockUnixClient.h>
#include <avdb/SQLReader.hpp>
#include <avtime/ConversionsTimespec.hpp>
#include <interfaces/Clock/Playback/PlaybackClockParams.h>
#include <interfaces/SampleMessage/SampleMessage.h>
#include <interfaces/baseTypes/Header.h>

#include <gtest/gtest.h>

#include <chrono>
#include <condition_variable>
#include <mutex>
#include <vector>

using namespace av;
using namespace std::chrono_literals;

namespace {

// String to use for the origin of all the test messages.
const std::string ORIGIN = "tester";
// String to use for the topic.
const std::string TOPIC = "sample";
// String to use for playback clock messages.
const std::string CLOCK_TOPIC = "PlaybackClockParams";
// String to use for the periodic function.
const std::string PERIODIC = "!periodic";
// Typedef for the handler functions.
using Handler = repro::MessageHandler<msg::SampleMessage>;

// Helper function to construct and serialize a SampleMessage.
std::string makeSample(size_t value, Duration time = Duration{0})
{
    interfaces::baseTypes::InterfaceHeader header;
    interfaces::baseTypes::setHeaderVersion(header, OTTO_INTERFACE_HEADER_VERSION);
    if (time == Duration{0}) {
        time = std::chrono::system_clock::now().time_since_epoch();
    }
    interfaces::baseTypes::setSendTimeStamp(header, timespecFromDuration<otimeSpec>(time));
    interfaces::baseTypes::setHeaderPublisherId(header, 0);
    interfaces::baseTypes::setHeaderMessageCounter(header, static_cast<uint32_t>(value + 1));

    std::string blob;
    msg::SampleMessage message(static_cast<int>(value));
    aptv::serializeODatumMessage(std::string(reinterpret_cast<char*>(&header), sizeof(header)), blob, message);
    return blob;
}

// Helper function to construct a playback clock message.
std::string makeClock(Duration time)
{
    interfaces::baseTypes::InterfaceHeader header;
    interfaces::baseTypes::setHeaderVersion(header, OTTO_INTERFACE_HEADER_VERSION);

    std::string blob;
    PlaybackClockParams message{};
    message.baseTime = timespecFromDuration<otimeSpec>(time);
    aptv::serializeODatumMessage(std::string(reinterpret_cast<char*>(&header), sizeof(header)), blob, message);
    return blob;
}

// Get the path for the log file.
std::string logPath()
{
    const char* tmpDir = std::getenv("TEST_TMPDIR");
    return (tmpDir ? tmpDir : "/tmp") + std::string("/EventFrameworkTest.events");
}

// Helper object to receive messages.
class Receiver
{
public:
    // Receive messages.
    void handle(const repro::Inputs<msg::SampleMessage>& inputs);

    // Wait until the given number of messages are received.
    // Get a copy of the accumulated buffer.
    std::vector<msg::SampleMessage> wait(size_t count = 1, std::chrono::nanoseconds timeout = 1s);

private:
    std::mutex mutex_;
    std::condition_variable condition_;
    std::vector<msg::SampleMessage> messages_;
};
void Receiver::handle(const repro::Inputs<msg::SampleMessage>& inputs)
{
    {
        std::scoped_lock lock(mutex_);
        for (const repro::Message<msg::SampleMessage>& it : inputs) {
            messages_.push_back(*it);
            EXPECT_EQ(it.topic(), TOPIC);
            EXPECT_EQ(it.origin(), ORIGIN);
            EXPECT_EQ(it.id(), messages_.size());
        }
    }
    condition_.notify_all();
}
std::vector<msg::SampleMessage> Receiver::wait(size_t count, std::chrono::nanoseconds timeout)
{
    // Note: with GCC 10, wait_for() reports a spurious TSAN issue:
    // https://github.com/google/sanitizers/issues/1259
    std::unique_lock lock(mutex_);
    condition_.wait_for(lock, timeout, [count, this] { return messages_.size() == count; });
    return std::move(messages_);
}

// Helper object to count timeout events.
class Counter
{
public:
    // Use a reference to the framework to query the current time.
    Counter(const repro::EventFramework& framework);

    // Receive a timeout event.
    void handle();

    // Wait until the given number of events are received.
    size_t wait(size_t count = 1, std::chrono::nanoseconds timeout = 1s);

    // Get the timestamps for each call to the handler.
    const std::vector<std::chrono::nanoseconds>& times() const;

private:
    std::mutex mutex_;
    std::condition_variable condition_;
    const repro::EventFramework& framework_;
    size_t count_{0};
    std::vector<std::chrono::nanoseconds> times_;
};
Counter::Counter(const repro::EventFramework& framework)
    : framework_(framework)
{}
void Counter::handle()
{
    {
        std::scoped_lock lock(mutex_);
        ++count_;
        times_.push_back(framework_.frozenTime());
    }
    condition_.notify_all();
}
size_t Counter::wait(size_t count, std::chrono::nanoseconds timeout)
{
    std::unique_lock lock(mutex_);
    condition_.wait_for(lock, timeout, [count, this] { return count_ >= count; });
    return count_;
}
const std::vector<std::chrono::nanoseconds>& Counter::times() const
{
    return times_;
}

} // namespace

class EventFramework : public testing::Test
{
public:
    static void SetUpTestSuite()
    {
        // Make sure avlog is running and printing to the console.
        avlog::AvLogOptions logOptions;
        logOptions.logToConsole = true;
        logOptions.globalLogLevel = avlog::LogLevel::debug;
        avlog::startup(logOptions);
    }
};

TEST(EventFramework, receive)
{
    auto client = std::make_shared<SimpleComms::MockUnixClient>();

    EXPECT_CALL(*client, connect());
    repro::EventFramework framework;
    std::string path = logPath();
    framework.setClient(client);
    framework.init(repro::EventFramework::Mode::frozen, path);

    EXPECT_CALL(*client, declare(TOPIC, SimpleComms::ChannelType::Subscriber, testing::_));
    Receiver receiver;
    Handler handler = std::bind(&Receiver::handle, &receiver, std::placeholders::_1);
    framework.setMessageHandler(TOPIC, handler);
    framework.start();

    EXPECT_EQ(framework.mode(), repro::EventFramework::Mode::frozen);

    const size_t REPS = 20;
    for (size_t i = 0; i < REPS; ++i) {
        framework.receive(makeSample(i), TOPIC, ORIGIN);
    }
    std::vector<msg::SampleMessage> messages = receiver.wait(REPS);

    // Make sure to shut down the framework before destructing the receiver.
    framework.quit();

    // Verify the values.
    EXPECT_EQ(messages.size(), REPS);
    int value = 0;
    for (const msg::SampleMessage& it : messages) {
        EXPECT_EQ(it.value, value++);
    }

    // Verify the log.
    using Log = db::SQLReader<int64_t, uint32_t, std::string_view, std::string_view, uint32_t, uint32_t>;
    Log log(path);
    ASSERT_TRUE(log);
    ASSERT_EQ(log.labels().size(), 6);
    size_t total = 0;
    size_t count = 0;
    for (; log; ++count) {
        auto [time, outID, topic, origin, firstID, lastID] = log.read();
        EXPECT_EQ(time, 0);
        EXPECT_EQ(outID, 1);
        EXPECT_EQ(topic, TOPIC);
        EXPECT_EQ(origin, ORIGIN);
        EXPECT_GT(firstID, 0);
        EXPECT_LE(firstID, lastID);
        total += (lastID - firstID + 1);
    }
    EXPECT_EQ(total, REPS);
    std::cout << "Received " << total << " messages in " << count << " batches." << std::endl;
}

TEST(EventFramework, periodic)
{
    repro::EventFramework framework;
    framework.init(repro::EventFramework::Mode::live);
    Counter counter(framework);
    auto start = std::chrono::steady_clock::now();
    framework.setTimeoutHandler(PERIODIC, std::bind(&Counter::handle, &counter), std::chrono::milliseconds(10));
    framework.start();

    EXPECT_EQ(framework.mode(), repro::EventFramework::Mode::live);

    static const size_t REPS = 20;
    ASSERT_EQ(counter.wait(REPS + 1), REPS + 1);
    auto duration = std::chrono::nanoseconds(std::chrono::steady_clock::now() - start);
    double frequency = (REPS * 1e9) / static_cast<double>(duration.count());
    std::cout << "Requested 100 Hz, ran at " << frequency << " Hz." << std::endl;

    // Make sure to shut down the framework before destructing the receiver.
    framework.quit();

    // It's an error if the scheduler ran too quickly. But, there is a small delay
    // in between when I started timing and when the handler ran for the first time.
    // That may reduce the frequency, but should never increase it above the max.
    EXPECT_LE(frequency, 100.);
    // We've seen cases where the frequency is as low as 50 Hz when running sanitizers,
    // so we can't have a test for a minimum frequency.
}

TEST(EventFramework, frozen)
{
    repro::EventFramework framework;
    framework.init(repro::EventFramework::Mode::frozen);
    Counter counter(framework);
    Receiver receiver;
    framework.setTimeoutHandler(PERIODIC, std::bind(&Counter::handle, &counter), 10ms);
    // If the framework hasn't been started yet, clock messages should be ignored.
    framework.receive(makeClock(50ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(1, 50ms), 0);

    framework.start();

    EXPECT_EQ(framework.mode(), repro::EventFramework::Mode::frozen);

    // In response to this first message, the periodic function should run once (at t = 10).
    framework.receive(makeClock(10ms), CLOCK_TOPIC, ORIGIN);
    // Wait for two messages, and make sure we actually only get one.
    ASSERT_EQ(counter.wait(2, 50ms), 1);

    // If messages are coming in slow enough, the framework has time to keep up.
    framework.receive(makeClock(25ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(2, 50ms), 2);
    framework.receive(makeClock(29ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(2, 50ms), 2);
    framework.receive(makeClock(35ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(3, 50ms), 3);
    // If they come in too quicky or with too big a gap, the periodic function should
    // reschedule at the current time.
    framework.receive(makeClock(60ms), CLOCK_TOPIC, ORIGIN);
    // Wait for six messages total, and make sure we actually only get five.
    ASSERT_EQ(counter.wait(6, 50ms), 5);

    // The handler is always called with the current timestamp according to the playback clock.
    // We wanted to trigger at 10, 20, and 30 ms, but the executive won't trigger until the clock
    // reaches or passes that point, and it does so at 25 and 35 ms. Then we have an executive
    // scheduled for 40 ms, so when the clock jumps to 60 ms it runs twice.
    using namespace std::chrono_literals;
    const std::vector<std::chrono::nanoseconds> expected = {10ms, 25ms, 35ms, 60ms, 60ms};
    EXPECT_EQ(counter.times(), expected);

    // Make sure to shut down the framework before destructing the receiver.
    framework.quit();
}

TEST(EventFramework, schedule)
{
    repro::EventFramework framework;
    framework.init(repro::EventFramework::Mode::frozen);
    Counter counter(framework);
    Receiver receiver;
    framework.setTimeoutHandler(PERIODIC, std::bind(&Counter::handle, &counter));
    framework.start();

    EXPECT_EQ(framework.mode(), repro::EventFramework::Mode::frozen);

    framework.schedule(PERIODIC, 15ms);
    // The very first message received starts the clock, so the event will actually be
    // scheduled at 10 + 15 = 25 milliseconds.
    framework.receive(makeClock(10ms), CLOCK_TOPIC, ORIGIN);
    // Make sure we don't receive a message yet.
    ASSERT_EQ(counter.wait(1, 50ms), 0);
    // Reschedule the timeout to 10 + 25 = 35 milliseconds.
    framework.schedule(PERIODIC, 25ms);
    // Make sure we still don't receive a message yet.
    ASSERT_EQ(counter.wait(1, 50ms), 0);
    // Now, advance the scheduler and make sure we receive one, but only one message.
    framework.receive(makeClock(40ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(2, 50ms), 1);

    // Schedule another timeout at 40 + 15 = 55 milliseconds, then cancel it.
    framework.schedule(PERIODIC, 15ms);
    framework.unschedule(PERIODIC);
    framework.receive(makeClock(60ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(2, 50ms), 1);

    // Check that the handler was called with the right timestamp each time.
    using namespace std::chrono_literals;
    const std::vector<std::chrono::nanoseconds> expected = {40ms};
    EXPECT_EQ(counter.times(), expected);

    // Make sure to shut down the framework before destructing the receiver.
    framework.quit();
}

TEST(EventFramework, rewind)
{
    repro::EventFramework framework;
    framework.init(repro::EventFramework::Mode::frozen);
    Counter counter(framework);
    Receiver receiver;
    framework.setTimeoutHandler(PERIODIC, std::bind(&Counter::handle, &counter), 10ms);
    framework.start();

    EXPECT_EQ(framework.mode(), repro::EventFramework::Mode::frozen);

    // Start "log playback" at 2 seconds.
    framework.receive(makeClock(2000ms), CLOCK_TOPIC, ORIGIN);
    // Wait for two messages, and make sure we actually only get one.
    ASSERT_EQ(counter.wait(2, 50ms), 1);
    // Advance by 25 ms, which should trigger the scheduler two more times.
    framework.receive(makeClock(2025ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(4, 50ms), 3);

    // "Rewind" to an earlier time. This should not trigger any scheduled events.
    // The periodic function was scheduled to run in 10ms, so now it's scheduled at 1040ms.
    framework.receive(makeClock(1030ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(4, 50ms), 3);

    // Step the clock forward to trigger more events.
    // Again, the scheduler will have to make up time by skipping to the current time.
    framework.receive(makeClock(1055ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(6, 50ms), 5);

    // "Fast forward" to a later time. This should not trigger any scheduled events.
    // Again, the next event is rescheduled to 10ms in the future, or 3070ms.
    framework.receive(makeClock(3060ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(6, 50ms), 5);

    // Step the clock forward to trigger more events.
    framework.receive(makeClock(3085ms), CLOCK_TOPIC, ORIGIN);
    ASSERT_EQ(counter.wait(8, 50ms), 7);

    // Check that the handler was called with the right timestamp each time.
    const std::vector<std::chrono::nanoseconds> expected = {2000ms, 2025ms, 2025ms, 1055ms, 1055ms, 3085ms, 3085ms};
    EXPECT_EQ(counter.times(), expected);

    // Make sure to shut down the framework before destructing the receiver.
    framework.quit();
}
