/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines unit tests for BackupRecoveryHandler
 */
#include <Clock/SystemClock.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <memory>

#include "BackupRecoveryTest.h"
#include "FakeIDbObject.hpp"
#include <task/BackupRecoveryHandler.h>

using namespace ::testing;
using namespace task;
using namespace task::TestUtils;

namespace BRHTest {
const std::string HostName = "127.0.0.1";

const std::string TaskBaseName = "TestTask";
const std::string PrimaryTaskName = TaskBaseName + "_" + "primary";
const std::string BackupTaskName = TaskBaseName + "_" + "backup";

const std::string OutputIntfcName = "SampleOutput";
const std::string InputOutputIntfcName = "SampleInputOutput";

const std::string OutputIntfcNameFull = "WriteOnlyInterfaces." + OutputIntfcName;
const std::string InputOutputIntfcNameFull = "ReadWriteInterfaces." + InputOutputIntfcName;

InterfaceParser makeInterfaceConfigParser(const std::string& configString)
{
    InterfaceParser parser;
    EXPECT_TRUE(parser.parseString(configString));
    return parser;
}

std::list<ConfigSection> makeWriteOnlyInterfacesConfigSection()
{
    std::string configString = "WriteOnlyInterfaces = { " + OutputIntfcName;
    configString += " = { interface =  \"" + OutputIntfcName + "\"";
    configString += R"(; implementation = "Channel";
            channel =
            {
                type = "SimpleComms";
                queueSize = 100;
                name = "SampleO";
            };
        };
    };
    )";
    return makeInterfaceConfigParser(configString).getAllWriteOnlyInterfaces();
}

std::list<ConfigSection> makeReadWriteInterfacesConfigSection()
{

    std::string configString = "ReadWriteInterfaces = { " + InputOutputIntfcName;
    configString += " = { interface =  \"" + InputOutputIntfcName + "\"";
    configString += R"(; implementation = "Channel";
            channel =
            {
                type = "SimpleComms";
                queueSize = 100;
                name = "SampleIO";
            };
        };
    };
    )";
    return makeInterfaceConfigParser(configString).getAllReadWriteInterfaces();
}

/**
 * \brief base class fixture for tesitng BackupRecoveryHandler
 */
class BackupRecoveryHandlerBasicTest : public ::testing::Test
{
protected:
    std::shared_ptr<MockSaferCommandInput> input_{nullptr};
    std::shared_ptr<MockSaferResponseOutput> output_{nullptr};
};

/**
 * \brief primary class fixture for tesitng BackupRecoveryHandler
 *
 * \details the convention for this test is that a Write interface is stored
 *   in the raw pointer interfaces map, and a Read/Write pointer is stored in
 *   the managed, shared_ptr map.
 */
class BackupRecoveryHandlerTest : public BackupRecoveryHandlerBasicTest
{
public:
    //! \brief set up tests
    void SetUp() override
    {
        if (!SystemClock::isInitialized()) {
            SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeSteady);
        }

        ASSERT_TRUE(SystemClock::isInitialized());
        input_ = std::make_shared<MockSaferCommandInput>();
        output_ = std::make_shared<MockSaferResponseOutput>();
    }

    //! \brief clean up the tests
    void TearDown() override
    {
        BackupRecoveryHandlerBasicTest::TearDown();
        for (auto intf : rawInterfaces_) {
            delete intf.second;
        }
    }

protected:
    /**
     * \brief Initialize the the interfaces and respective configs with the provided access
     *
     * \details For the sake of this test, we put the Write interface in our
     *   raw pointer map, and the Read/Write interface in our shared_ptr map.
     */
    void initInterfaces(task::Access wAcc, task::Access rwAcc)
    {
        wIntfcCs_ = makeWriteOnlyInterfacesConfigSection();
        rwIntfcCs_ = makeReadWriteInterfacesConfigSection();
        rawInterfaces_[OutputIntfcNameFull] = new FakeIDbObject(wAcc);
        smartInterfaces_[InputOutputIntfcNameFull] = std::shared_ptr<IDbObject>(new FakeIDbObject(rwAcc));
    }

    //! get the raw interface by name.
    FakeIDbObject* getRawInterface(std::string interfaceName)
    {
        return dynamic_cast<FakeIDbObject*>(rawInterfaces_.at(interfaceName));
    }

    //! get the managed interface by name.
    std::shared_ptr<FakeIDbObject> getManagedInterface(std::string interfaceName)
    {
        return std::dynamic_pointer_cast<FakeIDbObject>(smartInterfaces_.at(interfaceName));
    }

    std::map<std::string, IDbObject*> rawInterfaces_{};
    std::map<std::string, std::shared_ptr<IDbObject>> smartInterfaces_{};
    BackupRecoveryParameter handlingParams_{0, 0, 0};
    av::msg::SaferCommand mockCmd{};
    ConfigSectionList wIntfcCs_{};
    ConfigSectionList rwIntfcCs_{};
};

} // namespace BRHTest

using namespace BRHTest;

TEST_F(BackupRecoveryHandlerBasicTest, CanInstantiateAndSend)
{
    input_ = std::make_shared<MockSaferCommandInput>();
    output_ = std::make_shared<MockSaferResponseOutput>();

    BackupRecoveryHandler brh{input_, output_, TaskBaseName, HostName};
    EXPECT_TRUE(brh.initialize());
    EXPECT_CALL(*output_, publishV(_, _)).Times(Exactly(1)).WillOnce(Return(true));
    EXPECT_TRUE(brh.sendSaferResponse("A message"));
}

TEST_F(BackupRecoveryHandlerBasicTest, InvalidInstantiation_InputNull)
{
    input_ = nullptr;
    output_ = std::make_shared<MockSaferResponseOutput>();

    BackupRecoveryHandler brh{input_, output_, TaskBaseName, HostName};

    EXPECT_FALSE(brh.initialize());
    // can still send without an input
    EXPECT_CALL(*output_, publishV(_, _)).Times(Exactly(1)).WillOnce(Return(true));
    EXPECT_TRUE(brh.sendSaferResponse("A message"));
}

TEST_F(BackupRecoveryHandlerBasicTest, InvalidInstantiation_OutputNull)
{
    input_ = std::make_shared<MockSaferCommandInput>();
    output_ = nullptr;

    BackupRecoveryHandler brh{input_, output_, TaskBaseName, HostName};
    EXPECT_FALSE(brh.initialize());
    EXPECT_FALSE(brh.sendSaferResponse("A message"));
}

TEST_F(BackupRecoveryHandlerBasicTest, InvalidInstantiation_InputOutputNull)
{
    BackupRecoveryHandler brh{nullptr, nullptr, TaskBaseName, HostName};
    EXPECT_FALSE(brh.initialize());
    EXPECT_FALSE(brh.sendSaferResponse("A message"));
}

TEST_F(BackupRecoveryHandlerTest, HandleRecovery_HungTask_IAmHotBackup)
{
    // mock msg from safer indicating that primary task is hung
    mockCmd.messageType_ = av::msg::MessageId::TASKHUNG;
    mockCmd.taskName_ = PrimaryTaskName;
    EXPECT_CALL(*input_, get(_, _)).WillOnce(DoAll(SetArgReferee<0>(mockCmd), Return(true))).WillOnce(Return(false));

    handlingParams_.numReplicas = 1; // This task is backup
    handlingParams_.coldStandby = 0; // this task is hot

    initInterfaces(task::Access::Standby, task::Access::Read);
    BackupRecoveryHandler brh{input_, output_, BackupTaskName, HostName};

    EXPECT_TRUE(brh.handleBackupRecovery(wIntfcCs_, rwIntfcCs_, handlingParams_, rawInterfaces_, smartInterfaces_));

    // should have been set now to primary
    EXPECT_EQ(handlingParams_.numReplicas, 0);
    // should still be hot
    EXPECT_EQ(handlingParams_.coldStandby, 0);

    // my interfaces should now be set to their full access levels
    EXPECT_EQ(getRawInterface(OutputIntfcNameFull)->getAccessMode(), task::Access::Write);
    EXPECT_EQ(getManagedInterface(InputOutputIntfcNameFull)->getAccessMode(), task::Access::ReadWrite);
}

TEST_F(BackupRecoveryHandlerTest, HandleRecovery_HungTask_IAmColdBackup)
{
    // mock msg from safer indicating that primary task is hung
    mockCmd.messageType_ = av::msg::MessageId::TASKHUNG;
    mockCmd.taskName_ = PrimaryTaskName;
    EXPECT_CALL(*input_, get(_, _)).WillOnce(DoAll(SetArgReferee<0>(mockCmd), Return(true))).WillOnce(Return(false));

    handlingParams_.numReplicas = 1; // This task is backup
    handlingParams_.coldStandby = 1; // this task is cold

    initInterfaces(task::Access::Standby, task::Access::Read);

    BackupRecoveryHandler brh{input_, output_, BackupTaskName, HostName};

    EXPECT_TRUE(brh.handleBackupRecovery(wIntfcCs_, rwIntfcCs_, handlingParams_, rawInterfaces_, smartInterfaces_));

    // should have been set now to primary
    EXPECT_EQ(handlingParams_.numReplicas, 0);
    // should have turned me to hot
    EXPECT_EQ(handlingParams_.coldStandby, 0);

    // my interfaces should be set to their full access levels
    EXPECT_EQ(getRawInterface(OutputIntfcNameFull)->getAccessMode(), task::Access::Write);
    EXPECT_EQ(getManagedInterface(InputOutputIntfcNameFull)->getAccessMode(), task::Access::ReadWrite);
}

TEST_F(BackupRecoveryHandlerTest, HandleRecovery_HungTask_NoRecovNeeded_WrongNameConvention)
{
    // mock msg from safer indicating that the primary task is hung, but name it as a backup task convention
    mockCmd.messageType_ = av::msg::MessageId::TASKHUNG;
    mockCmd.taskName_ = BackupTaskName;
    EXPECT_CALL(*input_, get(_, _)).WillOnce(DoAll(SetArgReferee<0>(mockCmd), Return(true))).WillOnce(Return(false));

    handlingParams_.numReplicas = 1; // This task is functioning as backup task

    initInterfaces(task::Access::Standby, task::Access::Read);

    BackupRecoveryHandler brh{input_, output_, PrimaryTaskName, HostName};

    EXPECT_TRUE(brh.handleBackupRecovery(wIntfcCs_, rwIntfcCs_, handlingParams_, rawInterfaces_, smartInterfaces_));

    // should still be functioning as backup task
    EXPECT_EQ(handlingParams_.numReplicas, 1);
    // should be unchanged
    EXPECT_EQ(getRawInterface(OutputIntfcNameFull)->getAccessMode(), task::Access::Standby);
    EXPECT_EQ(getManagedInterface(InputOutputIntfcNameFull)->getAccessMode(), task::Access::Read);
}

TEST_F(BackupRecoveryHandlerTest, HandleRecovery_RejoinedTask_IAmBackup_NamedBackup)
{
    // mock msg from safer indicating that real primary task is now healthy
    mockCmd.messageType_ = av::msg::MessageId::TASKHEALTHY;
    mockCmd.taskName_ = PrimaryTaskName;
    EXPECT_CALL(*input_, get(_, _)).WillOnce(DoAll(SetArgReferee<0>(mockCmd), Return(true))).WillOnce(Return(false));

    handlingParams_.numReplicas = 0; // This task is currently functioning as primary task

    initInterfaces(task::Access::Write, task::Access::ReadWrite);

    BackupRecoveryHandler brh{input_, output_, BackupTaskName, HostName};

    EXPECT_TRUE(brh.handleBackupRecovery(wIntfcCs_, rwIntfcCs_, handlingParams_, rawInterfaces_, smartInterfaces_));

    // should have been set now to function as backup
    EXPECT_EQ(handlingParams_.numReplicas, 1);
    // my interfaces should now be restricted
    EXPECT_EQ(getRawInterface(OutputIntfcNameFull)->getAccessMode(), task::Access::Standby);
    EXPECT_EQ(getManagedInterface(InputOutputIntfcNameFull)->getAccessMode(), task::Access::Standby);
}

TEST_F(BackupRecoveryHandlerTest, HandleRecovery_UnexpectedSaferCmd)
{
    mockCmd.messageType_ = av::msg::MessageId::TASKHEALTHY;
    mockCmd.taskName_ = TaskBaseName;
    // mock an unexpected safer command
    mockCmd.messageType_ = av::msg::MessageId::PAUSE;
    EXPECT_CALL(*input_, get(_, _)).WillOnce(DoAll(SetArgReferee<0>(mockCmd), Return(true))).WillOnce(Return(false));
    handlingParams_.numReplicas = 1; // I am backup

    BackupRecoveryHandler brh{input_, output_, TaskBaseName, HostName};

    EXPECT_TRUE(brh.handleBackupRecovery(wIntfcCs_, rwIntfcCs_, handlingParams_, rawInterfaces_, smartInterfaces_));
    // should be unchanged
    EXPECT_EQ(handlingParams_.numReplicas, 1);
    EXPECT_EQ(handlingParams_.coldStandby, 0);
}
