/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief unit tests for ClosingSignalHandler.cpp
 */

#include "ClosingSignalHandler.h"
#include <chrono>
#include <gtest/gtest.h>

using namespace std::chrono_literals;
using namespace task;

namespace CSHTesting {

//! Default signals to use throughout testing
const std::vector<int> defClosingSignals{SIGINT, SIGUSR1, SIGUSR2};
constexpr int illegalClosingSignal{SIGSTOP};

//! Period if time to wait for signal handling thread to receive
//! signal and update status.  regular signals do not queue, and
//! handling thread will need some time to receive signal and set exit
//! flag.
constexpr auto sig_recept_buftime{100ms};

// The maximum time to wait() for csh to decide it should exit.
constexpr uint64_t csh_wait_timeout{100};

bool isSigMember(sigset_t& set, int sig)
{
    return sigismember(&set, sig);
}

bool isSigBlocked(int sig)
{
    sigset_t set;
    sigemptyset(&set);
    sigprocmask(SIG_BLOCK, nullptr, &set);
    return isSigMember(set, sig);
}

void blockSigs(const std::vector<int>& sigs)
{
    sigset_t set;
    ASSERT_EQ(sigemptyset(&set), 0);

    for (const auto sig : sigs) {
        ASSERT_EQ(sigaddset(&set, sig), 0);
    }

    ASSERT_EQ(pthread_sigmask(SIG_BLOCK, &set, NULL), 0);
}

void clearSigMasks(void)
{
    sigset_t mask;
    ASSERT_EQ(sigemptyset(&mask), 0);
    ASSERT_EQ(pthread_sigmask(SIG_SETMASK, &mask, nullptr), 0);
}

//! @brief Test Fixture for ClosingSignalHandler
class ClosingSignalHandlerTest : public ::testing::Test
{

protected:
    ClosingSignalHandler csh;

    void SetUp() override { clearSigMasks(); }

    void TearDown() override { clearSigMasks(); }
};

} // namespace CSHTesting

using namespace CSHTesting;

TEST_F(ClosingSignalHandlerTest, ShouldNotExitAfterConstruction)
{
    EXPECT_FALSE(csh.shouldExit());
}

TEST_F(ClosingSignalHandlerTest, ShouldNotExitAfterRegistration)
{
    EXPECT_TRUE(csh.setSignals(defClosingSignals));
    EXPECT_TRUE(csh.handleClosingSignals());
    EXPECT_FALSE(csh.shouldExit());
}

TEST_F(ClosingSignalHandlerTest, SignalsShowAsBlocked)
{
    EXPECT_TRUE(csh.setSignals(defClosingSignals));
    EXPECT_TRUE(csh.handleClosingSignals());

    for (const auto sig : defClosingSignals) {
        EXPECT_TRUE(isSigBlocked(sig));
    }
}

TEST_F(ClosingSignalHandlerTest, NoIllegalClosingSignal)
{
    std::vector<int> closingSignals{SIGINT};
    EXPECT_TRUE(csh.setSignals(closingSignals));
}

TEST_F(ClosingSignalHandlerTest, IllegalClosingSignalOnly)
{
    std::vector<int> closingSignals{illegalClosingSignal};
    EXPECT_FALSE(csh.setSignals(closingSignals));
}

TEST_F(ClosingSignalHandlerTest, IllegalClosingSignalAmongOthers)
{
    std::vector<int> closingSignals{illegalClosingSignal, SIGINT};
    EXPECT_FALSE(csh.setSignals(closingSignals));
}

TEST_F(ClosingSignalHandlerTest, ErrorOnInvalidSignalRegistration_InvalidSignalNumber)
{
    EXPECT_FALSE(csh.setSignals({SIGRTMAX + 1}));
    EXPECT_FALSE(csh.shouldExit());
}

TEST_F(ClosingSignalHandlerTest, ErrorOnInvalidSignalRegistration_InvalidSignalNumber_Recoverable)
{
    EXPECT_FALSE(csh.setSignals({SIGRTMAX + 1}));
    EXPECT_TRUE(csh.setSignals(defClosingSignals));
    EXPECT_TRUE(csh.handleClosingSignals());
    EXPECT_FALSE(csh.shouldExit());

    for (const auto sig : defClosingSignals) {
        EXPECT_TRUE(isSigBlocked(sig));
    }
}

TEST_F(ClosingSignalHandlerTest, ErrorOnInvalidSignalRegistration_EmptySignalsVec)
{
    EXPECT_FALSE(csh.setSignals({}));
    EXPECT_FALSE(csh.shouldExit());
}

TEST_F(ClosingSignalHandlerTest, ErrorOnInvalidSignalRegistration_EmptySignalsVec_Recoverable)
{
    EXPECT_FALSE(csh.setSignals({}));
    EXPECT_TRUE(csh.setSignals(defClosingSignals));
    EXPECT_TRUE(csh.handleClosingSignals());
    EXPECT_FALSE(csh.shouldExit());

    for (const auto sig : defClosingSignals) {
        EXPECT_TRUE(isSigBlocked(sig));
    }
}

TEST_F(ClosingSignalHandlerTest, UnableToHandleSigKill)
{
    const std::vector<int> closingSignals{SIGKILL, SIGINT};

    EXPECT_FALSE(csh.setSignals(closingSignals));
    EXPECT_FALSE(csh.shouldExit());

    for (const auto sig : closingSignals) {
        EXPECT_FALSE(isSigBlocked(sig));
    }
}

TEST_F(ClosingSignalHandlerTest, UnableToHandleSigStop)
{
    const std::vector<int> closingSignals{SIGINT, SIGSTOP};

    EXPECT_FALSE(csh.setSignals(closingSignals));
    EXPECT_FALSE(csh.shouldExit());

    for (const auto sig : closingSignals) {
        EXPECT_FALSE(isSigBlocked(sig));
    }
}

TEST_F(ClosingSignalHandlerTest, CombineSignalSets)
{
    EXPECT_TRUE(csh.setSignals({defClosingSignals[0], defClosingSignals[1]}));
    EXPECT_TRUE(csh.setSignals({defClosingSignals[2]}));

    EXPECT_TRUE(isSigBlocked(defClosingSignals[0]));
    EXPECT_TRUE(isSigBlocked(defClosingSignals[1]));
    EXPECT_TRUE(isSigBlocked(defClosingSignals[2]));
}

TEST_F(ClosingSignalHandlerTest, ErrorOnInvalidSignalStateInvariant)
{
    EXPECT_FALSE(csh.setSignals({SIGINT, SIGRTMAX + 1}));
    EXPECT_FALSE(csh.shouldExit());
    EXPECT_FALSE(isSigBlocked(SIGINT));
}

TEST_F(ClosingSignalHandlerTest, ShouldExitTrueAfterOneSignal)
{
    EXPECT_TRUE(csh.setSignals(defClosingSignals));
    EXPECT_TRUE(csh.handleClosingSignals());
    kill(getpid(), defClosingSignals[1]);
    csh.wait(csh_wait_timeout);
    EXPECT_TRUE(csh.shouldExit());
}

TEST_F(ClosingSignalHandlerTest, ShouldExitTrueAfterOneSignalNoTimeout)
{
    EXPECT_TRUE(csh.setSignals(defClosingSignals));
    EXPECT_TRUE(csh.handleClosingSignals());
    kill(getpid(), defClosingSignals[1]);
    csh.wait();
    EXPECT_TRUE(csh.shouldExit());
}

TEST_F(ClosingSignalHandlerTest, ShouldExitTrueAfterTwoSameSignals)
{
    EXPECT_TRUE(csh.setSignals(defClosingSignals));
    EXPECT_TRUE(csh.handleClosingSignals());
    kill(getpid(), defClosingSignals[1]);
    csh.wait(csh_wait_timeout);
    kill(getpid(), defClosingSignals[1]);
    std::this_thread::sleep_for(sig_recept_buftime);
    csh.wait(csh_wait_timeout);
    EXPECT_TRUE(csh.shouldExit());
}

TEST_F(ClosingSignalHandlerTest, ShouldExitTrueAfterTwoDifferentSignals)
{
    EXPECT_TRUE(csh.setSignals(defClosingSignals));
    EXPECT_TRUE(csh.handleClosingSignals());
    kill(getpid(), defClosingSignals[1]);
    csh.wait(csh_wait_timeout);
    kill(getpid(), defClosingSignals[2]);
    std::this_thread::sleep_for(sig_recept_buftime);
    csh.wait(csh_wait_timeout);
    EXPECT_TRUE(csh.shouldExit());
}

TEST(ClosingSignalHandlerDeathTest, DISABLED_ExitAfterThreeSameSigs)
{
    testing::FLAGS_gtest_death_test_style = "threadsafe";

    EXPECT_EXIT(
        {
            clearSigMasks();
            ClosingSignalHandler csh;
            EXPECT_TRUE(csh.setSignals(defClosingSignals));
            EXPECT_TRUE(csh.handleClosingSignals());
            kill(getpid(), SIGINT);
            csh.wait(csh_wait_timeout);
            EXPECT_TRUE(csh.shouldExit());
            kill(getpid(), SIGINT);
            csh.wait(csh_wait_timeout);
            EXPECT_TRUE(csh.shouldExit());
            kill(getpid(), SIGINT);
            csh.wait(csh_wait_timeout);
        },
        testing::ExitedWithCode(1),
        "");
}

TEST(ClosingSignalHandlerDeathTest, ExitAfterThreeDifferentSigs)
{
    testing::FLAGS_gtest_death_test_style = "threadsafe";

    EXPECT_EXIT(
        {
            clearSigMasks();
            ClosingSignalHandler csh;
            EXPECT_TRUE(csh.setSignals(defClosingSignals));
            EXPECT_TRUE(csh.handleClosingSignals());
            kill(getpid(), SIGINT);
            csh.wait(csh_wait_timeout);
            EXPECT_TRUE(csh.shouldExit());
            kill(getpid(), SIGUSR1);
            csh.wait(csh_wait_timeout);
            EXPECT_TRUE(csh.shouldExit());
            kill(getpid(), SIGUSR2);
            std::this_thread::sleep_for(sig_recept_buftime);
        },
        testing::ExitedWithCode(1),
        "");
}

TEST(ClosingSignalHandlerDeathTest, ExtraSigNoEffect)
{
    testing::FLAGS_gtest_death_test_style = "threadsafe";

    EXPECT_EXIT(
        {
            clearSigMasks();
            ClosingSignalHandler csh;

            // block 'decoy' sig to start so that it is a candidate on which
            // to be waited
            blockSigs({SIGUSR2});

            EXPECT_TRUE(csh.setSignals({SIGINT, SIGUSR1}));
            EXPECT_TRUE(csh.handleClosingSignals());

            // second call to handleClosingSignals is not applied, so handler
            // doesn't wait on SIGUSR2, and therefore should not indicate an exit.
            kill(getpid(), SIGUSR2);
            std::this_thread::sleep_for(sig_recept_buftime);
            EXPECT_FALSE(csh.shouldExit());

            // Ensure that the first call was, however, applied.
            kill(getpid(), SIGUSR1);
            csh.wait(csh_wait_timeout);
            EXPECT_TRUE(csh.shouldExit());

            // finish the test with a death
            kill(getpid(), SIGINT);
            csh.wait(csh_wait_timeout);
            kill(getpid(), SIGUSR1);
            std::this_thread::sleep_for(sig_recept_buftime);
        },
        testing::ExitedWithCode(1),
        "");
}
