/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief unit tests for CmdLineOptProcessor.cpp
 */

#include "CmdLineOptProcessor.h"
#include "VersionPrinter.h"
#include <cstring>
#include <gtest/gtest.h>

//! @brief Encapsulates a list of null-terminated strings
class ArgArray
{
public:
    ~ArgArray()
    {
        for (int i = 0; i < size_; i++) {
            delete[] data_[i];
        }
        delete[] data_;
    }

    ArgArray(int size)
    {
        data_ = new char*[size + 1];
        size_ = size;
        for (auto i = 0; i < size; i++) {
            data_[i] = new char[32];
            std::memset(data_[i], '\0', 32);
        }
        data_[size] = nullptr;
    }

    void clearArgs()
    {
        for (int i = 0; i < size_; i++) {
            std::memset(data_[i], '\0', 32);
        }
    }

    void setArgument(int i, const char* val) { std::strcpy(data_[i], val); }

    char** data() { return data_; }

private:
    char** data_;
    int size_;
};

//! @brief Test Fixture for CmdLineOptProcessor
class CmdLineOptProcessorTest : public ::testing::Test
{

protected:
    std::unique_ptr<task::CmdLineOptProcessor> clop_;
    std::shared_ptr<task::VersionPrinterInterface> versionPrinter_;

    void SetUp() override
    {
        versionPrinter_ = std::make_shared<task::VersionPrinter>();
        clop_ = std::make_unique<task::CmdLineOptProcessor>(versionPrinter_);
        clop_->setExitOnRequestFlag(false); // Don't invoke `std::exit(0)` upon encountering `--help`, etc
    }

    void TearDown() override
    { /* ... */
    }
};

TEST_F(CmdLineOptProcessorTest, ParseCmdLine_NullArgv)
{
    int argc{0};
    char** argv = nullptr;
    auto cmdMap = clop_->parseCmdLine(argc, argv);
    EXPECT_TRUE(cmdMap.optMap.empty());
}

TEST_F(CmdLineOptProcessorTest, ParseCmdLine_ZeroArguments)
{
    int argc{1};
    ArgArray args(argc);
    auto cmdMap = clop_->parseCmdLine(0, args.data());
    EXPECT_TRUE(cmdMap.optMap.empty());
}

TEST_F(CmdLineOptProcessorTest, AddOption_EnsureCharRange)
{
    // Only add an option if within the range of a char [-128, 127]
    option opt{};
    opt.name = "foo";
    opt.has_arg = no_argument;
    opt.flag = nullptr;
    opt.val = 256;
    EXPECT_FALSE(clop_->addOption(opt));
}

TEST_F(CmdLineOptProcessorTest, EnsureInvalidOptionsNotRecognized)
{
    // Ensure options are cleared before starting
    clop_->clearOptions();

    int argc{6};
    ArgArray args(argc);
    args.setArgument(0, "<program>");
    args.setArgument(1, "--longA");
    args.setArgument(2, "--longB");
    args.setArgument(3, "-a");
    args.setArgument(4, "-b");
    args.setArgument(5, "a");

    // Ensure none of the specified arguments are matched to the default options
    EXPECT_TRUE(clop_->addDefaultOptions());
    auto cmdMap = clop_->parseCmdLine(argc, args.data());
    EXPECT_TRUE(cmdMap.optMap.empty());
}

TEST_F(CmdLineOptProcessorTest, AddDefaultOptions)
{
    // Ensure options are cleared before starting
    clop_->clearOptions();

    int argc{6};
    ArgArray args(argc);
    args.setArgument(0, "<program>");
    args.setArgument(1, "--help");
    args.setArgument(2, "--instanceName");
    args.setArgument(3, "foo");
    args.setArgument(4, "--configTest");
    args.setArgument(5, "bar");

    EXPECT_TRUE(clop_->addDefaultOptions());
    auto cmdMap = clop_->parseCmdLine(argc, args.data());
    EXPECT_EQ(cmdMap.optMap.size(), 3);

    std::string value{};
    cmdMap.getOptValue("instanceName", value);
    EXPECT_EQ(value, "foo");

    cmdMap.getOptValue("configTest", value);
    EXPECT_EQ(value, "bar");
}

TEST_F(CmdLineOptProcessorTest, ClearOptions)
{
    // Ensure options are cleared before starting
    clop_->clearOptions();

    // Add 2 options, then clear
    // Ensure that no options are populated in the map returned from `parseCmdLine`
    option opt{};
    opt.name = "foo";
    opt.val = static_cast<int>('f');
    ASSERT_TRUE(clop_->addOption(opt));

    opt.name = "bar";
    opt.val = static_cast<int>('b');
    ASSERT_TRUE(clop_->addOption(opt));

    clop_->clearOptions();

    int argc = 4;
    ArgArray args(argc);
    args.setArgument(0, "<program>");
    args.setArgument(1, "--foo");
    args.setArgument(2, "-f");
    args.setArgument(3, "-b");

    // Should be empty since cleared above
    auto cmdMap = clop_->parseCmdLine(argc, args.data());
    EXPECT_TRUE(cmdMap.optMap.empty());
}
