/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines unit tests for HeartbeatSender
 */

#include "MockHeartbeatMsgPublisher.h"
#include "lib/avlog/avlog.hpp"
#include <Clock/SystemClock.h>
#include <task/HeartbeatSender.h>

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <chrono>
#include <memory>
#include <thread>

using ::testing::_;
using ::testing::Exactly;
using ::testing::Return;
using namespace task;
using namespace ottotime;

/**
 * \brief This class defines unit tests for HeartbeatSender class
 */
class HeartbeatSenderTest : public ::testing::Test
{
public:
    /**
     * \brief Set up the tests
     */
    void SetUp() override
    {
        // Init the system clock
        SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeSteady);
        // Check the system clock is initialized
        ASSERT_TRUE(SystemClock::isInitialized());
        // Instanticate the sender after the clock is initialized
        testSender_ = std::make_unique<HeartbeatSender>(nullptr, cycleTime_s_, intendedFreq_);
        ASSERT_NE(nullptr, testSender_);
        // Init the heartbeat message
        hbMessage_.processID = 0;
        hbMessage_.host = "hostName";
        hbMessage_.source = "taskName";
        hbMessage_.status = "alive";
        hbMessage_.intendedCycleRate_hz = intendedFreq_;
    }

    /**
     * \brief clean up the tests
     */
    void TearDown() override {}

    /**
     * \brief Simulate a running of AV task cycles until a heartbeat message needs to be sent and compute the average
     * frequency
     */
    void SimulateTask()
    {
        constexpr std::chrono::microseconds cycleTime_us(static_cast<int64_t>(hbSenderSecToMicroSecs / intendedFreq_));
        auto nextCycle = std::chrono::steady_clock::now();

        bool exitFlag{false};
        while (!exitFlag) {
            // sleep til start of next cycle
            nextCycle += cycleTime_us;
            std::this_thread::sleep_until(nextCycle);

            // Do the heart beat check after the timer, as this emulates what the task does.
            exitFlag = testSender_->needToSendHeartbeat();
        }
    }

protected:
    static constexpr double intendedFreq_{100.0};          ///< The task intended frequency
    static constexpr double cycleTime_s_{1.0};             ///< The sender cycle time
    av::msg::ProcessHealthMessage hbMessage_{};            ///< The heartbeat message to be sent
    std::unique_ptr<HeartbeatSender> testSender_{nullptr}; ///< The sender with null publisher under test
};

TEST_F(HeartbeatSenderTest, setAndGetHeartbeatMessage)
{
    testSender_->setHeartbeatMessage(hbMessage_);
    auto currentMsg = testSender_->getHeartbeatMessage();
    // Check the setting and getting
    EXPECT_EQ(currentMsg.timeStamp, av::TimePoint());
    EXPECT_EQ(currentMsg.processID, hbMessage_.processID);
    EXPECT_EQ(currentMsg.host, hbMessage_.host);
    EXPECT_EQ(currentMsg.source, hbMessage_.source);
    EXPECT_EQ(currentMsg.status, hbMessage_.status);
    EXPECT_EQ(currentMsg.intendedCycleRate_hz, hbMessage_.intendedCycleRate_hz);
}

TEST_F(HeartbeatSenderTest, needToSendHeartbeat_frequency)
{
    testSender_->setHeartbeatMessage(hbMessage_);
    // The first time check, always return true
    ASSERT_TRUE(testSender_->needToSendHeartbeat());

    // simulate a task with multiple cycles
    SimulateTask();

    // Get the update hb message
    auto currentMsg = testSender_->getHeartbeatMessage();
    AVLOG_INFO("The actual task frequency: {}Hz", AVLOG_ARGS("frequency"), currentMsg.cycleRate_hz);

    // Expect the actual freqs computed by needToSendHeartbeat() is close to the intended frequency,
    //
    // The requirement is that the heartbeat is never late and should be sent in the last possible task cycle.
    // And as there is always some level of delay in waking a task from a sleep, this results in the frequency
    // calculation of:
    //
    //   freq = (<intended freq> - 1) / (<heartbeat period> - (1.0/<intended freq>) + <wakeup delay> )
    //
    // The delay should be small, but on heavily loaded systems (eg Jenkins sanitizers) it can be significant,
    // so we just assume that the task will be woken up sometime in the last cycle.
    //
    //   worst case = (<intended freq> - 1) / (<heartbeat period>)
    //
    // Is should be much better than this, but this avoids having a flaky test.
    constexpr auto worstCaseFreq = (intendedFreq_ - 1.0) / cycleTime_s_;
    EXPECT_LE(currentMsg.cycleRate_hz, intendedFreq_);
    EXPECT_GE(currentMsg.cycleRate_hz, worstCaseFreq);
}

TEST_F(HeartbeatSenderTest, needToSendHeartbeat_timing)
{
    // Create a 3 second heartbeat for a 2.0 Hz task
    HeartbeatSender testSender(nullptr, 3.0, 2.0);
    testSender.setHeartbeatMessage(hbMessage_);

    // Always sent on the first check, this is time 0
    EXPECT_TRUE(testSender.needToSendHeartbeat());
    auto t0 = std::chrono::steady_clock::now();

    // The heartbeat should be sent in the last expected cycle time before it is due
    // that is between 2.5 and 3.0 seconds after the start time. even if this is the only
    // execution
    std::this_thread::sleep_until(t0 + std::chrono::milliseconds(2700));
    EXPECT_TRUE(testSender.needToSendHeartbeat());
    auto t1 = std::chrono::steady_clock::now();

    // An extra execution in the existing cycle, must not trigger another heartbeat
    EXPECT_FALSE(testSender.needToSendHeartbeat());

    // Confirm it doesn't fire early during this cycle
    std::this_thread::sleep_until(t1 + std::chrono::milliseconds(100));
    EXPECT_FALSE(testSender.needToSendHeartbeat());
    std::this_thread::sleep_until(t1 + std::chrono::milliseconds(1000));
    EXPECT_FALSE(testSender.needToSendHeartbeat());
    std::this_thread::sleep_until(t1 + std::chrono::milliseconds(1550));
    EXPECT_FALSE(testSender.needToSendHeartbeat());
    std::this_thread::sleep_until(t1 + std::chrono::milliseconds(2100));
    EXPECT_FALSE(testSender.needToSendHeartbeat());

    // But should be sent when you hit the next period
    //   - "t1" is just after the last heartbeat was sent, so that combined
    //      with the wakeup delay means this will just be over the boundary
    std::this_thread::sleep_until(t1 + std::chrono::milliseconds(3000));
    EXPECT_TRUE(testSender.needToSendHeartbeat());
}

TEST_F(HeartbeatSenderTest, sendHeartbeatNullPublisher)
{
    ASSERT_FALSE(testSender_->sendHeartbeat());
}

TEST_F(HeartbeatSenderTest, sendHeartbeat)
{
    // Set up a mocked publisher
    auto mockHeartbeatPublisher = std::make_unique<MockHeartbeatMsgPublisher>(nullptr);
    EXPECT_NE(nullptr, mockHeartbeatPublisher);
    auto publisherPtr = mockHeartbeatPublisher.get();
    // Instantiate a sender with the mocked publisher
    HeartbeatSender testSender{std::move(mockHeartbeatPublisher), cycleTime_s_, intendedFreq_};
    EXPECT_NE(nullptr, publisherPtr);
    testSender.setHeartbeatMessage(hbMessage_);
    // HB message publishing is failed
    EXPECT_CALL(*publisherPtr, publishMessage(_)).Times(Exactly(1)).WillOnce(Return(false));
    ASSERT_FALSE(testSender.sendHeartbeat());
    // HB message publishing is fine
    EXPECT_CALL(*publisherPtr, publishMessage(_)).Times(Exactly(1)).WillOnce(Return(true));
    ASSERT_TRUE(testSender.sendHeartbeat());
}
