/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines unit tests for the main function
 */

#include "../../main.def.h"
#include "MockTaskApplication.h"

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <chrono>
#include <cstdio>
#include <cstdlib>
#include <signal.h>
#include <spawn.h>
#include <thread>
#include <vector>

using namespace task;

#ifndef INFINITE_TASK_EXE
#define INFINITE_TASK_EXE "platform/middleware/libs/task/test/MainFunctionTest/InfiniteExeTaskTest"
#endif

/**
 * Task get implementation
 */
Task* Task::get(void)
{
    return nullptr;
}

/**
 * \brief This class defines unit tests for the main function
 */
class MainFunctionTest : public ::testing::Test
{
public:
    /**
     * \brief Set up the tests
     */
    void SetUp() override {}

    /**
     * \brief Clean up the tests
     */
    void TearDown() override { killTask(); }

    /**
     * \brief spawn a new task process
     *
     * \param fileName the binary file name
     *
     * \return True if the spawning is fine, false otherwise
     */
    bool spawnTask(const std::string& fileName)
    {
        bool result{false};
        const std::string arvStr{"--instanceName=MainFunctionTest"};
        std::vector<char*> charsVec{};
        charsVec.push_back(const_cast<char*>(arvStr.c_str()));
        charsVec.push_back(nullptr);
        // Spawn a process
        result = (posix_spawn(&pID_, fileName.c_str(), nullptr, nullptr, &charsVec[0], nullptr) == 0);
        return result;
    }

    /**
     * \brief Kill the task process
     */
    bool killTask()
    {
        bool result{false};
        if ((pID_ != -1) && (kill(pID_, SIGKILL) == 0)) {
            pID_ = -1;
            result = true;
        }
        return result;
    }

    /**
     * \brief Test executeTask() function
     *
     * \param initResult The return value of task initialization
     * \param runResult The return value of task running
     *
     * \return The return value from executeTask()
     */
    template<bool initResult, bool runResult>
    int testExecuteTask(int argc, char** argv)
    {
        const auto result = executeTask<ClosingSignalHandler,
                                        CmdLineOptProcessor,
                                        VersionPrinter,
                                        Task,
                                        MockTaskApplication<initResult, runResult>>(argc, argv);
        return result;
    }
    template<bool initResult, bool runResult>
    int testExecuteTask()
    {
        // Note: the argv[] passed to main() must contain at least one entry, followed by a null pointer.
        char programName[] = "MyTask";
        char* argv[] = {programName, nullptr};
        return testExecuteTask<initResult, runResult>(1, argv);
    }

protected:
    pid_t pID_{-1}; ///< The task process ID
};

TEST_F(MainFunctionTest, executeTask_FailedInit)
{
    auto exeResult = testExecuteTask<false, false>();
    EXPECT_EQ(exeResult, EXIT_FAILURE);
    exeResult = testExecuteTask<false, true>();
    EXPECT_EQ(exeResult, EXIT_FAILURE);
}

TEST_F(MainFunctionTest, executeTask_FailedRun)
{
    const auto exeResult = testExecuteTask<true, false>();
    EXPECT_EQ(exeResult, EXIT_FAILURE);
}

TEST_F(MainFunctionTest, executeTask_Normal)
{
    const auto exeResult = testExecuteTask<true, true>();
    EXPECT_EQ(exeResult, EXIT_FAILURE);
}

TEST_F(MainFunctionTest, executeTask_Named)
{
    char programName[] = "MyTask";
    char shortArg[] = "-i";
    char longArg[] = "--instanceName";
    char equalArg[] = "--instanceName=MyTask";
    char* argv[] = {programName, shortArg, programName, nullptr};
    const auto shortResult = testExecuteTask<true, true>(3, argv);
    EXPECT_EQ(shortResult, EXIT_FAILURE);
    argv[1] = longArg;
    const auto longResult = testExecuteTask<true, true>(3, argv);
    EXPECT_EQ(longResult, EXIT_FAILURE);
    argv[1] = equalArg;
    argv[2] = nullptr;
    const auto equalResult = testExecuteTask<true, true>(2, argv);
    EXPECT_EQ(equalResult, EXIT_FAILURE);
    argv[1] = shortArg;
    const auto badResult = testExecuteTask<true, true>(2, argv);
    EXPECT_EQ(badResult, EXIT_FAILURE);
}

TEST_F(MainFunctionTest, normalExecutionTask)
{
    // Spawn an infinite execution task
    EXPECT_TRUE(spawnTask(INFINITE_TASK_EXE));
    std::this_thread::sleep_for(std::chrono::microseconds(1000));
    EXPECT_TRUE(killTask());
}
