/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines unit tests for MonitoringStatusProvider
 */

#include "lib/avlog/avlog.hpp"
#include <avtime/Time.hpp>
#include <gtest/gtest.h>
#include <interfaces/AvTaskStatus/AvTaskStatus.h>
#include <memory>
#include <task/MonitoringStatusProvider.hpp>
#include <task/Task.h>

using namespace task;
using namespace av::msg;

// Define the Task get API for compilation
task::Task* task::Task::get()
{
    return nullptr;
}

/**
 * \brief This class defines unit tests for MonitoringStatusProvider class
 */
class MonitoringStatusProviderTest : public ::testing::Test
{
public:
    /**
     * \brief set up the tests
     */
    void SetUp() override
    {
        // Instantiate a MonitoringStatusProvider
        testProvider_ = std::make_shared<MonitoringStatusProvider>("MonitoringStatusProviderTest");
        ASSERT_NE(nullptr, testProvider_);
    }

    // Comparator for AvTaskStatus
    bool operator==(const AvTaskStatus& B)
    {
        return (avTaskStatus_.timestamp == B.timestamp && avTaskStatus_.taskName == B.taskName &&
                avTaskStatus_.state == B.state && avTaskStatus_.source == B.source &&
                avTaskStatus_.message == B.message && avTaskStatus_.type == B.type &&
                avTaskStatus_.parameter_name == B.parameter_name && avTaskStatus_.measured_value == B.measured_value &&
                avTaskStatus_.threshold_value == B.threshold_value);
    }
    /**
     * \brief clean up the tests
     */
    void TearDown() override {}

protected:
    AvTaskStatus avTaskStatus_{};                                     ///< AvTaskStatus message from the task
    std::shared_ptr<MonitoringStatusProvider> testProvider_{nullptr}; ///< test MonitoringStatusProvider
};

TEST_F(MonitoringStatusProviderTest, setAndGetStatus)
{
    {
        int64_t currTime = av::usecFromTimePoint(av::AvClock::now());
        testProvider_->setStatus(currTime, TaskState::UP, FaultSource::INTERNAL, "Task is working as expected");
        AvTaskStatus testStatus = {currTime,
                                   "MonitoringStatusProviderTest",
                                   TaskState::UP,
                                   FaultSource::INTERNAL,
                                   "Task is working as expected",
                                   ParameterType::INT,
                                   "",
                                   "",
                                   ""};
        AvTaskStatus receivedStatus = testProvider_->getStatus();
        EXPECT_EQ(testStatus, receivedStatus);
    }
    {
        int64_t currTime = av::usecFromTimePoint(av::AvClock::now());
        AvTaskStatus testStatus = {currTime,
                                   "MonitoringStatusProviderTest",
                                   TaskState::UP,
                                   FaultSource::INTERNAL,
                                   "Task is working as expected",
                                   ParameterType::INT,
                                   "",
                                   "",
                                   ""};
        testProvider_->setStatus(testStatus);
        AvTaskStatus receivedStatus = testProvider_->getStatus();
        EXPECT_EQ(testStatus, receivedStatus);
    }
}

TEST_F(MonitoringStatusProviderTest, testIsCurrentStateMoreSevereThan)
{
    // intended state is severe than current state
    {
        int64_t currTime = av::usecFromTimePoint(av::AvClock::now());
        testProvider_->setStatus(currTime, TaskState::UP, FaultSource::INTERNAL, "Task is working as expected");
        currTime = av::usecFromTimePoint(av::AvClock::now());
        TaskState intendedState = TaskState::DEGRADED;
        bool retVal = testProvider_->isCurrentStateMoreSevereThan(intendedState);
        EXPECT_EQ(retVal, true);
        // branch coverage
        intendedState = TaskState::DOWN;
        bool retVal2 = testProvider_->isCurrentStateMoreSevereThan(intendedState);
        EXPECT_EQ(retVal2, true);
        // branch coverage
        intendedState = TaskState::UNRECOVERABLE;
        bool retVal3 = testProvider_->isCurrentStateMoreSevereThan(intendedState);
        EXPECT_EQ(retVal3, true);
    }

    // intended state is not severe than current state
    {
        int64_t currTime = av::usecFromTimePoint(av::AvClock::now());
        testProvider_->setStatus(
            currTime, TaskState::UNRECOVERABLE, FaultSource::INTERNAL, "Task is working as expected");
        TaskState intendedState = TaskState::DOWN;
        bool retVal = testProvider_->isCurrentStateMoreSevereThan(intendedState);
        EXPECT_EQ(retVal, false);
    }
}

TEST_F(MonitoringStatusProviderTest, updateStatusAndGetStatus)
{
    int64_t currTime = av::usecFromTimePoint(av::AvClock::now());
    testProvider_->updateStatus(currTime,
                                av::msg::TaskState::DEGRADED,
                                av::msg::FaultSource::INTERNAL,
                                "Random Variable exceeded the threshold value",
                                av::msg::ParameterType::INT,
                                "randomValue",
                                std::to_string(10),
                                std::to_string(9));
    AvTaskStatus testStatus = {currTime,
                               "MonitoringStatusProviderTest",
                               TaskState::DEGRADED,
                               FaultSource::INTERNAL,
                               "Random Variable exceeded the threshold value",
                               ParameterType::INT,
                               "randomValue",
                               std::to_string(10),
                               std::to_string(9)};
    AvTaskStatus receivedStatus = testProvider_->getStatus();
    EXPECT_EQ(testStatus, receivedStatus);
}

int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
