/**
 * \copyright
 * Copyright (c) 2021- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines unit tests for MonitoringStatusSender
 */

#include "MockAvTaskStatusPublisher.h"
#include <config/ConfigParser.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <task/MonitoringStatusSender.h>

#include <memory>

using ::testing::_;
using ::testing::Exactly;
using ::testing::Return;
using namespace task;
using namespace av::msg;

namespace {
constexpr double defaultFreq_Hz{1.0};
constexpr double msSenderSecToMicroSecs{1000000.0};
} // namespace

/**
 * \brief This class defines unit tests for MonitoringStatusSender class
 */
class MonitoringStatusSenderTest : public ::testing::Test
{
public:
    /**
     * \brief set up the tests
     */
    void SetUp() override
    {
        // Set up a mocked publisher
        auto mockAvTaskStatusPublisher = std::make_unique<MockAvTaskStatusPublisher>(nullptr);
        ASSERT_NE(nullptr, mockAvTaskStatusPublisher);
        m_publisherPtr = mockAvTaskStatusPublisher.get();
        ASSERT_NE(nullptr, m_publisherPtr);
        // Instantiate a sender with the mocked publisher
        m_testSender = std::make_unique<MonitoringStatusSender>(std::move(mockAvTaskStatusPublisher));
        ASSERT_NE(nullptr, m_testSender);
        // initialize the status message
        m_statusMessage.timestamp = 0;
        m_statusMessage.state = TaskState::UP;
    }

    /**
     * \brief clean up the tests
     */
    void TearDown() override {}

    /**
     * \brief create task parameters
     *
     * \param taskRate the AV task rate
     * \param sendingRate the status sending rate
     *
     * \return void
     */
    void makeTaskParam(const double taskRate, const double sendingRate)
    {
        std::string configString = "params = {";
        // add the task rate
        configString += "cycleRate_hz = " + std::to_string(taskRate) + ";";
        // add the min rate
        configString += "monitoringStatusSendingRate_hz = " + std::to_string(sendingRate) + ";";
        // close
        configString += "}";

        // parse it
        ConfigParser parser{};
        ASSERT_TRUE(parser.parseString(configString));
        ASSERT_TRUE(parser.getSection("params", m_taskParams));
    }

protected:
    ConfigSection m_taskParams{};                                  ///< The task param for testing
    AvTaskStatus m_statusMessage{};                                ///< The status message to be sent
    std::unique_ptr<MonitoringStatusSender> m_testSender{nullptr}; ///< The sender with null publisher under test
    MockAvTaskStatusPublisher* m_publisherPtr{nullptr};            ///< The underlying publisher
};

TEST_F(MonitoringStatusSenderTest, initializeDefaultFreq)
{
    makeTaskParam(-1.0, -2.0);
    EXPECT_TRUE(m_testSender->initialize(m_taskParams));
    EXPECT_EQ(m_testSender->getSendingFrequency(), defaultFreq_Hz);
    // Already initialized
    EXPECT_TRUE(m_testSender->initialize(m_taskParams));
}

TEST_F(MonitoringStatusSenderTest, initializeTaskFreq)
{
    makeTaskParam(0.5, -2.0);
    EXPECT_TRUE(m_testSender->initialize(m_taskParams));
    EXPECT_EQ(m_testSender->getSendingFrequency(), 0.5);
}

TEST_F(MonitoringStatusSenderTest, initializeTaskFreqBiggerThanDefaultFreq)
{
    makeTaskParam(2.5, -2.0);
    EXPECT_TRUE(m_testSender->initialize(m_taskParams));
    EXPECT_EQ(m_testSender->getSendingFrequency(), defaultFreq_Hz);
}

TEST_F(MonitoringStatusSenderTest, initializeSendingFreq)
{
    makeTaskParam(-1.0, 2.0);
    EXPECT_TRUE(m_testSender->initialize(m_taskParams));
    EXPECT_EQ(m_testSender->getSendingFrequency(), 2.0);
}

TEST_F(MonitoringStatusSenderTest, initializeMinFreq)
{
    makeTaskParam(1.0, 2.0);
    EXPECT_TRUE(m_testSender->initialize(m_taskParams));
    EXPECT_EQ(m_testSender->getSendingFrequency(), 1.0);
}

TEST_F(MonitoringStatusSenderTest, sendStatusNullPublisher)
{
    MonitoringStatusSender nullPublisherSender{nullptr};
    EXPECT_FALSE(nullPublisherSender.sendStatus(m_statusMessage));
}

TEST_F(MonitoringStatusSenderTest, sendStatus)
{
    // Uninitialized sender
    EXPECT_FALSE(m_testSender->sendStatus(m_statusMessage));
    // Initialize the sender
    makeTaskParam(5.0, 2.0);
    EXPECT_TRUE(m_testSender->initialize(m_taskParams));
    // Send for the first time and fails
    EXPECT_CALL(*m_publisherPtr, publishMessage(_)).Times(Exactly(1)).WillOnce(Return(false));
    EXPECT_FALSE(m_testSender->sendStatus(m_statusMessage));
    // Send the same status but the timestamp advancing is smaller than 1/2.0 seconds, no publishing
    m_statusMessage.timestamp = static_cast<int64_t>(0.2 * msSenderSecToMicroSecs);
    EXPECT_CALL(*m_publisherPtr, publishMessage(_)).Times(Exactly(0));
    EXPECT_TRUE(m_testSender->sendStatus(m_statusMessage));
    // Advance the time more than 1/2.0 seconds
    m_statusMessage.timestamp = static_cast<int64_t>(0.9 * msSenderSecToMicroSecs);
    EXPECT_CALL(*m_publisherPtr, publishMessage(_)).Times(Exactly(1)).WillOnce(Return(true));
    EXPECT_TRUE(m_testSender->sendStatus(m_statusMessage));
    // Advance the time less than 1/2.0 seconds but the status state changes
    m_statusMessage.timestamp = static_cast<int64_t>(1.2 * msSenderSecToMicroSecs);
    m_statusMessage.state = TaskState::DOWN;
    EXPECT_CALL(*m_publisherPtr, publishMessage(_)).Times(Exactly(1)).WillOnce(Return(true));
    EXPECT_TRUE(m_testSender->sendStatus(m_statusMessage));
}

int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}