/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Defintions for resim test helper functions.
#include "resimHelpers.hpp"

#include "lib/avutils/process/Process.hpp"

#include <gtest/gtest.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

#include <experimental/filesystem>
#include <fstream>
#include <random>
#include <stdexcept>

namespace fs = std::experimental::filesystem;

namespace resim {

namespace file {
const fs::path tmp = fs::path(getenv("TEST_TMPDIR"));
const fs::path base = fs::path(getenv("TEST_SRCDIR")) / "av_stack/platform/middleware";
const fs::path scs = base / "tasks/SimpleComms/SimpleComms/scs";
} // namespace file

namespace {
std::vector<std::string> readOutput(av::Pipe& pipe)
{
    std::vector<std::string> lines;
    std::string line;
    while (pipe.getLine(line)) {
        if (line.find("Receiver received") != std::string::npos) {
            lines.push_back(line);
        }
    }
    return lines;
}
std::string scsPath;
} // namespace

std::string setup(const std::string& clock, const std::string& receiver)
{
    // Random SCS instance name:
    scsPath = setSCSPath();

    // path to Sanitizers suppression file is relative to file::workDir.
    // we should copy suppression file to the file::tmp , where other processess will be executed
    const char* suppressionsFile = getenv("SANITIZER_SUPPRESSION_FILE");
    if (suppressionsFile != nullptr) {
        fs::copy(fs::current_path() / suppressionsFile, file::tmp / suppressionsFile);
    }

    // Set the working directory where the event logs will be saved.
    EXPECT_EQ(chdir(file::tmp.c_str()), 0);

    // Config files:
    {
        std::ofstream out("SampleSender.cfg");
        out << "Parameters =\n{\n";
        out << "    loggerThreshold = \"info\";\n";
        out << "    clockType = \"" << clock << "\";\n";
        out << "    reproducible = True;\n";
        out << "    cycleRate_hz = 1618.034;\n};\n";
        out << "Execution =\n{\n    executableName = \"SampleSender\";\n};\n";
    }
    {
        std::ofstream out(receiver + ".cfg");
        out << "Parameters =\n{\n";
        out << "    loggerThreshold = \"info\";\n";
        out << "    clockType = \"" << clock << "\";\n";
        out << "    reproducible = True;\n";
        out << "    cycleRate_hz = 1000.0;\n";
        out << "    cycleLimit = 10000;\n};\n";
        out << "Execution =\n{\nexecutableName = \"" << receiver << "\";\n};\n";
    }
    const char* VAR = "OTTO_CONFIG_DIR";
    setenv(VAR, file::tmp.c_str(), 1);
    std::cout << VAR << "=" << getenv(VAR) << std::endl;

    // Return the SCS path.
    return scsPath;
}

std::string setSCSPath()
{
    // Random SCS instance name:
    std::string path = "scs-";
    std::random_device rd;
    std::default_random_engine gen(rd());
    std::uniform_int_distribution<char> dist('a', 'z');
    for (int i = 0; i < 12; ++i) {
        path += dist(gen);
    }
    const char* VAR = "SCS_PATH";
    setenv("SCS_PATH", path.c_str(), 1);
    std::cout << VAR << "=" << getenv(VAR) << std::endl;
    return path;
}

std::vector<std::string> runOriginal(const std::string& sampleSender,
                                     const std::string& sampleReceiver,
                                     const std::string& messageLog)
{
    std::vector<std::string> scsArgs = {"--maxExitSignalsToCatch", "1", "--unixPath", scsPath};
    if (!messageLog.empty()) {
        // Log messages for server-side resim.
        scsArgs.emplace_back("--log");
        scsArgs.emplace_back(messageLog);
    } else {
        // Run SCS without logging messages. (This is for client-side resim.)
        scsArgs.emplace_back("--staleFragmentPruningThreshold");
        scsArgs.emplace_back("0");
    }
    av::Process scs(file::scs, scsArgs);
    av::Process sender(sampleSender);
    // Delay 100 ms. before launching the receiver, so that some messages will be lost.
    usleep(100000);
    av::Process receiver(sampleReceiver);

    // Store all the lines that report received messages.
    std::vector<std::string> lines = readOutput(receiver.out());

    // The sender and SCS must be manually killed.
    sender.kill();
    scs.kill();

    return lines;
}

std::vector<std::string> runReplay(const std::string& sampleSender, const std::string& sampleReceiver)
{
    setenv("REPRODUCIBLE", getenv("TEST_TMPDIR"), 1);

    av::Process scs(
        file::scs, "--maxExitSignalsToCatch", "1", "--unixPath", scsPath, "--staleFragmentPruningThreshold", "0");
    av::Process receiver(sampleReceiver);

    // Wait for the receiver to be ready before sending messages.
    std::string line;
    while (receiver.out().getLine(line)) {
        if (line.find("Starting task in replay mode.") != std::string::npos) {
            break;
        }
    }

    // Now, it's safe to launch the sender.
    av::Process sender(sampleSender);

    // Store all the lines that report received messages.
    std::vector<std::string> lines = readOutput(receiver.out());

    // The sender and SCS must be manually killed.
    sender.kill();
    scs.kill();

    unsetenv("REPRODUCIBLE");

    return lines;
}

void compareOutputs(const std::vector<std::string>& original, const std::vector<std::string>& replay)
{
    // Allow the replay to be missing one event that was in the original.
    ASSERT_GE(original.size(), replay.size());
    ASSERT_LE(original.size(), replay.size() + 1);
    // Use asserts here so only the first mismatch is printed.
    ASSERT_GT(original.size(), 1000);
    for (size_t i = 0; i < replay.size(); ++i) {
        ASSERT_EQ(original[i], replay[i]);
    }
}

} // namespace resim
