/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief unit tests for TaskApplication.hpp
 */

#include "TaskApplication.h"
#include "platform/middleware/libs/Clock/include/Clock/SystemClock.h"
#include "platform/middleware/libs/TestUtils/include/TestUtils/SystemClockScopedInitializer.hpp"
#include "platform/middleware/libs/log/include/log/Logger.h"
#include "platform/middleware/libs/task/TaskUtil.h"
#include "platform/middleware/libs/task/test/TaskTestUtils/MockObjects.hpp"

#include <algorithm>
#include <cstdlib>
#include <gtest/gtest.h>

using namespace task;
using task::TestUtils::MockClosingSignalHandler;
using task::TestUtils::MockCmdLineOptProcessor;
using task::TestUtils::MockTask;
using task::TestUtils::MockVersionPrinter;

namespace {
std::unique_ptr<CmdLineOptProcessorInterface> makeCmdLineOptProcessor()
{
    auto mockOptRawPtr = new MockCmdLineOptProcessor();
    auto mockOpt = std::unique_ptr<CmdLineOptProcessorInterface>(mockOptRawPtr);

    EXPECT_CALL(*mockOptRawPtr, parseCmdLine).WillOnce([](int, char**) { return CmdLineOptMap{}; });
    EXPECT_CALL(*mockOptRawPtr, addOption).Times(0);
    EXPECT_CALL(*mockOptRawPtr, addDefaultOptions).WillOnce([] { return true; });
    EXPECT_CALL(*mockOptRawPtr, clearOptions).Times(0);
    return mockOpt;
}

bool buildAndRun(std::unique_ptr<ClosingSignalHandlerInterface> mockSignalHandler, Task* task)
{
    auto mockOpt = makeCmdLineOptProcessor();
    auto mockVp = std::unique_ptr<VersionPrinterInterface>(new MockVersionPrinter());

    MockVersionPrinter* mockVpRaw = static_cast<MockVersionPrinter*>(mockVp.get());
    EXPECT_CALL(*mockVpRaw, getVehicleName).WillOnce([](const std::string& /*configFile*/, const bool /*logMode*/) {
        return "Argo";
    });

    EXPECT_CALL(*mockVpRaw, getAVStackBuildNumber)
        .WillOnce([](const std::string& /*mode*/, const std::string& /*buildTagFile*/, const bool /*logMode*/) {
            AVStackBuildNumber buildNumber;
            buildNumber.avStackNumber = "e";
            buildNumber.avStackNumberFromLog = "17";
            return buildNumber;
        });

    TaskApplication app{std::move(mockSignalHandler), std::move(mockOpt), std::move(mockVp), task};

    // Can't just use "char* argv[]{"mocker"}, because string literals
    // are const. And we can't use const char* argv, because
    // CmdLineOptProcessorInterface presumes it's using getopt, which
    // modifies argv.
    char mockerStr[]{"mocker"};
    char* argv[]{mockerStr};
    app.initTask(sizeof(argv) / sizeof(argv[0]), argv);

    return app.runTask();
}

} // namespace

TEST(TaskApplication, runTask)
{
    auto mockSignalHandler = std::unique_ptr<ClosingSignalHandlerInterface>(new MockClosingSignalHandler());
    MockClosingSignalHandler* mockSignalHandlerRawPtr = static_cast<MockClosingSignalHandler*>(mockSignalHandler.get());

    // Note: The signal mask is set at the very top of main(), before the TaskApplication is created.
    EXPECT_CALL(*mockSignalHandlerRawPtr, shouldExit).WillRepeatedly([] { return false; });
    EXPECT_CALL(*mockSignalHandlerRawPtr, handleClosingSignals).WillOnce([] { return true; });

    int count{0};
    std::vector<int> records;

    MockTask mockTask{};
    EXPECT_CALL(mockTask, cleanup).WillOnce([] {});

    const auto action = [&count, &records]() {
        ++count;
        records.push_back(count * count);
        return true;
    };
    const auto finalAction = [&count, &records]() {
        ++count;
        records.push_back(count * count);
        return false;
    };

    EXPECT_CALL(mockTask, executive)
        .WillOnce(action)
        .WillOnce(action)
        .WillOnce(action)
        .WillOnce(action)
        .WillOnce(finalAction);

    EXPECT_TRUE(buildAndRun(std::move(mockSignalHandler), &mockTask));

    EXPECT_EQ(5, count);

    const std::vector<int> expectedRecords{1, 4, 9, 16, 25};
    EXPECT_EQ(expectedRecords, records);
    SystemClockScopedInitializer::uninitialize();
}

TEST(TaskApplication, runTaskExitWithSignal)
{
    auto mockSignalHandler = std::unique_ptr<ClosingSignalHandlerInterface>(new MockClosingSignalHandler());
    MockClosingSignalHandler* mockSignalHandlerRawPtr = static_cast<MockClosingSignalHandler*>(mockSignalHandler.get());

    // Note: The signal mask is set at the very top of main(), before the TaskApplication is created.
    bool shouldExitFlag{false};
    EXPECT_CALL(*mockSignalHandlerRawPtr, shouldExit).WillRepeatedly([&shouldExitFlag] { return shouldExitFlag; });
    EXPECT_CALL(*mockSignalHandlerRawPtr, handleClosingSignals).WillOnce([] { return true; });
    const auto mockIncomingSignal = [&shouldExitFlag](int /*sig*/) { shouldExitFlag = true; };

    int count{0};
    std::vector<int> records;
    {
        MockTask mockTask;
        EXPECT_CALL(mockTask, cleanup).WillOnce([] {});

        const auto action = [&count, &records, &mockIncomingSignal, mockSignalHandlerRawPtr]() {
            ++count;
            records.push_back(count * count);
            mockIncomingSignal(SIGTERM);
            return true;
        };

        EXPECT_CALL(mockTask, executive).WillOnce(action);

        EXPECT_FALSE(buildAndRun(std::move(mockSignalHandler), &mockTask));
    }

    EXPECT_EQ(1, count);

    const std::vector<int> expectedRecords{1};
    EXPECT_EQ(expectedRecords, records);
    SystemClockScopedInitializer::uninitialize();
}
