#include <gtest/gtest-param-test.h>
#include <gtest/gtest.h>

#include <ODatum/ODatum.h>
#include <SimpleComms/Constants.h>
#include <SimpleComms/ControlMessages.h>
#include <TimeStamp/otime.h>
#include <interfaces/SampleMessage/SampleMessage.h>
#include <interfaces/baseTypes/Header.h>
#include <task/CallbackSet.h>
#include <task/MessageCallback.h>

#include "MockReceiver.h"

#include <cstdint> // for int64_t
#include <exception>
#include <memory>
#include <string>
#include <vector>

namespace {

std::vector<std::string> getBlobData(const std::vector<int>& valMessages,
                                     int64_t startSec = 12345,
                                     int64_t startNsecs = 67890)
{
    std::vector<std::string> blobList;
    for (auto val : valMessages) {
        std::string blob;
        av::msg::SampleMessage payload;
        payload.value = val;
        interfaces::baseTypes::InterfaceHeader header;

        header.headerVersion = htonl(OTTO_INTERFACE_HEADER_VERSION);
        ottotime::otimeSpec currTime(startSec, startNsecs);
        interfaces::baseTypes::setSendTimeStamp(header, currTime);

        aptv::serializeODatumMessage(std::string(reinterpret_cast<char*>(&header), sizeof(header)), blob, payload);
        blobList.push_back(blob);
    }

    return blobList;
}

} // namespace

TEST(CallbackSetTest, testConstructorDestructorNoThrow)
{
    av::repro::EventFramework framework;
    std::unique_ptr<task::CallbackSet> ptr;
    EXPECT_NO_THROW(ptr = std::make_unique<task::CallbackSet>(framework));
    EXPECT_NO_THROW(ptr->timestamp());
    EXPECT_NO_THROW(ptr.reset());
}

TEST(CallbackSetTest, testInitStopNoRepNoThrow)
{
    av::repro::EventFramework framework;
    task::CallbackSet cs(framework);
    EXPECT_NO_THROW(cs.init(false));
    EXPECT_NO_THROW(cs.stop());
}

TEST(CallbackSetTest, testInitStopRepNoThrow)
{
    av::repro::EventFramework framework;
    task::CallbackSet cs(framework);
    EXPECT_NO_THROW(cs.init(true));
    EXPECT_NO_THROW(cs.stop());
}

TEST(CallbackSetTest, testSetNoThrow)
{
    av::repro::EventFramework framework;
    task::CallbackSet cs(framework);
    EXPECT_NO_THROW(cs.setStaleThreshold(1.0));
    EXPECT_NO_THROW(cs.setExecutiveBacklogLimit(10));
}

TEST(CallbackSetTest, testCallOnEmptyChannelNoThrow)
{
    av::repro::EventFramework framework;
    task::CallbackSet cs(framework);

    // The following passes a non-registered (empty) channel and an empty message.
    // The call method should return without throwing and do nothing
    task::CallbackSet::MessageKey msgKey;
    const std::string noChannel{};
    EXPECT_NO_THROW(cs.call(noChannel, "", msgKey));

    // An empty string in the internal channel map represents the executive function
    // Register an executive function and verify that this would call it
    // Add callback on channel
    Receiver<av::msg::SampleMessage> receiver;
    ASSERT_NO_THROW(cs.add(noChannel, &receiver, &Receiver<av::msg::SampleMessage>::receive));

    int64_t startSec = 12345;
    int64_t startNsecs = 67890;
    const std::vector<int> dataMsg = {10000};
    auto blobList = getBlobData(dataMsg, startSec, startNsecs);
    receiver.setExpectMsgs(1);

    ottotime::otimeSpec defaultTime;
    EXPECT_EQ(cs.timestamp(), defaultTime);
    ASSERT_NO_THROW(cs.call(noChannel, blobList.front(), msgKey));
    receiver.wait();

    // Check received messages
    ottotime::otimeSpec currTime(startSec, startNsecs);
    EXPECT_EQ(cs.timestamp(), currTime)
        << "Timestamp should be set to the current time of the message read from the queue";
    EXPECT_EQ(receiver.received.size(), 1);
    EXPECT_EQ(receiver.received[0].value, dataMsg.front());

    // Note: in actual code, we rely on Task::cleanup() to call callbacks_.stop().
    ASSERT_NO_THROW(cs.stop());
}

TEST(CallbackSetTest, testCallOnControlChannelReplayMsgNoThrow)
{
    av::repro::EventFramework framework;
    task::CallbackSet cs(framework);

    const std::string channel{SIMPLECOMMS_CONTROL_CHANNEL};
    task::CallbackSet::MessageKey mk;

    SimpleComms::ReplayMessage repMsg;
    std::string blob(reinterpret_cast<const char*>(&repMsg), sizeof(repMsg));
    ASSERT_NO_THROW(cs.call(channel, blob, mk));
    EXPECT_TRUE(cs.isReplaying()) << "CallbackSet should be replaying";

    // Note: in actual code, we rely on Task::cleanup() to call callbacks_.stop().
    ASSERT_NO_THROW(cs.stop());
}

TEST(CallbackSetTest, testCallOnControlChannelExecutiveMsgNoThrow)
{
    av::repro::EventFramework framework;
    task::CallbackSet cs(framework);

    const std::string channel{SIMPLECOMMS_CONTROL_CHANNEL};
    task::CallbackSet::MessageKey mk;

    SimpleComms::ExecutiveMessage exeMsg;
    std::string blob(reinterpret_cast<const char*>(&exeMsg), sizeof(exeMsg));

    // Register a callback that should get triggered on executive messages
    ReceiverNoInput<SimpleComms::ExecutiveMessage> receiver;
    ASSERT_NO_THROW(cs.add("", &receiver, &ReceiverNoInput<SimpleComms::ExecutiveMessage>::receive));

    // Test that an executive message calls an "executive" callback function
    receiver.setExpectMsgs(1);
    ASSERT_NO_THROW(cs.call(channel, blob, mk));
    receiver.wait();
    EXPECT_EQ(receiver.m_received, 1);

    // Note: in actual code, we rely on Task::cleanup() to call callbacks_.stop().
    ASSERT_NO_THROW(cs.stop());
}

TEST(CallbackSetTest, testCallOnRegisteredChannelNoThrow)
{
    av::repro::EventFramework framework;
    task::CallbackSet cs(framework);

    // Add callback on channel
    const std::string channel{"MyChannel"};
    Receiver<av::msg::SampleMessage> receiver;
    ASSERT_NO_THROW(cs.add(channel, &receiver, &Receiver<av::msg::SampleMessage>::receive));

    int64_t startSec = 12345;
    int64_t startNsecs = 67890;
    const std::vector<int> dataMsg = {1234, 4567, 8901};
    auto messageBlobs = getBlobData(dataMsg, startSec, startNsecs);
    receiver.setExpectMsgs(static_cast<int>(dataMsg.size()));

    ottotime::otimeSpec defaultTime;
    EXPECT_EQ(cs.timestamp(), defaultTime);
    for (const auto& blob : messageBlobs) {
        ASSERT_NO_THROW(cs.call(channel, blob, task::CallbackSet::MessageKey()));
    }
    receiver.wait();

    // Check received messages
    ottotime::otimeSpec currTime(startSec, startNsecs);
    EXPECT_EQ(cs.timestamp(), currTime)
        << "Timestamp should be set to the current time of the message read from the queue";
    EXPECT_EQ(receiver.received.size(), dataMsg.size());

    std::size_t ctr{0};
    for (auto data : dataMsg) {
        EXPECT_EQ(receiver.received[ctr++].value, data);
    }

    // Note: in actual code, we rely on Task::cleanup() to call callbacks_.stop().
    ASSERT_NO_THROW(cs.stop());
}
