#include <gtest/gtest-param-test.h>
#include <gtest/gtest.h>

#include <Clock/SystemClock.h>
#include <ODatum/ODatum.h>
#include <SimpleComms/test/MockUnixClient.h>
#include <interfaces/SampleMessage/SampleMessage.h>
#include <repro/EventFramework.hpp>
#include <task/CallbackSet.h>
#include <task/MessageOutput.h>

using ::testing::_;
using ::testing::Return;
using ::testing::SaveArg;

class MessageOutputTestFixture
{
public:
    std::shared_ptr<SimpleComms::MockUnixClient> client;
    av::repro::EventFramework framework;
    std::shared_ptr<task::CallbackSet> cset;

    MessageOutputTestFixture()
    {
        SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeSteady);
        client = std::make_shared<SimpleComms::MockUnixClient>();
        framework.setClient(client);
        framework.init(av::repro::EventFramework::Mode::live);
        cset = std::make_shared<task::CallbackSet>(framework);
    }
};

TEST(MessageOutputTest, testConstructorNoThrow)
{
    MessageOutputTestFixture msgTestFixt;

    EXPECT_NO_THROW(task::MessageOutput<av::msg::SampleMessage> moAux("", *msgTestFixt.cset));
}

TEST(MessageOutputTest, testPublish)
{
    MessageOutputTestFixture msgTestFixt;

    std::string blob;
    EXPECT_CALL(*msgTestFixt.client, sendMsgV(_, _, _))
        .WillOnce(testing::DoAll(SaveArg<1>(&blob), Return(SimpleComms::Status::Ok)));

    av::msg::SampleMessage in;
    in.value = 123;
    task::MessageOutput<av::msg::SampleMessage> mo("", *msgTestFixt.cset);
    EXPECT_NO_THROW(mo.publish(in));

    // Get the received data and check that it is effectively the published data
    interfaces::baseTypes::InterfaceHeader header;
    auto headerStr = std::string(reinterpret_cast<char*>(&header), sizeof(header));
    blob.erase(0, headerStr.size());

    av::msg::SampleMessage out;
    aptv::deserializeODatum(out, blob);
    EXPECT_EQ(in.value, out.value);
}

TEST(MessageOutputTest, testPublishSetMessageCounter)
{
    MessageOutputTestFixture msgTestFixt;

    std::string blob;
    EXPECT_CALL(*msgTestFixt.client, sendMsgV(_, _, _))
        .WillOnce(testing::DoAll(SaveArg<1>(&blob), Return(SimpleComms::Status::Ok)));

    av::msg::SampleMessage in;
    in.value = 1234;
    task::MessageOutput<av::msg::SampleMessage> mo("", *msgTestFixt.cset);

    // Publish message using a specific messageCounter value - used for reproducible execution
    uint32_t messageCounter = 4321;
    EXPECT_NO_THROW(mo.publish(messageCounter, in));

    // NOLINTNEXTLINE(cppcoreguidelines-pro-type-reinterpret-cast)
    const auto& recvHeader = *reinterpret_cast<const interfaces::baseTypes::InterfaceHeader*>(blob.data());
    EXPECT_EQ(ntohl(recvHeader.headerVersion), OTTO_INTERFACE_HEADER_VERSION);
    EXPECT_EQ(ntohl(recvHeader.messageCounter), messageCounter);
}
