#include <gtest/gtest-param-test.h>
#include <gtest/gtest.h>

#include <SimpleComms/Constants.h>
#include <SimpleComms/ControlMessages.h>
#include <interfaces/SampleMessage/SampleMessage.h>
#include <interfaces/baseTypes/Header.h>
#include <task/CallbackSet.h>
#include <task/MessageQueue.h>

#include <chrono>
#include <condition_variable>
#include <memory>
#include <mutex>
#include <string>
#include <thread>

namespace {
// Name of the channel.
const std::string SAMPLE_CHANNEL = "Sample";
const std::string FLUSH_CHANNEL = "Flush";

// A MessageQueue<> should only ever be used in the executive function.
// To test that, use this object to call get() in the executive.
class TaskStub
{
public:
    explicit TaskStub(task::CallbackSet& callbacks, size_t size = 0)
        : m_input(SAMPLE_CHANNEL, callbacks, size)
        , m_size(size)
    {
        callbacks.add("", this, &TaskStub::executive);
    }

    void expect(const std::vector<int>& values)
    {
        std::unique_lock<std::mutex> lock(m_mutex);
        // If given more values than the queue capacity, we only expect to
        // receive the most recent values.
        if (m_size && values.size() > m_size) {
            auto start = std::prev(values.end(), static_cast<std::vector<int>::difference_type>(m_size));
            m_expected.assign(start, values.end());
        } else {
            m_expected = values;
        }
        m_received = false;
    }

    void executive()
    {
        {
            std::unique_lock<std::mutex> lock(m_mutex);
            for (int value : m_expected) {
                ASSERT_FALSE(m_input.empty());
                EXPECT_EQ(m_input.pop().value, value);
            }
            m_received = true;
        }
        m_condition.notify_all();
    }

    // Wait until the callback is done.
    void wait()
    {
        std::unique_lock<std::mutex> lock(m_mutex);
        m_condition.wait(lock, [this] { return m_received; });
    }

private:
    task::MessageQueue<av::msg::SampleMessage> m_input;
    std::vector<int> m_expected;
    size_t m_size;
    bool m_received{true};
    std::mutex m_mutex;
    std::condition_variable m_condition;
};

std::string makeSerializedSample(int value)
{
    av::msg::SampleMessage payload;
    payload.value = value;

    std::string blob;
    interfaces::baseTypes::InterfaceHeader header;
    header.headerVersion = htonl(OTTO_INTERFACE_HEADER_VERSION);
    aptv::serializeODatumMessage(std::string(reinterpret_cast<char*>(&header), sizeof(header)), blob, payload);

    return blob;
}

std::string makeSerializedExecutive()
{
    SimpleComms::ExecutiveMessage exeMsg;
    return std::string(reinterpret_cast<const char*>(&exeMsg), sizeof(exeMsg));
}

class Waiter
{
public:
    void receive()
    {
        {
            std::unique_lock lock(mutex_);
            done_ = true;
        }
        condition_.notify_all();
    }
    void wait()
    {
        std::unique_lock lock(mutex_);
        condition_.wait_for(lock, std::chrono::seconds(1), [this] { return done_; });
        ASSERT_TRUE(done_);
        done_ = false;
    }

protected:
    std::condition_variable condition_;
    std::mutex mutex_;
    bool done_{false};
};
class CallbackSetWaiter : public Waiter
{
public:
    CallbackSetWaiter(task::CallbackSet& cs) { cs.add(FLUSH_CHANNEL, this, &CallbackSetWaiter::handler); }
    void handler(const av::msg::SampleMessage&) { receive(); }
};
class EventFrameworkWaiter : public Waiter
{
public:
    EventFrameworkWaiter(av::repro::EventFramework& framework)
    {
        framework.setMessageHandler(FLUSH_CHANNEL, this, &EventFrameworkWaiter::handler);
    }
    void handler(const av::repro::Inputs<av::msg::SampleMessage>&) { receive(); }
};

// Perform a set of tests with a given capacity.
void doTests(const std::vector<std::vector<int>>& tests, size_t capacity)
{
    av::repro::EventFramework framework;
    task::CallbackSet cs(framework);
    TaskStub task(cs, capacity);

    for (const std::vector<int>& test : tests) {
        // Tell the stub what messages to expect.
        task.expect(test);
        // Send the messages.
        for (int value : test) {
            cs.call(SAMPLE_CHANNEL, makeSerializedSample(value), task::CallbackSet::MessageKey());
        }
        // Call the executive function.
        cs.call(SIMPLECOMMS_CONTROL_CHANNEL, makeSerializedExecutive(), task::CallbackSet::MessageKey());
        // Wait for the executive to finish.
        task.wait();
    }

    // Note: in actual code, we rely on Task::cleanup() to call callbacks_.stop().
    // Make sure ThreadSanitizer knows the thread is done at this point.
    ASSERT_NO_THROW(cs.stop());
}

} // namespace

TEST(MessageQueue, fromCallbackSet)
{
    av::repro::EventFramework framework;
    task::CallbackSet cs(framework);
    task::MessageQueue<av::msg::SampleMessage> queue(SAMPLE_CHANNEL, cs, 10);
    EXPECT_TRUE(queue.empty());

    CallbackSetWaiter waiter(cs);

    cs.call(SAMPLE_CHANNEL, makeSerializedSample(1), task::CallbackSet::MessageKey());
    cs.call(FLUSH_CHANNEL, makeSerializedSample(0), task::CallbackSet::MessageKey());
    waiter.wait();

    EXPECT_FALSE(queue.empty());
    EXPECT_EQ(queue.pop().value, 1);
    EXPECT_TRUE(queue.empty());

    cs.call(SAMPLE_CHANNEL, makeSerializedSample(2), task::CallbackSet::MessageKey());
    cs.call(SAMPLE_CHANNEL, makeSerializedSample(3), task::CallbackSet::MessageKey());
    cs.call(FLUSH_CHANNEL, makeSerializedSample(0), task::CallbackSet::MessageKey());
    waiter.wait();

    EXPECT_FALSE(queue.empty());
    EXPECT_EQ(queue.pop().value, 2);
    EXPECT_FALSE(queue.empty());
    EXPECT_EQ(queue.pop().value, 3);
    EXPECT_TRUE(queue.empty());

    // Note: in actual code, we rely on Task::cleanup() to call callbacks_.stop().
    // Make sure ThreadSanitizer knows the thread is done at this point.
    ASSERT_NO_THROW(cs.stop());
}

TEST(MessageQueue, fromFramework)
{
    av::repro::EventFramework framework;
    framework.init(av::repro::EventFramework::Mode::live);
    framework.start();
    task::MessageQueue<av::msg::SampleMessage> queue(SAMPLE_CHANNEL, framework, 10);
    EXPECT_TRUE(queue.empty());

    EventFrameworkWaiter waiter(framework);

    framework.receive(makeSerializedSample(1), SAMPLE_CHANNEL, "origin");
    framework.receive(makeSerializedSample(0), FLUSH_CHANNEL, "origin");
    waiter.wait();

    EXPECT_FALSE(queue.empty());
    EXPECT_EQ(queue.pop().value, 1);
    EXPECT_TRUE(queue.empty());

    framework.receive(makeSerializedSample(2), SAMPLE_CHANNEL, "origin");
    framework.receive(makeSerializedSample(3), SAMPLE_CHANNEL, "origin");
    framework.receive(makeSerializedSample(0), FLUSH_CHANNEL, "origin");
    waiter.wait();

    EXPECT_FALSE(queue.empty());
    EXPECT_EQ(queue.pop().value, 2);
    EXPECT_FALSE(queue.empty());
    EXPECT_EQ(queue.pop().value, 3);
    EXPECT_TRUE(queue.empty());

    // Note: in actual code, we rely on Task::cleanup() to call callbacks_.stop().
    // Make sure ThreadSanitizer knows the thread is done at this point.
    ASSERT_NO_THROW(framework.quit());
}

TEST(MessageQueue, popNoThrow)
{
    std::vector<std::vector<int>> tests = {{}, {1}, {2, 3, 4}, {5, 6, 7, 8, 9}, {}};
    for (size_t capacity = 0; capacity < 3; ++capacity) {
        doTests(tests, capacity);
    }
}
