/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines a testing object with mocked components for Task class
 */

#ifndef MOCK_TESTING_TASK_H
#define MOCK_TESTING_TASK_H

#include "MockBackupHandler.h"
#include "MockHeartbeatSender.h"
#include <task/Task.h>

#include <memory>
#include <string>

/**
 * \brief This implements a testing class for Task
 */
class MockTestingTask : public task::Task
{
public:
    /**
     * \brief Default constructor, copy and assignment constructors are deleted
     */
    MockTestingTask() = delete;
    MockTestingTask(const MockTestingTask&) = delete;
    MockTestingTask& operator=(const MockTestingTask&) = delete;
    MockTestingTask(const MockTestingTask&&) = delete;
    MockTestingTask& operator=(const MockTestingTask&&) = delete;

    /**
     * \brief User-defined constructor
     *
     * @param taskName The task name
     */
    explicit MockTestingTask(const std::string& taskName)
        : Task(taskName)
    {}

    /**
     * @brief Default destructor
     */
    virtual ~MockTestingTask() = default;

    /**
     * \brief Initializes the task
     *
     * \param Void
     *
     * \return Void
     */
    inline bool initialize() override { return initializeResult_; }

    /**
     * \brief Execute a task cycle
     *
     * \param Void
     *
     * \return True if the execution is fine, false, otherwise
     */
    inline bool executive() override { return executiveResult_; }

    /**
     * \brief Cleans up when task stops
     *
     * \param Void
     *
     * \return Void
     */
    inline void cleanup() override {}

    /**
     * \brief Set the initialization result
     *
     * \param The result to be set
     *
     * \return Void
     */
    inline void setInitResult(const bool result) { initializeResult_ = result; }

    /**
     * \brief Set the cycle execution result
     *
     * \param The result to be set
     *
     * \return Void
     */
    inline void setExeResult(const bool result) { executiveResult_ = result; }

    /**
     * \brief Set the backup recovery handler pointer
     *
     * \param Void
     *
     * \return The pointer to the handler
     */
    inline void setBackupHandler(std::unique_ptr<task::BackupRecoveryHandlerInterface> bkHandler)
    {
        bkHandler_ = std::move(bkHandler);
    }

    /**
     * \brief Create a backup recovery handler. This method creates a standard handler for all AV tasks. It can be
     * overrode by a sub-class to have a specific handler. Thus, this follows the design pattern Factory Method
     *
     * \param Void
     *
     * \return A pointer to the created handler object
     */
    inline std::unique_ptr<task::BackupRecoveryHandlerInterface> createBackupRecoveryHandler() override
    {
        return std::move(bkHandler_);
    }

    /**
     * \brief Set the heartbeat sender pointer
     *
     * \param Void
     *
     * \return The pointer to the sender
     */
    inline void setHeartbeatSender(std::unique_ptr<task::HeartbeatSenderInterface> hbSender)
    {
        hbSender_ = std::move(hbSender);
    }

    /**
     * \brief Create a heartbeat sender. This method creates a standard heartbeat sender for all AV tasks. It can be
     * overrode by a sub-class to have a specific sender. Thus, this follows the design pattern Factory Method
     *
     * \param Void
     *
     * \return A pointer to the created sender object
     */
    inline std::unique_ptr<task::HeartbeatSenderInterface> createHeartbeatSender() override
    {
        return std::move(hbSender_);
    }

    /**
     * \brief Set the task scheduler pointer
     *
     * \param Void
     *
     * \return The pointer to the scheduler
     */
    inline void setTaskScheduler(std::unique_ptr<scheduler::TaskSchedulerInterface> taskScheduler)
    {
        taskScheduler_ = std::move(taskScheduler);
    }

    /**
     * \brief Create a task scheduler. This method creates a standard scheduler or simulation scheduler for all
     * AV tasks. It can be overrode by a sub-class to have a specific handler, for example testing. Thus, this
     * follows the design pattern Factory Method
     *
     * \param Void
     *
     * \return A pointer to the created handler object
     */
    inline std::unique_ptr<scheduler::TaskSchedulerInterface> createTaskScheduler() override
    {
        return std::move(taskScheduler_);
    }

private:
    bool initializeResult_{true}; ///< The result of the task initialization
    bool executiveResult_{true};  ///< The result of the task cycle execution
    std::unique_ptr<task::BackupRecoveryHandlerInterface> bkHandler_{nullptr}; ///< The testing handler
    std::unique_ptr<task::HeartbeatSenderInterface> hbSender_{nullptr};        ///< The testing handler
};

#endif // MOCk_TESTING_TASK_H
