/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines unit tests for Task
 */

#include "MockTestingTask.h"
#include "TestingTask.h"
#include "platform/middleware/libs/Clock/include/Clock/SystemClock.h"
#include "platform/middleware/libs/ProcessLauncher/include/ProcessLauncher/processUtil.h"
#include "platform/middleware/libs/TestUtils/include/TestUtils/Common.hpp"
#include "platform/middleware/libs/TestUtils/include/TestUtils/CopyInterfaces.hpp"
#include "platform/middleware/libs/TestUtils/include/TestUtils/CwdBackup.hpp"
#include "platform/middleware/libs/TestUtils/include/TestUtils/EnvBackup.hpp"
#include "platform/middleware/libs/task/TaskUtil.h"

#include <interfaces/Generic/Input/Channel/Impl.h>
#include <interfaces/Generic/Output/Channel/Impl.h>
#include <interfaces/SampleMessage/SampleMessage.h>

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <memory>
#include <string>

using ::testing::_;
using ::testing::Exactly;
using ::testing::Return;
using namespace task;
using namespace scheduler;
using namespace ottotime;
using ::TestUtils::CopyInterfaces;
using ::TestUtils::CwdBackup;
using ::TestUtils::EnvBackup;

/**
 * \brief Internal constants of the task framework
 */
namespace {
constexpr int unknownError{10001};                        ///< The unknown error code
const std::string configDirEnv{"OTTO_CONFIG_DIR"};        ///< The configuration directory environment variable
const std::string libDirEnv{"OTTO_DIR"};                  ///< The AV stack base directory environment variable
const std::string ldLibraryPathEnv{"LD_LIBRARY_PATH"};    ///< The LD_LIBRARY_PATH environment variable
const std::string testIntfLibDirEnv{"OTTO_TEST_LIB_DIR"}; ///< The path the interface lib directory environment variable
const std::string testingTaskName{"TestingTask"};         ///< The testing task name
const std::string invalidHostName{"invalidTestingName"};  ///< The invalid host name for testing
const std::string noParameterConfigFile{"NoParameters"};  ///< No parameter config file name
const std::string invalidReplicasPrimaryConfigFile{
    "InvalidReplicas_primary"}; ///< The invalid replicas parameter config file for primary task
const std::string invalidStandbyPrimaryConfigFile{
    "InvalidStandby_primary"}; ///< The invalid standby parameter config file for primary task
const std::string invalidReplicasBackupConfigFile{
    "InvalidReplicas_backup"}; ///< The invalid replicas parameter config file for backup task
const std::string invalidStandbyBackupConfigFile{
    "InvalidStandby_backup"}; ///< The invalid standby parameter config file for backup task
const std::string initInterfaceFailed{
    "InitInterfaceFailed"}; ///< The config file such that interfaces initialization is failed
} // namespace

/**
 * \brief Fake function for Task get method
 */
Task* Task::get()
{
    return nullptr;
}

/**
 * \brief This class defines unit tests for Task class
 */
class TaskBaseClassTest : public ::testing::Test
{
public:
    /**
     * \brief Set up the tests
     */
    void SetUp() override
    {
        // Init the system clock
        SystemClock::initialize(SystemClock::SystemTimeMode::TimeModeSteady);
        // Check the system clock is initialized
        ASSERT_TRUE(SystemClock::isInitialized());
        // Copy the interface lib files into testint location"};
        intfLibOk_ = copyInterfacesEnv.copyInterface("/SampleMessage/Input/Channel", "/SampleMessageInput");
        intfLibOk_ &= copyInterfacesEnv.copyInterface("/SampleMessage/Output/Channel", "/SampleMessageOutput");
        intfLibOk_ &= copyInterfacesEnv.copyInterface("/SaferCommand/Input/Channel", "/SaferCommandInput");
        intfLibOk_ &= copyInterfacesEnv.copyInterface("/SaferResponse/Output/Channel", "/SaferResponseOutput");
        intfLibOk_ &= copyInterfacesEnv.copyInterface("/ProcessHealth/Output/Channel", "/ProcessHealthOutput");

        ASSERT_TRUE(intfLibOk_);
        // Get the hostname
        std::string hostName{""};
        if (getHostMachineName(hostName)) {
            hostName_ = hostName;
        }
        // Create a valid config section
        createConfigSection();
        // Create an invalid config section
        createInvalidConfigSection();
    }

    /**
     * \brief Clean up the tests
     */
    void TearDown() override
    {
        testingTask_._cleanup();
        mockTestingTask_._cleanup();
    }

    /**
     * \brief Test the chmodInterfaces
     *
     * \param Void
     *
     * \return Void
     */
    void testChmod()
    {
        std::list<ConfigSection> configList{};
        // Add valid config interface
        configList.push_back(configSection_);
        ASSERT_EQ(configList.size(), 1);
        ASSERT_TRUE(testingTask_.testChmodInterfaces(configList, Access::Standby));
        // Add invalid config interface
        configList.push_back(invalidConfigSection_);
        ASSERT_EQ(configList.size(), 2);
        ASSERT_FALSE(testingTask_.testChmodInterfaces(configList, Access::Standby));
    }

protected:
    /**
     * \brief Create a valid config section
     *
     * \param Void
     *
     * \return Void
     */
    void createConfigSection()
    {
        const std::string interfaceName{"ProcessHealthOutput"};
        const std::string configString{
            interfaceName + " = {interface = \"" + interfaceName + "\"; " +
            "implementation = \"Channel\"; channel = {type = \"LOG\" queueSize = 100; name = \"ProcessHealth\";};};"};
        // Parse the config string
        ConfigParser parser{};
        EXPECT_TRUE(parser.parseString(configString));
        EXPECT_TRUE(parser.getSection(interfaceName, configSection_));
    }

    /**
     * \brief Create an invalid config section
     *
     * \param Void
     *
     * \return Void
     */
    void createInvalidConfigSection()
    {
        const std::string interfaceName{"ProcessHealthOutput"};
        const std::string configString{interfaceName + " = {implementation = \"Channel\"; channel = {type = \"LOG\" "
                                                       "queueSize = 100; name = \"ProcessHealth\";};};"};
        // Parse the config string
        ConfigParser parser{};
        EXPECT_TRUE(parser.parseString(configString));
        EXPECT_TRUE(parser.getSection(interfaceName, invalidConfigSection_));
    }

    /**
     * \brief Set-up the testing task with mocked components
     *
     * \param Void
     *
     * \return Void
     */
    void setupMockTestingTask()
    {
        // Backup recovery handler
        auto mockBkHandler = std::make_unique<MockBackupHandler>();
        EXPECT_NE(nullptr, mockBkHandler);
        // For EXPECT_CALL, the life-time is determined by mockTestingTask_ life-time
        mockBkHandlerPtr_ = mockBkHandler.get();
        mockTestingTask_.setBackupHandler(std::move(mockBkHandler));
        EXPECT_NE(nullptr, mockBkHandlerPtr_);
        // Heartbeat sender
        auto mockHbSender = std::make_unique<MockHeartbeatSender>();
        EXPECT_NE(nullptr, mockHbSender);
        // For EXPECT_CALL, the life-time is determined by mockTestingTask_ life-time
        mockHbSenderPtr_ = mockHbSender.get();
        mockTestingTask_.setHeartbeatSender(std::move(mockHbSender));
        EXPECT_NE(nullptr, mockHbSenderPtr_);
        // Task scheduler
        auto mockTaskScheduler = std::make_unique<MockTaskScheduler>();
        EXPECT_NE(nullptr, mockTaskScheduler);
        // For EXPECT_CALL, the life-time is determined by mockTestingTask_ life-time
        mockSchedulerPtr_ = mockTaskScheduler.get();
        mockTestingTask_.setTaskScheduler(std::move(mockTaskScheduler));
        EXPECT_NE(nullptr, mockSchedulerPtr_);
    }

    std::string hostName_{invalidHostName};            ///< The host name
    bool intfLibOk_{false};                            ///< True if the interface lib files are copied succesfully
    ConfigSection configSection_{};                    ///< The config section for testing
    ConfigSection invalidConfigSection_{};             ///< The invalid config section for testing
    TestingTask testingTask_{testingTaskName};         ///< The testing task object
    MockTestingTask mockTestingTask_{testingTaskName}; ///< The testing task object with mocked components
    MockBackupHandler* mockBkHandlerPtr_{nullptr};     ///< The mocked backup handler pointer
    MockHeartbeatSender* mockHbSenderPtr_{nullptr};    ///< The mocked heartbeat sender pointer
    MockTaskScheduler* mockSchedulerPtr_{nullptr};     ///< The mocked scheduler pointer
    logger::Logger logger_{};                          ///< The logger with defautl level
    CopyInterfaces copyInterfacesEnv{"TaskTest"};
};

TEST_F(TaskBaseClassTest, getTaskName)
{
    ASSERT_EQ(testingTask_.getTaskName(), testingTaskName);
    ASSERT_EQ(mockTestingTask_.getTaskName(), testingTaskName);
}

TEST_F(TaskBaseClassTest, setTaskName)
{
    // Empty input does not affect
    testingTask_.setTaskName("");
    ASSERT_EQ(testingTask_.getTaskName(), testingTaskName);
    // Non-empty input makes the name changed
    const std::string newName{"newTestingTask"};
    testingTask_.setTaskName(newName);
    ASSERT_EQ(testingTask_.getTaskName(), newName);
}

TEST_F(TaskBaseClassTest, setGlobalLogLevel)
{
    // Set invalid logging level
    ASSERT_FALSE(testingTask_.setGlobalLogLevel("none"));
    // Set the logging level to notice
    ASSERT_TRUE(testingTask_.setGlobalLogLevel("info"));
    logger_ = testingTask_.getLogger();
    ASSERT_EQ(logger_.getThreshold(), Priority::PriorityLevel::info);
}

TEST_F(TaskBaseClassTest, getConfigFileDirectory)
{
    // OTTO_CONFIG_DIR is not defined
    {
        const std::string defaultConfigDir{"."};
        // Unset the OTTO_CONFIG_DIR
        EnvBackup envBackup{configDirEnv.c_str(), ""};
        ASSERT_EQ(testingTask_.getConfigFileDirectory(), defaultConfigDir);
    }
    // OTTO_CONFIG_DIR is defined
    {
        const std::string definedConfigDir{"./config"};
        EnvBackup envBackup{configDirEnv.c_str(), definedConfigDir};
        ASSERT_EQ(testingTask_.getConfigFileDirectory(), definedConfigDir);
    }
}

TEST_F(TaskBaseClassTest, getUCDirectory)
{
    // OTTO_DIR is not defined
    {
        const std::string defaultOttoDir{"."};
        // Unset the OTTO_DIR
        EnvBackup envBackup{libDirEnv.c_str(), ""};
        ASSERT_EQ(testingTask_.getUCDirectory(), defaultOttoDir);
    }
    // OTTO_DIR is defined
    {
        const std::string definedOttoDir{"./"};
        EnvBackup envBackup{libDirEnv.c_str(), definedOttoDir};
        ASSERT_EQ(testingTask_.getUCDirectory(), definedOttoDir);
    }
}

TEST_F(TaskBaseClassTest, getHostName)
{
    if (hostName_ != invalidHostName) {
        ASSERT_EQ(testingTask_.getHostName(), hostName_);
    }
}

TEST_F(TaskBaseClassTest, createHeartbeatSender)
{
    // Set OTTO_CONFIG_DIR to the config directory
    EnvBackup configEnv{configDirEnv.c_str(), TestUtils::getTestingDataPath()};
    // Test createHeartbeatSender() is fine
    ASSERT_TRUE(testingTask_.testCreateHeartbeatSender());
    testingTask_._cleanup();
}

TEST_F(TaskBaseClassTest, createBackupRecoveryHandler)
{
    // Set OTTO_CONFIG_DIR to the config directory
    EnvBackup configEnv{configDirEnv.c_str(), TestUtils::getTestingDataPath()};
    // Test createHeartbeatSender() is fine
    ASSERT_TRUE(testingTask_.testCreateBackupRecoveryHandler());
    testingTask_._cleanup();
}

TEST_F(TaskBaseClassTest, getPublishedMessage)
{
    const std::string invalidIntfName{"NoName"};
    // Set OTTO_CONFIG_DIR to the config directory
    EnvBackup envBackup{configDirEnv.c_str(), TestUtils::getTestingDataPath()};
    ASSERT_TRUE(testingTask_.testInitParameters());

    testingTask_.setTaskName(testingTaskName);
    testingTask_._cleanup();
    // Everything is ok
    testingTask_.setInitResult(true);
    ASSERT_TRUE(testingTask_._initialize());

    ASSERT_TRUE(testingTask_.testGetPublishListMessage());

    testingTask_._cleanup();
}

TEST_F(TaskBaseClassTest, initParameters)
{
    // initParameters() returns false because of parsing config file
    ASSERT_FALSE(testingTask_.testInitParameters());
    // Set OTTO_CONFIG_DIR to the config directory
    EnvBackup envBackup{configDirEnv.c_str(), TestUtils::getTestingDataPath()};
    // initParameters() returns false because of parsing parameters section
    testingTask_.setTaskName(noParameterConfigFile);
    ASSERT_FALSE(testingTask_.testInitParameters());
    // initParameters() returns false because of parsing recovery parameters
    // 1. Primary task with invalid replicas parameter
    testingTask_.setTaskName(invalidReplicasPrimaryConfigFile);
    ASSERT_FALSE(testingTask_.testInitParameters());
    // 2. Primary task with invalid standby parameter
    testingTask_.setTaskName(invalidStandbyPrimaryConfigFile);
    ASSERT_FALSE(testingTask_.testInitParameters());
    // 3. Backup task with invalid replicas parameter
    testingTask_.setTaskName(invalidReplicasBackupConfigFile);
    ASSERT_FALSE(testingTask_.testInitParameters());
    // 4. Backup task with invalid standby parameter
    testingTask_.setTaskName(invalidStandbyBackupConfigFile);
    ASSERT_FALSE(testingTask_.testInitParameters());
    // Everything is ok
    testingTask_.setTaskName(testingTaskName);
    ASSERT_TRUE(testingTask_.testInitParameters());
}

TEST_F(TaskBaseClassTest, initInterfaces)
{
    // Set OTTO_CONFIG_DIR to the config directory
    EnvBackup envBackup{configDirEnv.c_str(), TestUtils::getTestingDataPath()};
    testingTask_.setTaskName(initInterfaceFailed);
    ASSERT_TRUE(testingTask_.testInitParameters());
    // interfaces initialization fails
    ASSERT_FALSE(testingTask_.testInitInterfaces());
    testingTask_._cleanup();
    // Everything is ok
    testingTask_.setTaskName(testingTaskName);
    ASSERT_TRUE(testingTask_.testInitParameters());
    ASSERT_TRUE(testingTask_.testInitInterfaces());
    // Test chmodInterfaces
    testChmod();
    testingTask_._cleanup();
}

TEST_F(TaskBaseClassTest, getInterface)
{
    const std::string invalidIntfName{"NoName"};
    // Set OTTO_CONFIG_DIR to the config directory
    EnvBackup envBackup{configDirEnv.c_str(), TestUtils::getTestingDataPath()};
    ASSERT_TRUE(testingTask_.testInitParameters());
    ASSERT_FALSE(testingTask_.testGetReadOnlyInterface<GenericInputChannel<av::msg::SampleMessage>>(invalidIntfName));
    ASSERT_TRUE(testingTask_.testGetReadOnlyInterface<GenericInputChannel<av::msg::SampleMessage>>("SampleInput"));
    ASSERT_FALSE(testingTask_.testGetWriteOnlyInterface<GenericOutputChannel<av::msg::SampleMessage>>(invalidIntfName));
    ASSERT_TRUE(testingTask_.testGetWriteOnlyInterface<GenericOutputChannel<av::msg::SampleMessage>>("SampleOutput"));
    testingTask_._cleanup();
}

TEST_F(TaskBaseClassTest, getSystemErrorCodeText)
{
    std::string errorStr = systemErrorCodeToString(EIO);
    std::string expectedResult{"Input/output error"};
    EXPECT_EQ(errorStr, expectedResult);
    // unknown error code
    errorStr = systemErrorCodeToString(unknownError);
    expectedResult = "Unknown error " + std::to_string(unknownError);
    EXPECT_EQ(errorStr, expectedResult);
}

TEST_F(TaskBaseClassTest, getSystemTime)
{
    // Task::get() returns null, expect system time from SystemClock::getSystemTime()
    otimePoint systemTime = testingTask_.getSystemTime();
    ASSERT_TRUE(systemTime.isTimeValid());
}

TEST_F(TaskBaseClassTest, _initialize)
{
    // initParameters() returns false
    ASSERT_FALSE(testingTask_._initialize());
    // Set OTTO_CONFIG_DIR to the config directory
    EnvBackup envBackup{configDirEnv.c_str(), TestUtils::getTestingDataPath()};
    testingTask_.setTaskName(initInterfaceFailed);
    // initInterfaces() returns false
    ASSERT_FALSE(testingTask_._initialize());
    testingTask_._cleanup();
    testingTask_.setTaskName(testingTaskName);
    // initialize() returns false
    testingTask_.setInitResult(false);
    ASSERT_FALSE(testingTask_._initialize());
    testingTask_._cleanup();
    // Everything is ok
    testingTask_.setInitResult(true);
    ASSERT_TRUE(testingTask_._initialize());
    // Test getTaskConfig
    auto taskParams = testingTask_.getTaskConfig();
    double freq_hz{0};
    if (taskParams.get("cycleRate_hz", freq_hz, 0)) {
        EXPECT_EQ(freq_hz, 1.0);
    }
    testingTask_._cleanup();
}

/**
 * Test methods with mocked heartbeat sender,
 * mocked backup recovery handler, and mocked task scheduler
 */
TEST_F(TaskBaseClassTest, hbSenderAndSchedulerAPIs)
{
    const std::string testResponse{"Test Reponse"};
    const int64_t executionTime{10};
    const timespec aTime{1, 1000};
    const otimePoint currentTime{aTime};
    const unsigned int executionInterval{9U};
    // Setup the mocked testing task
    setupMockTestingTask();
    // Set OTTO_CONFIG_DIR to the config directory
    EnvBackup envBackup{configDirEnv.c_str(), TestUtils::getTestingDataPath()};
    // Task initialization
    EXPECT_CALL(*mockBkHandlerPtr_, initialize()).Times(Exactly(1)).WillOnce(Return(true));
    EXPECT_CALL(*mockHbSenderPtr_, setHeartbeatMessage(_)).Times(Exactly(1));
    ASSERT_TRUE(mockTestingTask_._initialize());
    // sendSaferResponse() returns false
    EXPECT_CALL(*mockBkHandlerPtr_, sendSaferResponse(_)).Times(Exactly(1)).WillOnce(Return(false));
    ASSERT_FALSE(mockTestingTask_.sendStatusReport(testResponse));
    // sendSaferResponse() returns true
    EXPECT_CALL(*mockBkHandlerPtr_, sendSaferResponse(_)).Times(Exactly(1)).WillOnce(Return(true));
    ASSERT_TRUE(mockTestingTask_.sendStatusReport(testResponse));
    // getCurrentTime() returns currentTime
    EXPECT_CALL(*mockSchedulerPtr_, getCurrentTime()).Times(Exactly(1)).WillOnce(Return(currentTime));
    EXPECT_EQ(mockTestingTask_.getCurrentTaskTime(), currentTime);
    mockTestingTask_._cleanup();
}

TEST_F(TaskBaseClassTest, _executive)
{
    // Components are null
    ASSERT_FALSE(mockTestingTask_._executive());
    // Setup the mocked testing task
    setupMockTestingTask();
    // Set OTTO_CONFIG_DIR to the config directory
    EnvBackup envBackup{configDirEnv.c_str(), TestUtils::getTestingDataPath()};
    // Task initialization
    EXPECT_CALL(*mockBkHandlerPtr_, initialize()).Times(Exactly(1)).WillOnce(Return(true));
    EXPECT_CALL(*mockHbSenderPtr_, setHeartbeatMessage(_)).Times(Exactly(1));
    ASSERT_TRUE(mockTestingTask_._initialize());
    // executive() returns false
    mockTestingTask_.setExeResult(false);
    EXPECT_CALL(*mockSchedulerPtr_, handlePreScheduling()).Times(Exactly(1));
    ASSERT_FALSE(mockTestingTask_._executive());
    // executive() returns true, needToSendHeartbeat() returns false
    mockTestingTask_.setExeResult(true);
    EXPECT_CALL(*mockSchedulerPtr_, handlePreScheduling()).Times(Exactly(1));
    EXPECT_CALL(*mockSchedulerPtr_, doScheduling()).Times(Exactly(1));
    EXPECT_CALL(*mockHbSenderPtr_, needToSendHeartbeat()).Times(Exactly(1)).WillOnce(Return(false));
    EXPECT_CALL(*mockSchedulerPtr_, getRequiresSteadyClock()).Times(Exactly(1));
    EXPECT_CALL(*mockBkHandlerPtr_, handleBackupRecovery(_, _, _, _, _)).Times(Exactly(1)).WillOnce(Return(false));
    ASSERT_TRUE(mockTestingTask_._executive());
    // executive() returns true, needToSendHeartbeat() returns true
    mockTestingTask_.setExeResult(true);
    EXPECT_CALL(*mockSchedulerPtr_, handlePreScheduling()).Times(Exactly(1));
    EXPECT_CALL(*mockSchedulerPtr_, doScheduling()).Times(Exactly(1));
    EXPECT_CALL(*mockHbSenderPtr_, needToSendHeartbeat()).Times(Exactly(1)).WillOnce(Return(true));
    EXPECT_CALL(*mockHbSenderPtr_, sendHeartbeat()).Times(Exactly(1)).WillOnce(Return(true));
    EXPECT_CALL(*mockSchedulerPtr_, getRequiresSteadyClock()).Times(Exactly(1));
    EXPECT_CALL(*mockBkHandlerPtr_, handleBackupRecovery(_, _, _, _, _)).Times(Exactly(1)).WillOnce(Return(true));
    ASSERT_TRUE(mockTestingTask_._executive());
    mockTestingTask_._cleanup();
}
