/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines a testing object for Task class
 */

#ifndef TESTING_TASK_H
#define TESTING_TASK_H

#include "platform/middleware/libs/task/TaskUtil.h"
#include "platform/middleware/libs/task/include/task/Task.h"

#include <memory>
#include <string>

/**
 * \brief This implements a testing class for Task
 */
class TestingTask : public task::Task
{
public:
    /**
     * \brief Default constructor, copy and assignment constructors are deleted
     */
    TestingTask() = delete;
    TestingTask(const TestingTask&) = delete;
    TestingTask& operator=(const TestingTask&) = delete;
    TestingTask(const TestingTask&&) = delete;
    TestingTask& operator=(const TestingTask&&) = delete;

    /**
     * \brief User-defined constructor
     *
     * @param taskName The task name
     */
    explicit TestingTask(const std::string& taskName)
        : Task(taskName)
    {}

    /**
     * @brief Default destructor
     */
    virtual ~TestingTask() = default;

    /**
     * \brief Initializes the task
     *
     * \param Void
     *
     * \return Void
     */
    inline bool initialize() override { return initializeResult_; }

    /**
     * \brief Execute a task cycle
     *
     * \param Void
     *
     * \return True if the execution is fine, false, otherwise
     */
    inline bool executive() override { return executiveResult_; }

    /**
     * \brief Cleans up when task stops
     *
     * \param Void
     *
     * \return Void
     */
    inline void cleanup() override {}

    /**
     * \brief Set the initialization result
     *
     * \param The result to be set
     *
     * \return Void
     */
    inline void setInitResult(const bool result) { initializeResult_ = result; }

    /**
     * \brief Set the cycle execution result
     *
     * \param The result to be set
     *
     * \return Void
     */
    inline void setExeResult(const bool result) { executiveResult_ = result; }

    /**
     * \brief Help to test Task::initParameters() method
     *
     * \param Void
     *
     * \return True if the initialization is fine
     */
    inline bool testInitParameters() { return initParameters(getConfigFileDirectory()); }

    /**
     * \brief Help to test Task::initInterfaces() method
     *
     * \param Void
     *
     * \return True if the initialization is fine
     */
    inline bool testInitInterfaces() { return initInterfaces(); }

    /**
     * \brief Help to test Task::chmodInterfaces() method
     *
     * \param ifclist The interface list
     * \param acc The access mode (R/W/RW/Standby)
     *
     * \return True if the chmod is fine
     */
    inline bool testChmodInterfaces(ConfigSectionList ifclist, const task::Access& acc)
    {
        return chmodInterfaces(ifclist, acc);
    }

    /**
     * \brief Help to test Task::getWriteOnlyInterface() method
     *
     * \param interfaceName The name of the interface
     *
     * \return True if the interface is created well
     */
    template<typename InterfaceType>
    inline bool testGetWriteOnlyInterface(const std::string& interfaceName)
    {
        return (getWriteOnlyInterface<InterfaceType>(interfaceName) != nullptr);
    }

    /**
     * \brief Help to test Task::getReadOnlyInterface() method
     *
     * \param interfaceName The name of the interface
     *
     * \return True if the interface is created well
     */
    template<typename InterfaceType>
    inline bool testGetReadOnlyInterface(const std::string& interfaceName)
    {
        return (getReadOnlyInterface<InterfaceType>(interfaceName) != nullptr);
    }

    /**
     * \brief Help to test Task::createHeartbeatSender() method
     *
     * \param Void
     *
     * \return True if the sender is created succesfully
     */
    inline bool testCreateHeartbeatSender()
    {
        bool result{false};
        tp_.setTaskParamDirectory(getConfigFileDirectory());
        if (tp_.parseTask(taskName_)) {
            if (tp_.getTaskParameters(taskParams_)) {
                auto hbSender = createHeartbeatSender();
                result = (hbSender != nullptr);
            }
        }
        return result;
    }

    /**
     * \brief Help to test Task::createBackupRecoveryHandler() method
     *
     * \param Void
     *
     * \return True if the handler is created successfully
     */
    inline bool testCreateBackupRecoveryHandler()
    {
        bool result{false};
        tp_.setTaskParamDirectory(getConfigFileDirectory());
        if (tp_.parseTask(taskName_)) {
            if (tp_.getTaskParameters(taskParams_)) {
                auto bkHandler = createBackupRecoveryHandler();
                result = (bkHandler != nullptr);
            }
        }
        return result;
    }

    inline bool testGetPublishListMessage()
    {
        av::msg::PublishListMessage msg = getPublishListMessage();
        return msg.channels.empty();
    }

private:
    bool initializeResult_{true}; ///< The result of the task initialization
    bool executiveResult_{true};  ///< The result of the task cycle execution
};

#endif // TESTING_TASK_H
