/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief Defines unit tests for TaskUtil
 */

#include "TaskUtil.h"
#include "MockIDbObject.h"
#include <MockObjects.hpp>
#include <TestUtils/EnvBackup.hpp>
#include <config/ConfigParser.h>
#include <task/Task.h>

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <chrono>
#include <cstdlib>
#include <map>
#include <memory>
#include <thread>

using ::testing::_;
using ::testing::Exactly;
using ::testing::Return;
using ::TestUtils::EnvBackup;
using namespace task;
using namespace ottotime;

Task* Task::get(void)
{
    // A static variable inside a function is specified to be
    // initialized exactly once, in a thread-safe manner, the first
    // time the function is called.
    static task::TestUtils::MockTask mockTask;

    return &mockTask;
}

/**
 * \brief This class defines unit tests for TaskUtil class
 */
class TaskUtilTest : public ::testing::Test
{
public:
    /**
     * \brief Set up the tests
     */
    void SetUp() override
    {
        // Create a valid config section
        createConfigSection();
        // Create an invalid config section
        createInvalidConfigSection();
        // Create raw pointer interfaces map
        createRawPtrInterfaces();
        // Create shared pointer interfaces map
        createSharedPtrInterfaces();
    }

    /**
     * \brief Clean up the tests
     */
    void TearDown() override
    {
        for (auto& intf : interfaces_) {
            if (intf.second != nullptr) {
                delete intf.second;
                intf.second = nullptr;
            }
        }
    }

protected:
    /**
     * \brief Create a map from interface names to their raw pointers
     *
     * \param Void
     *
     * \return Void
     */
    void createRawPtrInterfaces()
    {
        IDbObject* processHealthOutputPtr = new MockIDbObject();
        ASSERT_NE(processHealthOutputPtr, nullptr);
        processHealthOutputPtr->chmod(Access::Write);
        interfaces_[interfaceName_] = processHealthOutputPtr;
    }

    /**
     * \brief Create a map from interface names to their shared pointers
     *
     * \param Void
     *
     * \return Void
     */
    void createSharedPtrInterfaces()
    {
        std::shared_ptr<IDbObject> processHealthOutputPtr = std::make_shared<MockIDbObject>();
        ASSERT_NE(processHealthOutputPtr, nullptr);
        processHealthOutputPtr->chmod(Access::Write);
        interfacesManaged_[interfaceName_] = processHealthOutputPtr;
    }

    /**
     * \brief Create a valid config section
     *
     * \param Void
     *
     * \return Void
     */
    void createConfigSection()
    {
        const std::string configString =
            interfaceName_ + " = {interface = \"" + interfaceName_ + "\"; " + "implementation = \"" + implName_ +
            "\"; channel = {type = \"SimpleComms\" queueSize = 100; name = \"ProcessHealth\";};};";
        // Parse the config string
        ConfigParser parser{};
        EXPECT_TRUE(parser.parseString(configString));
        EXPECT_TRUE(parser.getSection(interfaceName_, configSection_));
    }

    /**
     * \brief Create an invalid config section
     *
     * \param Void
     *
     * \return Void
     */
    void createInvalidConfigSection()
    {
        const std::string configString =
            interfaceName_ + " = {implementation = \"" + implName_ +
            "\"; channel = {type = \"SimpleComms\" queueSize = 100; name = \"ProcessHealth\";};};";
        // Parse the config string
        ConfigParser parser{};
        EXPECT_TRUE(parser.parseString(configString));
        EXPECT_TRUE(parser.getSection(interfaceName_, invalidConfigSection_));
    }

    int unknownError_{10001};                          ///< The unknown error code
    std::string interfaceName_{"ProcessHealthOutput"}; ///< The interface name
    std::string implName_{"Channel"};                  ///< The interface channel name
    std::string envName_{"MY_TASKUTIL_TEST"};          ///< The environment variable name for testing
    std::string envValue_{"MY_TASKUTIL_TEST_VALUE"};   ///< The environment variable value for testing
    ConfigSection configSection_{};                    ///< The config section for testing
    ConfigSection invalidConfigSection_{};             ///< The invalid config section for testing
    std::map<std::string, IDbObject*> interfaces_{};   ///< The list of interfaces
    std::map<std::string, std::shared_ptr<IDbObject>>
        interfacesManaged_{}; ///< The list of interfaces managed by shared ptr
};

TEST_F(TaskUtilTest, elapsedTimeInmilliSeconds)
{
    otimePoint startTime{timespec{0, 5900000}};
    otimePoint endTime{timespec{0, 10000000}};
    const long double expectedResult = 4.00;
    const long double elapsedTime_ms = elapsedTimeInmilliSeconds(startTime, endTime);
    AVLOG_INFO("Computed elapsed time: {}ms", AVLOG_ARGS("elapsed"), elapsedTime_ms);
    // The underlying otimeDuration::totalMilliSeconds cast to int64_t
    EXPECT_EQ(elapsedTime_ms, expectedResult);
}

TEST_F(TaskUtilTest, convertTomilliSeconds)
{
    otimePoint aTime{timespec{100, 5700100}};
    const long double expectedResult = 100005.00;
    const long double time_ms = convertTomilliSeconds(aTime);
    AVLOG_INFO("Computed time: {}ms", AVLOG_ARGS("time"), time_ms);
    // The underlying otimeDuration::totalMilliSeconds cast to int64_t
    EXPECT_EQ(time_ms, expectedResult);
}

TEST_F(TaskUtilTest, systemErrorCodeToString)
{
    std::string errorStr = systemErrorCodeToString(EIO);
    std::string expectedResult = "Input/output error";
    AVLOG_INFO("Computed error string: {}", AVLOG_ARGS("error"), errorStr);
    EXPECT_EQ(errorStr, expectedResult);
    // unknown error code
    errorStr = systemErrorCodeToString(unknownError_);
    AVLOG_INFO("Computed error string: {}", AVLOG_ARGS("error"), errorStr);
    expectedResult = "Unknown error " + std::to_string(unknownError_);
    EXPECT_EQ(errorStr, expectedResult);
}

TEST_F(TaskUtilTest, validateInterfaceConfig)
{
    std::string instanceName{""};
    std::string interfaceName{""};
    std::string implName{""};
    // Test valid interface config section
    ASSERT_TRUE(validateInterfaceConfig(configSection_, instanceName, interfaceName, implName));
    EXPECT_EQ(instanceName, interfaceName_);
    EXPECT_EQ(interfaceName, interfaceName_);
    EXPECT_EQ(implName, implName_);
    // Test invalid config section
    ASSERT_FALSE(validateInterfaceConfig(invalidConfigSection_, instanceName, interfaceName, implName));
}

TEST_F(TaskUtilTest, needRecoveryWithBackup)
{
    std::string primaryTaskName{"LocalizationTask"};
    const std::string backupTaskName{"LocalizationTask_backup"};
    ASSERT_FALSE(needRecoveryWithBackup(primaryTaskName, backupTaskName));
    // Update the primary task name w.r.t the convention
    primaryTaskName += "_primary";
    ASSERT_TRUE(needRecoveryWithBackup(primaryTaskName, backupTaskName));
}

TEST_F(TaskUtilTest, getEnvVariableValue)
{
    {
        const std::optional<std::string> envValue = getEnvVariableValue(envName_);
        ASSERT_FALSE(envValue);
    }

    // Set the envName_
    ASSERT_EQ(0, setenv(envName_.c_str(), envValue_.c_str(), 1));

    {
        const std::optional<std::string> envValue = getEnvVariableValue(envName_);
        ASSERT_TRUE(envValue);
        EXPECT_EQ(*envValue, envValue_);
    }
}

TEST_F(TaskUtilTest, getConfigFileDirectory)
{
    {
        EnvBackup ottoClearer("OTTO_CONFIG_DIR");
        EXPECT_EQ(".", Task::get()->getConfigFileDirectory());
    }

    {
        const std::string path{"/tmp/otto"};
        EnvBackup ottoClearer("OTTO_CONFIG_DIR", path);
        EXPECT_EQ(path, Task::get()->getConfigFileDirectory());
    }
}

TEST_F(TaskUtilTest, chmodTaskInterfaces)
{
    std::list<ConfigSection> configList{};
    // Add valid config interface
    configList.push_back(configSection_);
    ASSERT_EQ(configList.size(), 1);
    ASSERT_TRUE(chmodTaskInterfaces(configList, Access::Standby, interfaces_, interfacesManaged_));
    // Add invalid config interface
    configList.push_back(invalidConfigSection_);
    ASSERT_EQ(configList.size(), 2);
    ASSERT_FALSE(chmodTaskInterfaces(configList, Access::Standby, interfaces_, interfacesManaged_));
}

int main(int argc, char** argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
