/**
 * \copyright
 * Copyright (c) 2020- Motional. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Motional and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Motional and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Motional.
 *
 * \brief unit tests for VersionPrinter.hpp
 */

#include "platform/middleware/libs/task/include/task/VersionPrinter.h"
#include "platform/middleware/libs/TestUtils/include/TestUtils/CwdBackup.hpp"
#include "platform/middleware/libs/TestUtils/include/TestUtils/EnvBackup.hpp"
#include "platform/middleware/libs/task/TaskUtil.h"
#include <gtest/gtest.h>
#include <thread>

#include <string.h>

using namespace task;
using ::TestUtils::CwdBackup;
using ::TestUtils::EnvBackup;

namespace {

#ifndef VERSION_TEST_DIR
#define VERSION_TEST_DIR "platform/middleware/libs/task/test/VersionPrinterTest/data/config"
#endif
const std::string configPath = VERSION_TEST_DIR; //"./aptivware-core/src/task/test/VersionPrinterTest/data/config";

class ConfigPathBackup
{
public:
    ConfigPathBackup(const std::string& suffix = "")
        : m_configPath(configPath)
        , m_ottoConfigDir("OTTO_CONFIG_DIR", m_configPath + suffix)
        , m_activeConfigDir("ACTIVE_CONFIG_DIR", m_configPath + suffix + "/active")
    {}

    virtual ~ConfigPathBackup() = default;

private:
    std::string m_configPath;
    EnvBackup m_ottoConfigDir;
    EnvBackup m_activeConfigDir;
};

} // namespace

TEST(VersionPrinterTest, getVehicleName)
{
    ConfigPathBackup cpb;
    VersionPrinter vp;

    EXPECT_EQ("Model A", vp.getVehicleName());
}

TEST(VersionPrinterTest, getVehicleNameLogMode)
{
    ConfigPathBackup cpb;
    VersionPrinter vp;

    EXPECT_EQ("Model A", vp.getVehicleName(defaultVehicleConfigFile, true));
}

TEST(VersionPrinterTest, getVehicleNameEmptyConfigDir)
{
    // For the config dir actually to be empty, we have to clear out
    // the environment variables.
    EnvBackup ottoClearer("OTTO_CONFIG_DIR");
    EnvBackup activeClearer("ACTIVE_CONFIG_DIR");

    VersionPrinter vp;

    EXPECT_EQ("N/A", vp.getVehicleName());
}

TEST(VersionPrinterTest, getVehicleNameFileMissing)
{
    ConfigPathBackup cpb("/nonesuch");
    VersionPrinter vp;

    EXPECT_EQ("N/A", vp.getVehicleName());
}

TEST(VersionPrinterTest, getVehicleNameFileEmpty)
{
    ConfigPathBackup cpb("/empty");
    VersionPrinter vp;

    EXPECT_EQ("N/A", vp.getVehicleName());
}

TEST(VersionPrinterTest, getVehicleNameFirstLineEmpty)
{
    ConfigPathBackup cpb("/first-line-empty");
    VersionPrinter vp;

    EXPECT_EQ("N/A", vp.getVehicleName());
}

TEST(VersionPrinterTest, getAVStackBuildNumber)
{
    ConfigPathBackup cpb;
    VersionPrinter vp;

    const auto buildNumber = vp.getAVStackBuildNumber();

    EXPECT_EQ("local-build-fd495b8d", buildNumber.avStackNumber);
}

TEST(VersionPrinterTest, getAVStackBuildNumberLogMode)
{
    ConfigPathBackup cpb;
    VersionPrinter vp;

    const auto buildNumber = vp.getAVStackBuildNumber("", defaultAVBuildTags, true);

    EXPECT_EQ("local-build-fd495b8d", buildNumber.avStackNumber);
}

TEST(VersionPrinterTest, getAVStackBuildNumberFromLogs)
{
    ConfigPathBackup cpb;
    CwdBackup cwdb(configPath + "/..");

    ASSERT_EQ(0, cwdb.getConstructorErrno()) << "Could not change directory";

    VersionPrinter vp;

    const auto buildNumber = vp.getAVStackBuildNumber("OttoView_playback");

    EXPECT_EQ("VeryAlpha", buildNumber.avStackNumberFromLog);
}

TEST(VersionPrinterTest, getAVStackBuildNumberFromLogsNoBuildNumber)
{
    ConfigPathBackup cpb("/no-build-number");
    VersionPrinter vp;

    const auto buildNumber = vp.getAVStackBuildNumber("OttoView_playback");

    EXPECT_EQ("N/A", buildNumber.avStackNumberFromLog);
}

TEST(VersionPrinterTest, getAVStackBuildNumberFromLogsFileMissing)
{
    ConfigPathBackup cpb("/nonesuch");
    VersionPrinter vp;

    const auto buildNumber = vp.getAVStackBuildNumber("OttoView_playback", defaultAVBuildTags, true);

    EXPECT_EQ("N/A", buildNumber.avStackNumberFromLog);
}

TEST(VersionPrinterTest, getAVStackBuildNumberEmptyConfigDir)
{
    // For the config dir actually to be empty, we have to clear out
    // the environment variables.
    EnvBackup ottoClearer("OTTO_CONFIG_DIR");
    EnvBackup activeClearer("ACTIVE_CONFIG_DIR");

    VersionPrinter vp;

    const auto buildNumber = vp.getAVStackBuildNumber();

    EXPECT_EQ("N/A", buildNumber.avStackNumber);
}

TEST(VersionPrinterTest, getAVStackBuildNumberFileMissing)
{
    ConfigPathBackup cpb("/nonesuch");
    VersionPrinter vp;

    const auto buildNumber = vp.getAVStackBuildNumber();

    EXPECT_EQ("N/A", buildNumber.avStackNumber);
}

TEST(VersionPrinterTest, getAVStackBuildNumberFileEmpty)
{
    ConfigPathBackup cpb("/empty");
    VersionPrinter vp;

    const auto buildNumber = vp.getAVStackBuildNumber();

    EXPECT_EQ("N/A", buildNumber.avStackNumber);
}
