/*!
 ***********************************************************************************************************************
 * APTIV PRIVILEGED AND CONFIDENTIAL
 *
 * @file CalibrationConversion.cc
 * @product common
 * @version 1.0
 * @released TODO
 * @updated 2019-04-01
 * @brief Provides functions for converting between ottoware and nutonomy calibration frames
 *
 * @section License
 *
 * Copyright (c) 2019- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * @section Open Source License List
 *
 * This is a list of any open-source project used by this file.
 ***********************************************************************************************************************
 */

#include "CalibrationConversion.h"
#include <iomanip>
#include <iostream>

namespace utils {

Eigen::Matrix3d getMat3FromYPRInDeg(double yawInDeg, double pitchInDeg, double rollInDeg)
{
    Eigen::AngleAxisd qz(yawInDeg / 180.0 * M_PI, Eigen::Vector3d::UnitZ());
    Eigen::AngleAxisd qy(pitchInDeg / 180.0 * M_PI, Eigen::Vector3d::UnitY());
    Eigen::AngleAxisd qx(rollInDeg / 180.0 * M_PI, Eigen::Vector3d::UnitX());
    Eigen::Quaternion<double> rotation = qz * qy * qx;
    return rotation.matrix();
}

Eigen::Matrix4d getMat4FromYPRXYZ(const Eigen::Matrix3d& ypr, const Eigen::Vector3d& xyz)
{
    Eigen::Matrix4d mat;
    mat.block<3, 3>(0, 0) = ypr;
    mat.block<3, 1>(0, 3) = xyz;
    mat.row(3) << 0, 0, 0, 1;
    return mat;
}

const Eigen::Matrix3d Rx(double t)
{
    double st = sin(t), ct = cos(t);
    Eigen::Matrix3d m;
    m << 1, 0, 0, 0, ct, -st, 0, st, ct;
    return m;
}

const Eigen::Matrix3d Ry(double t)
{
    double st = sin(t), ct = cos(t);
    Eigen::Matrix3d m;
    m << ct, 0, st, 0, 1, 0, -st, 0, ct;
    return m;
}

const Eigen::Vector3d getRollPitchYaw(const Eigen::Matrix4d& rotation)
{
    const auto A = rotation.block<3, 3>(0, 0);

    const double x = -std::atan2(-A(2, 1), A(2, 2));
    const auto Qx = Rx(-x);
    const auto B = A * Qx;

    const double y = -std::atan2(B(2, 0), B(2, 2));
    const auto Qy = Ry(-y);
    const auto C = B * Qy;

    double z = -std::atan2(-C(1, 0), C(1, 1));

    return Eigen::Vector3d(x, y, z);
}

const XyzRpyInDeg convertCalibrationBetweenNutOtto(const XyzRpyInDeg& xyzRpyInDeg)
{
    Eigen::Matrix4d nutLidarOnNutOrigin{
        getMat4FromYPRXYZ(getMat3FromYPRInDeg(xyzRpyInDeg.yaw, xyzRpyInDeg.pitch, xyzRpyInDeg.roll),
                          Eigen::Vector3d(xyzRpyInDeg.x, xyzRpyInDeg.y, xyzRpyInDeg.z))};
    const auto tf = nutOriginOnOttoOrigin * nutLidarOnNutOrigin * ottoLidarOnNutLidar;
    const auto t = tf.block<3, 1>(0, 3);
    const auto r = getRollPitchYaw(tf);
    return {t.x(), t.y(), t.z(), r.x() / M_PI * 180.0, r.y() / M_PI * 180.0, r.z() / M_PI * 180.0};
}

} // namespace utils
