/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief A class to make it easy to apply "cooldowns" to certain events.

#pragma once

#include <avtime/AvClock.hpp>

#include <unordered_map>

/// @brief A class which will remember the last time `isCooledDown` returned true, for each condition
/// @note This is intended to easily implement a cooldown period, for instance to avoid spamming the log.
template<typename T>
class RateLimiter
{
public:
    /// @brief default constructor
    RateLimiter() = default;

    /// @brief Returns whether the condition is cooled down or not, and ensures the condition is active
    bool isCooledDown(const T& condition, const av::TimePoint& now, const av::Duration& cooldown)
    {
        return conditions_[condition].isCooledDown(now, cooldown);
    }

    /// @brief Returns whether the condition is active.
    bool isActive(const T& condition) { return conditions_[condition].isActive(); }

    /// @brief Sets the condition to be inactive.
    void reset(const T& condition) { conditions_[condition].reset(); };

private:
    /// @brief Implements the individual logic for the rate limiter.
    class Status
    {
    public:
        /// @brief default constructor
        Status() = default;

        /// @brief Returns if the status is "cooled down."
        /// @return True if active_ was previously false, or if the cooldown has elapsed since this function last
        /// returned true.
        bool isCooledDown(const av::TimePoint& now, const av::Duration& cooldown)
        {
            bool cooldownReached{!active_ || (now - lastMessage_ > cooldown)};
            active_ = true;
            if (cooldownReached) {
                lastMessage_ = now;
            }
            return cooldownReached;
        }

        /// @brief Return true if the status is active
        bool isActive() const { return active_; };

        /// @brief Set the status to be inactive.
        void reset() { active_ = false; };

    private:
        bool active_{false};          //< Whether the status has is "active"
        av::TimePoint lastMessage_{}; //< The last time `isCooledDown` returned true.
    };

    std::unordered_map<T, Status> conditions_{}; //< A map of all known conditions
};
