#pragma once

#include <avtime/Time.hpp>

#include <atomic>
#include <cstdint>
#include <functional>
#include <mutex>
#include <optional>
#include <string>
#include <utility>
#include <vector>

namespace task {
namespace stats {
struct MessageReport
{
    enum class EventType
    {
        Connected,
        Sent,
        Received,
        Disconnected
    };

    /// Name of the channel on which the event occurred.
    std::string channelName;

    // Type of event being reported.
    EventType eventType;

    /// Number of bytes in the message. Not initialized for event
    /// types other than Sent and Received.
    size_t messageSize;

    /// Time of the event.
    av::TimePoint timestamp;

    /// In a receive event, the time at which the message was sent.
    std::optional<av::TimePoint> sendTime;

    // Latency. Not present unless this is a report of receiving a
    // message, on a message system that lets us measure latency (SCS,
    // DDS).
    std::optional<av::Duration> latencyns;
};

/// @brief A callback to invoke when a message is reported.
using MessageCallback = std::function<void(const MessageReport&)>;

/// @brief An ID for a callback, so that it can be unregistered.
using MessageCallbackID = uint32_t;

class MessageCallbackRegistry
{
public:
    MessageCallbackRegistry()
    {
        for (size_t i = 0; i < s_maxCallbacks; ++i) {
            callbackIdsInUse_[i] = false;
            callbacks_[i] = nullptr;
        }
    }

    // Returns std::nullopt if there are no slots left in the registry.
    std::optional<MessageCallbackID> registerCallback(MessageCallback callback);

    void unregisterCallback(MessageCallbackID id);
    void invokeCallbacks(const MessageReport& report);

private:
    // If you get an error saying you've exceeded the maximum # of callbacks,
    // update this constant. If you need a very large number of callbacks,
    // we'll need to modify how the registry stores them--at present, it's
    // a simple fixed-size array of atomics, so that we can avoid having to lock
    // a mutex while invoking callbacks.
    static constexpr size_t s_maxCallbacks = 8;

    std::mutex mutex_{};
    // Guarded by mutex_
    std::array<bool, s_maxCallbacks> callbackIdsInUse_{};
    // Also guarded by mutex_
    std::array<MessageCallback, s_maxCallbacks> callbackStorage_{};

    // Not guarded.
    std::array<std::atomic<MessageCallback*>, s_maxCallbacks> callbacks_{};
};

class MessageCallbackRegistration
{
public:
    /// @brief Constructor.
    ///
    /// @param callback The callback to register. It will be
    /// unregistered when this object is destructed.
    explicit MessageCallbackRegistration(MessageCallbackRegistry& registry, MessageCallback callback)
        : registry_(registry)
        , id_(registry_.registerCallback(callback))
    {}

    /// @brief Destructor. Unregisters the callback.
    ~MessageCallbackRegistration()
    {
        if (id_) {
            registry_.unregisterCallback(*id_);
        }
    }

private:
    MessageCallbackRegistry& registry_;
    const std::optional<MessageCallbackID> id_;
};

} // namespace stats
} // namespace task
