/*!
 ***********************************************************
 * @file utility.h.
 * @brief A file implementing all the helper functions to replace C++ Boost
          library in Ottoware
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Yunpeng Xu (yunpeng.xu@aptiv.com)
 * @date 04/04/2018
 * @remark add initial source code for utility
 *
 ***********************************************************
 */

#ifndef _UTILITY_H_
#define _UTILITY_H_

#include <algorithm>
#include <execinfo.h>
#include <log/NamedVars.h>
#include <ostream>
#include <sstream>
#include <string> // std::string
#include <string_view>
#include <vector> // std::vector

#include <arpa/inet.h>
#include <sys/stat.h> // stat

/*!
 * @brief Split the given string with the given delim and store the result
 *
 * @param str: given string to split
 *        delim: given delimiter
 *        result: result to store
 *
 * @return void
 */
template<typename T>
void split(const std::string& str, const char delim, T result)
{
    std::stringstream ss(str);
    std::string item;

    while (getline(ss, item, delim)) {
        *(result++) = item;
    }
}

/*!
 * @brief Split the given string with the given delim.
 *
 * @param str: given string to split
 *        delim: given delimiter
 *
 * @return result as a vector of string
 */
std::vector<std::string> split(const std::string& str, const char delim);

namespace middleware {
namespace utility {

inline std::string ipv4ToString(const struct in_addr& addr)
{
    char addrStr[80];
    if (inet_ntop(AF_INET, &addr, addrStr, sizeof(addrStr))) {
        return addrStr;
    }
    return "<invalid.ipv4>";
}

inline std::string ipv4ToString(uint32_t addrNetworkOrder)
{
    struct in_addr addr;
    addr.s_addr = addrNetworkOrder;
    return ipv4ToString(addr);
}

/// @brief Join a container of items into a string.
///
/// T must be a container of items whose type can be <<ed to ostream&.
///
/// @param strings The container.
///
/// @param delimiter The delimiter to insert between the items.
///
/// @return A joined list of the items' string representations. Note
/// that the order of these items will be the same as the order of
/// iterating over the T. If T is not an ordered container (e.g., if
/// it's std::unordered_set), then the order of the string
/// representations may not be consistent from one run to the next.
template<typename T>
std::string join(const T& strings, std::string_view delimiter)
{
    // Copy into a string just once. We should be able to just << a
    // string_view, but it's conflicting with a "print any container"
    // operator in log/Logger.h.
    const std::string delim{delimiter};
    std::stringstream out;
    bool isFirst{true};

    for (const auto& s : strings) {
        if (isFirst) {
            isFirst = false;
        } else {
            out << delim;
        }

        out << s;
    }

    return out.str();
}
} // namespace utility
} // namespace middleware

/// Escapes the given string (if necessary) according to RFC-4180, the
/// IETF spec for CSV.
std::string escapeForCSV(const std::string& str);

/*!
 * @brief A lexical cast from string to the type specified (throw version)
 *
 * @param str: given string to cast
 *
 * @return casted result
 */
template<typename T>
T lexical_cast(const std::string& str)
{
    std::stringstream ss(str);

    T result;
    /**
     * throw bad_cast if the following:
     * 1. detect bad word like 'nana'
     * 2. detect white spaces but not the beginning or end of the word like
     *    '1.23  45'
     */
    if ((ss >> result).fail() || !(ss >> std::ws).eof()) {
        throw std::bad_cast();
    }

    return result;
}

/*!
 * @brief A lexical cast from string to the type specified (no-throw version)
 *
 * @param str: given string to cast
 *        T:   a reference of casted result
 *
 * @return true if successful, otherwise false
 */
template<typename T>
bool lexical_cast(const std::string& str, T& t)
{
    try {
        t = lexical_cast<T>(str);
    } catch (const std::bad_cast& e) {
        return false;
    }
    return true;
}

/*!
 * @brief Case sensitive implemantation of startsWith.
 * It checks if the string 'str' starts with given string 'prefix'
 *
 * @return true if starts with, otherwise false.
 */
bool startsWith(const std::string& str, const std::string& prefix);

/*!
 * @brief Case sensitive implemantation of endsWith.
 * It checks if the string 'str' ends with given string 'suffix'
 *
 * @return true if ends with, otherwise false.
 */
bool endsWith(const std::string& str, const std::string& suffix);

/// @brief Trims string from start
/// @param [in,out] str String to trim
static inline std::string& ltrim(std::string& str)
{
    str.erase(str.begin(), std::find_if(str.begin(), str.end(), std::not1(std::ptr_fun<int, int>(&std::isspace))));
    return str;
}

/// @brief Trim string from end
/// @param [in,out] str String to trim
static inline std::string& rtrim(std::string& str)
{
    str.erase(std::find_if(str.rbegin(), str.rend(), std::not1(std::ptr_fun<int, int>(&std::isspace))).base(),
              str.end());
    return str;
}

/// @brief Trims string from left and right
/// @param [in,out] str String to trim
static inline std::string& trim(std::string& str)
{
    return ltrim(rtrim(str));
}

/*!
 * @brief Check if a file exists with the given file name
 *
 * @return true if exists, otherwise false.
 */
bool fileExists(const std::string& name);

/*!
 * @brief Check if a path is a directory
 *
 * @return true if it is, otherwise false.
 */
bool isDirectory(const std::string& path);

#endif // ifndef _UTILITY_H_
