#include <gtest/gtest.h>
#include <utility/MessageCallback.hpp>

#include <map>
#include <string>
#include <utility>

namespace {

using namespace task::stats;

TEST(MessageCallbackTest, RegisterManually)
{
    MessageCallbackRegistry registry{};
    std::map<std::string, size_t> channelToTotalBytes{};
    const auto callback = [&channelToTotalBytes](const MessageReport& report) {
        if (report.eventType == MessageReport::EventType::Sent ||
            report.eventType == MessageReport::EventType::Received) {
            auto it = channelToTotalBytes.find(report.channelName);
            if (it == channelToTotalBytes.end()) {
                channelToTotalBytes[report.channelName] = report.messageSize;
            } else {
                it->second += report.messageSize;
            }
        }
    };

    registry.registerCallback(callback);

    const av::TimePoint zeroTime{};
    const av::Duration zeroDur{};

    registry.invokeCallbacks({"English", MessageReport::EventType::Sent, 17, zeroTime, {}, {}});
    registry.invokeCallbacks({"Panama", MessageReport::EventType::Received, 19, zeroTime, zeroTime, zeroDur});
    registry.invokeCallbacks({"Panama", MessageReport::EventType::Sent, 23, zeroTime, {}, {}});
    registry.invokeCallbacks({"Erie", MessageReport::EventType::Received, 29, zeroTime, zeroTime, zeroDur});

    EXPECT_EQ(3, channelToTotalBytes.size());
    EXPECT_EQ(17, channelToTotalBytes["English"]);
    EXPECT_EQ(42, channelToTotalBytes["Panama"]);
    EXPECT_EQ(29, channelToTotalBytes["Erie"]);
}

TEST(MessageCallbackTest, RegisterManuallyThenUnregisterManually)
{
    MessageCallbackRegistry registry{};
    std::map<std::string, size_t> channelToTotalBytes{};
    const auto callback = [&channelToTotalBytes](const MessageReport& report) {
        if (report.eventType == MessageReport::EventType::Sent ||
            report.eventType == MessageReport::EventType::Received) {
            auto it = channelToTotalBytes.find(report.channelName);
            if (it == channelToTotalBytes.end()) {
                channelToTotalBytes[report.channelName] = report.messageSize;
            } else {
                it->second += report.messageSize;
            }
        }
    };

    const auto id = registry.registerCallback(callback);
    ASSERT_TRUE(id.has_value());
    registry.unregisterCallback(id.value());

    const av::TimePoint zeroTime{};
    const av::Duration zeroDur{};

    registry.invokeCallbacks({"English", MessageReport::EventType::Sent, 17, zeroTime, {}, {}});
    registry.invokeCallbacks({"Panama", MessageReport::EventType::Received, 19, zeroTime, zeroTime, zeroDur});
    registry.invokeCallbacks({"Panama", MessageReport::EventType::Sent, 23, zeroTime, {}, {}});
    registry.invokeCallbacks({"Erie", MessageReport::EventType::Received, 29, zeroTime, zeroTime, zeroDur});

    EXPECT_EQ(0, channelToTotalBytes.size());
}

TEST(MessageCallbackTest, RegisterWithRAII)
{
    MessageCallbackRegistry registry{};
    std::map<std::string, size_t> channelToTotalBytes{};
    const auto callback = [&channelToTotalBytes](const MessageReport& report) {
        if (report.eventType == MessageReport::EventType::Sent ||
            report.eventType == MessageReport::EventType::Received) {
            auto it = channelToTotalBytes.find(report.channelName);
            if (it == channelToTotalBytes.end()) {
                channelToTotalBytes[report.channelName] = report.messageSize;
            } else {
                it->second += report.messageSize;
            }
        }
    };

    MessageCallbackRegistration registration(registry, callback);

    const av::TimePoint zeroTime{};
    const av::Duration zeroDur{};

    registry.invokeCallbacks({"English", MessageReport::EventType::Sent, 17, zeroTime, {}, {}});
    registry.invokeCallbacks({"Panama", MessageReport::EventType::Received, 19, zeroTime, zeroTime, zeroDur});
    registry.invokeCallbacks({"Panama", MessageReport::EventType::Sent, 23, zeroTime, {}, {}});
    registry.invokeCallbacks({"Erie", MessageReport::EventType::Received, 29, zeroTime, zeroTime, zeroDur});

    EXPECT_EQ(3, channelToTotalBytes.size());
    EXPECT_EQ(17, channelToTotalBytes["English"]);
    EXPECT_EQ(42, channelToTotalBytes["Panama"]);
    EXPECT_EQ(29, channelToTotalBytes["Erie"]);
}

TEST(MessageCallbackTest, RegisterWithRAIIThenUnregister)
{
    MessageCallbackRegistry registry{};
    std::map<std::string, size_t> channelToTotalBytes{};
    const auto callback = [&channelToTotalBytes](const MessageReport& report) {
        if (report.eventType == MessageReport::EventType::Sent ||
            report.eventType == MessageReport::EventType::Received) {
            auto it = channelToTotalBytes.find(report.channelName);
            if (it == channelToTotalBytes.end()) {
                channelToTotalBytes[report.channelName] = report.messageSize;
            } else {
                it->second += report.messageSize;
            }
        }
    };

    const av::TimePoint zeroTime{};
    const av::Duration zeroDur{};

    {
        MessageCallbackRegistration registration(registry, callback);
        registry.invokeCallbacks({"English", MessageReport::EventType::Sent, 17, zeroTime, {}, {}});
        registry.invokeCallbacks({"Panama", MessageReport::EventType::Received, 19, zeroTime, zeroTime, zeroDur});
    }

    registry.invokeCallbacks({"Panama", MessageReport::EventType::Sent, 23, zeroTime, {}, {}});
    registry.invokeCallbacks({"Erie", MessageReport::EventType::Received, 29, zeroTime, zeroTime, zeroDur});

    EXPECT_EQ(2, channelToTotalBytes.size());
    EXPECT_EQ(17, channelToTotalBytes["English"]);
    EXPECT_EQ(19, channelToTotalBytes["Panama"]);
}

TEST(MessageCallbackTest, ExceedCallbackLimit)
{
    MessageCallbackRegistry registry{};
    for (int i = 0; true; ++i) {
        // Make sure we exhaust the set of callbacks eventually.
        EXPECT_TRUE(i < 1000);
        std::optional<MessageCallbackID> id;
        EXPECT_NO_THROW(id = registry.registerCallback([](const MessageReport& /*report*/) {}));
        if (!id) {
            break;
        }
    }
}
} // namespace
