/// @copyright Copyright (c) 2021- Motional. All rights reserved.
///
/// NOTICE: Removing this heading is prohibited.
///
/// All information contained herein is, and remains the property of Motional,
/// its affiliates, and its suppliers (if any). The intellectual and technical
/// concepts contained herein are proprietary to Motional, its affiliates, and
/// its suppliers (if any) and may be covered by U.S. and Foreign Patents,
/// patents in process, and are protected by trade secrets or copyright law.
/// Dissemination of this information or reproduction of this material is
/// strictly forbidden unless prior written permission is obtained from Motional.
///
/// @file
/// @brief Test the RateLimiter class.

#include "platform/middleware/libs/utility/RateLimiter.hpp"

#include <avtime/ConversionsUsec.hpp>

#include <gtest/gtest.h>
#include <string>
#include <string_view>

TEST(RateLimiter, BasicUsage)
{
    RateLimiter<std::string> rl{};

    // Does not exist yet, should not be active.
    EXPECT_FALSE(rl.isActive("test"));

    // Turn active
    EXPECT_TRUE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(1000000l), std::chrono::seconds(5)));

    // Too soon! Not cooled down!
    EXPECT_FALSE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(2000000l), std::chrono::seconds(5)));

    // Other cooldowns are independent of the first
    EXPECT_TRUE(rl.isCooledDown("test2", av::timePointFromUsec<av::AvClock>(2000000l), std::chrono::seconds(1)));
    EXPECT_FALSE(rl.isCooledDown("test2", av::timePointFromUsec<av::AvClock>(3000000l), std::chrono::seconds(1)));
    EXPECT_TRUE(rl.isCooledDown("test2", av::timePointFromUsec<av::AvClock>(3000001l), std::chrono::seconds(1)));

    // After all that, we still cool down on schedule.
    EXPECT_FALSE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(6000000l), std::chrono::seconds(5)));
    EXPECT_TRUE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(6000001l), std::chrono::seconds(5)));

    // Now, timeout is five seconds from the last one, at 6000001us
    EXPECT_FALSE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(7000000l), std::chrono::seconds(5)));
    EXPECT_FALSE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(11000001l), std::chrono::seconds(5)));
    EXPECT_TRUE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(11000002l), std::chrono::seconds(5)));
}

TEST(RateLimiter, Resets)
{
    RateLimiter<std::string> rl{};

    // Does not exist yet, should not be active.
    EXPECT_FALSE(rl.isActive("test"));

    // Turn active
    EXPECT_TRUE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(1000000l), std::chrono::seconds(5)));
    EXPECT_FALSE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(2000000l), std::chrono::seconds(5)));
    EXPECT_TRUE(rl.isActive("test"));
    rl.reset("test");
    EXPECT_FALSE(rl.isActive("test"));
    EXPECT_TRUE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(2000000l), std::chrono::seconds(5)));
}

TEST(RateLimiter, ChangingTimeout)
{
    RateLimiter<std::string> rl{};

    EXPECT_TRUE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(1000000l), std::chrono::seconds(5)));
    EXPECT_FALSE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(3000000l), std::chrono::seconds(5)));

    // Shortening the timeout, it triggers
    EXPECT_TRUE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(3000000l), std::chrono::seconds(1)));

    // Lengthening the timeout, it does not trigger.
    EXPECT_FALSE(rl.isCooledDown("test", av::timePointFromUsec<av::AvClock>(12000000l), std::chrono::seconds(10)));
}
