/*!
 ***********************************************************
 * @file testUtility.cc.
 * @brief A file implementing all the testing of helper functions which are used
 *        to replace C++ Boost library in Ottoware
 *
 * @section License
 *
 * Copyright (c) 2015- Delphi Automotive. All rights
 * reserved.
 *
 * @section changeLogs Change Logs
 *
 * @author Yunpeng Xu (yunpeng.xu@aptiv.com)
 * @date 04/04/2018
 * @remark add initial source code for utility
 *
 ***********************************************************
 */

#ifndef _TESTUTILITY_CC_
#define _TESTUTILITY_CC_

#include <dirent.h> // dirent
#include <iostream> // std::cerr, std::cout
#include <unistd.h> // readlink

#include "gtest/gtest.h" // google test
#include <utility/utility.h>

using namespace std;

TEST(TESTUTILITY, SPLIT)
{
    // test case for white space
    {
        const vector<string> expected = {"porting", "ottoware", "to", "QNX", "platform"};
        const vector<string> words = split("porting ottoware to QNX platform", ' ');
        EXPECT_EQ(words, expected);
    }

    // test case for consecutive delimiters (there are two spaces
    // between "to" and "QNX")
    {
        const vector<string> expected = {"porting", "ottoware", "to", "", "QNX", "platform"};
        const vector<string> words = split("porting ottoware to  QNX platform", ' ');
        EXPECT_EQ(words, expected);
    }

    // test case for white spaces
    {
        const vector<string> words = split("", ' ');
        EXPECT_TRUE(words.empty());
    }

    // test for new line
    {
        const vector<string> expected = {"porting", "ottoware", "to", "QNX", "platform"};
        const vector<string> words = split("porting\nottoware\nto\nQNX\nplatform", '\n');
        EXPECT_EQ(words, expected);
    }

    // test for new line
    {
        const vector<string> expected = {"porting", "ottoware", "to", "QNX", "platform"};
        const vector<string> words = split("porting\nottoware\nto\nQNX\nplatform", '\t');
        EXPECT_NE(words, expected);
    }
}

TEST(TESTUTILITY, Join)
{
    // test case for white space
    {
        const vector<string> words = {"porting", "ottoware", "to", "QNX", "platform"};
        const string joined = middleware::utility::join(words, " ");
        const string expected = "porting ottoware to QNX platform";
        EXPECT_EQ(joined, expected);
    }

    // Comma-separated
    {
        const vector<string> words = {"porting", "ottoware", "to", "QNX", "platform"};
        const string joined = middleware::utility::join(words, ",");
        const string expected = "porting,ottoware,to,QNX,platform";
        EXPECT_EQ(joined, expected);
    }

    {
        const vector<int> nums = {1, 1, 2, 3, 5, 8, 13, 21};
        const string joined = middleware::utility::join(nums, ", ");
        const string expected = "1, 1, 2, 3, 5, 8, 13, 21";
        EXPECT_EQ(joined, expected);
    }

    // test case for consecutive delimiters (there are two spaces
    // between "to" and "QNX")
    {
        const vector<string> words = {"porting", "ottoware", "to", "", "QNX", "platform"};
        const string joined = middleware::utility::join(words, " ");
        const string expected = "porting ottoware to  QNX platform";
        EXPECT_EQ(joined, expected);
    }
}

TEST(TestUtility, escapeForCSV)
{
    EXPECT_EQ(escapeForCSV("foo"), "foo");
    EXPECT_EQ(escapeForCSV("foo\nbar"), "\"foo\nbar\"");
    EXPECT_EQ(escapeForCSV("foo, bar, baz"), "\"foo, bar, baz\"");
    EXPECT_EQ(escapeForCSV("foo says \"bar\", baz"), "\"foo says \"\"bar\"\", baz\"");
    EXPECT_EQ(escapeForCSV("foo \"bars\" the door.\nbaz, quux, and xyzzy knock."),
              "\"foo \"\"bars\"\" the door.\nbaz, quux, and xyzzy knock.\"");
}

TEST(TESTUTILITY, LEXICAL_CAST_THROW1)
{
    string str = "";
    bool isExceptionCatched = false;
    // test case for int
    str = "20180404";
    int intResult;
    int intExpected = 20180404;
    try {
        intResult = lexical_cast<int>(str);
    } catch (const bad_cast& e) {
        cerr << "lexical cast throw bad cast exception" << '\n';
        isExceptionCatched = true;
    }
    ASSERT_EQ(intResult, intExpected);
    ASSERT_FALSE(isExceptionCatched);

    // test case for int
    str = "-12345";
    intExpected = -12345;
    isExceptionCatched = false;
    try {
        intResult = lexical_cast<int>(str);
    } catch (const bad_cast& e) {
        cerr << "lexical cast throw bad cast exception" << '\n';
        isExceptionCatched = true;
    }
    ASSERT_EQ(intResult, intExpected);
    ASSERT_FALSE(isExceptionCatched);

    // test case for int
    str = "2018 nana";
    isExceptionCatched = false;
    try {
        intResult = lexical_cast<int>(str);
    } catch (const bad_cast& e) {
        cerr << "lexical cast throw bad cast exception" << '\n';
        isExceptionCatched = true;
    }
    ASSERT_TRUE(isExceptionCatched);

    // test case for float
    str = "2.832";
    isExceptionCatched = false;
    float floatResult;
    float floatExpected = 2.832f;
    try {
        floatResult = lexical_cast<float>(str);
    } catch (const bad_cast& e) {
        cerr << "lexical cast throw bad cast exception" << '\n';
        isExceptionCatched = true;
    }
    ASSERT_EQ(floatResult, floatExpected);
    ASSERT_FALSE(isExceptionCatched);

    // test case for float
    // this will pass although there's a white space at end
    str = "2.832 ";
    isExceptionCatched = false;
    try {
        floatResult = lexical_cast<float>(str);
    } catch (const bad_cast& e) {
        cerr << "lexical cast throw bad cast exception" << '\n';
        isExceptionCatched = true;
    }
    ASSERT_EQ(floatResult, floatExpected);
    ASSERT_FALSE(isExceptionCatched);

    // test case for double
    str = "-3.1415926";
    isExceptionCatched = false;
    double doubleResult;
    double doubleExpected = -3.1415926;
    try {
        doubleResult = lexical_cast<double>(str);
    } catch (const bad_cast& e) {
        cerr << "lexical cast throw bad cast exception" << '\n';
        isExceptionCatched = true;
    }
    ASSERT_EQ(doubleResult, doubleExpected);
    ASSERT_FALSE(isExceptionCatched);
}

// two more complicated cases
TEST(TESTUTILITY, LEXICAL_CAST_THROW2)
{
    // test case for float
    // this will pass although there're white space at the beginning and end
    string str = " 2.832  ";
    float floatResult;
    float floatExpected = 2.832f;
    bool isExceptionCatched = false;

    try {
        floatResult = lexical_cast<float>(str);
    } catch (const bad_cast& e) {
        cerr << "lexical cast throw bad cast exception" << '\n';
        isExceptionCatched = true;
    }
    ASSERT_EQ(floatResult, floatExpected);
    ASSERT_FALSE(isExceptionCatched);
}

TEST(TESTUTILITY, LEXICAL_CAST_THROW3)
{
    // test case for float
    // this will fail since there're white spaces between digits
    string str = "2.832       1456";
    bool isExceptionCatched = false;
    try {
        lexical_cast<float>(str);
    } catch (const bad_cast& e) {
        cerr << "lexical cast throw bad cast exception" << '\n';
        isExceptionCatched = true;
    }
    ASSERT_TRUE(isExceptionCatched);
}

TEST(TESTUTILITY, LEXICAL_CAST_NO_THROW)
{
    string str = "";
    int intResult;
    int intExpected;

    // test case for int
    str = "-12345";
    intExpected = -12345;

    ASSERT_TRUE(lexical_cast(str, intResult));
    ASSERT_EQ(intResult, intExpected);

    // test case for int
    // this will fail since there's an invalid 'nana'
    str = "2018 nana";
    intResult = 2018;
    ASSERT_FALSE(lexical_cast(str, intResult));
}

TEST(TESTUTILITY, STARTSWITH)
{
    string prefix = "Logger-";
    string str = "Logger-04042018-Safer-log";

    ASSERT_TRUE(startsWith(str, prefix));

    str = "logger-04042018-Safer-log";
    ASSERT_FALSE(startsWith(str, prefix));
}

TEST(TESTUTILITY, ENDSWITH)
{
    string suffix = "-log";
    string str = "Logger-04042018-Safer-log";

    ASSERT_TRUE(endsWith(str, suffix));

    str = "Logger-04042018-Safer-Log";
    ASSERT_FALSE(startsWith(str, suffix));

    str = "Logger-04042018-Safer-log   ";
    ASSERT_FALSE(startsWith(str, suffix));
}

TEST(TESTUTILITY, FILEEXISTS)
{
    string testFile = "/bin/chmod";
    cout << "test file is: " << testFile << endl;
    ASSERT_TRUE(fileExists(testFile));

    testFile = "/bin/abcdefg";
    cout << "test file is: " << testFile << endl;
    ASSERT_FALSE(fileExists(testFile));
}

TEST(TESTUTILITY, ISDIRECTORY)
{
    string testPath = "/bin";
    cout << "test path is: " << testPath << endl;
    ASSERT_TRUE(isDirectory(testPath));

    testPath = "/abcdefg";
    cout << "test path is: " << testPath << endl;
    ASSERT_FALSE(isDirectory(testPath));
}

TEST(TESTUTILITY, IPv4ToString)
{
    using middleware::utility::ipv4ToString;

    {
        const std::string expected{"127.0.0.1"};
        const uint32_t addr_num = htonl(0x7f000001);
        const struct in_addr addr
        {
            addr_num
        };

        EXPECT_EQ(ipv4ToString(addr), expected);
        EXPECT_EQ(ipv4ToString(addr_num), expected);
    }

    {
        const std::string expected{"1.2.3.4"};
        const uint32_t addr_num = htonl(0x01020304);
        const struct in_addr addr
        {
            addr_num
        };

        EXPECT_EQ(ipv4ToString(addr), expected);
        EXPECT_EQ(ipv4ToString(addr_num), expected);
    }
}

#endif // ifndef _TESTUTILITY_CC_
