#include <interfaces/AssistanceRequest/AssistanceRequest.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


AssistancePayload::AssistancePayload() = default;

AssistancePayload::AssistancePayload(
    Category category_arg ,
    SubCategory subcategory_arg ,
    std::string description_arg ,
    Priority priority_arg ,
    std::vector<int64_t> tracked_objs_list_arg ,
    std::vector<std::string> map_objs_list_arg ,
    std::vector<std::string> failed_sensors_list_arg 
    ):
    category { std::move(category_arg) } // NOLINT
      ,
    subcategory { std::move(subcategory_arg) } // NOLINT
      ,
    description { std::move(description_arg) } // NOLINT
      ,
    priority { std::move(priority_arg) } // NOLINT
      ,
    tracked_objs_list { std::move(tracked_objs_list_arg) } // NOLINT
      ,
    map_objs_list { std::move(map_objs_list_arg) } // NOLINT
      ,
    failed_sensors_list { std::move(failed_sensors_list_arg) } // NOLINT
      
    {
}

AssistancePayload::AssistancePayload(const AssistancePayload&) = default;

// NOLINTNEXTLINE
AssistancePayload::AssistancePayload(AssistancePayload&&) = default;

AssistancePayload& AssistancePayload::operator=(const AssistancePayload&) = default;

// NOLINTNEXTLINE
AssistancePayload& AssistancePayload::operator=(AssistancePayload&&) = default;

void AssistancePayload::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << static_cast<uint32_t>(category);out << ",";
    out << static_cast<uint32_t>(subcategory);out << ",";
    out << "\"";
        out << description;
        out << "\"";out << ",";
    out << static_cast<uint32_t>(priority);out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!tracked_objs_list.empty())
        {
            std::copy(tracked_objs_list.begin(), tracked_objs_list.end() - 1, std::ostream_iterator<int64_t>(out, "},{"));
            std::copy(tracked_objs_list.end()-1, tracked_objs_list.end(), std::ostream_iterator<int64_t>(out, ""));
        }
        out << "}}\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!map_objs_list.empty())
        {
            std::copy(map_objs_list.begin(), map_objs_list.end() - 1, std::ostream_iterator<std::string>(out, "},{"));
            std::copy(map_objs_list.end()-1, map_objs_list.end(), std::ostream_iterator<std::string>(out, ""));
        }
        out << "}}\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!failed_sensors_list.empty())
        {
            std::copy(failed_sensors_list.begin(), failed_sensors_list.end() - 1, std::ostream_iterator<std::string>(out, "},{"));
            std::copy(failed_sensors_list.end()-1, failed_sensors_list.end(), std::ostream_iterator<std::string>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream AssistancePayload::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void AssistancePayload::serialize(Archive& ar, unsigned int /*version*/)
{
    ar& CEREAL_NVP(category);
    ar& CEREAL_NVP(subcategory);
    ar& CEREAL_NVP(description);
    ar& CEREAL_NVP(priority);
    ar& CEREAL_NVP(tracked_objs_list);
    ar& CEREAL_NVP(map_objs_list);
    ar& CEREAL_NVP(failed_sensors_list);
}



template void AssistancePayload::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void AssistancePayload::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void AssistancePayload::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void AssistancePayload::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void AssistancePayload::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void AssistancePayload::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void AssistancePayload::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const AssistancePayload& assistancepayload)
{
    assistancepayload.toCsv(os);
    return os;
}

bool operator== (const AssistancePayload &lhs, const AssistancePayload &rhs)
{
    return (  lhs.category== rhs.category &&
     lhs.subcategory== rhs.subcategory &&
     lhs.description== rhs.description &&
     lhs.priority== rhs.priority &&
     lhs.tracked_objs_list== rhs.tracked_objs_list &&
     lhs.map_objs_list== rhs.map_objs_list &&
     lhs.failed_sensors_list== rhs.failed_sensors_list 
    );
}

std::size_t AssistancePayload::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<Category>()(category);
    result = result * 31 + std::hash<SubCategory>()(subcategory);
    result = result * 31 + std::hash<std::string>()(description);
    result = result * 31 + std::hash<Priority>()(priority);
    result = result * 31 + hashVector(tracked_objs_list);
    result = result * 31 + hashVector(map_objs_list);
    result = result * 31 + hashVector(failed_sensors_list);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::AssistancePayload)




 namespace av {
 namespace msg {


AssistanceRequest::AssistanceRequest() = default;

AssistanceRequest::AssistanceRequest(
    std::string id_arg ,
    av::msg::TimePoint timestamp_arg ,
    std::vector<AssistancePayload> payloads_arg 
    ):
    id { std::move(id_arg) } // NOLINT
      ,timestamp { std::move(timestamp_arg) } // NOLINT
      ,payloads { std::move(payloads_arg) } // NOLINT
      {
}

AssistanceRequest::AssistanceRequest(const AssistanceRequest&) = default;

// NOLINTNEXTLINE
AssistanceRequest::AssistanceRequest(AssistanceRequest&&) = default;

AssistanceRequest& AssistanceRequest::operator=(const AssistanceRequest&) = default;

// NOLINTNEXTLINE
AssistanceRequest& AssistanceRequest::operator=(AssistanceRequest&&) = default;

void AssistanceRequest::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << "\"";
        out << id;
        out << "\"";out << ",";
    out << timestamp;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!payloads.empty())
        {
            std::copy(payloads.begin(), payloads.end() - 1, std::ostream_iterator<AssistancePayload>(out, "},{"));
            std::copy(payloads.end()-1, payloads.end(), std::ostream_iterator<AssistancePayload>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream AssistanceRequest::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void AssistanceRequest::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(id);
        ar& CEREAL_NVP(timestamp);
        ar& CEREAL_NVP(payloads);
    }


template void AssistanceRequest::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void AssistanceRequest::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void AssistanceRequest::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void AssistanceRequest::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void AssistanceRequest::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void AssistanceRequest::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void AssistanceRequest::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const AssistanceRequest& assistancerequest)
{
    assistancerequest.toCsv(os);
    return os;
}

bool operator== (const AssistanceRequest &lhs, const AssistanceRequest &rhs)
{
    return (  lhs.id== rhs.id &&
     lhs.timestamp== rhs.timestamp &&
     lhs.payloads== rhs.payloads 
    );
}

std::size_t AssistanceRequest::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<std::string>()(id);
    result = result * 31 + std::hash<av::msg::TimePoint>()(timestamp);
    result = result * 31 + hashVector(payloads);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::AssistanceRequest)

