#include <interfaces/BaselabsDiagnosticData/BaselabsDiagnosticData.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace safetyhigh {


EgoMotionState::EgoMotionState() = default;

EgoMotionState::EgoMotionState(
    float yaw_rate_arg ,
    float velocity_arg ,
    float acceleration_arg ,
    uint64_t measurement_time_stamp_arg ,
    uint64_t arrival_time_stamp_arg ,
    bool invalid_status_arg 
    ):
    yaw_rate { std::move(yaw_rate_arg) } // NOLINT
      ,
    velocity { std::move(velocity_arg) } // NOLINT
      ,
    acceleration { std::move(acceleration_arg) } // NOLINT
      ,
    measurement_time_stamp { std::move(measurement_time_stamp_arg) } // NOLINT
      ,
    arrival_time_stamp { std::move(arrival_time_stamp_arg) } // NOLINT
      ,
    invalid_status { std::move(invalid_status_arg) } // NOLINT
      
    {
}

EgoMotionState::EgoMotionState(const EgoMotionState&) = default;

// NOLINTNEXTLINE
EgoMotionState::EgoMotionState(EgoMotionState&&) = default;

EgoMotionState& EgoMotionState::operator=(const EgoMotionState&) = default;

// NOLINTNEXTLINE
EgoMotionState& EgoMotionState::operator=(EgoMotionState&&) = default;

void EgoMotionState::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << yaw_rate;out << ",";
    out << velocity;out << ",";
    out << acceleration;out << ",";
    out << measurement_time_stamp;out << ",";
    out << arrival_time_stamp;out << ",";
    out << invalid_status;
    
}

std::stringstream EgoMotionState::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void EgoMotionState::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(yaw_rate);
        ar& CEREAL_NVP(velocity);
        ar& CEREAL_NVP(acceleration);
        ar& CEREAL_NVP(measurement_time_stamp);
        ar& CEREAL_NVP(arrival_time_stamp);
        ar& CEREAL_NVP(invalid_status);
    }


template void EgoMotionState::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void EgoMotionState::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void EgoMotionState::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void EgoMotionState::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void EgoMotionState::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void EgoMotionState::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void EgoMotionState::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const EgoMotionState& egomotionstate)
{
    egomotionstate.toCsv(os);
    return os;
}

bool operator== (const EgoMotionState &lhs, const EgoMotionState &rhs)
{
    return (  lhs.yaw_rate== rhs.yaw_rate &&
     lhs.velocity== rhs.velocity &&
     lhs.acceleration== rhs.acceleration &&
     lhs.measurement_time_stamp== rhs.measurement_time_stamp &&
     lhs.arrival_time_stamp== rhs.arrival_time_stamp &&
     lhs.invalid_status== rhs.invalid_status 
    );
}

std::size_t EgoMotionState::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<float>()(yaw_rate);
    result = result * 31 + std::hash<float>()(velocity);
    result = result * 31 + std::hash<float>()(acceleration);
    result = result * 31 + std::hash<uint64_t>()(measurement_time_stamp);
    result = result * 31 + std::hash<uint64_t>()(arrival_time_stamp);
    result = result * 31 + std::hash<bool>()(invalid_status);
    return result;
}

 } // namespace safetyhigh


CEREAL_BIND_TO_ARCHIVES(safetyhigh::EgoMotionState)

 namespace safetyhigh {


EgoMotion::EgoMotion() = default;

EgoMotion::EgoMotion(
    EgoMotionState ego_motion_state_data_arg ,
    uint64_t last_update_time_arg 
    ):
    ego_motion_state_data { std::move(ego_motion_state_data_arg) } // NOLINT
      ,
    last_update_time { std::move(last_update_time_arg) } // NOLINT
      
    {
}

EgoMotion::EgoMotion(const EgoMotion&) = default;

// NOLINTNEXTLINE
EgoMotion::EgoMotion(EgoMotion&&) = default;

EgoMotion& EgoMotion::operator=(const EgoMotion&) = default;

// NOLINTNEXTLINE
EgoMotion& EgoMotion::operator=(EgoMotion&&) = default;

void EgoMotion::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << ego_motion_state_data;out << ",";
    out << last_update_time;
    
}

std::stringstream EgoMotion::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void EgoMotion::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(ego_motion_state_data);
        ar& CEREAL_NVP(last_update_time);
    }


template void EgoMotion::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void EgoMotion::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void EgoMotion::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void EgoMotion::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void EgoMotion::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void EgoMotion::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void EgoMotion::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const EgoMotion& egomotion)
{
    egomotion.toCsv(os);
    return os;
}

bool operator== (const EgoMotion &lhs, const EgoMotion &rhs)
{
    return (  lhs.ego_motion_state_data== rhs.ego_motion_state_data &&
     lhs.last_update_time== rhs.last_update_time 
    );
}

std::size_t EgoMotion::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<EgoMotionState>()(ego_motion_state_data);
    result = result * 31 + std::hash<uint64_t>()(last_update_time);
    return result;
}

 } // namespace safetyhigh


CEREAL_BIND_TO_ARCHIVES(safetyhigh::EgoMotion)

 namespace safetyhigh {


OccupancyGrid::OccupancyGrid() = default;

OccupancyGrid::OccupancyGrid(
    float cell_size_x_arg ,
    float cell_size_y_arg ,
    uint32_t cell_count_x_arg ,
    uint32_t cell_count_y_arg ,
    float size_x_arg ,
    float size_y_arg ,
    float reference_point_in_grid_x_arg ,
    float reference_point_in_grid_y_arg ,
    float reference_point_in_world_x_arg ,
    float reference_point_in_world_y_arg ,
    float ego_position_diff_x_arg ,
    float ego_position_diff_y_arg ,
    float ego_heading_arg ,
    std::vector<float> grid_cell_occupancies_arg ,
    uint64_t time_stamp_arg 
    ):
    cell_size_x { std::move(cell_size_x_arg) } // NOLINT
      ,
    cell_size_y { std::move(cell_size_y_arg) } // NOLINT
      ,
    cell_count_x { std::move(cell_count_x_arg) } // NOLINT
      ,
    cell_count_y { std::move(cell_count_y_arg) } // NOLINT
      ,
    size_x { std::move(size_x_arg) } // NOLINT
      ,
    size_y { std::move(size_y_arg) } // NOLINT
      ,
    reference_point_in_grid_x { std::move(reference_point_in_grid_x_arg) } // NOLINT
      ,
    reference_point_in_grid_y { std::move(reference_point_in_grid_y_arg) } // NOLINT
      ,
    reference_point_in_world_x { std::move(reference_point_in_world_x_arg) } // NOLINT
      ,
    reference_point_in_world_y { std::move(reference_point_in_world_y_arg) } // NOLINT
      ,
    ego_position_diff_x { std::move(ego_position_diff_x_arg) } // NOLINT
      ,
    ego_position_diff_y { std::move(ego_position_diff_y_arg) } // NOLINT
      ,
    ego_heading { std::move(ego_heading_arg) } // NOLINT
      ,
    grid_cell_occupancies { std::move(grid_cell_occupancies_arg) } // NOLINT
      ,
    time_stamp { std::move(time_stamp_arg) } // NOLINT
      
    {
}

OccupancyGrid::OccupancyGrid(const OccupancyGrid&) = default;

// NOLINTNEXTLINE
OccupancyGrid::OccupancyGrid(OccupancyGrid&&) = default;

OccupancyGrid& OccupancyGrid::operator=(const OccupancyGrid&) = default;

// NOLINTNEXTLINE
OccupancyGrid& OccupancyGrid::operator=(OccupancyGrid&&) = default;

void OccupancyGrid::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << cell_size_x;out << ",";
    out << cell_size_y;out << ",";
    out << cell_count_x;out << ",";
    out << cell_count_y;out << ",";
    out << size_x;out << ",";
    out << size_y;out << ",";
    out << reference_point_in_grid_x;out << ",";
    out << reference_point_in_grid_y;out << ",";
    out << reference_point_in_world_x;out << ",";
    out << reference_point_in_world_y;out << ",";
    out << ego_position_diff_x;out << ",";
    out << ego_position_diff_y;out << ",";
    out << ego_heading;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!grid_cell_occupancies.empty())
        {
            std::copy(grid_cell_occupancies.begin(), grid_cell_occupancies.end() - 1, std::ostream_iterator<float>(out, "},{"));
            std::copy(grid_cell_occupancies.end()-1, grid_cell_occupancies.end(), std::ostream_iterator<float>(out, ""));
        }
        out << "}}\"";out << ",";
    out << time_stamp;
    
}

std::stringstream OccupancyGrid::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void OccupancyGrid::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(cell_size_x);
        ar& CEREAL_NVP(cell_size_y);
        ar& CEREAL_NVP(cell_count_x);
        ar& CEREAL_NVP(cell_count_y);
        ar& CEREAL_NVP(size_x);
        ar& CEREAL_NVP(size_y);
        ar& CEREAL_NVP(reference_point_in_grid_x);
        ar& CEREAL_NVP(reference_point_in_grid_y);
        ar& CEREAL_NVP(reference_point_in_world_x);
        ar& CEREAL_NVP(reference_point_in_world_y);
        ar& CEREAL_NVP(ego_position_diff_x);
        ar& CEREAL_NVP(ego_position_diff_y);
        ar& CEREAL_NVP(ego_heading);
        ar& CEREAL_NVP(grid_cell_occupancies);
        ar& CEREAL_NVP(time_stamp);
    }


template void OccupancyGrid::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void OccupancyGrid::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void OccupancyGrid::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void OccupancyGrid::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void OccupancyGrid::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void OccupancyGrid::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void OccupancyGrid::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const OccupancyGrid& occupancygrid)
{
    occupancygrid.toCsv(os);
    return os;
}

bool operator== (const OccupancyGrid &lhs, const OccupancyGrid &rhs)
{
    return (  lhs.cell_size_x== rhs.cell_size_x &&
     lhs.cell_size_y== rhs.cell_size_y &&
     lhs.cell_count_x== rhs.cell_count_x &&
     lhs.cell_count_y== rhs.cell_count_y &&
     lhs.size_x== rhs.size_x &&
     lhs.size_y== rhs.size_y &&
     lhs.reference_point_in_grid_x== rhs.reference_point_in_grid_x &&
     lhs.reference_point_in_grid_y== rhs.reference_point_in_grid_y &&
     lhs.reference_point_in_world_x== rhs.reference_point_in_world_x &&
     lhs.reference_point_in_world_y== rhs.reference_point_in_world_y &&
     lhs.ego_position_diff_x== rhs.ego_position_diff_x &&
     lhs.ego_position_diff_y== rhs.ego_position_diff_y &&
     lhs.ego_heading== rhs.ego_heading &&
     lhs.grid_cell_occupancies== rhs.grid_cell_occupancies &&
     lhs.time_stamp== rhs.time_stamp 
    );
}

std::size_t OccupancyGrid::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<float>()(cell_size_x);
    result = result * 31 + std::hash<float>()(cell_size_y);
    result = result * 31 + std::hash<uint32_t>()(cell_count_x);
    result = result * 31 + std::hash<uint32_t>()(cell_count_y);
    result = result * 31 + std::hash<float>()(size_x);
    result = result * 31 + std::hash<float>()(size_y);
    result = result * 31 + std::hash<float>()(reference_point_in_grid_x);
    result = result * 31 + std::hash<float>()(reference_point_in_grid_y);
    result = result * 31 + std::hash<float>()(reference_point_in_world_x);
    result = result * 31 + std::hash<float>()(reference_point_in_world_y);
    result = result * 31 + std::hash<float>()(ego_position_diff_x);
    result = result * 31 + std::hash<float>()(ego_position_diff_y);
    result = result * 31 + std::hash<float>()(ego_heading);
    result = result * 31 + hashVector(grid_cell_occupancies);
    result = result * 31 + std::hash<uint64_t>()(time_stamp);
    return result;
}

 } // namespace safetyhigh


CEREAL_BIND_TO_ARCHIVES(safetyhigh::OccupancyGrid)

 namespace safetyhigh {


Track::Track() = default;

Track::Track(
    uint64_t id_arg ,
    float existence_probability_arg ,
    float x_arg ,
    float y_arg ,
    float v_x_arg ,
    float v_y_arg ,
    float a_x_arg ,
    float a_y_arg ,
    float length_arg ,
    float length_variance_arg ,
    float width_arg ,
    float width_variance_arg ,
    std::vector<float> covariance_arg 
    ):
    id { std::move(id_arg) } // NOLINT
      ,
    existence_probability { std::move(existence_probability_arg) } // NOLINT
      ,
    x { std::move(x_arg) } // NOLINT
      ,
    y { std::move(y_arg) } // NOLINT
      ,
    v_x { std::move(v_x_arg) } // NOLINT
      ,
    v_y { std::move(v_y_arg) } // NOLINT
      ,
    a_x { std::move(a_x_arg) } // NOLINT
      ,
    a_y { std::move(a_y_arg) } // NOLINT
      ,
    length { std::move(length_arg) } // NOLINT
      ,
    length_variance { std::move(length_variance_arg) } // NOLINT
      ,
    width { std::move(width_arg) } // NOLINT
      ,
    width_variance { std::move(width_variance_arg) } // NOLINT
      ,
    covariance { std::move(covariance_arg) } // NOLINT
      
    {
}

Track::Track(const Track&) = default;

// NOLINTNEXTLINE
Track::Track(Track&&) = default;

Track& Track::operator=(const Track&) = default;

// NOLINTNEXTLINE
Track& Track::operator=(Track&&) = default;

void Track::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << id;out << ",";
    out << existence_probability;out << ",";
    out << x;out << ",";
    out << y;out << ",";
    out << v_x;out << ",";
    out << v_y;out << ",";
    out << a_x;out << ",";
    out << a_y;out << ",";
    out << length;out << ",";
    out << length_variance;out << ",";
    out << width;out << ",";
    out << width_variance;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!covariance.empty())
        {
            std::copy(covariance.begin(), covariance.end() - 1, std::ostream_iterator<float>(out, "},{"));
            std::copy(covariance.end()-1, covariance.end(), std::ostream_iterator<float>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream Track::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void Track::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(id);
        ar& CEREAL_NVP(existence_probability);
        ar& CEREAL_NVP(x);
        ar& CEREAL_NVP(y);
        ar& CEREAL_NVP(v_x);
        ar& CEREAL_NVP(v_y);
        ar& CEREAL_NVP(a_x);
        ar& CEREAL_NVP(a_y);
        ar& CEREAL_NVP(length);
        ar& CEREAL_NVP(length_variance);
        ar& CEREAL_NVP(width);
        ar& CEREAL_NVP(width_variance);
        ar& CEREAL_NVP(covariance);
    }


template void Track::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void Track::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void Track::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void Track::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void Track::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void Track::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void Track::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const Track& track)
{
    track.toCsv(os);
    return os;
}

bool operator== (const Track &lhs, const Track &rhs)
{
    return (  lhs.id== rhs.id &&
     lhs.existence_probability== rhs.existence_probability &&
     lhs.x== rhs.x &&
     lhs.y== rhs.y &&
     lhs.v_x== rhs.v_x &&
     lhs.v_y== rhs.v_y &&
     lhs.a_x== rhs.a_x &&
     lhs.a_y== rhs.a_y &&
     lhs.length== rhs.length &&
     lhs.length_variance== rhs.length_variance &&
     lhs.width== rhs.width &&
     lhs.width_variance== rhs.width_variance &&
     lhs.covariance== rhs.covariance 
    );
}

std::size_t Track::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<uint64_t>()(id);
    result = result * 31 + std::hash<float>()(existence_probability);
    result = result * 31 + std::hash<float>()(x);
    result = result * 31 + std::hash<float>()(y);
    result = result * 31 + std::hash<float>()(v_x);
    result = result * 31 + std::hash<float>()(v_y);
    result = result * 31 + std::hash<float>()(a_x);
    result = result * 31 + std::hash<float>()(a_y);
    result = result * 31 + std::hash<float>()(length);
    result = result * 31 + std::hash<float>()(length_variance);
    result = result * 31 + std::hash<float>()(width);
    result = result * 31 + std::hash<float>()(width_variance);
    result = result * 31 + hashVector(covariance);
    return result;
}

 } // namespace safetyhigh


CEREAL_BIND_TO_ARCHIVES(safetyhigh::Track)

 namespace safetyhigh {


TrackListData::TrackListData() = default;

TrackListData::TrackListData(
    uint32_t number_of_tracks_arg ,
    std::vector<Track> track_list_arg ,
    uint64_t time_stamp_arg 
    ):
    number_of_tracks { std::move(number_of_tracks_arg) } // NOLINT
      ,
    track_list { std::move(track_list_arg) } // NOLINT
      ,
    time_stamp { std::move(time_stamp_arg) } // NOLINT
      
    {
}

TrackListData::TrackListData(const TrackListData&) = default;

// NOLINTNEXTLINE
TrackListData::TrackListData(TrackListData&&) = default;

TrackListData& TrackListData::operator=(const TrackListData&) = default;

// NOLINTNEXTLINE
TrackListData& TrackListData::operator=(TrackListData&&) = default;

void TrackListData::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << number_of_tracks;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!track_list.empty())
        {
            std::copy(track_list.begin(), track_list.end() - 1, std::ostream_iterator<Track>(out, "},{"));
            std::copy(track_list.end()-1, track_list.end(), std::ostream_iterator<Track>(out, ""));
        }
        out << "}}\"";out << ",";
    out << time_stamp;
    
}

std::stringstream TrackListData::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void TrackListData::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(number_of_tracks);
        ar& CEREAL_NVP(track_list);
        ar& CEREAL_NVP(time_stamp);
    }


template void TrackListData::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void TrackListData::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void TrackListData::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void TrackListData::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void TrackListData::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void TrackListData::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void TrackListData::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const TrackListData& tracklistdata)
{
    tracklistdata.toCsv(os);
    return os;
}

bool operator== (const TrackListData &lhs, const TrackListData &rhs)
{
    return (  lhs.number_of_tracks== rhs.number_of_tracks &&
     lhs.track_list== rhs.track_list &&
     lhs.time_stamp== rhs.time_stamp 
    );
}

std::size_t TrackListData::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<uint32_t>()(number_of_tracks);
    result = result * 31 + hashVector(track_list);
    result = result * 31 + std::hash<uint64_t>()(time_stamp);
    return result;
}

 } // namespace safetyhigh


CEREAL_BIND_TO_ARCHIVES(safetyhigh::TrackListData)

 namespace safetyhigh {


LongitudinalBrakingRequest::LongitudinalBrakingRequest() = default;

LongitudinalBrakingRequest::LongitudinalBrakingRequest(
    BRAKING_REQ braking_request_arg ,
    uint64_t track_id_arg ,
    uint64_t grid_cell_index_arg ,
    uint64_t time_stamp_arg 
    ):
    braking_request { std::move(braking_request_arg) } // NOLINT
      ,
    track_id { std::move(track_id_arg) } // NOLINT
      ,
    grid_cell_index { std::move(grid_cell_index_arg) } // NOLINT
      ,
    time_stamp { std::move(time_stamp_arg) } // NOLINT
      
    {
}

LongitudinalBrakingRequest::LongitudinalBrakingRequest(const LongitudinalBrakingRequest&) = default;

// NOLINTNEXTLINE
LongitudinalBrakingRequest::LongitudinalBrakingRequest(LongitudinalBrakingRequest&&) = default;

LongitudinalBrakingRequest& LongitudinalBrakingRequest::operator=(const LongitudinalBrakingRequest&) = default;

// NOLINTNEXTLINE
LongitudinalBrakingRequest& LongitudinalBrakingRequest::operator=(LongitudinalBrakingRequest&&) = default;

void LongitudinalBrakingRequest::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << static_cast<uint32_t>(braking_request);out << ",";
    out << track_id;out << ",";
    out << grid_cell_index;out << ",";
    out << time_stamp;
    
}

std::stringstream LongitudinalBrakingRequest::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void LongitudinalBrakingRequest::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(braking_request);
        ar& CEREAL_NVP(track_id);
        ar& CEREAL_NVP(grid_cell_index);
        ar& CEREAL_NVP(time_stamp);
    }


template void LongitudinalBrakingRequest::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void LongitudinalBrakingRequest::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void LongitudinalBrakingRequest::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void LongitudinalBrakingRequest::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void LongitudinalBrakingRequest::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void LongitudinalBrakingRequest::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void LongitudinalBrakingRequest::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const LongitudinalBrakingRequest& longitudinalbrakingrequest)
{
    longitudinalbrakingrequest.toCsv(os);
    return os;
}

bool operator== (const LongitudinalBrakingRequest &lhs, const LongitudinalBrakingRequest &rhs)
{
    return (  lhs.braking_request== rhs.braking_request &&
     lhs.track_id== rhs.track_id &&
     lhs.grid_cell_index== rhs.grid_cell_index &&
     lhs.time_stamp== rhs.time_stamp 
    );
}

std::size_t LongitudinalBrakingRequest::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<BRAKING_REQ>()(braking_request);
    result = result * 31 + std::hash<uint64_t>()(track_id);
    result = result * 31 + std::hash<uint64_t>()(grid_cell_index);
    result = result * 31 + std::hash<uint64_t>()(time_stamp);
    return result;
}

 } // namespace safetyhigh


CEREAL_BIND_TO_ARCHIVES(safetyhigh::LongitudinalBrakingRequest)




 namespace safetyhigh {


BaselabsDiagnosticData::BaselabsDiagnosticData() = default;

BaselabsDiagnosticData::BaselabsDiagnosticData(
    EgoMotion ego_motion_grid_arg ,
    EgoMotion ego_motion_tracks_arg ,
    OccupancyGrid diag_occupancy_grid_arg ,
    TrackListData diag_track_list_arg ,
    LongitudinalBrakingRequest braking_request_obsa_arg ,
    LongitudinalBrakingRequest braking_request_sbsa_arg ,
    LongitudinalBrakingRequest braking_request_dm_arg ,
    uint64_t last_update_time_grid_arg ,
    uint64_t last_update_time_tracks_arg ,
    uint64_t time_stamp_arg 
    ):
    ego_motion_grid { std::move(ego_motion_grid_arg) } // NOLINT
      ,ego_motion_tracks { std::move(ego_motion_tracks_arg) } // NOLINT
      ,diag_occupancy_grid { std::move(diag_occupancy_grid_arg) } // NOLINT
      ,diag_track_list { std::move(diag_track_list_arg) } // NOLINT
      ,braking_request_obsa { std::move(braking_request_obsa_arg) } // NOLINT
      ,braking_request_sbsa { std::move(braking_request_sbsa_arg) } // NOLINT
      ,braking_request_dm { std::move(braking_request_dm_arg) } // NOLINT
      ,last_update_time_grid { std::move(last_update_time_grid_arg) } // NOLINT
      ,last_update_time_tracks { std::move(last_update_time_tracks_arg) } // NOLINT
      ,time_stamp { std::move(time_stamp_arg) } // NOLINT
      {
}

BaselabsDiagnosticData::BaselabsDiagnosticData(const BaselabsDiagnosticData&) = default;

// NOLINTNEXTLINE
BaselabsDiagnosticData::BaselabsDiagnosticData(BaselabsDiagnosticData&&) = default;

BaselabsDiagnosticData& BaselabsDiagnosticData::operator=(const BaselabsDiagnosticData&) = default;

// NOLINTNEXTLINE
BaselabsDiagnosticData& BaselabsDiagnosticData::operator=(BaselabsDiagnosticData&&) = default;

void BaselabsDiagnosticData::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << ego_motion_grid;out << ",";
    out << ego_motion_tracks;out << ",";
    out << diag_occupancy_grid;out << ",";
    out << diag_track_list;out << ",";
    out << braking_request_obsa;out << ",";
    out << braking_request_sbsa;out << ",";
    out << braking_request_dm;out << ",";
    out << last_update_time_grid;out << ",";
    out << last_update_time_tracks;out << ",";
    out << time_stamp;
    
}

std::stringstream BaselabsDiagnosticData::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void BaselabsDiagnosticData::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(ego_motion_grid);
        ar& CEREAL_NVP(ego_motion_tracks);
        ar& CEREAL_NVP(diag_occupancy_grid);
        ar& CEREAL_NVP(diag_track_list);
        ar& CEREAL_NVP(braking_request_obsa);
        ar& CEREAL_NVP(braking_request_sbsa);
        ar& CEREAL_NVP(braking_request_dm);
        ar& CEREAL_NVP(last_update_time_grid);
        ar& CEREAL_NVP(last_update_time_tracks);
        ar& CEREAL_NVP(time_stamp);
    }


template void BaselabsDiagnosticData::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void BaselabsDiagnosticData::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void BaselabsDiagnosticData::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void BaselabsDiagnosticData::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void BaselabsDiagnosticData::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void BaselabsDiagnosticData::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void BaselabsDiagnosticData::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const BaselabsDiagnosticData& baselabsdiagnosticdata)
{
    baselabsdiagnosticdata.toCsv(os);
    return os;
}

bool operator== (const BaselabsDiagnosticData &lhs, const BaselabsDiagnosticData &rhs)
{
    return (  lhs.ego_motion_grid== rhs.ego_motion_grid &&
     lhs.ego_motion_tracks== rhs.ego_motion_tracks &&
     lhs.diag_occupancy_grid== rhs.diag_occupancy_grid &&
     lhs.diag_track_list== rhs.diag_track_list &&
     lhs.braking_request_obsa== rhs.braking_request_obsa &&
     lhs.braking_request_sbsa== rhs.braking_request_sbsa &&
     lhs.braking_request_dm== rhs.braking_request_dm &&
     lhs.last_update_time_grid== rhs.last_update_time_grid &&
     lhs.last_update_time_tracks== rhs.last_update_time_tracks &&
     lhs.time_stamp== rhs.time_stamp 
    );
}

std::size_t BaselabsDiagnosticData::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<EgoMotion>()(ego_motion_grid);
    result = result * 31 + std::hash<EgoMotion>()(ego_motion_tracks);
    result = result * 31 + std::hash<OccupancyGrid>()(diag_occupancy_grid);
    result = result * 31 + std::hash<TrackListData>()(diag_track_list);
    result = result * 31 + std::hash<LongitudinalBrakingRequest>()(braking_request_obsa);
    result = result * 31 + std::hash<LongitudinalBrakingRequest>()(braking_request_sbsa);
    result = result * 31 + std::hash<LongitudinalBrakingRequest>()(braking_request_dm);
    result = result * 31 + std::hash<uint64_t>()(last_update_time_grid);
    result = result * 31 + std::hash<uint64_t>()(last_update_time_tracks);
    result = result * 31 + std::hash<uint64_t>()(time_stamp);
    return result;
}

 } // namespace safetyhigh


CEREAL_BIND_TO_ARCHIVES(safetyhigh::BaselabsDiagnosticData)

