#include <interfaces/DsrcMessage/DsrcMessage.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


Signal::Signal() = default;

Signal::Signal(
    uint64_t signalId_arg ,
    LIGHT_COLOR color_arg ,
    LIGHT_SHAPE shape_arg ,
    LIGHT_STATE state_arg 
    ):
    signalId { std::move(signalId_arg) } // NOLINT
      ,
    color { std::move(color_arg) } // NOLINT
      ,
    shape { std::move(shape_arg) } // NOLINT
      ,
    state { std::move(state_arg) } // NOLINT
      
    {
}

Signal::Signal(const Signal&) = default;

// NOLINTNEXTLINE
Signal::Signal(Signal&&) = default;

Signal& Signal::operator=(const Signal&) = default;

// NOLINTNEXTLINE
Signal& Signal::operator=(Signal&&) = default;

void Signal::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << signalId;out << ",";
    out << static_cast<uint32_t>(color);out << ",";
    out << static_cast<uint32_t>(shape);out << ",";
    out << static_cast<uint32_t>(state);
    
}

std::stringstream Signal::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void Signal::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(signalId);
        ar& CEREAL_NVP(color);
        ar& CEREAL_NVP(shape);
        ar& CEREAL_NVP(state);
    }


template void Signal::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void Signal::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void Signal::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void Signal::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void Signal::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void Signal::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void Signal::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const Signal& signal)
{
    signal.toCsv(os);
    return os;
}

bool operator== (const Signal &lhs, const Signal &rhs)
{
    return (  lhs.signalId== rhs.signalId &&
     lhs.color== rhs.color &&
     lhs.shape== rhs.shape &&
     lhs.state== rhs.state 
    );
}

std::size_t Signal::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<uint64_t>()(signalId);
    result = result * 31 + std::hash<LIGHT_COLOR>()(color);
    result = result * 31 + std::hash<LIGHT_SHAPE>()(shape);
    result = result * 31 + std::hash<LIGHT_STATE>()(state);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::Signal)

 namespace av {
 namespace msg {


IntersectionState::IntersectionState() = default;

IntersectionState::IntersectionState(
    uint64_t intersectionId_arg ,
    std::vector<Signal> signals_arg 
    ):
    intersectionId { std::move(intersectionId_arg) } // NOLINT
      ,
    signals { std::move(signals_arg) } // NOLINT
      
    {
}

IntersectionState::IntersectionState(const IntersectionState&) = default;

// NOLINTNEXTLINE
IntersectionState::IntersectionState(IntersectionState&&) = default;

IntersectionState& IntersectionState::operator=(const IntersectionState&) = default;

// NOLINTNEXTLINE
IntersectionState& IntersectionState::operator=(IntersectionState&&) = default;

void IntersectionState::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << intersectionId;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!signals.empty())
        {
            std::copy(signals.begin(), signals.end() - 1, std::ostream_iterator<Signal>(out, "},{"));
            std::copy(signals.end()-1, signals.end(), std::ostream_iterator<Signal>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream IntersectionState::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void IntersectionState::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(intersectionId);
        ar& CEREAL_NVP(signals);
    }


template void IntersectionState::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void IntersectionState::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void IntersectionState::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void IntersectionState::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void IntersectionState::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void IntersectionState::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void IntersectionState::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const IntersectionState& intersectionstate)
{
    intersectionstate.toCsv(os);
    return os;
}

bool operator== (const IntersectionState &lhs, const IntersectionState &rhs)
{
    return (  lhs.intersectionId== rhs.intersectionId &&
     lhs.signals== rhs.signals 
    );
}

std::size_t IntersectionState::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<uint64_t>()(intersectionId);
    result = result * 31 + hashVector(signals);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::IntersectionState)




 namespace av {
 namespace msg {


DsrcMessage::DsrcMessage() = default;

DsrcMessage::DsrcMessage(
    TimePoint timestamp_arg ,
    std::vector<IntersectionState> intersectionStates_arg 
    ):
    timestamp { std::move(timestamp_arg) } // NOLINT
      ,intersectionStates { std::move(intersectionStates_arg) } // NOLINT
      {
}

DsrcMessage::DsrcMessage(const DsrcMessage&) = default;

// NOLINTNEXTLINE
DsrcMessage::DsrcMessage(DsrcMessage&&) = default;

DsrcMessage& DsrcMessage::operator=(const DsrcMessage&) = default;

// NOLINTNEXTLINE
DsrcMessage& DsrcMessage::operator=(DsrcMessage&&) = default;

void DsrcMessage::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << timestamp;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!intersectionStates.empty())
        {
            std::copy(intersectionStates.begin(), intersectionStates.end() - 1, std::ostream_iterator<IntersectionState>(out, "},{"));
            std::copy(intersectionStates.end()-1, intersectionStates.end(), std::ostream_iterator<IntersectionState>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream DsrcMessage::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void DsrcMessage::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(timestamp);
        ar& CEREAL_NVP(intersectionStates);
    }


template void DsrcMessage::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void DsrcMessage::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void DsrcMessage::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void DsrcMessage::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void DsrcMessage::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void DsrcMessage::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void DsrcMessage::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const DsrcMessage& dsrcmessage)
{
    dsrcmessage.toCsv(os);
    return os;
}

bool operator== (const DsrcMessage &lhs, const DsrcMessage &rhs)
{
    return (  lhs.timestamp== rhs.timestamp &&
     lhs.intersectionStates== rhs.intersectionStates 
    );
}

std::size_t DsrcMessage::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + 0;
    result = result * 31 + hashVector(intersectionStates);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::DsrcMessage)

