#include <interfaces/H264Frame/H264Frame.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>





 namespace av {
 namespace camera {


H264Frame::H264Frame() = default;

H264Frame::H264Frame(
    std::vector<uint8_t> data_arg ,
    uint32_t sequenceNumber_arg ,
    H264FrameMeta metadata_arg 
    ):
    data { std::move(data_arg) } // NOLINT
      ,sequenceNumber { std::move(sequenceNumber_arg) } // NOLINT
      ,metadata { std::move(metadata_arg) } // NOLINT
      {
}

H264Frame::H264Frame(const H264Frame&) = default;

// NOLINTNEXTLINE
H264Frame::H264Frame(H264Frame&&) = default;

H264Frame& H264Frame::operator=(const H264Frame&) = default;

// NOLINTNEXTLINE
H264Frame& H264Frame::operator=(H264Frame&&) = default;

void H264Frame::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!data.empty())
        {
            std::copy(data.begin(), data.end() - 1, std::ostream_iterator<uint8_t>(out, "},{"));
            std::copy(data.end()-1, data.end(), std::ostream_iterator<uint8_t>(out, ""));
        }
        out << "}}\"";out << ",";
    out << sequenceNumber;out << ",";
    out << metadata;
    
}

std::stringstream H264Frame::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void H264Frame::serialize(Archive& ar, unsigned int version)
{
    ar& CEREAL_NVP(data);
    
    if (version >= 2) {
        ar& CEREAL_NVP(sequenceNumber);
        ar& CEREAL_NVP(metadata);
    } else {
        ar& cereal::make_nvp("cameraId", metadata.cameraId);
        ar& cereal::make_nvp("width", metadata.width);
        ar& cereal::make_nvp("height", metadata.height);
        ar& cereal::make_nvp("syncTriggerOffset_us", metadata.syncTriggerOffset_us);
        ar& cereal::make_nvp("captureTime_us", metadata.captureTime_us);
        ar& cereal::make_nvp("receiveTime_us", metadata.receiveTime_us);
        ar& cereal::make_nvp("ingestTime_us", metadata.ingestTime_us);
        ar& cereal::make_nvp("exposure", metadata.exposureTime_us[0]);
        ar& cereal::make_nvp("analogGain", metadata.analogGain[0]);
        ar& cereal::make_nvp("digitalGain", metadata.digitalGain[0]);
        ar& cereal::make_nvp("whiteBalance", metadata.whiteBalance.red);
    }
}


template void H264Frame::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void H264Frame::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void H264Frame::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void H264Frame::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void H264Frame::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void H264Frame::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void H264Frame::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const H264Frame& h264frame)
{
    h264frame.toCsv(os);
    return os;
}

bool operator== (const H264Frame &lhs, const H264Frame &rhs)
{
    return (  lhs.data== rhs.data &&
     lhs.sequenceNumber== rhs.sequenceNumber &&
     lhs.metadata== rhs.metadata 
    );
}

std::size_t H264Frame::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + hashVector(data);
    result = result * 31 + std::hash<uint32_t>()(sequenceNumber);
    result = result * 31 + std::hash<H264FrameMeta>()(metadata);
    return result;
}

 } // namespace av
 } // namespace camera


CEREAL_BIND_TO_ARCHIVES(av::camera::H264Frame)

