#include <interfaces/HostTaskList/HostTaskList.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


TaskIdentity::TaskIdentity() = default;

TaskIdentity::TaskIdentity(
    int32_t pid_arg ,
    int32_t status_arg ,
    std::string instanceName_arg ,
    std::string executableName_arg 
    ):
    pid { std::move(pid_arg) } // NOLINT
      ,
    status { std::move(status_arg) } // NOLINT
      ,
    instanceName { std::move(instanceName_arg) } // NOLINT
      ,
    executableName { std::move(executableName_arg) } // NOLINT
      
    {
}

TaskIdentity::TaskIdentity(const TaskIdentity&) = default;

// NOLINTNEXTLINE
TaskIdentity::TaskIdentity(TaskIdentity&&) = default;

TaskIdentity& TaskIdentity::operator=(const TaskIdentity&) = default;

// NOLINTNEXTLINE
TaskIdentity& TaskIdentity::operator=(TaskIdentity&&) = default;

void TaskIdentity::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << pid;out << ",";
    out << status;out << ",";
    out << "\"";
        out << instanceName;
        out << "\"";out << ",";
    out << "\"";
        out << executableName;
        out << "\"";
    
}

std::stringstream TaskIdentity::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void TaskIdentity::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(pid);
        ar& CEREAL_NVP(status);
        ar& CEREAL_NVP(instanceName);
        ar& CEREAL_NVP(executableName);
    }


template void TaskIdentity::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void TaskIdentity::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void TaskIdentity::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void TaskIdentity::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void TaskIdentity::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void TaskIdentity::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void TaskIdentity::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const TaskIdentity& taskidentity)
{
    taskidentity.toCsv(os);
    return os;
}

bool operator== (const TaskIdentity &lhs, const TaskIdentity &rhs)
{
    return (  lhs.pid== rhs.pid &&
     lhs.status== rhs.status &&
     lhs.instanceName== rhs.instanceName &&
     lhs.executableName== rhs.executableName 
    );
}

std::size_t TaskIdentity::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<int32_t>()(pid);
    result = result * 31 + std::hash<int32_t>()(status);
    result = result * 31 + std::hash<std::string>()(instanceName);
    result = result * 31 + std::hash<std::string>()(executableName);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::TaskIdentity)




 namespace av {
 namespace msg {


HostTaskList::HostTaskList() = default;

HostTaskList::HostTaskList(
    std::string hostname_arg ,
    std::vector<TaskIdentity> tasks_arg 
    ):
    hostname { std::move(hostname_arg) } // NOLINT
      ,tasks { std::move(tasks_arg) } // NOLINT
      {
}

HostTaskList::HostTaskList(const HostTaskList&) = default;

// NOLINTNEXTLINE
HostTaskList::HostTaskList(HostTaskList&&) = default;

HostTaskList& HostTaskList::operator=(const HostTaskList&) = default;

// NOLINTNEXTLINE
HostTaskList& HostTaskList::operator=(HostTaskList&&) = default;

void HostTaskList::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << "\"";
        out << hostname;
        out << "\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!tasks.empty())
        {
            std::copy(tasks.begin(), tasks.end() - 1, std::ostream_iterator<TaskIdentity>(out, "},{"));
            std::copy(tasks.end()-1, tasks.end(), std::ostream_iterator<TaskIdentity>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream HostTaskList::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void HostTaskList::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(hostname);
        ar& CEREAL_NVP(tasks);
    }


template void HostTaskList::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void HostTaskList::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void HostTaskList::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void HostTaskList::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void HostTaskList::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void HostTaskList::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void HostTaskList::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const HostTaskList& hosttasklist)
{
    hosttasklist.toCsv(os);
    return os;
}

bool operator== (const HostTaskList &lhs, const HostTaskList &rhs)
{
    return (  lhs.hostname== rhs.hostname &&
     lhs.tasks== rhs.tasks 
    );
}

std::size_t HostTaskList::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<std::string>()(hostname);
    result = result * 31 + hashVector(tasks);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::HostTaskList)

