#include <interfaces/LaneGraph/LaneGraph.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


GraphVertex::GraphVertex() = default;

GraphVertex::GraphVertex(
    MapId vertexID_arg 
    ):
    vertexID { std::move(vertexID_arg) } // NOLINT
      
    {
}

GraphVertex::GraphVertex(const GraphVertex&) = default;

// NOLINTNEXTLINE
GraphVertex::GraphVertex(GraphVertex&&) = default;

GraphVertex& GraphVertex::operator=(const GraphVertex&) = default;

// NOLINTNEXTLINE
GraphVertex& GraphVertex::operator=(GraphVertex&&) = default;

void GraphVertex::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << vertexID;
    
}

std::stringstream GraphVertex::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void GraphVertex::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(vertexID);
    }


template void GraphVertex::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void GraphVertex::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void GraphVertex::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void GraphVertex::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void GraphVertex::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void GraphVertex::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void GraphVertex::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const GraphVertex& graphvertex)
{
    graphvertex.toCsv(os);
    return os;
}

bool operator== (const GraphVertex &lhs, const GraphVertex &rhs)
{
    return (  lhs.vertexID== rhs.vertexID 
    );
}

std::size_t GraphVertex::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<MapId>()(vertexID);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::GraphVertex)

 namespace av {
 namespace msg {


GraphEdge::GraphEdge() = default;

GraphEdge::GraphEdge(
    MapId fromVertexID_arg ,
    MapId toVertexID_arg ,
    std::string edgeType_arg 
    ):
    fromVertexID { std::move(fromVertexID_arg) } // NOLINT
      ,
    toVertexID { std::move(toVertexID_arg) } // NOLINT
      ,
    edgeType { std::move(edgeType_arg) } // NOLINT
      
    {
}

GraphEdge::GraphEdge(const GraphEdge&) = default;

// NOLINTNEXTLINE
GraphEdge::GraphEdge(GraphEdge&&) = default;

GraphEdge& GraphEdge::operator=(const GraphEdge&) = default;

// NOLINTNEXTLINE
GraphEdge& GraphEdge::operator=(GraphEdge&&) = default;

void GraphEdge::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << fromVertexID;out << ",";
    out << toVertexID;out << ",";
    out << "\"";
        out << edgeType;
        out << "\"";
    
}

std::stringstream GraphEdge::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void GraphEdge::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(fromVertexID);
        ar& CEREAL_NVP(toVertexID);
        ar& CEREAL_NVP(edgeType);
    }


template void GraphEdge::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void GraphEdge::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void GraphEdge::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void GraphEdge::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void GraphEdge::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void GraphEdge::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void GraphEdge::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const GraphEdge& graphedge)
{
    graphedge.toCsv(os);
    return os;
}

bool operator== (const GraphEdge &lhs, const GraphEdge &rhs)
{
    return (  lhs.fromVertexID== rhs.fromVertexID &&
     lhs.toVertexID== rhs.toVertexID &&
     lhs.edgeType== rhs.edgeType 
    );
}

std::size_t GraphEdge::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<MapId>()(fromVertexID);
    result = result * 31 + std::hash<MapId>()(toVertexID);
    result = result * 31 + std::hash<std::string>()(edgeType);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::GraphEdge)




 namespace av {
 namespace msg {


LaneGraph::LaneGraph() = default;

LaneGraph::LaneGraph(
    uint32_t numVertices_arg ,
    uint32_t numEdges_arg ,
    std::vector<GraphVertex> vertices_arg ,
    std::vector<GraphEdge> edges_arg 
    ):
    numVertices { std::move(numVertices_arg) } // NOLINT
      ,numEdges { std::move(numEdges_arg) } // NOLINT
      ,vertices { std::move(vertices_arg) } // NOLINT
      ,edges { std::move(edges_arg) } // NOLINT
      {
}

LaneGraph::LaneGraph(const LaneGraph&) = default;

// NOLINTNEXTLINE
LaneGraph::LaneGraph(LaneGraph&&) = default;

LaneGraph& LaneGraph::operator=(const LaneGraph&) = default;

// NOLINTNEXTLINE
LaneGraph& LaneGraph::operator=(LaneGraph&&) = default;

void LaneGraph::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << numVertices;out << ",";
    out << numEdges;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!vertices.empty())
        {
            std::copy(vertices.begin(), vertices.end() - 1, std::ostream_iterator<GraphVertex>(out, "},{"));
            std::copy(vertices.end()-1, vertices.end(), std::ostream_iterator<GraphVertex>(out, ""));
        }
        out << "}}\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!edges.empty())
        {
            std::copy(edges.begin(), edges.end() - 1, std::ostream_iterator<GraphEdge>(out, "},{"));
            std::copy(edges.end()-1, edges.end(), std::ostream_iterator<GraphEdge>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream LaneGraph::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void LaneGraph::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(numVertices);
        ar& CEREAL_NVP(numEdges);
        ar& CEREAL_NVP(vertices);
        ar& CEREAL_NVP(edges);
    }


template void LaneGraph::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void LaneGraph::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void LaneGraph::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void LaneGraph::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void LaneGraph::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void LaneGraph::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void LaneGraph::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const LaneGraph& lanegraph)
{
    lanegraph.toCsv(os);
    return os;
}

bool operator== (const LaneGraph &lhs, const LaneGraph &rhs)
{
    return (  lhs.numVertices== rhs.numVertices &&
     lhs.numEdges== rhs.numEdges &&
     lhs.vertices== rhs.vertices &&
     lhs.edges== rhs.edges 
    );
}

std::size_t LaneGraph::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<uint32_t>()(numVertices);
    result = result * 31 + std::hash<uint32_t>()(numEdges);
    result = result * 31 + hashVector(vertices);
    result = result * 31 + hashVector(edges);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::LaneGraph)

