#include <interfaces/LidarScan/LidarScan.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


LidarReturn::LidarReturn() = default;

LidarReturn::LidarReturn(
    float x_arg ,
    float y_arg ,
    float z_arg ,
    float range_m_arg ,
    float azimuth_rad_arg ,
    float elevation_rad_arg ,
    float intensity_arg ,
    uint8_t layer_arg ,
    uint8_t echo_arg ,
    uint16_t echo_pulse_width_arg ,
    uint32_t flags_arg ,
    TimePoint timeStamp_arg ,
    TimePoint sensorTimeStamp_arg ,
    uint16_t weight_arg 
    ):
    x { std::move(x_arg) } // NOLINT
      ,
    y { std::move(y_arg) } // NOLINT
      ,
    z { std::move(z_arg) } // NOLINT
      ,
    range_m { std::move(range_m_arg) } // NOLINT
      ,
    azimuth_rad { std::move(azimuth_rad_arg) } // NOLINT
      ,
    elevation_rad { std::move(elevation_rad_arg) } // NOLINT
      ,
    intensity { std::move(intensity_arg) } // NOLINT
      ,
    layer { std::move(layer_arg) } // NOLINT
      ,
    echo { std::move(echo_arg) } // NOLINT
      ,
    echo_pulse_width { std::move(echo_pulse_width_arg) } // NOLINT
      ,
    flags { std::move(flags_arg) } // NOLINT
      ,
    timeStamp { std::move(timeStamp_arg) } // NOLINT
      ,
    sensorTimeStamp { std::move(sensorTimeStamp_arg) } // NOLINT
      ,
    weight { std::move(weight_arg) } // NOLINT
      
    {
}

LidarReturn::LidarReturn(const LidarReturn&) = default;

// NOLINTNEXTLINE
LidarReturn::LidarReturn(LidarReturn&&) = default;

LidarReturn& LidarReturn::operator=(const LidarReturn&) = default;

// NOLINTNEXTLINE
LidarReturn& LidarReturn::operator=(LidarReturn&&) = default;

void LidarReturn::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << x;out << ",";
    out << y;out << ",";
    out << z;out << ",";
    out << range_m;out << ",";
    out << azimuth_rad;out << ",";
    out << elevation_rad;out << ",";
    out << intensity;out << ",";
    // Cast to uint16_t to print as integer instead of char
        out << static_cast<uint16_t>(layer);out << ",";
    // Cast to uint16_t to print as integer instead of char
        out << static_cast<uint16_t>(echo);out << ",";
    out << echo_pulse_width;out << ",";
    out << flags;out << ",";
    out << timeStamp;out << ",";
    out << sensorTimeStamp;out << ",";
    out << weight;
    
}

std::stringstream LidarReturn::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void LidarReturn::serialize(Archive& ar, const unsigned int version)
{
	ar& CEREAL_NVP(x) & CEREAL_NVP(y) & CEREAL_NVP(z);
	ar& CEREAL_NVP(intensity);

	if (version < 7)
	{
		otimePoint otp;
		ar& cereal::make_nvp("timeStamp", otp);
		timeStamp = av::TimePoint(std::chrono::seconds(otp.tv_sec) + std::chrono::nanoseconds(otp.tv_nsec));
	}
	else
	{
		ar& CEREAL_NVP(timeStamp);
	}

	if (version > 1) {
	    ar& CEREAL_NVP(layer);
	    ar& CEREAL_NVP(echo) & CEREAL_NVP(echo_pulse_width);
	    ar& CEREAL_NVP(flags);
	}

	if (version > 3) {
	    ar& CEREAL_NVP(range_m);
	    ar& CEREAL_NVP(azimuth_rad) & CEREAL_NVP(elevation_rad);
	}

	if (version > 4) {
		if (version < 7)
		{
			otimePoint sensor_otp;
			ar& cereal::make_nvp("sensorTimeStamp", sensor_otp);
			sensorTimeStamp = av::TimePoint(std::chrono::seconds(sensor_otp.tv_sec) + std::chrono::nanoseconds(sensor_otp.tv_nsec));
		}
		else
		{
			ar& CEREAL_NVP(sensorTimeStamp);
		}
	}

	if (version > 5) {
	    ar& CEREAL_NVP(weight);
	}
}

}} // to close av::msg namespace temporary

struct LidarReturn : public av::msg::LidarReturn{};
CEREAL_CLASS_VERSION(LidarReturn, 7)
CEREAL_REGISTER_TYPE(LidarReturn)

namespace av {
namespace msg {



template void LidarReturn::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void LidarReturn::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void LidarReturn::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void LidarReturn::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void LidarReturn::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void LidarReturn::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void LidarReturn::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const LidarReturn& lidarreturn)
{
    lidarreturn.toCsv(os);
    return os;
}

bool operator== (const LidarReturn &lhs, const LidarReturn &rhs)
{
    return (  lhs.x== rhs.x &&
     lhs.y== rhs.y &&
     lhs.z== rhs.z &&
     lhs.range_m== rhs.range_m &&
     lhs.azimuth_rad== rhs.azimuth_rad &&
     lhs.elevation_rad== rhs.elevation_rad &&
     lhs.intensity== rhs.intensity &&
     lhs.layer== rhs.layer &&
     lhs.echo== rhs.echo &&
     lhs.echo_pulse_width== rhs.echo_pulse_width &&
     lhs.flags== rhs.flags &&
     lhs.timeStamp== rhs.timeStamp &&
     lhs.sensorTimeStamp== rhs.sensorTimeStamp &&
     lhs.weight== rhs.weight 
    );
}

std::size_t LidarReturn::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<float>()(x);
    result = result * 31 + std::hash<float>()(y);
    result = result * 31 + std::hash<float>()(z);
    result = result * 31 + std::hash<float>()(range_m);
    result = result * 31 + std::hash<float>()(azimuth_rad);
    result = result * 31 + std::hash<float>()(elevation_rad);
    result = result * 31 + std::hash<float>()(intensity);
    result = result * 31 + std::hash<uint8_t>()(layer);
    result = result * 31 + std::hash<uint8_t>()(echo);
    result = result * 31 + std::hash<uint16_t>()(echo_pulse_width);
    result = result * 31 + std::hash<uint32_t>()(flags);
    result = result * 31 + 0;
    result = result * 31 + 0;
    result = result * 31 + std::hash<uint16_t>()(weight);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::LidarReturn)




 namespace av {
 namespace msg {


LidarScan::LidarScan() = default;

LidarScan::LidarScan(
    bool hasIntensity_arg ,
    bool downSampled_arg ,
    std::string sensorName_arg ,
    std::vector<LidarReturn> data_arg 
    ):
    hasIntensity { std::move(hasIntensity_arg) } // NOLINT
      ,downSampled { std::move(downSampled_arg) } // NOLINT
      ,sensorName { std::move(sensorName_arg) } // NOLINT
      ,data { std::move(data_arg) } // NOLINT
      {
}

LidarScan::LidarScan(const LidarScan&) = default;

// NOLINTNEXTLINE
LidarScan::LidarScan(LidarScan&&) = default;

LidarScan& LidarScan::operator=(const LidarScan&) = default;

// NOLINTNEXTLINE
LidarScan& LidarScan::operator=(LidarScan&&) = default;

void LidarScan::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << hasIntensity;out << ",";
    out << downSampled;out << ",";
    out << "\"";
        out << sensorName;
        out << "\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!data.empty())
        {
            std::copy(data.begin(), data.end() - 1, std::ostream_iterator<LidarReturn>(out, "},{"));
            std::copy(data.end()-1, data.end(), std::ostream_iterator<LidarReturn>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream LidarScan::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void LidarScan::serialize(Archive& ar, const unsigned int version)
{
	ar& CEREAL_NVP(sensorName);
	if (version > 2) {
		ar& CEREAL_NVP(hasIntensity);
		ar& cereal::make_nvp("LidarReturn", data);
	}
	if (version > 3) {
		ar& CEREAL_NVP(downSampled);
	}
}

}} // to close av::msg namespace temporary

struct LidarScan : public av::msg::LidarScan{};
CEREAL_CLASS_VERSION(LidarScan, 5)
CEREAL_REGISTER_TYPE(LidarScan)

namespace av {
namespace msg {




template void LidarScan::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void LidarScan::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void LidarScan::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void LidarScan::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void LidarScan::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void LidarScan::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void LidarScan::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const LidarScan& lidarscan)
{
    lidarscan.toCsv(os);
    return os;
}

bool operator== (const LidarScan &lhs, const LidarScan &rhs)
{
    return (  lhs.hasIntensity== rhs.hasIntensity &&
     lhs.downSampled== rhs.downSampled &&
     lhs.sensorName== rhs.sensorName &&
     lhs.data== rhs.data 
    );
}

std::size_t LidarScan::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<bool>()(hasIntensity);
    result = result * 31 + std::hash<bool>()(downSampled);
    result = result * 31 + std::hash<std::string>()(sensorName);
    result = result * 31 + hashVector(data);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::LidarScan)

