#include <interfaces/MRMCommand/MRMCommand.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


Response::Response() = default;

Response::Response(
    Parameter parameter_arg ,
    MRMType mrmType_arg 
    ):
    parameter { std::move(parameter_arg) } // NOLINT
      ,
    mrmType { std::move(mrmType_arg) } // NOLINT
      
    {
}

Response::Response(const Response&) = default;

// NOLINTNEXTLINE
Response::Response(Response&&) = default;

Response& Response::operator=(const Response&) = default;

// NOLINTNEXTLINE
Response& Response::operator=(Response&&) = default;

void Response::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << parameter;out << ",";
    out << static_cast<uint32_t>(mrmType);
    
}

std::stringstream Response::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void Response::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(parameter);
        ar& CEREAL_NVP(mrmType);
    }


template void Response::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void Response::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void Response::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void Response::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void Response::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void Response::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void Response::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const Response& response)
{
    response.toCsv(os);
    return os;
}

bool operator== (const Response &lhs, const Response &rhs)
{
    return (  lhs.parameter== rhs.parameter &&
     lhs.mrmType== rhs.mrmType 
    );
}

std::size_t Response::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<Parameter>()(parameter);
    result = result * 31 + std::hash<MRMType>()(mrmType);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::Response)




 namespace av {
 namespace msg {


MRMCommand::MRMCommand() = default;

MRMCommand::MRMCommand(
    TimePoint timestamp_arg ,
    std::vector<Response> allReactions_arg ,
    Response activeReaction_arg 
    ):
    timestamp { std::move(timestamp_arg) } // NOLINT
      ,allReactions { std::move(allReactions_arg) } // NOLINT
      ,activeReaction { std::move(activeReaction_arg) } // NOLINT
      {
}

MRMCommand::MRMCommand(const MRMCommand&) = default;

// NOLINTNEXTLINE
MRMCommand::MRMCommand(MRMCommand&&) = default;

MRMCommand& MRMCommand::operator=(const MRMCommand&) = default;

// NOLINTNEXTLINE
MRMCommand& MRMCommand::operator=(MRMCommand&&) = default;

void MRMCommand::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << timestamp;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!allReactions.empty())
        {
            std::copy(allReactions.begin(), allReactions.end() - 1, std::ostream_iterator<Response>(out, "},{"));
            std::copy(allReactions.end()-1, allReactions.end(), std::ostream_iterator<Response>(out, ""));
        }
        out << "}}\"";out << ",";
    out << activeReaction;
    
}

std::stringstream MRMCommand::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void MRMCommand::serialize(Archive& ar, unsigned int version)
{
    if (version >= 2) {
        ar& CEREAL_NVP(timestamp);
    }
    
    ar& CEREAL_NVP(allReactions);
    ar& CEREAL_NVP(activeReaction);
}



template void MRMCommand::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void MRMCommand::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void MRMCommand::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void MRMCommand::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void MRMCommand::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void MRMCommand::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void MRMCommand::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const MRMCommand& mrmcommand)
{
    mrmcommand.toCsv(os);
    return os;
}

bool operator== (const MRMCommand &lhs, const MRMCommand &rhs)
{
    return (  lhs.timestamp== rhs.timestamp &&
     lhs.allReactions== rhs.allReactions &&
     lhs.activeReaction== rhs.activeReaction 
    );
}

std::size_t MRMCommand::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + 0;
    result = result * 31 + hashVector(allReactions);
    result = result * 31 + std::hash<Response>()(activeReaction);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::MRMCommand)

