#include <interfaces/MatrixImageSpaceInfo/MatrixImageSpaceInfo.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>





 namespace av {
 namespace horizon {


MatrixImageSpaceInfo::MatrixImageSpaceInfo() = default;

MatrixImageSpaceInfo::MatrixImageSpaceInfo(
    BoundingBox2D box2d_arg ,
    BoundingBox3D box3d_arg ,
    std::vector<Point3D> poly_arg ,
    double sigmaWidth_arg ,
    double sigmaHeight_arg ,
    double sigmaX_arg ,
    double sigmaY_arg 
    ):
    box2d { std::move(box2d_arg) } // NOLINT
      ,box3d { std::move(box3d_arg) } // NOLINT
      ,poly { std::move(poly_arg) } // NOLINT
      ,sigmaWidth { std::move(sigmaWidth_arg) } // NOLINT
      ,sigmaHeight { std::move(sigmaHeight_arg) } // NOLINT
      ,sigmaX { std::move(sigmaX_arg) } // NOLINT
      ,sigmaY { std::move(sigmaY_arg) } // NOLINT
      {
}

MatrixImageSpaceInfo::MatrixImageSpaceInfo(const MatrixImageSpaceInfo&) = default;

// NOLINTNEXTLINE
MatrixImageSpaceInfo::MatrixImageSpaceInfo(MatrixImageSpaceInfo&&) = default;

MatrixImageSpaceInfo& MatrixImageSpaceInfo::operator=(const MatrixImageSpaceInfo&) = default;

// NOLINTNEXTLINE
MatrixImageSpaceInfo& MatrixImageSpaceInfo::operator=(MatrixImageSpaceInfo&&) = default;

void MatrixImageSpaceInfo::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << box2d;out << ",";
    out << box3d;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!poly.empty())
        {
            std::copy(poly.begin(), poly.end() - 1, std::ostream_iterator<Point3D>(out, "},{"));
            std::copy(poly.end()-1, poly.end(), std::ostream_iterator<Point3D>(out, ""));
        }
        out << "}}\"";out << ",";
    out << sigmaWidth;out << ",";
    out << sigmaHeight;out << ",";
    out << sigmaX;out << ",";
    out << sigmaY;
    
}

std::stringstream MatrixImageSpaceInfo::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void MatrixImageSpaceInfo::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(box2d);
        ar& CEREAL_NVP(box3d);
        ar& CEREAL_NVP(poly);
        ar& CEREAL_NVP(sigmaWidth);
        ar& CEREAL_NVP(sigmaHeight);
        ar& CEREAL_NVP(sigmaX);
        ar& CEREAL_NVP(sigmaY);
    }


template void MatrixImageSpaceInfo::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void MatrixImageSpaceInfo::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void MatrixImageSpaceInfo::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void MatrixImageSpaceInfo::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void MatrixImageSpaceInfo::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void MatrixImageSpaceInfo::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void MatrixImageSpaceInfo::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const MatrixImageSpaceInfo& matriximagespaceinfo)
{
    matriximagespaceinfo.toCsv(os);
    return os;
}

bool operator== (const MatrixImageSpaceInfo &lhs, const MatrixImageSpaceInfo &rhs)
{
    return (  lhs.box2d== rhs.box2d &&
     lhs.box3d== rhs.box3d &&
     lhs.poly== rhs.poly &&
     lhs.sigmaWidth== rhs.sigmaWidth &&
     lhs.sigmaHeight== rhs.sigmaHeight &&
     lhs.sigmaX== rhs.sigmaX &&
     lhs.sigmaY== rhs.sigmaY 
    );
}

std::size_t MatrixImageSpaceInfo::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<BoundingBox2D>()(box2d);
    result = result * 31 + std::hash<BoundingBox3D>()(box3d);
    result = result * 31 + hashVector(poly);
    result = result * 31 + std::hash<double>()(sigmaWidth);
    result = result * 31 + std::hash<double>()(sigmaHeight);
    result = result * 31 + std::hash<double>()(sigmaX);
    result = result * 31 + std::hash<double>()(sigmaY);
    return result;
}

 } // namespace av
 } // namespace horizon


CEREAL_BIND_TO_ARCHIVES(av::horizon::MatrixImageSpaceInfo)

