#include <interfaces/MatrixRawObject/MatrixRawObject.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>





 namespace av {
 namespace horizon {


MatrixRawObject::MatrixRawObject() = default;

MatrixRawObject::MatrixRawObject(
    OBJECT_TYPE objectType_arg ,
    double confidence_arg ,
    BoundingBox2D box2d_arg ,
    std::vector<int32_t> objProperty_arg ,
    std::array<std::string,20> objPropertyName_arg ,
    std::vector<int32_t> objPropertyType_arg ,
    std::vector<float> objPropertyConf_arg 
    ):
    objectType { std::move(objectType_arg) } // NOLINT
      ,confidence { std::move(confidence_arg) } // NOLINT
      ,box2d { std::move(box2d_arg) } // NOLINT
      ,objProperty { std::move(objProperty_arg) } // NOLINT
      ,objPropertyName { std::move(objPropertyName_arg) } // NOLINT
      ,objPropertyType { std::move(objPropertyType_arg) } // NOLINT
      ,objPropertyConf { std::move(objPropertyConf_arg) } // NOLINT
      {
}

MatrixRawObject::MatrixRawObject(const MatrixRawObject&) = default;

// NOLINTNEXTLINE
MatrixRawObject::MatrixRawObject(MatrixRawObject&&) = default;

MatrixRawObject& MatrixRawObject::operator=(const MatrixRawObject&) = default;

// NOLINTNEXTLINE
MatrixRawObject& MatrixRawObject::operator=(MatrixRawObject&&) = default;

void MatrixRawObject::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << static_cast<uint32_t>(objectType);out << ",";
    out << confidence;out << ",";
    out << box2d;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!objProperty.empty())
        {
            std::copy(objProperty.begin(), objProperty.end() - 1, std::ostream_iterator<int32_t>(out, "},{"));
            std::copy(objProperty.end()-1, objProperty.end(), std::ostream_iterator<int32_t>(out, ""));
        }
        out << "}}\"";out << ",";
    // Enclose array in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!objPropertyName.empty())
        {
            std::copy(objPropertyName.begin(), objPropertyName.end() - 1, std::ostream_iterator<std::string>(out, "},{"));
            out << objPropertyName.back();
        }
        out << "}}\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!objPropertyType.empty())
        {
            std::copy(objPropertyType.begin(), objPropertyType.end() - 1, std::ostream_iterator<int32_t>(out, "},{"));
            std::copy(objPropertyType.end()-1, objPropertyType.end(), std::ostream_iterator<int32_t>(out, ""));
        }
        out << "}}\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!objPropertyConf.empty())
        {
            std::copy(objPropertyConf.begin(), objPropertyConf.end() - 1, std::ostream_iterator<float>(out, "},{"));
            std::copy(objPropertyConf.end()-1, objPropertyConf.end(), std::ostream_iterator<float>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream MatrixRawObject::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void MatrixRawObject::serialize(Archive& ar, unsigned int version)
{

    ar& CEREAL_NVP(objectType);
    ar& CEREAL_NVP(confidence);
    ar& CEREAL_NVP(box2d);
    if(version >= 2)
    {
        ar& CEREAL_NVP(objProperty);
        ar& CEREAL_NVP(objPropertyName);
        ar& CEREAL_NVP(objPropertyType);
        ar& CEREAL_NVP(objPropertyConf);
    }
}


template void MatrixRawObject::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void MatrixRawObject::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void MatrixRawObject::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void MatrixRawObject::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void MatrixRawObject::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void MatrixRawObject::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void MatrixRawObject::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const MatrixRawObject& matrixrawobject)
{
    matrixrawobject.toCsv(os);
    return os;
}

bool operator== (const MatrixRawObject &lhs, const MatrixRawObject &rhs)
{
    return (  lhs.objectType== rhs.objectType &&
     lhs.confidence== rhs.confidence &&
     lhs.box2d== rhs.box2d &&
     lhs.objProperty== rhs.objProperty &&
     lhs.objPropertyName== rhs.objPropertyName &&
     lhs.objPropertyType== rhs.objPropertyType &&
     lhs.objPropertyConf== rhs.objPropertyConf 
    );
}

std::size_t MatrixRawObject::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<OBJECT_TYPE>()(objectType);
    result = result * 31 + std::hash<double>()(confidence);
    result = result * 31 + std::hash<BoundingBox2D>()(box2d);
    result = result * 31 + hashVector(objProperty);
    result = result * 31 + hashVector(objPropertyName);
    result = result * 31 + hashVector(objPropertyType);
    result = result * 31 + hashVector(objPropertyConf);
    return result;
}

 } // namespace av
 } // namespace horizon


CEREAL_BIND_TO_ARCHIVES(av::horizon::MatrixRawObject)

