#include <interfaces/Path/Path.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


DubinsTransition::DubinsTransition() = default;

DubinsTransition::DubinsTransition(
    msg::SE2 startPose_arg ,
    DUBINS_TYPE type_arg ,
    double radius_arg ,
    std::array<double,3> length_arg 
    ):
    startPose { std::move(startPose_arg) } // NOLINT
      ,
    type { std::move(type_arg) } // NOLINT
      ,
    radius { std::move(radius_arg) } // NOLINT
      ,
    length { std::move(length_arg) } // NOLINT
      
    {
}

DubinsTransition::DubinsTransition(const DubinsTransition&) = default;

// NOLINTNEXTLINE
DubinsTransition::DubinsTransition(DubinsTransition&&) = default;

DubinsTransition& DubinsTransition::operator=(const DubinsTransition&) = default;

// NOLINTNEXTLINE
DubinsTransition& DubinsTransition::operator=(DubinsTransition&&) = default;

void DubinsTransition::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << startPose;out << ",";
    out << static_cast<uint32_t>(type);out << ",";
    out << radius;out << ",";
    // Enclose array in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!length.empty())
        {
            std::copy(length.begin(), length.end() - 1, std::ostream_iterator<double>(out, "},{"));
            out << length.back();
        }
        out << "}}\"";
    
}

std::stringstream DubinsTransition::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void DubinsTransition::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(startPose);
        ar& CEREAL_NVP(type);
        ar& CEREAL_NVP(radius);
        ar& CEREAL_NVP(length);
    }


template void DubinsTransition::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void DubinsTransition::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void DubinsTransition::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void DubinsTransition::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void DubinsTransition::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void DubinsTransition::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void DubinsTransition::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const DubinsTransition& dubinstransition)
{
    dubinstransition.toCsv(os);
    return os;
}

bool operator== (const DubinsTransition &lhs, const DubinsTransition &rhs)
{
    return (  lhs.startPose== rhs.startPose &&
     lhs.type== rhs.type &&
     lhs.radius== rhs.radius &&
     lhs.length== rhs.length 
    );
}

std::size_t DubinsTransition::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<msg::SE2>()(startPose);
    result = result * 31 + std::hash<DUBINS_TYPE>()(type);
    result = result * 31 + std::hash<double>()(radius);
    result = result * 31 + hashVector(length);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::DubinsTransition)




 namespace av {
 namespace msg {


Path::Path() = default;

Path::Path(
    std::vector<DubinsTransition> transitions_arg 
    ):
    transitions { std::move(transitions_arg) } // NOLINT
      {
}

Path::Path(const Path&) = default;

// NOLINTNEXTLINE
Path::Path(Path&&) = default;

Path& Path::operator=(const Path&) = default;

// NOLINTNEXTLINE
Path& Path::operator=(Path&&) = default;

void Path::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!transitions.empty())
        {
            std::copy(transitions.begin(), transitions.end() - 1, std::ostream_iterator<DubinsTransition>(out, "},{"));
            std::copy(transitions.end()-1, transitions.end(), std::ostream_iterator<DubinsTransition>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream Path::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void Path::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(transitions);
    }


template void Path::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void Path::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void Path::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void Path::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void Path::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void Path::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void Path::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const Path& path)
{
    path.toCsv(os);
    return os;
}

bool operator== (const Path &lhs, const Path &rhs)
{
    return (  lhs.transitions== rhs.transitions 
    );
}

std::size_t Path::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + hashVector(transitions);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::Path)

