#include <interfaces/PathInfoMessage/PathInfoMessage.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


PathState::PathState() = default;

PathState::PathState(
    double x_m_arg ,
    double y_m_arg ,
    double yaw_rad_arg ,
    double curvature_arg 
    ):
    x_m { std::move(x_m_arg) } // NOLINT
      ,
    y_m { std::move(y_m_arg) } // NOLINT
      ,
    yaw_rad { std::move(yaw_rad_arg) } // NOLINT
      ,
    curvature { std::move(curvature_arg) } // NOLINT
      
    {
}

PathState::PathState(const PathState&) = default;

// NOLINTNEXTLINE
PathState::PathState(PathState&&) = default;

PathState& PathState::operator=(const PathState&) = default;

// NOLINTNEXTLINE
PathState& PathState::operator=(PathState&&) = default;

void PathState::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << x_m;out << ",";
    out << y_m;out << ",";
    out << yaw_rad;out << ",";
    out << curvature;
    
}

std::stringstream PathState::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void PathState::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(x_m);
        ar& CEREAL_NVP(y_m);
        ar& CEREAL_NVP(yaw_rad);
        ar& CEREAL_NVP(curvature);
    }


template void PathState::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void PathState::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void PathState::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void PathState::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void PathState::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void PathState::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void PathState::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const PathState& pathstate)
{
    pathstate.toCsv(os);
    return os;
}

bool operator== (const PathState &lhs, const PathState &rhs)
{
    return (  lhs.x_m== rhs.x_m &&
     lhs.y_m== rhs.y_m &&
     lhs.yaw_rad== rhs.yaw_rad &&
     lhs.curvature== rhs.curvature 
    );
}

std::size_t PathState::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<double>()(x_m);
    result = result * 31 + std::hash<double>()(y_m);
    result = result * 31 + std::hash<double>()(yaw_rad);
    result = result * 31 + std::hash<double>()(curvature);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::PathState)




 namespace av {
 namespace msg {


PathInfoMessage::PathInfoMessage() = default;

PathInfoMessage::PathInfoMessage(
    std::vector<PathState> states_arg ,
    TimePoint timestamp_arg 
    ):
    states { std::move(states_arg) } // NOLINT
      ,timestamp { std::move(timestamp_arg) } // NOLINT
      {
}

PathInfoMessage::PathInfoMessage(const PathInfoMessage&) = default;

// NOLINTNEXTLINE
PathInfoMessage::PathInfoMessage(PathInfoMessage&&) = default;

PathInfoMessage& PathInfoMessage::operator=(const PathInfoMessage&) = default;

// NOLINTNEXTLINE
PathInfoMessage& PathInfoMessage::operator=(PathInfoMessage&&) = default;

void PathInfoMessage::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!states.empty())
        {
            std::copy(states.begin(), states.end() - 1, std::ostream_iterator<PathState>(out, "},{"));
            std::copy(states.end()-1, states.end(), std::ostream_iterator<PathState>(out, ""));
        }
        out << "}}\"";out << ",";
    out << timestamp;
    
}

std::stringstream PathInfoMessage::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void PathInfoMessage::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(states);
        ar& CEREAL_NVP(timestamp);
    }


template void PathInfoMessage::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void PathInfoMessage::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void PathInfoMessage::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void PathInfoMessage::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void PathInfoMessage::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void PathInfoMessage::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void PathInfoMessage::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const PathInfoMessage& pathinfomessage)
{
    pathinfomessage.toCsv(os);
    return os;
}

bool operator== (const PathInfoMessage &lhs, const PathInfoMessage &rhs)
{
    return (  lhs.states== rhs.states &&
     lhs.timestamp== rhs.timestamp 
    );
}

std::size_t PathInfoMessage::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + hashVector(states);
    result = result * 31 + 0;
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::PathInfoMessage)

