#include <interfaces/SafeStopCommand/SafeStopCommand.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


SafeStopPoint::SafeStopPoint() = default;

SafeStopPoint::SafeStopPoint(
    double x_m_arg ,
    double y_m_arg ,
    double speed_mps_arg ,
    double station_m_arg ,
    double heading_rad_arg 
    ):
    x_m { std::move(x_m_arg) } // NOLINT
      ,
    y_m { std::move(y_m_arg) } // NOLINT
      ,
    speed_mps { std::move(speed_mps_arg) } // NOLINT
      ,
    station_m { std::move(station_m_arg) } // NOLINT
      ,
    heading_rad { std::move(heading_rad_arg) } // NOLINT
      
    {
}

SafeStopPoint::SafeStopPoint(const SafeStopPoint&) = default;

// NOLINTNEXTLINE
SafeStopPoint::SafeStopPoint(SafeStopPoint&&) = default;

SafeStopPoint& SafeStopPoint::operator=(const SafeStopPoint&) = default;

// NOLINTNEXTLINE
SafeStopPoint& SafeStopPoint::operator=(SafeStopPoint&&) = default;

void SafeStopPoint::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << x_m;out << ",";
    out << y_m;out << ",";
    out << speed_mps;out << ",";
    out << station_m;out << ",";
    out << heading_rad;
    
}

std::stringstream SafeStopPoint::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void SafeStopPoint::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(x_m);
        ar& CEREAL_NVP(y_m);
        ar& CEREAL_NVP(speed_mps);
        ar& CEREAL_NVP(station_m);
        ar& CEREAL_NVP(heading_rad);
    }


template void SafeStopPoint::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void SafeStopPoint::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void SafeStopPoint::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void SafeStopPoint::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void SafeStopPoint::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void SafeStopPoint::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void SafeStopPoint::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const SafeStopPoint& safestoppoint)
{
    safestoppoint.toCsv(os);
    return os;
}

bool operator== (const SafeStopPoint &lhs, const SafeStopPoint &rhs)
{
    return (  lhs.x_m== rhs.x_m &&
     lhs.y_m== rhs.y_m &&
     lhs.speed_mps== rhs.speed_mps &&
     lhs.station_m== rhs.station_m &&
     lhs.heading_rad== rhs.heading_rad 
    );
}

std::size_t SafeStopPoint::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<double>()(x_m);
    result = result * 31 + std::hash<double>()(y_m);
    result = result * 31 + std::hash<double>()(speed_mps);
    result = result * 31 + std::hash<double>()(station_m);
    result = result * 31 + std::hash<double>()(heading_rad);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::SafeStopPoint)




 namespace av {
 namespace msg {


SafeStopCommand::SafeStopCommand() = default;

SafeStopCommand::SafeStopCommand(
    int64_t timeStamp_arg ,
    std::vector<SafeStopPoint> safeStopPoints_arg 
    ):
    timeStamp { std::move(timeStamp_arg) } // NOLINT
      ,safeStopPoints { std::move(safeStopPoints_arg) } // NOLINT
      {
}

SafeStopCommand::SafeStopCommand(const SafeStopCommand&) = default;

// NOLINTNEXTLINE
SafeStopCommand::SafeStopCommand(SafeStopCommand&&) = default;

SafeStopCommand& SafeStopCommand::operator=(const SafeStopCommand&) = default;

// NOLINTNEXTLINE
SafeStopCommand& SafeStopCommand::operator=(SafeStopCommand&&) = default;

void SafeStopCommand::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << timeStamp;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!safeStopPoints.empty())
        {
            std::copy(safeStopPoints.begin(), safeStopPoints.end() - 1, std::ostream_iterator<SafeStopPoint>(out, "},{"));
            std::copy(safeStopPoints.end()-1, safeStopPoints.end(), std::ostream_iterator<SafeStopPoint>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream SafeStopCommand::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void SafeStopCommand::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(timeStamp);
        ar& CEREAL_NVP(safeStopPoints);
    }


template void SafeStopCommand::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void SafeStopCommand::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void SafeStopCommand::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void SafeStopCommand::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void SafeStopCommand::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void SafeStopCommand::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void SafeStopCommand::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const SafeStopCommand& safestopcommand)
{
    safestopcommand.toCsv(os);
    return os;
}

bool operator== (const SafeStopCommand &lhs, const SafeStopCommand &rhs)
{
    return (  lhs.timeStamp== rhs.timeStamp &&
     lhs.safeStopPoints== rhs.safeStopPoints 
    );
}

std::size_t SafeStopCommand::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<int64_t>()(timeStamp);
    result = result * 31 + hashVector(safeStopPoints);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::SafeStopCommand)

