#include <interfaces/SimulationCtrl/SimulationCtrl.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace simulation {


GenericObjectCreation::GenericObjectCreation() = default;

GenericObjectCreation::GenericObjectCreation(
    std::vector<double> start_arg ,
    std::vector<double> end_arg ,
    bool shouldDelete_arg 
    ):
    start { std::move(start_arg) } // NOLINT
      ,
    end { std::move(end_arg) } // NOLINT
      ,
    shouldDelete { std::move(shouldDelete_arg) } // NOLINT
      
    {
}

GenericObjectCreation::GenericObjectCreation(const GenericObjectCreation&) = default;

// NOLINTNEXTLINE
GenericObjectCreation::GenericObjectCreation(GenericObjectCreation&&) = default;

GenericObjectCreation& GenericObjectCreation::operator=(const GenericObjectCreation&) = default;

// NOLINTNEXTLINE
GenericObjectCreation& GenericObjectCreation::operator=(GenericObjectCreation&&) = default;

void GenericObjectCreation::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!start.empty())
        {
            std::copy(start.begin(), start.end() - 1, std::ostream_iterator<double>(out, "},{"));
            std::copy(start.end()-1, start.end(), std::ostream_iterator<double>(out, ""));
        }
        out << "}}\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!end.empty())
        {
            std::copy(end.begin(), end.end() - 1, std::ostream_iterator<double>(out, "},{"));
            std::copy(end.end()-1, end.end(), std::ostream_iterator<double>(out, ""));
        }
        out << "}}\"";out << ",";
    out << shouldDelete;
    
}

std::stringstream GenericObjectCreation::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void GenericObjectCreation::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(start);
        ar& CEREAL_NVP(end);
        ar& CEREAL_NVP(shouldDelete);
    }


template void GenericObjectCreation::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void GenericObjectCreation::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void GenericObjectCreation::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void GenericObjectCreation::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void GenericObjectCreation::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void GenericObjectCreation::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void GenericObjectCreation::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const GenericObjectCreation& genericobjectcreation)
{
    genericobjectcreation.toCsv(os);
    return os;
}

bool operator== (const GenericObjectCreation &lhs, const GenericObjectCreation &rhs)
{
    return (  lhs.start== rhs.start &&
     lhs.end== rhs.end &&
     lhs.shouldDelete== rhs.shouldDelete 
    );
}

std::size_t GenericObjectCreation::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + hashVector(start);
    result = result * 31 + hashVector(end);
    result = result * 31 + std::hash<bool>()(shouldDelete);
    return result;
}

 } // namespace simulation


CEREAL_BIND_TO_ARCHIVES(simulation::GenericObjectCreation)

 namespace simulation {


TeleportMsg::TeleportMsg() = default;

TeleportMsg::TeleportMsg(
    std::vector<double> pos_arg ,
    double heading_arg 
    ):
    pos { std::move(pos_arg) } // NOLINT
      ,
    heading { std::move(heading_arg) } // NOLINT
      
    {
}

TeleportMsg::TeleportMsg(const TeleportMsg&) = default;

// NOLINTNEXTLINE
TeleportMsg::TeleportMsg(TeleportMsg&&) = default;

TeleportMsg& TeleportMsg::operator=(const TeleportMsg&) = default;

// NOLINTNEXTLINE
TeleportMsg& TeleportMsg::operator=(TeleportMsg&&) = default;

void TeleportMsg::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!pos.empty())
        {
            std::copy(pos.begin(), pos.end() - 1, std::ostream_iterator<double>(out, "},{"));
            std::copy(pos.end()-1, pos.end(), std::ostream_iterator<double>(out, ""));
        }
        out << "}}\"";out << ",";
    out << heading;
    
}

std::stringstream TeleportMsg::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void TeleportMsg::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(pos);
        ar& CEREAL_NVP(heading);
    }


template void TeleportMsg::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void TeleportMsg::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void TeleportMsg::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void TeleportMsg::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void TeleportMsg::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void TeleportMsg::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void TeleportMsg::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const TeleportMsg& teleportmsg)
{
    teleportmsg.toCsv(os);
    return os;
}

bool operator== (const TeleportMsg &lhs, const TeleportMsg &rhs)
{
    return (  lhs.pos== rhs.pos &&
     lhs.heading== rhs.heading 
    );
}

std::size_t TeleportMsg::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + hashVector(pos);
    result = result * 31 + std::hash<double>()(heading);
    return result;
}

 } // namespace simulation


CEREAL_BIND_TO_ARCHIVES(simulation::TeleportMsg)




 namespace simulation {


SimulationCtrl::SimulationCtrl() = default;

SimulationCtrl::SimulationCtrl(
    bool killGroot_arg ,
    std::string filePath_arg ,
    bool mockPerception_arg ,
    EGO_MODES egoModeToStart_arg 
    ):
    killGroot { std::move(killGroot_arg) } // NOLINT
      ,filePath { std::move(filePath_arg) } // NOLINT
      ,mockPerception { std::move(mockPerception_arg) } // NOLINT
      ,egoModeToStart { std::move(egoModeToStart_arg) } // NOLINT
      {
}

SimulationCtrl::SimulationCtrl(const SimulationCtrl&) = default;

// NOLINTNEXTLINE
SimulationCtrl::SimulationCtrl(SimulationCtrl&&) = default;

SimulationCtrl& SimulationCtrl::operator=(const SimulationCtrl&) = default;

// NOLINTNEXTLINE
SimulationCtrl& SimulationCtrl::operator=(SimulationCtrl&&) = default;

void SimulationCtrl::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << killGroot;out << ",";
    out << "\"";
        out << filePath;
        out << "\"";out << ",";
    out << mockPerception;out << ",";
    out << static_cast<uint32_t>(egoModeToStart);
    
}

std::stringstream SimulationCtrl::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void SimulationCtrl::serialize(Archive& ar, unsigned int version)
{
	if (version >= 1)
	{
		ar& CEREAL_NVP(killGroot);
		ar& CEREAL_NVP(filePath);
	}
	if (version >= 2)
	{
		ar& CEREAL_NVP(mockPerception);
	}
	if (version >= 3)
	{
		ar& CEREAL_NVP(egoModeToStart);
	}
}



template void SimulationCtrl::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void SimulationCtrl::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void SimulationCtrl::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void SimulationCtrl::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void SimulationCtrl::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void SimulationCtrl::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void SimulationCtrl::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const SimulationCtrl& simulationctrl)
{
    simulationctrl.toCsv(os);
    return os;
}

bool operator== (const SimulationCtrl &lhs, const SimulationCtrl &rhs)
{
    return (  lhs.killGroot== rhs.killGroot &&
     lhs.filePath== rhs.filePath &&
     lhs.mockPerception== rhs.mockPerception &&
     lhs.egoModeToStart== rhs.egoModeToStart 
    );
}

std::size_t SimulationCtrl::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<bool>()(killGroot);
    result = result * 31 + std::hash<std::string>()(filePath);
    result = result * 31 + std::hash<bool>()(mockPerception);
    result = result * 31 + std::hash<EGO_MODES>()(egoModeToStart);
    return result;
}

 } // namespace simulation


CEREAL_BIND_TO_ARCHIVES(simulation::SimulationCtrl)

