#include <interfaces/SystemStatus/SystemStatus.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {


MachineFailureInfo::MachineFailureInfo() = default;

MachineFailureInfo::MachineFailureInfo(
    std::string hostName_arg ,
    std::vector<std::string> appsOnMachine_arg ,
    SystemStatusSeverity severityLevel_arg 
    ):
    hostName { std::move(hostName_arg) } // NOLINT
      ,
    appsOnMachine { std::move(appsOnMachine_arg) } // NOLINT
      ,
    severityLevel { std::move(severityLevel_arg) } // NOLINT
      
    {
}

MachineFailureInfo::MachineFailureInfo(const MachineFailureInfo&) = default;

// NOLINTNEXTLINE
MachineFailureInfo::MachineFailureInfo(MachineFailureInfo&&) = default;

MachineFailureInfo& MachineFailureInfo::operator=(const MachineFailureInfo&) = default;

// NOLINTNEXTLINE
MachineFailureInfo& MachineFailureInfo::operator=(MachineFailureInfo&&) = default;

void MachineFailureInfo::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << "\"";
        out << hostName;
        out << "\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!appsOnMachine.empty())
        {
            std::copy(appsOnMachine.begin(), appsOnMachine.end() - 1, std::ostream_iterator<std::string>(out, "},{"));
            std::copy(appsOnMachine.end()-1, appsOnMachine.end(), std::ostream_iterator<std::string>(out, ""));
        }
        out << "}}\"";out << ",";
    out << static_cast<uint32_t>(severityLevel);
    
}

std::stringstream MachineFailureInfo::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void MachineFailureInfo::serialize(Archive& ar, unsigned int version)
{
    ar& CEREAL_NVP(hostName);
    ar& CEREAL_NVP(appsOnMachine);
    if (version >= 2) {
        ar& CEREAL_NVP(severityLevel);
    }
}


template void MachineFailureInfo::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void MachineFailureInfo::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void MachineFailureInfo::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void MachineFailureInfo::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void MachineFailureInfo::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void MachineFailureInfo::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void MachineFailureInfo::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const MachineFailureInfo& machinefailureinfo)
{
    machinefailureinfo.toCsv(os);
    return os;
}

bool operator== (const MachineFailureInfo &lhs, const MachineFailureInfo &rhs)
{
    return (  lhs.hostName== rhs.hostName &&
     lhs.appsOnMachine== rhs.appsOnMachine &&
     lhs.severityLevel== rhs.severityLevel 
    );
}

std::size_t MachineFailureInfo::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<std::string>()(hostName);
    result = result * 31 + hashVector(appsOnMachine);
    result = result * 31 + std::hash<SystemStatusSeverity>()(severityLevel);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::MachineFailureInfo)

 namespace av {
 namespace msg {


SensorFailureInfo::SensorFailureInfo() = default;

SensorFailureInfo::SensorFailureInfo(
    std::string sensorName_arg ,
    std::string sensorLocation_arg ,
    std::string sensorStatus_arg ,
    SystemStatusSeverity severityLevel_arg 
    ):
    sensorName { std::move(sensorName_arg) } // NOLINT
      ,
    sensorLocation { std::move(sensorLocation_arg) } // NOLINT
      ,
    sensorStatus { std::move(sensorStatus_arg) } // NOLINT
      ,
    severityLevel { std::move(severityLevel_arg) } // NOLINT
      
    {
}

SensorFailureInfo::SensorFailureInfo(const SensorFailureInfo&) = default;

// NOLINTNEXTLINE
SensorFailureInfo::SensorFailureInfo(SensorFailureInfo&&) = default;

SensorFailureInfo& SensorFailureInfo::operator=(const SensorFailureInfo&) = default;

// NOLINTNEXTLINE
SensorFailureInfo& SensorFailureInfo::operator=(SensorFailureInfo&&) = default;

void SensorFailureInfo::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << "\"";
        out << sensorName;
        out << "\"";out << ",";
    out << "\"";
        out << sensorLocation;
        out << "\"";out << ",";
    out << "\"";
        out << sensorStatus;
        out << "\"";out << ",";
    out << static_cast<uint32_t>(severityLevel);
    
}

std::stringstream SensorFailureInfo::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void SensorFailureInfo::serialize(Archive& ar, unsigned int version)
{
    ar& CEREAL_NVP(sensorName);
    ar& CEREAL_NVP(sensorLocation);
    ar& CEREAL_NVP(sensorStatus);
    if (version >= 2) {
        ar& CEREAL_NVP(severityLevel);
    }
}


template void SensorFailureInfo::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void SensorFailureInfo::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void SensorFailureInfo::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void SensorFailureInfo::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void SensorFailureInfo::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void SensorFailureInfo::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void SensorFailureInfo::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const SensorFailureInfo& sensorfailureinfo)
{
    sensorfailureinfo.toCsv(os);
    return os;
}

bool operator== (const SensorFailureInfo &lhs, const SensorFailureInfo &rhs)
{
    return (  lhs.sensorName== rhs.sensorName &&
     lhs.sensorLocation== rhs.sensorLocation &&
     lhs.sensorStatus== rhs.sensorStatus &&
     lhs.severityLevel== rhs.severityLevel 
    );
}

std::size_t SensorFailureInfo::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<std::string>()(sensorName);
    result = result * 31 + std::hash<std::string>()(sensorLocation);
    result = result * 31 + std::hash<std::string>()(sensorStatus);
    result = result * 31 + std::hash<SystemStatusSeverity>()(severityLevel);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::SensorFailureInfo)

 namespace av {
 namespace msg {


ActuatorFailureInfo::ActuatorFailureInfo() = default;

ActuatorFailureInfo::ActuatorFailureInfo(
    std::string actuatorName_arg ,
    std::string actuatorLocation_arg ,
    std::string actuatorStatus_arg ,
    SystemStatusSeverity severityLevel_arg 
    ):
    actuatorName { std::move(actuatorName_arg) } // NOLINT
      ,
    actuatorLocation { std::move(actuatorLocation_arg) } // NOLINT
      ,
    actuatorStatus { std::move(actuatorStatus_arg) } // NOLINT
      ,
    severityLevel { std::move(severityLevel_arg) } // NOLINT
      
    {
}

ActuatorFailureInfo::ActuatorFailureInfo(const ActuatorFailureInfo&) = default;

// NOLINTNEXTLINE
ActuatorFailureInfo::ActuatorFailureInfo(ActuatorFailureInfo&&) = default;

ActuatorFailureInfo& ActuatorFailureInfo::operator=(const ActuatorFailureInfo&) = default;

// NOLINTNEXTLINE
ActuatorFailureInfo& ActuatorFailureInfo::operator=(ActuatorFailureInfo&&) = default;

void ActuatorFailureInfo::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << "\"";
        out << actuatorName;
        out << "\"";out << ",";
    out << "\"";
        out << actuatorLocation;
        out << "\"";out << ",";
    out << "\"";
        out << actuatorStatus;
        out << "\"";out << ",";
    out << static_cast<uint32_t>(severityLevel);
    
}

std::stringstream ActuatorFailureInfo::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void ActuatorFailureInfo::serialize(Archive& ar, unsigned int version)
{
    ar& CEREAL_NVP(actuatorName);
    ar& CEREAL_NVP(actuatorLocation);
    ar& CEREAL_NVP(actuatorStatus);
    if (version >= 2) {
        ar& CEREAL_NVP(severityLevel);
    }
}


template void ActuatorFailureInfo::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void ActuatorFailureInfo::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void ActuatorFailureInfo::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void ActuatorFailureInfo::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void ActuatorFailureInfo::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void ActuatorFailureInfo::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void ActuatorFailureInfo::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const ActuatorFailureInfo& actuatorfailureinfo)
{
    actuatorfailureinfo.toCsv(os);
    return os;
}

bool operator== (const ActuatorFailureInfo &lhs, const ActuatorFailureInfo &rhs)
{
    return (  lhs.actuatorName== rhs.actuatorName &&
     lhs.actuatorLocation== rhs.actuatorLocation &&
     lhs.actuatorStatus== rhs.actuatorStatus &&
     lhs.severityLevel== rhs.severityLevel 
    );
}

std::size_t ActuatorFailureInfo::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<std::string>()(actuatorName);
    result = result * 31 + std::hash<std::string>()(actuatorLocation);
    result = result * 31 + std::hash<std::string>()(actuatorStatus);
    result = result * 31 + std::hash<SystemStatusSeverity>()(severityLevel);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::ActuatorFailureInfo)

 namespace av {
 namespace msg {


AlgorithmFailureInfo::AlgorithmFailureInfo() = default;

AlgorithmFailureInfo::AlgorithmFailureInfo(
    std::string taskName_arg ,
    std::string taskStatus_arg ,
    SystemStatusSeverity severityLevel_arg 
    ):
    taskName { std::move(taskName_arg) } // NOLINT
      ,
    taskStatus { std::move(taskStatus_arg) } // NOLINT
      ,
    severityLevel { std::move(severityLevel_arg) } // NOLINT
      
    {
}

AlgorithmFailureInfo::AlgorithmFailureInfo(const AlgorithmFailureInfo&) = default;

// NOLINTNEXTLINE
AlgorithmFailureInfo::AlgorithmFailureInfo(AlgorithmFailureInfo&&) = default;

AlgorithmFailureInfo& AlgorithmFailureInfo::operator=(const AlgorithmFailureInfo&) = default;

// NOLINTNEXTLINE
AlgorithmFailureInfo& AlgorithmFailureInfo::operator=(AlgorithmFailureInfo&&) = default;

void AlgorithmFailureInfo::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << "\"";
        out << taskName;
        out << "\"";out << ",";
    out << "\"";
        out << taskStatus;
        out << "\"";out << ",";
    out << static_cast<uint32_t>(severityLevel);
    
}

std::stringstream AlgorithmFailureInfo::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void AlgorithmFailureInfo::serialize(Archive& ar, unsigned int /*version*/)
{
    ar& CEREAL_NVP(taskName);
    ar& CEREAL_NVP(taskStatus);
    ar& CEREAL_NVP(severityLevel);
}


template void AlgorithmFailureInfo::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void AlgorithmFailureInfo::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void AlgorithmFailureInfo::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void AlgorithmFailureInfo::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void AlgorithmFailureInfo::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void AlgorithmFailureInfo::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void AlgorithmFailureInfo::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const AlgorithmFailureInfo& algorithmfailureinfo)
{
    algorithmfailureinfo.toCsv(os);
    return os;
}

bool operator== (const AlgorithmFailureInfo &lhs, const AlgorithmFailureInfo &rhs)
{
    return (  lhs.taskName== rhs.taskName &&
     lhs.taskStatus== rhs.taskStatus &&
     lhs.severityLevel== rhs.severityLevel 
    );
}

std::size_t AlgorithmFailureInfo::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<std::string>()(taskName);
    result = result * 31 + std::hash<std::string>()(taskStatus);
    result = result * 31 + std::hash<SystemStatusSeverity>()(severityLevel);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::AlgorithmFailureInfo)




 namespace av {
 namespace msg {


SystemStatus::SystemStatus() = default;

SystemStatus::SystemStatus(
    TimePoint timeStamp_arg ,
    std::vector<MachineFailureInfo> machineFailInfo_arg ,
    std::vector<SensorFailureInfo> sensorFailInfo_arg ,
    std::vector<ActuatorFailureInfo> actuatorFailInfo_arg ,
    std::vector<AlgorithmFailureInfo> algorithmFailInfo_arg 
    ):
    timeStamp { std::move(timeStamp_arg) } // NOLINT
      ,machineFailInfo { std::move(machineFailInfo_arg) } // NOLINT
      ,sensorFailInfo { std::move(sensorFailInfo_arg) } // NOLINT
      ,actuatorFailInfo { std::move(actuatorFailInfo_arg) } // NOLINT
      ,algorithmFailInfo { std::move(algorithmFailInfo_arg) } // NOLINT
      {
}

SystemStatus::SystemStatus(const SystemStatus&) = default;

// NOLINTNEXTLINE
SystemStatus::SystemStatus(SystemStatus&&) = default;

SystemStatus& SystemStatus::operator=(const SystemStatus&) = default;

// NOLINTNEXTLINE
SystemStatus& SystemStatus::operator=(SystemStatus&&) = default;

void SystemStatus::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << timeStamp;out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!machineFailInfo.empty())
        {
            std::copy(machineFailInfo.begin(), machineFailInfo.end() - 1, std::ostream_iterator<MachineFailureInfo>(out, "},{"));
            std::copy(machineFailInfo.end()-1, machineFailInfo.end(), std::ostream_iterator<MachineFailureInfo>(out, ""));
        }
        out << "}}\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!sensorFailInfo.empty())
        {
            std::copy(sensorFailInfo.begin(), sensorFailInfo.end() - 1, std::ostream_iterator<SensorFailureInfo>(out, "},{"));
            std::copy(sensorFailInfo.end()-1, sensorFailInfo.end(), std::ostream_iterator<SensorFailureInfo>(out, ""));
        }
        out << "}}\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!actuatorFailInfo.empty())
        {
            std::copy(actuatorFailInfo.begin(), actuatorFailInfo.end() - 1, std::ostream_iterator<ActuatorFailureInfo>(out, "},{"));
            std::copy(actuatorFailInfo.end()-1, actuatorFailInfo.end(), std::ostream_iterator<ActuatorFailureInfo>(out, ""));
        }
        out << "}}\"";out << ",";
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!algorithmFailInfo.empty())
        {
            std::copy(algorithmFailInfo.begin(), algorithmFailInfo.end() - 1, std::ostream_iterator<AlgorithmFailureInfo>(out, "},{"));
            std::copy(algorithmFailInfo.end()-1, algorithmFailInfo.end(), std::ostream_iterator<AlgorithmFailureInfo>(out, ""));
        }
        out << "}}\"";
    
}

std::stringstream SystemStatus::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void SystemStatus::serialize(Archive& ar, unsigned int version)
{    
    if (version < 4)
    {
        otimePoint otp;
        ar& cereal::make_nvp("timeStamp", otp);
        timeStamp = av::TimePoint(std::chrono::seconds(otp.tv_sec) + std::chrono::nanoseconds(otp.tv_nsec));
    }
    else
    {
        ar& CEREAL_NVP(timeStamp);
    }
    ar& CEREAL_NVP(machineFailInfo);
    ar& CEREAL_NVP(sensorFailInfo);
    ar& CEREAL_NVP(actuatorFailInfo);
    if (version >= 3) {
        ar& CEREAL_NVP(algorithmFailInfo);
    }
}

}} // to close av::msg namespace temporary

struct SystemStatusMessage : public av::msg::SystemStatus {};
CEREAL_CLASS_VERSION(SystemStatusMessage, 4)
CEREAL_REGISTER_TYPE(SystemStatusMessage)

namespace av {
namespace msg {


template void SystemStatus::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void SystemStatus::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void SystemStatus::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void SystemStatus::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void SystemStatus::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void SystemStatus::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void SystemStatus::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const SystemStatus& systemstatus)
{
    systemstatus.toCsv(os);
    return os;
}

bool operator== (const SystemStatus &lhs, const SystemStatus &rhs)
{
    return (  lhs.timeStamp== rhs.timeStamp &&
     lhs.machineFailInfo== rhs.machineFailInfo &&
     lhs.sensorFailInfo== rhs.sensorFailInfo &&
     lhs.actuatorFailInfo== rhs.actuatorFailInfo &&
     lhs.algorithmFailInfo== rhs.algorithmFailInfo 
    );
}

std::size_t SystemStatus::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + 0;
    result = result * 31 + hashVector(machineFailInfo);
    result = result * 31 + hashVector(sensorFailInfo);
    result = result * 31 + hashVector(actuatorFailInfo);
    result = result * 31 + hashVector(algorithmFailInfo);
    return result;
}

 } // namespace av
 } // namespace msg


CEREAL_BIND_TO_ARCHIVES(av::msg::SystemStatus)

