#include <interfaces/ValidationResults/ValidationResults.h>

#include <iterator>
#include <type_traits>

#include <cereal/cereal.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/archives/binary.hpp>

#include <cereal/types/map.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/array.hpp>
#include <cereal/types/string.hpp>
#include <cereal/types/complex.hpp>

#include <ODatum/IntrospectionArchive.h>
#include <ODatum/Hash.h>


 namespace av {
 namespace msg {
 namespace calibration {


SensorValidationData::SensorValidationData() = default;

SensorValidationData::SensorValidationData(
    std::string name_arg ,
    bool isValidationValid_arg ,
    bool isValidationAngleGood_arg ,
    float icpScore_arg ,
    float svdScore_arg ,
    float rotationErrorDegree_arg ,
    av::geometry::Quaterniond rotationError_arg ,
    av::geometry::Vector3d translationError_arg 
    ):
    name { std::move(name_arg) } // NOLINT
      ,
    isValidationValid { std::move(isValidationValid_arg) } // NOLINT
      ,
    isValidationAngleGood { std::move(isValidationAngleGood_arg) } // NOLINT
      ,
    icpScore { std::move(icpScore_arg) } // NOLINT
      ,
    svdScore { std::move(svdScore_arg) } // NOLINT
      ,
    rotationErrorDegree { std::move(rotationErrorDegree_arg) } // NOLINT
      ,
    rotationError { std::move(rotationError_arg) } // NOLINT
      ,
    translationError { std::move(translationError_arg) } // NOLINT
      
    {
}

SensorValidationData::SensorValidationData(const SensorValidationData&) = default;

// NOLINTNEXTLINE
SensorValidationData::SensorValidationData(SensorValidationData&&) = default;

SensorValidationData& SensorValidationData::operator=(const SensorValidationData&) = default;

// NOLINTNEXTLINE
SensorValidationData& SensorValidationData::operator=(SensorValidationData&&) = default;

void SensorValidationData::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    out << "\"";
        out << name;
        out << "\"";out << ",";
    out << isValidationValid;out << ",";
    out << isValidationAngleGood;out << ",";
    out << icpScore;out << ",";
    out << svdScore;out << ",";
    out << rotationErrorDegree;out << ",";
    out << rotationError;out << ",";
    out << translationError;
    
}

std::stringstream SensorValidationData::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void SensorValidationData::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(name);
        ar& CEREAL_NVP(isValidationValid);
        ar& CEREAL_NVP(isValidationAngleGood);
        ar& CEREAL_NVP(icpScore);
        ar& CEREAL_NVP(svdScore);
        ar& CEREAL_NVP(rotationErrorDegree);
        ar& CEREAL_NVP(rotationError);
        ar& CEREAL_NVP(translationError);
    }


template void SensorValidationData::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void SensorValidationData::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void SensorValidationData::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void SensorValidationData::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void SensorValidationData::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void SensorValidationData::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void SensorValidationData::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const SensorValidationData& sensorvalidationdata)
{
    sensorvalidationdata.toCsv(os);
    return os;
}

bool operator== (const SensorValidationData &lhs, const SensorValidationData &rhs)
{
    return (  lhs.name== rhs.name &&
     lhs.isValidationValid== rhs.isValidationValid &&
     lhs.isValidationAngleGood== rhs.isValidationAngleGood &&
     lhs.icpScore== rhs.icpScore &&
     lhs.svdScore== rhs.svdScore &&
     lhs.rotationErrorDegree== rhs.rotationErrorDegree &&
     lhs.rotationError== rhs.rotationError &&
     lhs.translationError== rhs.translationError 
    );
}

std::size_t SensorValidationData::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + std::hash<std::string>()(name);
    result = result * 31 + std::hash<bool>()(isValidationValid);
    result = result * 31 + std::hash<bool>()(isValidationAngleGood);
    result = result * 31 + std::hash<float>()(icpScore);
    result = result * 31 + std::hash<float>()(svdScore);
    result = result * 31 + std::hash<float>()(rotationErrorDegree);
    result = result * 31 + std::hash<av::geometry::Quaterniond>()(rotationError);
    result = result * 31 + std::hash<av::geometry::Vector3d>()(translationError);
    return result;
}

 } // namespace av
 } // namespace msg
 } // namespace calibration


CEREAL_BIND_TO_ARCHIVES(av::msg::calibration::SensorValidationData)




 namespace av {
 namespace msg {
 namespace calibration {


ValidationResults::ValidationResults() = default;

ValidationResults::ValidationResults(
    std::vector<SensorValidationData> sensors_arg ,
    av::msg::Duration validationDuration_arg ,
    av::msg::TimePoint validationDate_arg ,
    std::string vehicleName_arg 
    ):
    sensors { std::move(sensors_arg) } // NOLINT
      ,validationDuration { std::move(validationDuration_arg) } // NOLINT
      ,validationDate { std::move(validationDate_arg) } // NOLINT
      ,vehicleName { std::move(vehicleName_arg) } // NOLINT
      {
}

ValidationResults::ValidationResults(const ValidationResults&) = default;

// NOLINTNEXTLINE
ValidationResults::ValidationResults(ValidationResults&&) = default;

ValidationResults& ValidationResults::operator=(const ValidationResults&) = default;

// NOLINTNEXTLINE
ValidationResults& ValidationResults::operator=(ValidationResults&&) = default;

void ValidationResults::toCsv(std::ostream& out) const
{
    out.precision(16);
    out << std::boolalpha;
    // Enclose sequence in double quotations and curly brackets to keep CSV line length uniform
        
        out << "\"{{";
        
        if (!sensors.empty())
        {
            std::copy(sensors.begin(), sensors.end() - 1, std::ostream_iterator<SensorValidationData>(out, "},{"));
            std::copy(sensors.end()-1, sensors.end(), std::ostream_iterator<SensorValidationData>(out, ""));
        }
        out << "}}\"";out << ",";
    out << validationDuration;out << ",";
    out << validationDate;out << ",";
    out << "\"";
        out << vehicleName;
        out << "\"";
    
}

std::stringstream ValidationResults::toStringstream() const
{
    std::stringstream out;
    toCsv(out);
    return out;
}

template<class Archive>
    void ValidationResults::serialize(Archive& ar, [[maybe_unused]] unsigned int version)
    {

        ar& CEREAL_NVP(sensors);
        ar& CEREAL_NVP(validationDuration);
        ar& CEREAL_NVP(validationDate);
        ar& CEREAL_NVP(vehicleName);
    }


template void ValidationResults::serialize(cereal::BinaryInputArchive& ar, unsigned int version);
template void ValidationResults::serialize(cereal::BinaryOutputArchive& ar, unsigned int version);
template void ValidationResults::serialize(cereal::IntrospectionArchive& ar, unsigned int version);
template void ValidationResults::serialize(cereal::JSONInputArchive& ar, unsigned int version);
template void ValidationResults::serialize(cereal::JSONOutputArchive& ar, unsigned int version);
template void ValidationResults::serialize(cereal::PortableBinaryInputArchive& ar, unsigned int version);
template void ValidationResults::serialize(cereal::PortableBinaryOutputArchive& ar, unsigned int version);

std::ostream& operator<< (std::ostream& os, const ValidationResults& validationresults)
{
    validationresults.toCsv(os);
    return os;
}

bool operator== (const ValidationResults &lhs, const ValidationResults &rhs)
{
    return (  lhs.sensors== rhs.sensors &&
     lhs.validationDuration== rhs.validationDuration &&
     lhs.validationDate== rhs.validationDate &&
     lhs.vehicleName== rhs.vehicleName 
    );
}

std::size_t ValidationResults::hash() const noexcept
{
    std::size_t result = 17;
    result = result * 31 + hashVector(sensors);
    result = result * 31 + std::hash<av::msg::Duration>()(validationDuration);
    result = result * 31 + std::hash<av::msg::TimePoint>()(validationDate);
    result = result * 31 + std::hash<std::string>()(vehicleName);
    return result;
}

 } // namespace av
 } // namespace msg
 } // namespace calibration


CEREAL_BIND_TO_ARCHIVES(av::msg::calibration::ValidationResults)

