/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief This file contains a few shared utilities for fuzzing.
 */

#include <FuzzUtils/FuzzUtils.h>

#include <cmath>
#include <fstream>
#include <iostream>
#include <iterator>

/**
 * \brief Constructor to pre-allocate the underlying buffer.
 * \param[in] aLen: length to pre-allocate for the buffer.
 */
FuzzData::FuzzData(size_t aLen)
{
    data.reserve(aLen);
}

/**
 * \brief Constructor that takes a buffer
 * \param[in] aData: the buffer.
 */
FuzzData::FuzzData(std::vector<char> aData)
    : data(aData)
{}

/**
 * \brief Return the raw buffer.
 * \return a pointer to the underyling buffer.
 * \note
 * Return is consistent with signature of libfuzzer entry point:
 * int LLVMFuzzerTestOneInput(const uint8_t* data, size_t size)
 */
const uint8_t* FuzzData::getData() const
{
    return static_cast<const uint8_t*>(static_cast<const void*>(data.data()));
}

/**
 * \brief Return the length of the the data buffer.
 * \return the length
 * \note
 * Return is consistent with signature of libfuzzer entry point:
 * int LLVMFuzzerTestOneInput(const uint8_t* data, size_t size)
 */
size_t FuzzData::getSize() const
{
    return data.size();
}

/**
 * \brief Mutate the input string (in-place) based on the fuzz buffer.
 * \param[in] str: the input string to be mutated (in-place).
 * \param[in] data: the fuzz input buffer.
 * \param[in] size: the size of the buffer.
 * \note Internal linkage
 */
static void _fuzzStringData(std::string& str, const uint8_t data[], size_t size);

/**
 * \brief Repeat the input string, to reach at least a certain length.
 * \param[in] input: the input string to be repeated.
 * \param[in] length: the desired length.
 * \return the repeated string.
 * \note Internal linkage
 */
static std::string _repeatString(const std::string& input, size_t length);

/**
 * \brief Read the fuzz data from a file.
 * \param[in] fileName: the name of the file.
 * \return the contents of the file as a FuzzData struct.
 */
std::unique_ptr<FuzzData> getFuzzDataFromFile(const char* fileName)
{
    using std::ifstream;
    using std::ios;

    // flag ate: the get pointer will be positioned at the end of the file
    ifstream ifs(fileName, ios::in | ios::binary | ios::ate);

    ifstream::pos_type fileSize = ifs.tellg();
    ifs.seekg(0, std::ios::beg);

    std::vector<char> bytes(static_cast<size_t>(fileSize));
    bytes.insert(bytes.begin(), std::istream_iterator<char>(ifs), std::istream_iterator<char>());

    return std::make_unique<FuzzData>(bytes);
}

/**
 * \brief Read the fuzz data from the console.
 * \return the contents of console input as a FuzzData struct
 */
std::unique_ptr<FuzzData> getFuzzDataFromConsole()
{
    using std::cin;
    using std::cout;

    std::string line;
    std::getline(cin, line);

    size_t size = line.size();
    auto fuzzData = std::make_unique<FuzzData>(size);
    std::copy(line.begin(), line.end(), std::back_inserter(fuzzData->data));
    return fuzzData;
}

/**
 * \brief Mutate the input string (in-place) based on the fuzz buffer.
 * \param[in] str: the string to be mutated
 * \param[in] data: the fuzz input buffer.
 * \param[in] size: the size of the buffer.
 */
void _fuzzStringData(std::string& str, const uint8_t data[], size_t size)
{
    if (data == nullptr || size == 0) {
        return;
    }

    size_t j = 0;
    bool done = false;
    while (j < size && !done) {
        for (size_t i = 0; i < str.size(); i++) {
            // Take position into account. Also handle j==0.
            str[i] = static_cast<char>(
                str[i] +
                static_cast<char>((j + 1) * data[j])); // NOLINT(cppcoreguidelines-pro-bounds-pointer-arithmetic)
            j++;
            if (j == size) {
                j = 0;
                done = true;
            }
        }
    }
}

/**
 * \brief Compute a single integer value based on fuzz data.
 * \param[in] data: the fuzz input buffer.
 * \param[in] size: the size of the buffer.
 * \param[in] base: the base to add. With default value
 * \return: a single int value.
 */
uint32_t getFuzzInt(const uint8_t* data, size_t size, int base)
{
    uint32_t ret{};

    if (size >= sizeof(ret) && data != nullptr) {
        // Use the entire fuzz data.
        ret = *static_cast<const uint32_t*>(static_cast<const void*>(data));
        ret += static_cast<uint32_t>(base);
    }
    return ret;
}

/**
 * \brief Repeat the input string, to reach at least a certain length.
 * \param[in] input: the input string to be repeated.
 * \param[in] length: the desired length.
 * \return the repeated string.
 * \note If length == 0, return the input string.
 */
std::string _repeatString(const std::string& input, size_t length)
{
    std::string ret = input;
    const size_t inputSize = input.size();

    ret.reserve(std::max(inputSize, length));

    size_t num = inputSize;
    while (num < length) {
        ret += input;
        num += inputSize;
    }
    return ret;
}

/**
 * \brief Convert the raw fuzz buffer to a string object, without any changes.
 * \param[in] data: the fuzz input buffer.
 * \param[in] size: the size of the buffer.
 * \return the string object.
 */
std::string getFuzzString(const uint8_t* data, size_t size)
{
    std::string str(static_cast<const char*>(static_cast<const void*>(data)), size);
    return str;
}

/**
 * \brief Get a fuzz string object, based on a seed string.
 * \param[in] seed: the seed string.
 * \param[in] data: the fuzz input buffer.
 * \param[in] size: the size of the buffer.
 * \return the string object.
 */
std::string getFuzzMsg(const uint8_t* data, size_t size, const std::string& seed)
{
    std::string repeated = _repeatString(seed, size);
    _fuzzStringData(repeated, data, size);
    return repeated;
}

/**
 * \brief Get a fuzz timespec object, based on a seed int.
 * \param[in] data: the fuzz input buffer.
 * \param[in] size: the size of the buffer.
 * \param[in] seed: the constant added.
 * \return the timespec object.
 */
timespec getFuzzTime(const uint8_t* data, size_t size, uint8_t seed)
{
    // Use the entire fuzz data.
    timespec t{};
    if (size >= sizeof(t.tv_sec) && data != nullptr) {
        t.tv_sec = *static_cast<const uint32_t*>(static_cast<const void*>(data));
    }
    if (size >= sizeof(t.tv_nsec) && data != nullptr) {
        t.tv_nsec = *static_cast<const uint32_t*>(static_cast<const void*>(data));
        t.tv_nsec += static_cast<uint32_t>(seed);
    }
    return t;
}
