/**
 * \copyright
 * Copyright (c) 2020- Aptiv. All rights reserved.
 * NOTICE: Removing this heading is prohibited.
 * All information contained herein is, and remains the property of Aptiv and its suppliers, if any.
 * The intellectual and technical concepts contained herein are proprietary to Aptiv and its suppliers and may be
 * covered by U.S. and Foreign Patents, patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
 * permission is obtained from Aptiv.
 *
 * \brief This file contains a few shared utilities for fuzzing.
 */
#ifndef FUZZ_UTILS
#define FUZZ_UTILS

#include <cstddef>
#include <cstdint>
#include <memory>
#include <string>
#include <vector>

/**
 * \brief Represent fuzz buffer that is read either from a file or from the console.
 */
struct FuzzData
{
public:
    /**
     * \brief Constructor to pre-allocate the underlying buffer.
     * \param[in] aLen: length to pre-allocate for the buffer.
     */
    FuzzData(size_t aLen);

    /**
     * \brief Constructor that takes a buffer
     * \param[in] aData: the buffer.
     */
    FuzzData(std::vector<char> aData);

    /**
     * \brief Public field to store the fuzz data and its length
     */
    std::vector<char> data;

    /**
     * \brief Return the raw buffer.
     * \return a pointer to the underyling buffer.
     * \note
     * Return is consistent with signature of libfuzzer entry point:
     * int LLVMFuzzerTestOneInput(const uint8_t* data, size_t size)
     */
    const uint8_t* getData() const;

    /**
     * \brief Return the length of the the data buffer.
     * \return the length
     * \note
     * Return is consistent with signature of libfuzzer entry point:
     * int LLVMFuzzerTestOneInput(const uint8_t* data, size_t size)
     */
    size_t getSize() const;
};

/**
 * \brief Read fuzz data from a file.
 * \param[in] fileName: the name of the file.
 * \return the contents of the file as a FuzzData struct
 */
std::unique_ptr<FuzzData> getFuzzDataFromFile(const char* fileName);

/**
 * \brief Read fuzz data from the console.
 * \return the contents of console input as a FuzzData struct
 */
std::unique_ptr<FuzzData> getFuzzDataFromConsole();

/**
 * \brief Compute a single integer value based on fuzz buffer.
 * \param[in] data: the fuzz input buffer.
 * \param[in] size: the size of the buffer.
 * \param[in] base: the base used in the hash computation. With default value.
 * \return: a single int value.
 * \note The idea is to compute a hash of the entire fuzz buffer.
 */
uint32_t getFuzzInt(const uint8_t* data, size_t size, int base = 13);

/**
 * \brief Convert the raw fuzz buffer to a string object, without any changes.
 * \param[in] data: the fuzz input buffer.
 * \param[in] size: the size of the buffer.
 * \return the string object.
 */
std::string getFuzzString(const uint8_t* data, size_t size);

/**
 * \brief Get a fuzz string message, based on a seed string.
 * \param[in] data: the fuzz input buffer.
 * \param[in] size: the size of the buffer.
 * \param[in] seed: the seed string.
 * \return the string object.
 */
std::string getFuzzMsg(const uint8_t* data, size_t size, const std::string& seed = std::string());

/**
 * \brief Get a fuzz timespec object, based on a seed int.
 * \param[in] data: the fuzzed input buffer.
 * \param[in] size: the size of the buffer.
 * \param[in] seed: the constant used in the hash computation.
 * \return the timespec object.
 * \note The idea is to compute a hash of the entire fuzz buffer.
 */
timespec getFuzzTime(const uint8_t* data, size_t size, uint8_t seed);

#endif // FUZZ_UTILS